// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------


#include "Plateau.h"


#define MIN(x,y) ((x) < (y) ? (x) : (y))
#define MAX(x,y) ((x) > (y) ? (x) : (y))

extern	long near(float) ;

PlateauList *NewPlateauList() {
	PlateauList *list ;
	
	list = new PlateauList ;
	if (list != nil) {
		list->numPlateaus = 0 ;
		list->plateauLink = nil ;
	}
	return list ;
}

Plateau	*FindPlateau(PlateauList *list,Pixel *pixel) {
	PlateauLink *link ;
	PixelLink	*pixelLink ;
	//char		text[256] ;
	
	for(link = list->plateauLink ;link != nil; link = link->nextLink) {
		for (pixelLink = ((link->plateau)->pixelList)->pixLink; pixelLink != nil; 
				pixelLink = pixelLink->nextLink) {
			if (pixel->h == (pixelLink->pixel).h && pixel->v == (pixelLink->pixel).v)
				return link->plateau ;
		}
	}
	return nil ;
}

Boolean AddPlateau(PlateauList *list,Plateau *plateau) {
	PlateauLink *link ;
	
	// We save time by always adding to the beginning of the list
	link = new PlateauLink ;
	if (link == nil)
		return false ;
		
	link->nextLink = list->plateauLink ;
	link->plateau  = plateau ;
	list->plateauLink = link ;
	list->numPlateaus++ ;
	return true ;
}

void RemovePlateau(PlateauList *list,Plateau *plateau) {
	PlateauLink* link;
	PlateauLink* last;
	Plateau		*currPlateau ;
	
	last = nil;
	for (link = list->plateauLink ;link != nil; link = link->nextLink) {
		currPlateau = link->plateau ;
		if ( currPlateau == plateau ) {	// is the current plateau the one to be removed?
										// yes, remove it
			// if first item in list, just set first
				if ( last == nil )				
					list->plateauLink = link->nextLink;
				else
					last->nextLink = link->nextLink ;
			
			// free the Plateau		
				FreePlateau(plateau) ;				
				list->numPlateaus--;
				
			// we're done; no need to continue onto the next document
				return;
		}
		else		// no, go onto the next document in the list
			last = link;
	}
}

Plateau		*NewPlateau(Pixel *pixel,short number) {
	Plateau 	*plateau ;
	PixelLink	*pixlink ;
	PixelList	*pixlist ;
	
	plateau = new Plateau ;
	if (plateau == nil) {
		return nil ;
	}
	
	plateau->plateauNumber = number ;	
	
	pixlist = new PixelList ;
	
	if (pixlist == nil) {
		delete plateau ;
		return nil ;
	}
	plateau->pixelList = pixlist ;
	
	pixlink = new PixelLink ;
	if (pixlink == nil) {
		delete plateau ;
		delete plateau->pixelList ;
		return nil ;
	}
	(plateau->pixelList)->pixLink = pixlink ;
	(plateau->pixelList)->numPixels = 1 ;
	((plateau->pixelList)->pixLink)->pixel = *pixel ;
	((plateau->pixelList)->pixLink)->nextLink = nil ;
	
	
	return plateau ;
}

// This is the routine that must do all the work.  It checks the neighbors,
// finds the lowest plateau belonging to a neighbor, assigns the pixel to that 
// neighbor and also adds other adjoining plateaus to the plateau of lowest number
//
Boolean AssignPixelToPlateau(PlateauList *plateaulist,Pixel *pixel,
				float *pixaddress,short Width) {
	short	left,top,bottom,right,iright ;
	register int	i,j,di,dj ;
	float	fval ;
	Plateau *plateau,*thisPlateau ;
	Pixel	thisPixel ;
	Boolean isPlateauPeak[4],makeNewPlateau ;
	short	minPlateauNumber ;
	//char 	text[256] ;
	
	// We know that we are part of a plateau.  We need to decide if we are part of
	// an existing plateau or if we need to create a new one
	// First check to see if there are adjoining pixels belonging to a plateau
	//	-------------
	//	| x | x | x |			Check the marked pixels
	//  	| x | p |
	//
	
	makeNewPlateau = true ;
	
	left = pixel->h > 0 ? pixel->h-1 : pixel->h ;
	top = pixel->v > 0 ? pixel->v-1 : pixel->v ;
	right = pixel->h < Width-1 ? pixel->h+1 : pixel->h ;
	bottom = pixel->v ;
	
	for(i=0;i<4;i++) 
		isPlateauPeak[i] = false ;
		
	for(j=top;j <= bottom;j++) {
		dj = j - pixel->v ;
		iright = (dj == 0 ? left : right) ;
		for(i=left;i <= iright;i++) {
			di = i - pixel->h ;
			if (di != 0 || dj != 0) {
				fval = *(pixaddress+dj*Width+di) ;
				if (fval != 0.) {
					makeNewPlateau = false ;
					isPlateauPeak[(dj+1)*3+di+1] = true ;
				}
			}
		}
	}
	
	if (makeNewPlateau) {
		// Create a new plateau
		plateau = NewPlateau(pixel,plateaulist->numPlateaus) ;
		if (plateau != nil) {
			if (!AddPlateau(plateaulist,plateau)) {
				return false ;
			}
		}
		else {
			return false ;
		}
	}
	else {
		// Now for the hard work
		// Find out which plateau which we need to assign the pixel to
		minPlateauNumber = 10000 ;
		for(i=0;i<4;i++) {
			if (isPlateauPeak[i]) {
				switch(i) {
					case 0:
						thisPixel.h = pixel->h-1 ;
						thisPixel.v = pixel->v-1 ;
						thisPlateau = FindPlateau(plateaulist,&thisPixel) ;
						if (thisPlateau != nil) {
							minPlateauNumber = MIN(minPlateauNumber,thisPlateau->plateauNumber) ;
						}
						else {
							return false ;
						}
						break ;
					case 1:
						thisPixel.h = pixel->h ;
						thisPixel.v = pixel->v-1 ;
						thisPlateau = FindPlateau(plateaulist,&thisPixel) ;
						if (thisPlateau != nil) {
							minPlateauNumber = MIN(minPlateauNumber,thisPlateau->plateauNumber) ;
						}
						else {
							return false ;
						}
						break ;
					case 2:
						thisPixel.h = pixel->h+1 ;
						thisPixel.v = pixel->v-1 ;
						thisPlateau = FindPlateau(plateaulist,&thisPixel) ;
						if (thisPlateau != nil) {
							minPlateauNumber = MIN(minPlateauNumber,thisPlateau->plateauNumber) ;
						}
						else {
							return false ;
						}
						break ;
					case 3:
						thisPixel.h = pixel->h-1 ;
						thisPixel.v = pixel->v ;
						thisPlateau = FindPlateau(plateaulist,&thisPixel) ;
						if (thisPlateau != nil) {
							minPlateauNumber = MIN(minPlateauNumber,thisPlateau->plateauNumber) ;
						}
						else {
							return false ;
						}
						break ;
				}
			}
			
		}
		
		// Assign this pixel to the platea with minimum plateau number
		plateau = FindPlateauFromNumber(plateaulist,minPlateauNumber) ;
		if (plateau == nil) {
			return false ;
		}
			
		if (!AddPixelToPlateau(plateau,pixel))
			return false ;
		
		// Go through any remaining peaks and assign to the plateau of lowest number
		for(i=0;i<4;i++) {
			if (isPlateauPeak[i]) {
				switch(i) {
					case 0:
						thisPixel.h = pixel->h-1 ;
						thisPixel.v = pixel->v-1 ;
						thisPlateau = FindPlateau(plateaulist,&thisPixel) ;
						if (thisPlateau != nil) {
							if (thisPlateau->plateauNumber != minPlateauNumber) {
								AddPlateauToPlateau(plateau,thisPlateau) ;
								RemovePlateau(plateaulist,thisPlateau) ;
							}
						}
						else {
							return false ;
						}
						break ;
					case 1:
						thisPixel.h = pixel->h ;
						thisPixel.v = pixel->v-1 ;
						thisPlateau = FindPlateau(plateaulist,&thisPixel) ;
						if (thisPlateau != nil) {
							if (thisPlateau->plateauNumber != minPlateauNumber) {
								AddPlateauToPlateau(plateau,thisPlateau) ;
								RemovePlateau(plateaulist,thisPlateau) ;
							}
						}
						else {
							return false ;
						}
						break ;
					case 2:
						thisPixel.h = pixel->h+1 ;
						thisPixel.v = pixel->v-1 ;
						thisPlateau = FindPlateau(plateaulist,&thisPixel) ;
						if (thisPlateau != nil) {
							if (thisPlateau->plateauNumber != minPlateauNumber) {
								AddPlateauToPlateau(plateau,thisPlateau) ;
								RemovePlateau(plateaulist,thisPlateau) ;
							}
						}
						else {
							return false ;
						}
						break ;
					case 3:
						thisPixel.h = pixel->h-1 ;
						thisPixel.v = pixel->v ;
						thisPlateau = FindPlateau(plateaulist,&thisPixel) ;
						if (thisPlateau != nil) {
							if (thisPlateau->plateauNumber != minPlateauNumber) {
								AddPlateauToPlateau(plateau,thisPlateau) ;
								RemovePlateau(plateaulist,thisPlateau) ;
							}
						}
						else {
							return false ;
						}
						break ;
				}
			}
		}
	}
	return true ;
}

Boolean AddPixelToPlateau(Plateau *plateau,Pixel *pixel) {
	PixelLink 	*link ;
	
	// We save time by always adding to the beginning of the list
	link = new PixelLink ;
	if (link == nil) {
		return false ;
	}
		
	link->nextLink = (plateau->pixelList)->pixLink ;
	link->pixel = *pixel ;
	(plateau->pixelList)->pixLink = link ;
	(plateau->pixelList)->numPixels++ ;
	return true ;
}

void AddPlateauToPlateau(Plateau *destPlateau,Plateau *sourcePlateau) {
	PixelLink	*destLink ;
	PixelLink	*lastLink ;
	
	// Find the end of the destLink ;
	for(destLink = (destPlateau->pixelList)->pixLink ;destLink != nil; ) {
		lastLink = destLink ;
		destLink = destLink->nextLink ;
	}
	lastLink->nextLink = (sourcePlateau->pixelList)->pixLink ;
	(destPlateau->pixelList)->numPixels += (sourcePlateau->pixelList)->numPixels ;
	
}

Plateau	*FindPlateauFromNumber(PlateauList *list,short number) {
	PlateauLink *link ;
	short		pNumber ;
	//char 		text[256] ;
	
	for(link = list->plateauLink ;link != nil; link = link->nextLink) {
		pNumber = (link->plateau)->plateauNumber ;
		if (number == pNumber)
			return (link->plateau) ;
	}
	return nil ;
}

void ReducePlateaus(PlateauList *list,float *imagedata,short Width) {
	float sumx,sumy ;
	PlateauLink *link ;
	PixelLink	*pixelLink ;
	short 		cmh,cmv,nPixels ;
	
	// For every plateau
	for(link = list->plateauLink ;link != nil; link = link->nextLink) {
		// For every pixel in the plateau
		sumx = 0. ;
		sumy = 0. ;
		nPixels = 0 ;
		for (pixelLink = ((link->plateau)->pixelList)->pixLink; pixelLink != nil; 
				pixelLink = pixelLink->nextLink) {
			sumx += (pixelLink->pixel).h ;
			sumy += (pixelLink->pixel).v ;
			nPixels++ ;
			
		}
		sumx /= nPixels ;		// Center x of spot
		sumy /= nPixels ;		// Center y of spot
		(link->plateau)->cmx = near(sumx) ;
		(link->plateau)->cmy = near(sumy) ;
	}
	
	// Now set every pixel in the image except for the center pixel to 0.
	for(link = list->plateauLink ;link != nil; link = link->nextLink) {
		// For every pixel in the plateau
		cmh = (link->plateau)->cmx ;
		cmv = (link->plateau)->cmy ;
		for (pixelLink = ((link->plateau)->pixelList)->pixLink; pixelLink != nil; 
				pixelLink = pixelLink->nextLink) {
			if (cmh != (pixelLink->pixel).h  || cmv != (pixelLink->pixel).v) {
				*(imagedata + (pixelLink->pixel).v*Width + (pixelLink->pixel).h) = 0. ;
			}
			
		}
		// Also set the center pixel to 1 since the center may not be part of the plateau
		*(imagedata + cmv*Width + cmh) = 1. ;
	}
}
/* -----------------------------------------------------------	*/
/* Determine if the current peak is part of a plateau			*/
/* Return true if it is, else false								*/
/* -----------------------------------------------------------	*/
Boolean	IsPlateau(Pixel *pixel,float *pixaddress,long Width,long Height) {
	Boolean isplateau ;
	short	left,top,bottom,right ;
	short	i,j,di,dj ;
	float	sumPixels ;
	
	// First check to see if there are any adjoining pixels.  If there are we
	// need to assign a plateau to this pixel
	//	-------------
	//	| x | x | x |			Check the marked pixels
	//  	| x | p | x |
	//	| x | x | x |
	//	-------------
	
	isplateau = false ;
	
	left = pixel->h > 0 ? pixel->h-1 : 0 ;
	top = pixel->v > 0 ? pixel->v-1 : 0 ;
	right = pixel->h < Width-1 ? pixel->h+1 : Width-1 ;
	bottom = pixel->v < Height-1 ? pixel->v+1 : Height-1 ;
	sumPixels = 0. ;
	
	for(j=top;j <= bottom;j++) {
		dj = j - pixel->v ;
		for(i=left;i <= right;i++) {
			di = i - pixel->h ;
			sumPixels += *(pixaddress+dj*Width+di) ;
		}
	}
	if (sumPixels > 1.)
		isplateau = true ;	
		
	return(isplateau) ;
}

// --------------------------------------------
// Delete the entire plateau list including 
// plateaus
//
void DeletePlateauList(PlateauList *list) {
	PlateauLink *link ;
	PlateauLink *next ;
	
	
	// For every plateau
	link = list->plateauLink ;
	while (link != nil) {
		next = link->nextLink ;
		DeletePlateau(link->plateau) ;	// Free the plateau
		delete link ;					// Free the link
		link = next ;
	} ;
	delete list ;						// Free the list itself
}

// --------------------------------------------
// Delete the plateau, but the pixel list is
// now owned by another plateau
// plateaus
//
void FreePlateau(Plateau *plateau) {
	delete plateau->pixelList ;	
	delete plateau ;
}

// --------------------------------------------
// Delete the entire plateau including 
// pixels
//
void DeletePlateau(Plateau *plateau) {
	PixelLink	*link ;
	PixelLink	*next ;
	PixelList	*list ;
	
	// For every Pixel in the list
	list = plateau->pixelList ;
	link = (plateau->pixelList)->pixLink ;
	while (link != nil) {
		next = link->nextLink ;
		delete link ;					// Free the link of pixels
		link = next ;
	} ;
	delete list ;						// Free the list pointer itself
	// Free the plateau
	delete plateau ;
}
