// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------


#include "TApplicationWindow.h"
#include "XDrivers.h"
#include "globals.h"
#include "TMenuBar.h"

#include "MenuCommands.h"

void	TerminateApplication() ; 
TWindow	*GetTargetWindow() ;
void	SetTargetWindow(TWindow *window) ;
extern	short	gDebugMode ;
extern	Boolean	gFrontWindowIsModal ;

TMenuPane::TMenuPane(TPane *parent,short xloc,short yloc,short width,short height) 
	: TPane(parent, xloc, yloc, width, height) {

	SetFgColor(gBlackPixel) ;
	SetBkColor(gGrayPixel) ; 

}


TMenuPane::TMenuPane(TPane *parent,Rect *rect)
	: TPane(parent,rect) {

	SetFgColor(gBlackPixel) ;
	SetBkColor(gGrayPixel) ;

}

void	TMenuPane::Draw(){
	Rect	theRect = mPortRect ;
	SetFgColor(gWhitePixel) ;
	MoveTo(theRect.left,theRect.bottom) ;
	LineTo(theRect.left,theRect.top) ;
	LineTo(theRect.right,theRect.top) ;
	SetFgColor(gBlackPixel) ; 
	LineTo(theRect.right,theRect.bottom) ;
	LineTo(theRect.left,theRect.bottom) ;
	if (mMenuBar != nil)
		mMenuBar->Draw() ;
}

TApplicationWindow::TApplicationWindow(short xloc,short yloc,short width,short height,
			short depth,char *title) 
	: TFrameWindow( xloc,yloc,width,height,depth,title) {

	mWindowType = kFloatingWindow ; // So that we are never the target for the 
								// operations to be made on other windows

	mCanReceiveMenuCommands = false ;
	mBkColor = gGrayPixel ;
	XSetBackground(gDisplay,mGC,mBkColor) ;
	mMenuPane = nil ;
	mMenuBar  = nil ;
	CreateMenuBar() ;
	
}

TApplicationWindow::~TApplicationWindow()  {
	//if (mMenuBar != nil)
	//	delete mMenuBar ;
}


void	TApplicationWindow::CreateMenuBar() {
	mMenuPane = new TMenuPane(this,1,0,mPortRect.right-2,25) ;
	AddPane(mMenuPane) ;
	mMenuBar = new TMenuBar(mMenuPane,mPortRect.right,25) ;
	mMenuPane->SetMenuBar(mMenuBar) ;
}

void	TApplicationWindow::SetupMenuBar() {

	TMenu	*menu1 = new TMenu("File") ;
	menu1->AddMenuItem("New...",kNew) ;
	menu1->AddMenuItem("Open...",kOpen) ;
	menu1->AddMenuItem("Quit",kQuit) ;
	mMenuBar->AddMenu(menu1) ;

	menu1 = new TMenu("Edit") ;
	menu1->AddMenuItem("Undo...") ;
	menu1->AddMenuItem("Copy...") ;
	menu1->AddMenuItem("Cut") ;
	menu1->AddMenuItem("Paste") ;
	mMenuBar->AddMenu(menu1) ;
}

void	TApplicationWindow::DoCommand(TCommand command,Boolean *hasBeenHandled) {
	TWindow *theTargetWindow ;
	switch(command) {
		case kClose:
			theTargetWindow = GetTargetWindow() ;
			if (theTargetWindow != nil) {
				theTargetWindow->Close() ;
				SetTargetWindow(nil) ; // So we can pick up another target window
			}
			break ;			
		case kQuit:
			TerminateApplication() ;
			break ;
		default:
			break ;
	}
}

void	TApplicationWindow::GetCommandStatus(TCommand command,Boolean *isEnabled,Boolean *isChecked) {
	switch(command) {
		case kQuit:
			*isEnabled = true ;
			*isChecked = false ;
			break ;
		case kHelp:
			*isEnabled = true ;
			break ;
		default:
			break ;
	}	
}


void	TApplicationWindow::SetCommandStatus(TMenu *theMenu,TWindow	*theTargetWindow) {
	TMenuItem 	*theMenuItem ;
	TCommand	command ;
	Boolean	isEnabled = false,isChecked = false ;
	
	theMenuItem = theMenu->mMenuItems ;
	while(theMenuItem != nil) {
		if (theMenuItem->mIsHierarchicalMenu) {
			theMenuItem->Enable(true) ;
			SetCommandStatus(theMenuItem->mHierarchicalMenu,theTargetWindow) ;
		}
		else {
			command = theMenuItem->mCommand  ;
			if (command != 0) {
				if (theTargetWindow != nil) 
					theTargetWindow->GetCommandStatus(command,&isEnabled,&isChecked) ;
				GetCommandStatus(command,&isEnabled,&isChecked) ;	
				theMenuItem->Enable(isEnabled) ;
				theMenuItem->Check(isChecked) ;
				isEnabled = false ;	// Reset
				isChecked = false ;	// Reset
			}
		}
		theMenuItem = theMenuItem->mNextItem ;
	}
}

void	TApplicationWindow::DoMouseDownInMenuBar(Point *where) {
	Point thePoint = *where ;
	TCommand	command ;

	// We need to set the status of all the menus
	// Traverse all the menus and get each menuitem and command number
	// and then get the status of each item
	
	
	TWindow		*theTargetWindow = GetTargetWindow() ;

	if (gDebugMode) {
		if (theTargetWindow != nil) {
			char windowName[256] ;
			theTargetWindow->GetWindowTitle(windowName) ;
			fprintf(stderr,"\nDoMouseDownInMenuBar : Target Window = %s",windowName) ;
		}
		else
			fprintf(stderr,"\nDoMouseDownInMenuBar : Target Window = nil") ;
	}

	TMenu *theMenu = mMenuBar->mMenus ;
	short i ;
	for (i=0;i<mMenuBar->mNumMenus;i++) {
		if (theMenu->mEnabled) {
			SetCommandStatus(theMenu,theTargetWindow) ;
		} 
		theMenu = theMenu->mNextMenu ;
	}

	Boolean		hasBeenHandled = false ;
	
	command = mMenuBar->DoMouseDown(&thePoint) ;
	short	menuID = thePoint.x ;
	short	menuItem	= thePoint.y ;
	if (menuID != -1 && menuItem != 0) {
		if (theTargetWindow != nil)
			theTargetWindow->DoCommand(command,&hasBeenHandled) ;
		if (!hasBeenHandled) DoCommand(command,&hasBeenHandled) ;
	}
}

void	TApplicationWindow::Draw() {

	if (!mIsVisible) return ;

	TPane *pane ;
	GrafPtr		tPort ;
	GetPort(&tPort) ;
	SetPort(this) ;
		
	if (mPaneList != nil) {
		for(int i = 0 ; i < mPaneList->GetNumberOfPanes(); i++) {
			pane = mPaneList->GetPane(i) ;
			SetPort(pane) ;
			pane->Show() ;
			pane->Draw() ;
		}
	}	
	SetPort(tPort) ;
}

void	TApplicationWindow::DoMouseDown(XEvent *event) {
	Point thePoint ;
	TPane *pane ;
	GrafPtr	prevPort ;

	if (gFrontWindowIsModal) return ;

	// First find out if we are in one of the panes
	thePoint.x = event->xbutton.x ;
	thePoint.y = event->xbutton.y ;
	// The point is relative to the application window
	// and is not in local coordinates of the pane
	SetFocus(this) ;
	SetPort(this) ;
	if (mPaneList != nil) {
		for(int i = 0 ; i < mPaneList->GetNumberOfPanes(); i++) {
			pane = mPaneList->GetPane(i) ;
			if (pane->InPane(&thePoint)) {
				if (pane == mMenuPane) {
					DoMouseDownInMenuBar(&thePoint) ;
				}
				else {
					GetPort(&prevPort) ;
					SetPort(pane) ;
					pane->DoMouseDown(event) ;
					SetPort(prevPort) ;
				}
				return ;
			}
		}
	}	
}

void	TApplicationWindow::DoMenuKey(XEvent *theEvent) {
	TMenuKey	menuKey ;
	
	if (mMenuPane == nil) {
		return ;
	} 

	int 	count ;
	XComposeStatus compose ;
	char 	InputBuffer[256] ;

	
	count = XLookupString((XKeyEvent *)theEvent,InputBuffer,
			256,&menuKey,&compose) ;
	if (count <= 0)	return  ;
	if (count > 255) count = 255 ;
	InputBuffer[count] = '\0' ;

	if (menuKey == 0) {
		return ;
	}
	
	GrafPtr	tPort ;
	GetPort(&tPort) ;
	
	SetPort(this) ;
	
	
	// Translate the menu key press to a command

	TCommand	command ;
	TWindow		*theTargetWindow ;
	Boolean		hasBeenHandled = false ;
	
	// We need to see if the command is active if it exists
	int i ;
	TMenu *theMenu ;
	TMenuItem *theMenuItem ;
	theMenu = mMenuBar->mMenus ;
	Boolean	isEnabled = false ;
	Boolean	isChecked = false ;
	theTargetWindow = GetTargetWindow() ;
	for (i=0;i<mMenuBar->mNumMenus;i++) {
		if (theMenu->mEnabled) {
			theMenuItem = theMenu->mMenuItems ;
			while(theMenuItem != nil) {
				if (menuKey == theMenuItem->mMenuKey) {					
					command = theMenuItem->mCommand  ;
					if (command != 0) {
						if (theTargetWindow != nil) {
							theTargetWindow->GetCommandStatus(command,&isEnabled,&isChecked) ;
						}
						GetCommandStatus(command,&isEnabled,&isChecked) ;	
						if (isEnabled) {
							if (theTargetWindow != nil) {
								theTargetWindow->DoCommand(command,&hasBeenHandled) ;
							}
							if (!hasBeenHandled) DoCommand(command,&hasBeenHandled) ;
						}
						// Fudge
						SetControlKeyUp() ;
						SetCommandKeyUp() ;
						return ;
					}
				}
				theMenuItem = theMenuItem->mNextItem ;
			} ;
		}
		theMenu = theMenu->mNextMenu ;
	}
	SetPort(tPort) ;
}

void	TApplicationWindow::DoKeyDown(XEvent *theEvent) {
	TPane *pane ;

	// Check the panes
	if (mPaneList != nil) {
		for(int i = 0 ; i < mPaneList->GetNumberOfPanes(); i++) {
			pane = mPaneList->GetPane(i) ;
			pane->DoKeyDown(theEvent) ;
		}
	}	
}
