// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------

#ifndef __TArray__
#define __TArray__
#include <stdio.h>
#include <stdlib.h>
#include "types.h"
#include "Complex.h"
extern	long near(float) ;
typedef enum {kInheritType,kRealArray,kComplexArray} ArrayDataType ;

// 1 Dimensional array (vector) associated with possible two dimensions
class TArray {
	public:
		TArray() { mWidth = 0 ; mHeight = 0 ; mIsComplex = false ; mNumElements = 0 ; mData = nil ; } ;
		TArray(long width,long height,ArrayDataType dataType=kRealArray,float *data=nil) ;
		TArray(long width,ArrayDataType dataType=kRealArray,float *data=nil) ;
		TArray(const TArray& u,ArrayDataType dataType=kInheritType,Rect *rect = nil) ;
		TArray(const TArray* u,ArrayDataType dataType=kInheritType,Rect *rect = nil) ;
		~TArray() ;
		
		float *GetData(long n=0,long m=0) {if(mIsComplex)return (mData+2*(m*mWidth+n)) ; return (mData+m*mWidth+n) ;} ;
		long	Length() { return mNumElements ; } ;
		Boolean	IsComplex() { return mIsComplex ; } ;

		void	Print() ;
		void	SetData(float *data,long w=0,long h=0) ; 
		
		Boolean	IsValid() { return mData != 0 ; } ;
		
		long	GetSize() { return mWidth*mHeight ; } ;
		long    GetWidth() { return mWidth ; } ;
		long    GetHeight() { return mHeight ; } ;
		long	GetByteCount() {if(mIsComplex)return (2*mNumElements*sizeof(float)) ; return (mNumElements*sizeof(float)) ;} ;
		float	*GetReal(long n,long m=0) { return (mData+2*(m*mWidth+n)) ; } ;
		float	*GetImag(long n,long m=0) { return (mData+2*(m*mWidth+n)+1) ; } ;
		float	GetPixelAmplitude(float x,float y=0.) ;
		float	GetPixel(float x,float y) { return GetPixelAmplitude(x,y) ; } ;
		void	SetPixel(float x,float y,float val) ;
		float	GetPixelPhase(float x,float y=0.) { return GetPixelPhase(near(x),near(y)) ; } ;
		float	GetPixelAmplitude(long n,long m=0){
					if (!mIsComplex) return *(mData+m*mWidth+n);
					Cmplx *c=(Cmplx*)GetRealHK(n-mWidth/2,m-mHeight/2); 
					return c->Modulus(); } ;
		float	GetPixelPhase(long n,long m=0){
					if (!mIsComplex) return 0. ;
					Cmplx *c=(Cmplx*)GetRealHK(n-mWidth/2,m-mHeight/2); 
					return c->Phase(); } ;
		float	GetPixelReal(long n,long m=0){
					if (!mIsComplex) return *(mData+m*mWidth+n);
					Cmplx *c=(Cmplx*)GetRealHK(n-mWidth/2,m-mHeight/2); 
					return c->x; } ;
		float	GetPixelImag(long n,long m=0){
					if (!mIsComplex) return 0.;
					Cmplx *c=(Cmplx*)GetRealHK(n-mWidth/2,m-mHeight/2); 
					return c->y; } ;
		void	SetPixelReal(float val,long n,long m=0){
					if (!mIsComplex) return ;
					Cmplx *c=(Cmplx*)GetRealHK(n-mWidth/2,m-mHeight/2); 
					c->x = val ; } ;
		void	SetPixelImag(float val,long n,long m=0){
					if (!mIsComplex) return ;
					Cmplx *c=(Cmplx*)GetRealHK(n-mWidth/2,m-mHeight/2); 
					c->y = val ; } ;
		float	*GetRealHK(long h,long k) ;
		float	*GetImagHK(long h,long k) ;
		TArray	*Real() ;
		TArray	*Imag() ;

		void	MakeComplex() ;
		void	MakeReal() ;

		void	ConvertToComplex() { MakeComplex() ; } ;
		void	ConvertToFloat() { MakeReal() ;} ;

		float	Mean() ;
		float	Min() ;
		float	Max() ;
		float	Min(short &x,short &y) ;
		float	Max(short &x,short &y) ;
		float	Stdv() ;
		float	Sum() ;
		float	VectorLength() ;

		void	Modulus(bool centerImage = false) ;
		void	Amplitude(bool centerImage = false) { Modulus(centerImage) ; } ;
		void	Intensity(bool centerImage = false) ;
		void	Phase(bool centerImage = false) ;
		void	RealPart(bool centerImage = false) ;
		void	ImaginaryPart(bool centerImage = false) ;
		void	log() ;
		void	log10() ;
		void	pow(float x) ;
		void	sqrt() ;
		void	sq() ;
		void	exp() ;
		void	sin() ;
		void	cos() ;
		void	tan() ;
		void	asin() ;
		void	acos() ;
		void	atan() ;
		void	fft() ;									// FFTW Implementation 
		void	ifft(Boolean keepComplex = true) ;		// FFTW Implementation 
		void	FFT() ;
		void	IFFT(Boolean keepComplex = true) ;
		
		float	*GetDataHK(long h,long k) ;	// This is for a function that has 
											// the packing of a  Fourier transform
											// but has been made real
		TArray& operator=(const float u) ;
		TArray& operator=(const Cmplx& u) ;
		TArray& operator=(const TArray& u) ;

		TArray&	operator+=(const float v) ;
		TArray&	operator-=(const float v) ;
		TArray&	operator*=(const float v) ;
		TArray&	operator/=(const float v) ;
		TArray&	operator+=(const Cmplx& v) ;
		TArray&	operator-=(const Cmplx& v) ;
		TArray&	operator*=(const Cmplx& v) ;
		TArray&	operator/=(const Cmplx& v) ;
		TArray&	operator+=(const TArray& v) ;
		TArray&	operator-=(const TArray& v) ;
		TArray&	operator*=(const TArray& v) ;
		TArray&	operator/=(const TArray& v) ;
		
		TArray&	operator^=(const float v) ;


		friend TArray operator*(const TArray& u,const float v) ;
		friend TArray operator/(const TArray& u,const float v) ;
		friend TArray operator+(const TArray& u,const float v) ;
		friend TArray operator-(const TArray& u,const float v) ;
		friend TArray operator*(const float v,const TArray& u) ;
		friend TArray operator/(const float v,const TArray& u) ;
		friend TArray operator+(const float v,const TArray& u) ;
		friend TArray operator-(const float v,const TArray& u) ;

		friend TArray operator^(const TArray& u,const float v) ;

//LDM Test: was friend  float   InnerProduct(const TArray& u,const TArray& u) ;
		friend	float 	InnerProduct(const TArray& u,const TArray& v) ;
		friend	float	Min(TArray& u)  { return u.Min() ; }  ;
		friend	float	Max(TArray& u)  { return u.Max() ; } ;
		friend	float	Mean(TArray& u) { return u.Mean() ; }  ;
		friend	float	Stdv(TArray& u) { return u.Stdv() ; }  ;
		friend	float	Sum(TArray& u)  { return u.Sum() ; } ;

		friend TArray operator*(const TArray& u,const Cmplx& v) ;
		friend TArray operator/(const TArray& u,const Cmplx& v) ;
		friend TArray operator+(const TArray& u,const Cmplx& v) ;
		friend TArray operator-(const TArray& u,const Cmplx& v) ;
		friend TArray operator*(const Cmplx& v,const TArray& u) ;
		friend TArray operator/(const Cmplx& v,const TArray& u) ;
		friend TArray operator+(const Cmplx& v,const TArray& u) ;
		friend TArray operator-(const Cmplx& v,const TArray& u) ;

		friend TArray operator*(const TArray& u,const TArray& v) ;
		friend TArray operator/(const TArray& u,const TArray& v) ;
		friend TArray operator+(const TArray& u,const TArray& v) ;
		friend TArray operator-(const TArray& u,const TArray& v) ;

		friend TArray Modulus(TArray &u) ;
		friend TArray Intensity(TArray &u) ;
		friend TArray Phase(TArray &u) ;
		friend TArray log(TArray &u) ;
		friend TArray log10(TArray &u) ;
		friend TArray pow(TArray &u,float x) ;
		friend TArray sqrt(TArray &u) ;
		friend TArray sq(TArray &u) ;
		friend TArray exp(TArray &u) ;
		friend TArray sin(TArray &u) ;
		friend TArray cos(TArray &u) ;
		friend TArray tan(TArray &u) ;
		friend TArray asin(TArray &u) ;
		friend TArray acos(TArray &u) ;
		friend TArray atan(TArray &u) ;
		friend TArray Real(TArray &u) ;
		friend TArray Imag(TArray &u) ;
		
		friend	void	FourierTransform(TArray *u) { u->FFT() ; } ;
		friend	void	InverseFourierTransform(TArray *u) { u->IFFT() ; } ;
		
		friend	TArray *Intensity(TArray *v) { TArray *u = new TArray(v) ; u->Intensity() ; return u ; } ;
		friend	TArray *Amplitude(TArray *v) { TArray *u = new TArray(v) ; u->Modulus() ; return u ; } ;
		friend	TArray *Phase(TArray *v) { TArray *u = new TArray(v) ; u->Phase() ; return u ; } ;
		
		
	protected:
		float	*mData ;
		long	mNumElements ;
		long	mWidth ;
		long	mHeight ;
		Boolean	mIsComplex ;
} ;


Boolean	AssertArray(TArray **array) ;


#endif
