#ifndef __TButton_H
#define __TButton_H

#include "types.h"
extern	unsigned long 	gBlackPixel ;
extern	unsigned long 	gWhitePixel ;
extern	unsigned long 	gRedPixel ;
extern	unsigned long 	gGreenPixel ;
extern	unsigned long 	gBluePixel ;
extern	unsigned long 	gGrayPixel ;
extern	unsigned long 	gDarkerGray ;
extern	unsigned long 	gDarkGray ;
extern	unsigned long 	gLighterGray ;

#ifdef VMS
	#include <decw$include/keysym.h>
#else
	#include <X11/keysym.h>
#endif

#include <string.h>

typedef char	Str64[64] ;
typedef char	Str256[256] ;

typedef enum {kLabelInside,kLabelOutsideLeft,kLabelOutsideTop,kLabelOutsideRight,kLabelOutsideBottom}
LabelPosition ;
typedef enum {kHorizontal,kVertical,kDiagonal } Orientation ;

Boolean	PtInRect(Point *,Rect *) ;

class	TIcon {
	public:
		TIcon(short	width,short height,unsigned char *data=nil,
			Drawable drawable=nil) ;
		~TIcon() ;
		void	Draw(short x,short y) ;
		Pixmap	GetPixmap() { return mPixmap ; }  ;
		short	GetWidth() { return mWidth ; } ;
		short	GetHeight(){ return mHeight ; } ;
	private:
		Pixmap  mPixmap ;
		short	mWidth ;
		short	mHeight ;
} ;

class	TPane ;
class	TGrafPort ;
extern	void SetPort(TGrafPort *) ;
extern	void GetPort(TGrafPort **) ;

class TObject ;
typedef void (*ObjectActionRoutine)(TObject *) ;

class	TObject {
	public:
		TObject(Rect *rect=nil,char *label = nil) ;
		TObject(short x1,short y1, short x2, short y2,char *label = nil) ;
		
		virtual				~TObject() ;
		void				SetParentWindow(TPane *pane) { mParentWindow = pane ; } ;
		TPane				*GetParentWindow() { return mParentWindow ; } ;
		void				Enable() { mIsActive = true ; Update(); } ; 
		void				Disable() { mIsActive = false ; Update(); } ; 
		void				Activate(Boolean active) { mIsActive = active ; Update() ; } ;
		void				SetVisible(Boolean visible) { mIsVisible = visible ; Update() ; } ;
		void				SetRefCon(unsigned long refCon) { mRefCon = refCon ; } ;
		unsigned long 		GetRefCon() { return mRefCon ; } ;
		void				SetObjectRect(Rect &theRect) { mRect = theRect ; } ;
		Boolean				IsVisible() { return mIsVisible ; } ;
		Boolean				GetState() { return mState ; } ;
		Boolean				IsEnabled() { return mIsActive ; } ;
		virtual	void 		SetState(Boolean state) { mState = state; Update() ; } ;
		Rect				&GetRect() { return mRect ; } ;
		virtual Boolean 	InObject(Point *pt) { if (mIsActive && mIsVisible) return PtInRect(pt,&mRect); else return false ; } ;
		virtual	void		DoEvent(XEvent *) ;
		virtual	Boolean		DoMouseDown(Point *) {return true ;} ;
		virtual KeySym 		DoKeyDown(XEvent *) { return 1 ;} ;
		virtual void 		Draw() {} ;
		virtual void 		Erase() {} ;
		virtual	void		Update() ;
		void				SetItemText(char *text) ;
		void				GetItemText(char *text) ;
		void				SetLabelPosition(LabelPosition pos) {mLabelPosition = pos ;} ;
		void 				SetJustification(JustificatonType type) {mJustification=type; Draw();};
		void				SetCallBackRoutine(ObjectActionRoutine cb) { mActionRoutine = cb ; } ;
		void				SetTrackingRoutine(ObjectActionRoutine cb) { mTrackingRoutine = cb ; } ;
	protected:
		void				CreateLabel(char *text) ;
		virtual		void	DrawLabel() {} ;
		Rect				mRect ;
		unsigned	long	mRefCon ;
		Boolean				mIsActive ;
		Boolean				mState ;
		Boolean				mIsVisible ;
		ObjectActionRoutine	mActionRoutine ;
		ObjectActionRoutine	mTrackingRoutine ;
		Str256				*mLabel ;
		TPane				*mParentWindow ;
		short				mNumLabelLines ;
		LabelPosition		mLabelPosition ;
		JustificatonType	mJustification ;
} ;

class	TUserRect : public TObject {
	public:
		TUserRect(Rect *rect) : TObject(rect) { mIsActive = false ;} ;
		TUserRect(short x1,short y1,short x2,short y2):TObject(x1,y1,x2,y2) {mIsActive=false;};
		void Draw() {} ;
} ;

class	TFrame : public TObject {
	public:
		TFrame(Rect *rect=nil,unsigned long col = gWhitePixel) : TObject(rect) { 	mColorIndex = col ; mIsActive = false ;} ;
		TFrame(short x1,short y1, short x2, short y2,unsigned long col = 0) 
			: TObject(x1,y1,x2,y2) { mColorIndex = col ; mIsActive = false ;} ;
		void Draw() ;
		unsigned long	mColorIndex ;
} ;

class	TGroupBox : public TObject {
	public:
		TGroupBox(Rect *rect,char *label=nil) ;
		TGroupBox(short x1,short y1, short x2, short y2, char *label=nil) ;
		virtual	~TGroupBox() ;
		void Draw() ;
	protected:
		void	DrawLabel() ;
} ;


class TLine : public TObject { 
	public:
		TLine(short x1,short y1, short x2, short y2,unsigned long col = gWhitePixel) ;
		virtual void Draw() ;
	protected:
		float	mCosTheta ;
		float	mSinTheta ;
		unsigned long	mColorIndex ;
} ;

class TAnnotationLine : public TLine {
	public:
		TAnnotationLine(short x1,short y1,short x2,short y2,unsigned long col = gWhitePixel) 
			: TLine(x1,y1,x2,y2,col) { mIsActive = true ;}  ;
		Boolean DoMouseDown(Point *) ;
} ;

class	TAnnotationRect : public TFrame {
	public:
		TAnnotationRect(Rect *rect=nil,unsigned long col = gWhitePixel) : TFrame(rect,col) {mIsActive = true ;} ;
		TAnnotationRect(short x1,short y1,short x2,short y2,unsigned long col = 0) 
			: TFrame(x1,y1,x2,y2,col) { mIsActive = true ;}  ;
		Boolean DoMouseDown(Point *) ;
} ;

class	TAnnotationTextRect : public TAnnotationRect {
	public:
		TAnnotationTextRect(char *text,Rect *rect=nil,unsigned long col = gWhitePixel) : TAnnotationRect(rect,col) {mIsActive = true ; strcpy(mText,text) ; } ;
		TAnnotationTextRect(char *text,short x1,short y1,short x2,short y2,unsigned long col = 0) 
			: TAnnotationRect(x1,y1,x2,y2,col) { mIsActive = true ; strcpy(mText,text) ; }  ;
		void Draw() ;
	private:
		char mText[256] ;
} ;

class	TAnnotationCircle : public TObject {
	public:
		TAnnotationCircle(short xc,short yc, short rad,unsigned long col = gWhitePixel) ;
		Boolean InObject(Point *pt) ;
		Boolean DoMouseDown(Point *) ;
		void	GetCenterX(short &x,short &y) { x =  mXC ; y = mYC ; } ;
		short	GetRadius() { return mRad ; } ;
		void	SetCenter(short x,short y) { mXC = x ; mYC = y ; };
		void	SetRadius(short r) { mRad = r ; } ;
		void	Draw() ;
	protected:
		short	mXC ;
		short	mYC ;
		short	mRad ;
		unsigned long	mColorIndex ;

} ;

class	TControl : public TObject {
	public:
		TControl(short x1,short y1,short x2,short y2,char *title=nil,
			float minVal=0.,float maxVal=1.,float curVal=0.5) 
			: TObject(x1,y1,x2,y2,title),mMinValue(minVal),mMaxValue(maxVal),
			  mCurrentValue(curVal) {} ;
		void	SetControlValue(float val) { Erase() ; mCurrentValue=val ; Update() ; } ; 
		void	SetControlMinimum(float val) { Erase() ; mMinValue = val ;  Update() ; } ;
		void	SetControlMaximum(float val) { Erase() ; mMaxValue = val ;  Update() ; } ;
		float	GetControlValue() { return mCurrentValue ; } ;
		void	SetControlBounds(float val1,float val2) { mMinValue = val1 ; mMaxValue = val2 ; } ;
		void	SetControlIncrement(float val) { mStepValue = val ; } ;
		virtual	Boolean DoMouseDown(Point *) {return true ;} ;
		virtual void Draw() {} ;
	protected:
		virtual	void	DrawMinValue() {} ;
		virtual	void	DrawMaxValue() {} ;
		virtual	void	DrawCurrentValue() {} ;
		virtual	void	DrawControl() {} ;
		virtual	void EraseControl() {} ;
		float	mStepValue ;
		float	mMinValue ;
		float	mMaxValue ;
		float	mCurrentValue ;
} ;

class	TSlider : public TControl {
	public:
		TSlider(short x1,short y1, short x2, short y2,char *title=nil,
			float minVal=0.,float maxVal=1.,float curVal=0.5)
				: TControl(x1,y1,x2,y2,title,minVal,maxVal,curVal) { mOrientation = (mRect.top == mRect.bottom) ? kHorizontal : kVertical ; SetControlRect() ;}  ;
		virtual void Draw() ;
		Boolean InObject(Point *pt) {if (mIsVisible && mIsActive) return PtInRect(pt,&mControlRect) ; else return false ; } ;
		virtual	void	Update() ;
		Boolean DoMouseDown(Point *) ;
	protected:
		virtual	void	DrawMinValue() ;
		virtual	void	DrawMaxValue() ;
		virtual	void	DrawCurrentValue() ;
		virtual	void	DrawControl() ;	
		virtual	void	DrawLabel() ;
		virtual	void	MoveControl(short pos) ;
		virtual	void	SetControlRect() ;
		virtual void 	Erase() ;
		virtual	void 	EraseControl() ;
		Rect	mControlRect ;
		Orientation	mOrientation ;
} ;

class	TScrollBar : public TControl {
	public:
		TScrollBar(short x1,short y1, short x2, short y2,Orientation orientation = kVertical,
			float minVal=0.,float maxVal=1.,float curVal=0.5) ;
		virtual void Draw() ;
		void			Resize(short x1,short y1, short x2, short y2,float maxValue) ;
		virtual	void	Update() {if (mParentWindow==nil) return ;
					TGrafPort *tp;GetPort(&tp);
					SetPort((TGrafPort *)mParentWindow); SetControlRect() ; Draw();SetPort(tp);};
		Boolean DoMouseDown(Point *) ;
	protected:
		virtual	void	DrawControl() ;	
		virtual	void	MoveControl(short pos) ;
		virtual	void	SetControlRect() ;
		virtual	void 	EraseControl() ;
		virtual	void	DrawPageUp() ;
		virtual	void	DrawPageDown() ;
		Rect	mThumbRect ;
		Rect	mPageUpRect ;
		Rect	mPageDownRect ;
		Rect	mScrollRect ;
		short	mThumbSize ;
		Orientation	mOrientation ;
} ;

class TButton : public TObject {
	public:
		TButton(Rect *rect,char *title=nil,ObjectActionRoutine actionRoutine = nil) ;
		TButton(short x1,short y1, short x2, short y2,
			char *title=nil,ObjectActionRoutine actionRoutine = nil) ;
		virtual	~TButton() ;
		virtual void Draw() ;
		virtual	Boolean DoMouseDown(Point *) ;
		Boolean	GetControlValue() { return mState ; } ;
		void	SetControlValue(Boolean state) {mState=state;Update();} ;
	protected:
		Boolean	mIsDefault ;
		void	DrawLabel() ;
} ;

class TIconButton : public TButton { 
	public:
		TIconButton(Rect *rect,TIcon *icon=nil,char *title=nil,Boolean state = OFF) ;
		TIconButton(short x1,short y1, short x2, short y2,TIcon *icon=nil,
			char *title=nil,Boolean state = OFF) ;
		void Draw() ;
		virtual	~TIconButton() ;
	protected:
		TIcon	*mIcon ;
		short	mIconPosX ;
		short	mIconPosY ;
} ;

// Push button is a tool button and stays down when pressed
// It has a toolconstant ID that can be used to get which tool is
// selected through the (TPushButton *)gCurrentTool->GetToolID() ;
class TPushButton : public TButton { 
	public:
		TPushButton(Rect *rect,TIcon *icon=nil,short id=0,char *title=nil,Boolean state = OFF) ;
		TPushButton(short x1,short y1, short x2, short y2,TIcon *icon=nil,
			short id=0,char *title=nil,Boolean state = OFF) ;
		Boolean DoMouseDown(Point *) ;
		short	GetToolID() { return mToolID ; } ; 
		void Draw() ;
		virtual	~TPushButton() ;
	protected:
		TIcon	*mIcon ;
		short	mIconPosX ;
		short	mIconPosY ;
		short	mToolID ;
} ;

class TCheckBox : public TButton {
	public:
		TCheckBox(Rect *rect,char *title=nil,Boolean state = OFF) ;
		TCheckBox(short x1,short y1, short x2, short y2,
			char *title=nil,Boolean state = OFF) ;
		Boolean DoMouseDown(Point *) ;
		void Draw() ;
		virtual	~TCheckBox() ;
	protected:
		void DrawPressed(Boolean upDwn) ;
		void DrawLabel() ;
} ;

class TRadioButton : public TButton {
	public:
		TRadioButton(Rect *rect,char *title=nil,Boolean state = OFF) ;
		TRadioButton(short x1,short y1, short x2, short y2,
			char *title=nil,Boolean state = OFF) ;
		virtual	~TRadioButton() ;
		void Draw() ;
		Boolean DoMouseDown(Point *) ;
	protected:
		void DrawPressed(Boolean upDwn) ;
		void	DrawLabel() ;
} ;

class TMenuButton : public TButton {
	public:
		TMenuButton(Rect *rect,char *text=nil,char *label=nil,LabelPosition pos=kLabelOutsideLeft) ;
		TMenuButton(short x1,short y1, short x2, short y2,
			char *text=nil,char *label=nil,LabelPosition pos=kLabelOutsideLeft) ;
		virtual	~TMenuButton() ;
		void Draw() ;
		void	SetMenuText(char *text) { strcpy(mMenuText,text) ; } ; 
		virtual	Boolean DoMouseDown(Point *) ;
	protected:
		void	DrawLabel() ;
		void	DrawMenuText() ;
		char	mMenuText[256] ;
		
} ;

class TStaticText : public TObject {
	public:
		TStaticText(Rect *rect,char *title=nil) ;
		TStaticText(short x1,short y1, short x2, short y2,char *title=nil) ;
		virtual	~TStaticText() ;
		void SetItemText(char *text) ;
		void GetItemText(char *text) ;
		void Draw() ;
	protected:
		void	DrawLabel() ;
} ;
typedef enum {kNumericInput,kGenericInput} TextInputType ;

class TEditText : public TObject {
	public:
		TEditText(Rect *rect,char *title,char *text=nil) ;
		TEditText(short x1,short y1, short x2, short y2,char *title,char *text=nil) ;
		virtual	~TEditText() ;
		void 	SetEditText(char *text) ;
		void 	GetEditText(char *text) ;
		virtual	Boolean DoMouseDown(Point *) ;
		virtual	KeySym	DoKeyDown(XEvent *event) ;
		virtual	void 	Draw() ;
		void	SelectAll() ;
		void	SetFocus(Boolean focus) ;
		void	Deselect() { SetFocus(false);} ;
	protected:
		char	*mText ;
		Point	mInsertionPt ;
		Boolean	mIsFocus ;
		short	mFirstSelChar,mNumSelChars ;
		short	FindCharacterToLeft(short x) ;
		short	FindCharacterToRight(short x) ;
		TextInputType	mInputType ;
} ;

class TObjectEntry {	
	public:
		TObjectEntry() ;
		~TObjectEntry() ;
		friend class TObjectList ;
	private:	
		TObject			*mObject ;
		TObjectEntry	*mNextObjectEntry ;
}  ;

class TObjectList {
	public:
		TObjectList() ;
		~TObjectList() ;
		void 	AddObject(TObject *theObject) ;
		void	RemoveObject(TObject *theObject) ;
		short	GetNumberOfObjects(void) { return mNumObjects ; } ;
		TObject	*GetObject(int number) ;
		TObject *GetObject(Point *pt) ;
		void	DrawObjects() ;
	private:
		short	mNumObjects ;
		TObjectEntry	*mObjectEntry ;
	
}  ;


class TEditTextEntry {	
	public:
		TEditTextEntry() { mNextEditTextEntry = nil ; mObject = nil ; } ;
		~TEditTextEntry() { mNextEditTextEntry = nil ; } ;
		friend class TEditTextList ;
	private:	
		TObject			*mObject ;
		TEditTextEntry	*mNextEditTextEntry ;
}  ;

class TEditTextList {
	public:
		TEditTextList() { mNumObjects = 0 ; mEditTextEntry = nil ; } ;
		~TEditTextList() { if (mEditTextEntry != nil) delete mEditTextEntry ; } ;
		void 	AddObject(TObject *theObject) ;
		void	RemoveObject(TObject *theObject) ;
		TEditText	*GetNextField(TEditText *current) ;
		short	GetNumberOfObjects(void) { return mNumObjects ; } ;
		TObject	*GetObject(int number) ;
		TObject *GetObject(Point *pt) ;
		int		GetObjectIndex(TObject *obj) ;
	private:
		short	mNumObjects ;
		TEditTextEntry	*mEditTextEntry ;
	
}  ;
#endif // __TButton__ 

