// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------


#include "TCrystalDiffractionDataTable.h"
#include "TRefinementSet.h"
#include "DMMenuCommands.h"
#include "XDrivers.h" 
#include <math.h>
#include <stdio.h>
#define	MonitorWidth	1024
#define	MonitorHeight	768 
void	PutMessage(char *) ;
extern	Point	gImageDisplayPosition ;

#define log2(x) log(x)/log(2.0) 

#define max(x,y) (x) > (y) ? (x) : (y)
#define min(x,y) (x) < (y) ? (x) : (y)

extern	double Pi ;
extern	char	*gPlaneGroup[] ;

#define kUnAssignedPhase	-9999.

TCrystalDiffractionDataTable::TCrystalDiffractionDataTable(char *title,TDiffractionPeaks *diffractionPeaks,short planeGroup) 
	: TTable(title) {
	short 	h,k,hDataMax,kDataMax,hStart,groupIndex ;
	float	amplitude,astar,bstar,gammastar,twocsgastar,astarSq,bstarSq,ksqbstsq,
			astbstcsgast,kbst2csgast,csgastar ;
	long	i ;
	
	TArray	*dataArray = diffractionPeaks->mFourierData ;
	hDataMax = dataArray->GetWidth()/2 -1 ;
	kDataMax = dataArray->GetHeight()/2 -1 ;
	
	mRefinementSet = nil ;
	mPlaneGroup = planeGroup ;
	mGamma = 180 - diffractionPeaks->mGammaStar ;
	mCalibrationUnit = diffractionPeaks->mCalibrationUnit ;	
	
	// First find out how many diffraction spots we have
	mNumReflections = 0 ;
	for(k = 0; k < kDataMax; k++)  {
		hStart = k == 0 ? 1 : -hDataMax ;
		for(h = hStart; h < hDataMax; h++) {
			if (!(h == 0 && k == 0)) {
				amplitude = *dataArray->GetRealHK(h,k) ;
				if (amplitude > 0.) mNumReflections++ ;
			}
		}
	}

	mGVector 	= new float[mNumReflections] ;
	mAmplitude 	= new float[mNumReflections] ;
	mPhase 		= new float[mNumReflections] ;
	mH			= new short[mNumReflections] ;
	mK			= new short[mNumReflections] ;
	mIndex		= new short[mNumReflections] ;
	
	for (i=0; i < mNumReflections;  mIndex[i++] = 0.)  ;				

	switch(planeGroup) {
		case 1:	// p1
		case 2: // p2
			astar = diffractionPeaks->mAStar ;
			bstar = diffractionPeaks->mBStar ;
			gammastar = diffractionPeaks->mGammaStar ;
			mA =  diffractionPeaks->mA ;
			mB =  diffractionPeaks->mB ;
			break ;
		case 3: // pm -  
		case 4: // pg 
		case 5: // cm 
		case 6: // pmm  -
		case 7: // pmg  
		case 8: // pgg  
		case 9: // cmm  
			astar = diffractionPeaks->mAStar ;
			bstar = diffractionPeaks->mBStar ;
			gammastar = 90. ;
			mA =  diffractionPeaks->mA ;
			mB =  diffractionPeaks->mB ;
			break ;
		case 10: // p4
		case 11: // p4m
		case 12: // p4g
			astar = (diffractionPeaks->mAStar+diffractionPeaks->mBStar)/2. ;
			bstar = astar ;
			gammastar = 90. ;
			mA =  0.5*(diffractionPeaks->mA+ diffractionPeaks->mB) ;
			mB =  mA ;
			break ;
		case 13: // p3
		case 14: // p3m1
		case 15: // p31m
		case 16: // p6
		case 17: // p6m
			// Gamma = 120
			astar = (diffractionPeaks->mAStar+diffractionPeaks->mBStar)/2. ;
			bstar = astar ;
			gammastar = 60. ;
			mA =  0.5*(diffractionPeaks->mA+ diffractionPeaks->mB) ;
			mB =  mA ;
			break ;
	}
	
	mGamma = 180. - gammastar ;
	
	astarSq = astar*astar ;
	bstarSq = bstar*bstar ;
	csgastar = cos(gammastar*asin(1.0)/90.) ;
	twocsgastar = 2.*cos(gammastar*asin(1.0)/90.) ;
	
	i = 0 ;
	
	TArray	*IsSet = new TArray(dataArray->GetWidth(),dataArray->GetHeight()) ;
	*IsSet = 0. ;
	
	groupIndex = 1 ;
	astbstcsgast = astar*bstar*twocsgastar ;
	float	radToDegree = 90.0/asin(1.0) ;
	bool	isSymmetryRelated ;
	// Do only upper half plane
	for(k = 0; k < kDataMax; k++) {
		ksqbstsq = k*k*bstarSq ;
		kbst2csgast = k*astbstcsgast ;
		hStart = k == 0 ? 1 : -hDataMax ;
		for(h = hStart; h < hDataMax; h++) {
			if (!(h == 0 && k == 0)) {
				amplitude = *dataArray->GetRealHK(h,k) ;
				if (amplitude > 0.) {
					mGVector[i] = ::sqrt(h*h*astarSq + ksqbstsq + h*kbst2csgast) ;
					mAmplitude[i] = amplitude ;
					mPhase[i] = radToDegree*(*dataArray->GetImagHK(h,k)) ;
					mH[i] = h ;
					mK[i] = k ;
					if (*IsSet->GetDataHK(h,k) == 0.) {
						// We first need to check to see if the reflection is symmetry related
						// to one that has been set, but which did not set this one
						isSymmetryRelated = false ;
						switch(mPlaneGroup) {
							case 1:	// p1
							case 2: // p2
								break ;
							case 3: // pm	|F(h,k)| = |F(-h,k)|
							case 4: // pg	|F(h,k)| = |F(-h,k)|
							case 5: // cm	|F(h,k)| = |F(-h,k)|
							case 6: // pmm	|F(h,k)| = |F(-h,k)|
							case 7: // pmg  |F(h,k)| = |F(-h,k)|
							case 8: // pgg 	|F(h,k)| = |F(-h,k)|
							case 9: // cmm	|F(h,k)| = |F(-h,k)|
								if (*IsSet->GetDataHK(-h,k) != 0.)  { isSymmetryRelated = true ; mIndex[i] = *IsSet->GetDataHK(-h,k) ; }
								break ;
							case 10: // p4	|F(h,k)| = |F(-k,h)|
								if (*IsSet->GetDataHK(-k,h) != 0.)  { isSymmetryRelated = true ; mIndex[i] = *IsSet->GetDataHK(-k,h) ; }
								break ;
							case 11: // p4m |F(h,k)| = |F(-k,h)| = |F(-h,k)|
							case 12: // p4g	|F(h,k)| = |F(-k,h)| = |F(-h,k)|
								if (*IsSet->GetDataHK(-k,h) != 0.)  { isSymmetryRelated = true ; mIndex[i] = *IsSet->GetDataHK(-k,h) ; }
								if (*IsSet->GetDataHK(-h,k) != 0.)  { isSymmetryRelated = true ; mIndex[i] = *IsSet->GetDataHK(-h,k) ; }
								break ;
							case 13: // p3	|F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)|
								if (*IsSet->GetDataHK(k,-h-k) != 0.)  { isSymmetryRelated = true ; mIndex[i] = *IsSet->GetDataHK(k,-h-k) ; }
								if (*IsSet->GetDataHK(-h-k,h) != 0.)  { isSymmetryRelated = true ; mIndex[i] = *IsSet->GetDataHK(-h-k,h) ; }
								break ;
							case 14: // p3m1	|F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)| = |F(k,h)|
							case 15: // p31m	|F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)| = |F(k,h)|
							case 17: // p6m		|F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)| = |F(k,h)|
								if (*IsSet->GetDataHK(k,-h-k) != 0.)  { isSymmetryRelated = true ; mIndex[i] = *IsSet->GetDataHK(k,-h-k) ; }
								if (*IsSet->GetDataHK(-h-k,h) != 0.)  { isSymmetryRelated = true ; mIndex[i] = *IsSet->GetDataHK(-h-k,h) ; }
								if (*IsSet->GetDataHK(k,h) != 0.)  	  { isSymmetryRelated = true ; mIndex[i] = *IsSet->GetDataHK(k,h) ; }
								break ;
							case 16: // p6	|F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)|
								if (*IsSet->GetDataHK(k,-h-k) != 0.)  { isSymmetryRelated = true ; mIndex[i] = *IsSet->GetDataHK(k,-h-k) ; }
								if (*IsSet->GetDataHK(-h-k,h) != 0.)  { isSymmetryRelated = true ; mIndex[i] = *IsSet->GetDataHK(-h-k,h) ; }
								break ;
						}
						
						if (isSymmetryRelated) {
							*IsSet->GetDataHK(h,k) = mIndex[i] ;
							*IsSet->GetDataHK(-h,-k) = mIndex[i] ;
						}
						else {
							mIndex[i] = groupIndex ;
							*IsSet->GetDataHK(h,k) = (float)groupIndex ;
							// Set symmetry related reflection
							switch(mPlaneGroup) {
								case 1:	// p1
								case 2: // p2
									break ;
								case 3: // pm	|F(h,k)| = |F(-h,k)|
								case 4: // pg	|F(h,k)| = |F(-h,k)|
								case 5: // cm	|F(h,k)| = |F(-h,k)|
								case 6: // pmm	|F(h,k)| = |F(-h,k)|
								case 7: // pmg  |F(h,k)| = |F(-h,k)|
								case 8: // pgg 	|F(h,k)| = |F(-h,k)|
								case 9: // cmm	|F(h,k)| = |F(-h,k)|
									*IsSet->GetDataHK(-h,k) = groupIndex ;
									break ;
								case 10: // p4	|F(h,k)| = |F(-k,h)|
									*IsSet->GetDataHK(-k,h) = groupIndex ;
									*IsSet->GetDataHK(k,-h) = groupIndex ;
									break ;
								case 11: // p4m |F(h,k)| = |F(-k,h)| = |F(-h,k)|
								case 12: // p4g	|F(h,k)| = |F(-k,h)| = |F(-h,k)|
									*IsSet->GetDataHK(-k,h) = groupIndex ;
									*IsSet->GetDataHK(-h,k) = groupIndex ;
									*IsSet->GetDataHK(-h,-k) = groupIndex ;
									*IsSet->GetDataHK(k,-h) = groupIndex ;
									*IsSet->GetDataHK(h,-k) = groupIndex ;
									break ;
								case 13: // p3	|F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)|
									*IsSet->GetDataHK(k,-h-k) = groupIndex ;
									*IsSet->GetDataHK(-h-k,h) = groupIndex ;
									*IsSet->GetDataHK(-h,-k) = groupIndex ;
									*IsSet->GetDataHK(-k,h+k) = groupIndex ;
									*IsSet->GetDataHK(h+k,-h) = groupIndex ;
									break ;
								case 14: // p3m1	|F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)| = |F(k,h)|
								case 15: // p31m	|F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)| = |F(k,h)|
								case 17: // p6m		|F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)| = |F(k,h)|
									*IsSet->GetDataHK(k,-h-k) = groupIndex ;
									*IsSet->GetDataHK(-h-k,h) = groupIndex ;
									*IsSet->GetDataHK(k,h) = groupIndex ;
									*IsSet->GetDataHK(-h,-k) = groupIndex ;
									*IsSet->GetDataHK(-k,h+k) = groupIndex ;
									*IsSet->GetDataHK(h+k,-h) = groupIndex ;
									*IsSet->GetDataHK(-k,-h) = groupIndex ;
									break ;
								case 16: // p6	|F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)|
									*IsSet->GetDataHK(k,-h-k) = groupIndex ;
									*IsSet->GetDataHK(-h-k,h) = groupIndex ;
									*IsSet->GetDataHK(-h,-k) = groupIndex ;
									*IsSet->GetDataHK(-k,h+k) = groupIndex ;
									*IsSet->GetDataHK(h+k,-h) = groupIndex ;
									break ;
							}
							groupIndex++ ;
						}
					}
					else {
						mIndex[i] = (short) *IsSet->GetDataHK(h,k) ;
						switch(mPlaneGroup) {
							case 1:	// p1
							case 2: // p2
								break ;
							case 3: // pm	|F(h,k)| = |F(-h,k)|
							case 4: // pg	|F(h,k)| = |F(-h,k)|
							case 5: // cm	|F(h,k)| = |F(-h,k)|
							case 6: // pmm	|F(h,k)| = |F(-h,k)|
							case 7: // pmg  |F(h,k)| = |F(-h,k)|
							case 8: // pgg 	|F(h,k)| = |F(-h,k)|
							case 9: // cmm	|F(h,k)| = |F(-h,k)|
								*IsSet->GetDataHK(-h,k) = mIndex[i] ;
								break ;
							case 10: // p4	|F(h,k)| = |F(-k,h)|
								*IsSet->GetDataHK(-k,h) = mIndex[i] ;
								*IsSet->GetDataHK(k,-h) = mIndex[i] ;
								break ;
							case 11: // p4m |F(h,k)| = |F(-k,h)| = |F(-h,k)|
							case 12: // p4g	|F(h,k)| = |F(-k,h)| = |F(-h,k)|
								*IsSet->GetDataHK(-k,h) = mIndex[i] ;
								*IsSet->GetDataHK(-h,k) = mIndex[i] ;
								*IsSet->GetDataHK(-h,-k) = mIndex[i] ;
								*IsSet->GetDataHK(k,-h) = mIndex[i] ;
								*IsSet->GetDataHK(h,-k) = mIndex[i] ;
								break ;
							case 13: // p3	|F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)|
								*IsSet->GetDataHK(k,-h-k) = mIndex[i] ;
								*IsSet->GetDataHK(-h-k,h) = mIndex[i] ;
								*IsSet->GetDataHK(-h,-k) = mIndex[i] ;
								*IsSet->GetDataHK(-k,h+k) = mIndex[i] ;
								*IsSet->GetDataHK(h+k,-h) = mIndex[i] ;
								break ;
							case 14: // p3m1	|F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)| = |F(k,h)|
							case 15: // p31m	|F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)| = |F(k,h)|
							case 17: // p6m		|F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)| = |F(k,h)|
								*IsSet->GetDataHK(k,-h-k) = mIndex[i] ;
								*IsSet->GetDataHK(-h-k,h) = mIndex[i] ;
								*IsSet->GetDataHK(k,h) = mIndex[i] ;
								*IsSet->GetDataHK(-h,-k) = mIndex[i] ;
								*IsSet->GetDataHK(-k,h+k) = mIndex[i] ;
								*IsSet->GetDataHK(h+k,-h) = mIndex[i] ;
								*IsSet->GetDataHK(-k,-h) = mIndex[i] ;
								break ;
							case 16: // p6	|F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)|
								*IsSet->GetDataHK(k,-h-k) = mIndex[i] ;
								*IsSet->GetDataHK(-h-k,h) = mIndex[i] ;
								*IsSet->GetDataHK(-h,-k) = mIndex[i] ;
								*IsSet->GetDataHK(-k,h+k) = mIndex[i] ;
								*IsSet->GetDataHK(h+k,-h) = mIndex[i] ;
								break ;
						}
					}
					i++ ;
				}
			}
		}
	}
	
	delete IsSet ;
	
	mSortKey = kSortOnG ;
	SortOnG() ;
		
	
	mRows = mNumReflections ;
	mCols = 4 ;
	mCellDataSize = 4 ;
	mColWidth = 80 ;
	mRowHeight = 20 ;
	
	Int16 	xpos = gImageDisplayPosition.x ;
	Int16 	ypos = gImageDisplayPosition.y ;
	gImageDisplayPosition.x += 10 ;
	gImageDisplayPosition.y += 10 ;
				
	mParentPane = new TTableWindow(xpos,ypos,this) ;		
		
}

TCrystalDiffractionDataTable::~TCrystalDiffractionDataTable() {
	if (mGVector != nil) delete[] mGVector ;
	if (mAmplitude != nil) delete[] mAmplitude ;
	if (mPhase != nil) delete[] mPhase ;
	if (mH != nil) delete[] mH ;
	if (mK != nil) delete[] mK ;
	if (mIndex != nil) delete[] mIndex ;
	
	if (mRefinementSet != nil) {
		// Notify the refiment set that we have been deleted
		mRefinementSet->RemoveTable() ;
	}
}

void	TCrystalDiffractionDataTable::CreateImageFromDiffractionData() {
	// Put the data from the diffraction amplitudes and phases to create
	// an image with the specified number of unit cells. We are pretty free to 
	// specify the sampling (magnification) as we like as long as we include all 
	// spatial frequencies.
	
	long	i,hMax=0,kMax=0 ;
	double	degreeToRad = asin(1.0)/90. ;
	for(i=0; i < mNumReflections; i++) {
		hMax = max(hMax,mH[i]) ;
		kMax = max(kMax,mK[i]) ;
	}
	hMax = hMax/sin(degreeToRad*(180-mGamma)) + 1 ;
	kMax = kMax/sin(degreeToRad*(180-mGamma)) + 1 ;
	
	long n = 4*(log2((double)hMax) + 1) ; //LDM: Added double here
	n = max(pow(2.,(double)n),32) ;
	long m = 4*(log2((double)kMax) + 1) ; //LDM: Added double here
	m = max(pow(2.,(double)m),32) ;
	
	TArray *theDataArray = new TArray(n,m,kComplexArray) ;

	for(i=0; i < mNumReflections; i++) {
		*theDataArray->GetRealHK(mH[i],mK[i]) = mAmplitude[i]*cos(degreeToRad*mPhase[i]) ;
		*theDataArray->GetImagHK(mH[i],mK[i]) = mAmplitude[i]*sin(degreeToRad*mPhase[i]) ;
		*theDataArray->GetRealHK(-mH[i],-mK[i]) = *theDataArray->GetRealHK(mH[i],mK[i]) ;
		*theDataArray->GetImagHK(-mH[i],-mK[i]) = -*theDataArray->GetImagHK(mH[i],mK[i]) ;
	}
	
	theDataArray->ifft(false) ;
	TImage *theImage = new TImage(theDataArray,mGamma) ;
	theImage->SetSampling(mA/theImage->Width(),mB/theImage->Height()) ;
	theImage->Repeat(8,8) ;
	theImage->AdjustGamma() ;
	theImage->AdjustSampling() ;

	theImage->SetName(gPlaneGroup[mPlaneGroup-1]) ;
	theImage->Display() ;
}


void	TCrystalDiffractionDataTable::DoCommand(TCommand command,Boolean *hasBeenHandled) {
	switch(command) {
		case kHelp:
			OnCommandHelp() ;		
			*hasBeenHandled = true ;
			break ;
		case kIFFT:
			CreateImageFromDiffractionData() ;
			*hasBeenHandled = true ;
			break ;	
		case kRunFeasSet:	
			mRefinementSet = new TRefinementSet(this) ;
			mRefinementSet->RunFS98() ;
			*hasBeenHandled = true ;
			break ;
		case kCopy:
			DoCopy() ;
			*hasBeenHandled = true ;
			break ;	
		case kPaste:
			DoPaste() ;
			*hasBeenHandled = true ;
			break ;
		default:
			*hasBeenHandled = false ; 
			break ;
	}
}

void	TCrystalDiffractionDataTable::GetCommandStatus(TCommand command,Boolean *isEnabled,
			Boolean *isChecked) {
	switch(command) {
		case kIFFT:
			*isEnabled = true ;
			break ;	
		case kRunFeasSet:
			*isEnabled = mRefinementSet == nil ;
			break ;	
		case kCopy:
			*isEnabled = HasSelection() ;
			break ;	
		case kPaste:
			*isEnabled = HasSelection() && gTableClipBoardList != nil ;
			break ;	
		case kHelp:
			*isEnabled = true ;
			break ;
		default:
			*isEnabled = false ;
			break ;
	}
}

void	TCrystalDiffractionDataTable::DoCopy() {
	register	long i ;
	
	if (gTableClipBoardList != nil) delete gTableClipBoardList ;
	gTableClipBoardList = new TTableClipBoardList(mNumReflections) ;
	switch(mTableSelection.mTopLeftCell.col) {
		case 3:	// Amplitudes 
			gTableClipBoardList->mClipboardObjectType = 'ampl' ;
			for(i=0; i < mNumReflections; i++) {
				gTableClipBoardList->mIndex1[i] = mH[i] ;
				gTableClipBoardList->mIndex2[i] = mK[i] ;
				gTableClipBoardList->mValue[i] = mAmplitude[i] ;
			}
			break ;
		case 4: // Phases 
			gTableClipBoardList->mClipboardObjectType = 'phas' ;
			for(i=0; i < mNumReflections; i++) {
				gTableClipBoardList->mIndex1[i] = mH[i] ;
				gTableClipBoardList->mIndex2[i] = mK[i] ;
				gTableClipBoardList->mValue[i] = mPhase[i] ;
			}
			break ;
	}
	ZeroSelection() ;
}

void	TCrystalDiffractionDataTable::DoPaste() {
	// Fill in the phases from the table
	register long i,j ;
	short	h,k ;	
	float	scaleFactor,maxAmplitude = 0. ;	
	short	iH,iK ;
	long	II,JJ ;
	
	switch(mTableSelection.mTopLeftCell.col) {
		case 3:	// Amplitudes 

			
			for(j=0; j < mNumReflections; j++) {
				h = mH[j] ;
				k = mK[j] ;
				if (h != 0 && k != 0 ) {
					if (mAmplitude[j] > maxAmplitude) {
						for(i=0; i < gTableClipBoardList->mNumValues ; i++) {
							if (h == gTableClipBoardList->mIndex1[i] && k == gTableClipBoardList->mIndex2[i]) {
								maxAmplitude = mAmplitude[j] ;
								iH = h ;
								iK = k ;
								II = i ;
								JJ = j ;
								break ;
							}
						}
					}
				}
			}
				
			scaleFactor = mAmplitude[JJ]/gTableClipBoardList->mValue[II] ;
			for(i=0; i < mNumReflections; i++) {
				h = mH[i] ;
				k = mK[i] ;
				for(j=0; j < gTableClipBoardList->mNumValues ; j++) {
					if (gTableClipBoardList->mIndex1[j] == h && gTableClipBoardList->mIndex2[j] == k) {
						mAmplitude[i] = gTableClipBoardList->mValue[j]*scaleFactor ;
						break ;
					}
				}
			}
			break ;
		case 4: // Phases 
			for(i=0; i < mNumReflections; i++) {
				h = mH[i] ;
				k = mK[i] ;
				for(j=0; j < gTableClipBoardList->mNumValues ; j++) {
					if (gTableClipBoardList->mIndex1[j] == h && gTableClipBoardList->mIndex2[j] == k) {
						mPhase[i] = gTableClipBoardList->mValue[j] ;
						break ;
					}
				}
			}
			break ;
	}
		
	mTableSelection.ZeroSelection() ;
	Refresh() ;
}

void	TCrystalDiffractionDataTable::ClickColLabel(Int16 inCol) {
	TableCellT	firstSelectedCell,lastSelectedCell ;
	switch (inCol) {
		case 2:		// Sort on g-vector
			if (mSortKey != kSortOnG) {
				SortOnG() ;
				mSortKey = kSortOnG ;
				Refresh() ;
			}
			break ;
		case 3:		// Sort on amplitude
			if (!ControlKeyDown()) {
				if (mSortKey != kSortOnAmplitude) {
					SortOnAmpitude() ;
					mSortKey = kSortOnAmplitude ;
					Refresh() ;
				}
			}
			else {
				// Select the whole column fo copying/pasting
				if (!HasSelection()) {
					firstSelectedCell.col = inCol ;
					lastSelectedCell.col = inCol ;
					firstSelectedCell.row = 1 ;
					lastSelectedCell.row = mRows ;
					Select(firstSelectedCell,lastSelectedCell) ;
				}
				else {
					ZeroSelection() ;
				}
			}
			break ;
		case 4:
			// Select the whole column fo copying/pasting
			if (!HasSelection()) {
				firstSelectedCell.col = inCol ;
				lastSelectedCell.col = inCol ;
				firstSelectedCell.row = 1 ;
				lastSelectedCell.row = mRows ;
				Select(firstSelectedCell,lastSelectedCell) ;
			}
			else {
				ZeroSelection() ;
			}
			break ;
		default:
			break ;
	}
}


void	TCrystalDiffractionDataTable::Save() {
	// ------------------------------------------------------------------ 
	// Put up the standard file dialog Box to determine the file name 
	// ------------------------------------------------------------------ 
	FSInfo 	theFInfo ;
	
	strcpy(theFInfo.mFileName,"") ;
	ChooseSaveFile("Save Diffraction Table data as:",theFInfo) ;
	
	if (!theFInfo.mReply) 
		return ;

	char buffer[256] ;
	int i ;
	long inByteCount ;
	TableCellT theCell ;
	FILE	*fp ;

	if ((fp = fopen(theFInfo.mFileName,"w")) == nil) {
		sprintf(buffer,"Unable to open %s",theFInfo.mFileName) ;
		PutMessage(buffer) ;
		return ;
	}
	
	
	for(i=1;i<=mCols;i++) {
		switch (i) {
			case 1:
				sprintf(buffer,"h\tk\t") ;
				break ;
			case 2:
				sprintf(buffer,"u[A-1]\t") ;
				break ;
			case 3:
				sprintf(buffer,"Amp\t") ;
				break ;
			case 4:
				sprintf(buffer,"Phase\n") ;
				break ;
		}
		inByteCount = strlen(buffer) ;
		fwrite(buffer,inByteCount,1,fp) ;
	}
	
	for(i=1;i<=mRows;i++) {
		theCell.row = i ;
		for(theCell.col=1;theCell.col<=mCols;theCell.col++) {
			switch (theCell.col) {
				case 1:
					sprintf(buffer,"%d\t%d\t",mH[theCell.row-1],mK[theCell.row-1]) ;
					break ;
				case 2:
					sprintf(buffer,"%.3f\t",mGVector[theCell.row-1]) ;
					break ;
				case 3:
					sprintf(buffer,"%.2f\t",mAmplitude[theCell.row-1]) ;
					break ;
				case 4:
					sprintf(buffer,"%.2f\n",mPhase[theCell.row-1]) ;
					break ;
			}
			inByteCount = strlen(buffer) ;
			fwrite(buffer,inByteCount,1,fp) ;
		}
	}
	fclose(fp) ;
}

void TCrystalDiffractionDataTable::DrawRowLabelText(Int16 rowNumber) {
	char str[64] ;
	sprintf(str,"%d",rowNumber) ;
	DrawString(str) ;
}

void TCrystalDiffractionDataTable::ClickCell(
	TableCellT &inCell) 
{
	if ((mClickTime-mLastClick) < kDoubleClickTime) {
		EditCell(inCell) ;
	}	
}

void TCrystalDiffractionDataTable::EditCell(
	TableCellT &inCell) 
{
	/*
	MoveTo(4,10) ;
	char str[16] ;
	sprintf(str,"%d,%d",inCell.col,inCell.row) ;
	DrawString(str) ;
	*/
}

void TCrystalDiffractionDataTable::DrawColLabel(
	Int16 colNumber) 
{
	Rect	cellFrame;
	Rect	theSortRect  = {0,0,6,6} ;
	
	FPoint	imageOrigin = mParentPane->GetPortOrigin() ;
	
	if (mHasRowLabels)
		SetRect(&cellFrame,colNumber*mColWidth,0,(colNumber+1)*mColWidth,mRowHeight) ;
	else
		SetRect(&cellFrame,(colNumber-1)*mColWidth,0,colNumber*mColWidth,mRowHeight) ;
	if (colNumber != 0) 
		::OffsetRect(&cellFrame,near(imageOrigin.x),0) ;
	InsetRect(&cellFrame,1,1) ;
	SetFgColor(gGrayPixel) ;
	PaintRect(&cellFrame) ;
	SetFgColor(gDarkerGray) ;
	MoveTo(cellFrame.left,cellFrame.bottom) ;
	LineTo(cellFrame.left,cellFrame.top) ;
	LineTo(cellFrame.right,cellFrame.top) ;
	SetFgColor(gBlackPixel) ;
	MoveTo(cellFrame.right,cellFrame.top) ;
	LineTo(cellFrame.right,cellFrame.bottom) ;
	LineTo(cellFrame.left,cellFrame.bottom) ;
	InsetRect(&cellFrame,-1,-1) ;
	TextSize(0);
	MoveTo(cellFrame.left + 4, cellFrame.bottom - 4);
	if (colNumber > 0) {
		DrawColLabelText(colNumber) ;
		switch (colNumber) {
			case 2:
				if (mSortKey == kSortOnG) {
					OffsetRect(&theSortRect,cellFrame.right - 10,cellFrame.bottom - 10) ;
					PaintRect(&theSortRect) ;					
				}
				break ;
			case 3:
				if (mSortKey == kSortOnAmplitude) {
					OffsetRect(&theSortRect,cellFrame.right - 10,cellFrame.bottom - 10) ;
					PaintRect(&theSortRect) ;					
				}
				break ;
			default :
				break ;
		}
	}
}

void TCrystalDiffractionDataTable::DrawColLabelText(Int16 colNumber) {
	switch (colNumber) {
		case 1:
			DrawString("[h,k]") ;
			break ;
		case 2:
			DrawString("d-star") ;
			break ;
		case 3:
			DrawString("Amplitude") ;
			break ;
		case 4:
			DrawString("Phase") ;
			break ;
	}
}

void TCrystalDiffractionDataTable::DrawCellValue(
	TableCellT &inCell,
	Rect &cellFrame)
{
	char str[64] ;
	
	Rect	theRect = cellFrame ;
	InsetRect(&theRect,1,1) ;
	if (mIndex[inCell.row-1]%2 == 0) {
		SetFgColor(gGrayPixel) ;
		PaintRect(&theRect) ;
		SetFgColor(gBlackPixel) ;
	}
	else {	// We do this so that we can properly show up selected cells
		SetFgColor(gWhitePixel) ;
		PaintRect(&theRect) ;
		SetFgColor(gBlackPixel) ;
	}
	
	TextSize(0);
	MoveTo(cellFrame.left + 4, cellFrame.bottom - 4);
	switch (inCell.col) {
		case 1:
			sprintf(str,"[%3d,%3d]",mH[inCell.row-1],mK[inCell.row-1]) ;
			break ;
		case 2:
			sprintf(str,"%.3f",mGVector[inCell.row-1]) ;
			break ;
		case 3:
			sprintf(str,"%.2f",mAmplitude[inCell.row-1]) ;
			break ;
		case 4:
			if (mPhase[inCell.row-1] == kUnAssignedPhase) 
				sprintf(str,"-") ;
			else
				sprintf(str,"%.2f",mPhase[inCell.row-1]) ;
			break ;
	}
	DrawString(str) ;
}

void	TCrystalDiffractionDataTable::SortOnAmpitude() {
	long gap ;
	register long i,j ;
	register short itemp ;
	register float temp ;
	for (gap=mNumReflections/2;gap>0;gap/=2)
		for(i=gap;i<mNumReflections;i++)
			for(j=i-gap;j>=0;j-=gap)
				{
					if (mAmplitude[j] >= mAmplitude[j+gap])
						break ;
						
					temp = mGVector[j] ;
					mGVector[j] = mGVector[j+gap] ;
					mGVector[j+gap] = temp ;
					
					temp = mAmplitude[j] ;
					mAmplitude[j] = mAmplitude[j+gap] ;
					mAmplitude[j+gap] = temp ;
					
					temp = mPhase[j] ;
					mPhase[j] = mPhase[j+gap] ;
					mPhase[j+gap] = temp ;
					
					itemp = mH[j] ;
					mH[j] = mH[j+gap] ;
					mH[j+gap] = itemp ;
					
					itemp = mK[j] ;
					mK[j] = mK[j+gap] ;
					mK[j+gap] = itemp ;

					itemp = mIndex[j] ;
					mIndex[j] = mIndex[j+gap] ;
					mIndex[j+gap] = itemp ;
			  	}
			  	
	GroupSymmetryRelatedReflections() ;
}

void	TCrystalDiffractionDataTable::SortOnG() {
	long gap ;
	register long i,j ;
	register short itemp ;
	register float temp ;
	for (gap=mNumReflections/2;gap>0;gap/=2)
		for(i=gap;i<mNumReflections;i++)
			for(j=i-gap;j>=0;j-=gap)
				{
					if (mGVector[j] <= mGVector[j+gap])
						break ;
						
					temp = mGVector[j] ;
					mGVector[j] = mGVector[j+gap] ;
					mGVector[j+gap] = temp ;
					
					temp = mAmplitude[j] ;
					mAmplitude[j] = mAmplitude[j+gap] ;
					mAmplitude[j+gap] = temp ;
					
					temp = mPhase[j] ;
					mPhase[j] = mPhase[j+gap] ;
					mPhase[j+gap] = temp ;
					
					itemp = mH[j] ;
					mH[j] = mH[j+gap] ;
					mH[j+gap] = itemp ;
					
					itemp = mK[j] ;
					mK[j] = mK[j+gap] ;
					mK[j+gap] = itemp ;

					itemp = mIndex[j] ;
					mIndex[j] = mIndex[j+gap] ;
					mIndex[j+gap] = itemp ;
			  	}

	GroupSymmetryRelatedReflections() ;
}

void	TCrystalDiffractionDataTable::GroupSymmetryRelatedReflections() {
	// Set the index mIndex so that symmetry related reflections appear together
	// The index is already set but needs to be reset now that they have been sorted on 
	// either g-vector or amplitude
	
	short	numInGroup,currentGroupIndex,groupIndex = 1 ;
	
	long gap ;
	register long i,j,k ;
	register short itemp ;
	register float temp ;
	float	delta = 0.0001 ;
	for (gap=mNumReflections/2;gap>0;gap/=2)
		for(i=gap;i<mNumReflections;i++)
			for(j=i-gap;j>=0;j-=gap)
				{
					if (mSortKey == kSortOnG) {
						if ((mGVector[j+gap] - mGVector[j]) > delta)
							break ;
						if (mAmplitude[j] >= mAmplitude[j+gap])
							break ;
					} 
					else if (mSortKey == kSortOnAmplitude) {
						if (mAmplitude[j] > mAmplitude[j+gap])
							break ;
						if ((mGVector[j+gap] - mGVector[j]) > delta)
							break ;
					} 
						
					if (mIndex[j] == mIndex[j+gap])
						break ;
						
					temp = mGVector[j] ;
					mGVector[j] = mGVector[j+gap] ;
					mGVector[j+gap] = temp ;
					
					temp = mAmplitude[j] ;
					mAmplitude[j] = mAmplitude[j+gap] ;
					mAmplitude[j+gap] = temp ;
					
					temp = mPhase[j] ;
					mPhase[j] = mPhase[j+gap] ;
					mPhase[j+gap] = temp ;
					
					itemp = mH[j] ;
					mH[j] = mH[j+gap] ;
					mH[j+gap] = itemp ;
					
					itemp = mK[j] ;
					mK[j] = mK[j+gap] ;
					mK[j+gap] = itemp ;

					itemp = mIndex[j] ;
					mIndex[j] = mIndex[j+gap] ;
					mIndex[j+gap] = itemp ;
			  	}
	
	// Assign new group indicies
	for(i = 0; i < mNumReflections;) {
		currentGroupIndex = mIndex[i] ;
		for(j=i+1; j < mNumReflections && mIndex[j] == currentGroupIndex;j++) ;
		numInGroup = j - i ;
		for(k=i;k < i+numInGroup; k++) {
			mIndex[k] = groupIndex ;
		}
		i += numInGroup ;
		groupIndex++ ;		
	}
}
