// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------


#include "TDMApplicationWindow.h"
#include "XDrivers.h"
#include "globals.h"
#include "ToolConstants.h"
#include "hand.bmp"
#include "ruler.bmp"
#include "pointer.bmp"
#include "zoom.bmp"
#include "zoom_mask.bmp"
#include "selection.bmp"
#include "CircleMask.bmp"
#include "SpotMask.bmp"
#include "LineMask.bmp"
#include "WedgeMask.bmp"
#include "LatticeMask.bmp"
#include "lattice.bmp"
#include "FLattice.bmp"
#include "Aperture.bmp"
#include "Motif.bmp"
#include "TLattice.h"
#include "TTable.h"
#include "TDiffractionDataTable.h"
#include "TCrystal.h"

#include "DMMenuCommands.h"
#include "TImage.h"
#include "TImageWindow.h"
#include "config.h"

float	gDiffractionAperture ;
//TAnnotationCircle	gDiffractionApertureCircle ;

extern	XFontStruct	*gMenuFont ;
extern	XFontStruct	*gDialogFont ;
extern	XFontStruct	*gWindowFont ;
extern	Boolean	gWMFocusMove ;

extern	short	gImageDepth ;
extern	void	PutMessage(char *) ;
TWindow	*GetTargetWindow() ;
void	SetTargetWindow(TWindow *window) ;
extern void  OpenImageFile(FSInfo &theFSInfo);
TPushButton	*GetCurrentTool() ; 

void	AddWindow(TWindow *window) ;
void	SetCurrentTool(TPushButton	*) ;
void	ResetCurrentTool() ;
void	TerminateApplication() ; 
class	TMainPane ;

void	DisplayImageOnLogScale(TObject *) ;
void	FindEDDLattice(TObject *) ;
void	DeleteEDDLattice(TObject *) ;
void	ShowHideEDDLattice(TObject *) ;
void	DeleteTEMLattice(TObject *) ;
void	ShowHideTEMLattice(TObject *) ;
void	FindTEMLattice(TObject *) ;
void	FindTEMMotif(TObject *) ;
void	TEMCalculateFFT(TObject *) ;
void	FindDiffractionPeaks(TObject *) ;
extern void	SetMotifPars(TObject *) ;
void	DeleteDiffractionPeaks(TObject *) ;
void	SymmetrizeReflections(TObject *) ;
void	ShowHideDiffractionPeaks(TObject *) ;
void	SetPeakMotifVisibility(TObject *) ;
void	TestDiffractionPeaksSymmetry(TObject *) ;

void	SetContrast(TObject *) ;
void	SetBrightness(TObject *) ;

float 	GetTEMUhValue() ;
float 	GetTEMUkValue() ;
float 	GetTEMVhValue() ;
float 	GetTEMVkValue() ;
float 	GetEDDG1hValue() ;
float 	GetEDDG1kValue() ;
float 	GetEDDG2hValue() ;
float 	GetEDDG2kValue() ;
float 	GetEDDG3hValue() ;
float 	GetEDDG3kValue() ;

void 	FindEquivalentUnitCellChoices(TLattice *theLattice) ;

void	DisplayEDDLatticeParameters(TLattice *lattice,TImage *theImage) ;
void 	DisplayTEMLatticeParameters(TLattice *lattice,TImage *theImage) ;
float	GetPeakThreshold() ;
void	SetPeakThreshold(float value) ;
Boolean GetPeakMotifVisibility() ;
void	ChangeWorkingDirectory() ;
void	ChangeWorkingDirectory(FSInfo &theFSInfo) ;

TMainPane	*gTheMainPane ;
Boolean	FindInitialReciprocalLattice(TVector &g1,TVector &g2,TVector &g3) ;
extern	void	WriteObjectInfo(char *text) ;
void	RemoveWindow(TWindow *window) ;
extern	short	gDebugMode ;
extern  void	UnsetQuadPars() ;
extern	void	UnsetCubePars() ;

#include "FS98ControlRecord.h"
extern	FS98ControlRecord gFS98ControlRecord  ;

#define	kSpaceBetweenGroups	12

#include "TPeakList.h"
PeakCriteria GetPeakCriteria() { 
	return (PeakCriteria)gTheMainPane->mPeakCriteriaMenuButton->GetSelection() ;
}
IntensityMeasurementMethod GetIntensityMeasurementMethod() { 
	return (IntensityMeasurementMethod)gTheMainPane->mPeakIntensityMethodMenuButton->GetSelection() ;
}

bool	UseFriedelSymmetry() {
	return gTheMainPane->mEDDUseFriedel->GetState() ;
}

bool	UseZero() {
	return gTheMainPane->mUseZero->GetState() ;
}

TMenu	*gWindowsMenu = nil ;
void	AddWindowToWindowsMenu(TWindow *theWindow) {
	if (gWindowsMenu == nil) return ;

	// Get the name of the window
	char	*cp,text[256] ;
	theWindow->GetWindowTitle(text) ;
	if ((cp = strrchr(text,'/')) == nil) cp = text ; else cp++ ;
	
	if (gDebugMode) fprintf(stderr,"\n*** Adding window %d to windows menu **** ",(long)theWindow) ;
	gWindowsMenu->AddMenuItem(cp,(TCommand)theWindow) ;
}
void	RemoveWindowFromWindowsMenu(TWindow *theWindow) {
	if (gWindowsMenu == nil) return ;
	if (gDebugMode) fprintf(stderr,"\n*** Removing window %s (%d) from windows menu ****",theWindow->GetWindowTitle(),(long)theWindow) ;
	gWindowsMenu->RemoveMenuItem((TCommand)theWindow) ;
}


extern	char	*gPlaneGroup[] ;
extern	char	*gAlternatePlaneGroup[] ;

// The two string arrays below are related to the typedefs in TPeakList.h and need to be synchronized
char	*gIntensityMethod[] = {"XCF","XCF+Mask","XCF+Est","Integ","IntegBZM","IntegBZ",nil} ;	
//char	*gPeakCriteriaMethod[] = {"Image Mean","Local Mean","Local Std","None",nil} ;
char	*gPeakCriteriaMethod[] = {"Image Mean","Local Mean","None",nil} ;

TMainPane::TMainPane(TPane *parent,short xloc,short yloc,short width,short height)  
			: TPane(parent, xloc, yloc, width, height)  {
	mObjectList = new TObjectList ;
	mEditTextList = new TEditTextList ;

	short	x1,x2,y1,y2,toolAreaWidth,toolAreaHeight,xSpacing,ySpacing,distanceFromLeft ;
	short	numToolsPerRow,numToolsInSection,sizeOfTool,widthOfButton, 
		heightOfButton,
		numRows,maxToolsPerRow,yBorder = 8 ;
	TStaticText	*label ;

	strcpy(mObjectInfoString,"") ;
	
	mFgColor = gBlackPixel ;
	XSetForeground(gDisplay,mGC,mFgColor) ;

	mBkColor = gGrayPixel ;
	XSetBackground(gDisplay,mGC,mBkColor) ;

	XSetWindowAttributes attr ;
	attr.backing_store = WhenMapped ;
	XChangeWindowAttributes(gDisplay,mXWindow,CWBackingStore,&attr) ;

	mFont = gWindowFont ;
	XSetFont(gDisplay,mGC,mFont->fid) ;

	ySpacing = 4 ;

	sizeOfTool = 28 ;
	toolAreaWidth = width - 8 ;	// 4 Pixels around

	// Check to see how many tools we can fit
	maxToolsPerRow = toolAreaWidth/sizeOfTool ;

	// ------------- First Tool Section -----------------

	numToolsInSection = 11 ;
	numRows = (numToolsInSection-1)/maxToolsPerRow + 1 ;
	numToolsPerRow = numToolsInSection > maxToolsPerRow ? maxToolsPerRow : numToolsInSection ;

	toolAreaWidth = width - 8 ;
	xSpacing = (toolAreaWidth-numToolsPerRow*sizeOfTool)/(numToolsPerRow+1) ;	
	toolAreaHeight = 2*yBorder + numRows*sizeOfTool + (numRows-1)*xSpacing ;
	
	TGroupBox *grpBox = new TGroupBox(4,10,width - 4,
		6+toolAreaHeight,"Tools  ") ;
	AddObject(grpBox) ;

	Rect	theRect = grpBox->GetRect() ;
	toolAreaWidth = theRect.right - theRect.left ;
	
	xSpacing = (toolAreaWidth-numToolsPerRow*sizeOfTool)/(numToolsPerRow+1) ;	
	distanceFromLeft = (toolAreaWidth-numToolsPerRow*sizeOfTool-(numToolsPerRow-1)*xSpacing)/2 ;

	TIcon	*icon ;
	x1 = theRect.left + distanceFromLeft ;
	y1 = theRect.top + yBorder ;
	y2 = y1+sizeOfTool ;

	icon = nil ;
	icon = new TIcon(pointer_width,pointer_height,pointer_bits,mXWindow) ;
	mPointerTool = new TPushButton(x1,y1,x1+sizeOfTool,y2,icon,kPointer) ;
	AddObject(mPointerTool) ;

	x1 += sizeOfTool+xSpacing ;
	if ((x1+sizeOfTool+xSpacing) > width) {
		x1 = theRect.left + distanceFromLeft ;
		y1 += sizeOfTool + xSpacing ;
		y2 = y1+sizeOfTool ;
	}
	icon = new TIcon(selection_width,selection_height,selection_bits,mXWindow) ;
	mSelectionTool = new TPushButton(x1,y1,x1+sizeOfTool,y2,icon,kSelection) ;
	AddObject(mSelectionTool) ;

	x1 += sizeOfTool+xSpacing ;
	if ((x1+sizeOfTool+xSpacing) > width) {
		x1 = theRect.left + distanceFromLeft ;
		y1 += sizeOfTool + xSpacing ;
		y2 = y1+sizeOfTool ;
	}
	icon = new TIcon(ruler_width,ruler_height,ruler_bits,mXWindow) ;
	mRulerTool = new TPushButton(x1,y1,x1+sizeOfTool,y2,icon,kRuler) ;
	AddObject(mRulerTool) ;

	x1 += sizeOfTool+xSpacing ;
	if ((x1+sizeOfTool+xSpacing) > width) {
		x1 = theRect.left + distanceFromLeft ;
		y1 += sizeOfTool + xSpacing ;
		y2 = y1+sizeOfTool ;
	}
	icon = new TIcon(zoom_width,zoom_height,zoom_bits,mXWindow) ;
	mZoomTool = new TPushButton(x1,y1,x1+sizeOfTool,y2,icon,kZoom) ;
	AddObject(mZoomTool) ;

	x1 += sizeOfTool+xSpacing ;
	if ((x1+sizeOfTool+xSpacing) > width) {
		x1 = theRect.left + distanceFromLeft ;
		y1 += sizeOfTool + xSpacing ;
		y2 = y1+sizeOfTool ;
	}
	icon = new TIcon(hand_width,hand_height,hand_bits,mXWindow) ;
	mHandTool = new TPushButton(x1,y1,x1+sizeOfTool,y2,icon,kHand) ;
	AddObject(mHandTool) ;

	x1 += sizeOfTool+xSpacing ;
	if ((x1+sizeOfTool+xSpacing) > width) {
		x1 = theRect.left + distanceFromLeft ;
		y1 += sizeOfTool + xSpacing ;
		y2 = y1+sizeOfTool ;
	}
	icon = new TIcon(Aperture_width,Aperture_height,Aperture_bits,mXWindow) ;
	mTEMSetApertureTool = new TPushButton(x1,y1,x1+sizeOfTool,y2,icon,kTEMAperture) ;
	AddObject(mTEMSetApertureTool) ;

	x1 += sizeOfTool+xSpacing ;
	if ((x1+sizeOfTool+xSpacing) > width) {
		x1 = theRect.left + distanceFromLeft ;
		y1 += sizeOfTool + xSpacing ;
		y2 = y1+sizeOfTool ;
	}
	icon = new TIcon(SpotMask_width,SpotMask_height,SpotMask_bits,mXWindow) ;
	mSpotMaskTool = new TPushButton(x1,y1,x1+sizeOfTool,y2,icon,kSpotMask) ;
	AddObject(mSpotMaskTool) ;

	x1 += sizeOfTool+xSpacing ;
	if ((x1+sizeOfTool+xSpacing) > width) {
		x1 = theRect.left + distanceFromLeft ;
		y1 += sizeOfTool + xSpacing ;
		y2 = y1+sizeOfTool ;
	}
	icon = new TIcon(CircleMask_width,CircleMask_height,CircleMask_bits,mXWindow) ;
	mCircleMaskTool = new TPushButton(x1,y1,x1+sizeOfTool,y2,icon,kCircleMask) ;
	AddObject(mCircleMaskTool) ;

	x1 += sizeOfTool+xSpacing ;
	if ((x1+sizeOfTool+xSpacing) > width) {
		x1 = theRect.left + distanceFromLeft ;
		y1 += sizeOfTool + xSpacing ;
		y2 = y1+sizeOfTool ;
	}
	icon = new TIcon(LatticeMask_width,LatticeMask_height,LatticeMask_bits,mXWindow) ;
	mLatticeMaskTool = new TPushButton(x1,y1,x1+sizeOfTool,y2,icon,kLatticeMask) ;
	AddObject(mLatticeMaskTool) ;

	x1 += sizeOfTool+xSpacing ;
	if ((x1+sizeOfTool+xSpacing) > width) {
		x1 = theRect.left + distanceFromLeft ;
		y1 += sizeOfTool + xSpacing ;
		y2 = y1+sizeOfTool ;
	}
	icon = new TIcon(WedgeMask_width,WedgeMask_height,WedgeMask_bits,mXWindow) ;
	mWedgeMaskTool = new TPushButton(x1,y1,x1+sizeOfTool,y2,icon,kWedgeMask) ;
	AddObject(mWedgeMaskTool) ;

	x1 += sizeOfTool+xSpacing ;
	if ((x1+sizeOfTool+distanceFromLeft) > width) {
		x1 = theRect.left + xSpacing ;
		y1 += sizeOfTool + xSpacing ;
		y2 = y1+sizeOfTool ;
	}
	icon = new TIcon(LineMask_width,LineMask_height,LineMask_bits,mXWindow) ;
	mLineMaskTool = new TPushButton(x1,y1,x1+sizeOfTool,y2,icon,kLineMask) ;
	AddObject(mLineMaskTool) ;

	
	// --------- Image Info and Object Info --------------
	
	grpBox = new TGroupBox(4,theRect.bottom+kSpaceBetweenGroups,width - 4,
		theRect.bottom+kSpaceBetweenGroups+40,"Info ") ;
	AddObject(grpBox) ;
	mObjectInfoRect = theRect = grpBox->GetRect() ;


	// --------- Image Contrast & Brightness -------------

	grpBox = new TGroupBox(4,theRect.bottom+kSpaceBetweenGroups,width - 4,
		theRect.bottom+kSpaceBetweenGroups+50,"Image Adjustments ") ;
	AddObject(grpBox) ;
	theRect = grpBox->GetRect() ;
	
	mDisplayOnLogScale = new TCheckBox(theRect.left+2,theRect.top+10,
		theRect.left+70,theRect.top+30,"Log Scale") ;
	mDisplayOnLogScale->SetCallBackRoutine(DisplayImageOnLogScale) ;
	mDisplayOnLogScale->Disable() ;
	AddObject(mDisplayOnLogScale) ; 

	mContrastBar = new TContrastControl(theRect.left+130,
		theRect.top+20,
		theRect.right-8,theRect.top+20) ;
	//mContrastBar->SetCallBackRoutine(SetContrast) ;
	mContrastBar->SetTrackingRoutine(SetContrast) ;
	AddObject(mContrastBar) ;

	mBrightnessBar = new TBrightnessControl(theRect.left+130,theRect.top+40,
		theRect.right-8,theRect.top+40) ;
	//mBrightnessBar->SetCallBackRoutine(SetBrightness) ;
	mBrightnessBar->SetTrackingRoutine(SetBrightness) ;
	AddObject(mBrightnessBar) ;



	// --------- Electron Diffraction Section -------------------------------

	TGroupBox *EDDGroupBox = new TGroupBox(4,theRect.bottom+kSpaceBetweenGroups,width - 4,
		theRect.bottom+kSpaceBetweenGroups+314,"Electron Diffraction Quantification ") ;
	AddObject(EDDGroupBox) ;

	Rect theEDDRect = EDDGroupBox->GetRect() ;

	TGroupBox *EDDMarkBox = new TGroupBox(theEDDRect.left+4,theEDDRect.top+15,138,
		theEDDRect.top + 110,"Mark Lattice") ;
	AddObject(EDDMarkBox) ;
	Rect theEDDMarkRect = EDDMarkBox->GetRect() ;


	TGroupBox *EDDDefineBox = new TGroupBox(theEDDMarkRect.right+4,theEDDMarkRect.top,
			theEDDRect.right-4,theEDDMarkRect.bottom,"Define") ;
	AddObject(EDDDefineBox) ;
	Rect theEDDDefineRect = EDDDefineBox->GetRect() ;

	xSpacing = yBorder ;
	x1 = theEDDMarkRect.left + 4 ;
	y1 = theEDDMarkRect.top + 4 ;
	y2 = y1 ;

	mEDDSetG1Tool = new TPushButton(x1,y2+16,x1+32,y2+36,nil,kEDDSetG1,"G1:") ;
	mEDDSetG1Tool->Disable() ;
	AddObject(mEDDSetG1Tool) ;

	mEDDSetG2Tool = new TPushButton(x1,y2+40,x1+32,y2+60,nil,kEDDSetG2,"G2:") ;
	mEDDSetG2Tool->Disable() ;
	AddObject(mEDDSetG2Tool) ;

	mEDDSetG3Tool = new TPushButton(x1,y2+64,x1+32,y2+84,nil,kEDDSetG3,"G3:") ;
	mEDDSetG3Tool->Disable() ;
	AddObject(mEDDSetG3Tool) ;

	
	label = new TStaticText(60,y2+2,90,y2+14,"h") ; 
	label->SetLabelPosition(kLabelOutsideTop) ;
	label->SetJustification(kCenterJustified) ;
	AddObject(label) ;
	label = new TStaticText(100,y2+2,130,y2+14,"k") ;
	label->SetLabelPosition(kLabelOutsideTop) ;
	label->SetJustification(kCenterJustified) ;
	AddObject(label) ;
        //LDM: Set these to some numbers that are extreme -- to make user think
        //LDM: They just happen to be ones for Ann's 3x3...
	mEDDG3k = new TEditText(100,y2+64,130,y2+84,"","0") ;
	mEDDG3k->Disable() ;
	AddTextEditObject(mEDDG3k) ;
	mEDDG3h = new TEditText(60,y2+64,90,y2+84,"","-10") ;
	mEDDG3h->Disable() ;
	AddTextEditObject(mEDDG3h) ;
	mEDDG2k = new TEditText(100,y2+40,130,y2+60,"","-10") ;
	mEDDG2k->Disable() ;
	AddTextEditObject(mEDDG2k) ;
	mEDDG2h = new TEditText(60,y2+40,90,y2+60,"","10") ;
	mEDDG2h->Disable() ;
	AddTextEditObject(mEDDG2h) ;
	mEDDG1k = new TEditText(100,y2+16,130,y2+36,"","10") ;
	mEDDG1k->Disable() ;
	AddTextEditObject(mEDDG1k) ;
	mEDDG1h = new TEditText(60,y2+16,90,y2+36,"","0") ;
	mEDDG1h->Disable() ;
	AddTextEditObject(mEDDG1h) ;


	theRect = theEDDRect ;
	InsetRect(&theRect,8,8) ;
	xSpacing = yBorder ;

	y1 = y2 + 5 ;

	y1 += 22 ;
	mEDDFindLattice = new TButton(theEDDDefineRect.left+4,y2+16,
			theEDDDefineRect.right-4,y2+36,"Find") ;
	mEDDFindLattice->SetCallBackRoutine(FindEDDLattice) ;
	mEDDFindLattice->Disable() ;
	AddObject(mEDDFindLattice) ;

	y1 += 22 ;
	mEDDShowHideLattice = new TCheckBox(theEDDDefineRect.left+4,y2+40,
			theEDDDefineRect.right-4,y2+60,"Show ",true) ;
	mEDDShowHideLattice->SetCallBackRoutine(ShowHideEDDLattice) ;
	mEDDShowHideLattice->Disable() ;
	AddObject(mEDDShowHideLattice) ;

	y1 += 22 ;
	mEDDDeleteLattice = new TButton(theEDDDefineRect.left+4,y2+64,
			theEDDDefineRect.right-4,y2+84,"Delete") ;
	mEDDDeleteLattice->SetCallBackRoutine(DeleteEDDLattice) ;
	mEDDDeleteLattice->Disable() ;
	AddObject(mEDDDeleteLattice) ;

	grpBox = new TGroupBox(theEDDMarkRect.left,theEDDMarkRect.bottom+10,theEDDDefineRect.right,theEDDMarkRect.bottom+36,"Real Space Lattice Parameters") ;
	AddObject(grpBox) ;
	Rect theEDDParamRect = grpBox->GetRect();
	
	// Add the fields to show the lattice parameters below in the Lattice
	// box below the Mark/Define boxes
	theRect = EDDMarkBox->GetRect() ;
	y1 = theEDDParamRect.top + 10 ;
	x1 = theEDDParamRect.left + 4 ;
	x2 = theEDDMarkRect.right ;
	mEDDU = new TStaticText(x1,y1,theEDDDefineRect.right-2,y1+14,"a=0.0px, b=0.0px, ga=0.00deg") ;
	AddObject(mEDDU) ;

	TGroupBox *EDDPeakBox = new TGroupBox(theEDDParamRect.left,theEDDParamRect.bottom+12,
		theEDDParamRect.right,theEDDRect.bottom-4,"Peaks/Intensities") ;
	AddObject(EDDPeakBox) ;
	theRect = EDDPeakBox->GetRect() ;

	x1 = theRect.left + 4 ;
	y1 = theRect.top + 10 ;
		
	widthOfButton = 80 ;
	heightOfButton = 20 ;
	
	mEDMTestSymmetry = new TButton(theRect.left+4,y1,
			theRect.left+4+widthOfButton/2,y1+18,"Test") ;
	mEDMTestSymmetry->SetCallBackRoutine(TestDiffractionPeaksSymmetry) ;
	mEDMTestSymmetry->Disable() ;
	AddObject(mEDMTestSymmetry) ;

	mEDDUseFriedel = new TCheckBox(theRect.left+4+widthOfButton/2+2,y1,
			(theRect.right+theRect.left)/2-12,y1+18,"Friedel",false) ;
	mEDDUseFriedel->Activate(false) ;
	AddObject(mEDDUseFriedel) ;
		
	
	mSpacegroupMenuButton = new TCenteredPopupMenu(this,theRect.right - widthOfButton-12 ,y1,theRect.right - 4,
			y1+18,gAlternatePlaneGroup,1,"Group ") ;
	mSpacegroupMenuButton->Activate(false) ;
	AddObject(mSpacegroupMenuButton) ;

	// This will need to be fixed
	/*
	TMenu	*OneFold = new TMenu("1-fold") ;
	OneFold->AddMenuItem(gAlternatePlaneGroup[0]) ;
	TMenu	*TwoFold = new TMenu("2-fold") ;
	TwoFold->AddMenuItem(gAlternatePlaneGroup[1]) ;
	TwoFold->AddMenuItem(gAlternatePlaneGroup[2]) ;
	TwoFold->AddMenuItem(gAlternatePlaneGroup[3]) ;
	TwoFold->AddMenuItem(gAlternatePlaneGroup[4]) ;
	TwoFold->AddMenuItem(gAlternatePlaneGroup[5]) ;
	TwoFold->AddMenuItem(gAlternatePlaneGroup[6]) ;
	TwoFold->AddMenuItem(gAlternatePlaneGroup[7]) ;
	TwoFold->AddMenuItem(gAlternatePlaneGroup[8]) ;
	TwoFold->AddMenuItem(gAlternatePlaneGroup[9]) ;
	TwoFold->AddMenuItem(gAlternatePlaneGroup[10]) ;
	TwoFold->AddMenuItem(gAlternatePlaneGroup[11]) ;
	TwoFold->AddMenuItem(gAlternatePlaneGroup[12]) ;
	TMenu	*FourFold = new TMenu("4-fold") ;
	FourFold->AddMenuItem(gAlternatePlaneGroup[13]) ;
	FourFold->AddMenuItem(gAlternatePlaneGroup[14]) ;
	FourFold->AddMenuItem(gAlternatePlaneGroup[15]) ;
	FourFold->AddMenuItem(gAlternatePlaneGroup[16]) ;
	TMenu	*ThreeFold = new TMenu("3-fold") ;
	ThreeFold->AddMenuItem(gAlternatePlaneGroup[17]) ;
	ThreeFold->AddMenuItem(gAlternatePlaneGroup[18]) ;
	ThreeFold->AddMenuItem(gAlternatePlaneGroup[19]) ;
	TMenu	*SixFold = new TMenu("6-fold") ;
	SixFold->AddMenuItem(gAlternatePlaneGroup[20]) ;
	SixFold->AddMenuItem(gAlternatePlaneGroup[21]) ;
	mSpacegroupMenuButton = new TPopupMenu(this,theRect.right - widthOfButton-12 ,y1,theRect.right - 4,
										   y1+18,"p1","Group ") ;
	mSpacegroupMenuButton->AddMenuItem(OneFold) ;
	mSpacegroupMenuButton->AddMenuItem(TwoFold) ;
	mSpacegroupMenuButton->AddMenuItem(FourFold) ;
	mSpacegroupMenuButton->AddMenuItem(ThreeFold) ;
	mSpacegroupMenuButton->AddMenuItem(SixFold) ;
	mSpacegroupMenuButton->AddMenuItem(theMenu) ;
	mSpacegroupMenuButton->Activate(false) ;
	AddObject(mSpacegroupMenuButton) ;
	*/
	

	y1 += 22 ;

	mPeakIntensityMethodMenuButton = new TPopupMenu(this,(theRect.right + theRect.left)/2+20 -widthOfButton,y1,(theRect.right + theRect.left)/2+ 20,
			y1+18,gIntensityMethod,2,"Int Meas ") ;
	mPeakIntensityMethodMenuButton->Activate(false) ;
	AddObject(mPeakIntensityMethodMenuButton) ;

	mThresholdPeaks = new TEditText(theRect.right - widthOfButton/2,y1,
			theRect.right - 4,y1+20,"Threshold","0.25") ;
	mThresholdPeaks->Disable() ;
	AddTextEditObject(mThresholdPeaks) ;
	y1 += 20 ;
        //LDM: default true, really should be this way for general use
	mShowPeakMotif = new TCheckBox((theRect.right+theRect.left)/2+24,y1,theRect.right-4,y1+18,"Show Ave Peak",true) ;
	mShowPeakMotif->SetCallBackRoutine(SetPeakMotifVisibility) ;
	mShowPeakMotif->Disable() ;
	AddObject(mShowPeakMotif) ;

	mPeakCriteriaMenuButton = new TPopupMenu(this,(theRect.right + theRect.left)/2+20 -widthOfButton ,y1,(theRect.right + theRect.left)/2+ 20,
			y1+18,gPeakCriteriaMethod,2,"Criteria ") ;
	mPeakCriteriaMenuButton->Activate(false) ;
	AddObject(mPeakCriteriaMenuButton) ;

	y1 += 18 ;

	mEDDShowSymmetrizedPattern = new TCheckBox(x1,y1,(theRect.right+theRect.left)/2,y1+18,"Diff & Symm",false) ;
	mEDDShowSymmetrizedPattern->Disable() ;
	AddObject(mEDDShowSymmetrizedPattern) ;

	y1 += 18 ;
	mShowHidePeaks = new TCheckBox(x1,y1,(theRect.right+theRect.left)/2-38,y1+16,"Show Peaks",true) ;
	mShowHidePeaks->SetCallBackRoutine(ShowHideDiffractionPeaks) ;
	mShowHidePeaks->Disable() ;
	AddObject(mShowHidePeaks) ;

      int ldms = 16 ;
	mUseZero = new TCheckBox((theRect.right+theRect.left)/2-40,y1,
		(theRect.right+theRect.left)/2+15,y1+16,"Use All",false) ;
	mUseZero->Disable() ;
	AddObject(mUseZero) ;

	mAddSpot = new TPushButton((theRect.right+theRect.left)/2+xSpacing+ldms,y1+16-3*sizeOfTool/4,
		(theRect.right+theRect.left)/2+xSpacing+3*sizeOfTool/4+ldms,y1+16,nil,kEDDAddSpot,"+") ;
	mAddSpot->Disable() ;
	AddObject(mAddSpot) ;
	
	mRemoveSpot = new TPushButton((theRect.right+theRect.left)/2+xSpacing+3*sizeOfTool/4+ldms,y1+16-3*sizeOfTool/4,
		(theRect.right+theRect.left)/2+6*sizeOfTool/4+xSpacing+ldms,y1+16,nil,kEDDRemoveSpot,"-") ;
	mRemoveSpot->Disable() ;
	AddObject(mRemoveSpot) ;

	mTestSpot = new TPushButton((theRect.right+theRect.left)/2+xSpacing+6*sizeOfTool/4+ldms,y1+16-3*sizeOfTool/4,
		(theRect.right+theRect.left)/2+9*sizeOfTool/4+xSpacing+ldms,y1+16,nil,kEDDTestSpot,"T") ;
	mTestSpot->Disable() ;
	AddObject(mTestSpot) ;
	
	mMotifPars = new TButton((theRect.right+theRect.left)/2+xSpacing+9*sizeOfTool/4+ldms,y1+16-3*sizeOfTool/4,
		(theRect.right+theRect.left)/2+12*sizeOfTool/4+xSpacing+ldms,y1+16,"E") ;
	mMotifPars->SetCallBackRoutine(SetMotifPars) ;
	mMotifPars->Enable() ;
	AddObject(mMotifPars) ;

	y1 += 20 ;
	mEDDFindPeaks = new TButton(x1,y1,(theRect.right+theRect.left)/2-2,y1+heightOfButton,"Find Peaks") ;
	mEDDFindPeaks->SetCallBackRoutine(FindDiffractionPeaks) ;
	mEDDFindPeaks->Disable() ;
	AddObject(mEDDFindPeaks) ;

	mEDDSymmetrizeReflections = new TButton((theRect.right+theRect.left)/2+2,y1,theRect.right - 4,y1+heightOfButton,"Quantify / Symmetrize") ;
	mEDDSymmetrizeReflections->SetCallBackRoutine(SymmetrizeReflections) ;
	mEDDSymmetrizeReflections->Disable() ;
	AddObject(mEDDSymmetrizeReflections) ;

	y1 += heightOfButton + 2 ;
				
	mEDDDeletePeaks = new TButton(x1,y1,theRect.right - 4,y1+heightOfButton,"Delete Found Diffraction Peaks") ;
	mEDDDeletePeaks->SetCallBackRoutine(DeleteDiffractionPeaks) ;
	mEDDDeletePeaks->Disable() ;
	AddObject(mEDDDeletePeaks) ;

	theRect = theEDDRect ;

	// --------- Crystal Symmetry Section -----------------

	grpBox = new TGroupBox(4,theRect.bottom+kSpaceBetweenGroups,width - 4,
		height-4,"Crystal Symmetry Image Processing ") ;
	AddObject(grpBox) ;

	Rect	theSymmetrySectionRect = grpBox->GetRect() ;
	
	xSpacing = yBorder ;
	x1 = theSymmetrySectionRect.left + xSpacing ;
	y1 = theSymmetrySectionRect.top + 12 ;
	
	mTEMCalculateFFT = new TButton(theSymmetrySectionRect.left + 4,y1,theSymmetrySectionRect.right - 4,y1+24,"Calculate Hanning Windowed FFT") ;
	mTEMCalculateFFT->SetCallBackRoutine(TEMCalculateFFT) ;
	mTEMCalculateFFT->Disable() ;
	AddObject(mTEMCalculateFFT) ;
	
	y2 = y1 + 32 ;
	
	grpBox = new TGroupBox(theSymmetrySectionRect.left+4,y2,134,y2+68,"Mark Lattice") ;
	AddObject(grpBox) ;

	Rect	theMarkRect = grpBox->GetRect() ;
	y2 = theMarkRect.top + 4 ;	

	mTEMSetUTool = new TPushButton(x1,y2+16,x1+32,y2+36,nil,kTEMSetG1,"G1:") ;
	mTEMSetUTool->Disable() ;
	AddObject(mTEMSetUTool) ;

	mTEMSetVTool = new TPushButton(x1,y2+40,x1+32,y2+60,nil,kTEMSetG2,"G2:") ;
	mTEMSetVTool->Disable() ;
	AddObject(mTEMSetVTool) ;

	label = new TStaticText(60,y2+2,90,y2+14,"h") ;
	label->SetLabelPosition(kLabelOutsideTop) ;
	label->SetJustification(kCenterJustified) ;
	AddObject(label) ;
	label = new TStaticText(100,y2+2,130,y2+14,"k") ;
	label->SetLabelPosition(kLabelOutsideTop) ;
	label->SetJustification(kCenterJustified) ;
	AddObject(label) ;

	mTEMVk = new TEditText(100,y2+40,130,y2+60,"","10") ;
	mTEMVk->Disable() ;
	AddTextEditObject(mTEMVk) ;
	mTEMVh = new TEditText(60,y2+40,90,y2+60,"","0") ;
	mTEMVh->Disable() ;
	AddTextEditObject(mTEMVh) ;
	mTEMUk = new TEditText(100,y2+16,130,y2+36,"","0") ;
	mTEMUk->Disable() ;
	AddTextEditObject(mTEMUk) ;
	mTEMUh = new TEditText(60,y2+16,90,y2+36,"","10") ;
	mTEMUh->Disable() ;
	AddTextEditObject(mTEMUh) ;


	grpBox = new TGroupBox(138,theMarkRect.top,theSymmetrySectionRect.right-4,theMarkRect.bottom,"Define") ;
	AddObject(grpBox) ;

	Rect	theDefineRect = grpBox->GetRect() ;
	y2 = theDefineRect.top + 6 ;	
	mTEMRefineLattice = new TButton(theDefineRect.left+4,y2+2,theDefineRect.right-4,y2+22,"Find Lattice") ;
	mTEMRefineLattice->SetCallBackRoutine(FindTEMLattice) ;
	mTEMRefineLattice->Disable() ;
	AddObject(mTEMRefineLattice) ;

	mTEMShowHideLattice = new TCheckBox(theDefineRect.left+4,y2+22,theDefineRect.right-4,y2+40,"Show",false) ;
	mTEMShowHideLattice->SetCallBackRoutine(ShowHideTEMLattice) ;
	mTEMShowHideLattice->Disable() ;
	AddObject(mTEMShowHideLattice) ;

	mTEMDeleteLattice = new TButton(theDefineRect.left+4,y2+40,theDefineRect.right-4,y2+60,"Delete") ;
	mTEMDeleteLattice->SetCallBackRoutine(DeleteTEMLattice) ;
	mTEMDeleteLattice->Disable() ;
	AddObject(mTEMDeleteLattice) ;


	grpBox = new TGroupBox(theMarkRect.left,theMarkRect.bottom+10,theDefineRect.right,theMarkRect.bottom+36,"Unit Cell Lattice Parameters") ;
	AddObject(grpBox) ;
	Rect theParameterRect = grpBox->GetRect();
	
	y2 = theParameterRect.top+4 ;
	x1 = theParameterRect.left + 4 ;
	
	mU = new TStaticText(theParameterRect.left+4,y2+4,theParameterRect.right-2,y2+18,
		"a=0.00px, b=0.00px, ga=0.00deg") ;
	AddObject(mU) ;


	mTEMFindMotif = new TButton(theSymmetrySectionRect.left+4,theParameterRect.bottom + 8,
		theSymmetrySectionRect.right-4,theParameterRect.bottom + 32,"Find Symmetrized Unit Cell") ;
	mTEMFindMotif->SetCallBackRoutine(FindTEMMotif) ;
	mTEMFindMotif->Disable() ;
	AddObject(mTEMFindMotif) ;


	SetCurrentTool(mPointerTool) ;
}
 

TMainPane::TMainPane(TPane *parent,Rect *rect) : TPane(parent,rect)  {
	short	width = rect->right - rect->left ;
	mObjectList = new TObjectList ;
	mEditTextList = new TEditTextList ;
	SetFgColor(gBlackPixel) ;
	SetBkColor(gGrayPixel) ; 

	TButton	*theButton = new TButton(width-80,20,width-20,40,"Hi") ;
	AddObject(theButton) ;

	theButton = new TButton(width-80,50,width-20,70,"There") ;
	AddObject(theButton) ;


} 	

		
void	TMainPane::WriteObjectInfo(char *infoText) {
	
	if (strcmp(mObjectInfoString,infoText) == 0) return ;

	GrafPtr	tPort ;
	GetPort(&tPort) ;
	SetPort(this) ;


	strcpy(mObjectInfoString,infoText) ;
	WriteObjectInfoString() ;
	
	
	SetPort(tPort) ;
}

void TMainPane::Draw() {
	if (!mIsVisible) return ;
	TPane::Draw() ;
	WriteObjectInfoString() ;
}

void	WriteObjectInfo(char *text) {
	gTheMainPane->WriteObjectInfo(text) ; 
}

void	TMainPane::WriteObjectInfoString() {
	short	x,y ;
	char	*cpStart,*cpEnd ;
	Rect	theRect = mObjectInfoRect ;
	char	text[256] ;
	InsetRect(&theRect,2,2) ;
	theRect.top += 2 ;
	EraseRect(&theRect) ;
	
	x = mObjectInfoRect.left + 4 ;
	y = mObjectInfoRect.top + 20 ;
	
	strcpy(text,mObjectInfoString) ;
	cpStart = text ;
	while ((cpEnd = strpbrk(cpStart,"\n")) != nil) {
		*cpEnd = '\0' ;
		DrawText(x,y,cpStart) ;
		cpStart = cpEnd + 1 ;
		y += 14 ;
	}
	DrawText(x,y,cpStart) ;

}

void	ResetCurrentTool() {
	SetCurrentTool(gTheMainPane->mPointerTool) ;
}

void DisplayTEMLatticeParameters(TLattice *lattice,TImage *theImage) {
	float	ux,uy,vx,vy,u,v,gamma,gammastar,a,b,sigastar ;
	char	text[256] ;
	FPoint	U,V ;

	float	calibrationConstant = theImage->GetCalibration() ;
	CalibrationUnit calibrationUnit = theImage->GetCalibrationUnit() ;

	if (lattice == nil) {
		ux = 0. ;
		uy = 0. ;
		vx = 0. ;
		vy = 0. ;
		u  = 0. ;
		v  = 0. ;

		gammastar = 0. ;
		gamma = 0. ;

		a = 0. ;
		b = 0. ;
	}
	else {
		U = lattice->GetU() ;
		V = lattice->GetV() ;
		ux = calibrationConstant*U.x ;
		uy = calibrationConstant*U.y ;
		vx = calibrationConstant*V.x ;
		vy = calibrationConstant*V.y ;
		u  = sqrt(ux*ux+uy*uy) ;
		v  = sqrt(vx*vx+vy*vy) ;

		gammastar = acos((ux*vx+uy*vy)/(u*v)) ;
		sigastar = sin(gammastar) ;
		gammastar = 90.*gammastar/asin(1.0) ;
		gamma = 180. - gammastar ;

		a = 1/(u*sigastar) ;
		b = 1/(v*sigastar) ;
	}
	
	switch(calibrationUnit) {
		case kInversePixel:
			sprintf(text,"a=%.2fpx, b=%.2fpx, ga=%.2fdeg",a,b,gamma) ;
			break ;
		case kInverseAngstrom:
			sprintf(text,"a=%.2fA, b=%.2fA, ga=%.2fdeg",a,b,gamma) ;
			break ;
		case kInverseNanometer:
			sprintf(text,"a=%.2fnm, b=%.2fnm, ga=%.2fdeg",a,b,gamma) ;
			break ;
		default:
			//sprintf(text,"a=0.0px, b=0.0px, ga=0.00deg") ;
			sprintf(text,"a=%.2fpx, b=%.2fpx, ga=%.2fdeg",a,b,gamma) ;
			break ;
	}
	
	gTheMainPane->mU->SetItemText(text) ;
}

void DisplayEDDLatticeParameters(TLattice *lattice,TImage *theImage) {
	float	ux,uy,vx,vy,u,v,gamma,gammastar,a,b,sigastar ;
	char	text[256] ;
	FPoint	U,V ;

	float	calibrationConstant = theImage->GetCalibration() ;
	CalibrationUnit calibrationUnit = theImage->GetCalibrationUnit() ;

	if (lattice == nil) {
		ux = 0. ;
		uy = 0. ;
		vx = 0. ;
		vy = 0. ;
		u  = 0. ;
		v  = 0. ;

		gammastar = 0. ;
		gamma = 0. ;

		a = 0. ;
		b = 0. ;
	}
	else {
		U = lattice->GetU() ;
		V = lattice->GetV() ;
		ux = calibrationConstant*U.x ;
		uy = calibrationConstant*U.y ;
		vx = calibrationConstant*V.x ;
		vy = calibrationConstant*V.y ;
		u  = sqrt(ux*ux+uy*uy) ;
		v  = sqrt(vx*vx+vy*vy) ;

		gammastar = acos((ux*vx+uy*vy)/(u*v)) ;
		sigastar = sin(gammastar) ;
		gammastar = 90.*gammastar/asin(1.0) ;
		gamma = 180. - gammastar ;

		a = 1./(u*sigastar) ;
		b = 1./(v*sigastar) ;
	}

	switch(calibrationUnit) {
		case kPixel:
		case kInversePixel:
			sprintf(text,"a=%.2fpx, b=%.2fpx, ga=%.2fdeg",a,b,gamma) ;
			break ;
		case kAngstrom:
		case kInverseAngstrom:
			sprintf(text,"a=%.2fA, b=%.2fA, ga=%.2fdeg",a,b,gamma) ;
			break ;
		case kNanometer:
		case kInverseNanometer:
			sprintf(text,"a=%.2fnm, b=%.2fnm, ga=%.2fdeg",a,b,gamma) ;
			break ;
		default:
			sprintf(text,"a=0.0px, b=0.0px, ga=0.00deg") ;
			break ;
	}

	gTheMainPane->mEDDU->SetItemText(text) ;
}

void	FindEDDLattice(TObject *object) {
	char	text[256] ;
	TEDDRecord	*theEDDRecord ;

	TImageWindow *theImageWindow = (TImageWindow *)GetTargetWindow() ;
	if (theImageWindow == nil) {
		PutMessage("There is no target image to operate on") ;
		return ;
	}

	gDiffractionAperture = theImageWindow->GetDiffractionAperture() ;
	/*
	TEMRecord	*theRec = theImageWindow->GetTEMRecord() ;
	if (theRec != nil) {
		if (theRec->radCircle != nil) 
			gDiffractionApertureCircle = *theRec->radCircle ;
		else {
			gDiffractionApertureCircle.SetCenter(theImageWindow->GetImage()->Width()/2,theImageWindow->GetImage()->Height()/2) ;
			gDiffractionApertureCircle.SetRadius(gDiffractionAperture ;
		}
	}
	*/

	theEDDRecord = theImageWindow->GetEDDRecord() ;
	if (theEDDRecord == nil) {
		PutMessage("You haven't marked 3 independent diffraction spots yet") ;
		return ;
	}
	if (!theEDDRecord->g1Defined || !theEDDRecord->g2Defined 
			|| !theEDDRecord->g3Defined) {
		PutMessage("You haven't marked 3 independent diffraction spots yet") ;
		return ;
	}
	(gTheMainPane->mEDDG1h)->GetEditText(text) ;
	theEDDRecord->g1.a[2] = atoi(text) ;
	(gTheMainPane->mEDDG1k)->GetEditText(text) ;
	theEDDRecord->g1.a[3] = atoi(text) ;
	(gTheMainPane->mEDDG2h)->GetEditText(text) ;
	theEDDRecord->g2.a[2] = atoi(text) ;
	(gTheMainPane->mEDDG2k)->GetEditText(text) ;
	theEDDRecord->g2.a[3] = atoi(text) ;
	(gTheMainPane->mEDDG3h)->GetEditText(text) ;
	theEDDRecord->g3.a[2] = atoi(text) ;
	(gTheMainPane->mEDDG3k)->GetEditText(text) ;
	theEDDRecord->g3.a[3] = atoi(text) ;
	
	TVector	g1 = theEDDRecord->g1 ;
	TVector	g2 = theEDDRecord->g2 ;
	TVector	g3 = theEDDRecord->g3 ;

	if (!FindInitialReciprocalLattice(g1,g2,g3)) {
		PutMessage("Failed to find a lattice based on the marked spots") ;
		return ;
	}
	// Tell lattice drawing routines that quadratic fit is unset
	UnsetQuadPars() ;
	UnsetCubePars() ;
	TVector	O,U,V ;

	// The initial points have been destroyed and replaced with the solution
	O = g1 ;	
	U = g2 ;	
	V = g3 ;	

	// Create the lattice on the image
	TLattice *theLattice = new TLattice(O,U,V) ;
	
	float angle = theLattice->GetLatticeAngle() ;
	if (angle > 94.) {
		PutMessage("Please use a reciprocal lattice with an angle <= 90 degrees") ;
		delete theLattice ;
		return ;
	}

	FindEquivalentUnitCellChoices(theLattice) ;
	
	theImageWindow->AddEDDLattice(theLattice) ;

	// Let us immediately try to find peaks and to refine it.
	theImageWindow->RefineEDDLattice() ;

	theImageWindow->ShowEDDLattice() ;
	theImageWindow->SetControls() ;

	gTheMainPane->mEDDShowHideLattice->SetState(true) ;
	SetCurrentTool(gTheMainPane->mPointerTool) ;
	
}

void	DeleteEDDLattice(TObject *object) {
	//TButton *button = (TButton *)object ;

	TImageWindow *theImageWindow = (TImageWindow *)GetTargetWindow() ;
	if (theImageWindow == nil) {
		PutMessage("There is no target image to operate on") ;
		return ;
	}

	theImageWindow->RemoveEDDLattice() ;
	theImageWindow->SetControls() ;
}

void	ShowHideEDDLattice(TObject *object) {
	TButton *button = (TButton *)object ;

	Boolean state = button->GetState() ;
	// We will use this to remove the lattice
	TImageWindow *theImageWindow = (TImageWindow *)GetTargetWindow() ;
	if (theImageWindow == nil) {
		PutMessage("There is no target image to operate on") ;
		return ;
	}
	if (state)
		theImageWindow->ShowEDDLattice() ;
	else
		theImageWindow->HideEDDLattice() ;

}

void	DisplayImageOnLogScale(TObject *object) {
	TButton *button = (TButton *)object ;
	TImageWindow *theImageWindow = (TImageWindow *)GetTargetWindow() ;
	if (theImageWindow == nil) {
		PutMessage("There is no target image to operate on") ;
		return ;
	}
	theImageWindow->DisplayOnLogScale(button->GetControlValue()) ;
	gTheMainPane->mContrastBar->SetControlValue(atan(theImageWindow->GetContrastValue())*90./asin(1.0)) ;
	gTheMainPane->mBrightnessBar->SetControlValue(theImageWindow->GetBrightnessValue()) ;
	theImageWindow->Update(nil) ;
}


void	FindDiffractionPeaks(TObject *object) {
	//TButton *button = (TButton *)object ;
	TImageWindow *theImageWindow = (TImageWindow *)GetTargetWindow() ;
	//fprintf(stderr,"Went to 1\n");
	if (theImageWindow == nil) {
		PutMessage("There is no target image to operate on") ;
		return ; 
	}
	theImageWindow->FindDiffractionPeaks() ;
}
		
void	DeleteDiffractionPeaks(TObject *object) {
	//TButton *button = (TButton *)object ;
	TImageWindow *theImageWindow = (TImageWindow *)GetTargetWindow() ;
	if (theImageWindow == nil) {
		PutMessage("There is no target image to operate on") ;
		return ;
	}
	theImageWindow->DeleteDiffractionPeaks() ;
}
	
void	SymmetrizeReflections(TObject *object) {
	//TButton *button = (TButton *)object ;
	TImageWindow *theImageWindow = (TImageWindow *)GetTargetWindow() ;
	if (theImageWindow == nil) {
		PutMessage("There is no target image to operate on") ;
		return ;
	}
	// button->Disable() ;
	theImageWindow->SetControls() ;
	theImageWindow->SymmetrizeReflections(gTheMainPane->mSpacegroupMenuButton->GetSelection(),
		gTheMainPane->mEDDShowSymmetrizedPattern->GetState()) ;
}
	
void	TestDiffractionPeaksSymmetry(TObject *object) {
	//TButton *button = (TButton *)object ;
	TImageWindow *theImageWindow = (TImageWindow *)GetTargetWindow() ;
	if (theImageWindow == nil) {
		PutMessage("There is no image to operate on") ;
		return ;
	}
	theImageWindow->CalculateAmplitudeResidual() ;
}

void	ShowHideDiffractionPeaks(TObject *object) {
	TButton *button = (TButton *)object ;
	Boolean state = button->GetState() ;
	TImageWindow *theImageWindow = (TImageWindow *)GetTargetWindow() ;
	if (theImageWindow == nil) {
		PutMessage("There is no image to operate on") ;
		return ;
	}
	if (state)
		theImageWindow->ShowPeakAnnotations() ;
	else
		theImageWindow->HidePeakAnnotations() ;
}

void	SetPeakMotifVisibility(TObject *object) {
	TButton *button = (TButton *)object ;
	Boolean state = button->GetState() ;
	TImageWindow *theImageWindow = (TImageWindow *)GetTargetWindow() ;
	if (theImageWindow == nil) {
		PutMessage("There is no image to operate on") ;
		return ;
	}
	theImageWindow->SetPeakMotifVisibility(state) ;
}	

Boolean GetPeakMotifVisibility() {
	return gTheMainPane->mShowPeakMotif->GetState() ;
}	

void	FindTEMLattice(TObject *object) {
	//TButton *button = (TButton *)object ;
	char	text[256] ;
	TTEMRecord	*theTEMRecord ;
	float	h1,k1,h2,k2 ;

	TImageWindow *theImageWindow = (TImageWindow *)GetTargetWindow() ;
	if (theImageWindow == nil) {
		PutMessage("There is no target image to operate on") ;
		return ;
	}

	theTEMRecord = theImageWindow->GetTEMRecord() ;
	if (theTEMRecord == nil) {
		PutMessage("You haven't marked 2 independent diffraction spots yet") ;
		return ;
	}
	if (!theTEMRecord->g1Defined || !theTEMRecord->g2Defined) {
		PutMessage("You haven't marked 3 independent diffraction spots yet") ;
		return ;
	}
	(gTheMainPane->mTEMUh)->GetEditText(text) ;
	h1 = atoi(text) ;
	(gTheMainPane->mTEMUk)->GetEditText(text) ;
	k1 = atoi(text) ;
	(gTheMainPane->mTEMVh)->GetEditText(text) ;
	h2 = atoi(text) ;
	(gTheMainPane->mTEMVk)->GetEditText(text) ;
	k2 = atoi(text) ;

	// The initial lattice is just defined as follows
	// g1 - O = (h1*ux + k1*vx)X + (h1*uy + k1*vy)Y
	// g2 - O = (h2*ux + k2*vx)X + (h2*uy + k2*vy)Y
	TVector	O,U,V ;
	short	width,height ;
	width = theImageWindow->GetImage()->Width() ;
	height = theImageWindow->GetImage()->Height() ;
	
	O.a[0] = width/2 ;
	O.a[1] = height/2 ;
	TVector	u1 = theTEMRecord->g1 - O ; 	// u1.a[0] = (h1*ux + k1*vx)
										// u1.a[1] = (h1*uy + k1*vy)
	TVector	u2 = theTEMRecord->g2 - O ;	// u2.a[0] = (h2*ux + k2*vx)
										// u2.a[1] = (h2*uy + k2*vy)
	float	det = h1*k2 - k1*h2 ;
	if (fabs(det) < 1.0e-10) {
		PutMessage("Cannot find a lattice based on the marked spots") ;

		return ;
	}
	
	
	U.a[0] = (k2*u1.a[0] - k1*u2.a[0])/det ;
	V.a[0] = (h1*u2.a[0] - h2*u1.a[0])/det ;
	
	U.a[1] = (k2*u1.a[1] - k1*u2.a[1])/det ;
	V.a[1] = (h1*u2.a[1] - h2*u1.a[1])/det ;

	gDiffractionAperture = theImageWindow->GetDiffractionAperture() ;
	/*
	TEMRecord	*theRec = theImageWindow->GetTEMRecord() ;
	if (theRec != nil) {
		if (theRec->radCircle != nil) 
			gDiffractionApertureCircle = *theRec->radCircle ;
		else {
			gDiffractionApertureCircle.SetCenter(theImageWindow->GetImage()->Width()/2,theImageWindow->GetImage()->Height()/2) ;
			gDiffractionApertureCircle.SetRadius(gDiffractionAperture ;
		}
	}
	*/


	// Create the lattice on the image
	TLattice *theLattice = new TLattice(O,U,V) ;

	float angle = theLattice->GetLatticeAngle() ;
	if (angle > 94.) {
		PutMessage("Please use a reciprocal lattice with an angle <= 90 degrees") ;
		delete theLattice ;
		return ;
	}

	FindEquivalentUnitCellChoices(theLattice) ;

	theImageWindow->AddTEMLattice(theLattice) ;

	// Let us immediately try to find peaks and to refine it.
	theImageWindow->RefineTEMLattice() ;

	theImageWindow->ShowTEMLattice() ;
	
	theImageWindow->SetControls() ;

	gTheMainPane->mTEMShowHideLattice->SetState(true) ;
	SetCurrentTool(gTheMainPane->mPointerTool) ;
}

void	DeleteTEMLattice(TObject *object) {

	TImageWindow *theImageWindow = (TImageWindow *)GetTargetWindow() ;
	if (theImageWindow == nil) {
		PutMessage("There is no target image to operate on") ;
		return ;
	}

	theImageWindow->RemoveTEMLattice() ;
	theImageWindow->SetControls() ;
}

void	ShowHideTEMLattice(TObject *object) {
	TButton *button = (TButton *)object ;
	Boolean state = button->GetState() ;
	// We will use this to remove the lattice
	TImageWindow *theImageWindow = (TImageWindow *)GetTargetWindow() ;
	if (theImageWindow == nil) {
		PutMessage("There is no target image to operate on") ;
		return ;
	}
	if (state)
		theImageWindow->ShowTEMLattice() ;
	else
		theImageWindow->HideTEMLattice() ;

}

void	FindTEMMotif(TObject *object) {
	// We will use this to remove the lattice
	TImageWindow *theImageWindow = (TImageWindow *)GetTargetWindow() ;
	if (theImageWindow == nil) {
		PutMessage("There is no target image to operate on") ;
		return ;
	}
	theImageWindow->FindMotif() ;
}

void	TEMCalculateFFT(TObject *object) {
	TButton *button = (TButton *)object ;
	SetCurrentTool(gTheMainPane->mPointerTool) ;
	TImageWindow *theImageWindow = (TImageWindow *) GetTargetWindow() ;
	if (theImageWindow == nil) {
		PutMessage("There is no target image to operate on") ;
		return ;
	}
	theImageWindow->CalculateMaskedFFT() ;
	button->Disable() ;
}

float GetTEMUhValue() {
	char	text[256] ;
	gTheMainPane->mTEMUh->GetEditText(text) ;
	return atof(text) ;
}

float GetTEMUkValue() {
	char	text[256] ;
	gTheMainPane->mTEMUk->GetEditText(text) ;
	return atof(text) ;
}

float GetTEMVhValue() {
	char	text[256] ;
	gTheMainPane->mTEMVh->GetEditText(text) ;
	return atof(text) ;
}

float GetTEMVkValue() {
	char	text[256] ;
	gTheMainPane->mTEMVk->GetEditText(text) ;
	return atof(text) ;
}

float GetEDDG1hValue() {
	char	text[256] ;
	gTheMainPane->mEDDG1h->GetEditText(text) ;
	return atof(text) ;
}

float GetEDDG1kValue() {
	char	text[256] ;
	gTheMainPane->mEDDG1k->GetEditText(text) ;
	return atof(text) ;
}

float GetEDDG2hValue() {
	char	text[256] ;
	gTheMainPane->mEDDG2h->GetEditText(text) ;
	return atof(text) ;
}

float GetEDDG2kValue() {
	char	text[256] ;
	gTheMainPane->mEDDG2k->GetEditText(text) ;
	return atof(text) ;
}

float GetEDDG3hValue() {
	char	text[256] ;
	gTheMainPane->mEDDG3h->GetEditText(text) ;
	return atof(text) ;
}

float GetEDDG3kValue() {
	char	text[256] ;
	gTheMainPane->mEDDG3k->GetEditText(text) ;
	return atof(text) ;
}

float	GetPeakThreshold() {
	char	text[256] ;
	gTheMainPane->mThresholdPeaks->GetEditText(text) ;
	return atof(text) ;
}

void	SetPeakThreshold(float value) {
	char	text[256] ;
	sprintf(text,"%.2f",value) ;
	gTheMainPane->mThresholdPeaks->SetEditText(text) ;
}

void	SetContrast(TObject *object) {
	TSlider *slider = (TSlider *)object ;
	TImageWindow *theImageWindow = (TImageWindow *)GetTargetWindow() ;
	if (theImageWindow != nil && theImageWindow->GetWindowKind() == 'iwnd') {
		float	angle = slider->GetControlValue() ;
		float	contrast = tan(angle*asin(1.0)/90.) ;
		theImageWindow->SetContrast(contrast) ;
	}
}

void	SetBrightness(TObject *object) {
	TSlider *slider = (TSlider *)object ;
	TImageWindow *theImageWindow = (TImageWindow *)GetTargetWindow() ;
	if (theImageWindow != nil) {
		float	brightnessValue = slider->GetControlValue() ;
		theImageWindow->SetBrighness(brightnessValue) ;
	}
}

TDMApplicationWindow::TDMApplicationWindow(short xloc,short yloc,short width,short height,
			short depth,char *title) 
	: TApplicationWindow( xloc,yloc,width,height,depth,title) {
	
	SetupMenuBar() ;
	gTheMainPane = new TMainPane(this,0,25,width,height-25) ;
	AddPane(gTheMainPane) ;
	SetPort(this) ;
	Show() ;

}

TDMApplicationWindow::~TDMApplicationWindow()  {
}

void	TDMApplicationWindow::ActivateListener(TWindow *win) {

	if (!mIsVisible) return ;

	if (win->GetWindowKind() != 'iwnd') {

		gTheMainPane->mContrastBar->Disable() ;
		gTheMainPane->mBrightnessBar->Disable() ;
		gTheMainPane->mDisplayOnLogScale->Disable() ;

		gTheMainPane->mTEMCalculateFFT->Disable() ;
		gTheMainPane->mTEMSetUTool->Disable() ;
		gTheMainPane->mTEMSetVTool->Disable() ;
		gTheMainPane->mTEMUh->Disable() ;
		gTheMainPane->mTEMUk->Disable() ;
		gTheMainPane->mTEMVh->Disable() ;
		gTheMainPane->mTEMVk->Disable() ;
		gTheMainPane->mTEMFindMotif->Disable() ;
		gTheMainPane->mTEMDeleteLattice->Disable() ;
		gTheMainPane->mTEMShowHideLattice->Disable() ;
		gTheMainPane->mTEMRefineLattice->Disable() ;
		
		gTheMainPane->mAddSpot->Disable() ;
		gTheMainPane->mRemoveSpot->Disable() ;
		gTheMainPane->mTestSpot->Disable() ;
		gTheMainPane->mEDMTestSymmetry->Disable() ;		
		if (gTheMainPane->mEDDFindLattice->IsEnabled())
			gTheMainPane->mEDDFindLattice->Disable() ;
		if (gTheMainPane->mEDDFindPeaks->IsEnabled())
			gTheMainPane->mEDDFindPeaks->Disable() ;
		if (gTheMainPane->mEDDDeleteLattice->IsEnabled())
			gTheMainPane->mEDDDeleteLattice->Disable() ;
		if (gTheMainPane->mEDDShowHideLattice->IsEnabled())
			gTheMainPane->mEDDShowHideLattice->Disable() ;
		if (gTheMainPane->mEDDDeletePeaks->IsEnabled())
			gTheMainPane->mEDDDeletePeaks->Disable() ;
		if (gTheMainPane->mShowHidePeaks->IsEnabled())
			gTheMainPane->mShowHidePeaks->Disable() ;
		if (gTheMainPane->mEDDSymmetrizeReflections->IsEnabled())
			gTheMainPane->mEDDSymmetrizeReflections->Disable() ;
		if (gTheMainPane->mUseZero->IsEnabled())
			gTheMainPane->mUseZero->Disable() ;
		gTheMainPane->mSpacegroupMenuButton->Activate(false) ;
		gTheMainPane->mPeakCriteriaMenuButton->Activate(false) ;
		gTheMainPane->mPeakIntensityMethodMenuButton->Activate(false) ;
		gTheMainPane->mThresholdPeaks->Disable() ;
		gTheMainPane->mShowPeakMotif->Disable() ;
		gTheMainPane->mEDDShowSymmetrizedPattern->Disable() ;
			
		gTheMainPane->mEDDG1h->Disable() ;
		gTheMainPane->mEDDG1k->Disable() ;
		gTheMainPane->mEDDG2h->Disable() ;
		gTheMainPane->mEDDG2k->Disable() ;
		gTheMainPane->mEDDG3h->Disable() ;
		gTheMainPane->mEDDG3k->Disable() ;
		gTheMainPane->mEDDSetG1Tool->Disable() ;
		gTheMainPane->mEDDSetG2Tool->Disable() ;
		gTheMainPane->mEDDSetG3Tool->Disable() ;
		gTheMainPane->mEDDUseFriedel->Disable() ;
		//Overkill cases when we might want the expert mode enabled
		if (	gTheMainPane->mEDDFindPeaks->IsEnabled()		||
				gTheMainPane->mEDDDeleteLattice->IsEnabled()	||
				gTheMainPane->mEDDShowHideLattice->IsEnabled()	||
				gTheMainPane->mEDDDeletePeaks->IsEnabled()		||
				gTheMainPane->mShowHidePeaks->IsEnabled()		||
				gTheMainPane->mEDDSymmetrizeReflections->IsEnabled() ) {
				gTheMainPane->mMotifPars->Enable() ;
				gTheMainPane->mPeakIntensityMethodMenuButton->Activate(true) ;
				gTheMainPane->mPeakCriteriaMenuButton->Activate(true) ;
				gTheMainPane->mUseZero->Enable() ;
			}
		else {
				gTheMainPane->mMotifPars->Disable() ;
				gTheMainPane->mPeakIntensityMethodMenuButton->Activate(false) ;
				gTheMainPane->mPeakCriteriaMenuButton->Activate(false) ;
			}
		return ;
	}

	if (!gTheMainPane->mContrastBar->IsEnabled())
		gTheMainPane->mContrastBar->Enable() ;
	if (!gTheMainPane->mBrightnessBar->IsEnabled())
		gTheMainPane->mBrightnessBar->Enable() ;
	
	TImageWindow *theImageWindow = (TImageWindow *)win ;
	// Set up controls as required
	if (theImageWindow == nil) return ;
	Boolean	commandEnabled,isChecked ;
	theImageWindow->GetCommandStatus(kFFT,&commandEnabled,&isChecked) ;
	if (commandEnabled) {
		gTheMainPane->mTEMCalculateFFT->Enable() ;
	}
	else {
		gTheMainPane->mTEMCalculateFFT->Disable() ;
	}
	
	if (theImageWindow->GetWindowContent() == 'hfft') {
		gTheMainPane->mPeakCriteriaMenuButton->Activate(false) ;
		gTheMainPane->mPeakIntensityMethodMenuButton->Activate(false) ;
		gTheMainPane->mAddSpot->Disable() ;
		gTheMainPane->mRemoveSpot->Disable() ;
		gTheMainPane->mTestSpot->Disable() ;
		
		gTheMainPane->mTEMSetUTool->Enable() ;
		gTheMainPane->mTEMSetVTool->Enable() ;
		gTheMainPane->mTEMUh->Enable() ;
		gTheMainPane->mTEMUk->Enable() ;
		gTheMainPane->mTEMVh->Enable() ;
		gTheMainPane->mTEMVk->Enable() ;
	}
	else {
		
		gTheMainPane->mTEMSetUTool->Disable() ;
		gTheMainPane->mTEMSetVTool->Disable() ;
		gTheMainPane->mTEMUh->Disable() ;
		gTheMainPane->mTEMUk->Disable() ;
		gTheMainPane->mTEMVh->Disable() ;
		gTheMainPane->mTEMVk->Disable() ;
	}
	
	gTheMainPane->mContrastBar->SetControlValue(atan(theImageWindow->GetContrastValue())*90./asin(1.0)) ;
	gTheMainPane->mBrightnessBar->SetControlValue(theImageWindow->GetBrightnessValue()) ;
	gTheMainPane->mDisplayOnLogScale->SetControlValue(theImageWindow->IsDisplayOnLogScale()) ;
	if (theImageWindow->ImageIsRealSpace()) 
		gTheMainPane->mDisplayOnLogScale->Enable() ;
	else 
		gTheMainPane->mDisplayOnLogScale->Disable() ;


	TEDDRecord *theEDDRecord = theImageWindow->GetEDDRecord() ;
	if (theEDDRecord != nil) {
		if (theEDDRecord->g1Defined && theEDDRecord->g2Defined 
				&& theEDDRecord->g3Defined) {
			if (!gTheMainPane->mEDDFindLattice->IsEnabled())
				gTheMainPane->mEDDFindLattice->Enable() ;
		}
		else {
			if (gTheMainPane->mEDDFindLattice->IsEnabled())
				gTheMainPane->mEDDFindLattice->Disable() ;
		}
	}
	else {
		if (gTheMainPane->mEDDFindLattice->IsEnabled())
			gTheMainPane->mEDDFindLattice->Disable() ;
	}
	TTEMRecord *theTEMRecord = theImageWindow->GetTEMRecord() ;
	if (theTEMRecord != nil) {
		if (theTEMRecord->g1Defined && theTEMRecord->g2Defined) {
			if (!gTheMainPane->mTEMRefineLattice->IsEnabled())
				gTheMainPane->mTEMRefineLattice->Enable() ;
		}
		else {
			if (gTheMainPane->mTEMRefineLattice->IsEnabled())
				gTheMainPane->mTEMRefineLattice->Disable() ;
		}
	}
	else {
		if (gTheMainPane->mTEMRefineLattice->IsEnabled())
			gTheMainPane->mTEMRefineLattice->Disable() ;
	}
	TLattice *theLattice = theImageWindow->GetEDDLattice() ;
	TPeakList	*peakList = theImageWindow->GetPeakList() ;
	if (theLattice != nil) {
		if (peakList == nil) {
			gTheMainPane->mPeakCriteriaMenuButton->Activate(true) ;
			gTheMainPane->mPeakIntensityMethodMenuButton->Activate(true) ;
			gTheMainPane->mShowPeakMotif->Enable() ;
			gTheMainPane->mAddSpot->Disable() ;
			gTheMainPane->mRemoveSpot->Disable() ;
			gTheMainPane->mTestSpot->Disable() ;
		}
		else {
			gTheMainPane->mPeakCriteriaMenuButton->Activate(false) ;
			gTheMainPane->mPeakIntensityMethodMenuButton->Activate(false) ;
			gTheMainPane->mShowPeakMotif->Disable() ;
			gTheMainPane->mAddSpot->Enable() ;
			gTheMainPane->mRemoveSpot->Enable() ;
			gTheMainPane->mTestSpot->Enable() ;
		}
		if (!gTheMainPane->mEDDDeleteLattice->IsEnabled()) { 
			gTheMainPane->mEDDDeleteLattice->Enable() ;
		}
		if (!gTheMainPane->mEDDFindPeaks->IsEnabled()) 
			gTheMainPane->mEDDFindPeaks->Enable() ;
		if (!gTheMainPane->mEDDShowHideLattice->IsEnabled())
			gTheMainPane->mEDDShowHideLattice->Enable() ;
			
		// Let us disable any items that are not possible planegroups
		float angle = theLattice->GetLatticeAngle() ;
		float u = theLattice->GetLatticeParameterU() ;
		float v = theLattice->GetLatticeParameterV() ;
		Boolean angleIs90 = fabs(90.0-angle) < 4. ? true : false ;
		Boolean	angleIs60 = fabs(60.0-angle) < 4. ? true : false ;
		Boolean AequalB = fabs((u-v)/((u+v)/2.)) < 0.03 ? true : false ;
		if (angleIs90) {
			EnableMenuItem(gTheMainPane->mSpacegroupMenuButton,3) ;
			EnableMenuItem(gTheMainPane->mSpacegroupMenuButton,4) ;
			EnableMenuItem(gTheMainPane->mSpacegroupMenuButton,5) ;
			EnableMenuItem(gTheMainPane->mSpacegroupMenuButton,6) ;
			EnableMenuItem(gTheMainPane->mSpacegroupMenuButton,7) ;
			EnableMenuItem(gTheMainPane->mSpacegroupMenuButton,8) ;
			EnableMenuItem(gTheMainPane->mSpacegroupMenuButton,9) ;
			EnableMenuItem(gTheMainPane->mSpacegroupMenuButton,10) ;
			EnableMenuItem(gTheMainPane->mSpacegroupMenuButton,11) ;
			EnableMenuItem(gTheMainPane->mSpacegroupMenuButton,12) ;
			EnableMenuItem(gTheMainPane->mSpacegroupMenuButton,13) ;
			if (AequalB) {
				EnableMenuItem(gTheMainPane->mSpacegroupMenuButton,14) ;
				EnableMenuItem(gTheMainPane->mSpacegroupMenuButton,15) ;
				EnableMenuItem(gTheMainPane->mSpacegroupMenuButton,16) ;
				EnableMenuItem(gTheMainPane->mSpacegroupMenuButton,17) ;
			}
			else {
				DisableMenuItem(gTheMainPane->mSpacegroupMenuButton,14) ;
				DisableMenuItem(gTheMainPane->mSpacegroupMenuButton,15) ;
				DisableMenuItem(gTheMainPane->mSpacegroupMenuButton,16) ;
				DisableMenuItem(gTheMainPane->mSpacegroupMenuButton,17) ;
			}
		}
		else {
			DisableMenuItem(gTheMainPane->mSpacegroupMenuButton,3) ;
			DisableMenuItem(gTheMainPane->mSpacegroupMenuButton,4) ;
			DisableMenuItem(gTheMainPane->mSpacegroupMenuButton,5) ;
			DisableMenuItem(gTheMainPane->mSpacegroupMenuButton,6) ;
			DisableMenuItem(gTheMainPane->mSpacegroupMenuButton,7) ;
			DisableMenuItem(gTheMainPane->mSpacegroupMenuButton,8) ;
			DisableMenuItem(gTheMainPane->mSpacegroupMenuButton,9) ;
			DisableMenuItem(gTheMainPane->mSpacegroupMenuButton,10) ;
			DisableMenuItem(gTheMainPane->mSpacegroupMenuButton,11) ;
			DisableMenuItem(gTheMainPane->mSpacegroupMenuButton,12) ;
			DisableMenuItem(gTheMainPane->mSpacegroupMenuButton,13) ;
			DisableMenuItem(gTheMainPane->mSpacegroupMenuButton,14) ;
			DisableMenuItem(gTheMainPane->mSpacegroupMenuButton,15) ;
			DisableMenuItem(gTheMainPane->mSpacegroupMenuButton,16) ;
			DisableMenuItem(gTheMainPane->mSpacegroupMenuButton,17) ;
		}
		if (angleIs60) {
			EnableMenuItem(gTheMainPane->mSpacegroupMenuButton,18) ;
			EnableMenuItem(gTheMainPane->mSpacegroupMenuButton,19) ;
			EnableMenuItem(gTheMainPane->mSpacegroupMenuButton,20) ;
			EnableMenuItem(gTheMainPane->mSpacegroupMenuButton,21) ;
			EnableMenuItem(gTheMainPane->mSpacegroupMenuButton,22) ;
		}
		else {
			DisableMenuItem(gTheMainPane->mSpacegroupMenuButton,18) ;
			DisableMenuItem(gTheMainPane->mSpacegroupMenuButton,19) ;
			DisableMenuItem(gTheMainPane->mSpacegroupMenuButton,20) ;
			DisableMenuItem(gTheMainPane->mSpacegroupMenuButton,21) ;
			DisableMenuItem(gTheMainPane->mSpacegroupMenuButton,22) ;
		}
		gTheMainPane->mSpacegroupMenuButton->Activate(true) ;
		gTheMainPane->mThresholdPeaks->Enable() ;
		gTheMainPane->mShowPeakMotif->Enable() ;
		gTheMainPane->mEDDShowSymmetrizedPattern->Enable() ;
	}
	else {	
		gTheMainPane->mEDDShowHideLattice->Disable() ;
		if (gTheMainPane->mEDDFindPeaks->IsEnabled())
			gTheMainPane->mEDDFindPeaks->Disable() ;
		if (gTheMainPane->mEDDDeleteLattice->IsEnabled())
			gTheMainPane->mEDDDeleteLattice->Disable() ;
		if (gTheMainPane->mEDDDeletePeaks->IsEnabled())
			gTheMainPane->mEDDDeletePeaks->Disable() ;
		if (gTheMainPane->mShowHidePeaks->IsEnabled())
			gTheMainPane->mShowHidePeaks->Disable() ;
		if (gTheMainPane->mUseZero->IsEnabled())
			gTheMainPane->mUseZero->Disable() ;
		//if (gTheMainPane->mEDDSymmetrizeReflections->IsEnabled())
		//	gTheMainPane->mEDDSymmetrizeReflections->Disable() ;

		gTheMainPane->mEDMTestSymmetry->Disable() ;
		gTheMainPane->mSpacegroupMenuButton->Activate(false) ;
		gTheMainPane->mThresholdPeaks->Disable() ;
		gTheMainPane->mShowPeakMotif->Disable() ;
		gTheMainPane->mEDDShowSymmetrizedPattern->Disable() ;
	}
	DisplayEDDLatticeParameters(theLattice,theImageWindow->GetImage()) ;

	theLattice = theImageWindow->GetTEMLattice() ;
	if (theLattice != nil) {
		if (!gTheMainPane->mTEMDeleteLattice->IsEnabled())
			gTheMainPane->mTEMDeleteLattice->Enable() ;
		if (!gTheMainPane->mTEMFindMotif->IsEnabled())
			gTheMainPane->mTEMFindMotif->Enable() ;
		gTheMainPane->mTEMShowHideLattice->Enable() ;
	}
	else {	
		if (gTheMainPane->mTEMFindMotif->IsEnabled())
			gTheMainPane->mTEMFindMotif->Disable() ;
		if (gTheMainPane->mTEMDeleteLattice->IsEnabled())
			gTheMainPane->mTEMDeleteLattice->Disable() ;
		gTheMainPane->mTEMShowHideLattice->Disable() ;
	}
	DisplayTEMLatticeParameters(theLattice,theImageWindow->GetImage()) ;

	if (theImageWindow->ImageIsRealSpace()) {

		gTheMainPane->mEDDG1h->Enable() ;
		gTheMainPane->mEDDG1k->Enable() ;
		gTheMainPane->mEDDG2h->Enable() ;
		gTheMainPane->mEDDG2k->Enable() ;
		gTheMainPane->mEDDG3h->Enable() ;
		gTheMainPane->mEDDG3k->Enable() ;
		gTheMainPane->mEDDSetG1Tool->Enable() ;
		gTheMainPane->mEDDSetG2Tool->Enable() ;
		gTheMainPane->mEDDSetG3Tool->Enable() ;


		gTheMainPane->mEDDShowHideLattice->SetControlValue(theImageWindow->IsEDDLatticeVisible()) ;
		gTheMainPane->mShowHidePeaks->SetControlValue(theImageWindow->IsPeakAnnotationsVisisble()) ;
		gTheMainPane->mTEMShowHideLattice->SetControlValue(theImageWindow->IsTEMLatticeVisible()) ;

		if (peakList != nil) {
			if (gTheMainPane->mEDDFindPeaks->IsEnabled())
				gTheMainPane->mEDDFindPeaks->Disable() ;
			if (!gTheMainPane->mEDDDeletePeaks->IsEnabled())
				gTheMainPane->mEDDDeletePeaks->Enable() ;
			if (!gTheMainPane->mShowHidePeaks->IsEnabled())
				gTheMainPane->mShowHidePeaks->Enable() ;
			if (!gTheMainPane->mEDDFindPeaks->IsEnabled() /*&& !theImageWindow->HasSymmetrizedPeaks()*/) {
				// We have found peaks
				gTheMainPane->mEDDUseFriedel->Enable() ;
				gTheMainPane->mEDMTestSymmetry->Enable() ;
				gTheMainPane->mUseZero->Enable() ;
			}
			else {
				gTheMainPane->mEDMTestSymmetry->Disable() ;
			}
			gTheMainPane->mUseZero->Enable() ;
			gTheMainPane->mAddSpot->Enable() ;
			gTheMainPane->mRemoveSpot->Enable() ;
			gTheMainPane->mTestSpot->Enable() ;
			if (!gTheMainPane->mEDDSymmetrizeReflections->IsEnabled())
				gTheMainPane->mEDDSymmetrizeReflections->Enable() ;
		}
		else {
			gTheMainPane->mEDDUseFriedel->Disable() ;
			gTheMainPane->mAddSpot->Disable() ;
			gTheMainPane->mRemoveSpot->Disable() ;
			gTheMainPane->mTestSpot->Disable() ;
			gTheMainPane->mEDDUseFriedel->Disable() ;
			gTheMainPane->mUseZero->Disable() ;
			gTheMainPane->mEDMTestSymmetry->Disable() ;
			if (gTheMainPane->mEDDDeletePeaks->IsEnabled())
				gTheMainPane->mEDDDeletePeaks->Disable() ;
			if (gTheMainPane->mShowHidePeaks->IsEnabled())
				gTheMainPane->mShowHidePeaks->Disable() ;
			if (gTheMainPane->mEDDSymmetrizeReflections->IsEnabled())
				gTheMainPane->mEDDSymmetrizeReflections->Disable() ;
		}
	}
	else {
		gTheMainPane->mEDDUseFriedel->Disable() ;
		gTheMainPane->mEDDDeletePeaks->Disable() ;
		gTheMainPane->mShowHidePeaks->Disable() ;
		gTheMainPane->mEDDSymmetrizeReflections->Disable() ;
		gTheMainPane->mEDDFindPeaks->Disable() ;
		gTheMainPane->mEDDG1h->Disable() ;
		gTheMainPane->mEDDG1k->Disable() ;
		gTheMainPane->mEDDG2h->Disable() ;
		gTheMainPane->mEDDG2k->Disable() ;
		gTheMainPane->mEDDG3h->Disable() ;
		gTheMainPane->mEDDG3k->Disable() ;
		gTheMainPane->mEDDSetG1Tool->Disable() ;
		gTheMainPane->mEDDSetG2Tool->Disable() ;
		gTheMainPane->mEDDSetG3Tool->Disable() ;
		gTheMainPane->mUseZero->Disable() ;

	}
		//Overkill cases when we might want the expert mode enabled
		if (	gTheMainPane->mEDDFindPeaks->IsEnabled()		||
				gTheMainPane->mEDDDeleteLattice->IsEnabled()	||
				gTheMainPane->mEDDShowHideLattice->IsEnabled()	||
				gTheMainPane->mEDDDeletePeaks->IsEnabled()		||
				gTheMainPane->mShowHidePeaks->IsEnabled()		||
				gTheMainPane->mEDDSymmetrizeReflections->IsEnabled() ) {
				gTheMainPane->mMotifPars->Enable() ;
				gTheMainPane->mPeakIntensityMethodMenuButton->Activate(true) ;
				gTheMainPane->mPeakCriteriaMenuButton->Activate(true) ;
				gTheMainPane->mUseZero->Enable() ;
			}
		else {
				gTheMainPane->mMotifPars->Disable() ;
				gTheMainPane->mPeakIntensityMethodMenuButton->Activate(false) ;
				gTheMainPane->mPeakCriteriaMenuButton->Activate(false) ;
			}
		//Enable find if delete is active
		if (gTheMainPane->mEDDDeleteLattice->IsEnabled()) { 
			gTheMainPane->mEDDFindLattice->Enable() ;
		}

}
void	TDMApplicationWindow::SetupMenuBar() {
	TMenu	*theMenu = new TMenu("File") ;
	theMenu->AddMenuItem("Open Image...",kOpen,'o') ;
	theMenu->AddMenuItem("Open HKL File...",kOpenHKLFile) ;
    //theMenu->AddMenuItem("Import HKL File...",kImportHKL) ;
	theMenu->AddMenuItem("Save",kSave,'s') ;
	theMenu->AddMenuItem("Save As...",kSaveAs) ;
	theMenu->AddMenuItem("Save As HKL file...",kSaveAsHKL) ;
    //theMenu->AddMenuItem("Export HKL File...",kExportHKL) ;
	theMenu->AddMenuItem("-") ;	// Divider
#ifdef HAVE_CYGWIN
	theMenu->AddMenuItem("Open File",kWinOpen) ;
	theMenu->AddMenuItem("Open Directory",kWinDir) ;
	theMenu->AddMenuItem("Open Wordpad",kWinWordpad);
	theMenu->AddMenuItem("Open Notepad",kWinNotepad);
        theMenu->AddMenuItem("Windows Print",kWinprint);
#endif
        theMenu->AddMenuItem("-") ; // Divider
        theMenu->AddMenuItem("Open xterm",kxterm) ;
#ifdef  HAVE_CYGWIN
        theMenu->AddMenuItem("Open bash shell",kWinbash) ;
        // theMenu->AddMenuItem("Print File",kWinprint) ;
#endif
#ifdef HAVE_XV
        theMenu->AddMenuItem("Open xv",kxv) ;
#endif
	theMenu->AddMenuItem("-") ;	// Divider
	theMenu->AddMenuItem("Open NCEMSS File...",kOpenNCEMSSFile) ;
	theMenu->AddMenuItem("Open CBED File...",kOpenCBEDFile) ;
	theMenu->AddMenuItem("-") ;	// Divider
	theMenu->AddMenuItem("Export...",kExportImage,'e') ;
	theMenu->AddMenuItem("-") ;	// Divider
	theMenu->AddMenuItem("Close",kClose,'w') ;
	theMenu->AddMenuItem("Print",kPrint,'p') ;
	theMenu->AddMenuItem("-") ;	// Divider
	theMenu->AddMenuItem("Set Working Directory...",kSetDirectory) ;
	theMenu->AddMenuItem("-") ;	// Divider
	theMenu->AddMenuItem("Help",kHelp);
	theMenu->AddMenuItem("Quit",kQuit,'q') ;
	mMenuBar->AddMenu(theMenu) ;

	theMenu = new TMenu("Edit") ;
	theMenu->AddMenuItem("Copy...",kCopy,'c') ;
	theMenu->AddMenuItem("Cut",kCut,'x') ;
	theMenu->AddMenuItem("Paste",kPaste,'v') ;
	theMenu->AddMenuItem("-") ;	// Divider
	theMenu->AddMenuItem("Select All",kSelectAll,'a') ;
	theMenu->AddMenuItem("-") ;	// Divider
	theMenu->AddMenuItem("Calibrate...",kCalibrate) ;
	theMenu->AddMenuItem("-") ;	// Divider
	theMenu->AddMenuItem("Click-To-Focus",kClickToFocus) ;
	theMenu->AddMenuItem("-") ;	// Divider
	theMenu->AddMenuItem("Help",kHelp);
	//theMenu->AddMenuItem("Edit Motif Size",kEditMotifSize) ;
	//theMenu->AddMenuItem("Edit Motif Mask Size",kEditMotifMask) ;
	//theMenu->AddMenuItem("Edit Saturation Level",kEditSaturation) ;
	//theMenu->AddMenuItem("Edit Image Noise Level",kEditTooLow) ;
	//theMenu->AddMenuItem("Edit Peak Displacement Limit",kEditOffTol) ;
	//theMenu->AddMenuItem("Edit Peak Outer Radius",kEditPeakRMax);
	//theMenu->AddMenuItem("Edit Peak Inner Radius",kEditPeakRMin);
	//theMenu->AddMenuItem("-") ;	// Divider
	mMenuBar->AddMenu(theMenu) ;

	theMenu = new TMenu("Process") ;
	theMenu->AddMenuItem("FFT",kFFT,'f') ;
	theMenu->AddMenuItem("Inverse FFT",kIFFT,'i') ;
	theMenu->AddMenuItem("Hanning Masked FFT",kHFFT,'h') ;
	theMenu->AddMenuItem("-") ;	// Divider
	theMenu->AddMenuItem("Edit Mask...",kEditMask) ;
	theMenu->AddMenuItem("Apply Mask...",kApplyMask,'m') ;
	theMenu->AddMenuItem("-") ;	// Divider
	theMenu->AddMenuItem("Soft WienerFilter",kWienerFilter) ;
	theMenu->AddMenuItem("Strong WienerFilter",kWienerFilter2) ;
	theMenu->AddMenuItem("Adjustable WienerFilter",kWienerFiltera) ;
	theMenu->AddMenuItem("-") ;	// Divider
	theMenu->AddMenuItem("Invert",kInvert) ;
	theMenu->AddMenuItem("-") ;	// Divider
	theMenu->AddMenuItem("Scale...",kScale) ;
	theMenu->AddMenuItem("Rotate...",kRotate,'r') ;
	theMenu->AddMenuItem("Flip Horizontal",kFlipHorizontal) ;
	theMenu->AddMenuItem("Flip Vertical",kFlipVertical) ;
	theMenu->AddMenuItem("-") ;	// Divider
	theMenu->AddMenuItem("Sharpen",kSharpen) ;
	theMenu->AddMenuItem("Blur",kBlur,'b') ;
	theMenu->AddMenuItem("-") ;	// Divider
	theMenu->AddMenuItem("Convert...",kConvert) ;
	theMenu->AddMenuItem("-") ;	// Divider
	theMenu->AddMenuItem("Set Reciprocal Lattice",kDefineRecLatt) ;
	theMenu->AddMenuItem("Edit Lattice",kEditRecLatt) ;
	theMenu->AddMenuItem("Refine Lattice",kRefineRecLatt) ;
	theMenu->AddMenuItem("Paste Amp/Phases",kPasteAmpPhs) ;
	theMenu->AddMenuItem("-") ;
	theMenu->AddMenuItem("Help",kHelp) ;
	mMenuBar->AddMenu(theMenu) ;

	theMenu = new TMenu("Direct Methods") ;
	theMenu->AddMenuItem("Run Feasible Sets...",kRunFeasSet) ;

	//theMenu->AddMenuItem("Symmetrize HKL data...",kSymmHKLData) ;
	theMenu->AddMenuItem("-") ;	// Divider
    TMenu   *importMenu = new TMenu("Import HKL Data") ;
		importMenu->AddMenuItem("H K L Intensity",kImport01) ;
		importMenu->AddMenuItem("L H K Intensity (ELD)",kImport02) ;
		importMenu->AddMenuItem("H K Intensity",kImport03) ;
		importMenu->AddMenuItem("H K F",kImport04) ;
		theMenu->AddMenuItem(importMenu) ;
	TMenu   *exportMenu = new TMenu("Export HKL Data") ;
		exportMenu->AddMenuItem("H K L Intensity",kExport01) ;
		exportMenu->AddMenuItem("L H K Intensity (ELD)",kExport02) ;
		exportMenu->AddMenuItem("H K Intensity",kExport03) ;
		exportMenu->AddMenuItem("H K F",kExport04) ;
		theMenu->AddMenuItem(exportMenu) ;


	TMenu	*symmMenu = new TMenu("Symmetrize HKL data") ;
	symmMenu->AddMenuItem(gAlternatePlaneGroup[0],801) ;
	symmMenu->AddMenuItem("-") ;	// Divider
	symmMenu->AddMenuItem(gAlternatePlaneGroup[1],802) ;
	symmMenu->AddMenuItem("-") ;	// Divider
	symmMenu->AddMenuItem(gAlternatePlaneGroup[2],803) ;
	symmMenu->AddMenuItem(gAlternatePlaneGroup[3],804) ;
	symmMenu->AddMenuItem(gAlternatePlaneGroup[4],805) ;
	symmMenu->AddMenuItem(gAlternatePlaneGroup[5],806) ;
	symmMenu->AddMenuItem(gAlternatePlaneGroup[6],807) ;
	symmMenu->AddMenuItem(gAlternatePlaneGroup[7],808) ;
	symmMenu->AddMenuItem("-") ;	// Divider
	symmMenu->AddMenuItem(gAlternatePlaneGroup[8],809) ;
	symmMenu->AddMenuItem(gAlternatePlaneGroup[9],810) ;
	symmMenu->AddMenuItem(gAlternatePlaneGroup[10],811) ;
	symmMenu->AddMenuItem(gAlternatePlaneGroup[11],812) ;
	symmMenu->AddMenuItem(gAlternatePlaneGroup[12],813) ;
	symmMenu->AddMenuItem("-") ;	// Divider
	symmMenu->AddMenuItem(gAlternatePlaneGroup[13],814) ;
	symmMenu->AddMenuItem(gAlternatePlaneGroup[14],815) ;
	symmMenu->AddMenuItem(gAlternatePlaneGroup[15],816) ;
	symmMenu->AddMenuItem(gAlternatePlaneGroup[16],817) ;
	symmMenu->AddMenuItem("-") ;	// Divider
	symmMenu->AddMenuItem(gAlternatePlaneGroup[17],818) ;
	symmMenu->AddMenuItem(gAlternatePlaneGroup[18],819) ;
	symmMenu->AddMenuItem(gAlternatePlaneGroup[19],820) ;
	symmMenu->AddMenuItem("-") ;	// Divider
	symmMenu->AddMenuItem(gAlternatePlaneGroup[20],821) ;
	symmMenu->AddMenuItem(gAlternatePlaneGroup[21],822) ;

	theMenu->AddMenuItem(symmMenu) ;

	theMenu->AddMenuItem("-") ;	// Divider
	theMenu->AddMenuItem("Run MEM...",kRunMEM) ;
	theMenu->AddMenuItem("-") ; // Divider ;
	theMenu->AddMenuItem("Trim a dataset"); // Placeholder for trim2D
	theMenu->AddMenuItem("Merge datasets"); // Placeholder for mergehkl
	theMenu->AddMenuItem("-");
	theMenu->AddMenuItem("Help",kHelp);
	mMenuBar->AddMenu(theMenu) ;
	
	gWindowsMenu = new TMenu("Windows") ;
	mMenuBar->AddMenu(gWindowsMenu) ;

	theMenu = new TMenu("?") ;
	mMenuBar->AddMenu(theMenu) ;

}

void	TDMApplicationWindow::DoCommand(TCommand command,Boolean *hasBeenHandled) {
	TWindow	*theWindow ;
	FSInfo	theFSInfo ;
	TImage	*theImage ;
	TImage u(128,128) ;
	TImage v(128,128) ;
	char	*cp,fileName[256], docommand[256] ;
	int	status = 0 ;
    int     ImportMode ;
	
	switch(command) {
		case kClose:
			theWindow = GetTargetWindow() ;
			if (theWindow == this) {
				TerminateApplication() ;
			}				
			else if (theWindow != nil) {
				RemoveWindow(theWindow) ;
				*hasBeenHandled = true ;
			}
			break ;	
		case kHelp:
			OnCommandHelp() ;		
			*hasBeenHandled = true ;
			break ;
		case kPaste:
			theImage = new TImage(*(TImage *)GetClipboardObject(0)) ;
			theImage->SetName("untitled") ;
			theImage->Display() ;
			*hasBeenHandled = true ;
			break ;
		case kClickToFocus:
			gWMFocusMove = 1 - gWMFocusMove ;
			*hasBeenHandled = true ;
			break ;
		case kSetDirectory:
			ChangeWorkingDirectory() ;
			*hasBeenHandled = true ;
			break ;
		case kQuit:
			TerminateApplication() ;
			break ;
		case kNew:
			*hasBeenHandled = true ;
			break ;
		case kOpen:
			ChooseFile("Choose Image File","*",theFSInfo) ;
			if (theFSInfo.mReply) {
				OpenImageFile(theFSInfo) ;
				theWindow = GetTargetWindow() ;
				if (theWindow != nil) {
					theWindow->SetCommander(this) ;
					theWindow->SetControls() ;
					// Set the base name for fs98 in case it has not been set by now
					if (strcmp(gFS98ControlRecord.baseName,"untitled") == 0) {
						// Create the base name
						strcpy(fileName,theFSInfo.mFileName) ;
						cp = strrchr(fileName,'.') ;
						if (cp != NULL) *cp = '\0' ;
						// FInd any directory path name
						if ((cp = strrchr(fileName,'/')) != NULL) {
							strcpy(gFS98ControlRecord.baseName,cp+1) ;
						}
						else {
							strcpy(gFS98ControlRecord.baseName,fileName) ;
						}
					}
				}
			}
			*hasBeenHandled = true ;
			break ;
		case kOpenHKLFile:
			ChooseFile("Choose HKL File","*.hkl",theFSInfo) ;
			if (theFSInfo.mReply) {
				TDiffractionDataTable *theDPTable = new TDiffractionDataTable(&theFSInfo) ;
				theWindow = (TWindow *)theDPTable->GetTablePane() ;
				if (theWindow != nil) {
					theWindow->SetCommander(this) ;
					AddWindow(theWindow) ;
				}
				// Set this as the current working directory as well since we will also pick up
				// the control files from this directory
				ChangeWorkingDirectory(theFSInfo) ;
			}
			*hasBeenHandled = true ;
			break ;
//#ifdef HAVE_CYGWIN
		case kWinOpen:
			ChooseFile("Choose File","*",theFSInfo) ;
			if (theFSInfo.mReply) {
				sprintf(docommand,"cygstart %s",theFSInfo.mFileName);
				status = system(docommand) ;
			}
			*hasBeenHandled = true ;
			break ;
		case kWinWordpad:
			//ChooseFile("Choose File","*",theFSInfo) ;
			//if (theFSInfo.mReply) {
				sprintf(docommand,"cygstart wordpad");
				status = system(docommand) ;
			//}
			*hasBeenHandled = true ;
			break ;
                // Import-Output ?
                case kImport01:
                        ImportMode = 1 ;
                case kImport02:
                        ImportMode = 2 ;
                case kImport03:
                        ImportMode = 3 ;
                case kImport04:
                        ImportMode = 4 ;
                ChooseFile("Choose HKL File","*.hkl",theFSInfo) ;
				if (theFSInfo.mReply) {
					TDiffractionDataTable *theDPTable = new TDiffractionDataTable(&theFSInfo, ImportMode) ;
					theWindow = (TWindow *)theDPTable->GetTablePane() ;
					if (theWindow != nil) {
						theWindow->SetCommander(this) ;
						AddWindow(theWindow) ;
					}
					// Set this as the current working directory as well since we will also pick up
					// the control files from this directory
					ChangeWorkingDirectory(theFSInfo) ;
				}
			*hasBeenHandled = true ;
			break ;
		case kWinNotepad:
			ChooseFile("Choose File","*",theFSInfo) ;
			if (theFSInfo.mReply) {
				//sprintf(docommand,"cygstart notepad");
                                sprintf(docommand,"cygstart notepad /cygwin%s",theFSInfo.mFileName);
                                //fprintf(stderr,"? %s\n",docommand) ;
				status = system(docommand) ;
			}
			//else {
                        //	sprintf(docommand,"cygstart notepad ");
			//	status = system(docommand) ;
			//}
			*hasBeenHandled = true ;
			break ;
		case kWinprint:
			ChooseFile("Choose File","*",theFSInfo) ;
			if (theFSInfo.mReply) {
				sprintf(docommand,"cygstart --print %s",theFSInfo.mFileName);
				status = system(docommand) ;
                                //fprintf(stderr," %s",theFSInfo.mFileName);
			}
			*hasBeenHandled = true ;
			break ;
		case kWinbash:
			sprintf(docommand,"cygstart bash -i ");
			status = system(docommand) ;
			*hasBeenHandled = true ;
			break ;
		case kxv:
			sprintf(docommand,"xv &");
			status = system(docommand) ;
			*hasBeenHandled = true ;
			break ;
		case kWinDir:
			ChooseFile("Choose a File in the Directory","*",theFSInfo) ;
			if (theFSInfo.mReply) {
				char workingDirectory[256] ;
				strcpy(workingDirectory,theFSInfo.mFileName) ;
				// Find the path excluding the file name
				char *fs = strrchr(workingDirectory,'/') ;
				if (fs != NULL) *fs = '\0' ;
				sprintf(docommand,"cygstart %s",workingDirectory);
				status = system(docommand) ;
			}
			*hasBeenHandled = true ;
			break ;
//#endif
		case kxterm:
			sprintf(docommand,"xterm &");
			status = system(docommand) ;
			*hasBeenHandled = true ;
			break ;
		case kOpenNCEMSSFile:
			ChooseFile("Choose NCEMSS Structure File","*.at",theFSInfo) ;
			if (theFSInfo.mReply) {
				TCrystal *theCrystal ;
				Boolean succeeded = true ;
				try {
					theCrystal = new TCrystal(theFSInfo) ;
				}
				catch (const char *string) {
					PutMessage("Failed in creating the crystal from the file") ;
					succeeded = false ;
				}
				if (succeeded) {
					theWindow = (TWindow *)theCrystal->GetWindow() ;
					if (theWindow != nil) {
						theWindow->SetCommander(this) ;
						AddWindow(theWindow) ;
					}
				}
			}
			*hasBeenHandled = true ;
			break ;
		case kOpenCBEDFile:
			ChooseFile("Choose CBED Control File","*.txt",theFSInfo) ;
			if (theFSInfo.mReply) {
				TCrystal *theCrystal ;
				Boolean succeeded = true ;
				try {
					theCrystal = new TCrystal(theFSInfo) ;
				}
				catch (const char *string) {
					PutMessage("Failed in creating the crystal from the file") ;
					succeeded = false ;
				}
				if (succeeded) {
					theWindow = (TWindow *)theCrystal->GetWindow() ;
					if (theWindow != nil) {
						theWindow->SetCommander(this) ;
						AddWindow(theWindow) ;
					}
				}
			}
			*hasBeenHandled = true ;
			break ;		
		default:
			break ;
	}

	if (command > 32768) {
		// This is a window
		theWindow = (TWindow *)command ;
		if (theWindow->CanReceiveMenuCommands()) {
			SetTargetWindow(theWindow) ;
			BringToFront(theWindow) ;
			gFrontWindow = theWindow ;
		}
	}


	SetFocus(this) ;
}

void	TDMApplicationWindow::GetCommandStatus(TCommand command,
		Boolean *isEnabled,Boolean *isChecked) {
		
	TBaseObject *clipboardObject ;
	
	if (command > 32768) {
		// This is a window
		*isEnabled = true ;
		TWindow *theWindow = (TWindow *)command ;
		*isChecked = theWindow == gFrontWindow ;
		return ;
	}
	
	switch(command) {
		case kQuit:
			*isEnabled = true ;
			break ;
		case kNew:
			*isEnabled = true ;
			break ; // We don't create windows on out own
		case kOpen:
		case kOpenHKLFile:
		case kOpenNCEMSSFile:
		case kOpenCBEDFile:
		case kxterm:
		case kImport01:
		case kImport02:
		case kImport03:
		case kImport04:
#ifdef HAVE_CYGWIN
		case kWinbash:
		case kWinOpen:
		case kWinWordpad:
		case kWinNotepad:
		case kWinDir:
		case kWinprint:
#endif
#ifdef HAVE_XV
		case kxv:
#endif
		case kHelp:
		case kSetDirectory:
			*isEnabled = true ;
			break ;
		case kPaste:
			if (!*isEnabled && !ClipboardEmpty()) {
				// check to see if we have an image on the clipboard
				if (GetNumberOfClipboardObjects() == 1) {
					clipboardObject = GetClipboardObject(0) ;
					if (clipboardObject->GetSignature() == 'imag') 
						*isEnabled = true ;
				}
			}
			break ;
		case kClickToFocus:
			*isEnabled = true ;
			*isChecked = gWMFocusMove == false ;
			break ;
		default:
			break ;
	}	
}
