// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------

#include <ctype.h>
#include "TDiffractionDataTable.h"
#include "TDMApplicationWindow.h"
#include "DMMenuCommands.h"
#include "XDrivers.h"
#include "TRefinementSet.h"
#include <math.h>
#include <stdio.h>
#define	MonitorWidth	1024
#define	MonitorHeight	768 
#define kUnAssignedPhase	-9999.
void	PutMessage(char *) ;
char	*ReadLine(char *buffer,long numChars,FILE *fp) ;

extern	class TDMApplicationWindow *gTheControlWindow ;
extern	void	AddWindow(TWindow *window) ; 


void  pvgs (long *he, long *ke, long *lse, long h, long k, long &ne, char *symm ) ;
void pgs (long *he, long *ke, long *lse, long h, long k, long &ne, char *symm ) ;
void p6symm (long *he, long *ke, long *lse, long h, long k, long &ne, char *symm ) ;
void p4symm (long *he, long *ke, long *lse, long h, long k, long &ne, char *symm ) ;
void p4mg (long *he, long *ke, long *lse, long h, long k, long &ne) ;
void p4 (long *he, long *ke, long *lse, long h, long k, long &ne, char *symm ) ;
void p3symm (long *he, long *ke, long *lse, long h, long k, long &ne, char *symm ) ;
void p2symm (long *he, long *ke, long *lse, long h, long k, long &ne, char *symm ) ; 
void p2mgs (long *he, long *ke, long *lse, long h, long k, long &ne, char *symm ) ; 
void p2ggs (long *he, long *ke, long *lse, long h, long k, long &ne, char *symm ) ;
void p1symm (long *he, long *ke, long *lse, long h, long k, long &ne, char *symm ) ; 
void ms (long *he, long *ke, long *lse, long h, long k, long &ne, char *symm ) ;
void f2symm (long *he, long *ke, long *lse, long h, long k, long &ne, char *symm ) ;
void equiv2 (long *he, long *ke, long *lse, long h, long k, long &ne, char *symm ) ; 
void averagehkl() ;

#define log2(x) log(x)/log(2.0) 
#define	iabs(x) (x) < 0 ? -(x) : (x) 

#define	CR	'\r'
#define LF	'\n'

#define max(x,y) (x) > (x) > (y) ? (x) : (y)
#define min(x,y) (x) > (x) < (y) ? (x) : (y)

extern	double Pi ;
extern	char	*gPlaneGroup[] ;
extern	char	*gPlaneGroupSymbols[] ;

char	*gOldPlaneGroupSymbols[] = {"P1  ","P2  ","Pm  ","Pg  ","Cm  ","P2mm","P2mg","P2gg",
				"C2mm","P4  ","P4mm","P4gm","P3  ","P3m1","P31m","P6  ","P6mm",nil} ;

extern	Point	gImageDisplayPosition ;
float	SymmetryRelatedPhase(short h,short k,short h1,short k1,float phase, short planegroup ) ;

class	TRunDirectMethodsDialog ;
extern	TRunDirectMethodsDialog *gRunDirectMethodsDialog ;

extern	FS98ControlRecord gFS98ControlRecord ;


TTableClipBoardList::TTableClipBoardList(long	numElements) {
	mNumValues	= numElements ;
	mCalibration.isCalibrated = false ;
	mIndex1	= new short[mNumValues] ;
	mIndex2	= new short[mNumValues] ;
	mValue	= new float[mNumValues] ;
}

TTableClipBoardList::~TTableClipBoardList() {
	delete[] mIndex1 ;
	delete[] mIndex2 ;
	delete[] mValue ;
}

TTableClipBoardList	*gTableClipBoardList = nil ;

TDiffractionDataTable::TDiffractionDataTable(FSInfo	*theFSInfo) 
	: TTable(theFSInfo->mFileName) {
	
	long 	i,groupIndex,j ;
	char	c,c0,buffer[256] ;
	char 	*cp,temp[32] ;
	char	fileName[256] ;

	mRefinementSet = nil ;
	
	mPlaneGroup			= 1 ;	// Since we don't know what the plane group is
	mUseAllReflections	= true ;
	mUseAllPhases		= false ;
	mAllUnknown			= true ;
	
	mCalibrationUnit	= kInversePixel ;

	// Create the base name
	strcpy(fileName,theFSInfo->mFileName) ;
	cp = strstr(fileName,".hkl") ;
	*cp = '\0' ;
	// FInd any directory path name
	if ((cp = strrchr(fileName,'/')) != NULL) {
		strcpy(gFS98ControlRecord.baseName,cp+1) ;
	}
	else {
		strcpy(gFS98ControlRecord.baseName,fileName) ;
	}
	
	// Adjust the title of the table
	strcpy(mTitle,gFS98ControlRecord.baseName) ;
	
	// First count the number of lines
	
	FILE *fp = fopen(theFSInfo->mFileName,"r") ;
	if (fp == nil) return ;
	mNumReflections = 0 ;
	//LDM Get first character
	c0 = fgetc(fp) ;	
	for (; (c = fgetc(fp)) != EOF ; ) {
		if (c == CR || c == LF) mNumReflections++ ;
	}	
	
	if (mNumReflections == 0) {
		PutMessage("Did not find any reflection data") ;
		return ;
	}

	mGVector 	= new float[mNumReflections] ;
	mAmplitude 	= new float[mNumReflections] ;
	mPhase 		= new float[mNumReflections] ;
	mH			= new short[mNumReflections] ;
	mK			= new short[mNumReflections] ;
	mIndex		= new short[mNumReflections] ;
	mIncludeReflection	= new bool[mNumReflections] ;
	mUsePhase	= new bool[mNumReflections] ;
	
	// Now rewind and read in the data
	rewind(fp) ;
	short l ;
	groupIndex = 1 ;
	bool edm_format = false ;
	for (i = 0 ;  i < mNumReflections; i++) {
		if (ReadLine(buffer,256,fp) == NULL) break ;
		//LDM skip first line if it starts with h
		if ( i == 0 && buffer[0] == 'h' ) {
			//We have an edm format file, which has more information
			edm_format = true ;
			//fprintf(stderr , "Inside logical\n" );
			if (ReadLine(buffer,256,fp) == NULL) break ;
		}
		//LDM: BUG here sometimes if there is more at the end!
		sscanf(buffer,"%hd %hd %hd %f %f",&mH[i],&mK[i],&l,&mAmplitude[i],&mPhase[i]) ;	
		mGVector[i] = 0. ;
		mIndex[i] = groupIndex++ ;
		mIncludeReflection[i] = true ;
		mUsePhase[i] = mPhase[i] == 0. ? false : true ;
		if (!mUsePhase[i]) mPhase[i] = kUnAssignedPhase ;
	}	
	fclose(fp) ;
	//LDM Correct the number of reflections if needed
	if (c0 == 'h') mNumReflections = mNumReflections - 1 ;
	// Now read the 'ins' file to get the rest of the information
	strcpy(fileName,theFSInfo->mFileName) ;
	cp = strstr(fileName,".hkl") ;
	*(cp+1) = '\0' ;
	strcat(fileName,"ins") ;

	mSortKey = kSortOnG ;
	
	gFS98ControlRecord.powerOfAmplitude = 1.0 ;
	gFS98ControlRecord.DefineOrigin = false ;
	
	long	numAtoms, Other ;
	char	planeGroupSymbol[5] ;
        Other = -1 ;
	fp = fopen(fileName,"r") ;
	if (fp == nil) {
		PutMessage("Unknown calibration. d-values will be incorrect") ;
		mA = gFS98ControlRecord.a ;
		mB = gFS98ControlRecord.b ;
		mGamma = gFS98ControlRecord.ga ;
		gFS98ControlRecord.contentKnown = 0 ;//false ;	
		gFS98ControlRecord.symm = gPlaneGroupSymbols[0] ;			
	}
	else {
		gFS98ControlRecord.useElectron = false ; 
		gFS98ControlRecord.floatBeams = false ;
		gFS98ControlRecord.verbose = false ;
		gFS98ControlRecord.contentKnown = 0 ; //true ;
        gFS98ControlRecord.Other = -1 ;
		if (gFS98ControlRecord.sfac != nil) delete[] gFS98ControlRecord.sfac ;
		if (gFS98ControlRecord.unit != nil) delete[] gFS98ControlRecord.unit ;
		if (gFS98ControlRecord.tfac != nil) delete[] gFS98ControlRecord.tfac ;
		if (gFS98ControlRecord.aocc != nil) delete[] gFS98ControlRecord.aocc ;

		// We will set off space for maximum of 32 entries	
		gFS98ControlRecord.sfac = new Str5[32] ;
		gFS98ControlRecord.unit = new long[32] ;
		gFS98ControlRecord.tfac = new float[32] ;
		gFS98ControlRecord.aocc = new float[32] ;

		for(;;) {
			if (ReadLine(buffer,256,fp) == NULL) break ;
			//fprintf(stderr,"\nRead:  %s",buffer) ; fflush(stderr) ;
			if (strstr(buffer,"TITL") != NULL) {
				sscanf(buffer,"TITL\t%s",gFS98ControlRecord.titl) ;
			}
			else if (strstr(buffer,"CELL") != NULL) {
				sscanf(buffer,"CELL\t%f %f %f %f %f %f %f",
					&gFS98ControlRecord.wavl,&gFS98ControlRecord.a,&gFS98ControlRecord.b,&gFS98ControlRecord.c,
					&gFS98ControlRecord.al,&gFS98ControlRecord.be,&gFS98ControlRecord.ga) ;
				mA = gFS98ControlRecord.a ;
				mB = gFS98ControlRecord.b ;
				mGamma = gFS98ControlRecord.ga ;				
				mCalibrationUnit	= kInverseAngstrom ;
			}
			else if (strstr(buffer,"SYMM") != NULL) {
				sscanf(buffer,"SYMM\t%s",planeGroupSymbol) ;
				// Check for our new symbols
				mPlaneGroup = -1 ;
				// Compare all in lower case
				for(cp = planeGroupSymbol; *cp; cp++) { *cp = tolower(*cp) ; } 
				for(i=0;i < 22; i++) {
					strcpy(temp,gPlaneGroupSymbols[i]) ;
					for(cp = temp; *cp; cp++) { *cp = tolower(*cp) ; } 
					if (strncmp(planeGroupSymbol,temp,strlen(planeGroupSymbol)) == 0) {
						gFS98ControlRecord.symm = gPlaneGroupSymbols[i] ;
						mPlaneGroup = i+1 ;
						break ;
					}
				}
				if (mPlaneGroup == -1) {
					// Check our old range 1-17 symbols
					for(i=0;i < 17; i++) {
						strcpy(temp,gOldPlaneGroupSymbols[i]) ;
						for(cp = temp; *cp; cp++) { *cp = tolower(*cp) ; } 
						if (strncmp(planeGroupSymbol,temp,strlen(planeGroupSymbol)) == 0) {
							mPlaneGroup = i+1 ;
							if (mPlaneGroup == 4) 
								mPlaneGroup == 5 ;
							else if (mPlaneGroup == 5) 
								mPlaneGroup == 7 ;
							else if (mPlaneGroup == 6) 
								mPlaneGroup == 9 ;
							else if (mPlaneGroup == 7) 
								mPlaneGroup == 10 ;
							else if (mPlaneGroup == 8) 
								mPlaneGroup == 12 ;
							else if (mPlaneGroup == 9) 
								mPlaneGroup == 13 ;
							else if (mPlaneGroup == 10) 
								mPlaneGroup == 14 ;
							else if (mPlaneGroup == 11) 
								mPlaneGroup == 15 ;
							else if (mPlaneGroup == 12) 
								mPlaneGroup == 16 ;
							else if (mPlaneGroup == 13) 
								mPlaneGroup == 18 ;
							else if (mPlaneGroup == 14) 
								mPlaneGroup == 19 ;
							else if (mPlaneGroup == 15) 
								mPlaneGroup == 20 ;
							else if (mPlaneGroup == 16) 
								mPlaneGroup == 21 ;
							else if (mPlaneGroup == 17) 
								mPlaneGroup == 22 ;								
							gFS98ControlRecord.symm = gPlaneGroupSymbols[mPlaneGroup-1] ;
							break ;
						}
					}
				}
				if (mPlaneGroup == -1) mPlaneGroup = 1 ;				
			}
			else if (strstr(buffer,"SFAC") != NULL) {
				// Find out the number of atoms we have
				numAtoms = sscanf(buffer,"SFAC\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s",
					gFS98ControlRecord.sfac[0],gFS98ControlRecord.sfac[1],gFS98ControlRecord.sfac[2],gFS98ControlRecord.sfac[3],
					gFS98ControlRecord.sfac[4],gFS98ControlRecord.sfac[5],gFS98ControlRecord.sfac[6],gFS98ControlRecord.sfac[7],
					gFS98ControlRecord.sfac[8],gFS98ControlRecord.sfac[9],gFS98ControlRecord.sfac[10],gFS98ControlRecord.sfac[11],
					gFS98ControlRecord.sfac[12],gFS98ControlRecord.sfac[13],gFS98ControlRecord.sfac[14],gFS98ControlRecord.sfac[15]) ;
			}
			else if (strstr(buffer,"UNIT") != NULL) {
				numAtoms = sscanf(buffer,"UNIT\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld",
					&gFS98ControlRecord.unit[0],&gFS98ControlRecord.unit[1],&gFS98ControlRecord.unit[2],&gFS98ControlRecord.unit[3],
					&gFS98ControlRecord.unit[4],&gFS98ControlRecord.unit[5],&gFS98ControlRecord.unit[6],&gFS98ControlRecord.unit[7],
					&gFS98ControlRecord.unit[8],&gFS98ControlRecord.unit[9],&gFS98ControlRecord.unit[10],&gFS98ControlRecord.unit[11],
					&gFS98ControlRecord.unit[12],&gFS98ControlRecord.unit[13],&gFS98ControlRecord.unit[14],&gFS98ControlRecord.unit[15]) ;
			}
			else if (strstr(buffer,"TFAC") != NULL) {
				numAtoms = sscanf(buffer,"TFAC\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f",
					&gFS98ControlRecord.tfac[0],&gFS98ControlRecord.tfac[1],&gFS98ControlRecord.tfac[2],&gFS98ControlRecord.tfac[3],
					&gFS98ControlRecord.tfac[4],&gFS98ControlRecord.tfac[5],&gFS98ControlRecord.tfac[6],&gFS98ControlRecord.tfac[7],
					&gFS98ControlRecord.tfac[8],&gFS98ControlRecord.tfac[9],&gFS98ControlRecord.tfac[10],&gFS98ControlRecord.tfac[11],
					&gFS98ControlRecord.tfac[12],&gFS98ControlRecord.tfac[13],&gFS98ControlRecord.tfac[14],&gFS98ControlRecord.tfac[15]) ;
			}
			else if (strstr(buffer,"AOCC") != NULL) {
				numAtoms = sscanf(buffer,"AOCC\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f",
					&gFS98ControlRecord.aocc[0],&gFS98ControlRecord.aocc[1],&gFS98ControlRecord.aocc[2],&gFS98ControlRecord.aocc[3],
					&gFS98ControlRecord.aocc[4],&gFS98ControlRecord.aocc[5],&gFS98ControlRecord.aocc[6],&gFS98ControlRecord.aocc[7],
					&gFS98ControlRecord.aocc[8],&gFS98ControlRecord.aocc[9],&gFS98ControlRecord.aocc[10],&gFS98ControlRecord.aocc[11],
					&gFS98ControlRecord.aocc[12],&gFS98ControlRecord.aocc[13],&gFS98ControlRecord.aocc[14],&gFS98ControlRecord.aocc[15]) ;
			}
			else if (strstr(buffer,"MODE") != NULL) {
				sscanf(buffer,"MODE\t%4s",&gFS98ControlRecord.mode) ;
				//fprintf(stderr,"\n MODE %s",&gFS98ControlRecord.mode) ;
			}
			else if (strstr(buffer,"ELECTRON") != NULL) {
				gFS98ControlRecord.useElectron = true ;
			}
			else if (strstr(buffer,"XRAY") != NULL) {
				gFS98ControlRecord.useElectron = false ;
			}
			else if (strstr(buffer,"PLAMBDA") != NULL) {
				sscanf(buffer,"PLAMBDA\t%f",&gFS98ControlRecord.plambda) ;
			}
			else if (strstr(buffer,"LAMBDA") != NULL) {
				if (strstr(buffer,"AUTO") != NULL) {
					gFS98ControlRecord.autoSpeed = true ;
					sscanf(buffer,"LAMBDA\t%f",&gFS98ControlRecord.lambda) ;
				}
				else {
					gFS98ControlRecord.autoSpeed = false ;
					sscanf(buffer,"LAMBDA\t%f\t%f",&gFS98ControlRecord.lambda,&gFS98ControlRecord.speed) ;
				}
			}
			else if (strstr(buffer,"CAMP") != NULL) {
				sscanf(buffer,"CAMP\t%f",&gFS98ControlRecord.camp) ;
			}
			else if (strstr(buffer,"ORIG") != NULL) 
				gFS98ControlRecord.DefineOrigin = true ;
			else if (strstr(buffer,"TIMES") != NULL) {
				sscanf(buffer,"TIMES\t%f",&gFS98ControlRecord.times) ;
			}
			else if (strstr(buffer,"WIDTH") != NULL) {
				sscanf(buffer,"WIDTH\t%ld",&gFS98ControlRecord.width) ;
			}
			else if (strstr(buffer,"HEIGHT") != NULL) {
				sscanf(buffer,"HEIGHT\t%ld",&gFS98ControlRecord.height) ;
			}
			else if (strstr(buffer,"POWER") != NULL) {
				sscanf(buffer,"POWER\t%f",&gFS98ControlRecord.powerOfAmplitude) ;
			}
			else if (strstr(buffer,"FLOAT") != NULL) {
				gFS98ControlRecord.floatBeams = true ;
			}
			else if (strstr(buffer,"VERB") != NULL) {
				gFS98ControlRecord.verbose = true ;
			}
			else if (strstr(buffer,"REM") != NULL ) {
                                short dummy = 0 ;
                        }
			//Adding the PUNT command
			else if (strstr(buffer,"PUNT") !=NULL ) {
				gFS98ControlRecord.contentKnown = 1 ;
			}
			//Adding the PUNU command
			else if (strstr(buffer,"PUNU") !=NULL ) {
				gFS98ControlRecord.contentKnown = 2 ;
			}

			else if (strstr(buffer,"END") != NULL) {
				break ;
                        }
                        else {
                                Other = Other + 1 ;   // Increment
                                for ( j=0; j < 256 ; j++ ) {
                                gFS98ControlRecord.OtherLines[Other][j] = buffer[j] ;
                                       }
                                //fprintf(stderr,"%s\n",buffer);
		        }
                }
		gFS98ControlRecord.Other = Other ;
		//fprintf(stderr,"Found %d \n", Other);
		gFS98ControlRecord.numAtoms = numAtoms ;
		
		fclose(fp) ;
	}
	

	// Create the g-vectors	
	float	gaSt  = 180.0 - mGamma ;
	float	csGaSt	= cos(asin(1.0)*gaSt/90.) ;
	float	siGaSt	= sin(asin(1.0)*gaSt/90.) ;
	float	aStar = 1./(mA*siGaSt) ;	
	float	bStar = 1./(mB*siGaSt) ;	
	float	aStarSq = aStar*aStar ;
	float	bStarSq = bStar*bStar ;
	for(i = 0 ; i < mNumReflections; i++) {
		mGVector[i] = sqrt(mH[i]*mH[i]*aStarSq + mK[i]*mK[i]*bStarSq + 2*mH[i]*mK[i]*aStar*bStar*csGaSt) ;
	}
	//LDM, removed the sort on G SortOnG() ;

	
	mRows = mNumReflections ;
	mCols = 6 ;
	mCellDataSize = 4 ;
	mColWidth = 80 ;
	mRowHeight = 20 ;
		
	Int16 	xpos = gImageDisplayPosition.x ;
	Int16 	ypos = gImageDisplayPosition.y ;
	gImageDisplayPosition.x += 10 ;
	gImageDisplayPosition.y += 10 ;
				
	mParentPane = new TTableWindow(xpos,ypos,this) ;	
	((TWindow *)mParentPane)->SetWindowTitle(gFS98ControlRecord.baseName) ;
	

}
TDiffractionDataTable::TDiffractionDataTable(FSInfo	*theFSInfo, int ImportMode ) 
	: TTable(theFSInfo->mFileName) {
	
	long 	i,groupIndex,j ;
	char	c,c0,buffer[256] ;
	char 	*cp,temp[32] ;
	char	fileName[256] ;
	float tmp ;
	mRefinementSet = nil ;
	
	mPlaneGroup			= 1 ;	// Since we don't know what the plane group is
	mUseAllReflections	= true ;
	mUseAllPhases		= false ;
	mAllUnknown		= true ;
	
	mCalibrationUnit	= kInversePixel ;

	// Create the base name
	strcpy(fileName,theFSInfo->mFileName) ;
	cp = strstr(fileName,".hkl") ;
	*cp = '\0' ;
	// FInd any directory path name
	if ((cp = strrchr(fileName,'/')) != NULL) {
		strcpy(gFS98ControlRecord.baseName,cp+1) ;
	}
	else {
		strcpy(gFS98ControlRecord.baseName,fileName) ;
	}
	
	// Adjust the title of the table
	strcpy(mTitle,gFS98ControlRecord.baseName) ;
	
	// First count the number of lines
	
	FILE *fp = fopen(theFSInfo->mFileName,"r") ;
	if (fp == nil) return ;
	mNumReflections = 0 ;
	//LDM Get first character
	c0 = fgetc(fp) ;	
	for (; (c = fgetc(fp)) != EOF ; ) {
		if (c == CR || c == LF) mNumReflections++ ;
	}	
	
	if (mNumReflections == 0) {
		PutMessage("Did not find any reflection data") ;
		return ;
	}

	mGVector 	= new float[mNumReflections] ;
	mAmplitude 	= new float[mNumReflections] ;
	mPhase 		= new float[mNumReflections] ;
	mH		= new short[mNumReflections] ;
	mK		= new short[mNumReflections] ;
	mIndex		= new short[mNumReflections] ;
	mIncludeReflection	= new bool[mNumReflections] ;
	mUsePhase	= new bool[mNumReflections] ;
	
	// Now rewind and read in the data
	rewind(fp) ;
	short l ;
	groupIndex = 1 ;
	bool edm_format = false ;
	for (i = 0 ;  i < mNumReflections; i++) {
		if (ReadLine(buffer,256,fp) == NULL) break ;
		//LDM skip first line if it starts with h
		if ( i == 0 && buffer[0] == 'h' ) {
			//We have an edm format file, which has more information
			edm_format = true ;
			//fprintf(stderr , "Inside logical\n" );
			if (ReadLine(buffer,256,fp) == NULL) break ;
		}
		//LDM: BUG here sometimes if there is more at the end!
		switch(ImportMode) {
		case 1:
			sscanf(buffer,"%hd %hd %hd %f",&mH[i],&mK[i],&l, &tmp) ;
			mPhase[i] = 0 ;
			mAmplitude[i] = sqrt(tmp) ;	
			break ;
		case 2:
			sscanf(buffer,"%hd %hd %hd %f",&l,&mH[i],&mK[i], &tmp) ;
			mPhase[i] = 0 ;
			mAmplitude[i] = sqrt(tmp) ;	
			break ;
		case 3:
			sscanf(buffer,"%hd %hd %f",&mH[i],&mK[i], &tmp) ;
			mPhase[i] = 0 ;
			mAmplitude[i] = sqrt(tmp) ;	
			break ;
		case 4:
			sscanf(buffer,"%hd %hd %f",&mH[i],&mK[i], &mAmplitude[i]) ;
			mPhase[i] = 0 ;
			mGVector[i] = 0. ;
			break ;
		}

		mGVector[i] = 0. ;
		mIndex[i] = groupIndex++ ;
		mIncludeReflection[i] = true ;
		mUsePhase[i] = mPhase[i] == 0. ? false : true ;
		if (!mUsePhase[i]) mPhase[i] = kUnAssignedPhase ;
	}	
	fclose(fp) ;
	//LDM Correct the number of reflections if needed
	if (c0 == 'h') mNumReflections = mNumReflections - 1 ;
	// Now read the 'ins' file to get the rest of the information
	strcpy(fileName,theFSInfo->mFileName) ;
	cp = strstr(fileName,".hkl") ;
	*(cp+1) = '\0' ;
	strcat(fileName,"ins") ;

	mSortKey = kSortOnG ;
	
	gFS98ControlRecord.powerOfAmplitude = 1.0 ;
	gFS98ControlRecord.DefineOrigin = false ;
	
	long	numAtoms, Other ;
	char	planeGroupSymbol[5] ;
        Other = -1 ;
	fp = fopen(fileName,"r") ;
	if (fp == nil) {
		PutMessage("Unknown calibration. d-values will be incorrect") ;
		mA = gFS98ControlRecord.a ;
		mB = gFS98ControlRecord.b ;
		mGamma = gFS98ControlRecord.ga ;
		gFS98ControlRecord.contentKnown = 0 ;//false ;	
		gFS98ControlRecord.symm = gPlaneGroupSymbols[0] ;			
	}
	else {
		gFS98ControlRecord.useElectron = false ; 
		gFS98ControlRecord.floatBeams = false ;
		gFS98ControlRecord.verbose = false ;
		gFS98ControlRecord.contentKnown = 0 ; //true ;
        gFS98ControlRecord.Other = -1 ;
		if (gFS98ControlRecord.sfac != nil) delete[] gFS98ControlRecord.sfac ;
		if (gFS98ControlRecord.unit != nil) delete[] gFS98ControlRecord.unit ;
		if (gFS98ControlRecord.tfac != nil) delete[] gFS98ControlRecord.tfac ;
		if (gFS98ControlRecord.aocc != nil) delete[] gFS98ControlRecord.aocc ;

		// We will set off space for maximum of 32 entries	
		gFS98ControlRecord.sfac = new Str5[32] ;
		gFS98ControlRecord.unit = new long[32] ;
		gFS98ControlRecord.tfac = new float[32] ;
		gFS98ControlRecord.aocc = new float[32] ;

		for(;;) {
			if (ReadLine(buffer,256,fp) == NULL) break ;
			//fprintf(stderr,"\nRead:  %s",buffer) ; fflush(stderr) ;
			if (strstr(buffer,"TITL") != NULL) {
				sscanf(buffer,"TITL\t%s",gFS98ControlRecord.titl) ;
			}
			else if (strstr(buffer,"CELL") != NULL) {
				sscanf(buffer,"CELL\t%f %f %f %f %f %f %f",
					&gFS98ControlRecord.wavl,&gFS98ControlRecord.a,&gFS98ControlRecord.b,&gFS98ControlRecord.c,
					&gFS98ControlRecord.al,&gFS98ControlRecord.be,&gFS98ControlRecord.ga) ;
				mA = gFS98ControlRecord.a ;
				mB = gFS98ControlRecord.b ;
				mGamma = gFS98ControlRecord.ga ;				
				mCalibrationUnit	= kInverseAngstrom ;
			}
			else if (strstr(buffer,"SYMM") != NULL) {
				sscanf(buffer,"SYMM\t%s",planeGroupSymbol) ;
				// Check for our new symbols
				mPlaneGroup = -1 ;
				// Compare all in lower case
				for(cp = planeGroupSymbol; *cp; cp++) { *cp = tolower(*cp) ; } 
				for(i=0;i < 22; i++) {
					strcpy(temp,gPlaneGroupSymbols[i]) ;
					for(cp = temp; *cp; cp++) { *cp = tolower(*cp) ; } 
					if (strncmp(planeGroupSymbol,temp,strlen(planeGroupSymbol)) == 0) {
						gFS98ControlRecord.symm = gPlaneGroupSymbols[i] ;
						mPlaneGroup = i+1 ;
						break ;
					}
				}
				if (mPlaneGroup == -1) {
					// Check our old range 1-17 symbols
					for(i=0;i < 17; i++) {
						strcpy(temp,gOldPlaneGroupSymbols[i]) ;
						for(cp = temp; *cp; cp++) { *cp = tolower(*cp) ; } 
						if (strncmp(planeGroupSymbol,temp,strlen(planeGroupSymbol)) == 0) {
							mPlaneGroup = i+1 ;
							if (mPlaneGroup == 4) 
								mPlaneGroup == 5 ;
							else if (mPlaneGroup == 5) 
								mPlaneGroup == 7 ;
							else if (mPlaneGroup == 6) 
								mPlaneGroup == 9 ;
							else if (mPlaneGroup == 7) 
								mPlaneGroup == 10 ;
							else if (mPlaneGroup == 8) 
								mPlaneGroup == 12 ;
							else if (mPlaneGroup == 9) 
								mPlaneGroup == 13 ;
							else if (mPlaneGroup == 10) 
								mPlaneGroup == 14 ;
							else if (mPlaneGroup == 11) 
								mPlaneGroup == 15 ;
							else if (mPlaneGroup == 12) 
								mPlaneGroup == 16 ;
							else if (mPlaneGroup == 13) 
								mPlaneGroup == 18 ;
							else if (mPlaneGroup == 14) 
								mPlaneGroup == 19 ;
							else if (mPlaneGroup == 15) 
								mPlaneGroup == 20 ;
							else if (mPlaneGroup == 16) 
								mPlaneGroup == 21 ;
							else if (mPlaneGroup == 17) 
								mPlaneGroup == 22 ;								
							gFS98ControlRecord.symm = gPlaneGroupSymbols[mPlaneGroup-1] ;
							break ;
						}
					}
				}
				if (mPlaneGroup == -1) mPlaneGroup = 1 ;				
			}
			else if (strstr(buffer,"SFAC") != NULL) {
				// Find out the number of atoms we have
				numAtoms = sscanf(buffer,"SFAC\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s",
					gFS98ControlRecord.sfac[0],gFS98ControlRecord.sfac[1],gFS98ControlRecord.sfac[2],gFS98ControlRecord.sfac[3],
					gFS98ControlRecord.sfac[4],gFS98ControlRecord.sfac[5],gFS98ControlRecord.sfac[6],gFS98ControlRecord.sfac[7],
					gFS98ControlRecord.sfac[8],gFS98ControlRecord.sfac[9],gFS98ControlRecord.sfac[10],gFS98ControlRecord.sfac[11],
					gFS98ControlRecord.sfac[12],gFS98ControlRecord.sfac[13],gFS98ControlRecord.sfac[14],gFS98ControlRecord.sfac[15]) ;
			}
			else if (strstr(buffer,"UNIT") != NULL) {
				numAtoms = sscanf(buffer,"UNIT\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld",
					&gFS98ControlRecord.unit[0],&gFS98ControlRecord.unit[1],&gFS98ControlRecord.unit[2],&gFS98ControlRecord.unit[3],
					&gFS98ControlRecord.unit[4],&gFS98ControlRecord.unit[5],&gFS98ControlRecord.unit[6],&gFS98ControlRecord.unit[7],
					&gFS98ControlRecord.unit[8],&gFS98ControlRecord.unit[9],&gFS98ControlRecord.unit[10],&gFS98ControlRecord.unit[11],
					&gFS98ControlRecord.unit[12],&gFS98ControlRecord.unit[13],&gFS98ControlRecord.unit[14],&gFS98ControlRecord.unit[15]) ;
			}
			else if (strstr(buffer,"TFAC") != NULL) {
				numAtoms = sscanf(buffer,"TFAC\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f",
					&gFS98ControlRecord.tfac[0],&gFS98ControlRecord.tfac[1],&gFS98ControlRecord.tfac[2],&gFS98ControlRecord.tfac[3],
					&gFS98ControlRecord.tfac[4],&gFS98ControlRecord.tfac[5],&gFS98ControlRecord.tfac[6],&gFS98ControlRecord.tfac[7],
					&gFS98ControlRecord.tfac[8],&gFS98ControlRecord.tfac[9],&gFS98ControlRecord.tfac[10],&gFS98ControlRecord.tfac[11],
					&gFS98ControlRecord.tfac[12],&gFS98ControlRecord.tfac[13],&gFS98ControlRecord.tfac[14],&gFS98ControlRecord.tfac[15]) ;
			}
			else if (strstr(buffer,"AOCC") != NULL) {
				numAtoms = sscanf(buffer,"AOCC\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f",
					&gFS98ControlRecord.aocc[0],&gFS98ControlRecord.aocc[1],&gFS98ControlRecord.aocc[2],&gFS98ControlRecord.aocc[3],
					&gFS98ControlRecord.aocc[4],&gFS98ControlRecord.aocc[5],&gFS98ControlRecord.aocc[6],&gFS98ControlRecord.aocc[7],
					&gFS98ControlRecord.aocc[8],&gFS98ControlRecord.aocc[9],&gFS98ControlRecord.aocc[10],&gFS98ControlRecord.aocc[11],
					&gFS98ControlRecord.aocc[12],&gFS98ControlRecord.aocc[13],&gFS98ControlRecord.aocc[14],&gFS98ControlRecord.aocc[15]) ;
			}
			else if (strstr(buffer,"MODE") != NULL) {
				sscanf(buffer,"MODE\t%s",&gFS98ControlRecord.mode) ;
			}
			else if (strstr(buffer,"ELECTRON") != NULL) {
				gFS98ControlRecord.useElectron = true ;
			}
			else if (strstr(buffer,"XRAY") != NULL) {
				gFS98ControlRecord.useElectron = false ;
			}
			else if (strstr(buffer,"PLAMBDA") != NULL) {
				sscanf(buffer,"PLAMBDA\t%f",&gFS98ControlRecord.plambda) ;
			}
			else if (strstr(buffer,"LAMBDA") != NULL) {
				if (strstr(buffer,"AUTO") != NULL) {
					gFS98ControlRecord.autoSpeed = true ;
					sscanf(buffer,"LAMBDA\t%f",&gFS98ControlRecord.lambda) ;
				}
				else {
					gFS98ControlRecord.autoSpeed = false ;
					sscanf(buffer,"LAMBDA\t%f\t%f",&gFS98ControlRecord.lambda,&gFS98ControlRecord.speed) ;
				}
			}
			else if (strstr(buffer,"CAMP") != NULL) {
				sscanf(buffer,"CAMP\t%f",&gFS98ControlRecord.camp) ;
			}
			else if (strstr(buffer,"ORIG") != NULL) 
				gFS98ControlRecord.DefineOrigin = true ;
			else if (strstr(buffer,"TIMES") != NULL) {
				sscanf(buffer,"TIMES\t%f",&gFS98ControlRecord.times) ;
			}
			else if (strstr(buffer,"WIDTH") != NULL) {
				sscanf(buffer,"WIDTH\t%ld",&gFS98ControlRecord.width) ;
			}
			else if (strstr(buffer,"HEIGHT") != NULL) {
				sscanf(buffer,"HEIGHT\t%ld",&gFS98ControlRecord.height) ;
			}
			else if (strstr(buffer,"POWER") != NULL) {
				sscanf(buffer,"POWER\t%f",&gFS98ControlRecord.powerOfAmplitude) ;
			}
			else if (strstr(buffer,"FLOAT") != NULL) {
				gFS98ControlRecord.floatBeams = true ;
			}
			else if (strstr(buffer,"VERB") != NULL) {
				gFS98ControlRecord.verbose = true ;
			}
			else if (strstr(buffer,"REM") != NULL ) {
                                short dummy = 0 ;
                        }
			//Adding the PUNT command
			else if (strstr(buffer,"PUNT") !=NULL ) {
				gFS98ControlRecord.contentKnown = 1 ;
			}
			//Adding the PUNU command
			else if (strstr(buffer,"PUNU") !=NULL ) {
				gFS98ControlRecord.contentKnown = 2 ;
			}

			else if (strstr(buffer,"END") != NULL) {
				break ;
                        }
                        else {
                                Other = Other + 1 ;   // Increment
                                for ( j=0; j < 256 ; j++ ) {
                                gFS98ControlRecord.OtherLines[Other][j] = buffer[j] ;
                                       }
                                //fprintf(stderr,"%s\n",buffer);
		        }
                }
		gFS98ControlRecord.Other = Other ;
		//fprintf(stderr,"Found %d \n", Other);
		gFS98ControlRecord.numAtoms = numAtoms ;
		
		fclose(fp) ;
	}
	

	// Create the g-vectors	
	float	gaSt  = 180.0 - mGamma ;
	float	csGaSt	= cos(asin(1.0)*gaSt/90.) ;
	float	siGaSt	= sin(asin(1.0)*gaSt/90.) ;
	float	aStar = 1./(mA*siGaSt) ;	
	float	bStar = 1./(mB*siGaSt) ;	
	float	aStarSq = aStar*aStar ;
	float	bStarSq = bStar*bStar ;
	for(i = 0 ; i < mNumReflections; i++) {
		mGVector[i] = sqrt(mH[i]*mH[i]*aStarSq + mK[i]*mK[i]*bStarSq + 2*mH[i]*mK[i]*aStar*bStar*csGaSt) ;
	}
	//LDM, removed the sort on G SortOnG() ;

	
	mRows = mNumReflections ;
	mCols = 6 ;
	mCellDataSize = 4 ;
	mColWidth = 80 ;
	mRowHeight = 20 ;
		
	Int16 	xpos = gImageDisplayPosition.x ;
	Int16 	ypos = gImageDisplayPosition.y ;
	gImageDisplayPosition.x += 10 ;
	gImageDisplayPosition.y += 10 ;
				
	mParentPane = new TTableWindow(xpos,ypos,this) ;	
	((TWindow *)mParentPane)->SetWindowTitle(gFS98ControlRecord.baseName) ;

}

TDiffractionDataTable::TDiffractionDataTable(TDiffractionDataTable *s,short symmetryGroup,bool friedelSymmetry) 
		: TTable("Symmetrized") {
	long 	i,groupIndex ;
	//char	c ;

	mRefinementSet = nil ;
	
	mPlaneGroup			= symmetryGroup ;	// The new desired group
	
	// Might need to adjust for settings
	
	// The following are defaults fot the nre set
	mUseAllReflections	= true ;
	mUseAllPhases		= false ;
	mAllUnknown			= true ;

	mNumReflections		= s->mNumReflections ;	
	mCalibrationUnit	= s->mCalibrationUnit ;

		
	mGVector 	= new float[mNumReflections] ; 
	mAmplitude 	= new float[mNumReflections] ; 
	mH			= new short[mNumReflections] ;
	mK			= new short[mNumReflections] ; 

	mPhase 		= new float[mNumReflections] ;
	mIndex		= new short[mNumReflections] ;
	mIncludeReflection	= new bool[mNumReflections] ;
	mUsePhase	= new bool[mNumReflections] ;

	groupIndex = 1 ;
	for (i = 0 ;  i < mNumReflections; i++) {
		mIndex[i] = groupIndex++ ;
		mIncludeReflection[i] = true ;
		mUsePhase[i] = false ;
		mPhase[i] = kUnAssignedPhase ;
	}	
	
	// Our mAmplitude becomes the new set which is averaged from s->mAmplitude (the original set)
	// Symmetrize the data set
	SymmetrizeHKLDataSet(s,gPlaneGroupSymbols[symmetryGroup-1],false,kNoHKLError) ;

	// Create the new table		
	mRows = mNumReflections ;
	mCols = 6 ;
	mCellDataSize = 4 ;
	mColWidth = 80 ;
	mRowHeight = 20 ;
		
	Int16 	xpos = gImageDisplayPosition.x ;
	Int16 	ypos = gImageDisplayPosition.y ;
	gImageDisplayPosition.x += 10 ;
	gImageDisplayPosition.y += 10 ;
				
	mParentPane = new TTableWindow(xpos,ypos,this) ;
	
	char newTitle[256] ;
	char addOn[256] ;
	((TWindow *)s->mParentPane)->GetWindowTitle(newTitle) ;
	sprintf(addOn," [%s]",gPlaneGroupSymbols[symmetryGroup-1]) ;
	strcat(newTitle,addOn) ;
	((TWindow *)mParentPane)->SetWindowTitle(newTitle) ;
	
}

void TDiffractionDataTable::SymmetrizeHKLDataSet(TDiffractionDataTable *s,char *symm,Boolean testSigma,HKLErrorFlag errorType) {
	
	long	n,numOldReflections,numNewReflections,m,h,k,i,j,klow,khigh,ne,i1,i2 ;
	short	groupIndex = 1 ;

	double	s1,t,val ; //,e,s2,sm ;

	TArray	*data ; 	// hkl data set 2D  nb = 80 data(-nb:nb,-nb:nb) ;
	//TArray	*err ; 		// hkl data set errors 2D  nb = 80 data(-nb:nb,-nb:nb) ;
	//float	*error ;	// This is the space for the new errors
	long	*ngot ; 	//This is space for how many reflections were found for a [hkl] set 

	long	he[20],ke[20],ise[20] ;	
	float	wrk1[40] ;
	//float	wrk2[40] ;

	bool	listMinusGReflections = false ;
	
	float	astar,bstar,gammastar,csgammastar,sigammastar,gsquare ;
	switch(mPlaneGroup) {
		case 1:	// p1
			//if (!friedelSymmetry) listMinusGReflections = true ;
		case 2: // pg
			mGamma = s->mGamma ;
			gammastar = 180-mGamma ;
			csgammastar = cos(gammastar*asin(1.0)/90.) ;
			sigammastar = sin(gammastar*asin(1.0)/90.) ;
			astar = 1./(s->mA*sigammastar) ;
			bstar = 1./(s->mB*sigammastar) ;
			break ;
		case 3: // pm 
		case 4: // p11m 
		case 5: // pg 
		case 6: // p11g
		case 7: // cm  
		case 8: // c11m 
		case 9: // p2mm  
		case 10: // p2mg  
		case 11: // p2gm
		case 12: // pgg
		case 13: // cmm
			mA	= s->mA ;
			mB	= s->mB ;
			astar = 1./mA ;
			bstar = 1./mB ;
			mGamma = 90.0 ;
			gammastar = 90. ;
			csgammastar = 0. ;
			sigammastar = 1.0 ;
			break ;
		case 14: // p4
		case 15: // p4mm
		case 16: // p4g
		case 17: // p4mg
			mA =  (s->mA+s->mB)/2 ;
			mB =  mA ;
			astar = 1./mA ;
			bstar = 1./mB ;
			gammastar = 90. ;
			mGamma = 90.0 ;
			csgammastar = 0. ;
			sigammastar = 1.0 ;
			break ;
		case 18: // p3
		case 19: // p3m1
		case 20: // p31m
			//if (!friedelSymmetry) listMinusGReflections = true ;
		case 21: // p6
		case 22: // p6mm
			// Gamma = 120
			mA =  (s->mA+s->mB)/2 ;
			mB =  mA ;
			gammastar = 60. ;
			mGamma = 120.0 ;
			csgammastar = cos(60.0*asin(1.0)/90.) ;
			sigammastar = sin(60.0*asin(1.0)/90.) ;
			astar = 1./(mA*sigammastar) ;	
			bstar = 1./(mB*sigammastar) ;	
			break ;
	}
		
	float	astarSq = astar*astar ;
	float	bstarSq = bstar*bstar ;

	numOldReflections	=	s->mNumReflections ;
	//fprintf(stderr,"Symm %s Group %d \n",symm,mPlaneGroup);
	
	// Check the limits on the possible array dimensions
	long	hMax = -100, kMax = -100, ival ;
	for (i = 0 ;  i < numOldReflections; i++) {
		ival = iabs(s->mH[i]) ;
		if(ival < 0 ) ival = -ival ;
		if (hMax < ival) hMax = ival ;
		ival = iabs(s->mK[i]) ;
		if(ival < 0 ) ival = -ival ;
		if (kMax < ival) kMax = ival ;		
	}
		
	long	ndim = 2*(hMax+kMax+2) ;	// Make sure we have a large enough data set for all indicies
	data = new TArray(ndim,ndim) ;
	//err = new TArray(ndim,ndim) ;
	
	long maxIndex = ndim/2-2 ;
	
	*data = -1 ;	// Initialize to all not set
	//*err = 1 ;		// Currently no errors		errorType = kNoHKLError
	
	for (i = 0 ;  i < s->mNumReflections; i++) {
		h = s->mH[i] ;
		k = s->mK[i] ;
		val = s->mAmplitude[i] ;
		*data->GetDataHK(h,k) = val ;
		//fprintf(stderr,"Input %d %d %d %f \n",i,h,k,val);
		//fprintf(stderr,"\nSetting [%d,%d] = %f  G = %f",h,k,val,s->mGVector[i]) ;
		//if (errorType == kErrorSquareRootAmplitude)
		//	*err->GetDataHK(h,k) = sqrt(val) ;
	}	

	numNewReflections = 0 ;	
	// Loop over over "diffraction data"
	for(j = 0 ; j < maxIndex; j++) {
		//	Half plane, p1, p2	 ; 
		klow	= -maxIndex ; 
		khigh	= maxIndex ; 
		//	Limits for the second index ; 
		if(strcmp(symm,"p6mm") == 0 || strcmp(symm,"p3m1") == 0 ||  strcmp(symm,"p31m") == 0 ||  
						strncmp(symm,"p4m",3) == 0 || strncmp(symm,"p4g",3) == 0 ) {
			klow	= 0 ; 
			khigh	= j	 ; 
		}
		else if (strncmp(symm,"pm",2) == 0 || strncmp(symm,"p2m",3) == 0 || strncmp(symm,"pg",2) == 0 || 
						strncmp(symm,"cm",2) == 0 || strncmp(symm,"p2g",3) == 0 || strcmp(symm,"c2mm") == 0 ||
						strncmp(symm,"p4",2) == 0 || strncmp(symm,"p3",2) == 0 || strncmp(symm,"p6",2) == 0) { 
			klow	= 0 ; 
			khigh	= maxIndex ; 
		}
		//fprintf(stderr,"Loop %d %d %d \n",j,klow,khigh);
		for(k = klow ; k <= khigh ; k++) {
			//if (*data->GetDataHK(j,k) > -0.5) { 
				//	Find equivalent reflections ; 
				i1	= j ; 
				i2	= k ; 
				equiv2(he,ke,ise,i1,i2,ne,symm) ; 
				m	= 0 ; 
				s1	= 0 ; 
				//s2	= 0 ; 
				//sm	= 0 ; 
				//fprintf(stderr,"\nNumber Equivalent Reflections for [%d,%d] = %d",i1,i2,ne) ;
				for(n=0; n < ne; n++) { 
					//fprintf(stderr,"[%d,%d]  ",he[n],ke[n]) ;
					t	= *data->GetDataHK(he[n],ke[n]) ; 
					//e	= *err->GetDataHK(he[n],ke[n]) ; 
					if(t > -0.5) {
						s1	= s1 + t ; // /e ; 
						//s2	= s2 + 1.0/e ; 
						*data->GetDataHK(he[n],ke[n])	= -1 ; 
						wrk1[m]	= t ; 
						//wrk2[m]	= e ; 
						m++ ;
					} 
				}
				if ( m > 0 ) {
				val		=	s1/m ; //s2 ; 
				//fprintf(stderr,"\nSymmetrized Amplitude for [%d,%d] = %f",i1,i2,val) ;
				mAmplitude[numNewReflections]	= val ;
				//LDM: This should be replaced by a bi-weight average if m > 3
				
				// -------------------------------------------
				// I don't currently set off space for the errors, but
				// this is where they get set
				// -------------------------------------------
				/*
				// Note: sqrt 2 is about right (?) for robust (or was it 1.8) ; 
				er	= sqrt(double(m))/s2/.80231 ; 
				error[numNewReflections]	= er ; 
				// Chi trap ; 
				if(m > 1) {
					s1=0 ; 
					elarge = -999 ; 
					for(i=0; i < m; i++) {
						s1 = s1+abs(val-wrk1[i])/er ; 
						elarge = max(abs(val-wrk1[i])/er,elarge) ; 
					} 
					s2 = max(1.0,s1/m) ; 
					if(testSigma) s2 = max(s2,elarge/3.0) ;
					error[numNewReflections] = error[numNewReflections]*s2 ; 
				}
				*/
				gsquare = j*j*astarSq + k*k*bstarSq + 2.0*j*k*astar*bstar*csgammastar ;
				mGVector[numNewReflections] = sqrt(gsquare) ;
				mH[numNewReflections]	= j ; 
				mK[numNewReflections]	= k ; 
				mIndex[numNewReflections] = groupIndex ;
				// ngot[numNewReflections]	= m ; 
				numNewReflections++ ; 
				groupIndex++ ;
				//fprintf(stderr,"Averaged %d %d %d %f \n",j,k,m,val);
				}
				//else
				//	fprintf(stderr,"Nothing for %d %d \n",j,k);
		}
	} 
	mNumReflections = numNewReflections ;
	
	delete data ;
	//delete err ;
}


TDiffractionDataTable::TDiffractionDataTable(char *title,TDiffractionPeaks *diffractionPeaks,short planeGroup,bool friedelSymmetry) 
	: TTable(title) {
	short 	h,k,hDataMax,kDataMax,hStart,groupIndex ; //,n,m,imageWidth,imageHeight ;
	float	amplitude,intensity,astar,bstar,gammastar,twocsgastar,astarSq,bstarSq,ksqbstsq,
			astbstcsgast,kbst2csgast,csgastar ; //,x,y,a,b ;
	long	i ; //,ih ;
	
	TArray	*dataArray = diffractionPeaks->mFourierData ;
	hDataMax = dataArray->GetWidth()/2 -1 ;
	kDataMax = dataArray->GetHeight()/2 -1 ;
	
	mRefinementSet = nil ;
	
	mPlaneGroup			= planeGroup ;
	mUseAllReflections	= true ;
	mUseAllPhases		= false ;
	mAllUnknown			= true ;

	bool	listMinusGReflections = false ;
	switch(planeGroup) {
		case 1:	// p1
			if (!friedelSymmetry) listMinusGReflections = true ;
		case 2: // p2
			astar = diffractionPeaks->mAStar ;
			bstar = diffractionPeaks->mBStar ;
			gammastar = diffractionPeaks->mGammaStar ;
			mA =  diffractionPeaks->mA ;
			mB =  diffractionPeaks->mB ;
			mGamma = 180. - gammastar ;
			break ;
		case 3: // pm 
		case 4: // p11m 
		case 5: // pg 
		case 6: // p11g
		case 7: // cm  
		case 8: // c11m 
		case 9: // p2mm  
		case 10: // p2mg  
		case 11: // p2gm
		case 12: // pgg
		case 13: // cmm
			astar = diffractionPeaks->mAStar ;
			bstar = diffractionPeaks->mBStar ;
			mA =  diffractionPeaks->mA ;
			mB =  diffractionPeaks->mB ;
			gammastar = 90. ;
			mGamma = 90.0 ;
			break ;
		case 14: // p4
		case 15: // p4mm
		case 16: // p4g
		case 17: // p4mg
			astar = (diffractionPeaks->mAStar+diffractionPeaks->mBStar)/2. ;
			bstar = astar ;
			gammastar = 90. ;
			mA =  1./astar ;
			mB =  mA ;
			mGamma = 90.0 ;
			break ;
		case 18: // p3
		case 19: // p3m1
		case 20: // p31m
			if (!friedelSymmetry) listMinusGReflections = true ;
		case 21: // p6
		case 22: // p6mm
			// Gamma = 120
			astar = (diffractionPeaks->mAStar+diffractionPeaks->mBStar)/2. ;
			bstar = astar ;
			gammastar = 60. ;
			mA =  1./(astar*sin(60.0*asin(1.0)/90.)) ;
			mB =  mA ;
			mGamma = 120.0 ;
			break ;
	}
	
	// First find out how many diffraction spots we have
	mNumReflections = 0 ;
	if (!listMinusGReflections) {
		for(k = 0; k < kDataMax; k++)  {
			hStart = k == 0 ? 1 : -hDataMax ;
			for(h = hStart; h < hDataMax; h++) {
				if (!(h == 0 && k == 0)) {
					intensity = *dataArray->GetRealHK(h,k) ;
					if (intensity > 0.) mNumReflections++ ;
				}
			}
		}
	}
	else {
		for(k = -kDataMax; k < kDataMax; k++)  {
			for(h = -hDataMax; h < hDataMax; h++) {
				if (!(h == 0 && k == 0)) {
					intensity = *dataArray->GetRealHK(h,k) ;
					if (intensity > 0.) mNumReflections++ ;
				}
			}
		}
	}
	mGVector 	= new float[mNumReflections] ;
	mAmplitude 	= new float[mNumReflections] ;
	mPhase 		= new float[mNumReflections] ;
	mH		= new short[mNumReflections] ;
	mK		= new short[mNumReflections] ;
	mIndex		= new short[mNumReflections] ;
	mIncludeReflection	= new bool[mNumReflections] ;
	mUsePhase	= new bool[mNumReflections] ;
	
	for (i=0; i < mNumReflections;  mIndex[i++] = 0.)  ;				
	
	astarSq = astar*astar ;
	bstarSq = bstar*bstar ;
	csgastar = cos(gammastar*asin(1.0)/90.) ;
	twocsgastar = 2.*cos(gammastar*asin(1.0)/90.) ;

	mCalibrationUnit = diffractionPeaks->mCalibrationUnit ;	
	
	i = 0 ;
	
	TArray	*IsSet = new TArray(dataArray->GetWidth(),dataArray->GetHeight()) ;
	*IsSet = 0. ;
	
	groupIndex = 1 ;
	astbstcsgast = astar*bstar*twocsgastar ;
	// Do only upper half plane
	bool	isSymmetryRelated ;
	if (!listMinusGReflections) {
		for(k = 0; k < kDataMax; k++) {
			ksqbstsq = k*k*bstarSq ;
			kbst2csgast = k*astbstcsgast ;
			hStart = k == 0 ? 1 : -hDataMax ;
			for(h = hStart; h < hDataMax; h++) {
				if (!(h == 0 && k == 0)) {
					amplitude = ::sqrt(*dataArray->GetRealHK(h,k)) ;
					if (amplitude > 0.) {
						mGVector[i] = ::sqrt(h*h*astarSq + ksqbstsq + h*kbst2csgast) ;
						mAmplitude[i] = amplitude ;
						mPhase[i] = kUnAssignedPhase ;
						mH[i] = h ;
						mK[i] = k ;
						mIncludeReflection[i] = true ;
						mUsePhase[i] = false ;
						if (*IsSet->GetDataHK(h,k) == 0.) {
							// We first need to check to see if the reflection is symmetry related
							// to one that has been set, but which did not set this one
							isSymmetryRelated = false ;
							switch(mPlaneGroup) {
								case 1:	// p1
								case 2: // p2
									break ;
								case 3: // pm		|F(h,k)| = |F(-h,k)|
								case 4: // p11m		|F(h,k)| = |F(-h,k)|
								case 5: // pg		|F(h,k)| = |F(-h,k)|
								case 6: // p11g		|F(h,k)| = |F(-h,k)|
								case 7: // cm		|F(h,k)| = |F(-h,k)|  
								case 8: // c11m		|F(h,k)| = |F(-h,k)| 
								case 9: // p2mm		|F(h,k)| = |F(-h,k)|  
								case 10: // p2mg	|F(h,k)| = |F(-h,k)|  
								case 11: // p2gm	|F(h,k)| = |F(-h,k)|
								case 12: // pgg		|F(h,k)| = |F(-h,k)|
								case 13: // cmm		|F(h,k)| = |F(-h,k)|
									if (*IsSet->GetDataHK(-h,k) != 0.)  { isSymmetryRelated = true ; mIndex[i] = *IsSet->GetDataHK(-h,k) ; }
									break ;
								case 14: // p4	|F(h,k)| = |F(-k,h)|
									if (*IsSet->GetDataHK(-k,h) != 0.)  { isSymmetryRelated = true ; mIndex[i] = *IsSet->GetDataHK(-k,h) ; }
									break ;
								case 15: // p4mm 	|F(h,k)| = |F(-k,h)| = |F(-h,k)|
								case 16: // p4g		|F(h,k)| = |F(-k,h)| = |F(-h,k)|
								case 17: // p4mg 	|F(h,k)| = |F(-k,h)| = |F(-h,k)|
									if (*IsSet->GetDataHK(-k,h) != 0.)  { isSymmetryRelated = true ; mIndex[i] = *IsSet->GetDataHK(-k,h) ; }
									if (*IsSet->GetDataHK(-h,k) != 0.)  { isSymmetryRelated = true ; mIndex[i] = *IsSet->GetDataHK(-h,k) ; }
									break ;
								case 18: // p3	|F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)|
									if (*IsSet->GetDataHK(k,-h-k) != 0.)  { isSymmetryRelated = true ; mIndex[i] = *IsSet->GetDataHK(k,-h-k) ; }
									if (*IsSet->GetDataHK(-h-k,h) != 0.)  { isSymmetryRelated = true ; mIndex[i] = *IsSet->GetDataHK(-h-k,h) ; }
									break ;
								case 19: // p3m1	|F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)| = |F(k,h)|
								case 20: // p31m	|F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)| = |F(k,h)|
								case 22: // p6m		|F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)| = |F(k,h)|
									if (*IsSet->GetDataHK(k,-h-k) != 0.)  { isSymmetryRelated = true ; mIndex[i] = *IsSet->GetDataHK(k,-h-k) ; }
									if (*IsSet->GetDataHK(-h-k,h) != 0.)  { isSymmetryRelated = true ; mIndex[i] = *IsSet->GetDataHK(-h-k,h) ; }
									if (*IsSet->GetDataHK(k,h) != 0.)  	  { isSymmetryRelated = true ; mIndex[i] = *IsSet->GetDataHK(k,h) ; }
									break ;
								case 21: // p6	|F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)|
									if (*IsSet->GetDataHK(k,-h-k) != 0.)  { isSymmetryRelated = true ; mIndex[i] = *IsSet->GetDataHK(k,-h-k) ; }
									if (*IsSet->GetDataHK(-h-k,h) != 0.)  { isSymmetryRelated = true ; mIndex[i] = *IsSet->GetDataHK(-h-k,h) ; }
									break ;
							}
							
							if (isSymmetryRelated) {
								*IsSet->GetDataHK(h,k) = mIndex[i] ;
								*IsSet->GetDataHK(-h,-k) = mIndex[i] ;
							}
							else {
								mIndex[i] = groupIndex ;
								*IsSet->GetDataHK(h,k) = (float)groupIndex ;
								// Set symmetry related reflection
								switch(mPlaneGroup) {
									case 1:	// p1
									case 2: // p2
										break ;
									case 3: // pm		|F(h,k)| = |F(-h,k)|
									case 4: // p11m		|F(h,k)| = |F(-h,k)|
									case 5: // pg		|F(h,k)| = |F(-h,k)|
									case 6: // p11g		|F(h,k)| = |F(-h,k)|
									case 7: // cm		|F(h,k)| = |F(-h,k)|  
									case 8: // c11m		|F(h,k)| = |F(-h,k)| 
									case 9: // p2mm		|F(h,k)| = |F(-h,k)|  
									case 10: // p2mg	|F(h,k)| = |F(-h,k)|  
									case 11: // p2gm	|F(h,k)| = |F(-h,k)|
									case 12: // pgg		|F(h,k)| = |F(-h,k)|
									case 13: // cmm		|F(h,k)| = |F(-h,k)|
										*IsSet->GetDataHK(-h,k) = groupIndex ;
										*IsSet->GetDataHK(-h,-k) = groupIndex ;
										*IsSet->GetDataHK(h,-k) = groupIndex ;
										break ;
									case 14: // p4	|F(h,k)| = |F(-k,h)|
										*IsSet->GetDataHK(-h,-k) = groupIndex ;
										*IsSet->GetDataHK(-k,h) = groupIndex ;
										*IsSet->GetDataHK(k,-h) = groupIndex ;
										break ;
									case 15: // p4mm 	|F(h,k)| = |F(-k,h)| = |F(-h,k)|
									case 16: // p4g		|F(h,k)| = |F(-k,h)| = |F(-h,k)|
									case 17: // p4mg 	|F(h,k)| = |F(-k,h)| = |F(-h,k)|
										*IsSet->GetDataHK(-k,h) = groupIndex ;
										*IsSet->GetDataHK(-h,k) = groupIndex ;
										*IsSet->GetDataHK(-h,-k) = groupIndex ;
										*IsSet->GetDataHK(k,-h) = groupIndex ;
										*IsSet->GetDataHK(h,-k) = groupIndex ;
										break ;
									case 18: // p3	|F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)|
										*IsSet->GetDataHK(k,-h-k) = groupIndex ;
										*IsSet->GetDataHK(-h-k,h) = groupIndex ;
										*IsSet->GetDataHK(-h,-k) = groupIndex ;
										*IsSet->GetDataHK(-k,h+k) = groupIndex ;
										*IsSet->GetDataHK(h+k,-h) = groupIndex ;
										break ;
									case 19: // p3m1	|F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)| = |F(k,h)|
									case 20: // p31m	|F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)| = |F(k,h)|
									case 22: // p6m		|F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)| = |F(k,h)|
										*IsSet->GetDataHK(k,-h-k) = groupIndex ;
										*IsSet->GetDataHK(-h-k,h) = groupIndex ;
										*IsSet->GetDataHK(k,h) = groupIndex ;
										*IsSet->GetDataHK(-h,-k) = groupIndex ;
										*IsSet->GetDataHK(-k,h+k) = groupIndex ;
										*IsSet->GetDataHK(h+k,-h) = groupIndex ;
										*IsSet->GetDataHK(-k,-h) = groupIndex ;
										break ;
									case 21: // p6	|F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)|
										*IsSet->GetDataHK(k,-h-k) = groupIndex ;
										*IsSet->GetDataHK(-h-k,h) = groupIndex ;
										*IsSet->GetDataHK(-h,-k) = groupIndex ;
										*IsSet->GetDataHK(-k,h+k) = groupIndex ;
										*IsSet->GetDataHK(h+k,-h) = groupIndex ;
										break ;
								}
								groupIndex++ ;
							}
						}
						else {
							mIndex[i] = (short) *IsSet->GetDataHK(h,k) ;
							// We need to make sure any straggling reflections are set as well
							// Set symmetry related reflection
							switch(mPlaneGroup) {
								case 1:	// p1
								case 2: // p2
									break ;
								case 3: // pm		|F(h,k)| = |F(-h,k)|
								case 4: // p11m		|F(h,k)| = |F(-h,k)|
								case 5: // pg		|F(h,k)| = |F(-h,k)|
								case 6: // p11g		|F(h,k)| = |F(-h,k)|
								case 7: // cm		|F(h,k)| = |F(-h,k)|  
								case 8: // c11m		|F(h,k)| = |F(-h,k)| 
								case 9: // p2mm		|F(h,k)| = |F(-h,k)|  
								case 10: // p2mg	|F(h,k)| = |F(-h,k)|  
								case 11: // p2gm	|F(h,k)| = |F(-h,k)|
								case 12: // pgg		|F(h,k)| = |F(-h,k)|
								case 13: // cmm		|F(h,k)| = |F(-h,k)|
										*IsSet->GetDataHK(-h,k) = mIndex[i] ;
									*IsSet->GetDataHK(-h,-k) = mIndex[i] ;
									*IsSet->GetDataHK(h,-k) = mIndex[i] ;
									break ;
								case 14: // p4	|F(h,k)| = |F(-k,h)|
									*IsSet->GetDataHK(-h,-k) = mIndex[i] ;
									*IsSet->GetDataHK(-k,h) = mIndex[i] ;
									*IsSet->GetDataHK(k,-h) = mIndex[i] ;
									break ;
								case 15: // p4mm 	|F(h,k)| = |F(-k,h)| = |F(-h,k)|
								case 16: // p4g		|F(h,k)| = |F(-k,h)| = |F(-h,k)|
								case 17: // p4mg 	|F(h,k)| = |F(-k,h)| = |F(-h,k)|
									*IsSet->GetDataHK(-k,h) = mIndex[i] ;
									*IsSet->GetDataHK(-h,k) = mIndex[i] ;
									*IsSet->GetDataHK(-h,-k) = mIndex[i] ;
									*IsSet->GetDataHK(k,-h) = mIndex[i] ;
									*IsSet->GetDataHK(h,-k) = mIndex[i] ;
									break ;
								case 18: // p3	|F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)|
									*IsSet->GetDataHK(k,-h-k) = mIndex[i] ;
									*IsSet->GetDataHK(-h-k,h) = mIndex[i] ;
									*IsSet->GetDataHK(-h,-k) = mIndex[i] ;
									*IsSet->GetDataHK(-k,h+k) = mIndex[i] ;
									*IsSet->GetDataHK(h+k,-h) = mIndex[i] ;
									break ;
								case 19: // p3m1	|F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)| = |F(k,h)|
								case 20: // p31m	|F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)| = |F(k,h)|
								case 22: // p6m		|F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)| = |F(k,h)|
									*IsSet->GetDataHK(k,-h-k) = mIndex[i] ;
									*IsSet->GetDataHK(-h-k,h) = mIndex[i] ;
									*IsSet->GetDataHK(k,h) = mIndex[i] ;
									*IsSet->GetDataHK(-h,-k) = mIndex[i] ;
									*IsSet->GetDataHK(-k,h+k) = mIndex[i] ;
									*IsSet->GetDataHK(h+k,-h) = mIndex[i] ;
									*IsSet->GetDataHK(-k,-h) = mIndex[i] ;
									break ;
								case 21: // p6	|F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)|
									*IsSet->GetDataHK(k,-h-k) = mIndex[i] ;
									*IsSet->GetDataHK(-h-k,h) = mIndex[i] ;
									*IsSet->GetDataHK(-h,-k) = mIndex[i] ;
									*IsSet->GetDataHK(-k,h+k) = mIndex[i] ;
									*IsSet->GetDataHK(h+k,-h) = mIndex[i] ;
									break ;
							}
						}
						i++ ;
					}
				}
			}
		}
	}
	else {
		// Only true for group 1,13(18),14(19),15(20)
		for(k = -kDataMax; k < kDataMax; k++) {
			ksqbstsq = k*k*bstarSq ;
			kbst2csgast = k*astbstcsgast ;
			for(h = -hDataMax; h < hDataMax; h++) {
				if (!(h == 0 && k == 0)) {
					amplitude = ::sqrt(*dataArray->GetRealHK(h,k)) ;
					if (amplitude > 0.) {
						mGVector[i] = ::sqrt(h*h*astarSq + ksqbstsq + h*kbst2csgast) ;
						mAmplitude[i] = amplitude ;
						mPhase[i] = kUnAssignedPhase ;
						mH[i] = h ;
						mK[i] = k ;
						mIncludeReflection[i] = true ;
						mUsePhase[i] = false ;
						if (*IsSet->GetDataHK(h,k) == 0.) {
							// We first need to check to see if the reflection is symmetry related
							// to one that has been set, but which did not set this one
							isSymmetryRelated = false ;
							switch(mPlaneGroup) {
								case 1:	// p1
									break ;
								case 18: // p3	|F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)|
									if (*IsSet->GetDataHK(k,-h-k) != 0.)  { isSymmetryRelated = true ; mIndex[i] = *IsSet->GetDataHK(k,-h-k) ; }
									if (*IsSet->GetDataHK(-h-k,h) != 0.)  { isSymmetryRelated = true ; mIndex[i] = *IsSet->GetDataHK(-h-k,h) ; }
									break ;
								case 19: // p3m1	|F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)| = |F(k,h)|
								case 20: // p31m	|F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)| = |F(k,h)|
									if (*IsSet->GetDataHK(k,-h-k) != 0.)  { isSymmetryRelated = true ; mIndex[i] = *IsSet->GetDataHK(k,-h-k) ; }
									if (*IsSet->GetDataHK(-h-k,h) != 0.)  { isSymmetryRelated = true ; mIndex[i] = *IsSet->GetDataHK(-h-k,h) ; }
									if (*IsSet->GetDataHK(k,h) != 0.)  	  { isSymmetryRelated = true ; mIndex[i] = *IsSet->GetDataHK(k,h) ; }
									break ;
							}
							
							if (isSymmetryRelated) {
								*IsSet->GetDataHK(h,k) = mIndex[i] ;
							}
							else {
								mIndex[i] = groupIndex ;
								*IsSet->GetDataHK(h,k) = (float)groupIndex ;
								// Set symmetry related reflection
								switch(mPlaneGroup) {
									case 1:	// p1
										break ;
									case 18: // p3	|F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)|
										*IsSet->GetDataHK(k,-h-k) = groupIndex ;
										*IsSet->GetDataHK(-h-k,h) = groupIndex ;
										break ;
									case 19: // p3m1	|F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)| = |F(k,h)|
									case 20: // p31m	|F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)| = |F(k,h)|
										*IsSet->GetDataHK(k,-h-k) = groupIndex ;
										*IsSet->GetDataHK(-h-k,h) = groupIndex ;
										*IsSet->GetDataHK(k,h) = groupIndex ;
										break ;
								}
								groupIndex++ ;
							}
						}
						else {
							mIndex[i] = (short) *IsSet->GetDataHK(h,k) ;
							// We need to make sure any straggling reflections are set as well
							// Set symmetry related reflection
							switch(mPlaneGroup) {
								case 1:	// p1
									break ;
								case 18: // p3	|F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)|
									*IsSet->GetDataHK(k,-h-k) = mIndex[i] ;
									*IsSet->GetDataHK(-h-k,h) = mIndex[i] ;
									break ;
								case 19: // p3m1	|F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)| = |F(k,h)|
								case 20: // p31m	|F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)| = |F(k,h)|
									*IsSet->GetDataHK(k,-h-k) = mIndex[i] ;
									*IsSet->GetDataHK(-h-k,h) = mIndex[i] ;
									*IsSet->GetDataHK(k,h) = mIndex[i] ;
									break ;
							}
						}
						i++ ;
					}
				}
			}
		}
	}
	delete IsSet ;
	
	mSortKey = kSortOnG ;
	SortOnG() ;


	mRows = mNumReflections ;
	mCols = 6 ;
	mCellDataSize = 4 ;
	mColWidth = 80 ;
	mRowHeight = 20 ;
		
	Int16 	xpos = gImageDisplayPosition.x ;
	Int16 	ypos = gImageDisplayPosition.y ;
	gImageDisplayPosition.x += 10 ;
	gImageDisplayPosition.y += 10 ;
				
	mParentPane = new TTableWindow(xpos,ypos,this) ;	
		
}

TDiffractionDataTable::~TDiffractionDataTable() {
	if (mGVector != nil) delete[] mGVector ;
	if (mAmplitude != nil) delete[] mAmplitude ;
	if (mPhase != nil) delete[] mPhase ;
	if (mH != nil) delete[] mH ;
	if (mK != nil) delete[] mK ;
	if (mIndex != nil) delete[] mIndex ;
	if (mIncludeReflection != nil) delete[] mIncludeReflection ; 
	if (mUsePhase != nil) delete[] mUsePhase ; 

	if (mRefinementSet != nil) {
		// Notify the refiment set that we have been deleted
		mRefinementSet->RemoveTable() ;
	}
}

void	TDiffractionDataTable::CreateImageFromDiffractionData() {
	// Put the data from the diffraction amplitudes and phases to create
	// an image with the specified number of unit cells. We are pretty free to 
	// specify the sampling (magnification) as we like as long as we include all 
	// spatial frequencies.
	
	long	i,hMax=0,kMax=0 ;
	double	degreeToRad = asin(1.0)/90. ;
	for(i=0; i < mNumReflections; i++) {
		hMax = max(hMax,mH[i]) ;
		kMax = max(kMax,mK[i]) ;
	}
	hMax = hMax/sin(degreeToRad*(180-mGamma)) + 1 ;
	kMax = kMax/sin(degreeToRad*(180-mGamma)) + 1 ;
	
	long n = 4*(log2((double)hMax) + 1) ;
	n = max(pow(2.,(double)n),32) ;
	long m = 4*(log2((double)kMax) + 1) ;
	m = max(pow(2.,(double)m),32) ;

	
	TArray *theDataArray = new TArray(n,m,kComplexArray) ;
	for(i=0; i < mNumReflections; i++) {
		*theDataArray->GetRealHK(mH[i],mK[i]) = mAmplitude[i]*cos(degreeToRad*mPhase[i]) ;
		*theDataArray->GetImagHK(mH[i],mK[i]) = mAmplitude[i]*sin(degreeToRad*mPhase[i]) ;
		*theDataArray->GetRealHK(-mH[i],-mK[i]) = *theDataArray->GetRealHK(mH[i],mK[i]) ;
		*theDataArray->GetImagHK(-mH[i],-mK[i]) = -*theDataArray->GetImagHK(mH[i],mK[i]) ;
	}
	theDataArray->ifft(false) ;
	TImage *theImage = new TImage(theDataArray,mGamma) ;
	theImage->SetSampling(mA/theImage->Width(),mB/theImage->Height()) ;
	theImage->Repeat(3,3) ;
	theImage->AdjustGamma() ;
	theImage->AdjustSampling() ;

	theImage->SetName(gPlaneGroupSymbols[mPlaneGroup-1]) ;
	theImage->Display() ;
}

void	TDiffractionDataTable::DoCommand(TCommand command,Boolean *hasBeenHandled) {	
	TRefinementSet	*refinementSet ;
	TDiffractionDataTable *newDiffractionDataTable ;
	TWindow *theWindow ;
	
	if (command > 800 && command < 825) {
		// We need to get the new symmetry
		newDiffractionDataTable = new TDiffractionDataTable(this,command-800,true) ;
		theWindow = (TWindow *)newDiffractionDataTable->GetTablePane() ;
		theWindow->SetCommander(gTheControlWindow) ;	
		AddWindow(theWindow) ;
		*hasBeenHandled = true ;
		return ;
	}
	
	switch(command) {
		case kHelp:
			OnCommandHelp() ;		
			*hasBeenHandled = true ;
			break ;
		case kSaveAsHKL:
			SaveAsHKL() ;
			*hasBeenHandled = true ;
			break ;
                case kExport01:
                case kExport02:
                case kExport03:
                case kExport04:
                case kExport05:
                        //ExportHKL(...) ;
                        *hasBeenHandled = true ;
                        break ;
		case kIFFT:
			CreateImageFromDiffractionData() ;
			*hasBeenHandled = true ;
			break ;
		case kRunFeasSet:	
			mRefinementSet = new TRefinementSet(this) ;
			mRefinementSet->RunFS98() ;
			*hasBeenHandled = true ;
			break ;
		case kCopy:
			DoCopy() ;
			*hasBeenHandled = true ;
			break ;	
		case kPaste:
			DoPaste() ;
			*hasBeenHandled = true ;
			break ;
		default:
			*hasBeenHandled = false ;
			break ;
	}
}

void	TDiffractionDataTable::DoCopy() {
	register	long i ;
	
	if (gTableClipBoardList != nil) delete gTableClipBoardList ;
	gTableClipBoardList = new TTableClipBoardList(mNumReflections) ;
	switch(mTableSelection.mTopLeftCell.col) {
		case 3:	// Amplitudes 
			gTableClipBoardList->mClipboardObjectType = 'ampl' ;
			for(i=0; i < mNumReflections; i++) {
				gTableClipBoardList->mIndex1[i] = mH[i] ;
				gTableClipBoardList->mIndex2[i] = mK[i] ;
				gTableClipBoardList->mValue[i] = mAmplitude[i] ;
			}
			break ;
		case 4: // Phases 
			gTableClipBoardList->mClipboardObjectType = 'phas' ;
			for(i=0; i < mNumReflections; i++) {
				gTableClipBoardList->mIndex1[i] = mH[i] ;
				gTableClipBoardList->mIndex2[i] = mK[i] ;
				gTableClipBoardList->mValue[i] = mPhase[i] ;
			}
			break ;
	}
	ZeroSelection() ;
}

void	TDiffractionDataTable::DoPaste() {
	// Fill in the phases from the table
	register long i,j ;
	float	scaleFactor,maxAmplitude = 0. ;	
	short	iH,iK ;
	long	II,JJ ;
	short	h,k ;	
	
	switch(mTableSelection.mTopLeftCell.col) {
		case 3:	// Amplitudes 
			
			for(j=0; j < mNumReflections; j++) {
				h = mH[j] ;
				k = mK[j] ;
				if (h != 0 && k != 0 ) {
					if (mAmplitude[j] > maxAmplitude) {
						for(i=0; i < gTableClipBoardList->mNumValues ; i++) {
							if (h == gTableClipBoardList->mIndex1[i] && k == gTableClipBoardList->mIndex2[i]) {
								maxAmplitude = mAmplitude[j] ;
								iH = h ;
								iK = k ;
								II = i ;
								JJ = j ;
								break ;
							}
						}
					}
				}
			}
				
			scaleFactor = mAmplitude[JJ]/gTableClipBoardList->mValue[II] ;
	
			for(i=0; i < mNumReflections; i++) {
				h = mH[i] ;
				k = mK[i] ;
				for(j=0; j < gTableClipBoardList->mNumValues ; j++) {
					if (gTableClipBoardList->mIndex1[j] == h && gTableClipBoardList->mIndex2[j] == k) {
						mAmplitude[i] = gTableClipBoardList->mValue[j]*scaleFactor ;
						break ;
					}
				}
			}
			break ;
		case 4: // Phases 
			mAllUnknown = false ;
			for(i=0; i < mNumReflections; i++) {
				h = mH[i] ;
				k = mK[i] ;
				for(j=0; j < gTableClipBoardList->mNumValues ; j++) {
					if (gTableClipBoardList->mIndex1[j] == h && gTableClipBoardList->mIndex2[j] == k) {
						mPhase[i] = gTableClipBoardList->mValue[j] ;
						break ;
					}
				}
			}
			// We should probably set "use phase" to false, since these are all new phases
			mUseAllPhases = false ;
			SetUseAllPhases(mUseAllPhases) ;
			Refresh() ;
			break ;
	}

	if (mCalibrationUnit != kInverseAngstrom) {		
		// We have uncalibrated data
		// Check to see if we are pasting calibrated data
		IndexCalibrationRecord	cbr ;
		if (gTableClipBoardList->HasCalibration()) {
			cbr = gTableClipBoardList->mCalibration ;
			if (mTableSelection.mTopLeftCell.col == 3 || mTableSelection.mTopLeftCell.col == 4) {
				// Set the g-vectors
				float astar = cbr.astar ;
				float bstar = cbr.bstar ;
				float csgammastar = cos(cbr.gammastar*asin(1.0)/90.) ;
				float sigammastar = sin(cbr.gammastar*asin(1.0)/90.) ;
				mA = 1./(astar*sigammastar) ;
				mB = 1./(bstar*sigammastar) ;
				mGamma = 180 - cbr.gammastar ;
				for(i=0; i < mNumReflections; i++) {
					h = mH[i] ;
					k = mK[i] ;
					mGVector[i] = sqrt(h*h*astar*astar+k*k*bstar*bstar+2*h*k*astar*bstar*csgammastar) ;
				}
			}
			Refresh() ;
			mCalibrationUnit = kInverseAngstrom ;
		}
	}
	mTableSelection.ZeroSelection() ;
	Refresh() ;
}

void	TDiffractionDataTable::GetCommandStatus(TCommand command,Boolean *isEnabled,
			Boolean *isChecked) {
			
	Boolean	aEqualB=false,gammaIs90=false,gammaIs120=false ;
	short	planeGroup ;
	
	if (command > 800 && command < 823) {
		//if (mPlaneGroup > 2) {
		//	*isEnabled = false ;
		//	return ;
		//}
		planeGroup = command - 800 ;
		// Check to see if this planegroup is consistent with the data
		if (fabs((mA-mB)/((mA+mB)/2)) < 0.03) aEqualB = true ;	// 3 percent tolerance
		if (fabs(mGamma-90.0) < 4.0) gammaIs90 = true ;
		if (fabs(mGamma-120.0) < 4.0) gammaIs120 = true ;

		if (planeGroup == 1 || planeGroup == 2) 
			*isEnabled = true ;
		else if (planeGroup > 2 && planeGroup < 18) {
			if (!gammaIs90)
				*isEnabled = false ;
			else {	// Angle is 90
				if (!aEqualB && planeGroup > 13)
					*isEnabled = false ;
				else
					*isEnabled = true ;
			}
		}
		else { // planegroup > 17  p3,p6
			if (gammaIs120)
				*isEnabled = true ;
			else
				*isEnabled = false ;
		}
		return ;
	}
	
	switch(command) {
		case kSaveAsHKL:
			*isEnabled = true ;
			break ;
                case kExport01:
                case kExport02:
                case kExport03:
                case kExport04:
                        *isEnabled = true ;
                        break ;	
		case kIFFT:
			*isEnabled = true ;
			break ;	
		case kRunFeasSet:
			// we only allow one refinement dialog to be active at one time
			// We could delete the old one and create a new one
			*isEnabled = mRefinementSet == nil && gRunDirectMethodsDialog == nil ;
			break ;	
		case kSymmHKLData:
			//*isEnabled = mPlaneGroup == 1 || mPlaneGroup == 2 ;
			*isEnabled = true ;
			break ;
		case kCopy:
			*isEnabled = HasSelection() ;
			break ;	
		case kPaste:
			*isEnabled = HasSelection() && gTableClipBoardList != nil ;
			break ;	
		case kHelp:
			*isEnabled = true ;
			break ;
		default:
			*isEnabled = false ;
			break ;
	}
}

void TDiffractionDataTable::GetCellText(TableCellT &inCell,char *str) {
	switch (inCell.col) {
		case 1:
			sprintf(str,"[%3d,%3d]",mH[inCell.row-1],mK[inCell.row-1]) ;
			break ;
		case 2:
			sprintf(str,"%.3f",mGVector[inCell.row-1]) ;
			break ;
		case 3:
			sprintf(str,"%10g",mAmplitude[inCell.row-1]) ; //LDM Was .3f
			break ;
		case 4:
		// LDM	if (mPhase[inCell.row-1] == kUnAssignedPhase) {
			if (mPhase[inCell.row-1] <= -9000.) {
				str[0] = '\0' ;				
			}
			else {
				sprintf(str,"%.2f",mPhase[inCell.row-1]) ;
			}
			break ;
		case 5:
			if (mIncludeReflection[inCell.row-1]) {
				sprintf(str,"-") ;				
			}
			else {
				sprintf(str,"Include") ;				
			}
			break ;
		case 6:
			if (mUsePhase[inCell.row-1]) {
				sprintf(str,"Variable") ;				
			}
			else {
				sprintf(str,"Fixed") ;				
			}
			break ;
			
	}
}

void TDiffractionDataTable::SetCellText(TableCellT &inCell,char *str) {
	// This actually sets the table entries according to the text
	short	h,k,groupIndex ;
	register long i ;
	float	value,calibration ;
	if (str[0] == '\0') return ;
	switch (inCell.col) {
		case 2:	// We must calibrate all the d-spacings based on the value entered
			sscanf(str,"%f",&value) ;
			// If we have symmetrized, we must set symmetry related reflections as well
			calibration = value/mGVector[inCell.row-1] ;
			mA	/= calibration ;
			mB	/= calibration ;
			mCalibrationUnit = kInverseAngstrom ;	// We hope
			for(i=0; i < mNumReflections; i++) 
				mGVector[i] *= calibration ;
			break ;
		case 3:
			sscanf(str,"%f",&value) ;
			// If we have symmetrized, we must set symmetry related reflections as well
			groupIndex = mIndex[inCell.row-1] ;
			for(i=0; i < mNumReflections; i++) 
				if (mIndex[i] == groupIndex) mAmplitude[i] = value ;
			break ;
		case 4:
			sscanf(str,"%f",&value) ;
			groupIndex = mIndex[inCell.row-1] ;
			if (str[0] == '\0' || (str[0] == '-' && str[1] == '\0')) {
				//fprintf(stderr,"\nUnassigned phase") ;
				value = kUnAssignedPhase ;
				for(i=0; i < mNumReflections; i++) {
					if (mIndex[i] == groupIndex) 
						mPhase[i] = kUnAssignedPhase ;
				}
			}
			else {
				sscanf(str,"%f",&value) ;
				mAllUnknown = false ;

				//fprintf(stderr,"\nphase %f",value) ;
				// If we have symmetrized, we must set symmetry related reflections as well
				h = mH[inCell.row-1] ;
				k = mK[inCell.row-1] ;
				for(i=0; i < mNumReflections; i++) {
					if (mIndex[i] == groupIndex) 
						mPhase[i] = SymmetryRelatedPhase( h,k,mH[i],mK[i],value,mPlaneGroup ) ;
				}
			}
			break ;
	}
}

void	TDiffractionDataTable::ClickColLabel(Int16 inCol) {
	TableCellT	firstSelectedCell,lastSelectedCell ;
	switch (inCol) {
		case 2:		// Sort on g-vector
			if (mSortKey != kSortOnG) {
				SortOnG() ;
				mSortKey = kSortOnG ;
				Refresh() ;
			}
			break ;
		case 3:		// Sort on amplitude
			if (!ControlKeyDown()) {
				if (mSortKey != kSortOnAmplitude) {
					SortOnAmpitude() ;
					mSortKey = kSortOnAmplitude ;
					Refresh() ;
				}
			}
			else {
				// Select the whole column fo copying/pasting
				if (!HasSelection()) {
					firstSelectedCell.col = inCol ;
					lastSelectedCell.col = inCol ;
					firstSelectedCell.row = 1 ;
					lastSelectedCell.row = mRows ;
					Select(firstSelectedCell,lastSelectedCell) ;
				} 
				else {
					ZeroSelection() ;
				}
			}
			break ;
		case 4:
			// Select the whole column fo copying/pasting
			if (!HasSelection()) {
				firstSelectedCell.col = inCol ;
				lastSelectedCell.col = inCol ;
				firstSelectedCell.row = 1 ;
				lastSelectedCell.row = mRows ;
				Select(firstSelectedCell,lastSelectedCell) ;
			}
			else {
				ZeroSelection() ;
			}
			break ;
		case 5:
			if (mUseAllReflections) {
				mUseAllReflections = false ;
				SetIncludeAllReflections(mUseAllReflections) ;
				Refresh() ;
			}
			else {
				mUseAllReflections = true ;
				SetIncludeAllReflections(mUseAllReflections) ;
				Refresh() ;
			}
			break ;
		case 6:
			if (mUseAllPhases) {
				mUseAllPhases = false ;
				SetUseAllPhases(mUseAllPhases) ;
				Refresh() ;
			}
			else {
				mUseAllPhases = true ;
				SetUseAllPhases(mUseAllPhases) ;
				Refresh() ;
			}
			break ;
		default:
			break ;
	}
}


void	TDiffractionDataTable::Save() {
	// ------------------------------------------------------------------ 
	// Put up the standard file dialog Box to determine the file name 
	// ------------------------------------------------------------------ 
	FSInfo 	theFInfo ;
	
	strcpy(theFInfo.mFileName,mTitle) ;
	strcat(theFInfo.mFileName,".txt") ;
	ChooseSaveFile("Save Diffraction Table data as:",theFInfo) ;
	
	if (!theFInfo.mReply) 
		return ;

	char buffer[256] ;
	int i ;
	long inByteCount ;
	char str[64] ;
	TableCellT theCell ;
	FILE	*fp ;

	if ((fp = fopen(theFInfo.mFileName,"w")) == nil) {
		sprintf(buffer,"Unable to open the file %s",theFInfo.mFileName) ;
		PutMessage(buffer) ;
		return ;
	}
	
	
	for(i=1;i<=mCols;i++) {
		switch (i) {
			case 1:
				sprintf(buffer,"h\tk\t") ;
				break ;
			case 2:
				sprintf(buffer,"u\t") ;
				break ;
			case 3:
				sprintf(buffer,"     Amp\t") ;
				break ;
			case 4:
				sprintf(buffer,"Phase\t") ;
				break ;
			case 5:
				sprintf(buffer,"Include\t") ;
				break ;
			case 6:
  				sprintf(buffer,"Use Phase\n") ;
				break ;
		}
		inByteCount = strlen(buffer) ;
		fwrite(buffer,inByteCount,1,fp) ;
	}
	
	for(i=1;i<=mRows;i++) {
		theCell.row = i ;
		for(theCell.col=1;theCell.col<=mCols;theCell.col++) {
			switch (theCell.col) {
				case 1:
					sprintf(buffer,"%d\t%d\t",mH[theCell.row-1],mK[theCell.row-1]) ;
					break ;
				case 2:
					sprintf(buffer,"%.2f\t",mGVector[theCell.row-1]) ;
					break ;
				case 3:
				// Change here, LDM
					sprintf(buffer,"%10g\t",mAmplitude[theCell.row-1]) ;
					break ;
				case 4:
					// Change here, LDM
					if (mPhase[theCell.row-1] <= -9000.) {
						sprintf(buffer,"  0\t") ; // ,mPhase[theCell.row-1]) ;
					}
					else {
						sprintf(buffer,"%.2f\t",mPhase[theCell.row-1]) ;
					}
					break ;
				case 5:
					if (mIncludeReflection[theCell.row-1]) {
						sprintf(buffer,"  Yes\t") ; // ,mPhase[theCell.row-1]) ;					
					}
					else {
						sprintf(buffer,"   No\t") ; // ,mPhase[theCell.row-1]) ;					
					}
					break ;
				case 6:
			       if (mUsePhase[theCell.row-1]) {
					   sprintf(buffer,"  Yes\n") ; // ,mPhase[theCell.row-1]) ;					
					}
					else {
						sprintf(buffer,"   No\n") ; // ,mPhase[theCell.row-1]) ;					
					}
					break ;
			}
			inByteCount = strlen(buffer) ;
			fwrite(buffer,inByteCount,1,fp) ;
		}
	}
	fclose(fp) ;
}

void	TDiffractionDataTable::SaveAsHKL() {
	// ------------------------------------------------------------------ 
	// Put up the standard file dialog Box to determine the file name 
	// ------------------------------------------------------------------ 
	FSInfo 	theFInfo ;
	
	strcpy(theFInfo.mFileName,mTitle) ;
	strcat(theFInfo.mFileName,".hkl") ;
	ChooseSaveFile("Save Diffraction Table data as:",theFInfo) ;
	
	if (!theFInfo.mReply) 
		return ;

	char buffer[256] ;
	int i ;
	long inByteCount ;
	char str[64] ;
	TableCellT theCell ;
	FILE	*fp ;

	if ((fp = fopen(theFInfo.mFileName,"w")) == nil) {
		sprintf(buffer,"Unable to open the file %s",theFInfo.mFileName) ;
		PutMessage(buffer) ;
		return ;
	}
	
	
	for(i=1;i<=mCols;i++) {
		switch (i) {
			case 1:
				sprintf(buffer,"h\tk\t") ;
				break ;
			case 2:
				sprintf(buffer,"l\t") ; //LDM sprintf(buffer,"u\t") 
				break ;
			case 3:
				sprintf(buffer,"     Amp\t") ;
				break ;
			case 4:
				sprintf(buffer,"Phase\t") ;
				break ;
			case 5:
				sprintf(buffer,"Include\t") ;
				break ;
			case 6:
  				sprintf(buffer,"Use Phase\n") ;
				break ;
		}
		inByteCount = strlen(buffer) ;
		fwrite(buffer,inByteCount,1,fp) ;
	}
	
	for(i=1;i<=mRows;i++) {
		theCell.row = i ;
		for(theCell.col=1;theCell.col<=mCols;theCell.col++) {
			switch (theCell.col) {
				case 1:
					sprintf(buffer,"%d\t%d\t",mH[theCell.row-1],mK[theCell.row-1]) ;
					break ;
				case 2:
//LDM					sprintf(buffer,"%.2f\t",mGVector[theCell.row-1]) ;
					sprintf(buffer,"0\t") ;
					break ;
				case 3:
				// Change here, LDM
					sprintf(buffer,"%10g\t",mAmplitude[theCell.row-1]) ;
					break ;
				case 4:
					// Change here, LDM
					if (mPhase[theCell.row-1] <= -9000.) {
						sprintf(buffer,"  0\t") ; // ,mPhase[theCell.row-1]) ;
					}
					else {
						sprintf(buffer,"%.2f\t",mPhase[theCell.row-1]) ;
					}
					break ;
				case 5:
					if (mIncludeReflection[theCell.row-1]) {
						sprintf(buffer,"  Yes\t") ; // ,mPhase[theCell.row-1]) ;					
					}
					else {
						sprintf(buffer,"   No\t") ; // ,mPhase[theCell.row-1]) ;					
					}
					break ;
				case 6:
			       if (mUsePhase[theCell.row-1]) {
						sprintf(buffer,"  Yes\n") ; // ,mPhase[theCell.row-1]) ;					
					}
					else {
						sprintf(buffer,"   No\n") ; // ,mPhase[theCell.row-1]) ;					
					}
					break ;
			}
			inByteCount = strlen(buffer) ;
			fwrite(buffer,inByteCount,1,fp) ;
		}
	}
	fclose(fp) ;
}

void TDiffractionDataTable::DrawRowLabelText(Int16 rowNumber) {
	char str[64] ;
	sprintf(str,"%d",rowNumber) ;
	DrawString(str) ;
}

void TDiffractionDataTable::ClickCell(
	TableCellT &inCell) 
{
	if ((mClickTime-mLastClick) < kDoubleClickTime && (inCell.col == 4 || inCell.col == 2)) {
		EditCell(inCell) ;
		return ;
	}	
	if (inCell.col == 5) {
		bool	status ;
		status = mIncludeReflection[inCell.row-1] ? false : true ;
		long	index = mIndex[inCell.row-1] ;
		for(long i=0; i < mNumReflections; i++) {
			if (mIndex[i] == index) {
				 mIncludeReflection[i] = status ;
			}
		}
		Refresh() ;
	}
	if (inCell.col == 6) {
		bool	status ;
		status = mUsePhase[inCell.row-1] ? false : true ;
		long	index = mIndex[inCell.row-1] ;
		for(long i=0; i < mNumReflections; i++) {
			if (mIndex[i] == index) {
				 mUsePhase[i] = status ;
			}
		}
		if (status) mAllUnknown	= false ;
		Refresh() ;
	}
	
}

void TDiffractionDataTable::EditCell(
	TableCellT &inCell) 
{
	TTable::EditCell(inCell) ;
}

void TDiffractionDataTable::DrawColLabel(
	Int16 colNumber) 
{
	Rect	cellFrame;
	Rect	theSortRect  = {0,0,6,6} ;
	FPoint	imageOrigin = mParentPane->GetPortOrigin() ;
	
	if (mHasRowLabels)
		SetRect(&cellFrame,colNumber*mColWidth,0,(colNumber+1)*mColWidth,mRowHeight) ;
	else
		SetRect(&cellFrame,(colNumber-1)*mColWidth,0,colNumber*mColWidth,mRowHeight) ;
	if (colNumber != 0) 
		::OffsetRect(&cellFrame,near(imageOrigin.x),0) ;
	InsetRect(&cellFrame,1,1) ;
	SetFgColor(gGrayPixel) ;
	PaintRect(&cellFrame) ;
	SetFgColor(gDarkerGray) ;
	MoveTo(cellFrame.left,cellFrame.bottom) ;
	LineTo(cellFrame.left,cellFrame.top) ;
	LineTo(cellFrame.right,cellFrame.top) ;
	SetFgColor(gBlackPixel) ;
	MoveTo(cellFrame.right,cellFrame.top) ;
	LineTo(cellFrame.right,cellFrame.bottom) ;
	LineTo(cellFrame.left,cellFrame.bottom) ;
	InsetRect(&cellFrame,-1,-1) ;
	TextSize(0);
	MoveTo(cellFrame.left + 4, cellFrame.bottom - 4);
	if (colNumber > 0) {
		DrawColLabelText(colNumber) ;
		switch (colNumber) {
			case 2:
				if (mSortKey == kSortOnG) {
					OffsetRect(&theSortRect,cellFrame.right - 10,cellFrame.bottom - 10) ;
					PaintRect(&theSortRect) ;					
				}
				break ;
			case 3:
				if (mSortKey == kSortOnAmplitude) {
					OffsetRect(&theSortRect,cellFrame.right - 10,cellFrame.bottom - 10) ;
					PaintRect(&theSortRect) ;					
				}
				break ;
			case 5:
				OffsetRect(&theSortRect,cellFrame.right - 20,cellFrame.bottom - 10) ;
				MoveTo((theSortRect.left+theSortRect.right)/2 - 4,(theSortRect.top+theSortRect.bottom)/2-2) ;
				LineTo((theSortRect.left+theSortRect.right)/2,(theSortRect.top+theSortRect.bottom)/2+2) ;					
				LineTo((theSortRect.left+theSortRect.right)/2 + 4,(theSortRect.top+theSortRect.bottom)/2-6) ;
			case 6:
				OffsetRect(&theSortRect,cellFrame.right - 20,cellFrame.bottom - 10) ;
				MoveTo((theSortRect.left+theSortRect.right)/2 - 4,(theSortRect.top+theSortRect.bottom)/2-2) ;
				LineTo((theSortRect.left+theSortRect.right)/2,(theSortRect.top+theSortRect.bottom)/2+2) ;					
				LineTo((theSortRect.left+theSortRect.right)/2 + 4,(theSortRect.top+theSortRect.bottom)/2-6) ;
			default :
				break ;
		}
	}
}

void TDiffractionDataTable::DrawColLabelText(Int16 colNumber) {
	switch (colNumber) {
		case 1:
			DrawString("[h,k]") ;
			break ;
		case 2:
			DrawString("d-star") ;
			break ;
		case 3:
			DrawString("Amplitude") ;
			break ;
		case 4:
			DrawString("Phase") ;
			break ;
		case 5:
			DrawString("Include") ;
			break ;
		case 6:
			DrawString("Use Phase") ;
			break ;
	}
}

void TDiffractionDataTable::DrawCellValue(
	TableCellT &inCell,
	Rect &cellFrame)
{
	char str[64] ;
	
	Rect	theRect = cellFrame ;
	InsetRect(&theRect,1,1) ;
	if (mIndex[inCell.row-1]%2 == 0) {
		SetFgColor(gGrayPixel) ;
		PaintRect(&theRect) ;
		SetFgColor(gBlackPixel) ;
	}
	else {	// We do this so that we can properly show up selected cells
		SetFgColor(gWhitePixel) ;
		PaintRect(&theRect) ;
		SetFgColor(gBlackPixel) ;
	}
	
	TextSize(0);
	MoveTo(cellFrame.left + 4, cellFrame.bottom - 4);
	switch (inCell.col) {
		case 1:
			sprintf(str,"[%3d,%3d]",mH[inCell.row-1],mK[inCell.row-1]) ;
			DrawString(str) ;
			break ;
		case 2:
			sprintf(str,"%.3f",mGVector[inCell.row-1]) ;
			DrawString(str) ;
			break ;
		case 3:
			sprintf(str,"%.2f",mAmplitude[inCell.row-1]) ;
			DrawString(str) ;
			break ;
		case 4:
			if (mPhase[inCell.row-1] == kUnAssignedPhase) {
				MoveTo((cellFrame.left+cellFrame.right)/2 - 4,(cellFrame.top+cellFrame.bottom)/2) ;
				LineTo((cellFrame.left+cellFrame.right)/2 + 4,(cellFrame.top+cellFrame.bottom)/2) ;					
			}
			else {
				sprintf(str,"%.2f",mPhase[inCell.row-1]) ;
				DrawString(str) ;
			}
			break ;
		case 5:
			if (mIncludeReflection[inCell.row-1]) {
				MoveTo((cellFrame.left+cellFrame.right)/2 - 4,(cellFrame.top+cellFrame.bottom)/2-2) ;
				LineTo((cellFrame.left+cellFrame.right)/2,(cellFrame.top+cellFrame.bottom)/2+2) ;					
				LineTo((cellFrame.left+cellFrame.right)/2 + 4,(cellFrame.top+cellFrame.bottom)/2-6) ;					
			}
			else {
				MoveTo((cellFrame.left+cellFrame.right)/2 - 4,(cellFrame.top+cellFrame.bottom)/2) ;
				LineTo((cellFrame.left+cellFrame.right)/2 + 4,(cellFrame.top+cellFrame.bottom)/2) ;					
			}
			break ;
		case 6:
			if (mUsePhase[inCell.row-1]) {
				MoveTo((cellFrame.left+cellFrame.right)/2 - 4,(cellFrame.top+cellFrame.bottom)/2-2) ;
				LineTo((cellFrame.left+cellFrame.right)/2,(cellFrame.top+cellFrame.bottom)/2+2) ;					
				LineTo((cellFrame.left+cellFrame.right)/2 + 4,(cellFrame.top+cellFrame.bottom)/2-6) ;					
			}
			else {
				MoveTo((cellFrame.left+cellFrame.right)/2 - 4,(cellFrame.top+cellFrame.bottom)/2) ;
				LineTo((cellFrame.left+cellFrame.right)/2 + 4,(cellFrame.top+cellFrame.bottom)/2) ;					
			}
			break ;
	}
}

void	TDiffractionDataTable::SortOnAmpitude() {
	long gap ;
	register long i,j ;
	register short itemp ;
	register bool	btemp ;
	register float temp ;
	for (gap=mNumReflections/2;gap>0;gap/=2)
		for(i=gap;i<mNumReflections;i++)
			for(j=i-gap;j>=0;j-=gap)
				{
					if (mAmplitude[j] >= mAmplitude[j+gap])
						break ;
						
					temp = mGVector[j] ;
					mGVector[j] = mGVector[j+gap] ;
					mGVector[j+gap] = temp ;
					
					temp = mAmplitude[j] ;
					mAmplitude[j] = mAmplitude[j+gap] ;
					mAmplitude[j+gap] = temp ;
					
					temp = mPhase[j] ;
					mPhase[j] = mPhase[j+gap] ;
					mPhase[j+gap] = temp ;
					
					itemp = mH[j] ;
					mH[j] = mH[j+gap] ;
					mH[j+gap] = itemp ;
					
					itemp = mK[j] ;
					mK[j] = mK[j+gap] ;
					mK[j+gap] = itemp ;

					itemp = mIndex[j] ;
					mIndex[j] = mIndex[j+gap] ;
					mIndex[j+gap] = itemp ;

					btemp = mIncludeReflection[j] ;
					mIncludeReflection[j] = mIncludeReflection[j+gap] ;
					mIncludeReflection[j+gap] = btemp ;

					btemp = mUsePhase[j] ;
					mUsePhase[j] = mUsePhase[j+gap] ;
					mUsePhase[j+gap] = btemp ;
			  	}
			  	
	GroupSymmetryRelatedReflections() ;
}

void	TDiffractionDataTable::SortOnG() {
	long gap ;
	register long i,j ;
	register short itemp ;
	register bool	btemp ;
	register float temp ;
	for (gap=mNumReflections/2;gap>0;gap/=2)
		for(i=gap;i<mNumReflections;i++)
			for(j=i-gap;j>=0;j-=gap)
				{
					if (mGVector[j] <= mGVector[j+gap])
						break ;
						
					temp = mGVector[j] ;
					mGVector[j] = mGVector[j+gap] ;
					mGVector[j+gap] = temp ;
					
					temp = mAmplitude[j] ;
					mAmplitude[j] = mAmplitude[j+gap] ;
					mAmplitude[j+gap] = temp ;
					
					temp = mPhase[j] ;
					mPhase[j] = mPhase[j+gap] ;
					mPhase[j+gap] = temp ;
					
					itemp = mH[j] ;
					mH[j] = mH[j+gap] ;
					mH[j+gap] = itemp ;
					
					itemp = mK[j] ;
					mK[j] = mK[j+gap] ;
					mK[j+gap] = itemp ;

					itemp = mIndex[j] ;
					mIndex[j] = mIndex[j+gap] ;
					mIndex[j+gap] = itemp ;

					btemp = mIncludeReflection[j] ;
					mIncludeReflection[j] = mIncludeReflection[j+gap] ;
					mIncludeReflection[j+gap] = btemp ;

					btemp = mUsePhase[j] ;
					mUsePhase[j] = mUsePhase[j+gap] ;
					mUsePhase[j+gap] = btemp ;
			  	}

	GroupSymmetryRelatedReflections() ;
}

void	TDiffractionDataTable::GroupSymmetryRelatedReflections() {
	// Set the index mIndex so that symmetry related reflections appear together
	// The index is already set but needs to be reset now that they have been sorted on 
	// either g-vector or amplitude
	
	short	numInGroup,currentGroupIndex,groupIndex = 1 ;
	
	long gap ;
	register long i,j,k ;
	register short itemp ;
	register bool	btemp ;
	register float temp ;
	float	delta = 0.0001 ;
	for (gap=mNumReflections/2;gap>0;gap/=2)
		for(i=gap;i<mNumReflections;i++)
			for(j=i-gap;j>=0;j-=gap)
				{
					if (mSortKey == kSortOnG) {
						if ((mGVector[j+gap] - mGVector[j]) > delta)
							break ;
						if (mAmplitude[j] >= mAmplitude[j+gap])
							break ;
					} 
					else if (mSortKey == kSortOnAmplitude) {
						if (mAmplitude[j] > mAmplitude[j+gap])
							break ;
						if ((mGVector[j+gap] - mGVector[j]) > delta)
							break ;
					} 
						
					if (mIndex[j] == mIndex[j+gap])
						break ;
						
					temp = mGVector[j] ;
					mGVector[j] = mGVector[j+gap] ;
					mGVector[j+gap] = temp ;
					
					temp = mAmplitude[j] ;
					mAmplitude[j] = mAmplitude[j+gap] ;
					mAmplitude[j+gap] = temp ;
					
					temp = mPhase[j] ;
					mPhase[j] = mPhase[j+gap] ;
					mPhase[j+gap] = temp ;
					
					itemp = mH[j] ;
					mH[j] = mH[j+gap] ;
					mH[j+gap] = itemp ;
					
					itemp = mK[j] ;
					mK[j] = mK[j+gap] ;
					mK[j+gap] = itemp ;

					itemp = mIndex[j] ;
					mIndex[j] = mIndex[j+gap] ;
					mIndex[j+gap] = itemp ;

					btemp = mIncludeReflection[j] ;
					mIncludeReflection[j] = mIncludeReflection[j+gap] ;
					mIncludeReflection[j+gap] = btemp ;

					btemp = mUsePhase[j] ;
					mUsePhase[j] = mUsePhase[j+gap] ;
					mUsePhase[j+gap] = btemp ;
			  	}
	
	// Assign new group indicies
	for(i = 0; i < mNumReflections;) {
		currentGroupIndex = mIndex[i] ;
		for(j=i+1; j < mNumReflections && mIndex[j] == currentGroupIndex;j++) ;
		numInGroup = j - i ;
		for(k=i;k < i+numInGroup; k++) {
			mIndex[k] = groupIndex ;
		}
		i += numInGroup ;
		groupIndex++ ;		
	}
}

void	TDiffractionDataTable::SetIncludeAllReflections(bool status) {
	for(long i = 0; i < mNumReflections;i++) {
		mIncludeReflection[i] = status ;
	}
}

void	TDiffractionDataTable::SetUseAllPhases(bool status) {
	for(long i = 0; i < mNumReflections;i++) {
		// Let us not set phases that are unassigned
		if (mPhase[i] != kUnAssignedPhase)
			mUsePhase[i] = status ;
		else
			mUsePhase[i] = false ;	// regardless
	}
}

// --------------------------------------------------------------------------
// return the phase that is symmetry related to the phase of reflection h,k
// --------------------------------------------------------------------------
float	SymmetryRelatedPhase(short h,short k,short h1,short k1,float phase, short planegroup ) {
	// Possibilities are:
	//	:	Unrelated
	//	:	Same
	//	:	Complex Conjugate (either because (-h,-k) or because of symmetry
	//	:	phase + 180
	//	:	-phase + 180
	
	if (h1 == h && k1 == k) return phase ;
	
	// We'll use 0 for unrelated
	switch (planegroup) {
		case 1:	// p1
			// Shouldn't even be here
			return phase ;	// unrelated
			break ;
		case 2:	// p2
			// Shouldn't even be here
			return phase ;	// unrelated
			break ;
		case 3:	// p1m1
			// P(h,k) = P(-h,k) = -P(h,-k)
			if (h1 == -h && k1 == k) return phase ;
			return -phase ;
			break ;
		case 4:	// p11m	-- check
			// P(h,k) = P(-h,k) = -P(h,-k)
			if (h1 == -h && k1 == k) return phase ;
			return -phase ;
			break ;
		case 5:	// pg - p1g1
			// P(h,k) = P(-h,k)+k*180
			if (h1 == -h && k1 == k) {
				if (k1%2 == 0)
					return phase ;
				else 
					return phase+180 ;
			}
			if (h1 == h && k1 == -k) {
				if (k1%2 == 0)
					return -phase ;
				else 
					return -phase + 180 ;
			}
			return -phase ;
			break ;
		case 6:	// p11g	-- check
			// P(h,k) = P(-h,k)+k*180
			if (h1 == -h && k1 == k) {
				if (k1%2 == 0)
					return phase ;
				else 
					return phase+180 ;
			}
			if (h1 == h && k1 == -k) {
				if (k1%2 == 0)
					return -phase ;
				else 
					return -phase + 180 ;
			}
			return -phase ;
			break ;
		case 7:	// cm (c1m1)
			// P(h,k) = P(-h,k) = -P(h,-k)
			if (h1 == -h && k1 == k) return phase ;
			return -phase ;
			break ;
		case 8:	// c11m	-- check
			// P(h,k) = P(-h,k) = -P(h,-k)
			if (h1 == -h && k1 == k) return phase ;
			return -phase ;
			break ;
		case 9:	// pmm (p2mm)
			// P(h,k) = P(-h,k) = -P(h,-k)
			if (h1 == -h && k1 == k) return phase ;
			return -phase ;
			break ;
		case 10:	// p2mg
			// P(h,k) = P(-h,k)+k*180
			if (h1 == -h && k1 == k) {
				if (k%2 == 0)
					return phase ;
				else 
					return phase+180 ;
			}
			if (h1 == h && k1 == -k) {
				if (k%2 == 0)
					return -phase ;
				else 
					return -phase + 180 ;
			}
			return -phase ;
			break ;
		case 11:	// p2gm	-- check
			// P(h,k) = P(-h,k)+k*180
			if (h1 == -h && k1 == k) {
				if (k%2 == 0)
					return phase ;
				else 
					return phase+180 ;
			}
			if (h1 == h && k1 == -k) {
				if (k%2 == 0)
					return -phase ;
				else 
					return -phase + 180 ;
			}
			return -phase ;
			break ;
		case 12:	// p2gg
			// P(h,k) = P(-h,k)+(h+k)*180
			if (h1 == -h && k1 == k) {
				if ((h+k)%2 == 0)
					return phase ;
				else 
					return phase + 180 ;
			}
			if (h1 == -h && k1 == k) {
				if ((h+k)%2 == 0)
					return -phase ;
				else 
					return -phase + 180 ;
			}
			return -phase ;
			break ;
		case 13:	// c2mm
			// P(h,k) = P(-h,k) = -P(h,-k)
			if (h1 == -h && k1 == k) return phase ;
			return -phase ;
			break ;
		case 14:	// p4
			// P(h,k) = P(-k,h) = -P(k,-h)
			if (h1 == -k && k1 == h) return phase ;
			return -phase ;
			break ;
		case 15:	// p4mm
			// P(h,k) = P(-k,h) = P(-h,k) = -P(k,-h) = -P(h,-k)
			if (h1 == -k && k1 == h) return phase ;
			if (h1 == -h && k1 == k) return phase ;
			return -phase ;
			break ;
		case 16:	// p4g (p4mg)
			// P(h,k) = P(-k,h) = P(-h,k)+k*180
			if (h1 == -h && k1 == k) {
				if (k%2 == 0)
					return phase ;
				else 
					return phase + 180 ;
			}
			if (h1 == h && k1 == -k) {
				if (k%2 == 0)
					return -phase ;
				else 
					return -phase + 180 ;
			}
			if (h1 == -k && k1 == h) return phase ;
			return -phase ;
			break ;
		case 17:	// p4gm	-- check
			// P(h,k) = P(-k,h) = P(-h,k)+k*180
			if (h1 == -h && k1 == k) {
				if (k%2 == 0)
					return phase ;
				else 
					return phase + 180 ;
			}
			if (h1 == h && k1 == -k) {
				if (k%2 == 0)
					return -phase ;
				else 
					return -phase + 180 ;
			}
			if (h1 == -k && k1 == h) return phase ;
			return -phase ;
			break ;
		case 18:	// p3
			// P(h,k) = P(k,-h-k) = P(-h-k,h)
			if (h1 == k && k1 == -h-k) return phase ;
			if (h1 == -h-k && k1 == h) return phase ;
			return -phase ;
			break ;
		case 19:	// p3m1
			// P(h,k) = P(k,-h-k) = P(-h-k,h) = P(-k,-h)
			if (h1 == k && k1 == -h-k) return phase ;
			if (h1 == -h-k && k1 == h) return phase ;
			if (h1 == -k && k1 == -h) return phase ;
			return -phase ;
			break ;
		case 20:	// p31m
			// P(h,k) = P(k,-h-k) = P(-h-k,h) = P(k,h)
			if (h1 == k && k1 == -h-k) return phase ;
			if (h1 == -h-k && k1 == h) return phase ;
			if (h1 == k && k1 == h) return phase ;
			return -phase ;
			break ;
		case 21:	// p6
			// P(h,k) = P(k,-h-k) = P(-h-k,h)
			if (h1 == k && k1 == -h-k) return phase ;
			if (h1 == -h-k && k1 == h) return phase ;
			return -phase ;
			break ;
		case 22:	// p6mm
			// P(h,k) = P(k,-h-k) = P(-h-k,h) = P(k,h)
			if (h1 == k && k1 == -h-k) return phase ;
			if (h1 == -h-k && k1 == h) return phase ;
			if (h1 == k && k1 == h) return phase ;
			return -phase ;
			break ;
	}
}

