// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------


#include "TImageWindow.h"
#include <stdio.h>
#include <math.h>
#include "globals.h"
#include "XDrivers.h"
#include "DMMenuCommands.h"
#include "TScaleDialog.h"
#include "ToolConstants.h"
#include "TPeakList.h" 
#include "TMotifWindow.h" 
#include "TDiffractionDataTable.h"
#include "TCalibrationDialog.h"
#include "TConvertImageDialog.h"
#include "TTable.h"
#include "TDMApplicationWindow.h"

float	FindFTLatticePeak(TImage *image,FPoint *peakPos) ;
float   ReturnOffTol() ;

#define	ABS(x) ((x) < 0 ? -(x) : (x))

#define kUnAssignedPhase	-9999.
extern	void	WriteObjectInfo(char *text) ;

extern	Cursor	gZoomCursor ;
extern	Cursor	gHandCursor ;
extern	Cursor	gSelectionCursor ;
extern	Cursor	gSpotCursor ;
extern	Cursor	gSpotMaskCursor ;
extern	Cursor	gLatticeMaskCursor ;
extern	Cursor	gWedgeMaskCursor ;
extern	Cursor	gCircleMaskCursor ;
extern	Cursor	gLineMaskCursor ;
extern	Cursor	gRulerCursor ;
extern	Cursor	gApertureCursor ;

extern	short	gDebugMode ;

extern  void    PutMessage(char *) ;

extern	XFontStruct	*gMenuFont ;
extern	XFontStruct	*gDialogFont ;
extern	XFontStruct	*gWindowFont ;

typedef enum { PSEUDOCOLOR, TRUECOLOR } MonitorDisplayType ;
extern MonitorDisplayType  gMonitorDisplayType ;

extern	GC	gGC ;
extern	Pixmap	gStipple ;
extern	unsigned char	gStippleData[]  ;

void	SetCurrentTool(TPushButton	*) ;

Boolean WriteImageDataToByteTiff(char *filename,float *data,int ncols,int nrows,Boolean flipImage) ;
#include "TDMApplicationWindow.h"

char	*gPlaneGroup[] = {"p1","p2","pm","pg","cm","pmm","pmg","pgg",
				"cmm","p4","p4m","p4g","p3","p3m1","p31m","p6","p6m",nil} ;

char	*gAlternatePlaneGroup[] = {"p1","p2 (p211)","pm (p1m1)","p11m","pg (p1g1)","p11g","cm( c1m1)","c11m",
				"pmm (p2mm)","pmg (p2mg)","p2gm","pgg (p2gg)","cmm (c2mm)","p4","p4m (p4mm)","p4g (p4gm)","p4mg","p3","p3m1","p31m","p6","p6m (p6mm)",nil} ;
	
char	*gPlaneGroupSymbols[] = {"p1  ","p2  ","pm  ","p11m","p1g1","p11g","c1m1","c11m",
				"p2mm","p2mg","p2gm","p2gg","c2mm","p4  ","p4mm","p4mg","p4gm","p3  ","p3m1","p31m","p6  ","p6mm",nil} ;
				

#define min(x,y) ((x) < (y) ? (x) : (y))
#define max(x,y) ((x) > (y) ? (x) : (y))
#define	log2(x)	log(x)/log(2.0)

Boolean	IsDoubleClick() ;
Boolean	gDebug = false ;

float	gMeanOfMaskedImage ;
float BX[10] ; 
float BY[10] ;
bool  GotQuad = false ;

void	MaximumEntropyConvolution(TImageWindow *window) ;

int IsPowerOfTwo(long value) ; 
TPushButton	*GetCurrentTool() ;
extern	long near(float) ;
extern	void	ResetCurrentTool() ;
extern	float	GetPeakThreshold() ;
extern	class   TDMApplicationWindow *gTheControlWindow ;
extern	void	SetTargetWindow(TWindow *window) ;
extern	void	SendApplicationCommand(TCommand command) ; 
extern	void	AddWindow(TWindow *window) ; 
extern	void	PutMessage(char *) ;
extern  bool	ReturnQuadFit() ;
extern  bool	ReturnCubicFit() ;
void			SetOffTol(float Tol) ;

extern	float	FindPeak(TImage *image,Rect *areaToSearch,FPoint *peakPos,Boolean threshold=true) ;
extern	TPeakList *FindPeaksAlongLattice(TImage *image,TLattice *lattice,Boolean	threshold=true) ;
extern	TPeakList *FindPeaksAtLatticePoints(TImage *image,TLattice *lattice,Boolean	threshold=true) ;
extern	TPeakList *FindTEMPeaksAlongLattice(TImage *image,TLattice *lattice) ;
extern	TPeakList *FindDiffractionSpotsAtLatticePoints(TImage *image,TLattice *lattice) ;
extern	TPeakList *FindDiffractionSpotsAtLatticePoints(TImage *image,TLattice *lattice,TImage *peakMotif,float ccThreshold) ;
extern  TPeakList *FindDiffractionSpotsAtLatticePoints(TImage *image,TLattice *lattice,TImage *peakMotif,
							float ccThreshold,float *BX, float *BY)	;
extern  TPeakList *FindDiffractionSpotsAtLatticePointsC(TImage *image,TLattice *lattice,TImage *peakMotif,
							float ccThreshold,float *BX, float *BY)	;
extern	TPeak	*FindLatticePeak(TImage *image,Point &pos,TImage *peakMotif,TImage *weightImage,TLattice *lattice) ;
extern	TPeak	*TestLatticePeak(TImage *image,Point &pos,TImage *peakMotif,TImage *weightImage,TLattice *lattice) ;

TPeak *FindDiffractionSpotAtLatticePoint(Point &mousePt,TImage *image,TLattice *lattice,TImage *peakMotif) ;
TPeak *TestDiffractionSpotAtLatticePoint(Point &mousePt,TImage *image,TLattice *lattice,TImage *peakMotif) ;

void	ExtractAmplitudePhase(TArray *imageArray,long &n,long &m,float &x,float &y) ;
TImage *FindAveragePeakMotif(TImage *image,TLattice *lattice,TPeakArray *PeakArray,short peakRadius = 0) ;
TImage *FindAveragePeakMotifNoIter(TImage *image,TLattice *lattice,TPeakArray *PeakArray,short peakRadius = 0) ;
void 	FitAndSubtractPeakBackground(TImage *image,TImage *weightImage) ;

Boolean GetPeakMotifVisibility() ;

double	PiHalf = ::asin(1.0) ;
double	Pi = 2.0*PiHalf ;
double	TwoPi = 2.0*Pi ;

inline	float Mod2Pi(float x) { float xx = x - TwoPi*((int)(x/TwoPi)) ; return xx ; } ;

float 	GetTEMUhValue() ;
float 	GetTEMUkValue() ;
float 	GetTEMVhValue() ;
float 	GetTEMVkValue() ;
float 	GetEDDG1hValue() ;
float 	GetEDDG1kValue() ;
float 	GetEDDG2hValue() ; 
float 	GetEDDG2kValue() ;
float 	GetEDDG3hValue() ;
float 	GetEDDG3kValue() ;

extern	float	gDiffractionAperture ;

extern	TDMApplicationWindow *gTheControlWindow ;

extern	PeakForm gPeakForm ;
TEDDRecord::TEDDRecord() {
	g1Rect	= nil ; ;
	g2Rect	= nil ; ;
	g3Rect	= nil ; ;
	g1Defined	= false ;
	g2Defined	= false ;
	g3Defined	= false ;
} ;

TTEMRecord::TTEMRecord() {
	g1Rect	= nil ; ;
	g2Rect	= nil ; ;
	g1Defined	= false ;
	g2Defined	= false ;
	radCircle = nil ;
} ;


class	TAmplitudeResidualTable : public TTable {
	public:
		TAmplitudeResidualTable(TArray *dataArray) ;
		~TAmplitudeResidualTable() ;
	private:
		TArray	*mDataArray ;
		void	DrawCellValue(TableCellT &inCell,Rect &cellFrame) ;
		void 	DrawColLabelText(Int16 colNumber) ;
		void 	DrawRowLabelText(Int16 rowNumber) ;	
} ;


TAmplitudeResidualTable::TAmplitudeResidualTable(TArray *dataArray)
	: TTable(dataArray->GetHeight(),dataArray->GetWidth(),sizeof(float),64,16,"R-val") {
		mDataArray = new TArray(dataArray) ;
}

TAmplitudeResidualTable::~TAmplitudeResidualTable() {
	delete mDataArray ;
}

void 
TAmplitudeResidualTable::DrawCellValue(
	TableCellT &inCell,
	Rect &cellFrame)
{
	char str[64] ;
	
	TextSize(0);
	MoveTo(cellFrame.left + 4, cellFrame.bottom - 4);
	float val = *mDataArray->GetData(inCell.row-1,inCell.col-1) ;
	if (val == -1.) 
		sprintf(str,"-") ;
	else {
		if (val < 0.15) SetFgColor(gRedPixel) ;		 
		sprintf(str,"%f",*mDataArray->GetData(inCell.row-1,inCell.col-1)) ;
	}
	DrawString(str) ;
	SetFgColor(gBlackPixel) ;
}

void	
TAmplitudeResidualTable::DrawRowLabelText(
	Int16 rowNumber) 
{
	char str[64] ;
	sprintf(str,"%s",gPlaneGroupSymbols[rowNumber-1]) ;
	DrawString(str) ;
}

void 
TAmplitudeResidualTable::DrawColLabelText(
	Int16 colNumber) 
{
	char str[64] ;
	if (colNumber == 1)
		sprintf(str,"Amp Error Column #%d",colNumber) ;
	else
		sprintf(str,"C%d",colNumber) ;
	DrawString(str) ;
}
	

TImageWindow::TImageWindow(short xloc,short yloc,short width,
	short height,short depth,char *title,Boolean hasOffscreen,TImage *image)
	: TWindow(xloc,yloc,width,height,depth,title,hasOffscreen,true) {

 	mIsTarget = true ;
	mCanReceiveMenuCommands = true ;
	Show() ;	
	mImage = image ;
	mMask = nil ;
	mRuler = nil ;
	mReciprocalVectors = nil ;
	mReciprocalLattice = nil ;

	mTEMLattice = nil ;
	mEDDLattice = nil ;
	mPeakShapeImage = nil ; 
	mShowPeakAnnotations = true ;
	mPeakAnnotations = nil ;
	mPeakList = nil ;
	mObjectList = new TObjectList ;	

	mEDDRecord = nil ;
	mTEMRecord = nil ;
	mImageSelection = nil ;
	mDiffractionPeaks = nil ;	
	mCCThreshold = 0.5 ;
	mShowPeakMotif = false ;
	mHasSymmetrizedPeaks = false ;
	mDiffractionAperture = width/2. < height/2. ? width/2. : height/2. ;
	
	mWindowKind = 'iwnd' ;
	for(int i=0;i<17;i++) mPhaseResiduals[i] = nil ;
	
	mCommander = gTheControlWindow ; 
	
	short	w2 = image->Width()/2 ;
	short	h2 = image->Height()/2 ;
	gDiffractionAperture = w2 < h2 ? w2 : h2 ;
		
	SetControls() ;
	
	
	// Check to see what the size of the window is
	// If it is larger that the screen size, zoom out and set the magnification
	long	maxDim = width > height ? width : height ;
	short	dx = 0, dy = 0 ;
	
	// I have two situations. If my image is very large it will be impossible to move around in a zoomed state
	// The best moving around is accomplished with mag == 1, but then I won't see the whole image
	// I will  demagnify only if the image is less than 1024
	if (maxDim > 700) {
		if (maxDim > 1024) {
			if (width > 700) { width = 700 ; dx = (700 - mImage->Width())/2 ; } 
			if (height > 700) { height = 700 ; dy = (700 - mImage->Height())/2 ; } 
			XResizeWindow(gDisplay,mXWindow,width,height) ;
			SetRect(&mPortRect,0,0,width,height) ;
			// Set the center of the image in the center of the window
			mPortOrigin.x	-= dx ;
			mPortOrigin.y	-= dy ;
		}
		else {
			mMagnification = 700.0/maxDim ;
			// Try to set the magnification to powers of two
			float	mag = 1./mMagnification ;
			float	rp2 = log2(mag) ; 
			long n = near(rp2) ;
			mMagnification = 1./pow(2.0,n) ;
			width = near(mMagnification*width) ;
			height = near(mMagnification*height) ;
			XResizeWindow(gDisplay,mXWindow,width,height) ;
			SetRect(&mPortRect,0,0,width,height) ;
		}
	}
	
	Show() ;

	char	text[256] ;
	sprintf(text,"Image: W = %d, H = %d",mImage->Width(),mImage->Height()) ;
	::WriteObjectInfo(text) ;
	
}

TImageWindow::~TImageWindow() {
	if (mImage != nil) delete mImage ;
	if (mPeakAnnotations != nil) delete mPeakAnnotations ;
	if (mDiffractionPeaks != nil) delete mDiffractionPeaks ;
	if (mPeakShapeImage != nil) delete mPeakShapeImage ; 
	if (mMask != nil) delete mMask ; 
	if (mReciprocalVectors != nil) delete mReciprocalVectors ; 
	if (mReciprocalLattice != nil) delete mReciprocalLattice ; 

	if (mTEMLattice != nil) delete mTEMLattice ;
	if (mEDDLattice != nil) delete mEDDLattice ;
	if (mEDDRecord != nil) {
		if (mEDDRecord->g1Rect != nil)
			mObjectList->RemoveObject(mEDDRecord->g1Rect) ;
		if (mEDDRecord->g2Rect != nil)
			mObjectList->RemoveObject(mEDDRecord->g2Rect) ;
		if (mEDDRecord->g3Rect != nil)
			mObjectList->RemoveObject(mEDDRecord->g3Rect) ;
		delete mEDDRecord ;
		mEDDRecord = nil ;
	}	
	if (mTEMRecord != nil) {
		if (mTEMRecord->g1Rect != nil)
			mObjectList->RemoveObject(mTEMRecord->g1Rect) ;
		if (mTEMRecord->g2Rect != nil)
			mObjectList->RemoveObject(mTEMRecord->g2Rect) ;
		delete mTEMRecord ;
		mTEMRecord = nil ;
	}
	for(int i=0;i<17;i++) 
		if (mPhaseResiduals[i] != nil)
			delete mPhaseResiduals[i] ;
}

void TImageWindow::ResizeWindow(long width,long height) {
	GrafPtr	tPort ;
	GetPort(&tPort) ;
	SetPort(this) ;
	SetRect(&mPortRect,0,0,width,height) ;
	SetClipRegion(mPortRect) ;
	SetPort(tPort) ;
}

void	TImageWindow::DoWMRequestDelete() {
	// Send the application the close command for this window
	SetTargetWindow(this) ;
	SendApplicationCommand(kClose) ;
}

void	TImageWindow::Close() {
	// First prompt for saving
	// if (OkayToSave)
	RemoveEDDAnnotations() ;
	RemoveTEMAnnotations() ;
	TWindow::Close() ;
}

void	TImageWindow::DoCommand(TCommand command,Boolean *hasBeenHandled) {
	TImage 				*theImage,*tempImage ;
	FSInfo				theFSInfo ;
	TPeakArray			*peakArray ;
	float 				deviation ;
	float				numDegrees = 0. ;
	Rect				*imageRect = nil ;
	Rect				selectionRect ;
	DialogReturnCode 	retCode ;
	TScaleDialog		*theScaleDialog ;
	TCalibrationRecord 	calibration ;
	TCalibrationDialog 	*calibrationDialog ;
	TConvertImageDialog	*convertDialog ;
	ImageConvertType	convertToType ;
	TRuler 				*theRuler ;
	TMask 				*theMask ;
	TLattice			*theLattice ;
	unsigned	long	signature ;
	short				numClipboardObjects,i ;
	bool				needToUpdate,selectAllImage ;
	char				currentImageName[256],imageName[256],fullImageName[256] ;
	TScaleDialogRecord sRec = {mImage->Width(),mImage->Height(),100.,100.} ;
	
	
	mImage->Name(fullImageName) ;
	mImage->GetAbbreviatedName(currentImageName) ;
	
	switch(command) {
		case kHelp:
			OnCommandHelp() ;		
			*hasBeenHandled = true ;
			break ;
		case kCopy:
			ClearClipboard() ;
			if (mImageSelection != nil && mImageSelection->IsSelected()) {
				mImageSelection->GetSelectionRect(selectionRect) ;
				theImage = new TImage(*mImage,&selectionRect) ;
				AddToClipboard(theImage) ;
			}
			else {
				if (mMask != nil && mMask->IsSelected()) {
					theMask = (TMask *)mMask->Copy() ;
					AddToClipboard(theMask) ;
				}
				if (mRuler != nil && mRuler->IsSelected()) {
					theRuler = (TRuler *)mRuler->Copy() ;
					AddToClipboard(theRuler) ;
				}
				if (mTEMLattice != nil && mTEMLattice->IsSelected()) {
					theLattice = (TLattice *)mTEMLattice->Copy() ;
					AddToClipboard(theLattice) ;
				}
				if (mEDDLattice != nil && mEDDLattice->IsSelected()) {
					theLattice = (TLattice *)mEDDLattice->Copy() ;
					AddToClipboard(theLattice) ;
				}
				if (mReciprocalLattice != nil && mReciprocalLattice->IsSelected()) {
					theLattice = (TLattice *)mReciprocalLattice->Copy() ;
					AddToClipboard(theLattice) ;
				}
			}
			*hasBeenHandled = true ;
			break ;	
		case kSelectAll:
			selectAllImage = true ;
			if (mMask != nil) { mMask->Select(true) ; selectAllImage = false ; }
			if (mRuler != nil) { mRuler->Select(true) ; selectAllImage = false ; }
			if (mTEMLattice != nil) { mTEMLattice->Select(true) ; selectAllImage = false ; }
			if (mEDDLattice != nil) { mEDDLattice->Select(true) ; selectAllImage = false ; }
			if (mReciprocalLattice != nil) { mReciprocalLattice->Select(true) ; selectAllImage = false ; }
			if (mImageSelection != nil) { mImageSelection->Select(true) ; selectAllImage = false ; }
			if (selectAllImage) {
				mImageSelection = new CSelection ;
				SetRect(&selectionRect,0,0,mImage->Width(),mImage->Height()) ;
				mImageSelection->SetSelectionRect(selectionRect) ;
				mImageSelection->SetWindow(this) ;
				mImageSelection->Select(true) ;
			}
			Update(nil) ;
			*hasBeenHandled = true ;
			break ;	
		case kPaste:
			needToUpdate = false ;
			numClipboardObjects = GetNumberOfClipboardObjects() ;
			if (numClipboardObjects == 1) {
				signature = GetClipboardObject(0)->GetSignature() ;
				if (signature == 'imag') {
					theImage = new TImage(*(TImage *)GetClipboardObject(0)) ;
					theImage->SetName("untitled") ;
					theImage->Display() ;
				}
				else if (signature == 'mask') {
					if (mMask != nil) delete mMask ;
					mMask = (TMask *)GetClipboardObject(0)->Copy() ;
					mMask->SetWindow(this) ;
					mMask->Select(true) ;
					needToUpdate = true ;
				}
				else if (signature == 'rulr') {
					if (mRuler != nil) delete mRuler ;
					mRuler = (TRuler *)GetClipboardObject(0)->Copy() ;
					mRuler->SetWindow(this) ;
					mRuler->Select(true) ;
					needToUpdate = true ;
				}
				else if (signature == 'latt') {
					if (mReciprocalLattice != nil) delete mReciprocalLattice ;
					mReciprocalLattice = (TLattice *)GetClipboardObject(0)->Copy() ;
					mReciprocalLattice->SetWindow(this) ;
					mReciprocalLattice->Select(true) ;
					needToUpdate = true ;
				}
				else if (signature == 'isel') {
					if (mImageSelection != nil) delete mImageSelection ;
					mImageSelection = (CSelection *)GetClipboardObject(0)->Copy() ;
					mImageSelection->SetWindow(this) ;
					mImageSelection->Select(true) ;
					needToUpdate = true ;
				}
			}
			else {
				for(i=0; i < numClipboardObjects; i++) {
					signature = GetClipboardObject(i)->GetSignature() ;
					if (signature == 'mask') {
						if (mMask != nil) delete mMask ;
						mMask = (TMask *)GetClipboardObject(i)->Copy() ;
						mMask->SetWindow(this) ;
						mMask->Select(true) ;
						needToUpdate = true ;
					}
					else if (signature == 'rulr') {
						if (mRuler != nil) delete mRuler ;
						mRuler = (TRuler *)GetClipboardObject(i)->Copy() ;
						mRuler->SetWindow(this) ;
						mRuler->Select(true) ;
						needToUpdate = true ;
					}
					else if (signature == 'latt') {
						if (mReciprocalLattice != nil) delete mReciprocalLattice ;
						mReciprocalLattice = (TLattice *)GetClipboardObject(0)->Copy() ;
						mReciprocalLattice->SetWindow(this) ;
						mReciprocalLattice->Select(true) ;
						needToUpdate = true ;
					}
					else if (signature == 'isel') {
						if (mImageSelection != nil) delete mImageSelection ;
						mImageSelection = (CSelection *)GetClipboardObject(0)->Copy() ;
						mImageSelection->SetWindow(this) ;
						mImageSelection->Select(true) ;
						needToUpdate = true ;
					}
				}
			}
			if (needToUpdate) Update(nil) ;
			*hasBeenHandled = true ;
			break ;	
		case kCut:
			ClearClipboard() ;
			if (mMask != nil && mMask->IsSelected()) {
				AddToClipboard(mMask) ;
				mMask = nil ;
			}
			if (mRuler != nil && mRuler->IsSelected()) {
				AddToClipboard(mRuler) ;
				mRuler = nil ;
			}
			if (mTEMLattice != nil && mTEMLattice->IsSelected()) {
				AddToClipboard(mTEMLattice) ;
				mTEMLattice = nil ;
			}
			if (mEDDLattice != nil && mEDDLattice->IsSelected()) {
				AddToClipboard(mEDDLattice) ;
				mEDDLattice = nil ;
			}
			if (mReciprocalLattice != nil && mReciprocalLattice->IsSelected()) {
				AddToClipboard(mReciprocalLattice) ;
				mReciprocalLattice = nil ;
			}
			if (mImageSelection != nil && mImageSelection->IsSelected()) {
				AddToClipboard(mImageSelection) ;
				mImageSelection = nil ;
			}
			Update(nil) ;
			*hasBeenHandled = true ;
			break ;	
		case kCalibrate:
			calibration.pixelLength = mRuler->GetLength() ;
			calibration.calibrationConstant = mImage->mCalibration ;
			calibration.calibrationUnit = mImage->mCalibrationUnit ;
			calibration.calibratedLength = calibration.pixelLength*calibration.calibrationConstant ;
			calibrationDialog = new TCalibrationDialog(&calibration) ;
			if (calibrationDialog->DoModalDialog()) {
				calibrationDialog->GetDialogData(&calibration) ;
				mImage->mCalibration = calibration.calibrationConstant ;
				mImage->mCalibrationUnit = calibration.calibrationUnit ;
			}
			delete calibrationDialog ;
			*hasBeenHandled = true ;
			break ;
		case kRunMEM:
			MaximumEntropyConvolution(this) ;
			*hasBeenHandled = true ;
			break ;
		case kDefineRecLatt:
			SetPort(this) ;
			if (mReciprocalVectors != nil) { delete mReciprocalVectors ; mReciprocalVectors = nil ; } 
			if (mReciprocalLattice != nil) { delete mReciprocalLattice ; mReciprocalLattice = nil ; } 
			if (mPeakList != nil) { delete mPeakList ; mPeakList = nil ; } 
			mReciprocalVectors = new TReciprocalVectors(this) ;
			Update(nil) ;
			*hasBeenHandled = true ;
			break ;
		case kEditRecLatt:
			mReciprocalVectors->EditMaskparameters() ;
			Update(nil) ;
			*hasBeenHandled = true ;
			break ;
		case kRefineRecLatt:
			if (mReciprocalLattice != nil) { delete mReciprocalLattice ; mReciprocalLattice = nil ; } 
			mReciprocalLattice = mReciprocalVectors->GetLattice() ;
			
			gPeakForm = kHanningPeak ;	
			if (mPeakList != nil) delete mPeakList ;
			gDiffractionAperture = mDiffractionAperture ;
			mPeakList = FindTEMPeaksAlongLattice(mImage,mReciprocalLattice) ;
			if (mPeakList == nil) return ;
			if (mPeakList->GetNumberOfPeaks() == 0) {
				PutMessage("Didn't succeed in creating peaks") ;
				return ; 
			}
			
			peakArray = new TPeakArray(mPeakList) ;
			deviation = mReciprocalLattice->RefineUV(peakArray,0.1) ;	
			delete peakArray ;
			delete mPeakList ; mPeakList = nil ;
			mPeakList = FindTEMPeaksAlongLattice(mImage,mReciprocalLattice) ;
			peakArray = new TPeakArray(mPeakList) ;
			mReciprocalLattice->RefineUV(peakArray,0.03) ;
			
			delete mReciprocalVectors ;
			mReciprocalVectors = nil ;
			
			delete peakArray ;
			delete mPeakList ; mPeakList = nil ;
			
			Update(nil) ;
			*hasBeenHandled = true ;
			break ;
		case kPasteAmpPhs:
			PasteAmplitudePhases() ;
			Update(nil) ;
			*hasBeenHandled = true ;
			break ;
		case kHFFT:
			if (mImageSelection != nil) {
				mImageSelection->GetSelectionRect(selectionRect) ;
				imageRect = &selectionRect ;
			}
			theImage = new TImage(*mImage,imageRect) ;

			theImage->Mask(kHanningMask) ;
			theImage->ZeroStatistics() ;
			gMeanOfMaskedImage = theImage->Mean() ;
			*theImage -= gMeanOfMaskedImage ;
			theImage->fft() ;
			
			sprintf(imageName,"FFT of Hanning windowed %s",currentImageName) ;
			theImage->SetName(imageName) ;

			theImage->Display() ;	
			gDiffractionAperture = theImage->Width()/2. ; // In case it doesn't get set 	
			
			// Just to make sure the controls are set up properly for the new window
			theImage->GetWindow()->SetWindowContent('hfft') ;
			SetTargetWindow(theImage->GetWindow()) ;
			theImage->GetWindow()->SetControls() ;
			*hasBeenHandled = true ;
			break ;
		case kFFT:
			if (mImageSelection != nil) {
				mImageSelection->GetSelectionRect(selectionRect) ;
				imageRect = &selectionRect ;
			}
			theImage = new TImage(*mImage,imageRect) ;
			theImage->fft() ;
			sprintf(imageName,"FFT of %s",currentImageName) ;
			theImage->SetName(imageName) ;
			theImage->Display() ;
			*hasBeenHandled = true ;
			break ;
		case kIFFT:
			theImage = new TImage(*mImage) ;
			theImage->ifft() ;
			sprintf(imageName,"IFFT of %s",currentImageName) ;
			theImage->SetName(imageName) ;
			theImage->Display() ;
			*hasBeenHandled = true ;
			break ;
		case kEditMask:
			// Set the propeties for the mask
			mMask->EditMaskparameters() ;
			Update(nil) ;
			*hasBeenHandled = true ;
			break ;
		case kApplyMask:
			theImage = new TImage(*mImage) ;
			mMask->Apply(theImage) ;
			sprintf(imageName,"Masked %s",currentImageName) ;
			theImage->SetName(imageName) ;
			theImage->Display() ;
			*hasBeenHandled = true ;
			break ;
		case kWienerFilter:
			if (mImageSelection != nil) {
				mImageSelection->GetSelectionRect(selectionRect) ;
				imageRect = &selectionRect ;
			}
			tempImage = new TImage(*mImage,imageRect) ;
			theImage = WienerFilter(tempImage) ;
			sprintf(imageName,"Soft Wiener Filtered %s",currentImageName) ;
			theImage->SetName(imageName) ;
			delete tempImage ;
			if (theImage != nil)
				theImage->Display() ;
			*hasBeenHandled = true ;
			break ;
		case kWienerFilter2:
			if (mImageSelection != nil) {
				mImageSelection->GetSelectionRect(selectionRect) ;
				imageRect = &selectionRect ;
			}
			tempImage = new TImage(*mImage,imageRect) ;
			theImage = WienerFilter2(tempImage) ;
			sprintf(imageName,"Strong Wiener Filtered %s",currentImageName) ;
			theImage->SetName(imageName) ;
			delete tempImage ;
			if (theImage != nil)
				theImage->Display() ;
			*hasBeenHandled = true ;
			break ;
		case kWienerFiltera:
			if (mImageSelection != nil) {
				mImageSelection->GetSelectionRect(selectionRect) ;
				imageRect = &selectionRect ;
			}
			tempImage = new TImage(*mImage,imageRect) ;
			theImage = WienerFiltera(tempImage) ;
			sprintf(imageName,"Adjustable Wiener Filtered %s",currentImageName) ;
			theImage->SetName(imageName) ;
			delete tempImage ;
			if (theImage != nil)
				theImage->Display() ;
			*hasBeenHandled = true ;
			break ;
		case kInvert:
			mImage->Invert() ;
			Update(nil) ;
			SetControls() ; // To reset brightness and contrast
			*hasBeenHandled = true ;
			break ;
		case kScale:
			theScaleDialog = new TScaleDialog(&sRec) ;
			retCode = theScaleDialog->DoModalDialog() ;
			if (retCode == kDialogCancel) {
				delete theScaleDialog ;
				*hasBeenHandled = true ;
				break ;
			}
			theScaleDialog->GetDialogData(&sRec) ;
			delete theScaleDialog ;
			theImage = new TImage(*mImage) ;
			theImage->Scale(sRec.widthInPercent,sRec.heightInPercent) ;
			sprintf(imageName,"Scaled %s",currentImageName) ;
			theImage->SetName(imageName) ;
			theImage->Display() ;
			*hasBeenHandled = true ;
			break ;
		case kRotate:
			numDegrees = GetRealNumberWithPrompt("Degrees to rotate clockwise",0.) ;
			if (numDegrees != 0.) {
				theImage = new TImage(*mImage) ;
				theImage->RotateClockWise(numDegrees) ;
				sprintf(imageName,"Rotated %s",currentImageName) ;
				theImage->SetName(imageName) ;
				theImage->Display() ;
			}
			*hasBeenHandled = true ;
			break ;
		case kFlipHorizontal:
			mImage->FlipHorizontal() ;
			break ;
		case kFlipVertical:
			mImage->FlipVertical() ;
			break ;
		case kSharpen:
			mImage->Sharpen() ;
			*hasBeenHandled = true ;
			break ;
		case kBlur:
			mImage->Blur() ;
			*hasBeenHandled = true ;
			break ;
		case kConvert:
			convertDialog = new TConvertImageDialog ;
			if (convertDialog->DoModalDialog()) {
				convertDialog->GetDialogData(&convertToType) ;
				delete convertDialog ;
				switch(convertToType) {
					case kConvertModulusSquared:
						mImage->Intensity() ;
						break ;
					case kConvertModulus:
						mImage->Modulus() ;
						break ;
					case kConvertPhase:
						mImage->Phase() ;
						break ;
					case kConvertReal:
						mImage->Real() ;
						break ;
					case kConvertImaginary:
						mImage->Imaginary() ;
						break ;
				}
			}
			else 
				delete convertDialog ;
			*hasBeenHandled = true ;
			break ;
		case kSave:		
			SaveGrayscaleTiffImage(mImage) ;
			*hasBeenHandled = true ;
			break ;
		case kExportImage:
			SaveBinaryImage(mImage) ;
			*hasBeenHandled = true ;
			break ;
		default:
			*hasBeenHandled = false ;
			break ;
	}
}

void	TImageWindow::GetCommandStatus(TCommand command,Boolean *isEnabled,
			Boolean *isChecked) {
	long	width = mImage->Width() ;
	long	height = mImage->Height() ;
	Rect	selectionRect ;
	Boolean	isRealSpace = mImage->IsRealSpace() ;
	switch(command) {
		case kHelp:
			*isEnabled = true ;
			break ;
		case kInvert:
			*isEnabled = isRealSpace ;
			break ;	
		case kClose:
			*isEnabled = true ;
			break ;	
		case kCopy:
			*isEnabled = false ;
			if (mMask != nil && mMask->IsSelected())
				*isEnabled = true ;
			else if (mImageSelection != nil && mImageSelection->IsSelected())
				*isEnabled = true ;
			else if (mRuler != nil && mRuler->IsSelected())
				*isEnabled = true ;
			else if (mTEMLattice != nil && mTEMLattice->IsSelected())
				*isEnabled = true ;
			else if (mEDDLattice != nil && mEDDLattice->IsSelected())
				*isEnabled = true ;
			else if (mReciprocalLattice != nil && mReciprocalLattice->IsSelected())
				*isEnabled = true ;
			break ;	
		case kPaste:
			*isEnabled = ClipboardEmpty() ? false : true ;
			break ;	
		case kCut:
			*isEnabled = false ;
			if (mImageSelection != nil && mImageSelection->IsSelected())
				*isEnabled = true ;
			if (mMask != nil && mMask->IsSelected())
				*isEnabled = true ;
			else if (mRuler != nil && mRuler->IsSelected())
				*isEnabled = true ;
			else if (mTEMLattice != nil && mTEMLattice->IsSelected())
				*isEnabled = true ;
			else if (mEDDLattice != nil && mEDDLattice->IsSelected())
				*isEnabled = true ;
			else if (mReciprocalLattice != nil && mReciprocalLattice->IsSelected())
				*isEnabled = true ;
			break ;	
		case kSelectAll:
			*isEnabled = true ;
			break ;
		case kSave:	
		case kExportImage:
			*isEnabled = true ;
			break ;		
		case kCalibrate:
			// Enable if we have a ruler drawn
			*isEnabled = mRuler != nil ;
			break ;
		case kRunMEM:
			if (width == height && isRealSpace &&
					IsPowerOfTwo(width) && IsPowerOfTwo(height)) {
				if (mImage->mCalibrationUnit == kAngstrom || mImage->mCalibrationUnit == kNanometer)
					*isEnabled = true ;
				else
					*isEnabled = false ;
			}
			else
				*isEnabled = false ;
			break ;
		case kHFFT:
		case kFFT:
			if (mImageSelection != nil) {
				mImageSelection->GetSelectionRect(selectionRect) ;
				width = selectionRect.right - selectionRect.left ;
				height = selectionRect.bottom - selectionRect.top ; 
			}
			if (width == height && isRealSpace)
				*isEnabled = true ;
			else
				*isEnabled = false ;
			break ;
		case kDefineRecLatt:
		case kIFFT:
			if (width == height && !isRealSpace)
				*isEnabled = true ;
			else
				*isEnabled = false ;
			break ;
		case kEditRecLatt:
			*isEnabled = mReciprocalVectors != nil ;
			break ;
		case kRefineRecLatt:
			*isEnabled = mReciprocalVectors != nil ;
			break ;
		case kPasteAmpPhs:
			*isEnabled = !isRealSpace && (mReciprocalLattice != nil || mTEMLattice != nil) && gTableClipBoardList != nil ;
			break ;
		case kApplyMask:
			*isEnabled = (mMask != nil) ? true : false ;
			break ;
		case kEditMask:
			if (mMask == nil)
				*isEnabled = false ;
			else {
				*isEnabled = mMask->IsSelected() ;
			}
			break ;
		case kWienerFilter:
			if (mImageSelection != nil) {
				mImageSelection->GetSelectionRect(selectionRect) ;
				width = selectionRect.right - selectionRect.left ;
				height = selectionRect.bottom - selectionRect.top ; 
			}
			if (width == height && isRealSpace)
				*isEnabled = true ;
			else
				*isEnabled = false ;
			break ;
		case kWienerFilter2:
			if (mImageSelection != nil) {
				mImageSelection->GetSelectionRect(selectionRect) ;
				width = selectionRect.right - selectionRect.left ;
				height = selectionRect.bottom - selectionRect.top ;
			}
			if (width == height && isRealSpace)
				*isEnabled = true ;
			else
				*isEnabled = false ;
			break ;
		case kWienerFiltera:
			if (mImageSelection != nil) {
				mImageSelection->GetSelectionRect(selectionRect) ;
				width = selectionRect.right - selectionRect.left ;
				height = selectionRect.bottom - selectionRect.top ;
			}
			if (width == height && isRealSpace)
				*isEnabled = true ;
			else
				*isEnabled = false ;
			break ;		
		case kScale:
			*isEnabled = true ;
			break ;
		case kRotate:
			*isEnabled = true ;
			break ;
		case kFlipHorizontal:
			*isEnabled = true ;
			break ;
		case kFlipVertical:
			*isEnabled = true ;
			break ;
		case kSharpen:
			if (isRealSpace)
				*isEnabled = true ;
			break ;
		case kBlur:
			if (isRealSpace)
				*isEnabled = true ;
			break ;
		case kConvert:
			*isEnabled = !isRealSpace ;
			break ;
		default:
			break ;
	}
}

TPeakAnnotations::TPeakAnnotations(TImageWindow *imageWindow,TPeakList *peakList) {
	mNumAnnotations = peakList->GetNumberOfPeaks() ;
	if (mNumAnnotations == 0) return ;
	unsigned	long	color ;
	
	mPeakAnnotationRects = new TAnnotationRect *[mNumAnnotations] ;
	short	i,x,y ; 
	TPeakEntry	*peakEntry ;
	TPeak		*peak ;
	for (i = 0, peakEntry = peakList->GetFirstPeakEntry() ;peakEntry != nil; peakEntry = peakEntry->GetNextPeakEntry(), i++) {
		peak = peakEntry->GetPeak() ;
		x = near( peak->pos.x) ;
		y = near(peak->pos.y) ;
		switch(peak->peakType) {
			case kFoundPeak:
				color = gWhitePixel ;
				break ;
			case kAddedPeak:
				color = gGreenPixel ;
				break ;
			case kDeletedPeak:
				color = gRedPixel ;
				break ;
		}
		mPeakAnnotationRects[i] = new TAnnotationRect(x-4,y-4,x+2,y+2,color) ;
		mPeakAnnotationRects[i]->SetParentWindow(imageWindow) ;
	}
}

TPeakAnnotations::~TPeakAnnotations() {
	int i ;
	if (mPeakAnnotationRects != nil) {
		for (i = 0; i < mNumAnnotations; i++) {
			delete  mPeakAnnotationRects[i] ;
		}
		delete[] mPeakAnnotationRects ;
	}
}

void TImageWindow::UpdatePeakAnnotations() {
	if (mPeakAnnotations != nil) delete mPeakAnnotations ;	
	mPeakAnnotations = new TPeakAnnotations(this,mPeakList) ;
	Update(nil) ;
}

void TImageWindow::Draw() {

	// we will draw everything to an offscreen environment and then move everything to the window
	// depending on 
	
	SetPort(mOffscreen) ;
	mImage->Draw() ;
	if (mMask != nil)
		mMask->Draw() ;
	if (mRuler != nil)
		mRuler->Draw() ;
	if (mReciprocalLattice != nil)
		mReciprocalLattice->Draw2DGrid(mImage) ;
		
	if (mReciprocalVectors != nil)
		mReciprocalVectors->Draw() ;
		
	if (mTEMLattice != nil && mShowTEMLattice)
		DrawTEMLattice() ;
	if (mEDDLattice != nil && mShowEDDLattice)
		DrawEDDLattice() ;
	if (mPeakAnnotations != nil && mShowPeakAnnotations) {
		for(int i = 0; i < mPeakAnnotations->mNumAnnotations; i++) {
			mPeakAnnotations->mPeakAnnotationRects[i]->Draw() ;
		}
	}
	if (mImageSelection != nil)
		mImageSelection->Draw() ;
	TPane::Draw() ;
	
	SetPort(this) ;	
	
	// Calculate the part of the offscreen that fits on the image window
	FRect	fSourceRect ;
	FRect	fDestRect ;
	fDestRect.left = mPortRect.left ;
	fDestRect.top = mPortRect.top ;
	fDestRect.right = mPortRect.right ;
	fDestRect.bottom =  mPortRect.bottom ;
	fSourceRect = fDestRect ;
	WindowToImageRect(fSourceRect) ;	// This is the part of the image that should display on the window
	
	if (fSourceRect.left < 0.) fSourceRect.left = 0. ;
	if (fSourceRect.top < 0.) fSourceRect.top = 0. ;
	if (fSourceRect.right > mImage->Width()) fSourceRect.right =  mImage->Width() ;
	if (fSourceRect.bottom > mImage->Height()) fSourceRect.bottom =  mImage->Height() ;
	
	fDestRect = fSourceRect ;
	ImageToWindowRect(fDestRect) ;	// This is the part of the window occupied by the image 

	Rect destRect ;
	Rect sourceRect ;
	sourceRect.left 	= near(fSourceRect.left) ;
	sourceRect.top 		= near(fSourceRect.top) ;
	sourceRect.right 	= near(fSourceRect.right) ;
	sourceRect.bottom 	= near(fSourceRect.bottom) ;

	destRect.left 		= near(fDestRect.left) ;
	destRect.top 		= near(fDestRect.top) ;
	destRect.right 		= near(fDestRect.right) ;
	destRect.bottom 	= near(fDestRect.bottom) ;
	
	
	// If there is no intersection between the destination rectangle and the window
	// then there is nothing to draw
	if (RectISectRect(&destRect,&mPortRect)) {			
		TStOffscreen	stackOffscreen(this,&mPortRect) ;
		CopyBits(mOffscreen,&stackOffscreen,&sourceRect,&destRect,srcCopy) ;
	}
	
}


void TImageWindow::DoRuler(Point &location) {
	if (mRuler != nil) {
		delete mRuler ;
		mRuler = nil ;
	}
	mRuler = new TRuler(location,this) ;
	if (!mRuler->IsValid()) {
		delete mRuler ;
		mRuler = nil ;
	}
}

void TImageWindow::PasteAmplitudePhases() {
	long 	i,j,h,k ;
	float	phase,meanOfMaskedImage ;
	float	x,y ;
	FPoint		O,U,V ;
	TDiffractionPeaks	*diffractionPeaks ;
	
	if (mTEMLattice != nil) {
		O = mTEMLattice->GetOrigin() ;
		U = mTEMLattice->GetU() ;
		V = mTEMLattice->GetV() ;

		gPeakForm = kHanningPeak ;	
		gDiffractionAperture = mDiffractionAperture ;
			
		diffractionPeaks = new TDiffractionPeaks(mImage,mTEMLattice) ;
		
		if (diffractionPeaks->mFourierPeaks->mNumPeaks == 0) {
			PutMessage("Didn't succeed in finding peaks") ;
			delete diffractionPeaks ;
			delete mTEMLattice ;
			mTEMLattice = nil ;
			return ; 
		}
		
		meanOfMaskedImage = 0. ;
		
		delete mTEMLattice ;
		mTEMLattice = nil ;
	}
	else {
		O = mReciprocalLattice->GetOrigin() ;
		U = mReciprocalLattice->GetU() ;
		V = mReciprocalLattice->GetV() ;
		
		// Create a new work image which we will use to find the peaks
		// by doing a Hanning windowed fft
			
		TImage	*workImage = new TImage(*mImage) ;
		workImage->ifft() ;
		workImage->Mask(kHanningMask) ;
		workImage->ZeroStatistics() ;
		meanOfMaskedImage = workImage->Mean() ;
		*workImage -= meanOfMaskedImage ;
		workImage->fft() ;
		
		gPeakForm = kHanningPeak ;	
		gDiffractionAperture = mDiffractionAperture ;
			
		diffractionPeaks = new TDiffractionPeaks(workImage,mReciprocalLattice) ;
				
		delete workImage ;

		
		if (diffractionPeaks->mFourierPeaks->mNumPeaks == 0) {
			PutMessage("Didn't succeed in finding peaks") ;
			delete diffractionPeaks ;
			delete mReciprocalLattice ;
			mReciprocalLattice = nil ;
			return ; 
		}
		
		delete mReciprocalLattice ;
		mReciprocalLattice = nil ;
	}

	TArray *hkArray = diffractionPeaks->mFourierData ;
	
	float	maxAmplitude = 0. ;	
	short	iH,iK ;
	long	II,JJ ;
	
	if (gTableClipBoardList->mClipboardObjectType == 'ampl') {
		for(j=0; j < diffractionPeaks->mFourierPeaks->mNumPeaks ; j++) {
			h = diffractionPeaks->mFourierPeaks->mLatticeH[j] ;
			k = diffractionPeaks->mFourierPeaks->mLatticeK[j] ;
			if (h != 0 && k != 0 ) {
				if (diffractionPeaks->mFourierPeaks->mAmplitude[j] > maxAmplitude) {
					for(i=0; i < gTableClipBoardList->mNumValues ; i++) {
						if (h == gTableClipBoardList->mIndex1[i] && k == gTableClipBoardList->mIndex2[i]) {
							maxAmplitude = diffractionPeaks->mFourierPeaks->mAmplitude[j] ;
							iH = h ;
							iK = k ;
							II = i ;
							JJ = j ;
							break ;
						}
					}
				}
			}
		}
			
		short	hMax = hkArray->GetWidth()/2 ;
		short	kMax = hkArray->GetHeight()/2 ;
		float	scaleFactor = diffractionPeaks->mFourierPeaks->mAmplitude[JJ]/gTableClipBoardList->mValue[II] ;
		long	width = mImage->Width() ;
		long	height = mImage->Height() ;
		
		// --------------
		// Put the diffraction data down at the lattice points
		float	ux,uy,vx,vy,amp,phase ;
		ux = U.x ;
		uy = U.y ;
		vx = V.x ;
		vy = V.y ;
		
		float	ax,bx,ay,by ;
		// a = (bst * cst) / Vst
		// b = (cst * ast) / Vst
		
		ux /= width ;
		vx /= width ;

		uy /= height ;
		vy /= height ;
		
		float	Vst = fabs(ux*vy - uy*vx) ;
		ax = vy/Vst ;
		ay = -vx/Vst ;
		bx = -uy/Vst ;
		by = ux/Vst ;
		

		ux = U.x ;
		uy = U.y ;
		vx = V.x ;
		vy = V.y ;
		
		float	fnx = width/fabs(ax+bx) ;
		float	fny = width/fabs(ay+by)  ;
		long nx = fnx ;
		long ny = fny ;
		
		if ((fnx-nx) > 0.01) nx++ ;
		if ((fny-ny) > 0.01) ny++ ;
		
		float scaleFactorX = nx*fabs(ax+bx)/width ;
		float scaleFactorY = ny*fabs(ay+by)/height ;
				
		long imageWidth = 2*(near(scaleFactorX*width)/2) ;
		long imageHeight = 2*(near(scaleFactorY*height)/2) ;

		ux *= scaleFactorX ;
		uy *= scaleFactorY ;
		vx *= scaleFactorX ;
		vy *= scaleFactorY ;

		ux = near(ux) ;
		uy = near(uy) ;
		vx = near(vx) ;
		vy = near(vy) ;
		
		TArray *imageArray = new TArray(imageWidth,imageHeight,kComplexArray) ;
		*imageArray = 0. ;
		
		register long ih,ik ;
		
		// Scale all the amplitudes in the array
		for(j=0; j < gTableClipBoardList->mNumValues ; j++) {
			h = gTableClipBoardList->mIndex1[j] ;		// h value in the 'measured' difraction patterns
			if (ABS(h) > hMax) continue ;
			k = gTableClipBoardList->mIndex2[j] ;		// k value in the 'measured' difraction patterns
			if (ABS(k) > kMax) continue ;
			if (*hkArray->GetRealHK(h,k) > 0.) {
				*hkArray->GetRealHK(h,k) = gTableClipBoardList->mValue[j]*scaleFactor ;
			}
		}
		
		for(h = -hMax+1; h < hMax; h++) {
			for(k=-kMax+1; k < kMax; k++) {
				amp = *hkArray->GetRealHK(h,k) ;
				if (amp > 0. ) {
					ih = near(h*ux+k*vx) ;
					ik = near(h*uy+k*vy) ;
					phase = *hkArray->GetImagHK(h,k) ;
					*imageArray->GetRealHK(ih,ik) 	= amp*::cos(phase) ;
					*imageArray->GetImagHK(ih,ik) 	= amp*::sin(phase) ;
				}
			}
		}
		
		// Make sure that the whole plane is set in case we only had a half-plane of amplitudes to be pasted
		long hstart ;
		for(k=0; k < kMax; k++) {
			hstart = k == 0 ? 0 : -hMax+1 ;
			for(h = hstart; h < hMax; h++) {
				*imageArray->GetRealHK(-h,-k) 	= *imageArray->GetRealHK(h,k) ;
				*imageArray->GetImagHK(-h,-k) 	= -*imageArray->GetImagHK(h,k) ;
			}
		}
		*imageArray->GetImagHK(0,0) 	= 0. ;
		
		delete diffractionPeaks ;	
		imageArray->ifft(false) ;
					
		Rect	imageRect ;
		SetRect(&imageRect,imageWidth-width,imageHeight-height,
			imageWidth,imageHeight) ;
				
		TImage *theImage = new TImage(*imageArray,&imageRect) ;
		theImage->SetCalibration(1.0) ;
		theImage->SetCalibrationUnit(kPixel) ;
		
		*theImage += meanOfMaskedImage ;
		
		
		// ----------
		// I don't think I need this one
		/* 
		// Make sure we take all the phases from the original image
		theImage->fft() ;
		TArray	*amplitudeArray = theImage->GetDataArray() ;
		TArray	*phaseArray = mImage->GetDataArray() ;
		for(iy=0; iy < height; iy++)
			for(ix=0; ix < height; ix++) {
				amplitude = amplitudeArray->GetPixelAmplitude(ix,iy) ;
				phase = phaseArray->GetPixelPhase(ix,iy) ;
				amplitudeArray->SetPixelReal(amplitude*cos(phase),ix,iy) ;
				amplitudeArray->SetPixelImag(amplitude*sin(phase),ix,iy) ;
			}
		
		theImage->ifft() ;
		 */
		// -----------------------
		
		theImage->SetName("Image from substituted ampitudes") ;
		theImage->Display() ;	
			
		delete 	imageArray ;	

// -------------------		
		
		
	}
	else if (gTableClipBoardList->mClipboardObjectType == 'phas') {
		for(j=0; j < gTableClipBoardList->mNumValues ; j++) {
			h = gTableClipBoardList->mIndex1[j] ;
			k = gTableClipBoardList->mIndex2[j] ;
			phase = gTableClipBoardList->mValue[j] ;
			x = O.x + h*U.x + k*V.x ;
			y = O.y + h*U.y + k*V.y ;
		}
	}
}

void TImageWindow::DoSelection(Point &thePoint) {
	Rect	portRect ;
	portRect = mPortRect ;
	if (mImageSelection != nil) {
		delete mImageSelection ;
		mImageSelection = nil ;
	}
	mImageSelection = new CSelection(thePoint,this) ;
	if (mImageSelection->IsNullSelection()) {
		delete mImageSelection ;
		mImageSelection = nil ;
	}
	Update(nil) ;
}

void TImageWindow::DoZoom(Point &location) {
	short	xScreen,yScreen ;
	float	oldMagnification,fractionalChange ;
	
	Rect	selectionRect ;
	bool	hasSelection = false ;
	CSelection *theSelection = new CSelection(location,this) ;
	if (!theSelection->IsNullSelection()) {
		theSelection->GetSelectionRect(selectionRect) ;
		// Impose some constraint on the size
		if (selectionRect.right - selectionRect.left > 4 && selectionRect.bottom - selectionRect.top > 4) {
			hasSelection = true ;
		}
	}
	delete theSelection ;
	
	if (hasSelection) {
		// We want to zoom the selection onto the window
		long windowWidth = mPortRect.right - mPortRect.left ;
		long windowHeight = mPortRect.bottom - mPortRect.top ;
		long selectionWidth = selectionRect.right - selectionRect.left ;
		long selectionHeight = selectionRect.bottom - selectionRect.top ;
		float magx,magy ;
		mPortOrigin.x	 = selectionRect.left ;
		mPortOrigin.y	 = selectionRect.top ;
		magx	 = windowWidth/selectionWidth ;
		magy	 = windowHeight/selectionHeight ;
		if (magx < magy) {
			mMagnification = magx ;
			mPortOrigin.y = (selectionRect.bottom + selectionRect.top - windowHeight/mMagnification)/2.0 ;
		}
		else {
			mMagnification = magy ;
			mPortOrigin.x = (selectionRect.right + selectionRect.left - windowWidth/mMagnification)/2.0 ;
		}
	}
	else {
		ImageToWindowPoint(location) ;
		xScreen	= location.x ;
		yScreen	= location.y ;

		WindowToImagePoint(location) ;
				
		oldMagnification = 	mMagnification ;
		// Changed by LDM, also just factors of 2
		// Check to see if control/Shift/Command=Alt key is down
		if (ControlKeyDown() || ShiftKeyDown() || CommandKeyDown() ) {
			  fractionalChange = 0.5 ;
		}
		else {
			  fractionalChange = 2.0 ;				
		}
		
		
		mMagnification *= fractionalChange ;
		if (mMagnification > 0.95 && mMagnification < 1.05) mMagnification = 1.0 ;
		
		// Old code which makes the point flicked stay at a fixed position
		// mPortOrigin.x	 = mPortOrigin.x+xScreen*(mMagnification-oldMagnification)/(oldMagnification*mMagnification) ;
		// mPortOrigin.y	 = mPortOrigin.y+yScreen*(mMagnification-oldMagnification)/(oldMagnification*mMagnification) ;
		
		// In screen coordinates we have the location of the cursor become the center of the window
		// Thus the location is in image coordinates	
		mPortOrigin.x	 = location.x - (mPortRect.right+mPortRect.left)/(2*mMagnification) ;
		mPortOrigin.y	 = location.y - (mPortRect.bottom+mPortRect.top)/(2*mMagnification) ;
	}	
	
	Update(nil) ;
}

void TImageWindow::DoHandTool(Point &location) {	

	Point 	newLoc,oldLoc ;
	float	dx,dy ;
	GetMouse(&oldLoc) ; newLoc = oldLoc ;
	while (Button()) {
		GetMouse(&newLoc) ;
		if (newLoc.x != oldLoc.x || newLoc.y != oldLoc.y) {
			dx = (newLoc.x - oldLoc.x)/mMagnification ;
			dy = (newLoc.y - oldLoc.y)/mMagnification ;
			if (dx != 0. || dy != 0.) {
			
			
				mPortOrigin.x	-= dx ;
				mPortOrigin.y	-= dy ;
				
				Update(nil) ;
				oldLoc = newLoc ;
			}
		}
	}
}

void	TImageWindow::RefineTEMLattice() {

	gPeakForm = kHanningPeak ;	
	TPeakList *peakList = FindTEMPeaksAlongLattice(mImage,mTEMLattice) ;

	if (peakList == nil)  {
		PutMessage("Didn't succeed in finding peaks") ;
		return ; 
	}

	TPeakArray *peakArray = new TPeakArray(peakList) ;
	float deviation = mTEMLattice->RefineUV(peakArray,0.1) ;	
	delete peakArray ;
	delete peakList ;
	
	Update(nil) ;
}

void	TImageWindow::CalculateMaskedFFT() {
	TImage *theImage ;
	Rect	*imageRect = nil ;
	Rect	selectionRect ;

	if (mImageSelection != nil) {
		mImageSelection->GetSelectionRect(selectionRect) ;
		imageRect = &selectionRect ;
	}
	theImage = new TImage(*mImage,imageRect) ;
	
	theImage->Mask(kHanningMask) ;
	theImage->ZeroStatistics() ;
	gMeanOfMaskedImage = theImage->Mean() ;
	*theImage -= gMeanOfMaskedImage ;
	theImage->fft() ;

	char	currentImageName[256],imageName[256] ;
	mImage->GetAbbreviatedName(currentImageName) ;
	
	sprintf(imageName,"FFT of Hanning windowed %s",currentImageName) ;
	theImage->SetName(imageName) ;

	theImage->Display() ;	
	gDiffractionAperture = theImage->Width()/2. ; // In case it doesn't get set 	
	
	// Just to make sure the controls are set up properly for the new window
	theImage->GetWindow()->SetWindowContent('hfft') ;
	SetTargetWindow(theImage->GetWindow()) ;
	theImage->GetWindow()->SetControls() ;
}

void	TImageWindow::RemoveTEMAnnotations() {
	if (mTEMRecord == nil) return ;
	if (mTEMRecord->g1Rect != nil)
		mObjectList->RemoveObject(mTEMRecord->g1Rect) ;
	if (mTEMRecord->g2Rect != nil)
		mObjectList->RemoveObject(mTEMRecord->g2Rect) ;
	if (mTEMRecord->radCircle != nil)
		mObjectList->RemoveObject(mTEMRecord->radCircle) ;
	mTEMRecord->radCircle = nil ;
	mTEMRecord->g1Rect = nil ;
	mTEMRecord->g2Rect = nil ;
	delete mTEMRecord ;
	mTEMRecord = nil ;
	Update((XEvent *)nil) ;
}


void	TImageWindow::DrawTEMLattice() {
	mTEMLattice->Draw2DGrid(mImage) ;	
}

void	TImageWindow::FindMotif() {

	gPeakForm = kHanningPeak ;

	RemoveTEMAnnotations() ;

	TDiffractionPeaks *diffractionPattern = new TDiffractionPeaks(mImage,mTEMLattice) ;
	
	TMotifWindow *theMotifWindow = new TMotifWindow(diffractionPattern) ;
	theMotifWindow->SetCommander(gTheControlWindow) ;		
	AddWindow(theMotifWindow) ; 
	delete diffractionPattern ;
}


// EDD routines

void	TImageWindow::RefineEDDLattice() { 
	float Tol1, Tol2 ;
	Tol1 = ReturnOffTol();
	Tol2 = max(36.,Tol1*5) ; //Double the tolerance -- units of R**2
	if (mEDDLattice == nil) return ;
	gPeakForm = kGaussianPeak ;
	//Simple search here for largest local pixel used -- not super accurate 
	TPeakList *peakList = FindDiffractionSpotsAtLatticePoints(mImage,mEDDLattice) ;
	if (peakList == nil) {
		fprintf(stderr,"\nDidn't find any suitable peaks using FindPeaksAlongLattice(mImage,mEDDLattice)") ;
		return ;
	}

	TPeakArray *peakArray = new TPeakArray(peakList) ;
	float deviation = mEDDLattice->Refine(peakArray,Tol2) ;	
	deviation = mEDDLattice->Refine(peakArray,Tol1) ;	
	delete peakArray ;
	delete peakList ;
	Update(nil) ;
}

void	TImageWindow::FindDiffractionPeaks() {	
	float Tol1, Tol2, Tol3, TolBase ;
	int failed ;
	Tol1 = ReturnOffTol();
	Tol2 = max(25.,Tol1*2.5) ;
	Tol3 = max(36.,Tol1*5) ;
	TolBase = Tol1 ;
	if (mEDDLattice == nil) return ;
	
	// Find peaks along refined lattice
	if (mPeakList != nil)
		delete mPeakList ;
		
	if (mTEMRecord != nil) {
		if (mTEMRecord->radCircle != nil)
			mObjectList->RemoveObject(mTEMRecord->radCircle) ;
	}
	else {
		gDiffractionAperture = mDiffractionAperture ;
	}
		
	// First pass along the lattice
	// Turn off Zero here maybe ?
	bool useZero = UseZero() ;
	if (useZero) gTheMainPane->mUseZero->SetState(false) ;
	SetOffTol(Tol3) ; // Loose Tolerance here
	mPeakList = FindDiffractionSpotsAtLatticePoints(mImage,mEDDLattice) ;
	SetOffTol(Tol2) ; // Tighten it slightly for the next step
	if (mPeakList == nil) return ;
	
	TPeakArray	*peakArray = new TPeakArray(mPeakList) ;
	//Refine the lattice, cannot hurt!
	mEDDLattice->Refine(peakArray,Tol3) ;	
	//mEDDLattice->Print() ;
	if(ReturnQuadFit()){
		mEDDLattice->RefineQuad(peakArray,Tol3,BX,BY) ;
		//mEDDLattice->RefineQuadBiweight(peakArray,Tol2,BX,BY) ;
	}
	else if(ReturnCubicFit()){
		//mEDDLattice->RefineQuad(peakArray,Tol3,BX,BY) ;
		mEDDLattice->RefineCubic(peakArray,Tol3,BX,BY,failed) ;
		if (failed < 0) {
			//PutMessage("Warning: Cubic fit failed, using quadratic") ;
			mEDDLattice->RefineQuad(peakArray,Tol3,BX,BY) ;
		}
		//mEDDLattice->RefineCubicBiweight(peakArray,Tol2,BX,BY) ;
	}
	else {
		mEDDLattice->Refine(peakArray,Tol3) ;	
		//mEDDLattice->Print() ;
		mEDDLattice->RefineBiweight(peakArray,Tol3) ;
	}
	//mEDDLattice->Print() ;

	peakArray->Sort(kSortDescending) ;	
	// Use inner 50 percent of reflections to determine peak shape	
	TImage *peakShapeImage = FindAveragePeakMotif(mImage,mEDDLattice,peakArray) ;

	//*peakShapeImage /= peakShapeImage->VectorLength() ;
	
	// Go back and do a 'template matching' with the average motif with each spot on the lattice
	delete peakArray ;
	
	mCCThreshold = GetPeakThreshold() ;
	// Don't use too weak peaks in this pass
	if ( mCCThreshold < 0.1 ) mCCThreshold = 0.1 ;
	gDebug = true ;
	if( ReturnCubicFit() ) {
	mPeakList = FindDiffractionSpotsAtLatticePointsC(mImage,mEDDLattice,peakShapeImage,mCCThreshold,
		BX, BY) ;
	}
	else if (ReturnQuadFit()) {
	mPeakList = FindDiffractionSpotsAtLatticePoints(mImage,mEDDLattice,peakShapeImage,mCCThreshold,
		BX, BY) ;
	}
	else {
	mPeakList = FindDiffractionSpotsAtLatticePoints(mImage,mEDDLattice,peakShapeImage,mCCThreshold) ;
	}
	//delete peakShapeImage ;
	gDebug = false ;
	mCCThreshold = GetPeakThreshold() ;
	
	if (mPeakList == nil) return ;
	SetOffTol(Tol1) ; // Tighten it to the used value
	// --------------------- We will do the whole thing once more -----------------
	// -- This is to see if we can improve the accuracy and pick up more spots ----
	peakArray = new TPeakArray(mPeakList) ;
	if(ReturnQuadFit()){
		mEDDLattice->RefineQuad(peakArray,Tol2,BX,BY) ;
	//fprintf(stderr,"Quad Terms 2 X O = %.2f,%.2f,%.2f\n",BX[3],BX[4],BX[5]) ;
	//fprintf(stderr,"Quad Terms 2 Y O = %.2f,%.2f,%.2f\n",BY[3],BY[4],BY[5]) ;
	}
	else if(ReturnCubicFit()){
		mEDDLattice->RefineQuad(peakArray,Tol2,BX,BY) ;
		mEDDLattice->RefineCubic(peakArray,Tol2,BX,BY,failed) ;
		if (failed < 0 ) {
			//PutMessage("Warning: Cubic fit failed, using quadratic") ;
			mEDDLattice->RefineQuad(peakArray,Tol2,BX,BY) ;
		}
		//mEDDLattice->RefineCubicBiweight(peakArray,Tol2,BX,BY) ;
	}
	else {
		mEDDLattice->RefineBiweight(peakArray,Tol2) ;
	}
	//mEDDLattice->RefineBiweight(peakArray,Tol2) ;
	//mEDDLattice->Print() ;

	//LDM Addition, show the new lattice ?
	TWindow *GetTargetWindow();
	TImageWindow *theImageWindow = (TImageWindow *)GetTargetWindow() ;
	theImageWindow->ShowEDDLattice() ;

	// Redo the search routine all over again
	delete peakArray ;
	//mPeakList = FindDiffractionSpotsAtLatticePoints(mImage,mEDDLattice) ;
	if (ReturnQuadFit()){
		mPeakList = FindDiffractionSpotsAtLatticePoints(mImage,mEDDLattice,peakShapeImage,mCCThreshold,BX,BY) ;
	}
	else if(ReturnCubicFit()){
		mPeakList = FindDiffractionSpotsAtLatticePointsC(mImage,mEDDLattice,peakShapeImage,mCCThreshold,BX,BY) ;
	}
	else {
		mPeakList = FindDiffractionSpotsAtLatticePoints(mImage,mEDDLattice,peakShapeImage,mCCThreshold) ;
	}
	if (mPeakList == nil) return ;
	peakArray = new TPeakArray(mPeakList) ;
	//Don't refine again -- we have added weak spots!
	//if(ReturnQuadFit()){
	//	mEDDLattice->RefineQuad(peakArray,Tol1,BX,BY) ;
	//fprintf(stderr,"Quad Terms 3 X O = %.2f,%.2f,%.2f\n",BX[3],BX[4],BX[5]) ;
	//fprintf(stderr,"Quad Terms 3 Y O = %.2f,%.2f,%.2f\n",BY[3],BY[4],BY[5]) ;
	//}
	//else if(ReturnCubicFit()){
	//	mEDDLattice->RefineCubic(peakArray,Tol1,BX,BY,failed) ;
	//	if (failed < 0 ) {
	//		PutMessage("Warning: Cubic fit failed, using quadratic") ;
	//		mEDDLattice->RefineQuad(peakArray,Tol1,BX,BY) ;
	//	}
	//}
	//else {
	//	mEDDLattice->RefineBiweight(peakArray,Tol1) ;
	//}
	//LDM Addition, show the new lattice ?
	theImageWindow->ShowEDDLattice() ;

	// Use inner 50 percent of reflections to determine peak shape
	peakArray->Sort(kSortDescending) ;
	delete peakShapeImage ;
	peakShapeImage = FindAveragePeakMotifNoIter(mImage,mEDDLattice,peakArray) ;
        //
        //
	// Estimate S/N level
        // Take the signal as proportional to the number of peaks > Max/3, NLarge
        // The noise in these peaks will give a contribution proportional to sqrt(NLarge)
        // Hence S/N will scale as sqrt(PeakMax)
        // Threshold ~ Signal**2/(Signal**2 + Noise**2)
        // With a S/N cutoff of 3, this then gives 1/(1+NLarge/9)
        //
        //
	int PeakHeight = peakShapeImage->Height();
	int PeakWidth  = peakShapeImage->Width();
	float *p1 = peakShapeImage->GetData() ;
	//Account for the fact that we might be -ve at the edges
	float ZeroOff = *p1 ;
	float PeakMax = (peakShapeImage->Max() -ZeroOff )/3. +ZeroOff ;
	int NLarge = 0 ;
	int i ;
	for (i=0; i<PeakHeight*PeakWidth ; i++ ) {
		if (*p1++ > PeakMax) NLarge = NLarge + 1 ;
		}
	// Estimated R value for S/N of 3, reduced slightly
	float Rest = 2./(1.+NLarge/9.)/3. ;
	//fprintf(stderr,"Estimated Threshold to use %f %d %d \n",Rest,NLarge,PeakWidth) ;
	//restore Zero
	if (useZero) gTheMainPane->mUseZero->SetState(true) ;
	// Go back and do a 'template matching' with the average motif with each spot on the lattice

	delete peakArray ;
	delete mPeakList ;
	//fprintf(stderr,"Pass 2 threshold %f \n",mCCThreshold);
	if (ReturnQuadFit()){
		mPeakList = FindDiffractionSpotsAtLatticePoints(mImage,mEDDLattice,peakShapeImage,mCCThreshold,BX,BY) ;
	}
	else if(ReturnCubicFit()){
		mPeakList = FindDiffractionSpotsAtLatticePointsC(mImage,mEDDLattice,peakShapeImage,mCCThreshold,BX,BY) ;
	}
	else {
		mPeakList = FindDiffractionSpotsAtLatticePoints(mImage,mEDDLattice,peakShapeImage,mCCThreshold) ;
	}
	if (mPeakList == nil) {
		delete peakShapeImage ;
		return ;
	}

	// Now finally put up the found peaks
	mPeakAnnotations = new TPeakAnnotations(this,mPeakList) ;
	Update(nil) ;

	mShowPeakMotif = ::GetPeakMotifVisibility() ;		
	
	if (mPeakShapeImage != nil) { delete mPeakShapeImage ; mPeakShapeImage = nil ; } 
	mPeakShapeImage = new TImage(*peakShapeImage) ;
	
	RemoveEDDAnnotations() ;
	if (mTEMRecord != nil) RemoveTEMAnnotations() ;
	SetControls() ;
	
	if (mShowPeakMotif) {	
		char	*cp,text[256],imageTitle[256] ;
		if ((cp = strrchr(mWindowTitle,'/')) != NULL) 
			strcpy(text,cp+1) ;
		else
			strcpy(text,mWindowTitle) ;
			
		sprintf(imageTitle,"Motif with Threshold ~ %.2f for %s",Rest,text) ;
		peakShapeImage->SetName(imageTitle) ;
	
	
		peakShapeImage->Scale(800,800) ;
		peakShapeImage->Display() ;
	}
	else {
		delete peakShapeImage ;
	}
	//Put out some information about the motif
		char text[256] ;
		sprintf(text," Motif Size W=%d, H=%d\n Approximate Threshold %.3f",PeakWidth,PeakHeight,Rest) ;
                //SetFgColor(gRedPixel) ;
		::WriteObjectInfo(text) ;
                //SetFgColor(gBlackPixel) ;
}

void	TImageWindow::DeleteDiffractionPeaks() {
	if (mPeakList == nil) return ;
	if (mPeakAnnotations != nil) delete mPeakAnnotations ;
	mPeakAnnotations = nil ; 
	delete mPeakList ;
	mPeakList = nil ;
	if (mDiffractionPeaks != nil) { delete mDiffractionPeaks ; mDiffractionPeaks = nil ; } 
	mHasSymmetrizedPeaks = false ;
	Update(nil) ; 
	SetControls() ;
} 

void	TImageWindow::ExtractReflections() {
	// This takes the peaks with amplitudes that we have already found
	// and places them in mDiffractionPeaks so that they can be symmetrized
	// if needed
	 
	if (mEDDLattice == nil) return ;
	if (mPeakList == nil) return ;
	
	if (mDiffractionPeaks != nil) delete mDiffractionPeaks ;
	mDiffractionPeaks = nil ;
	
	TPeakArray	*peakArray = new TPeakArray(mPeakList) ;	
	mDiffractionPeaks = new TDiffractionPeaks(peakArray,mEDDLattice,mImage->GetCalibration(),mImage->GetCalibrationUnit()) ;
	delete peakArray ;

}

void	TImageWindow::CalculateAmplitudeResidual() {	
	ExtractReflections() ;		
	TArray	ampResidual(1,22) ;
	for(int i = 1; i <= 22; i++ ) {
		*ampResidual.GetData(0,i-1) = mDiffractionPeaks->CalculateAmplitudeResidual(i) ;
	}
	
	TAmplitudeResidualTable *theTable = new TAmplitudeResidualTable(&ampResidual) ;
	TWindow *theWindow = (TWindow *)theTable->GetTablePane() ;
	theWindow->SetCommander(gTheControlWindow) ;	
	AddWindow(theWindow) ;
	SetTargetWindow(theWindow) ;
}

void	TImageWindow::SymmetrizeReflections(short numPlaneGroup,Boolean showSymmetrizedPattern) {
	
	ExtractReflections() ;
		
	mDiffractionPeaks->Symmetrize(numPlaneGroup) ;	 
	
	short 	h,k,hDataMax,kDataMax,n,m,imageWidth,imageHeight ;
	float	x,y ;
	TArray	*dataArray = mDiffractionPeaks->mFourierData ;
	hDataMax = dataArray->GetWidth()/2 -1 ;
	kDataMax = dataArray->GetHeight()/2 -1 ;

	mHasSymmetrizedPeaks = true ;

	ResetCurrentTool() ;
	SetControls() ;

	char	*cp,text[256],imageTitle[256] ; 
	if ((cp = strrchr(mWindowTitle,'/')) != NULL) 
		strcpy(text,cp+1) ;
	else
		strcpy(text,mWindowTitle) ;
	
	
	sprintf(imageTitle,"Symmetrized reflections [%s] for %s",gPlaneGroupSymbols[numPlaneGroup-1],text) ;
	TDiffractionDataTable *dataTable = new TDiffractionDataTable(imageTitle,mDiffractionPeaks,numPlaneGroup,UseFriedelSymmetry()) ;
	TWindow *theWindow = (TWindow *)dataTable->GetTablePane() ;
	theWindow->SetCommander(gTheControlWindow) ;	
	AddWindow(theWindow) ;
	if (!showSymmetrizedPattern) SetTargetWindow(theWindow) ;


	if (showSymmetrizedPattern) {
		imageWidth = mImage->Width() ;
		imageHeight = mImage->Height() ;
		TImage *symmetrizedPattern = new TImage(imageWidth,imageHeight) ;
		
			
		sprintf(imageTitle,"Symmetrized [%s] Pattern of %s",gPlaneGroupSymbols[numPlaneGroup-1],text) ;
		symmetrizedPattern->SetName(imageTitle) ;
		
		FPoint	U,V,O ;
		float	ux,uy,vx,vy,ox,oy ;
		
		U = mEDDLattice->GetU() ;
		V = mEDDLattice->GetV() ;
		O = mEDDLattice->GetOrigin() ;
		ux = U.x ;
		uy = U.y ;
		vx = V.x ;
		vy = V.y ;
		ox = O.x ;
		oy = O.y ;
		Rect	sourceRect = {0,0,mPeakShapeImage->Width(),mPeakShapeImage->Height()} ;
		Rect	destRect ;
		float	halfWidth = mPeakShapeImage->Width()/2. ;
		float	halfHeight = mPeakShapeImage->Height()/2. ;
		//Add the value at the origin to the Peak Image -- so it is > 0
		long x0=0 ; long y0=0 ;
		float MakePositive = mPeakShapeImage->GetPixel(x0,y0) ;
		*mPeakShapeImage -= MakePositive ;
		//*mPeakShapeImage -= mPeakShapeImage->Min() ;	// Make it all positive
		for(k = -kDataMax; k < kDataMax; k++) 
			for(h = -hDataMax; h < hDataMax; h++) {
				x = ox + h*ux + k*vx ;
				y = oy + h*uy + k*vy ; 
				n = near(x-halfWidth) ;
				m = near(y-halfHeight) ;
				destRect = sourceRect ;
				OffsetRect(&destRect,n,m) ;
				float amp = *dataArray->GetRealHK(h,k) ; 
				if (n >= 0 && n < imageWidth && m >= 0 && m < imageHeight && amp > 1e-3) {
					CopyImageToImage(mPeakShapeImage,&sourceRect,symmetrizedPattern,&destRect,amp) ;
				}
			}
		symmetrizedPattern->SetDisplayLogScale(true) ; 
		symmetrizedPattern->Display() ;
		//LDM: COMPLETE THIS
		//LDM: Add here stuff for difference image
		//This is close, but not quite right. We need to correct for shifts from the lattice in the peaklist
		TImage *differencePattern = new TImage(imageWidth,imageHeight) ;
		sprintf(imageTitle,"Difference Pattern of %s",text) ;
		differencePattern->SetName(imageTitle) ;
		//short ijunk = -1 ;
		//if(ijunk > 0 ) {
		//	float *p1 = mImage->GetData() ;
		//	float *p2 = symmetrizedPattern->GetData() ;
		//	float *p3 = differencePattern->GetData() ;
		//	long size = imageWidth*imageHeight ;
		//	for (long i=0; i<size ; i++){
		//		*p3++ = *p1++ - *p2++ ;
		//		}
		//	}
		//else {
			TPeakArray *peakArray = new TPeakArray(mPeakList) ;
			for(long i=0; i<peakArray->mNumPeaks; i++) {
				x = peakArray->mX[i] ;
				y = peakArray->mY[i] ;
				float amp = peakArray->mValue[i];
				n = near(x-halfWidth) ;
				m = near(y-halfHeight) ;
				destRect = sourceRect ;
				OffsetRect(&destRect,n,m) ;
				if (n >= 0 && n < imageWidth && m >= 0 && m < imageHeight && amp > 1e-3) {
					CopyImageToImage(mPeakShapeImage,&sourceRect,differencePattern,&destRect,amp) ;
					}
				}
		symmetrizedPattern->SetDisplayLogScale(true) ; 
			float *p1 = mImage->GetData() ;
			float *p3 = differencePattern->GetData() ;
			long size = imageWidth*imageHeight ;
			for (long i=0; i<size ; i++){
				*p3++ = *p1++ - *p3 ;
				}
		//	}
		differencePattern->SetDisplayLogScale(true) ; 
		differencePattern->Display() ;
	}
	
}  

void	TImageWindow::DrawEDDLattice() {
	mEDDLattice->Draw2DGrid(mImage) ;
}

void	TImageWindow::RemoveEDDAnnotations() {
	if (mEDDRecord == nil) return ;

	if (mEDDRecord->g1Rect != nil)
		mObjectList->RemoveObject(mEDDRecord->g1Rect) ;
	if (mEDDRecord->g2Rect != nil)
		mObjectList->RemoveObject(mEDDRecord->g2Rect) ;
	if (mEDDRecord->g3Rect != nil)
		mObjectList->RemoveObject(mEDDRecord->g3Rect) ;

	mEDDRecord->g1Rect = nil ;
	mEDDRecord->g2Rect = nil ;
	mEDDRecord->g3Rect = nil ;

	delete mEDDRecord ;
	mEDDRecord = nil ;

	Update((XEvent *)nil) ;
}

void TImageWindow::DoKeyDown(XEvent *theEvent) {
	// We will respond to the delete key and cut/copy keys
	// so that we can remove maks
	KeySym 	keysym ;
	XComposeStatus compose ;
	char 	InputBuffer[80] ;
	int		count ;

	// What is the character
	count = XLookupString((XKeyEvent *)theEvent,InputBuffer,80,&keysym,&compose) ; 
	if (keysym == XK_Delete || keysym == XK_BackSpace) {
		if (mMask != nil)
			if (mMask->IsSelected()) {
				delete mMask ;
				mMask = nil ;
				Update((XEvent *)nil) ;
			}
		if (mRuler != nil)
			if (mRuler->IsSelected()) {
				delete mRuler ;
				mRuler = nil ;
				Update((XEvent *)nil) ;
			}
		if (mImageSelection != nil)
			if (mImageSelection->IsSelected()) {
				delete mImageSelection ;
				mImageSelection = nil ;
				Update((XEvent *)nil) ;
			}
	}
}

void TImageWindow::SetCursor() {
	short	toolID = GetCurrentTool()->GetToolID() ;
	switch(toolID) { 
		case kPointer:
			XUndefineCursor(gDisplay,mXWindow) ;
			break ;
		case kSelection:
			XDefineCursor(gDisplay,mXWindow,gSelectionCursor) ;			
			break ;
		case kRuler:
			XDefineCursor(gDisplay,mXWindow,gRulerCursor) ;
			break ;
		case kZoom:
			XDefineCursor(gDisplay,mXWindow,gZoomCursor) ;
			break ;
		case kHand:
			XDefineCursor(gDisplay,mXWindow,gHandCursor) ;			
			break ;
		case kSpotMask:
			XDefineCursor(gDisplay,mXWindow,gSpotMaskCursor) ;			
			break ;
		case kCircleMask:
			XDefineCursor(gDisplay,mXWindow,gCircleMaskCursor) ;			
			break ;
		case kLatticeMask:
			XDefineCursor(gDisplay,mXWindow,gLatticeMaskCursor) ;			
			break ;
		case kWedgeMask:
			XDefineCursor(gDisplay,mXWindow,gWedgeMaskCursor) ;			
			break ;
		case kLineMask:
			XDefineCursor(gDisplay,mXWindow,gLineMaskCursor) ;			
			break ; 
		case kTEMSetG1:
			XDefineCursor(gDisplay,mXWindow,gSpotCursor) ;			
			break ;
		case kTEMSetG2:
			XDefineCursor(gDisplay,mXWindow,gSpotCursor) ;
			break ;
		case kTEMAperture:
			XDefineCursor(gDisplay,mXWindow,gApertureCursor) ;		
			break ;
		case kEDDSetG1:
			XDefineCursor(gDisplay,mXWindow,gSpotCursor) ;
			break ;
		case kEDDSetG2:
			XDefineCursor(gDisplay,mXWindow,gSpotCursor) ;
			break ;
		case kEDDSetG3:
			XDefineCursor(gDisplay,mXWindow,gSpotCursor) ;
			break ;
	}
}

void TImageWindow::UnsetCursor() {
}

void TImageWindow::DoMouseDown(XEvent *event) {
	Point	thePoint ;
	FPoint	thePeak ;
	FPoint	theCenter ;
	GrafPtr	tPort ;
	Rect	peakSearchRect ;
	TPeak	*theFoundPeak ;
	float	thePeakValue,frad ;
	int		width,height,irad,rad = 3 ;
	char	objectInfo[256] ;
	Rect	objectRect ;
	bool	needToUpdate ;
	char	text[256] ;
	Boolean	inObject ;

	short	toolID = GetCurrentTool()->GetToolID() ;

	thePoint.x = event->xbutton.x ;
	thePoint.y = event->xbutton.y ;
	
	BringToFront(this) ;
	gFrontWindow = this ;
	GetPort(&tPort) ;
	SetPort(this) ;
	
	// Convert the point to a point on the image
	WindowToImagePoint(thePoint) ;
	
	needToUpdate = false ;
	
	switch(toolID) {
		case kPointer:
			inObject = false ;
			if (mMask != nil) {
				if (mMask->InMask(thePoint)) {
					inObject = true ;
					if (mMask->IsSelected())
						mMask->DoMouseDown(thePoint) ;
					else 
						mMask->Select(true) ;
				}
				else 
					mMask->Select(false) ;
				if (mMask->IsSelected()) {
					mMask->GetObjectInfo(objectInfo) ;
					::WriteObjectInfo(objectInfo) ;
				}
				needToUpdate = true ;
			}
			if (mReciprocalVectors != nil) {
				if (mReciprocalVectors->InMask(thePoint)) {
					inObject = true ;
					if (mReciprocalVectors->IsSelected())
						mReciprocalVectors->DoMouseDown(thePoint) ;
					else 
						mReciprocalVectors->Select(true) ;
				}
				else 
					mReciprocalVectors->Select(false) ;
				if (mReciprocalVectors->IsSelected()) {
					mReciprocalVectors->GetObjectInfo(objectInfo) ;
					::WriteObjectInfo(objectInfo) ;
				}
				needToUpdate = true ;
			}
			if (mRuler != nil) {
				if (mRuler->InRuler(thePoint)) {
					inObject = true ;
					if (mRuler->IsSelected())
						mRuler->DoMouseDown(thePoint) ;
					else 
						mRuler->Select(true) ;
				}
				else 
					mRuler->Select(false) ;
				if (mRuler->IsSelected()) 
					mRuler->ShowInfo() ;
				needToUpdate = true ;
			}
			if (mTEMLattice != nil) {
				if (mTEMLattice->InLattice(thePoint)) {
					inObject = true ;
					if (!mTEMLattice->IsSelected())
						mTEMLattice->Select(true) ;
				}
				else 
					mTEMLattice->Select(false) ;
				needToUpdate = true ;
			}
			if (mEDDLattice != nil) {
				if (mEDDLattice->InLattice(thePoint)) {
					inObject = true ;
					if (!mEDDLattice->IsSelected()) 
						mEDDLattice->Select(true) ;
				}
				else 
					mEDDLattice->Select(false) ;
				needToUpdate = true ;
			}
			if (mReciprocalLattice != nil) {
				if (mReciprocalLattice->InLattice(thePoint)) {
					inObject = true ;
					if (!mReciprocalLattice->IsSelected())
						mReciprocalLattice->Select(true) ;
				}
				else 
					mReciprocalLattice->Select(false) ;
				needToUpdate = true ;
			}
			if (mImageSelection != nil) {
				if (mImageSelection->InSelection(thePoint) || mImageSelection->IsInside(thePoint)) {
					inObject = true ;
					if (mImageSelection->IsSelected())
						mImageSelection->DoMouseDown(thePoint) ;
					else {
						if (mImageSelection->InSelection(thePoint))
							mImageSelection->Select(true) ;
						else
							mImageSelection->Select(false) ;
					}
					if (mImageSelection->IsSelected()) {
						mImageSelection->GetSelectionRect(objectRect) ;
						sprintf(objectInfo,"Selection: W = %d, H = %d\nSelection Rect: (%d,%d,%d,%d)",
							objectRect.right-objectRect.left,objectRect.bottom-objectRect.top,
							objectRect.left,objectRect.top,objectRect.right,objectRect.bottom) ;
						::WriteObjectInfo(objectInfo) ;
					}

				}
				else 
					mImageSelection->Select(false) ;
				needToUpdate = true ;
			}
			if (needToUpdate) Update(event) ;
			if (!inObject) DoHandTool(thePoint) ; // Treat as hand tool so we can move the image
			break ;
		case kSelection:
			// Deselect any mask and any that might be drawn
			if (mRuler != nil) mRuler->Select(false) ;
			if (mMask != nil) mMask->Select(false) ;
			if (mReciprocalVectors != nil) { delete mReciprocalVectors ; mReciprocalVectors = nil ; } ;
			DoSelection(thePoint) ;
			if (mImageSelection != nil) inObject = true ;
			ResetCurrentTool() ;
			XUndefineCursor(gDisplay,mXWindow) ;
			break ;
		case kRuler:
			if (mMask != nil) mMask->Select(false) ;
			if (mReciprocalVectors != nil) { delete mReciprocalVectors ; mReciprocalVectors = nil ; } ;
			if (mImageSelection != nil) mImageSelection->Select(false) ;
			DoRuler(thePoint) ;
			if (mRuler != nil) inObject = true ;
			ResetCurrentTool() ;
			Update(event) ;
			XUndefineCursor(gDisplay,mXWindow) ;
			break ;
		case kZoom:
			if (IsDoubleClick()) {
				ResetPanAndZoom() ;
				ResetCurrentTool() ;
				XUndefineCursor(gDisplay,mXWindow) ;
			}
			else 
				DoZoom(thePoint) ;
			break ;
		case kHand:
			DoHandTool(thePoint) ;
			break ;
		case kSpotMask:
			if (mRuler != nil) mRuler->Select(false) ;
			if (mReciprocalVectors != nil) { delete mReciprocalVectors ; mReciprocalVectors = nil ; } ;
			if (mImageSelection != nil) mImageSelection->Select(false) ;
			if (mMask != nil) delete mMask ;
			mMask = (TSpotMask *)new TSpotMask(this) ;
			inObject = true ;
			Update(event) ;
			ResetCurrentTool() ;
			XUndefineCursor(gDisplay,mXWindow) ;
			break ;
		case kCircleMask:
			if (mRuler != nil) mRuler->Select(false) ;
			if (mReciprocalVectors != nil) { delete mReciprocalVectors ; mReciprocalVectors = nil ; } ;
			if (mImageSelection != nil) mImageSelection->Select(false) ;
			if (mMask != nil) delete mMask ;
			mMask = (TCircleMask *)new TCircleMask(this) ;
			inObject = true ;
			Update(event) ;
			ResetCurrentTool() ;
			XUndefineCursor(gDisplay,mXWindow) ;
			break ;
		case kLatticeMask:
			if (mRuler != nil) mRuler->Select(false) ;
			if (mReciprocalVectors != nil) { delete mReciprocalVectors ; mReciprocalVectors = nil ; } ;
			if (mImageSelection != nil) mImageSelection->Select(false) ;
			if (mMask != nil) delete mMask ;
			mMask = (TLatticeMask *)new TLatticeMask(this) ;
			inObject = true ;
			Update(event) ;
			ResetCurrentTool() ;
			XUndefineCursor(gDisplay,mXWindow) ;
			break ;
		case kWedgeMask:
			if (mReciprocalVectors != nil) { delete mReciprocalVectors ; mReciprocalVectors = nil ; } ;
			if (mMask != nil) delete mMask ;
			mMask = (TWedgeMask *)new TWedgeMask(this) ;
			inObject = true ;
			Update(event) ;
			ResetCurrentTool() ;
			XUndefineCursor(gDisplay,mXWindow) ;
			break ;
		case kLineMask:
			if (mReciprocalVectors != nil) { delete mReciprocalVectors ; mReciprocalVectors = nil ; } ;
			if (mRuler != nil) mRuler->Select(false) ;
			if (mImageSelection != nil) mImageSelection->Select(false) ;
			if (mMask != nil) delete mMask ;
			mMask = (TLineMask *)new TLineMask(this) ;
			inObject = true ;
			Update(event) ;
			ResetCurrentTool() ;
			XUndefineCursor(gDisplay,mXWindow) ;
			break ; 
		case kTEMSetG1:
			if (mMask != nil) mMask->Select(false) ;
			if (mRuler != nil) mRuler->Select(false) ;
			if (mImageSelection != nil) mImageSelection->Select(false) ;
			if (mTEMRecord == nil) mTEMRecord = new TTEMRecord ;

			mTEMRecord->g1.a[2] = GetTEMUhValue() ;
			mTEMRecord->g1.a[3] = GetTEMUkValue() ;
			mTEMRecord->g2.a[2] = GetTEMVhValue() ;
			mTEMRecord->g2.a[3] = GetTEMVkValue() ;

			if (mTEMRecord->g1Defined) {
				// Redefine g1
				mObjectList->RemoveObject(mTEMRecord->g1Rect) ;
				mTEMRecord->g1Rect = nil ;
				mTEMRecord->g1Defined = false ;
			}

			SetRect(&peakSearchRect,thePoint.x-rad,thePoint.y-rad,
				thePoint.x+rad,thePoint.y+rad) ;
			thePeakValue = FindPeak(mImage,&peakSearchRect,&thePeak,true) ;
			if (thePeakValue == 0.) {
				Beep() ;
				break ;
			}
			mTEMRecord->g1Rect = new TAnnotationTextRect("1",near(thePeak.x-4),
								near(thePeak.y-4),near(thePeak.x+2),
								near(thePeak.y+2),gWhitePixel) ;
			mTEMRecord->g1.a[0] = thePeak.x ;
			mTEMRecord->g1.a[1] = thePeak.y ;			
			AddObject(mTEMRecord->g1Rect) ;

			mTEMRecord->g1Defined = true ;

			Update(event) ;
			SetCurrentTool(gTheMainPane->mTEMSetVTool) ;
			SetControls() ;
			break ;
		case kTEMSetG2:
			if (mMask != nil) mMask->Select(false) ;
			if (mRuler != nil) mRuler->Select(false) ;
			if (mImageSelection != nil) mImageSelection->Select(false) ;
			if (mTEMRecord == nil) mTEMRecord = new TTEMRecord ;

			mTEMRecord->g1.a[2] = GetTEMUhValue() ;
			mTEMRecord->g1.a[3] = GetTEMUkValue() ;
			mTEMRecord->g2.a[2] = GetTEMVhValue() ;
			mTEMRecord->g2.a[3] = GetTEMVkValue() ;

			if (mTEMRecord->g2Defined) {
				// Redefine g2
				mObjectList->RemoveObject(mTEMRecord->g2Rect) ;
				mTEMRecord->g2Rect = nil ;
				mTEMRecord->g2Defined = false ;
			}
			SetRect(&peakSearchRect,thePoint.x-rad,thePoint.y-rad,
				thePoint.x+rad,thePoint.y+rad) ;
			thePeakValue = FindPeak(mImage,&peakSearchRect,&thePeak,true) ;
			if (thePeakValue == 0.) {
				Beep() ;
				break ;
			}
			// Make sure that somehow we don't mark the same spot twice
			if (mTEMRecord->g1.a[0] == thePeak.x && mTEMRecord->g1.a[1] == thePeak.y) {
				Beep() ;
				break ;
			}
			mTEMRecord->g2Rect = new TAnnotationTextRect("2",near(thePeak.x-4),
								near(thePeak.y-4),near(thePeak.x+2),
								near(thePeak.y+2),gWhitePixel) ;
			mTEMRecord->g2.a[0] = thePeak.x ;
			mTEMRecord->g2.a[1] = thePeak.y ;			
			AddObject(mTEMRecord->g2Rect) ;
			mTEMRecord->g2Defined = true ;
			Update(event) ;
			ResetCurrentTool() ;
			SetControls() ;
			XUndefineCursor(gDisplay,mXWindow) ;
			break ;
		case kTEMAperture:
			if (mMask != nil) mMask->Select(false) ;
			if (mRuler != nil) mRuler->Select(false) ;
			if (mImageSelection != nil) mImageSelection->Select(false) ;
			width = mImage->Width() ;
			height =  mImage->Height() ;
			if (mTEMRecord == nil) mTEMRecord = new TTEMRecord ;
			if (mTEMRecord->radCircle != nil) {
				mObjectList->RemoveObject(mTEMRecord->radCircle) ;
				mTEMRecord->radCircle = nil ;
			}
			if (mEDDLattice == nil) {
				frad = sqrt((float)(thePoint.x-width/2)*(thePoint.x-width/2)
								+(thePoint.y-height/2)*(thePoint.y-height/2)) ;
				irad = (short)frad ;
				mDiffractionAperture = frad ;
				mTEMRecord->radCircle = new TAnnotationCircle(width/2,height/2,irad) ;
				AddObject(mTEMRecord->radCircle) ;
				//Put out some information about the radius
				sprintf(text,"Aperture Radius %.3f Pixels",frad) ;	::WriteObjectInfo(text) ;
			}
			else {
				theCenter = mEDDLattice->GetOrigin() ;
				float latu = mEDDLattice->GetLatticeParameterU();
				float latv = mEDDLattice->GetLatticeParameterV();
				frad = sqrt((float)(thePoint.x-theCenter.x)*(thePoint.x-theCenter.x)
								+(thePoint.y-theCenter.y)*(thePoint.y-theCenter.y)) ;
				irad = (short)frad ;
				mDiffractionAperture = frad ;
				mTEMRecord->radCircle = new TAnnotationCircle(theCenter.x,theCenter.y,irad) ;
				AddObject(mTEMRecord->radCircle) ;
				//Put out some information about the radius
				latu = frad*latu/(latu*latu+1e-8) ; latv = frad*latv/(latv*latv+1e-8) ;
				sprintf(text,"Diffraction Aperture Radius %.3f Pixels\n%.1f along U, %.1f along V",frad,latu,latv) ;	
				::WriteObjectInfo(text) ;
				gDiffractionAperture = frad ;
			}
			//gDiffractionApertureCircle = *mTEMRecord->radCircle ;
			Update(event) ;
			ResetCurrentTool() ;
			XUndefineCursor(gDisplay,mXWindow) ;
			break ;
		case kEDDSetG1:
			if (mMask != nil) mMask->Select(false) ;
			if (mRuler != nil) mRuler->Select(false) ;
			if (mImageSelection != nil) mImageSelection->Select(false) ;
			if (mEDDRecord == nil) mEDDRecord = new TEDDRecord ;
			mEDDRecord->g1.a[2] = GetEDDG1hValue() ;
			mEDDRecord->g1.a[3] = GetEDDG1kValue() ;
			mEDDRecord->g2.a[2] = GetEDDG2hValue() ;
			mEDDRecord->g2.a[3] = GetEDDG2kValue() ;
			mEDDRecord->g3.a[2] = GetEDDG3hValue() ;
			mEDDRecord->g3.a[3] = GetEDDG3kValue() ;
			if (mEDDRecord->g1Defined) {
				// Redefine g1
				mObjectList->RemoveObject(mEDDRecord->g1Rect) ;
				mEDDRecord->g1Rect = nil ;
				mEDDRecord->g1Defined = false ;
			}

			/*
			switch(gPeakForm) {
				case kGaussianPeak:
					// Gaussian peak
					fprintf(stderr,"\ngPeakForm = kGaussianPeak") ;
					break ;
				case kSincPeak:
					// Sinc Peak
					fprintf(stderr,"\ngPeakForm = kSincPeak") ;
					break ;
				case kCMPeak:
					fprintf(stderr,"\ngPeakForm = kCMPeak") ;
					break ;
				case kRawPeak:
					fprintf(stderr,"\ngPeakForm = kRawPeak") ;
					break ;
				case kHanningPeak:		
					fprintf(stderr,"\ngPeakForm = kHanningPeak") ;
					break ;
			}
			*/

			SetRect(&peakSearchRect,thePoint.x-rad,thePoint.y-rad,
				thePoint.x+rad,thePoint.y+rad) ;
			thePeakValue = FindPeak(mImage,&peakSearchRect,&thePeak,true) ;
			if (thePeakValue == 0.) {
				Beep() ;
				break ;
			}
			mEDDRecord->g1Rect = new TAnnotationTextRect("1",near(thePeak.x-4),
								near(thePeak.y-4),near(thePeak.x+2),
								near(thePeak.y+2),gWhitePixel) ;
			mEDDRecord->g1.a[0] = thePeak.x ;
			mEDDRecord->g1.a[1] = thePeak.y ;			
			mEDDRecord->g1Rect->Enable() ;
			AddObject(mEDDRecord->g1Rect) ;
			mEDDRecord->g1Defined = true ;
			Update(event) ;
			SetCurrentTool(gTheMainPane->mEDDSetG2Tool) ;
			SetControls() ;
			break ;
		case kEDDSetG2:
			if (mMask != nil) mMask->Select(false) ;
			if (mRuler != nil) mRuler->Select(false) ;
			if (mImageSelection != nil) mImageSelection->Select(false) ;
			if (mEDDRecord == nil) mEDDRecord = new TEDDRecord ;
			mEDDRecord->g1.a[2] = GetEDDG1hValue() ;
			mEDDRecord->g1.a[3] = GetEDDG1kValue() ;
			mEDDRecord->g2.a[2] = GetEDDG2hValue() ;
			mEDDRecord->g2.a[3] = GetEDDG2kValue() ;
			mEDDRecord->g3.a[2] = GetEDDG3hValue() ;
			mEDDRecord->g3.a[3] = GetEDDG3kValue() ;
			if (mEDDRecord->g2Defined) {
				// Redefine g2
				mObjectList->RemoveObject(mEDDRecord->g2Rect) ;
				mEDDRecord->g2Rect = nil ;
				mEDDRecord->g2Defined = false ;
			}
			SetRect(&peakSearchRect,thePoint.x-rad,thePoint.y-rad,
				thePoint.x+rad,thePoint.y+rad) ;
			thePeakValue = FindPeak(mImage,&peakSearchRect,&thePeak,true) ;
			if (thePeakValue == 0.) {
				Beep() ;
				break ;
			}
			// Make sure that somehow we don't mark the same spot twice
			if (mEDDRecord->g1.a[0] == thePeak.x && mEDDRecord->g1.a[1] == thePeak.y) {
				Beep() ;
				break ;
			}
				
			mEDDRecord->g2Rect = new TAnnotationTextRect("2",near(thePeak.x-4),
								near(thePeak.y-4),near(thePeak.x+2),
								near(thePeak.y+2),gWhitePixel) ;
			mEDDRecord->g2.a[0] = thePeak.x ;
			mEDDRecord->g2.a[1] = thePeak.y ;			
			mEDDRecord->g2Rect->Enable() ;
			AddObject(mEDDRecord->g2Rect) ;
			mEDDRecord->g2Defined = true ;
			Update(event) ;
			SetCurrentTool(gTheMainPane->mEDDSetG3Tool) ;
			SetControls() ;
			break ;
		case kEDDSetG3:
			if (mMask != nil) mMask->Select(false) ;
			if (mRuler != nil) mRuler->Select(false) ;
			if (mImageSelection != nil) mImageSelection->Select(false) ;
			if (mEDDRecord == nil) mEDDRecord = new TEDDRecord ;
			mEDDRecord->g1.a[2] = GetEDDG1hValue() ;
			mEDDRecord->g1.a[3] = GetEDDG1kValue() ;
			mEDDRecord->g2.a[2] = GetEDDG2hValue() ;
			mEDDRecord->g2.a[3] = GetEDDG2kValue() ;
			mEDDRecord->g3.a[2] = GetEDDG3hValue() ;
			mEDDRecord->g3.a[3] = GetEDDG3kValue() ;
			if (mEDDRecord->g3Defined) {
				// Redefine g3
				mObjectList->RemoveObject(mEDDRecord->g3Rect) ;
				mEDDRecord->g3Rect = nil ;
				mEDDRecord->g3Defined = false ;
			}
			SetRect(&peakSearchRect,thePoint.x-rad,thePoint.y-rad,
				thePoint.x+rad,thePoint.y+rad) ;
			thePeakValue = FindPeak(mImage,&peakSearchRect,&thePeak,true) ;
			if (thePeakValue == 0.) {
				Beep() ;
				break ;
			}
			if (mEDDRecord->g1.a[0] == thePeak.x && mEDDRecord->g1.a[1] == thePeak.y) {
				Beep() ;
				break ;
			}
			if (mEDDRecord->g2.a[0] == thePeak.x && mEDDRecord->g2.a[1] == thePeak.y) {
				Beep() ;
				break ;
			}
			mEDDRecord->g3Rect = new TAnnotationTextRect("3",near(thePeak.x-4),
								near(thePeak.y-4),near(thePeak.x+2),
								near(thePeak.y+2),gWhitePixel) ;
			mEDDRecord->g3.a[0] = thePeak.x ;
			mEDDRecord->g3.a[1] = thePeak.y ;			
			mEDDRecord->g3Rect->Enable() ;
			AddObject(mEDDRecord->g3Rect) ;
			mEDDRecord->g3Defined = true ;
			Update(event) ;
			ResetCurrentTool() ;
			SetControls() ;
			XUndefineCursor(gDisplay,mXWindow) ;
			break ;
		case kEDDAddSpot:
			if (mPeakList == nil) break ;
			if (mPeakShapeImage == nil) break ;
			if (mEDDLattice == nil) break ;
			// Check to see if we already have a peak at the position which has been deleted
			theFoundPeak = mPeakList->GetPeak(thePoint) ;
			if (theFoundPeak != nil) {
				theFoundPeak->peakType = kAddedPeak ;
			}
			else {
				theFoundPeak = FindDiffractionSpotAtLatticePoint(thePoint,mImage,mEDDLattice,mPeakShapeImage) ;
				if (theFoundPeak != nil) {
					theFoundPeak->peakType = kAddedPeak ;
					mPeakList->AddPeak(theFoundPeak) ;
				}
			}
			if (theFoundPeak == nil) {
				Beep() ;
				break ;
			}
			// Update Peak annotations
			UpdatePeakAnnotations() ;
			break ;
		case kEDDTestSpot:
			if (mPeakList == nil) break ;
			if (mPeakShapeImage == nil) break ;
			if (mEDDLattice == nil) break ;

			theFoundPeak = TestDiffractionSpotAtLatticePoint(thePoint,mImage,mEDDLattice,mPeakShapeImage) ;
			if (theFoundPeak == nil) {
				Beep() ;
			}
			break ;
		case kEDDRemoveSpot:
			if (mPeakList == nil) break ;
			theFoundPeak = mPeakList->GetPeak(thePoint) ;
			if (theFoundPeak != nil) {
				// Update Peak annotations
				theFoundPeak->peakType = kDeletedPeak ;
				UpdatePeakAnnotations() ;
			}
			else {
				Beep() ;
				break ;
			}
			break ;
	}

	SetPort(tPort) ;
	if (!inObject) {
		sprintf(text,"Image: W = %d, H = %d",mImage->Width(),mImage->Height()) ;
		::WriteObjectInfo(text) ;
	}
}

void TImageWindow::Update(XEvent *event) {
    GrafPtr tPort ;
 
    GetPort(&tPort) ;
    SetPort(this) ;
    Draw() ;
    XFlush(gDisplay) ;
    SetPort(tPort) ;
}
