// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------


#include <math.h>
#include "TDialog.h"
#include "TList.h"
#include "types.h"
#include "TLattice.h"
#include "TVector.h"
#include "TMatrix.h"
#include "TImage.h"
#include "XDrivers.h"
#include "globals.h"
extern	long near(float) ;
int hindex, kindex ;

bool Solve3LinearEquation(TMatrix *A,TVector *B,TVector *C) ;
int SolveLinearEquations(float **Amatrix,int Ndimension,float *BXsolution) ;
void UnsetQuadPars() ;
void UnsetCubePars() ;
float BXHold[13]={0.0}, BYHold[13]={0.0} ;
bool  QuadSet = false ;
bool  CubeSet = false ;
extern bool ReturnQuadFit() ;
extern bool ReturnCubicFit() ;
extern int ReturnXTics() ; extern int ReturnYTics() ;
float	RealXYValueQ( int h, int k, float BX[13] ) ;
float	RealXYValueC(int h, int k, float BX[13] ) ;
float	RealHKValueQ ( float BX[13], float BY[13], float x, float y) ;
float	RealHKValueC ( float BX[13], float BY[13], float x, float y) ;
// ---------------------------------------------------------------
//	void Solve3LinearEquation(TMatrix *A,TVector *B,TVector *C)
//	
//	Solves a set of 3 linear equations of the type Ax = B
//	with the solution returned in C
// ---------------------------------------------------------------

void UnsetQuadPars() {
	QuadSet = false ;
}
void UnsetCubePars() {
	CubeSet = false ;
}
bool Solve3LinearEquation(TMatrix *A,TVector *B,TVector *C) {
	double det,a,b,c ;
	
	det = A->m[0][0]*A->m[1][1]*A->m[2][2] +  A->m[0][1]*A->m[1][2]*A->m[2][0] +  
		A->m[0][2]*A->m[1][0]*A->m[2][1] - 
		A->m[0][1]*A->m[1][0]*A->m[2][2] -  A->m[0][0]*A->m[1][2]*A->m[2][1] -  
		A->m[0][2]*A->m[1][1]*A->m[2][0] ;
	
	
	a = B->a[0]*A->m[1][1]*A->m[2][2] +  A->m[0][1]*A->m[1][2]*B->a[2] +  
		A->m[0][2]*B->a[1]*A->m[2][1] - 
		A->m[0][1]*B->a[1]*A->m[2][2] -  B->a[0]*A->m[1][2]*A->m[2][1] -  
		A->m[0][2]*A->m[1][1]*B->a[2] ;
		
	b = A->m[0][0]*B->a[1]*A->m[2][2] +  B->a[0]*A->m[1][2]*A->m[2][0] +  
		A->m[0][2]*A->m[1][0]*B->a[2] - 
		B->a[0]*A->m[1][0]*A->m[2][2] -  A->m[0][0]*A->m[1][2]*B->a[2] -  
		A->m[0][2]*B->a[1]*A->m[2][0] ;
		
	c = A->m[0][0]*A->m[1][1]*B->a[2] +  A->m[0][1]*B->a[1]*A->m[2][0] +  
		B->a[0]*A->m[1][0]*A->m[2][1] - 
		A->m[0][1]*A->m[1][0]*B->a[2] -  A->m[0][0]*B->a[1]*A->m[2][1] -  
		B->a[0]*A->m[1][1]*A->m[2][0] ;
	
	if (fabs(det) < 1.0e-10) {
		fprintf(stderr,"\nDeteminant = 0 in Solve3LinearEquation") ;
		return false ;
	}
		
	C->a[0] = a/det ;
	C->a[1] = b/det ;
	C->a[2] = c/det ;
	return true ;
}

TLattice::TLattice(FPoint &o, FPoint &u, FPoint &v) 
	: mOrigin(o),mU(u),mV(v) {
	mSignature = 'latt' ; 
	mWindow = nil ;
}


TLattice::TLattice(TVector &o, TVector &u, TVector &v) {
	mOrigin.x = o.a[0] ;
	mOrigin.y = o.a[1] ;
	mU.x = u.a[0] ;
	mU.y = u.a[1] ;
	mV.x = v.a[0] ;
	mV.y = v.a[1] ;
	mSignature = 'latt' ; 
	mWindow = nil ;
}

bool 	TLattice::InLattice(Point &pt) {

	float	ox = mOrigin.x ;
	float	oy = mOrigin.y ;
	float	k,b,x1,y1,x2,y2,d ;
	float	imageDiagonal = sqrt(ox*ox+oy*oy) ;
	short	n, nMax ;
	float	vx,vy,kk,tolsq = 16 ;
	
	
	x1 = pt.x ;
	y1 = pt.y ;
	float	lineDistance = sqrt(mV.x*mV.x+mV.y*mV.y) ;
	nMax = near(imageDiagonal/lineDistance + 1) ;
	if (mU.x != 0.) {
		k = mU.y/mU.x ;
		if (k != 0.) {
			kk = 1+k*k ;
			// Test the shortest distance to one of the lines
			for(n = -nMax; n <= nMax; n++) {
				vx = n*mV.x ;
				vy = n*mV.y ;
				b = (oy+vy) - k*(ox+vx) ;
				
				// Check distance to line y = kx + b
				// The closest point on the line is (x2,y2)
				x2 = (k*y1+x1-k*b)/kk ;
				y2 = k*x2+b ;
				d = (x2-x1)*(x2-x1)+(y2-y1)*(y2-y1) ;
				if (d <= tolsq) {
					return true ;
				}
			}
		}
		else {
			// Line parallel to x-axis
			// Test the shortest distance to one of the lines
			for(n = -nMax; n <= nMax; n++) {
				y2 = oy+n*mV.y ;
				d = (y2-y1)*(y2-y1) ;
				if (d <= tolsq) {
					return true ;
				}
			}
		}
	}
	else {
		// Line parallel to y-axis
			// Test the shortest distance to one of the lines
		for(n = -nMax; n <= nMax; n++) {
			x2 = ox+n*mV.x ;
			d = (x2-x1)*(x2-x1) ;
			if (d <= tolsq) {
				return true ;
			}
		}
	}

	lineDistance = sqrt(mU.x*mU.x+mU.y*mU.y) ;
	nMax = near(imageDiagonal/lineDistance + 1) ;
	if (mV.x != 0.) {
		k = mV.y/mV.x ;
		if (k != 0.) {
			kk = 1+k*k ;
			// Test the shortest distance to one of the lines
			for(n = -nMax; n <= nMax; n++) {
				vx = n*mU.x ;
				vy = n*mU.y ;
				b = (oy+vy) - k*(ox+vx) ;
				
				// Check distance to line y = kx + b
				// The closest point on the line is (x2,y2)
				x2 = (k*y1+x1-k*b)/kk ;
				y2 = k*x2+b ;
				d = (x2-x1)*(x2-x1)+(y2-y1)*(y2-y1) ;
				if (d <= tolsq) {
					return true ;
				}
			}
		}
		else {
			// Line parallel to x-axis
			// Test the shortest distance to one of the lines
			for(n = -nMax; n <= nMax; n++) {
				y2 = oy+n*mU.y ;
				d = (y2-y1)*(y2-y1) ;
				if (d <= tolsq) {
					return true ;
				}
			}
		}
	}
	else {
		// Line parallel to y-axis
			// Test the shortest distance to one of the lines
		for(n = -nMax; n <= nMax; n++) {
			x2 = ox+n*mU.x ;
			d = (x2-x1)*(x2-x1) ;
			if (d <= tolsq) {
				return true ;
			}
		}
	}

	return false ;
}

void TLattice::Print() {
	float	ux,uy,vx,vy,ox,oy,u,v,gamma ;

	ux = mU.x ;
	uy = mU.y ;
	vx = mV.x ;
	vy = mV.y ;
	ox = mOrigin.x ;
	oy = mOrigin.y ;
	
	u  = sqrt(ux*ux+uy*uy) ;
	v  = sqrt(vx*vx+vy*vy) ;

	gamma = acos((ux*vx+uy*vy)/(u*v)) ;
	gamma = 90.*gamma/asin(1.0) ;
	

	fprintf(stderr,"\n --- Lattice at origin (%.3f,%.3f) --------\n",ox,oy) ;
	fprintf(stderr,"U = %.3f,%.3f |%.3f|\n",ux,uy,u) ;
	fprintf(stderr,"V = %.3f,%.3f |%.3f|\n",vx,vy,v) ;
	fprintf(stderr,"Ga = %.3f\n",gamma) ;

}

float TLattice::Refine(TPeakArray *peakList,float maxPeakDeviation) {
	register int i,j,k ;
	TMatrix A ;
	TVector BX,BY,C ;
	float ax,ay,bx,by,wx,wy,x,y,dx,dy,dsq,dn,dm ;
	float fn,fm,wt ;
	int n,m,nPeaks ;
	
	//Do it twice, should not hurt...
	for(k=0; k<2; k++) {	
		for (i=0;i<3;i++) {
			BX.a[i] = 0. ;
			BY.a[i] = 0. ;
			for(j=0;j<3;j++) {
				A.m[i][j] = 0. ;
			}
		}
		ax = mU.x ;
		ay = mU.y ;
		bx = mV.x ;
		by = mV.y ;
		wx = mOrigin.x ;
		wy = mOrigin.y ;
		for(i=0;i<peakList->mNumPeaks;i++) {
			x = peakList->mX[i] ;
			y = peakList->mY[i] ;
			wt = peakList->mValue[i];
			fn = (by*(x-wx)-bx*(y-wy))/(ax*by-ay*bx) ;
			fm = (ay*(x-wx)-ax*(y-wy))/(ay*bx-ax*by) ;
			n = near(fn) ;
			m = near(fm) ;
			if (fabs(fn-n) < maxPeakDeviation && fabs(fm-m) < maxPeakDeviation) {
				A.m[0][1] += n*wt ;
				A.m[0][2] += m*wt ;
				A.m[1][1] += n*n*wt ;
				A.m[1][2] += n*m*wt ;
				A.m[2][2] += m*m*wt ;
				A.m[0][0] += wt ;
				BX.a[0] += x*wt ;
				BX.a[1] += n*x*wt ;
				BX.a[2] += m*x*wt ;
				BY.a[0] += y*wt ;
				BY.a[1] += n*y*wt ;
				BY.a[2] += m*y*wt ;
			}
		}

		A.m[0][1] /= A.m[0][0] ;
		A.m[0][2] /= A.m[0][0] ;
		A.m[1][1] /= A.m[0][0] ;
		A.m[1][2] /= A.m[0][0] ;
		A.m[2][2] /= A.m[0][0] ;
		BX.a[0] /= A.m[0][0] ;
		BX.a[1] /= A.m[0][0] ;
		BX.a[2] /= A.m[0][0] ;
		BY.a[0] /= A.m[0][0] ;
		BY.a[1] /= A.m[0][0] ;
		BY.a[2] /= A.m[0][0] ;
		A.m[0][0] = 1. ;
		
		A.m[1][0] = A.m[0][1] ;
		A.m[2][0] = A.m[0][2] ;
		A.m[2][1] = A.m[1][2] ;
		
		if (!Solve3LinearEquation(&A,&BX,&C)) return 0. ;
		mOrigin.x = C.a[0] ;
		mU.x = C.a[1] ;
		mV.x = C.a[2] ;
		
		if (!Solve3LinearEquation(&A,&BY,&C)) return 0. ;
		mOrigin.y = C.a[0] ;
		mU.y = C.a[1] ;
		mV.y = C.a[2] ;
		
		// Compute the summed square difference
		ax = mU.x ;
		ay = mU.y ;
		bx = mV.x ;
		by = mV.y ;
		wx = mOrigin.x ;
		wy = mOrigin.y ;
		}
	dsq = 0. ;
	nPeaks = 0 ;
	for(i=0;i<peakList->mNumPeaks;i++) {
		x = peakList->mX[i] ;
		y = peakList->mY[i] ;
		fn = (by*(x-wx)-bx*(y-wy))/(ax*by-ay*bx) ;
		fm = (ay*(x-wx)-ax*(y-wy))/(ay*bx-ax*by) ;
		n = near(fn) ;
		m = near(fm) ;
		dn = fabs(fn-n) ;
		dm = fabs(fm-m) ;
		if (dn < maxPeakDeviation && dm < maxPeakDeviation) {
			dx = x-(wx+n*ax+m*bx) ;
			dy = y-(wy+n*ay+m*by) ;
			dsq += dx*dx+dy*dy ;
			nPeaks++ ;
		}
	}
	if (nPeaks > 0)
		dsq = sqrt(dsq)/nPeaks ;	// RMS deviation
	else
		dsq = 0. ;

	return dsq ;
}	

float TLattice::RefineQuad(TPeakArray *peakList,float maxPeakDeviation,float *BX,float *BY) {
	register int i,j,i1,i2 ;
	float **A , **B, T[6] ;
	float ax,ay,bx,by,wx,wy,x,y,dx,dy,dsq,dn,dm ;
	float rbest,wt;
	int n,h,k,nPeaks ;
	float BX0[13],BY0[13] ;
	n = 6 ;
	A = new float *[n] ;
	B = new float *[n] ;
	for(j=0;j<n;j++) {
		A[j] = new float[n] ;
		B[j] = new float[n] ;
		BX0[j] = BXHold[j];
		BY0[j] = BYHold[j];
	}
	for (j=3; j<13;j++) {
		BX0[j]=0 ; BY0[j]=0 ;
	}

	//Do it three times, should not hurt...
	for(k=0; k<3; k++) {	
		for (i=0;i<6;i++) {
			BX[i] = 0. ;
			BY[i] = 0. ;
			for(j=0;j<6;j++) {
				A[i][j] = 0. ;
			}
		}
		BX0[1] = mU.x ;
		BY0[1] = mU.y ;
		BX0[2] = mV.x ;
		BY0[2] = mV.y ;
		BX0[0] = mOrigin.x ; 
		BY0[0] = mOrigin.y ; 
		for(i=0;i<peakList->mNumPeaks;i++) {
			x = peakList->mX[i] ;
			y = peakList->mY[i] ;
			wt = peakList->mValue[i];
			rbest = RealHKValueQ ( BX0, BY0, x, y) ;

			if (rbest < maxPeakDeviation ) {
				T[0]=1 ; 
				T[1]=hindex ;
				T[2]=kindex ;
				T[3]=hindex*hindex ;
				T[4]=kindex*kindex ;
				T[5]=hindex*kindex ;
				for (i1=0; i1<6; i1++) {
					for (i2=0; i2<6 ; i2++) {
						A[i1][i2]+= T[i1]*T[i2]*wt ;
					}
					BX[i1] += x*T[i1]*wt;
					BY[i1] += y*T[i1]*wt;
				}
			}
		}
		float AA=A[0][0];
		for (i1=0; i1<6; i1++) {
			for (i2=0; i2<6 ; i2++) {
				A[i1][i2] = A[i1][i2]/AA ;
				B[i1][i2]=A[i1][i2] ;
			}
			BX[i1] = BX[i1]/AA;
			BY[i1] = BY[i1]/AA;
		}

		int Ndim = 6;
		if (SolveLinearEquations(A,Ndim,BX) != 1) return 0. ;
		mOrigin.x = BX[0] ;
		mU.x =  BX[1] ;
		mV.x =  BX[2] ;
		
		if (SolveLinearEquations(B,Ndim,BY) != 1) return 0. ;
		mOrigin.y = BY[0] ;
		mU.y = BY[1] ;
		mV.y = BY[2] ;
		for (i=0; i<6; i++){
			BX0[i] = BX[i] ;
			BY0[i] = BY[i] ;
		}
		}
	//fprintf(stderr,"Quad Terms X O = %.3f,%.3f,%.3f\n",BX[3],BX[4],BX[5]) ;
	//fprintf(stderr,"Quad Terms Y O = %.3f,%.3f,%.3f\n",BY[3],BY[4],BY[5]) ;
	//Save the values for later
	for (i=0 ; i<6 ; i++) {
		BXHold[i] = BX[i] ;
		BYHold[i] = BY[i] ;
	}
        for (i=6 ; i<12 ; i++) {
                BXHold[i] = 0.0 ;
                BYHold[i] = 0.0 ;
        }
	QuadSet = true ;
    CubeSet = false ;
	dsq = 0. ;

	return dsq ;
}	
float TLattice::RefineCubic(TPeakArray *peakList,float maxPeakDeviation,float *BX,float *BY
							, int failed) {
	register int i,j,k, i1,i2 ;
	float **A , **B, T[13] ;
	float ax,ay,bx,by,wx,wy,x,y,dx,dy,dsq,dn,dm ;
	float fn,fm,wt ;
	int n,m,nPeaks ;
	float BX0[13],BY0[13] ;
	failed = -1 ;
	n = 13 ;
	A = new float *[n] ;
	B = new float *[n] ;
	for(j=0;j<n;j++) {
		A[j] = new float[n] ;
		B[j] = new float[n] ;
		BX0[j] = BXHold[j];
		BY0[j] = BYHold[j];
	}
	for (j=3; j<13;j++) {
		BX0[j]=0 ; BY0[j]=0 ;
	}
	
	//Do it three times, should not hurt...
	for(k=0; k<3; k++) {	
		for (i=0;i<13;i++) {
			BX[i] = 0. ;
			BY[i] = 0. ;
			for(j=0;j<13;j++) {
				A[i][j] = 0. ;
			}
		}
		BX0[1] = mU.x ;
		BY0[1] = mU.y ;
		BX0[2] = mV.x ;
		BY0[2] = mV.y ;
		BX0[0] = mOrigin.x ; 
		BY0[0] = mOrigin.y ; 
		float rbest ;
		for(i=0;i<peakList->mNumPeaks;i++) {
			x = peakList->mX[i] ;
			y = peakList->mY[i] ;
			wt = peakList->mValue[i];
			rbest = RealHKValueC ( BX0, BY0, x, y) ;
			n = hindex ; m = kindex ;
			if (rbest < maxPeakDeviation ) {
				T[0]=1 ; 
				T[1]=n ;
				T[2]=m ;
				T[3]=n*n ;
				T[4]=m*m ;
				T[5]=n*m ;
				T[6]=n*n*n ;
				T[7]=n*n*m ;
				T[8]=n*m*m ;
				T[9]=m*m*m ;
				T[10]=n*n*n*n;
				T[11]=m*m*m*m;
				T[12]=n*n*m*m;
				for (i1=0; i1<13; i1++) {
					for (i2=0; i2<13 ; i2++) {
						A[i1][i2]+= T[i1]*T[i2]*wt ;
					}
					BX[i1] += x*T[i1]*wt;
					BY[i1] += y*T[i1]*wt;
				}
			}
		}
		float AA=A[0][0];
		for (i1=0; i1<13; i1++) {
			for (i2=0; i2<13 ; i2++) {
				A[i1][i2] = A[i1][i2]/AA ;
				B[i1][i2]=A[i1][i2] ;
			}
			BX[i1] = BX[i1]/AA;
			BY[i1] = BY[i1]/AA;
		}

		int Ndim = 13;
		if (SolveLinearEquations(A,Ndim,BX) != 1) return 0. ;
		mOrigin.x = BX[0] ;
		mU.x =  BX[1] ;
		mV.x =  BX[2] ;		
		if (SolveLinearEquations(B,Ndim,BY) != 1) return 0. ;
		mOrigin.y = BY[0] ;
		mU.y = BY[1] ;
		mV.y = BY[2] ;
		for (i=0; i<13; i++){
			BX0[i] = BX[i] ;
			BY0[i] = BY[i] ;
		}
		//fprintf(stderr,"Cube Terms X O = %.3f,%.3f,%.3f\n",BX[3],BX[4],BX[5]) ;
		//fprintf(stderr,"Cube Terms Y O = %.3f,%.3f,%.3f\n",BY[3],BY[4],BY[5]) ;
		}
	//Save the values for later
	for (i=0 ; i<13 ; i++) {
		BXHold[i] = BX[i] ;
		BYHold[i] = BY[i] ;
	}
	CubeSet = true ;
    QuadSet = false ;
	dsq = 0. ;
	failed = 1 ;
	return dsq ;
}	

float TLattice::RefineBiweight(TPeakArray *peakList,float maxPeakDeviation) {
	register int i,j,k ;
	TMatrix A ;
	TVector BX,BY,C ;
	double ax,ay,bx,by,wx,wy,x,y,dx,dy,dsq,dn,dm,SumW,Dev,SumDev ;
	double fn,fm,wt ;
	int n,m,nPeaks,AllPeaks ;
	//Biweight array
	AllPeaks = peakList->mNumPeaks ;
	float biweight[AllPeaks] ;
	for (i=0; i<AllPeaks; i++)
		biweight[i] = 1 ;

	//Do it ten times, should not hurt...
	for(k=0; k<10; k++) {	

		for (i=0;i<3;i++) {
			BX.a[i] = 0. ;
			BY.a[i] = 0. ;
			for(j=0;j<3;j++) {
				A.m[i][j] = 0. ;
			}
		}

		ax = mU.x ;
		ay = mU.y ;
		bx = mV.x ;
		by = mV.y ;
		wx = mOrigin.x ;
		wy = mOrigin.y ;
		for(i=0;i<peakList->mNumPeaks;i++) {
			x = peakList->mX[i] ;
			y = peakList->mY[i] ;
			wt = biweight[i]*peakList->mValue[i];
                        // if( ZeroBack ) wt = wt*biweight[i] ;
			fn = (by*(x-wx)-bx*(y-wy))/(ax*by-ay*bx) ;
			fm = (ay*(x-wx)-ax*(y-wy))/(ay*bx-ax*by) ;
			n = near(fn) ;
			m = near(fm) ;
			if (fabs(fn-n) < maxPeakDeviation && fabs(fm-m) < maxPeakDeviation) {
				A.m[0][1] += n*wt ;
				A.m[0][2] += m*wt ;
				A.m[1][1] += n*n*wt ;
				A.m[1][2] += n*m*wt ;
				A.m[2][2] += m*m*wt ;
				A.m[0][0] += wt ;
				BX.a[0] += x*wt ;
				BX.a[1] += n*x*wt ;
				BX.a[2] += m*x*wt ;
				BY.a[0] += y*wt ;
				BY.a[1] += n*y*wt ;
				BY.a[2] += m*y*wt ;
			}
		}

		A.m[0][1] /= A.m[0][0] ;
		A.m[0][2] /= A.m[0][0] ;
		A.m[1][1] /= A.m[0][0] ;
		A.m[1][2] /= A.m[0][0] ;
		A.m[2][2] /= A.m[0][0] ;
		BX.a[0] /= A.m[0][0] ;
		BX.a[1] /= A.m[0][0] ;
		BX.a[2] /= A.m[0][0] ;
		BY.a[0] /= A.m[0][0] ;
		BY.a[1] /= A.m[0][0] ;
		BY.a[2] /= A.m[0][0] ;
		A.m[0][0] = 1. ;
		
		A.m[1][0] = A.m[0][1] ;
		A.m[2][0] = A.m[0][2] ;
		A.m[2][1] = A.m[1][2] ;
		
		if (!Solve3LinearEquation(&A,&BX,&C)) return 0. ;
		mOrigin.x = C.a[0] ;
		mU.x = C.a[1] ;
		mV.x = C.a[2] ;
		
		if (!Solve3LinearEquation(&A,&BY,&C)) return 0. ;
		mOrigin.y = C.a[0] ;
		mU.y = C.a[1] ;
		mV.y = C.a[2] ;
		
		// Update the errors
		ax = mU.x ;
		ay = mU.y ;
		bx = mV.x ;
		by = mV.y ;
		wx = mOrigin.x ;
		wy = mOrigin.y ;
	
		//dsq = 0. ;
		//nPeaks = 0 ;
		SumDev = 0 ;
		SumW   = 0 ;
		for(i=0;i<peakList->mNumPeaks;i++) {
			x = peakList->mX[i] ;
			y = peakList->mY[i] ;
			fn = (by*(x-wx)-bx*(y-wy))/(ax*by-ay*bx) ;
			fm = (ay*(x-wx)-ax*(y-wy))/(ay*bx-ax*by) ;
			n = near(fn) ;
			m = near(fm) ;
			dn = fabs(fn-n) ;
			dm = fabs(fm-m) ;
			if (dn < maxPeakDeviation && dm < maxPeakDeviation) {
				dx = x-(wx+n*ax+m*bx) ;
				dy = y-(wy+n*ay+m*by) ;
				//dsq += dx*dx+dy*dy ;
				//nPeaks++ ;
				//Deviation, take half of both (why?)
				Dev = 0.5*(fabs(dx)+fabs(dy)) ;
				SumDev += biweight[i]*Dev*peakList->mValue[i] ;
				SumW   += biweight[i]*peakList->mValue[i] ;
				biweight[i] = Dev ;
			}
			else
				biweight[i]= 1e3 ;
		}
		//if (nPeaks > 0)
		//	dsq = sqrt(dsq)/nPeaks ;	// RMS deviation
		//else
		//	dsq = 0. ;
		//Biweighting
		SumDev = SumDev / SumW ;
		//fprintf(stderr,"Error = %f Iteration %d \n",SumDev,k);
		float BtuneL=4.685 ;
		float Btune2L=1./(BtuneL*BtuneL) ;
		for(i=0;i<peakList->mNumPeaks;i++) {
				Dev = biweight[i]/SumDev ;
				if ( Dev < BtuneL ) {
					Dev= 1.-Dev*Dev*Btune2L ;
					biweight[i]=Dev*Dev ; // Note, this is the weight (not squared)
					}
				else {
					biweight[i]=0 ;
					}
		}
	}
	dsq = SumDev ;
	return dsq ;
}	
float TLattice::RefineUV(TPeakArray *peakList,float maxPeakDeviation) {
	register int i,j ;
	TMatrix A ;
	TVector BX,BY ;
	double ax,ay,bx,by,wx,wy,x,y,dx,dy,dsq,dn,dm,det ;
	float fn,fm ;
	int n,m,nPeaks ;
	
	for (i=0;i<3;i++) {
		BX.a[i] = 0. ;
		BY.a[i] = 0. ;
		for(j=0;j<3;j++) {
			A.m[i][j] = 0. ;
		}
	}
	
	ax = mU.x ;
	ay = mU.y ;
	bx = mV.x ;
	by = mV.y ;
	wx = mOrigin.x ;
	wy = mOrigin.y ;

	//fprintf(stderr,"Initial Lattice: O = (%.2f,%.2f), U = (%.2f,%.2f),  V = (%.2f,%.2f\n",wx,wy,ax,ay,bx,by) ;
	
	for(i=0;i<peakList->mNumPeaks;i++) {
		x = peakList->mX[i] ;
		y = peakList->mY[i] ;
		fn = (by*(x-wx)-bx*(y-wy))/(ax*by-ay*bx) ;
		fm = (ay*(x-wx)-ax*(y-wy))/(ay*bx-ax*by) ;
		n = near(fn) ;
		m = near(fm) ;
		if (fabs(fn-n) < maxPeakDeviation && fabs(fm-m) < maxPeakDeviation) {
			A.m[0][0] += n*n ;
			A.m[0][1] += n*m ;
			A.m[1][1] += m*m ;
			A.m[2][0] += n ;
			A.m[2][1] += m ;
			BX.a[0] += n*x ;
			BX.a[1] += m*x ;
			BY.a[0] += n*y ;
			BY.a[1] += m*y ;
		}
	}

	BX.a[0] -= wx*A.m[2][0] ;
	BX.a[1] -= wy*A.m[2][1] ;
	BY.a[0] -= wy*A.m[2][0] ;
	BY.a[1] -= wy*A.m[2][1] ;

	det = A.m[0][0]*A.m[1][1] - A.m[0][1]*A.m[0][1] ; 	// A.m[1][0] = A.m[0][1] 
	
	if (fabs(det) < 1.0e-10) {
		fprintf(stderr,"\nDeteminant = 0 in TLattice::RefineUV") ;
		return 0. ;
	}

	mU.x = (BX.a[0]*A.m[1][1] - BX.a[1]*A.m[0][1])/det ;
	mV.x = (BX.a[1]*A.m[0][0] - BX.a[0]*A.m[0][1])/det ;

	mU.y = (BY.a[0]*A.m[1][1] - BY.a[1]*A.m[0][1])/det ;
	mV.y = (BY.a[1]*A.m[0][0] - BY.a[0]*A.m[0][1])/det ;

	// Compute the summed square difference
	ax = mU.x ;
	ay = mU.y ;
	bx = mV.x ;
	by = mV.y ;
	wx = mOrigin.x ;
	wy = mOrigin.y ;

	//fprintf(stderr,"refined Lattice: O = (%.2f,%.2f), U = (%.2f,%.2f),  V = (%.2f,%.2f\n",wx,wy,ax,ay,bx,by) ;
	
	dsq = 0. ;
	nPeaks = 0 ;
	for(i=0;i<peakList->mNumPeaks;i++) {
		x = peakList->mX[i] ;
		y = peakList->mY[i] ;
		fn = (by*(x-wx)-bx*(y-wy))/(ax*by-ay*bx) ;
		fm = (ay*(x-wx)-ax*(y-wy))/(ay*bx-ax*by) ;
		n = near(fn) ;
		m = near(fm) ;
		dn = fabs(fn-n) ;
		dm = fabs(fm-m) ;
		if (dn < maxPeakDeviation && dm < maxPeakDeviation) {
			dx = x-(wx+n*ax+m*bx) ;
			dy = y-(wy+n*ay+m*by) ;
			dsq += dx*dx+dy*dy ;
			nPeaks++ ;
		}
	}
	if (nPeaks > 0)
		dsq = sqrt(dsq)/nPeaks ;	// RMS deviation
	else
		dsq = 0. ;

	//fprintf(stderr,"Lattice RMS = %f\n",dsq) ;
	return dsq ;
}	

Boolean TLattice::InterceptY0(long Width,long Height,int n,Point *iSect,AxisType whichAxis) {
	float ax,bx,ay,by,wx,wy ;
	float x ;
	float height,width ;
	Boolean intercepts ;
	
	height = Height ;
	width = Width ;
	
	ax = mU.x ;
	ay = mU.y ;
	bx = mV.x ;
	by = mV.y ;
	wx = mOrigin.x ;
	wy = mOrigin.y ;
	intercepts = false ;
	
	switch(whichAxis) {
		case kXaxis:
			x = wx+n*ax ;
			if (bx == 0.) {
				if (x >= 0 && x < width-1) {
					intercepts = true ;
					iSect->x = near(x) ;
					iSect->y = 0 ;
				}
			}
			else if (by != 0.) { 
				x -= bx*(wy + n*ay)/by ;
				if (x >= 0 && x < width-1) {
					intercepts = true ;
					iSect->x = near(x) ;
					iSect->y = 0 ;
				}
			}
			break ;
		case kYaxis:
			x = wx+n*bx ;
			if (ax == 0.) {
				if (x >= 0 && x < width-1) {
					intercepts = true ;
					iSect->x = near(x) ;
					iSect->y = 0 ;
				}
			}
			else if (ay != 0.) { 
				x -= ax*(wy + n*by)/ay ;
				if (x >= 0 && x < width-1) {
					intercepts = true ;
					iSect->x = near(x) ;
					iSect->y = 0 ;
				}
			}
			break ;
	}
	return intercepts ;
}

Boolean TLattice::InterceptYH(long Width,long Height,int n,Point *iSect,AxisType whichAxis) {
	float ax,bx,ay,by,wx,wy ;
	float x ;
	float height,width ;
	Boolean intercepts ;
	
	height = Height ;
	width = Width ;
	
	ax = mU.x ;
	ay = mU.y ;
	bx = mV.x ;
	by = mV.y ;
	wx = mOrigin.x ;
	wy = mOrigin.y ;
	intercepts = false ;
	
	switch(whichAxis) {
		case kXaxis:
			x = wx+n*ax ;
			if (bx == 0.) {
				if (x >= 0 && x < width-1) {
					intercepts = true ;
					iSect->x = near(x) ;
					iSect->y = Height-1 ;
				}
			}
			else if (by != 0.) { 
				x -= bx*(wy + n*ay - height +1)/by ;
				if (x >= 0 && x < width-1) {
					intercepts = true ;
					iSect->x = near(x) ;
					iSect->y = Height-1 ;
				}
			}
			break ;
		case kYaxis:
			x = wx+n*bx ;
			if (ax == 0.) {
				if (x >= 0 && x < width-1) {
					intercepts = true ;
					iSect->x = near(x) ;
					iSect->y = Height-1 ;
				}
			}
			else if (ay != 0.) { 
				x -= ax*(wy + n*by - height +1)/ay ;
				if (x >= 0 && x < width-1) {
					intercepts = true ;
					iSect->x = near(x) ;
					iSect->y = Height-1 ;
				}
			}
			break ;
	}
	return intercepts ;
}

Boolean TLattice::InterceptX0(long Width,long Height,int n,Point *iSect,AxisType whichAxis) {
	float ax,bx,ay,by,wx,wy ;
	float y ;
	float height,width ;
	Boolean intercepts ;
	
	height = Height ;
	width = Width ;
	
	ax = mU.x ;
	ay = mU.y ;
	bx = mV.x ;
	by = mV.y ;
	wx = mOrigin.x ;
	wy = mOrigin.y ;
	intercepts = false ;
	
	switch(whichAxis) {
		case kXaxis:
			if (bx != 0.) {
				y = wy+n*ay -by/bx*(wx+n*ax) ;
				if (y >= 0 && y < height-1) {
					intercepts = true ;
					iSect->x = 0 ;
					iSect->y = near(y) ;
				}
			}
			break ;
		case kYaxis:
			if (ax != 0.) {
				y = wy+n*by - ay/ax*(wx+n*bx) ;
				if (y >= 0 && y < height-1) {
					intercepts = true ;
					iSect->x = 0 ;
					iSect->y = near(y) ;
				}
			}
			break ;
	}
	return intercepts ;
}

Boolean TLattice::InterceptXH(long Width,long Height,int n,Point *iSect,AxisType whichAxis) {
	float ax,bx,ay,by,wx,wy ;
	float y ;
	float height,width ;
	Boolean intercepts ;
	
	height = Height ;
	width = Width ;
	
	ax = mU.x ;
	ay = mU.y ;
	bx = mV.x ;
	by = mV.y ;
	wx = mOrigin.x ;
	wy = mOrigin.y ;
	intercepts = false ;
	
	switch(whichAxis) {
		case kXaxis:
			if (bx != 0.) {
				y = by/bx*(width-1) + wy+n*ay -by/bx*(wx+n*ax) ;
				if (y >= 0 && y < height-1) {
					intercepts = true ;
					iSect->x = width-1 ;
					iSect->y = near(y) ;
				}
			}
			break ;
		case kYaxis:
			if (ax != 0.) {
				y = ay/ax*(width-1) + wy+n*by - ay/ax*(wx+n*bx) ;
				if (y >= 0 && y < height-1) {
					intercepts = true ;
					iSect->x = width-1 ;
					iSect->y = near(y) ;
				}
			}
			break ;
	}
	return intercepts ;
}

void TLattice::DrawUnitVectors(TImage *image) {
	SetFgColor(gGreenPixel) ;
	MoveTo(near(mOrigin.x),near(mOrigin.y)) ;
	LineTo(near(mOrigin.x+mU.x),near(mOrigin.y+mU.y)) ;
	MoveTo(near(mOrigin.x),near(mOrigin.y)) ;
	LineTo(near(mOrigin.x+mV.x),near(mOrigin.y+mV.y)) ;
	LineTo(near(mOrigin.x+mV.x+mU.x),near(mOrigin.y+mV.y+mU.y)) ;
	LineTo(near(mOrigin.x+mU.x),near(mOrigin.y+mU.y)) ;
	DrawText(near(mOrigin.x)-3,near(mOrigin.y)-4,"O") ;
	DrawText(near(mOrigin.x+mU.x)-3,near(mOrigin.y+mU.y)-4,"U") ;
	DrawText(near(mOrigin.x+mV.x)-3,near(mOrigin.y+mV.y)-4,"V") ;


	// Draw arrows
	double	k,b,x0,y0,x1,x2,y1,y2,A,B,L,d,a1,a2,a3,a4,xw1,yw1,xw2,yw2,w ;
	float	tol = 0.0001 ;
	w = 4 ;
	d = 8 ;
	x1 = mOrigin.x ;
	y1 = mOrigin.y ;

	x2 = mOrigin.x + mU.x ;
	y2 = mOrigin.y + mU.y ;
	L = sqrt((x1-x2)*(x1-x2)+(y1-y2)*(y1-y2)) ;
	if (fabs(mU.x) >  tol) {
		k = (y2-y1)/(x2-x1) ;
		b = (y1*x2 - x1*y2)/(x2-x1) ;
		if (k != 0.) {
			A = k*(y2-y1) + (x2-x1) ;
			B = x2*x2+y2*y2-x1*x2-y1*y2-b*(y2-y1)-L*d ;
			x0 = B/A ;
			y0 = k*x0+b ;
			a1 = 1 + k*k ;
			a2 = -2*a1*x0 ;
			a3 = a1*x0*x0 - k*k*w*w ;
			a4 = a2*a2-4*a1*a3 ;
			if (a4 <= 0.) {
				xw1 = -a2/(2*a1) ;
				xw2 = xw1 ;
			}
			else {
				xw1 = (-a2 + sqrt(a4))/(2*a1) ;
				xw2 = (-a2 - sqrt(a4))/(2*a1) ;
			} 
			yw1 = (-1./k)*xw1 + y0 + (1./k)*x0 ;
			yw2 = (-1./k)*xw2 + y0 + (1./k)*x0 ;
		}
		else {
			// Line is parallel to the x-axis
			xw1 = xw2 = (x2*x2-x1*x2-L*d)/(x2-x1) ;
			yw1 = mOrigin.y+w ;
			yw2 = mOrigin.y-w ;
		}
	}
	else {
		// Line is parallel to the y-axis
		xw1 = mOrigin.x+w ;
		xw2 = mOrigin.x-w ;
		yw1 = yw2 = (y2*y2-y1*y2-L*d)/(y2-y1) ;
	}

	MoveTo(near(x2),near(y2)) ;
	LineTo(near(xw1),near(yw1)) ;
	MoveTo(near(x2),near(y2)) ;
	LineTo(near(xw2),near(yw2)) ;
		
	x2 = mOrigin.x + mV.x ;
	y2 = mOrigin.y + mV.y ;
	L = sqrt((x1-x2)*(x1-x2)+(y1-y2)*(y1-y2)) ;
	if (fabs(mV.x) >  tol) {
		k = (y2-y1)/(x2-x1) ;
		b = (y1*x2 - x1*y2)/(x2-x1) ;
		if (k != 0.) {
			A = k*(y2-y1) + (x2-x1) ;
			B = x2*x2+y2*y2-x1*x2-y1*y2-b*(y2-y1)-L*d ;
			x0 = B/A ;
			y0 = k*x0+b ;
			a1 = 1 + k*k ;
			a2 = -2*a1*x0 ;
			a3 = a1*x0*x0 - k*k*w*w ;
			a4 = a2*a2-4*a1*a3 ;
			if (a4 <= 0.) {
				xw1 = -a2/(2*a1) ;
				xw2 = xw1 ;
			}
			else {
				xw1 = (-a2 + sqrt(a4))/(2*a1) ;
				xw2 = (-a2 - sqrt(a4))/(2*a1) ;
			} 
			yw1 = (-1./k)*xw1 + y0 + (1./k)*x0 ;
			yw2 = (-1./k)*xw2 + y0 + (1./k)*x0 ;
		}
		else {
			// Line is parallel to the x-axis
			xw1 = xw2 = (x2*x2-x1*x2-L*d)/(x2-x1) ;
			yw1 = mOrigin.y+w ;
			yw2 = mOrigin.y-w ;
		}
	}
	else {
		// Line is parallel to the y-axis
		xw1 = mOrigin.x+w ;
		xw2 = mOrigin.x-w ;
		yw1 = yw2 = (y2*y2-y1*y2-L*d)/(y2-y1) ;
	}
	MoveTo(near(x2),near(y2)) ;
	LineTo(near(xw1),near(yw1)) ;
	MoveTo(near(x2),near(y2)) ;
	LineTo(near(xw2),near(yw2)) ;


	SetFgColor(gBlackPixel) ;
}

void TLattice::Draw2DVectors(TImage *image) {

	long Width,Height ;
	Point endPoint[2],iSect[4] ;
	Boolean interSects[4],draw,withinImage; 
	AxisType whichAxis ;
	long n,m,ndx,i ;
	
	float	uLength,vLength ;
	uLength = sqrt(mU.x*mU.x+mU.y*mU.y) ;		
	vLength = sqrt(mV.x*mV.x+mV.y*mV.y) ;
	if (uLength < 1.) return ;
	if (vLength < 1.) return ;
		
	Width = image->Width() ;
	Height = image->Height() ;

	// Draw the two main axes n == 0 && m == 0
	// The line will (if not parallel to one of the x and y axis)
	// need to intercept one of the lines x=0,x=W-1,y=0,y=H-1
	// within the area of the image

	whichAxis = kXaxis ;	// A or U
	interSects[0] = InterceptY0(Width,Height,0,&iSect[0],whichAxis) ;
	interSects[1] = InterceptYH(Width,Height,0,&iSect[1],whichAxis) ;
	interSects[2] = InterceptX0(Width,Height,0,&iSect[2],whichAxis) ;
	interSects[3] = InterceptXH(Width,Height,0,&iSect[3],whichAxis) ;
		
	ndx = 0 ;
	for(i=0;i<4 && ndx < 2;i++) {
		if (interSects[i]) {
			endPoint[ndx++] = iSect[i] ;
		}
	}
		
	withinImage = (ndx == 2 ? true : false ) ;		
		
	if (withinImage) {
		// Draw the line, but check for null line
		if ((endPoint[0].y != endPoint[1].y) || (endPoint[0].x != endPoint[1].x)) {
			// Draw the line
			DrawLine(endPoint[0].x,endPoint[0].y,endPoint[1].x,endPoint[1].y) ;
		}
	}

	whichAxis = kYaxis ;	// B or V
	interSects[0] = InterceptY0(Width,Height,0,&iSect[0],whichAxis) ;
	interSects[1] = InterceptYH(Width,Height,0,&iSect[1],whichAxis) ;
	interSects[2] = InterceptX0(Width,Height,0,&iSect[2],whichAxis) ;
	interSects[3] = InterceptXH(Width,Height,0,&iSect[3],whichAxis) ;
		
	ndx = 0 ;
	for(i=0;i<4 && ndx < 2;i++) {
		if (interSects[i]) {
			endPoint[ndx++] = iSect[i] ;
		}
	}
		
	withinImage = (ndx == 2 ? true : false ) ;		
		
	if (withinImage) {
		// Draw the line, but check for null line
		if ((endPoint[0].y != endPoint[1].y) || (endPoint[0].x != endPoint[1].x)) {
			DrawLine(endPoint[0].x,endPoint[0].y,endPoint[1].x,endPoint[1].y) ;
		}
	}


	// Change n in a positive fashion until we fall out of the image
	n = 1 ;
	draw = true ;
	whichAxis = kXaxis ;
	while (draw ) {
		
		// The line will (if not parallel to one of the x and y axis)
		// need to intercept one of the lines x=0,x=W-1,y=0,y=H-1
		// within the area of the image
		interSects[0] = InterceptY0(Width,Height,n,&iSect[0],whichAxis) ;
		interSects[1] = InterceptYH(Width,Height,n,&iSect[1],whichAxis) ;
		interSects[2] = InterceptX0(Width,Height,n,&iSect[2],whichAxis) ;
		interSects[3] = InterceptXH(Width,Height,n,&iSect[3],whichAxis) ;
		
		ndx = 0 ;
		for(i=0;i<4 && ndx < 2;i++) {
			if (interSects[i]) {
				endPoint[ndx++] = iSect[i] ;
			}
		}
		
		withinImage = (ndx == 2 ? true : false ) ;		
		
		if (withinImage) {
			// Draw the line, but check for null line
			if ((endPoint[0].y != endPoint[1].y) || (endPoint[0].x != endPoint[1].x)) {
				// We should place markers on the m == 0 line
				//DrawLine(endPoint[0].x,endPoint[0].y,endPoint[1].x,endPoint[1].y) ;
			}
		}
		else
			draw = false ;
		n++ ;
	} ;
	
	// Change n in a negative fashion until we fall out of the image
	n = -1 ;
	draw = true ;
	whichAxis = kXaxis ;
	while (draw) {
		interSects[0] = InterceptY0(Width,Height,n,&iSect[0],whichAxis) ;
		interSects[1] = InterceptYH(Width,Height,n,&iSect[1],whichAxis) ;
		interSects[2] = InterceptX0(Width,Height,n,&iSect[2],whichAxis) ;
		interSects[3] = InterceptXH(Width,Height,n,&iSect[3],whichAxis) ;
		
		ndx = 0 ;
		for(i=0;i<4 && ndx < 2;i++) {
			if (interSects[i]) {
				endPoint[ndx++] = iSect[i] ;
			}
		}
		
		withinImage = (ndx == 2 ? true : false ) ;		
		
		if (withinImage) {
			// Draw the line, but check for null line
			if ((endPoint[0].y != endPoint[1].y) || (endPoint[0].x != endPoint[1].x)) {
				// We should only draw markers on the m == 0 line
				//DrawLine(endPoint[0].x,endPoint[0].y,endPoint[1].x,endPoint[1].y) ;
			}
		}
		else
			draw = false ;
		n-- ;
	} ;

	// Change m in a positive fashion until we fall out of the image
	m = 1 ;
	draw = true ;
	whichAxis = kYaxis ;
	while (draw) {
		interSects[0] = InterceptY0(Width,Height,m,&iSect[0],whichAxis) ;
		interSects[1] = InterceptYH(Width,Height,m,&iSect[1],whichAxis) ;
		interSects[2] = InterceptX0(Width,Height,m,&iSect[2],whichAxis) ;
		interSects[3] = InterceptXH(Width,Height,m,&iSect[3],whichAxis) ;
		
		ndx = 0 ;
		for(i=0;i<4 && ndx < 2;i++) {
			if (interSects[i]) {
				endPoint[ndx++] = iSect[i] ;
			}
		}
		
		withinImage = (ndx == 2 ? true : false ) ;		
		
		if (withinImage) {
			// Draw the line, but check for null line
			if ((endPoint[0].y != endPoint[1].y) || (endPoint[0].x != endPoint[1].x)) {
				//DrawLine(endPoint[0].x,endPoint[0].y,endPoint[1].x,endPoint[1].y) ;
			}
		}
		else
			draw = false ;
		m++ ;
	} ;
	
	// Change m in a negative fashion until we fall out of the image
	m = -1 ;
	draw = true ;
	whichAxis = kYaxis ;
	while (draw) {
		interSects[0] = InterceptY0(Width,Height,m,&iSect[0],whichAxis) ;
		interSects[1] = InterceptYH(Width,Height,m,&iSect[1],whichAxis) ;
		interSects[2] = InterceptX0(Width,Height,m,&iSect[2],whichAxis) ;
		interSects[3] = InterceptXH(Width,Height,m,&iSect[3],whichAxis) ;
		
		ndx = 0 ;
		for(i=0;i<4 && ndx < 2;i++) {
			if (interSects[i]) {
				endPoint[ndx++] = iSect[i] ;
			}
		}
		
		withinImage = (ndx == 2 ? true : false ) ;		
		
		if (withinImage) {
			// Draw the line, but check for null line
			if ((endPoint[0].y != endPoint[1].y) || (endPoint[0].x != endPoint[1].x)) {
				//DrawLine(endPoint[0].x,endPoint[0].y,endPoint[1].x,endPoint[1].y) ;
			}
		}
		else
			draw = false ;
		m-- ;
	} ;

}

void TLattice::Draw2DGrid(TImage *image) {
	long Width,Height ;
	Point endPoint[2],iSect[4] ;
	Boolean interSects[4],draw,withinImage; 
	AxisType whichAxis ;
	long n,m,ndx,i, n1, m1 ;
	float tx, ty ;
	char label[8] ;
	int xlat , ylat ;
	
	float	uLength,vLength ;
	uLength = sqrt(mU.x*mU.x+mU.y*mU.y) ;		
	vLength = sqrt(mV.x*mV.x+mV.y*mV.y) ;
	if (uLength < 1.) return ;
	if (vLength < 1.) return ;

	Width = image->Width() ;
	Height = image->Height() ;
	xlat = ReturnXTics () ;
	ylat = ReturnYTics () ;
	//fprintf(stderr,"\nLattice %d %d",xlat,ylat) ;

	SetFgColor(gRedPixel) ;

	// Change n in a positive fashion until we fall out of the image
	n = 0 ;
	draw = true ;
	whichAxis = kXaxis ;
	if( ReturnQuadFit() && QuadSet ) {

	SetFgColor(gRedPixel) ;
	for (n=-30; n<31 ; n++){
		for (m=-30 ; m<31 ; m++) {
			n1=n+1 ;
			endPoint[0].x=RealXYValueQ( n, m, BXHold) ; 
			endPoint[0].y=RealXYValueQ( n, m, BYHold) ; 
			endPoint[1].x=RealXYValueQ( n1,m, BXHold) ; 
			endPoint[1].y=RealXYValueQ( n1,m, BYHold) ; 
			DrawLine(endPoint[0].x,endPoint[0].y,endPoint[1].x,endPoint[1].y) ;
		}
	}

	SetFgColor(gBluePixel) ;
	for (n=-30; n<31 ; n++){
		for (m=-30 ; m<31 ; m++) {
		    m1=m+1 ;
			endPoint[0].x=RealXYValueQ( n, m, BXHold) ; 
			endPoint[0].y=RealXYValueQ( n, m, BYHold) ; 
			endPoint[1].x=RealXYValueQ( n, m1,BXHold) ; 
			endPoint[1].y=RealXYValueQ( n, m1,BYHold) ;
			DrawLine(endPoint[0].x,endPoint[0].y,endPoint[1].x,endPoint[1].y) ;
		}
	}
        // LDM Addition Annotate some of the peaks
       SetFgColor(gGreenPixel) ;
	   if (xlat > 0 || ylat > 0 ) {
		int xnmax=near(60/xlat) ; int ynmax=near(60/ylat) ;
		for (n1=-xnmax; n1<xnmax ; n1++){
               for (m1=-ynmax ; m1<ynmax ; m1++) {
                       n = xlat*n1 ;
                       m = ylat*m1 ;
                       sprintf(label,"%d,%d",n,m) ;
                       tx=RealXYValueQ(n, m, BXHold) ;
                       ty=RealXYValueQ(n, m, BYHold) ;
                       DrawText(near(tx-8),near(ty-8),label) ;
                        }
                }
	   }
	}
	else if( ReturnCubicFit() && CubeSet) {
	SetFgColor(gRedPixel) ;
	for (n=-30; n<31 ; n++){
		for (m=-30 ; m<31 ; m++) {
			n1=n+1 ;
			endPoint[0].x=RealXYValueC( n, m, BXHold) ; 
			endPoint[0].y=RealXYValueC( n, m, BYHold) ; 
			endPoint[1].x=RealXYValueC( n1,m, BXHold) ; 
			endPoint[1].y=RealXYValueC( n1,m, BYHold) ; 
			DrawLine(endPoint[0].x,endPoint[0].y,endPoint[1].x,endPoint[1].y) ;
		}
	}

	SetFgColor(gBluePixel) ;
	for (n=-30; n<31 ; n++){
		for (m=-30 ; m<31 ; m++) {
		    m1=m+1 ;
			endPoint[0].x=RealXYValueC( n, m, BXHold) ; 
			endPoint[0].y=RealXYValueC( n, m, BYHold) ; 
			endPoint[1].x=RealXYValueC( n ,m1,BXHold) ; 
			endPoint[1].y=RealXYValueC( n ,m1,BYHold) ; 
			DrawLine(endPoint[0].x,endPoint[0].y,endPoint[1].x,endPoint[1].y) ;
		}
	}
        // LDM Addition Annotate some of the peaks
       SetFgColor(gGreenPixel) ;
	   if (xlat > 0 || ylat > 0 ) {	
		int xnmax=near(60/xlat) ; int ynmax=near(60/ylat) ;
		for (n1=-xnmax; n1<xnmax ; n1++){
               for (m1=-ynmax ; m1<ynmax ; m1++) {
					   n = xlat*n1 ;
                       m = ylat*m1 ;
                       sprintf(label,"%d,%d",n,m) ;
                       tx=RealXYValueC(n, m, BXHold) ;
                       ty=RealXYValueC(n, m, BYHold) ;
                       DrawText(near(tx-8),near(ty-8),label) ;
                        }
                }
	   }
	}
	else {

	while (draw ) {
		
		// The line will (if not parallel to one of the x and y axis)
		// need to intercept one of the lines x=0,x=W-1,y=0,y=H-1
		// within the area of the image
		interSects[0] = InterceptY0(Width,Height,n,&iSect[0],whichAxis) ;
		interSects[1] = InterceptYH(Width,Height,n,&iSect[1],whichAxis) ;
		interSects[2] = InterceptX0(Width,Height,n,&iSect[2],whichAxis) ;
		interSects[3] = InterceptXH(Width,Height,n,&iSect[3],whichAxis) ;
		
		ndx = 0 ;
		for(i=0;i<4 && ndx < 2;i++) {
			if (interSects[i]) {
				endPoint[ndx++] = iSect[i] ;
			}
		}
		
		withinImage = (ndx == 2 ? true : false ) ;		
		
		if (withinImage) {
			// Draw the line, but check for null line
			if ((endPoint[0].y != endPoint[1].y) || (endPoint[0].x != endPoint[1].x)) {
				// Draw the line
				DrawLine(endPoint[0].x,endPoint[0].y,endPoint[1].x,endPoint[1].y) ;
			}
		}
		else
			draw = false ;
		n++ ;
	} ;
	
	// Change n in a negative fashion until we fall out of the image
	n = -1 ;
	draw = true ;
	whichAxis = kXaxis ;
	while (draw) {
		interSects[0] = InterceptY0(Width,Height,n,&iSect[0],whichAxis) ;
		interSects[1] = InterceptYH(Width,Height,n,&iSect[1],whichAxis) ;
		interSects[2] = InterceptX0(Width,Height,n,&iSect[2],whichAxis) ;
		interSects[3] = InterceptXH(Width,Height,n,&iSect[3],whichAxis) ;
		
		ndx = 0 ;
		for(i=0;i<4 && ndx < 2;i++) {
			if (interSects[i]) {
				endPoint[ndx++] = iSect[i] ;
			}
		}
		
		withinImage = (ndx == 2 ? true : false ) ;		
		
		if (withinImage) {
			// Draw the line, but check for null line
			if ((endPoint[0].y != endPoint[1].y) || (endPoint[0].x != endPoint[1].x)) {
				DrawLine(endPoint[0].x,endPoint[0].y,endPoint[1].x,endPoint[1].y) ;
			}
		}
		else
			draw = false ;
		n-- ;
	} ;

	SetFgColor(gBluePixel) ;

	// Change m in a positive fashion until we fall out of the image
	m = 0 ;
	draw = true ;
	whichAxis = kYaxis ;
	while (draw) {
		interSects[0] = InterceptY0(Width,Height,m,&iSect[0],whichAxis) ;
		interSects[1] = InterceptYH(Width,Height,m,&iSect[1],whichAxis) ;
		interSects[2] = InterceptX0(Width,Height,m,&iSect[2],whichAxis) ;
		interSects[3] = InterceptXH(Width,Height,m,&iSect[3],whichAxis) ;
		
		ndx = 0 ;
		for(i=0;i<4 && ndx < 2;i++) {
			if (interSects[i]) {
				endPoint[ndx++] = iSect[i] ;
			}
		}
		
		withinImage = (ndx == 2 ? true : false ) ;		
		
		if (withinImage) {
			// Draw the line, but check for null line
			if ((endPoint[0].y != endPoint[1].y) || (endPoint[0].x != endPoint[1].x)) {
				DrawLine(endPoint[0].x,endPoint[0].y,endPoint[1].x,endPoint[1].y) ;
			}
		}
		else
			draw = false ;
		m++ ;
	} ;
	
	// Change m in a negative fashion until we fall out of the image
	m = -1 ;
	draw = true ;
	whichAxis = kYaxis ;
	while (draw) {
		interSects[0] = InterceptY0(Width,Height,m,&iSect[0],whichAxis) ;
		interSects[1] = InterceptYH(Width,Height,m,&iSect[1],whichAxis) ;
		interSects[2] = InterceptX0(Width,Height,m,&iSect[2],whichAxis) ;
		interSects[3] = InterceptXH(Width,Height,m,&iSect[3],whichAxis) ;
		
		ndx = 0 ;
		for(i=0;i<4 && ndx < 2;i++) {
			if (interSects[i]) {
				endPoint[ndx++] = iSect[i] ;
			}
		}
		
		withinImage = (ndx == 2 ? true : false ) ;		
		
		if (withinImage) {
			// Draw the line, but check for null line
			if ((endPoint[0].y != endPoint[1].y) || (endPoint[0].x != endPoint[1].x)) {
				DrawLine(endPoint[0].x,endPoint[0].y,endPoint[1].x,endPoint[1].y) ;
			}
		}
		else
			draw = false ;
		m-- ;
	} 
        // LDM Addition Annotate some of the peaks
       SetFgColor(gGreenPixel) ;
	   if (xlat > 0 || ylat > 0 ) {	
		int xnmax=near(60/xlat) ; int ynmax=near(60/ylat) ;
		for (n1=-xnmax; n1<xnmax ; n1++){
               for (m1=-ynmax ; m1<ynmax ; m1++) {
					   n = xlat*n1 ;
                       m = ylat*m1 ;
                       sprintf(label,"%d,%d",n,m) ;
                       tx=mOrigin.x+n*mU.x+m*mV.x;
                       ty=mOrigin.y+n*mU.y+m*mV.y;
                       DrawText(near(tx-8),near(ty-8),label) ;
                        }
                }
	   }
	}
	SetFgColor(gBlackPixel) ;

	DrawUnitVectors(image) ;

}


void FindEquivalentUnitCellChoices(TLattice *theLattice) ;
void trace2D(float ux,float uy,float vx,float vy,float ap[8][2][2], long &nn) ;
void psymcheck(float a[2][2], long ism[3]) ;
void mat2mult(float a[2][2],float b[2][2],float c[2][2]) ;

class	TLatticeList : public TList {
	public :
		TLatticeList(Rect	*theListRect,short	dataSize) : TList(theListRect,dataSize) {} ;
		void 	DrawCellValue(ListCellT &inCell,Rect &cellFrame) ;
} ;

class	TSelectLatticeDialog : public TDialog {
	public:
		TSelectLatticeDialog(short xloc,short yloc,short width,short height,char *title = nil) ;
		void	AddLatticeChoice(TLattice *latticeChoice) ;
		void	SelectLattice(short i) { mLatticeList->SelectRow(i+1) ; } ;
		short	GetSelectedLattice() { return mLatticeList->GetSelectedRow() - 1 ; } ;
		
		TLatticeList	*mLatticeList ;
		
} ;

void TLatticeList::DrawCellValue(
	ListCellT &inCell,
	Rect &cellFrame)
{
	char str[256] ;
	TLattice latticeChoice ;
	FPoint	u,v ;
		
	TextSize(0);
	FetchCellData(inCell,&latticeChoice) ;
	u = latticeChoice.GetU() ;
	v = latticeChoice.GetV() ;


	MoveTo(cellFrame.left + 4, cellFrame.bottom - 4);
	sprintf(str,"%.2f,%.2f",u.x,u.y) ;
	DrawString(str) ;

	MoveTo(cellFrame.left + 75, cellFrame.bottom - 4);
	sprintf(str,"%.2f,%.2f",v.x,v.y) ;
	DrawString(str) ;

	MoveTo(cellFrame.left + 150, cellFrame.bottom - 4);
	sprintf(str,"%.2f",latticeChoice.GetLatticeParameterU()) ;
	DrawString(str) ;

	MoveTo(cellFrame.left + 195, cellFrame.bottom - 4);
	sprintf(str,"%.2f",latticeChoice.GetLatticeParameterV()) ;
	DrawString(str) ;

	MoveTo(cellFrame.left + 240, cellFrame.bottom - 4);
	sprintf(str,"%.2f",latticeChoice.GetLatticeAngle()) ;
	DrawString(str) ;

}

TSelectLatticeDialog::TSelectLatticeDialog(short xloc,short yloc,short width,short height,char *title)
		: TDialog(xloc,yloc,width,height,title) {

		SetPort(this) ;
		
		TStaticText *theText = new TStaticText(20,10,390,30,"Alternate settings for the reciprocal lattice have been found!") ;
		AddObject(theText) ;
		theText = new TStaticText(20,30,390,50,"Please choose which to use. Your original lattice is the first entry.") ;
		AddObject(theText) ;

		theText = new TStaticText(32,57,105,73,"ux,uy") ;
		AddObject(theText) ;

		theText = new TStaticText(105,57,175,73,"vx,vy") ;
		AddObject(theText) ;

		theText = new TStaticText(175,57,220,73,"|a*|") ;
		AddObject(theText) ;

		theText = new TStaticText(220,57,260,73,"|b*|") ;
		AddObject(theText) ;

		theText = new TStaticText(260,57,300,73,"Gamma*") ;
		AddObject(theText) ;
		
		TButton	*theButton = new TButton(326,135,420,155,"Select") ;
		AddObject(theButton) ;
		SetOKButton(theButton) ;
		SetDefaultButton(theButton) ;
				
		Rect	theRect ;
		SetRect(&theRect,20,72,310,155) ;
		
		mLatticeList = new TLatticeList(&theRect,sizeof(TLattice)) ;
		AddObject(mLatticeList) ;


}


void	TSelectLatticeDialog::AddLatticeChoice(TLattice *latticeChoice) {

	mLatticeList->AddRow(latticeChoice) ;	
	
}

void FindEquivalentUnitCellChoices(TLattice *theLattice) {
    float 	ap[8][2][2],cc[2][2] ;
    long 	nn,i ;
    short	whichlattice ;
    
    FPoint	u = theLattice->GetU() ;
    FPoint	v = theLattice->GetV() ;
    FPoint u0 = u ;
    FPoint v0 = v ;

	trace2D(u.x,u.y,v.x,v.y,ap,nn) ;
			
	if (nn > 0) {
	
		// Show alternate unit cell choices and have the user pick the one he/she wants
		
		TSelectLatticeDialog *theDialog = new TSelectLatticeDialog(kDialogPosX,kDialogPosY,440,190,"Lattice") ;
		
		TLattice *latticeChoice[8] = {nil,nil,nil,nil,nil,nil,nil,nil} ; 
		
		latticeChoice[0] = new TLattice(theLattice) ;		
		theDialog->AddLatticeChoice(latticeChoice[0]) ;
		
		for(i = 0; i < nn; i++) {

	        u.x = cc[0][0] = ap[i][0][0] ;
	        u.y = cc[0][1] = ap[i][0][1] ;
	        v.x = cc[1][0] = ap[i][1][0] ;
	        v.y = cc[1][1] = ap[i][1][1] ; 
	        
			latticeChoice[i+1] = new TLattice(theLattice->GetOrigin(),u,v) ;

	        /*
	        psymcheck(cc,ism) ;
			fprintf(stderr,"\nAlternate #%d",i+1) ;
			
			if (ism[0] == 1) {
				fprintf(stderr,"\n\tgamma* = 90 Degrees") ;
			}
			else if (ism[2] == 1) {
				fprintf(stderr,"\n\tgamma* = 60 Degrees") ;
			}
			if (ism[1] == 1) {
				fprintf(stderr,"\n\t|a*| = |b*| = %.4f",sqrt(u.x*u.x+u.y*u.y)) ;
			}
			else {
				fprintf(stderr,"\n\t|a*| = %.4f",sqrt(u.x*u.x+u.y*u.y)) ;
				fprintf(stderr,"\n\t|b*| = %.4f",sqrt(v.x*v.x+v.y*v.y)) ;
			}
	        fprintf(stderr,"\n\ta* = (%12.6f,%12.6f)",u.x,u.y) ;
	        fprintf(stderr,"\n\tb* = (%12.6f,%12.6f)",v.x,v.y) ;
	        fprintf(stderr,"\n") ;
	        */

			theDialog->AddLatticeChoice(latticeChoice[i+1]) ;

		}		
		
		theDialog->SelectLattice(0) ;
		
		if (theDialog->DoModalDialog()) {
			whichlattice = theDialog->GetSelectedLattice() ;
			if (whichlattice >= 0 && whichlattice <= nn)
				*theLattice = *latticeChoice[theDialog->GetSelectedLattice()] ;
			delete theDialog ;
		}
		else 
			delete theDialog ;

		for(i=0; i < nn+1; i++) delete latticeChoice[i] ;
			
	}
	
}

// -----------------------------------------------------------------------------
//  Copyright (c) 2004 Wharton Sinkler
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------

// -------------------------------------------------------------------------------------------------
// checks for alternate unit cells (2D) with higher symmetry.   
// Sent: ux,uy,vx,vy (origin and points defining end of u and v basis 
// vectors of the diffraction pattern - all cartesian, e.g. pixel coordinates relative to origin).   

// returns: array ap(2,2,nnmax), i.e. a bunch of 2 x 2 arrays whose columns are cartesian  
// vectors giving new representations of u and v.  The long nn indicates how many are found, i.e. 
// how many new unit basis vector sets we should give the user the choice of.  
// 
// To determine whether the symmetry is higher or not it sets integers in array ism(3) in which 
// ism(1)=1 if the angle is 90 degrees (to within tolerances specified in psymcheck). 
// Ism(2)=1 if the cell is equiaxed, and ism(3)=1 if the angle is 60 degrees. 
// -------------------------------------------------------------------------------------------------
void trace2D(float ux,float uy,float vx,float vy,float ap[8][2][2], long &nn) {

	float 	a0[2][2],bb[2][2],cc[2][2] ;
	long 	ism0[3],ism[3],m,j ;
    float	pi ;

	// convert to a matrix representation: 

	a0[0][0] = ux ;
	a0[1][0] = uy ;
	a0[0][1] = vx ;
	a0[1][1] = vy ;

	psymcheck(a0,ism0) ;

	pi =  3.14159274 ;

	// shear along y direction and check for consistency: 
	nn = 0 ;

	bb[0][0] = 1. ;
	bb[0][1] = 0. ;
	bb[1][1] = 1. ;
	for(m = -8; m <= 8; m++) {
		if(m == 0) continue ;
		bb[1][0] = m ;
		// do the transform (shear) to get cc:  
		mat2mult(a0,bb,cc) ;
		// check of possible symmetries based on cell constants in cc.  returns ism (see psymcheck).   
		psymcheck(cc,ism) ;

		// now check if c* represents possibly higher symmetry cell than a0: 
		for(j=0; j < 3; j++) {
			if(ism[j] > ism0[j]) {
			    ap[nn][0][0] = cc[0][0] ;
			    ap[nn][0][1] = cc[0][1] ;
			    ap[nn][1][0] = cc[1][0] ;
			    ap[nn][1][1] = cc[1][1] ;
			    nn++ ;
		        break ;
			}
		}
	}

	bb[1][0] = 0. ;
	for(m = -8; m <= 8; m++) {
        if(m == 0) continue ;
        bb[0][1] = m ;
		// do the transform (shear) to get cc:  ;
        mat2mult(a0,bb,cc) ;
		// check of possible symmetries based on cell constants in cc.  returns ism (see psymcheck).   
        psymcheck(cc,ism) ;

		// now check if c// represents possibly higher symmetry cell than a0: 
		for(j=0; j < 3; j++) {
			if(ism[j] > ism0[j]) {
			    ap[nn][0][0] = cc[0][0] ;
			    ap[nn][0][1] = cc[0][1] ;
			    ap[nn][1][0] = cc[1][0] ;
			    ap[nn][1][1] = cc[1][1] ;
			    nn++ ;
		        break ;
	        }
		}
	}
}

// -------------------------------------------------------------------------------------------------
// check for compatibility with rectangular, square and hexagonal grids: 
// return long array ism, which indicates rectangular if ism(1) = 1, square if  
// ism(2) = 1 and hexagonal if ism(3) = 1.   
// -------------------------------------------------------------------------------------------------

void psymcheck(float a[2][2], long ism[3]) {

	float 	pi,tol1,tol2,gam,prod,xm1,xm2 ;
	Boolean equiax ;

	pi =  3.14159274	 ;

	// allow error of 2 degrees on angle and 2% on ratio: 
	tol1 = 2./180*pi ;
	tol2 = 0.02 ;

	equiax = false ;
	ism[0] = 0 ;
	ism[1] = 0 ;
	ism[2] = 0 ;

	xm1 = sqrt(a[0][0]*a[0][0]+a[1][0]*a[1][0]) ;
	xm2 = sqrt(a[0][1]*a[0][1]+a[1][1]*a[1][1]) ;

	prod = a[0][0]*a[0][1]+a[1][0]*a[1][1] ;

	gam = acos(prod/(xm1*xm2)) ;

	// condition for rectangular: gam = 90 degrees 
    if(fabs(gam-pi/2.) < tol1)ism[0] = 1 ;

	// condition for square: gam = 90 degrees and a = b 
	equiax = (fabs((xm1-xm2)/xm1) < tol2) ;

	if(equiax && ism[0] == 1) ism[1] = 1 ;
	if(equiax && fabs(gam-pi/3.) < tol1) ism[2] = 1 ;

}

// -------------------------------------------------------------------------------------------------
// this routine premultiplies a 2x2 matrix b with a matrix a and returns 
// the result a*b = c 
// -------------------------------------------------------------------------------------------------
void mat2mult(float a[2][2],float b[2][2],float c[2][2]) {
	long m ;
	for(m=0; m < 2; m++) {
		c[0][m] = a[0][0]*b[0][m] + a[0][1]*b[1][m] ;
		c[1][m] = a[1][0]*b[0][m] + a[1][1]*b[1][m] ;
	}
}

// Return the true value of x, y for different lattice refinement options
float RealXYValueQ( int h, int k, float BX[13] ) {
	float tx=BX[0]+h*(BX[1]+h*BX[3]) + k*(BX[2]+k*BX[4]+h*BX[5]);
	return tx ;
	   }
float RealXYValueC(int h, int k, float BX[13] ) {
		int hh = h*h ; int kk = k*k ;
		float tx=BX[0]+h*(BX[1]+h*BX[3]) + k*(BX[2]+k*BX[4]+h*BX[5])
				+ hh*(h*BX[6]+k*BX[7]+hh*BX[10])
				+ kk*(h*BX[8]+k*BX[9]+kk*BX[11] + hh*BX[12]) ;
		return tx ;
}

// Return the proper indices h,k for a given x,y
float RealHKValue ( float BX[13], float BY[13], float x, float y ) {
	float fh, fk, rbest ;
	float LX = (BX[1]*BX[1]+BY[1]*BY[1]) ;
	float LY = (BX[2]*BX[2]+BY[2]*BY[2]) ;
	int   h, k, ht, kt;
		fh = (BY[2]*(x-BX[0])-BX[2]*(y-BY[0]))/(BX[1]*BY[2]-BY[1]*BX[2]) ;
		fk = (BY[1]*(x-BX[0])-BX[1]*(y-BY[0]))/(BY[1]*BX[2]-BX[1]*BY[2]) ;
		h = near(fh) ;
		k = near(fk) ;
		rbest = 9999 ; hindex = h ; kindex = k ;
		if( ReturnQuadFit() && QuadSet) {
			for (ht=h-1; ht<h+2; ht++) {
				for (kt=k-1 ; kt<k+2; kt++){
					float xh = RealXYValueQ( ht, kt, BX ) ;
					float yh = RealXYValueQ( ht, kt, BY ) ;
					float res = (xh-x)*(xh-x)+(yh-y)*(yh-y) ;
					if ( res < rbest ) {
							hindex = ht ; 
							kindex = kt ;
							//Fractiohal distahce off via dot products
							fh = ((xh-x)*BX[1]+(yh-y)*BY[1])/LX ;
							fk = ((xh-x)*BX[2]+(yh-y)*BY[2])/LY ;
							rbest = res ;
					}
				}
			}
		}
		else if ( ReturnCubicFit() & CubeSet) {		
			for (ht=h-1; ht<h+2; ht++) {
				for (kt=k-1 ; kt<k+2; kt++){
					float xh = RealXYValueC( ht, kt, BX ) ;
					float yh = RealXYValueC( ht, kt, BY ) ;
					float res = (xh-x)*(xh-x)+(yh-y)*(yh-y) ;
					if ( res < rbest ) {
							hindex = ht ; 
							kindex = kt ;
							//Fractiohal distahce off via dot products
							fh = ((xh-x)*BX[1]+(yh-y)*BY[1])/LX ;
							fk = ((xh-x)*BX[2]+(yh-y)*BY[2])/LY ;
							rbest = res ;
					}
				}
			}
		}
		else {
			rbest = (fh-hindex)*(fh-hindex)+(fk-kindex)*(fk-kindex) ;
		}
	return rbest ;
}
float RealHKValueQ ( float BX[13], float BY[13], float x, float y ) {
	float fh, fk, rbest ;
	//float LX = (BX[1]*BX[1]+BY[1]*BY[1]) ;
	//float LY = (BX[2]*BX[2]+BY[2]*BY[2]) ;
	int   h, k, ht, kt ;
	fh = (BY[2]*(x-BX[0])-BX[2]*(y-BY[0]))/(BX[1]*BY[2]-BY[1]*BX[2]) ;
	fk = (BY[1]*(x-BX[0])-BX[1]*(y-BY[0]))/(BY[1]*BX[2]-BX[1]*BY[2]) ;
	h = near(fh) ;
	k = near(fk) ;
	rbest = 999999.0 ; hindex = h ; kindex = k ;
	for (ht=h-1; ht<h+2; ht++) {
		for (kt=k-1 ; kt<k+2; kt++){
			float xh = RealXYValueQ( ht, kt, BX ) ;
			float yh = RealXYValueQ( ht, kt, BY ) ;
			float res = (xh-x)*(xh-x)+(yh-y)*(yh-y) ;
			if ( res < rbest ) {
					hindex = ht ; 
					kindex = kt ;
					rbest = res ;
			}
		}
	}
	return rbest ;
}
float RealHKValueC ( float BX[13], float BY[13], float x, float y ) {
	float fh, fk, rbest ;
	int   h, k, ht, kt ;
	fh = (BY[2]*(x-BX[0])-BX[2]*(y-BY[0]))/(BX[1]*BY[2]-BY[1]*BX[2]) ;
	fk = (BY[1]*(x-BX[0])-BX[1]*(y-BY[0]))/(BY[1]*BX[2]-BX[1]*BY[2]) ;
	h = near(fh) ;
	k = near(fk) ;
	rbest = 999999.0 ; hindex = h ; kindex = k ;
	for (ht=h-1; ht<h+2; ht++) {
		for (kt=k-1 ; kt<k+2; kt++){
			float xh = RealXYValueC( ht, kt, BX ) ;
			float yh = RealXYValueC( ht, kt, BY ) ;
			float res = (xh-x)*(xh-x)+(yh-y)*(yh-y) ;
			if ( res < rbest ) {
					hindex = ht ; 
					kindex = kt ;
					rbest = res ;
			}
		}
	}
	return rbest ;
}
float RealHKValueS ( float BX[13], float BY[13], float x, float y ) {
	float fh = (BY[2]*(x-BX[0])-BX[2]*(y-BY[0]))/(BX[1]*BY[2]-BY[1]*BX[2]) ;
	float fk = (BY[1]*(x-BX[0])-BX[1]*(y-BY[0]))/(BY[1]*BX[2]-BX[1]*BY[2]) ;
	hindex = near(fh) ;
	kindex = near(fk) ;
	float rbest = (fh-hindex)*(fh-hindex)+(fk-kindex)*(fk-kindex) ;
	return rbest ;
}
