#include "TRuler.h"
#include <math.h>
#include "TImageWindow.h"
#include "globals.h"
inline  int 	SIGN(float x) { return (x < 0. ? -1   : 1) ; }
inline float   	FABS(float a) { return (a < 0. ? -a : a ) ; }
extern  Boolean ShiftKeyDown() ;
long	near(float) ;
//extern	Pattern	gDash ;

extern	void	WriteObjectInfo(char *text) ;

TRuler::TRuler(Point &thePoint,TImageWindow *window)
	 : mIsSelected(true),mWindow(window) {
	 
	mSignature = 'rulr' ;
	// We come in as coordinates in the image
	mEndPoint1.x = thePoint.x ;
	mEndPoint2.x = thePoint.x ;
	mEndPoint1.y = thePoint.y ;
	mEndPoint2.y = thePoint.y ;
	
	Modify(thePoint) ;

}

void TRuler::Modify(Point &location) {
	float	dx,dy,ratio ;
	
	mWindow->ImageToWindowPoint(mEndPoint1) ;
	mWindow->ImageToWindowPoint(mEndPoint2) ;
	mWindow->ImageToWindowPoint(location) ;
	
	Rect	theRect ;
	FPoint 	tempPt ;
	Point	newPoint,lastPoint ;
	
	mIsSelected = false ;
	
	GetMouse(&newPoint) ;
	lastPoint = newPoint ;
	PenMode(kXOrMode) ;
	Draw() ;

	SetRect(&theRect,near(mEndPoint1.x-4),near(mEndPoint1.y-4),near(mEndPoint1.x+4),near(mEndPoint1.y+4)) ;
	if (PtInRect(&location,&theRect)) {
		// Switch the end points wo we don't have to do the loop twice
 		tempPt = mEndPoint1 ;
 		mEndPoint1 = mEndPoint2 ;
 		mEndPoint2 = tempPt ;
	}
	
	SetRect(&theRect,near(mEndPoint2.x-4),near(mEndPoint2.y-4),near(mEndPoint2.x+4),near(mEndPoint2.y+4)) ;
	if (PtInRect(&location,&theRect)) {
		while(Button()) {
			GetMouse(&newPoint) ;
			if (ShiftKeyDown()) {
			    // We need to adjust the line
			    dx = newPoint.x - mEndPoint1.x ;
			    dy = newPoint.y - mEndPoint1.y ;
			    if (dx != 0.) {
			        ratio = fabs(dy/dx) ;
			        if (ratio < 0.5) {
			            if (mEndPoint2.x != newPoint.x) {
			                Draw() ;
			                mEndPoint2.y = mEndPoint1.y ;
			                mEndPoint2.x = newPoint.x ;
					        ShowInfo() ; 
			                Draw() ;
			            }
			        }
			        else if (ratio > 1.5) {
			            if (mEndPoint2.y != newPoint.y) {
			                Draw() ;
			                mEndPoint2.x = mEndPoint1.x ;
			                mEndPoint2.y = newPoint.y ;
					        ShowInfo() ; 
			                Draw() ;
			            }
			        }
			        else {
			            dy = FABS(dx)*SIGN(dy) ;
			            newPoint.x = mEndPoint1.x + dx ;
			            newPoint.y = mEndPoint1.y + dy ;
			            if (mEndPoint2.x != newPoint.x || mEndPoint2.y != newPoint.y) {
			                Draw() ;
			 		        mEndPoint2.x = newPoint.x ;
			 		        mEndPoint2.y = newPoint.y ;
			                Draw() ;
					        ShowInfo() ; 
			            }
			        }
			    }
			    else {
			        if (mEndPoint2.y != newPoint.y) {
			            Draw() ;
			            mEndPoint2.x = mEndPoint1.x ;
			            mEndPoint2.y = newPoint.y ;
			            Draw() ;
			            ShowInfo() ;
			        }
			    }
			}
			else {
			 	if (mEndPoint2.x != newPoint.x || mEndPoint2.y != newPoint.y) {
			 		Draw() ;
			 		mEndPoint2.x = newPoint.x ;
			 		mEndPoint2.y = newPoint.y ;
			 		Draw() ;
			        ShowInfo() ; 
			 	}
			}
		}
	}
	else {
		while(Button()) {
			GetMouse(&newPoint) ;
			// Shift the line
			dx = newPoint.x - lastPoint.x ;
			dy = newPoint.y - lastPoint.y ;
			if (dx != 0. || dy != 0.) {
				Draw() ;
				Shift(near(dx),near(dy)) ;
				lastPoint = newPoint ;
				Draw() ;
		        ShowInfo() ; 
			}	
	   }	    
	}
	mIsValid = mEndPoint2.x != mEndPoint1.x || mEndPoint2.y != mEndPoint1.y ;
	if (mIsValid) {
		if (mEndPoint1.x > mEndPoint2.x) {
			FPoint tempPt = mEndPoint1 ;
			mEndPoint1 = mEndPoint2 ;
			mEndPoint2 = tempPt ;
		}
	}	 
	mWindow->WindowToImagePoint(mEndPoint1) ;
	mWindow->WindowToImagePoint(mEndPoint2) ;
	PenNormal() ;
	mIsSelected = true ;
}

void TRuler::ShowInfo() {
	char text[256] ;
	
	CalibrationUnit calibrationUnit = mWindow->GetImage()->GetCalibrationUnit() ;
	
	float	lineLength = GetLength() ;
	float	calibrationConstant = mWindow->GetImage()->GetCalibration() ;
	switch(calibrationUnit) {
		case kPixel:
			sprintf(text,"L = %.2f Px, Angle = %.2f Deg.",lineLength,GetAngle()) ;
			break ;
		case kInversePixel:
			sprintf(text,"L = %.4f 1/Px, Angle = %.2f Deg.\nL = %.2f Px",lineLength*calibrationConstant,GetAngle(),lineLength) ;
			break ;
		case kAngstrom:
			sprintf(text,"L = %.2f A, Angle = %.2f Deg.\nL = %.2f Px",lineLength*calibrationConstant,GetAngle(),lineLength) ;
			break ;
		case kInverseAngstrom:
			sprintf(text,"L = %.4f 1/A, Angle = %.2f Deg.\nL = %.2f A, L = %.2f Px",lineLength*calibrationConstant,GetAngle(),
				1/(lineLength*calibrationConstant),lineLength) ;
			break ;
		case kNanometer:
			sprintf(text,"L = %.2f nm, Angle = %.2f Deg.\nL = %.2f Px",lineLength*calibrationConstant,GetAngle(),lineLength) ;
			break ;
		case kInverseNanometer:
			sprintf(text,"L = %.4f 1/nm, Angle = %.2f Deg.\nL = %.2f nm, L = %.2f Px",lineLength*calibrationConstant,GetAngle(),
				1/(lineLength*calibrationConstant),lineLength) ;
			break ;
	}
	::WriteObjectInfo(text) ;
}

void TRuler::DoMouseDown(Point &location) {

	if (!mIsSelected) {
		if (InRuler(location)) {
			mIsSelected = true ;
		}
		else 
			return ;
	}
	else {
		if (!InRuler(location)) {
			mIsSelected = false ;
			return ;
		}
	}
	Modify(location) ;
}

bool	TRuler::InRuler(Point &pt) {
	if (InHandles(pt))
		return true ;
		
	float	x1,x2,y1,y2,k,b,x,y,x0,y0 ;
	float	dx,dy,kk,d ;
	Rect	theRect ;
	
	float	tolsq = 16. ;
	
	x1 = mEndPoint1.x ;
	y1 = mEndPoint1.y ;
	x2 = mEndPoint2.x ;
	y2 = mEndPoint2.y ;
	dx = x2 - x1 ;
	dy = y2 - y1 ;
	
	short	left = x1 < x2 ? near(x1) : near(x2) ;
	short	right = x1 > x2 ? near(x1) : near(x2) ;
	short	top = y1 < y2 ? near(y1) : near(y2) ;
	short	bottom = y1 > y2 ? near(y1) : near(y2) ;
	SetRect(&theRect,left,top,right,bottom) ;

	x = pt.x ;
	y = pt.y ;
	
	if (dx != 0.) {
		k =dy/dx ;
		if (k != 0.) {
			kk = 1+k*k ;
			b = y2 - k*x2 ;			
			// Check distance to line y = kx + b
			// The closest point on the line is (x0,y0)
			x0 = (k*y+x-k*b)/kk ;
			y0 = k*x0+b ;
			d = (x-x0)*(x-x0)+(y-y0)*(y-y0) ;
			if (d < tolsq && PtInRect(&pt,&theRect)) return true ;
		}
		else {
			// Line parallel to x-axis
			y0 = y2 ;
			d = (y-y0)*(y-y0) ;
			if (d < tolsq && PtInRect(&pt,&theRect)) return true ;
		}
	}
	else {
		// Line parallel to y-axis
		x0 = x2 ;
		d = (x-x0)*(x-x0) ;
		if (d < tolsq && PtInRect(&pt,&theRect)) return true ;
	}
		
	return false ;
	
}

bool	TRuler::InHandles(Point &point) {
	Rect	theRect ;
	
	SetRect(&theRect,near(mEndPoint1.x-4),near(mEndPoint1.y-4),near(mEndPoint1.x+4),near(mEndPoint1.y+4)) ;
	if (PtInRect(&point,&theRect)) return true ;
	SetRect(&theRect,near(mEndPoint2.x-4),near(mEndPoint2.y-4),near(mEndPoint2.x+4),near(mEndPoint2.y+4)) ;
	if (PtInRect(&point,&theRect)) return true ;
	
	return false ;
}

void TRuler::DrawHandles() {
	Rect	theRect ;
	SetFgColor(gGreenPixel) ;
	SetRect(&theRect,near(mEndPoint1.x-3),near(mEndPoint1.y-3),near(mEndPoint1.x+3),near(mEndPoint1.y+3)) ;
	PaintRect(&theRect) ;
	SetRect(&theRect,near(mEndPoint2.x-3),near(mEndPoint2.y-3),near(mEndPoint2.x+3),near(mEndPoint2.y+3)) ;
	PaintRect(&theRect) ;
	SetFgColor(gBlackPixel) ;		
}

void TRuler::Shift(short dx, short dy) { 
    mEndPoint1.x += dx ;
    mEndPoint2.x += dx ;
    mEndPoint1.y += dy ;
    mEndPoint2.y += dy ;
}		

float TRuler::GetLength() {
	float 	x,y ;
	
	x = mEndPoint1.x - mEndPoint2.x ;
	y = mEndPoint1.y - mEndPoint2.y ;
	return sqrt(x*x+y*y) ;
}

float TRuler::GetAngle() {
	float 	x,y ;
	Point p1,p2 ;
	float angle ;
	
	p1.x = (mEndPoint1.x < mEndPoint2.x ? mEndPoint1.x : mEndPoint2.x) ;
	p1.y = (mEndPoint1.x < mEndPoint2.x ? mEndPoint1.y : mEndPoint2.y) ;
	p2.x = (mEndPoint1.x > mEndPoint2.x ? mEndPoint1.x : mEndPoint2.x) ;
	p2.y = (mEndPoint1.x > mEndPoint2.x ? mEndPoint1.y : mEndPoint2.y) ;
	x = p2.x - p1.x ;
	y = p2.y - p1.y ;
	if (x == 0.) {
		angle = 90. ;
	}
	else if (y == 0.) 
		angle = 0. ;
	else if (y == x) {
		if (y < 0.)
			angle = 45. ;
		else 
			angle = -45. ;
	}
	else {
		angle = -atan2(y,x) ;
		angle = angle*90./asin(1.) ;
	}
	return(angle) ;
}

void TRuler::Draw() {
	
	SetFgColor(gRedPixel) ;
	MoveTo(mEndPoint1.x,mEndPoint1.y) ;
	LineTo(mEndPoint2.x,mEndPoint2.y) ;	
	if (mIsSelected) DrawHandles() ;
	SetFgColor(gBlackPixel) ;		
}

