#ifndef __TTable_h
#define __TTable_h
#include "types.h"
#include "TWindow.h"
#include "LArray.h"
#include "UMemoryMgr.h"

extern	uLong	gWhitePixel,gBlackPixel,gDarkerGray,gGrayPixel ;

#define	kDoubleClickTime	200

long	near(float) ;
typedef struct {
	short row ;
	short col ;
} TableCellT ;

typedef struct {
	Int16	nRows ;
	Int16	nCols ;
	Int16	DataSize ;
	Int16	ColWidth ;
	Int16	RowHeight ;
	Int16	WindowWidth ;
	Int16	WindowHeight ;
	Boolean	AdornColumns ;
	Boolean	AdornRows ;
	Boolean	HasColumnLabels ;
	Boolean	HasRowLabels ;
	char	*WindowTitle ;
}	TableAttributes ;

class	TTable ;
class	TTableWindow ;
class	TTableEditField ;

typedef struct {
	Boolean	isCalibrated ;
	float	astar ;
	float	bstar ;
	float	gammastar ;
} IndexCalibrationRecord ;

class	TTableClipBoardList {
	public:
		TTableClipBoardList(long	numPhases) ;
		~TTableClipBoardList() ;
		
		void	Calibrate(IndexCalibrationRecord &cbr) { mCalibration = cbr ;} ;
		Boolean	HasCalibration() { return mCalibration.isCalibrated ; } ;
		
		short	mNumValues ;
		short	*mIndex1 ;
		short	*mIndex2 ;
		float	*mValue ;

		unsigned long mClipboardObjectType ;

		IndexCalibrationRecord	mCalibration ;	
		

} ;

extern	TTableClipBoardList	*gTableClipBoardList ;

class	TTableSelection {
	public:
		TTableSelection() { mHasSelection = false ; } ;
		bool	HasSelection() { return mHasSelection ; } ;
		void	ZeroSelection() {  mHasSelection = false ; } ;
		void	Select(TableCellT &topLeft, TableCellT &bottomRight) {
					mTopLeftCell = topLeft ; mBottomRightCell = bottomRight ; mHasSelection = true ; 
				} ;
				
		bool	InSelection(TableCellT &cell) { 
					if (!mHasSelection) return false ;
					return (cell.col >= mTopLeftCell.col 
						&& cell.col <= mBottomRightCell.col && cell.row >= mTopLeftCell.row
						&&   cell.row <= mBottomRightCell.row) ; 
				} ;
			
		
		TableCellT	mTopLeftCell ;
		TableCellT	mBottomRightCell ;
		
		bool	mHasSelection ; 
} ;

class TTableWindow : public TWindow {

	public:
		TTableWindow(short x,short y,TTable *table) ;
		virtual	~TTableWindow() ;

		virtual	void	DoCommand(TCommand command,Boolean *hasBeenHandled) ;
		virtual	void	GetCommandStatus(TCommand,Boolean *isEnabled,
							Boolean *isChecked);

		virtual void 	Draw() ;
		virtual	void 	Refresh() ;
		virtual void	Print() ;
		virtual	void	Save() ;
		virtual	bool	RequestClose() ;
		void			DoWMRequestDelete() ;
		void			ResizeWindow(long width,long height) ;
		virtual void	Update(XEvent *event) ;
		virtual	void	DoMouseDown(XEvent *event) ;
		virtual	void 	DoGrow(XEvent *inEvent) ;
		virtual	void	DoKeyDown(XEvent *event) ;
		static	void	DoHScrollBar(TObject *) ;
		static	void	DoVScrollBar(TObject *) ;
				
	protected:
		friend	class	TTable ;
		TTable			*mTable ;

		virtual void 	DoClick(XEvent *event) ;
		
		bool 			FetchLocalCellFrame(const TableCellT &inCell,Rect &outCellFrame) ;
		void 			ScrollTableBy(float deltaX,float deltaY) ;
		void 			ScrollTableTo(float deltaX,float deltaY) ;
		
		void			AdjustScrollBars() ;		

} ;

class TTablePane : public TPane {

	public:
		TTablePane(TPane *parent,short xpos,short ypos,short width,short height,TTable *table) ;
		virtual	~TTablePane() ;

		virtual void 	Draw() ;
		virtual	void 	Refresh() ;
		TTable			*GetTable() { return mTable ; } ;
		virtual void	Update(XEvent *event) ;
		virtual	void	DoMouseDown(XEvent *event) ;
		virtual	void	DoKeyDown(XEvent *event) ;
		static	void	DoHScrollBar(TObject *) ;
		static	void	DoVScrollBar(TObject *) ;
		virtual	void	SetTableSize(Int16 nrows, Int16 ncols) ;
				
	protected:
		friend	class	TTable ;
		TTable			*mTable ;

		virtual void 	DoClick(XEvent *event) ;
		
		bool 			FetchLocalCellFrame(const TableCellT &inCell,Rect &outCellFrame) ;
		void 			ScrollTableBy(float deltaX,float deltaY) ;
		void 			ScrollTableTo(float deltaX,float deltaY) ;
		
		void			AdjustScrollBars() ;		
		Rect			mWindowRect ;

} ;

class TTable {
	public:
		TTable(Int16 rows,Int16 cols,Int16 cellSize,Int16 colWidth,Int16 rowHeight,char *tableTitle) ;
		TTable(char *title) ;
		TTable(TableAttributes *attributes) ;
		
		virtual			~TTable() ;
		virtual void 	SetRowHeight(Int16 height) ;
		virtual void 	SetColWidth(Int16 width) ;
		virtual void 	Draw(Region) ;
		virtual void	Print() ;
		virtual	void	Save() ;
		
		virtual	void	SetTableSize(Int16 nrows, Int16 ncols) ;
		
		void			SetCanvasRect(Rect &theRect) { mCanvasRect = theRect ; } ;
		virtual	void	DoCommand(TCommand command,Boolean *hasBeenHandled)  ;
		virtual	void	GetCommandStatus(TCommand,Boolean *isEnabled,Boolean *isChecked)  ;
		
		virtual	void	OnCommandHelp() { EDMHelp("TTable") ; } ;

		virtual void 	SetCellData(TableCellT &inCell,void *data) ;
		virtual bool 	FetchCellData(TableCellT &inCell,void *data) ;
		virtual void 	SetDataCells(TableCellT &inFromCell, TableCellT &inToCell,void *data) ; 
		virtual void 	FetchDataCells(TableCellT &inFromCell, TableCellT &inToCell,void *data) ; 
		virtual void 	ClickCell(TableCellT &inCell) ;
		virtual void 	EditCell(TableCellT &inCell) ;
		virtual void	SetColLabel(Int16 inCol,char *title) ;
		virtual void	SetRowLabel(Int16 inRow,char *title) ;
		virtual void	ClickColLabel(Int16 inCol) ;
		virtual void	ClickRowLabel(Int16 inRow) ;
		virtual	void 	DoClick(XEvent *inEvent) ;
		virtual	bool 	DoClickInLabels(XEvent *inEvent) ;
		virtual	void	DoKeyDown(XEvent *theEvent) ;
		virtual	bool	CellIsEditable(TableCellT &) { return false ; } ;
		
		TableCellT		&GetSelectedCell() { return mSelectedCell ; } ;
		void			SetSelectedCell(TableCellT &theCell) ;
		TPane			*GetTablePane() { return mParentPane ; } ;
		
		void			GetTitle(char *name) { strcpy(name,mTitle) ; } ;
		
		Int16			GetNumRows() { return mRows ; } ;
		Int16			GetNumCols() { return mCols ; } ;
	
	protected:
	
		friend	class	TTableWindow ;
		friend	class	TTablePane ;

		bool 			ImageRectIntersectsFrame(Int16 left,Int16 top,Int16 right,Int16 bottom) ;
		
		virtual void	GetCellText(TableCellT &inCell,char *text) ;
		virtual void	SetCellText(TableCellT &inCell,char *text) ;

		void			Refresh() { mParentPane->Refresh() ; } ;
		bool			HasSelection() { return mTableSelection.HasSelection() ; } ;
		void			Select(TableCellT &inCell1,TableCellT &inCell2) {
								mTableSelection.Select(inCell1,inCell2) ; Refresh() ; } ;
		void			ZeroSelection() { mTableSelection.ZeroSelection() ; Refresh() ; } 
		virtual	void 	DrawCellValue(TableCellT &inCell,Rect &cellFrame) ;
		virtual void 	DrawColLabel(Int16 colNumber) ;
		virtual void 	DrawRowLabel(Int16 rowNumber) ;
		virtual void 	DrawCell(TableCellT &inCell) ;
		virtual void 	HighlightCell(TableCellT &inCell) ;
		virtual void	DrawRowLabelText(Int16 rowNumber) ;
		virtual void	DrawColLabelText(Int16 colNumber) ;
		virtual void	DoPaste() ;
		virtual void	DoCopy() ;
		
		void			FetchCellHitBy(const Point &inImagePt,TableCellT &outCell) ;
		bool 			IsValidCell(const TableCellT	&inCell) const ;
		bool 			FetchLocalCellFrame(const TableCellT &inCell,Rect &outCellFrame) ;

		TPane			*mParentPane ;
		
		char			mTitle[256] ;

		Boolean			mHasColLabels ;
		Boolean			mHasRowLabels ;
		Boolean			mAdornRows ;
		Boolean			mAdornColumns ;
		Pattern			mAdornPattern;
		
		// these are arrays of 16 characters (max of labels)
		LArray			mRowLabels;
		LArray			mColLabels;
		
		LArray			*mCellData;
		
		Int16			mRows ;
		Int16			mCols ;
		Int16			mColWidth ;
		Int16			mRowHeight ;
		Int16			mCellDataSize ;
		TableCellT		mSelectedCell ;
		long			mClickTime ;
		long			mLastClick ;
		Rect			mCanvasRect ;
		Rect			mTableRect ;
		
		bool			mTableCellsAreSelectable ;
		
		TTableSelection	mTableSelection ;
				
		TTableEditField	*mEditField ;

} ;


class	TTableEditField : public TEditText {
	public:
		TTableEditField(Rect *rect,char *title,char *text=nil,TWindow *win=nil) : TEditText(rect,title,text) {mParentWindow = win ;  mIsFocus = true ;}  ;
		TTableEditField(short x1,short y1, short x2, short y2,char *title,char *text=nil,TWindow *win=nil) 
			: TEditText(x1,y1,x2,y2,title,text) {mParentWindow = win ; mIsFocus = true ;} ;
		Boolean DoMouseDown(Point *) ;
		KeySym	DoKeyDown(XEvent *event) ;
		void 	Draw() ;
} ;

#endif
