// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------


#include "TXForm.h"
#include "globals.h"
void	ProcessModalEvent(XEvent *event) ;
void	SetFocus(FL_FORM *form) ;

Boolean	TXForm::DefinedAtClose = false ;
int TXForm::CloseWindow(FL_FORM *form,void *data) {
	TXForm *theForm = (TXForm *)form->u_vdata ;
	if (theForm->RequestClose())
		delete theForm ;
	return FL_IGNORE ;
}

TXForm::TXForm(char *title) {
	mXForm = (XFormRecord *) fl_calloc(1, sizeof(XFormRecord));
	mXForm->mForm = nil ;

	mIsValid = true ;
	if (!TXForm::DefinedAtClose) {
		fl_set_atclose(TXForm::CloseWindow,0) ;
		TXForm::DefinedAtClose = true ;
	}

	mOKButton = (FL_OBJECT *)-1 ;
	mCancelButton = (FL_OBJECT *)-1 ;

	mPlacement = FL_PLACE_FREE ;
	mBorder = FL_FULLBORDER ;
	if (title == nil)
		strcpy(mTitle,"") ;
	else 
		strcpy(mTitle,title) ;
}

Boolean TXForm::DoModal() {
	FL_OBJECT	*theObject ;
	XEvent		event ;

	// By now we have set the data, but we haven't been shown
	Show() ;
	do {
		theObject = fl_do_forms() ;
		if (theObject == FL_EVENT) {
			fl_XNextEvent(&event) ;
			ProcessModalEvent(&event) ;
		}
		else {
			DoObject(theObject) ;
		}
	} while (theObject != mOKButton && theObject != mCancelButton) ;
	if (theObject == mCancelButton) return false ;
	return true ;
}

void TXForm::DoObject(FL_OBJECT *obj) {
}

void	TXForm::Update() { 
	Draw() ; 
} 

void TXForm::Show() {
	// If we haven't defined a form by now, let us define a 
	// dummy form on the fly
	if (mXForm->mForm == nil) {
		mXForm->mForm = fl_bgn_form(FL_NO_BOX, 300, 200);
		mWindowBox = fl_add_box(FL_UP_BOX,0,0,300,200,"");
		mOKButton = fl_add_button(FL_NORMAL_BUTTON,230,165,60,25,"OK");
		mCancelButton = fl_add_button(FL_NORMAL_BUTTON,130,165,90,25,"Cancel");
		fl_end_form();
		mXForm->mForm->fdui = mXForm ;
	}
	mXForm->mForm->u_vdata = (void *)this ;
	fl_show_form(mXForm->mForm,mPlacement,mBorder,mTitle) ;
	SetFocus(mXForm->mForm) ;
}

TXForm::~TXForm() {
	if (mXForm->mForm != nil) {
		fl_hide_form(mXForm->mForm) ;
		fl_free_form(mXForm->mForm) ;
	}
	else 
		fl_free(mXForm) ;
}

void	SetFocus(FL_FORM *form) {
	Window	focusWindow ;
	int	revert_to ;
	XGetInputFocus(gDisplay,&focusWindow,&revert_to) ;
	XSetInputFocus(gDisplay,fl_winget(),RevertToParent,CurrentTime) ;
}



