// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------


#include <math.h>
#include <stdio.h>
#include "TImage.h"
#include "TDialog.h"


extern	void 	fft1d(long m,long inc,float *data,float *table) ;
extern	void 	binrev(long m,long inc,float *c)  ;
extern	void 	fft2d(float data[],long n,long m,long isign) ;
extern	long 	near(float) ;
extern	void	PutMessage(char *) ;
int IsPowerOfTwo(long value) ;

void vclr(float *a1,long inc, long numElements) ;
void vsmul(float *a1,long inc1,float scalar,float *a2,long inc2, long numElements) ;
void vmov(float *a1,long inc1,float *a2,long inc2,long numElements) ;
void polar(float *a1,long inc1,float *a2,long inc2,long numElements) ;
void vsq(float *a1,long inc1,float *a2,long inc2,long numElements) ;
void vsqrt(float *a1,long inc1,float *a2,long inc2,long numElements) ;
void rect(float *a1,long inc1,float *a2,long inc2,long numElements) ;

#define log2(x)	log(x)/log(2.0)
void FFTW(float *data,long width,long height) ;
void IFFTW(float *data,long width,long height) ;

Boolean ApplyWienerFilter(float *data,long width,long height, float strength) ;
Boolean DoApplyWienerFilter(float *data,float *PSNoise,float *PSData,long width,long height, float strength) ;
Boolean FindMedianBackground(float *data,long width,float *median) ;
long near(float fx) ;
void sort(float data[],long n) ;

TImage *WienerFilter(TImage *u) {
	TImage *filteredImage = new TImage(*u) ;
        float fstrength = 1 ;
	if (!ApplyWienerFilter(filteredImage->GetData(), 
		filteredImage->Width(), filteredImage->Height(), fstrength)) {
		PutMessage("Failure in ApplyWienerFilter") ;
		delete filteredImage ;
		filteredImage = nil ;
	}
	return filteredImage ;
	
}
TImage *WienerFilter2(TImage *u) {
	TImage *filteredImage = new TImage(*u) ;
        float fstrength = 4 ;
	if (!ApplyWienerFilter(filteredImage->GetData(),
		filteredImage->Width(), filteredImage->Height(), fstrength)) {
		PutMessage("Failure in ApplyWienerFilter") ;
		delete filteredImage ;
		filteredImage = nil ;
	}
	return filteredImage ;
}
TImage *WienerFiltera(TImage *u) {
	TImage *filteredImage = new TImage(*u) ;
        float fstrength = GetRealNumberWithPrompt("Filter Strength, 0.5-10.0",1.) ;
        if (fstrength < 0.5) fstrength = 0.5 ;
        if (fstrength > 10.0) fstrength = 10.0 ;
	if (!ApplyWienerFilter(filteredImage->GetData(),
		filteredImage->Width(), filteredImage->Height(), fstrength)) {
		PutMessage("Failure in ApplyWienerFilter") ;
		delete filteredImage ;
		filteredImage = nil ;
	}
	return filteredImage ;
}
Boolean ApplyWienerFilter(float *data,long width,long height, float strength ) {
	float *fftdata,*ampdata,*avdata,scale,*p1,*PSData ;
	long numPixels ;
	
	numPixels = width*height ;
	
	// to hold the fourier transform
	fftdata = new float[2*numPixels] ;
	if (fftdata == NULL) {
		PutMessage("Unable to allocate memory for 'fftdata' in 'ApplyWienerFilter'") ;
		return false ;
	}
	
	ampdata = new float[2*numPixels] ;
	if (ampdata == NULL) {
		PutMessage("Unable to allocate memory for 'ampdata' in 'ApplyWienerFilter'") ;
		return false ;
	}
	PSData = new float[numPixels] ;
	if (PSData == NULL) {
		PutMessage("Unable to allocate memory for 'PSData' in 'ApplyWienerFilter'") ;
		return false ;
	}
	
	avdata = new float[width/2] ;
	if (avdata == NULL) {
		PutMessage("Unable to allocate memory for 'avdata' in 'ApplyWienerFilter'") ;
		return false ;
	}
	
	// Zero the complex array in case the allocate storage is not zeroed
	p1 = fftdata+1 ;
	vclr(p1,2,numPixels) ;
	
	// Move the real image data into a complex and do in place fft
	vmov(data,1,fftdata,2,numPixels) ;
	FFTW(fftdata,width,height) ;
	
	
	// Convert to polar data, amplitude stored in real part, phase in imaginary part
	polar(fftdata,1,ampdata,1,numPixels) ;

	// Simplifying by using the Median as the estimate for the noise
	if (!FindMedianBackground(ampdata,width,avdata)) {
		delete[] ampdata ;
		delete[] avdata ;
		delete[] fftdata ;
		delete[] PSData ;
		return(false) ;
	}

	// square the data to get the estimate for the power-spectrum of the noise
	vsq(avdata,1,avdata,1,width/2) ;
	
	// Copy the original signal 
	vmov(ampdata,2,PSData,1,numPixels) ;
	vsq(PSData,1,PSData,1,numPixels) ;
	
	if (!DoApplyWienerFilter(ampdata,avdata,PSData,width,height,strength)) {
		delete[] ampdata ;
		delete[] avdata ;
		delete[] fftdata ;
		delete[] PSData ;
		return(false) ;
	}
	
	
	rect(ampdata,1,fftdata,1,numPixels) ;
	
	IFFTW(fftdata,width,height) ;
	
	// Move our result into image and scale
	scale = 1./numPixels ;
	vsmul(fftdata,2,scale,data,1,numPixels) ;
	
	delete[] ampdata ;
	delete[] avdata ;
	delete[] fftdata ;
	delete[] PSData ;
	return(true) ;
	

}

Boolean DoApplyWienerFilter(float *data,float *PSNoise,float *PSData,long width,long height,float strength) {
	float *pi ;
	register int h,k,i,j ;
	int	kmax,hmax ;
	float ksq,hsq,filter,signal ;
	int ir,length ;

	length = width/2 ;	// The input image has width = height
	pi = data ;			// Point to our data, which contains the amplitude and phase
	kmax = length ;
	hmax = kmax ;
	PSNoise[0] = 0. ;
	
	for(j=0;j<height;j++) {
		k = (j > kmax ? j-height : j) ;
		ksq = k*k ;
		for (i=0;i<width;i++) {
			h = (i > hmax ? i-width : i) ;
			hsq = h*h ;
			ir = (int)(sqrt(hsq+ksq)+0.5) ;
			if (ir < length) {
				signal = PSData[j*width+i] ;
				if (signal > 0) 
					filter = (signal-strength*PSNoise[ir])/signal ;
				else
					filter = 0. ;
				*pi *= filter ;
				if (*pi < 0. ) *pi = 0. ;
			}
			else {
				*pi = 0. ;
			}
			pi += 2 ;
		}
	}
	return(true) ;

}


void vclr(float *a1,long inc, long numElements) {
	long i ;
	float *p1 ;
	
	for(i=0,p1=a1;i<numElements;i++) {
		*p1 = 0. ;
		p1 += inc ;
	}
}

void vsmul(float *a1,long inc1,float scalar,float *a2,long inc2, long numElements) {
	long i ;
	float *p1,*p2 ;
	
	for(i=0,p1=a1,p2=a2;i<numElements;i++) {
		*p2 = scalar*(*p1) ;
		p1 += inc1 ;
		p2 += inc2 ;
	}
}

void vmov(float *a1,long inc1,float *a2,long inc2,long numElements) {
	long i ;
	float *p1,*p2 ;
	
	for(i=0,p1=a1,p2=a2;i<numElements;i++) {
		*p2 = *p1 ;
		p1 += inc1 ;
		p2 += inc2 ;
	}
}

void polar(float *a1,long inc1,float *a2,long inc2,long numElements) {
	long i,cInc1,cInc2 ;
	float *p1,*p2,rl,im ;
	
	cInc1 = 2*inc1 ;
	cInc2 = 2*inc2 ;
	for(i=0,p1=a1,p2=a2;i<numElements;i++) {
		rl = *p1 ;
		im = *(p1+1) ;
		*p2 = sqrt(rl*rl+im*im) ;
		*(p2+1) = atan2(im,rl) ;
		p1 += cInc1 ;
		p2 += cInc2 ;
	}
}
		
void vsq(float *a1,long inc1,float *a2,long inc2,long numElements) {
	long i ;
	float *p1,*p2 ;
	
	for(i=0,p1=a1,p2=a2;i<numElements;i++) {
		*p2 = *p1 * (*p1) ;
		p2 += inc2 ;
		p1 += inc1 ;
	}
}

void vsqrt(float *a1,long inc1,float *a2,long inc2,long numElements) {
	long i ;
	float *p1,*p2 ;
	
	for(i=0,p1=a1,p2=a2;i<numElements;i++) {
		*p2 = sqrt(*p1) ;
		p2 += inc2 ;
		p1 += inc1 ;
	}
}

void rect(float *a1,long inc1,float *a2,long inc2,long numElements) {
	long i,cInc1,cInc2 ;
	float *p1,*p2,amp,phs ;
	
	cInc1 = 2*inc1 ;
	cInc2 = 2*inc2 ;
	for(i=0,p1=a1,p2=a2;i<numElements;i++) {
		amp = *p1 ;
		phs = *(p1+1) ;
		*p2 = amp*cos(phs) ;
		*(p2+1) = amp*sin(phs) ;
		p1 += cInc1 ;
		p2 += cInc2 ;
	}
}

// -----------------------------------------------------------------------------------------
// Calculate the rotational average intensity of the 'image' stored in data of size 'width'
// by 'height.  The data input is the amplitude and phase of a complex image
// -----------------------------------------------------------------------------------------
Boolean FindMedianBackground(float *data,long width,float *median) {
	float *pi ;
	register int h,k,i,j ;
	int	*count,kmax,hmax,numValues ;
	float ksq,hsq ;
	int ir,length,n2 ;
	float **radialValues ;
	
	length = width/2 ;			// The input image has width = height
	count = new int[length] ;
	if (count == NULL) {
		PutMessage("Unable to allocate memory for 'count' in 'FindMedianBackground'") ;
		return(false) ;
	}
	
	for (i=0; i< length;i++) {
		count[i] = 0 ;
	}
	
	pi = data ;	// Point to our data, either amplitude or intensity
	kmax = length ;
	hmax = kmax ;
	for(k=0;k<kmax;k++) {		// Only half-plane
		ksq = k*k ;
		for (i=0;i<width;i++) {
			h = (i > hmax ? i-width : i) ;
			hsq = h*h ;
			ir = (int)(sqrt(hsq+ksq)+0.5) ;
			if (ir < length /* && (*pi) > 0.*/) {
				count[ir]++ ;
			}
			pi += 2 ;
		}
	}

	radialValues = new float *[length] ;
	if (radialValues == NULL) {
		PutMessage("Unable to allocate memory for 'radialValues' in 'FindAverageBackground'") ;
		delete[] count ;
		return(false) ;
	}
	
	for (i=0; i< length;i++) {
		numValues = (count[i] < 16 ? 16 : count[i]) ;
		*(radialValues+i) = new float[numValues] ;
		if (*(radialValues+i) == NULL) {
			PutMessage("Unable to allocate memory for 'radialValues[i]' in 'FindAverageBackground'") ;
			delete[] count ;
			for(j=0;j<i;j++) {
				delete[] *(radialValues+j) ;
			}
			return(false) ;
		}
		count[i] = 0 ;
	}
	
	pi = data ;	// Point to our data, either amplitude or intensity
	for(k=0;k<kmax;k++) {		// Only half-plane
		ksq = k*k ;
		for (i=0;i<width;i++) {
			h = (i > hmax ? i-width : i) ;
			hsq = h*h ;
			ir = (int)(sqrt(hsq+ksq)+0.5) ;
			if (ir < length /* && (*pi) > 0. */) {
				*(*(radialValues+ir)+count[ir])  = *pi ;
				count[ir]++ ;
			}
			pi += 2 ;
		}
	}
	
	// Try to find the median of the data
	for(i=0;i<length;i++) {
		if (count[i] > 1) {
			sort(*(radialValues+i),count[i]) ;
			n2 = count[i]/2 ;
			if(2*n2 == count[i]) {
				median[i] = 0.5*(*(*(radialValues+i)+n2-1)+ *(*(radialValues+i)+n2)) ;
			}
			else {
				median[i] = *(*(radialValues+i)+n2) ;			
			}
		}
		else if (count[i] == 1) {
			median[i] = *(*(radialValues+i)) ;
		}
		else {
			median[i] = 0. ;
		}
	}
	
	// Free our allocated storage and return
	delete[] count ;
	for(i=0;i<length;i++) {
		delete[] *(radialValues+i) ;
	}
	delete[] radialValues ;
	return(true) ;
}

void sort(float data[],long n) {
	long gap ;
	register long i,j ;
	register float temp ;
	
	for (gap=n/2;gap>0;gap/=2)
		for(i=gap;i<n;i++)
			for(j=i-gap;j>=0;j-=gap)
				{
					if (data[j] <= data[j+gap])
						break ;
					temp = data[j] ;
					data[j] = data[j+gap] ;
					data[j+gap] = temp ;
			  	}
}
				
