// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------
#include "numbers.h"
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <math.h>
//#include <iostream.h> Temp hack
#define	SIGN(x)	((x) < 0. ? -1 : 1)

#define	EOS		'\0'
#define NEWLINE	'\n'
#define	CARRIAGE_RETURN	'\r'
#define	BLANK	' '
#define	CR	'\r'
#define LF	'\n'

// -----------------------------
// Ascii to octal number
// -----------------------------
long atoo(char *cp){
	long n ;
	n = 0 ;
	while (*cp >='0' && *cp <= '7') {
		n = 8*n + *cp - '0' ;
		cp++ ;
	} ;
	return(n) ;
}
// -----------------------------
// ascii to hex number
// -----------------------------
long atox(char *cp) {
	long n ;

	n = 0 ;
	while (*cp != '\0' && *cp != ' ') {
		if (*cp >= '0' && *cp <= '9')
			n = 16*n + *cp - '0' ;
		else if (toupper(*cp) >= 'A' && toupper(*cp) <= 'F')
			n = 16*n + *cp - 55 ;
		cp++ ;
	} ;
	return(n) ;
}

// -----------------------------
// Round real to nearest integer
// -----------------------------
long near(float x) ;
long near(float x) {
	float y ;
	long i ;

	y = x+.5 ;
	i = y ;
	if (y < 0.) i-- ;
	return(i) ;
}
// --------------------------------------------------------------------------
// Reduce real number to values < 1. >= 0. 
// --------------------------------------------------------------------------
float one(float *x) ;
float one(float *x) {
    float y ;
    y = *x ;
    y = y - (int)y + 1. ;
    y = y - (int)y ;
    if (1.-y < 0.000001)
	y = 0. ;
    return(y) ;
  }
float one(float x) {
    float y ;
    y = x ;
    y = y - (int)y + 1. ;
    y = y - (int)y ;
    if (1.-y < 0.000001)
	y = 0. ;
    return(y) ;
  }

float	CheckZeroOne(float x) {
	float	delta = 0.000001 ;

	if (fabs(x) < delta)
		return 0. ;
	if (fabs(1.0-x) < delta)
		return 1. ;
	if (fabs(1.0+x) < delta)
		return -1. ;
	return x ;
}

void	DegreeToRad(float *degree) {
	double pi = 2.0*asin(1.0) ;
	*degree = (*degree)*pi/180. ;
}

void	RadToDegree(float *degree) {
	double pi = 2.0*asin(1.0) ;
	*degree = (*degree)*180/pi ;
}

Boolean	NumberIsInteger(float x) {
	long 	ival,sign ;
	float	tval,fracPart,ftol=0.00001 ;

	sign = SIGN(x) ;
	tval = fabs(x)+0.5 ;
	ival = (long)tval ;

	fracPart = fabs(tval-ival-0.5) ;
	if (fracPart < ftol) 
		return true ;
	else
		return false ;
}

char *ReadLine(char *buffer,long numChars,FILE *fp) {
	char c ;
	int i ;
	
	for(i=0;i<numChars;i++) {
		c = fgetc(fp) ;
		if (c == EOF) {
			buffer[i] = '\0' ;
			return NULL ;
		}	
		buffer[i] = c ;
		if (c == LF || c == CR) {
			buffer[i] = '\0' ;
			return buffer ;
		}
	}
	buffer[numChars-1] = '\0' ;
	return buffer ;
}

// This routine reads the next input line and skips any lines that are considered to be comment lines
// The position in the file immediately following the line read is returned
// Comment lines start with either # % or are all enclosed by COMMENT{  }COMMENT
// Maybe I also should add here the normal comments like /* */ and //
char *ReadLine(char *inputStream,char *lineRead,long numChars) {
	char 	*cpStart,*cp ;
	int 	i,numCharRead,numValidChars ;
	char	buffer[256] ;
	
	Boolean validLine = false ;
	short isInsideCommentBlock = 0;
	Boolean	isCommentLine ;
	
	cp = inputStream ;
	while (!validLine) {
		numCharRead = 0 ;
		cpStart = cp ;
		// Read to the end of the line
		while (*cp != CR && *cp != LF && *cp != EOS && numCharRead < numChars) {
			numCharRead++ ;
			cp++ ;
		}
		if (numCharRead == 0 && *cp == EOS) return NULL ;
		if (numCharRead == numChars) *(cp-1) = '\0' ;
		
		// Now skip past the end of line
		while (*cp == CR || *cp == LF) cp++ ;
		
		if (numCharRead != 0) {	// We have a line with characters
			// Copy it to temporary storage
			strncpy(buffer,cpStart,numCharRead) ;
			buffer[numCharRead] = '\0' ;
			
			// Check to see if this is a comment line only
			if (strstr(buffer,"COMMENT{") != NULL) isInsideCommentBlock++ ;
			if (strstr(buffer,"}COMMENT") != NULL) isInsideCommentBlock-- ;
			
			isCommentLine = isInsideCommentBlock == 0 ? false : true ;
			
			if (!isCommentLine) {
				// We still might have a comment line 
				numValidChars = 0 ;
				for(i=0; i < numCharRead && !isCommentLine ; i++) {
					if (buffer[i] != '#' && buffer[i] != '%') {
						if (buffer[i] != ' ') {
							numValidChars++ ;
							break ;
						}
					}
					else if (numValidChars == 0) {
						isCommentLine = true ;
					}
				}
				if (numValidChars == 0) {
					isCommentLine = true ;
				}
			}
			if (!isCommentLine) {
				strcpy(lineRead,buffer) ;
				validLine = true ;
			}
		}
	} ;
	
	if (!validLine) return nil ;
	
	return cp ;	
}

