dnl

 
AC_DEFUN(
        [CHECK_GNU_MAKE], [ AC_CACHE_CHECK( for GNU make,_cv_gnu_make_command,
                _cv_gnu_make_command='' ;
dnl Search all the common names for GNU make
                for a in "$MAKE" make gmake gnumake ; do
                        if test -z "$a" ; then continue ; fi ;
                        if  ( sh -c "$a --version" 2> /dev/null | grep GNU  2>&1 > /dev/null ) ;  then
                                _cv_gnu_make_command=$a ;
                                break;
                        fi
                done ;
        ) ;
dnl If there was a GNU version, then set @ifGNUmake@ to the empty string, 'No' otherwise
        if test  "x$_cv_gnu_make_command" != "x"  ; then
                ifGNUmake='' ;
        else
                ifGNUmake='No' ;
                AC_MSG_RESULT("Not found");
		echo "I did not find a GNU make" >> edm.log ;
		echo "You might run into problems during the make phase" >> edm.log ;
        fi
dnl        AC_SUBST(ifGNUmake)
] )
dnl Modified macro to ask a question
AC_DEFUN([AC_PROMPT_USER],
[
echo $ECHO_N "$2 (for affirmative, enter $3) $ECHO_C"
tmpinput=""
read tmpinput
eval $1=$tmpinput
]
) dnl

AC_DEFUN(ACX_CHECK_CC_FLAGS,
[
AC_REQUIRE([AC_PROG_CC])
ac_save_CFLAGS=$CFLAGS
CFLAGS="$1"
AC_CACHE_CHECK(whether $CC accepts $1, ac_cv_$2,
               [AC_COMPILE_IFELSE([AC_LANG_PROGRAM()], [ac_cv_$2=yes], 
						       [ac_cv_$2=no])])
CFLAGS=$ac_save_CFLAGS
if test "$ac_cv_$2" = yes; then
	:
	$3
else
	:
	$4
fi
])

AC_DEFUN(ACX_PROG_CC_MAXOPT,
[
AC_REQUIRE([AC_PROG_CC])
AC_REQUIRE([ACX_PROG_CC_EGCS])
AC_REQUIRE([AC_CANONICAL_HOST])

# Try to determine "good" native compiler flags if none specified on command
# line
if test "$ac_test_CFLAGS" != "set"; then
  CFLAGS=""
  case "${host_cpu}-${host_os}" in

  *linux*) if test "$CC" = icc; then
                    CFLAGS="-O2 -ip -w"
	        fi;;
  sparc-solaris2*) if test `basename "$CC"` = cc; then
                    CFLAGS="-native -fast -xO5 -dalign $WSUN"
                 fi;;

  alpha*-osf*)  if test "$CC" = cc; then
                    CFLAGS="-newc $WALPH -O5 -ansi_alias -ansi_args -fp_reorder -tune host -arch host"
                fi;;

  hppa*-hpux*)  if test "$ac_compiler_gnu" != yes; then
                    CFLAGS="-Ae $WHPUX"
                    LDFLAGS="$LDFLAGS -L/usr/local/include"
                fi;;

  *sgi-irix*)	if test "$CC" = cc; then
		    CFLAGS="-O3 $WSGI"
		fi;;

   *-aix*)
	if test "$CC" = cc -o "$CC" = xlc -o "$CC" = ccc; then
                ACX_CHECK_CC_FLAGS(-qarch=auto $WAIX -qtune=auto, qarch_auto,
                        CFLAGS="-O3 -qansialias $WAIX -qarch=auto -qtune=auto",
                        [CFLAGS="-O3 -qansialias $WAIX"
                echo "*******************************************************">> edm.log
                echo "*  You seem to have AIX and the IBM compiler.  It is  *">> edm.log
                echo "*  recommended for best performance that you use:     *">> edm.log
                echo "*                                                     *">> edm.log
                echo "*    CFLAGS=-O3 -qarch=xxx -qtune=xxx -qansialias -w  *">> edm.log
		echo "*    Similarly for CXXFLAGS & FFLAGS                  *">> edm.log
                echo "*                      ^^^        ^^^                 *">> edm.log
                echo "*  where xxx is pwr2, pwr3, 604, or whatever kind of  *">> edm.log
                echo "*  CPU you have.  (Set the CFLAGS environment var.    *">> edm.log
                echo "*  and re-run configure.)  For more info, man cc.     *">> edm.log
                echo "*******************************************************">> edm.log
		]
                )
        fi;;
  esac

  # use default flags for gcc on all systems
  if test $ac_cv_prog_gcc = yes; then
     CFLAGS="$WGCC"
  fi

  # the egcs scheduler is too smart and destroys our own schedule.
  # Disable the first instruction scheduling pass.  The second
  # scheduling pass (after register reload) is ok.
  if test "$acx_prog_egcs" = yes; then
     if test "$1" = fftw; then
        CFLAGS="$CFLAGS -fno-schedule-insns"
     fi
  fi

# test for gcc-specific flags:
  if test $ac_cv_prog_gcc = yes; then
    # -malign-double for x86 systems
    ACX_CHECK_CC_FLAGS(-malign-double,align_double,
	CFLAGS="$CFLAGS -malign-double")
    # -fstrict-aliasing for gcc-2.95+, don't use
#    ACX_CHECK_CC_FLAGS(-fstrict-aliasing,fstrict_aliasing,
#	CFLAGS="$CFLAGS -fstrict-aliasing")
    ACX_CHECK_CC_FLAGS(-mpreferred-stack-boundary=4, m_psb_4,
        CFLAGS="$CFLAGS -mpreferred-stack-boundary=4")
  fi

# test if the compiler accepts static
  if test "$static" = "yes"; then
    ACX_CHECK_CC_FLAGS(-static, m_static_cc,CFLAGS="$CFLAGS -static")
  fi

  CPU_FLAGS=""
  CPU_OPTIM=""
  CODELET_OPTIM=""
  if test "$GCC" = "yes"; then
	  CPU_OPTIM=-O3
	  CODELET_OPTIM=-O3
	  dnl try to guess correct CPU flags, at least for linux
	  case "${host_cpu}" in
	  i586*)  ACX_CHECK_CC_FLAGS(-mcpu=pentium,cpu_pentium,
			[CPU_FLAGS=-mcpu=pentium],
			[ACX_CHECK_CC_FLAGS(-mpentium,pentium,
				[CPU_FLAGS=-mpentium])])
		  if test "$1" = fftw; then
	            CODELET_OPTIM=-O
                  fi
		  if test "$1" = benchfft; then
	            CPU_OPTIM=-O2
                  fi
		  ;;
	  i686*)  ACX_CHECK_CC_FLAGS(-mcpu=pentiumpro,cpu_pentiumpro,
			[CPU_FLAGS=-mcpu=pentiumpro],
			[ACX_CHECK_CC_FLAGS(-mpentiumpro,pentiumpro,
				[CPU_FLAGS=-mpentiumpro])])
		  if test "$1" = fftw; then
	            CODELET_OPTIM=-O
                  fi
		  if test "$1" = benchfft; then
	            CPU_OPTIM=-O2
                  fi
		  ;;
	  sparc*)  
dnl  Removed          ACX_CHECK_CC_FLAGS(-mcpu=ultrasparc,cpu_ultrasparc,
dnl			[CPU_FLAGS=-mcpu=ultrasparc])
		  ;;
	  alphaev67)  ACX_CHECK_CC_FLAGS(-mcpu=ev67,cpu_ev67,
			[CPU_FLAGS=-mcpu=ev67])
		  ;;
	  alphaev6)  ACX_CHECK_CC_FLAGS(-mcpu=ev6,cpu_ev6,
			[CPU_FLAGS=-mcpu=ev6])
		  ;;
	  alphaev56)  ACX_CHECK_CC_FLAGS(-mcpu=ev56,cpu_ev56,
			[CPU_FLAGS=-mcpu=ev56],
			[ACX_CHECK_CC_FLAGS(-mcpu=ev5,cpu_ev5,
				[CPU_FLAGS=-mcpu=ev5])])
		  ;;
	  alphaev5)  ACX_CHECK_CC_FLAGS(-mcpu=ev5,cpu_ev5,
			[CPU_FLAGS=-mcpu=ev5])
		  ;;
             hppa*)  ACX_CHECK_CC_FLAGS(-march=2.0,,
                        [CPU_FLAGS=-march=2.0])          
                  ;;
	  powerpc*)
		cputype=`((grep cpu /proc/cpuinfo | head -1 | cut -d: -f2 | sed 's/ //g') ; /usr/bin/machine ; /bin/machine) 2> /dev/null`
		cputype=`echo $cputype | sed -e s/ppc//g`
		is60x=`echo $cputype | egrep "^60[[0-9]]e?$"`
		is750=`echo $cputype | grep "750"`
		is74xx=`echo $cputype | egrep "^74[[0-9]][[0-9]]$"`
		if test -n "$is60x"; then
			ACX_CHECK_CC_FLAGS(-mcpu=$cputype,m_cpu_60x,
				CPU_FLAGS=-mcpu=$cputype)
		elif test -n "$is750"; then
			ACX_CHECK_CC_FLAGS(-mcpu=750,m_cpu_750,
				CPU_FLAGS=-mcpu=750)
		elif test -n "$is74xx"; then
			ACX_CHECK_CC_FLAGS(-mcpu=$cputype,m_cpu_74xx,
				CPU_FLAGS=-mcpu=$cputype)
		fi
		if test -z "$CPU_FLAGS"; then
		        ACX_CHECK_CC_FLAGS(-mcpu=powerpc,m_cpu_powerpc,
				CPU_FLAGS=-mcpu=powerpc)
		fi
		if test -z "$CPU_FLAGS"; then
			ACX_CHECK_CC_FLAGS(-mpowerpc,m_powerpc,
				CPU_FLAGS=-mpowerpc)
		fi
	  esac
# Added material if wincpu has been defined
        echo Testing for PC specific option -- $wincpu 
	if test "$wincpu" != "None"; then
	   CPU_FLAGS="-march=i386"
	   case "$wincpu" in
		portable)
	   		ACX_CHECK_CC_FLAGS(-mfpmath=387,math_387,
			[CPU_FLAGS="$CPU_FLAGS -mfpmath=387"])
			;;			  	   		
	   	i586)
	   		ACX_CHECK_CC_FLAGS(-march=i586,cpu_i586,
			[CPU_FLAGS=-march=i586])
			;;			  	   
	   	pentium)
	   		ACX_CHECK_CC_FLAGS(-march=pentium,cpu_pentium,
			[CPU_FLAGS=-march=pentium])
			;;			  
	   	pentium2)
	   		ACX_CHECK_CC_FLAGS(-march=pentium2,cpu_pentium2,
			[CPU_FLAGS=-march=pentium2])
	   		ACX_CHECK_CC_FLAGS(-mmmx,mmmx,
			[CPU_FLAGS="$CPU_FLAGS -mmmx"])
			;;
	   	pentium3)
	   		ACX_CHECK_CC_FLAGS(-march=pentium3,cpu_pentium3,
			[CPU_FLAGS=-march=pentium3])
	   		ACX_CHECK_CC_FLAGS(-mmmx,mmmx,
			[CPU_FLAGS="$CPU_FLAGS -mmmx"])
			;;			  
	   	pentium4)
	   		ACX_CHECK_CC_FLAGS(-march=pentium4,cpu_pentium4,
			[CPU_FLAGS=-march=pentium4])
	   		ACX_CHECK_CC_FLAGS(-mmmx,mmmx,
			[CPU_FLAGS="$CPU_FLAGS -mmmx"])
			;;			  
	   	Xeon*)
	   		ACX_CHECK_CC_FLAGS(-march=pentium4,cpu_pentium4,
			[CPU_FLAGS=-march=pentium4])
	   		ACX_CHECK_CC_FLAGS(-mmmx,mmmx,
			[CPU_FLAGS="$CPU_FLAGS -mmmx"])
			;;			  
	   	k6)
	   		ACX_CHECK_CC_FLAGS(-march=k6,cpu_k6,
			[CPU_FLAGS=-march=k6])
			;;			  
	   	athlon)
	   		ACX_CHECK_CC_FLAGS(-march=athlon,cpu_athlon,
			[CPU_FLAGS=-march=athlon])
	   		ACX_CHECK_CC_FLAGS(-m3dnow,m3dnow,
			[CPU_FLAGS="$CPU_FLAGS -m3dnow"])
			;;			  
	   	athlon-xp)
	   		ACX_CHECK_CC_FLAGS(-march=athlon-xp,cpu_athlon-xp,
			[CPU_FLAGS=-march=athlon-xp])
	   		ACX_CHECK_CC_FLAGS(-m3dnow,m3dnow,
			[CPU_FLAGS="$CPU_FLAGS -m3dnow"])
			;;			  
	   	Opteron*)
	   		ACX_CHECK_CC_FLAGS(-march=athlon,cpu_athlon,
			[CPU_FLAGS=-march=athlon])
	   		ACX_CHECK_CC_FLAGS(-m3dnow,m3dnow,
			[CPU_FLAGS="$CPU_FLAGS -m3dnow"])
			;;			  
	   esac
	fi
  fi


  if test -n "$CPU_OPTIM"; then
        CFLAGS="$CPU_OPTIM $CFLAGS"
  fi

  if test -n "$CPU_FLAGS"; then
        CFLAGS="$CFLAGS $CPU_FLAGS"
  fi

  AC_SUBST(CODELET_OPTIM)

  if test -z "$CFLAGS"; then
	echo "" >> edm.log
	echo "********************************************************" >> edm.log
        echo "* WARNING: Don't know the best CFLAGS for this system  *" >> edm.log
        echo "* Use  make CFLAGS=..., or edit the top level Makefile *" >> edm.log
	echo "* (otherwise, a default of CFLAGS=-O3 will be used)    *" >> edm.log
	echo "********************************************************" >> edm.log
	echo "" >> edm.log
        CFLAGS="-O3"
  fi

  ACX_CHECK_CC_FLAGS(${CFLAGS}, guessed_cflags, , [
	echo "" >> edm.log
        echo "********************************************************" >> edm.log
        echo "* WARNING: The guessed CFLAGS don't seem to work with  *" >> edm.log
        echo "* your compiler.                                       *" >> edm.log
        echo "* Use  make CFLAGS=..., or edit the top level Makefile *" >> edm.log
        echo "********************************************************" >> edm.log
        echo "" >> edm.log
        CFLAGS=""
  ])

fi
])

AC_DEFUN(CHECK_WITH_CYGWIN,
[
  case $host_os in
  cygwin* | mingw* | pw32* )
    if test "$GCC" = yes; then
      # Ensure MSVC-compatible struct packing convention.
      # Depends on GCC version. gcc2 uses -fnative-struct while
      # gcc3 uses -mms-bitfields.
      #
      msnative_struct=''
      AC_MSG_CHECKING([how to get MSVC-compatible struct packing])
      case `$CC --version | sed -e 's,\..*,.,' -e q` in
      2.)
        if $CC -v --help 2>/dev/null | grep fnative-struct > /dev/null; then
        msnative_struct='-fnative-struct'
        fi
        ;;
      *)
        if $CC -v --help 2>/dev/null | grep ms-bitfields > /dev/null; then
        msnative_struct='-mms-bitfields'
        fi
        ;;
      esac

      if test x"$msnative_struct" = x; then
        AC_MSG_RESULT([no way])
        AC_MSG_WARN([produced libraries might be incompatible with MSVC libs])
      else
        CXXFLAGS="$CXXFLAGS $msnative_struct"
	CFLAGS="$CFLAGS $msnative_struct"
	FFLAGS="$FFLAGS $msnative_struct"
        AC_MSG_RESULT([${msnative_struct}])
      fi
    fi

    # Export all symbols to Win32 DLL using MinGW 2.0 ld.
    WIN32_LD_EXPORT_ALL_SYMBOLS=''
    AC_MSG_CHECKING([whether ld accepts --export-all-symbols])
    if $LD --help 2>&1 | egrep 'export-all-symbols' > /dev/null; then
      WIN32_LD_EXPORT_ALL_SYMBOLS='-Wl,--export-all-symbols'
      AC_MSG_RESULT([yes])
    else
      AC_MSG_RESULT([no])
    fi
    if test x"$WIN32_LD_EXPORT_ALL_SYMBOLS" != x; then
      LDFLAGS="$LDFLAGS $WIN32_LD_EXPORT_ALL_SYMBOLS"
    fi
    ;;
  esac
])
AC_DEFUN(ACX_PROG_GCC_VERSION,
[
AC_REQUIRE([AC_PROG_CC])
AC_CACHE_CHECK(whether we are using gcc $1.$2 or later, ac_cv_prog_gcc_$1_$2,
[
dnl The semicolon after "yes" below is to pacify NeXT's syntax-checking cpp.
cat > conftest.c <<EOF
#ifdef __GNUC__
#  if (__GNUC__ > $1) || (__GNUC__ == $1 && __GNUC_MINOR__ >= $2)
     yes;
#  endif
#endif
EOF
if AC_TRY_COMMAND(${CC-cc} -E conftest.c) | egrep yes >/dev/null 2>&1; then
  ac_cv_prog_gcc_$1_$2=yes
else
  ac_cv_prog_gcc_$1_$2=no
fi
])
if test "$ac_cv_prog_gcc_$1_$2" = yes; then
	:
	$3
else
	:
	$4
fi
])

AC_DEFUN(ACX_PROG_CC_EGCS,
[ACX_PROG_GCC_VERSION(2,90,acx_prog_egcs=yes,acx_prog_egcs=no)])

# Check to see if we are using a version of gcc that aligns the stack
# (true in gcc-2.95+, which have the -mpreferred-stack-boundary flag).
# Also check for stack alignment bug in gcc-2.95.x
# (see http://egcs.cygnus.com/ml/gcc-bugs/1999-11/msg00259.html), and
# whether main() is correctly aligned by the OS/libc/loader.
AC_DEFUN(ACX_GCC_ALIGNS_STACK,
[
AC_REQUIRE([AC_PROG_CC])
acx_gcc_aligns_stack=no
if test "$GCC" = "yes"; then
ACX_CHECK_CC_FLAGS(-mpreferred-stack-boundary=4, m_pref_stack_boundary_4)
if test "$ac_m_pref_stack_boundary_4" = "yes"; then
	AC_MSG_CHECKING([whether the stack is correctly aligned by gcc])
	save_CFLAGS="$CFLAGS"
	CFLAGS="-O -malign-double"
	AC_TRY_RUN([#include <stdlib.h>
#       include <stdio.h>
	struct yuck { int blechh; };
	int one(void) { return 1; }
	struct yuck ick(void) { struct yuck y; y.blechh = 3; return y; }
#       define CHK_ALIGN(x) if ((((long) &(x)) & 0x7)) { fprintf(stderr, "bad alignment of " #x "\n"); exit(1); }
	void blah(int foo) { double foobar; CHK_ALIGN(foobar); }
	int main2(void) {double ok1; struct yuck y; double ok2; CHK_ALIGN(ok1);
                         CHK_ALIGN(ok2); y = ick(); blah(one()); return 0;}
	int main(void) { if ((((long) (__builtin_alloca(0))) & 0x7)) __builtin_alloca(4); return main2(); }
	], [acx_gcc_aligns_stack=yes; acx_gcc_stack_align_bug=no], 
	acx_gcc_stack_align_bug=yes, acx_gcc_stack_align_bug=yes)
	CFLAGS="$save_CFLAGS"
	AC_MSG_RESULT($acx_gcc_aligns_stack)
fi
fi
if test "$acx_gcc_aligns_stack" = yes; then
	:
	$1
else
	:
	$2
fi
]) 
