/*
 * atomf.c
 *   by KT
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "utils.h"
#include "mbconfig.h"
#include "atomf.h"

/* extern DEBUG_INFO dbg; */
extern EPS_CTRL	epsilon;

/* #define ATOMF_FILE		"mbfit.asf" */
#define XDT_FILE		"mbfitxdt.dat"


double atom_scat_f_dt( double sq_s, EDT dyl )
{
  /*
   * atomic scattering factor for electron (elastic part)
   *  (Angstrom unit)
   *
   *  not including gamma factor (relativistic correction)
   *
   *  <input>    s = sin(theta) / lambda
   *         dyl : Doyle & Turner's parameters
   */
  return  dyl.a1 * exp( - dyl.b1 * sq_s )
    + dyl.a2 * exp( - dyl.b2 * sq_s )
      + dyl.a3 * exp( - dyl.b3 * sq_s )
        + dyl.a4 * exp( - dyl.b4 * sq_s );
}


double atom_scat_f_dtx( double sq_s, XDT xdyl )
{
  /*
   * atomic scattering factor for electron (elastic part)
   *   (Angstrom unit)
   *  calc. from x-ray's values (Mott formula)
   *
   *  The return value does not include
   *   the gamma factor for relativistic correction.
   *
   *  <input>  sq_s = sq(s); s = sin(theta) / lambda
   *           xdyl : Doyle & Turner's parameters for X-ray
   */
  double xray_f;

  if (sq_s > epsilon.eps){
    xray_f =	xdyl.a1 * exp( - xdyl.b1 * sq_s )
      + xdyl.a2 * exp( - xdyl.b2 * sq_s )
        + xdyl.a3 * exp( - xdyl.b3 * sq_s )
          + xdyl.a4 * exp( - xdyl.b4 * sq_s )
            + xdyl.c;
    
    /* z_number = xdyl.a1 + xdyl.a2 + xdyl.a3 + xdyl.a4 + xdyl.c; */
    return  0.023934 * ( xdyl.znum - xray_f ) / sq_s;
    /* Mott formula */
  }
  else if (0.0 <= sq_s && sq_s <= epsilon.eps){
    return 0.023934 * (
      xdyl.a1 * xdyl.b1 + xdyl.a2 * xdyl.b2
      + xdyl.a3 * xdyl.b3 + xdyl.a4 * xdyl.b4
      ); /* Not recommended */
  }
  else{
    printf("$$ ERROR in atom_scat_f_dtx. sq_s = %lf \n",sq_s);
    printf("   sq_s = sq(sin(theta)/lambda) >= 0.0\n");
    exit(EXIT_FAILURE);
    return 0.0;  /* dummy */
  }
}


double xray_asf( double sq_s, XDT xdyl )
{
  /*
   * X-ray atomic scattering factor (Angstrom unit)
   *  sq_s = sq(s); s = sin(theta) / lambda
   *  xdyl : Doyle & Turner's parameters for X-ray
   */
  double xray_f;

  xray_f = xdyl.a1 * exp( - xdyl.b1 * sq_s )
    + xdyl.a2 * exp( - xdyl.b2 * sq_s )
      + xdyl.a3 * exp( - xdyl.b3 * sq_s )
        + xdyl.a4 * exp( - xdyl.b4 * sq_s )
          + xdyl.c;
  return xray_f;
}


#if 0
XDT read_xray_asf_table(char *element)
{
  /*
   * Read coefs. of X-ray atomic scattering factor
   * from MBFIT.ASF
   */
  char buf[128];
  char *delims=" ,\t";
  char *token;
  XDT xd;
  FILE *fp;

  token = strtok(element,delims);
  strcpy(element, token); /* Remove spaces */
  /* Read atomic-scattering-factor table. */
  fp = fopen_chk_rmode(ATOMF_FILE);
  /*  if( (fp = fopen(ATOMF_FILE,"r")) == NULL ){
    printf("Cannot open %s\n",ATOMF_FILE);
    exit(EXIT_FAILURE);
  } */

  while( comment_skip_fgets(100, buf, fp) != NULL ){	/* element */
    token = strtok(buf,delims);
    if( strcmp(element, rm_cr(token)) == 0 ){		/* Found */
      comment_skip_fgets(100, buf, fp);
      token = strtok(buf, delims);	xd.a1 = atof(token);
      token = strtok(NULL,delims);	xd.b1 = atof(token);
      token = strtok(NULL,delims);	xd.a2 = atof(token);
      token = strtok(NULL,delims);	xd.b2 = atof(token);
      token = strtok(NULL,delims);	xd.a3 = atof(token);
      token = strtok(NULL,delims);	xd.b3 = atof(token);
      token = strtok(NULL,delims);	xd.a4 = atof(token);
      token = strtok(NULL,delims);	xd.b4 = atof(token);
      token = strtok(NULL,delims);	xd.c  = atof(token);
      fclose(fp);
      return	xd;
    }
    else{
      comment_skip_fgets(100, buf, fp);	/* skip */
      comment_skip_fgets(100, buf, fp);	/* skip */
    }
  }
  printf("$$ Not found the element in read_xray_d_t.\n");
  fclose(fp);
  exit(EXIT_FAILURE);
}
#endif


XDT read_xdt_table(char *element)
{
  /*
   * Read coefs. of X-ray atomic scattering factor with Z-number
   * from mbfitxdt.dat
   */
  char buf[128];
  char *delims=" ,\t";
  char *token;
  XDT xd;
  FILE *fp;
  char mbfit_XDT_file[256];
  sprintf(mbfit_XDT_file,"%s/mbfit/mbfitxdt.dat", PREFIX);  


  token = strtok(element, delims);
  strcpy(element, token);  /* Remove spaces */
/*  if( (fp = fopen(XDT_FILE,"r")) == NULL ){
    printf("Cannot open %s\n",XDT_FILE);
    exit(EXIT_FAILURE);
  } */

   fp=fopen(mbfit_XDT_file,"r");
/*  fp = fopen_chk_rmode(XDT_FILE); */

  while( comment_skip_fgets(100, buf, fp) != NULL ){	/* element */
    token = strtok(buf,delims);
    if( strcmp(element, rm_cr(token)) == 0 ){		/* Found */
      token = strtok(NULL, delims);	xd.znum = atof(token);
      token = strtok(NULL, delims);	xd.a1 = atof(token);
      token = strtok(NULL, delims);	xd.b1 = atof(token);
      token = strtok(NULL, delims);	xd.a2 = atof(token);
      token = strtok(NULL, delims);	xd.b2 = atof(token);
      token = strtok(NULL, delims);	xd.a3 = atof(token);
      token = strtok(NULL, delims);	xd.b3 = atof(token);
      token = strtok(NULL, delims);	xd.a4 = atof(token);
      token = strtok(NULL, delims);	xd.b4 = atof(token);
      token = strtok(NULL, delims);	xd.c  = atof(token);
      /*
      printf(
        "$$%lf\t%lf\t%lf\t%lf\t%lf\t%lf\t%lf\t%lf\t%lf\n",
        xd.a1, xd.b1, xd.a2, xd.b2, xd.a3, xd.b3, xd.a4, xd.b4, xd.c
        );
      */
      fclose(fp);
      return xd;
    }
  }
  printf("$$ Not found the element in read_xdt_table.\n");
  fclose(fp);
  exit(EXIT_FAILURE);
  return xd;  /* dummy */
}


double atomsf_xdt( double sq_s, XDT xdyl )
{
  /*
   * atomic scattering factor for electron (elastic part)
   *   (Angstrom unit)
   *  calculated from X-ray's values 
   *  using an improved Mott formula (Peng and Cowley 1988)
   *
   *  [available for both of neutral and ionized atoms.]
   *
   *  The return value does not include
   *   the gamma factor for relativistic correction.
   *  <input>  sq_s = sq(s); s = sin(theta) / lambda
   *           xdyl : Doyle & Turner's parameters for X-ray
   *  Coef.: $ 8\pi me^{2}/h^{2} = 0.02393367 $
   */
  double asf;

  if (sq_s > epsilon.eps){
    asf = 0.02393367 * (
      (double) xdyl.znum 
      - xdyl.a1 * exp( - xdyl.b1 * sq_s ) 
      - xdyl.a2 * exp( - xdyl.b2 * sq_s ) 
      - xdyl.a3 * exp( - xdyl.b3 * sq_s ) 
      - xdyl.a4 * exp( - xdyl.b4 * sq_s )
      - xdyl.c
      ) / sq_s;
/*    asf = 0.02393367 * (
      xdyl.a1 * ( 1.0 - exp( - xdyl.b1 * sq_s ) )
      + xdyl.a2 * ( 1.0 - exp( - xdyl.b2 * sq_s ) )
      + xdyl.a3 * ( 1.0 - exp( - xdyl.b3 * sq_s ) )
      + xdyl.a4 * ( 1.0 - exp( - xdyl.b4 * sq_s ) ) 
      ) / sq_s; */
  }
  else if (0.0 <= sq_s && sq_s <= epsilon.eps){
    asf = 0.023934 * (
      xdyl.a1 * xdyl.b1 + xdyl.a2 * xdyl.b2
      + xdyl.a3 * xdyl.b3 + xdyl.a4 * xdyl.b4
      );
  }
  else{
    printf("$$ ERROR in atomsf_dtxn. sq_s = %lf \n",sq_s);
    printf("   sq_s = sq(sin(theta)/lambda) >= 0.0\n");
    exit(EXIT_FAILURE);
  }
  return asf;
}


double atom_f_absorption( double sq_s, double f_absrp_0, double b_factor )
{ 
  /*
   * Atomic absorption factor for electron
   *  (Ichimiya's formula)
   */
  return f_absrp_0 * exp( - 2.0 * b_factor * sq_s );
}


double atom_z_xdt( XDT xdyl )
{
  /*
   * Z number
   *  <input>  xdyl : Doyle & Turner's parameters for X-ray
   * Caution: only for neutral atoms!
   */
  return xdyl.a1 + xdyl.a2 + xdyl.a3 + xdyl.a4 + xdyl.c;
}
