#include	<stdio.h>
#include	<stdlib.h>
#include	<math.h>
#include	<string.h>
#include	"complex.h"
#include	"cmatutil.h"
#include	"vecmat.h"

/*#define AMACH 1e-8*/			/* Recommended value for eps */
/*#define ITER_MAX 30*/			/* Iteration limit */


void	eispack_cg( int n, double epsilon, complex **a, complex *eigenval, complex **eigenvec );
void	comqr( int n,int low, int high, double **ar, double **ai,double *eigenvalr,double *eigenvali,int ierr );
void	comqr2( int n, int low, int high, double *ortr, double *orti, complex **a,
                complex *eigenval, complex **eigenvec, int ierr);
void	corth( int n, int low, int high, complex **a, double *ortr, double *orti );
void	cdiv_cg(double ar, double ai, double br, double bi, double *cr, double *ci);
void	csroot(double xr, double xi, double *yr, double *yi);
double	pythag(double a, double b);
void	cbabk2( int n, int low, int high, double *scale, int m, complex **eigenvec );
void	matexchange_rowcol( int n, int j, int k, int l, int m, complex **a );
void	cbal( int n, complex **a, int *low, int *high, double *scale );
double	dmax1( double d1, double d2 );
int		min0( int d1, int d2 );
double	dmin1( double d1, double d2 );
void	matprint_cg(int n, int perline, char *format, double **vr, double **vi);
void	vecprint_cg(int n, int perline, char *format, double *vr, double *vi);
void	sort_eigval_eigvec( int ne, complex *eigenval, complex **eigenvec );

#define MAX2(x, y) ((x) > (y) ? (x) : (y))
#define MIN2(x, y) ((x) < (y) ? (x) : (y))
double eps;


double	dmax1( double d1, double d2 )
{
  if( d1 >= d2 )	return d1;
  else			return d2;
}


int		min0( int d1, int d2 )
{
  if( d1 <= d2 )	return d1;
  else			return d2;
}


double	dmin1( double d1, double d2 )
{
  if( d1 <= d2 )	return d1;
  else			return d2;
}


void dbg_print_cmat(int flag, char *comment, int n, complex **a);	/* <<<<< */


void eispack_cg(
  int n, double epsilon, complex **a, complex *eigenval, complex **eigenvec
  )
{
  /*
   * EISPACK
   */
  int		is1,is2,ierr=1;
  double	*fv1,*fv2,*fv3;
  
  /* dbg_print_cmat(1, "$$a", n, a); */
  fv1 = new_vector( double, n+1 );  chk_vector( fv1 );
  fv2 = new_vector( double, n+1 );  chk_vector( fv2 );
  fv3 = new_vector( double, n+1 );  chk_vector( fv3 );
  eps = epsilon;	/*printf("$$ eps = %lg\n", eps);*/
  
  cbal( n, a, &is1, &is2, fv1 );
  /*dbg_print_cmat(1, "$$a", n, a);*/
  corth( n, is1, is2, a, fv2, fv3 );
  
  /*.......... find both eigenvalues and eigenvectors ..........*/
  comqr2( n, is1, is2, fv2, fv3, a, eigenval, eigenvec, ierr );
  
  if ( ierr == 0 ) { cbabk2( n, is1, is2, fv1, n, eigenvec ); }

  sort_eigval_eigvec( n, eigenval, eigenvec );  /* in descending order */

  free_vector( fv1 );
  free_vector( fv2 );
  free_vector( fv3 );
}


void comqr(
  /* int nm, */int n, int low, int high, double **ar, double **ai,
  double *eigenvalr,double *eigenvali,int ierr
  )
{
  /*
   * calculation of Eigenvalue only
   */
  int		i,j,l,en,ll,itn,its,lp1,enm1;
  double	si,sr,ti,tr,xi,xr,yi,yr,zzi,zzr,norm,tst1,tst2;
  
  ierr = 0;
  
  if (low != high) {
    /*     .......... create real subdiagonal elements ..........*/
    l = low + 1;
    
    for( i=l; i<=high; i++ ) {
      ll = min0( i+1, high );
      if ( fabs(ai[i][i-1]) < eps ) break;
      norm = pythag( ar[i][i-1], ai[i][i-1] );
      yr = ar[i][i-1] / norm;
      yi = ai[i][i-1] / norm;
      ar[i][i-1] = norm;
      ai[i][i-1] = 0.0;
      
      for( j=i; j<=high; j++ ) {
        si = yr * ai[i-1][j-1] - yi * ar[i-1][j-1];
        ar[i-1][j-1] = yr * ar[i-1][j-1] + yi * ai[i-1][j-1];
        ai[i-1][j-1] = si;
      }
      for( j=low; j<=ll; j++ ) {
        si = yr * ai[j-1][i-1] + yi * ar[j-1][i-1];
        ar[j-1][i-1] = yr * ar[j-1][i-1] - yi * ai[j-1][i-1];
        ai[j-1][i-1] = si;
      }
    }
  }
  /*    .......... store roots isolated by cbal ..........*/
  for( i=1; i<=n; i++ ) {
    if ( i >= low && i <= high ) break;
    eigenvalr[i] = ar[i-1][i-1];
    eigenvali[i] = ai[i-1][i-1];
  }
  en = high; tr = 0.0; ti = 0.0; itn = 30 * n;
  
  /***		search for next eigenvalue		***/
  its = 0;
  enm1 = en - 1;
  while ( en >= low ) {
    /***	look for single small sub-diagonal element		***/
    /*			 for l=en step -1 until low  -- ..........*/
    for( ll=low; ll<=en; ll++ ) {
      l = en + low - ll;
      if ( l == low ) break;
      tst1 = fabs(ar[l-1][l-1]) + fabs(ai[l-1][l-1]) + fabs(ar[l-1][l-1]) + fabs(ai[l-1][l-1]);
      tst2 = tst1 + fabs(ar[l-1][l-2]);
      if ( fabs(ar[l-1][l-2]) < eps ) break;
    }
    
    /***	form shift		***/
    if ( l != en ) {/*goto 660*/
      
      if ( itn == 0 ) { ierr = en; return; }
      
      if ( its == 10 || its == 20 ) {
        /***	form exceptional shift		***/
        sr = fabs( ar[en-1][enm1-1] ) + fabs( ar[enm1-1][en-3] );
        si = 0.0;
      }
      else {
        sr = ar[en-1][en-1];
        si = ai[en-1][en-1];
        xr = ar[enm1-1][en-1] * ar[en-1][enm1-1];
        xi = ai[enm1-1][en-1] * ar[en-1][enm1-1];
        
        if ( fabs( xr ) > eps || fabs( xi ) > eps ) {/*goto 340*/
          yr = ( ar[enm1-1][enm1-1] - sr ) / 2.0;
          yi = ( ai[enm1-1][enm1-1] - si ) / 2.0;
          csroot( yr*yr-yi*yi+xr, 2.0*yr*yi+xi, &zzr, &zzi );
          if ( yr * zzr + yi * zzi < 0.0 ) {
            zzr = -zzr;
            zzi = -zzi;
          }
          cdiv_cg( xr, xi, yr+zzr, yi+zzi, &xr, &xi );
          sr -= xr;
          si -= xi;
        }
      }
      /*340*/		for( i=low; i<=en; i++ ) {
        ar[i-1][i-1] -= sr;
        ai[i-1][i-1] -= si;
      }
      tr += sr;
      ti += si;
      its++;
      itn--;
      
      /*     .......... reduce to triangle (rows) ..........	*/
      lp1 = l + 1;
      
      for( i=lp1; i<=en; i++ ) {
        sr = ar[i][i-1];
        ar[i][i-1] = 0.0;
        norm = pythag( pythag( ar[i-1][i-1], ai[i-1][i-1]), sr );
        xr = ar[i-1][i-1] / norm;	eigenvalr[i-1] = xr;
        xi = ai[i-1][i-1] / norm;	eigenvali[i-1] = xi;
        ar[i-1][i-1] = norm;
        ai[i-1][i-1] = 0.0;	ai[i][i-1] = sr / norm;
        for( j=i; j<=en; j++ ) {
          yr = ar[i-1][j];	yi = ai[i-1][j];
          zzr = ar[i-1][j-1];		zzi = ai[i-1][j-1];
          ar[i-1][j] = xr * yr + xi * yi + ai[i][i-1] * zzr;
          ai[i-1][j] = xr * yi - xi * yr + ai[i][i-1] * zzi;
          ar[i-1][j-1] = xr * zzr - xi * zzi - ai[i][i-1] * yr;
          ai[i-1][j-1] = xr * zzi + xi * zzr - ai[i][i-1] * yi;
        }
      }
      
      si = ai[en-1][en-1];
      if ( fabs( si ) > eps ) {
        norm = pythag( ar[en-1][en-1], si );
        sr = ar[en-1][en-1] / norm;	si = si / norm;
        ar[en-1][en-1] = norm;
        ai[en-1][en-1] = 0.0;
      }
      
      /***	inverse operation (columns)		***/
      /*540*/
      for( j=lp1; j<=en; j++ ) {
        xr = eigenvalr[j-1];	xi = eigenvali[j-1];
        
        for( i=l; i<=j; i++ ) {
          yr = ar[i][j-1];
          yi = 0.0;
          zzr = ar[i-1][j-1];
          zzi = ai[i-1][j-1];
          if ( i != j ) {
            yi = ai[i][j-1];
            ai[i][j-1] = xr * yi + xi * yr + ai[j][j-1] * zzi;
          }
          ar[i][j-1] = xr * yr - xi * yi + ai[j][j-1] * zzr;
          ar[i-1][j-1] = xr * zzr + xi * zzi - ai[j][j-1] * yr;
          ai[i-1][j-1] = xr * zzi - xi * zzr - ai[j][j-1] * yi;
        }
      }
      
      if ( fabs( si ) < eps ) continue;
      
      for( i=l; i<=en; i++ ) {
        yr = ar[i][en];
        yi = ai[i][en];
        ar[i][en] = sr * yr - si * yi;
        ai[i][en] = sr * yi + si * yr;
      }
      continue;
      /***	a root found		***/
    }/*660*/
    eigenvalr[en-1] = ar[en-1][en-1] + tr;
    eigenvali[en-1] = ai[en-1][en-1] + ti;
    
    en = enm1;
    its = 0;
    enm1 = en - 1;
  }
  
  return;
  /***	set error -- all eigenvalues have not converged after 30*n iterations 	***/
  ierr = en;
}



void	comqr2(
  int n, int low, int high, double *ortr, double *orti,
  complex **a, complex *eigenval, complex **eigenvec, int ierr
  )
{
  /*
  calculation of Eigenvalue and Eigenvector by complex QR method
   */
  int		i,j,k,l,m,en,ii,jj,ll,nn,ip1,itn,its,lp1,enm1,iend;
  double	si,sr,ti,tr,xi,xr,yi,yr,zzi,zzr,norm,tst1,tst2;
  double	length_sq;
  
  ierr = 0;
  /* initialize eigenvector matrix */
  for( j=1; j<=n; j++ ) {
    for( i=1; i<=n; i++ ) {
      eigenvec[i-1][j-1].re = 0.0;
      eigenvec[i-1][j-1].im = 0.0;
    }
    eigenvec[j-1][j-1].re = 1.0;
  }
  /*
  form the matrix of accumulated transformations
  from the information left by corth
  */
  iend = high - low - 1;
  
  /****	iend > 0 ***/
  if( iend > 0 ) {
    /*	for i=high-1 step -1 until low+1 do --	*/
    for( ii=1; ii<=iend; ii++ ) {
      i = high - ii;
      if ( fabs( ortr[i] ) < eps && fabs( orti[i] ) < eps )
        { continue;}
      if ( fabs( a[i-1][i-2].re ) < eps && fabs( a[i-1][i-2].im ) < eps )
        { continue;}
      /***	norm below is negative of h formed in corth		***/
      norm = a[i-1][i-2].re * ortr[i] + a[i-1][i-2].im * orti[i];
      ip1 = i + 1;
      
      for( k=ip1; k<=high; k++ ) {
        ortr[k] = a[k-1][i-2].re;
        orti[k] = a[k-1][i-2].im;
      }
      
      for( j=i; j<=high; j++ ) {
        sr = 0.0;
        si = 0.0;
        for( k=i; k<=high; k++ ) {
          sr = sr + ortr[k] * eigenvec[k-1][j-1].re
            + orti[k] * eigenvec[k-1][j-1].im;
          si = si + ortr[k] * eigenvec[k-1][j-1].im
            - orti[k] * eigenvec[k-1][j-1].re;
        }
        sr = sr / norm;
        si = si / norm;
        for( k=i; k<=high; k++ ) {
          eigenvec[k-1][j-1].re
            = eigenvec[k-1][j-1].re + sr * ortr[k] - si * orti[k];
          eigenvec[k-1][j-1].im
            = eigenvec[k-1][j-1].im + sr * orti[k] + si * ortr[k];
        }
      }
    }
  }

  /* dbg_print_cmat(1, "$$a", n, a);*/

  /****	iend = 0 ***/
  if( iend >= 0 ) {
    /***	create real subdiagonal elements 	***/
    l = low + 1;
    for( i=l; i<=high; i++ ) {
      ll = min0( i+1, high );
      if ( fabs(a[i-1][i-2].im) < eps ) continue;
      /* printf("$$ i=%d, %lf %lf\n", i, a[i-1][i-2].re, a[i-1][i-2].im);*/
      norm = pythag( a[i-1][i-2].re, a[i-1][i-2].im );
      yr = a[i-1][i-2].re / norm;
      yi = a[i-1][i-2].im / norm;
      a[i-1][i-2].re = norm;
      a[i-1][i-2].im = 0.0;
      
      for( j=i; j<=n; j++ ) {
        si = yr * a[i-1][j-1].im - yi * a[i-1][j-1].re;
        a[i-1][j-1].re = yr * a[i-1][j-1].re + yi * a[i-1][j-1].im;
        a[i-1][j-1].im = si;
      }
      for( j=1; j<=ll; j++ ) {
        si = yr * a[j-1][i-1].im + yi * a[j-1][i-1].re;
        a[j-1][i-1].re = yr * a[j-1][i-1].re - yi * a[j-1][i-1].im;
        a[j-1][i-1].im = si;
      }
      for( j=low; j<=high; j++ ) {
        si = yr * eigenvec[j-1][i-1].im + yi * eigenvec[j-1][i-1].re;
        eigenvec[j-1][i-1].re = yr * eigenvec[j-1][i-1].re
          - yi * eigenvec[j-1][i-1].im;
        eigenvec[j-1][i-1].im = si;
      }
    }
  }
  
  /* dbg_print_cmat(1, "$$a", n, a); */
  
  /****	iend < 0 ***/
  /***	store roots isolated by cbal	***/
  for( i=1; i<=n; i++ ) {
    if ( i >= low && i <= high ) continue;
    eigenval[i-1].re = a[i-1][i-1].re;
    eigenval[i-1].im = a[i-1][i-1].im;
  }
  en = high;		/*printf("$$ en=high=%d\n",en);*/
  tr = 0.0;
  ti = 0.0;
  itn = 30 * n;
  
  /* Search for next eigen value */
  while( 1 ) {
    /*printf("$$ en low =%d\n", en);*/
    /*220*/	if( en < low ) break;/*goto 680*/
    
    its = 0;
    enm1 = en - 1;
    
    while( 1 ) {
      /***	look for single small sub-diagonal element		***/
      /*		for l=en step -1 until low do --	*/
      /*printf("$$ low, en = %d %d\n", low, en);*/
      /*240*/		for( ll=low; ll<=en; ll++ ) {
        l = en + low - ll;
        if ( l == low ) break;
        tst1 = fabs( a[l-2][l-2].re ) + fabs( a[l-2][l-2].im )
          + fabs( a[l-1][l-1].re ) + fabs( a[l-1][l-1].im );
        tst2 = tst1 + fabs( a[l-1][l-2].re );
        /*printf( "$$ll=%d, fabs %lf\n", ll, fabs(a[l-1][l-2].re) );*/
        if (fabs(a[l-1][l-2].re) < eps) break;
      }
      
      /***	form shift	***/
      
      /*printf("$$ itn, its, l = %d %d %d\n", itn, its, l);*/
      
      /*300*/		if ( l == en ) break;
      else {
        if ( itn == 0 ) { ierr = en; return; }
        if ( its == 10 || its == 20 ) {
          /***	form exceptional shift		***/
          if( en<3 ){
            printf("$$ en=%d, abort.\n",en); exit(EXIT_FAILURE);
          }
          sr = fabs( a[en-1][enm1-1].re ) + fabs( a[enm1-1][en-3].re );
          si = 0.0;
        }
        else {
          sr = a[en-1][en-1].re;
          si = a[en-1][en-1].im;
          xr = a[enm1-1][en-1].re * a[en-1][enm1-1].re;
          xi = a[enm1-1][en-1].im * a[en-1][enm1-1].re;
          
          if ( fabs( xr ) > eps || fabs( xi ) > eps ) {/*goto 340*/
            yr = ( a[enm1-1][enm1-1].re - sr ) / 2.0;
            yi = ( a[enm1-1][enm1-1].im - si ) / 2.0;
            csroot( yr*yr-yi*yi+xr, 2.0*yr*yi+xi, &zzr, &zzi );
            if ( yr * zzr + yi * zzi < 0.0 ) {
              zzr = -zzr;
              zzi = -zzi;
            }
            cdiv_cg( xr, xi, yr+zzr, yi+zzi, &xr, &xi );
            sr -= xr;
            si -= xi;
          }
        }
        
        /*340*/
        for( i=low; i<=en; i++ ) {
          a[i-1][i-1].re -= sr;
          a[i-1][i-1].im -= si;
        }
        
        tr += sr;
        ti += si;
        its++;
        itn--;
        
        /***	reduce to triangle (rows)	***/
        lp1 = l + 1;
        
        for( i=lp1; i<=en; i++ ) {
          sr = a[i-1][i-2].re;
          a[i-1][i-2].re = 0.0;
          norm = pythag( pythag( a[i-2][i-2].re, a[i-2][i-2].im ), sr );
          xr = a[i-2][i-2].re / norm;	eigenval[i-2].re = xr;
          xi = a[i-2][i-2].im / norm;	eigenval[i-2].im = xi;
          a[i-2][i-2].re = norm;
          a[i-2][i-2].im = 0.0;
          a[i-1][i-2].im = sr / norm;
          
          for( j=i; j<=n; j++ ) {
            yr = a[i-2][j-1].re;
            yi = a[i-2][j-1].im;
            zzr = a[i-1][j-1].re;
            zzi = a[i-1][j-1].im;
            a[i-2][j-1].re = xr * yr + xi * yi + a[i-1][i-2].im * zzr;
            a[i-2][j-1].im = xr * yi - xi * yr + a[i-1][i-2].im * zzi;
            a[i-1][j-1].re = xr * zzr - xi * zzi - a[i-1][i-2].im * yr;
            a[i-1][j-1].im = xr * zzi + xi * zzr - a[i-1][i-2].im * yi;
            
          }
        }
        
        si = a[en-1][en-1].im;
        if ( fabs(si) > eps ) {
          norm = pythag( a[en-1][en-1].re, si );
          sr = a[en-1][en-1].re / norm;
          si = si / norm;
          a[en-1][en-1].re = norm;
          a[en-1][en-1].im = 0.0;
          if ( en != n ) {
            ip1 = en + 1;
            
            for( j=ip1; j<=n; j++ ) {
              yr = a[en-1][j-1].re;
              yi = a[en-1][j-1].im;
              a[en-1][j-1].re = sr * yr + si * yi;
              a[en-1][j-1].im = sr * yi - si * yr;
            }
          }
        }
        
        /***	inverse operation	***/
        for( j=lp1; j<=en; j++ ) {
          xr = eigenval[j-2].re;
          xi = eigenval[j-2].im;
          
          for( i=1; i<=j; i++ ) {
            yr = a[i-1][j-2].re;
            yi = 0.0;
            zzr = a[i-1][j-1].re;
            zzi = a[i-1][j-1].im;
            if ( i != j ) {
              yi = a[i-1][j-2].im;
              a[i-1][j-2].im = xr * yi + xi * yr + a[j-1][j-2].im * zzi;
            }
            a[i-1][j-2].re = xr * yr - xi * yi + a[j-1][j-2].im * zzr;
            a[i-1][j-1].re = xr * zzr + xi * zzi - a[j-1][j-2].im * yr;
            a[i-1][j-1].im = xr * zzi - xi * zzr - a[j-1][j-2].im * yi;
          }
          for( i=low; i<=high; i++ ) {
            yr = eigenvec[i-1][j-2].re;
            yi = eigenvec[i-1][j-2].im;
            zzr = eigenvec[i-1][j-1].re;
            zzi = eigenvec[i-1][j-1].im;
            eigenvec[i-1][j-2].re = xr * yr - xi * yi + a[j-1][j-2].im * zzr;
            eigenvec[i-1][j-2].im = xr * yi + xi * yr + a[j-1][j-2].im * zzi;
            eigenvec[i-1][j-1].re = xr * zzr + xi * zzi - a[j-1][j-2].im * yr;
            eigenvec[i-1][j-1].im = xr * zzi - xi * zzr - a[j-1][j-2].im * yi;
          }
        }
        
        if ( fabs( si ) < eps ) {continue;}
        
        for( i=1; i<= en; i++ ) {
          yr = a[i-1][en-1].re;
          yi = a[i-1][en-1].im;
          a[i-1][en-1].re = sr * yr - si * yi;
          a[i-1][en-1].im = sr * yi + si * yr;
        }
        for( i=low; i<=high; i++ ) {
          yr = eigenvec[i-1][en-1].re;
          yi = eigenvec[i-1][en-1].im;
          eigenvec[i-1][en-1].re = sr * yr - si * yi;
          eigenvec[i-1][en-1].im = sr * yi + si * yr;
        }
        continue;
      }
    }/*goto 240*/
    
    /***	a root found		***/
    /*660*/	a[en-1][en-1].re += tr;	eigenval[en-1].re = a[en-1][en-1].re;
    a[en-1][en-1].im += ti;	eigenval[en-1].im = a[en-1][en-1].im;
    
    en = enm1;  		/*printf("$$ en=enm1=%d\n",en);*/
    /*		its = 0;*/
    /*		enm1 = en - 1;*/
  }	/*goto 220*/
  
  /*
     all roots found.
     backsubstitute to find vectors of upper triangular form
   */
  /*680*/
  norm = 0.0;
  
  for( i=1; i<=n; i++ ) {
    for( j=i; j<=n; j++ ) {
      tr = fabs( a[i-1][j-1].re ) + fabs( a[i-1][j-1].im );
      if ( tr > norm ) norm = tr;
    }
  }
  
  if ( n == 1 || fabs(norm) < eps ) return;
  
  /*		for en=n step -1 until 2 do -- 		*/
  for( nn=2; nn<=n; nn++ ) {
    en = n + 2 - nn;			/*printf("$$ en = n + 2 - nn = %d\n", en);*/
    xr = eigenval[en-1].re;
    xi = eigenval[en-1].im;
    a[en-1][en-1].re = 1.0;
    a[en-1][en-1].im = 0.0;
    enm1 = en - 1;
    
    /*		for i=en-1 step -1 until 1 do --	*/
    for( ii=1; ii<=enm1; ii++ ) {
      i = en - ii;
      zzr = 0.0;
      zzi = 0.0;
      ip1 = i + 1;
      
      for( j=ip1; j<=en; j++ ) {
        zzr = zzr + a[i-1][j-1].re * a[j-1][en-1].re
          - a[i-1][j-1].im * a[j-1][en-1].im;
        zzi = zzi + a[i-1][j-1].re * a[j-1][en-1].im
          + a[i-1][j-1].im * a[j-1][en-1].re;
      }
      
      yr = xr - eigenval[i-1].re;
      yi = xi - eigenval[i-1].im;
      
      if ( fabs( yr ) == 0.0 && fabs( yi ) == 0.0 ) {/* continue;*/
        tst1 = norm;
        yr = tst1;
        
        while(1) {
          yr = 0.01 * yr;
          tst2 = norm + yr;
          if( tst2 <= tst1 ) break;
        }
      }
      cdiv_cg( zzr, zzi, yr, yi, &a[i-1][en-1].re, &a[i-1][en-1].im);
      
      /* overflow control */
      tr = fabs(a[i-1][en-1].re) + fabs(a[i-1][en-1].im);
      if ( fabs( tr ) < eps ) continue;
      tst1 = tr;
      tst2 = tst1 + 1.0/tst1;
      if (tst2 > tst1) continue;
      
      for( j=i; j<=en; j++ ) {
        a[j-1][en-1].re = a[j-1][en-1].re / tr;
        a[j-1][en-1].im = a[j-1][en-1].im / tr;
      }
    }
  }
  
  /* end backsubstitution vectors of isolated roots */
  
  for( i=1; i<=n; i++ ) {
    if ( i >= low && i <= high ) continue;
    
    for( j=i; j<=n; j++ ) {
      eigenvec[i-1][j-1].re = a[i-1][j-1].re;
      eigenvec[i-1][j-1].im = a[i-1][j-1].im;
    }
  }
  
  /*
	for( i=0; i<n; i++ ) {
		for( j=0; j<n; j++ ) {
			printf( "eigenvec[%d][%d]=%lg,%lg ", i, j, eigenvec[i][j].re, eigenvec[i][j].im );
			if( j%10 == 9 ) printf( "\n" );
		}
		printf( "\n" );
	}
   */	
  /*
   multiply by transformation matrix to give vectors of original full matrix.
  */
  /*		for j=n step -1 until low do --	*/
  for( jj=low; jj<=n; jj++ ) {
    j = n + low - jj;
    m = min0( j, high );
    for( i=low; i<=high; i++ ) {
      zzr = 0.0;
      zzi = 0.0;
      for( k=low; k<=m; k++ ) {
        zzr = zzr + eigenvec[i-1][k-1].re * a[k-1][j-1].re
          - eigenvec[i-1][k-1].im * a[k-1][j-1].im;
        zzi = zzi + eigenvec[i-1][k-1].re * a[k-1][j-1].im
          + eigenvec[i-1][k-1].im * a[k-1][j-1].re;
      }
      eigenvec[i-1][j-1].re = zzr;
      eigenvec[i-1][j-1].im = zzi;
    }
  }
  /***	normalization of the eigenvectors	***/
  for( j=1; j<=n; j++ ) {
    length_sq = 0.0;
    for( i=1; i<=n; i++ )
      length_sq += eigenvec[i-1][j-1].re*eigenvec[i-1][j-1].re 
        + eigenvec[i-1][j-1].im*eigenvec[i-1][j-1].im;
    
    if( length_sq < eps ) continue;
    
    for( i=1; i<=n; i++ ) {
      eigenvec[i-1][j-1].re = eigenvec[i-1][j-1].re / sqrt( length_sq );
      eigenvec[i-1][j-1].im = eigenvec[i-1][j-1].im / sqrt( length_sq );
    }
  }
  /* set error -- all eigenvalues have not converged after 30*n iterations */
  ierr = en;
  /*printf("$$ ierr = en = %d\n", en);*/
  
  return;
}


void corth(
  int n, int low, int high, complex **a, double *ortr, double *orti
  )
{
  /*  
    Transformation to Hessenberg matrix
   */
  int		i,j,m,ii,jj,mp;
  double	f,g,h,fi,fr,scale;
  
  /*dbg_print_cmat(1, "$$a", n, a);*/
  /*printf("$$ n=%d, low=%d, high=%d\n", n, low, high);*/
  
  if ( high-1 < low+1 ) return;
  
  for( m=low+1; m<=high-1; m++ ) {	/*	1 <= m <= n-1	*/
    h = 0.0;
    ortr[m] = 0.0;
    orti[m] = 0.0;
    scale = 0.0;
    
    /***		scale column (algol tol then not needed)	***/
    /*printf("$$ m=%d\n", m);*/
    for( i=m; i<=high; i++ ){
      scale += fabs( a[i-1][m-2].re ) + fabs( a[i-1][m-2].im );
      /*printf("$$ a[%d][%d].re=%lf, im=%lf, scale=%lf\n",
				i-1, m-2, a[i-1][m-2].re, a[i-1][m-2].im, scale );*/
    }
    
    /*printf("$$ scale=%lf\n", scale);*/
    if ( fabs(scale) < eps ) continue;
    
    mp = m + high;
    /*		for i=high step -1 until m do --		*/
    for( ii=m; ii<=high; ii++ ) {
      i = mp - ii;
      ortr[i] = a[i-1][m-2].re / scale;
      orti[i] = a[i-1][m-2].im / scale;
      h += ortr[i] * ortr[i] + orti[i] * orti[i];
      /*printf("$$ i=%d, ortr=%lf, orti=%lf, h=%lf\n", i, ortr[i], orti[i], h);*/
    }
    
    /*		for( i=m; i<high; i++ ) printf( "%.3lf+%.3lfi ", ortr[i],orti[i] ); printf( "\n" );*/
    g = sqrt(h);
    f = pythag( ortr[m], orti[m] );
    /*printf("$$ h=%lf, g=%lg, f=%lg, ortr=%lf, orti=%lf\n", h, g, f, ortr[m], orti[m]);
		printf("$$ eps = %lg\n", eps);*/
    if ( fabs( f ) < eps ) {
      ortr[m] = g;
      a[m-1][m-2].re = scale;
    }
    else {
      h += f * g;
      g = g / f;
      ortr[m] = (1.0 + g) * ortr[m];
      orti[m] = (1.0 + g) * orti[m];
    }
    /*printf("$$ h=%lf, g=%lg, f=%lg, ortr=%lf, orti=%lf\n", h, g, f, ortr[m], orti[m]);*/
    
    /***	form (i-(u*ut)/h) * a	***/
    for( j=m; j<=n; j++ ) {
      fr = 0.0;
      fi = 0.0;
      /*	for i=high step -1 until m do --	*/
      for( ii=m; ii<=high; ii++ ) {
        i = mp - ii;
        fr = fr + ortr[i] * a[i-1][j-1].re + orti[i] * a[i-1][j-1].im;
        fi = fi + ortr[i] * a[i-1][j-1].im - orti[i] * a[i-1][j-1].re;
      }
      fr = fr / h;
      fi = fi / h;
      for( i=m; i<=high; i++) {
        a[i-1][j-1].re = a[i-1][j-1].re - fr * ortr[i] + fi * orti[i];
        a[i-1][j-1].im = a[i-1][j-1].im - fr * orti[i] - fi * ortr[i];
      }
    }
    
    /***	form (i-(u*ut)/h)*a*(i-(u*ut)/h)	***/
    for ( i=1; i<=high; i++ ) {
      fr = 0.0;
      fi = 0.0;
      /*		for j=high step -1 until m do --	*/
      for( jj=m; jj<=high; jj++ ) {
        j = mp - jj;
        fr = fr + ortr[j] * a[i-1][j-1].re - orti[j] * a[i-1][j-1].im;
        fi = fi + ortr[j] * a[i-1][j-1].im + orti[j] * a[i-1][j-1].re;
      }
      fr = fr / h;
      fi = fi / h;
      for( j=m; j<=high; j++ ) {
        a[i-1][j-1].re = a[i-1][j-1].re - fr * ortr[j] - fi * orti[j];
        a[i-1][j-1].im = a[i-1][j-1].im + fr * orti[j] - fi * ortr[j];
      }
    }
    ortr[m] = scale * ortr[m];
    orti[m] = scale * orti[m];
    a[m-1][m-2].re = -g * a[m-1][m-2].re;
    a[m-1][m-2].im = -g * a[m-1][m-2].im;
    
    /*printf("$$ a[%d][%d].re=%lf, im=%lf, g=%lf\n",
			m-1, m-2, a[m-1][m-2].re, a[m-1][m-2].im, g);*/
  }
  
}


void	cdiv_cg(double ar, double ai, double br, double bi, double *cr, double *ci)
{
  /*
   *     complex division, (cr,ci) = (ar,ai)/(br,bi)
   */
  double s,ars,ais,brs,bis;
  
  s	= fabs( br ) + fabs( bi );
  /*
	if( fabs( s ) < eps ) {
		printf( "alart! division by 0 (cdiv_cg()) \n");
	}
   */
  ars = ar / s;	ais = ai / s;
  brs = br / s;	bis = bi / s;
  s	= brs * brs + bis * bis;
  *cr	= ( ars * brs + ais * bis ) / s;
  *ci	= ( ais * brs - ars * bis ) / s;
  return;
}


void	csroot(double xr, double xi, double *yr, double *yi)
{
  /*
   *		(yr,yi) = complex sqrt(xr,xi) 
   *		branch chosen so that yr >= 0.0 and sign(yi) == sign(xi)
   */
  double s,tr,ti;
  
  tr = xr;
  ti = xi;
  s = sqrt(0.5*(pythag(tr,ti) + fabs(tr)));
  if (tr >= 0.0) *yr = s;
  if (ti < 0.0) s = -s;
  if (tr <= 0.0) *yi = s;
  if (tr < 0.0) *yr = 0.5*( ti / *yi );
  if (tr > 0.0) *yi = 0.5*( ti / *yr );
}



double	pythag(double a, double b)
{
  /*
   *    finds sqrt(a**2+b**2) without overflow or destructive underflow
   */
  double p,r,s,t,u;
  
  p = dmax1( fabs(a), fabs(b) );
  if (p == 0.0) return( 0.0 );
  r = ( dmin1(fabs(a),fabs(b)) / p ) * ( dmin1(fabs(a),fabs(b)) / p );
  
  while(1) {
    t = 4.0 + r;
    if (t == 4.0) break;
    s = r/t;
    u = 1.0 + 2.0*s;
    p = u*p;
    r = (s/u)*(s/u) * r;
  }
  return( p );
}



void	cbabk2( int n, int low, int high, double *scale, int m, complex **eigenvec )
{
  int i,j,k,ii;
  double s;
  
  if (m == 0) return;
  
  if ( high != low ) {
    for( i=low; i<=high; i++ ) {
      s = scale[i];
      /***
		left hand eigenvectors are back transformed
			 if the foregoing statement is replaced by
			 s=1.0/scale(i).
       ***/
      for( j=1; j<=m; j++ ) {
        eigenvec[i-1][j-1].re = eigenvec[i-1][j-1].re * s;
        eigenvec[i-1][j-1].im = eigenvec[i-1][j-1].im * s;
      }
    }
  }
  /*	for i=low-1 step -1 until 1,
			high+1 step 1 until n do --	*/
  else {
    for( ii=1; ii<=n; ii++ ) {
      i = ii;
      if (i >= low && i <= high) break;
      if (i < low) i = low - ii;
      k = (int) scale[i];
      if (k == i) break;
      
      for( j=1; j<=m; j++ ) {
        s = eigenvec[i-1][j-1].re;
        eigenvec[i-1][j-1].re = eigenvec[k-1][j-1].re;
        eigenvec[k-1][j-1].re = s;
        s = eigenvec[i-1][j-1].im;
        eigenvec[i-1][j-1].im = eigenvec[k-1][j-1].im;
        eigenvec[k-1][j-1].im = s;
      }
    }
  }
}



void	matexchange_rowcol( int n, int j, int k, int l, int m, complex **a )
     /*
 n:order(rank) of matrix a
       
      */
{
  int		i;
  double	f;
  
  if ( j != m ) {
    for( i=1; i<=l; i++ ) {
      f = a[i-1][j-1].re;	a[i-1][j-1].re = a[i-1][m-1].re;	a[i-1][m-1].re = f;
      f = a[i-1][j-1].im;	a[i-1][j-1].im = a[i-1][m-1].im;	a[i-1][m-1].im = f;
    }
    for( i=k; i<=n; i++ ) {
      f = a[j-1][i-1].re;	a[j-1][i-1].re = a[m-1][i-1].re;	a[m-1][i-1].re = f;
      f = a[j-1][i-1].im;	a[j-1][i-1].im = a[m-1][i-1].im;	a[m-1][i-1].im = f;
    }
  }
}



void	cbal( int n, complex **a, int *low, int *high, double *scale )
{
  int		i,j,k,l,m;
  double	c,f,g,r,s,b2,radix;
  int		noconv=0, all_zero=1;
  
  radix = 16.0;
  
  b2 = radix * radix;
  k = 1;
  l = n;
  
  for( j=l; j>=1; j-- ) {
    for( i=1; i<=l; i++ ) {
      if ( i == j ) continue;
      /*	check all ar[j-1][i-1] & ai[j-1][i-1] = 0 or not (0<i<l)	*/
      if ( fabs( a[j-1][i-1].re ) > eps || fabs( a[j-1][i-1].im ) > eps ) { all_zero = 0; break;}
    }
    /*	case of all ar[j-1][i-1] & ai[j-1][i-1] = 0 (0<i<l)	*/
    if( all_zero == 1 ) {
      m = l;
      scale[m] = j;
      /***	in-line procedure for row and column exchange	***/
      if( j != m ) {
        matexchange_rowcol( n, j, k, l, m, a );
        j = l;
      }
      
      if ( l == 1 ) {
        *low = k;
        *high = l;
        return;
      }
      l--;
    }
    all_zero = 1;
  }
  
  for( j=k; j<=l; j++ ) {
    
    for( i=k; i<=l; i++ ) {
      if ( i == j ) continue;
      if ( fabs( a[i-1][j-1].re ) > eps || fabs( a[i-1][j-1].im ) > eps ) { all_zero = 0; break; }
    }
    /*	case of all ar[i-1][j-1] & ai[i-1][j-1] = 0 (k<i<l)	*/
    if( all_zero == 1 ) {
      m = k;
      scale[m] = j;
      /***	in-line procedure for row and column exchange	***/
      if( j != m ) {
        matexchange_rowcol( n, j, k, l, m, a );
        j = k;
      }
      k++;
    }
    all_zero = 1;
  }
  
  /***	 now balance the submatrix in rows k to l	***/
  for( i=k; i<=l; i++ ) scale[i] = 1.0;
  
  /***	iterative loop for norm reduction	***/
  /*190*/
  while( 1 ) {
    for( i=k; i<=l; i++ ) {
      c = 0.0;
      r = 0.0;
      
      for( j=k; j<=l; j++ ) {
        if ( j == i ) continue;
        c += fabs( a[j-1][i-1].re ) + fabs( a[j-1][i-1].im );
        r += fabs( a[i-1][j-1].re ) + fabs( a[i-1][j-1].im );
      }
      
      /***	guard against zero c or r due to underflow 		***/
      if ( fabs(c) < eps || fabs(r) < eps ) continue;
      
      g = r / radix;
      f = 1.0;
      s = c + r;
      
      while( 1 ) {
        if ( c >= g ) break;
        else {
          f = f * radix;
          c = c * b2;
        }
      }
      
      g = r * radix;
      
      while( 1 ) {
        if ( c < g ) break;
        else {
          f = f / radix;
          c = c / b2;
        }
      }
      
      /***	now balance		***/
      if ( (c + r) / f >= 0.95 * s) {
        continue;
      }
      g = 1.0 / f;
      scale[i] = scale[i] * f;
      noconv = 1;
      for( j=k; j<=n; j++ ) { a[i-1][j-1].re = a[i-1][j-1].re * g;	a[i-1][j-1].im = a[i-1][j-1].im * g; }
      for( j=1; j<=l; j++ ) { a[j-1][i-1].re = a[j-1][i-1].re * f;	a[j-1][i-1].im = a[j-1][i-1].im * f; }
    }
    
    if( noconv != 1 ) break;
  }
  
  *low = k;
  *high = l;
}


/*******************************************************************************************
	Sorting eigenvalues and eigenvectors according to "eigenval[i].re"
 *******************************************************************************************/
void	sort_eigval_eigvec( int ne, complex *eigenval, complex **eigenvec )
{
  int		i,j,k,l,minimum_value;
  complex *eigenval_sort, **eigenvec_sort;
  
  eigenval_sort = new_vector( complex, ne );		chk_vector( eigenval_sort );
  eigenvec_sort = new_matrix( complex, ne, ne );	chk_matrix( eigenvec_sort );
  
  /*	for 0th eigenvalue and eigenvector	*/
  eigenval_sort[0] = eigenval[0];
  for( l=0; l<ne; l++ ) eigenvec_sort[l][0] = eigenvec[l][0];
  
  /* for 1st to ne-th eigenvalues and eigenvectors	*/
  for( i=1; i<ne; i++ ) {
    
    /*	compare with already sort eigenvalues and eigenvectors	*/
    minimum_value = 1;
    for( j=0; j<i; j++ ) {
      if( eigenval[i].re > eigenval_sort[j].re ) {
        /*	shift	( [i-1] -> [i] )	*/
        for( k=i-1; k>=j; k-- ) {
          eigenval_sort[k+1] = eigenval_sort[k];
          for( l=0; l<ne; l++ ) eigenvec_sort[l][k+1] = eigenvec_sort[l][k];
        }
        eigenval_sort[j] = eigenval[i];
        for( l=0; l<ne; l++ ) eigenvec_sort[l][j] = eigenvec[l][i];
        minimum_value = 0;
        break;
      }
    }
    if( minimum_value == 1 ) {
      eigenval_sort[i] = eigenval[i];
      for( l=0; l<ne; l++ ) eigenvec_sort[l][i] = eigenvec[l][i];
    }
  }
  
  for( i=0; i<ne; i++ ) {
    eigenval[i] = eigenval_sort[i];
    for( j=0; j<ne; j++ ) eigenvec[i][j] = eigenvec_sort[i][j];
  }
  
  free_vector( eigenval_sort );
  free_matrix( eigenvec_sort );
}
