/*
  e_density.c
  */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "mbconfig.h"
#include "fitting.h"
#include "mb_input.h"
#include "data_out.h"
#include "mb_sim.h"
#include "mb_mpi.h"
#define  BUFSIZE  256
extern DEBUG_INFO dbg;
extern EPS_CTRL	epsilon;


/* ****************************** */
/* *****   E_DENSITY part   ***** */
/* ****************************** */

void fget_sd_beams_3d(SIM_DATA *sd, CRYSTAL_STRCT_DATA *csd, FILE *fi)
     /*
      *  Beam search 3D updated on 2001.06.16
      */
{
  HKL *g_exact, *g_bethe;
  int nr;  /*, g_table_size; */
  
  sd->beam_search.flag_bs = fget_i(fi);
  if(
    sd->beam_search.flag_bs == 0
    ){	/* Manual search */
    fget_2i(&(sd->num_exact), &(sd->num_bethe), fi);
    printf("Exact = %d, Bethe = %d\n", sd->num_exact, sd->num_bethe);
    nr = sd->num_exact + sd->num_bethe;
    sd->reflex = new_vector(HKL, nr);	chk_vector(sd->reflex);
    fget_beams( sd->num_exact, sd->num_bethe, sd->reflex, fi );
    printf("$$ vg_select_flag is set to 0.\n");
    epsilon.vg_select_flag = 0;
  }
  else if(
    sd->beam_search.flag_bs == 1
    ){	/* Automatic search (wg) */
    fget_beam_search_wg(fi, &(sd->beam_search));
    
    /*g_table_size = SIZE_GTABLE;*/
    g_exact = new_vector(HKL, sd->beam_search.nglim);	chk_vector(g_exact);
    g_bethe = new_vector(HKL, sd->beam_search.nglim);	chk_vector(g_bethe);
    printf("Beam search (wg)... ");	fflush(stdout);	
    beam_search_sphere(
      &(sd->num_exact), &(sd->num_bethe), g_exact, g_bethe,
      sd->beam_search, sd->incidence, sd->zolzbase1, sd->H_tilt, sd->G_tilt,
      sd->accel_kvol, csd
      );
    printf("Exact = %d, Bethe = %d\n", sd->num_exact, sd->num_bethe);
    nr = sd->num_exact + sd->num_bethe;
    sd->reflex = new_vector(HKL, nr);	chk_vector(sd->reflex);
    beam_merge(sd->num_exact, sd->num_bethe, g_exact, g_bethe, sd->reflex);
    free_vector(g_bethe);
    free_vector(g_exact);
  }
  else if(
    sd->beam_search.flag_bs == 2
    ){	/* Automatic search (sg) */
    fget_beam_search_sg(fi, &(sd->beam_search) );
    
    /*g_table_size = SIZE_GTABLE;*/
    g_exact = new_vector(HKL, sd->beam_search.nglim);	chk_vector(g_exact);
    g_bethe = new_vector(HKL, sd->beam_search.nglim);	chk_vector(g_bethe);
    printf("Beam search (sg)... ");	fflush(stdout);	
    beam_search_sphere(
      &(sd->num_exact), &(sd->num_bethe), g_exact, g_bethe,
      sd->beam_search, sd->incidence, sd->zolzbase1, sd->H_tilt, sd->G_tilt,
      sd->accel_kvol, csd
      );
    printf("Exact = %d, Bethe = %d\n", sd->num_exact, sd->num_bethe);
    nr = sd->num_exact + sd->num_bethe;
    sd->reflex = new_vector(HKL, nr);	chk_vector(sd->reflex);
    beam_merge(sd->num_exact, sd->num_bethe, g_exact, g_bethe, sd->reflex);
    free_vector(g_bethe);
    free_vector(g_exact);
  }
  else if(
    sd->beam_search.flag_bs == 3 || sd->beam_search.flag_bs == 13
    ){	/* Automatic search (sg - multi) */
    fget_beam_search_sg_multi(fi, &(sd->beam_search) );
    
    /*g_table_size = SIZE_GTABLE;*/
    g_exact = new_vector(HKL, sd->beam_search.nglim);	chk_vector(g_exact);
    g_bethe = new_vector(HKL, sd->beam_search.nglim);	chk_vector(g_bethe);
    printf("Beam search (sg)... ");	fflush(stdout);	
    beam_search_sg_multi(
      &(sd->num_exact), &(sd->num_bethe), g_exact, g_bethe,
      sd->beam_search, sd->incidence, sd->zolzbase1, sd->H_tilt, sd->G_tilt,
      sd->accel_kvol, csd
      );
    printf("Exact = %d, Bethe = %d\n", sd->num_exact, sd->num_bethe);
    nr = sd->num_exact + sd->num_bethe;
    sd->reflex = new_vector(HKL, nr);	chk_vector(sd->reflex);
    beam_merge(sd->num_exact, sd->num_bethe, g_exact, g_bethe, sd->reflex);
    free_vector(g_bethe);
    free_vector(g_exact);
  }
  else{
    fprintf(
      stderr, "Flag_bs = %d is invalid for 3D Density calc.\n",
      sd->beam_search.flag_bs
      );
    exit(EXIT_FAILURE);
  }
}


#if 0
void input_density_proj_data_old(SIM_DATA *sd, CRYSTAL_STRCT_DATA *csd, FILE *fi)
{
  CELL_PAR  bv;

  bv = bravai(csd->lp);

  fget_inc_wave(
    &(sd->accel_kvol), &(sd->incidence), &(sd->surf_norm), &(sd->zolzbase1), fi
    );
  fget_2d( &(sd->H_tilt), &(sd->G_tilt), fi );

  zolz_ort_base_two_uvw(
    sd->incidence, sd->zolzbase1, bv, &(sd->x1), &(sd->x2)
    );
  printf("real-space base-vector ( length = 1/|zolzbase1| )\n");
  printf(
    "x1 = %lf %lf %lf, x2 = %lf %lf %lf\n",
    sd->x1.u, sd->x1.v, sd->x1.w, sd->x2.u, sd->x2.v, sd->x2.w
    );

  sd->calc_type = 0;	sd->num_points = 1;
  fget_2d(&(sd->H_begin), &(sd->G_begin), fi);
  fget_2d(&(sd->x_begin), &(sd->x_end), fi);
  fget_2d(&(sd->y_begin), &(sd->y_end), fi);
  fget_2i(&(sd->num_points_x), &(sd->num_points_y), fi);
  chk_data("CALC_END",fi);

  /* Beam search */
  fget_sd_beams( sd, csd, fi );
  chk_data("BEAM_END",fi);
  reflex_dump_csd(
    0, sd->num_exact, sd->num_bethe, sd->reflex,
    csd, sd->beam_search, sd->incidence, sd->zolzbase1, sd->accel_kvol, bv
    );
  
  /* Branches */
  fget_2i(&(sd->branch_begin), &(sd->branch_end), fi);
  if (
    sd->branch_begin < 1 || sd->branch_begin > sd->num_exact
    || sd->branch_end < 1 || sd->branch_end > sd->num_exact
    || sd->branch_begin > sd->branch_end
    ){
    printf(
      "$$ Invalid paramter: branch_begin = %d, branch_end = %d. Abort.\n",
      sd->branch_begin, sd->branch_end
      );
    exit(EXIT_FAILURE);
  }
  chk_data("BLOCH_END",fi);
  fclose(fi);
  
  /* dummy array for free_sd */
  sd->num_thickness = 0;
  /*	sd->calc_type = -1; */
  sd->intensity = new_tensor(double, 1, 1, 1);
  chk_tensor(sd->intensity);
}
#endif


void input_density_proj_data(SIM_DATA *sd, CRYSTAL_STRCT_DATA *csd, FILE *fi)
{
  char  message[BUFSIZE];
  CELL_PAR  bv;

  bv = bravai(csd->lp);

  chk_data("E_DENSITY_CALCULATION_CONDITION{", fi);

  chk_data("INCIDENT_BEAM{", fi);
  sd->incidence = fget_uvw(fi);
  sd->surf_norm = fget_uvw(fi);
  sd->zolzbase1 = fget_r_hkl(fi);
  sd->H_tilt = fget_d(fi);
  sd->G_tilt = fget_d(fi);
  sd->accel_kvol = fget_d(fi);
  chk_data("}INCIDENT_BEAM", fi);
  
  chk_data("BEAM_SEARCH{", fi);
  fget_sd_beams( sd, csd, fi );
  reflex_dump_csd(
    0, sd->num_exact, sd->num_bethe, sd->reflex,
    csd, sd->beam_search, sd->incidence, sd->zolzbase1, sd->accel_kvol, bv
    );
  sd->beam_search.disksize = 1.0;  /* caution */
  mb_fprintf(stdout, "$$disksize = %f\n", sd->beam_search.disksize);
  chk_data("}BEAM_SEARCH", fi);

  chk_data("CALC_RXY_RANGE{", fi);
  zolz_ort_base_two_uvw(
    sd->incidence, sd->zolzbase1, bv, &(sd->x1), &(sd->x2)
    );
  mb_fprintf(stdout, "real-space base-vector ( length = 1/|zolzbase1| )\n");
  mb_fprintf(
    stdout, "x1 = %lf %lf %lf, x2 = %lf %lf %lf\n",
    sd->x1.u, sd->x1.v, sd->x1.w, sd->x2.u, sd->x2.v, sd->x2.w
    );
  fget_2i(&(sd->num_points_x), &(sd->num_points_y), fi);
  fget_2d(&(sd->x_begin), &(sd->x_end), fi);
  fget_2d(&(sd->y_begin), &(sd->y_end), fi);
  chk_data("}CALC_RXY_RANGE", fi);

  chk_data("AREA_PROJ_CELL{", fi);
  sd->area_proj_cell = fget_d(fi); /* the area of the projected unit cell [A2]*/
  chk_data("}AREA_PROJ_CELL", fi);

  chk_data("}E_DENSITY_CALCULATION_CONDITION",fi);
  fclose(fi);
  
  /* dummy for free_sd */
  sd->num_g = 0;
  sd->num_thickness = 0;
  sd->calc_type = -1;
  /*
  sd->intensity = new_tensor(double, 1, 1, 1);
  chk_tensor(sd->intensity);
    */
}


void mb_electron_density_proj( FILE *fi, FILE *fo_log, char *infile )
{
  /*
   * Electron density and potential distribution
   */
  int  nr, nx, ny, nmx, nmy, prog, ix, iy;
  double  x_scale, y_scale;
  CRYSTAL_STRCT_DATA csd;
  SIM_DATA sd;
  CELL_PAR bv;
  complex  *vgv, *fxgv, projp, rho;
  R_UVW  x1, x2, rx;
  char  pp_outfile[BUFSIZE], ed_outfile[BUFSIZE];
  char  pp_pltfile[BUFSIZE], ed_pltfile[BUFSIZE];
  FILE  *fo1, *fo2, *fo_pp_plt, *fo_ed_plt;
  
  input_csd_cstr(&csd, fi, infile);    /* crystal structure data */
  input_density_proj_data(&sd, &csd, fi);
  bv = bravai(csd.lp);
  
  nr = sd.num_exact + sd.num_bethe;
  vgv = new_vector(complex, nr);  chk_vector(vgv);
  fxgv = new_vector(complex, nr);  chk_vector(fxgv);
  
  printf("Fourier coef. calculation....\n");
  vgv_fxgv_csd(
    sd.num_exact, sd.num_bethe, sd.reflex,
    &csd, sd.accel_kvol, bv, vgv, fxgv
    );
  
  /* ----- Projected potential and electron density ----- */
  sprintf(pp_outfile, "%s_pp.txt", infile);
  printf("Projected potential        ==> %s\n", pp_outfile );
  fo1 = fopen_chk_wmode(pp_outfile);
  sprintf(ed_outfile, "%s_ed.txt", infile);
  printf("Projected electron density ==> %s\n", ed_outfile );
  fo2 = fopen_chk_wmode(ed_outfile);
  
  fprintf(fo1, "# Projected potential (formed by %d beams)\n", nr);
  fprintf(fo1, "# nx  ny\n");
  fprintf(fo1, "# %d  %d\n", sd.num_points_x, sd.num_points_y);
  /*fprintf(fo1, "# x\ty\tix\tiy\tpotential.re[V])\n");*/
  fprintf(fo1, "# x\ty\tpotential.re[V])\n");
  
  fprintf(fo2, "# Projected electron density (formed by %d beams)\n", nr);
  fprintf(fo2, "# nx  ny\n");
  fprintf(fo2, "# %d  %d\n", sd.num_points_x, sd.num_points_y);
  /*fprintf(fo2, "# x\ty\tix\tiy\trho.re\n");*/
  fprintf(fo2, "# x\ty\tpotential.re[V])\n");

  nx = sd.num_points_x;
  ny = sd.num_points_y;
  nmx = nx - 1;
  nmy = ny - 1;
  x1 = sd.x1;  x2 = sd.x2;

  for(ix=0; ix<nx; ix++){
    progress_indicator(ix, sd.num_points_x);
    if(nmx == 0){
      x_scale = sd.x_begin;
    }
    else{
      x_scale =
        sd.x_begin + (sd.x_end - sd.x_begin) * (double) ix / (double) nmx;
    }
    for(iy=0; iy<ny; iy++){
      if(nmy == 0){
        y_scale = sd.y_begin;
      }
      else{
        y_scale =
          sd.y_begin + (sd.y_end - sd.y_begin) * (double) iy / (double) nmy;
      }
      rx.u = x_scale * x1.u + y_scale * x2.u;
      rx.v = x_scale * x1.v + y_scale * x2.v;
      rx.w = x_scale * x1.w + y_scale * x2.w;
      projp = calc_inv_fourier( rx, nr, sd.reflex, vgv );
      rho =   calc_inv_fourier( rx, nr, sd.reflex, fxgv );
/*      fprintf(
        fo1, "%lf\t%lf\t%d\t%d\t%lf\n",
        x_scale, y_scale, ix, iy, projp.re
        );
      fprintf(
        fo2, "%lf\t%lf\t%d\t%d\t%lf\n",
        x_scale, y_scale, ix, iy, rho.re / sd.area_proj_cell
        );*/
      fprintf(
        fo1, "%lf\t%lf\t%lf\n",
        x_scale, y_scale, projp.re
        );
      fprintf(
        fo2, "%lf\t%lf\t%lf\n",
        x_scale, y_scale, rho.re / sd.area_proj_cell
        );
    }
    fprintf(fo1, "\n");
    fprintf(fo2, "\n");
  }
  printf("100%%.\n");
  fclose(fo1);  fclose(fo2);

  /* plt files */
  sprintf(pp_pltfile, "%s_pp.plt", infile);
  fo_pp_plt = fopen_chk_wmode(pp_pltfile);
  sprintf(ed_pltfile, "%s_ed.plt", infile);
  fo_ed_plt = fopen_chk_wmode(ed_pltfile);
  fprintf(fo_ed_plt, "#");
  fprintf(fo_ed_plt, "#set size square\n");
  fprintf(fo_ed_plt, "#set pm3d\n");
  fprintf(fo_ed_plt, "set view 90,0\n");
  fprintf(fo_ed_plt, "set term postscript\n");
  fprintf(fo_ed_plt, "set output \"%s_ed.ps\"\n", infile);
  fprintf(fo_ed_plt, "splot \"%s\" using 1:2:3 with lines\n", ed_outfile);
  fprintf(fo_ed_plt, "#pause -1\n");
  fprintf(fo_pp_plt, "#");
  fprintf(fo_pp_plt, "#set size square\n");
  fprintf(fo_pp_plt, "#set pm3d\n");
  fprintf(fo_pp_plt, "set view 90,0\n");
  fprintf(fo_pp_plt, "set term postscript\n");
  fprintf(fo_pp_plt, "set output \"%s_pp.ps\"\n", infile);
  fprintf(fo_pp_plt, "splot \"%s\" using 1:2:3 with lines\n", pp_outfile);
  fprintf(fo_pp_plt, "#pause -1\n");
  fclose(fo_pp_plt);  fclose(fo_ed_plt);

  free_vector(vgv);  free_vector(fxgv);
  free_sd(&sd);
  free_csd(&csd);
}


void mb_xrd_electron_density_proj( FILE *fi, FILE *fo_log, char *infile )
{
  /*
   Electron density calculated from X-ray's structure factors
   */
  CRYSTAL_STRCT_DATA csd;
  SIM_DATA sd;
  CELL_PAR bv;
  int		nr, nmx, nmy, prog, ix, iy;
  double	x_scale, y_scale;
  char	pp_outfile[BUFSIZE];
  complex	*fxgv, rho;
  R_UVW	x1, x2, rx;
  FILE	*fo1;

  input_csd_cstr(&csd, fi, infile);    /* crystal structure data */
  input_density_proj_data(&sd, &csd, fi);
  bv = bravai(csd.lp);

  nr = sd.num_exact + sd.num_bethe;
  fxgv = new_vector(complex, nr);	chk_vector(fxgv);
  
  printf("X-ray's structure factors.\n");
  xrd_fxgv_csd(
    sd.num_exact, sd.num_bethe, sd.reflex, &csd, sd.accel_kvol, bv, fxgv
    );
  
  /* ----- Projected potential ----- */
  sprintf(pp_outfile, "%s.pp", infile);
  printf("Projected charge density (XRD): output to %s\n", pp_outfile );
  fo1 = fopen_chk_wmode(pp_outfile);

  fprintf(fo1, "%% Electron density distribution calculated from X-ray Fg.\n");
  fprintf(fo1, "%% (formed of %d beams)\n", nr);
  fprintf(fo1, "x\ty\tix\tiy\trho.re\trho.im\n");
  
  nmx = sd.num_points_x - 1;	nmy = sd.num_points_y - 1;
  x1 = sd.x1;	x2 = sd.x2;
/*  prog = 0; */
  for(ix=0; ix<sd.num_points_x; ix++){
    progress_indicator(ix, sd.num_points_x);
/*    exec_progress(&prog, ix, sd.num_points_x); */
    if(nmx == 0){
      x_scale = sd.x_begin;
    }
    else{
      x_scale = sd.x_begin
        + (sd.x_end - sd.x_begin) * (double) ix / (double) nmx;
    }
    for(iy=0; iy<sd.num_points_y; iy++){
      if(nmy == 0){ y_scale = sd.y_begin; }
      else{
        y_scale = sd.y_begin 
          + (sd.y_end - sd.y_begin) * (double) iy / (double) nmy;
      }
      rx.u = x_scale * x1.u + y_scale * x2.u;
      rx.v = x_scale * x1.v + y_scale * x2.v;
      rx.w = x_scale * x1.w + y_scale * x2.w;
      rho =   calc_inv_fourier( rx, nr, sd.reflex, fxgv );
      fprintf(
        fo1, "%lf\t%lf\t%d\t%d\t%lf\t%lf\n",
        x_scale, y_scale, ix, iy, rho.re, rho.im
        );
    }
  }
  printf("100%%.\n");
  fclose(fo1);
  
  free_vector(fxgv);
  free_sd(&sd);
  free_csd(&csd);
}


void mb_nuclear_charge_density_proj( FILE *fi, FILE *fo_log, char *infile )
{
  /*
   Nuclear charge density distribution (Z-factor)
   */
  CRYSTAL_STRCT_DATA csd;
  SIM_DATA  sd;
  CELL_PAR  bv;
  int       nr, nmx, nmy, prog, ix, iy;
  double    x_scale, y_scale;
  char      pp_outfile[BUFSIZE];
  complex   *ncgv, nuclear_charge;
  R_UVW     x1, x2, rx;
  FILE      *fo1;

  input_csd_cstr(&csd, fi, infile);    /* crystal structure data */
  input_density_proj_data(&sd, &csd, fi);
  bv = bravai(csd.lp);

  nr = sd.num_exact + sd.num_bethe;
  ncgv = new_vector(complex, nr);	chk_vector(ncgv);

  printf("$$ Z structure factors.\n");
  ncgv_csd(
    sd.num_exact, sd.num_bethe, sd.reflex, &csd, sd.accel_kvol, bv, ncgv
    );
  
  /* ----- Projected potential ----- */
  sprintf(pp_outfile, "%s.pp", infile);
  printf("Projected nuclear charge density: output to %s\n", pp_outfile );
  fo1 = fopen_chk_wmode(pp_outfile);

  fprintf(fo1, "%% Nuclear charge density distribtution\n");
  fprintf(fo1, "%% (formed of %d beams)\n", nr);
  fprintf(fo1, "x\ty\tix\tiy\trho.re\trho.im\n");
  
  nmx = sd.num_points_x - 1;	nmy = sd.num_points_y - 1;
  x1 = sd.x1;	x2 = sd.x2;
/*  prog = 0;*/
  for(ix=0; ix<sd.num_points_x; ix++){
    progress_indicator(ix, sd.num_points_x);
/*    exec_progress(&prog, ix, sd.num_points_x); */
    if(nmx == 0){ x_scale = sd.x_begin; }
    else{ x_scale = sd.x_begin + (sd.x_end - sd.x_begin) * (double) ix / (double) nmx; }
    for(iy=0; iy<sd.num_points_y; iy++){
      if(nmy == 0){ y_scale = sd.y_begin; }
      else{ y_scale = sd.y_begin + (sd.y_end - sd.y_begin) * (double) iy / (double) nmy; }
      rx.u = x_scale * x1.u + y_scale * x2.u;
      rx.v = x_scale * x1.v + y_scale * x2.v;
      rx.w = x_scale * x1.w + y_scale * x2.w;
      nuclear_charge = calc_inv_fourier( rx, nr, sd.reflex, ncgv );
      fprintf(
        fo1, "%lf\t%lf\t%d\t%d\t%lf\t%lf\n",
        x_scale, y_scale, ix, iy, nuclear_charge.re, nuclear_charge.im
        );
    }
  }
  printf("100%%.\n");
  fclose(fo1);
  
  free_vector(ncgv);
  free_sd(&sd);
  free_csd(&csd);
}


/* --------------- */

void input_density_3d_data(SIM_DATA *sd, CRYSTAL_STRCT_DATA *csd, FILE *fi)
{
  char  buf[BUFSIZE];
  CELL_PAR	bv;
  
  bv = bravai(csd->lp);
  
  chk_data("E_DENSITY_CALCULATION_CONDITION{", fi);

  chk_data("INCIDENT_BEAM{", fi);
  sd->incidence = fget_uvw(fi);
  sd->surf_norm = fget_uvw(fi);
  sd->zolzbase1 = fget_r_hkl(fi);
  sd->H_tilt = fget_d(fi);
  sd->G_tilt = fget_d(fi);
  sd->accel_kvol = fget_d(fi);
  chk_data("}INCIDENT_BEAM", fi);
  
  /* Beam search */
  chk_data("BEAM_SEARCH_3D{", fi);
  fget_sd_beams_3d( sd, csd, fi );
  reflex_dump_csd(
    0, sd->num_exact, sd->num_bethe, sd->reflex,
    csd, sd->beam_search, sd->incidence, sd->zolzbase1, sd->accel_kvol, bv
    );
  sd->beam_search.disksize = 1.0;  /* caution */
  mb_fprintf(stdout, "$$disksize = %f\n", sd->beam_search.disksize);
  chk_data("}BEAM_SEARCH_3D", fi);

  zolz_ort_base_two_uvw(
    sd->incidence, sd->zolzbase1, bv, &(sd->x1), &(sd->x2)
    );

  mb_fprintf(stdout, "real-space base-vector ( length = 1/|zolzbase1| )\n");
  mb_fprintf(
    stdout, "x1 = %lf %lf %lf, x2 = %lf %lf %lf\n",
    sd->x1.u, sd->x1.v, sd->x1.w, sd->x2.u, sd->x2.v, sd->x2.w
    );

  chk_data("CALC_RXY_RANGE_3D{", fi);
  comment_skip_fgets(sizeof buf, buf, fi);  /* UNITCELL_XYZ option for VEND */
  if( strncmp( "UNITCELL_XYZ", rm_cr(buf), strlen("UNITCELL_XYZ") )==0 ){
    sd->unitcell_xyz_flag = 1; /* flag */
    fget_3i( &(sd->num_points_x), &(sd->num_points_y), &(sd->num_points_z), fi);
    sd->x_begin = 0.0;  sd->x_end = 1.0;
    sd->y_begin = 0.0;  sd->y_end = 1.0;
    sd->z_begin = 0.0;  sd->z_end = 1.0;
  }
  else{
    if (
      sscanf(
        buf, "%d %d %d", &(sd->num_points_x),
        &(sd->num_points_y), &(sd->num_points_z), fi
        ) != 3
      ){
      mb_fprintf(stdout, "$$Data format error in CALC_RXY_RANGE_3D. Abort.\n");
      mb_fprintf(stdout, "$$buf=%s", buf);
      exit(EXIT_FAILURE);
    }
    sd->unitcell_xyz_flag = 0; /* flag */
    fget_2d(&(sd->x_begin), &(sd->x_end), fi);
    fget_2d(&(sd->y_begin), &(sd->y_end), fi);
    fget_2d(&(sd->z_begin), &(sd->z_end), fi);
  }
  chk_data("}CALC_RXY_RANGE_3D", fi);

  chk_data("}E_DENSITY_CALCULATION_CONDITION",fi);
  fclose(fi);
  
  /* dummy for free_sd */
  sd->num_g = 0;
  sd->num_thickness = 0;
  sd->calc_type = -1;
  /*
  sd->intensity = new_tensor(double, 1, 1, 1);
  chk_tensor(sd->intensity);
    */
}


void mb_electron_density_3d( FILE *fi, FILE *fo_log, char *infile )
{
  /*
   Electron density and potential distribution
   */
  CRYSTAL_STRCT_DATA csd;
  SIM_DATA sd;
  CELL_PAR bv;
  int     nr, nmx, nmy, nmz, prog, ix, iy, iz;
  double  x_scale, y_scale, z_scale;
  double  ***eld, ***pot;
  char	ed_file[BUFSIZE], pt_file[BUFSIZE];
  complex	*vgv, *fxgv, potential, rho;
  R_UVW	x1, x2, x3, rx;
  FILE	*fo_ed, *fo_pt;
  
  input_csd_cstr(&csd, fi, infile);    /* crystal structure data */
  input_density_3d_data(&sd, &csd, fi);
  bv = bravai(csd.lp);
  
  nr = sd.num_exact + sd.num_bethe;
  vgv = new_vector(complex, nr);	chk_vector(vgv);
  fxgv = new_vector(complex, nr);	chk_vector(fxgv);
  if (sd.num_points_x > 0 && sd.num_points_y > 0 && sd.num_points_z > 0){
    eld = new_tensor(
      double, sd.num_points_x, sd.num_points_y, sd.num_points_z
      );
    chk_tensor(eld);
    pot = new_tensor(
      double, sd.num_points_x, sd.num_points_y, sd.num_points_z
      );
    chk_tensor(pot);
  }
  else{
    mb_fprintf(
      stdout, "$$ number of pixels should be greater than 0. Abort.\n"
      );
    mb_fprintf(
      stdout, "$$ x,y,z = %d %d %d\n",
      sd.num_points_x, sd.num_points_y, sd.num_points_z
      );
    exit(EXIT_FAILURE);
  }
  
  mb_fprintf(stdout, "Potential.\n");
  vgv_fxgv_csd(
    sd.num_exact, sd.num_bethe, sd.reflex, &csd, sd.accel_kvol, bv, vgv, fxgv
    );
  
  nmx = sd.num_points_x - 1;
  nmy = sd.num_points_y - 1;
  nmz = sd.num_points_z - 1;

  if (sd.unitcell_xyz_flag == 1){  /* for VEND */
    x1.u = 1.0;  x1.v = 0.0;  x1.w = 0.0;
    x2.u = 0.0;  x2.v = 1.0;  x2.w = 0.0;
    x3.u = 0.0;  x3.v = 0.0;  x3.w = 1.0;
  }
  else{
    x1 = sd.x1;
    x2 = sd.x2;
    x3.u = (double) sd.incidence.u;
    x3.v = (double) sd.incidence.v;
    x3.w = (double) sd.incidence.w;
  }

  mb_fprintf(stdout, "Electron density and potential distributions.\n");
  for(ix=0; ix<sd.num_points_x; ix++){
    if(ix==0) {exec_time_forecast(-1);}
    progress_indicator(ix, sd.num_points_x);
    if(nmx == 0){
      x_scale = sd.x_begin;
    }
    else{
      x_scale = sd.x_begin
        + (sd.x_end - sd.x_begin) * (double) ix / (double) nmx;
    }
    for(iy=0; iy<sd.num_points_y; iy++){
      if(nmy == 0){
        y_scale = sd.y_begin;
      }
      else{
        y_scale = sd.y_begin
          + (sd.y_end - sd.y_begin) * (double) iy / (double) nmy;
      }
      for(iz=0; iz<sd.num_points_z; iz++){
        if(nmz == 0){
          z_scale = sd.z_begin;
        }
        else{
          z_scale = sd.z_begin
            + (sd.z_end - sd.z_begin) * (double) iz / (double) nmz;
        }
        rx.u = x_scale * x1.u + y_scale * x2.u + z_scale * x3.u;
        rx.v = x_scale * x1.v + y_scale * x2.v + z_scale * x3.v;
        rx.w = x_scale * x1.w + y_scale * x2.w + z_scale * x3.w;
        potential = calc_inv_fourier( rx, nr, sd.reflex, vgv );
        rho       = calc_inv_fourier( rx, nr, sd.reflex, fxgv );
        eld[ix][iy][iz] = rho.re / bv.vc;
        pot[ix][iy][iz] = potential.re;
      }
    }
    if(ix==0){exec_time_forecast(sd.num_points_x);}
  }
  exec_time_forecast(0);

  /* VEND-format Output */
  sprintf(ed_file, "%s_ed.3ed", infile);
  printf("Electron density (VEND-format): ==> %s\n", ed_file );
  fo_ed = fopen_chk_wmode(ed_file);
  sprintf(pt_file, "%s_pt.3ed", infile);
  printf("Potential (VEND-format):        ==> %s\n", pt_file );
  fo_pt = fopen_chk_wmode(pt_file);

  fprintf(fo_ed, "#Electron density [e/A^3]\n");
  fprintf(fo_pt, "#Electrostatic potential [V]\n");
  /* UNITCELL_XYZ option */
  if (sd.unitcell_xyz_flag == 1){
    fprintf(
      fo_ed, "%lg %lg %lg %lg %lg %lg\n",
      csd.lp.a, csd.lp.b, csd.lp.c, csd.lp.alpha, csd.lp.beta, csd.lp.gamma
      );
    fprintf(
      fo_pt, "%lg %lg %lg %lg %lg %lg\n",
      csd.lp.a, csd.lp.b, csd.lp.c, csd.lp.alpha, csd.lp.beta, csd.lp.gamma
      );
  }
  else{
    fprintf(
      fo_ed, "%lg %lg %lg %lg %lg %lg\n",
      abs_r_uvw(x1, bv) * (sd.x_end - sd.x_begin), 
      abs_r_uvw(x2, bv) * (sd.y_end - sd.y_begin), 
      abs_r_uvw(x3, bv) * (sd.z_end - sd.z_begin), 
      90.0, 90.0, 90.0
      );
    fprintf(
      fo_pt, "%lg %lg %lg %lg %lg %lg\n",
      abs_r_uvw(x1, bv) * (sd.x_end - sd.x_begin), 
      abs_r_uvw(x2, bv) * (sd.y_end - sd.y_begin), 
      abs_r_uvw(x3, bv) * (sd.z_end - sd.z_begin), 
      90.0, 90.0, 90.0
      );
  }
  fprintf(
    fo_ed, "%d  %d  %d\n", sd.num_points_x, sd.num_points_y, sd.num_points_z
    );
  fprintf(
    fo_pt, "%d  %d  %d\n", sd.num_points_x, sd.num_points_y, sd.num_points_z
    );
  for(ix=0; ix<sd.num_points_x; ix++){
    for(iy=0; iy<sd.num_points_y; iy++){
      for(iz=0; iz<sd.num_points_z; iz++){
        fprintf( fo_ed, "%lg  ", eld[ix][iy][iz]);
        fprintf( fo_pt, "%lg  ", pot[ix][iy][iz]);
      }
      fprintf( fo_ed, "\n" );
      fprintf( fo_pt, "\n" );
    }
  }
  fclose(fo_ed);  fclose(fo_pt);

  /* T3D-format Output */
  if(sd.unitcell_xyz_flag == 0){  /* if UNITCELL_XYZ is not defined */
    sprintf(ed_file, "%s_ed.t3d", infile);
    printf("Electron density (T3D-format):  ==> %s\n", ed_file );
    fo_ed = fopen_chk_wmode(ed_file);
    sprintf(pt_file, "%s_pt.t3d", infile);
    printf("Potential (T3D-format):         ==> %s\n", pt_file );
    fo_pt = fopen_chk_wmode(pt_file);
    for(iz=0; iz<sd.num_points_z; iz++){
      for(iy=0; iy<sd.num_points_y; iy++){
        for(ix=0; ix<sd.num_points_x; ix++){
          fprintf( fo_ed, "%lg  ", eld[ix][iy][iz]);
          fprintf( fo_pt, "%lg  ", pot[ix][iy][iz]);
        }
        fprintf( fo_ed, "\n" );
        fprintf( fo_pt, "\n" );
      }
    }
    fclose(fo_ed);  fclose(fo_pt);
  }
  
  free_vector(vgv);	 free_vector(fxgv);
  free_tensor(eld);  free_tensor(pot);
  free_sd(&sd);
  free_csd(&csd);
}
