/*
 * fitfunc.c
 * Lots of small functions for fitting process
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "mbconfig.h"
#include "fitting.h"
#include "mb_input.h"
#include "data_out.h"
#include "utils.h"
#include "svd.h"
#include "mb_mpi.h"

#define BUFLEN 256

extern DEBUG_INFO dbg;
extern EPS_CTRL	epsilon;


double cal_weight(
  double obs_sig, double obs_intensity, double cal_intensity, double g,
  double art_weight
  );
double cal_weight_factor(
  double obs_sig, double obs_intensity, double cal_intensity,
  double artificial_weight
  );
FIT_CTRL search_csd_fed(
  int ntp, CRYSTAL_STRCT_DATA *csd, int num_exp_dataset, FIT_EXP_DATASET *fed
  );


void calc_avrpix_wt(int num_avrpix, double **pix_wt, double gaussian_par)
{
  int 	ix, iy, cx, cy;
  double	sum, distance, t, t_par;
  FILE	*fp;

  if(num_avrpix == 0 || num_avrpix == 1){ return; }	/* No convolution */

  if(num_avrpix > 1 && num_avrpix % 2 == 0){
    mb_fprintf(
      stdout, "$$ The number of averaging points must be odd number. Abort.\n"
      );
    mb_fprintf(stdout, "$$ (calc_avrpix_wt)\n");
    exit(EXIT_FAILURE);
  }

  if (gaussian_par == 0.0){
    mb_fprintf(stdout, "$$ Gaussian_par must not be 0.0. Abort.\n");
    exit(EXIT_FAILURE);
  }
  else if(gaussian_par < 0.0){	/* ***** test ***** */
    
    cx = cy = (num_avrpix - 1) / 2; /* (cx, cy) center: ix,iy:0~num_avrpix-1 */
    t_par = - gaussian_par;
    
    /* Weight for each pixel */
    sum = 0.0;
    for(ix=0; ix<num_avrpix; ix++){
      for(iy=0; iy<num_avrpix; iy++){
        distance = sqrt( (double) ( sq(ix - cx) + sq(iy - cy) ) );
        if ( distance < epsilon.eps ){
          t = 1.0;
        }
        else{
          t = sin(PI * distance / t_par) / (PI * distance / t_par);
        }
        pix_wt[ix][iy] = sq(t);
        sum += pix_wt[ix][iy];
      }
    }
    fp = mb_fopen_chk_wmode("convlt.dat");

    mb_fprintf(stdout, "$$Weights for convolution --> convlt.dat\n");

    /* Normalize */
    for(iy=0; iy<num_avrpix; iy++){
      for(ix=0; ix<num_avrpix; ix++){
        pix_wt[ix][iy] = pix_wt[ix][iy] / sum;
        mb_fprintf(fp, "%lf\t", pix_wt[ix][iy]);
      }
      mb_fprintf(fp, "\n");
    }
    mb_fclose(fp);
    return;
  }		/* ********** */
  
  
  /* Weight for Gaussian */
  cx = cy = (num_avrpix - 1) / 2;/* (cx, cy) center: ix,iy:0~num_avrpix-1 */
  
  /* Weight for each pixel */
  sum = 0.0;
  for(ix=0; ix<num_avrpix; ix++){
    for(iy=0; iy<num_avrpix; iy++){
      distance = sqrt((double) ( sq(ix - cx) + sq(iy - cy) ));	/* <<<<< */
      pix_wt[ix][iy] = exp( - sq(distance / gaussian_par) );
      sum += pix_wt[ix][iy];
    }
  }

  fp = mb_fopen_chk_wmode("convlt.dat");
  mb_fprintf(stdout, "$$Weights for convolution --> convlt.dat\n");

  /* Normalize */
  for(iy=0; iy<num_avrpix; iy++){
    for(ix=0; ix<num_avrpix; ix++){
      pix_wt[ix][iy] = pix_wt[ix][iy] / sum;
      mb_fprintf(fp, "%lf\t", pix_wt[ix][iy]);
    }
    mb_fprintf(fp, "\n");
  }
  mb_fclose(fp);
}


/* ------------------------------------------------------------------------- */
/* ----- alloc and free of csd and fed ------------------------------------- */
/* ------------------------------------------------------------------------- */

void alloc_csd(
  int num_sym_op, int n_atom, int num_lofg, CRYSTAL_STRCT_DATA *csd_new
  )
{
  csd_new->sym_op = new_vector(SYM_OP, num_sym_op);
  chk_vector(csd_new->sym_op);
  csd_new->atm	= new_vector(ATOM, n_atom);
  chk_vector(csd_new->atm);
  csd_new->atm_f	= new_vector(ATOM_FIT, n_atom);
  chk_vector(csd_new->atm_f);
  csd_new->xdyl		= new_vector(XDT, n_atom);
  chk_vector(csd_new->xdyl);
  csd_new->ftbl		= new_vector(ATOMF_BK, n_atom);
  chk_vector(csd_new->ftbl);
  
  if( num_lofg > 0){
    csd_new->g_lofg	= new_vector(HKL, num_lofg);
    chk_vector(csd_new->g_lofg);
    csd_new->fg_lofg	= new_vector(LO_FG, num_lofg);
    chk_vector(csd_new->fg_lofg);
    csd_new->fg_lofg_f	= new_vector(LO_FG_FIT, num_lofg);
    chk_vector(csd_new->fg_lofg);
  }
}


void alloc_csd_sub_cstr(
  int n_atom, int num_lofg, int lof_type, CRYSTAL_STRCT_DATA *csd_new
  )
{
  int i;
  
  if( num_lofg > 0 ){
    if (lof_type == 1){ 
      ;
    }
    else if (lof_type == 2){
      ;
    }
    else if (lof_type == 3 || lof_type == 4){
      for(i=0; i<num_lofg; i++){
        csd_new->fg_lofg[i].fe		= new_vector(double, n_atom);
        chk_vector(csd_new->fg_lofg[i].fe);
        csd_new->fg_lofg[i].fi		= new_vector(double, n_atom);
        chk_vector(csd_new->fg_lofg[i].fi);
        csd_new->fg_lofg_f[i].fe	= new_vector(FIT_CTRL, n_atom);
        chk_vector(csd_new->fg_lofg_f[i].fe);
        csd_new->fg_lofg_f[i].fi	= new_vector(FIT_CTRL, n_atom);
        chk_vector(csd_new->fg_lofg_f[i].fi);
      }
    }
    else if (lof_type == 5){
      mb_fprintf(stdout, "$$ Sorry, lof_type = 5 is not available. Abort.\n");
      exit(EXIT_FAILURE);
    }
  }
}


void cpy_csd(CRYSTAL_STRCT_DATA *csd, CRYSTAL_STRCT_DATA *csd_new)
{
  int i, ia, ns, na, nl, type;

  csd_new->space_group = csd->space_group;
  ns = csd_new->num_sym_op = csd->num_sym_op;
  for(i=0; i<ns; i++){
    csd_new->sym_op[i] = csd->sym_op[i];
  }
  strcpy(csd_new->lattice_type, csd->lattice_type);
  csd_new->lp = csd->lp;
  csd_new->lp_f = csd->lp_f;

  na = csd_new->n_atom = csd->n_atom;
  for(i=0; i<na; i++){
    csd_new->atm[i]		= csd->atm[i];
    csd_new->atm_f[i]	= csd->atm_f[i];
    csd_new->xdyl[i]		= csd->xdyl[i];
    csd_new->ftbl[i]		= csd->ftbl[i];
  }

  nl = csd_new->num_lofg = csd->num_lofg;
  type = csd_new->lof_type = csd->lof_type;

  for(i=0; i<nl; i++){
    csd_new->g_lofg[i] = csd->g_lofg[i];
    
    if (type == 1 || type == 2){/* Fg */
      csd_new->fg_lofg[i] = csd->fg_lofg[i];
      csd_new->fg_lofg_f[i] = csd->fg_lofg_f[i];
    }
    else if(type == 3 || type == 4){/* f */
      for(ia=0; ia<na; ia++){
        csd_new->fg_lofg[i].fe[ia] = csd->fg_lofg[i].fe[ia];
        csd_new->fg_lofg_f[i].fe[ia] = csd->fg_lofg_f[i].fe[ia];
      }
    }
    else{
      mb_fprintf(stdout, "$$ Invalid lof_type %d. Abort.\n", type);	exit(EXIT_FAILURE);
    }
  }
}


void free_csd(CRYSTAL_STRCT_DATA *csd)
{
  int i, natm, nlg;
  
  free_vector(csd->sym_op);
  natm = csd->n_atom;
  free_vector(csd->atm);			free_vector(csd->atm_f);
  free_vector(csd->xdyl);			free_vector(csd->ftbl);
  
  if(csd->num_lofg > 0){
    nlg = csd->num_lofg;
    free_vector(csd->g_lofg);
    if(csd->lof_type == 3 || csd->lof_type == 4){
      for(i=0; i<nlg; i++){
        free_vector(csd->fg_lofg[i].fe);
        free_vector(csd->fg_lofg[i].fi);
        free_vector(csd->fg_lofg_f[i].fe);
        free_vector(csd->fg_lofg_f[i].fi);
      }
    }
    free_vector(csd->fg_lofg);		free_vector(csd->fg_lofg_f);
  }
}


void alloc_ed(
  BEAM_SEARCH_PAR beam_search, int nr, int ng, int nt, int nk[],
  FIT_EXP_DATASET *ed
  )
{
  /*	int it, ig, nd; */
  
  if(	beam_search.flag_bs == 0 ||	beam_search.flag_bs == 10 ){		/* <<<<< */
    ed->reflex = new_vector(HKL, nr);	chk_vector(ed->reflex);
  }
  else if(
    beam_search.flag_bs == 1 || beam_search.flag_bs == 11 || 
    beam_search.flag_bs == 2 || beam_search.flag_bs == 12 || 
    beam_search.flag_bs == 3 || beam_search.flag_bs == 13
    ){
    ;
  }
  
  ed->scale_factor = new_vector(double, nt);		chk_vector(ed->scale_factor);
  ed->scale_factor_f = new_vector(FIT_CTRL, nt);	chk_vector(ed->scale_factor_f);
  ed->thickness = new_vector(double, nt);			chk_vector(ed->thickness);
  ed->thickness_f = new_vector(FIT_CTRL, nt);		chk_vector(ed->thickness_f);
  ed->thickwidth = new_vector(double, nt);		chk_vector(ed->thickwidth);
  ed->thickwidth_f = new_vector(FIT_CTRL, nt);	chk_vector(ed->thickwidth_f);
  ed->num_avrpix = new_vector(int, nt);			chk_vector(ed->num_avrpix);
  ed->gaussian_par = new_vector(double, nt);		chk_vector(ed->gaussian_par);
  ed->avrpix_wt = new_tensor_element(nt);			chk_tensor(ed->avrpix_wt);
  
  /*	ed->obs_rg = new_vector(HKL, ng);				chk_vector(ed->obs_rg); */
  ed->g = new_vector(double, ng);					chk_vector(ed->g);
  ed->art_weight = new_vector(double, ng);		chk_vector(ed->art_weight);
  
  ed->H_shift = new_matrix(double, ng, nt);		chk_matrix(ed->H_shift);
  ed->G_shift = new_matrix(double, ng, nt);		chk_matrix(ed->G_shift);
  ed->H_shift_f = new_matrix(FIT_CTRL, ng, nt);	chk_matrix(ed->H_shift_f);
  ed->G_shift_f = new_matrix(FIT_CTRL, ng, nt);	chk_matrix(ed->G_shift_f);
  /*	ed->num_data_points = new_vector(int, ng);		chk_vector(ed->num_data_points); */
}


void alloc_ei(
  int nr, int ng, int nt, int nk[], FIT_EXP_INTSET *ei
  )
{
  int it, ig, nd;
  
  /* Technique */
  ei->obs_H			= new_tensor_element(ng);	chk_tensor(ei->obs_H);
  ei->obs_G			= new_tensor_element(ng);	chk_tensor(ei->obs_G);
  ei->obs_intensity	= new_tensor_element(ng);	chk_tensor(ei->obs_intensity);
  ei->cal_intensity	= new_tensor_element(ng);	chk_tensor(ei->cal_intensity);
  ei->obs_sig			= new_tensor_element(ng);	chk_tensor(ei->obs_sig);
  
  for(ig=0; ig<ng; ig++){
    ei->obs_H[ig]			= new_matrix_element(nt);	chk_matrix(ei->obs_H[ig]);
    ei->obs_G[ig]			= new_matrix_element(nt);	chk_matrix(ei->obs_G[ig]);
    ei->obs_intensity[ig]	= new_matrix_element(nt);	chk_matrix(ei->obs_intensity[ig]);
    ei->cal_intensity[ig]	= new_matrix_element(nt);	chk_matrix(ei->cal_intensity[ig]);
    ei->obs_sig[ig]			= new_matrix_element(nt);	chk_matrix(ei->obs_sig[ig]);
    nd = nk[ig];
    for(it=0; it<nt; it++){
      ei->obs_H[ig][it]		  = new_vector_element(double, nd);	chk_vector(ei->obs_H[ig][it]);
      ei->obs_G[ig][it]		  = new_vector_element(double, nd);	chk_vector(ei->obs_G[ig][it]);
      ei->obs_intensity[ig][it] = new_vector_element(double, nd);	chk_vector(ei->obs_intensity[ig][it]);
      ei->cal_intensity[ig][it] = new_vector_element(double, nd);	chk_vector(ei->cal_intensity[ig][it]);
      ei->obs_sig[ig][it]		  = new_vector_element(double, nd);	chk_vector(ei->obs_sig[ig][it]);
    }
  }
}


void cpy_fed(int num_exp_dataset, FIT_EXP_DATASET *fed, FIT_EXP_DATASET *fed_new)
{
  int i, it, nt, ig, ng, ir, nr;
  
  for(i=0; i<num_exp_dataset; i++){
    fed_new[i].incidence = fed[i].incidence;
    fed_new[i].surf_norm = fed[i].surf_norm;
    fed_new[i].zolzbase1 = fed[i].zolzbase1;
    fed_new[i].H_tilt = fed[i].H_tilt;
    fed_new[i].H_tilt_f = fed[i].H_tilt_f;
    fed_new[i].G_tilt = fed[i].G_tilt;
    fed_new[i].G_tilt_f = fed[i].G_tilt_f;
    fed_new[i].calc_type = fed[i].calc_type;
    fed_new[i].accel_kvol = fed[i].accel_kvol;
    fed_new[i].accel_kvol_f = fed[i].accel_kvol_f;
    fed_new[i].beam_search.flag_bs = fed[i].beam_search.flag_bs;
    fed_new[i].beam_search = fed[i].beam_search;
    
    if(
      fed[i].beam_search.flag_bs == 0 || fed[i].beam_search.flag_bs == 10
      ){		/* <<<<< */
      fed_new[i].num_exact = fed[i].num_exact;
      fed_new[i].num_bethe = fed[i].num_bethe;
      nr = fed[i].num_exact + fed[i].num_bethe;
      for(ir=0; ir<nr; ir++){
        fed_new[i].reflex[ir] = fed[i].reflex[ir];
      }
    }
    else if(
      fed[i].beam_search.flag_bs == 1 || fed[i].beam_search.flag_bs == 11 || 
      fed[i].beam_search.flag_bs == 2 || fed[i].beam_search.flag_bs == 12 || 
      fed[i].beam_search.flag_bs == 3 || fed[i].beam_search.flag_bs == 13
      ){
      ;
    }
    
    nt = fed_new[i].num_thickness = fed[i].num_thickness;
    ng = fed_new[i].num_rg = fed[i].num_rg;

    fed_new[i].calc_range_flag = fed[i].calc_range_flag;
    fed_new[i].calc_numh = fed[i].calc_numh;
    fed_new[i].calc_hbgn = fed[i].calc_hbgn;
    fed_new[i].calc_hend = fed[i].calc_hend;
    fed_new[i].calc_gbgn = fed[i].calc_gbgn;
    fed_new[i].calc_gend = fed[i].calc_gend;
    strcpy(fed_new[i].intensity_file, fed[i].intensity_file);/* <<<<< */

    for(it=0; it<nt; it++){
      fed_new[i].scale_factor[it]   = fed[i].scale_factor[it];
      fed_new[i].scale_factor_f[it] = fed[i].scale_factor_f[it];
      fed_new[i].thickness[it] = fed[i].thickness[it];
      fed_new[i].thickness_f[it] = fed[i].thickness_f[it];
      fed_new[i].thickwidth[it] = fed[i].thickwidth[it];
      fed_new[i].thickwidth_f[it] = fed[i].thickwidth_f[it];
      fed_new[i].num_avrpix[it] = fed[i].num_avrpix[it];
      fed_new[i].gaussian_par[it] = fed[i].gaussian_par[it];
      if(fed_new[i].num_avrpix[it] > 0){
        fed_new[i].avrpix_wt[it]
          = new_matrix(
            double, fed_new[i].num_avrpix[it], fed_new[i].num_avrpix[it]
            );
        chk_matrix(fed_new[i].avrpix_wt[it]);
        calc_avrpix_wt(
          fed_new[i].num_avrpix[it], fed_new[i].avrpix_wt[it],
          fed_new[i].gaussian_par[it]
          );
      }

      for(ig=0; ig<ng; ig++){
        fed_new[i].g[ig]			= fed[i].g[ig];
        fed_new[i].art_weight[ig]   = fed[i].art_weight[ig];
        fed_new[i].H_shift[ig][it] = fed[i].H_shift[ig][it];
        fed_new[i].H_shift_f[ig][it] = fed[i].H_shift_f[ig][it];
        fed_new[i].G_shift[ig][it] = fed[i].G_shift[ig][it];
        fed_new[i].G_shift_f[ig][it] = fed[i].G_shift_f[ig][it];
      }
    }
  }
}


void free_ed(FIT_EXP_DATASET *ed)
{
  int it, nt;
  
  if( /* manual beam selection */
      ed->beam_search.flag_bs == 0 ||
      ed->beam_search.flag_bs == 10
      ){
    free_vector(ed->reflex);
    /* Caution:
        In case of auto-beam-selection, the area of .reflex
        should be released in the same level as auto_ed_beams(). */
  }
  free_vector(ed->scale_factor);
  free_vector(ed->scale_factor_f);
  free_vector(ed->thickness);
  free_vector(ed->thickness_f);
  free_vector(ed->thickwidth);
  free_vector(ed->thickwidth_f);
  nt = ed->num_thickness;
  for(it=0; it<nt; it++){
    if(ed->num_avrpix[it] > 0){
      free_tensor(ed->avrpix_wt);
    }
    else{
      free_vector(ed->avrpix_wt);
    }
  }
  free_vector(ed->num_avrpix);
  free_vector(ed->gaussian_par);
  /*  free_vector(ed->obs_rg); */
  free_vector(ed->g);
  free_vector(ed->art_weight);
  free_matrix(ed->H_shift);
  free_matrix(ed->G_shift);
  free_matrix(ed->H_shift_f);
  free_matrix(ed->G_shift_f);
}


void free_ei(FIT_EXP_INTSET *ei)
{
  int ig, it, ng, nt;
  
  ng = ei->num_rg;
  nt = ei->num_thickness;
  for(ig=0; ig<ng; ig++){
    for(it=0; it<nt; it++){
      free_vector(ei->obs_H[ig][it]);
      free_vector(ei->obs_G[ig][it]);
      free_vector(ei->obs_intensity[ig][it]);
      free_vector(ei->cal_intensity[ig][it]);
      free_vector(ei->obs_sig[ig][it]);
    }
    free_vector((void *) ei->obs_H[ig]);
    free_vector((void *) ei->obs_G[ig]);
    free_vector((void *) ei->obs_intensity[ig]);
    free_vector((void *) ei->cal_intensity[ig]);
    free_vector((void *) ei->obs_sig[ig]);
  }
  free_vector((void *) ei->obs_H);
  free_vector((void *) ei->obs_G);
  free_vector((void *) ei->obs_intensity);
  free_vector((void *) ei->obs_sig);
  free_vector((void *) ei->cal_intensity);
}


void free_fed_vector(int num_exp_dataset, FIT_EXP_DATASET *fed)
{
  int i, it;
  
  for(i=0; i<num_exp_dataset; i++){
    free_ed(&(fed[i]));
    /*
    if(
        fed[i].beam_search.flag_bs == 0 ||
        fed[i].beam_search.flag_bs == 10
      ){
      free_vector(fed[i].reflex);
    }
    free_vector(fed[i].scale_factor);
    free_vector(fed[i].scale_factor_f);
    free_vector(fed[i].thickness);
    free_vector(fed[i].thickness_f);
    free_vector(fed[i].thickwidth);
    free_vector(fed[i].thickwidth_f);
    for(it=0; it<fed[i].num_thickness; it++){
      if(fed[i].num_avrpix[it] > 0){
        free_tensor(fed[i].avrpix_wt);
      }
      else{
        free_vector(fed[i].avrpix_wt);
      }
    }
    free_vector(fed[i].num_avrpix);
    free_vector(fed[i].gaussian_par);
    free_vector(fed[i].g);
    free_vector(fed[i].art_weight);
    free_matrix(fed[i].H_shift);	free_matrix(fed[i].G_shift);
    free_matrix(fed[i].H_shift_f);	free_matrix(fed[i].G_shift_f);
  */
  }
  free_vector(fed);
}


void free_fei_vector(int num_exp_dataset, FIT_EXP_INTSET *fei)
{
  int i, ig, it, ng, nt;
  
  for(i=0; i<num_exp_dataset; i++){
    free_ei(&(fei[i]));
/*    ng = fei[i].num_rg;
    nt = fei[i].num_thickness;
    for(ig=0; ig<ng; ig++){
      for(it=0; it<nt; it++){
        free_vector(fei[i].obs_H[ig][it]);
        free_vector(fei[i].obs_G[ig][it]);
        free_vector(fei[i].obs_intensity[ig][it]);
        free_vector(fei[i].cal_intensity[ig][it]);
        free_vector(fei[i].obs_sig[ig][it]);
      }
      free_vector((void *) fei[i].obs_H[ig]);
      free_vector((void *) fei[i].obs_G[ig]);
      free_vector((void *) fei[i].obs_intensity[ig]);
      free_vector((void *) fei[i].cal_intensity[ig]);
      free_vector((void *) fei[i].obs_sig[ig]);
    }
    free_vector((void *) fei[i].obs_H);
    free_vector((void *) fei[i].obs_G);
    free_vector((void *) fei[i].obs_intensity);
    free_vector((void *) fei[i].obs_sig);
    free_vector((void *) fei[i].cal_intensity); */
  }
  free_vector(fei);
}


/* ------------------------------------------------------------------------- */
/* ----- fitpar ------------------------------------------------------------ */
/* ------------------------------------------------------------------------- */


void chk_fitpar(int idx, int *nth_fitpar, int *ct_fitpar)
{
  if(idx == 1 || idx == 2){/* treated as a fitting parameter */
    *nth_fitpar = *ct_fitpar;
    (*ct_fitpar)++;
  }
  else{/* treated as a fixed parameter */
    *nth_fitpar = -1;
  }
}


void count_num_fitpar(
  CRYSTAL_STRCT_DATA *csd, int num_exp_dataset, FIT_EXP_DATASET *fed, 
  int *num_fitpar, int *num_par, int *num_perturb
  )
{
  /*
   * Count the number of parameters
   *  idx = 2: free
   *        1: restraint
   *        0: fix
   *       -1: constraint
   */
  int i, it, ig, nt, ng, ct_fitpar, ct_par, ct_perturb, ia;

  /* Initialize */
  ct_fitpar = ct_par = ct_perturb = 0;

  /* CRYSTAL_STRCT_DATA *csd  */

  /* lattice parameter */
  ct_par += 6;
  chk_fitpar( csd->lp_f.a.flg, &(csd->lp_f.a.ntp), &ct_fitpar );
  chk_fitpar( csd->lp_f.b.flg, &(csd->lp_f.b.ntp), &ct_fitpar );
  chk_fitpar( csd->lp_f.c.flg, &(csd->lp_f.c.ntp), &ct_fitpar );
  chk_fitpar( csd->lp_f.alpha.flg, &(csd->lp_f.alpha.ntp), &ct_fitpar );
  chk_fitpar( csd->lp_f.beta.flg,  &(csd->lp_f.beta.ntp),  &ct_fitpar );
  chk_fitpar( csd->lp_f.gamma.flg, &(csd->lp_f.gamma.ntp), &ct_fitpar );

  /* ATOM */
  for(i=0; i<csd->n_atom; i++){
    ct_par += 3;
    chk_fitpar( csd->atm_f[i].x.flg, &(csd->atm_f[i].x.ntp), &ct_fitpar );
    chk_fitpar( csd->atm_f[i].y.flg, &(csd->atm_f[i].y.ntp), &ct_fitpar );
    chk_fitpar( csd->atm_f[i].z.flg, &(csd->atm_f[i].z.ntp), &ct_fitpar );

    if(csd->atm[i].bf == 0){/* Isotropic B */
      ct_par++;
      chk_fitpar( csd->atm_f[i].b.flg, &(csd->atm_f[i].b.ntp), &ct_fitpar );
    }
    else if(csd->atm[i].bf == 1){/* Anisotropic B */
      ct_par += 6;
      chk_fitpar(
        csd->atm_f[i].b11.flg, &(csd->atm_f[i].b11.ntp), &ct_fitpar
        );
      chk_fitpar(
        csd->atm_f[i].b22.flg, &(csd->atm_f[i].b22.ntp), &ct_fitpar
        );
      chk_fitpar(
        csd->atm_f[i].b33.flg, &(csd->atm_f[i].b33.ntp), &ct_fitpar
        );
      chk_fitpar(
        csd->atm_f[i].b12.flg, &(csd->atm_f[i].b12.ntp), &ct_fitpar
        );
      chk_fitpar(
        csd->atm_f[i].b23.flg, &(csd->atm_f[i].b23.ntp), &ct_fitpar
        );
      chk_fitpar(
        csd->atm_f[i].b13.flg, &(csd->atm_f[i].b13.ntp), &ct_fitpar
        );
    }
    
    ct_par++;
    chk_fitpar( csd->atm_f[i].occ.flg, &(csd->atm_f[i].occ.ntp),   &ct_fitpar );
    
    if(csd->atm[i].af == 0){/* X-ray + Mott and Ichimiya */
      ct_par++;
      chk_fitpar(
        csd->atm_f[i].fabs.flg, &(csd->atm_f[i].fabs.ntp), &ct_fitpar
        );
    }
    else if (csd->atm[i].af == 1){/* BK */
      ;
    }
    else if (csd->atm[i].af == 2){/* BK + db */
      ct_par++;
      chk_fitpar(
        csd->atm_f[i].db.flg, &(csd->atm_f[i].db.ntp), &ct_fitpar
        );
    }
    else{
      mb_fprintf(stdout, "$$ count_num_fitpar: Invalid af. Abort.\n");
      exit(EXIT_FAILURE);
    }
  }
  
  for(i=0; i<csd->num_lofg; i++){
    if(csd->lof_type == 1){	
      ct_par += 2;
      chk_fitpar(
        csd->fg_lofg_f[i].re.flg, &(csd->fg_lofg_f[i].re.ntp), &ct_fitpar
        );
      chk_fitpar(
        csd->fg_lofg_f[i].im.flg, &(csd->fg_lofg_f[i].im.ntp), &ct_fitpar
        );
    }
    else if(csd->lof_type == 2){
      ct_par += 2;
      chk_fitpar(
        csd->fg_lofg_f[i].ela_re.flg, &(csd->fg_lofg_f[i].ela_re.ntp),
        &ct_fitpar
        );
      chk_fitpar(
        csd->fg_lofg_f[i].ela_im.flg, &(csd->fg_lofg_f[i].ela_im.ntp),
        &ct_fitpar
        );
    }
    else if(csd->lof_type == 3 || csd->lof_type == 4){
      for(ia=0; ia<csd->n_atom; ia++){
        ct_par += 2;
        chk_fitpar(
          csd->fg_lofg_f[i].fe[ia].flg, &(csd->fg_lofg_f[i].fe[ia].ntp),
          &ct_fitpar
          );
        chk_fitpar(
          csd->fg_lofg_f[i].fi[ia].flg, &(csd->fg_lofg_f[i].fi[ia].ntp),
          &ct_fitpar
          );
      }
    }
    else{
      mb_fprintf(stdout, "$$ Invalid lof_type %d. Abort.\n", csd->lof_type);
      exit(EXIT_FAILURE);
    }
  }

  ct_perturb = ct_fitpar;

  /* FIT_EXP_DATASET *fed */
  
  for(i=0; i<num_exp_dataset; i++){
    ct_par++;
    chk_fitpar(
      fed[i].H_tilt_f.flg, &(fed[i].H_tilt_f.ntp), &ct_fitpar
      );
    chk_fitpar(
      fed[i].G_tilt_f.flg, &(fed[i].G_tilt_f.ntp), &ct_fitpar
      );
    chk_fitpar(
      fed[i].accel_kvol_f.flg, &(fed[i].accel_kvol_f.ntp), &ct_fitpar
      );
    /* if(fed[i].accel_kvol_n>=0){ ct_perturb++; }*/
    nt = fed[i].num_thickness;
    ng = fed[i].num_rg;
    for(it=0; it<nt; it++){
      ct_par += 2;
      chk_fitpar(
        fed[i].scale_factor_f[it].flg, &(fed[i].scale_factor_f[it].ntp),
        &ct_fitpar
        );
      chk_fitpar(
        fed[i].thickness_f[it].flg, &(fed[i].thickness_f[it].ntp),
        &ct_fitpar
        );
      chk_fitpar(
        fed[i].thickwidth_f[it].flg, &(fed[i].thickwidth_f[it].ntp),
        &ct_fitpar
        );
      for(ig=0; ig<ng; ig++){
        ct_par += 2;
        chk_fitpar(
          fed[i].H_shift_f[ig][it].flg, &(fed[i].H_shift_f[ig][it].ntp),
          &ct_fitpar
          );
        chk_fitpar(
          fed[i].G_shift_f[ig][it].flg, &(fed[i].G_shift_f[ig][it].ntp),
          &ct_fitpar
          );
      }
    }
  }

  /* Results */
  *num_fitpar = ct_fitpar;
  *num_par = ct_par;
  *num_perturb = ct_perturb;
}


void csd_fed_to_fitpar(
  CRYSTAL_STRCT_DATA *csd, int num_exp_dataset, FIT_EXP_DATASET fed[],
  int num_fitpar, double fitpar[]
  )
{
  int i, it, j, ig, nt, ng, ia;

  /* CRYSTAL_STRCT_DATA *csd */

  /* lattice parameter */
  j = csd->lp_f.a.ntp;
  if(j >= 0){ fitpar[j] = csd->lp.a; }
  j = csd->lp_f.b.ntp;
  if(j >= 0){ fitpar[j] = csd->lp.b; }
  j = csd->lp_f.c.ntp;
  if(j >= 0){ fitpar[j] = csd->lp.c; }
  j = csd->lp_f.alpha.ntp;
  if(j >= 0){ fitpar[j] = csd->lp.alpha; }
  j = csd->lp_f.beta.ntp;
  if(j >= 0){ fitpar[j] = csd->lp.beta;  }
  j = csd->lp_f.gamma.ntp;
  if(j >= 0){ fitpar[j] = csd->lp.gamma; }

  /* ATOM */
  for(i=0; i<csd->n_atom; i++){
    j = csd->atm_f[i].x.ntp;
    if(j >= 0){ fitpar[j] = csd->atm[i].x; }
    j = csd->atm_f[i].y.ntp;
    if(j >= 0){ fitpar[j] = csd->atm[i].y; }
    j = csd->atm_f[i].z.ntp;
    if(j >= 0){ fitpar[j] = csd->atm[i].z; }
    if(csd->atm[i].bf == 0){
      j = csd->atm_f[i].b.ntp;
      if(j >= 0){fitpar[j] = csd->atm[i].b;}
    }
    else if(csd->atm[i].bf == 1){ 
      j = csd->atm_f[i].b11.ntp;
      if(j >= 0){ fitpar[j] = csd->atm[i].b11; }
      j = csd->atm_f[i].b22.ntp;
      if(j >= 0){ fitpar[j] = csd->atm[i].b22; }
      j = csd->atm_f[i].b33.ntp;
      if(j >= 0){ fitpar[j] = csd->atm[i].b33; }
      j = csd->atm_f[i].b12.ntp;
      if(j >= 0){ fitpar[j] = csd->atm[i].b12; }
      j = csd->atm_f[i].b23.ntp;
      if(j >= 0){ fitpar[j] = csd->atm[i].b23; }
      j = csd->atm_f[i].b13.ntp;
      if(j >= 0){ fitpar[j] = csd->atm[i].b13; }
    }
    j = csd->atm_f[i].occ.ntp;
    if(j >= 0){ fitpar[j] = csd->atm[i].occ; }
  }

  /* Low order structure factors */
  for(i=0; i<csd->num_lofg; i++){
    if(csd->lof_type == 1){
      j = csd->fg_lofg_f[i].re.ntp;
      if(j >= 0){fitpar[j] = csd->fg_lofg[i].re;}
      j = csd->fg_lofg_f[i].im.ntp;
      if(j >= 0){fitpar[j] = csd->fg_lofg[i].im;}
    }
    else if(csd->lof_type == 2){
      j = csd->fg_lofg_f[i].ela_re.ntp;
      if(j >= 0){fitpar[j] = csd->fg_lofg[i].ela_re;}
      j = csd->fg_lofg_f[i].ela_im.ntp;
      if(j >= 0){fitpar[j] = csd->fg_lofg[i].ela_im;}
    }
    else if(csd->lof_type == 3 || csd->lof_type == 4){
      for(ia=0; ia<csd->n_atom; ia++){
        j = csd->fg_lofg_f[i].fe[ia].ntp;
        if(j >= 0){fitpar[j] = csd->fg_lofg[i].fe[ia];}
        j = csd->fg_lofg_f[i].fi[ia].ntp;
        if(j >= 0){fitpar[j] = csd->fg_lofg[i].fi[ia];}
      }
    }
    else{
      mb_fprintf(stdout, "$$ Invalid lof_type %d. Abort.\n", csd->lof_type);
      exit(EXIT_FAILURE);
    }
  }

  /* FIT_EXP_DATASET *fed */

  for(i=0; i<num_exp_dataset; i++){
    j = fed[i].H_tilt_f.ntp;
    if(j >= 0){fitpar[j] = fed[i].H_tilt;}
    j = fed[i].G_tilt_f.ntp;
    if(j >= 0){fitpar[j] = fed[i].G_tilt;}
    j = fed[i].accel_kvol_f.ntp;
    if(j >= 0){fitpar[j] = fed[i].accel_kvol;}
    nt = fed[i].num_thickness;
    ng = fed[i].num_rg;
    for(it=0; it<nt; it++){
      j = fed[i].scale_factor_f[it].ntp;
      if(j >= 0){fitpar[j] = fed[i].scale_factor[it];}
      j = fed[i].thickness_f[it].ntp;
      if(j >= 0){fitpar[j] = fed[i].thickness[it];}
      j = fed[i].thickwidth_f[it].ntp;
      if(j >= 0){fitpar[j] = fed[i].thickwidth[it];}
      for(ig=0; ig<ng; ig++){
        j = fed[i].H_shift_f[ig][it].ntp;
        if(j >= 0){fitpar[j] = fed[i].H_shift[ig][it];}
        j = fed[i].G_shift_f[ig][it].ntp;
        if(j >= 0){fitpar[j] = fed[i].G_shift[ig][it];}
      }
    }
  }
}


void fitpar_to_csd_fed(
  CRYSTAL_STRCT_DATA *csd, int num_exp_dataset, FIT_EXP_DATASET fed[],
  int num_fitpar, double fitpar[]
  )
{
  int i, it, j, ig, nt, ng, ia;

  /* lattice parameter */
  j = csd->lp_f.a.ntp;
  if(j >= 0){ csd->lp.a = fitpar[j]; }
  j = csd->lp_f.b.ntp;
  if(j >= 0){ csd->lp.b = fitpar[j]; }
  j = csd->lp_f.c.ntp;
  if(j >= 0){ csd->lp.c = fitpar[j]; }
  j = csd->lp_f.alpha.ntp;
  if(j >= 0){ csd->lp.alpha = fitpar[j]; }
  j = csd->lp_f.beta.ntp;
  if(j >= 0){ csd->lp.beta  = fitpar[j]; }
  j = csd->lp_f.gamma.ntp;
  if(j >= 0){ csd->lp.gamma = fitpar[j]; }

  /* ATOM */
  for(i=0; i<csd->n_atom; i++){
    j = csd->atm_f[i].x.ntp;
    if(j >= 0){ csd->atm[i].x = fitpar[j]; }
    j = csd->atm_f[i].y.ntp;
    if(j >= 0){ csd->atm[i].y = fitpar[j]; }
    j = csd->atm_f[i].z.ntp;
    if(j >= 0){ csd->atm[i].z = fitpar[j]; }
    if(csd->atm[i].bf == 0){
      j = csd->atm_f[i].b.ntp;
      if(j >= 0){ csd->atm[i].b = fitpar[j];}
    }
    else if(csd->atm[i].bf == 1){ 
      j = csd->atm_f[i].b11.ntp;
      if(j >= 0){ csd->atm[i].b11 = fitpar[j]; }
      j = csd->atm_f[i].b22.ntp;
      if(j >= 0){ csd->atm[i].b22 = fitpar[j]; }
      j = csd->atm_f[i].b33.ntp;
      if(j >= 0){ csd->atm[i].b33 = fitpar[j]; }
      j = csd->atm_f[i].b12.ntp;
      if(j >= 0){ csd->atm[i].b12 = fitpar[j]; }
      j = csd->atm_f[i].b23.ntp;
      if(j >= 0){ csd->atm[i].b23 = fitpar[j]; }
      j = csd->atm_f[i].b13.ntp;
      if(j >= 0){ csd->atm[i].b13 = fitpar[j]; }
    }
    j = csd->atm_f[i].occ.ntp;
    if(j >= 0){ csd->atm[i].occ = fitpar[j]; }
  }
  
  /* Low order structure factors */
  for(i=0; i<csd->num_lofg; i++){
    if(csd->lof_type == 1){
      j = csd->fg_lofg_f[i].re.ntp;
      if(j >= 0){csd->fg_lofg[i].re = fitpar[j];}
      j = csd->fg_lofg_f[i].im.ntp;
      if(j >= 0){csd->fg_lofg[i].im = fitpar[j];}
    }
    else if(csd->lof_type == 2){
      j = csd->fg_lofg_f[i].ela_re.ntp;
      if(j >= 0){csd->fg_lofg[i].ela_re = fitpar[j];}
      j = csd->fg_lofg_f[i].ela_im.ntp;
      if(j >= 0){csd->fg_lofg[i].ela_im = fitpar[j];}
    }
    else if(csd->lof_type == 3 || csd->lof_type == 4){
      for(ia=0; ia<csd->n_atom; ia++){
        j = csd->fg_lofg_f[i].fe[ia].ntp;
        if(j >= 0){csd->fg_lofg[i].fe[ia] = fitpar[j];}
        j = csd->fg_lofg_f[i].fi[ia].ntp;
        if(j >= 0){csd->fg_lofg[i].fi[ia] = fitpar[j];}
      }
    }
    else{
      mb_fprintf(stdout, "$$ Invalid lof_type %d. Abort.\n", csd->lof_type);
      exit(EXIT_FAILURE);
    }
  }

  /* FIT_EXP_DATASET *fed */
  for(i=0; i<num_exp_dataset; i++){
    j = fed[i].H_tilt_f.ntp;
    if(j >= 0){ fed[i].H_tilt = fitpar[j];}
    j = fed[i].G_tilt_f.ntp;
    if(j >= 0){ fed[i].G_tilt = fitpar[j];}
    j = fed[i].accel_kvol_f.ntp;
    if(j >= 0){ fed[i].accel_kvol = fitpar[j];}
    nt = fed[i].num_thickness;
    ng = fed[i].num_rg;
    for(it=0; it<nt; it++){
      j = fed[i].scale_factor_f[it].ntp;
      if(j >= 0){ fed[i].scale_factor[it] = fitpar[j];}
      j = fed[i].thickness_f[it].ntp;
      if(j >= 0){ fed[i].thickness[it] = fitpar[j];}
      j = fed[i].thickwidth_f[it].ntp;
      if(j >= 0){ fed[i].thickwidth[it] = fitpar[j];}
      for(ig=0; ig<ng; ig++){
        j = fed[i].H_shift_f[ig][it].ntp;
        if(j >= 0){ fed[i].H_shift[ig][it] = fitpar[j];}
        j = fed[i].G_shift_f[ig][it].ntp;
        if(j >= 0){ fed[i].G_shift[ig][it] = fitpar[j];}
      }
    }
  }
}


/* ----- step   ------------------------------------------------------------ */
/* ------------------------------------------------------------------------- */


double eval_step( double par )
{
     /*
      * Evaluate step for each free parameter
      * ( for calc. of jacobian )
      */
  double x;
  x = fabs( par ) * epsilon.perturb_step;
  if( x < epsilon.eps ){ x = epsilon.eps; }
  return x;
}


void thickness_step(FIT_EXP_DATASET *ed, double *tmp_thickness )
     /*
      * For calc. of jacobian elements of thicknesses.
      */
{
  int i, nt;
  
  nt = ed->num_thickness;
  
  for(i=0; i<nt; i++){
    tmp_thickness[i] = ed->thickness[i];
  }
  for(i=0; i<nt; i++){
    tmp_thickness[i + nt] = ed->thickness[i] + ed->thickness_f[i].dlt;
  }
}


void thickness_set(
  FIT_EXP_DATASET *ed, int num_thickwidth, double *tmp_thickness
  )
     /*
      * For calc. of jacobian elements of thicknesses.
      */
{
  int it, itw, nt, ntw, nn;
  
  nt = ed->num_thickness;
  ntw = num_thickwidth;
  nn = 2 * ntw + 1;
  
  if( ntw < 0 ){
    mb_fprintf(stdout, "$$ num_thickwidth = %d. Abort.\n", ntw);
    exit(EXIT_FAILURE);
  }
  
  for(it=0; it<nt; it++){
    /*mb_fprintf(stdout, "$$ ed->thickness[it] = %lf\n", ed->thickness[it]);*/
    for(itw=0; itw<nn; itw++){
      if(fabs(ed->thickwidth[it]) < EPS){				/* Without thickwidth */
        tmp_thickness[it * nn + itw] = ed->thickness[it];
        tmp_thickness[(it + nt) * nn + itw]
          = tmp_thickness[it * nn + itw] + ed->thickness_f[it].dlt;
      }
      else{                       /* With thickwidth */
        tmp_thickness[it * nn + itw]
          = ed->thickness[it]
            + ed->thickwidth[it] * (double) (itw - ntw)/ (double) ntw;
        tmp_thickness[(it + nt) * nn + itw]
          = tmp_thickness[it * nn + itw] + ed->thickness_f[it].dlt;
      }
      /*
      mb_fprintf(stdout,
        "$$tmp_thickness[it*nn+itw], tmp_thickness[(it+nt)*nn+itw]=%lf %lf\n",
        tmp_thickness[it * nn + itw], tmp_thickness[(it + nt) * nn + itw]
        );
        */
    }
  }
  
}


void thickweight_set( int num_thickwidth, double *thickness_weight )
     /*
      * Set the weight of intensities with thicknesses
      *  for thickness convolution
      */
{
  int i, n;
  double sum, wt, sq_sig;
  
  n = num_thickwidth * 2 + 1;
  sum=0.0;
  
  if(num_thickwidth == 0){
    thickness_weight[0] = 1.0;
    return;
  }
  
  if(epsilon.thickwidth_type == 0){					/* Rectangle */
    wt = 1.0 / (double) n;
    for(i=0; i<n; i++){
      thickness_weight[i] = wt;
    }
  }
  else if(epsilon.thickwidth_type == 1){				/* Triangle */
    mb_fprintf(stdout, "$$ num_thickwidth = %d\n", num_thickwidth);
    for(i=0; i<n; i++){
      thickness_weight[i]
        = 1.0 - fabs((double) i - (double) num_thickwidth)
          / (double) num_thickwidth;
      sum += thickness_weight[i];
    }
    for(i=0; i<n; i++){ thickness_weight[i] /= sum; }
  }
  else if(epsilon.thickwidth_type == 2){				/* Gaussian */
    /*
     * Gaussian distribution:
     *  P(e) = \exp( - e^2 / sig~2 /2 ) / \sqrt(2 * PI * sig^2)
     *  ( \sum_{-\inf}~{\inf} P(e) de = 1 )
     *
     * caution:  "thickwidth" corresponds to 3sigma.
     */
    sq_sig = sq( (double) num_thickwidth / 3.0 );
    for(i=0; i<n; i++){
      thickness_weight[i]
        = exp( - sq((double) i - (double) num_thickwidth) / ( 2.0 * sq_sig ) );
      sum += thickness_weight[i];
    }
    for(i=0; i<n; i++){ thickness_weight[i] /= sum; }
  }
  else{
    mb_fprintf(
      stdout, "$$ Sorry, thickwidth_type=%d is not supported. Abort.\n",
      epsilon.thickwidth_type
      );
    exit(EXIT_FAILURE);
  }
  
  /*	t = 0;
	for(i=0; i<n; i++){
		mb_fprintf(stdout, "$$ tw[%d] = %lf\n", i, thickness_weight[i]);
		t += thickness_weight[i];
	}
	mb_fprintf(stdout, "$$ Sum of weight = %lf\n", t);	*/
}


double thickness_average(
  int num_thickwidth, double thickwidth, double *tv_intensity, double *thickness_weight
  )
{
  int i, n;
  double av_int;
  
  n = num_thickwidth * 2 + 1;
  av_int = 0.0;
  
  if( fabs(thickwidth) < epsilon.eps ){
    av_int += tv_intensity[0];
  }
  else{
    for(i=0; i<n; i++){
      av_int += tv_intensity[i] * thickness_weight[i];
    }
  }
  
  return av_int;
}


void scale_and_average_2d(
  int num_thickwidth, double scale_factor, double thickwidth,
  int it, int nt, int ig, int ik, int num_perturb, double *thickness_weight, 
  double **tmp_intensity, double ***tmp_perturb_intensity,
  double ***intensity, double ****p_intensity
  )
{
  /* Scaling and Averaging in 2D */
  int iv, nv, ip;
  double *tv_intensity;

  if(num_thickwidth < 0){
    mb_fprintf(stdout, "$$ num_thickwidth must not be less than 0. Abort.\n");
    exit(EXIT_FAILURE);
  }

  nv = num_thickwidth * 2 + 1;
  tv_intensity = new_vector(double, nv);  chk_vector(tv_intensity);
  
  for(iv=0; iv<nv; iv++){
    tv_intensity[iv]
      = epsilon.intensity_scaling * scale_factor
        * tmp_intensity[ig][it * nv + iv];
  }
  intensity[ik][ig][it]
    = thickness_average(
      num_thickwidth, thickwidth, tv_intensity, thickness_weight
      );

  for(iv=0; iv<nv; iv++){/* thickness elements of jacobian calc. */
    tv_intensity[iv]
      = epsilon.intensity_scaling * scale_factor
        * tmp_intensity[ig][(it + nt) * nv + iv];
  }
  intensity[ik][ig][it + nt]
    = thickness_average(
      num_thickwidth, thickwidth, tv_intensity, thickness_weight
      );

  for(ip=0; ip<num_perturb; ip++){
    for(iv=0; iv<nv; iv++){
      tv_intensity[iv]
        = epsilon.intensity_scaling * scale_factor
          * tmp_perturb_intensity[ip][ig][it * nv + iv];
    }
    p_intensity[ik][ig][it][ip]
      = thickness_average(
      num_thickwidth, thickwidth, tv_intensity, thickness_weight
        );
  }
  free_vector(tv_intensity);
}


#if 0
void scale_and_av(
  int num_thickwidth, double scale_factor, double thickwidth,
  int it, int nt, int tmp_g, int num_perturb, double *thickness_weight, 
  double **tmp_intensity, double ***tmp_perturb_intensity,
  double **t_intensity, double ***tp_intensity
  )
     /* Scaling and Averaging */
{
  int iv, nv, ip;
  double *tv_intensity;
  
  if(num_thickwidth < 0){
    mb_fprintf(stdout, "$$ num_thickwidth must not be less than 0. Abort.\n");
    exit(EXIT_FAILURE);
  }
  
  nv = num_thickwidth * 2 + 1;
  tv_intensity = new_vector(double, nv);	chk_vector(tv_intensity);
  
  for(iv=0; iv<nv; iv++){
    tv_intensity[iv]
      = epsilon.intensity_scaling * scale_factor
        * tmp_intensity[tmp_g][it * nv + iv];
  }
  t_intensity[tmp_g][it]
    = thickness_average( num_thickwidth, thickwidth, tv_intensity, thickness_weight );
  
  for(iv=0; iv<nv; iv++){
    tv_intensity[iv]
      = epsilon.intensity_scaling * scale_factor
        * tmp_intensity[tmp_g][(it + nt) * nv + iv];
  }							/* thickness elements of jacobian calc. */
  t_intensity[tmp_g][it + nt]
    = thickness_average( num_thickwidth, thickwidth, tv_intensity, thickness_weight);
  
  for(ip=0; ip<num_perturb; ip++){
    for(iv=0; iv<nv; iv++){
      tv_intensity[iv]
        = epsilon.intensity_scaling * scale_factor
          * tmp_perturb_intensity[ip][tmp_g][it * nv + iv];
    }
    tp_intensity[ip][tmp_g][it]
      = thickness_average( num_thickwidth, thickwidth, tv_intensity, thickness_weight);
  }
  
  free_vector(tv_intensity);
}
#endif


void set_csd_fed_step( CRYSTAL_STRCT_DATA *csd, int num_exp_dataset, FIT_EXP_DATASET *fed )
     /*
      * Set csd and fed step for perturbation
      */
{
  int	i, ig, it, ng, nt, ia;
  double sqrt_eps;
  
  sqrt_eps = sqrt( epsilon.perturb_step );
  
  csd->lp_f.a.dlt = eval_step( csd->lp.a );
  csd->lp_f.b.dlt = eval_step( csd->lp.b );
  csd->lp_f.c.dlt = eval_step( csd->lp.c );
  csd->lp_f.alpha.dlt = eval_step( csd->lp.alpha );
  csd->lp_f.beta.dlt  = eval_step( csd->lp.beta );
  csd->lp_f.gamma.dlt = eval_step( csd->lp.gamma );
  
  for(i=0; i<csd->n_atom; i++){
    csd->atm_f[i].x.dlt = eval_step( csd->atm[i].x );
    csd->atm_f[i].y.dlt = eval_step( csd->atm[i].y );
    csd->atm_f[i].z.dlt = eval_step( csd->atm[i].z );
    if(csd->atm[i].bf == 0){
      csd->atm_f[i].b.dlt = eval_step( csd->atm[i].b );
    }
    else if(csd->atm[i].bf == 1){ 
      csd->atm_f[i].b11.dlt = eval_step( csd->atm[i].b11 );
      csd->atm_f[i].b22.dlt = eval_step( csd->atm[i].b22 );
      csd->atm_f[i].b33.dlt = eval_step( csd->atm[i].b33 );
      csd->atm_f[i].b12.dlt = eval_step( csd->atm[i].b12 );
      csd->atm_f[i].b23.dlt = eval_step( csd->atm[i].b23 );
      csd->atm_f[i].b13.dlt = eval_step( csd->atm[i].b13 );
    }
    csd->atm_f[i].occ.dlt  = eval_step( csd->atm[i].occ );
  }
  for(i=0; i<csd->num_lofg; i++){
    if(csd->lof_type == 1){
      csd->fg_lofg_f[i].re.dlt = eval_step( csd->fg_lofg[i].re );
      csd->fg_lofg_f[i].im.dlt = eval_step( csd->fg_lofg[i].im );
    }
    else if(csd->lof_type == 2){
      csd->fg_lofg_f[i].ela_re.dlt = eval_step( csd->fg_lofg[i].ela_re );
      csd->fg_lofg_f[i].ela_im.dlt = eval_step( csd->fg_lofg[i].ela_im );
    }
    else if(csd->lof_type == 3 || csd->lof_type == 4){
      for(ia=0; ia<csd->n_atom; ia++){
        csd->fg_lofg_f[i].fe[ia].dlt = eval_step( csd->fg_lofg[i].fe[ia] );
        csd->fg_lofg_f[i].fi[ia].dlt = eval_step( csd->fg_lofg[i].fi[ia] );
      }
    }
    else{
      mb_fprintf(stdout, "$$ Invalid lof_type %d. Abort.\n", csd->lof_type);
      exit(EXIT_FAILURE);
    }
  }
  
  for(i=0; i<num_exp_dataset; i++){
    fed[i].H_tilt_f.dlt = eval_step( fed[i].H_tilt );
    fed[i].G_tilt_f.dlt = eval_step( fed[i].G_tilt );
    fed[i].accel_kvol_f.dlt = eval_step( fed[i].accel_kvol );
    nt = fed[i].num_thickness;
    ng = fed[i].num_rg;
    for(it=0; it<nt; it++){
      fed[i].scale_factor_f[it].dlt = eval_step( fed[i].scale_factor[it] );
      fed[i].thickness_f[it].dlt = eval_step( fed[i].thickness[it] );
      /* fed[i].thickness_step[it] = 5.0;	*/ /* <<<<< */
      for(ig=0; ig<ng; ig++){
        fed[i].H_shift_f[ig][it].dlt = eval_step( fed[i].H_shift[ig][it] );
        fed[i].G_shift_f[ig][it].dlt = eval_step( fed[i].G_shift[ig][it] );
      }
    }
  }
}


/* ------------------------------------------------------------------------- */
/* ----- Jacobian ---------------------------------------------------------- */
/* ------------------------------------------------------------------------- */
void ith_csd_prtb(
  int ith_par, int nth, double *par_perturb, double par, double step
  )
{
  if(ith_par == nth){
    *par_perturb = par + step;
  }
  else{
    *par_perturb = par;
  }
}


void make_csd_perturb(
  int ith_prtb, CRYSTAL_STRCT_DATA *csd, CRYSTAL_STRCT_DATA *csd_prtb
  )
{
  /* Only for ith_parameter to be perturbed */
  int i, ia;
  /* .flg --> .ntp */
  
  /* Lattice parameter */
  ith_csd_prtb(
    ith_prtb, csd->lp_f.a.ntp, &(csd_prtb->lp.a), csd->lp.a, csd->lp_f.a.dlt
    );
  ith_csd_prtb(
    ith_prtb, csd->lp_f.b.ntp, &(csd_prtb->lp.b), csd->lp.b, csd->lp_f.b.dlt
    );
  ith_csd_prtb(
    ith_prtb, csd->lp_f.c.ntp, &(csd_prtb->lp.c), csd->lp.c, csd->lp_f.c.dlt
    );
  ith_csd_prtb(
    ith_prtb, csd->lp_f.alpha.ntp, &(csd_prtb->lp.alpha),
    csd->lp.alpha, csd->lp_f.alpha.dlt
    );
  ith_csd_prtb(
    ith_prtb, csd->lp_f.beta.ntp,  &(csd_prtb->lp.beta),
    csd->lp.beta,  csd->lp_f.beta.dlt
    );
  ith_csd_prtb(
    ith_prtb, csd->lp_f.gamma.ntp, &(csd_prtb->lp.gamma),
    csd->lp.gamma, csd->lp_f.gamma.dlt
    );

  /* ATOM */
  for(i=0; i<csd->n_atom; i++){
    ith_csd_prtb(
      ith_prtb, csd->atm_f[i].x.ntp, &(csd_prtb->atm[i].x),
      csd->atm[i].x, csd->atm_f[i].x.dlt
      );
    ith_csd_prtb(
      ith_prtb, csd->atm_f[i].y.ntp, &(csd_prtb->atm[i].y),
      csd->atm[i].y, csd->atm_f[i].y.dlt
      );
    ith_csd_prtb(
      ith_prtb, csd->atm_f[i].z.ntp, &(csd_prtb->atm[i].z),
      csd->atm[i].z, csd->atm_f[i].z.dlt
      );
    if(csd->atm[i].bf == 0){
      ith_csd_prtb(
        ith_prtb, csd->atm_f[i].b.ntp, &(csd_prtb->atm[i].b),
        csd->atm[i].b, csd->atm_f[i].b.dlt
        );
    }
    else if(csd->atm[i].bf == 1){ 
      ith_csd_prtb(
        ith_prtb, csd->atm_f[i].b11.ntp, &(csd_prtb->atm[i].b11),
        csd->atm[i].b11, csd->atm_f[i].b11.dlt
        );
      ith_csd_prtb(
        ith_prtb, csd->atm_f[i].b22.ntp, &(csd_prtb->atm[i].b22),
        csd->atm[i].b22, csd->atm_f[i].b22.dlt
        );
      ith_csd_prtb(
      ith_prtb, csd->atm_f[i].b33.ntp, &(csd_prtb->atm[i].b33),
        csd->atm[i].b33, csd->atm_f[i].b33.dlt
        );
      ith_csd_prtb(
        ith_prtb, csd->atm_f[i].b12.ntp, &(csd_prtb->atm[i].b12),
        csd->atm[i].b12, csd->atm_f[i].b12.dlt
        );
      ith_csd_prtb(
        ith_prtb, csd->atm_f[i].b23.ntp, &(csd_prtb->atm[i].b23),
        csd->atm[i].b23, csd->atm_f[i].b23.dlt
        );
      ith_csd_prtb(
        ith_prtb, csd->atm_f[i].b13.ntp, &(csd_prtb->atm[i].b13),
        csd->atm[i].b13, csd->atm_f[i].b13.dlt
        );
    }
    ith_csd_prtb(
      ith_prtb, csd->atm_f[i].occ.ntp,  &(csd_prtb->atm[i].occ),
      csd->atm[i].occ,  csd->atm_f[i].occ.dlt
      );
  }

  /* Low order structure factors */
  for(i=0; i<csd->num_lofg; i++){
    if(csd->lof_type == 1){
      ith_csd_prtb(
        ith_prtb, csd->fg_lofg_f[i].re.ntp, &(csd_prtb->fg_lofg[i].re),
        csd->fg_lofg[i].re, csd->fg_lofg_f[i].re.dlt
        );
      ith_csd_prtb(
        ith_prtb, csd->fg_lofg_f[i].im.ntp, &(csd_prtb->fg_lofg[i].im),
        csd->fg_lofg[i].im, csd->fg_lofg_f[i].im.dlt
        );
    }
    else if(csd->lof_type == 2){
      ith_csd_prtb(
        ith_prtb,
        csd->fg_lofg_f[i].ela_re.ntp, &(csd_prtb->fg_lofg[i].ela_re),
        csd->fg_lofg[i].ela_re, csd->fg_lofg_f[i].ela_re.dlt
        );
      ith_csd_prtb(
        ith_prtb,
        csd->fg_lofg_f[i].ela_im.ntp, &(csd_prtb->fg_lofg[i].ela_im),
        csd->fg_lofg[i].ela_im, csd->fg_lofg_f[i].ela_im.dlt
        );
    }
    else if(csd->lof_type == 3 || csd->lof_type == 4){
      for(ia=0; ia<csd->n_atom; ia++){
        ith_csd_prtb(
          ith_prtb,
          csd->fg_lofg_f[i].fe[ia].ntp, &(csd_prtb->fg_lofg[i].fe[ia]),
          csd->fg_lofg[i].fe[ia], csd->fg_lofg_f[i].fe[ia].dlt
          );
        ith_csd_prtb(
          ith_prtb,
          csd->fg_lofg_f[i].fi[ia].ntp, &(csd_prtb->fg_lofg[i].fi[ia]),
          csd->fg_lofg[i].fi[ia], csd->fg_lofg_f[i].fi[ia].dlt
          );
      }
    }
    else{
      mb_fprintf(stdout, "$$ Invalid lof_type %d. Abort.\n", csd->lof_type);
      exit(EXIT_FAILURE);
    }
  }
}


double ****alloc_jacobian(
  int num_thickness, int num_rg, int *num_data_points, int num_fitpar
  )
{
  int it, ig, ifp, nt, ng, nd, ik, nd_sum;
  double ****jacobian, *z;
  void ***x, **y;
  
  nt = num_thickness;
  ng = num_rg;
  
  jacobian = new_tensor4_element(num_fitpar);	chk_tensor4(jacobian);
  x = new_tensor_element(num_fitpar * ng);	chk_tensor(x);
  y = new_matrix_element(num_fitpar * ng * nt);	chk_matrix(y);
  nd_sum = 0;	for(ig=0; ig<ng; ig++){	nd_sum += num_data_points[ig]; }
  z = new_vector_element(double, num_fitpar * ng * nt * nd_sum );	chk_vector(z);
  for(ifp=0; ifp<num_fitpar; ifp++){
    jacobian[ifp] = (double ***)(x);	x += ng;		/* <<<<< 19 Oct. 2000 */
    for(ig=0; ig<ng; ig++){
      jacobian[ifp][ig] = (double **)(y); y += nt;	/* <<<<< 19 Oct. 2000 */
      nd = num_data_points[ig];
      for(it=0; it<nt; it++){
        jacobian[ifp][ig][it] = (double *)(z); z += nd;
      }
    }
  }
  
  /* init. & chk */
  for(ifp=0; ifp<num_fitpar; ifp++){
    for(ig=0; ig<ng; ig++){
      nd = num_data_points[ig];
      for(it=0; it<nt; it++){
        for(ik=0; ik<nd; ik++){
          jacobian[ifp][ig][it][ik] = 0.0;
        }
      }
    }
  }
  return jacobian;
}


void free_jacobian(int num_fitpar, FIT_EXP_DATASET *ed, double ****jacobian)
{
  free_tensor4(jacobian);
}


void init_jacobian(
  int num_fitpar, int ng, int nt, int *nk_g, double ****jacobian
  )
{
  /* Initialize jacobian */
  int	ifp, ig, it, ik, nk;
  
  for(ifp=0; ifp<num_fitpar; ifp++){
    for(ig=0; ig<ng; ig++){
      nk = nk_g[ig];
      for(it=0; it<nt; it++){
        for(ik=0; ik<nk; ik++){
          /*
           mb_fprintf(stdout,
             "$$ jacobian init : ifp,ig,it,ik = %d %d %d %d\n",
              ifp, ig, it, ik
             );
           */
          jacobian[ifp][ig][it][ik] = 0.0;
        }
      }
    }
  }
}


double ***alloc_jacobian_j( int num_rg, int num_thickness, int *num_data_points )
{
  int it, ig, ik, nk;
  double ***jacobian;
  
  jacobian = new_tensor_element(num_rg);	chk_tensor(jacobian);
  for(ig=0; ig<num_rg; ig++){
    jacobian[ig] = new_matrix_element(num_thickness);	chk_matrix(jacobian[ig]);
  }
  for(ig=0; ig<num_rg; ig++){
    nk = num_data_points[ig];
    for(it=0; it<num_thickness; it++){
      jacobian[ig][it] = new_vector_element(double, nk);
      chk_vector(jacobian[ig][it]);
    }
  }
  
  /* init. & chk */
  for(ig=0; ig<num_rg; ig++){
    nk = num_data_points[ig];
    for(it=0; it<num_thickness; it++){
      for(ik=0; ik<nk; ik++){
        jacobian[ig][it][ik] = 0.0;
      }
    }
  }	
  return jacobian;
}


void free_jacobian_j(double ***jacobian)
{
  free_tensor(jacobian);
}


void init_jacobian_j(int ng, int nt, int *nk_g, double ***jacobian)
     /* Initialize jacobian */
{
  int	ig, it, ik, nk;
  
  for(ig=0; ig<ng; ig++){
    nk = nk_g[ig];
    for(it=0; it<nt; it++){
      for(ik=0; ik<nk; ik++){
        jacobian[ig][it][ik] = 0.0;
      }
    }
  }
}


/* ------------------------------------------------------------------------- */
/* ----- interpolations ---------------------------------------------------- */
/* ------------------------------------------------------------------------- */
double cubic_intrpl(double x[], double xx, double f[])
     /* Use Lagrange interpolating cubic */
{
  double x1, x2, x3, x4, x12, x13, x14, x23, x24, x34, res;
  x1 = xx - x[0];
  x2 = xx - x[1];
  x3 = xx - x[2];
  x4 = xx - x[3];
  x12 = x[0] - x[1];
  x13 = x[0] - x[2];
  x14 = x[0] - x[3];
  x23 = x[1] - x[2];
  x24 = x[1] - x[3];
  x34 = x[2] - x[3];
  res =
    f[0] * x2 * x3 * x4 / (x12 * x13 * x14) -
      f[1] * x1 * x3 * x4 / (x12 * x23 * x24) +
        f[2] * x1 * x2 * x4 / (x13 * x23 * x34) -
          f[3] * x1 * x2 * x3 / (x14 * x24 * x34);
  return res;
}


double bicubic_intrpl(
  double xx, double yy, int nx, int ny, double *x, double *y, double **val
  )
     /* 2 dimensional */
{
  double h[4], f[4], hh[4], ff[4], res;
  int i, j, ix, iy;
  
  ix = 1; do{ ix++; } while(xx >= x[ix] && ix < nx - 2);
  iy = 1; do{ iy++; } while(yy >= y[iy] && iy < ny - 2);
  /*mb_fprintf(stdout, "$$ ix, iy = %d %d \n", ix, iy);*/
  
  for(j=0; j<4; j++){
    for(i=0; i<4; i++){
      h[i] = x[i + ix - 2];
      f[i] = val[i + ix - 2][j + iy - 2];
      /*mb_fprintf(stdout, "$$ i f h = %d %lf %lf\n", i, f[i], h[i]);*/
    }
    ff[j] = cubic_intrpl(h, xx, f);
    hh[j] = y[j + iy - 2];
    /*mb_fprintf(stdout, "$$ j ff hh = %d %lf %lf\n", j, ff[j], hh[j]);*/
  }
  res = cubic_intrpl(hh, yy, ff);
  
  return res;
}


#if 0
void  intrpl_2d(
  FIT_EXP_DATASET *ed, double ***intensity, R_HKL *k_xy, HGPIX *k_pix,
  double H_tilt, double G_tilt, int ng, int nt, int nk, int n
  )
{
     /* n: number of pixels in H (or G) direction */
  int		i, j, ig, it, ik;
  double **image, *hh, *gg, bh, bg;
  
  image = new_matrix(double, n, n);	chk_matrix(image);
  hh = new_vector(double, n);			chk_vector(hh);
  gg = new_vector(double, n);			chk_vector(gg);
  
  for(i=0; i<n; i++){
    for(j=0; j<n; j++){
      image[i][j] = 0.0;		/* initialize */
    }
  }
  
  for(ig=0; ig<ng; ig++){
    for(it=0; it<nt; it++){
      for(ik=0; ik<nk; ik++){
        i = k_pix[ik].ih;		j = k_pix[ik].ig;
        hh[i] = k_pix[ik].H;	gg[j] = k_pix[ik].G;
        image[i][j] = intensity[ig][it][ik];
      }
      /* plan: chk hh, gg */
      for(ik=0; ik<nk; ik++){
        bh = H_tilt + ed->obs_H[ig][it][ik];
        bg = G_tilt + ed->obs_G[ig][it][ik];
        ed->cal_intensity[ig][it][ik] = bicubic_intrpl(bh, bg, n, n, hh, gg, image);
      }
    }
  }
  
  free_matrix(image);
  free_vector(hh);	free_vector(gg);
}
#endif


void gaussian_convolution(
  int npix, double **intensity, int num_avrpix, double **pixwt
  )
{
  int ix, iy, cx, cy, avx, avy, mx, my;
  double **tpix;
  
  if( num_avrpix > 1 && (num_avrpix % 2) == 0 ){
    mb_fprintf(
      stdout, "$$num_avrpix = %d: num_avrpix must be odd number. Abort.\n",
      num_avrpix
      );
    mb_fprintf(stdout, "$$(gaussian_convolution)\n");
    exit(EXIT_FAILURE);
  }
  
  cx = cy = (num_avrpix -1) / 2; /* ix:0~num_avrpix-1, iy:0~num_avrpix-1 */
  tpix = new_matrix(double, npix, npix);	chk_matrix(tpix);
  
  for(ix=0; ix<npix; ix++){
    for(iy=0; iy<npix; iy++){
      tpix[ix][iy] = 0.0;
    }
  }
  
  for(avx=0; avx<num_avrpix; avx++){
    for(avy=0; avy<num_avrpix; avy++){
      
      for(ix=0; ix<npix; ix++){
        for(iy=0; iy<npix; iy++){
          mx = ix-(avx-cx);	my = iy-(avy-cy);
          if(mx >= 0 && mx < npix && my >=0 && my < npix){
            tpix[ix][iy] += intensity[mx][my] * pixwt[avx][avy];
          }
        }
      }
      
    }
  }
  for(ix=0; ix<npix; ix++){
    for(iy=0; iy<npix; iy++){
      /*
        mb_fprintf(
          stdout, "$$ intensity, tpix = %lf %lf\n",
          intensity[ix][iy], tpix[ix][iy]
          );
       */
      intensity[ix][iy] = tpix[ix][iy];
    }
  }
  free_matrix(tpix);
}


void gaussian_conv(
  int nx_pix, int ny_pix, double **intensity, int num_avrpix, double **pixwt
  )
{
  int ix, iy, cx, cy, avx, avy, mx, my;
  double **tpix;
  
  if( num_avrpix > 1 && (num_avrpix % 2) == 0 ){
    mb_fprintf(
      stdout, "$$num_avrpix = %d: num_avrpix must be odd number. Abort.\n",
      num_avrpix
      );
    mb_fprintf(stdout, "$$(gaussian_conv)\n");
    exit(EXIT_FAILURE);
  }
  
  cx = cy = (num_avrpix -1) / 2;  /* ix:0~num_avrpix-1, iy:0~num_avrpix-1 */
  tpix = new_matrix(double, nx_pix, ny_pix);  chk_matrix(tpix);
  
  for(ix=0; ix<nx_pix; ix++){
    for(iy=0; iy<ny_pix; iy++){
      tpix[ix][iy] = 0.0;
    }
  }
  
  for(avx=0; avx<num_avrpix; avx++){
    for(avy=0; avy<num_avrpix; avy++){
      for(ix=0; ix<nx_pix; ix++){
        for(iy=0; iy<ny_pix; iy++){
          mx = ix-(avx-cx);  my = iy-(avy-cy);
          if(mx >= 0 && mx < nx_pix && my >=0 && my < ny_pix){
            tpix[ix][iy] += intensity[mx][my] * pixwt[avx][avy];
          }
        }
      }
      
    }
  }
  for(ix=0; ix<nx_pix; ix++){
    for(iy=0; iy<ny_pix; iy++){
      intensity[ix][iy] = tpix[ix][iy];
    }
  }
  free_matrix(tpix);
}


void interpolation_2d(
  double *itp_intensity, double *intensity,
  R_HKL *k_xy, HGPIX *k_pix,
  double H_shift, double G_shift, double *obs_H, double *obs_G,
  int obs_num_points, int npixels, int nk,
  int num_avrpix, double **avrpix_wt
  )
{
  /*
   * 2D interpolation of 1D-array image.
   *  (2D intensity distribution is expressed as 1D-array image in mbfit.)
   *  npixels: number of pixels in H (or G) direction
   */
  int  ix, iy, ik;
  double  **image, *hh, *gg, bh, bg;

  image = new_matrix(double, npixels, npixels);  chk_matrix(image);
  hh = new_vector(double, npixels);  chk_vector(hh);
  gg = new_vector(double, npixels);  chk_vector(gg);

  /* Initialize */
  for(ix=0; ix<npixels; ix++){
    hh[ix] = 0.0;  gg[ix] = 0.0;
    for(iy=0; iy<npixels; iy++){
      image[ix][iy] = 0.0;
    }
  }
  /* Transformation from 1D array to 2D array */
  for(ik=0; ik<nk; ik++){
    ix = k_pix[ik].ih;     iy = k_pix[ik].ig;
    hh[ix] = k_pix[ik].H;  gg[iy] = k_pix[ik].G;
    image[ix][iy] = intensity[ik];
  }
  /* ==> chk hh, gg */
  if( num_avrpix > 1 ){
    gaussian_convolution( npixels, image, num_avrpix, avrpix_wt );
  }

  for(ik=0; ik<obs_num_points; ik++){
    bh = H_shift + obs_H[ik];  bg = G_shift + obs_G[ik];
    itp_intensity[ik] = bicubic_intrpl(
      bh, bg, npixels, npixels, hh, gg, image
      );
  }
  free_matrix(image);
  free_vector(hh);  free_vector(gg);
}


void interpolate2d_of_1d_array(
  /* orginal intensity (1D) ==> input */
  double *org_intensity, int org_num_k_points, int *org_hpix, int *org_gpix,
  int nx_pixels, int ny_pixels, double *org_H, double *org_G,
  /* interpolated intensity (1D) ==> output */
  double *itp_intensity, int itp_num_k_points, double *itp_H, double *itp_G, 
  /* shift and averaging parameters */
  double H_shift, double G_shift, int num_avrpix, double **avrpix_wt
  )
{
  /*
    2D interpolation of 1D-array image.
      (2D intensity distribution is expressed as 1D-array image in mbfit.)
      (Caution: npixels is the number of pixels in one H (or G) direction)
   */
  int  ix, iy, ik;
  double  **image, *xpos, *ypos;
  double  bigh, bigg;

  image = new_matrix(double, nx_pixels, ny_pixels);  chk_matrix(image);
  xpos = new_vector(double, nx_pixels);  chk_vector(xpos);
  ypos = new_vector(double, ny_pixels);  chk_vector(ypos);

  /* Initialize */
  for(ix=0; ix<nx_pixels; ix++){
    xpos[ix] = 0.0;
  }
  for(iy=0; iy<ny_pixels; iy++){
    ypos[ix] = 0.0;
  }
  for(ix=0; ix<nx_pixels; ix++){
    for(iy=0; iy<ny_pixels; iy++){
      image[ix][iy] = 0.0;
    }
  }
  /* Transformation from 1D array to 2D array */
  for(ik=0; ik<org_num_k_points; ik++){
    ix = org_hpix[ik];  iy = org_gpix[ik];
    xpos[ix] = org_H[ik];  ypos[iy] = org_G[ik];
    image[ix][iy] = org_intensity[ik];
    /*
    printf(
      "$$ image[%d][%d] = %lf, H, G=%lf %lf\n",
      ix, iy, image[ix][iy], xpos[ix], ypos[iy]
      );
     */
  }
  /* xpos, ypos should be checked here..*/
  if( num_avrpix > 1 ){
    gaussian_conv( nx_pixels, ny_pixels, image, num_avrpix, avrpix_wt );
  }

  for(ik=0; ik<itp_num_k_points; ik++){
    bigh = H_shift + itp_H[ik];  bigg = G_shift + itp_G[ik];
    itp_intensity[ik] = bicubic_intrpl(
      bigh, bigg, nx_pixels, ny_pixels, xpos, ypos, image
      );
    /*
    printf(
      "itp_intensity = %lf, bigh bigg=%lf %lf\n",
      itp_intensity[ik], bigh, bigg
      );
     */
  }
  free_matrix(image);
  free_vector(xpos);  free_vector(ypos);
}


/* ------------------------------------------------------------------------- */
/* ----- data processing --------------------------------------------------- */
/* ------------------------------------------------------------------------- */
void mb_thinout( void )
     /* 
      * Thin out number of data points of intensity data
      * coded on 981222
      */
{
  int		it, nt, ig, ng, ik, nd;
  int		new_nd, increment, ct;
  /*	int		num_beams; */
  double	obsH, obsG, obsint, obssig;
  HKL		rg;
  char	infile[BUFLEN], outfile[BUFLEN], buf[BUFLEN];
  FILE	*fi, *fo;

  mb_fprintf(stdout, "***** THINOUT *****\n");
  mb_fprintf(stdout, " Thin out data points of intensity data.\n");
  mb_fprintf(stdout, " Intensity file name to be thinned out (with .int) >>");
  fgets(infile, BUFLEN, stdin);
  fi = fopen_chk_rmode(infile); /* res file (input) */

  mb_fprintf(stdout, " File name for output (with.int) >>");
  fgets(outfile, BUFLEN, stdin);
  if( strcmp(infile, outfile) == 0 ){
    mb_fprintf(
      stdout, "$$ Output file name must be different from the previous one.\n"
      );
    exit(EXIT_FAILURE);
  }
  fo = mb_fopen_chk_wmode(outfile);

  mb_fprintf(stdout, " Increment for thinning out (integer, not zero) ?>>");
  /*gets(buf); increment = atoi(buf);*/
  fgets(buf, sizeof buf, stdin); increment = atoi(rm_cr(buf));
  if(increment <= 0){
    mb_fprintf(stdout, "$$ increment must be larger than zero.\n");
    exit(EXIT_FAILURE);
  }
  mb_fprintf(stdout, "$$ increment = %d\n", increment);

  /* Data I/O */
  mb_fprintf(fo, "%% %s\n", outfile);
  mb_fprintf(fo, "%% Number of thicknesses\n");
  nt = fget_i(fi);
  mb_fprintf(fo, "%d\n", nt);
  
  ng = fget_i(fi); /* Number of reflections to be treated as experimental data */
  mb_fprintf(fo, "%% Number of reflecitons\n");
  mb_fprintf(fo, "%d\n", ng);
  
  for(it=0; it<nt; it++){
    mb_fprintf(fo, "%% Thickness %d\n", it);
    for(ig=0; ig<ng; ig++){
      rg = fget_hkl(fi);
      mb_fprintf(fo, "%% hkl\n");
      mb_fprintf(fo, "%d\t%d\t%d\n", rg.h, rg.k, rg.l);
      
      nd = fget_i(fi);
      if( nd % increment == 0 ){
        new_nd = (int) (nd / increment);		/* <<<<< */
      }
      else{
        new_nd = (int) (nd / increment) + 1;		/* <<<<< */
      }
      mb_fprintf(fo, "%% Number of data points\n");
      mb_fprintf(fo, "%d\n", new_nd);

      mb_fprintf(fo, "%%H\tG\tint\tsig\n");
      ct = 0;
      for(ik=0; ik<nd; ik++){
        fget_4d( &(obsH), &(obsG), &(obsint), &(obssig), fi );
        if(obssig <= 0.0){ mb_fprintf(stdout, "$$ warning !!: sig <= 0.0\n"); }
        if( ik % increment == 0 ){
          mb_fprintf(fo, "%lf\t%lf\t%lf\t%lf\n", obsH, obsG, obsint, obssig);
          ct++;
        }
      }
      if(ct != new_nd){
        mb_fprintf(
          stdout, "$$ ct = %d does not agree new_nd = %d. Please chk.\n",
          ct, new_nd
          );
        exit(EXIT_FAILURE);
      }
      chk_data("RF_END",fi);
      mb_fprintf(fo, "RF_END\n");
    }
    chk_data("THICK_END",fi);

    mb_fprintf(fo, "THICK_END\n");
  }
  chk_data("INTENSITY_DATA_END",fi);
  mb_fprintf(fo, "INTENSITY_DATA_END\n");
  
  fclose(fi);
  mb_fclose(fo);
}


/* output ------------------------------------------------------------------ */
/* ------------------------------------------------------------------------- */
void print_int_ratio(
  FILE *fo, int num_exp_dataset, FIT_EXP_DATASET *fed, FIT_EXP_INTSET *fei
  )
{
  /*
    Intensities: ZOLZ/HOLZ Ratio
   */
  int  ie, ig, ng, it, nt, ik, nk, nlz;
  long int total_num_data, numz_data, numh_data;
  double  obs, sig, g, aw;
  /*	double	weight; */
  double  intg_int, intg_intz, intg_inth;
  double  sq_intg_intw, sq_intg_intwz, sq_intg_intwh;

  total_num_data = numz_data = numh_data = 0;
  intg_int = intg_intz = intg_inth = 0.0;
  sq_intg_intw = sq_intg_intwz = sq_intg_intwh = 0.0;

  for(ie=0; ie<num_exp_dataset; ie++){ /* each set of experimental data */
    ng = fed[ie].num_rg;
    nt = fed[ie].num_thickness;
    for(it=0; it<nt; it++){
      for(ig=0; ig<ng; ig++){
        nk = fei[ie].num_data_points[ig];
        total_num_data += nk;

        nlz = fed[ie].incidence.u * fei[ie].obs_rg[ig].h
          + fed[ie].incidence.v * fei[ie].obs_rg[ig].k
            + fed[ie].incidence.w * fei[ie].obs_rg[ig].l; /* Laue zone */

        if( nlz == 0 ){/* ZOLZ */
          numz_data += nk;
        }
        else if (nlz > 0) {/* HOLZ */
          numh_data += nk;
        }
        else{
          mb_fprintf(stdout, "nlz < 0. Abort.\n");
          exit(EXIT_FAILURE);
        }

        for(ik=0; ik<nk; ik++){
          obs = fei[ie].obs_intensity[ig][it][ik];
          sig = fei[ie].obs_sig[ig][it][ik];
          g = fed[ie].g[ig];
          aw = fed[ie].art_weight[ig];

          intg_int += obs;
          sq_intg_intw += sq(obs) * aw / sq(sig);

          if( nlz == 0 ){/* ZOLZ */
            intg_intz += obs;
            sq_intg_intwz += sq(obs) * aw / sq(sig);
          }
          else if (nlz > 0) {/* HOLZ */
            intg_inth += obs;
            sq_intg_intwh += sq(obs) * aw / sq(sig);
          }
        }
      }
    }
  }

  mb_fprintf(fo, "\nNumber of data points:\n");
  mb_fprintf(
    fo, "  (ZOLZ)= %ld, (HOLZ)= %ld, (Total)= %ld\n",
    numz_data, numh_data, total_num_data
    );
  mb_fprintf(fo, "Integrated intensity:\n");
  mb_fprintf(
    fo, "  (ZOLZ)= %lf, (HOLZ)= %lf, (Total)= %lf\n",
    intg_intz, intg_inth, intg_int
    );
  if(intg_inth > 0.0){
    mb_fprintf(
      fo, "  Ratio: ZOLZ/HOLZ = %lf\n",  intg_intz / intg_inth
      );
  }
  mb_fprintf(fo, "Squares of weighted integrated intensities:\n");
  mb_fprintf(
    fo, "  (ZOLZ)= %lf, (HOLZ)= %lf, (Total)= %lf\n",
    sq_intg_intwz, sq_intg_intwh, sq_intg_intw
    );
  if(sq_intg_intwh > 0.0){
    mb_fprintf(
      fo, "  Ratio: ZOLZ/HOLZ = %lf\n", sq_intg_intwz / sq_intg_intwh
      );
  }
  mb_fprintf(fo, "\n");
}


void fit_intensity_output(
  int num_exp_dataset, FIT_EXP_DATASET *fed, FIT_EXP_INTSET *fei,
  char *outfile
  )
{
  /*
   * intensities output (mbfit-output-format)
   */
  int ie, it, nt, ig, ng, ik, nk;
  HKL	g;
  FILE *fo;

  fo = mb_fopen_chk_wmode(outfile);

  for(ie=0; ie<num_exp_dataset; ie++){
    ng = fed[ie].num_rg;
    nt = fed[ie].num_thickness;
    for(it=0; it<nt; it++){
      for(ig=0; ig<ng; ig++){
        g = fei[ie].obs_rg[ig];
        mb_fprintf(fo, "%%h k l\n");
        mb_fprintf(fo, "%d\t%d\t%d\n", g.h, g.k, g.l);
        nk = fei[ie].num_data_points[ig];

        mb_fprintf(fo, "%%number of pixels\n");
        mb_fprintf(fo, "%d\n", nk);
        mb_fprintf(fo, "%%H\tG\tObs\tCal\tSig\n");
        for(ik=0; ik<nk; ik++) {
          mb_fprintf(
            fo, "%lf\t%lf\t%lf\t%lf\t%lf\n",
            fei[ie].obs_H[ig][it][ik], fei[ie].obs_G[ig][it][ik],
            fei[ie].obs_intensity[ig][it][ik], fei[ie].cal_intensity[ig][it][ik],
            fei[ie].obs_sig[ig][it][ik]
            );
        }
      }
    }
  }

  mb_fclose(fo);
}


void print_scaled_r_factors(
  FILE *fo, int num_exp_dataset, FIT_EXP_DATASET *fed, FIT_EXP_INTSET *fei
  )
{
  int ie, ig, ng, it, nt, ik, nk;
  double obs, cal, sig, ruw, rdw, swu, swd, g, aw, weight;
  /*	double rd, ru, scaled_R; */
  double scale_factor_w, scaled_Rw;
  
  ruw = 0.0;	rdw = 0.0;
  for(ie=0; ie<num_exp_dataset; ie++){			/* each set of experimental data */
    ng = fed[ie].num_rg;
    nt = fed[ie].num_thickness;
    for(it=0; it<nt; it++){
      /* ----- scale factor recommended ----- */
      swu = 0.0;
      swd = 0.0;
      for(ig=0; ig<ng; ig++){
        nk = fei[ie].num_data_points[ig];
        g = fed[ie].g[ig];
        aw = fed[ie].art_weight[ig];
        for(ik=0; ik<nk; ik++){
          obs = fei[ie].obs_intensity[ig][it][ik];
          cal = fei[ie].cal_intensity[ig][it][ik];
          sig = fei[ie].obs_sig[ig][it][ik];
          weight = cal_weight( sig, obs, cal, g, aw );
          swu += obs * cal * weight;
          swd += sq(cal) * weight;
        }
      }
      scale_factor_w = swu / swd;
      mb_fprintf(
        fo, "Data set %d: temporary scale = %lf,", ie, scale_factor_w
        );

      for(ig=0; ig<ng; ig++){
        nk = fei[ie].num_data_points[ig];
        g = fed[ie].g[ig];
        aw = fed[ie].art_weight[ig];
        for(ik=0; ik<nk; ik++){
          obs = fei[ie].obs_intensity[ig][it][ik];
          cal = fei[ie].cal_intensity[ig][it][ik];
          sig = fei[ie].obs_sig[ig][it][ik];
          weight = cal_weight( sig, obs, cal, g, aw );	/* <<<<< */
          swu += obs * cal * weight;
          swd += sq(cal) * weight;		
          ruw += sq(obs - scale_factor_w * cal) * weight;
          rdw += sq(obs) * weight;
        }
      }
    }
  }
  scaled_Rw = sqrt( ruw / rdw );
  mb_fprintf( fo, " scaled R_w = %lf\n", scaled_Rw );
}


double rf_r_factor(double *obs, double *cal, int num_points)
{
  /* For individual R-factor */
  int i;
  double ru, rd, r;

  ru = 0.0; rd = 0.0;
  for(i=0; i<num_points; i++){
    ru += fabs(obs[i] - cal[i]);
    rd += fabs(obs[i]);
  }
  
  if (rd > 0){
    r = ru / rd;
  }
  else{
    r = 1e+3;
  }
  return r;
}


void print_r_factors(
  FILE *fo, int num_exp_dataset, FIT_EXP_DATASET *fed, FIT_EXP_INTSET *fei
  )
{
  /*
   * R-factors
   */
  int ie, ig, ng, it, nt, ik, nk, total_num_data, numz_data, numh_data, nlz;
  double obs, cal, sig, weight, g, aw, dy;
  double ru, rd, ruw, rdw, ruh, rdh, ruz, rdz, ruwh, rdwh, ruwz, rdwz;
  double chisq, chisq_holz, chisq_zolz, gof, gof_zolz, gof_holz;
  double ruwf, rdwf, ruws, rdws;

  ru = 0.0;	rd = 0.0;	ruw = 0.0;	rdw = 0.0;
  ruz = 0.0; rdz = 0.0;ruwz = 0.0;	rdwz = 0.0;  /* ZOLZ */
  ruh = 0.0;	rdh = 0.0;	ruwh = 0.0;	rdwh = 0.0;  /* HOLZ */
  ruwf = 0.0;	rdwf = 0.0; ruws = 0.0; rdws = 0.0;	 /* FOLZ and SOLZ */

  total_num_data = numz_data = numh_data = 0;

  for(ie=0; ie<num_exp_dataset; ie++){/* each set of experimental data */
    ng = fed[ie].num_rg;
    nt = fed[ie].num_thickness;
    for(it=0; it<nt; it++){
      for(ig=0; ig<ng; ig++){
        nk = fei[ie].num_data_points[ig];
        total_num_data += nk;

        nlz = fed[ie].incidence.u * fei[ie].obs_rg[ig].h
          + fed[ie].incidence.v * fei[ie].obs_rg[ig].k
            + fed[ie].incidence.w * fei[ie].obs_rg[ig].l;/* Laue zone */
        if( nlz == 0 ){/* ZOLZ */
          numz_data += nk;
        }
        else if (nlz > 0) { /* HOLZ */
          numh_data += nk;
        }
        else{
          mb_fprintf(stdout, "nlz < 0. Abort.\n");
          exit(EXIT_FAILURE);
        }

        g = fed[ie].g[ig];
        aw = fed[ie].art_weight[ig];

        for(ik=0; ik<nk; ik++){
          obs = fei[ie].obs_intensity[ig][it][ik];
          cal = fei[ie].cal_intensity[ig][it][ik];
          sig = fei[ie].obs_sig[ig][it][ik];
          dy = obs - cal;

          ru += fabs(dy);
          rd += fabs(obs);
          weight = cal_weight( sig, obs, cal, g, aw );
          ruw += sq(dy) * weight;
          rdw += sq(obs) * weight;

          if( nlz == 0 ){/* ZOLZ */
            ruz += fabs(dy);
            rdz += fabs(obs);
            ruwz += sq(dy) * weight;
            rdwz += sq(obs) * weight;
          }
          else if (nlz > 0) { /* HOLZ */
            ruh += fabs(dy);
            rdh += fabs(obs);
            ruwh += sq(dy) * weight;
            rdwh += sq(obs) * weight;
            if (nlz == 1) {/* FOLZ */
              ruwf += sq(dy) * weight;
              rdwf += sq(obs) * weight;
            }
            else if (nlz == 2) {/* SOLZ */
              ruws += sq(dy) * weight;
              rdws += sq(obs) * weight;
            }
          }
        }
      }
    }
  }
  chisq = ruw;
  chisq_holz = ruwh;
  chisq_zolz = ruwz;

  mb_fprintf( fo, "Goodness of fit indices:\n");
  mb_fprintf(
    fo, " Chisq = %lf, Chisq_ZOLZ = %lf, Chisq_HOLZ = %lf\n",
    chisq, chisq_zolz, chisq_holz
    );
  if(chisq_holz > 0){
    mb_fprintf(
      fo, "  Ratio chisq_zolz / chisq_holz = %lf\n",
      chisq_zolz / chisq_holz
      );
  }

  if(rd > 0){
    mb_fprintf(fo, " R = %lf", ru / rd);
  }
  if(rdz > 0){
    mb_fprintf(fo, "  R_ZOLZ = %lf", ruz / rdz);
  }
  if(rdh > 0){
    mb_fprintf(fo, "  R_HOLZ = %lf", ruh / rdh);
  }
  mb_fprintf(fo, "\n");

  if(rdw > 0){
    mb_fprintf(fo, " Rw = %lf", sqrt(ruw / rdw) );
  }
  if(rdwz > 0){
    mb_fprintf(fo, "  Rw_ZOLZ = %lf", sqrt(ruwz / rdwz) );
  }
  if(rdwh > 0){
    mb_fprintf(fo, "  Rw_HOLZ = %lf", sqrt(ruwh / rdwh) );
  }
  mb_fprintf(fo, "\n");

/*
  if(rdwf > 0){
    mb_printf( fo, "Rw_FOLZ = %lf", sqrt(ruwf / rdwf) );
  }
  if(rdws > 0){
    mb_fprintf( fo, "  Rw_SOLZ = %lf", sqrt(ruws / rdws) );
  }
  mb_fprintf( fo, "\n");
 */

  if(total_num_data > 0){
    gof = sqrt( chisq / (double) total_num_data );
    mb_fprintf( fo, " GOF = %lf", gof);
  }
  if(numz_data > 0){
    gof_zolz = sqrt( chisq_zolz / (double) numz_data );
    mb_fprintf( fo, "  GOF_ZOLZ = %lf", gof_zolz);
  }
  if(numh_data > 0){
    gof_holz = sqrt( chisq_holz / (double) numh_data );
    mb_fprintf( fo, "  GOF_HOLZ = %lf", gof_holz);
  }
  mb_fprintf( fo, "\n");
}


void fprintf_fitpar( FIT_CTRL fc, double value, FILE *fp_log )
{
  if (fc.flg == 2){/* fitting parameter */
    mb_fprintf(
      fp_log, " [%d] %s= %lf, esd= %lf\n",
      fc.ntp, fc.name, value, fc.esd
      );
  }
}


void par_output_short(
  CRYSTAL_STRCT_DATA *csd, int num_exp_dataset, FIT_EXP_DATASET *fed,
  FILE *fp_log
  )
{
  /* Parameter output in a short format */
  int  i, ia, it, nt, ig, ng, ir, nr, j;
  /* int		ik, nk */
  HKL  g;

  mb_fprintf(fp_log, "Fitting parameters:\n");

  fprintf_fitpar( csd->lp_f.a, csd->lp.a, fp_log );
  fprintf_fitpar( csd->lp_f.b, csd->lp.b, fp_log );
  fprintf_fitpar( csd->lp_f.c, csd->lp.c, fp_log );
  fprintf_fitpar( csd->lp_f.alpha, csd->lp.alpha, fp_log );
  fprintf_fitpar( csd->lp_f.beta,  csd->lp.beta, fp_log  );
  fprintf_fitpar( csd->lp_f.gamma, csd->lp.gamma, fp_log );

  for(i=0; i<csd->n_atom; i++){
    fprintf_fitpar( csd->atm_f[i].x, csd->atm[i].x, fp_log );
    fprintf_fitpar( csd->atm_f[i].y, csd->atm[i].y, fp_log );
    fprintf_fitpar( csd->atm_f[i].z, csd->atm[i].z, fp_log );
    if(csd->atm[i].bf == 0){/* Isotropic B */
      fprintf_fitpar( csd->atm_f[i].b, csd->atm[i].b, fp_log );
    }
    else if ( csd->atm[i].bf == 1 ){/* Anisotropic B */
      fprintf_fitpar( csd->atm_f[i].b11, csd->atm[i].b11, fp_log );
      fprintf_fitpar( csd->atm_f[i].b22, csd->atm[i].b22, fp_log );
      fprintf_fitpar( csd->atm_f[i].b33, csd->atm[i].b33, fp_log );
      fprintf_fitpar( csd->atm_f[i].b12, csd->atm[i].b12, fp_log );
      fprintf_fitpar( csd->atm_f[i].b13, csd->atm[i].b13, fp_log );
      fprintf_fitpar( csd->atm_f[i].b23, csd->atm[i].b23, fp_log );
    }
    else {
      mb_fprintf(stdout, "Invalid b_flag.\n");
      exit(EXIT_FAILURE);
    }
    fprintf_fitpar( csd->atm_f[i].occ, csd->atm[i].occ, fp_log );
  }

  if(csd->num_lofg == 0){
    ;
  }
  else if(csd->lof_type == 1){/* Fg.re, .im */
    for(i=0; i<csd->num_lofg; i++){
      fprintf_fitpar( csd->fg_lofg_f[i].re, csd->fg_lofg[i].re, fp_log);
      fprintf_fitpar( csd->fg_lofg_f[i].im, csd->fg_lofg[i].im, fp_log);
    }
  }
  else if(csd->lof_type == 2){/* Fg.re, .im */
    for(i=0; i<csd->num_lofg; i++){
      fprintf_fitpar(
        csd->fg_lofg_f[i].ela_re, csd->fg_lofg[i].ela_re, fp_log
        );
      fprintf_fitpar(
        csd->fg_lofg_f[i].ela_im, csd->fg_lofg[i].ela_im, fp_log
        );
    }
  }
  else if (csd->lof_type == 3){ /* f */
    for(i=0; i<csd->num_lofg; i++){
      for(ia=0; ia<csd->n_atom; ia++){
        fprintf_fitpar(
          csd->fg_lofg_f[i].fe[ia], csd->fg_lofg[i].fe[ia], fp_log
          );
        fprintf_fitpar(
          csd->fg_lofg_f[i].fi[ia], csd->fg_lofg[i].fi[ia], fp_log
          );
      }
    }
  }
  else if ( csd->lof_type == 4){ /* f */
    for(i=0; i<csd->num_lofg; i++){
      for(ia=0; ia<csd->n_atom; ia++){
        fprintf_fitpar(
          csd->fg_lofg_f[i].fe[ia], csd->fg_lofg[i].fe[ia], fp_log
          );
      }
    }
  }
  else{
    mb_fprintf(stdout, "$$ invalid lof_type (=%d).\n", csd->lof_type);
  }

  for(i=0; i<num_exp_dataset; i++){
    fprintf_fitpar( fed[i].H_tilt_f, fed[i].H_tilt, fp_log );
    fprintf_fitpar( fed[i].G_tilt_f, fed[i].G_tilt, fp_log );
    fprintf_fitpar( fed[i].accel_kvol_f, fed[i].accel_kvol, fp_log );
    nt = fed[i].num_thickness;
    for(it=0; it<nt; it++){
      fprintf_fitpar(
        fed[i].scale_factor_f[it], fed[i].scale_factor[it], fp_log
        );
      fprintf_fitpar(
        fed[i].thickness_f[it], fed[i].thickness[it], fp_log
        );
      fprintf_fitpar(
        fed[i].thickwidth_f[it], fed[i].thickwidth[it], fp_log
        );
      ng = fed[i].num_rg;
      for(ig=0; ig<ng; ig++){
        fprintf_fitpar(
          fed[i].H_shift_f[ig][it], fed[i].H_shift[ig][it], fp_log
          );
        fprintf_fitpar(
          fed[i].G_shift_f[ig][it], fed[i].G_shift[ig][it], fp_log
          );
      }
    }
  }
  mb_fflush(fp_log);
}


void fprintf_cstr(FIT_CTRL fc, double value, char *comment, FILE *fo)
{
  if (fc.flg == 0){/* fix */
    mb_fprintf( fo, "%lf  FIX  %s  %%%s\n", value, fc.name, comment );
  }
  else if( fc.flg == 2){ /* fitting */
    mb_fprintf(
      fo, "%lf  ADJ  %s  %%%s:esd=%lf\n", value, fc.name, comment, fc.esd
      );
  }
  else if (fc.flg == -1){/* constraint */
    mb_fprintf(fo, "%s  CST  %s  %%%s\n", fc.cstr, fc.name, comment);
  }
  else{
    mb_fprintf(stdout, "$$ (fprintf_cstr) invalid flg=%d. Abort.\n", fc.flg);
    exit(EXIT_FAILURE);
  }
}


void output_result_parfile(
  char *out_parfile, CRYSTAL_STRCT_DATA *csd,
  int maxnum_iter, int num_exp_dataset, 
  FIT_EXP_DATASET *fed, FIT_EXP_INTSET *fei
  )
{
  /*
   * Output mbfit-format parameter file
   */
  int  i, ia, it, nt, ig, ng, ir, nr, j;
  /* int  ik, nk */
  HKL  g;
  FILE  *fo;

  fo = mb_fopen_chk_wmode(out_parfile);

  mb_fprintf(fo, "%% %s\n", out_parfile);
  mb_fprintf(fo, "%% data format for mbfit ver 3.*\n");
  mb_fprintf(fo, "%% ==================================================\n");
  mb_fprintf(fo, "%% ==================================================\n");
  mb_fprintf(fo, "%% job type\n");
  mb_fprintf(fo, "%%  FIT_MODE, SIM_MODE, BLOCH_MODE, DISPSURF_MODE,\n");
  mb_fprintf(fo, "%%  T_SCAN_MODE, SHIFT_SCAN_MODE\n");
  mb_fprintf(fo, "FIT_MODE\n");
  mb_fprintf(fo, "%% ==================================================\n");
  mb_fprintf(fo, "%% ==================================================\n");

  /*  fprintf_eps_ctrl_params(fo);*/
  mb_fprintf(fo, "EPS_CONTROL_PARAMETERS{\n");
  mb_fprintf(fo, "intensity_scaling  %lf", epsilon.intensity_scaling);
  mb_fprintf(fo, "  %%Scaling incident beam intensity\n");
  mb_fprintf(fo, "vg_select_flag  %d", epsilon.vg_select_flag);
  mb_fprintf(fo, "  %%vg selection for each tilt\n");
  mb_fprintf(fo, "svd_flag  %d", epsilon.svd_flag);
  mb_fprintf(fo, "  %%1:svd, 0:gaussj\n");
  mb_fprintf(fo, "no_fitting_flag  %d", epsilon.no_fitting_flag);
  mb_fprintf(fo, "  %%0: fitting, 1: no fitting\n");
  mb_fprintf(fo, "num_thickwidth  %d", epsilon.num_thickwidth);
  mb_fprintf(fo, "  %%num_thickwidth * 2 + 1 --> calc. points\n");
  mb_fprintf(fo, "thickwidth_type  %d", epsilon.thickwidth_type);
  mb_fprintf(fo, "  %%0:rectangle, 1:triangle, 2:Gaussian\n");
  mb_fprintf(fo, "auto_multiplicity_off  %d", epsilon.auto_multiplicity_off);
  mb_fprintf(fo, "  %%0:normal 1:manual (for the split-atom method)\n");
  mb_fprintf(fo, "svd_flag  %d", epsilon.svd_flag);
  mb_fprintf(fo, "  %%0:gauss-jordan, 1:SVD\n");
  mb_fprintf(fo, "}EPS_CONTROL_PARAMETERS\n");
  mb_fprintf(fo, "%% ==================================================\n");
  mb_fprintf(fo, "%% ==================================================\n");

  mb_fprintf(fo, "%% fitting parameter:\n");
  mb_fprintf(fo, "%%  value (or constraint), fitting flag, parameter's name\n");
  mb_fprintf(fo, "%%  - constraint:\n");
  mb_fprintf(fo, "%%  - fitting flag:   FIX:fix, ADJ:fit, CST:constrained\n");
  mb_fprintf(fo, "%%  - parameter's name: \n");
  mb_fprintf(fo, "%%\n");

  /* Crystal strucural parameters */
  mb_fprintf(fo, "CRYSTAL_STRUCTURE_DATA{\n");
  mb_fprintf(fo, "%%\n");
  mb_fprintf(fo, "LATTICE{\n");
  mb_fprintf(fo, "%d    %% space group number\n", csd->space_group);
  mb_fprintf(fo, "%% lattice parameter\n");
  fprintf_cstr( csd->lp_f.a, csd->lp.a, "a[A]", fo );
  fprintf_cstr( csd->lp_f.b, csd->lp.b, "b[A]", fo );
  fprintf_cstr( csd->lp_f.c, csd->lp.c, "c[A]", fo );
  fprintf_cstr( csd->lp_f.alpha, csd->lp.alpha, "alpha[deg.]", fo );
  fprintf_cstr( csd->lp_f.beta,  csd->lp.beta,  "beta[deg.]",  fo );
  fprintf_cstr( csd->lp_f.gamma, csd->lp.gamma, "gamma[deg.]", fo );
  mb_fprintf(fo, "}LATTICE\n");
  mb_fprintf(fo, "%%\n");

  mb_fprintf(fo, "ATOM{\n");
  mb_fprintf(fo, "%d    %% number of independent sites\n", csd->n_atom);
  for(i=0; i<csd->n_atom; i++){
    mb_fprintf(fo, "SITE{\n");
    mb_fprintf(fo, "%s\n", csd->atm[i].element );
    fprintf_cstr( csd->atm_f[i].x, csd->atm[i].x, "x (frac. coord.)", fo );
    fprintf_cstr( csd->atm_f[i].y, csd->atm[i].y, "y (frac. coord.)", fo );
    fprintf_cstr( csd->atm_f[i].z, csd->atm[i].z, "z (frac. coord.)", fo );
    mb_fprintf(fo, "%% B-factor type (type=0:iso, 1:aniso)\n" );
    mb_fprintf(
      fo, "%d    %% B type (0: isotropic, 1:anisotropic)\n", csd->atm[i].bf
      );
    if(csd->atm[i].bf == 0){  /* Isotropic B */
      fprintf_cstr( csd->atm_f[i].b, csd->atm[i].b, "B [A^{2}]", fo );
    }
    else if ( csd->atm[i].bf == 1 ){  /* Anisotropic B */
      mb_fprintf(fo, "%%  anisotropic beta_ij\n");
      fprintf_cstr( csd->atm_f[i].b11, csd->atm[i].b11, "beta11", fo );
      fprintf_cstr( csd->atm_f[i].b22, csd->atm[i].b22, "beta22", fo );
      fprintf_cstr( csd->atm_f[i].b33, csd->atm[i].b33, "beta33", fo );
      fprintf_cstr( csd->atm_f[i].b12, csd->atm[i].b12, "beta12", fo );
      fprintf_cstr( csd->atm_f[i].b13, csd->atm[i].b13, "beta13", fo );
      fprintf_cstr( csd->atm_f[i].b23, csd->atm[i].b23, "beta23", fo );
    }
    else {
      mb_fprintf(stdout, "$$ Invalid b_flag. Abort.\n");
      exit(EXIT_FAILURE);
    }
    fprintf_cstr( csd->atm_f[i].occ, csd->atm[i].occ, "occupancy (0-1)", fo );
    mb_fprintf(fo, "%%\n");
    mb_fprintf(fo, "%d     %% f table (0:Mott+BK, 1:B&K)\n", csd->atm[i].af);
    mb_fprintf(fo, "}SITE\n");
  }
  mb_fprintf(fo, "}ATOM\n");

  mb_fprintf(fo, "%%\n");
  mb_fprintf(fo, "%% --------------------------------------------------\n");
  mb_fprintf(fo, "%% Low-order Fg for electrons\n");
  mb_fprintf(fo, "%%  - Angstrom unit without relativistic correction\n");
  mb_fprintf(fo, "%%  - LOF type\n");
  mb_fprintf(fo, "%%     1: total (ela + inela) Fg.re, Fg.im)\n");
  mb_fprintf(fo, "%%     2: elastic Fg.re and Fg.im)\n");
  mb_fprintf(fo, "%%  - The values for symmetrically-equivalent \n");
  mb_fprintf(fo, "%%     reflections are automatically generated.\n");
  mb_fprintf(fo, "%% --------------------------------------------------\n");
  mb_fprintf(fo, "LOW_ORDER_FG{\n");
  mb_fprintf(fo, "%d  %% Number of low-order g\n", csd->num_lofg);
  if(csd->num_lofg == 0){
    ;
  }
  else if(csd->lof_type == 1){  /* Fg.re, .im */
    mb_fprintf(fo, "%d    %%lof_type\n", csd->lof_type);
    for(i=0; i<csd->num_lofg; i++){
      mb_fprintf(fo, "%% low-order g\n" );
      mb_fprintf(fo, "%% h k l	\n" );
      mb_fprintf(
        fo, "%d\t%d\t%d\n",
        csd->g_lofg[i].h, csd->g_lofg[i].k, csd->g_lofg[i].l
        );
      mb_fprintf(fo, "%% Fg.re	[A]\n" );
      fprintf_cstr( csd->fg_lofg_f[i].re, csd->fg_lofg[i].re, "Fg.re", fo );
      mb_fprintf(fo, "%% Fg.im	[A]\n" );
      fprintf_cstr( csd->fg_lofg_f[i].im, csd->fg_lofg[i].im, "Fg.im", fo );
    }
  }
  else if(csd->lof_type == 2){  /* Fg.re, .im */
    mb_fprintf(fo, "%% lof_type\n");
    mb_fprintf(fo, "%%  2: Fg.re, Fg.im: elastic part   (inelastic --> BK) \n");
    mb_fprintf(fo, "%d\n", csd->lof_type);
    for(i=0; i<csd->num_lofg; i++){
      mb_fprintf(fo, "%% low-order g\n" );
      mb_fprintf(fo, "%% h k l	\n" );
      mb_fprintf(
        fo, "%d\t%d\t%d\n",
        csd->g_lofg[i].h, csd->g_lofg[i].k, csd->g_lofg[i].l
        );
      mb_fprintf(fo, "%% Fg.re	[A]\n" );
      fprintf_cstr(
        csd->fg_lofg_f[i].ela_re, csd->fg_lofg[i].ela_re, "Fg_ela.re", fo );
      mb_fprintf(fo, "%% Fg.im	[A]\n" );
      fprintf_cstr(
        csd->fg_lofg_f[i].ela_im, csd->fg_lofg[i].ela_im, "Fg_ela.im", fo );
    }
  }
  else if (csd->lof_type == 3){/* f */
    mb_fprintf(fo, "%% lof_type\n");
    mb_fprintf(fo, "%%  3: f : elastic ang inelastic \n");
    mb_fprintf(fo, "%d\n", csd->lof_type);
    for(i=0; i<csd->num_lofg; i++){
      mb_fprintf(fo, "%% low-order g\n" );
      mb_fprintf(fo, "%% h k l	\n" );
      mb_fprintf(
        fo, "%d\t%d\t%d\n",
        csd->g_lofg[i].h, csd->g_lofg[i].k, csd->g_lofg[i].l
        );
      for(ia=0; ia<csd->n_atom; ia++){
        mb_fprintf(fo, "%% fe	[A]\n" );
        fprintf_cstr(
          csd->fg_lofg_f[i].fe[ia], csd->fg_lofg[i].fe[ia], "f_ela", fo
          );
        mb_fprintf(fo, "%% fi	[A]\n" );
        fprintf_cstr(
          csd->fg_lofg_f[i].fi[ia], csd->fg_lofg[i].fi[ia], "f_inela", fo
          );
      }
    }
  }
  else if ( csd->lof_type == 4){/* f */
    mb_fprintf(fo, "%% lof_type\n");
    mb_fprintf(fo, "%%  4: f : elastic part (inelastic --> BK) \n");
    mb_fprintf(fo, "%d\n", csd->lof_type);
    for(i=0; i<csd->num_lofg; i++){
      mb_fprintf(fo, "%% low-order g\n" );
      mb_fprintf(fo, "%% h k l	\n" );
      mb_fprintf(
        fo, "%d\t%d\t%d\n",	csd->g_lofg[i].h, csd->g_lofg[i].k, csd->g_lofg[i].l
        );
      for(ia=0; ia<csd->n_atom; ia++){
        mb_fprintf(fo, "%% fe	[A]\n" );
        fprintf_cstr(
          csd->fg_lofg_f[i].fe[ia], csd->fg_lofg[i].fe[ia], "f_ela", fo
          );
      }
    }
  }
  else{
    mb_fprintf(stdout, "$$ invalid lof_type (=%d). Abort.\n", csd->lof_type);
    exit(EXIT_FAILURE);
  }
  mb_fprintf(fo, "}LOW_ORDER_FG\n");
  mb_fprintf(fo, "%%\n");
  mb_fprintf(fo, "}CRYSTAL_STRUCTURE_DATA\n");
  mb_fprintf(fo, "%% ==================================================\n");
  mb_fprintf(fo, "%% ==================================================\n");

  /* Calc. parameters */
  mb_fprintf(fo, "FIT_CALCULATION_CONDITION{\n");

  mb_fprintf(fo, "ITER_EXP_NUM{\n");
  mb_fprintf(fo, "%d    %%Max. number of iteration\n", maxnum_iter);
  mb_fprintf(fo, "%d    %%Number of experimental data sets\n", num_exp_dataset);
  mb_fprintf(fo, "}ITER_EXP_NUM\n");
  mb_fprintf(fo, "%%\n");

  for(i=0; i<num_exp_dataset; i++){
    mb_fprintf(fo, "DATA_SET{\n");
    mb_fprintf(fo, "%% Data set no. %d\n", i);

    mb_fprintf(fo, "INCIDENT_BEAM{\n");
    mb_fprintf(
      fo, "%d  %d  %d  %%Incidence	(upward)\n",
      fed[i].incidence.u, fed[i].incidence.v, fed[i].incidence.w
      );
    mb_fprintf(
      fo, "%d  %d  %d  %%Surface normal (upward)\n",
      fed[i].surf_norm.u, fed[i].surf_norm.v, fed[i].surf_norm.w
      ); 
    mb_fprintf(
      fo, "%lf  %lf  %lf  %%ZOLZ base vector\n",
      fed[i].zolzbase1.h, fed[i].zolzbase1.k, fed[i].zolzbase1.l
      );
    fprintf_cstr( fed[i].H_tilt_f, fed[i].H_tilt, "H_tilt", fo );
    fprintf_cstr( fed[i].G_tilt_f, fed[i].G_tilt, "G_tilt", fo );
    fprintf_cstr(
      fed[i].accel_kvol_f, fed[i].accel_kvol, "Accelerating Voltage [kV]", fo
      );
    mb_fprintf(fo, "}INCIDENT_BEAM\n");

    mb_fprintf(fo, "%% ----- Beam search -------------------------\n");
    mb_fprintf(fo, "%% search method\n");
    mb_fprintf(fo, "%%   0: manual\n");
    mb_fprintf(fo, "%%   1,2,3: automatic + generalized Bethe approx.\n");
    mb_fprintf(fo, "%%          [1:wg, 2:sg, 3:sg-multi] \n");
    mb_fprintf(fo, "%%   11,12,13: automatic + simple Bethe approx.\n");
    mb_fprintf(fo, "%%             [11:wg, 12:sg, 13:sg-multi] \n");
    mb_fprintf(fo, "BEAM_SEARCH{\n");
    mb_fprintf(fo, "%d  %%search method\n", fed[i].beam_search.flag_bs);
    if ( fed[i].beam_search.flag_bs == 0 ){
      mb_fprintf(fo, "%%Manual type\n");
      mb_fprintf(
        fo, "%d  %d    %%num_exact_beam num_bethe_beam\n",
        fed[i].num_exact, fed[i].num_bethe
        );
      mb_fprintf(fo, "%% hkl\n");
      nr = fed[i].num_exact + fed[i].num_bethe;
      for(ir=0; ir<nr; ir++){
        mb_fprintf(
          fo,"%d  %d  %d\n",
          fed[i].reflex[ir].h, fed[i].reflex[ir].k, fed[i].reflex[ir].l
          );
      }
    }
    else if (
      fed[i].beam_search.flag_bs == 1 ||
      fed[i].beam_search.flag_bs == 11
      ){ /* wg */
      mb_fprintf(fo, "%%Auto type [wg]\n");
      mb_fprintf(
        fo,"%% num_g_limit  g_max[A^{-1}]  lz_min  lz_max  disksize\n"
        );
      mb_fprintf(
        fo,"%d  %lf  %d  %d  %lf\n",
        fed[i].beam_search.nglim, 
        fed[i].beam_search.gmax, fed[i].beam_search.lzmin,
        fed[i].beam_search.lzmax, fed[i].beam_search.disksize
        );
      mb_fprintf(fo,"%% w_g_bethe_min,  w_g_max,  xi_g_max\n");
      mb_fprintf(
        fo,"%lf  %lf  %lf\n", fed[i].beam_search.wgbmin, 
        fed[i].beam_search.wgmax, fed[i].beam_search.xigmax
        );
    }
    else if (
      fed[i].beam_search.flag_bs == 2 ||
      fed[i].beam_search.flag_bs == 12
      ){  /* sg */
      mb_fprintf(fo, "%%Auto type [sg]\n");
      mb_fprintf(
        fo, "%% num_g_limit  g_max[A^{-1}]  lz_min  lz_max  disksize\n"
        );
      mb_fprintf(
        fo,"%d  %lf  %d  %d  %lf\n",
        fed[i].beam_search.nglim, 
        fed[i].beam_search.gmax, fed[i].beam_search.lzmin,
        fed[i].beam_search.lzmax, fed[i].beam_search.disksize
        );
      mb_fprintf(fo,"%% s_g_bethe_min,  s_g_max,  xi_g_max\n");
      mb_fprintf(
        fo,"%lf  %lf  %lf\n", fed[i].beam_search.sgbmin, 
        fed[i].beam_search.sgmax, fed[i].beam_search.xigmax
        );
    }
    else if (
      fed[i].beam_search.flag_bs == 3 ||
      fed[i].beam_search.flag_bs == 13
      ){/* sg-multi */
      mb_fprintf(fo, "%%Auto type [sg-multi]\n");
      mb_fprintf(
        fo, "%% num_g_limit  g_max[A^{-1}]  lz_min  lz_max  disksize\n"
        );
      mb_fprintf(
        fo,"%d  %lf  %d  %d  %lf\n",
        fed[i].beam_search.nglim, 
        fed[i].beam_search.gmax, fed[i].beam_search.lzmin,
        fed[i].beam_search.lzmax, fed[i].beam_search.disksize
        );
      mb_fprintf(fo,"%% s_g_bethe_min,  s_g_max,  xi_g_max\n");
      for(j=0; j<=fed[i].beam_search.lzmax; j++){
        mb_fprintf(
          fo, "%lf  %lf  %lf  %%d-LZ:\n", fed[i].beam_search.sge_max[j], 
          fed[i].beam_search.sgb_max[j], fed[i].beam_search.xig_max[j], j
          );
      }
    }
    mb_fprintf(fo,"}BEAM_SEARCH\n");

    mb_fprintf(fo, "%%\n");
    mb_fprintf(fo, "CALC_KXY_RANGE{\n");
    mb_fprintf(
      fo, "%d  %%cal_range_flag  [0:manual, 1:auto]\n", fed[i].calc_range_flag
      );  /* $$ 031003 */
    mb_fprintf(
      fo, "%d  %%0: 0D, 1: 1D, 2: 2D disk, 3: 2D square\n", fed[i].calc_type
      );
    mb_fprintf(
      fo, "%d  %%numh: number of calc. points > 5 (total~sq(numh))\n",
      fed[i].calc_numh
      );
    mb_fprintf(
      fo, "%lf  %lf  %%hbgn, hend\n", fed[i].calc_hbgn, fed[i].calc_hend
      );
    mb_fprintf(
      fo, "%lf  %lf  %%gbgn, gend\n", fed[i].calc_gbgn, fed[i].calc_gend
      );
    mb_fprintf(fo, "}CALC_KXY_RANGE\n");

    mb_fprintf(fo, "%%\n");
    mb_fprintf(fo, "INTENSITY_DATA{\n");
    mb_fprintf(fo, "%% Intensity file\n");
    mb_fprintf(fo, "%s\n", fed[i].intensity_file);
    mb_fprintf(fo, "%d    %% Number of thicknesses\n", fed[i].num_thickness);
    mb_fprintf(fo, "%d    %% Number of reflections\n", fed[i].num_rg);
    mb_fprintf(fo, "}INTENSITY_DATA\n");

    nt = fed[i].num_thickness;
    for(it=0; it<nt; it++){
      mb_fprintf(fo, "THICKNESS_BLOCK{\n");
      mb_fprintf(fo, "%% Thickness No. %d \n", it);

      mb_fprintf(fo, "STW_PAR{\n");
      fprintf_cstr(
        fed[i].scale_factor_f[it], fed[i].scale_factor[it], "scale_factor", fo
        );
      fprintf_cstr(
        fed[i].thickness_f[it], fed[i].thickness[it], "thickness", fo
        );
      fprintf_cstr(
        fed[i].thickwidth_f[it], fed[i].thickwidth[it],
        "thickness_width[plus-minus]", fo
        );
      mb_fprintf(
        fo,
        "%% num_thickwidth  0  %% num_thickwidth * 2 + 1 --> calc. points\n"
        );
      mb_fprintf(
        fo,"%% thickwidth_type  1  %% 0:rectangle, 1:triangle, 2:Gaussian\n"
        );
      mb_fprintf(fo,"%%\n");
      mb_fprintf(fo,"%% Gaussian convolution:\n");
      mb_fprintf(
        fo, "%%   number of pixels (odd), Gaussian par. (a: exp(-x^2/a^2) )\n"
        );
      mb_fprintf(fo, "%%   if num.=0 then no convolution\n");
      mb_fprintf(
        fo, "%d  %lf\n", fed[i].num_avrpix[it], fed[it].gaussian_par[it]
        );
      mb_fprintf(fo, "}STW_PAR\n");

      ng = fed[i].num_rg;
      mb_fprintf(fo, "REFLEX_SHIFT_PAR{\n");
      for(ig=0; ig<ng; ig++){
        mb_fprintf(fo, "%% hkl no. %d\n", ig);
        g = fei[i].obs_rg[ig];
/*        if( fed[i].art_weight[ig] == 1.0 ){
          mb_fprintf(fo, "%d  %d  %d  %%h k l\n", g.h, g.k, g.l);
        }
        else{*/
        mb_fprintf(
          fo, "%d  %d  %d  %lf  %%h k l weight\n",
          g.h, g.k, g.l, fed[i].art_weight[ig]
          );
/*        }*/
        fprintf_cstr(
          fed[i].H_shift_f[ig][it], fed[i].H_shift[ig][it], "H_shift", fo );
        fprintf_cstr(
          fed[i].G_shift_f[ig][it], fed[i].G_shift[ig][it], "G_shift", fo );
      }
      mb_fprintf(fo, "}REFLEX_SHIFT_PAR\n");
      mb_fprintf(fo,"}THICKNESS_BLOCK\n");
    }
    mb_fprintf(fo, "}DATA_SET\n");
  }

  mb_fprintf(fo, "}FIT_CALCULATION_CONDITION\n");
  mb_fclose(fo);
}
