/*
 * mb_mpi.c
 */
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include "mbconfig.h"
#include "mb_mpi.h"

void mb_fprintf(FILE *fp, const char *fmt, ...)
{
#if defined(MPI_MODE)
  if(proc_id == 0){
    va_list ap;
    va_start(ap, fmt);
    vfprintf(fp, fmt, ap);
    va_end(ap);
  }
#else
  va_list ap;
  va_start(ap, fmt);
  vfprintf(fp, fmt, ap);
  va_end(ap);
#endif
}


/*
void mb_2fprintf(FILE *fp, const char *fmt, ...)
{
#if defined(MPI_MODE)
  if(proc_id == 0){
    va_list ap;
    va_start(ap, fmt);
    vfprintf(fp, fmt, ap);
    vfprintf(stdout, fmt, ap);
    va_end(ap);
  }
#else
  va_list ap;
  va_start(ap, fmt);
  vfprintf(fp, fmt, ap);
  vfprintf(stdout, fmt, ap);
  va_end(ap);
#endif
}
*/


void mb_fflush(FILE *fp)
{
#if defined(MPI_MODE)
  if(proc_id == 0){  
    fflush(fp);
  }
#else
  fflush(fp);
#endif
}


void progress_indicator(long ith, long num_iter)
{
  /* The value of ith should be between 0 and num_iter-1. */
  int  prog_ratio;
  static int five_counter;

  if (ith == 0) {  /* Initialize */
    five_counter = 0;
    /* mb_fprintf( stdout, "%d%%", five_counter );  mb_fflush(stdout); */
    five_counter += 5;
  }

  prog_ratio = (int) (100.0 * (double) ith / (double) num_iter );
  if ( prog_ratio > five_counter ){
    mb_fprintf( stdout, "%d%%", five_counter );  mb_fflush(stdout);
    five_counter += 5;
    while(five_counter <= prog_ratio){
      mb_fprintf( stdout, "%d%%", five_counter );
      mb_fflush(stdout);
      five_counter += 5;
    }
  }

  if ( ith == num_iter-1 ){
    while(five_counter <= 95){
      mb_fprintf( stdout, "%d%%", five_counter );
      mb_fflush(stdout);
      five_counter += 5;
    }
  }
}


FILE *mb_fopen_chk_wmode(char *filename)
{
  FILE *fp;
#if defined(MPI_MODE)
  if(proc_id == 0){  
    if( (fp = fopen(filename, "w")) == NULL ){
      /* fprintf(stderr, "$$ Cannot open %s. Abort.\n", filename); */
      perror("$$ Cannot open. Abort.\n");
      exit(EXIT_FAILURE);
    }
    return fp;
  }
  else{
    return NULL;
  }
#else
  if( (fp = fopen(filename, "w")) == NULL ){
    /* fprintf(stderr, "$$ Cannot open %s. Abort.\n", filename); */
    perror("$$ Cannot open. Abort.\n");
    exit(EXIT_FAILURE);
  }
  return fp;
#endif
}


void mb_fclose(FILE *fp)
{
#if defined(MPI_MODE)
  if(proc_id == 0){  
    fclose(fp);
  }
#else
  fclose(fp);
#endif
}
