/*
  mb_sim.c
  Many-Beam Dynamical Calculation
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "mbconfig.h"
#include "fitting.h"
#include "mb_input.h"
#include "data_out.h"
#include "mb_sim.h"
#include "mb_mpi.h"
#include "tiff.h"
#define BUFLEN 256
#define GETLEN 128

extern DEBUG_INFO dbg;
extern EPS_CTRL	epsilon;

// -----------------------------
// Round real to nearest integer
// -----------------------------
long near(float x) ;
long near(float x) {
	float y ;
	long i ;

	y = x+.5 ;
	i = y ;
	if (y < 0.) i-- ;
	return(i) ;
}

void findmaxmin2(int nx, int ny, double **x, double *xmax, double *xmin);
/*void bmp_out(
  char *output_filename, double **val, int nx, int ny, int fmag
  ); */
/*void ass_out(char *outfil, double **mat, int nx, int ny);*/
/*void output_pgm_image(
  int image_width, int image_height, double **double_image,
  double maxval, double minval, double gamma_coef, int mag_coef,
  int contrast_reverse_flag, char *fn_out
  );*/


/* SIM_MODE ------------------------------------------------------------- */
/* ---------------------------------------------------------------------- */
void fget_sd_beams(SIM_DATA *sd, CRYSTAL_STRCT_DATA *csd, FILE *fi)
{
  /*
   *  Beam search
   */
  HKL *g_exact, *g_bethe;
  int nr, g_table_size;

  sd->beam_search.flag_bs = fget_i(fi);
  if(sd->beam_search.flag_bs == 0 || sd->beam_search.flag_bs == 10){
    /* Manual search */
    /* fget_1s(beamfile, fi); */
    fget_2i(&(sd->num_exact), &(sd->num_bethe), fi);
    mb_fprintf(
      stdout, "Exact = %d, Bethe = %d\n", sd->num_exact, sd->num_bethe
      );
    nr = sd->num_exact + sd->num_bethe;
    sd->reflex = new_vector(HKL, nr);  chk_vector(sd->reflex);
    fget_beams( sd->num_exact, sd->num_bethe, sd->reflex, fi );
    mb_fprintf(stdout, "$$ vg_select_flag is set to 0.\n");
    epsilon.vg_select_flag = 0;
  }
  else if(
    sd->beam_search.flag_bs == 1 || sd->beam_search.flag_bs == 11
    ){ /* Automatic search (wg) */
    fget_beam_search_wg(fi, &(sd->beam_search));

    g_table_size = SIZE_GTABLE;
    g_exact = new_vector(HKL, sd->beam_search.nglim);	chk_vector(g_exact);
    g_bethe = new_vector(HKL, sd->beam_search.nglim);	chk_vector(g_bethe);
    mb_fprintf(stdout, "Beam search (wg)... ");  mb_fflush(stdout);
    simple_beam_search_wg(
      &(sd->num_exact), &(sd->num_bethe), g_exact, g_bethe,
      sd->beam_search, sd->incidence, sd->zolzbase1, sd->H_tilt, sd->G_tilt,
      sd->accel_kvol, csd
      );
    mb_fprintf(
      stdout, "Exact = %d, Bethe = %d\n", sd->num_exact, sd->num_bethe
      );
    nr = sd->num_exact + sd->num_bethe;
    sd->reflex = new_vector(HKL, nr);	chk_vector(sd->reflex);
    beam_merge(sd->num_exact, sd->num_bethe, g_exact, g_bethe, sd->reflex);
    free_vector(g_bethe);
    free_vector(g_exact);
  }
  else if(
    sd->beam_search.flag_bs == 2 || sd->beam_search.flag_bs == 12
    ){ /* Automatic search (sg) */
    fget_beam_search_sg(fi, &(sd->beam_search) );
    
    g_table_size = SIZE_GTABLE;
    g_exact = new_vector(HKL, sd->beam_search.nglim);  chk_vector(g_exact);
    g_bethe = new_vector(HKL, sd->beam_search.nglim);  chk_vector(g_bethe);
    mb_fprintf(stdout, "Beam search (sg)... ");  mb_fflush(stdout);
    simple_beam_search_sg(
      &(sd->num_exact), &(sd->num_bethe), g_exact, g_bethe,
      sd->beam_search, sd->incidence, sd->zolzbase1, sd->H_tilt, sd->G_tilt,
      sd->accel_kvol, csd
      );
    mb_fprintf(
      stdout, "Exact = %d, Bethe = %d\n", sd->num_exact, sd->num_bethe
      );
    nr = sd->num_exact + sd->num_bethe;
    sd->reflex = new_vector(HKL, nr);	chk_vector(sd->reflex);
    beam_merge(sd->num_exact, sd->num_bethe, g_exact, g_bethe, sd->reflex);
    free_vector(g_bethe);
    free_vector(g_exact);
  }
  else if(
    sd->beam_search.flag_bs == 3 || sd->beam_search.flag_bs == 13
    ){ /* Automatic search (sg - multi) */
    fget_beam_search_sg_multi(fi, &(sd->beam_search) );
    
    g_table_size = SIZE_GTABLE;
    g_exact = new_vector(HKL, sd->beam_search.nglim);  chk_vector(g_exact);
    g_bethe = new_vector(HKL, sd->beam_search.nglim);  chk_vector(g_bethe);
    mb_fprintf(stdout, "Beam search (sg)... ");  mb_fflush(stdout);	
    beam_search_sg_multi(
      &(sd->num_exact), &(sd->num_bethe), g_exact, g_bethe,
      sd->beam_search, sd->incidence, sd->zolzbase1, sd->H_tilt, sd->G_tilt,
      sd->accel_kvol, csd
      );
    mb_fprintf(
      stdout, "Exact = %d, Bethe = %d\n", sd->num_exact, sd->num_bethe
      );
    nr = sd->num_exact + sd->num_bethe;
    sd->reflex = new_vector(HKL, nr);	chk_vector(sd->reflex);
    beam_merge(sd->num_exact, sd->num_bethe, g_exact, g_bethe, sd->reflex);
    free_vector(g_bethe);
    free_vector(g_exact);
  }
  else{
    mb_fprintf(stdout, "Invalid flag_bs\n"); exit(EXIT_FAILURE);
  }
}


void free_sd(SIM_DATA *sd)
{
  int ig, it;
  
  if(sd->num_exact + sd->num_bethe > 0) {free_vector(sd->reflex);}
  if(sd->num_g > 0){free_vector(sd->calc_g);}
  if(sd->num_thickness > 0 ){free_vector(sd->thickness);}
  if(
    sd->calc_type == 0 || sd->calc_type == 1 ||
    sd->calc_type == 2 || sd->calc_type == 3 ||
    sd->calc_type == 5 
    ){
    free_vector(sd->h_pos);  free_vector(sd->g_pos);
    free_vector(sd->H);	free_vector(sd->G);
    free_vector(sd->k_xy);
    free_tensor(sd->intensity);
  }
  else if(sd->calc_type < 0){/* E-density */
    ;
  }
}


void calc_kxy_tilt( SIM_DATA *sd, CELL_PAR bv )
{
  /*
    Size of $k_xy[]$ must be less than $(num_points)^{2}$. (calc_type 3)
   */
  int		i, ix, iy, j, la, nm, np, npx, npy, nmx, nmy, ct, nd;
  double	bh, bg, tx, ty, t[4];
  char	buf[BUFLEN];
  HKL		g;
  R_HKL	g_proj, zb1, zb2, tilt;
  FILE	*fp;
  
  sd->zolzbase2 = zolz_ort_base(sd->incidence, sd->zolzbase1, bv);
  /* zolzbase2 is orthogonal to zolzbase1 and has the same norm. */
  
  
  if (sd->calc_type == 0){				/* 0 Dim. ( 1 point ) */
    nd = 1;			/* caution: sq */
    sd->k_xy	= new_vector(R_HKL, nd);	chk_vector(sd->k_xy);
    sd->h_pos	= new_vector(int, nd);		chk_vector(sd->h_pos);
    sd->g_pos	= new_vector(int, nd);		chk_vector(sd->g_pos);
    sd->H		= new_vector(double, nd);	chk_vector(sd->H);
    sd->G		= new_vector(double, nd);	chk_vector(sd->G);
    ct = 0;
    bg = sd->G_begin + sd->G_tilt;
    bh = sd->H_begin + sd->H_tilt;
    sd->k_xy[0].h = 0.5 * ( bh * sd->zolzbase1.h + bg * sd->zolzbase2.h );
    sd->k_xy[0].k = 0.5 * ( bh * sd->zolzbase1.k + bg * sd->zolzbase2.k );
    sd->k_xy[0].l = 0.5 * ( bh * sd->zolzbase1.l + bg * sd->zolzbase2.l );
    sd->h_pos[0] = 0;	sd->g_pos[0] = 0;
    sd->H[0] = bh;		sd->G[0] = bg;
    ct = 1;
  }
  else if (sd->calc_type == 1){			/* 1 Dim. (Line prof.) */
    np = nd = sd->num_points;
    nm = sd->num_points - 1;
    sd->k_xy	= new_vector(R_HKL, nd);	chk_vector(sd->k_xy);
    sd->h_pos	= new_vector(int, nd);		chk_vector(sd->h_pos);
    sd->g_pos	= new_vector(int, nd);		chk_vector(sd->g_pos);
    sd->H		= new_vector(double, nd);	chk_vector(sd->H);
    sd->G		= new_vector(double, nd);	chk_vector(sd->G);
    ct = 0;
    bg = sd->G_tilt + sd->G_begin;
    for(j=0; j<np; j++){
      bh = sd->H_tilt + sd->H_begin + ( sd->H_end - sd->H_begin )
        * (double) j / (double) nm;
      sd->k_xy[ct].h = 0.5 * ( bh * sd->zolzbase1.h + bg * sd->zolzbase2.h );
      sd->k_xy[ct].k = 0.5 * ( bh * sd->zolzbase1.k + bg * sd->zolzbase2.k );
      sd->k_xy[ct].l = 0.5 * ( bh * sd->zolzbase1.l + bg * sd->zolzbase2.l );
      sd->h_pos[ct] = j - np / 2;
      sd->g_pos[ct] = 0;
      sd->H[ct] = bh;		sd->G[ct] = bg;
      ct++;
    }
  }
  else if (sd->calc_type == 2){			/* 2 Dim. (Disk) */
    nd = sq(sd->num_points);			/* caution: sq */
    sd->k_xy	= new_vector(R_HKL, nd);	chk_vector(sd->k_xy);
    sd->h_pos	= new_vector(int, nd);		chk_vector(sd->h_pos);
    sd->g_pos	= new_vector(int, nd);		chk_vector(sd->g_pos);
    sd->H		= new_vector(double, nd);	chk_vector(sd->H);
    sd->G		= new_vector(double, nd);	chk_vector(sd->G);
    np = sd->num_points;
    nm = sd->num_points - 1;
    ct = 0;
    for(i=0; i<np; i++){
      bg = sd->G_tilt + sd->G_begin + ( sd->G_end - sd->G_begin )
        * (double) i / (double) nm;
      la = 2 * (int)( sqrt( sq((double)(np/2) + 0.5) - sq((double)(i-np/2)) ) );
      /* la = 2 * (int)( sqrt( sq((double)(np/2)) - sq((double)(i-np/2)) ) ); */
      for(j=0;j<la;j++){
        bh = sd->H_tilt + sd->H_begin + ( sd->H_end - sd->H_begin ) *
          (double)(np/2 - la/2 + j) / (double) nm;
        sd->k_xy[ct].h = 0.5 * ( bh * sd->zolzbase1.h + bg * sd->zolzbase2.h );
        sd->k_xy[ct].k = 0.5 * ( bh * sd->zolzbase1.k + bg * sd->zolzbase2.k );
        sd->k_xy[ct].l = 0.5 * ( bh * sd->zolzbase1.l + bg * sd->zolzbase2.l );
        sd->h_pos[ct] = j - la / 2;
        sd->g_pos[ct] = i - np / 2;
        sd->H[ct] = bh;		sd->G[ct] = bg;
        ct++;
      }
    }
  }
  else if (sd->calc_type == 3){/* 2 Dim. (lectangle) */
    nd = sd->num_points_x * sd->num_points_y;			/* caution: sq */
    sd->k_xy	= new_vector(R_HKL, nd);	chk_vector(sd->k_xy);
    sd->h_pos	= new_vector(int, nd);		chk_vector(sd->h_pos);
    sd->g_pos	= new_vector(int, nd);		chk_vector(sd->g_pos);
    sd->H		= new_vector(double, nd);	chk_vector(sd->H);
    sd->G		= new_vector(double, nd);	chk_vector(sd->G);
    npx = sd->num_points_x;		npy = sd->num_points_y;
    nmx = sd->num_points_x - 1;	nmy = sd->num_points_y - 1;
    ct = 0;
    for(iy=0; iy<npy; iy++){
      bg = sd->G_tilt + sd->G_begin + ( sd->G_end - sd->G_begin )
        * (double) iy / (double) nmy;
      for(ix=0; ix<npx; ix++){
        bh = sd->H_tilt + sd->H_begin + ( sd->H_end - sd->H_begin )
          * (double) ix / (double) nmx;
        sd->k_xy[ct].h = 0.5 * ( bh * sd->zolzbase1.h + bg * sd->zolzbase2.h );
        sd->k_xy[ct].k = 0.5 * ( bh * sd->zolzbase1.k + bg * sd->zolzbase2.k );
        sd->k_xy[ct].l = 0.5 * ( bh * sd->zolzbase1.l + bg * sd->zolzbase2.l );
        sd->h_pos[ct] = ix - npx / 2;
        sd->g_pos[ct] = iy - npy / 2;
        sd->H[ct] = bh;		sd->G[ct] = bg;
        ct++;
      }
    }
  }
  else if (sd->calc_type == 5){/* 1 Dim. (lnp mode only for DISPSURF) */
    nd = sq(sd->num_points);			/* caution: sq */
    sd->k_xy	= new_vector(R_HKL, nd);	chk_vector(sd->k_xy);
    sd->h_pos	= new_vector(int, nd);		chk_vector(sd->h_pos);
    sd->g_pos	= new_vector(int, nd);		chk_vector(sd->g_pos);
    sd->H		= new_vector(double, nd);	chk_vector(sd->H);
    sd->G		= new_vector(double, nd);	chk_vector(sd->G);
    np = sd->num_points;
    nm = sd->num_points - 1;
    ct = 0;
    if (sd->num_g != 1){
      mb_fprintf(
        stdout, "$$ num_g must be 1 for calc_type=4 in DISPSURF mode. Abort."
        );
      exit(EXIT_FAILURE);
    }
    g = sd->reflex[sd->calc_g[0]];
    if( g.h == 0 && g.k == 0 && g.l == 0 ){
      mb_fprintf(
        stdout, "$$ g must not be 0 for calc_type=4 in DISPSURF mode. Abort."
        );
      exit(EXIT_FAILURE);
    }
    g_proj = proj_to_zolz(g, sd->incidence, sd->zolzbase1, bv, &tx, &ty);
    zb1 = normalize_vec(g_proj, sd->zolzbase1, bv);
    zb1.h = - zb1.h; zb1.k = - zb1.k; zb1.l = - zb1.l; 
    zb2 = zolz_ort_base(sd->incidence, zb1, bv);
    /* zb2 is orthogonal to zb1 and has the same norm. */
    
    /* tilt by H_tilt and G_tilt */
    sd->zolzbase2 = zolz_ort_base(sd->incidence, sd->zolzbase1, bv);
    tilt.h = 0.5 * ( sd->H_tilt * sd->zolzbase1.h + sd->G_tilt * sd->zolzbase2.h );
    tilt.k = 0.5 * ( sd->H_tilt * sd->zolzbase1.k + sd->G_tilt * sd->zolzbase2.k );
    tilt.l = 0.5 * ( sd->H_tilt * sd->zolzbase1.l + sd->G_tilt * sd->zolzbase2.l );
    
    bg = sd->G_begin;
    for(j=0; j<np; j++){
      bh = sd->H_begin + ( sd->H_end - sd->H_begin ) * (double) j / (double) nm;
      sd->k_xy[ct].h = tilt.h + 0.5 * ( bh * zb1.h + bg * zb2.h );
      sd->k_xy[ct].k = tilt.k + 0.5 * ( bh * zb1.k + bg * zb2.k );
      sd->k_xy[ct].l = tilt.l + 0.5 * ( bh * zb1.l + bg * zb2.l );
      sd->h_pos[ct] = j - np / 2;
      sd->g_pos[ct] = 0;
      sd->H[ct] = bh;		sd->G[ct] = bg;
      ct++;
    }
  }
  else if (sd->calc_type == 7){/* Arb. HG data (read from a file) */
    fp = fopen_chk_rmode(sd->hg_file);
    nd = fget_i(fp);
    sd->k_xy	= new_vector(R_HKL, nd);	chk_vector(sd->k_xy);
    sd->h_pos	= new_vector(int, nd);		chk_vector(sd->h_pos);
    sd->g_pos	= new_vector(int, nd);		chk_vector(sd->g_pos);
    sd->H		= new_vector(double, nd);	chk_vector(sd->H);
    sd->G		= new_vector(double, nd);	chk_vector(sd->G);
    for(i=0; i<nd; i++){
      comment_skip_fgets(GETLEN, buf, fp);	str_to_fs(4, buf, t);
      sd->H[i] = t[0];	sd->G[i] = t[1];
      sd->h_pos[i] = (int) t[2];	sd->g_pos[i] = (int) t[3];
      bh = sd->H_tilt + sd->H[i];		bg = sd->G_tilt + sd->G[i];
      sd->k_xy[i].h = 0.5 * ( bh * sd->zolzbase1.h + bg * sd->zolzbase2.h );
      sd->k_xy[i].k = 0.5 * ( bh * sd->zolzbase1.k + bg * sd->zolzbase2.k );
      sd->k_xy[i].l = 0.5 * ( bh * sd->zolzbase1.l + bg * sd->zolzbase2.l );
    }
    chk_data("HGDATA_END", fp);
    fclose(fp);
  }
  else{
    mb_fprintf(
      stdout, "$$ (calc_kxy_tilt) calc_type=%d is invalid. Abort.\n",
      sd->calc_type
      );
    exit(EXIT_FAILURE);
  }

  sd->num_k_points = ct;
}


#if 0
void calc_k_xy_radial_lnp( int ig, SIM_DATA *sd, CELL_PAR bv )
{
  /*
    Size of $k_xy[]$ must be less than $(num_points)^{2}$.
     (calc_type 3)
   */
  int j, nm, nd;
  double tx, ty, bh, bg;
  HKL g;
  R_HKL g_proj, zb1, zb2, tilt;
  
  /* tilt by H_tilt and G_tilt */
  sd->zolzbase2 = zolz_ort_base(sd->incidence, sd->zolzbase1, bv);
  tilt.h = 0.5 * ( sd->H_tilt * sd->zolzbase1.h + sd->G_tilt * sd->zolzbase2.h );
  tilt.k = 0.5 * ( sd->H_tilt * sd->zolzbase1.k + sd->G_tilt * sd->zolzbase2.k );
  tilt.l = 0.5 * ( sd->H_tilt * sd->zolzbase1.l + sd->G_tilt * sd->zolzbase2.l );
  
  /* convert g to zolzbase1 */
  g = sd->reflex[sd->calc_g[ig]];
  
  if( g.h == 0 && g.k == 0 && g.l == 0 ){
    zb1.h = sd->zolzbase1.h; zb1.k = sd->zolzbase1.k; zb1.l = sd->zolzbase1.l;
  }
  else{
    g_proj = proj_to_zolz(g, sd->incidence, sd->zolzbase1, bv, &tx, &ty);
    zb1 = normalize_vec(g_proj, sd->zolzbase1, bv);
    zb1.h = - zb1.h; zb1.k = - zb1.k; zb1.l = - zb1.l; 
  }
  zb2 = zolz_ort_base(sd->incidence, zb1, bv);
  /* zb2 is orthogonal to zb1 and has the same norm. */
  
  if (sd->calc_type != 4){
    mb_fprintf(
      stdout, "$$ calc_type error. Abort at calc_k_xy_radial_lnp().\n"
      );
    exit(EXIT_FAILURE);
  }
  
  bg = sd->G_lnp[ig];
  nd = sd->num_points_lnp[ig];
  nm = nd - 1;
  for(j=0; j<nd; j++){
    bh = sd->H_begin_lnp[ig]
      + ( sd->H_end_lnp[ig] - sd->H_begin_lnp[ig] ) * (double) j / (double) nm;
    sd->k_xy_lnp[ig][j].h = tilt.h + 0.5 * ( bh * zb1.h + bg * zb2.h );
    sd->k_xy_lnp[ig][j].k = tilt.k + 0.5 * ( bh * zb1.k + bg * zb2.k );
    sd->k_xy_lnp[ig][j].l = tilt.l + 0.5 * ( bh * zb1.l + bg * zb2.l );
  }
}
#endif


void sim_result_mode_output(
  SIM_DATA *sd, CELL_PAR bv, char *infile, FILE *fo_log
  )
{
  /* Updated on 030729 */
  int disk_width, disk_height,image_width,image_height,ix,iy,ndx,centerPoint;
  int  ik, it, ig, nt, nk, ng;
  int  min_hpix, max_hpix, min_gpix, max_gpix, ival;
  double  max_val, min_val, val;
  double  min_H, max_H, min_G, max_G, rx, ry, scale;
  double  gamma_coef;
  char  rsl_out[BUFLEN], scr_out[BUFLEN],tiff_out[BUFLEN] ;
  HKL   g;
  FILE  *fo, *fo_scr;
  float *data, *cData,max_HC,max_GC;
  
  ng = sd->num_g;
  nt = sd->num_thickness;
  nk = sd->num_k_points;

  /* Find max and min values of h_pos, g_pos, H, G */
  min_hpix = max_hpix = sd->h_pos[0];
  min_gpix = max_gpix = sd->g_pos[0];
  min_H = max_H = sd->H[0];
  min_G = max_G = sd->G[0];
  for(ik=0; ik<nk; ik++){
    ival = sd->h_pos[ik];
    if(min_hpix > ival){min_hpix = ival;}
    if(max_hpix < ival){max_hpix = ival;}
    ival = sd->g_pos[ik];
    if(min_gpix > ival){min_gpix = ival;}
    if(max_gpix < ival){max_gpix = ival;}
    val = sd->H[ik];
    if(min_H > val){min_H = val;}
    if(max_H < val){max_H = val;}
    val = sd->G[ik];
    if(min_G > val){min_G = val;}
    if(max_G < val){max_G = val;}
  }


  scale = (double) sd->num_points * 2.0 / fabs(sd->H_end - sd->H_begin);

  /* Script for mbutl */
  sprintf(scr_out, "%s_mbutl.txt", infile);
  fo_scr = mb_fopen_chk_wmode(scr_out);
  mb_fprintf(fo_scr, "MBUTL_SCRIPT\n");
  mb_fprintf(fo_scr, "%% script for mbutl.\n");
  mb_fprintf(fo_scr, "SIM_VISUAL_MODE\n");
  mb_fprintf(fo_scr, "%%\n");
  mb_fprintf(fo_scr, "%% VISUAL\n");
  mb_fprintf(fo_scr, "%% rs_file gamma_coef mag(int) cont_reverse_flag\n");
  mb_fprintf(fo_scr, "%%\n");
  mb_fprintf(fo_scr, "HTML_OPEN\n");
  mb_fprintf(fo_scr, "%s.html\n", infile);
  mb_fprintf(fo_scr, "%%\n");

  gamma_coef = 0.5;
  for(it=0; it<nt; it++){
    /* Calc. result */
    sprintf(rsl_out, "%s_rs_%d.txt", infile, it);
    /* mb_fprintf(stdout, "Calc. result: %s\n", rsl_out); */
    fo = mb_fopen_chk_wmode(rsl_out);

    /* script (for mbutl) */
    mb_fprintf(fo_scr, "SIM_VISUAL\n", rsl_out);
    mb_fprintf(fo_scr, "%s %g 1 0\n", rsl_out, gamma_coef);
    
    mb_fprintf(fo, "SIM_RESULT\n");
    mb_fprintf(fo, "%% %s\n", rsl_out);
    /* Find the max. and min. values of intensities */
    max_val = min_val = sd->intensity[0][it][0];
    for(ig=0; ig<ng; ig++){
      for(ik=0; ik<nk; ik++){
        val = sd->intensity[ig][it][ik];
        if (val > max_val) {max_val = val;}
        if (val < min_val) {min_val = val;}
      }
    }
    mb_fprintf(fo, "%%\n");
    mb_fprintf(fo, "VISUAL{\n");
    mb_fprintf(fo, "%% default max and min [>0]\n");
    mb_fprintf(fo, "%lf\t%lf\n", max_val, min_val);

    mb_fprintf(fo, "%% default gamma [>0]\n");
    mb_fprintf(fo, "1.0\n");
    mb_fprintf(fo, "%% defalut magnification (integer only)\n");
    mb_fprintf(fo, "1\n");
    mb_fprintf(fo, "%% default contrast-reverse-flag\n");
    mb_fprintf(fo, "%%  1:reverse, 0:not reverse\n");
    mb_fprintf(fo, "0\n");
    mb_fprintf(fo, "}VISUAL\n");

    mb_fprintf(fo, "%%\n");
    mb_fprintf(fo, "CR_PARAMETERS{\n");

    /* VVVVV */
    /*
    mb_fprintf(fo, "%% lattice parameter\n");
     */
    
    mb_fprintf(fo, "%% bv parameters\n");
    mb_fprintf(fo, "%% S11,S22,S33,S12,S13,S23\n");
    mb_fprintf(
      fo, "%% metric tensor (reciprocal) aas, bbs, ccs, abs, bcs, acs\n"
      );
    mb_fprintf(fo, "%% metric tensor (real) aa, bb, cc, ab, bc, ac\n");
    mb_fprintf(fo, "%lf\t%lf\t%lf\t%lf\t%lf\t%lf\t%lf\n",
            bv.s11, bv.s22, bv.s33, bv.s12, bv.s13, bv.s23, bv.vc);
    mb_fprintf(fo, "%lf\t%lf\t%lf\t%lf\t%lf\t%lf\n",
            bv.aas, bv.bbs, bv.ccs, bv.abs, bv.bcs, bv.acs);
    mb_fprintf(fo, "%lf\t%lf\t%lf\t%lf\t%lf\t%lf\n",
            bv.aa, bv.bb, bv.cc, bv.ab, bv.bc, bv.ac);
    mb_fprintf(fo, "%% incident direction\n");
    mb_fprintf(
      fo, "%d\t%d\t%d\n", sd->incidence.u, sd->incidence.v, sd->incidence.w
      );
    mb_fprintf(fo, "%% zolzbase1\n");
    mb_fprintf(
      fo, "%lf\t%lf\t%lf\n", sd->zolzbase1.h, sd->zolzbase1.k, sd->zolzbase1.l
      );
    mb_fprintf(fo, "%% thickness [Angstrom]\n");
    mb_fprintf(fo, "%lf\n", sd->thickness[it]);
    mb_fprintf(fo, "}CR_PARAMETERS\n");

    mb_fprintf(fo, "%%\n");
    mb_fprintf(fo, "RF_INDEX{\n");
    mb_fprintf(fo, "%% reflections\n");
    mb_fprintf(fo, "%% number of reflections\n");
    mb_fprintf(fo, "%d\n", ng);
    mb_fprintf(fo, "%% h k l, position (pixel coord)\n");
    max_HC = -10. ;
    max_GC = -10. ;
    for(ig=0; ig<ng; ig++){
      g = sd->reflex[ sd->calc_g[ig] ];
      proj_to_zolz( g, sd->incidence, sd->zolzbase1, bv, &rx, &ry );
      mb_fprintf(
        fo, "%d\t%d\t%d\t%lf\t%lf\n", g.h, g.k, g.l, scale * rx, scale * ry
        );
      max_HC = fabs(scale*rx) > max_HC ? fabs(scale*rx) : max_HC ;
      max_GC = fabs(scale*ry) > max_GC ? fabs(scale*ry) : max_GC ;
    }
    mb_fprintf(fo, "}RF_INDEX\n");

    mb_fprintf(fo, "COORD{\n");
    mb_fprintf(fo, "%% coordinates\n");
    mb_fprintf(fo, "%%\n");
    mb_fprintf(fo, "%% number of pixels\n");
    mb_fprintf(fo, "%d\n", nk);
    mb_fprintf(fo, "%% max and min of H\n");
    mb_fprintf(fo, "%lf\t%lf\n", max_H, min_H);
    mb_fprintf(fo, "%% max and min of G\n");
    mb_fprintf(fo, "%lf\t%lf\n", max_G, min_G);
    mb_fprintf(fo, "%% max and min of H pixel coordinate\n");
    mb_fprintf(fo, "%d\t%d\n", max_hpix, min_hpix);
    mb_fprintf(fo, "%% max and min of G pixel coordinate\n");
    mb_fprintf(fo, "%d\t%d\n", max_gpix, min_gpix);
    mb_fprintf(fo, "%% H, G, ih, ig\n");
    for(ik=0; ik<nk; ik++){
      mb_fprintf(
        fo, "%lf\t%lf\t%d\t%d\n",
        sd->H[ik], sd->G[ik], sd->h_pos[ik], sd->g_pos[ik]
        );
    }
    mb_fprintf(fo, "}COORD\n");

    mb_fprintf(fo, "%% intensities\n");

    for(ig=0; ig<ng; ig++){
      mb_fprintf(fo, "RF{\n");
      g = sd->reflex[ sd->calc_g[ig] ];
      mb_fprintf(fo, "%% hkl\n");
      mb_fprintf(fo, "%d\t%d\t%d\n", g.h, g.k, g.l);

      for(ik=0; ik<nk; ik++){
        mb_fprintf(fo, "%lf\n", sd->intensity[ig][it][ik]);
      }
      mb_fprintf(fo, "}RF\n");
    }
    mb_fclose(fo);
    /* Create the tiff file, but first set the data */
    disk_width = max_hpix - min_hpix + 2 ;
    disk_height = max_gpix - min_gpix + 2 ;
    image_width = 2*(max_HC + disk_width/2) ;
    image_height = 2*(max_GC + disk_height/2) ;
    if (image_height > image_width) image_width = image_height ;
    if (image_width > image_height) image_height = image_width ;
    data = malloc(image_width*image_height*sizeof(float)) ;
    for(ix = 0 ; ix < image_width*image_height; ix++) {
	*(data+ix) = 0. ;
    }
    centerPoint = (image_height/2)*image_width + image_width/2 ;
    cData = data + centerPoint ;
    /* Loop over every pixel */
    for(ig = 0 ; ig < ng; ig++) {
     	g = sd->reflex[ sd->calc_g[ig] ];
     	proj_to_zolz( g, sd->incidence, sd->zolzbase1, bv, &rx, &ry );
    	for(ik = 0 ; ik < nk; ik++) {
		ix = near(scale*rx + sd->h_pos[ik]) ;
		iy = near(scale*ry + sd->g_pos[ik]) ;
		ndx = -iy*image_width + ix ;
		if (ndx >= -centerPoint && ndx < centerPoint) { 
		if (cData+ndx < data || cData + ndx >= (data + image_width*image_height)) {
			fprintf(stderr,"\nOutside picture bounds") ;
			fprintf(stderr,"\nix = %d, iy = %d\n",ix,iy) ;
		}
		else {
			*(cData + ndx) += 
			sd->intensity[ig][it][ik] ;
		}
		}
	}
    }
    /* Create the Tiff file name */
    sprintf(tiff_out, "%s_rs_%d.tif", infile, it);
    /* Write the tiff file */
    WriteImageDataToByteTiff(tiff_out,data,image_width,image_height,0) ;
    /* Free the data */
    free(data) ;
  }

  mb_fprintf(fo_scr, "HTML_CLOSE\n");
  mb_fclose(fo_scr);
  mb_fprintf(stdout, "\nThe script for visualization ==> %s\n", scr_out);
  mb_fprintf(stdout, "Usage: mbutl %s\n", scr_out);
}


void visual_mode_output( SIM_DATA *sd, CELL_PAR bv, char *infile )
{
  /* Updated on 010824 (for bmp) */
  int		ik, it, ig, nt, nk, ng;
  /*	int		itg;	*/
  int		min_hpix, max_hpix, min_gpix, max_gpix, ival;
  double	max_val, min_val, val;
  double	min_H, max_H, min_G, max_G, rx, ry, scale;
  char	vis_out[BUFLEN], htm_out[BUFLEN], bat_out[BUFLEN];
  HKL		g;
  FILE	*fo, *fo_htm, *fo_bat;

  ng = sd->num_g;
  nt = sd->num_thickness;
  nk = sd->num_k_points;

  /* Find max and min values of h_pos, g_pos, H, G */
  min_hpix = max_hpix = sd->h_pos[0];
  min_gpix = max_gpix = sd->g_pos[0];
  min_H = max_H = sd->H[0];
  min_G = max_G = sd->G[0];
  for(ik=0; ik<nk; ik++){
    ival = sd->h_pos[ik];
    if(min_hpix > ival){min_hpix = ival;}
    if(max_hpix < ival){max_hpix = ival;}
    ival = sd->g_pos[ik];
    if(min_gpix > ival){min_gpix = ival;}
    if(max_gpix < ival){max_gpix = ival;}
    val = sd->H[ik];
    if(min_H > val){min_H = val;}
    if(max_H < val){max_H = val;}
    val = sd->G[ik];
    if(min_G > val){min_G = val;}
    if(max_G < val){max_G = val;}
  }

  scale = (double) sd->num_points * 2.0 / fabs(sd->H_end - sd->H_begin);

  /* html */
  sprintf(htm_out, "%s.htm", infile);
  mb_fprintf(stdout, "==> %s\n", htm_out);

  fo_htm = mb_fopen_chk_wmode(htm_out);
  mb_fprintf(fo_htm, "<html>\n");
  mb_fprintf(fo_htm, "<title>%s</title>\n", infile);
  mb_fprintf(fo_htm, "<body style=\"background:#ffffff\">\n");
  mb_fprintf(fo_htm, "<h2>%s</h2>\n", infile);
  mb_fprintf(fo_htm, "<hr>\n");

  /* batch (mbfit) */
  sprintf(bat_out, "%s.bat", infile);
  mb_fprintf(stdout, "For intel --> %s\n", bat_out);
  fo_bat = mb_fopen_chk_wmode(bat_out);
  mb_fprintf(fo_bat, "REM %s\n", bat_out);
  mb_fprintf(fo_bat, "REM\n");
  mb_fprintf(
    fo_bat, "REM mbfit VISUAL rsl_file gamma_coef mag(int) cont_reverse_flag\n"
    );
  mb_fprintf(fo_bat, "REM\n");

  for(it=0; it<nt; it++){
    /* Find the max. and min. value of intensities */
    max_val = min_val = sd->intensity[0][it][0];
    for(ig=0; ig<ng; ig++){
      for(ik=0; ik<nk; ik++){
        val = sd->intensity[ig][it][ik];
        if (val > max_val) {max_val = val;}
        if (val < min_val) {min_val = val;}
      }
    }

    /* output */
    sprintf(vis_out, "%s_%d.rsl", infile, it);
    mb_fprintf(stdout, "%s\n", vis_out);
    fo = mb_fopen_chk_wmode(vis_out);

    mb_fprintf(fo, "%% file:%s\n", infile);
    mb_fprintf(fo, "%% SIM result (disp_type=7)\n");
    mb_fprintf(fo, "VISUAL\n");

    mb_fprintf(fo, "%% default max and min [>0]\n");
    mb_fprintf(fo, "%lf\t%lf\n", max_val, min_val);
    mb_fprintf(fo, "%% default gamma [>0]\n");
    mb_fprintf(fo, "1.0\n");
    mb_fprintf(fo, "%% defalut magnification (integer only)\n");
    mb_fprintf(fo, "1\n");
    mb_fprintf(fo, "%% default contrast-reverse-flag\n");
    mb_fprintf(fo, "%%  1:reverse, 0:not reverse\n");
    mb_fprintf(fo, "0\n");

    mb_fprintf(fo, "%%\n");
    mb_fprintf(fo, "%% bv parameters\n");
    mb_fprintf(fo, "%% S11,S22,S33,S12,S13,S23\n");
    mb_fprintf(
      fo, "%% metric tensor (reciprocal) aas, bbs, ccs, abs, bcs, acs\n"
      );
    mb_fprintf(fo, "%% metric tensor (real) aa, bb, cc, ab, bc, ac\n");
    mb_fprintf(fo, "%lf\t%lf\t%lf\t%lf\t%lf\t%lf\t%lf\n",
            bv.s11, bv.s22, bv.s33, bv.s12, bv.s13, bv.s23, bv.vc);
    mb_fprintf(fo, "%lf\t%lf\t%lf\t%lf\t%lf\t%lf\n",
            bv.aas, bv.bbs, bv.ccs, bv.abs, bv.bcs, bv.acs);
    mb_fprintf(fo, "%lf\t%lf\t%lf\t%lf\t%lf\t%lf\n",
            bv.aa, bv.bb, bv.cc, bv.ab, bv.bc, bv.ac);

    mb_fprintf(fo, "%% incident direction\n");
    mb_fprintf(
      fo, "%d\t%d\t%d\n", sd->incidence.u, sd->incidence.v, sd->incidence.w
      );
    mb_fprintf(fo, "%% zolzbase1\n");
    mb_fprintf(
      fo, "%lf\t%lf\t%lf\n", sd->zolzbase1.h, sd->zolzbase1.k, sd->zolzbase1.l
      );

    mb_fprintf(fo, "%% thickness [Angstrom]\n");
    mb_fprintf(fo, "%lf\n", sd->thickness[it]);
    mb_fprintf(fo, "PARAMETERS_END\n");

    mb_fprintf(fo, "%%\n");
    mb_fprintf(fo, "%% reflections\n");
    mb_fprintf(fo, "%% number of reflections\n");
    mb_fprintf(fo, "%d\n", ng);
    mb_fprintf(fo, "%% h k l, position (pixel coord)\n");
 /* find max and min of positions of reflections */   

    for(ig=0; ig<ng; ig++){
      g = sd->reflex[ sd->calc_g[ig] ];
      proj_to_zolz( g, sd->incidence, sd->zolzbase1, bv, &rx, &ry );
      mb_fprintf(
        fo, "%d\t%d\t%d\t%lf\t%lf\n", g.h, g.k, g.l, scale * rx, scale * ry
        );
    }
    mb_fprintf(fo, "RF_INDEX_END\n");

    mb_fprintf(fo, "%% coordinates\n");
    mb_fprintf(fo, "%%\n");
    mb_fprintf(fo, "%% number of pixels\n");
    mb_fprintf(fo, "%d\n", nk);
    mb_fprintf(fo, "%% max and min of H\n");
    mb_fprintf(fo, "%lf\t%lf\n", max_H, min_H);
    mb_fprintf(fo, "%% max and min of G\n");
    mb_fprintf(fo, "%lf\t%lf\n", max_G, min_G);
    mb_fprintf(fo, "%% max and min of H pixel coordinate\n");
    mb_fprintf(fo, "%d\t%d\n", max_hpix, min_hpix);
    mb_fprintf(fo, "%% max and min of G pixel coordinate\n");
    mb_fprintf(fo, "%d\t%d\n", max_gpix, min_gpix);
    mb_fprintf(fo, "%% H, G, ih, ig\n");
    for(ik=0; ik<nk; ik++){
      mb_fprintf(
        fo, "%lf\t%lf\t%d\t%d\n",
        sd->H[ik], sd->G[ik], sd->h_pos[ik], sd->g_pos[ik]
        );
    }
    mb_fprintf(fo, "COORD_END\n");

    mb_fprintf(fo, "%% intensities\n");

    for(ig=0; ig<ng; ig++){
      g = sd->reflex[ sd->calc_g[ig] ];
      mb_fprintf(fo, "%% hkl\n");
      mb_fprintf(fo, "%d\t%d\t%d\n", g.h, g.k, g.l);

      for(ik=0; ik<nk; ik++){
        mb_fprintf( fo, "%lf\n", sd->intensity[ig][it][ik]	);
      }
      mb_fprintf(fo, "RF_END\n");
    }

    mb_fclose(fo);

    /* html */
    mb_fprintf(
      fo_htm, "<p><img src=\"%s_%d.png\" align=\"middle\">\n", infile, it
      );
    mb_fprintf(fo_htm, "t=%lg[A]</p>\n", sd->thickness[it]);
    /* batch (mbfit) */
    mb_fprintf(fo_bat, "mbfit VISUAL %s_%d 0.5 1 0\n", infile, it);
    /*	mb_fprintf(fo_bat, "b2j32 %s_%d.bmp /q100 /a /d\n", infile, it); */
    /*	mb_fprintf(fo_bat_alpha, "mbfit VISUAL %s_%d 0.5 1 0\n", infile, it);
      mb_fprintf(fo_bat_alpha, "b2j32 %s_%d.bmp /q100 /a /d\n", infile, it); */
  }

  mb_fprintf(fo_htm, "</body>\n</html>\n");
  mb_fclose(fo_htm);
  mb_fclose(fo_bat);
}


void visual_mode_output_coherent(
  SIM_DATA *sd, complex ***amp, CELL_PAR bv, char *infile
  )
{
  int		ik, it, ig, nt, nk, ng;	
  int		min_hpix, max_hpix, min_gpix, max_gpix, ival;
  double	max_val, min_val, val, intensity_sf;
  double	min_H, max_H, min_G, max_G, rx, ry, scale;
  char	vis_out[BUFLEN], htm_out[BUFLEN], bat_out[BUFLEN];
  HKL		g;
  FILE	*fo, *fo_htm, *fo_bat;
  
  ng = sd->num_g;
  nt = sd->num_thickness;
  nk = sd->num_k_points;
  
  /* Find max and min values of h_pos, g_pos, H, G */
  min_hpix = max_hpix = sd->h_pos[0];
  min_gpix = max_gpix = sd->g_pos[0];
  min_H = max_H = sd->H[0];
  min_G = max_G = sd->G[0];
  for(ik=0; ik<nk; ik++){
    ival = sd->h_pos[ik];
    if(min_hpix > ival){min_hpix = ival;}
    if(max_hpix < ival){max_hpix = ival;}
    ival = sd->g_pos[ik];
    if(min_gpix > ival){min_gpix = ival;}
    if(max_gpix < ival){max_gpix = ival;}
    val = sd->H[ik];
    if(min_H > val){min_H = val;}
    if(max_H < val){max_H = val;}
    val = sd->G[ik];
    if(min_G > val){min_G = val;}
    if(max_G < val){max_G = val;}
  }
  
  scale = (double) sd->num_points * 2.0 / fabs(sd->H_end - sd->H_begin);
  
  /* html */
  sprintf(htm_out, "%s.htm", infile);
  mb_fprintf(stdout, "--> %s\n", htm_out);
  fo_htm = mb_fopen_chk_wmode(htm_out);

  mb_fprintf(fo_htm, "<html>\n");
  mb_fprintf(fo_htm, "<title>%s</title>\n", infile);
  mb_fprintf(fo_htm, "<body style=\"background:#ffffff\">\n");
  mb_fprintf(fo_htm, "<h2>%s</h2>\n", infile);
  mb_fprintf(fo_htm, "<hr>\n");

  /* batch (mbfit) */
  sprintf(bat_out, "%s.bat", infile);
  mb_fprintf(stdout, "For intel --> %s\n", bat_out);
  fo_bat = mb_fopen_chk_wmode(bat_out);
  mb_fprintf(fo_bat, "REM %s\n", bat_out);
  mb_fprintf(fo_bat, "REM\n");
  mb_fprintf(
    fo_bat, "REM mbfit VISUAL rsl_file gamma_coef mag(int) cont_reverse_flag\n"
    );
  mb_fprintf(fo_bat, "REM\n");

  for(it=0; it<nt; it++){
    /* Find the max. and min. value of intensities */
    /* tmp: hard to know */
    max_val = 2.0 * epsilon.intensity_scaling;	min_val = 0.0;

    /* output */
    sprintf(vis_out, "%s_%d.rsl", infile, it);
    printf("%s\n", vis_out);
    fo = mb_fopen_chk_wmode(vis_out);

    mb_fprintf(fo, "%% file:%s\n", infile);
    mb_fprintf(fo, "%% SIM result (disp_type=7)\n");
    mb_fprintf(fo, "VISUAL_COHERENT\n");
    
    mb_fprintf(fo, "%% default max and min [>0]\n");
    mb_fprintf(fo, "%lf\t%lf\n", max_val, min_val);
    mb_fprintf(fo, "%% default gamma [>0]\n");
    mb_fprintf(fo, "1.0\n");
    mb_fprintf(fo, "%% defalut magnification (integer only)\n");
    mb_fprintf(fo, "1\n");
    mb_fprintf(fo, "%% default contrast-reverse-flag\n");
    mb_fprintf(fo, "%%  1:reverse, 0:not reverse\n");
    mb_fprintf(fo, "0\n");

    mb_fprintf(fo, "%%\n");
    mb_fprintf(fo, "%% bv parameters\n");
    mb_fprintf(
      fo, "%lf\t%lf\t%lf\t%lf\t%lf\t%lf\t%lf\n",
      bv.s11, bv.s22, bv.s33, bv.s12, bv.s13, bv.s23, bv.vc
      );
    mb_fprintf(
      fo, "%lf\t%lf\t%lf\t%lf\t%lf\t%lf\n",
      bv.aas, bv.bbs, bv.ccs, bv.abs, bv.bcs, bv.acs
      );
    mb_fprintf(
      fo, "%lf\t%lf\t%lf\t%lf\t%lf\t%lf\n",
      bv.aa, bv.bb, bv.cc, bv.ab, bv.bc, bv.ac
      );
    
    mb_fprintf(fo, "%% incident direction\n");
    mb_fprintf(
      fo, "%d\t%d\t%d\n", sd->incidence.u, sd->incidence.v, sd->incidence.w
      );
    mb_fprintf(fo, "%% zolzbase1\n");
    mb_fprintf(
      fo, "%lf\t%lf\t%lf\n", sd->zolzbase1.h, sd->zolzbase1.k, sd->zolzbase1.l
      );

    mb_fprintf(fo, "%% thickness [Angstrom]\n");
    mb_fprintf(fo, "%lf\n", sd->thickness[it]);
    mb_fprintf(fo, "PARAMETERS_END\n");

    mb_fprintf(fo, "%%\n");
    mb_fprintf(fo, "%% reflections\n");
    mb_fprintf(fo, "%% number of reflections\n");
    mb_fprintf(fo, "%d\n", ng);
    mb_fprintf(fo, "%% h k l, position (pixel coord)\n");
    for(ig=0; ig<ng; ig++){
      g = sd->reflex[ sd->calc_g[ig] ];
      proj_to_zolz( g, sd->incidence, sd->zolzbase1, bv, &rx, &ry );
      mb_fprintf(
        fo, "%d\t%d\t%d\t%lf\t%lf\n", g.h, g.k, g.l, scale * rx, scale * ry
        );
    }
    mb_fprintf(fo, "RF_INDEX_END\n");

    mb_fprintf(fo, "%% coordinates\n");
    mb_fprintf(fo, "%%\n");
    mb_fprintf(fo, "%% number of pixels\n");
    mb_fprintf(fo, "%d\n", nk);
    mb_fprintf(fo, "%% max and min of H\n");
    mb_fprintf(fo, "%lf\t%lf\n", max_H, min_H);
    mb_fprintf(fo, "%% max and min of G\n");
    mb_fprintf(fo, "%lf\t%lf\n", max_G, min_G);
    mb_fprintf(fo, "%% max and min of H pixel coordinate\n");
    mb_fprintf(fo, "%d\t%d\n", max_hpix, min_hpix);
    mb_fprintf(fo, "%% max and min of G pixel coordinate\n");
    mb_fprintf(fo, "%d\t%d\n", max_gpix, min_gpix);
    mb_fprintf(fo, "%% H, G, ih, ig\n");
    for(ik=0; ik<nk; ik++){
      mb_fprintf(
        fo, "%lf\t%lf\t%d\t%d\n",
        sd->H[ik], sd->G[ik], sd->h_pos[ik], sd->g_pos[ik]
        );
    }
    mb_fprintf(fo, "COORD_END\n");

    mb_fprintf(fo, "%% intensities\n");

    intensity_sf = sqrt(epsilon.intensity_scaling);
    for(ig=0; ig<ng; ig++){
      g = sd->reflex[ sd->calc_g[ig] ];
      mb_fprintf(fo, "%% hkl\n");
      mb_fprintf(fo, "%d\t%d\t%d\n", g.h, g.k, g.l);

      for(ik=0; ik<nk; ik++){
        mb_fprintf(
          fo, "%lf\t%lf\n",
          amp[ig][it][ik].re * intensity_sf, amp[ig][it][ik].im * intensity_sf
          );
      }
      mb_fprintf(fo, "RF_END\n");
    }

    mb_fclose(fo);

    /* html */
    mb_fprintf(
      fo_htm, "<p><img src=\"%s_%d.bmp\" align=\"middle\">\n", infile, it
      );
    mb_fprintf(
      fo_htm, "t=%lg[A]<p>\n", sd->thickness[it]
      );
    /* batch (mbfit) */
    mb_fprintf(
      fo_bat, "mbfit VISUAL %s_%d 0.5 1 0\n", infile, it
      );
    /*		mb_fprintf(fo_bat, "b2j32 %s_%d.bmp /q100 /a /d\n", infile, it); */
    /*		mb_fprintf(fo_bat_alpha, "mbfit VISUAL %s_%d 0.5 1 0\n", infile, it);
		fprintf(fo_bat_alpha, "b2j32 %s_%d.bmp /q100 /a /d\n", infile, it); */
  }

  mb_fprintf(fo_htm, "</body>\n</html>\n");
  mb_fclose(fo_htm);
  mb_fclose(fo_bat);
}


void calc_intensity_output(
  SIM_DATA *sd, CELL_PAR bv, char *infile, FILE *fo_log
  )
{
  /*
    disp_type:
     0: all in one
     1: each file
   */
  int		ik, it, ig, itg, nt, nk;
  int		posx, posy;
  double	scale, rx, ry;
  /*	double	**mat; */
  HKL		g;
  char	outfile[BUFLEN];
  FILE	*fo;

  if(dbg.vg_sel_chk_flag == 1){return;}

#if 0
  if(sd->disp_type == 0){	/* All-in-one */
    sprintf(outfile, "%s.txt", infile);
    mb_fprintf(stdout, "intensity_output to %s\n", outfile);
    fo = mb_fopen_chk_wmode(outfile);

    for(it=0; it< sd->num_thickness; it++){
      for(ig=0; ig< sd->num_g; ig++){
        mb_fprintf(fo, "%%mb\n");
        itg = sd->calc_g[ig];
        mb_fprintf(
          fo,"Rf: %d %d %d\n",
          sd->reflex[itg].h, sd->reflex[itg].k, sd->reflex[itg].l
          );
        mb_fprintf(fo, "Th: %lf [A]\n", sd->thickness[it]);
        mb_fprintf(fo, "x\ty\tIntensity\n");
        for(ik=0; ik<sd->num_k_points; ik++){
          mb_fprintf(
            fo, "%lf\t%lf\t%12.8lf\n",
            sd->H[ik], sd->G[ik], sd->intensity[ig][it][ik]
            );
        }
      }
    }
    mb_fclose(fo);
  }
  else if(sd->disp_type == 1){
    scale = (double) sd->num_points * 2.0 / fabs(sd->H_end - sd->H_begin);
    for(it=0; it<sd->num_thickness; it++){
      sprintf(outfile, "%s.%.0lf", infile, sd->thickness[it]);
      mb_fprintf(stdout, "intensity_output to %s\n",outfile);
      fo = mb_fopen_chk_wmode(outfile);

      mb_fprintf(fo,"%%mb\n");
      mb_fprintf(fo,"Th: %lf [A]\n", sd->thickness[it]);
      mb_fprintf(fo,"x\ty\tIntensity\n");
      for(ig=0; ig<sd->num_g; ig++){
        itg = sd->calc_g[ig];
        proj_to_zolz(
          sd->reflex[itg], sd->incidence, sd->zolzbase1, bv, &rx, &ry
          );
        posx = (int) (scale * rx);	posy = (int) (scale * ry);
        for(ik=0; ik<sd->num_k_points; ik++){
          mb_fprintf(
            fo, "%d\t%d\t%12.8lf\n",
            posx + sd->h_pos[ik], posy + sd->g_pos[ik],
            sd->intensity[ig][it][ik]
            );
        }
      }
      mb_fclose(fo);
    }
  }
  else if(sd->disp_type == 2){	/* Systematic line profile mode */
    sprintf(outfile, "%s.txt", infile);
    mb_fprintf(stdout, "intensity_output to %s\n",outfile);
    mb_fprintf(stdout, "(systematic line profile mode)\n");

    fo = mb_fopen_chk_wmode(outfile);
    nt = sd->num_thickness;
    scale = (double) sd->num_points * 2.0 / fabs(sd->H_end - sd->H_begin);

    mb_fprintf(fo,"%% systematic line profiles\n");
    mb_fprintf(fo,"H");
    for(it=0; it<nt; it++){
      mb_fprintf( fo,"\t%lf[A]", sd->thickness[it] );
    }
    mb_fprintf(fo,"\n");

    for(ig=0; ig<sd->num_g; ig++){
      itg = sd->calc_g[ig];
      proj_to_zolz( sd->reflex[itg], sd->incidence, sd->zolzbase1, bv, &rx, &ry );
      posx = (int) (scale * rx);	posy = (int) (scale * ry);

      mb_fprintf(fo, "%d", posx + sd->h_pos[0]);
      for(it=0; it<nt; it++){ mb_fprintf(fo, "\t%12.8lf", 0.0); }
      mb_fprintf(fo, "\n");
      
      nk = sd->num_k_points;
      for(ik=0; ik<nk; ik++){
        mb_fprintf(fo, "%d", posx + sd->h_pos[ik]);
        for(it=0; it<nt; it++){
          mb_fprintf(fo, "\t%12.8lf", sd->intensity[ig][it][ik]);
        }
        mb_fprintf(fo, "\n");
      }

      mb_fprintf(fo, "%d", posx + sd->h_pos[nk - 1]);
      for(it=0; it<nt; it++){ mb_fprintf(fo, "\t%12.8lf", 0.0); }
      mb_fprintf(fo, "\n");
    }

    mb_fclose(fo);
  }
  else if(sd->disp_type == 3){	/* fit-int-file format */
    sprintf(outfile, "%s.txt",infile);
    mb_fprintf(stdout, "intensity_output to %s\n", outfile);
    fo = mb_fopen_chk_wmode(outfile);
    mb_fprintf(fo, "%%(SIM output: fit-int-format)\n");
    mb_fprintf(fo, "%% Number of thicknesses\n");
    mb_fprintf(fo, "%d\n", sd->num_thickness);
    mb_fprintf(fo, "%% Number of reflections\n");
    mb_fprintf(fo, "%d\n", sd->num_g);
    for(it=0; it<sd->num_thickness; it++){
      mb_fprintf(fo,"%%thickness=%lf\n", sd->thickness[it]);
      for(ig=0; ig<sd->num_g; ig++){
        mb_fprintf(fo,"%% reflection no. %d\n", ig+1);
        mb_fprintf(fo,"%% hkl\n");
        itg = sd->calc_g[ig];
        mb_fprintf(
          fo,"%d\t%d\t%d\n",
          sd->reflex[itg].h, sd->reflex[itg].k, sd->reflex[itg].l
          );
        mb_fprintf(fo,"%% Number of data points\n");
        mb_fprintf(fo,"%d\n", sd->num_k_points);
        mb_fprintf(fo,"%%H\tG\tintensity\tsigma\n");
        
        for(ik=0; ik<sd->num_k_points; ik++){
          mb_fprintf(
            fo,"%lg\t%lg\t%lg\t%lg\n",
            sd->H[ik], sd->G[ik], sd->intensity[ig][it][ik], 1.0
            );
        }
        mb_fprintf(fo,"RF_END\n");
      }
      mb_fprintf(fo,"THICK_END\n");
    }
    mb_fprintf(fo,"INTENSITY_DATA_END\n");
    mb_fclose(fo);
  }
  else if(sd->disp_type == 4){	/* mbfit-output format */
    sprintf(outfile, "%s.txt",infile);
    mb_fprintf(stdout, "intensity_output to %s\n", outfile);
    fo = mb_fopen_chk_wmode(outfile);

    for(it=0; it<sd->num_thickness; it++){
      for(ig=0; ig<sd->num_g; ig++){
        mb_fprintf(fo,"%% hkl\n");
        itg = sd->calc_g[ig];
        mb_fprintf(
          fo, "%d\t%d\t%d\n",
          sd->reflex[itg].h, sd->reflex[itg].k, sd->reflex[itg].l
          );
        mb_fprintf(fo,"%% Number of data points\n");
        mb_fprintf(fo,"%d\n", sd->num_k_points);
        mb_fprintf(fo,"%% H\tG\tobs(dum)\tcal\tsig(dum)\n");

        for(ik=0; ik<sd->num_k_points; ik++){
          mb_fprintf(
            fo,"%lf\t%lf\t%lf\t%lf\t%lf\n",
            sd->H[ik], sd->G[ik], 1.0, sd->intensity[ig][it][ik], 1.0
            );
        }
      }
    }
    mb_fclose(fo);
  }
  else if(sd->disp_type == 5){	/* supercell mode */
    for(it=0; it<sd->num_thickness; it++){
      sprintf(outfile, "%s.%.0lf",infile, sd->thickness[it]);
      mb_fprintf(stdout, "intensity_output to %s\n", outfile);
      fo = mb_fopen_chk_wmode(outfile);

      mb_fprintf(fo,"%%mb\n");
      mb_fprintf(fo,"Th: %lf [A]\n", sd->thickness[it]);
      mb_fprintf(fo,"h\tk\tl\tIntensity\n");

      for(ig=0; ig<sd->num_g; ig++){
        itg = sd->calc_g[ig];
        g = sd->reflex[ig];
        ik = 0;
        mb_fprintf(
          fo, "%d\t%d\t%d\t%12.8lf\n",
          g.h, g.k, g.l, sd->intensity[ig][it][ik]
          );
      }
      mb_fclose(fo);
    }

  }
  else if(sd->disp_type == 7){/* VISUAL mode */
    visual_mode_output(sd, bv, infile);
  }
#endif
  if(sd->disp_type == 8){/* SIM_RESULT mode */
    sim_result_mode_output(sd, bv, infile, fo_log);
  }
  else{
    mb_fprintf(stdout, "Invalid disp_type\n");
    exit(EXIT_FAILURE);
  }
}


#if 0
void coherent_cbed_output(
  SIM_DATA *sd, complex ***amplitude, CELL_PAR bv, char *infile
  )
{
  int		ik, it, ig, itg, nt, nk, ng, ix, iy, nx, ny, np, npx, npy, mag;
  int		posx, posy, center_x, center_y, maxx, maxy, minx, miny;
  double	scale, rx, ry, **mat, maxval, minval;
  complex	**amp, ta;
  char	outhtml[BUFLEN], pgm_file[BUFLEN];
  FILE	*fo;
  
  ng = sd->num_g;
  nt = sd->num_thickness;
  nk = sd->num_k_points;
  
  /* Estimate nx and ny */
  if (sd->calc_type == 2){ /* disk */
    scale = (double) sd->num_points * 2.0 / fabs(sd->H_end - sd->H_begin);
    np = sd->num_points;
    maxx = maxy = minx = miny = np /2;
    for(ig=0; ig<ng; ig++){
      proj_to_zolz(
        sd->reflex[sd->calc_g[ig]], sd->incidence, sd->zolzbase1, bv, &rx, &ry
        );
      posx = (int) (scale * rx);	posy = (int) (scale * ry);
      ix = posx + np / 2;	if( maxx < ix ) { maxx = ix; }
      ix = posx - np / 2;	if( minx > ix ) { minx = ix; }
      iy = posy + np / 2;	if( maxy < iy ) { maxy = iy; }
      iy = posy - np / 2;	if( miny > iy ) { miny = iy; }
    }
  }
  else if (sd->calc_type == 3){ /* lectangle */
    scale = (double) sd->num_points_x * 2.0 / fabs(sd->H_end - sd->H_begin);
    npx = sd->num_points_x;	npy = sd->num_points_y;
    maxx = maxy = -1000000;	minx = miny = 1000000;
    for(ig=0; ig<ng; ig++){
      proj_to_zolz(
        sd->reflex[sd->calc_g[ig]], sd->incidence, sd->zolzbase1, bv, &rx, &ry
        );
      posx = (int) (scale * rx);	posy = (int) (scale * ry);
      ix = posx + npx / 2;	if( maxx < ix ) { maxx = ix; }
      ix = posx - npx / 2;	if( minx > ix ) { minx = ix; }
      iy = posy + npy / 2;	if( maxy < iy ) { maxy = iy; }
      iy = posy - npy / 2;	if( miny > iy ) { miny = iy; }
    }
  }
  mb_fprintf(stdout, "$$ npx, npy = %d %d\n", npx, npy);
  mb_fprintf(
    stdout, "$$ maxx, minx, maxy, miny = %d %d %d %d.\n", maxx, minx, maxy, miny
    ); 
  nx = (maxx - minx) / 4 * 4 + 8;
  ny = (maxy - miny) / 4 * 4 + 8;
  /* nx and ny must be a multiple of 4. */
  center_x = - minx + 1;	center_y = - miny + 1;
  /* mb_fprintf(stdout, "$$ (coherent_cbed_output) nx, ny, cx, cy = %d, %d, %d, %d\n",
		nx, ny, center_x, center_y);*/
  
  amp = new_matrix(complex, nx, ny );	chk_matrix(amp);
  mat = new_matrix(double, nx, ny );	chk_matrix(mat);
  
  /* html file */
  sprintf(outhtml, "%s.htm", infile);
  mb_fprintf(stdout, "Output images to %s\n",outhtml);
  fo = mb_fopen_chk_wmode(outhtml);
  mb_fprintf(fo, "<html>\n");
  mb_fprintf(fo, "<title>%s</title>\n", infile);
  mb_fprintf(fo, "<body>\n");
  mb_fprintf(fo, "Data file: %s.cr and %s.cal<br>\n", infile, infile);

  for(it=0; it<nt; it++){
    /* Initialize */
    for(ix = 0; ix < nx; ix++){
      for(iy = 0; iy < ny; iy++){
        amp[ix][iy] = c_conv(0.0, 0.0);
      }
    }
    
    for(ig=0; ig<ng; ig++){
      itg = sd->calc_g[ig];
      proj_to_zolz(
        sd->reflex[itg], sd->incidence, sd->zolzbase1, bv, &rx, &ry
        );
      posx = (int) (scale * rx);	posy = (int) (scale * ry);
      for(ik=0; ik<nk; ik++){
        ix = center_x + posx + sd->h_pos[ik];
        iy = center_y + posy + sd->g_pos[ik];
        if(ix >= 0 && ix < nx && iy >= 0 && iy < ny){
          amp[ix][iy] = c_add( amp[ix][iy], amplitude[ig][it][ik] );
        }
        else{
          mb_fprintf(
            stdout,
            "$$ WARNING: pixel posion (%d,%d) is out of range.\n", ix, iy
            );
        }
      }
    }
    
    /* Transfer amp. to intensity. */
    for(ix = 0; ix < nx; ix++){
      for(iy = 0; iy < ny; iy++){
        ta = amp[ix][iy];
        mat[ix][iy] = ( sq(ta.re) + sq(ta.im) ) * epsilon.intensity_scaling;
      }
    }
    
    /* html */
    mb_fprintf(fo, "t=%lf[A] ", sd->thickness[it]);
    mb_fprintf(fo, "(%s_%d)<br>", infile, it);
    findmaxmin2(nx, ny, mat, &maxval, &minval);
    mb_fprintf(fo, "\t[max, min]=[%.0lf, %.0lf]<br>\n", minval, maxval);
    mb_fprintf(fo, "<img src=\"%s_%d.png\"> <p>\n", infile, it);

    sprintf(pgm_file, "%s_%d.pgm", infile, it);
    mag = 1;
    output_pgm_image(
      nx, ny, mat, maxval, 0.0, 1.0, mag, 0, pgm_file
      );
    /*
    sprintf(outass, "%s_%d.ass", infile, it);
    ass_out(outass, mat, nx, ny);
     */
  }
  
  mb_fclose(fo);
}
#endif


void calc_cbed_pattern(
  CRYSTAL_STRCT_DATA *csd, SIM_DATA *sd, complex **vg, char *infile,
  FILE *fo_log
  )
{
  int    ig, it, ik, nr, ng, nk, nt, progress_ratio;
  double **tmp_intensity;
  double abs_incidence, abs_surf_norm, cosfai_inc_surf, wave_length;
  double ev_dash, abs_K;
  CELL_PAR bv;
  WORK_CAL_INT wci;
  int     num_perturb = 0;  /* no perturbation */
  complex ***vgs_dummy;
  double  ***int_dummy;
#if defined(MPI_MODE)
  MPI_Status mpistatus;
  int a, b, j, mulgt, *start, *end;
  double *tmpoutput;
  double  time_start, time_finish;
#endif

  bv = bravai(csd->lp);
  /* ----- Incident directions ----- */
  calc_kxy_tilt( sd, bv );
  mb_fprintf(stdout, "$$ num_k_points = %d\n", sd->num_k_points);

  /* ----- Intensity calculation ----- */
  pre_calc_intensity(
    sd->accel_kvol, sd->incidence, sd->surf_norm, bv,
    &abs_incidence, &abs_surf_norm, &cosfai_inc_surf,
    &wave_length, &ev_dash
    );
  abs_K = calc_abs_K( csd, sd->accel_kvol, bv );

  ng = sd->num_g;
  nt = sd->num_thickness;
  nk = sd->num_k_points;
  sd->intensity = new_tensor(double, ng, nt, nk);  chk_tensor(sd->intensity);
  tmp_intensity = new_matrix(double, ng, nt);  chk_matrix(tmp_intensity);

  nr = sd->num_exact + sd->num_bethe;
  num_perturb = 0;
  alloc_work_cal_int( nr, ng, num_perturb, &wci );

  mb_fprintf(stdout, "Intensity calculation\n");

#if defined(MPI_MODE)
  a = nk / num_procs;
  b = nk % num_procs;
  mulgt = ng * nt;
  start = (int *)malloc( sizeof(int) * num_procs);
  end = (int *)malloc( sizeof(int) * num_procs);
  tmpoutput = (double *)malloc( sizeof(double) * nk * mulgt);

  start[0] = 0;
  end[0] = a;
  if ( b > 0 ) end[0]++;

  for ( j = 1; j < num_procs; j++){
    start[j] = end[j-1];
    end[j] = start[j] + a;
    if ( b > j ) end[j]++;
  }

#endif

#if defined(MPI_MODE)  
  MPI_Barrier( MPI_COMM_WORLD );
  time_start = MPI_Wtime();
  mb_fprintf(
    stdout, "\n[%d] **** Starting MPI Diagonalization ****\n", proc_id
    );
  for( ik = start[proc_id]; ik < end[proc_id]; ik++ ){
#else
  for(ik=0; ik<nk; ik++){
#endif
    if(ik==0){ exec_time_forecast(-1); }
    progress_indicator(ik, nk);
    dbg_sel_vg_pos(sd->h_pos[ik], sd->g_pos[ik]);

    calc_intensity_with_perturb_ug(
      sd->k_xy[ik], abs_K,
      sd->beam_search, sd->num_exact, sd->num_bethe, sd->reflex, vg,
      num_perturb, vgs_dummy,
      sd->num_g, sd->calc_g, sd->num_thickness, sd->thickness,
      sd->accel_kvol, sd->incidence, sd->surf_norm, bv,
      abs_incidence, abs_surf_norm, cosfai_inc_surf,
      tmp_intensity, int_dummy, &wci
      );
    /* mb_fprintf(stdout, "$$ug: %lf\n", tmp_intensity[0][0]); */
#if defined(MPI_MODE)
    for(ig=0; ig<ng; ig++){
      for(it=0; it<nt; it++){
        tmpoutput[ik * mulgt + ig * nt + it]
          = tmp_intensity[ig][it] * epsilon.intensity_scaling;
      }
    }
#else
    for(ig=0; ig<ng; ig++){
      for(it=0; it<nt; it++){
        sd->intensity[ig][it][ik]
          = tmp_intensity[ig][it] * epsilon.intensity_scaling;
      }
    }
#endif
    if(ik==0){exec_time_forecast(sd->num_k_points);}
  }
#if defined(MPI_MODE)
  mb_fprintf(stdout, "\n[%d] **** Ending MPI Diagonalization ****\n", proc_id);
#endif

#if defined(MPI_MODE)
  MPI_Barrier(MPI_COMM_WORLD);
  mb_fprintf(
    stdout, "\n\t[%d] **** Starting MPI Communication ****\n", proc_id               );
  if( proc_id == 0 ){
    for ( j = 1; j < num_procs; j++){
      MPI_Recv(
        tmpoutput + start[j] * mulgt, (end[j] - start[j]) * mulgt,
        MPI_DOUBLE, j, 100, MPI_COMM_WORLD, &mpistatus
        );
    }
  }
  else{
    MPI_Ssend(  /* <<<<< */
      tmpoutput + start[proc_id] * mulgt,
      (end[proc_id] - start[proc_id]) * mulgt,
      MPI_DOUBLE, 0, 100, MPI_COMM_WORLD
      );
  }
  if( proc_id == 0 ){
    for(ik=0; ik<nk; ik++){
      for(ig=0; ig<ng; ig++){
        for(it=0; it<nt; it++){
          sd->intensity[ig][it][ik] = tmpoutput[ik * mulgt + ig * nt + it];
        }
      }
    }
  }
  free(start);
  free(end);
  free(tmpoutput);
  MPI_Barrier(MPI_COMM_WORLD);
  time_finish = MPI_Wtime();
  mb_fprintf(
    stdout, "\n[%d] **** Ending MPI Communication ****\n", proc_id
    );
  time_finish -= time_start;
  if( proc_id == 0 ){
    mb_fprintf(
      stdout, "\n\nElapsed time = %e seconds = %lf hours\n\n",
      time_finish, time_finish / 3600.0
      );
  }
#endif
  
  exec_time_forecast(0);
  /*
  if(dbg.eigval_dump_flag == 1 || dbg.eigvec_dump_flag == 1){
    fclose(dbg.fp);
  }
   */
  /* ----- Output ----- */
  calc_intensity_output( sd, bv, infile, fo_log );
  free_work_cal_int( &wci );
  free_matrix(tmp_intensity);
}


void calc_initial_phase(
  SIM_DATA *sd, double lambda, complex initial_phase[], CELL_PAR bv
  )
{
  /* For coherent CBED mode.  Phases of incident wave at the upper surface. */
  int	ik, nk;
  double	defocus, cs, lambda_cub, orgx, orgy, orgz, sq_abs_kxy, chikt;
  R_HKL	kxy;
  
  defocus = sd->df;	cs = sd->coef_cs;
  orgx = sd->org_u;	orgy = sd->org_v;	orgz = sd->org_w;
  lambda_cub = lambda * lambda * lambda;
  nk = sd->num_k_points;
  
  for(ik=0; ik<nk; ik++){
    kxy = sd->k_xy[ik];
    sq_abs_kxy = sq_abs(kxy, bv);
    chikt =
      PI * (
        defocus * lambda * sq_abs_kxy
        + 0.5 * cs * lambda_cub * sq(sq_abs_kxy)
        + 2.0 * (orgx * kxy.h + orgy * kxy.k + orgz * kxy.l)
        );  /* <<<<< */
    initial_phase[ik] = c_exp(c_conv(0, chikt));
  }
}


void calc_coherent_cbed_pattern (
  CRYSTAL_STRCT_DATA *csd, SIM_DATA *sd, complex **vg, char *infile
  )
{
  int		ig, it, ik, nr, ng, nk, nt, progress_ratio;
  double	abs_incidence, abs_surf_norm, cosfai_inc_surf, wave_length, ev_dash, abs_K;
  CELL_PAR bv;
  WORK_CAL_INT wci;
  int num_perturb = 0;		/* dummy arguments for calc_intensity_with_perturb */
  complex	**tmp_amplitude, ***amplitude, *initial_phase;
  complex ***vgs_dummy, ***p_amp_dum;
  
  bv = bravai(csd->lp);
  /* ----- Incident directions ----- */
  calc_kxy_tilt( sd, bv );
  mb_fprintf(stdout, "$$ num_k_points = %d\n", sd->num_k_points);
  
  /* ----- Intensity calculation ----- */
  pre_calc_intensity(
    sd->accel_kvol, sd->incidence, sd->surf_norm, bv,
    &abs_incidence, &abs_surf_norm, &cosfai_inc_surf,
    &wave_length, &ev_dash
    );
  abs_K = calc_abs_K( csd, sd->accel_kvol, bv );				/* <<<<< */
  
  /* ----- Calculate initial phases of the incident beams ----- */
  initial_phase = new_vector(complex, sd->num_k_points);
  chk_vector(initial_phase);
  calc_initial_phase( sd, wave_length, initial_phase, bv );
  
  ng = sd->num_g;
  nt = sd->num_thickness;
  nk = sd->num_k_points;
  sd->intensity = new_tensor(double, 1, 1, 1);
  chk_tensor(sd->intensity);	/* dummy */
  amplitude = new_tensor(complex, ng, nt, nk);	chk_tensor(amplitude);
  tmp_amplitude = new_matrix(complex, ng, nt);	chk_matrix(tmp_amplitude);
  
  nr = sd->num_exact + sd->num_bethe;		num_perturb = 0;
  alloc_work_cal_int( nr, ng, num_perturb, &wci );
  
  mb_fprintf(stdout, "Intensity calculation\n");
  
  /*  progress_ratio = 0; */
  for(ik=0; ik<nk; ik++){
    
    if(ik==0) {exec_time_forecast(-1);}
    progress_indicator(ik, nk);
    /*  exec_progress(&progress_ratio, ik, nk); */
    
    dbg_sel_vg_pos(sd->h_pos[ik], sd->g_pos[ik]);		/* debug */
    
    calc_intensity_ug_coherent(
      sd->k_xy[ik], abs_K, initial_phase[ik],
      sd->beam_search, sd->num_exact, sd->num_bethe, sd->reflex, vg,
      num_perturb, vgs_dummy,
      sd->num_g, sd->calc_g, sd->num_thickness, sd->thickness,
      sd->accel_kvol, sd->incidence, sd->surf_norm, bv,
      abs_incidence, abs_surf_norm, cosfai_inc_surf,
      tmp_amplitude, p_amp_dum, &wci
      );
    
    for(ig=0; ig<ng; ig++){
      for(it=0; it<nt; it++){
        amplitude[ig][it][ik]
          = tmp_amplitude[ig][it];	/* <<<<< amplitudes of diffracted waves */
      }
    }
    
    if(ik==0){exec_time_forecast(sd->num_k_points);}
  }
  exec_time_forecast(0);
  /*
  if(dbg.eigval_dump_flag == 1 || dbg.eigvec_dump_flag == 1){
    fclose(dbg.fp);
  }
   */
  
  /* ----- Output ----- */
  /* coherent_cbed_output( sd, amplitude, bv, infile ); */
  visual_mode_output_coherent( sd, amplitude, bv, infile );
  
  free_work_cal_int( &wci );
  free_matrix(tmp_amplitude);	free_tensor(amplitude);
  free_vector(initial_phase);
}


double calc_probe_intensity( R_UVW rx, int ne, HKL *reflex, complex *probe )
{
  /*
    Calculate square of probe func.
   */
  int ig;
  complex p;
  double gr;
  
  p = c_conv(0.0,0.0);
  for(ig=0; ig<ne; ig++){
    gr	= (double) reflex[ig].h * rx.u
      + (double) reflex[ig].k * rx.v
        + (double) reflex[ig].l * rx.w;
    p = c_add( p, c_mul( probe[ig], c_exp( c_conv(0.0, PI * 2.0 * gr) ) ) );
  }
  return sq(p.re) + sq(p.im);
}


void output_probe_projpot(
  char *infile, int ne, int nb, complex *probe, HKL *reflex, complex **vg,
  double x_begin, double x_end, double y_begin, double y_end,
  int num_points_x, int num_points_y, R_UVW x1, R_UVW x2
  )
{
  int		ix, iy, nmx, nmy, prog;
  double	probe_intensity, x_scale, y_scale;
  complex	projp;
  R_UVW	rx;
  /*	char	pp_outfile[128]; */
  char	probe_outfile[BUFLEN];
  FILE	*fo_probe;
  
  sprintf(probe_outfile, "%s.prb", infile);
  /* ----- Projected potential and sq. of probe func. ----- */

  mb_fprintf(
    stdout, "Proj. pot and sq. of probe function: ==> %s\n",
    probe_outfile
    );

  fo_probe = mb_fopen_chk_wmode(probe_outfile);

  mb_fprintf(fo_probe, "%% Projected potential and probe function\n");
  mb_fprintf(fo_probe, "%% (formed by using %d beams)\n", ne+nb);
  mb_fprintf(
    fo_probe,
    "x\ty\tix\tiy\tpotential.re[V]\tpotential.im[V]\tprobe_intensity\n"
    );

  nmx = num_points_x - 1;	nmy = num_points_y - 1;
/*  prog = 0; */
  for(ix=0; ix<num_points_x; ix++){
/*    exec_progress(&prog, ix, num_points_x); */
    progress_indicator(ix, num_points_x);
    if(nmx == 0){
      x_scale = x_begin;
    }
    else{
      x_scale = x_begin + (x_end - x_begin) * (double) ix / (double) nmx;
    }
    for(iy=0; iy<num_points_y; iy++){
      if(nmy == 0){
        y_scale = y_begin;
      }
      else{
        y_scale = y_begin + (y_end - y_begin) * (double) iy / (double) nmy;
      }
      rx.u = x_scale * x1.u + y_scale * x2.u;
      rx.v = x_scale * x1.v + y_scale * x2.v;
      rx.w = x_scale * x1.w + y_scale * x2.w;
      /*	printf(
				"$$ x_scale, y_scale=%lf %lf, rx = %lf %lf %lf\n",
				x_scale, y_scale, rx.u, rx.v, rx.w
			);*/
      projp = calc_projpot( rx, ne + nb, reflex, vg );
      probe_intensity = calc_probe_intensity( rx, ne, reflex, probe );
      mb_fprintf(
        fo_probe, "%lf\t%lf\t%d\t%d\t%lf\t%lf\t%lf\n",
        x_scale, y_scale, ix, iy, projp.re, projp.im, probe_intensity
        );
    }
  }
  mb_fprintf(stdout, "100%%.\n");
  mb_fclose(fo_probe);

}

#if 0
void calc_cbed_pattern_spcell(
  CRYSTAL_STRCT_DATA *csd, SIM_DATA *sd, complex **vg, char *infile
  )
{
  /* Artificial superlattice method */
  int     i, ne, nb, ig, it, ik, nr, ng, nk, nt;
  /*	int		progress_ratio; */
  double  **tmp_intensity, abs_g, abs_g_zolzbase, chikt, kt, lambda;
  double  abs_incidence, abs_surf_norm, cosfai_inc_surf;
  double  wave_length, ev_dash, abs_K;
  CELL_PAR bv;
  WORK_CAL_INT wci;
  R_HKL	rg;
  int      num_perturb = 0; /* dummy arg. for calc_intensity_with_perturb */
  complex  ***vgs_dummy, *probe;
  double   ***int_dummy;

  bv = bravai(csd->lp);
  lambda = wavelength(sd->accel_kvol);
  
  /* ----- Incident directions ----- */
  calc_kxy_tilt( sd, bv );
  
  /* ----- Intensity calculation ----- */
  pre_calc_intensity(
    sd->accel_kvol, sd->incidence, sd->surf_norm, bv,
    &abs_incidence, &abs_surf_norm, &cosfai_inc_surf,
    &wave_length, &ev_dash
    );
  abs_K = calc_abs_K( csd, sd->accel_kvol, bv );
  
  ng = sd->num_g;
  nt = sd->num_thickness;
  nk = sd->num_k_points;
  sd->intensity = new_tensor(double, ng, nt, nk);	chk_tensor(sd->intensity);
  tmp_intensity = new_matrix(double, ng, nt);		chk_matrix(tmp_intensity);
  
  nr = sd->num_exact + sd->num_bethe;		num_perturb = 0;
  alloc_work_cal_int( nr, ng, num_perturb, &wci );
  
  /* Probe function */
  ne = sd->num_exact;		nb = sd->num_bethe;
  abs_g_zolzbase = sqrt( sq_abs( sd->zolzbase1, bv) );
  probe = new_vector(complex, ne);	chk_vector(probe);
  for(i=0; i<ne; i++){
    rg = hkl2rhkl(sd->reflex[i]);
    abs_g = sqrt( sq_abs( rg, bv) );
    kt = abs_g;													/* <<<<< */
    chikt =
      PI * (
        sd->df * lambda * sq(kt)
        + 0.5 * sd->coef_cs * sq(lambda) * lambda * sq(sq(kt))
        + 2.0 * (sd->org_u * rg.h + sd->org_v * rg.k + sd->org_w * rg.l)
        );													/* <<<<< */
    if(0.5 * abs_g / abs_g_zolzbase < sd->H_aperture){
      /* <<<<< Tentatively */
      probe[i] = c_exp(c_conv(0, chikt));
    }
    else{
      probe[i] = c_conv(0.0, 0.0);
    }
  }
  output_probe_projpot(
    infile, sd->num_exact, sd->num_bethe, probe, sd->reflex, vg,
    sd->x_begin, sd->x_end, sd->y_begin, sd->y_end,
    sd->num_points_x, sd->num_points_y, sd->x1, sd->x2
    );

  mb_fprintf(stdout, "*Intensity calculation\n");

  ik = 0;
  exec_time_forecast(-1);

  dbg_sel_vg_pos(sd->h_pos[ik], sd->g_pos[ik]);		/* debug */

  calc_intensity_ug_spcell(
    sd->k_xy[ik], abs_K, probe,
    sd->beam_search, sd->num_exact, sd->num_bethe, sd->reflex, vg,
    num_perturb, vgs_dummy,
    sd->num_g, sd->calc_g, sd->num_thickness, sd->thickness,
    sd->accel_kvol, sd->incidence, sd->surf_norm, bv,
    abs_incidence, abs_surf_norm, cosfai_inc_surf,
    tmp_intensity, int_dummy, &wci
    );
  /*	mb_fprintf(stdout, "$$ug: %lf\n", tmp_intensity[0][0]);		*/
  
  for(ig=0; ig<ng; ig++){
    for(it=0; it<nt; it++){
      sd->intensity[ig][it][ik]
        = tmp_intensity[ig][it] * epsilon.intensity_scaling;
    }
  }
  
  exec_time_forecast(0);
  
  /* ----- Output ----- */
  calc_intensity_output( sd, bv, infile );
  
  free_vector(probe);
  free_work_cal_int( &wci );
  free_matrix(tmp_intensity);
}
#endif

#if 0
void calc_each_line_profile (CRYSTAL_STRCT_DATA *csd, SIM_DATA *sd, complex **vg, char *infile)
{
  int ig, it, ik, tmp_num_g, *tmp_calc_g, ng, nt, nd, total_nd, ct_d, tmp_ig, progress_ratio;
  double	**tmp_intensity;
  double	abs_incidence, abs_surf_norm, cosfai_inc_surf, wave_length, ev_dash, abs_K;
  CELL_PAR bv;
  WORK_CAL_INT wci;
  int		nr, num_perturb = 0;	/* dummy arguments for calc_intensity_with_perturb */
  complex ***vgs_dummy;
  double	***int_dummy;
  
  bv = bravai(csd->lp);
  
  /* ----- Intensity calculation ----- */
  pre_calc_intensity(
    sd->accel_kvol, sd->incidence, sd->surf_norm, bv,
    &abs_incidence, &abs_surf_norm, &cosfai_inc_surf,
    &wave_length, &ev_dash
    );
  abs_K = calc_abs_K( csd, sd->accel_kvol, bv );			/* <<<<< */
  
  mb_fprintf(stdout, "Intensity calculation\n");
  
  ng = sd->num_g;
  nt = sd->num_thickness;
  sd->intensity = new_tensor_element(ng);	chk_tensor(sd->intensity);
  sd->k_xy_lnp = new_matrix_element(ng);	chk_matrix(sd->k_xy_lnp);
  
  tmp_num_g = 1;
  tmp_calc_g = new_vector(int, tmp_num_g);	chk_vector(tmp_calc_g);
  tmp_intensity = new_matrix(double, tmp_num_g, nt);	chk_matrix(tmp_intensity);
  nr = sd->num_exact + sd->num_bethe;
  num_perturb = 0;
  alloc_work_cal_int( nr, tmp_num_g, num_perturb, &wci );					/* <<<<< */
  
  total_nd = 0;
  for(ig=0; ig<ng; ig++){ total_nd += sd->num_points_lnp[ig]; }
  ct_d = 0;
  progress_ratio = 0;
  
  for(ig=0; ig<ng; ig++){
    
    nd = sd->num_points_lnp[ig];
    sd->k_xy_lnp[ig] = new_vector(R_HKL, nd);	chk_vector(sd->k_xy_lnp[ig]);
    
    calc_k_xy_radial_lnp( ig, sd, bv );		/* Incident directions */
    
    sd->intensity[ig] = new_matrix_element(nt); chk_matrix(sd->intensity[ig]);
    for(it=0; it<nt; it++){
      sd->intensity[ig][it] = new_vector(double, nd);
      chk_vector(sd->intensity[ig][it]);
    }
    
    tmp_ig = 0;
    tmp_calc_g[tmp_ig] = sd->calc_g[ig];
    
    for(ik=0; ik<nd; ik++){
      
      if(ig==0 && ik==0) {exec_time_forecast(-1);}
      progress_indicator(ct_d++, total_nd);
      /* exec_progress(&progress_ratio, ct_d++, total_nd); */
      calc_intensity_with_perturb_ug(
        sd->k_xy_lnp[ig][ik], abs_K,
        sd->beam_search, sd->num_exact, sd->num_bethe, sd->reflex, vg,
        num_perturb, vgs_dummy,
        tmp_num_g, tmp_calc_g, sd->num_thickness, sd->thickness,
        sd->accel_kvol, sd->incidence, sd->surf_norm, bv,
        abs_incidence, abs_surf_norm, cosfai_inc_surf,
        tmp_intensity, int_dummy, &wci
        );
      /* mb_fprintf(stdout, "$$ug: %lf\n", tmp_intensity[0][0]); */
      for(it=0; it<nt; it++){
        sd->intensity[ig][it][ik]
          = tmp_intensity[tmp_ig][it] * epsilon.intensity_scaling;
      }
      
      if(ig==0 && ik==0){ exec_time_forecast(total_nd);}
    }
  }
  exec_time_forecast(0);
  
  /* if(dbg.eigval_dump_flag == 1 || dbg.eigvec_dump_flag == 1){ fclose(dbg.fp); } */
  
  /* ----- Output ----- */
  if(sd->disp_type <0 && sd->disp_type >1){ sd->disp_type = 1; }
  calc_intensity_output_lnp( sd, bv, infile );
  
  free_work_cal_int( &wci );
  free_matrix(tmp_intensity);
}
#endif


double eval_disksize(
  double H_begin, double H_end, double G_begin, double G_end
  )
{
  double disksize;
  disksize = 0.0;
  if(fabs(H_begin) > disksize){ disksize = fabs(H_begin); }
  if(fabs(H_end) > disksize){ disksize = fabs(H_end); }
  if(fabs(G_begin) > disksize){ disksize = fabs(G_begin); }
  if(fabs(G_end) > disksize){ disksize = fabs(G_end); }
  return disksize;
}


void input_sim_data(SIM_DATA *sd, CRYSTAL_STRCT_DATA *csd, FILE *fi)
{
  /*
    data read & array allocation
   */
  int  i;
  char  buf[BUFLEN];
  char  message[BUFLEN];
  CELL_PAR  bv;

  bv = bravai(csd->lp);

  chk_data("SIM_CALCULATION_CONDITION{", fi);

  /* INCIDENT_BEAM block */
  chk_data("INCIDENT_BEAM{", fi);
  sd->incidence = fget_uvw(fi);
  sd->surf_norm = fget_uvw(fi);
  sd->zolzbase1 = fget_r_hkl(fi);
  sd->H_tilt = fget_d(fi);
  sd->G_tilt = fget_d(fi);
  sd->accel_kvol = fget_d(fi);
  chk_data("}INCIDENT_BEAM", fi);

  /* CALC_KXY_RANGE block */
  /*chk_data("CALC_KXY_RANGE{", fi);*/
  sprintf(
    message, "$$$$$ Caution!!! CALC_KXY_RANGE must be before BEAM_SEARCH. $$$$$"
    );
  chk_data_info("CALC_KXY_RANGE{", message, fi);
  sd->calc_range_flag = fget_i(fi);
  sd->calc_type = fget_i(fi);
  if(
    sd->calc_type == 0 || sd->calc_type == 1 ||
    sd->calc_type == 2 || sd->calc_type == 3
    ){
    sd->num_points = fget_i(fi);
    fget_2d( &(sd->H_begin), &(sd->H_end), fi );
    fget_2d( &(sd->G_begin), &(sd->G_end), fi );
  }
  else{
    mb_fprintf(stdout, "$$ Invalid calc_type=%d. Abort.\n", sd->calc_type);
    exit(EXIT_FAILURE);
  }
  chk_data("}CALC_KXY_RANGE", fi);
  sd->beam_search.disksize
    = eval_disksize( sd->H_begin, sd->H_end, sd->G_begin, sd->G_end );
  mb_fprintf(stdout, "$$disksize = %f\n", sd->beam_search.disksize);

  /* BEAM_SEARCH block */
  chk_data("BEAM_SEARCH{", fi);
  fget_sd_beams( sd, csd, fi );
  reflex_dump_csd(
    0, sd->num_exact, sd->num_bethe, sd->reflex,
    csd, sd->beam_search, sd->incidence, sd->zolzbase1, sd->accel_kvol, bv
    );
  chk_data("}BEAM_SEARCH", fi);

  /* OUTPUT_REFLEX block */
  chk_data("OUTPUT_REFLEX{", fi);
  sd->num_g = fget_i(fi);  /* number of reflections */

  if(sd->num_g > 0){
    mb_fprintf(stdout, "num_g = %d\n", sd->num_g);
    sd->calc_g = new_vector(int, sd->num_g);
    chk_vector(sd->calc_g);
    if(
      sd->calc_type == 0 || sd->calc_type == 1 ||
      sd->calc_type == 2 || sd->calc_type == 3
      ){  /* normal */
      fget_cal_g(
        sd->num_g, sd->num_exact, sd->num_bethe, sd->calc_g, sd->reflex, fi
        );
    }
    else{
      mb_fprintf(
        stdout, "$$ invalid calc_type=%d. Abort. (input_sim_data)\n",
        sd->calc_type
        );
      exit(EXIT_FAILURE);
    }
  }
  else if(sd->num_g == 0){  /* all reflections */
    sd->num_g = sd->num_exact;
    mb_fprintf(stdout, "num_g = %d\n", sd->num_g);
    sd->calc_g = new_vector(int, sd->num_g);
    chk_vector(sd->calc_g);
    if(
      sd->calc_type == 0 || sd->calc_type == 1 ||
      sd->calc_type == 2 || sd->calc_type == 3
      ){ /* normal */
      for(i=0; i<sd->num_g; i++){ sd->calc_g[i] = i; }
    }
    else{
      mb_fprintf(
        stdout, "$$ invalid calc_type=%d. Abort. (input_sim_data)\n",
        sd->calc_type
        );
      exit(EXIT_FAILURE);
    }
  }
  chk_data("}OUTPUT_REFLEX", fi);

  chk_data("OUTPUT_THICKNESS{", fi);
  sd->num_thickness = fget_i(fi);
  mb_fprintf(stdout, "num_thickness = %d\n", sd->num_thickness);
  sd->thickness = new_vector(double, sd->num_thickness);
  chk_vector(sd->thickness);
  fget_cal_thick(sd->num_thickness, sd->thickness, fi);
  chk_data("}OUTPUT_THICKNESS", fi);

/*
  chk_data("DISPLAY{", fi);
  fget_1s( buf, fi );
  if(strcmp(buf, "VISUAL") == 0){
    sd->disp_type = 7;
  }
  else{
    sd->disp_type = atoi(buf);
  }
  chk_data("}DISPLAY",fi);
  fget_1s(buf, fi);
*/
  /*sd->disp_type = 7;*/
  sd->disp_type = 8;  /* default */

  chk_data("}SIM_CALCULATION_CONDITION", fi);
}


void mb_simulation( FILE *fi, FILE *fo_log, char *infile )
{
  /*
   * Simulation using the dynamical diffraction theory
   *  *fi: parameter file
   *  *fo_log: log file
   */
  CRYSTAL_STRCT_DATA csd;
  int nr;
  complex **vg;
  SIM_DATA sd;
  CELL_PAR bv;

  input_csd_cstr(&csd, fi, infile);    /* crystal structure data */
  input_sim_data(&sd, &csd, fi);       /* sim data */
  mb_fclose(fi);
  bv = bravai(csd.lp);

  /* ----- Potential matrix ----- */
  mb_fprintf(stdout, "Potential :\n");  mb_fflush(stdout);
  nr = sd.num_exact + sd.num_bethe;
  vg = new_matrix(complex, nr, nr);	chk_matrix(vg);
  potential_matrix_csd(
    sd.num_exact, sd.num_bethe, sd.reflex, &csd, sd.accel_kvol, bv, vg
    );
  debug_vg_dump( vg, sd.num_exact + sd.num_bethe );

  /* debug */
  dbg_dump_ini_sim();
  dbg_sel_vg_init(sd.num_exact, sd.num_bethe);

  /* ----- Intensity calculation ----- */
  calc_cbed_pattern(&csd, &sd, vg, infile, fo_log);

  dbg_dump_term();
  free_matrix(vg);
  free_sd(&sd);
  free_csd(&csd);
}


/* ****************************** */
/* *****   COHERENT part    ***** */
/* ****************************** */

void input_sim_spcell_data(SIM_DATA *sd, CRYSTAL_STRCT_DATA *csd, FILE *fi)
{
  /*
    data read & array allocation
   */
  sd->disp_type = 5;
  fget_3i( &(sd->Ncell_a), &(sd->Ncell_b), &(sd->Ncell_c), fi);
  sd->H_aperture = fget_d(fi);
  sd->coef_cs = fget_d(fi);
  sd->df = fget_d(fi);
  fget_3d( &(sd->org_u), &(sd->org_v), &(sd->org_w), fi );
  fget_2d(&(sd->x_begin), &(sd->x_end), fi);
  fget_2d(&(sd->y_begin), &(sd->y_end), fi);
  fget_2i(&(sd->num_points_x), &(sd->num_points_y), fi);
  chk_data("SPCELL_END",fi);
  
}


void input_sim_coherent_data(SIM_DATA *sd, CRYSTAL_STRCT_DATA *csd, FILE *fi)
     /*
      * data read for coherent CBED mode
      */
{
  sd->disp_type = 1;
  sd->coef_cs = fget_d(fi);	/* spherical aberration coef. (mm unit) */
  sd->df = fget_d(fi);		/* defocus (Angstrom unit) */
  fget_3d( &(sd->org_u), &(sd->org_v), &(sd->org_w), fi );	/* probe position */
  chk_data("COHERENT_END",fi);
  
}


void mb_sim_coherent( FILE *fi, char *infile )
{
  CRYSTAL_STRCT_DATA csd;
  int nr;
  complex **vg;
  SIM_DATA sd;
  CELL_PAR bv;
  

  /* ----- Data read, array alloc. and beam search ----- */
  input_csd_cstr(&csd, fi, infile);    /* crystal structure data */
  input_sim_data(&sd, &csd, fi);
  input_sim_coherent_data(&sd, &csd, fi);
  fclose(fi);
  bv = bravai(csd.lp);

  /* ----- Potential matrix ----- */
  mb_fprintf(stdout, "Potential :");  mb_fflush(stdout);
  nr = sd.num_exact + sd.num_bethe;
  vg = new_matrix(complex, nr, nr);	chk_matrix(vg);
  potential_matrix_csd(
    sd.num_exact, sd.num_bethe, sd.reflex, &csd, sd.accel_kvol, bv, vg
    );
  debug_vg_dump( vg, sd.num_exact + sd.num_bethe );
  
  /* debug */
  dbg_dump_ini_sim();
  dbg_sel_vg_init(sd.num_exact, sd.num_bethe);
  
  /* ----- Intensity calculation ----- */
  if(sd.calc_type == 2 || sd.calc_type == 3){
    calc_coherent_cbed_pattern(&csd, &sd, vg, infile);	/* 2D */
  }
  else{
    mb_fprintf(stdout, "$$ invalid calc_type=%d for coherent CBED. Abort.\n",
           sd.calc_type);
    exit(EXIT_FAILURE);
  }
  
  dbg_dump_term();
  
  free_matrix(vg);
  free_sd(&sd);
  free_csd(&csd);
}


#if 0
void convert_spcell(
  CRYSTAL_STRCT_DATA *csd, CRYSTAL_STRCT_DATA *csd_sp, SIM_DATA *sd
  )
     /* Convert normal cell to supercell */
{
  int	n_atm_unit, n_atm_super, ct_atm, natm, i, j, ix, iy, iz, *num_nthatm;
  XYZ **pos;
  HKL *g_exact, *g_bethe;
  CELL_PAR	bv;
  
  /* lattice parameter */
  csd_sp->lp.a = (double) sd->Ncell_a * csd->lp.a;
  csd_sp->lp.b = (double) sd->Ncell_b * csd->lp.b;
  csd_sp->lp.c = (double) sd->Ncell_c * csd->lp.c;
  csd_sp->lp.alpha = csd->lp.alpha;
  csd_sp->lp.beta = csd->lp.beta;
  csd_sp->lp.gamma = csd->lp.gamma;
  
  /* number of atoms:  */
  csd_sp->space_group = 1;					/* P1 */
  csd_sp->num_sym_op = 1;
  csd_sp->sym_op = new_vector(SYM_OP, 1);	chk_vector(csd_sp->sym_op);
  csd_sp->sym_op[0].r11 = csd_sp->sym_op[0].r22 = csd_sp->sym_op[0].r33 = 1.0;
  csd_sp->sym_op[0].r21 = csd_sp->sym_op[0].r31 = csd_sp->sym_op[0].r12 = 0.0;
  csd_sp->sym_op[0].r32 = csd_sp->sym_op[0].r13 = csd_sp->sym_op[0].r23 = 0.0;
  csd_sp->sym_op[0].t1 = csd_sp->sym_op[0].t2 = csd_sp->sym_op[0].t3 = 0.0;
  strcpy(csd_sp->lattice_type, csd->lattice_type);
  
  /* atom sites */
  num_nthatm = new_vector(int, csd->n_atom);
  chk_vector(num_nthatm);
  pos = new_matrix(XYZ, csd->n_atom, csd->num_sym_op);
  chk_matrix(pos);
  newpos_sym_op(
    csd->n_atom, csd->atm, num_nthatm, pos, csd->num_sym_op, csd->sym_op
    );
  ct_atm = 0;
  for(i=0; i<csd->n_atom; i++){
    for(j=0; j<num_nthatm[i]; j++){
      ct_atm++;
    }
  }
  n_atm_unit = ct_atm;
  
  natm = n_atm_unit * sd->Ncell_a * sd->Ncell_b * sd->Ncell_c;
  mb_fprintf(stdout, "$$ n_atom = %d\n", natm);
  csd_sp->atm = new_vector(ATOM, natm);
  chk_vector(csd_sp->atm);
  csd_sp->atm_f	= new_vector(ATOM_FIT, natm);		chk_vector(csd_sp->atm_f);
  csd_sp->xdyl		= new_vector(XDT, natm);		chk_vector(csd_sp->xdyl);
  csd_sp->ftbl		= new_vector(ATOMF_BK, natm);	chk_vector(csd_sp->ftbl);
  
  ct_atm = 0;
  for(i=0; i<csd->n_atom; i++){
    for(j=0; j<num_nthatm[i]; j++){
      csd_sp->atm[ct_atm] = csd->atm[i];
      csd_sp->xdyl[ct_atm] = csd->xdyl[i];
      csd_sp->ftbl[ct_atm] = csd->ftbl[i];
      csd_sp->atm[ct_atm].x = pos[i][j].x / (double) sd->Ncell_a;
      csd_sp->atm[ct_atm].y = pos[i][j].y / (double) sd->Ncell_b;
      csd_sp->atm[ct_atm].z = pos[i][j].z / (double) sd->Ncell_c;
      ct_atm++;
    }
  }
  n_atm_unit = ct_atm;
  
  /* supercell */
  ct_atm=0;
  for(ix=0; ix<sd->Ncell_a; ix++){
    for(iy=0; iy<sd->Ncell_b; iy++){
      for(iz=0; iz<sd->Ncell_c; iz++){
        for(i=0; i<n_atm_unit; i++){
          csd_sp->atm[ct_atm] = csd_sp->atm[i];
          csd_sp->xdyl[ct_atm] = csd_sp->xdyl[i];
          csd_sp->ftbl[ct_atm] = csd_sp->ftbl[i];
          csd_sp->atm[ct_atm].x
            = csd_sp->atm[i].x + (double)ix / (double) sd->Ncell_a;
          csd_sp->atm[ct_atm].y
            = csd_sp->atm[i].y + (double)iy / (double) sd->Ncell_b;
          csd_sp->atm[ct_atm].z
            = csd_sp->atm[i].z + (double)iz / (double) sd->Ncell_c;
          /*
          mb_fprintf(
            stdout, "$$ atom[%d]: x,y,z = %lf %lf %lf\n",
            ct_atm, csd_sp->atm[ct_atm].x,
            csd_sp->atm[ct_atm].y, csd_sp->atm[ct_atm].z 
            );
           */
          ct_atm++;
        }
      }
    }
  }
  csd_sp->n_atom = n_atm_super = ct_atm;
  
  atom_multiplicity(
    csd_sp->n_atom, csd_sp->atm, csd_sp->num_sym_op, csd_sp->sym_op
    );
  
  free_vector(num_nthatm);
  free_matrix(pos);
  
  /* low-order g */
  csd_sp->num_lofg = 0;	/* Ignore the parameters for low-order structure factors */
  
  /* beam search */
  free_vector(sd->reflex);
  bv = bravai(csd_sp->lp);
  sd->zolzbase2 = zolz_ort_base(sd->incidence, sd->zolzbase1, bv);
  
  g_exact = new_vector(HKL, sd->beam_search.nglim);  chk_vector(g_exact);
  g_bethe = new_vector(HKL, sd->beam_search.nglim);  chk_vector(g_bethe);
  mb_fprintf(stdout, "Beam search (sg)... ");  mb_fflush(stdout);
  simple_beam_search_sg(
    &(sd->num_exact), &(sd->num_bethe), g_exact, g_bethe,
    sd->beam_search, sd->incidence, sd->zolzbase1, sd->H_tilt, sd->G_tilt,
    sd->accel_kvol, csd_sp
    );
  mb_fprintf(stdout, "Exact = %d, Bethe = %d\n", sd->num_exact, sd->num_bethe);
  sd->reflex = new_vector(HKL, sd->num_exact + sd->num_bethe);
  chk_vector(sd->reflex);
  beam_merge(sd->num_exact, sd->num_bethe, g_exact, g_bethe, sd->reflex);
  free_vector(g_bethe);
  free_vector(g_exact);
  
  
  /* num_g, calc_g */
  free_vector(sd->calc_g);
  sd->calc_g = new_vector(int, sd->num_exact);	chk_vector(sd->calc_g);
  sd->num_g = sd->num_exact;
  for(i=0; i<sd->num_g; i++){
    sd->calc_g[i] = i;
  }
}


void mb_sim_spcell( FILE *fi, char *infile )
{
  CRYSTAL_STRCT_DATA csd, csd_sp;
  int nr;
  complex **vg;
  SIM_DATA sd;
  CELL_PAR bv;
  
  epsilon.vg_select_flag = 0;	/* No vg selection for each tilt */
  
  /* ----- Data read, array alloc. and beam search ----- */
  input_csd_cstr(&csd, fi, infile);    /* crystal structure data */
  input_sim_data(&sd, &csd, fi);
  input_sim_spcell_data(&sd, &csd, fi);
  fclose(fi);

  /* Convert a normal cell to a super cell. */				/* <<<<< */
  convert_spcell(&csd, &csd_sp, &sd);
  bv = bravai(csd_sp.lp);
  zolz_ort_base_two_uvw( sd.incidence, sd.zolzbase1, bv, &(sd.x1), &(sd.x2) );
  mb_fprintf(stdout, "real-space base-vector ( length = 1/|zolzbase1| )\n");
  mb_fprintf(
    stdout, "x1 = %lf %lf %lf, x2 = %lf %lf %lf\n",
    sd.x1.u, sd.x1.v, sd.x1.w, sd.x2.u, sd.x2.v, sd.x2.w
    );
  
  /* data chk */  /* <<<<< */
  if( sd.calc_type != 0){
    mb_fprintf(stdout, "$$ calc_type must be 0 for the SPCELL mode. Abort.\n");
    exit(EXIT_FAILURE);
  }
  
  /* ----- Potential matrix ----- */
  mb_fprintf(stdout, "Potential :");  mb_fflush(stdout);
  nr = sd.num_exact + sd.num_bethe;
  vg = new_matrix(complex, nr, nr);	chk_matrix(vg);
  potential_matrix_csd(
    sd.num_exact, sd.num_bethe, sd.reflex, &csd_sp, sd.accel_kvol, bv, vg
    );
  debug_vg_dump( vg, sd.num_exact + sd.num_bethe );
  
  /* debug */
  dbg_dump_ini_sim();
  dbg_sel_vg_init(sd.num_exact, sd.num_bethe);
  
  /* ----- Intensity calculation ----- */
  if(sd.calc_type != 0){
    mb_fprintf(stdout, "$$ invalid calc_type=%d. Abort.\n", sd.calc_type);
    exit(EXIT_FAILURE);
  }
  calc_cbed_pattern_spcell(&csd_sp, &sd, vg, infile);	/* Normal */
  
  dbg_dump_term();
  
  free_matrix(vg);
  free_sd(&sd);
  free_csd(&csd_sp);
  free_csd(&csd);
}
#endif


/* ************************************************* */
/* *****   DISPSURF part: Dispersion surface   ***** */
/* ************************************************* */
void input_dispsurf_data(SIM_DATA *sd, CRYSTAL_STRCT_DATA *csd, FILE *fi)
{
  /*
    Data read & array allocation
   */
  int i;
  char buf[BUFLEN], message[BUFLEN];
  CELL_PAR	bv;

  bv = bravai(csd->lp);

  chk_data("DISPSURF_CALCULATION_CONDITION{", fi);

  /* INCIDENT_BEAM block */
  chk_data("INCIDENT_BEAM{", fi);
  sd->incidence = fget_uvw(fi);
  sd->surf_norm = fget_uvw(fi);
  sd->zolzbase1 = fget_r_hkl(fi);
  sd->H_tilt = fget_d(fi);
  sd->G_tilt = fget_d(fi);
  sd->accel_kvol = fget_d(fi);
  chk_data("}INCIDENT_BEAM", fi);
  
  /* CALC_KXY_RANGE block */
  /*chk_data("CALC_KXY_RANGE{", fi);*/
  sprintf(
    message, "$$$$$ Caution!!! CALC_KXY_RANGE must be before BEAM_SEARCH. $$$$$"
    );
  chk_data_info("CALC_KXY_RANGE{", message, fi);
  sd->calc_range_flag = fget_i(fi);
  sd->calc_type = fget_i(fi);
  if(
    sd->calc_type == 0 || sd->calc_type == 1 ||
    sd->calc_type == 2 || sd->calc_type == 3
    ){
    sd->num_points = fget_i(fi);
    fget_2d( &(sd->H_begin), &(sd->H_end), fi );
    fget_2d( &(sd->G_begin), &(sd->G_end), fi );
  }
  else{
    mb_fprintf(stdout, "$$ Invalid calc_type=%d. Abort.\n", sd->calc_type);
    exit(EXIT_FAILURE);
  }
  chk_data("}CALC_KXY_RANGE", fi);
  sd->beam_search.disksize
    = eval_disksize( sd->H_begin, sd->H_end, sd->G_begin, sd->G_end );
  mb_fprintf(stdout, "$$disksize = %f\n", sd->beam_search.disksize);

  /* BEAM_SEARCH block */
  chk_data("BEAM_SEARCH{", fi);
  fget_sd_beams( sd, csd, fi );
  reflex_dump_csd(
    0, sd->num_exact, sd->num_bethe, sd->reflex,
    csd, sd->beam_search, sd->incidence, sd->zolzbase1, sd->accel_kvol, bv
    );
  chk_data("}BEAM_SEARCH", fi);

  /* Output */
  chk_data("OUTPUT_BRANCH{", fi);
  fget_2i(&(sd->branch_begin), &(sd->branch_end), fi);
  if (
    sd->branch_begin < 1 || sd->branch_begin > sd->num_exact
    || sd->branch_end < 1 || sd->branch_end > sd->num_exact
    || sd->branch_begin > sd->branch_end
    ){
    mb_fprintf(
      stdout,
      "$$ Invalid paramter: branch_begin = %d, branch_end = %d. Abort.\n",
      sd->branch_begin, sd->branch_end
      );
    exit(EXIT_FAILURE);
  }
  chk_data("}OUTPUT_BRANCH", fi);
  
  /*
  sd->num_g = fget_i(fi);
  mb_fprintf(stdout, "num_g = %d\n", sd->num_g);
  sd->calc_g = new_vector(int, sd->num_g + 1);	chk_vector(sd->calc_g);
  if(
    sd->calc_type == 0 || sd->calc_type == 1 || sd->calc_type == 2
    || sd->calc_type == 3 || sd->calc_type == 4 || sd->calc_type == 5
    ){
    fget_cal_g(
      sd->num_g, sd->num_exact, sd->num_bethe, sd->calc_g, sd->reflex, fi
      );
  }
  else{
    mb_fprintf(
      stdout,"$$ invalid calc_type=%d. Abort. (input_disp_surf_data)\n",
      sd->calc_type
      );
    exit(EXIT_FAILURE);
  }

  sd->intensity = new_tensor(double, 1, 1, 1);
  
  sd->num_thickness = fget_i(fi);
  mb_fprintf(stdout, "num_thickness = %d\n", sd->num_thickness);
  sd->thickness = new_vector(double, sd->num_thickness + 1);
  chk_vector(sd->thickness);
  fget_cal_thick(sd->num_thickness, sd->thickness, fi);
   */

  /* OUTPUT_REFLEX block */
  chk_data("OUTPUT_REFLEX{", fi);
  sd->num_g = fget_i(fi);  /* number of reflections */
  mb_fprintf(stdout, "num_g = %d\n", sd->num_g);
  sd->calc_g = new_vector(int, sd->num_g);
  chk_vector(sd->calc_g);
  if(
    sd->calc_type == 0 || sd->calc_type == 1 ||
    sd->calc_type == 2 || sd->calc_type == 3
    ){
    fget_cal_g(
      sd->num_g, sd->num_exact, sd->num_bethe, sd->calc_g, sd->reflex, fi
      );
  }
  else{
    mb_fprintf(
      stdout, "$$ invalid calc_type=%d. Abort. (input_sim_data)\n",
      sd->calc_type
      );
    exit(EXIT_FAILURE);
  }
  chk_data("}OUTPUT_REFLEX", fi);

  sd->intensity = new_tensor(double, 1, 1, 1);

  chk_data("OUTPUT_THICKNESS{", fi);
  sd->num_thickness = fget_i(fi);
  mb_fprintf(stdout, "num_thickness = %d\n", sd->num_thickness);
  sd->thickness = new_vector(double, sd->num_thickness);
  chk_vector(sd->thickness);
  fget_cal_thick(sd->num_thickness, sd->thickness, fi);
  chk_data("}OUTPUT_THICKNESS", fi);

  chk_data("}DISPSURF_CALCULATION_CONDITION", fi);
  fclose(fi);
}


void mb_dispersion_surface( FILE *fi, FILE *fo_log, char *infile )
{
  /*
    Dispersion surface
   */
  CRYSTAL_STRCT_DATA csd;
  SIM_DATA sd;
  CELL_PAR bv;
  int		nr;
  double	abs_K;
  complex **vg;

  input_csd_cstr(&csd, fi, infile);    /* crystal structure data */
  input_dispsurf_data( &sd, &csd, fi );
  bv = bravai(csd.lp);

  abs_K = calc_abs_K( &csd, sd.accel_kvol, bv );

  /* ----- Potential ----- */
  nr = sd.num_exact + sd.num_bethe;
  vg = new_matrix(complex, nr, nr);  chk_matrix(vg);

  mb_fprintf(stdout, "Potential.\n");
  potential_matrix_csd(
    sd.num_exact, sd.num_bethe, sd.reflex, &csd, sd.accel_kvol, bv, vg
    );
  /* debug */ debug_vg_dump( vg, sd.num_exact + sd.num_bethe );

  /* ----- Incident directions ----- */
  calc_kxy_tilt( &sd, bv );
  mb_fprintf(stdout, "$$ num_k_points = %d\n", sd.num_k_points);

  /* ----- Dispersion surface ----- */
  calc_disp_surf_ug(
    sd.num_exact, sd.num_bethe, sd.num_k_points,
    sd.accel_kvol, csd.lp, bv, sd.H, sd.G, 
    sd.k_xy, abs_K, sd.incidence, sd.surf_norm, 
    sd.reflex, vg, sd.branch_begin, sd.branch_end,
    sd.calc_g, sd.num_g, sd.beam_search, sd.num_thickness, sd.thickness,
    infile
    );

  free_matrix(vg);
  free_sd(&sd);
  free_csd(&csd);
}


/* ******************************************** */
/* *****     BLOCH part : Bloch states    ***** */
/* ******************************************** */
void input_bloch_data(SIM_DATA *sd, CRYSTAL_STRCT_DATA *csd, FILE *fi)
{
  char  buf[BUFLEN];
  char  message[BUFLEN];
  CELL_PAR  bv;

  bv = bravai(csd->lp);

  chk_data("BLOCH_CALCULATION_CONDITION{", fi);

  /* INCIDENT_BEAM block */
  chk_data("INCIDENT_BEAM{", fi);
  sd->incidence = fget_uvw(fi);
  sd->surf_norm = fget_uvw(fi);
  sd->zolzbase1 = fget_r_hkl(fi);
  sd->H_tilt = fget_d(fi);
  sd->G_tilt = fget_d(fi);
  sd->accel_kvol = fget_d(fi);
  chk_data("}INCIDENT_BEAM", fi);
  
  /* CALC_KXY_RANGE block */
  /*chk_data("CALC_KXY_RANGE{", fi);*/
  sprintf(
    message, "$$$$$ Caution!!! CALC_KXY_RANGE must be before BEAM_SEARCH. $$$$$"
    );
  chk_data_info("CALC_KXY_RANGE{", message, fi);
  sd->calc_range_flag = fget_i(fi);
  sd->calc_type = fget_i(fi);
  if(
    sd->calc_type == 0 || sd->calc_type == 1 ||
    sd->calc_type == 2 || sd->calc_type == 3
    ){
    sd->num_points = fget_i(fi);
    fget_2d( &(sd->H_begin), &(sd->H_end), fi );
    fget_2d( &(sd->G_begin), &(sd->G_end), fi );
  }
  else{
    mb_fprintf(stdout, "$$ Invalid calc_type=%d. Abort.\n", sd->calc_type);
    exit(EXIT_FAILURE);
  }
  chk_data("}CALC_KXY_RANGE", fi);
  sd->beam_search.disksize
    = eval_disksize( sd->H_begin, sd->H_end, sd->G_begin, sd->G_end );
  mb_fprintf(stdout, "$$disksize = %f\n", sd->beam_search.disksize);

  /* BEAM_SEARCH block */
  chk_data("BEAM_SEARCH{", fi);
  fget_sd_beams( sd, csd, fi );
  reflex_dump_csd(
    0, sd->num_exact, sd->num_bethe, sd->reflex,
    csd, sd->beam_search, sd->incidence, sd->zolzbase1, sd->accel_kvol, bv
    );
  chk_data("}BEAM_SEARCH", fi);

  chk_data("CALC_RXY_RANGE{", fi);
  zolz_ort_base_two_uvw(
    sd->incidence, sd->zolzbase1, bv, &(sd->x1), &(sd->x2)
    );
  mb_fprintf(stdout, "real-space base-vector ( length = 1/|zolzbase1| )\n");
  mb_fprintf(
    stdout, "x1 = %lf %lf %lf, x2 = %lf %lf %lf\n",
    sd->x1.u, sd->x1.v, sd->x1.w, sd->x2.u, sd->x2.v, sd->x2.w
    );
  fget_2i(&(sd->num_points_x), &(sd->num_points_y), fi);
  fget_2d(&(sd->x_begin), &(sd->x_end), fi);
  fget_2d(&(sd->y_begin), &(sd->y_end), fi);
  chk_data("}CALC_RXY_RANGE", fi);

/* Branches */
  chk_data("OUTPUT_BRANCH{", fi);
  fget_2i(&(sd->branch_begin), &(sd->branch_end), fi);
  if (
    sd->branch_begin < 1 || sd->branch_begin > sd->num_exact
    || sd->branch_end < 1 || sd->branch_end > sd->num_exact
    || sd->branch_begin > sd->branch_end
    ){
    mb_fprintf(
      stdout, "$$ Invalid paramter: branch_begin = %d, branch_end = %d\n",
      sd->branch_begin, sd->branch_end
      );
    exit(EXIT_FAILURE);
  }
  chk_data("}OUTPUT_BRANCH", fi);
  chk_data("}BLOCH_CALCULATION_CONDITION", fi);
  fclose(fi);

  /* dummy array for free_sd */
  sd->num_thickness = 0;
  sd->intensity = new_tensor(double, 1, 1, 1);
  chk_tensor(sd->intensity);
}


void mb_bloch_state( FILE *fi, FILE *fo_log, char *infile )
{
  /*
    Bloch state
   */
  CRYSTAL_STRCT_DATA csd;
  SIM_DATA sd;
  CELL_PAR bv;
  int      nr;
  double   abs_K;
  complex  **vg;

  input_csd_cstr(&csd, fi, infile);    /* crystal structure data */
  input_bloch_data(&sd, &csd, fi);
  bv = bravai(csd.lp);

  /* ----- Potential ----- */
  nr = sd.num_exact + sd.num_bethe;
  vg = new_matrix(complex, nr, nr);  chk_matrix(vg);
  mb_fprintf(stdout, "Potential calc.\n");
  potential_matrix_csd(
    sd.num_exact, sd.num_bethe, sd.reflex, &csd, sd.accel_kvol, bv, vg
    );
  debug_vg_dump( vg, sd.num_exact + sd.num_bethe );

  abs_K = calc_abs_K( &csd, sd.accel_kvol, bv );
  calc_kxy_tilt( &sd, bv ); /* Incident direction */

  /* Bloch states */
  sd.num_g = 0;
  calc_bloch_state_ug(
    sd.accel_kvol, csd.lp, bv, sd.k_xy[0],
    abs_K, sd.incidence, sd.surf_norm, 
    sd.num_exact, sd.num_bethe, sd.reflex, vg, sd.beam_search,
    sd.x_begin, sd.x_end, sd.y_begin, sd.y_end, sd.x1, sd.x2,
    sd.num_points_x, sd.num_points_y, sd.branch_begin, sd.branch_end,
    infile, fo_log
    );

  free_matrix(vg);
  free_sd(&sd);
  free_csd(&csd);
}

/* ------------------------------------------------------------------------ */
/* ------------------------------------------------------------------------ */
/* TFRINGE_MODE */
void input_tfringe_data(SIM_DATA *sd, CRYSTAL_STRCT_DATA *csd, FILE *fi)
{
  int  i;
  double  t_begin, t_end;
  char  buf[BUFLEN];
  char  message[BUFLEN];
  CELL_PAR	bv;

  bv = bravai(csd->lp);

  chk_data("TFRINGE_CALCULATION_CONDITION{", fi);

  /* INCIDENT_BEAM block */
  chk_data("INCIDENT_BEAM{", fi);
  sd->incidence = fget_uvw(fi);
  sd->surf_norm = fget_uvw(fi);
  sd->zolzbase1 = fget_r_hkl(fi);
  sd->H_tilt = fget_d(fi);
  sd->G_tilt = fget_d(fi);
  sd->accel_kvol = fget_d(fi);
  chk_data("}INCIDENT_BEAM", fi);
  
  /* CALC_KXY_RANGE block */
  /*chk_data("CALC_KXY_RANGE{", fi);*/
  sprintf(
    message, "$$$$$ Caution!!! CALC_KXY_RANGE must be before BEAM_SEARCH. $$$$$"
    );
  chk_data_info("CALC_KXY_RANGE{", message, fi);
  sd->calc_range_flag = fget_i(fi);
  sd->calc_type = fget_i(fi);
  if(
    sd->calc_type == 0 || sd->calc_type == 1 ||
    sd->calc_type == 2 || sd->calc_type == 3
    ){
    sd->num_points = fget_i(fi);
    fget_2d( &(sd->H_begin), &(sd->H_end), fi );
    fget_2d( &(sd->G_begin), &(sd->G_end), fi );
  }
  else{
    mb_fprintf(stdout, "$$ Invalid calc_type=%d. Abort.\n", sd->calc_type);
    exit(EXIT_FAILURE);
  }
  chk_data("}CALC_KXY_RANGE", fi);
  sd->beam_search.disksize
    = eval_disksize( sd->H_begin, sd->H_end, sd->G_begin, sd->G_end );
  mb_fprintf(stdout, "$$disksize = %f\n", sd->beam_search.disksize);

  /* BEAM_SEARCH block */
  chk_data("BEAM_SEARCH{", fi);
  fget_sd_beams( sd, csd, fi );
  reflex_dump_csd(
    0, sd->num_exact, sd->num_bethe, sd->reflex,
    csd, sd->beam_search, sd->incidence, sd->zolzbase1, sd->accel_kvol, bv
    );
  chk_data("}BEAM_SEARCH", fi);

  /* Thicknesses */
  chk_data("CALC_T_RANGE{", fi);
  sd->num_thickness = fget_i(fi);
  printf("  sd->num_thickness = %d\n",   sd->num_thickness);
  sd->thickness = new_vector(double, sd->num_thickness);
  chk_vector(sd->thickness);
  fget_2d(&t_begin, &t_end, fi);
  for(i=0; i<sd->num_thickness; i++){
    sd->thickness[i] = t_begin + (t_end - t_begin)
      * (double)i / (double) (sd->num_thickness - 1);
  }
  chk_data("}CALC_T_RANGE", fi);

  /* OUTPUT_REFLEX block */
  chk_data("OUTPUT_REFLEX{", fi);
  sd->num_g = fget_i(fi);  /* number of reflections */
  if(sd->num_g > 0){
    mb_fprintf(stdout, "num_g = %d\n", sd->num_g);
    sd->calc_g = new_vector(int, sd->num_g);
    chk_vector(sd->calc_g);
    if(
      sd->calc_type == 0 || sd->calc_type == 1 ||
      sd->calc_type == 2 || sd->calc_type == 3
      ){  /* normal */
      fget_cal_g(
        sd->num_g, sd->num_exact, sd->num_bethe, sd->calc_g, sd->reflex, fi
        );
    }
    else{
      mb_fprintf(
        stdout, "$$ invalid calc_type=%d. Abort. (input_sim_data)\n",
        sd->calc_type
        );
      exit(EXIT_FAILURE);
    }
  }
  else if(sd->num_g == 0){  /* all reflections */
    sd->num_g = sd->num_exact;
    mb_fprintf(stdout, "num_g = %d\n", sd->num_g);
    sd->calc_g = new_vector(int, sd->num_g);
    chk_vector(sd->calc_g);
    if(
      sd->calc_type == 0 || sd->calc_type == 1 ||
      sd->calc_type == 2 || sd->calc_type == 3
      ){ /* normal */
      for(i=0; i<sd->num_g; i++){ sd->calc_g[i] = i; }
    }
    else{
      mb_fprintf(
        stdout, "$$ invalid calc_type=%d. Abort. (input_sim_data)\n",
        sd->calc_type
        );
      exit(EXIT_FAILURE);
    }
  }
  chk_data("}OUTPUT_REFLEX", fi);

  chk_data("}TFRINGE_CALCULATION_CONDITION", fi);
  fclose(fi);

  /* dummy array for free_sd */
  sd->intensity = new_tensor(double, 1, 1, 1);
  chk_tensor(sd->intensity);
}


void mb_tfringe( FILE *fi, FILE *fo_log, char *infile )
{
  /*
    THICKNESS FRINGES
   */
  CRYSTAL_STRCT_DATA csd;
  SIM_DATA sd;
  CELL_PAR bv;
  int      nr;
  double   abs_K;
  complex  **vg;

  input_csd_cstr(&csd, fi, infile);    /* crystal structure data */
  input_tfringe_data(&sd, &csd, fi);
  bv = bravai(csd.lp);

  /* ----- Potential ----- */
  nr = sd.num_exact + sd.num_bethe;
  vg = new_matrix(complex, nr, nr);	chk_matrix(vg);
  mb_fprintf(stdout, "Potential.\n");
  potential_matrix_csd(
    sd.num_exact, sd.num_bethe, sd.reflex, &csd, sd.accel_kvol, bv, vg
    );
  /* debug */ debug_vg_dump( vg, sd.num_exact + sd.num_bethe );

  abs_K = calc_abs_K( &csd, sd.accel_kvol, bv );

  /* ----- Incident directions ----- */
  calc_kxy_tilt( &sd, bv );

  /* ----- Dynamical calculation ----- */
  calc_tfringe_ug(
    sd.accel_kvol, csd.lp, bv, sd.k_xy[0], 
    abs_K, sd.incidence, sd.surf_norm, 
    sd.num_exact, sd.num_bethe, sd.reflex, vg, sd.beam_search, 
    sd.num_g, sd.calc_g, sd.num_thickness, sd.thickness, infile 
    );

  free_matrix(vg);
  free_sd(&sd);
  free_csd(&csd);
}
