/*
 * rfsearch.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "mbconfig.h"
#include "fitting.h"
#include "mb_input.h"

extern DEBUG_INFO dbg;
extern EPS_CTRL	epsilon;

typedef struct{int h,k,l; double d;} HKLD;

void evaluate_hklmax(double gmax, int *hmax, int *kmax, int *lmax, CELL_PAR bv);
int  evaluate_nth_lz_max(double g_max, double abs_chi, double abs_incidence);
double excitation_error( HKL g, double g_z, R_HKL chi_proj, double abs_chi, CELL_PAR bv );
double abs_edge_s_g(
  int nth_lz, double abs_incidence,
  double abs_chi, double cbed_disksize,
  double *x, double *y, double *s_g_start, double *s_g_end,
  HKL g, UVW incidence, R_HKL zolzbase1, CELL_PAR bv
  );
double abs_edge_s_g2(
  double s_g_start, double s_g_end,
  double s_g_target_start, double s_g_target_end
  );

/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= */

int find_n_g( int nr, HKL g, HKL *reflex )
     /*
      * Find the potision of g in reflex[]
      *  return value:  nth of reflex[]
      *  ret. val. is -1: not found.
      */
{
  int i;
  
  for(i=0;i<nr;i++){
    if(g.h==reflex[i].h && g.k==reflex[i].k && g.l==reflex[i].l){
      return i;
    }
  }
  /* printf("$$ g (=%d %d %d) is not found in reflex.\n", g.h, g.k, g.l); */
  return -1;
}


void evaluate_hklmax(double gmax, int *hmax, int *kmax, int *lmax, CELL_PAR bv)
     /*
      * evaluation of max number of h, k, l
      */
{
  int i;
  R_HKL g;
  
  i=0;
  do{
    i++; g.h = (double)i; g.k = 0.0; g.l = 0.0;
  }while( gmax > sqrt(sq_abs(g,bv)) );
  *hmax = i;
  
  i=0;
  do{
    i++; g.h = 0.0; g.k = (double)i; g.l = 0.0;
  }while( gmax > sqrt(sq_abs(g,bv)) );
  *kmax = i;
  
  i=0;
  do{
    i++; g.h = 0.0; g.k = 0.0; g.l = (double)i;
  }while( gmax > sqrt(sq_abs(g,bv)) );
  *lmax = i;
}


int evaluate_nth_lz_max(double g_max, double abs_chi, double abs_incidence)
{
  return
    (int) ( ( abs_chi - sqrt( sq(abs_chi) - sq(g_max) ) )
            * abs_incidence )
      + 1;
}


#if 0
double excitation_error(
  HKL g, double g_z, R_HKL k_xy, double abs_chi, CELL_PAR bv
  )
     /*
      * calc. of excitation error sg
      *
      *  s_g = ( \chi^{2} - ( {\bf \chi} + {\bf g} )^{2} ) / ( 2 \chi )
      *      = ( - \chi_{z} g_{z} - {\bf \chi}_{proj} {\bf g}_{proj} - 0.5 g^{2} ) / \chi
      *      = ( - \chi_{z} g_{z} - {\bf \chi}_{proj} {\bf g} - 0.5 g^{2} ) / \chi
      */
{
  R_HKL t;
  double chi_z;
  
  chi_z = - sqrt( sq(abs_chi) - sq_abs(k_xy, bv) ); /* chi: downwards. */
  t.h = k_xy.h + 0.5 * (double) g.h;
  t.k = k_xy.k + 0.5 * (double) g.k;
  t.l = k_xy.l + 0.5 * (double) g.l;
  
  return ( - chi_z * g_z - inner_product(t, hkl2rhkl(g), bv) ) / abs_chi;
}
#endif


#if 0
double abs_edge_s_g(
  int nth_lz, double abs_incidence,
  double abs_chi, double cbed_disksize,
  double *x, double *y, double *s_g_start, double *s_g_end,
  HKL g, UVW incidence, R_HKL zolzbase1, CELL_PAR bv
  )
     /*
      * Absolute value of minimum excitation errors
      *  at the two edges of a CBED disk
      *  (on the line along the radial direction through the center
      *   of the CBED disks).
      */
{
  R_HKL  k_proj, g_proj;
  double g_z, s_g, tx, ty;
  
  g_z = (double) nth_lz / abs_incidence;
  g_proj = proj_to_zolz(g, incidence, zolzbase1, bv, &tx, &ty);
  g_proj = normalize_vec(g_proj, zolzbase1, bv);
  *x = tx; *y = ty;
  
  k_proj.h = cbed_disksize * 0.5 * g_proj.h;
  k_proj.k = cbed_disksize * 0.5 * g_proj.k;
  k_proj.l = cbed_disksize * 0.5 * g_proj.l;
  *s_g_start = excitation_error( g, g_z, k_proj, abs_chi, bv );
  
  k_proj.h = - cbed_disksize * 0.5 * g_proj.h;
  k_proj.k = - cbed_disksize * 0.5 * g_proj.k;
  k_proj.l = - cbed_disksize * 0.5 * g_proj.l;
  *s_g_end = excitation_error( g, g_z, k_proj, abs_chi, bv );
  
  if (*s_g_start * *s_g_end < 0.0) {
    s_g = 0.0;
  }
  else if ( fabs(*s_g_start) < fabs(*s_g_end) ){
    s_g = *s_g_start;
  }
  else{
    s_g = *s_g_end;
  }
  return fabs(s_g);	
}
#endif


double abs_edge_s_g_tilt(
  HKL g, UVW incidence, R_HKL zolzbase1, R_HKL zolzbase2,
  double H_tilt, double G_tilt,
  double abs_incidence, double abs_K, double disksize,
  double *s_g_start, double *s_g_end, CELL_PAR bv
  )
     /*
      * Absolute value of minimum excitation errors
      *  at the two edges of a CBED disk
      *  (on the line along the radial direction through the center
      *   of the CBED disks).
      */
{
  int		nth_lz;
  R_HKL  k_xy, g_xy;
  double g_z, s_g, tx, ty, k_z;
  
  nth_lz =  g.h * incidence.u + g.k * incidence.v + g.l * incidence.w;
  g_z = (double) nth_lz / abs_incidence;
  g_xy = proj_to_zolz(g, incidence, zolzbase1, bv, &tx, &ty);
  g_xy = normalize_vec(g_xy, zolzbase1, bv);
  
  /* start edge */
  k_xy.h = 0.5 * ( disksize * g_xy.h + H_tilt * zolzbase1.h + G_tilt * zolzbase2.h );
  k_xy.k = 0.5 * ( disksize * g_xy.k + H_tilt * zolzbase1.k + G_tilt * zolzbase2.k );
  k_xy.l = 0.5 * ( disksize * g_xy.l + H_tilt * zolzbase1.l + G_tilt * zolzbase2.l );
  k_z = - sqrt( sq(abs_K) - sq_abs(k_xy, bv) );
  /*  *s_g_start = exct_err( k_z, abs_incidence, abs_K, g, k_xy, incidence, bv );  */
  *s_g_start = calc_excitation_error( k_xy, k_z, abs_K, g, g_z, bv );
  
  /* end edge */
  k_xy.h = 0.5 * ( - disksize * g_xy.h + H_tilt * zolzbase1.h + G_tilt * zolzbase2.h );
  k_xy.k = 0.5 * ( - disksize * g_xy.k + H_tilt * zolzbase1.k + G_tilt * zolzbase2.k );
  k_xy.l = 0.5 * ( - disksize * g_xy.l + H_tilt * zolzbase1.l + G_tilt * zolzbase2.l );
  k_z = - sqrt( sq(abs_K) - sq_abs(k_xy, bv) );
  /* *s_g_end = exct_err( k_z, abs_incidence, abs_K, g, k_xy, incidence, bv ); */
  *s_g_end = calc_excitation_error( k_xy, k_z, abs_K, g, g_z, bv );
  
  if (*s_g_start * *s_g_end < 0.0) { s_g = 0.0; }
  else if ( fabs(*s_g_start) < fabs(*s_g_end) ){ s_g = *s_g_start; }
  else{ s_g = *s_g_end; }
  
  return fabs(s_g);	
}


#if 0
double abs_edge_s_g2(
  double s_g_start, double s_g_end,
  double s_g_target_start, double s_g_target_end
  )
{
  double s_g2, s_g2_start, s_g2_end;
  
  s_g2_start = s_g_target_start - s_g_start;
  s_g2_end   = s_g_target_end   - s_g_end;
  
  if (s_g2_start * s_g2_end < 0.0){
    s_g2 = 0.0;
  }
  else if ( fabs(s_g2_start) < fabs(s_g2_end) ){
    s_g2 = s_g2_start;
  }
  else{
    s_g2 = s_g2_end;
  }
  return fabs(s_g2);
}
#endif


double spacing(HKL t, CELL_PAR bv)
     /*
      * Lattice spacing of (h,k,l)
      */
{
  R_HKL g;
  
  if (t.h == 0 && t.k == 0 && t.l == 0){
    return 0.0;
  }
  else{
    g.h = (double) t.h; g.k = (double) t.k; g.l = (double) t.l;
    return  1.0 / sqrt( sq_abs( g, bv ) );
  }
}


int hkld_comp( const void *x, const void *y )
     /*
      * comparison for HKLD for qsort()
      */
{
  double xd, yd;
  
  xd = ((HKLD *)x)->d;
  yd = ((HKLD *)y)->d;
  /* descending order */
  if ( (yd - xd) > 0.0 )
    return 1;
  else if ( (yd - xd) < 0.0 )
    return -1;
  else
    return 0;
}


void sort_reflex(int num_reflex, HKL *g, CELL_PAR bv)
{
  int i;
  HKLD *gd;
  
  gd = new_vector(HKLD, num_reflex + 1);	chk_vector(gd);		/* +1: dummy */
  
  for(i=0;i<num_reflex;i++){
    gd[i].h = g[i].h; gd[i].k = g[i].k; gd[i].l = g[i].l;
    if ( g[i].h == 0 && g[i].k == 0 && g[i].l == 0 ){
      gd[i].d = 1e+10;
    }
    else{
      gd[i].d = spacing( g[i], bv );
    }
  }
  qsort( gd, num_reflex, sizeof(HKLD), hkld_comp );
  for(i=0;i<num_reflex;i++){
    g[i].h = gd[i].h; g[i].k = gd[i].k; g[i].l = gd[i].l;
  }
  free_vector(gd);
}


#if 0
int reflex_search(
  int ind, int *num_g_exact, int *num_g_bethe, HKL *g_exact, HKL *g_bethe,
  LAT_PAR lp, char *lattice_type, int num_sym_op, SYM_OP *sym_op,
  int n_atom, ATOM *atm, XDT *xdyl,
  int num_low_order_g, HKL *g_low_order, complex *fg_low_order,
  double accel_kvol,
  BEAM_SEARCH_PAR beam_search, UVW incidence, R_HKL zolzbase1,
  HKL g_target, int g_table_size
  )
     /*
      * Search excited reflections:
      *   in accordance with w_g, w_g2, xi_g, xi_g2, lz_min and lz_max.
      *   The reflections found are divided into "exact" and "Bethe".
      *   Return value is i_g_target.
      *
      * ind<0: initialization
      * ind>0: calculation
      * ind==0: free & termination
      *
      */
{
  int h,k,l,cte, ctb, nth_lz, i_g_target=0;
  int hmax, kmax, lmax;
  double abs_s_g, s_g_start, s_g_end, abs_s_g2;
  double s_g_target_start, s_g_target_end;
  double abs_w_g, abs_w_g2;
  double tx, ty;
  double abs_chi, abs_g, xi_g, xi_g2;
  double abs_incidence;
  complex fg;
  R_HKL  rg;
  HKL g, g2;
  static HKL *g_table;		/* <<<<< */
  static complex *fg_table;	/* <<<<< */
  static int ct_fg_table;		/* <<<<< */
  CELL_PAR bv;
  
  if(ind < 0){	/* initialization */
    fg_table = new_vector(complex, g_table_size);	chk_vector(fg_table);
    g_table  = new_vector(HKL, g_table_size);		chk_vector(g_table);
    str_fct(ind = -1, g, lattice_type, num_sym_op, sym_op,
            n_atom, atm, xdyl, num_low_order_g, g_low_order, fg_low_order,
            &ct_fg_table, g_table, fg_table, g_table_size, bv);
    return 0;
  }
  else if( ind ==0 ){		/* free & termination */
    /* Free */
    str_fct(ind = 0, g, lattice_type, num_sym_op, sym_op,
            n_atom, atm, xdyl, num_low_order_g, g_low_order, fg_low_order,
            &ct_fg_table, g_table, fg_table, g_table_size, bv);
    free_vector(fg_table); free_vector(g_table);
    return 0;
  }
  
  bv = bravai(lp);
  abs_incidence = abs_uvw(incidence, lp);
  abs_chi = 1.0 / wavelength(accel_kvol);
  evaluate_hklmax(beam_search.gmax,&hmax,&kmax,&lmax,bv);
  cte = 0; ctb = 0; /* Counters for reflections exact and Bethe */
  g_exact[0].h = 0; g_exact[0].k = 0; g_exact[0].l = 0; /* g_exact[0] must be 000 */
  cte = 1;
  /* for g_target */
  nth_lz = g_target.h * incidence.u + g_target.k * incidence.v + g_target.l * incidence.w;
  
  abs_edge_s_g( nth_lz, abs_incidence, abs_chi, beam_search.disksize,
                &tx, &ty, &s_g_target_start, &s_g_target_end,
                g_target, incidence, zolzbase1, bv );
  /*printf("s_g_target_start=%lf, s_g_target_end=%lf\n", s_g_target_start, s_g_target_end);*/
  
  for(h = - hmax; h <= hmax; h++){
    for(k = - kmax; k <= kmax; k++){
      for(l = - lmax; l <= lmax; l++){
        
        g.h = h; g.k = k; g.l = l;
        rg.h = (double) h; rg.k = (double) k; rg.l = (double) l;
        nth_lz =  g.h * incidence.u + g.k * incidence.v + g.l * incidence.w;
        abs_g = sqrt( sq_abs(rg, bv) );
        
        if (   ( h!=0 || k!=0 || l!=0 )
               && ( ( beam_search.lzmin <= nth_lz ) && ( nth_lz <= beam_search.lzmax ) )
               && (  abs_g < beam_search.gmax )
               && ( extinction_rule(g,lattice_type) == 0 )   ){
          
          /* For 1st order diffraction (g) */
          fg = str_fct(ind = 1, g, lattice_type, num_sym_op, sym_op,
                       n_atom, atm, xdyl, num_low_order_g, g_low_order, fg_low_order,
                       &ct_fg_table, g_table, fg_table, g_table_size, bv);
          /* printf("F(%d %d %d)=(%lf,%lf)\n",g.h,g.k,g.l,fg.re,fg.im); */
          xi_g = fg_to_xig( fg, accel_kvol, bv );
          
          if ( g_target.h==0 && g_target.k==0 && g_target.l==0 ){
            /* only 1st order diffraction has to be considered. */
            if ( (xi_g < beam_search.xigmax) ){		/* Condition for excitation */
              nth_lz =  g.h * incidence.u + g.k * incidence.v + g.l * incidence.w;
              abs_s_g = abs_edge_s_g( nth_lz, abs_incidence, abs_chi, beam_search.disksize,
                                      &tx, &ty, &s_g_start, &s_g_end, g, incidence, zolzbase1, bv );
              abs_w_g  = abs_s_g * xi_g;
              /* printf("g=%d %d %d, s_g_start=%lf, s_g_end=%lf\n",
						g.h, g.k, g.l, s_g_start, s_g_end);*/
              if ( abs_w_g < beam_search.wgbmin ){
                g_exact[cte].h  = h; g_exact[cte].k = k; g_exact[cte].l = l;/* Exact */
                cte++;
                /* printf("%d\n",cte); */
              }
              else if ( abs_w_g < beam_search.wgmax ){
                g_bethe[ctb].h  = h; g_bethe[ctb].k = k; g_bethe[ctb].l = l;/* Bethe */
                ctb++;
                /* printf("%d B\n",ctb); */
              }
            }
            i_g_target = 0;		/* 000 beam */
          }
          else{			/* Include 2nd order diffraction (g2) */
            g2.h = g_target.h - g.h;
            g2.k = g_target.k - g.k;
            g2.l = g_target.l - g.l;
            if ( g2.h == 0 && g2.k == 0 && g2.l == 0 ){
              xi_g2 = 0.0;	/* Dummy */
            }
            else{
              fg = str_fct(ind = 1, g2, lattice_type, num_sym_op, sym_op,
                           n_atom, atm, xdyl, num_low_order_g, g_low_order, fg_low_order,
                           &ct_fg_table, g_table, fg_table, g_table_size, bv); /* <<<<< */
              xi_g2 = fg_to_xig( fg, accel_kvol, bv );
            }
            /* printf("#DEBUG:g2=%d %d %d, xi_g2=%lf\n",g2.h,g2.k,g2.l,xi_g2); */
            /* Condition for excitation */
            if ( (xi_g < beam_search.xigmax) && (xi_g2 < beam_search.xig2max) ){
              nth_lz =  g.h * incidence.u + g.k * incidence.v + g.l * incidence.w; /* g */
              abs_s_g = abs_edge_s_g( nth_lz, abs_incidence, abs_chi, beam_search.disksize,
                                      &tx, &ty, &s_g_start, &s_g_end, g, incidence, zolzbase1, bv );
              abs_w_g  = abs_s_g  * xi_g;
              /* printf("s_g_start=%lf, s_g_end=%lf\n", s_g_start, s_g_end);*/
              /* printf("#DEBUG:g=%d %d %d, abs_s_g=%lf, abs_w_g=%lf\n",
						g.h,g.k,g.l, abs_s_g, abs_w_g);*/
              abs_s_g2 = abs_edge_s_g2( s_g_start, s_g_end,
                                        s_g_target_start, s_g_target_end );		/* g2 */
              abs_w_g2 = abs_s_g2 * xi_g2;
              /* printf("#DEBUG:g2=%d %d %d, abs_s_g2=%lf, abs_w_g2=%lf\n",
						g2.h,g2.k,g2.l, abs_s_g2, abs_w_g2);*/
              if ( (abs_w_g < beam_search.wgbmin) && (abs_w_g2 < beam_search.wg2bmin) ){
                if(g_target.h==h && g_target.k==k && g_target.l==l){
                  i_g_target = cte + 1;
                }
                g_exact[cte].h  = h; g_exact[cte].k = k; g_exact[cte].l = l; /* Exact */
                cte++;
                /* printf("%d\n",cte); */
              }
              else if ( (abs_w_g < beam_search.wgmax) && (abs_w_g2 < beam_search.wg2max) ){
                g_bethe[ctb].h  = h; g_bethe[ctb].k = k; g_bethe[ctb].l = l; /* Bethe */
                ctb++;
                /* printf("%d B\n",ctb); */
              }
            }
          }
          
          if ( (cte >= beam_search.nglim) || (ctb >= beam_search.nglim) ){
            printf("$$ num_g has exceeded num_g_max.\n");
            printf("$$ Please set a greater value into num_g. Abort.\n");
            exit(EXIT_FAILURE);
          }
        }
        
      }
    }
  }
LOOP_EXIT:
  *num_g_exact = cte;
  *num_g_bethe = ctb;
  sort_reflex(*num_g_exact, g_exact, bv);
  sort_reflex(*num_g_bethe, g_bethe, bv);
  return i_g_target;
}


int reflex_search_s(
  int ind, int *num_g_exact, int *num_g_bethe, HKL *g_exact, HKL *g_bethe,
  LAT_PAR lp, char *lattice_type, int num_sym_op, SYM_OP *sym_op,
  int n_atom, ATOM *atm, XDT *xdyl,
  int num_low_order_g, HKL *g_low_order, complex *fg_low_order,
  double accel_kvol,
  BEAM_SEARCH_PAR beam_search, UVW incidence, R_HKL zolzbase1,
  HKL g_target, int g_table_size
  )
     /*
      * Search excited reflections:
      *   in accordance with w_g, w_g2, xi_g, xi_g2, lz_min and lz_max.
      *   special for ZOLZ reflections.
      *   The reflections found are divided into "exact" and "Bethe".
      *   Return value is i_g_target.
      */
{
  int h,k,l,cte, ctb, nth_lz, i_g_target=0;
  int hmax, kmax, lmax;
  double abs_s_g, s_g_start, s_g_end, abs_s_g2;
  double s_g_target_start, s_g_target_end;
  double abs_w_g, abs_w_g2;
  double tx, ty;
  double abs_chi, abs_g, xi_g, xi_g2;
  double abs_incidence;
  complex fg;
  R_HKL  rg;
  HKL g, g2;
  static HKL *g_table;		/* <<<<< */
  static complex *fg_table;	/* <<<<< */
  static int ct_fg_table;		/* <<<<< */
  CELL_PAR bv;
  
  if(ind < 0){	/* initialization */
    fg_table = new_vector(complex, g_table_size);	chk_vector(fg_table);
    g_table  = new_vector(HKL, g_table_size);		chk_vector(g_table);
    str_fct(ind = -1, g, lattice_type, num_sym_op, sym_op,
            n_atom, atm, xdyl, num_low_order_g, g_low_order, fg_low_order,
            &ct_fg_table, g_table, fg_table, g_table_size, bv);
    return 0;
  }
  else if( ind ==0 ){		/* free & termination */
    str_fct(ind = 0, g, lattice_type, num_sym_op, sym_op,
            n_atom, atm, xdyl, num_low_order_g, g_low_order, fg_low_order,
            &ct_fg_table, g_table, fg_table, g_table_size, bv);
    free_vector(fg_table); free_vector(g_table);
    return 0;
  }
  
  bv = bravai(lp);
  abs_incidence = abs_uvw(incidence, lp);
  abs_chi = 1.0 / wavelength(accel_kvol);
  evaluate_hklmax(beam_search.gmax,&hmax,&kmax,&lmax,bv);
  cte = 0; ctb = 0; /* Counters for reflections exact and Bethe */
  g_exact[0].h = 0; g_exact[0].k = 0; g_exact[0].l = 0; /* g_exact[0] must be 000 */
  cte = 1;
  /* for g_target */
  nth_lz = g_target.h * incidence.u + g_target.k * incidence.v + g_target.l * incidence.w;
  
  abs_edge_s_g( nth_lz, abs_incidence, abs_chi, beam_search.disksize,
                &tx, &ty, &s_g_target_start, &s_g_target_end,
                g_target, incidence, zolzbase1, bv );
  /*printf("s_g_target_start=%lf, s_g_target_end=%lf\n", s_g_target_start, s_g_target_end);*/
  
  for(h = - hmax; h <= hmax; h++){
    for(k = - kmax; k <= kmax; k++){
      for(l = - lmax; l <= lmax; l++){
        
        g.h = h; g.k = k; g.l = l;
        rg.h = (double) h; rg.k = (double) k; rg.l = (double) l;
        nth_lz =  g.h * incidence.u + g.k * incidence.v + g.l * incidence.w;
        abs_g = sqrt( sq_abs(rg, bv) );
        
        if (   ( h!=0 || k!=0 || l!=0 )
               && ( ( beam_search.lzmin <= nth_lz ) && ( nth_lz <= beam_search.lzmax ) )
               && (  abs_g < beam_search.gmax )
               && ( extinction_rule(g,lattice_type) == 0 )   ){
          /* For 1st order diffraction (g) */
          fg = str_fct(ind = 1, g, lattice_type, num_sym_op, sym_op,
                       n_atom, atm, xdyl, num_low_order_g, g_low_order, fg_low_order,
                       &ct_fg_table, g_table, fg_table, g_table_size, bv);
          /* printf("F(%d %d %d)=(%lf,%lf)\n",g.h,g.k,g.l,fg.re,fg.im); */
          xi_g = fg_to_xig( fg, accel_kvol, bv );
          if ( g_target.h==0 && g_target.k==0 && g_target.l==0 ){
            /* only 1st order diffraction has to be considered. */
            if ( (xi_g < beam_search.xigmax) ){		/* Condition for excitation */
              nth_lz =  g.h * incidence.u + g.k * incidence.v + g.l * incidence.w;
              abs_s_g = abs_edge_s_g( nth_lz, abs_incidence, abs_chi, beam_search.disksize,
                                      &tx, &ty, &s_g_start, &s_g_end, g, incidence, zolzbase1, bv );
              abs_w_g  = abs_s_g * xi_g;
              /* printf("g=%d %d %d, s_g_start=%lf, s_g_end=%lf\n",
						g.h, g.k, g.l, s_g_start, s_g_end);*/
              if ( abs_w_g < beam_search.wgbmin ){	/* Exact */
                g_exact[cte].h  = h; g_exact[cte].k = k; g_exact[cte].l = l;
                cte++;	/* printf("%d\n",cte); */
              }
              else if ( abs_w_g < beam_search.wgmax ){	/* Bethe */
                g_bethe[ctb].h  = h; g_bethe[ctb].k = k; g_bethe[ctb].l = l;
                ctb++;	/* printf("%d B\n",ctb); */
              }
            }
            i_g_target = 0;	/* 000 beam */
          }
          else{			/* Include 2nd order diffraction (g2) */
            g2.h = g_target.h - g.h;
            g2.k = g_target.k - g.k;
            g2.l = g_target.l - g.l;
            if ( g2.h == 0 && g2.k == 0 && g2.l == 0 ){
              xi_g2 = 0.0;	/* Dummy */
            }
            else{
              fg = str_fct(ind = 1, g2, lattice_type, num_sym_op, sym_op,
                           n_atom, atm, xdyl, num_low_order_g, g_low_order, fg_low_order,
                           &ct_fg_table, g_table, fg_table, g_table_size, bv); /* <<<<< */
              xi_g2 = fg_to_xig( fg, accel_kvol, bv );
            }
            nth_lz =  g.h * incidence.u + g.k * incidence.v + g.l * incidence.w;
            if ( (nth_lz == 0) && (xi_g < beam_search.xigmax) ){ /* only for ZOLZ */
              nth_lz =  g.h * incidence.u + g.k * incidence.v + g.l * incidence.w;
              abs_s_g = abs_edge_s_g( nth_lz, abs_incidence, abs_chi, beam_search.disksize,
                                      &tx, &ty, &s_g_start, &s_g_end, g, incidence, zolzbase1, bv );
              abs_w_g  = abs_s_g * xi_g;
              /* printf("g=%d %d %d, s_g_start=%lf, s_g_end=%lf\n",
						g.h, g.k, g.l, s_g_start, s_g_end);*/
              if ( abs_w_g < beam_search.wgbmin ){	/* Exact */
                g_exact[cte].h  = h; g_exact[cte].k = k; g_exact[cte].l = l;
                cte++;
              }
              else if ( abs_w_g < beam_search.wgmax ){	/* Bethe */
                g_bethe[ctb].h  = h; g_bethe[ctb].k = k; g_bethe[ctb].l = l;
                ctb++;
              }
            }
            else if ( (xi_g < beam_search.xigmax) && (xi_g2 < beam_search.xig2max) ){
              nth_lz =  g.h * incidence.u + g.k * incidence.v + g.l * incidence.w; /* g */
              abs_s_g = abs_edge_s_g( nth_lz, abs_incidence, abs_chi, beam_search.disksize,
                                      &tx, &ty, &s_g_start, &s_g_end, g, incidence, zolzbase1, bv );
              abs_w_g  = abs_s_g  * xi_g;
              /* printf("s_g_start=%lf, s_g_end=%lf\n", s_g_start, s_g_end);*/
              /* printf("#DEBUG:g=%d %d %d, abs_s_g=%lf, abs_w_g=%lf\n",
						g.h,g.k,g.l, abs_s_g, abs_w_g);*/
              abs_s_g2 = abs_edge_s_g2( s_g_start, s_g_end,
                                        s_g_target_start, s_g_target_end );		/* g2 */
              abs_w_g2 = abs_s_g2 * xi_g2;
              /* printf("#DEBUG:g2=%d %d %d, abs_s_g2=%lf, abs_w_g2=%lf\n",
						g2.h,g2.k,g2.l, abs_s_g2, abs_w_g2);*/
              if ( (abs_w_g < beam_search.wgbmin) && (abs_w_g2 < beam_search.wg2bmin) ){
                if(g_target.h==h && g_target.k==k && g_target.l==l){
                  i_g_target = cte + 1;
                }
                g_exact[cte].h  = h; g_exact[cte].k = k; g_exact[cte].l = l; /* Exact */
                cte++;
                /* printf("%d\n",cte); */
              }
              else if ( (abs_w_g < beam_search.wgmax) && (abs_w_g2 < beam_search.wg2max) ){
                g_bethe[ctb].h  = h; g_bethe[ctb].k = k; g_bethe[ctb].l = l; /* Bethe */
                ctb++;
                /* printf("%d B\n",ctb); */
              }
            }
          }
          
          if ( (cte >= beam_search.nglim) || (ctb >= beam_search.nglim) ){
            printf("$$ num_g has exceeded num_g_max.\n");
            printf("$$ Please set a greater value into num_g. Abort.\n");
            exit(EXIT_FAILURE);
          }
        }
        
      }
    }
  }
LOOP_EXIT:
  *num_g_exact = cte;
  *num_g_bethe = ctb;
  sort_reflex(*num_g_exact, g_exact, bv);
  sort_reflex(*num_g_bethe, g_bethe, bv);
  return i_g_target;
}


void reflex_g_chk( int nth_g, HKL g_target, int nexact, HKL exact_reflex[] )
{
  int ng;
  
  ng = find_n_g( nexact, g_target, exact_reflex );
  
  if( nth_g == -1 ){
    printf( "$$ Target reflection %d %d %d is not found. Abort.\n",
            g_target.h, g_target.k, g_target.l );
    exit(EXIT_FAILURE);
  }
  else if( nth_g != ng ){
    printf( "$$ Target reflection %d %d %d\n", g_target.h, g_target.k, g_target.l );
    printf( "$$ nth_g = %d, ng = %d. Different. Abort.\n", nth_g, ng );
    exit(EXIT_FAILURE);
  }
  else if( nth_g >= nexact ){
    printf( "$$ Target reflection %d %d %d is not in the exact ones. Abort.\n",
            g_target.h, g_target.k, g_target.l );
    exit(EXIT_FAILURE);
  }
}


int reflex_search_s_rfsym(
  int ind, int *num_g_exact, int *num_g_bethe, HKL *g_exact, HKL *g_bethe,
  LAT_PAR lp, char *lattice_type, int num_sym_op, SYM_OP *sym_op,
  int n_atom, ATOM *atm, XDT *xdyl, ATOMF_BK *ftbl,
  int num_low_order_g, HKL *g_low_order, int lof_type, LO_FG *fg_low_order,
  double accel_kvol, BEAM_SEARCH_PAR beam_search, UVW incidence, R_HKL zolzbase1,
  HKL g_target, int g_table_size
  )
     /*
      * Search excited reflections:
      *   in accordance with w_g, w_g2, xi_g, xi_g2, lz_min and lz_max.
      *   special treatment only for ZOLZ reflections.
      *   The reflections found are divided into "exact" and "Bethe".
      *   Return value is i_g_target.
      */
{
  int i, h, k, l,cte, ctb, nth_lz, i_g_target;
  int hmax, kmax, lmax;
  double abs_s_g, s_g_start, s_g_end, abs_s_g2;
  double s_g_target_start, s_g_target_end;
  double abs_w_g, abs_w_g2;
  double tx, ty, abs_chi, abs_g, xi_g, xi_g2, abs_incidence;
  complex fg;
  R_HKL  rg;
  HKL g, g2;
  static HKL *g_table;		/* <<<<< */
  static complex *fg_table;	/* <<<<< */
  static int ct_fg_table;		/* <<<<< */
  CELL_PAR bv;
  
  if(ind < 0){	/* initialization */
    fg_table = new_vector(complex, g_table_size);	chk_vector(fg_table);
    g_table  = new_vector(HKL, g_table_size);		chk_vector(g_table);
    ct_fg_table = 0;
    for(i=0; i<num_low_order_g; i++){
      stock_fg_tbl_sym(
        g_low_order[i], lofg_to_fg_0(lof_type, fg_low_order[i]), &ct_fg_table,
        g_table_size, g_table, fg_table, num_sym_op, sym_op
        );
    }
    return 0;
  }
  else if( ind ==0 ){		/* free & termination */
    free_vector(fg_table); free_vector(g_table);
    return 0;
  }
  
  bv = bravai(lp);
  abs_incidence = abs_uvw(incidence, lp);
  abs_chi = 1.0 / wavelength(accel_kvol);
  evaluate_hklmax(beam_search.gmax,&hmax,&kmax,&lmax,bv);
  cte = 0; ctb = 0; /* Counters for reflections exact and Bethe */
  g_exact[0].h = 0; g_exact[0].k = 0; g_exact[0].l = 0; /* g_exact[0] must be 000 */
  cte = 1;
  /* for g_target */
  nth_lz = g_target.h * incidence.u + g_target.k * incidence.v + g_target.l * incidence.w;
  
  abs_edge_s_g( nth_lz, abs_incidence, abs_chi, beam_search.disksize,
                &tx, &ty, &s_g_target_start, &s_g_target_end,
                g_target, incidence, zolzbase1, bv );
  /*printf("s_g_target_start=%lf, s_g_target_end=%lf\n", s_g_target_start, s_g_target_end);*/
  
  for(h = - hmax; h <= hmax; h++){
    for(k = - kmax; k <= kmax; k++){
      for(l = - lmax; l <= lmax; l++){
        
        g.h = h; g.k = k; g.l = l;
        rg.h = (double) h; rg.k = (double) k; rg.l = (double) l;
        nth_lz =  g.h * incidence.u + g.k * incidence.v + g.l * incidence.w;
        abs_g = sqrt( sq_abs(rg, bv) );
        
        if (   ( h!=0 || k!=0 || l!=0 )
               && ( ( beam_search.lzmin <= nth_lz ) && ( nth_lz <= beam_search.lzmax ) )
               && (  abs_g < beam_search.gmax )
               && ( extinction_rule(g,lattice_type) == 0 )   ){
          /* 1st order diffraction (g) */
          fg = structure_factor_ext_tbl(
            g, num_sym_op, sym_op, n_atom, atm, xdyl, ftbl, accel_kvol, bv, lattice_type,
            &ct_fg_table, g_table_size, g_table, fg_table
            );
          xi_g = fg_to_xig( fg, accel_kvol, bv );
          if ( g_target.h==0 && g_target.k==0 && g_target.l==0 ){
            /* only for 1st order diffraction */
            if ( (xi_g < beam_search.xigmax) ){
              nth_lz =  g.h * incidence.u + g.k * incidence.v + g.l * incidence.w;
              abs_s_g = abs_edge_s_g(
                nth_lz, abs_incidence, abs_chi, beam_search.disksize,
                &tx, &ty, &s_g_start, &s_g_end, g, incidence, zolzbase1, bv
                );
              abs_w_g  = abs_s_g * xi_g;
              /* printf("g=%d %d %d, s_g_start=%lf, s_g_end=%lf\n",
						g.h, g.k, g.l, s_g_start, s_g_end);*/
              if ( abs_w_g < beam_search.wgbmin ){	/* Exact */
                g_exact[cte].h  = h; g_exact[cte].k = k; g_exact[cte].l = l;
                cte++;
              }
              else if ( abs_w_g < beam_search.wgmax ){	/* Bethe */
                g_bethe[ctb].h  = h; g_bethe[ctb].k = k; g_bethe[ctb].l = l;
                ctb++;
              }
            }
            /* i_g_target = 0; */
          }
          else{			/* with 2nd order diffraction (g2) */
            g2.h = g_target.h - g.h;
            g2.k = g_target.k - g.k;
            g2.l = g_target.l - g.l;
            if ( g2.h == 0 && g2.k == 0 && g2.l == 0 ){
              xi_g2 = 0.0;	/* Dummy */
            }
            else{
              fg = structure_factor_ext_tbl(
                g2, num_sym_op, sym_op, n_atom, atm, xdyl, ftbl, accel_kvol,
                bv, lattice_type, &ct_fg_table, g_table_size, g_table, fg_table
                );
              xi_g2 = fg_to_xig( fg, accel_kvol, bv );
            }
            nth_lz =  g.h * incidence.u + g.k * incidence.v + g.l * incidence.w;
            if ( (nth_lz == 0) && (xi_g < beam_search.xigmax) ){ /* only for ZOLZ */
              nth_lz =  g.h * incidence.u + g.k * incidence.v + g.l * incidence.w;
              abs_s_g = abs_edge_s_g( nth_lz, abs_incidence, abs_chi, beam_search.disksize,
                                      &tx, &ty, &s_g_start, &s_g_end, g, incidence, zolzbase1, bv );
              abs_w_g  = abs_s_g * xi_g;
              /* printf("g=%d %d %d, s_g_start=%lf, s_g_end=%lf\n",
						g.h, g.k, g.l, s_g_start, s_g_end);*/
              if ( abs_w_g < beam_search.wgbmin ){	/* Exact */
                g_exact[cte].h  = h; g_exact[cte].k = k; g_exact[cte].l = l;
                cte++;
              }
              else if ( abs_w_g < beam_search.wgmax ){	/* Bethe */
                g_bethe[ctb].h  = h; g_bethe[ctb].k = k; g_bethe[ctb].l = l;
                ctb++;
              }
            }
            else if ( (xi_g < beam_search.xigmax) && (xi_g2 < beam_search.xig2max) ){
              nth_lz =  g.h * incidence.u + g.k * incidence.v + g.l * incidence.w; /* g */
              abs_s_g = abs_edge_s_g( nth_lz, abs_incidence, abs_chi, beam_search.disksize,
                                      &tx, &ty, &s_g_start, &s_g_end, g, incidence, zolzbase1, bv );
              abs_w_g  = abs_s_g  * xi_g;
              /* printf("s_g_start=%lf, s_g_end=%lf\n", s_g_start, s_g_end);*/
              /* printf("#DEBUG:g=%d %d %d, abs_s_g=%lf, abs_w_g=%lf\n",
						g.h,g.k,g.l, abs_s_g, abs_w_g);*/
              abs_s_g2 = abs_edge_s_g2( s_g_start, s_g_end,
                                        s_g_target_start, s_g_target_end );		/* g2 */
              abs_w_g2 = abs_s_g2 * xi_g2;
              /* printf("#DEBUG:g2=%d %d %d, abs_s_g2=%lf, abs_w_g2=%lf\n",
						g2.h,g2.k,g2.l, abs_s_g2, abs_w_g2);*/
              if ( (abs_w_g < beam_search.wgbmin) && (abs_w_g2 < beam_search.wg2bmin) ){
                /* if(g_target.h==h && g_target.k==k && g_target.l==l){
							i_g_target = cte + 1;
						}*/
                g_exact[cte].h  = h; g_exact[cte].k = k; g_exact[cte].l = l; /* Exact */
                cte++;
                /* printf("%d\n",cte); */
              }
              else if ( (abs_w_g < beam_search.wgmax) && (abs_w_g2 < beam_search.wg2max) ){
                g_bethe[ctb].h  = h; g_bethe[ctb].k = k; g_bethe[ctb].l = l; /* Bethe */
                ctb++;
                /* printf("%d B\n",ctb); */
              }
            }
          }
          
          if ( (cte >= beam_search.nglim) || (ctb >= beam_search.nglim) ){
            printf("$$ num_g has exceeded num_g_max.\n");
            printf("$$ Please set a greater value into num_g. Abort.\n");
            exit(EXIT_FAILURE);
          }
        }
        
      }
    }
  }
LOOP_EXIT:
  *num_g_exact = cte; *num_g_bethe = ctb;
  sort_reflex(*num_g_exact, g_exact, bv);
  sort_reflex(*num_g_bethe, g_bethe, bv);
  i_g_target = find_n_g( cte, g_target, g_exact );
  if (i_g_target == -1) {
    printf("$$ rfsearch: g_target is not found in the exact beams. Abort.\n"); 
    exit(EXIT_FAILURE);
  }
  
  return i_g_target;
}
#endif


#if 0
void wg_xig_list(
  int num_exact, int num_bethe, HKL g_exact[], HKL g_bethe[],
  double wge[], double xige[], double wgb[], double xigb[]
  )
{
  char outfile[128];
  FILE *fo;
  int i, nr, nth, *index;
  double *wg, *xig;
  HKL	*rflx;
  
  nr = num_exact + num_bethe;
  rflx = new_vector(HKL, nr);		chk_vector(rflx);
  index = new_vector(int, nr);	chk_vector(index);
  wg = new_vector(double, nr);	chk_vector(wg);
  xig = new_vector(double, nr);	chk_vector(xig);
  
  for(i=0; i<num_exact; i++){
    rflx[i] = g_exact[i]; wg[i] = wge[i]; xig[i] = xige[i];
  }
  for(i=0; i<num_bethe; i++){
    rflx[i + num_exact] = g_bethe[i];
    wg[i + num_exact] = wgb[i];
    xig[i + num_exact] = xigb[i];
  }
  
  d_heapsort_ascend(nr, wg, index);
  
  strcpy(outfile, dbg.infile);	strcat(outfile, ".wg");
  if((fo = fopen(outfile, "w")) == NULL){
    printf("$$ Cannot open %s. Abort.\n", outfile);
    exit(EXIT_FAILURE);
  }
  printf("($$ wg & xig ==> %s) ..", outfile);
  
  fprintf(fo,"%% wg and xig list\n");
  fprintf(fo,"    i:  h   k   l\t\twg\t\txig\n");
  for(i=0; i<nr; i++){
    nth = index[i];
    fprintf(
      fo,"%5d: %3d %3d %3d\t%8.1lf\t%10.1lf\n",
      i, rflx[nth].h, rflx[nth].k, rflx[nth].l, wg[i], xig[nth]
      );
  }

  fclose(fo);
  free_vector(rflx);	free_vector(index);
  free_vector(wg);	free_vector(xig);
}


void simple_beam_search_wg(
  int *num_g_exact, int *num_g_bethe, HKL *g_exact, HKL *g_bethe,
  BEAM_SEARCH_PAR beam_search, UVW incidence, R_HKL zolzbase1,
  double H_tilt, double G_tilt, double accel_kvol, CRYSTAL_STRCT_DATA *csd
  )
     /*
      * Search excited reflections:
      *   according to w_g, xi_g.
      *   The reflections found are divided into "exact" and "Bethe".
      */
{
  int		i, h, k, l, cte, ctb, nth_lz, ng, hmax, kmax, lmax;
  int		nr, ct, ct_fg_table, g_table_size;
  double	abs_s_g, s_g_start, s_g_end, s_g_target_start, s_g_target_end;
  double	tx, ty, abs_chi, abs_g, abs_w_g, xi_g, abs_incidence;
  double	relativistic_correction, wave_length;
  double  *wge, *xige, *wgb, *xigb;
  complex	fg, *fg_table;
  R_HKL	rg, zolzbase2;
  HKL		g, *g_table;
  CELL_PAR	bv;
  
  /* ----- Initialization ----- */
  g_table_size = SIZE_GTABLE;
  fg_table = new_vector(complex, g_table_size);	chk_vector(fg_table);
  g_table  = new_vector(HKL, g_table_size);		chk_vector(g_table);
  ct_fg_table = 0;
  for(i=0; i<csd->num_low_order_g; i++){
    stock_fg_tbl_sym(
      csd->g_low_order[i], csd->fg_low_order[i], &ct_fg_table,
      g_table_size, g_table, fg_table, csd->num_sym_op, csd->sym_op
      );
  }
  
  /* For debug */
  wge = new_vector(double, beam_search.nglim);	chk_vector(wge);
  xige = new_vector(double, beam_search.nglim);	chk_vector(xige);
  wgb = new_vector(double, beam_search.nglim);	chk_vector(wgb);
  xigb = new_vector(double, beam_search.nglim);	chk_vector(xigb);
  
  
  /* Beam search */
  bv = bravai(csd->lp);
  zolzbase2 = zolz_ort_base(incidence, zolzbase1, bv);
  /* zolzbase2 is orthogonal to zolzbase1, and has the same norm. */
  abs_incidence = abs_uvw(incidence, csd->lp);
  wave_length = wavelength(accel_kvol);
  abs_chi = 1.0 / wave_length;
  relativistic_correction = relativ_corr(accel_kvol);
  evaluate_hklmax(beam_search.gmax,&hmax,&kmax,&lmax,bv);
  
  cte = 0; ctb = 0; 					/* Counters for reflections exact and Bethe */
  g_exact[0].h = 0; g_exact[0].k = 0; g_exact[0].l = 0; /* g_exact[0] must be 000 */
  fg = structure_factor_ext_tbl(					/* Structure factor */
                                                    g_exact[0], csd->num_sym_op, csd->sym_op, csd->n_atom, csd->atm,
                                                    csd->xdyl, csd->ftbl, bv, csd->lattice_type,
                                                    &ct_fg_table, g_table_size, g_table, fg_table
                                                    );
  xi_g = fg_to_xig2(
    fg, relativistic_correction, wave_length, bv
    );									/* Extinction distance */
  xige[0] = xi_g;		wge[0] = 0.0;
  cte = 1;
  
  
  for(h = - hmax; h <= hmax; h++){
    for(k = - kmax; k <= kmax; k++){
      for(l = - lmax; l <= lmax; l++){
        
        g.h = h; g.k = k; g.l = l;
        rg.h = (double) h; rg.k = (double) k; rg.l = (double) l;
        nth_lz =  g.h * incidence.u + g.k * incidence.v + g.l * incidence.w;
        abs_g = sqrt( sq_abs(rg, bv) );
        
        if (   ( h!=0 || k!=0 || l!=0 )
               && ( ( beam_search.lzmin <= nth_lz ) && ( nth_lz <= beam_search.lzmax ) )
               && (  abs_g < beam_search.gmax )
               && ( extinction_rule(g, csd->lattice_type) == 0 )
               ){
          fg = structure_factor_ext_tbl(					/* Structure factor */
                                                            g, csd->num_sym_op, csd->sym_op, csd->n_atom, csd->atm,
                                                            csd->xdyl, csd->ftbl, bv, csd->lattice_type,
                                                            &ct_fg_table, g_table_size, g_table, fg_table
                                                            );
          xi_g = fg_to_xig2(
            fg, relativistic_correction, wave_length, bv
            );									/* Extinction distance */
          
          if ( (xi_g < beam_search.xigmax) ){
            nth_lz =  g.h * incidence.u + g.k * incidence.v + g.l * incidence.w;
            abs_s_g = abs_edge_s_g_tilt(
              g, incidence, zolzbase1, zolzbase2, H_tilt, G_tilt,
              abs_incidence, abs_chi, beam_search.disksize,
              &s_g_start, &s_g_end, bv
              );
            abs_w_g  = abs_s_g * xi_g;
            if ( abs_w_g < beam_search.wgbmin ){		/* Exact */
              g_exact[cte].h  = h; g_exact[cte].k = k; g_exact[cte].l = l;
              xige[cte] = xi_g; wge[cte] = abs_w_g;
              cte++;
              
            }
            else if ( abs_w_g < beam_search.wgmax ){	/* Bethe */
              g_bethe[ctb].h  = h; g_bethe[ctb].k = k; g_bethe[ctb].l = l;
              xigb[ctb] = xi_g; wgb[ctb] = abs_w_g;
              ctb++;
            }
          }
          
          if ( (cte >= beam_search.nglim) || (ctb >= beam_search.nglim) ){
            printf("$$ num_g has exceeded num_g_max.\n");
            printf("$$ Please set a greater value into num_g. Abort.\n");
            exit(EXIT_FAILURE);
          }
        }
        
      }
    }
  }
LOOP_EXIT:
  
  /* Debug: dump wg & xig list */
  if(dbg.wg_xig_list_flag == 1){
    wg_xig_list( cte, ctb, g_exact, g_bethe, wge, xige, wgb, xigb );
  }
  
  *num_g_exact = cte; *num_g_bethe = ctb;
  sort_reflex(*num_g_exact, g_exact, bv);
  sort_reflex(*num_g_bethe, g_bethe, bv);
  
  free_vector(fg_table);	free_vector(g_table);
  free_vector(wge);		free_vector(xige);
  free_vector(wgb);		free_vector(xigb);
  
}
#endif


void simple_beam_search_wg(
  int *num_g_exact, int *num_g_bethe, HKL *g_exact, HKL *g_bethe,
  BEAM_SEARCH_PAR beam_search, UVW incidence, R_HKL zolzbase1,
  double H_tilt, double G_tilt, double accel_kvol, CRYSTAL_STRCT_DATA *csd
  )
{
  /*
   * Search excited reflections:
   *   according to w_g, xi_g.
   *   The reflections found are divided into "exact" and "Bethe".
   */
  int		i, h, k, l, cte, ctb, nth_lz, hmax, kmax, lmax;
  int		ct_fg_table, g_table_size;
  double	abs_s_g, s_g_start, s_g_end;
  double	abs_chi, abs_g, abs_w_g, xi_g, abs_incidence;
  double	relativistic_correction, wave_length, abs_K;
  complex	fg, *fg_table;
  R_HKL	rg, zolzbase2;
  HKL		g, *g_table;
  CELL_PAR	bv;
  
  /* ----- Initialization ----- */
  bv = bravai(csd->lp);
  /*
	if(csd->num_low_order_g > 0){
		calc_cst_lofg(
			csd->n_atom, csd->num_low_order_g, csd->lof_type,
			csd->g_low_order, csd->fg_low_order, &(csd->cl)
		);
	}
   */
  g_table_size = SIZE_GTABLE;
  fg_table = new_vector(complex, g_table_size);	chk_vector(fg_table);
  g_table  = new_vector(HKL, g_table_size);		chk_vector(g_table);
  ct_fg_table = 0;
  for(i=0; i<csd->num_lofg; i++){
    stock_fg_tbl_sym(
      csd->g_lofg[i],
      lofg_to_fg(
        csd->g_lofg[i], csd->lof_type, &(csd->fg_lofg[i]),
        csd->num_sym_op, csd->sym_op,
        csd->n_atom, csd->atm, csd->ftbl, csd->xdyl, accel_kvol, bv
        ),
      &ct_fg_table,
      g_table_size, g_table, fg_table, csd->num_sym_op, csd->sym_op
      );
  }
  
  /* ----- Beam search ----- */
  zolzbase2 = zolz_ort_base(incidence, zolzbase1, bv);
  /* zolzbase2 is orthogonal to zolzbase1, and has the same norm. */
  abs_incidence = abs_uvw(incidence, bv);
  wave_length = wavelength(accel_kvol);
  abs_chi = 1.0 / wave_length;
  abs_K = calc_abs_K( csd, accel_kvol, bv );/* <<<<< */
  relativistic_correction = relativ_corr(accel_kvol);
  evaluate_hklmax(beam_search.gmax,&hmax,&kmax,&lmax,bv);
  
  cte = 0; ctb = 0;/* Counters for reflections exact and Bethe */
  /* g_exact[0] must be 000 */
  g_exact[0].h = 0; g_exact[0].k = 0; g_exact[0].l = 0;
  fg = structure_factor_ext_tbl(
    /* Structure factor */
    g_exact[0], csd->num_sym_op, csd->sym_op, csd->n_atom, csd->atm,
    csd->xdyl, csd->ftbl, accel_kvol, bv, csd->lattice_type,
    &ct_fg_table, g_table_size, g_table, fg_table
    );
  xi_g = fg_to_xig2(
    fg, relativistic_correction, wave_length, bv
    );/* Extinction distance */
  cte++;
  
  
  for(h = - hmax; h <= hmax; h++){
    for(k = - kmax; k <= kmax; k++){
      for(l = - lmax; l <= lmax; l++){
        
        g.h = h; g.k = k; g.l = l;
        rg.h = (double) h; rg.k = (double) k; rg.l = (double) l;
        nth_lz =  g.h * incidence.u + g.k * incidence.v + g.l * incidence.w;
        abs_g = sqrt( sq_abs(rg, bv) );
        
        if (   ( h!=0 || k!=0 || l!=0 )
               && (
                 ( beam_search.lzmin <= nth_lz )
                 && ( nth_lz <= beam_search.lzmax )
                 )
               && (  abs_g < beam_search.gmax )
               && ( extinction_rule(g, csd->lattice_type) == 0 )
               ){
          fg = structure_factor_ext_tbl(
            /* Structure factor */
            g, csd->num_sym_op, csd->sym_op, csd->n_atom, csd->atm,
            csd->xdyl, csd->ftbl, accel_kvol, bv, csd->lattice_type,
            &ct_fg_table, g_table_size, g_table, fg_table
            );
          xi_g = fg_to_xig2(
            fg, relativistic_correction, wave_length, bv
            );/* Extinction distance */
          
          if ( (xi_g < beam_search.xigmax) ){
            nth_lz =  g.h * incidence.u + g.k * incidence.v + g.l * incidence.w;
            abs_s_g = abs_edge_s_g_tilt(
              g, incidence, zolzbase1, zolzbase2, H_tilt, G_tilt,
              abs_incidence, abs_K, beam_search.disksize,
              &s_g_start, &s_g_end, bv
              );
            abs_w_g  = abs_s_g * xi_g;
            if ( abs_w_g < beam_search.wgbmin ){		/* Exact */
              g_exact[cte].h  = h; g_exact[cte].k = k; g_exact[cte].l = l;
              cte++;
              
            }
            else if ( abs_w_g < beam_search.wgmax ){	/* Bethe */
              g_bethe[ctb].h  = h; g_bethe[ctb].k = k; g_bethe[ctb].l = l;
              ctb++;
            }
          }
          
          if ( (cte >= beam_search.nglim) || (ctb >= beam_search.nglim) ){
            printf("$$ num_g has exceeded num_g_max.\n");
            printf("$$ Please set a greater value into num_g. Abort.\n");
            exit(EXIT_FAILURE);
          }
        }
        
      }
    }
  }
  /*	LOOP_EXIT: */
  
  *num_g_exact = cte; *num_g_bethe = ctb;
  sort_reflex(*num_g_exact, g_exact, bv);
  sort_reflex(*num_g_bethe, g_bethe, bv);
  
  free_vector(fg_table);	free_vector(g_table);
}


void simple_beam_search_sg(
  int *num_g_exact, int *num_g_bethe, HKL *g_exact, HKL *g_bethe,
  BEAM_SEARCH_PAR beam_search, UVW incidence, R_HKL zolzbase1,
  double H_tilt, double G_tilt, double accel_kvol, CRYSTAL_STRCT_DATA *csd
  )
{
  /*
   * Search excited reflections:
   *   according to s_g, xi_g.
   *   The reflections found are divided into "exact" and "Bethe".
   */
  int     i, h, k, l, cte, ctb, nth_lz, hmax, kmax, lmax;
  int     ct_fg_table, g_table_size;
  double  abs_s_g, s_g_start, s_g_end;
  double  abs_chi, abs_g, xi_g, abs_incidence;
  double  relativistic_correction, wave_length, abs_K;
  complex  fg, *fg_table;
  R_HKL    rg, zolzbase2;
  HKL      g, *g_table;
  CELL_PAR  bv;

  /* ----- Initialization ----- */
  bv = bravai(csd->lp);
  /*
	if(csd->num_low_order_g > 0){
		calc_cst_lofg(
			csd->n_atom, csd->num_low_order_g, csd->lof_type,
			csd->g_low_order, csd->fg_low_order, &(csd->cl)
		);
	}
   */
  g_table_size = SIZE_GTABLE;
  fg_table = new_vector(complex, g_table_size);	chk_vector(fg_table);
  g_table  = new_vector(HKL, g_table_size);		chk_vector(g_table);
  ct_fg_table = 0;
  for(i=0; i<csd->num_lofg; i++){
    stock_fg_tbl_sym(
      csd->g_lofg[i],
      lofg_to_fg(
        csd->g_lofg[i], csd->lof_type, &(csd->fg_lofg[i]),
        csd->num_sym_op, csd->sym_op,
        csd->n_atom, csd->atm, csd->ftbl, csd->xdyl, accel_kvol, bv
        ),
      &ct_fg_table, g_table_size, g_table, fg_table,
      csd->num_sym_op, csd->sym_op
      );
  }
  
  /* ----- Beam search ----- */
  zolzbase2 = zolz_ort_base(incidence, zolzbase1, bv);
  /* zolzbase2 is orthogonal to zolzbase1, and has the same norm. */
  abs_incidence = abs_uvw(incidence, bv);
  wave_length = wavelength(accel_kvol);
  abs_chi = 1.0 / wave_length;
  abs_K = calc_abs_K( csd, accel_kvol, bv );
  relativistic_correction = relativ_corr(accel_kvol);
  evaluate_hklmax(beam_search.gmax, &hmax, &kmax,& lmax, bv);
  
  cte = 0; ctb = 0;  /* Counters for reflections exact and Bethe */
  /* g_exact[0] must be 000 */
  g_exact[0].h = 0; g_exact[0].k = 0; g_exact[0].l = 0;
  /* Structure factor */
  fg = structure_factor_ext_tbl(
    g_exact[0], csd->num_sym_op, csd->sym_op, csd->n_atom, csd->atm,
    csd->xdyl, csd->ftbl, accel_kvol, bv, csd->lattice_type,
    &ct_fg_table, g_table_size, g_table, fg_table
    );
  /* Extinction distance */
  xi_g = fg_to_xig2( fg, relativistic_correction, wave_length, bv );
  cte++;

  for(h = - hmax; h <= hmax; h++){
    for(k = - kmax; k <= kmax; k++){
      for(l = - lmax; l <= lmax; l++){
        g.h = h; g.k = k; g.l = l;
        rg.h = (double) h; rg.k = (double) k; rg.l = (double) l;
        nth_lz =  g.h * incidence.u + g.k * incidence.v + g.l * incidence.w;
        abs_g = sqrt( sq_abs(rg, bv) );
        if ( ( h!=0 || k!=0 || l!=0 )
             && (
               ( beam_search.lzmin <= nth_lz )
               && ( nth_lz <= beam_search.lzmax )
               )
             && ( abs_g < beam_search.gmax )
             && ( extinction_rule(g, csd->lattice_type) == 0 )
             ){
          if(beam_search.xigmax >= XIGMAX){
            xi_g = XIGMAX;
          }
          else{
            fg = structure_factor_ext_tbl(
              /* Structure factor */
              g, csd->num_sym_op, csd->sym_op, csd->n_atom, csd->atm,
              csd->xdyl, csd->ftbl, accel_kvol, bv, csd->lattice_type,
              &ct_fg_table, g_table_size, g_table, fg_table
              );
            xi_g = fg_to_xig2( fg, relativistic_correction, wave_length, bv );
          }
          /* Extinction distance */
          if ( (xi_g < beam_search.xigmax) ){
            abs_s_g = abs_edge_s_g_tilt(
              g, incidence, zolzbase1, zolzbase2, H_tilt, G_tilt,
              abs_incidence, abs_K, beam_search.disksize,
              &s_g_start, &s_g_end, bv
              );
            if ( abs_s_g < beam_search.sgbmin ){		/* Exact */
              g_exact[cte].h  = h; g_exact[cte].k = k; g_exact[cte].l = l;
              cte++;
            }
            else if ( abs_s_g < beam_search.sgmax ){	/* Bethe */
              g_bethe[ctb].h  = h; g_bethe[ctb].k = k; g_bethe[ctb].l = l;
              ctb++;
            }
          }
          if ( (cte >= beam_search.nglim) || (ctb >= beam_search.nglim) ){
            printf("$$ num_g has exceeded num_g_max.\n");
            printf("$$ Please set a greater value into num_g. Abort.\n");
            exit(EXIT_FAILURE);
          }
        }

      }
    }
  }
  /* LOOP_EXIT: */

  *num_g_exact = cte; *num_g_bethe = ctb;
  sort_reflex(*num_g_exact, g_exact, bv);
  sort_reflex(*num_g_bethe, g_bethe, bv);

  free_vector(fg_table);	free_vector(g_table);
}


void beam_search_sg_multi(
  int *num_g_exact, int *num_g_bethe, HKL *g_exact, HKL *g_bethe,
  BEAM_SEARCH_PAR beam_search, UVW incidence, R_HKL zolzbase1,
  double H_tilt, double G_tilt, double accel_kvol, CRYSTAL_STRCT_DATA *csd
  )
{
  /*
   * Search excited reflections:
   *   according to s_g, xi_g.
   *   The reflections found are divided into "exact" and "Bethe".
   *   Multi parameters for Laue zones
   */
  int		i, h, k, l, cte, ctb, nth_lz, hmax, kmax, lmax;
  int		ct_fg_table, g_table_size;
  int		lzmin, lzmax;
  double	abs_s_g, s_g_start, s_g_end;
  double	abs_chi, abs_g, xi_g, abs_incidence;
  double	relativistic_correction, wave_length, abs_K, gmax;
  complex	fg, *fg_table;
  R_HKL	rg, zolzbase2;
  HKL		g, *g_table;
  CELL_PAR	bv;
  
  /* ----- Initialization ----- */
  bv = bravai(csd->lp);
  /*
	if(csd->num_low_order_g > 0){
		calc_cst_lofg(
			csd->n_atom, csd->num_low_order_g, csd->lof_type,
			csd->g_low_order, csd->fg_low_order, &(csd->cl)
		);
	}
   */
  g_table_size = SIZE_GTABLE;
  fg_table = new_vector(complex, g_table_size);	chk_vector(fg_table);
  g_table  = new_vector(HKL, g_table_size);		chk_vector(g_table);
  ct_fg_table = 0;
  for(i=0; i<csd->num_lofg; i++){
    stock_fg_tbl_sym(
      csd->g_lofg[i],
      lofg_to_fg(
        csd->g_lofg[i], csd->lof_type, &(csd->fg_lofg[i]),
        csd->num_sym_op, csd->sym_op,
        csd->n_atom, csd->atm, csd->ftbl, csd->xdyl, accel_kvol, bv
        ),
      &ct_fg_table, g_table_size, g_table, fg_table, csd->num_sym_op, csd->sym_op
      );
  }
  
  /* ----- Beam search ----- */
  zolzbase2 = zolz_ort_base(incidence, zolzbase1, bv);
  /* zolzbase2 is orthogonal to zolzbase1, and has the same norm. */
  abs_incidence = abs_uvw(incidence, bv);
  wave_length = wavelength(accel_kvol);
  abs_chi = 1.0 / wave_length;
  abs_K = calc_abs_K( csd, accel_kvol, bv );
  relativistic_correction = relativ_corr(accel_kvol);
  evaluate_hklmax(beam_search.gmax, &hmax, &kmax,& lmax, bv);
  
  cte = 0; ctb = 0;  /* Counters for reflections exact and Bethe */
  g_exact[0].h = 0;
  g_exact[0].k = 0;
  g_exact[0].l = 0; /* g_exact[0] must be 000 */
  fg = structure_factor_ext_tbl(
    /* Structure factor */
    g_exact[0], csd->num_sym_op, csd->sym_op, csd->n_atom, csd->atm,
    csd->xdyl, csd->ftbl, accel_kvol, bv, csd->lattice_type,
    &ct_fg_table, g_table_size, g_table, fg_table
    );
  xi_g = fg_to_xig2( fg, relativistic_correction, wave_length, bv );
  /* Extinction distance */
  cte++;
  
  lzmin = beam_search.lzmin;
  lzmax = beam_search.lzmax;
  gmax = beam_search.gmax;
  
  for(h = - hmax; h <= hmax; h++){
    for(k = - kmax; k <= kmax; k++){
      for(l = - lmax; l <= lmax; l++){
        
        g.h = h; g.k = k; g.l = l;
        rg.h = (double) h; rg.k = (double) k; rg.l = (double) l;
        nth_lz =  g.h * incidence.u + g.k * incidence.v + g.l * incidence.w;
        abs_g = sqrt( sq_abs(rg, bv) );
        
        if (   ( h!=0 || k!=0 || l!=0 )
               && ( ( lzmin <= nth_lz ) && ( nth_lz <= lzmax ) )
               && ( abs_g < gmax )
               && ( extinction_rule(g, csd->lattice_type) == 0 )
               ){
          if(beam_search.xig_max[nth_lz] >= XIGMAX){	/* <<<<< 9 Dec. 2000 */
            xi_g = XIGMAX;
          }
          else{
            fg = structure_factor_ext_tbl(
              /* Structure factor */
              g, csd->num_sym_op, csd->sym_op, csd->n_atom, csd->atm,
              csd->xdyl, csd->ftbl, accel_kvol, bv, csd->lattice_type,
              &ct_fg_table, g_table_size, g_table, fg_table
              );
            xi_g = fg_to_xig2( fg, relativistic_correction, wave_length, bv );
          }
          
          abs_s_g = abs_edge_s_g_tilt(
            g, incidence, zolzbase1, zolzbase2, H_tilt, G_tilt,
            abs_incidence, abs_K, beam_search.disksize,
            &s_g_start, &s_g_end, bv
            );
          
          if ( (xi_g < beam_search.xig_max[nth_lz]) ){
            if ( abs_s_g < beam_search.sge_max[nth_lz] ){		/* Exact */
              g_exact[cte].h = h;
              g_exact[cte].k = k;
              g_exact[cte].l = l;
              cte++;
            }
            else if ( abs_s_g < beam_search.sgb_max[nth_lz] ){	/* Bethe */
              g_bethe[ctb].h = h;
              g_bethe[ctb].k = k;
              g_bethe[ctb].l = l;
              ctb++;
            }
          }
          
          if ( (cte >= beam_search.nglim) || (ctb >= beam_search.nglim) ){
            printf("$$ num_g has exceeded num_g_max.\n");
            printf("$$ Please set a greater value into num_g. Abort.\n");
            exit(EXIT_FAILURE);
          }
        }
        
      }
    }
  }
  /* LOOP_EXIT: */
  
  *num_g_exact = cte; *num_g_bethe = ctb;
  sort_reflex(*num_g_exact, g_exact, bv);
  sort_reflex(*num_g_bethe, g_bethe, bv);
  
  free_vector(fg_table);	free_vector(g_table);
}



void beam_search_sphere(
  int *num_g_exact, int *num_g_bethe, HKL *g_exact, HKL *g_bethe,
  BEAM_SEARCH_PAR beam_search, UVW incidence, R_HKL zolzbase1,
  double H_tilt, double G_tilt, double accel_kvol, CRYSTAL_STRCT_DATA *csd
  )
     /*
      * Search reflections within the sphere of beam_search.gmax
      *  updated on 2001.06.14
      */
{
  int		i, h, k, l, cte, ctb, nth_lz, hmax, kmax, lmax;
  int		ct_fg_table, g_table_size;
  double	abs_chi, abs_g, abs_incidence;
  double	relativistic_correction, wave_length, abs_K;
  complex	fg, *fg_table;
  R_HKL	rg, zolzbase2;
  HKL		g, *g_table;
  CELL_PAR	bv;
  
  bv = bravai(csd->lp);
  g_table_size = SIZE_GTABLE;
  fg_table = new_vector(complex, g_table_size);	chk_vector(fg_table);
  g_table  = new_vector(HKL, g_table_size);		chk_vector(g_table);
  ct_fg_table = 0;
  for(i=0; i<csd->num_lofg; i++){
    stock_fg_tbl_sym(
      csd->g_lofg[i],
      lofg_to_fg(
        csd->g_lofg[i], csd->lof_type, &(csd->fg_lofg[i]),
        csd->num_sym_op, csd->sym_op,
        csd->n_atom, csd->atm, csd->ftbl, csd->xdyl, accel_kvol, bv
        ),
      &ct_fg_table, g_table_size, g_table, fg_table, csd->num_sym_op, csd->sym_op
      );
  }
  
  /* ----- Beam search ----- */
  zolzbase2 = zolz_ort_base(incidence, zolzbase1, bv);
  /* zolzbase2 is orthogonal to zolzbase1, and has the same norm. */
  abs_incidence = abs_uvw(incidence, bv);
  wave_length = wavelength(accel_kvol);
  abs_chi = 1.0 / wave_length;
  abs_K = calc_abs_K( csd, accel_kvol, bv );
  relativistic_correction = relativ_corr(accel_kvol);
  evaluate_hklmax(beam_search.gmax, &hmax, &kmax,& lmax, bv);
  
  cte = 0; ctb = 0; /* Counters for reflections exact and Bethe */
  g_exact[0].h = 0; g_exact[0].k = 0; g_exact[0].l = 0;
  /* g_exact[0] must be 000 */
  cte++;
  
  for(h = - hmax; h <= hmax; h++){
    for(k = - kmax; k <= kmax; k++){
      for(l = - lmax; l <= lmax; l++){
        
        g.h = h; g.k = k; g.l = l;
        rg.h = (double) h; rg.k = (double) k; rg.l = (double) l;
        nth_lz =  g.h * incidence.u + g.k * incidence.v + g.l * incidence.w;
        abs_g = sqrt( sq_abs(rg, bv) );
        
        if (   ( h!=0 || k!=0 || l!=0 )
               && ( abs_g < beam_search.gmax )
               && ( extinction_rule(g, csd->lattice_type) == 0 )
               ){
          g_exact[cte].h  = h; g_exact[cte].k = k; g_exact[cte].l = l;
          cte++;
          if ( (cte >= beam_search.nglim) || (ctb >= beam_search.nglim) ){
            printf("$$ num_g has exceeded num_g_max.\n");
            printf("$$ Please set a greater value into num_g. Abort.\n");
            exit(EXIT_FAILURE);
          }
        }
        
      }
    }
  }
  
  *num_g_exact = cte; *num_g_bethe = ctb;
  sort_reflex(*num_g_exact, g_exact, bv);
  sort_reflex(*num_g_bethe, g_bethe, bv);
  
  free_vector(fg_table);  free_vector(g_table);
}
