/*
 * utils.c
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "utils.h"

#define STRLEN 256


FILE *fopen_chk_rmode(char *filename)
{
  FILE *fp;
  if( (fp = fopen(filename, "r")) == NULL ){
    fprintf(
      stderr, "$$ (fopen_chk_rmode) Cannot open %s. Abort.\n", filename
      );
    /* perror("$$ (fopen_chk_rmode)\n");*/
    exit(EXIT_FAILURE);
  }
  return fp;
}


FILE *fopen_chk_wmode(char *filename)
{
  FILE *fp;
  if( (fp = fopen(filename, "w")) == NULL ){
    fprintf(
      stderr, "$$ (fopen_chk_wmode) Cannot open %s. Abort.\n", filename
      );
    /* perror("$$ (fopen_chk_wmode)\n"); */
    exit(EXIT_FAILURE);
  }
  return fp;
}


void two_fprintf(FILE *fp, FILE *fp2, char *fmt,...)
{
  va_list ap;
  va_start(ap, fmt);
  vfprintf(fp, fmt, ap);
  vfprintf(fp2, fmt, ap);
  va_end(ap);
}


int str_to_fs( int max_num, char *str, double x[])
{
  /*
   * breaks str into x[]
   * and returns number of strs elements
   * (x[0],x[1],...,x[n-1]).
   * max_num: < array size of x[]
   */
  int i, ct;
  char *delims=" ,\t", *token;

  for(i=0;i<max_num;i++) {x[i] = 0.0;} /* initialize */

  token = strtok(str, delims);
  ct = 0;
  while ( (token != NULL) && (ct < max_num) ){
    x[ct] = atof(token); ct++;
    token = strtok(NULL, delims);
  }
  return ct;
}


char *remove_extention(char *filename)
{
  /* Remove the extention part of the parfile */
  char buf[STRLEN], *tp;

  strcpy(buf, filename);
  tp = strtok(buf, ".");
  return tp;
}


char *rm_cr(char *str)
{
  /*
   * Remove cr chr. (\n) at the end of str
   * (useful to fgets() )
   */
  

  if ( strpbrk(str,"\n") == NULL )
    return str;
  else{
    str[ strlen(str) - 1] = '\0';
    return str;
  }

}


char *comment_skip_fgets(int n, char *buf, FILE *fp)
{
  /*
   * Skip comment lines starting from "%" and "#"
   */
  char *s;

  do{
    s = fgets(buf, n, fp);
    if(
      strncmp(buf, "%", 1)!=0 && strncmp(buf, "#", 1)!=0
      && s!=NULL && strncmp(buf, "COMMENT{",8)!=0
      ){
      break;
    }
    if(s == NULL){
      break;
    }
    if(strncmp(buf, "COMMENT{",8) == 0 ){
      while(strncmp(buf, "}COMMENT", 8) != 0){
        s = fgets(buf, n, fp);
      }
    }
  } while(1);

  /*
  do{
    s = fgets(buf, n, fp);
  }while(
    (strncmp(buf, "%", 1)==0 || strncmp(buf, "#", 1)==0) && s!=NULL
    );

  if(strncmp(buf, "COMMENT{",8) == 0 ){
    do{
      s = fgets(buf, n, fp);
    } while(strncmp(buf, "}COMMENT", 8) != 0);
    do{
      s = fgets(buf, n, fp);
    }while(
      (strncmp(buf, "%", 1)==0 || strncmp(buf, "#", 1)==0) && s!=NULL
      );
  }
 */
  return s;
}


void findmaxmin(int n, double x[], double *xmax, double *xmin)
     /* evaluation of max and min value for a 1D array */
{
  int j;
  double max,min;
  
  max = min = x[0];
  
  for (j=1;j<n;j++){
    if (x[j] > max) max = x[j];
    if (x[j] < min) min = x[j];
  }
  *xmax = max; *xmin = min;
}


void findmaxmin2(int nx, int ny, double **x, double *xmax, double *xmin)
     /* evaluation of max and min value for a 2D array */
{
  int i, j;
  double max,min;
  
  max = min = x[0][0];
  
  for (i=1; i<nx; i++){
    for (j=1; j<ny; j++){
      if (x[i][j] > max) max = x[i][j];
      if (x[i][j] < min) min = x[i][j];
    }
  }
  
  *xmax = max; *xmin = min;
}


double machine_eps(void)
{
  /*
   * evaluation of the machine epsilon
   * for double
   */
  double e, w;

  e = 1.0;	w = 1.0 + e;
  while(w > 1.0){
    e = e / 2.0;	 w = 1.0 + e;
  }
  return e * 2.0;
}


double quadratic_interpol(
  int ik, double h_shift, double *intensity, int num_k_points, double *bh
  )
     /*
      * 1D quadratic interpolation of intensity data
      */
{
  int i;
  double py, xx, x1, x2, x3, x12, x13, x23;
  
  xx = bh[ik] - h_shift;
  i = 1;
  do{ i++; }while ( xx > bh[i] && i < num_k_points-1 );
  
  x1 = xx - bh[i-2];
  x2 = xx - bh[i-1];
  x3 = xx - bh[i];
  x12 = bh[i-2] - bh[i-1];
  x13 = bh[i-2] - bh[i];
  x23 = bh[i-1] - bh[i];
  
  py	= intensity[i-2] * x2 * x3 / ( x12 * x13 )
    - intensity[i-1] * x1 * x3 / ( x12 * x23 )
      + intensity[i]   * x1 * x2 / ( x13 * x23 );
  return py;
}


void d_heapsort(int n, double a[], int index[])
{
  /*
   * heapsort
   *  Sort a[0] - a[n-1] in descending order
   *  return sorted a[]
   *  and index[] (reference number to the previous order)
   */
  int i, j, k, istock;
  double x;
  
  for (k = 0; k < n; k++){ index[k] = k; }	/* init. */
  for (k = n / 2 - 1; k >= 1; k--) {
    i = k;	x = a[i - 1];	istock = index[i - 1];
    while ((j = 2 * i) <= n) {
      if (j < n && a[j - 1] > a[j]) { j++; }
      if (x <= a[j - 1]) { break; }
      a[i - 1] = a[j - 1];	index[i - 1] = index[j - 1];
      i = j;
    }
    a[i - 1] = x;	index[i - 1] = istock;
  }
  while (n > 1) {
    x = a[n - 1];	istock = index[n - 1];
    a[n - 1] = a[0];	index[n - 1] = index[0];	n--;
    i = 1;
    while ((j = 2 * i) <= n) {
      if (j < n && a[j - 1] > a[j]) {j++;}
      if (x <= a[j - 1]) { break; }
      a[i - 1] = a[j - 1];	index[i - 1] = index[j - 1];
      i = j;
    }
    a[i - 1] = x;	index[i - 1] = istock;
  }
}


void d_heapsort_ascend(int n, double a[], int index[])
{
  /*
   * heapsort
   *  Sort a[0] - a[n-1] in ascending order
   *  return sorted a[]
   *  and index[] (reference number to the previous order)
   */
  int i, j, k, istock;
  double x;
  
  for (k = 0; k < n; k++){ index[k] = k; }	/* init. */
  for (k = n / 2 - 1; k >= 1; k--) {
    i = k;	x = a[i - 1];	istock = index[i - 1];
    while ((j = 2 * i) <= n) {
      if (j < n && a[j - 1] < a[j]) { j++; }
      if (x >= a[j - 1]) { break; }
      a[i - 1] = a[j - 1];	index[i - 1] = index[j - 1];
      i = j;
    }
    a[i - 1] = x;	index[i - 1] = istock;
  }
  while (n > 1) {
    x = a[n - 1];	istock = index[n - 1];
    a[n - 1] = a[0];	index[n - 1] = index[0];	n--;
    i = 1;
    while ((j = 2 * i) <= n) {
      if (j < n && a[j - 1] < a[j]) {j++;}
      if (x >= a[j - 1]) { break; }
      a[i - 1] = a[j - 1];	index[i - 1] = index[j - 1];
      i = j;
    }
    a[i - 1] = x;	index[i - 1] = istock;
  }
}


void i_heapsort(int n, int a[], int index[])
{
  /*
   * heapsort
   *  Sort a[0] - a[n-1] in descending order
   *  return sorted a[]
   *  and index[] (reference number to the previous order)
   */
  int i, j, k, istock;
  int x;
  
  for (k = 0; k < n; k++){ index[k] = k; }	/* init. */
  for (k = n / 2 - 1; k >= 1; k--) {
    i = k;	x = a[i - 1];	istock = index[i - 1];
    while ((j = 2 * i) <= n) {
      if (j < n && a[j - 1] > a[j]) { j++; }
      if (x <= a[j - 1]) { break; }
      a[i - 1] = a[j - 1];	index[i - 1] = index[j - 1];
      i = j;
    }
    a[i - 1] = x;	index[i - 1] = istock;
  }
  while (n > 1) {
    x = a[n - 1];	istock = index[n - 1];
    a[n - 1] = a[0];	index[n - 1] = index[0];	n--;
    i = 1;
    while ((j = 2 * i) <= n) {
      if (j < n && a[j - 1] > a[j]) {j++;}
      if (x <= a[j - 1]) { break; }
      a[i - 1] = a[j - 1];	index[i - 1] = index[j - 1];
      i = j;
    }
    a[i - 1] = x;	index[i - 1] = istock;
  }
}


void i_heapsort_ascend(int n, int a[], int index[])
{
  /*
   * heapsort
   *  Sort a[0] - a[n-1] in ascending order
   *  return sorted a[]
   *  and index[] (reference number to the previous order)
   */
  int i, j, k, istock;
  int x;
  
  for (k = 0; k < n; k++){ index[k] = k; }	/* init. */
  for (k = n / 2 - 1; k >= 1; k--) {
    i = k;	x = a[i - 1];	istock = index[i - 1];
    while ((j = 2 * i) <= n) {
      if (j < n && a[j - 1] < a[j]) { j++; }
      if (x >= a[j - 1]) { break; }
      a[i - 1] = a[j - 1];	index[i - 1] = index[j - 1];
      i = j;
    }
    a[i - 1] = x;	index[i - 1] = istock;
  }
  while (n > 1) {
    x = a[n - 1];	istock = index[n - 1];
    a[n - 1] = a[0];	index[n - 1] = index[0];	n--;
    i = 1;
    while ((j = 2 * i) <= n) {
      if (j < n && a[j - 1] < a[j]) {j++;}
      if (x >= a[j - 1]) { break; }
      a[i - 1] = a[j - 1];	index[i - 1] = index[j - 1];
      i = j;
    }
    a[i - 1] = x;	index[i - 1] = istock;
  }
}
