// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------
#include <math.h>
#include <stdio.h>
#include "funcdefs.h"
#include "files.h"
#include "PhasegratingRecord.h"
#include "ProjCellRecord.h"
#include "MslicRecord.h"
#include "filenames.h"
#include "errors.h"
#include "storage.h"
#include "atoms.h"
#include "xtal.h"
#include "header.h"
#include "devices.h"
#include "numbers.h"
#include "globals.h"
#define fabs(x) 	((x) < 0. ? -(x) : (x))
FILE 	*fp_print,*fp_pout ;
struct 	{ float ax[MAXTYPES][4],bx[MAXTYPES][4],cx[MAXTYPES] ; } xrf ;
extern  short gUseElScattFactors ;
extern 	long ERASE,LAYERED ;
extern 	struct { float wavel; long nslice ; float sigma ; } gMSRec ;

typedef struct RMatrix { 
	float m[3][3] ;
} RMatrix ;
extern	RMatrix	gTMatrix ;

char 	AtomNames[MAXTYPES][8] ;
long		noz[MAXTYPES] ;

extern Boolean gNoGUI ;

// --------------------------------------------------------------------------
// Entry for the calculation of the projected potential. 
// --------------------------------------------------------------------------
void CalculatePhasegrating(long *ierr) {

    long 	jmax;
	*ierr = 0 ;
    if (!gNoGUI) SetWatchCursor() ; 
	
    /* Open the print file */
    fp_print = fopen(flname.pprint,"w") ;
    if (fp_print==NULL) {
		ReportFileOpenError(flname.pprint) ;
		*ierr = -1 ;
		if (!gNoGUI) InitCursor() ;
		return ;
    }
    /* Open the output file */
    if (!open_output_file(flname.pout)) {
		*ierr = -1 ;
		if (!gNoGUI) InitCursor() ;
		return ;
    }
	
	
    /* Set the required parameters */
    SetPhasegratingParameters(AtomNames,noz) ;
    if (gIOError) {
		close_output_file() ;
		*ierr = -1 ;
		if (!gNoGUI) InitCursor() ;
		return ;
    }
    jmax = gPhsgrtRec->nbeam*gPhsgrtRec->mbeam ;
    /* Write to the printfile */
    WritePhasegratingOutput(AtomNames) ; 

    if (gIOError) {
		close_output_file() ;
		*ierr = -1 ;
		if (!gNoGUI) InitCursor() ;
		return ;
    }
	
    /* --------------------------------------------------------------------- */
    /* Calculate the fourier coefficients of the potential and the potential */
    /* --------------------------------------------------------------------- */
    CalculateProjectedPotential(gWorkSpace,noz,jmax) ;
	  
    /* -- Close the open files -- */
    fclose(fp_print) ;
    close_output_file() ;
	if (!gNoGUI) InitCursor() ;
		
}

void SetPhasegratingParameters(char name[MAXTYPES][8],long *noz) {
    /* --------------------------------------------------------	*/
    /* Just a little bit of set up work.			*/
    /* --------------------------------------------------------	*/

    float 	ewavel,evolt ;
    FILE 	*fp_scatt ;
    long 	element,jmax,ignore,i2,ierr ;
    register 	long i,k,j ;
    char 	SfactFile[80] ;
    extern 	char *NcemssDir ;
	
	evolt = gMsliceRec.volt ;
	ewavel = 1.2263652e-7/sqrt((evolt+9.785025e-4*evolt*evolt)*1000.) ;
	gMSRec.sigma = 2.08886e+6*ewavel*sqrt(1.0+5.886579e-20/(ewavel*ewavel)) ;
	gMSRec.wavel = 1.0e+8*ewavel ;  // Wavelength in angstroms
	
    /* --------------------------------------------------------------------- */
    /* Find the new unit cell that corresponds to the incoming electron beam */
    /* --------------------------------------------------------------------- */
    SetProjectedCell(&ierr) ;
    SetCalculationDimensions(&gProjCell->a,&gProjCell->b,&gPhsgrtRec->gmax,&gPhsgrtRec->nbeam,&gPhsgrtRec->mbeam) ;
    jmax = gPhsgrtRec->nbeam*(gPhsgrtRec->mbeam) ;
	
    gPhsgrtRec->slcth = gProjCell->c/gPhsgrtRec->nz ;
    
    /* --- Open the scatt.dat file for reading ---- */
    strcpy(SfactFile,NcemssDir) ;
#ifdef VMS
    if (gUseElScattFactors)
		strcat(SfactFile,"elfact.dat") ;
    else
		strcat(SfactFile,"scatt.dat") ;
#else
    if (gUseElScattFactors)
		strcat(SfactFile,"/elfact.dat") ;
    else
    	strcat(SfactFile,"/scatt.dat") ;
#endif
    fp_scatt = fopen(SfactFile,"r") ;

    if (fp_scatt==NULL) {
		ReportFileOpenError(SfactFile) ;
		gIOError = ERR_OP ;
		return ;
    }
	  
    for(i=0;i<gPhsgrtRec->natom;i++)
		noz[gPhsgrtRec->ntype[i]-1] = gPhsgrtRec->atno[i] ;
		
    /* Read in all the scattering factors from the data-file.	*/
    /* This would be much easier if we just positioned within	*/
    /* the file instead of rewinding the file after we get the	*/
    /* data for each atom.					*/
    /* Maybe I will do it one day. It doesn't do much though.	*/

	for(k=0;k<gPhsgrtRec->iatom;k++) {
		i2 = noz[k] ;
		for(i=0;i<i2;i++) {
			if (gUseElScattFactors)
				fscanf(fp_scatt,"%d",&element) ;
			fscanf(fp_scatt,"%s",&name[k][0]) ;
			for (j=0;j<4;j++) {
				fscanf(fp_scatt,"%f",&xrf.ax[k][j]) ;
				fscanf(fp_scatt,"%f",&xrf.bx[k][j]) ;
			}
			if (!gUseElScattFactors)
				fscanf(fp_scatt,"%f",&xrf.cx[k]) ;
		}
		/* Rewind the file */
		rewind(fp_scatt) ;
	}
    fclose(fp_scatt) ;
}

void WritePhasegratingOutput(char name[MAXTYPES][8]) {
    /* -------------------------------------------------------	*/
    /* Print various information to a printfile so the user can	*/
    /* get some numbers out of the calculation and so that it 	*/
    /* is not totally a black box. File = <structure>.p_prnt	*/
    /* --------------------------------------------------------	*/

    register 	long i,j ;
    float 	dx,dy ;
	
	
    dx = gProjCell->a/gPhsgrtRec->nbeam ;
    dy = gProjCell->b/gPhsgrtRec->mbeam ;
	
    fprintf(fp_print,"%s\n\n\n",gPhsgrtRec->title) ;
    fprintf(fp_print,"Lattice constants are :\n") ;
    fprintf(fp_print,"A = %.4f ,   B = %.4f ,   C = %.4f \n",
    		gProjCell->a,gProjCell->b,gProjCell->c) ;
    fprintf(fp_print,"GAMMA = %.2f Degrees\n\n",gProjCell->ga) ;
    fprintf(fp_print,"Reciprocal Space:\n") ;
    fprintf(fp_print,"AST = %.4f, BST = %.4f, CST = %.4f\n",
	gProjCell->ast,gProjCell->bst,gProjCell->cst) ;
    fprintf(fp_print,"\nVolume of the new unit cell: %f\n\n",gProjCell->vol) ;
    fprintf(fp_print,"\n ------ Transformation Matrix ---------\n\n") ;
    for (j=0;j<3;j++)
    	fprintf(fp_print,"   %6.3f  %6.3f  %6.3f\n",gProjCell->hkltr[j][0],
	         gProjCell->hkltr[j][1],gProjCell->hkltr[j][2]) ;
    fprintf(fp_print,"\n\n- Parameter fit for the electron formfactors -- \n") ;
    if (gUseElScattFactors) {
    	fprintf(fp_print,"- using 8-parameter fit for electrons -- \n") ;
    }
    else {
    	fprintf(fp_print,"- using 9-parameter fit for xrays -- \n") ;
    }
	
    for (i=0;i<gPhsgrtRec->iatom;i++) {
    	fprintf(fp_print,"\n Atom Type # %d  Name: %s",i+1,&name[i][0]) ;
    	fprintf(fp_print,"\n A1: %7.4f    B1: %7.4f",xrf.ax[i][0],
    						     xrf.bx[i][0]) ;
    	fprintf(fp_print,"\n A2: %7.4f    B2: %7.4f",xrf.ax[i][1],
    						     xrf.bx[i][1]) ;
    	fprintf(fp_print,"\n A3: %7.4f    B3: %7.4f",xrf.ax[i][2],
    						     xrf.bx[i][2]) ;
    	fprintf(fp_print,"\n A4: %7.4f    B4: %7.4f",xrf.ax[i][3],
    						     xrf.bx[i][3]) ;
	if (!gUseElScattFactors)
    		fprintf(fp_print,"\n C : %7.4f",xrf.cx[i]) ;
    }
    fprintf(fp_print,"\n\n --- Atom positions are as listed ------ \n\n") ;
    for(i=0;i<gProjCell->natom;i++)
    fprintf(fp_print," %d %s %f %f %f %f %f \n",i+1,name[*(gProjCell->ntype+i)-1],
	*(gProjCell->x+i),*(gProjCell->y+i),*(gProjCell->z+i),
	*(gProjCell->dw+i),*(gProjCell->occ+i)) ;
    fprintf(fp_print,"\n\n The unit cell contains %d atoms\n\n",gProjCell->natom) ;
    fprintf(fp_print,
      "\n\n The x-axis is divided into %d points of %.4f A",gPhsgrtRec->nbeam,dx) ;
    fprintf(fp_print,
      "\n\n The y-axis is divided into %d points of %.4f A",gPhsgrtRec->mbeam,dy) ;
    fprintf(fp_print,
      "\n\n The calculation extends out to %.2f Angstrom-1",gPhsgrtRec->gmax) ;
    fprintf(fp_print,
      "\n\n The unit cell is divided into %d slices of thickness %.2f A",
      gPhsgrtRec->nz,gPhsgrtRec->slcth) ;
}

void Find3DHKLIndicies(long ih,long ik,long il,long *jh,long *jk,long *jl,float r[3][3]) {
    *jh = r[0][0]*ih + r[1][0]*ik + r[2][0]*il ;
    *jk = r[0][1]*ih + r[1][1]*ik + r[2][1]*il ;
    *jl = r[0][2]*ih + r[1][2]*ik + r[2][2]*il ;
}

void ViewOriginalUnitCell(float *rview,long *ierr) {
    /* ---------------------------------------------------------------	*/
    /* View the original unit cell from a specified direction.		*/
    /* ----------------------------------------------------------------	*/

    float 	av[3],bv[3],cv[3],atv[3],btv[3],ctv[3],r[3][3] ;
    long 	istore[MAXTYPES],jz[MAXTYPES] ;
    float 	*xv ;         /* Pointer to an area of xv[gMaxAtoms][3] */
    float 	*xn ;         /* Pointer to an area of xn[gMaxAtoms][3] */
    float 	*xp,*yp,*zp ; /* Pointers to an area of --[gMaxAtoms]   */
    float 	*rad ;        /* Pointer to an area of ---[gMaxAtoms]   */
    float	*spaceHolder ;
    long		*icolor ;
    float 	*fp ;
    long 	nvuat,ndx,stored, (*icompare)() ;
    register 	long i,j,k ;
    float 	radmax,radmin,drad ;
	
    /* Use the workarea for the atom-positions etc. */
    xv = gWorkSpace + gMaxArraySize/2;
    xn = xv + 3*gMaxAtoms ;
    xp = xn + 3*gMaxAtoms ;
    yp = xp + gMaxAtoms ;
    zp = yp + gMaxAtoms ;
    rad = zp + gMaxAtoms ;
    spaceHolder = rad + gMaxAtoms ;
    icolor = (long *)spaceHolder ;
	
    av[0] = 1. ; av[1] = 0. ; av[2] = 0. ;
    bv[0] = 0. ; bv[1] = 1. ; bv[2] = 0. ;
    cv[0] = 0. ; cv[1] = 0. ; cv[2] = 1. ;
	
    /* Add the corner and face atoms for viewing. NVUAT will contain */
    /* the number of of viewable atoms. xv contains the x,y,z        */
    /* coordinates of all atoms	                                     */
    AddUnitCellAtoms((float (*)[3])xv,gPhsgrtRec->x,gPhsgrtRec->y,gPhsgrtRec->z,gPhsgrtRec->atno,
	gPhsgrtRec->natom,&nvuat) ;
	
    /* Set the parameters for the original cell */
    SetOriginalUnitCellParameters(ierr) ;
    if (*ierr < 0) return ;
	
    /* Set up the transformation matrix */
    SetCoordinateTransformationMatrix(rview,0.,1.,gPhsgrtRec->a,gPhsgrtRec->b,gPhsgrtRec->c,xtal.sial,xtal.sibe,xtal.siga,
	   xtal.csal,xtal.csbe,xtal.csga,r) ;
    /* Find the tree axis of the cell */
    TransfVectorToProjectedCell(av,atv,r) ;
    TransfVectorToProjectedCell(bv,btv,r) ;
    TransfVectorToProjectedCell(cv,ctv,r) ;
	
    /* Find the transformed positions of the atoms */
    for(i=0;i<nvuat;i++)
    	TransfVectorToProjectedCell(xv+3*i,xn+3*i,r) ;

    long colorIndex ;
	  	  
    for(i=0,fp=xn;i<nvuat;i++) {
		*(xp+i) = *fp++ ;
		*(yp+i) = *fp++ ;
		*(zp+i) = *fp++ ;
		colorIndex = gPhsgrtRec->atno[i] % 39 ;
		*(icolor+i) = gColors[gNumGrays-1+colorIndex].pixel ;
		/* Assign the atomic radii */
		rad[i]  = 0.5*covrad[gPhsgrtRec->atno[i]-1] ;
    }
	
	
    SortAtomsOnZ(xp,yp,zp,rad,icolor,nvuat) ;  /* sort according to the lowest Z */
    if (ERASE && OutDevice->type == WS)
		ClearDisplay() ;
    cell3c(atv,btv,ctv) ;
    DrawAtoms(xp,yp,zp,rad,icolor,nvuat) ;
    cell3c(atv,btv,ctv) ;
}

void SetProjectedCell(long *ierr) {
    /* ----------------------------------------------------------------	*/
    /* This routine gets called before the projected potential is 	*/
    /* calculated. It sets up a new unit cell with the c-axis in the	*/
    /* direction of the incident electron beam. The transformed cell	*/
    /* is stored in the struct 'gProjCell-> and transferred to the struct	*/
    /* 'gPhsgrtRec-> before the potential is calculated. The entries for	*/
    /* the coordinates could probalbly be eliminated in the 'gPhsgrtRec->	*/
    /* since I think they are all present in 'gProjCell->. When I get time	*/
    /* I should check this. It should free up some memory.		*/
    /* ---------------------------------------------------------------	*/
    register 	long i,j ;
    float 	gsq[2000] ;
    long 	inx[2],iny[2],inz[2],ihkl[2000][3],icount ;
    float 	atv[3],btv[3],ctv[3],r[3][3] ;


	
    /* Setup the parameters for the original cell */
    SetOriginalUnitCellParameters(ierr) ;
    if (*ierr < 0) return ;
	
    /* If it is trivial (ZONE=001 and all angles 90.0 degrees) then */
    /* we just use the old cell					    */
    if (gPhsgrtRec->iuvw[0]==0 && gPhsgrtRec->iuvw[1]==0 && gPhsgrtRec->iuvw[2]==1 &&
			gPhsgrtRec->alpha==90. && gPhsgrtRec->beta==90.) {
		ProjectedCellEqualsOriginal() ;
		for(j=0;j<3;j++)
			for(i=0;i<3;i++) {
				if (i!=j) {
					gProjCell->hkltr[i][j] = 0. ;
					gTMatrix.m[i][j] = 0. ;
				}
				else {
					gProjCell->hkltr[i][i] = 1. ;
					gTMatrix.m[i][i] = 1. ;
				}
			}
		set_ohkl_nhkl(gProjCell->hkltr) ;/* Set the inverse tramsformation matrix */
		     /* to transform form old hkl to new hkl  */
		return ;							  
	}
	
    /* Set all the hkl that are perpendicular to the zone axis */
    SetPerpendicularReflections(ihkl,gsq,&icount,gPhsgrtRec->iuvw) ;
	
    /* sort out the g_vectors in increasing order of magnitude */
    /* and find the two shortest ones                          */
    SortHKLVectors(ihkl,gsq,icount) ;
	
    /* Find a new projected unit cell */
    CalculateProjectedCell(atv,btv,ctv,ihkl,gsq,icount,r,inx,iny,inz) ;
    for(i=0;i<3;i++) {
	gTMatrix.m[0][i] = atv[i] ;
	gTMatrix.m[1][i] = btv[i] ;
	gTMatrix.m[2][i] = ctv[i] ;
    }
	
    if (LAYERED)  return ;  /* We are only interested in te unit cell */
			    /* There are no atoms in this case		*/
							  
    /*  INX		The min and max values of the times we have to go in */
    /*  INY     the X,Y and Z directions to entirely enclose the new         */
    /*  INZ     unit cells.						     */
	
    /* Get all the atoms within the new cell by keeping the ones within */
    /* the spanning supercell that lies within the new cell		*/
    gProjCell->natom = gPhsgrtRec->natom ;
    ExpandUnitCell(gPhsgrtRec->x,gPhsgrtRec->y,gPhsgrtRec->z,gPhsgrtRec->atno,gPhsgrtRec->dw,
	gPhsgrtRec->occf,gPhsgrtRec->ntype,
    	gProjCell->x,gProjCell->y,gProjCell->z,gProjCell->atno,gProjCell->dw,gProjCell->occ,
    	gProjCell->ntype,&gProjCell->natom,inx,iny,inz,r) ;
	
    /* Sort the atoms such that all atoms of the same type are listed together*/
    SortAtoms(gProjCell->x,gProjCell->y,gProjCell->z,gProjCell->atno,gProjCell->dw,gProjCell->occ,
    gProjCell->ntype,gProjCell->natom) ;
	
    set_ohkl_nhkl(gProjCell->hkltr) ;
}

void SortAtoms(float *x,float *y,float *z,short *iz,float *dw,float *occ,short *ntype,long n) {
    /* ---------------------------------------------------- */
    /* SORT THE ATOMS ACCORDING TO LOWEST 'TYPE' OF ATOM    */
    /* ---------------------------------------------------- */
    long 	gap ;
    register 	long i,j,itemp ;
    register 	float temp ;
	
    for (gap=n/2;gap>0;gap/=2)
    	for(i=gap;i<n;i++)
		for(j=i-gap;j>=0;j-=gap) {
			if (ntype[j] <= ntype[j+gap])
				break ;
			temp = x[j] ;
			x[j] = x[j+gap] ;
			x[j+gap] = temp ;
			temp = y[j] ;
			y[j] = y[j+gap] ;
			y[j+gap] = temp ;
			temp = z[j] ;
			z[j] = z[j+gap] ;
			z[j+gap] = temp ;
			temp = dw[j] ;
			dw[j] = dw[j+gap] ;
			dw[j+gap] = temp ;
			temp = occ[j] ;
			occ[j] = occ[j+gap] ;
			occ[j+gap] = temp ;
			itemp = ntype[j] ;
			ntype[j] = ntype[j+gap] ;
			ntype[j+gap] = itemp ;
			itemp = iz[j] ;
			iz[j] = iz[j+gap] ;
			iz[j+gap] = itemp ;
		}
}

void SortGVectors(float *f,float *g,short *ihk,long n) {
    /* ---------------------------------------------------- */
    /* SORT THE REFLECTIONS ACCORDING TO THEIR G-VALUE      */
    /* STARTING WITH THE LOWEST G			    */
    /* ---------------------------------------------------- */
    long 	gap ;
    register 	long i,j,itemp ;
    register 	float temp ;
	
    for (gap=n/2;gap>0;gap/=2)
    	for(i=gap;i<n;i++)
    		for(j=i-gap;j>=0;j-=gap) {
			if (g[j] <= g[j+gap])
				break ;
			temp = g[j] ;
			g[j] = g[j+gap] ;
			g[j+gap] = temp ;
			temp = f[j] ;
			f[j] = f[j+gap] ;
			f[j+gap] = temp ;
			itemp = *(ihk+2*j) ;
			*(ihk+2*j) = *(ihk+2*(j+gap)) ;
			*(ihk+2*(j+gap)) = itemp ;
			itemp = *(ihk+2*j+1) ;
			*(ihk+2*j+1) = *(ihk+2*(j+gap)+1) ;
			*(ihk+2*(j+gap)+1) = itemp ;
		}
}

void SortAtomsOnZ(float *x,float *y,float *z,float *r,long *icolor,long n) {
    /* ---------------------------------------------------- */
    /* SORT THE ATOMS ACCORDING TO	HIGHEST Z VALUE     */
    /* ---------------------------------------------------- */
    long gap ;
    register long i,j,itemp ;
    register float temp ;
	
    for (gap=n/2;gap>0;gap/=2)
    	for(i=gap;i<n;i++)
		for(j=i-gap;j>=0;j-=gap) {
			if (z[j] >= z[j+gap])
				break ;
			temp = z[j] ;
			z[j] = z[j+gap] ;
			z[j+gap] = temp ;
			temp = x[j] ;
			x[j] = x[j+gap] ;
			x[j+gap] = temp ;
			temp = y[j] ;
			y[j] = y[j+gap] ;
			y[j+gap] = temp ;
			temp = r[j] ;
			r[j] = r[j+gap] ;
			r[j+gap] = temp ;
			itemp = icolor[j] ;
			icolor[j] = icolor[j+gap] ;
			icolor[j+gap] = itemp ;
		}
}

void ProjectedCellEqualsOriginal() {
    /* -------------------------------------------------------	*/
    /* In the case that we are down the [001] and the angles	*/
    /* are 90 degrees, we do not have to perform any transf.	*/
    /* on the cell. We just set up the structure 'gProjCell-> 	*/
    /* --------------------------------------------------------	*/
    register long i ;

    float	aa = 10.4 ;
    gProjCell->natom = gPhsgrtRec->natom ;
    gProjCell->a = gPhsgrtRec->a ;
    gProjCell->b = gPhsgrtRec->b ;
    gProjCell->c = gPhsgrtRec->c ;
    gProjCell->al = gPhsgrtRec->alpha ;
    gProjCell->be = gPhsgrtRec->beta ;
    gProjCell->ga = gPhsgrtRec->gamma ;
    gProjCell->ast = xtal.ast ;
    gProjCell->bst = xtal.bst ;
    gProjCell->cst = xtal.cst ;
    gProjCell->siga = xtal.siga ;
    gProjCell->csga = xtal.csga ;
    gProjCell->vol = xtal.v ;
    for(i=0;i<gProjCell->natom;i++) {
		*(gProjCell->x+i) = *(gPhsgrtRec->x+i) ;
		*(gProjCell->y+i) = *(gPhsgrtRec->y+i) ;
		*(gProjCell->z+i) = *(gPhsgrtRec->z+i) ;
		*(gProjCell->atno+i) = *(gPhsgrtRec->atno+i) ;
		*(gProjCell->dw+i) = *(gPhsgrtRec->dw+i) ;
		*(gProjCell->occ+i) = *(gPhsgrtRec->occf+i) ;
		*(gProjCell->ntype+i) = *(gPhsgrtRec->ntype+i) ;
    }
}

void DrawNewUnitCell(float *rview,long *ierr) {
    /* --------------------------------------------------------	*/
    /* This routine prepares the atoms so they can be viewed	*/
    /* in the context of the transformed cell which the 	*/
    /* calculation is going to use. This allows the user to see	*/
    /* how the unit cell looks in the direction of the electron */
    /* beam.							*/
    /* -------------------------------------------------------- */

    float 	av[3],bv[3],cv[3],atv[3],btv[3],ctv[3],r[3][3] ;
    float 	*xv ;         /* Pointer to an area of xv[gMaxAtoms][3] */
    float 	*xn ;         /* Pointer to an area of xn[gMaxAtoms][3] */
    float 	*xp,*yp,*zp ; /* Pointers to an area of --[gMaxAtoms]   */
    float 	*rad ;        /* Pointer to an area of ---[gMaxAtoms]   */
    float 	*fp ;
    float 	*spaceHolder ;
    long	*icolor ;
    long 	nvuat ;
    register 	long i ;
	
    /* Use the workarea for the atom-positions etc. */
    xv = gWorkSpace + gMaxArraySize/2 ;
    xn = xv + 3*gMaxAtoms ;
    xp = xn + 3*gMaxAtoms ;
    yp = xp + gMaxAtoms ;
    zp = yp + gMaxAtoms ;
    rad = zp + gMaxAtoms ;
    spaceHolder = rad + gMaxAtoms ;
    icolor = (long *)spaceHolder ;
	
    av[0] = 1. ; av[1] = 0. ; av[2] = 0. ;
    bv[0] = 0. ; bv[1] = 1. ; bv[2] = 0. ;
    cv[0] = 0. ; cv[1] = 0. ; cv[2] = 1. ;
	
	
    /* Set the parameters for the original cell */
    SetProjectedCell(ierr) ;
    if (*ierr < 0) return ;
	
    /* -----------------------------------------------------------------  */
    /* Add the corner and face atoms for viewing. NVUAT will contain the  */
    /* number of of viewable atoms. xv contains the x,y,z coordinates of  */
    /* all atoms	    						  */
    /* -----------------------------------------------------------------  */
    AddUnitCellAtoms((float (*)[3])xv,gProjCell->x,gProjCell->y,gProjCell->z,gProjCell->atno,
		gProjCell->natom,&nvuat) ;

    /* Set up the transformation matrix */
    SetCoordinateTransformationMatrix(rview,1.,0.,gProjCell->a,gProjCell->b,gProjCell->c,1.,1.,gProjCell->siga,
    	   0.,0.,gProjCell->csga,r) ;
    /* Find the tree axis of the cell */
    TransfVectorToProjectedCell(av,atv,r) ;
    TransfVectorToProjectedCell(bv,btv,r) ;
    TransfVectorToProjectedCell(cv,ctv,r) ;
	
    /* Find the transformed positions of the atoms */
    for(i=0;i<nvuat;i++)
	  TransfVectorToProjectedCell(xv+3*i,xn+3*i,r) ;
	
    long colorIndex ;  
    for(i=0,fp=xn;i<nvuat;i++) {
	*(xp+i) = *fp++ ;
	*(yp+i) = *fp++ ;
	*(zp+i) = *fp++ ;
	colorIndex = gProjCell->atno[i] % 39 ;
	*(icolor+i) = gColors[gNumGrays-1+colorIndex].pixel ;
	/* Assign the atomic radii */
	*(rad+i)  = 0.5*covrad[*(gProjCell->atno+i)-1] ;
    }
	
	
    SortAtomsOnZ(xp,yp,zp,rad,icolor,nvuat) ;  /* sort according to the lowest Z */
    cell3c(atv,btv,ctv) ;
    DrawAtoms(xp,yp,zp,rad,icolor,nvuat) ;
    cell3c(atv,btv,ctv) ;
}

void AddUnitCellAtoms(float x[][3],float *x0,float *y0,float *z0,short *iz0,
	long num_atoms,long *nview_atoms) {
    /* ----------------------------------------------------------------	*/
    /* This routine just adds the atoms at the corners and at the faces	*/
    /* of the unit cell for viewing purposes. That is, for the atom at 	*/
    /* (0,0,0) it adds atoms at (1,0,0), (0,1,0) etc....		*/
    /* ----------------------------------------------------------------	*/
    register long i,j ;
	
    for(i=0,j=0;i<num_atoms;i++) {
	x[i][0] = x0[i] ;
	x[i][1] = y0[i] ;
	x[i][2] = z0[i] ;
	if (fabs(x[i][0]) < 0.001) x[i][0] = 0. ;
	if (fabs(x[i][1]) < 0.001) x[i][1] = 0. ;
	if (fabs(x[i][2]) < 0.001) x[i][2] = 0. ;
	if (x[i][0] == 0. || x[i][1] == 0. || x[i][2] == 0.) {
		if (x[i][0]==0. && x[i][1] != 0. && x[i][2] != 0.) {
			x[num_atoms+j][0] = 1. ;
			x[num_atoms+j][1] = x[i][1] ;
			x[num_atoms+j][2] = x[i][2] ;
			iz0[num_atoms+j] = iz0[i] ;
			j++ ;
		}
		else if (x[i][0]==0. && x[i][1]==0. && x[i][2]!=0.) {
			x[num_atoms+j][0] = 1. ;
			x[num_atoms+j][1] = x[i][1] ;
			x[num_atoms+j][2] = x[i][2] ;
			iz0[num_atoms+j] = iz0[i] ;
			j++ ;
			x[num_atoms+j][0] = x[i][0] ;
			x[num_atoms+j][1] = 1. ;
			x[num_atoms+j][2] = x[i][2] ;
			iz0[num_atoms+j] = iz0[i] ;
			j++ ;
			x[num_atoms+j][0] = 1. ;
			x[num_atoms+j][1] = 1. ;
			x[num_atoms+j][2] = x[i][2] ;
			iz0[num_atoms+j] = iz0[i] ;
			j++ ;
		}
		else if (x[i][0]==0. && x[i][1]==0. && x[i][2]==0.) {
			x[num_atoms+j][0] = 1. ;
			x[num_atoms+j][1] = x[i][1] ;
			x[num_atoms+j][2] = x[i][2] ;
			iz0[num_atoms+j] = iz0[i] ;
			j++ ;
			x[num_atoms+j][0] = 1. ;
			x[num_atoms+j][1] = 1. ;
			x[num_atoms+j][2] = x[i][2] ;
			iz0[num_atoms+j] = iz0[i] ;
			j++ ;
			x[num_atoms+j][0] = x[i][0] ;
			x[num_atoms+j][1] = 1. ;
			x[num_atoms+j][2] = x[i][2] ;
			iz0[num_atoms+j] = iz0[i] ;
			j++ ;
			x[num_atoms+j][0] = x[i][0] ;
			x[num_atoms+j][1] = x[i][1] ;
			x[num_atoms+j][2] = 1. ;
			iz0[num_atoms+j] = iz0[i] ;
			j++ ;
			x[num_atoms+j][0] = 1. ;
			x[num_atoms+j][1] = x[i][1] ;
			x[num_atoms+j][2] = 1. ;
			iz0[num_atoms+j] = iz0[i] ;
			j++ ;
			x[num_atoms+j][0] = x[i][0] ;
			x[num_atoms+j][1] = 1. ;
			x[num_atoms+j][2] = 1. ;
			iz0[num_atoms+j] = iz0[i] ;
			j++ ;
			x[num_atoms+j][0] = 1. ;
			x[num_atoms+j][1] = 1. ;
			x[num_atoms+j][2] = 1. ;
			iz0[num_atoms+j] = iz0[i] ;
			j++ ;
    		}
		else if (x[i][1]==0. && x[i][0]!=0. &&x[i][2]!=0.) {
			x[num_atoms+j][0] = x[i][0] ;
			x[num_atoms+j][1] = 1. ;
			x[num_atoms+j][2] = x[i][2] ;
			iz0[num_atoms+j] = iz0[i] ;
			j++ ;
		}
		else if (x[i][1]==0. && x[i][0]!=0. && x[i][2]==0.) {
			x[num_atoms+j][0] = x[i][0] ;
			x[num_atoms+j][1] = 1. ;
			x[num_atoms+j][2] = x[i][2] ;
			iz0[num_atoms+j] = iz0[i] ;
			j++ ;
			x[num_atoms+j][0] = x[i][0] ;
			x[num_atoms+j][1] = 1. ;
			x[num_atoms+j][2] = 1. ;
			iz0[num_atoms+j] = iz0[i] ;
			j++ ;
			x[num_atoms+j][0] = x[i][0] ;
			x[num_atoms+j][1] = x[i][1] ;
			x[num_atoms+j][2] = 1. ;
			iz0[num_atoms+j] = iz0[i] ;
			j++ ;
		}
		else if (x[i][2]==0. && x[i][0]!=0. && x[i][1]!=0.) {
			x[num_atoms+j][0] = x[i][0] ;
			x[num_atoms+j][1] = x[i][1] ;
			x[num_atoms+j][2] = 1. ;
			iz0[num_atoms+j] = iz0[i] ;
			j++ ;
		}
		else if (x[i][2]==0. && x[i][0]==0. && x[i][1]!=0.) {
			x[num_atoms+j][0] = 1. ;
			x[num_atoms+j][1] = x[i][1] ;
			x[num_atoms+j][2] = x[i][2] ;
			iz0[num_atoms+j] = iz0[i] ;
			j++ ;
			x[num_atoms+j][0] = x[i][0] ;
			x[num_atoms+j][1] = x[i][1] ;
			x[num_atoms+j][2] = 1. ;
			iz0[num_atoms+j] = iz0[i] ;
			j++ ;
			x[num_atoms+j][0] = 1. ;
			x[num_atoms+j][1] = x[i][1] ;
			x[num_atoms+j][2] = 1. ;
			iz0[num_atoms+j] = iz0[i] ;
			j++ ;
		}
	}
    }
    *nview_atoms = num_atoms+j ;
}

void ExpandUnitCell(float *x0,float *y0,float *z0,short *iaz0,float *dw0,float *occ0,
	short *nt0,float *x,float *y,float *z,short *iaz,float *dw,float *occ,short *nt,
	long *natom,long *inx,long *iny,long *inz,float r[3][3]) {
  /* ----------------------------------------------------------------- */
  /*   This routine will create a supercell spanned by the vectors     */
  /*   (inx(1),inx(2))A, (iny(1),iny(2))B and (inz(1),inz(2))C         */
  /*   and calculate all the atoms in this volume, discarding the      */
  /*   ones that fall outside the new unit cell. It transforms the     */
  /*   coordinates according to the new cell in the process            */
  /* ---------------------------------------------------------------   */
    register long j,iz,ix,iy,ndx ;
    float rx,ry,rz ;
	
    for(ndx=0,iz=inz[0];iz<=inz[1];iz++)
	for(iy=iny[0];iy<=iny[1];iy++)
	  	for(ix=inx[0];ix<=inx[1];ix++)
		  	for(j=0;j<*natom;j++) {
				rx = x0[j] + (float)(ix-1) ; 
				ry = y0[j] + (float)(iy-1) ; 
				rz = z0[j] + (float)(iz-1) ;
				if(ix < 1 && inx[1] < 1 && ((rx-(long)rx)==0.))
    						 rx += 1. ;
				if(iy < 1 && iny[1] < 1 && ((ry-(long)ry)==0.))
    						 ry += 1. ;
				if(iz < 1 && inz[1] < 1 && ((rz-(long)rz)==0.))
    						 rz += 1. ;
				TransformAtomCoordinates(&rx,&ry,&rz,1,r) ;
				if (rx>=0. && rx<0.999 && ry >=0. && 
    				    ry<0.999 && rz>=0. && rz < 0.999) {
						x[ndx] = rx ;
						y[ndx] = ry ;
						z[ndx] = rz ;
						iaz[ndx] = iaz0[j] ;
						dw[ndx] = dw0[j] ;
						occ[ndx] = occ0[j] ;
						nt[ndx] = nt0[j] ;
						ndx++ ;
				}
			}
    *natom = ndx ;
}

void TranslateIntoUnitCell(float *x,float *y,float *z,long natom) {
    /* ------------------------------------------------------------------ */
    /* This little routine makes sure all atoms lie within the unit cell  */
    /* ------------------------------------------------------------------ */
    register long i,ix,iy,iz ;

    for(i=0;i<natom;i++) {
	ix = near(x[i]) ;
	iy = near(y[i]) ;
	iz = near(z[i]) ;
	if (fabs(x[i]-ix) < 0.001) x[i] = 0. ;
	if (fabs(y[i]-iy) < 0.001) y[i] = 0. ;
	if (fabs(z[i]-iz) < 0.001) z[i] = 0. ;
	x[i] = one(x[i]) ;
	y[i] = one(y[i]) ;
	z[i] = one(z[i]) ;
    }
}

void TransformAtomCoordinates(float *x,float *y,float *z,long n,float r[3][3]) {
    /* --------------------------------------------------------	*/
    /* Apply the transformation matrix 'r' on the 'n' atoms	*/
    /* whose coordinates are stored in 'x','y' and 'z'.		*/
    /* --------------------------------------------------------	*/
    register long i ;
    register float tx,ty,tz ;
	
    for(i=0;i<n;i++) {
	tx = x[i] ;
	ty = y[i] ;
	tz = z[i] ;
	x[i] = r[0][0]*tx + r[1][0]*ty + r[2][0]*tz ;
	y[i] = r[0][1]*tx + r[1][1]*ty + r[2][1]*tz ;
	z[i] = r[0][2]*tx + r[1][2]*ty + r[2][2]*tz ;
	if (fabs(x[i]) < 0.000001)
		x[i] = 0. ;
	if (fabs(y[i]) < 0.000001)
		y[i] = 0. ;
	if (fabs(z[i]) < 0.000001)
		z[i] = 0. ;
    }
}

void SetOriginalUnitCellParameters(long *ierr) {
    float pi=3.14159265,alr,ber,gar,arg ;
	
    if (gPhsgrtRec->alpha != 90. || gPhsgrtRec-> beta != 90. || gPhsgrtRec->gamma != 90.) {
		alr = pi*gPhsgrtRec->alpha/180. ;
		ber = pi*gPhsgrtRec->beta/180. ;
		gar = pi*gPhsgrtRec->gamma/180. ;
		xtal.csal = cos(alr) ;
		xtal.sial = sin(alr) ;
		xtal.csbe = cos(ber) ;
		xtal.sibe = sin(ber) ;
		xtal.csga = cos(gar) ;
		xtal.siga = sin(gar) ;
		arg = 1.-xtal.csal*xtal.csal-xtal.csbe*xtal.csbe-xtal.csga*xtal.csga +
				2*xtal.csal*xtal.csbe*xtal.csga ;
		if(arg > 0.)
			xtal.v = gPhsgrtRec->a*gPhsgrtRec->b*gPhsgrtRec->c*sqrt(arg) ;
		else {
			*ierr = -1 ;
			return ;
	    	}
		xtal.csalst = (xtal.csbe*xtal.csga-xtal.csal)/(xtal.sibe*xtal.siga) ;
		xtal.csbest = (xtal.csga*xtal.csal-xtal.csbe)/(xtal.siga*xtal.sial) ;
		xtal.csgast = (xtal.csal*xtal.csbe-xtal.csga)/(xtal.sial*xtal.sibe) ;
		xtal.sialst = sqrt(1.-xtal.csalst*xtal.csalst) ;
		xtal.sibest = sqrt(1.-xtal.csbest*xtal.csbest) ;
		xtal.sigast = sqrt(1.-xtal.csgast*xtal.csgast) ;
		xtal.ast = gPhsgrtRec->b*gPhsgrtRec->c*xtal.sial/xtal.v ;
		xtal.bst = gPhsgrtRec->c*gPhsgrtRec->a*xtal.sibe/xtal.v ;
		xtal.cst = gPhsgrtRec->a*gPhsgrtRec->b*xtal.siga/xtal.v ;
		arg = 1.-xtal.csalst*xtal.csalst-xtal.csbest*xtal.csbest-
	                xtal.csgast*xtal.csgast+
			2.*xtal.csalst*xtal.csbest*xtal.csgast ;
		if (arg > 0.) 
			xtal.vst = xtal.ast*xtal.bst*xtal.cst*sqrt(arg) ;
		else {
			*ierr = -2 ;
			return ;
		}
		*ierr = 0 ;
		return ;
    }
    xtal.sial = 1. ;
    xtal.sibe = 1. ;
    xtal.siga = 1. ;
    xtal.csal = 0. ;
    xtal.csbe = 0. ;
    xtal.csga = 0. ;
    xtal.v = gPhsgrtRec->a*gPhsgrtRec->b*gPhsgrtRec->c ;
    xtal.ast = 1./gPhsgrtRec->a ;
    xtal.bst = 1./gPhsgrtRec->b ;
    xtal.cst = 1./gPhsgrtRec->c ;
    xtal.alst = 90. ;
    xtal.best = 90. ;
    xtal.gast = 90. ;
    xtal.sialst = 1. ;
    xtal.sibest = 1. ;
    xtal.sigast = 1. ;
    xtal.csalst = 0. ;
    xtal.csbest = 0. ;
    xtal.csgast = 0. ;
    xtal.vst = 1./xtal.v ;
	
    *ierr = 0 ;
}

void SortHKLVectors(long ihkl[][3],float *gsq,long n) {
    long gap ;
    register long i,j,l,itemp ;
    register float temp ;

    for (gap=n/2;gap>0;gap/=2)
	for(i=gap;i<n;i++)
		for(j=i-gap;j>=0;j-=gap) {
			if (gsq[j] <= gsq[j+gap])
				break ;
			temp = gsq[j] ;
			gsq[j] = gsq[j+gap] ;
			gsq[j+gap] = temp ;
			for(l=0;l<3;l++) {
				itemp = ihkl[j][l] ;
				ihkl[j][l] = ihkl[j+gap][l] ;
				ihkl[j+gap][l] = itemp ;
			}
		  }
}	

void CrossProduct(float *v1,float *v2,float vol,float *v3) {
    /* ----------------------------------------------	*/
    /* Take the crossproduct of two vectors.		*/
    /* ----------------------------------------------	*/

    v3[0] = (v1[1]*v2[2]-v1[2]*v2[1])*vol ;
    v3[1] = (v1[2]*v2[0]-v1[0]*v2[2])*vol ;
    v3[2] = (v1[0]*v2[1]-v1[1]*v2[0])*vol ;
}

void DotProduct(float *v1,float *v2,float a,float b,float c,
	float csa,float csb,float csg,float *dot) {
    /* -----------------------------------------------	*/
    /* Calculate the dot-product of two vectors		*/
    /* ------------------------------------------------	*/

    *dot = v1[0]*v2[0]*a*a+v1[1]*v2[1]*b*b+v1[2]*v2[2]*c*c +
    	(v1[1]*v2[2]+v1[2]*v2[1])*b*c*csa+
	(v1[0]*v2[2]+v1[2]*v2[0])*a*c*csb+
	(v1[1]*v2[0]+v1[0]*v2[1])*a*b*csg ;
}

void SetHKLTransformationMatrix(float *v1,float *v2,float *v3,float r[3][3]) {
    /* -----------------------------------------------	*/
    /* Set the transformation matrix that transforms	*/
    /* hkl's in the conventional unit cell into two	*/
    /* dimensional hk's in the projected cell in the	*/
    /* zone axis orientation.				*/
    /* ------------------------------------------------	*/

    r[0][0] = v1[0] ;
    r[0][1] = v1[1] ;
    r[0][2] = v1[2] ;
    r[1][0] = v2[0] ;
    r[1][1] = v2[1] ;
    r[1][2] = v2[2] ;
    r[2][0] = v3[0] ;
    r[2][1] = v3[1] ;
    r[2][2] = v3[2] ;
}

void SetPerpendicularReflections(long ihkl[][3],float *gsq,long *ic,long *iuvw) {
    /* -----------------------------------------------	*/
    /* Set the set of k,k,l's that are perpendicular to */
    /* the zone axis orientation iuvw[]			*/
    /* ------------------------------------------------	*/

    long 	h,k,l,count,nl,nh,nk ;
    long 	maxh,maxk,maxl,iu,iv,iw ;
    float 	astsq,bstsq,cstsq,rl,rk,rh ;
	
    maxh = 16 ;
    maxk = 16 ;
    maxl = 16 ;
    iu = iuvw[0] ;
    iv = iuvw[1] ;
    iw = iuvw[2] ;

    astsq = xtal.ast*xtal.ast ;
    bstsq = xtal.bst*xtal.bst ;
    cstsq = xtal.cst*xtal.cst ;
	
    count=0 ;
    if (iu==0)  {		/* U is zero */
    	if(iv==0) {
		l = 0 ;
		for(h= (-maxh);h<=maxh;h++)
			for (k= (-maxk);k<=maxk;k++) {
				ihkl[count][0] = h ;
				ihkl[count][1] = k ;
				ihkl[count][2] = l ;
				gsq[count++] = h*h*astsq+k*k*bstsq+l*l*cstsq+
				  	 2*h*k*xtal.ast*xtal.bst*xtal.csgast+
				  	 2*h*l*xtal.ast*xtal.cst*xtal.csbest+
				  	 2*k*l*xtal.bst*xtal.cst*xtal.csalst  ;
				if (count == 2000)  {
					*ic = count ;
					return ;
				}
			}
	}
	else if (iw==0) {
		k = 0 ;
		for(h= (-maxh);h<=maxh;h++)
			for(l= (-maxl);l<=maxl;l++) {
				ihkl[count][0] = h ;
				ihkl[count][1] = k ;
				ihkl[count][2] = l ;
				gsq[count++] = h*h*astsq+k*k*bstsq+l*l*cstsq+
				  	 2*h*k*xtal.ast*xtal.bst*xtal.csgast+
				  	 2*h*l*xtal.ast*xtal.cst*xtal.csbest+
				  	 2*k*l*xtal.bst*xtal.cst*xtal.csalst  ;
				if (count == 2000)  {
					*ic = count ;
					return ;
				}
			}
	}
	else {
		for(h= (-maxh);h<=maxh;h++)
			for(k= (-maxk);k<=maxk;k++) {
				rl = -(float)(iv*k)/iw ;
				l = rl ;
				if((rl-l) ==0.) {
					ihkl[count][0] = h ;
					ihkl[count][1] = k ;
					ihkl[count][2] = l ;
					gsq[count++] = h*h*astsq+k*k*bstsq+
    					l*l*cstsq+
    					2*h*k*xtal.ast*xtal.bst*xtal.csgast+
		  	 		2*h*l*xtal.ast*xtal.cst*xtal.csbest+
		  	 		2*k*l*xtal.bst*xtal.cst*xtal.csalst  ;
					if (count == 2000)  {
						*ic = count ;
						return ;
					}
				}
			}
       }
    }
    else if (iv==0)    /* U is non_zero and V is zero */ {
	if (iw==0) {
		h = 0 ;
		for(k= (-maxk);k<=maxk;k++)
			for(l= (-maxl);l<=maxl;l++) {
				ihkl[count][0] = h ;
				ihkl[count][1] = k ;
				ihkl[count][2] = l ;
				gsq[count++] = h*h*astsq+k*k*bstsq+l*l*cstsq+
				  	 2*h*k*xtal.ast*xtal.bst*xtal.csgast+
				  	 2*h*l*xtal.ast*xtal.cst*xtal.csbest+
				  	 2*k*l*xtal.bst*xtal.cst*xtal.csalst  ;
    				if (count == 2000)  {
					*ic = count ;
					return ;
				}
			}
	}
	else {
		for(k= (-maxk);k<=maxk;k++)
		for(h= (-maxh);h<=maxh;h++) {
				rl = -(float)(iu*h)/iw ;
				l = rl ;
				if((rl-l) ==0.) {
					ihkl[count][0] = h ;
					ihkl[count][1] = k ;
					ihkl[count][2] = l ;
					gsq[count++] = h*h*astsq+k*k*bstsq+
    					l*l*cstsq+
    					2*h*k*xtal.ast*xtal.bst*xtal.csgast+
		  	 		2*h*l*xtal.ast*xtal.cst*xtal.csbest+
		  	 		2*k*l*xtal.bst*xtal.cst*xtal.csalst  ;
					if (count == 2000) {
						*ic = count ;
						return ;
					}
				}
		}
    	}  
    }
    else     /* U is non-zero and V is non-zero */ {
	if (iw==0) {
		for(l= (-maxl);l<=maxl;l++)
		for(h= (-maxh);h<=maxh;h++) {
			rk = -(float)(iu*h)/iv ;
			k = rk ;
			if ((rk-k) == 0.) {
				ihkl[count][0] = h ;
				ihkl[count][1] = k ;
				ihkl[count][2] = l ;
				gsq[count++] = 	h*h*astsq+k*k*bstsq+l*l*cstsq+
    				 	2*h*k*xtal.ast*xtal.bst*xtal.csgast+
	  	 			2*h*l*xtal.ast*xtal.cst*xtal.csbest+
	  	 			2*k*l*xtal.bst*xtal.cst*xtal.csalst  ;
    					if (count == 2000)  {
						*ic = count ;
						return ;
					}
			}
		}
	}
	else {
		for(h= (-maxh);h<=maxh;h++)
			for(k= (-maxk);k<=maxk;k++) {
				rl = -(float)(iu*h+iv*k)/iw ;
				l = rl ;
				if ((rl-l)==0.) {
					ihkl[count][0] = h ;
					ihkl[count][1] = k ;
					ihkl[count][2] = l ;
					gsq[count++] = h*h*astsq+k*k*bstsq
    					+l*l*cstsq+
    					2*h*k*xtal.ast*xtal.bst*xtal.csgast+
		  	 		2*h*l*xtal.ast*xtal.cst*xtal.csbest+
		  	 		2*k*l*xtal.bst*xtal.cst*xtal.csalst  ;
    					if (count == 2000)  {
						*ic = count ;
						return ;
					}
    				}
			}
	}
    }
    *ic = count ;
}

void SetCoordinateTransformationMatrix(float *ruvw,float siang,float csang,float a,float b,float c,
	float sial,float sibe,float siga,float csal,float csbe,
	float csga,float r[3][3]) {
    float 	rt[3][3],aa[3][3],bb[3][3],cc[3][3],u,v,w,siphi,
		csphi,sith,csth ;
    register 	long i,j,k ;
	
    u = ruvw[1]*b*siga+ruvw[2]*c*(csal-csga*csbe)/siga ;
    v = ruvw[2]*c*sqrt(siga*siga-csal*csal-csbe*csbe+2.*csal*csbe*csga)/siga ;
    w = ruvw[0]*a+ruvw[1]*b*csga+ruvw[2]*c*csbe ;
	
    if(u==0. && w==0.) {
	siphi = 0. ;
	csphi = 1. ;
    }
    else {
	siphi = u/sqrt(u*u+w*w) ;
	csphi = w/sqrt(u*u+w*w) ;
    }
    sith = v/sqrt(u*u+v*v+w*w) ;
    csth = sqrt(u*u+w*w)/sqrt(u*u+v*v+w*w) ;
	
    aa[0][0] = csang ;
    aa[1][0] = -siang ;
    aa[2][0] = 0. ;
    aa[0][1] = siang ;
    aa[1][1] = csang ;
    aa[2][1] = 0. ;
    aa[0][2] = 0. ;
    aa[1][2] = 0. ;
    aa[2][2] = 1. ;
	
    bb[0][0] = csphi ;
    bb[1][0] = 0. ;
    bb[2][0] = -siphi ;
    bb[0][1] = -sith*siphi ;
    bb[1][1] = csth ;
    bb[2][1] = -sith*csphi ;
    bb[0][2] = csth*siphi ;
    bb[1][2] = sith ;
    bb[2][2] = csth*csphi ;
	
    cc[0][0] = 0. ;
    cc[1][0] = b*siga ;
    cc[2][0] = c*(csal-csga*csbe)/siga ;
    cc[0][1] = 0. ;
    cc[1][1] = 0. ;
    cc[2][1] = c*sqrt(siga*siga-csal*csal-csbe*csbe+2.*csal*csbe*csga)/siga ;
    cc[0][2] = a ;
    cc[1][2] = b*csga ;
    cc[2][2] = c*csbe ;
	
    for(i=0;i<3;i++)
    	for(j=0;j<3;j++) {
		rt[j][i] = 0. ;
		for(k=0;k<3;k++) {
			rt[j][i] += bb[k][i]*cc[j][k] ;
		}
	}
			
    for(i=0;i<3;i++)
	for(j=0;j<3;j++) {
		r[j][i] = 0. ;
		for(k=0;k<3;k++) {
			r[j][i] += aa[k][i]*rt[j][k] ;
		}
	}
}

void TransfVectorToProjectedCell(float *x,float *xtr,float r[3][3]){
    /* ----------------------------------------------	*/
    /* Rotate a set of coordinates by applying the	*/
    /* ratation matrix 'r'.				*/
    /* ------------------------------------------------	*/

    register long i,j ;
	
    for(i=0;i<3;i++) {
	xtr[i] = 0. ;
	for(j=0;j<3;j++) {
		xtr[i] += r[j][i]*x[j] ;
	}
    }
}
