// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------
#include <stdio.h> 
#include <math.h>
#include "funcdefs.h"
#include "files.h"
#include "header.h"
#include "PhasegratingRecord.h"
#include "ProjCellRecord.h"
#include "storage.h"
#include "filenames.h"
#include "atoms.h"
#include "xtal.h"
#define min(x,y)	((x) < (y) ? (x) : (y))
#define fabs(x) 	((x) < 0. ? -(x) : (x))
extern char stext[] ;

// --------------------------------------------------------------------------
// Calculate the real space potential generated by the atoms 
// within the unit cell. The potential is generated by summing	
// over the scattering factors of the atoms in reciprocal space	
// and computed from the fourier coefficients of the potential	
// --------------------------------------------------------------------------
void CalculateProjectedPotential(float *zvg,long *noz,long jmax) {
    extern FILE *fp_print ;
    extern short gSliceThreeD ;
    extern long gCentroSymmetric ;
    long hmax,kmax,iout,lmax,num_elements,ndx,last_type,nslc_cell ;
    long nmax,mmax,nh,n,il,ih,ik,nk,jh,jk,jl ;
    float umax,umax1,umax2,factnl,time,*out[400],*faddr ;
    short ihout[400],ikout[400],ignore ;
    register long h,k,l,i,j ;
    float dh,dk,dl,hksq,usq,arg,rm2csga,dhsq,dhga,zf,u ;
    float sumr,sumi,zim,rcmplx[2],*rpp,*rpn ;
    float pi=3.141592654,cnst,pi2,fact,usqmax ;
    float zsumatr,zsumati,zsumlr[8],zsumli[8],cer[8],cei[8],
          zfactr,zfacti ;
		  

    /*  the numbers nmax and mmax corresponds to number of intervals  */
    /*  along the unit cell axes.                                     */
    if (gSliceThreeD)
       nslc_cell = gPhsgrtRec->nz ;
    else
       nslc_cell = 1 ;
    
    cnst = 47.874/gProjCell->vol ;
    nmax = gPhsgrtRec->nbeam ;
    mmax = gPhsgrtRec->mbeam ;
    hmax = nmax/2 ;
    kmax = mmax/2 ;
    
    umax1 = (hmax-1)*gProjCell->ast*gProjCell->siga ;
    umax2 = (kmax-1)*gProjCell->bst*gProjCell->siga ;
    umax = min(umax1,umax2) ;
    usqmax = umax*umax ;   

    n = nslc_cell ;
    pi2 = 2*pi ;
    lmax = 8 ;
    fact = pi/n ;

    ndx = 0 ;

    /* Zero all the coefficients */
    num_elements = 2*jmax*nslc_cell ;
    for (j=0,rpp=zvg;j<num_elements;j++)
      *rpp++ = 0. ;

    rm2csga = -2.*gProjCell->csga ;
	
    fprintf(fp_print,"\n\n     OLD HKL    NEW HKL    DSTAR   V(REAL)    IMAG") ;
	
	long	outputInterval = nmax/10 ;
	
    for(i=1;i<nslc_cell;i++)
	fprintf(fp_print,"   REAL     IMAG") ;
    fprintf(fp_print,"\n") ;
    erase_message() ;
    if (nslc_cell == 1) {  // Simple 2-dimensional projection
        // Loop over the half plane -hmax+1 < h < hmax; k >= 0 
        // and fill in the bottom half as we go along       
        for (h=0;h<nmax;h++)  { // Go along the h-axis  
			// Check for operators intervention =
			if (abandn()) {
				erase_message() ;
				return ;
			}
			
			nh = h ;			// nh represents the index h
			if (nh > hmax)
			nh -= nmax ;		// The index is negative       

			dh = nh*gProjCell->ast ;
			dhsq = dh*dh ;
			dhga = dh*rm2csga ;
			for (k=0;k<= kmax;k++)  {  // Loop over k >= 0  
				dk = k*gProjCell->bst ;
				hksq = dhsq+dk*dk+dk*dhga ;		// the length of g-square
				if (hksq <= usqmax)	{			// Do not bother if too large
					sumr = 0. ;
					sumi = 0. ;
					last_type = -1 ;
					for (i=0;i<gProjCell->natom;i++)  { // Sum over atoms
						j = *(gProjCell->ntype+i) ;
						if (j != last_type)  // Get scattering factor 
							AtomicFormFactor(hksq,noz,j-1,i,&zf) ;
						arg = nh*(*(gProjCell->x+i))+k*(*(gProjCell->y+i)) ;
						arg *= pi2 ;
						sumr += zf*(*(gProjCell->occ+i))*cos(arg) ;
						if (!gCentroSymmetric)
							sumi -= zf*(*(gProjCell->occ+i))*sin(arg) ;
						last_type = j ;
					}
					rpp = zvg + 2*(k*nmax+h) ;
					*rpp = cnst*sumr ;
					if (!gCentroSymmetric)    // Set the imaginary part
						*(rpp+1) = cnst*sumi ;
					// Set the bottom half of the diffraction plane
					if (k > 0) {
						if (nh <= 0) {  // Negative h's, positive k's
							rpn = zvg - 2*nh + 2*(mmax-k)*nmax ;
							*rpn = *rpp ;
							if (!gCentroSymmetric)
							*(rpn+1) = - *(rpp+1) ;
						}
						else if (nh > 0) { 
							rpn = zvg + 2*(nmax-nh) + 
								   2*(mmax-k)*nmax ;
							*rpn = *rpp ;
							if (!gCentroSymmetric)
								*(rpn+1) = - *(rpp+1) ;
						}
					}
					if (fabs(*rpp) > 0.5 && ndx < 400) {
						ihout[ndx] = nh ;
						ikout[ndx] = k  ;
						out[ndx++] = rpp ;
					}
				}
			}
			// Print out a status report/
			if (h != 0 && h%outputInterval == 0) {
				sprintf(stext," STATUS:  Potential Completed %d Percent", 10*near((float)h/outputInterval)) ;
				write_message(stext) ;
			}
		}
	}
    else {
        // 3 Dimensional calculation by summing over L        
        // Loop over the half plane -hmax+1 < h < hmax; k >= 0
        // and fill in the bottom half as we go along          
        for (h=0;h<nmax;h++)  {  // Go along the h-axis         
			if (abandn()) {
				return ;
			}
            nh = h ;           // nh represents the index h  
            if (nh > hmax)
                nh -= nmax ;   // The index is negative    

            dh = nh*gProjCell->ast ;
            dhsq = dh*dh ;
            dhga = dh*rm2csga ;
            for (k=0;k<= kmax;k++) {   // Loop over k >= 0   
                dk = k*gProjCell->bst ;
                hksq = dhsq+dk*dk+dk*dhga ;		// the length of g-square 
                if (hksq <= usqmax) {			// Do not bother if too large
					for(l=0;l<nslc_cell;l++) {
						zsumlr[l] = 0. ;
						zsumli[l] = 0. ;
					}
                    for(l= (-lmax);l<=lmax;l++) {  // Sum over L
						if((l%nslc_cell)!= 0 || l==0){
                            dl = l/gProjCell->c ;
                            usq = hksq + dl*dl ;
                            if(usq <= usqmax) {
								sumr = 0. ;
								sumi = 0. ;
								last_type = -1 ;
								for (i=0;i<gProjCell->natom;i++) {  // Sum over atoms 
									j = *(gProjCell->ntype+i) ;
									if (j != last_type)    
									   AtomicFormFactor(usq,noz,j-1,i,&zf) ;
									arg = nh*(*(gProjCell->x+i))+
											k*(*(gProjCell->y+i))+
											  l*(*(gProjCell->z+i)) ;
									arg *= pi2 ;
									sumr += zf*(*(gProjCell->occ+i))*cos(arg) ;
									sumi -= zf*(*(gProjCell->occ+i))*sin(arg) ;
									last_type = j ;
								}
                                zsumatr = sumr ;
                                zsumati = sumi ;

                                // Sum over all slices in the cell 
                                factnl = fact*l ;
                                if (l!=0) {
									zfactr = zsumatr*sin(factnl)/factnl ;
									zfacti = zsumati*sin(factnl)/factnl ;
								}
                                for(il=0;il<nslc_cell;il++) {
									if (l==0) {
										zsumlr[il] += zsumatr ;
										zsumli[il] += zsumati ;
									}
                                    else {
										arg = (2*il+1)*factnl ;
										cer[il] = cos(arg) ;
										cei[il] = sin(arg) ;
										zsumlr[il] += (zfactr*cer[il]-
												  zfacti*cei[il]) ;
										zsumli[il] += (zfactr*cei[il]+
												  zfacti*cer[il]) ;
									}
								}
							}
						}
					}
                    for (il=0;il<nslc_cell;il++) {
						rpp = zvg + 2*(k*nmax+h+nmax*mmax*il) ;
						*rpp     = cnst*zsumlr[il] ;
						*(rpp+1) = cnst*zsumli[il] ;
						// Set the bottom half of diffraction plane
						if (k > 0) {
							if (nh <= 0) {  // Negative h, positive k
								rpn = zvg - 2*nh + 2*(mmax-k)*nmax +
												  2*(nmax*mmax*il) ;
								*rpn     = *rpp ;
								*(rpn+1) =   -*(rpp+1) ;
							}
							else if (nh > 0) { 
								rpn = zvg + 2*(nmax-nh) + 
								2*(mmax-k)*nmax +
										   2*(nmax*mmax*il) ;
								*rpn     = *rpp ;
								*(rpn+1) = -*(rpp+1) ;
							}
						}
						if (fabs(*rpp) > 0.5 && ndx < 400) {
							iout = TRUE ;
							faddr = zvg + 2*(k*nmax+h) ;
						}
					}
                    if (iout && ndx < 400) {
						ihout[ndx] = nh ;
						ikout[ndx] = k  ;
						out[ndx++] = faddr ;
						iout = FALSE ;
					}
				}
			}
			// Print out a status report/
			if (h != 0 && h%outputInterval == 0) {
				sprintf(stext," STATUS:  Potential Completed %d Percent", 10*near((float)h/outputInterval)) ;
				write_message(stext) ;
			}	
		}
	}
	
	// --------------------------------------------------
	// Check for coefficients to write to the output file
	// --------------------------------------------------
	for (i=0;i<ndx;i++) {
		ih = ihout[i] ;
		ik = ikout[i] ;
		usq = ih*ih*gProjCell->ast*gProjCell->ast + 
			ik*ik*gProjCell->bst*gProjCell->bst +
			ih*ik*gProjCell->ast*gProjCell->bst*rm2csga ;
		u = sqrt(usq) ;
		Find3DHKLIndicies(ih,ik,0,&jh,&jk,&jl,gProjCell->hkltr) ;
		
		// Print out the coefficients */  
		fprintf(fp_print,"\n %3d %3d %3d   %3d %3d     %6.4f  %6.2f  %7.2f",
				jh,jk,jl,ih,ik,u,*out[i],*(out[i]+1)) ;
		for(j=1;j<nslc_cell;j++)
			fprintf(fp_print,"  %6.2f  %7.2f",*(out[i]+2*nmax*mmax*j),*(out[i]+2*nmax*mmax*j+1)) ;
	}

	// -----------------------------------------------------
	// Fourier transform to get the potential and write the
	// real part to the output file                       
	// -----------------------------------------------------
	for (i=0;i<nslc_cell;i++) {
		rpp = zvg + 2*nmax*mmax*i ;
		mmffti(rpp,nmax,mmax,&time) ;
		fprintf(fp_print,"\n\n Time to do a %d by %d complex FFT : %f",
			  nmax,mmax,time) ;
		fprintf(fp_print,"\n\n Potential:\n %f %f %f %f %f %f %f %f %f %f",
			  *(rpp),*(rpp+1),*(rpp+2),*(rpp+3),*(rpp+4),*(rpp+5),
			  *(rpp+6),*(rpp+7),*(rpp+8),*(rpp+9)) ;
		condense(rpp,jmax) ;  /* Condense to get the real part into  */
							/* successive array elements           */
		fprintf(fp_print,"\n\n Potential:\n %f %f %f %f %f %f %f %f %f %f",
			  *(rpp),*(rpp+1),*(rpp+2),*(rpp+3),*(rpp+4),*(rpp+5),
			  *(rpp+6),*(rpp+7),*(rpp+8),*(rpp+9)) ;
		write_output_file((char *)rpp,jmax) ;
	}
	erase_message() ;
}

extern struct { float ax[MAXTYPES][4],bx[MAXTYPES][4],cx[MAXTYPES] ; } xrf ;
void AtomicFormFactor(float usq,long *noz,long j,long ino,float *elfrmf) {
	
	long h ;
	register float xrff,elff ;
	register long i ;
	extern short gUseElScattFactors ;
	float ssq,cnst=0.0239340,elScattFact ;
	
	ssq = usq/4. ;
	if (gUseElScattFactors) {
	 	elScattFact = 0. ;
    		for(i=0;i<4;i++)
			elScattFact += xrf.ax[j][i]*exp(-xrf.bx[j][i]*ssq) ;
		*elfrmf = elScattFact*exp(-(gPhsgrtRec->dw[ino])*ssq) ;
	}
	else {
	    if (ssq == 0.) {
    		elff = 0. ;
    		for(i=0;i<4;i++)
		    elff =  elff + xrf.ax[j][i]*xrf.bx[j][i] ;
		elff += xrf.cx[j] ;
		*elfrmf = cnst*elff ;
	    }
	    else {
	  	xrff = 0. ;
    		for(i=0;i<4;i++)
			xrff = xrff + xrf.ax[j][i]*exp(-xrf.bx[j][i]*ssq) ;
		xrff += xrf.cx[j] ;
		*elfrmf = cnst*(noz[j]-xrff)/ssq*exp(-(gPhsgrtRec->dw[ino])*ssq) ;
	    }
	}
  }
