// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------
#include "TAtomRefinementListDialog.h"
extern	char	*chemsymb[] ;

#include "funcdefs.h"
extern Boolean	gModifyOneByOne ;
#define	checkSymbol	'x'
#define	uncheckSymbol	' '
TAtomRefinementListDialog::TAtomRefinementListDialog(TCrystal *crystal,
		RefinementList *list,FPVII callback) {

	DefineForm() ;

	mCallback	= callback ;
	mCrystal	= crystal ;
	mList		= list ;
	mAtoms		= nil ;
	mXAllOn 	= true ;
	mYAllOn 	= true ;
	mZAllOn 	= true ;
	mDWAllOn 	= true ;
	mOCCAllOn 	= true ;

	mPlacement	= FL_PLACE_MOUSE ;
	mBorder		= FL_FULLBORDER ;
	Initialize() ;

}

void	TAtomRefinementListDialog::DefineForm() {
	FL_OBJECT *obj;

	mXForm->mForm = fl_bgn_form(FL_NO_BOX, 471, 251);
	obj = fl_add_box(FL_UP_BOX,0,0,471,251,"");
	mOK = fl_add_button(FL_NORMAL_BUTTON,360,225,90,20,"OK");
	mCancel = fl_add_button(FL_NORMAL_BUTTON,250,225,90,20,"Cancel");
	mAtomList = fl_add_browser(FL_SELECT_BROWSER,15,40,440,140,"");
	obj = fl_add_text(FL_NORMAL_TEXT,5,20,70,20,"Name");
	fl_set_object_lalign(obj,FL_ALIGN_CENTER|FL_ALIGN_INSIDE);
	obj = fl_add_text(FL_NORMAL_TEXT,68,20,20,20,"X");
	fl_set_object_lalign(obj,FL_ALIGN_CENTER|FL_ALIGN_INSIDE);
	obj = fl_add_text(FL_NORMAL_TEXT,140,20,20,20,"Y");
	fl_set_object_lalign(obj,FL_ALIGN_CENTER|FL_ALIGN_INSIDE);
	obj = fl_add_text(FL_NORMAL_TEXT,215,20,20,20,"Z");
	fl_set_object_lalign(obj,FL_ALIGN_CENTER|FL_ALIGN_INSIDE);
	obj = fl_add_text(FL_NORMAL_TEXT,290,20,30,20,"DW");
	fl_set_object_lalign(obj,FL_ALIGN_CENTER|FL_ALIGN_INSIDE);
	obj = fl_add_text(FL_NORMAL_TEXT,365,20,30,20,"OCC");
	fl_set_object_lalign(obj,FL_ALIGN_CENTER|FL_ALIGN_INSIDE);
	mONx = fl_add_button(FL_NORMAL_TEXT,100,20,25,20,"on");
	mONy = fl_add_button(FL_NORMAL_TEXT,175,20,25,20,"on");
	mONz = fl_add_button(FL_NORMAL_TEXT,248,20,25,20,"on");
	mONdw = fl_add_button(FL_NORMAL_TEXT,328,20,25,20,"on");
	mONocc = fl_add_button(FL_NORMAL_TEXT,400,20,25,20,"on");
	obj = fl_add_text(FL_NORMAL_TEXT,20,180,430,20,
		"Clicking in the \"Clicking on the \"on\" buttons toggles refinement for all atoms in the list");
	fl_set_object_lalign(obj,FL_ALIGN_LEFT|FL_ALIGN_INSIDE);
	mTestEach = obj = fl_add_checkbutton(FL_PUSH_BUTTON,20,205,210,20,"Test each variable separately");	
	fl_end_form();
	mXForm->mForm->fdui = mXForm;

	fl_set_button(mTestEach,gModifyOneByOne) ;

	// Create the call back
	fl_set_object_callback(mOK,TAtomRefinementListDialog::DoClick,(long)this) ;
	fl_set_object_callback(mCancel,TAtomRefinementListDialog::DoClick,(long)this) ;
	fl_set_object_callback(mAtomList,TAtomRefinementListDialog::DoClick,(long)this) ;
	fl_set_object_callback(mONx,TAtomRefinementListDialog::DoClick,(long)this) ;
	fl_set_object_callback(mONy,TAtomRefinementListDialog::DoClick,(long)this) ;
	fl_set_object_callback(mONz,TAtomRefinementListDialog::DoClick,(long)this) ;
	fl_set_object_callback(mONdw,TAtomRefinementListDialog::DoClick,(long)this) ;
	fl_set_object_callback(mONocc,TAtomRefinementListDialog::DoClick,(long)this) ;

	fl_set_clipping(0,0,471, 251) ;

}

void	TAtomRefinementListDialog::Initialize() {
 	long 	i,j,k,ndx,ngroup ;
	TAtom	*theAtom ;
	char	text[256] ;
	char	refx,refy,refz,refdw,refocc ;

	// Create a list of all the atoms that are to be refined
	// First count the number of atoms
	mNumAtoms = 0 ;
	for(j=0; j < mList->numGroups;j++) {
		for(k=0; k < (mList->groupPtr+j)->numInGroup ; k++) {
			mNumAtoms++ ;
		}
	}
	mAtoms = new TAtom[mNumAtoms] ;
	for(i=0,j=0; j < mList->numGroups;j++) {
		for(k=0; k < (mList->groupPtr+j)->numInGroup ; k++) {
			// This is the index into gProjCell or mAtomPtr
			ndx = *((mList->groupPtr+j)->atomIndex+k) ;			
			mAtoms[i] = *mCrystal->GetAtom(ndx) ;
			if ((mAtoms[i].refine & kRefineX) == 0) mXAllOn = false ;
			if ((mAtoms[i].refine & kRefineY) == 0) mYAllOn = false ;
			if ((mAtoms[i].refine & kRefineZ) == 0) mZAllOn = false ;
			if ((mAtoms[i].refine & kRefineDW) == 0) mDWAllOn = false ;
			if ((mAtoms[i].refine & kRefineOCC) == 0) mOCCAllOn = false ;
			i++ ;
		}
	}
	
	fl_freeze_form(mXForm->mForm);
	fl_clear_browser(mAtomList);

	for(i=0;i<mNumAtoms;i++) {
		theAtom = mAtoms+i ;
		ngroup = theAtom->groupnumber ;
		//ngroup = theAtom->groupnumber > 100 ? theAtom->groupnumber-100 :
		//		theAtom->groupnumber ;
		refx = (theAtom->refine & kRefineX) != 0 ? checkSymbol : uncheckSymbol ;
		refy = (theAtom->refine & kRefineY) != 0 ? checkSymbol : uncheckSymbol ;
		refz = (theAtom->refine & kRefineZ) != 0 ? checkSymbol : uncheckSymbol ;
		refdw = (theAtom->refine & kRefineDW) != 0 ? checkSymbol : uncheckSymbol ;
		refocc = (theAtom->refine & kRefineOCC) != 0 ? checkSymbol : uncheckSymbol ;
		sprintf(text,"%d %s\t%.4f\t   %c\t%.4f\t   %c\t%.4f\t   %c\t%.4f\t   %c\t%.4f\t   %c",
			ngroup,theAtom->name,theAtom->xc,refx,theAtom->yc,refy,theAtom->zc,refz,
			theAtom->dw,refdw,theAtom->occ,refocc) ;
		fl_add_browser_line(mAtomList,text) ;
	}
	fl_unfreeze_form(mXForm->mForm);
}


void	TAtomRefinementListDialog::GetRefinementListProperties() {

	long 	i,j,k,ndx ;
	TAtom	*theAtom ;

	gModifyOneByOne = fl_get_button(mTestEach) ;

	for(i=0,j=0; j < mList->numGroups;j++) {
		for(k=0; k < (mList->groupPtr+j)->numInGroup ; k++) {
			// This is the index into gProjCell or mAtomPtr
			ndx = *((mList->groupPtr+j)->atomIndex+k) ;						
			theAtom = mCrystal->GetAtom(ndx) ;
			*theAtom = mAtoms[i] ;
			i++ ;
			// Also set the refinement list if we do a reentry
			*((mList->groupPtr+j)->refine + k) = theAtom->refine ;
		}
	}
}

void 	TAtomRefinementListDialog::DoClick(FL_OBJECT *obj,long arg) {
	long	i,lineSelected,ngroup ;
	FL_Coord	x,y ;
	TAtomRefinementListDialog	*theForm = (TAtomRefinementListDialog *)arg ;
	char	text[256],refx,refy,refz,refdw,refocc ;
	TAtom	*theAtom ;

	if (obj == theForm->mOK) {
		(*theForm->mCallback)(arg,1) ;
		return ;
	}
	if (obj == theForm->mCancel) {
		(*theForm->mCallback)(arg,0) ;
		return ;
	}
	if (obj == theForm->mONx) {
		// Turn all X refinements either on or off
		fl_freeze_form(theForm->mXForm->mForm);
		for (i = 0; i < theForm->mNumAtoms;i++) {
			theAtom = theForm->mAtoms+i ;
			ngroup = theAtom->groupnumber ;
			//ngroup = theAtom->groupnumber > 100 ? theAtom->groupnumber-100 :
			//		theAtom->groupnumber ;
			if (theForm->mXAllOn) {
				theAtom->refine &= ~kRefineX ;
			}
			else {
				theAtom->refine |= kRefineX ;
			}
			refx = (theAtom->refine & kRefineX) != 0 ? checkSymbol : uncheckSymbol ;
			refy = (theAtom->refine & kRefineY) != 0 ? checkSymbol : uncheckSymbol ;
			refz = (theAtom->refine & kRefineZ) != 0 ? checkSymbol : uncheckSymbol ;
			refdw = (theAtom->refine & kRefineDW) != 0 ? checkSymbol : uncheckSymbol ;
			refocc = (theAtom->refine & kRefineOCC) != 0 ? checkSymbol : uncheckSymbol ;
			sprintf(text,"%d %s\t%.4f\t   %c\t%.4f\t   %c\t%.4f\t   %c\t%.4f\t   %c\t%.4f\t   %c",
			ngroup,theAtom->name,theAtom->xc,refx,theAtom->yc,refy,theAtom->zc,refz,
			theAtom->dw,refdw,theAtom->occ,refocc) ;
			fl_replace_browser_line(theForm->mAtomList,i+1,text) ;
		}
		fl_unfreeze_form(theForm->mXForm->mForm);
		theForm->mXAllOn = 1 - theForm->mXAllOn ;
		return ;
	}
	if (obj == theForm->mONy) {
		// Turn all Y refinements either on or off
		fl_freeze_form(theForm->mXForm->mForm);
		for (i = 0; i < theForm->mNumAtoms;i++) {
			theAtom = theForm->mAtoms+i ;
			ngroup = theAtom->groupnumber ;
			//ngroup = theAtom->groupnumber > 100 ? theAtom->groupnumber-100 :
			//		theAtom->groupnumber ;
			if (theForm->mYAllOn) {
				theAtom->refine &= ~kRefineY ;
			}
			else {
				theAtom->refine |= kRefineY ;
			}
			refx = (theAtom->refine & kRefineX) != 0 ? checkSymbol : uncheckSymbol ;
			refy = (theAtom->refine & kRefineY) != 0 ? checkSymbol : uncheckSymbol ;
			refz = (theAtom->refine & kRefineZ) != 0 ? checkSymbol : uncheckSymbol ;
			refdw = (theAtom->refine & kRefineDW) != 0 ? checkSymbol : uncheckSymbol ;
			refocc = (theAtom->refine & kRefineOCC) != 0 ? checkSymbol : uncheckSymbol ;
			sprintf(text,"%d %s\t%.4f\t   %c\t%.4f\t   %c\t%.4f\t   %c\t%.4f\t   %c\t%.4f\t   %c",
			ngroup,theAtom->name,theAtom->xc,refx,theAtom->yc,refy,theAtom->zc,refz,
			theAtom->dw,refdw,theAtom->occ,refocc) ;
			fl_replace_browser_line(theForm->mAtomList,i+1,text) ;
		}
		fl_unfreeze_form(theForm->mXForm->mForm);
		theForm->mYAllOn = 1 - theForm->mYAllOn ;
		return ;
	}
	if (obj == theForm->mONz) {
		// Turn all Z refinements either on or off
		fl_freeze_form(theForm->mXForm->mForm);
		for (i = 0; i < theForm->mNumAtoms;i++) {
			theAtom = theForm->mAtoms+i ;
			ngroup = theAtom->groupnumber ;
			//ngroup = theAtom->groupnumber > 100 ? theAtom->groupnumber-100 :
			//		theAtom->groupnumber ;
			if (theForm->mZAllOn) {
				theAtom->refine &= ~kRefineZ ;
			}
			else {
				theAtom->refine |= kRefineZ ;
			}
			refx = (theAtom->refine & kRefineX) != 0 ? checkSymbol : uncheckSymbol ;
			refy = (theAtom->refine & kRefineY) != 0 ? checkSymbol : uncheckSymbol ;
			refz = (theAtom->refine & kRefineZ) != 0 ? checkSymbol : uncheckSymbol ;
			refdw = (theAtom->refine & kRefineDW) != 0 ? checkSymbol : uncheckSymbol ;
			refocc = (theAtom->refine & kRefineOCC) != 0 ? checkSymbol : uncheckSymbol ;
			sprintf(text,"%d %s\t%.4f\t   %c\t%.4f\t   %c\t%.4f\t   %c\t%.4f\t   %c\t%.4f\t   %c",
			ngroup,theAtom->name,theAtom->xc,refx,theAtom->yc,refy,theAtom->zc,refz,
			theAtom->dw,refdw,theAtom->occ,refocc) ;
			fl_replace_browser_line(theForm->mAtomList,i+1,text) ;
		}
		fl_unfreeze_form(theForm->mXForm->mForm);
		theForm->mZAllOn = 1 - theForm->mZAllOn ;
		return ;
	}
	if (obj == theForm->mONdw) {
		// Turn all DW refinements either on or off
		fl_freeze_form(theForm->mXForm->mForm);
		for (i = 0; i < theForm->mNumAtoms;i++) {
			theAtom = theForm->mAtoms+i ;
			ngroup = theAtom->groupnumber ;
			//ngroup = theAtom->groupnumber > 100 ? theAtom->groupnumber-100 :
			//		theAtom->groupnumber ;
			if (theForm->mDWAllOn) {
				theAtom->refine &= ~kRefineDW ;
			}
			else {
				theAtom->refine |= kRefineDW ;
			}
			refx = (theAtom->refine & kRefineX) != 0 ? checkSymbol : uncheckSymbol ;
			refy = (theAtom->refine & kRefineY) != 0 ? checkSymbol : uncheckSymbol ;
			refz = (theAtom->refine & kRefineZ) != 0 ? checkSymbol : uncheckSymbol ;
			refdw = (theAtom->refine & kRefineDW) != 0 ? checkSymbol : uncheckSymbol ;
			refocc = (theAtom->refine & kRefineOCC) != 0 ? checkSymbol : uncheckSymbol ;
			sprintf(text,"%d %s\t%.4f\t   %c\t%.4f\t   %c\t%.4f\t   %c\t%.4f\t   %c\t%.4f\t   %c",
			ngroup,theAtom->name,theAtom->xc,refx,theAtom->yc,refy,theAtom->zc,refz,
			theAtom->dw,refdw,theAtom->occ,refocc) ;
			fl_replace_browser_line(theForm->mAtomList,i+1,text) ;
		}
		fl_unfreeze_form(theForm->mXForm->mForm);
		theForm->mDWAllOn = 1 - theForm->mDWAllOn ;
		return ;
	}
	if (obj == theForm->mONocc) {
		// Turn all OCC refinements either on or off
		fl_freeze_form(theForm->mXForm->mForm);
		for (i = 0; i < theForm->mNumAtoms;i++) {
			theAtom = theForm->mAtoms+i ;
			ngroup = theAtom->groupnumber ;
			//ngroup = theAtom->groupnumber > 100 ? theAtom->groupnumber-100 :
			//		theAtom->groupnumber ;
			if (theForm->mOCCAllOn) {
				theAtom->refine &= ~kRefineOCC ;
			}
			else {
				theAtom->refine |= kRefineOCC ;
			}
			refx = (theAtom->refine & kRefineX) != 0 ? checkSymbol : uncheckSymbol ;
			refy = (theAtom->refine & kRefineY) != 0 ? checkSymbol : uncheckSymbol ;
			refz = (theAtom->refine & kRefineZ) != 0 ? checkSymbol : uncheckSymbol ;
			refdw = (theAtom->refine & kRefineDW) != 0 ? checkSymbol : uncheckSymbol ;
			refocc = (theAtom->refine & kRefineOCC) != 0 ? checkSymbol : uncheckSymbol ;
			sprintf(text,"%d %s\t%.4f\t   %c\t%.4f\t   %c\t%.4f\t   %c\t%.4f\t   %c\t%.4f\t   %c",
			ngroup,theAtom->name,theAtom->xc,refx,theAtom->yc,refy,theAtom->zc,refz,
			theAtom->dw,refdw,theAtom->occ,refocc) ;
			fl_replace_browser_line(theForm->mAtomList,i+1,text) ;
		}
		fl_unfreeze_form(theForm->mXForm->mForm);
		theForm->mOCCAllOn = 1 - theForm->mOCCAllOn ;
		return ;
	}
	if (obj == theForm->mAtomList) {
		lineSelected = fl_get_browser(theForm->mAtomList) ;
		if (lineSelected > 0) {
			// We have to add logic that sets the refinement tags 
			// for all the atoms within a group
			// We got a click in the list box.
			// Find which item got clicked and turn refinement on 
			// or off for this item
			theAtom = theForm->mAtoms+lineSelected-1 ;
			long	groupNumber = theAtom->groupnumber ;
			ngroup = groupNumber ;
			//ngroup = groupNumber > 100 ? groupNumber-100 :
			//	groupNumber ;
			unsigned int keys_buttons = Button1Mask ;
			fl_get_win_mouse(fl_winget(),&x,&y,&keys_buttons) ;
			if (x >= theForm->mONx->x && x < (theForm->mONx->x+theForm->mONx->w)) {
				if ((theAtom->refine & kRefineX) != 0) {
					theAtom->refine &= ~kRefineX ;
				}
				else {
					theAtom->refine |= kRefineX ;
				}
			}
			else if (x >= theForm->mONy->x && x < (theForm->mONy->x+theForm->mONy->w)) {
				if ((theAtom->refine & kRefineY) != 0) {
					theAtom->refine &= ~kRefineY ;
				}
				else {
						theAtom->refine |= kRefineY ;
				}
			}
			else if (x >= theForm->mONz->x && x < (theForm->mONz->x+theForm->mONz->w)) {
				if ((theAtom->refine & kRefineZ) != 0) {
					theAtom->refine &= ~kRefineZ ;
				}
				else {
					theAtom->refine |= kRefineZ ;
				}
			}
			else if (x >= theForm->mONdw->x && x < (theForm->mONdw->x+theForm->mONdw->w)) {	
				if ((theAtom->refine & kRefineDW) != 0) {
					theAtom->refine &= ~kRefineDW ;
				}
				else {
					theAtom->refine |= kRefineDW ;
				}
			}
			else if (x >= theForm->mONocc->x && x < (theForm->mONocc->x+theForm->mONocc->w)) {
				if ((theAtom->refine & kRefineOCC) != 0) {
					theAtom->refine &= ~kRefineOCC ;
				}
				else {
					theAtom->refine |= kRefineOCC ;
				}
			}
			refx = (theAtom->refine & kRefineX) != 0 ? checkSymbol : uncheckSymbol ;
			refy = (theAtom->refine & kRefineY) != 0 ? checkSymbol : uncheckSymbol ;
			refz = (theAtom->refine & kRefineZ) != 0 ? checkSymbol : uncheckSymbol ;
			refdw = (theAtom->refine & kRefineDW) != 0 ? checkSymbol : uncheckSymbol ;
			refocc = (theAtom->refine & kRefineOCC) != 0 ? checkSymbol : uncheckSymbol ;
			sprintf(text,"%d %s\t%.4f\t   %c\t%.4f\t   %c\t%.4f\t   %c\t%.4f\t   %c\t%.4f\t   %c",
					ngroup,theAtom->name,theAtom->xc,refx,theAtom->yc,refy,theAtom->zc,refz,
					theAtom->dw,refdw,theAtom->occ,refocc) ;
			fl_replace_browser_line(theForm->mAtomList,lineSelected,text) ;
			// Replace any other line as well

			fl_freeze_form(theForm->mXForm->mForm);
			TAtom *nextAtom ;
			for (i = 0; i < theForm->mNumAtoms;i++) {
				nextAtom = theForm->mAtoms+i ;
				ngroup = nextAtom->groupnumber ; 
				//ngroup = nextAtom->groupnumber > 100 ? 
				//	nextAtom->groupnumber-100 :
				//	nextAtom->groupnumber ;
				if (nextAtom != theAtom && 
						nextAtom->groupnumber == theAtom->groupnumber) {
					nextAtom->refine = theAtom->refine ;
					refx = (nextAtom->refine & kRefineX) != 0 ? 
						checkSymbol : uncheckSymbol ;
					refy = (nextAtom->refine & kRefineY) != 0 ? 
						checkSymbol : uncheckSymbol ;
					refz = (nextAtom->refine & kRefineZ) != 0 ? 
						checkSymbol : uncheckSymbol ;
					refdw = (nextAtom->refine & kRefineDW) != 0 ? 
						checkSymbol : uncheckSymbol ;
					refocc = (nextAtom->refine & kRefineOCC) != 0 ? 
						checkSymbol : uncheckSymbol ;
sprintf(text,"%d %s\t%.4f\t   %c\t%.4f\t   %c\t%.4f\t   %c\t%.4f\t   %c\t%.4f\t   %c",
						ngroup,nextAtom->name,nextAtom->xc,refx,nextAtom->yc,refy,
						nextAtom->zc,refz,nextAtom->dw,refdw,
						nextAtom->occ,refocc) ;
					fl_replace_browser_line(theForm->mAtomList,i+1,text) ;
				}
			}
			fl_unfreeze_form(theForm->mXForm->mForm);
		} 
		return ;
	}
}

TAtomRefinementListDialog::~TAtomRefinementListDialog() {
	if (mAtoms != nil) delete[] mAtoms ;
}
