#ifndef __TButton_H__
#define __TButton_H__
#include "types.h"
#ifdef VMS
#include <decw$include/keysym.h>
#else
#include <X11/keysym.h>
#endif

typedef char	Str64[64] ;
typedef char	Str256[256] ;

typedef enum {kLabelInside,kLabelOutsideLeft,kLabelOutsideTop,kLabelOutsideRight,kLabelOutsideBottom}
LabelPosition ;

#include "funcdefs.h"

class	TIcon {
	public:
		TIcon(short	width,short height,unsigned char *data=nil) ;
		~TIcon() ;
		void	Draw(short x,short y) ;
		Pixmap	GetPixmap() { return mPixmap ; }  ;
	private:
		Pixmap  mPixmap ;
		short	mWidth ;
		short	mHeight ;
} ;

class	TPage ;
class	TObject {
	public:
		TObject(Rect *rect,char *label = nil) ;
		TObject(short x1,short y1, short x2, short y2,char *label = nil) ;
		virtual	~TObject() ;
		void	SetParentWindow(TPage *page) { mParentWindow = page ; } ;
		TPage	*GetParentWindow() { return mParentWindow ; } ;
		void	Enable() { mIsActive = true ; Draw(); } ; 
		void	Disable() { mIsActive = false ; Draw(); } ; 
		void	SetVisible(Boolean visible) { mIsVisible = visible ; Draw() ; } ;
		Boolean	IsVisible() { return mIsVisible ; } ;
		Boolean	GetState() { return mState ; } ;
		Boolean	IsEnabled() { return mIsActive ; } ;
		virtual	void SetState(Boolean state) { mState = state; Draw() ; } ;
		Rect	&GetRect() { return mRect ; } ;
		Boolean InObject(Point *pt) { if (mIsActive) return PtInRect(pt,&mRect);else return false ; } ;
		virtual	void	DoEvent(XEvent *) ;
		virtual	Boolean DoMouseDown(Point *) {return true ;} ;
		virtual KeySym 	DoKeyDown(XEvent *) { return 1 ;} ;
		virtual void 	Draw() {} ;
		void	SetItemText(char *text) ;
		void	GetItemText(char *text) ;
		void	SetLabelPosition(LabelPosition pos) {mLabelPosition = pos ;} ;
		void 	SetJustification(JustificatonType type) {mJustification=type; Draw();};
	protected:
		void	CreateLabel(char *text) ;
		virtual	void	DrawLabel() {} ;
		Rect	mRect ;
		Boolean	mIsActive ;
		Boolean	mState ;
		Boolean	mIsVisible ;
		Str256	*mLabel ;
		TPage	*mParentWindow ;
		short	mNumLabelLines ;
		LabelPosition	mLabelPosition ;
		JustificatonType	mJustification ;
} ;

class	TUserRect : public TObject {
	public:
		TUserRect(Rect *rect) : TObject(rect) { mIsActive = false ;} ;
		TUserRect(short x1,short y1,short x2,short y2):TObject(x1,y1,x2,y2) {mIsActive=false;};
		void Draw() {} ;
} ;

class	TFrame : public TObject {
	public:
		TFrame(Rect *rect) : TObject(rect) { 	mIsActive = false ;} ;
		TFrame(short x1,short y1, short x2, short y2) : TObject(x1,y1,x2,y2){mIsActive = false ;} ;
		void Draw() ;
} ;

class	TGroupBox : public TObject {
	public:
		TGroupBox(Rect *rect,char *label=nil) ;
		TGroupBox(short x1,short y1, short x2, short y2, char *label=nil) ;
		virtual	~TGroupBox() ;
		void Draw() ;
	protected:
		void	DrawLabel() ;
} ;

class TButton ;
typedef void (*ButtonActionRoutine)(TButton *) ;

class TButton : public TObject {
	public:
		TButton(Rect *rect,char *title=nil,ButtonActionRoutine actionRoutine = nil) ;
		TButton(short x1,short y1, short x2, short y2,
			char *title=nil,ButtonActionRoutine actionRoutine = nil) ;
		virtual	~TButton() ;
		virtual void Draw() ;
		virtual	Boolean DoMouseDown(Point *) ;
		Boolean	GetControlValue() { return mState ; } ;
		void	SetControlValue(Boolean state) {mState=state;Draw();} ;
	protected:
		ButtonActionRoutine		mActionRoutine ;
		Boolean	mIsDefault ;
		void	DrawLabel() ;
} ;

class TPushButton : public TButton { 
	public:
		TPushButton(Rect *rect,TIcon *icon=nil,char *title=nil,Boolean state = OFF) ;
		TPushButton(short x1,short y1, short x2, short y2,TIcon *icon=nil,
			char *title=nil,Boolean state = OFF) ;
		Boolean DoMouseDown(Point *) ;
		void Draw() ;
		virtual	~TPushButton() ;
	protected:
		TIcon	*mIcon ;
} ;

class TCheckBox : public TButton {
	public:
		TCheckBox(Rect *rect,char *title=nil,Boolean state = OFF) ;
		TCheckBox(short x1,short y1, short x2, short y2,
			char *title=nil,Boolean state = OFF) ;
		Boolean DoMouseDown(Point *) ;
		void Draw() ;
		virtual	~TCheckBox() ;
	protected:
		void DrawPressed(Boolean upDwn) ;
		void DrawLabel() ;
} ;

class TRadioButton : public TButton {
	public:
		TRadioButton(Rect *rect,char *title=nil,Boolean state = OFF) ;
		TRadioButton(short x1,short y1, short x2, short y2,
			char *title=nil,Boolean state = OFF) ;
		virtual	~TRadioButton() ;
		void Draw() ;
		Boolean DoMouseDown(Point *) ;
	protected:
		void DrawPressed(Boolean upDwn) ;
		void	DrawLabel() ;
} ;

class TMenuButton : public TButton {
	public:
		TMenuButton(Rect *rect,char *text=nil,char *label=nil,LabelPosition pos=kLabelOutsideLeft) ;
		TMenuButton(short x1,short y1, short x2, short y2,
			char *text=nil,char *label=nil,LabelPosition pos=kLabelOutsideLeft) ;
		virtual	~TMenuButton() ;
		void Draw() ;
		void	SetMenuText(char *text) { strcpy(mMenuText,text) ; } ; 
		virtual	Boolean DoMouseDown(Point *) ;
	protected:
		void	DrawLabel() ;
		void	DrawMenuText() ;
		char	mMenuText[256] ;
		
} ;

class TStaticText : public TObject {
	public:
		TStaticText(Rect *rect,char *title=nil) ;
		TStaticText(short x1,short y1, short x2, short y2,char *title=nil) ;
		virtual	~TStaticText() ;
		void SetItemText(char *text) ;
		void GetItemText(char *text) ;
		void Draw() ;
	protected:
		void	DrawLabel() ;
} ;
typedef enum {kNumericInput,kGenericInput} TextInputType ;

class TEditText : public TObject {
	public:
		TEditText(Rect *rect,char *title,char *text=nil) ;
		TEditText(short x1,short y1, short x2, short y2,char *title,char *text=nil) ;
		virtual	~TEditText() ;
		void 	SetEditText(char *text) ;
		void 	GetEditText(char *text) ;
		Boolean DoMouseDown(Point *) ;
		KeySym	DoKeyDown(XEvent *event) ;
		void 	Draw() ;
		void	SelectAll() ;
		void	SetFocus(Boolean focus) ;
	protected:
		char	*mText ;
		Point	mInsertionPt ;
		Boolean	mIsFocus ;
		short	mFirstSelChar,mNumSelChars ;
		short	FindCharacterToLeft(short x) ;
		short	FindCharacterToRight(short x) ;
		TextInputType	mInputType ;
} ;

class TObjectEntry {	
	public:
		TObjectEntry() ;
		~TObjectEntry() ;
		friend class TObjectList ;
	private:	
		TObject			*mObject ;
		TObjectEntry	*mNextObjectEntry ;
}  ;

class TObjectList {
	public:
		TObjectList() ;
		~TObjectList() ;
		void 	AddObject(TObject *theObject) ;
		void	RemoveObject(TObject *theObject) ;
		short	GetNumberOfObjects(void) { return mNumObjects ; } ;
		TObject	*GetObject(long number) ;
		TObject *GetObject(Point *pt) ;
		void	DrawObjects() ;
	private:
		short	mNumObjects ;
		TObjectEntry	*mObjectEntry ;
	
}  ;


class TEditTextEntry {	
	public:
		TEditTextEntry() { mNextEditTextEntry = nil ; mObject = nil ; } ;
		~TEditTextEntry() { mNextEditTextEntry = nil ; } ;
		friend class TEditTextList ;
	private:	
		TObject			*mObject ;
		TEditTextEntry	*mNextEditTextEntry ;
}  ;

class TEditTextList {
	public:
		TEditTextList() { mNumObjects = 0 ; mEditTextEntry = nil ; } ;
		~TEditTextList() { if (mEditTextEntry != nil) delete mEditTextEntry ; } ;
		void 	AddObject(TObject *theObject) ;
		void	RemoveObject(TObject *theObject) ;
		TEditText	*GetNextField(TEditText *current) ;
		short	GetNumberOfObjects(void) { return mNumObjects ; } ;
		TObject	*GetObject(long number) ;
		TObject *GetObject(Point *pt) ;
		long		GetObjectIndex(TObject *obj) ;
	private:
		short	mNumObjects ;
		TEditTextEntry	*mEditTextEntry ;
	
}  ;
#endif // __TButton__
 
