// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "graphics.h"
#include "TCrystal.h"
#include "globals.h"

#include "funcdefs.h"
#include "atoms.h"
#include "symops.h"
extern float one(float x);
extern void encode_symop(char *string,struct GENERATOR *sym_op) ;
extern 	long gSecondSetting,gHexagonalIndices,gAlternateSetting,gCentroSymmetric ;
extern	uLong	gWhitePixel,gBlackPixel,gDarkerGray,gGrayPixel,gGreenPixel,gRedPixel,gBluePixel ;

extern GENERATOR gen[20],temp_sym,temp_gen ;
extern GENERATOR inv_gen ;
extern GENERATOR identity_matr ;
extern GENERATOR p_matr,q_matr ;

static char symb[] = {'x','y','z'} ;
extern char micname[] ;

#define	EOS	'\0'
#define	NEWLINE	'\n'

#include "numbers.h"
extern char 	*chemsymb[] ;
// extern functions
extern	Boolean NumberIsInteger(float fval) ;
extern short	gNumActiveElements ;
extern short 	gActiveElementNumber[20] ;	// Maximum number of active elements
void DeleteRefinementList(RefinementList *list) ;

extern	RefinementList gRefinementList ;
extern TMatrix	*gSymOps  ;	// symetry operators for this "unit cell" 
RMatrix gTMatrix ; 

typedef struct {
	short numBasisAtoms ;
	TAtom *atom ;
} TBasis ;
extern TBasis gBasis ;


void	WriteStatus(char *text) ;

#define BlackColor 193
#define WhiteColor 0

void PenPat(Pattern *pattern=nil) ;

RGBColor RGBwhite = {65535,65535,65535} ;
RGBColor RGBblack = {0,0,0} ;
extern	RGBColor RGBred ;
extern	RGBColor RGBgreen ;
extern	RGBColor RGBblue ;

float	CheckZeroOne(float x) ;
inline  long SIGN(float x) { return (x < 0. ? -1   : 1) ; }
inline  float min(float x,float y) { return (x < y ? x : y) ; }
inline  long min(long x,long y) { return (x < y ? x : y) ; }
inline  float max(float x,float y) { return (x > y ? x : y) ; }
inline  long max(long x,long y) { return (x > y ? x : y) ; }

// ------------------------------------------------------
// Create an instance of the class TCrystal.  It creates
// a default crystal, that can be displayed and serves a
// basis for the create crystal routine.  All the entries
// are meant to be changed either through reading in a 
// crystal from file or by the user modifying the default
// Crystal
//

TCrystal::TCrystal(TWindow *window,Rect &theDisplayRect, ProjCellRecord * ProjCell)
{
		
	mDocWindow 		= window ;
	mDisplayRect	= theDisplayRect ;
	mSelection 		= nil ;
	mHasAtomsSelected = false ;
	Initialize(ProjCell) ;
}

void 
TCrystal::Initialize(ProjCellRecord * cell)
{	
	long i ;
	
	mAtomsInUCell 	= cell->natom ;
	mZone.a[0]		= 0. ;
	mZone.a[1]		= 0. ;
	mZone.a[2]		= 1. ;
	mZone.a[3]		= 0. ;
	mA 				= cell->a ;
	mB 				= cell->b ;
	mC				= cell->c ;
	mAlpha 			= cell->al ;
	mBeta			= cell->be ;
	mGamma			= cell->ga ;
	mLightX 		= 1. ;
	mLightY 		= 1. ;
	mLightZ			= 3. ;
	mAtomPtr		= nil ;
	mXViewCells		= 1 ;
	mXViewCells		= 1 ;
	mXViewCells		= 1 ;
	mShiftX = 0 ;
	mShiftY = 0 ;
	
	mAtomScaleFactor = 0.5 ;	
	
	mScreenAtom		= nil ;
	mDrawOutline = true ;
	mDrawOrigin = true ;
	mPlanePt		= kNoPlanePt ;
	
	mCurrentGroup = 1 ;
	mUseSymmetries = false ;
	mWAtom = kFirstAtom ;
	mAnglePt = kNoAnglePt ;
	mNumUnitCellAtoms 	= cell->natom ;
	mAtomsInUCell		= mNumUnitCellAtoms ;
	mNumBasisAtoms		= 0 ;
	mSymOps = nil ;
	mBasisPtr = nil ;
	mAllocatedAtoms = 0 ;
	
	// Allocate space for the atoms in the whole structure
	mAtomPtr = new TAtom[cell->natom] ;	// This is the maximum we can have, so we don't need
	Assert(mAtomPtr) ;					// to reallocate this storage
	mAllocatedAtoms = cell->natom ;
	for(i=0;i<cell->natom;i++) {
		(mAtomPtr+i)->xc = *(cell->x+i) ;
		(mAtomPtr+i)->yc = *(cell->y+i) ;
		(mAtomPtr+i)->zc = *(cell->z+i) ;
		(mAtomPtr+i)->dw = *(cell->dw+i) ;
		(mAtomPtr+i)->occ = *(cell->occ+i) ;
		(mAtomPtr+i)->type = *(cell->ntype+i) ;
		(mAtomPtr+i)->no = *(cell->atno+i) ;
		(mAtomPtr+i)->name = chemsymb[(mAtomPtr+i)->no-1] ;
		(mAtomPtr+i)->rad = covrad[(mAtomPtr+i)->no-1]  ;
		(mAtomPtr+i)->color 	= RGBgreen ;
		(mAtomPtr+i)->refine 	= kRefineX + kRefineY ;
	}

	// Allocate space for a possible basis for the whole structure
	mBasisPtr = new TAtom[cell->natom] ;
	Assert(mBasisPtr) ;

	
	DegreeToRad(&mAlpha) ;
	DegreeToRad(&mBeta) ;
	DegreeToRad(&mGamma) ;
     
 	// Allocate the atoms required for the view
	CreateView(1,1,1) ;	
    
	SetTMatrix() ;
	SetSMatrix() ;
    
	mWidth = mDisplayRect.right-mDisplayRect.left ;
	mHeight = mDisplayRect.bottom-mDisplayRect.top ;
	    
	mMagnification = 1. ;
	
	TVector tZone = mZone ;
			
	SetTransformationMatrix(tZone) ;
	
	CreateUnitCellViewAtoms() ;
	mRotMatrix = 1.0 ;
	
}

void TCrystal::SwitchSymmetryView(Boolean useSymmetry) {
	extern ProjCellRecord * gProjCell ;
	long	i,numAtomsToAllocate ;
	TAtom debugAtom ;
	
	mUseSymmetries = useSymmetry ;
	if (useSymmetry) {
		if (mBasisPtr != nil) {
			// We don't have to delete the mAtomPtr since we can have it point to the
			// original number of atoms which cannot be exceeded
			// We could in principle have many more atoms if we move an atom away from a special position
			numAtomsToAllocate = mNumBasisAtoms*mNumSymOps ;
			if (numAtomsToAllocate > mAllocatedAtoms) {
				if (mAtomPtr != nil) delete[] mAtomPtr ;
				mAtomPtr = new TAtom[numAtomsToAllocate] ;
				Assert(mAtomPtr) ;
				mAllocatedAtoms = numAtomsToAllocate ;
			}
			mAtomsInUCell		= mNumBasisAtoms ;
			for(i=0;i<mAtomsInUCell;i++) {
				*(mAtomPtr+i) = *(mBasisPtr+i) ;
				(mAtomPtr+i)->name = chemsymb[(mAtomPtr+i)->no-1] ;
				(mAtomPtr+i)->rad 	= 	covrad[(mAtomPtr+i)->no-1] ;
				(mAtomPtr+i)->color 	= RGBgreen ;
				(mAtomPtr+i)->refine 	= kRefineX + kRefineY ;
			}
		}
		if (gBasis.atom != nil) {
			delete[] gBasis.atom ;
			gBasis.atom = nil ;
			gBasis.numBasisAtoms = 0 ;
		}
		gBasis.atom = new TAtom[mAtomsInUCell] ;
		if (gBasis.atom != nil) {
			gBasis.numBasisAtoms = mAtomsInUCell ;
			for(i=0;i<mAtomsInUCell;i++) {
				*(gBasis.atom+i) = *(mBasisPtr+i) ;
			}
		}
		
		if (OptionKeyDown()) {
			FILE *fp ;
			char tt[256] ;
			long k,l ;
			GENERATOR tSym ;
			fp = fopen("Symmops.debug","w") ;
			if (fp != nil) {
				for (i=0;i<mNumBasisAtoms;i++) {
					fprintf(fp,"%d %s %f %f %f %f %f\n",(mBasisPtr+i)->type,(mBasisPtr+i)->name,
						(mBasisPtr+i)->xc,(mBasisPtr+i)->yc,(mBasisPtr+i)->zc,(mBasisPtr+i)->dw,(mBasisPtr+i)->occ) ;
				}
			  	for (i=0;i<mNumSymOps;i++) {
	  				for(k=0;k<4;k++) {
						for(l=0;l<4;l++) {
							tSym.matr[k][l] = mSymOps[i].m[k][l] ;
						}
					}
			  		encode_symop(tt,&tSym) ;
			  		fprintf(fp,"%s\n",tt) ;
				}
				fclose(fp) ;
			}
		}
	}
	else {
		mAtomsInUCell		= mNumUnitCellAtoms ;		
		if (gBasis.atom != nil) {
			delete[] gBasis.atom ;
			gBasis.atom = nil ;
			gBasis.numBasisAtoms = 0 ;
		}

		for(i=0;i<mAtomsInUCell;i++) {
			(mAtomPtr+i)->xc = *(gProjCell->x+i) ;
			(mAtomPtr+i)->yc = *(gProjCell->y+i) ;
			(mAtomPtr+i)->zc = *(gProjCell->z+i) ;
			(mAtomPtr+i)->dw = *(gProjCell->dw+i) ;
			(mAtomPtr+i)->occ = *(gProjCell->occ+i) ;
			(mAtomPtr+i)->no = *(gProjCell->atno+i) ;
			(mAtomPtr+i)->type = *(gProjCell->ntype+i) ;
			(mAtomPtr+i)->name = chemsymb[(mAtomPtr+i)->no-1] ;
			(mAtomPtr+i)->rad = covrad[(mAtomPtr+i)->no-1] ;
			(mAtomPtr+i)->color 	= RGBgreen ;
			(mAtomPtr+i)->refine 	= kRefineX + kRefineY ;
		}
	}
 	// Allocate the atoms required for the view
	mZone.a[0]		= 0. ;
	mZone.a[1]		= 0. ;
	mZone.a[2]		= 1. ;
	mZone.a[3]		= 0. ;
	CreateView(1,1,1) ;	
	mMagnification = 1. ;
	
	TVector tZone = mZone ;
			
	SetTransformationMatrix(tZone) ;
	
	CreateUnitCellViewAtoms() ;
	mRotMatrix = 1.0 ;
}

Boolean
TCrystal::NewSymop(GENERATOR *p_symop,long num_symop) {
	register long i,j ;
	long new_gen ;
	float *p_sym,*p_old_sym,test ;

	new_gen = true ;
	for (i=0;i< num_symop;i++) {
		p_sym = (float *)(&p_symop->matr[0][0]) ;
		p_old_sym = (float *)&mSymOps[i] ;
		for(j=0; j < 16 ;j++) {
			test = *p_sym - *p_old_sym ;
			test = (test >= 0. ? test : -test) ;
			if (test > 0.01) 
				break ;
			p_sym++ ;
			p_old_sym++ ;
		}
		if (j==16) {
			new_gen = false ;
			break ;
		}
	}
	return(new_gen) ;
}

void 
TCrystal::LoadSpacegroup(char *spacegroup,short spcgrp_number) {
	FILE	*fp ;
	char	SpcgrpFile[255] ;
	long	i ;
	extern	char *NcemssDir ;
	char	text[256] ;

	strcpy(SpcgrpFile,NcemssDir) ;
#ifdef VMS
	strcat(SpcgrpFile,"spcgrp.dat") ;
#else
	strcat(SpcgrpFile,"/spcgrp.dat") ;
#endif
	if ((fp=fopen(SpcgrpFile,"r")) == NULL) {
		eraser(0,0,1000,32) ;
		sprintf(text,"Unable to open the file %s , Using Group # 1 !",SpcgrpFile) ;
		PutMessage(text) ;
		strcpy(spacegroup,"1 0 1 1 0 0") ;
		generate_spcgrp(spacegroup) ;
		return ;
	}
	for (i=0;i<spcgrp_number;i++)
		fgets(spacegroup,132,fp) ;
	fclose(fp) ;
	i = 0 ;
	while (spacegroup[i++] != NEWLINE && i < 132) ;
	spacegroup[--i] = EOS ;
}

void	
TCrystal::SetSymmops(short spacegroupNumber,Boolean secondSetting,RMatrixPtr transfMatrix) {
	register long i,j,m,n ;
	long  num_gen ;
	long  inversion,num_sym ;
	long ione ;
	char spacegroup[132] ;
	float xc,yc,zc,x,y,z ;
	float xnew,ynew,znew ;
	TMatrix TInverse ;
	short	numGeneratorIteration ;
    
	LoadSpacegroup(spacegroup,spacegroupNumber) ;    

	if (mSymOps != nil) 
		delete[] mSymOps ;
	mSymOps = nil ;
		
	mSymOps = new TMatrix[MAXSYMOPS] ;
	if (mSymOps == nil)
		return ;
   
	// Set up the transformation matrix T based on "transfMatrix"
	mT = 1. ;
	for(j=0;j<3;j++) {
		for(i=0;i<3;i++) {
			mT.m[i][j] = transfMatrix->m[i][j] ;
		}
	}
	InvertMatrix(mT,TInverse) ;
	TransposeMatrix(TInverse,mU) ;
	
	mSymOps[0] = 1. ;
	mNumSymOps = 1 ;
	decode_spcgrp(spacegroup,&num_gen,&inversion) ;

	// Transform the generators and then see how many symmetry elements
	// we end up with
	for (i=0;i<num_gen;i++) {
    		TransformGenerator(&gen[i]) ;
	}

	for (i=0;i<num_gen;i++) {
        	num_sym = mNumSymOps ;
        	temp_gen = gen[i] ;
      		numGeneratorIteration = 0 ;
        	do {
			for (j=0;j< num_sym;j++) {
				apply_gen(&temp_gen,(GENERATOR *)&mSymOps[j],&temp_sym) ;
				if (NewSymop(&temp_sym,mNumSymOps)) {
					for(n=0;n<4;n++)
						for(m=0;m<4;m++)
							mSymOps[mNumSymOps].m[n][m] = temp_sym.matr[n][m] ;
					mNumSymOps++ ;
				}
			}
			apply_gen(&gen[i],&temp_gen,&temp_sym) ;
			if (!(ione=ident(&temp_sym)))
                		temp_gen = temp_sym ;
			numGeneratorIteration++ ;
		} while (!ione && numGeneratorIteration < 8 && mNumSymOps <= MAXSYMOPS) ;
	}
	num_sym = mNumSymOps ;
	if (inversion) {
		for (j=0;j< num_sym;j++) {
			apply_gen(&inv_gen,(GENERATOR *)&mSymOps[j],&temp_sym) ;
         	for(n=0;n<4;n++)
         		for(m=0;m<4;m++)
         			mSymOps[mNumSymOps].m[n][m] = temp_sym.matr[n][m] ;
			mNumSymOps++ ;
		}
	}

	
	// If we have a trivial unit cell, there isn't anything really to do	
	if (mNumSymOps == 1) {	// The basis and the atoms are the same
		mNumBasisAtoms = mAtomsInUCell ;
		for(i=0;i<mNumBasisAtoms;i++) {
			*(mBasisPtr+i) = *(mAtomPtr+i) ;
		}
		return ;
	}

	// Allocate space for the global symmetries		
	if (gSymOps != nil) {
		delete[] gSymOps ;
	}
	if (gSymOps == nil) {
		gSymOps = new TMatrix[mNumSymOps] ;
		if (gSymOps == nil)
			return ;
	}
	
	// Find the new basis based on the atoms in the unit cell and
	// the symmetry operators that we have just found	
	mNumBasisAtoms = 0 ;
	long iz = -1 ;
	float dw = -1. ;
	long	istart ;
	Boolean	isEquivalent ;
	Boolean xtest,ytest,ztest ;
	float 	xdiff,ydiff,zdiff ;
	for (j=0;j<mAtomsInUCell;j++) {
		xc = (mAtomPtr+j)->xc ;
	  	yc = (mAtomPtr+j)->yc ;
	  	zc = (mAtomPtr+j)->zc ;
	  	
	  	// SInce the atoms in the unit cell are sorted on atomic number, we know
	  	// that if we have a new atomic number, we have a new basis atom
	  	if (iz != (mAtomPtr+j)->no || dw != (mAtomPtr+j)->dw)  {
	  		dw = (mAtomPtr+j)->dw  ;
	  		iz = (mAtomPtr+j)->no  ;
			*(mBasisPtr+mNumBasisAtoms) = *(mAtomPtr+j) ;
		    istart = mNumBasisAtoms ;	// Always signal a start of a new atom
		    mNumBasisAtoms++ ;	  		
	  	}
	  	else {
	  		// Can this atom be generated by the atoms already in the basis
	  		isEquivalent = false ;
	  		for(long k = istart ; k < mNumBasisAtoms;k++) {
				x = (mBasisPtr+k)->xc ;
			  	y = (mBasisPtr+k)->yc ;
			  	z = (mBasisPtr+k)->zc ;
			  	for (i=0;i<mNumSymOps;i++) {
			    	xnew = mSymOps[i].m[0][0]*x + mSymOps[i].m[0][1]*y +
			           		mSymOps[i].m[0][2]*z + mSymOps[i].m[0][3] ;
			    	ynew = mSymOps[i].m[1][0]*x + mSymOps[i].m[1][1]*y +
			           		mSymOps[i].m[1][2]*z + mSymOps[i].m[1][3] ;
			    	znew = mSymOps[i].m[2][0]*x + mSymOps[i].m[2][1]*y +
			           		mSymOps[i].m[2][2]*z + mSymOps[i].m[2][3] ;
			    	xnew = one(xnew) ;
			    	ynew = one(ynew) ;
			    	znew = one(znew) ;
			      	xdiff = (xc-xnew)*gProjCell->a ;
			      	ydiff = (yc-ynew)*gProjCell->b ;
			      	zdiff = (zc-znew)*gProjCell->c ;      
			      	xtest = false ; 
			      	ytest = false ; 
			      	ztest = false ;
			      	if (fabs(xdiff) < 0.01) xtest = true ;
			      	if (fabs(ydiff) < 0.01) ytest = true ;
			      	if (fabs(zdiff) < 0.01) ztest = true ;
			      	if (xtest && ytest && ztest) {
			         	isEquivalent = true ;
			         	break ;
					}
			  	}
			}
		  	if (!isEquivalent) {
				*(mBasisPtr+mNumBasisAtoms) = *(mAtomPtr+j) ;
	    		mNumBasisAtoms++ ;
		  	}
	  	}
	}	
}

void 
TCrystal::TransformGenerator(GENERATOR *symop) {
	TMatrix	UInverse,S,Sprime,Identity,tempMatrix ;
	TVector	ST,STprime ;
	float	value ;
	long	i,j ;
	GENERATOR	tSymopMatr ;
	
	InvertMatrix(mU,UInverse) ;
	
	Identity = 1. ;
	
	S = 1. ;
	Sprime = 1. ;
	for(i=0;i<3;i++) {
		for(j=0;j<3;j++) {
			S.m[i][j] = symop->matr[i][j] ;
		}
		ST.a[i] =  symop->matr[i][3] ;
	}
	ST.a[3] = 0. ;
	Sprime = S*UInverse ;
	Sprime = mU*Sprime ;					// S' = USU-1
	tempMatrix = (S-Identity) ;
	tempMatrix = mU*tempMatrix ;
	STprime = tempMatrix*mO + mU*ST;	// ST' = U(S-1.0)*O + UST
	
	tSymopMatr.matr[3][0] = 0. ;
	tSymopMatr.matr[3][1] = 0. ;
	tSymopMatr.matr[3][2] = 0. ;
	tSymopMatr.matr[3][3] = 1. ;
	for(i=0;i<3;i++) {
		for(j=0;j<3;j++) {
			tSymopMatr.matr[i][j] = Sprime.m[i][j] ;
		}
		// Reduce the translation to lie between 0 and 1
		value = one(STprime.a[i]) ;
		tSymopMatr.matr[i][3] = value ;
	}
	// Check to see if we are close to 0 or 1
	for(i=0;i<4;i++) {
		for(j=0;j<4;j++) {
			value = tSymopMatr.matr[i][j] ;
			if (fabs(value) < 0.00001) value = 0. ;
			tSymopMatr.matr[i][j] = value ;
		}
	}	
	*symop = tSymopMatr ;
}
void TCrystal::SetSymmops(short numSymops,GENERATOR *symop,RMatrixPtr transfMatrix) {
	TMatrix	T,TInverse,U,UInverse,S,Sprime,Identity,tempMatrix ;
	TVector	O,ST,STprime ;
	float	value ;
	long	i,j,n ;
	float xc,yc,zc,x,y,z ;
	float xnew,ynew,znew ;

	if (mSymOps != nil) 
		delete[] mSymOps ;
	mSymOps = nil ;

	mNumSymOps = numSymops ;
		
	if (mNumSymOps == 1) {	// The basisptr and the atomptr are the same
		mNumBasisAtoms = mAtomsInUCell ;
		for(i=0;i<mNumBasisAtoms;i++) {
			*(mBasisPtr+i) = *(mAtomPtr+i) ;
		}
		return ;
	}

	mSymOps = new TMatrix[mNumSymOps] ;
	if (mSymOps == nil)
		return ;
		
	if (gSymOps == nil) {
		gSymOps = new TMatrix[mNumSymOps] ;
		if (gSymOps == nil)
			return ;
	}

	// Set up the transformation matrix T based on "transfMatrix"
	T = 1. ;
	for(j=0;j<3;j++) {
		for(i=0;i<3;i++) {
			T.m[i][j] = transfMatrix->m[i][j] ;
		}
	}
	// Calculate the appropriate matrices
	for(i=0;i<3;i++) {
		O.a[i] = T.m[i][3] ;
		T.m[i][3] = 0. ;
	}
	O.a[3] = 0. ;
	InvertMatrix(T,TInverse) ;
	TransposeMatrix(TInverse,U) ;
	InvertMatrix(U,UInverse) ;
	
	Identity = 1. ;
	for(n=0;n < numSymops; n++) {
		S = 1. ;
		Sprime = 1. ;
		for(i=0;i<3;i++) {
			for(j=0;j<3;j++) {
				S.m[i][j] = symop[n].matr[i][j] ;
			}
			ST.a[i] =  symop[n].matr[i][3] ;
		}
		ST.a[3] = 0. ;
		Sprime = S*UInverse ;
		Sprime = U*Sprime ;
		tempMatrix = (S-Identity) ;
		tempMatrix = U*tempMatrix ;
		STprime = tempMatrix*O + U*ST;
		mSymOps[n] = 1. ;
		for(i=0;i<3;i++) {
			for(j=0;j<3;j++) {
				mSymOps[n].m[i][j] = Sprime.m[i][j] ;
			}
			// Reduce the translation to lie between 0 and 1
			value = one(STprime.a[i]) ;
			mSymOps[n].m[i][3] = value ;
		}
		if (gSymOps != nil)
			*(gSymOps+n) = *(mSymOps+n) ;
		// EncodeSymop(mSymopTxt[n],&mSymop[n]) ;		
	}
	// Find the new basis based on the atoms in the unit cell and
	// the symmetry operators that we have just found	
	mNumBasisAtoms = 0 ;
	long iz = -1 ;
	float dw = -1. ;
	long	istart ;
	Boolean	isEquivalent ;
	Boolean xtest,ytest,ztest ;
	float 	xdiff,ydiff,zdiff ;
	for (j=0;j<mAtomsInUCell;j++) {
		xc = (mAtomPtr+j)->xc ;
	  	yc = (mAtomPtr+j)->yc ;
	  	zc = (mAtomPtr+j)->zc ;
	  	
	  	// SInce the atoms in the unit cell are sorted on atomic number, we know
	  	// that if we have a new atomic number, we have a new basis atom
	  	if (iz != (mAtomPtr+j)->no || dw != (mAtomPtr+j)->dw)  {
	  		dw = (mAtomPtr+j)->dw  ;
	  		iz = (mAtomPtr+j)->no  ;
			*(mBasisPtr+mNumBasisAtoms) = *(mAtomPtr+j) ;
		    istart = mNumBasisAtoms ;	// Always signal a start of a new atom
		    mNumBasisAtoms++ ;	  		
	  	}
	  	else {
	  		// Can this atom be generated by the atoms already in the basis
	  		isEquivalent = false ;
	  		for(long k = istart ; k < mNumBasisAtoms;k++) {
				x = (mBasisPtr+k)->xc ;
			  	y = (mBasisPtr+k)->yc ;
			  	z = (mBasisPtr+k)->zc ;
			  	for (i=0;i<mNumSymOps;i++) {
			    	xnew = mSymOps[i].m[0][0]*x + mSymOps[i].m[0][1]*y +
			           		mSymOps[i].m[0][2]*z + mSymOps[i].m[0][3] ;
			    	ynew = mSymOps[i].m[1][0]*x + mSymOps[i].m[1][1]*y +
			           		mSymOps[i].m[1][2]*z + mSymOps[i].m[1][3] ;
			    	znew = mSymOps[i].m[2][0]*x + mSymOps[i].m[2][1]*y +
			           		mSymOps[i].m[2][2]*z + mSymOps[i].m[2][3] ;
			    	xnew = one(xnew) ;
			    	ynew = one(ynew) ;
			    	znew = one(znew) ;
			      	xdiff = (xc-xnew)*gProjCell->a ;
			      	ydiff = (yc-ynew)*gProjCell->b ;
			      	zdiff = (zc-znew)*gProjCell->c ;      
			      	xtest = false ; 
			      	ytest = false ; 
			      	ztest = false ;
			      	if (fabs(xdiff) < 0.01) xtest = true ;
			      	if (fabs(ydiff) < 0.01) ytest = true ;
			      	if (fabs(zdiff) < 0.01) ztest = true ;
			      	if (xtest && ytest && ztest) {
			         	isEquivalent = true ;
			         	break ;
					}
			  	}
			}
		  	if (!isEquivalent) {
	    		*(mBasisPtr+mNumBasisAtoms) = *(mAtomPtr+j) ;
	    		mNumBasisAtoms++ ;
		  	}
	  	}
	}					
}
Boolean TCrystal::Equivalent(TAtom *atomList,float x,float y,float z,long istart,long ndx) {
  /* ----------------------------------------------------------	*/
  /* Routine to check if the atom whose coordinates were just	*/
  /* passed is a new one. The search is only through the list 	*/
  /* of atoms of the same type. Thus if too many atoms were put	*/
  /* into the basis, the application of the symmetry operators	*/
  /* will produce too many atoms in the total structure. Beware	*/
  /* Returns (false) if this is a 'new' atom, (true) otherwise  */
  /* ----------------------------------------------------------	*/

	long 	k ;
	Boolean xtest,ytest,ztest ;
	float 	xdiff,ydiff,zdiff ;

	for (k=istart;k<ndx;k++) {
      		xdiff = (x-(atomList+k)->xc)*mA ;
      		ydiff = (y-(atomList+k)->yc)*mB ;
      		zdiff = (z-(atomList+k)->zc)*mC ;      
      		xtest = false ; 
      		ytest = false ; 
      		ztest = false ;
      		if (fabs(xdiff) < 0.1) xtest = true ;
      		if (fabs(ydiff) < 0.1) ytest = true ;
      		if (fabs(zdiff) < 0.1) ztest = true ;
      		if (xtest && ytest && ztest) 
         		return(true) ;              /* The atom already exists */
	}
	return(false) ;                   	/* New atom                */
} 

void TCrystal::RedisplayShiftedAtoms() {
	long i ;
	if (!mUseSymmetries) {
		for(i=0;i<mAtomsInUCell;i++) {
			(mAtomPtr+i)->xc = *(gProjCell->x+i) ;
			(mAtomPtr+i)->yc = *(gProjCell->y+i) ;
			(mAtomPtr+i)->zc = *(gProjCell->z+i) ;
			(mAtomPtr+i)->dw = *(gProjCell->dw+i) ;
			(mAtomPtr+i)->occ = *(gProjCell->occ+i) ;
			(mAtomPtr+i)->no = *(gProjCell->atno+i) ;
			(mAtomPtr+i)->name = chemsymb[(mAtomPtr+i)->no-1] ;
			(mAtomPtr+i)->rad = covrad[(mAtomPtr+i)->no-1] ;
			(mAtomPtr+i)->color 	= RGBgreen ;
		}
		CreateUnitCellViewAtoms() ;
		DrawCrystal() ;
	}
	else {
		long savedNumAtoms = mAtomsInUCell ;
		mAtomsInUCell = gProjCell->natom ;
		for(i=0;i<mAtomsInUCell;i++) {
			(mAtomPtr+i)->xc = *(gProjCell->x+i) ;
			(mAtomPtr+i)->yc = *(gProjCell->y+i) ;
			(mAtomPtr+i)->zc = *(gProjCell->z+i) ;
			(mAtomPtr+i)->dw = *(gProjCell->dw+i) ;
			(mAtomPtr+i)->occ = *(gProjCell->occ+i) ;
			(mAtomPtr+i)->no = *(gProjCell->atno+i) ;
			(mAtomPtr+i)->name = chemsymb[(mAtomPtr+i)->no-1] ;
			(mAtomPtr+i)->rad = covrad[(mAtomPtr+i)->no-1] ;
			(mAtomPtr+i)->color 	= RGBgreen ;
		}
		CreateUnitCellViewAtoms() ;
		DrawCrystal() ;
		mAtomsInUCell = savedNumAtoms ;
		// Restore the atoms to the original link to the basis
		for(i=0;i<mAtomsInUCell;i++) {
			*(mAtomPtr+i) = *(gBasis.atom+i) ;
		}	
	}
}	
	
void TCrystal::Assert(void *ptr) {
	if (ptr == nil) {
		::PutMessage("Assertion Failure in TCrystal. We ran out of memory") ;
		exit(0) ;
	}
}

TCrystal::~TCrystal() {

	
	if (mAtomPtr != nil) 
		delete[] mAtomPtr ;
	if (mBasisPtr != nil && mBasisPtr != mAtomPtr)
		delete[] mBasisPtr ;
	if (mScreenAtom != nil) 
		delete[] mScreenAtom ;
	if (mSymOps != nil) 
		delete[] mSymOps ;	
	// delete mRecipLattice ;
}

void 
TCrystal::ApplySymmops() {

  	/* ------------------------------------------------------------	*/
  	/* Applies the symmetry operators on the atoms in the basis 	*/
  	/* ------------------------------------------------------------	*/
    
	register long i,j,ndx,icount ;
	float xc,yc,zc ;
	float xnew,ynew,znew ;

	ndx = 0 ;
	for (j=0;j<mNumBasisAtoms;j++) {
		icount = 0 ;
		xc = (mBasisPtr+j)->xc ;
		yc = (mBasisPtr+j)->yc ; 
		zc = (mBasisPtr+j)->zc ;
		for (i=0;i<mNumSymOps;i++) {
			xnew = (mSymOps+i)->m[0][0]*xc + (mSymOps+i)->m[0][1]*yc +
				   (mSymOps+i)->m[0][2]*zc + (mSymOps+i)->m[0][3] ;
			ynew = (mSymOps+i)->m[1][0]*xc + (mSymOps+i)->m[1][1]*yc +
				   (mSymOps+i)->m[1][2]*zc + (mSymOps+i)->m[1][3] ;
			znew = (mSymOps+i)->m[2][0]*xc + (mSymOps+i)->m[2][1]*yc +
				   (mSymOps+i)->m[2][2]*zc + (mSymOps+i)->m[2][3] ;
			xnew = one(xnew) ;
			ynew = one(ynew) ;
			znew = one(znew) ;
			if (!Equivalent(mAtomPtr,xnew,ynew,znew,icount,ndx)) {
				*(mAtomPtr+ndx) = *(mBasisPtr+j) ;
				ndx++ ;
				icount++ ;
			}
		}
	}
	mNumUnitCellAtoms = ndx ;
}

void	
TCrystal::WriteInfo(char *text) {
	WriteStatus(text) ;	
}

void	
TCrystal::SizeAtoms(Boolean reduceSize) {
	if (reduceSize) {
		mAtomScaleFactor /= 1.1 ;
	}
	else {
		mAtomScaleFactor *= 1.1 ;
	}
	UpdateViewAtoms() ;
	DrawCrystal() ;
}

// The vector izone is to be interpreted as a plane normal, thus it is a vector
// in reciprocal space (hkl)
void 
TCrystal::CalculateRealZoneFromReciprocal(TVector vec) {
	TVector 	recZone ;
	long		 i ;
	float 	lengthRecipVector,dSpacing,lengthRealVector,
			minVal,factor,absVal ;		
	
			
	recZone = vec ;
	
	RecipToRlVector(recZone,vec) ;
	
	lengthRecipVector = LengthOfRecipVector(recZone) ;
	dSpacing = 1./lengthRecipVector ;
	lengthRealVector = LengthOfRealVector(vec) ;
	
	factor = dSpacing/lengthRealVector ;
	
	vec *= factor ;
	
	// We probably still want to reduce so that we have no more than 1 
	// non-integer value ;
	minVal = 10000. ;
	for(i=0;i<3;i++) {
		absVal = fabs(vec.a[i]) ;
		minVal = (absVal < minVal && absVal != 0. ? absVal : minVal) ;
	}
	vec /= minVal ;
	
	// Now one of the numbers are an integer.  
	// Multiply with an integer number until all are or until one more is
	// integer and we have exceeded a maximum on the integer value
	long	maxInt = 10 ;
	Boolean allIntegers ;
	for(long j=1;j<=maxInt;j++) {
		vec.a[i] *= j ;
		allIntegers = false ;
		for (i=0;i<3;i++) {
			allIntegers = (NumberIsInteger(vec.a[i]) && allIntegers) ;
		}
		if (allIntegers) 
			break ;
	}
}

void 
TCrystal::RecipToRlVector(TVector &rcpVec,TVector &rlVec) {
	TMatrix sinv,sinvt ;

	/* We have (a,b,c) = S (ast,bst,cst), such that (ast,bst,cst) = S(-1) (a,b,c)	*/
	/* For coordinates we then have (xa,xb,xc) = [S(-1)] (T) (xast,xbst,xcst)		*/
	InvertMatrix(mSmatrix,sinv) ;
	TransposeMatrix(sinv,sinvt) ;
	rlVec = sinvt*rcpVec ;
	
}

void 
TCrystal::RlToRecipVector(TVector &rlVec,TVector &rcpVec) {
    TMatrix st ;

    /* We have (a,b,c) = S (ast,bst,cst)								*/
    /* For coordinates we then have (xast,xbst,xcst) = S (T) (xa,xb,xc)	*/
	TransposeMatrix(mSmatrix,st) ;
	rcpVec = st*rlVec ;
}

float 
TCrystal::LengthOfRecipVector(TVector &vec) {

  	float 	gsq,ast,bst,cst,csgast,csalst,csbest,h,k,l ;
  	float 	astsq,bstsq,cstsq,Vol ;
	float	sial,sibe,siga,csal,csbe,csga ;
	
  	sial 	= sin(mAlpha) ;
	sibe 	= sin(mBeta) ;
	siga 	= sin(mGamma) ;
	csal 	= cos(mAlpha) ;
	csbe 	= cos(mBeta) ;
	csga 	= cos(mGamma) ;
	Vol		= mA*mB*mC*sqrt(1.-csal*csal-
						csbe*csbe-csga*csga+2.*csal*csbe*csga) ;
	ast 	= mB*mC*sial/Vol ;
	bst 	= mC*mA*sibe/Vol ;
	cst 	= mA*mB*siga/Vol ;
	csalst	= (csbe*csga-csal)/(sibe*siga) ;
	csbest	= (csga*csal-csbe)/(siga*sial) ;
	csgast	= (csal*csbe-csga)/(sial*sibe) ;

	astsq = ast*ast ;
	bstsq = bst*bst ;
	cstsq = cst*cst ;
    
	h = vec.a[0] ;
	k = vec.a[1] ;
	l = vec.a[2] ;
    
	gsq =  h*h*astsq + k*k*bstsq + l*l*cstsq +
				2*h*k*ast*bst*csgast+
				2*h*l*ast*cst*csbest+
				2*k*l*bst*cst*csalst  ;
	return(sqrt(gsq)) ;
}

float 
TCrystal::LengthOfRealVector(TVector &vec) {

  	float 	lsq,u,v,w ;
  	float 	a,b,c,asq,bsq,csq ;
	float	csal,csbe,csga ;
	
  	csal 	= cos(mAlpha) ;
	csbe 	= cos(mBeta) ;
	csga 	= cos(mGamma) ;
	
	a		= mA ;
	b		= mB ;
	c		= mC ;
	asq 	= a*a ;
	bsq		= b*b ;
	csq		= c*c ;
	
	u 		= vec.a[0] ;
	v		= vec.a[1] ;
	w		= vec.a[2] ;
	lsq =  u*u*asq + v*v*bsq + w*w*csq +
				2*u*v*a*b*csga+
				2*u*w*a*c*csbe+
				2*v*w*b*c*csal  ;
	return(sqrt(lsq)) ;
}


void
TCrystal::SetScale() {
	float 	scale,ax,bx,cx,ay,by,cy,
			xlength,ylength,scalex,scaley ;
	// Draw the outline.  The projection of the three axis on the screen
	ax = mXyzToScreenMatrix.m[0][0]; ay = mXyzToScreenMatrix.m[1][0]; 
	bx = mXyzToScreenMatrix.m[0][1]; by = mXyzToScreenMatrix.m[1][1]; 
	cx = mXyzToScreenMatrix.m[0][2]; cy = mXyzToScreenMatrix.m[1][2] ;

	xlength = fabs(mXViewCells*ax) + fabs(mYViewCells*bx) + fabs(mZViewCells*cx) ;	
	/* Span of view in ngstrm horizontally	*/
	
	ylength = fabs(mXViewCells*ay) + fabs(mYViewCells*by) + fabs(mZViewCells*cy) ;	
	/* Span of unit cell in ngstrm vertically		*/

	scale = max(xlength,ylength) ;
	scalex = 0.8*mWidth/scale ;				/* Num pixels per Angstrom 	*/
	scaley = 0.8*mHeight/scale ;			/* Num pixels per Angstrom 	*/
	scale  = min(scalex,scaley) ;			/* Final scale				*/
	
	scale *= mMagnification ;
	mXSection.AngstrPrPixel	= 1./scale ;
	 	
	SetCenter() ;
	
}
void 
TCrystal::ResetView() {
	mShiftX = 0 ;
	mShiftY = 0 ;
}

void	
TCrystal::PanView(Point &loc) {
	Point lastPoint,currentPoint ;
	long dx,dy ;
	
	lastPoint = loc ;
	currentPoint = loc ;
	while(Button()) {
		GetMouse(&currentPoint) ;
		dx = currentPoint.x - lastPoint.x ;
		dy = -(currentPoint.y - lastPoint.y) ;
		if (dx != 0 || dy != 0) {
			mShiftX += dx ;
			mShiftY += dy ;
			
			mXyzToScreenMatrix.m[0][3] += dx ;
			mXyzToScreenMatrix.m[1][3] += dy ;
			
		    mXSection.Origin.x		+= dx ;
		    mXSection.Origin.y		+= dy ;
		    UpdateViewAtoms() ;
			DrawCrystal() ;
			if (mSelection != nil) {
				mSelection->Pan(dx,-dy) ;
				mSelection->Update(&mDisplayRect) ;
			}			
			lastPoint = currentPoint ;
		}
	}
}

void 
TCrystal::SetCenter(void)
{
	short 	iorigox,iorigoy,iorigoz ;
	float	scale,xpr,ypr,zpr ;
	
	scale = 1./mXSection.AngstrPrPixel ;
	
	/* We want the coordinates 1/2*nA, 1/2*nB, 1/2*nC to fall in the center of the screen */
	xpr = 0.5*(mXyzToScreenMatrix.m[0][0]*mXViewCells + mXyzToScreenMatrix.m[0][1]*mYViewCells + 
			mXyzToScreenMatrix.m[0][2]*mZViewCells) ;
	ypr = 0.5*(mXyzToScreenMatrix.m[1][0]*mXViewCells + mXyzToScreenMatrix.m[1][1]*mYViewCells +
			 mXyzToScreenMatrix.m[1][2]*mZViewCells) ;
	zpr = 0.5*(mXyzToScreenMatrix.m[2][0]*mXViewCells + mXyzToScreenMatrix.m[2][1]*mYViewCells + 
			mXyzToScreenMatrix.m[2][2]*mZViewCells) ;
	iorigox = (short)(mWidth/2 - xpr*scale) + mShiftX ;
	iorigoy = (short)(mHeight/2 - ypr*scale) + mShiftY ;	
	iorigoz = (short)(-zpr*scale) ;
	mXyzToScreenMatrix.m[0][3] = iorigox ;
	mXyzToScreenMatrix.m[1][3] = iorigoy ;
	mXyzToScreenMatrix.m[2][3] = iorigoz ;
	
	mXSection.Origin.x		= iorigox ;
	mXSection.Origin.y		= iorigoy ;
}

void
TCrystal::SetTransformationMatrix(TVector &view) {

	TVector 	screenUVW ;
	TMatrix 	temp ;
	short 		i,j ;
	float		ax,ay,pi,phipr ;
	float 		uvwmag,theta,phi,xymag ;
	
	
	for(i=0;i<4;i++) 				// Reset rotation matrix
		for(j=0;j<4;j++)
			mRotMatrix.m[i][j] = 0. ;
	for(i=0;i<4;i++) mRotMatrix.m[i][i] = 1. ;
	
	temp = mTmatrix*mRotMatrix ;
	TransposeMatrix(temp,mXyzToScreenMatrix) ;
	
	
	// Find the appropriate angles to rotate the crystal
	screenUVW = mXyzToScreenMatrix*view ;	// The UVW given in screen ccordinates
		
	uvwmag = screenUVW.length() ;
	
	xymag = screenUVW.a[0]*screenUVW.a[0]+screenUVW.a[1]*screenUVW.a[1] ;
	xymag = xymag > 0. ? sqrt(xymag) : 0. ;
	
	pi = 2*asin(1.0) ;
	if (screenUVW.a[0] == 0. && screenUVW.a[1] == 0.) {
		phi = 0. ;
		theta = 0. ;
	}
	else if (screenUVW.a[0] == 0.) {
		phi = (screenUVW.a[1] > 0. ?  pi/2. : -pi/2. ) ;
		theta = acos(screenUVW.a[2]/uvwmag) ;
	}
	else if (screenUVW.a[1] == 0.) {
		phi = (screenUVW.a[0] > 0. ?  0. : pi ) ;
		theta = acos(screenUVW.a[2]/uvwmag) ;
	}
	else {
		theta = acos(screenUVW.a[2]/uvwmag) ;
		phi = acos(screenUVW.a[0]/xymag) ;
		phi = (screenUVW.a[1] > 0. ?  phi : -phi ) ;
	}
	
	SetDisRotMatrix(zRotAxis,-phi) ;
	SetDisRotMatrix(yRotAxis,-theta) ;
	
	
	
	// Find the angle that a makes with the x-axis
	ax = mXyzToScreenMatrix.m[0][0]; ay = mXyzToScreenMatrix.m[1][0]; 
	xymag = ax*ax+ay*ay ;
	xymag = xymag > 0. ? sqrt(xymag) : 0. ;
	if (ax == 0. && ay == 0.) {
		phipr = 0. ;
	}
	else if (ay == 0.) {
		phipr = (ax > 0. ?  0. : pi ) ;
	}
	else if (ax == 0.) {
		phipr = (ay > 0. ?  pi/2. : -pi/2. ) ;
	}
	else {
		phipr = acos(ax/xymag) ;
	}
	
	SetDisRotMatrix(zRotAxis,-phipr) ;
	temp = mTmatrix*mRotMatrix ;	// MultiplyMatrix(gTMatrix,gDisplayRotMatrix,temp) ;
	
	TransposeMatrix(temp,mXyzToScreenMatrix) ;

	SetScale() ;
	
    
}



void 
TCrystal::CreateUnitCellViewAtoms()
{
	short 		i,n,m,l ;
	float		TZ0,fn,fm,fl ;
	float 		x,y,z,scale,rad ;
	RGBColor	color ;
	
	
	scale	=	1/mXSection.AngstrPrPixel ;	
	TZ0		= 	1.0 ;
	
	SetUCellFaces() ;
	
	mSnumAtoms = 0 ;
	fn = (float)mXViewCells ;
	fm = (float)mYViewCells ;
	fl = (float)mZViewCells ;
	for(l=0;l<mZViewCells;l++) {
		for(m=0;m<mYViewCells;m++) {
			for(n=0;n<mXViewCells;n++) {
				/* Transform all the coordinates of the atoms in the cell to screen coordinates */
				for(i=0;i< mAtomsInUCell;i++) {
					x = (mAtomPtr+i)->xc + n ;
					y = (mAtomPtr+i)->yc + m ;
					z = (mAtomPtr+i)->zc + l ;
					rad = (mAtomPtr+i)->rad;	// bjf 2/96  for custom radii
					color	= (mAtomPtr+i)->color ;
					
					if (fabs(x) < 0.00001) x = 0. ;
					if (fabs(y) < 0.00001) y = 0. ;
					if (fabs(z) < 0.00001) z = 0. ;
					
					AddViewAtom(x,y,z,TZ0,scale,i,rad,&color) ;
					
					if (x == 0. || y == 0. || z == 0.) {
						if (x==0. && y != 0. && z != 0.) {
							AddViewAtom(fn,y,z,TZ0,scale,i,rad,&color) ;
						}
						else if (x==0. && y==0. && z!=0.) {
							AddViewAtom(fn,y,z,TZ0,scale,i,rad,&color) ;
							AddViewAtom(x,fm,z,TZ0,scale,i,rad,&color) ;
							AddViewAtom(fn,fm,z,TZ0,scale,i,rad,&color) ;
						}
						else if (x==0. && y==0. && z==0.) {
							AddViewAtom(fn,y,z,TZ0,scale,i,rad,&color) ;
							AddViewAtom(fn,fm,z,TZ0,scale,i,rad,&color) ;
							AddViewAtom(x,fm,z,TZ0,scale,i,rad,&color) ;
							AddViewAtom(x,y,fl,TZ0,scale,i,rad,&color) ;
							AddViewAtom(fn,y,fl,TZ0,scale,i,rad,&color) ;
							AddViewAtom(x,fm,fl,TZ0,scale,i,rad,&color) ;
							AddViewAtom(fn,fm,fl,TZ0,scale,i,rad,&color) ;
						}
						else if (y==0. && x!=0. &&z!=0.) {
							AddViewAtom(x,fm,z,TZ0,scale,i,rad,&color) ;
						}
						else if (y==0. && x!=0. && z==0.) {
							AddViewAtom(x,fm,z,TZ0,scale,i,rad,&color) ;
							AddViewAtom(x,fm,fl,TZ0,scale,i,rad,&color) ;
							AddViewAtom(x,y,fl,TZ0,scale,i,rad,&color) ;
						}
						else if (z==0. && x!=0. && y!=0.) {
							AddViewAtom(x,y,fl,TZ0,scale,i,rad,&color) ;
						}
						else if (z==0. && x==0. && y!=0.) {
							AddViewAtom(fn,y,z,TZ0,scale,i,rad,&color) ;
							AddViewAtom(x,y,fl,TZ0,scale,i,rad,&color) ;
							AddViewAtom(fn,y,fl,TZ0,scale,i,rad,&color) ;
						}
					}
				}
			}
		}
	}
		
	// Before we can draw the atoms, we must sort the atoms according to the z-position
	SortScreenAtoms() ;
}

void 
TCrystal::DrawUnitCell() 
{
	if (mDrawOutline) {
		DrawInvisibleUCellFaces() ;
	}
	// Draw the Crystal 
	Draw2DScreenBalls() ;
	
	if (mDrawOutline) {
		DrawVisibleUCellFaces() ;
	}
	
	if(mDrawOrigin)
		DrawOrigin(); // bjf 9/96
	
}

void TCrystal::DrawCrystal() {
	//GrafPtr tPort ;

	//GetPort(&tPort) ;
	SetPort((GrafPtr)mDocWindow) ;			
	Draw() ;	
	//SetPort(tPort) ;
}

void 
TCrystal::Draw() 
{
	
	InsetRect(&mDisplayRect,1,1) ;
	TStOffscreen offScreen(mDocWindow,&mDisplayRect) ;
	if (offScreen.GetPixmap() != nil) {
		InsetRect(&mDisplayRect,-1,-1) ;
		DrawUnitCell() ;
	}
	
}

void
TCrystal::DrawOrigin() // test function to see if it is easy to id the origin
{

	Point screenPt0, screenPtA, screenPtB, screenPtC, A, B, C;
	TVector V;
	
	V.a[0] = V.a[1] = V.a[2] = 0.0;
	screenPt0 = VectorToPoint(V);
	
	V.a[0] = 0.5;
	V.a[1] = V.a[2] = 0.0;
	screenPtA = VectorToPoint(V);
	V.a[0] = 0.6;
	A = VectorToPoint(V);
	
	V.a[1] = 0.5;
	V.a[0] = V.a[2] = 0.0;
	screenPtB = VectorToPoint(V);
	V.a[1] = 0.6;
	B = VectorToPoint(V);

	V.a[2] = 0.5;
	V.a[0] = V.a[1] = 0.0;
	screenPtC = VectorToPoint(V);
	V.a[2] = 0.6;
	C = VectorToPoint(V);
		
	::PenSize(3,3);
	SetFgColor(gBluePixel) ;
	DrawLine(screenPt0.x,screenPt0.y,screenPtA.x,screenPtA.y); 
	DrawLine(screenPt0.x,screenPt0.y,screenPtB.x,screenPtB.y);
	DrawLine(screenPt0.x,screenPt0.y,screenPtC.x,screenPtC.y); 

	::MoveTo(A.x, A.y);
	::DrawString("A");
	//::DrawText("A", 0, 1);
	::MoveTo(B.x, B.y);
	::DrawString("B");
	//::DrawText("B", 0, 1);
	::MoveTo(C.x, C.y);
	//::DrawText("C", 0, 1);
	::DrawString("C");
	::PenNormal();	
	SetFgColor(gBlackPixel) ;

}


// this gets called alot so I made a subroutine 
// just takes a vector in crystal space and maps to screen coords
Point 
TCrystal::VectorToPoint(TVector inV)
{
	Point outPt;
	float x,y,z,xt,yt,zt, scale	= 1/mXSection.AngstrPrPixel;
	
	x = inV.a[0] ;
	y = inV.a[1] ;
	z = inV.a[2] ;
	
	xt = mXyzToScreenMatrix.m[0][0]*x + mXyzToScreenMatrix.m[0][1]*y + mXyzToScreenMatrix.m[0][2]*z ;
	yt = mXyzToScreenMatrix.m[1][0]*x + mXyzToScreenMatrix.m[1][1]*y + mXyzToScreenMatrix.m[1][2]*z ;
	zt = mXyzToScreenMatrix.m[2][0]*x + mXyzToScreenMatrix.m[2][1]*y + mXyzToScreenMatrix.m[2][2]*z ;	
	xt = scale*xt+mXyzToScreenMatrix.m[0][3] ;
	yt = scale*yt+mXyzToScreenMatrix.m[1][3] ;
	zt = scale*zt+mXyzToScreenMatrix.m[2][3] ;	
	outPt.x = near(xt) ;
	outPt.y = mHeight - near(yt) ;
	
	//outPt.x += mDisplayRect.left ;
	//outPt.y += mDisplayRect.top ;
	return outPt;
}


void
TCrystal::SortScreenAtoms(void) {
	long gap,n ;
	register long i,j;
	SAtom tAtom ; 
	
	n = mSnumAtoms ;
	for (gap=n/2;gap>0;gap/=2) {
		for(i=gap;i<n;i++) {
			for(j=i-gap;j>=0;j-=gap) {
				if ((mScreenAtom+j)->location.d <= (mScreenAtom+j+gap)->location.d)
					break ;					
				tAtom = *(mScreenAtom+j) ;				
				*(mScreenAtom+j) = *(mScreenAtom+j+gap) ;				
				*(mScreenAtom+j+gap) = tAtom ;
			}
		}
	}
}

void TCrystal::AddViewAtom(float x,float y, float z,float TZ0,float scale,short i,float atomrad,RGBColor *color)
{
	float 	xt,yt,zt ;
	short	x0,y0,z0 ;
	float 	rad ;
	SAtom *whichAtom ;
	
	xt = mXyzToScreenMatrix.m[0][0]*x + mXyzToScreenMatrix.m[0][1]*y + mXyzToScreenMatrix.m[0][2]*z ;
	yt = mXyzToScreenMatrix.m[1][0]*x + mXyzToScreenMatrix.m[1][1]*y + mXyzToScreenMatrix.m[1][2]*z ;
	zt = mXyzToScreenMatrix.m[2][0]*x + mXyzToScreenMatrix.m[2][1]*y + mXyzToScreenMatrix.m[2][2]*z ;

	rad = mAtomScaleFactor*atomrad*scale;	// Easier to work with
	if(atomrad <= 0.0)
	{
		rad = mAtomScaleFactor*covrad[(mAtomPtr+i)->no-1]*scale ; 
	}
	whichAtom = (mScreenAtom+mSnumAtoms) ;
	(mScreenAtom+mSnumAtoms)->atom = *(mAtomPtr+i) ;
	(mScreenAtom+mSnumAtoms)->atom.xc = x ; 
	(mScreenAtom+mSnumAtoms)->atom.yc = y ; 
	(mScreenAtom+mSnumAtoms)->atom.zc = z ;
	(mScreenAtom+mSnumAtoms)->position.h = xt ;
	(mScreenAtom+mSnumAtoms)->position.v = yt ;
	(mScreenAtom+mSnumAtoms)->position.d = zt ;
	(mScreenAtom+mSnumAtoms)->index = i ; 			// index into (mAtomPtr+i) ;
	
	xt = scale*xt+mXyzToScreenMatrix.m[0][3] ;
	yt = scale*yt+mXyzToScreenMatrix.m[1][3] ;
	zt = scale*zt+mXyzToScreenMatrix.m[2][3] ;
	
	x0 = near(xt) ;
	y0 = near(yt) ;
	z0 = near(zt) ;
	(mScreenAtom+mSnumAtoms)->location.h = x0 ;
	(mScreenAtom+mSnumAtoms)->location.v = y0 ;
	(mScreenAtom+mSnumAtoms)->location.d = z0 ;
	(mScreenAtom+mSnumAtoms)->rad = near(rad) ;
	(mScreenAtom+mSnumAtoms)->state = (mAtomPtr+i)->state ;
	(mScreenAtom+mSnumAtoms)->groupnumber = (mAtomPtr+i)->groupnumber ;
	mSnumAtoms++ ;
}

void TCrystal::SetTMatrix() {
	float csa,csb,csg,sia,sib,sig ;

	csa = CheckZeroOne(cos(mAlpha)) ;
	csb = CheckZeroOne(cos(mBeta)) ;
	csg = CheckZeroOne(cos(mGamma)) ;
	sia = CheckZeroOne(sin(mAlpha)) ;
	sib = CheckZeroOne(sin(mBeta)) ;
	sig = CheckZeroOne(sin(mGamma)) ;
    
    
	mTmatrix.m[0][0] = mA ;
	mTmatrix.m[0][1] = 0. ;
	mTmatrix.m[0][2] = 0. ;
	mTmatrix.m[0][3] = 0. ;
	mTmatrix.m[1][0] = mB*csg ;
	mTmatrix.m[1][1] = mB*sig ;
	mTmatrix.m[1][2] = 0. ;
	mTmatrix.m[1][3] = 0. ;
	mTmatrix.m[2][0] = mC*csb ;
	mTmatrix.m[2][1] = mC*(csa-csg*csb)/sig ;
	mTmatrix.m[2][2] = mC*sqrt(1.-csa*csa-csb*csb-csg*csg+2.*csa*csb*csg)/sig ;
	mTmatrix.m[2][3] = 0. ;
	mTmatrix.m[3][0] = 0. ;
	mTmatrix.m[3][1] = 0. ;
	mTmatrix.m[3][2] = 0. ;
	mTmatrix.m[3][3] = 1. ;
}

void TCrystal::SetSMatrix() {
	float csa,csb,csg,aa,ab,ac,bb,bc,cc ;
	csa = CheckZeroOne(cos(mAlpha)) ;
	csb = CheckZeroOne(cos(mBeta)) ;
	csg = CheckZeroOne(cos(mGamma)) ;
	aa 	= mA*mA ;
	ab	= mA*mB ;
	ac 	= mA*mC ;
	bb	= mB*mB ;
	bc	= mB*mC ;
	cc	= mC*mC ;
	
	mSmatrix.m[0][0] 	= aa ;
	mSmatrix.m[0][1]	= ab*csg ;
	mSmatrix.m[0][2]	= ac*csb ;
	mSmatrix.m[0][3] 	= 0. ;
	mSmatrix.m[1][0]	= ab*csg ;
	mSmatrix.m[1][1]	= bb ;
	mSmatrix.m[1][2]	= bc*csa ;
	mSmatrix.m[1][3] 	= 0. ;
	mSmatrix.m[2][0]	= ac*csb ;
	mSmatrix.m[2][1]	= bc*csa ;
	mSmatrix.m[2][2]	= cc ;
	mSmatrix.m[2][3] 	= 0. ;
	mSmatrix.m[3][0] 	= 0. ;
	mSmatrix.m[3][1] 	= 0. ;
	mSmatrix.m[3][2] 	= 0. ;
	mSmatrix.m[3][3] 	= 1. ;
}

void TCrystal::SetDisRotMatrix(TypeRotAxis whichAxis,float angle) {
	float 	csangle,siangle ;
	TMatrix 	tRotMatrix(1.0),oldMatrix ;
	
	oldMatrix = mRotMatrix ;
			
	csangle = cos(angle) ;
	siangle = sin(angle) ;
	
	switch(whichAxis) {
		case xRotAxis:
			tRotMatrix.m[0][0] = 1. ;
			tRotMatrix.m[0][1] = 0. ;
			tRotMatrix.m[0][2] = 0. ;
			tRotMatrix.m[1][0] = 0. ;
			tRotMatrix.m[2][0] = 0. ;
			tRotMatrix.m[1][1] = csangle ;
			tRotMatrix.m[1][2] = siangle ;
			tRotMatrix.m[2][1] = -siangle ;
			tRotMatrix.m[2][2] = csangle ;
			break ;
		case yRotAxis :
			tRotMatrix.m[1][1] = 1. ;
			tRotMatrix.m[0][1] = 0. ;
			tRotMatrix.m[1][0] = 0. ;
			tRotMatrix.m[1][2] = 0. ;
			tRotMatrix.m[2][1] = 0. ;
			tRotMatrix.m[0][0] = csangle ;
			tRotMatrix.m[2][0] = siangle ;
			tRotMatrix.m[0][2] = -siangle ;
			tRotMatrix.m[2][2] = csangle ;
			break ;
		case zRotAxis:
			tRotMatrix.m[2][2] = 1. ;
			tRotMatrix.m[0][2] = 0. ;
			tRotMatrix.m[1][2] = 0. ;
			tRotMatrix.m[2][0] = 0. ;
			tRotMatrix.m[2][1] = 0. ;
			tRotMatrix.m[0][0] = csangle ;
			tRotMatrix.m[0][1] = siangle ;
			tRotMatrix.m[1][0] = -siangle ;
			tRotMatrix.m[1][1] = csangle ;
			break ;
	}
	
	mRotMatrix = oldMatrix*tRotMatrix ;
}



void TCrystal::Draw2DScreenBalls(void) {
	register long i ;
	
	SetFgColor(gBlackPixel) ;
	for(i=0;i< mSnumAtoms;i++) {
		DrawAtom(mScreenAtom+i) ;
	}
	SetFgColor(gBlackPixel) ;

}


void TCrystal::DrawInvisibleUCellFaces() {
	long i ;
	
	
	SetLineStyle(DOT) ;
	for (i=0;i<6;i++) {
		if (!IsFaceVisible(mUCellFace[i])) 
			DrawFace(mUCellFace[i]) ;
	}
	SetLineStyle(SOLID) ;
}


void TCrystal::DrawVisibleUCellFaces() {
	long i ;
	for (i=0;i<6;i++) {
		if (IsFaceVisible(mUCellFace[i])) 
			DrawFace(mUCellFace[i]) ;
	}
}

// -------------------------------------------------------------------
// Sorts the faces according to their height on the screen           
// 
void TCrystal::SortFaces() {
  

	long 	i,j,gap,n ;
	UCellFace	temp ;
 
 	n = 6 ;
	for (gap = n/2; gap > 0 ;gap /= 2)
		for (i = gap; i < n; i++)
			for (j = i-gap;j >= 0;j -= gap) {
          			if ((mUCellFace[j].cz-mUCellFace[j+gap].cz) <= 0.) 
             				break ;  
          			temp = mUCellFace[j] ;
          			mUCellFace[j] = mUCellFace[j+gap] ;
          			mUCellFace[j+gap] = temp ;
        		}
}


Boolean TCrystal::IsFaceVisible(UCellFace face) {
	return(face.normal[2] > 0.) ;
}


void TCrystal::DrawFace(UCellFace face) {
	
	DrawLine(near(face.corner[0].h),near(mHeight-face.corner[0].v),
				near(face.corner[1].h),near(mHeight-face.corner[1].v)) ;
	DrawLine(near(face.corner[1].h),near(mHeight-face.corner[1].v),
				near(face.corner[2].h),near(mHeight-face.corner[2].v)) ;
	DrawLine(near(face.corner[2].h),near(mHeight-face.corner[2].v),
				near(face.corner[3].h),near(mHeight-face.corner[3].v)) ;
	DrawLine(near(face.corner[3].h),near(mHeight-face.corner[3].v),
				near(face.corner[0].h),near(mHeight-face.corner[0].v)) ;
}

void TCrystal::SetUCellFaces() {
	register long i,j,k ;
	register float ax,bx,cx ;
	float x1,y1,x2,y2,z1,z2 ;
	TMatrix M ;
	
	M = mXyzToScreenMatrix ;
	
	mUCellFace[0].corner[0].h = 0. ;
	mUCellFace[0].corner[0].v = 0. ;
	mUCellFace[0].corner[0].d = 0. ;
	mUCellFace[0].corner[1].h = 0. ;
	mUCellFace[0].corner[1].v = mYViewCells ;
	mUCellFace[0].corner[1].d = 0. ;
	mUCellFace[0].corner[2].h = mXViewCells ;
	mUCellFace[0].corner[2].v = mYViewCells ;
	mUCellFace[0].corner[2].d = 0. ;
	mUCellFace[0].corner[3].h = mXViewCells ;
	mUCellFace[0].corner[3].v = 0. ;
	mUCellFace[0].corner[3].d = 0. ;
	
	
	mUCellFace[1].corner[0].h = 0. ;
	mUCellFace[1].corner[0].v = 0. ;
	mUCellFace[1].corner[0].d = 0. ;
	mUCellFace[1].corner[1].h = mXViewCells ;
	mUCellFace[1].corner[1].v = 0. ;
	mUCellFace[1].corner[1].d = 0. ;
	mUCellFace[1].corner[2].h = mXViewCells ;
	mUCellFace[1].corner[2].v = 0. ;
	mUCellFace[1].corner[2].d = mZViewCells ;
	mUCellFace[1].corner[3].h = 0. ;
	mUCellFace[1].corner[3].v = 0. ;
	mUCellFace[1].corner[3].d = mZViewCells ;
	
	mUCellFace[2].corner[0].h = 0. ;
	mUCellFace[2].corner[0].v = 0. ;
	mUCellFace[2].corner[0].d = 0. ;
	mUCellFace[2].corner[1].h = 0. ;
	mUCellFace[2].corner[1].v = 0. ;
	mUCellFace[2].corner[1].d = mZViewCells ;
	mUCellFace[2].corner[2].h = 0. ;
	mUCellFace[2].corner[2].v = mYViewCells ;
	mUCellFace[2].corner[2].d = mZViewCells ;
	mUCellFace[2].corner[3].h = 0. ;
	mUCellFace[2].corner[3].v = mYViewCells ;
	mUCellFace[2].corner[3].d = 0. ;
	
	mUCellFace[3].corner[0].h = 0. ;
	mUCellFace[3].corner[0].v = 0. ;
	mUCellFace[3].corner[0].d = mZViewCells ;
	mUCellFace[3].corner[1].h = mXViewCells ;
	mUCellFace[3].corner[1].v = 0. ;
	mUCellFace[3].corner[1].d = mZViewCells ;
	mUCellFace[3].corner[2].h = mXViewCells ;
	mUCellFace[3].corner[2].v = mYViewCells ;
	mUCellFace[3].corner[2].d = mZViewCells ;
	mUCellFace[3].corner[3].h = 0. ;
	mUCellFace[3].corner[3].v = mYViewCells ;
	mUCellFace[3].corner[3].d = mZViewCells ;
	
	mUCellFace[4].corner[0].h = mXViewCells ;
	mUCellFace[4].corner[0].v = 0. ;
	mUCellFace[4].corner[0].d = 0. ;
	mUCellFace[4].corner[1].h = mXViewCells ;
	mUCellFace[4].corner[1].v = mYViewCells ;
	mUCellFace[4].corner[1].d = 0. ;
	mUCellFace[4].corner[2].h = mXViewCells ;
	mUCellFace[4].corner[2].v = mYViewCells ;
	mUCellFace[4].corner[2].d = mZViewCells ;
	mUCellFace[4].corner[3].h = mXViewCells ;
	mUCellFace[4].corner[3].v = 0. ;
	mUCellFace[4].corner[3].d = mZViewCells ;
	
	mUCellFace[5].corner[0].h = mXViewCells ;
	mUCellFace[5].corner[0].v = mYViewCells ;
	mUCellFace[5].corner[0].d = 0. ;
	mUCellFace[5].corner[1].h = 0. ;
	mUCellFace[5].corner[1].v = mYViewCells ;
	mUCellFace[5].corner[1].d = 0. ;
	mUCellFace[5].corner[2].h = 0. ;
	mUCellFace[5].corner[2].v = mYViewCells ;
	mUCellFace[5].corner[2].d = mZViewCells ;
	mUCellFace[5].corner[3].h = mXViewCells ;
	mUCellFace[5].corner[3].v = mYViewCells ;
	mUCellFace[5].corner[3].d = mZViewCells ;
	
	for(j=0;j<6;j++) {
		mUCellFace[j].cx = 0. ;
		mUCellFace[j].cy = 0. ;
		mUCellFace[j].cz = 0. ;
		for(i=0;i<4;i++) {
			mUCellFace[j].cx += mUCellFace[j].corner[i].h/4 ;
			mUCellFace[j].cy += mUCellFace[j].corner[i].v/4 ;
			mUCellFace[j].cz += mUCellFace[j].corner[i].d/4 ;
		}
	}
	
	// Transform the vertices
	for(i=0;i<6;i++) {
		for (k=0;k<4;k++) {
			ax = M.m[0][0]*mUCellFace[i].corner[k].h + M.m[0][1]*mUCellFace[i].corner[k].v + M.m[0][2]*mUCellFace[i].corner[k].d  ;
			bx = M.m[1][0]*mUCellFace[i].corner[k].h + M.m[1][1]*mUCellFace[i].corner[k].v + M.m[1][2]*mUCellFace[i].corner[k].d  ;
			cx = M.m[2][0]*mUCellFace[i].corner[k].h + M.m[2][1]*mUCellFace[i].corner[k].v + M.m[2][2]*mUCellFace[i].corner[k].d  ;
			mUCellFace[i].corner[k].h = ax/mXSection.AngstrPrPixel + M.m[0][3] ;
			mUCellFace[i].corner[k].v = bx/mXSection.AngstrPrPixel + M.m[1][3];
			mUCellFace[i].corner[k].d = cx/mXSection.AngstrPrPixel + M.m[2][3];
		}
		x1 = mUCellFace[i].corner[1].h - mUCellFace[i].corner[0].h ;
		y1 = mUCellFace[i].corner[1].v - mUCellFace[i].corner[0].v ;
		z1 = mUCellFace[i].corner[1].d - mUCellFace[i].corner[0].d ;
		x2 = mUCellFace[i].corner[2].h - mUCellFace[i].corner[1].h ;
		y2 = mUCellFace[i].corner[2].v - mUCellFace[i].corner[1].v ;
		z2 = mUCellFace[i].corner[2].d - mUCellFace[i].corner[1].d ;
		mUCellFace[i].normal[0] = y1*z2-z1*y2 ;
		mUCellFace[i].normal[1] = x2*z1-x1*z2 ;
		mUCellFace[i].normal[2] = x1*y2-y1*x2 ;
	}
	
	// Translate the faces to the appropriate location
	for(i=0;i<6;i++) {
		for (k=0;k<4;k++) {
			// mUCellFace[i].corner[k].h += mDisplayRect.left ;
			// mUCellFace[i].corner[k].v -= mDisplayRect.top ;	// Because it is flipped
		}
	}
	SortFaces() ; 
}

void TCrystal::DrawAtom(SAtom *sAtom) {
	Rect theRect ;
	long 	x,y,colorIndex,rad ;

	if ((sAtom->atom).occ == 0.) 
		return ;					// No drawing of atoms with occupancy factor 0
		
	x = sAtom->location.h ;
	y = mHeight - sAtom->location.v ;
	rad = sAtom->rad ;
	colorIndex = (sAtom->atom).no - 1 ;

	// x += mDisplayRect.left ;
	// y += mDisplayRect.top ;
	
	theRect.top = y - rad ;
	theRect.left = x - rad ;
	theRect.bottom = theRect.top+2*rad ;
	theRect.right = theRect.left+2*rad ;

	
	// PmForeColor(MyAtoms->colorindex[colorIndex]) ;
	switch(sAtom->state) {
		case kNormalState:
			colorIndex = (sAtom->atom).no % 39 ;
			SetFgColor(gColors[gNumGrays-1+colorIndex].pixel) ;
			FillCircle(x,y,rad) ;
			SetFgColor(gWhitePixel) ;
			FillCircle(x+rad/3,y-rad/3,rad/4) ;
			SetFgColor(gBlackPixel) ;
			DrawCircle(x,y,rad) ;
			break ;
		case kDeSelectedState:	
			colorIndex = (sAtom->atom).no % 39 ;
			SetFgColor(gColors[gNumGrays-1+colorIndex].pixel) ;
			FillCircle(x,y,rad) ;
			SetFgColor(gWhitePixel) ;
			FillCircle(x+rad/3,y-rad/3,rad/4) ;
			DrawCircle(x,y,rad-1) ;
			SetFgColor(gGreenPixel) ;
			DrawCircle(x,y,rad) ;
			SetFgColor(gBlackPixel) ;
			DrawCircle(x,y,rad+1) ;
			break ;
		case kSelectedState:
			colorIndex = (sAtom->atom).no % 39 ;
			SetFgColor(gColors[gNumGrays-1+colorIndex].pixel) ;
			PenPat() ;
			FillCircle(x,y,rad) ;
			PenNormal() ;
			SetFgColor(gWhitePixel) ;
			FillCircle(x+rad/3,y-rad/3,rad/4) ;
			DrawCircle(x,y,rad-1) ;
			SetFgColor(gRedPixel) ;
			DrawCircle(x,y,rad) ;
			SetFgColor(gBlackPixel) ;
			DrawCircle(x,y,rad+1) ;
			break ;
	}
}

void TCrystal::Print() {
	
	GrafPtr currPort ;
	GetPort(&currPort) ;
		
	if (mDrawOutline) {
		DrawInvisibleUCellFaces() ;
	}
	
	/* Draw the Crystal  */
	Draw2DScreenBalls() ;
	
	if (mDrawOutline) {
		DrawVisibleUCellFaces() ;
	}
	
}

void TCrystal::CreateView(short na,short nb,short nc)
{
	short numUCells ;
	long numAtomsInCells,atomsToAllocate,atomsInUCell ;
	
	numUCells = na*nb*nc ;	
	atomsInUCell = mNumUnitCellAtoms ;	// We set off enough to hold space for 
									// either the basis or all the atoms
	
	
	numAtomsInCells = numUCells*atomsInUCell ;	// Total number of atoms in the
												// expanded unit cells, excluding
												// atoms at the faces added for 
												// viewing
	
	mNumFaceAtoms = CalculateNumFaceAtoms() ;
	atomsToAllocate = numAtomsInCells + numUCells*mNumFaceAtoms ;
	
	if (mScreenAtom != nil) 
		delete[] mScreenAtom ;
	mScreenAtom = new SAtom[atomsToAllocate] ;
	Assert(mScreenAtom) ;
		
	mXViewCells = na ;
	mYViewCells = nb ;
	mZViewCells = nc ;
	
}

void	
TCrystal::SetMagnification(float magnification) { 

	mMagnification = magnification ;
	SetScale() ;
	UpdateViewAtoms() ; // CreateUnitCellViewAtoms() ; 
	DrawCrystal() ; 
}
		
void	
TCrystal::ChangeMagnification(float factor) { 
	float	zoomfactor  = 1./factor ;

	mXSection.AngstrPrPixel *= factor ;
	mMagnification /= factor ;
	SetCenter() ; 
	
	UpdateViewAtoms() ; // CreateUnitCellViewAtoms() ; 
	DrawCrystal() ; 
	if (mSelection != nil) {
		mSelection->Zoom(zoomfactor,0,0) ;
		mSelection->Update(&mDisplayRect) ;
	}			

}

long TCrystal::CalculateNumFaceAtoms(void)
{
	register 	long i,j ;
	register	float x,y,z ;
	 
	j = mNumUnitCellAtoms ;
	
	for(i=0;i<mNumUnitCellAtoms;i++) {
		x = (mAtomPtr+i)->xc ;
		y = (mAtomPtr+i)->yc ;
		z = (mAtomPtr+i)->zc ;
		if (fabs(x) < 0.001) x = 0. ;
		if (fabs(y) < 0.001) y = 0. ;
		if (fabs(z) < 0.001) z = 0. ;
		
		if (x == 0. || y == 0. || z == 0.) {
			if (x==0. && y != 0. && z != 0.) {
				j++ ;
			}
			else if (x==0. && y==0. && z!=0.) {
				j += 3 ;
			}
			else if (x==0. && y==0. && z==0.) {
				j += 7 ;
			}
			else if (y==0. && x!=0. &&z!=0.) {
				j++ ;
			}
			else if (y==0. && x!=0. && z==0.) {
				 j += 3 ;
			}
			else if (z==0. && x!=0. && y!=0.) {
				j++ ;
			}
			else if (z==0. && x==0. && y!=0.) {
				j += 3 ;
			}
		}
	}
	return (j-mAtomsInUCell) ;
}



void TCrystal::UpdateViewAtoms() {
	float 	x,y,z,xt,yt,zt ;
	short	x0,y0,z0 ;
	float 	rad,iorigox,iorigoy,iorigoz,scale ;
	float	txx,txy,txz,tyx,tyy,tyz,tzx,tzy,tzz ;
	long 	i ;

	scale	=	1/mXSection.AngstrPrPixel ;	
	
	txx = mXyzToScreenMatrix.m[0][0] ;
	txy = mXyzToScreenMatrix.m[0][1] ;
	txz = mXyzToScreenMatrix.m[0][2] ;
	
	tyx = mXyzToScreenMatrix.m[1][0] ;
	tyy = mXyzToScreenMatrix.m[1][1] ;
	tyz = mXyzToScreenMatrix.m[1][2] ;
	
	tzx = mXyzToScreenMatrix.m[2][0] ;
	tzy = mXyzToScreenMatrix.m[2][1] ;
	tzz = mXyzToScreenMatrix.m[2][2] ;
	
	SetUCellFaces() ;
	iorigox	=	mXyzToScreenMatrix.m[0][3] ;
	iorigoy =	mXyzToScreenMatrix.m[1][3] ;
	iorigoz =	mXyzToScreenMatrix.m[2][3] ;
	for(i=0;i < mSnumAtoms;i++) {
		x = (mScreenAtom+i)->atom.xc ;
		y = (mScreenAtom+i)->atom.yc ; ;
		z = (mScreenAtom+i)->atom.zc ; ;
		xt = txx*x + txy*y + txz*z ;
		yt = tyx*x + tyy*y + tyz*z ;
		zt = tzx*x + tzy*y + tzz*z ;
		(mScreenAtom+i)->position.h = xt ;
		(mScreenAtom+i)->position.v = yt ;
		(mScreenAtom+i)->position.d = zt ;
		
		xt = scale*xt + iorigox ;
		yt = scale*yt + iorigoy ;
		zt = scale*zt + iorigoz ;
		rad = mAtomScaleFactor*scale*(mScreenAtom+i)->atom.rad ;
		
		x0 = near(xt) ;
		y0 = near(yt) ;
		z0 = near(zt) ;
		(mScreenAtom+i)->location.h = x0 ;
		(mScreenAtom+i)->location.v = y0 ;
		(mScreenAtom+i)->location.d = z0 ;
		(mScreenAtom+i)->rad = near(rad) ;
	}
	SortScreenAtoms() ;

}

void		
TCrystal::ShowSelectedAtom(TAtom *atom) {
	// Find the atom(s) in the view that has this coordinate in the unit cell
	// If there are any, highlight them
	float 	x,y,z ;
	long		l,m,n ;
	float 	scale,TZ0,fn,fm,fl ;
	TAtom	theAtom ;
	
	Boolean removeOldHighlight = true ;
	theAtom.color = atom->color ;
	theAtom.rad = atom->rad ;
	scale = 1./mXSection.AngstrPrPixel ;	
	TZ0		= 1.0 ;
	
	// We need to set the proper port for our drawing
	GrafPtr		originalPort ;
	GetPort(&originalPort) ;
	SetPort((GrafPtr)mDocWindow) ;

	fn = (float)mXViewCells ;
	fm = (float)mYViewCells ;
	fl = (float)mZViewCells ;

	for(l=0;l<mZViewCells;l++) {
		for(m=0;m<mYViewCells;m++) {
			for(n=0;n<mXViewCells;n++) {
				/* Transform the coordinates of the atom(s) in the cell to screen coordinates */
				x = atom->xc + n ;
				y = atom->yc + m ;
				z = atom->zc + l ;
				if (fabs(x) < 0.0001) x = 0. ;
				if (fabs(y) < 0.0001) y = 0. ;
				if (fabs(z) < 0.0001) z = 0. ;	
				
				if (removeOldHighlight) {
					DrawCrystal() ;
					removeOldHighlight = false ;
				}
				SelectThisAtom(x,y,z,TZ0,scale,&theAtom) ;
				if (x == 0. || y == 0. || z == 0.) {
					if (x==0. && y != 0. && z != 0.) {
						SelectThisAtom(fn,y,z,TZ0,scale,&theAtom) ;
					}
					else if (x==0. && y==0. && z!=0.) {
						SelectThisAtom(fn,y,z,TZ0,scale,&theAtom) ;
						SelectThisAtom(x,fm,z,TZ0,scale,&theAtom) ;
						SelectThisAtom(fn,fm,z,TZ0,scale,&theAtom) ;
					}
					else if (x==0. && y==0. && z==0.) {
						SelectThisAtom(fn,y,z,TZ0,scale,&theAtom) ;
						SelectThisAtom(fn,fm,z,TZ0,scale,&theAtom) ;
						SelectThisAtom(x,fm,z,TZ0,scale,&theAtom) ;
						SelectThisAtom(x,y,fl,TZ0,scale,&theAtom) ;
						SelectThisAtom(fn,y,fl,TZ0,scale,&theAtom) ;
						SelectThisAtom(x,fm,fl,TZ0,scale,&theAtom) ;
						SelectThisAtom(fn,fm,fl,TZ0,scale,&theAtom) ;
					}
					else if (y==0. && x!=0. &&z!=0.) {
						SelectThisAtom(x,fm,z,TZ0,scale,&theAtom) ;
					}
					else if (y==0. && x!=0. && z==0.) {
						SelectThisAtom(x,fm,z,TZ0,scale,&theAtom) ;
						SelectThisAtom(x,fm,fl,TZ0,scale,&theAtom) ;
						SelectThisAtom(x,y,fl,TZ0,scale,&theAtom) ;
					}
					else if (z==0. && x!=0. && y!=0.) {
						SelectThisAtom(x,y,fl,TZ0,scale,&theAtom) ;
					}
					else if (z==0. && x==0. && y!=0.) {
						SelectThisAtom(fn,y,z,TZ0,scale,&theAtom) ;
						SelectThisAtom(x,y,fl,TZ0,scale,&theAtom) ;
						SelectThisAtom(fn,y,fl,TZ0,scale,&theAtom) ;
					}
				}

			}
		}
	}
	SetPort(originalPort) ;
}

void
TCrystal::SelectThisAtom(float x,float y,float z,float TZ0,float scale,TAtom *theAtom) {
	float 	xt,yt,zt ;
	short	x0,y0 ;
	float 	rad ;
	Point	where ;
		
	xt = mXyzToScreenMatrix.m[0][0]*x + mXyzToScreenMatrix.m[0][1]*y + mXyzToScreenMatrix.m[0][2]*z ;
	yt = mXyzToScreenMatrix.m[1][0]*x + mXyzToScreenMatrix.m[1][1]*y + mXyzToScreenMatrix.m[1][2]*z ;
	zt = mXyzToScreenMatrix.m[2][0]*x + mXyzToScreenMatrix.m[2][1]*y + mXyzToScreenMatrix.m[2][2]*z ;
	xt = scale*xt+mXyzToScreenMatrix.m[0][3] ;
	yt = scale*yt+mXyzToScreenMatrix.m[1][3] ;
	zt = scale*zt+mXyzToScreenMatrix.m[2][3] ;
	rad = theAtom->rad ;
	
	theAtom->xc = xt ;
	theAtom->yc = yt ;
	theAtom->zc = zt ;
	theAtom->rad = rad ;
	x0 = near(xt) ;
	y0 = near(yt) ;
	where.x = x0 ;
	where.y = mHeight - y0 ;
	HiliteAtom(theAtom, where);
}

SAtom	*TCrystal::GetAtom(Point& mouseLoc)  {
	long i,dsq,dx,dy,radsq,rad,ndx,num ;
	Boolean found ;
	Point	pt ;
	SAtom *whichAtom ;

	ndx = 0 ;
	whichAtom = nil ;
	
	// Position as it is referred to for the atom positions
	pt.x = mouseLoc.x -  mDisplayRect.left ;
	pt.y = mHeight - (mouseLoc.y - mDisplayRect.top) ;

	num = 1 ;
	for (i=mSnumAtoms-1;i >= 0;i--) {
		dx = pt.x-(mScreenAtom+i)->location.h ;
		dy = pt.y-(mScreenAtom+i)->location.v ;
		dsq = dx*dx+dy*dy ;
		rad = (mScreenAtom+i)->rad ;
		radsq = rad*rad ;
		if (dsq < radsq) {
			found = true ;
			ndx = i ;
			mouseLoc.x = (mScreenAtom+i)->location.h + mDisplayRect.left ;
			mouseLoc.y = mHeight - (mScreenAtom+i)->location.v + mDisplayRect.top ;
			whichAtom = (mScreenAtom+i) ;
			break ;
		}
	}
	return (whichAtom) ; 

}

void		
TCrystal::DeselectAllAtomsToBeRefined() {
	for (long i=0;i < mSnumAtoms ;i++) {
		(mScreenAtom+i)->state = kNormalState ;
		(mAtomPtr+(mScreenAtom+i)->index)->state = kNormalState ;
		(mScreenAtom+i)->groupnumber = -1 ;
		(mAtomPtr+(mScreenAtom+i)->index)->groupnumber = -1 ;
	}
	mCurrentGroup = 1 ;
	mHasAtomsSelected = false ;
	DrawCrystal() ;
}

	
void		
TCrystal::AddAtomsToBeRefined(FRect &rect) {
	long index,i,j,k ;
	float x,y ;
	FRect	refinementRect = rect ;
	Boolean foundAnAtom = false ;
	short atomicNumber ;
	
	// because the image coordinate (y) goes down in the image and up with respect to
	// the model we have to switch the meaning and values of top / bottom
	refinementRect.bottom = 1.0 - refinementRect.bottom ;
	refinementRect.top = 1.0 - refinementRect.top ;
	for (i=0;i < mSnumAtoms ;i++) {
		index = (mScreenAtom+i)->index ;
		x = (mAtomPtr+index)->xc ;
		y = (mAtomPtr+index)->yc ;
		if (x >= refinementRect.left && x <= refinementRect.right 
				 && y >= refinementRect.bottom && y <= refinementRect.top) {
			atomicNumber = (mAtomPtr+index)->no ;
			for (j=0;j<gNumActiveElements;j++) {
				if (atomicNumber == gActiveElementNumber[j]) {
					(mScreenAtom+i)->state = kSelectedState ;
					(mScreenAtom+i)->groupnumber = 0 ;
					foundAnAtom = true ;
					for(k=0;k< mSnumAtoms ;k++) {
						if (k != i && index == (mScreenAtom+k)->index) 
						{
						(mScreenAtom+k)->state = kSelectedState ;
						(mScreenAtom+k)->groupnumber = 0 ;
						}
					}
				}
			}
		}
	}
	if (foundAnAtom) mHasAtomsSelected = true ;
}

void		
TCrystal::AddAtomsToBeRefined(Rect &rect) {
	Point atomPosition ;
	long index ;
	Boolean foundAnAtom = false ;
	short atomicNumber ;
	long	i,j,k ;

	for (i=0;i < mSnumAtoms ;i++) {
		atomPosition.x = (mScreenAtom+i)->location.h + mDisplayRect.left ;
		atomPosition.y = mHeight - (mScreenAtom+i)->location.v + mDisplayRect.top ;
		if (PtInRect(&atomPosition,&rect)) {
			index = (mScreenAtom+i)->index ;
			atomicNumber = (mAtomPtr+index)->no ;		
			for (j=0;j<gNumActiveElements;j++) {
				if (atomicNumber == gActiveElementNumber[j]) {
					(mScreenAtom+i)->state = kSelectedState ;
					(mScreenAtom+i)->groupnumber = 0 ;
					foundAnAtom = true ;
					// Find equivalent atoms
					for(k=0;k< mSnumAtoms ;k++) {
						if (k != i && index == (mScreenAtom+k)->index) 
						{
						(mScreenAtom+k)->state = kSelectedState ;
						(mScreenAtom+k)->groupnumber = 0 ;
						}
					}
				}
			}
		}
	}
	if (foundAnAtom) mHasAtomsSelected = true ;
}

void		
TCrystal::AddGroupedAtomsToBeRefined(Rect &rect) {
	Point atomPosition ;
	long index ;
	Boolean foundAnAtom = false ;
	short atomicNumber ;
	long	i,j,k ;
	
	for (i=0;i < mSnumAtoms ;i++) {
		atomPosition.x = (mScreenAtom+i)->location.h + mDisplayRect.left ;
		atomPosition.y = mHeight - (mScreenAtom+i)->location.v + mDisplayRect.top ;
		if (PtInRect(&atomPosition,&rect)) {
			index = (mScreenAtom+i)->index ;
			atomicNumber = (mAtomPtr+index)->no ;
			for (j=0;j<gNumActiveElements;j++) {
				if (atomicNumber == gActiveElementNumber[j]) {
					(mScreenAtom+i)->state = kSelectedState ;
					(mScreenAtom+i)->groupnumber = mCurrentGroup ;
					foundAnAtom = true ;
					// Find equivalent atoms
					for(k=0;k< mSnumAtoms ;k++) {
						if (k != i && index == (mScreenAtom+k)->index) 
						{
						(mScreenAtom+k)->state = kSelectedState ;
						(mScreenAtom+k)->groupnumber = mCurrentGroup ;
						}
					}
				}
			}
		}
	}
	if (foundAnAtom) {
		mHasAtomsSelected = true ;
		mCurrentGroup++ ;
	}
}

Boolean TCrystal::FindAtom(Point& mouseLoc,SAtom *whichAtom) {
	long i,dsq,dx,dy,radsq,rad,ndx,num ;
	Boolean found ;
	Point	pt ;

	found = false ;
	ndx = 0 ;
	
	// Position as it is referred to for the atom positions
	pt.x = mouseLoc.x -  mDisplayRect.left ;
	pt.y = mHeight - (mouseLoc.y - mDisplayRect.top) ;

	num = 1 ;
	for (i=mSnumAtoms-1;i >= 0;i--) {
		dx = pt.x-(mScreenAtom+i)->location.h ;
		dy = pt.y-(mScreenAtom+i)->location.v ;
		dsq = dx*dx+dy*dy ;
		rad = (mScreenAtom+i)->rad ;
		radsq = rad*rad ;
		if (dsq < radsq) {
			found = true ;
			ndx = i ;
			mouseLoc.x = (mScreenAtom+i)->location.h + mDisplayRect.left ;
			mouseLoc.y = mHeight - (mScreenAtom+i)->location.v + mDisplayRect.top ;
			*whichAtom = *(mScreenAtom+i) ;
			break ;
		}
	}
	return found ;
}


void		
TCrystal::AddAtomToBeRefined(const Point& loc) {
	SAtom	*whichAtom ;
	Point	where = loc ;
	long index,i,j ;
	Boolean	foundAtom = false ;
	short atomicNumber ;

	if ((whichAtom = GetAtom(where)) != nil) {
		index = whichAtom->index ;
		atomicNumber = (mAtomPtr+index)->no ;
		for (j=0;j<gNumActiveElements;j++) {
			if (atomicNumber == gActiveElementNumber[j]) {
				whichAtom->state = kSelectedState ;
				whichAtom->groupnumber = 0 ;
				foundAtom = true ;
				for(i=0;i< mSnumAtoms ;i++) {
					if (index == (mScreenAtom+i)->index) {
						(mScreenAtom+i)->state = kSelectedState ;
						(mScreenAtom+i)->groupnumber = 0 ;
					}
				}
			}
		}
		if (!foundAtom) {
			PutMessage("Atom type is not an active element") ;
		}
		else {
			mHasAtomsSelected = true ;
			DrawCrystal() ;
		}
	}
}

void		
TCrystal::RemoveAtomToBeRefined(const Point& loc) {
	SAtom	*whichAtom ;
	Point	where = loc ;
	long index,i ;

	if ((whichAtom = GetAtom(where)) != nil) {
		index = whichAtom->index ;
		whichAtom->state = kDeSelectedState ;
		whichAtom->groupnumber = -1 ;
		(mAtomPtr+index)->state = kDeSelectedState ;
		(mAtomPtr+index)->groupnumber = -1 ;
		for(i=0;i< mSnumAtoms ;i++) {
			if (index == (mScreenAtom+i)->index) {
				(mScreenAtom+i)->state = kDeSelectedState ;
				(mScreenAtom+i)->groupnumber = -1 ;
			}
		}
		DrawCrystal() ;
	}
}	

// -----------------------------------------------------------
// void TCrystal::Inquire(Point *mouseLoc)
// Report back information on the clicked atom	
//
void TCrystal::Inquire(const Point& loc) {
	float 	x,y,z,xt,yt,zt ;
	char 	text[80] ;
	Point	where ;
	where = loc ;
	TMatrix iXyzToScreenMatrix ;
	SAtom	whichAtom ;
	
	if (!FindAtom(where,&whichAtom)) {
		return ;
	}
	
	DrawCrystal(); // to remove previous hilite
	HiliteAtom(&whichAtom.atom, where);
	
	//EncodeAtomName(name,whichAtom.atom.name,whichAtom.atom.tag) ;

	x = whichAtom.position.h ;
	y = whichAtom.position.v ;
	z = whichAtom.position.d ;
	
	InvertMatrix(mXyzToScreenMatrix,iXyzToScreenMatrix) ;
	xt = iXyzToScreenMatrix.m[0][0]*x + iXyzToScreenMatrix.m[0][1]*y + iXyzToScreenMatrix.m[0][2]*z ;
	yt = iXyzToScreenMatrix.m[1][0]*x + iXyzToScreenMatrix.m[1][1]*y + iXyzToScreenMatrix.m[1][2]*z ;
	zt = iXyzToScreenMatrix.m[2][0]*x + iXyzToScreenMatrix.m[2][1]*y + iXyzToScreenMatrix.m[2][2]*z ;

	//sprintf(text,"%s: at (%.3f,%.3f,%.3f)",name,xt,yt,zt) ;
	sprintf(text,"%s: at (%.3f,%.3f,%.3f)",whichAtom.atom.name,xt,yt,zt) ;
	WriteInfo(text) ;
}

void TCrystal::DrawCrystal(TVector &zone) {
	mMagnification = 1. ;
	
	mZone = zone ;			
	SetTransformationMatrix(zone) ;
	
	CreateUnitCellViewAtoms() ;
	DrawCrystal() ;

}
void TCrystal::HiliteAtom(TAtom *theAtom, Point AtomPt) 
{
	
	float scale	=	1/mXSection.AngstrPrPixel ;	
	float atomrad = theAtom->rad * scale;
	if(atomrad <= 0.0)
	{
		short ndx = (mAtomPtr+theAtom->no)->no-1 ;
		atomrad = covrad[ndx]*scale ; 
	}
	long Rad = near(mAtomScaleFactor*atomrad) + 2;
	
	RgnHandle oldClip = NewRgn() ;
	RectRgn(oldClip,&mDocWindow->GetPortRect()) ;
	//GetClip(oldClip) ;
	
	InsetRect(&mDisplayRect,1,1) ;
	RgnHandle newClip = NewRgn() ;
	RectRgn(newClip,&mDisplayRect) ;
	SetClip(newClip) ;	
	InsetRect(&mDisplayRect,-1,-1) ;
	
	RGBColor Hcolor = {56000, 6000, 1500}; // nice red color
	::RGBForeColor(&Hcolor);
	DrawCircle(AtomPt.x+1, AtomPt.y+1, Rad);
	RGBColor black = {0, 0, 0}; 
	::RGBForeColor(&black);
	
	SetClip(oldClip) ;
	DisposeRgn(oldClip) ;
	DisposeRgn(newClip) ;
}


// -----------------------------------------------------------
// void TCrystal::measureDistance(Point *mouseLoc)
// Measure distance between two atoms that are marked
// consecutively	
//
void TCrystal::MeasureDistance(const Point& loc) {
	float 	distance,x1,y1,z1,x2,y2,z2,projDist,dx,dy,dz ;
	SAtom	whichAtom ;
	char 	text[80] ;
	Point	where ;
	where = loc ;

	switch(mWAtom) {
		case kFirstAtom:
			if (!FindAtom(where,&whichAtom)) {
				//SysBeep(1) ;
				return ;
			}
			mWAtom = kSecondAtom ;
			mFirstAtom = whichAtom ;
			SetClip(mDisplayRect) ;
			DrawCross(where.x,where.y) ;
			mFirstPt = where ;
			SetClip(gCurrentPort->GetPortRect()) ;
			return ;
			break ;
		case kSecondAtom:
			if (!FindAtom(where,&whichAtom)) {
				//SysBeep(1) ;
				return ;
			}
			mWAtom = kFirstAtom ;
			mSecondAtom = whichAtom ;
			SetClip(mDisplayRect) ;
			DrawCross(where.x,where.y) ;
			mSecondPt = where ;
			break ;
		default:
			break ;
	}
    
	DrawLine(mFirstPt.x,mFirstPt.y,mSecondPt.x,mSecondPt.y) ;
	SetClip(gCurrentPort->GetPortRect()) ;

	//EncodeAtomName(name1,mFirstAtom.atom.name,mFirstAtom.atom.tag) ;
	//EncodeAtomName(name2,mSecondAtom.atom.name,mSecondAtom.atom.tag) ;
	x1 = mFirstAtom.position.h ;
	y1 = mFirstAtom.position.v ;
	z1 = mFirstAtom.position.d ;
	x2 = mSecondAtom.position.h ;
	y2 = mSecondAtom.position.v ;
	z2 = mSecondAtom.position.d ;
	dx = x2-x1 ;
	dy = y2-y1 ;
	dz = z2-z1 ;
	distance = sqrt(dx*dx+dy*dy+dz*dz) ;
	projDist = sqrt(dx*dx+dy*dy) ;
	sprintf(text,"Dist. between %s' and '%s' :Real: %.2fA,  Proj : %.2fA",mFirstAtom.atom.name,
    		mSecondAtom.atom.name,distance,projDist) ;
	WriteInfo(text) ;
}

void TCrystal::MeasureAngle(const Point &loc) {
	float 	x1,y1,z1,x2,y2,z2,x3,y3,z3,dx1,dy1,dx2,dy2,dz1,dz2,
			length1,length2,csth,theta,projTheta ;
	SAtom 	whichAtom ;
	char 	text[80] ;
	Point	endPoint ;
	Point	where ;
	where = loc ;

	switch(mAnglePt) {
		case kNoAnglePt:
			if (!FindAtom(where,&whichAtom)) {
				//SysBeep(1)  ;
				return ;
			}
			mAnglePt = kfirstEndPoint ;
			SetClip(mDisplayRect) ;
			mFirstAtom = whichAtom ;
			DrawCross(where.x,where.y) ;
			SetClip(gCurrentPort->GetPortRect()) ;
			mFirstPt = where ;
			return ;
			break ;
		case kfirstEndPoint:
			if (!FindAtom(where,&whichAtom)) {
				//SysBeep(1)  ;
				return ;
			}
			mAnglePt = kVertex ;
			mSecondAtom = whichAtom ;
			SetClip(mDisplayRect) ;
			DrawCross(where.x,where.y) ;
			mSecondPt = where ;
			DrawLine(mFirstPt.x,mFirstPt.y,mSecondPt.x,mSecondPt.y) ;
			SetClip(gCurrentPort->GetPortRect()) ;
			return ;
			break ;
		case kVertex:
			if (!FindAtom(where,&whichAtom)) {
				//SysBeep(1)  ;
				return ;
			}
			endPoint = where ;
			SetClip(mDisplayRect) ;
			DrawCross(where.x,where.y) ;
			DrawLine(mSecondPt.x,mSecondPt.y,endPoint.x,endPoint.y) ;
			SetClip(gCurrentPort->GetPortRect()) ;
			mAnglePt = kNoAnglePt ;
			break ;
	}
	
	x1 = mFirstAtom.position.h ;
	y1 = mFirstAtom.position.v ;
	z1 = mFirstAtom.position.d ;
	x2 = mSecondAtom.position.h ;
	y2 = mSecondAtom.position.v ;
	z2 = mSecondAtom.position.d ;
	x3 = whichAtom.position.h ;
	y3 = whichAtom.position.v ;
	z3 = whichAtom.position.d ;
	
	dx1 = x1-x2 ;
	dy1 = y1-y2 ;
	dx2 = x3-x2 ;
	dy2 = y3-y2 ;
	dz1 = z1-z2 ;
	dz2 = z3-z2 ;

	
	length1 = sqrt(dx1*dx1+dy1*dy1+dz1*dz1) ;
	length2 = sqrt(dx2*dx2+dy2*dy2+dz2*dz2) ;
	csth = (dx1*dx2+dy1*dy2+dz1*dz2)/(length1*length2) ;
	theta = acos(csth)*180./3.1415926 ;
	length1 = sqrt(dx1*dx1+dy1*dy1) ;
	length2 = sqrt(dx2*dx2+dy2*dy2) ;
	csth = (dx1*dx2+dy1*dy2)/(length1*length2) ;
	projTheta = acos(csth)*180./3.1415926 ;
	sprintf(text,"Measured Angle:  %.2f Deg.",theta) ;
	WriteInfo(text) ;
}



void TCrystal::Rotate() {
	float 	dAngle = 0.75,angle ;
	Point 	newLoc,oldLoc ;
	TMatrix	t,tr ;
	long	dx,dy ;
	TMatrix	temp,tInv ; ;
	
	GetMouse(&oldLoc) ;
	while (Button()) {
		GetMouse(&newLoc) ;
		if (newLoc.x != oldLoc.x || newLoc.y != oldLoc.y) {
			dx = newLoc.x - oldLoc.x ;
			dy = newLoc.y - oldLoc.y ;
			if (ShiftKeyDown()) {	/* Rotation about the z-axis	*/
				angle = (dx-dy)*dAngle ;
				if (angle != 0.) {
					DegreeToRad(&angle) ;
					SetDisRotMatrix(zRotAxis,angle) ; 
				}
			}
			else { 	// Rotation about the other two axis
					// Always do mA 'horizontal' rotation first
				angle = (newLoc.x - oldLoc.x)*dAngle ;
				if (angle != 0.) {
					DegreeToRad(&angle) ;
					SetDisRotMatrix(yRotAxis,angle) ; 
				}
				angle = (newLoc.y - oldLoc.y)*dAngle ;
				if (angle != 0.) {
					DegreeToRad(&angle) ;
					SetDisRotMatrix(xRotAxis,angle) ; 
				}
			}
			// Draw the axis
			oldLoc = newLoc ;
			temp = mTmatrix*mRotMatrix ;		// Current transformation matrix of	
			TransposeMatrix(temp,mXyzToScreenMatrix) ;	// between mA,mB,mC and screen system
			SetCenter() ;
			UpdateViewAtoms() ;
			DrawCrystal() ;
			
		}
	}
	DrawCrystal() ;
}


void 
TCrystal::RecipToCartVector(TVector &reciprocalVector,TVector &cartesianVector) {
	TMatrix sInverse,tt,t ;

	InvertMatrix(mSmatrix,sInverse) ;
	tt = sInverse*mTmatrix ;
	TransposeMatrix(tt,t) ;
	cartesianVector = t*reciprocalVector ;
}

void	
TCrystal::CreateRefinementList(RefinementList *list) {
	long i,j,ndx,ndx2,currentGroupNumber ;
	float x,y,x2,y2,asq,bsq,drsq ;
	long numNormalGroups ;
	
	
	// First let us delete the old list if there is one to delete
	DeleteRefinementList(list) ;
	// we will traverse the list of atoms a few times to decide 
	// how many atoms and groups to move
	
	list->numGroups = 0 ;
	asq = mA*mA ;
	bsq = mB*mB ;
	
	// First let us go through ungrouped atoms and see if some should be grouped
	// Also let us get rid of duplicates in the blocks
	for(i=0;i<mSnumAtoms;i++) {
		if ((mScreenAtom+i)->state != kDeSelectedState) {
			if ((mScreenAtom+i)->groupnumber == -1) {
				// Are we going to be selected by another atom later
				ndx = (mScreenAtom+i)->index ;
				x = (mAtomPtr+ndx)->xc ;
				y = (mAtomPtr+ndx)->yc ;
				for(j=i+1;j< mSnumAtoms;j++) {
					ndx2 = (mScreenAtom+j)->index ;
					x2 = (mAtomPtr+ndx2)->xc ;
					y2 = (mAtomPtr+ndx2)->yc ;
					drsq = (x2-x)*(x2-x)*asq+(y2-y)*(y2-y)*bsq ;
					if (drsq < 0.01) {	// 0.1
						 if ((mScreenAtom+j)->state == kSelectedState 
								&& (mScreenAtom+j)->groupnumber == 0) {
							// Yes we do get picked up. However, only set 
							// ourselves if we are different 
							if (ndx == ndx2) {
								// Equivalent 
								(mScreenAtom+i)->state = kNormalState ;	
								break ;
							}
							else {
								// Select ourselves by proximity
								// to a selected atom. We may get deselected later
								(mScreenAtom+i)->state = kSelectedState ;
								(mScreenAtom+i)->groupnumber = 0 ;
							}
						}
					}
				}
			}
			else {
				if ((mScreenAtom+i)->groupnumber > 0) {
					ndx = (mScreenAtom+i)->index ;
					for(j=i+1;j< mSnumAtoms;j++) {
						ndx2 = (mScreenAtom+j)->index ;
						if (ndx == ndx2) {
							(mScreenAtom+i)->groupnumber = -1 ;	
							break ;
						}							
					}
				}
			}
		}
	}

	
	// Now we are all selected. Now group ourselves 
	currentGroupNumber = 100 ; // the first group that is not part of a block
	for(i=0;i<mSnumAtoms;i++) {
		if ((mScreenAtom+i)->state == kSelectedState) {
			if ((mScreenAtom+i)->groupnumber == 0) {	
				// This atom has not been set yet
				// Set the groupnumbers for this atom
				(mScreenAtom+i)->groupnumber = currentGroupNumber ;
				ndx = (mScreenAtom+i)->index ;
				x = (mAtomPtr+ndx)->xc ;
				y = (mAtomPtr+ndx)->yc ;
				// Set the groupnumbers for all other atoms 
				// that belong to this atom
				for(j=i+1;j< mSnumAtoms;j++) {
					ndx2 = (mScreenAtom+j)->index ;
					x2 = (mAtomPtr+ndx2)->xc ;
					y2 = (mAtomPtr+ndx2)->yc ;
					drsq = (x2-x)*(x2-x)*asq+(y2-y)*(y2-y)*bsq ;
					if (drsq < 0.01) {	// 0.1
						 if ((mScreenAtom+j)->state == kSelectedState 
								&& (mScreenAtom+j)->groupnumber == 0) {
							(mScreenAtom+j)->groupnumber = 
								(mScreenAtom+i)->groupnumber ;
							(mAtomPtr+(mScreenAtom+j)->index)->state = kSelectedState ;
							(mAtomPtr+(mScreenAtom+j)->index)->groupnumber =
								(mScreenAtom+j)->groupnumber ;
						}
					}
				}
				currentGroupNumber++ ;
			}
		}
	}
	
	numNormalGroups = currentGroupNumber - 100 ;
	list->numGroups = numNormalGroups + mCurrentGroup-1 ;	
	if (list->numGroups == 0) {
		return ;
	}
	
	// Allocate space for the number of groups that we have
	list->groupPtr = new RefinementGroup[list->numGroups] ;
	if (list->groupPtr == NULL) {
		PutMessage("Out of memory. Do not continue with refinement") ;
		return ;
	}
	for(i=0;i<list->numGroups;i++) {
		(list->groupPtr+i)->numInGroup = 0 ;
	}
	
	// All the screen atoms should now have their group index st as either a number >= 100
	// or a number staring at 1 -> mCurrentGroup-1
	// Set the equivalent group index in the atom list, but reduce it to a 
	// number from 1 -> numgroups

	long groupIndex,groupNumber ;
	for(i=0;i<mSnumAtoms;i++) {
		if ((mScreenAtom+i)->state == kSelectedState) {
			currentGroupNumber = (mScreenAtom+i)->groupnumber ;
			if (currentGroupNumber >= 100) 	
				groupNumber = currentGroupNumber - 99 ;	// 100 -> 1  101 -> 2
			else 
				groupNumber = numNormalGroups + currentGroupNumber ; // 1 -> numNormalGroups + 1

			(mScreenAtom+i)->groupnumber = groupNumber ;
			ndx = (mScreenAtom+i)->index ;
			(mAtomPtr+ndx)->state = kSelectedState ;
			(mAtomPtr+ndx)->groupnumber = groupNumber ;
		}
	}

	// Now go through the list of atoms in the structure and count the number in each group
	for(i=0;i<mAtomsInUCell;i++) {
		if ((mAtomPtr+i)->state == kSelectedState) {
			groupIndex = (mAtomPtr+i)->groupnumber-1 ;
			(list->groupPtr+groupIndex)->numInGroup++ ;
		}
	}
	
	// We know how many atoms are in each group
	// Allocate sufficient space
	long	numElements ;
	for(i=0;i<list->numGroups;i++) {
		//fprintf(stderr,"\n **** Group #%d with %d atoms *******\n",
		//	i,(list->groupPtr+i)->numInGroup) ;
		numElements = (list->groupPtr+i)->numInGroup ;
		(list->groupPtr+i)->atomIndex = new long[numElements] ;
		if ((list->groupPtr+i)->atomIndex == NULL) {
			PutMessage("Out of memory. Do not continue with refinement") ;
			return ;
		}
		(list->groupPtr+i)->refine =  new unsigned char[numElements] ;
		if ((list->groupPtr+i)->refine == NULL) {
			PutMessage("Out of memory. Do not continue with refinement") ;
			return ;
		}
	}
	
	// Now go and fill in atom indices for the group
	for(i=0;i<list->numGroups;i++) {
		(list->groupPtr+i)->numInGroup = 0 ; // so we can use as index
	}
	
	for(i=0;i<mAtomsInUCell;i++) {
		if ((mAtomPtr+i)->state == kSelectedState) {
			groupIndex = (mAtomPtr+i)->groupnumber - 1 ;
			ndx = i ;	// Index into the atom list, which is the same as the gProjCell list
			*((list->groupPtr+groupIndex)->atomIndex + (list->groupPtr+groupIndex)->numInGroup) = ndx ;
			*((list->groupPtr+groupIndex)->refine + (list->groupPtr+groupIndex)->numInGroup) = (mAtomPtr+ndx)->refine ;
			(list->groupPtr+groupIndex)->numInGroup++ ;
		}
	}
	// We should now succesfully have set the list
	// Print out the group
	//for(i=0;i<list->numGroups;i++) {
	//	fprintf(stderr,"\n\nGroup #%d  ------ ",i) ;
	//	for( j=0; j < (list->groupPtr+i)->numInGroup ; j++) {
	//		fprintf(stderr,"\nAtom #%d  atomIndex = %d (Atom Groop number = %d) ",
	//			j,*((list->groupPtr+i)->atomIndex+j), 
	//		(mAtomPtr+*((list->groupPtr+i)->atomIndex+j))->groupnumber) ;
	//	}
	//}
	mCurrentGroup = list->numGroups + 1 ;
}

void	
TCrystal::UseRefinementList(RefinementList *list) {
	long i,j,k,ndx  ;


	for(j=0; j < list->numGroups;j++) {
		for(k=0; k < (list->groupPtr+j)->numInGroup ; k++) {
			ndx = *((list->groupPtr+j)->atomIndex+k) ;			// This is the index into gProjCell or mAtomPtr
			// Search the list of screen atoms for this index
			for(i=0;i<mSnumAtoms;i++) {
				if (ndx == (mScreenAtom+i)->index) {
					mHasAtomsSelected = true ;				
					// Yes. Set the highlight state of this one and move to the next atom
					(mScreenAtom+i)->state = kSelectedState ;
					(mScreenAtom+i)->groupnumber = j+1 ;
					(mScreenAtom+i)->atom.refine  = *((list->groupPtr+j)->refine + k) ;
					(mAtomPtr+ndx)->state = kSelectedState ;
					(mAtomPtr+ndx)->groupnumber = j+1 ;
					(mAtomPtr+ndx)->refine  = *((list->groupPtr+j)->refine + k) ;
				}
			}
		}
	}
	mCurrentGroup = list->numGroups + 1 ;
}
