// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------
#include "TDialog.h"
#include "funcdefs.h"
#include <forms.h>
#include "globals.h"
#include "devices.h"

void	AddWindow(TWindow *window) ;
void	RemoveWindow(TWindow *window) ;
void	ProcessModalEvent(XEvent *event) ;
extern	short	gImageDepth ;
extern	Time	gClickTime ;
extern 	Boolean	gIsDoubleClick ;
TDialog::TDialog(short xloc,short yloc,short width,short height) {

	mHasOwnWindow = true ;
	mWindow = new TWindow(xloc,yloc,width,height,gImageDepth,"",false) ;
	mWindow->SetFgColor(gGrayPixel) ;

	mOKButton = nil ;
	mCancelButton = nil ;
	mDone = false ;
	DefineDialog() ;
	mDefaultButton = nil ;

}

TDialog::~TDialog() {
	delete mWindow ;
}

void	TDialog::Show() {
	BringToFront(mWindow) ;
	SetPort(mWindow) ;
	SetBkColor(gGrayPixel) ;
}

void	TDialog::DoUpdate(XEvent *) {
	GrafPtr tPort ;
	GetPort(&tPort) ;
	SetPort(mWindow) ;
	Draw() ;
	SetPort(tPort) ; 
}

void	TDialog::Draw() {

	if (!mIsVisible) {
		mIsVisible = true ;
		XMapRaised(gDisplay,mWindow->GetXWindow()) ;
		XSetInputFocus(gDisplay,mWindow->GetXWindow(),RevertToNone,CurrentTime) ;
	}

	Rect theRect = mWindow->GetPortRect() ;
	SetClip(theRect) ;
	EraseRect(&theRect) ;
	mObjectList->DrawObjects() ; 
	if (mDefaultButton != nil) {
		Rect tRect = mDefaultButton->GetRect() ;
		InsetRect(&tRect,-2,-2) ;
		PenSize(2,2) ;
		FrameRect(&tRect) ;
		PenSize(1,1) ;
	}
}

DialogReturnCode TDialog::DoModalDialog() {
	XEvent event ;
	FL_OBJECT	*theFormsObject ;
	GrafPtr	tPort ;

	GetPort(&tPort) ;
	Show() ;
	SetPort(mWindow) ;
	while (!mDone) {

		theFormsObject = fl_do_forms() ;
		if (theFormsObject == FL_EVENT) {
			fl_XNextEvent(&event) ;
			DoEvent(&event) ;
		}
	}
	SetPort(tPort) ;
	return mDialogReturnCode ;
}

void	TDialog::DoMouseDown(XEvent *theEvent) {
	Point thePoint ;
	TObject *theObject = nil ;
	TEditText	*editText ;

	if (mObjectList->GetNumberOfObjects() > 0) {
		thePoint.x = theEvent->xbutton.x ;
		thePoint.y = theEvent->xbutton.y ;
		theObject = FindObject(&thePoint) ;
		if (theObject != nil) {
			if ((editText=(TEditText *)mEditTextList->GetObject(&thePoint)) != nil) {
				if (mCurrentTextEditField != editText) {
					mCurrentTextEditField->SetFocus(false) ;
				}
				mCurrentTextEditField = editText ;
				mCurrentTextEditField->SetFocus(true) ;
			}
			DoMouseDownInObject(theObject,&thePoint) ;
		}
	}
}


void	TDialog::DoMouseDownInObject(TObject *obj,Point *pt) {
	if (pt != nil)
		if (!obj->DoMouseDown(pt)) return ;
	if (obj == mOKButton) {
		mDone = true ;
		mDialogReturnCode = kDialogOK ;
	}
	else if (obj == mCancelButton) {
		mDone = true ;
		mDialogReturnCode = kDialogCancel ;
	}
}


void	TDialog::DoEvent(XEvent *theEvent) {
	// If the event doesn't belong to us
	// send it to ProcessModalEvent so that other windows
	// can get updated
	if (theEvent->xany.window != mWindow->GetXWindow()) {
		ProcessModalEvent(theEvent) ;
		return ;
	}
	XButtonEvent *buttEvent = (XButtonEvent *)theEvent ;
	KeySym theKey ;
	switch(theEvent->type) {
		case ButtonPress:
			gIsDoubleClick = false ; 
			if ((buttEvent->time - gClickTime) < 400) gIsDoubleClick = true ;
			gClickTime	= buttEvent->time ;
			DoMouseDown(theEvent) ;
			break ;
		case KeyPress:
			theKey = GetEventKey(theEvent) ;
			if (theKey == XK_Return && mDefaultButton != nil) {
				DoMouseDownInObject(mDefaultButton,nil) ;
				break ;
			}
			DoKeyDown(theEvent) ;
			break ;
		case Expose:
			DoUpdate(theEvent) ;
			break ;
	}
}

TMessageDialog::TMessageDialog(short xloc,short yloc,short width,
	short height,char *text) : TDialog(xloc,yloc,width,height) {

	TStaticText *theText = new TStaticText(10,20,width-10,height-50,text) ;
	AddObject(theText) ;
	mOKButton = new TButton(width-60,height-40,width-10,height-10,"OK") ;
	AddObject(mOKButton) ;

}

TMessageDialog::TMessageDialog(char *text) : TDialog(20,20,200,100) {
	TStaticText *theText = new TStaticText(10,10,180,30,text) ;
	AddObject(theText) ;
	mOKButton = new TButton(150,60,190,90,"OK") ;
	AddObject(mOKButton) ;
}

void TMessageDialog::DefineDialog() {
}

TCancelOKDialog::TCancelOKDialog(short xloc,short yloc,short width,
	short height,char *text) : TDialog(xloc,yloc,width,height) {

	TStaticText *theText = new TStaticText(10,20,width-10,height-50,text) ;
	AddObject(theText) ;
	mCancelButton = new TButton(width-160,height-40,width-80,height-10,"Cancel") ;
	AddObject(mCancelButton) ;
	mOKButton = new TButton(width-60,height-40,width-10,height-10,"OK") ;
	AddObject(mOKButton) ;

}

TCancelOKDialog::TCancelOKDialog(char *text) : TDialog(20,20,200,100) {
	TStaticText *theText = new TStaticText(10,10,180,30,text) ;
	AddObject(theText) ;
	mCancelButton = new TButton(80,60,140,90,"Cancel") ;
	AddObject(mCancelButton) ;
	mOKButton = new TButton(150,60,190,90,"OK") ;
	AddObject(mOKButton) ;
}

Boolean	CancelOKDialog(char *text) {
	Point pos ;
	short	width,height,x,y ;
        Boolean status ; 
	GetGlobalMouse(&pos) ;
	width = 220 ;
	height = 100 ;
	x = pos.x - width + 35 ;
	y = pos.y - height + 15 ;
	TCancelOKDialog *theDialog = new TCancelOKDialog(x,y,width,height,text) ;
	status = theDialog->DoModalDialog() ;
	delete theDialog ;
        return status ;

}

void DisplayMessage(char *text) {
	Point pos ;
	short	width,height,x,y ;

	GetGlobalMouse(&pos) ;
	width = 220 ;
	height = 100 ;
	x = pos.x - width + 35 ;
	y = pos.y - height + 15 ;
	TMessageDialog *theDialog = new TMessageDialog(x,y,width,height,text) ;
	theDialog->DoModalDialog() ;
	delete theDialog ;
}

