// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------
#include <stdlib.h>
#include "TWindowList.h"
#include "TResidualImageDisplay.h"
#include "types.h"
#include "PhasegratingRecord.h"
#include "ProjCellRecord.h"
#include "MslicRecord.h"
#include "displ.h"
#include "filenames.h"
#include "funcdefs.h"
#include "XDrivers.h"
#include "globals.h"
#include "storage.h"
void WindowToFile(TGrafPort *window) ;

typedef enum {kSave,kNoSave,kCancelClose} SaveCode ;
SaveCode PromptSave(char *) ;

typedef enum { PSEUDOCOLOR, TRUECOLOR } MonitorDisplayType ;
extern MonitorDisplayType  gMonitorDisplayType ;
extern Visual	*gVisual ;

extern TImage *gExperimentalImage;
extern TImage *gStandardDeviationImage;
Boolean AdjustImageSize(TImage **image,long width,long height) ;

Boolean HasExperimentalImage() ;
Boolean HasStandardDeviationImage() ;

extern	UInt32	gRedPixel, gBluePixel, gGreenPixel ;
extern short	gImageDepth ;

extern long	gBlackPixelIndex ;
extern long	gLighterGrayIndex ;
extern long	gGrayPixelIndex ;
extern long	gDarkerGrayIndex ;
extern long	gWhitePixelIndex ;
extern long	gRedPixelIndex ;
extern long	gGreenPixelIndex ;
extern long	gBluePixelIndex ;

extern	Atom	gDeleteWindowAtom ;

#define MIN(x,y)	((x) < (y) ? (x) : (y))

TResidualImageDisplay::TResidualImageDisplay(long imageWidth,long imageHeight,
				long numImagesHoriz,long numImagesVert) {
				
	long maxWidth,maxHeight ;
	float fractionX,fractionY,fraction ;
	// extern PaletteHandle MyPalette ;
	
	mWidth = imageWidth*numImagesHoriz + numImagesHoriz + 1 ;
	mHeight = imageHeight*numImagesVert + numImagesVert + 1 ;

	
	mNumHorizontalImages = numImagesHoriz ;
	mNumVerticalImages = numImagesVert ;

	mCurrentImageNumber = 0 ;
	mPseudoColor = true ;	// By default
	mIsDirty = true ;		// by default
	
	// We don't know the size of the screen, so we will pretend it is no smaller than
	// 800 by 600
	maxWidth = 800 - 40 ;
	maxHeight = 600 -80 ;
	mImageWidth = imageWidth ;
	mImageHeight = imageHeight ;
	if (mWidth > maxWidth || mHeight > maxHeight) {
		fractionX = (float)maxWidth/mWidth ;
		fractionY = (float)maxHeight/mHeight ;
		fraction = MIN(fractionX,fractionY) ;
		mWidth = near(mWidth*fraction) ;
		mHeight = near(mHeight*fraction) ;
		mImageWidth = (mWidth-(numImagesHoriz + 1))/numImagesHoriz ;
		mImageHeight = (mHeight-(numImagesVert + 1))/numImagesVert ;
	}
	
	long	winX,winY ;
	long	borderWidth ;
    
	XSizeHints sizeHints ;
	XGCValues	gcv ;

	mIsValid = false ;
	mUpdateRegion	= XCreateRegion() ;

	borderWidth = 4 ;
	winX = 40 ;
	winY = 80 ;
  
	SetRect(&mPortRect,0,0,mWidth,mHeight) ;
  	mDepth = gImageDepth ;
 	switch (gMonitorDisplayType) { 
		case PSEUDOCOLOR :
		    mXWindow = XCreateWindow(gDisplay,RootWindow(gDisplay,gScreen),
		    				winX,winY,mWidth,mHeight,
		    				borderWidth,gImageDepth,InputOutput,
		    				gVisualInfo.visual,
		    				CWBackPixel | CWBorderPixel | CWColormap,
		    				&gImageAttributes) ;
		    break ;
		case TRUECOLOR :
		    mXWindow = XCreateWindow(gDisplay,RootWindow(gDisplay,gScreen),
		    				winX,winY,mWidth,mHeight,
		    				borderWidth,gImageDepth,InputOutput,
		    				gVisualInfo.visual,
		    				CWBackPixel | CWBorderPixel | CWColormap,
		    				&gImageAttributes) ;
		    break ;
	}

	mDrawable	= mXWindow ;

	if (mXWindow != nil) mIsValid = true ;

	sizeHints.flags 	= PPosition | PSize | PMinSize | PMaxSize ;
	sizeHints.x	= winX ;
	sizeHints.y	= winY ;
	sizeHints.width	= mWidth ;
	sizeHints.height	= mHeight ;
	sizeHints.min_width 	= mWidth ;
	sizeHints.min_height	= mHeight ;
	sizeHints.max_width 	= mWidth ;
	sizeHints.max_height	= mHeight ;
	XSetStandardProperties(gDisplay,mXWindow,"Residual Image Display",0,
    			0,0,0,&sizeHints) ;
	XSetWMProtocols(gDisplay,mXWindow,&gDeleteWindowAtom,1) ;

	strcpy(mWindowTitle,"Residual Image Display") ;

	if (gMonitorDisplayType == PSEUDOCOLOR) {
	    // If the color map exists, set it up
	    XSetWindowColormap(gDisplay,mXWindow,gColorMap) ;
	    gcv.foreground = gBlackPixel ;
	    gcv.background = gWhitePixel ;
	}
	else {
	    XSetWindowColormap(gDisplay,mXWindow,gColorMap) ;
	    gcv.foreground = BlackPixel(gDisplay,gScreen) ;
	    gcv.background = WhitePixel(gDisplay,gScreen) ;
	}


	mFgColor = gBlackPixel ;
	mBkColor = gWhitePixel ;
    
	mGC = XCreateGC(gDisplay,mXWindow,GCForeground |
    			GCBackground, &gcv) ;
    
	XSetFont(gDisplay,mGC,gFontStructure->fid) ;
	XSelectInput(gDisplay,mXWindow,ExposureMask | KeyPressMask | StructureNotifyMask |
    			KeyReleaseMask | ButtonPressMask | ButtonReleaseMask) ;


	if ((mOffscreen = new TOffscreen(this)) == nil) {
		printf("\n Unable to set up an offscreen environment") ;
		exit(-1) ;
	}

	XMapWindow(gDisplay,mXWindow) ;
	
	SetRect(&mImageRect,0,0,mImageWidth,mImageHeight) ;
		
}

TResidualImageDisplay::~TResidualImageDisplay() {
}

void 
TResidualImageDisplay::Save() {
	WindowToFile(mOffscreen) ;
	mIsDirty = false ;
}

void	
TResidualImageDisplay::DoWMRequestDelete() {
	if (mIsDirty) {
		SaveCode status = PromptSave("Do you want to save the window?") ;
		switch(status) {
			case kCancelClose:
				return ;
			case kSave:
				Save() ;
			case kNoSave:
				break ;
		}
	}
	Close() ;
}

void TResidualImageDisplay::Close() {
	// We need to allow for saving and also cancelling the close
	// For now
	TWindow::Close() ;
}

void 
TResidualImageDisplay::Update(XEvent *event) {
	
	GrafPtr tPort ;
	XRectangle updateRect,clipRect ;
	GetPort(&tPort) ;
	updateRect.x = event->xexpose.x ;
	updateRect.y = event->xexpose.y ;
	updateRect.width = event->xexpose.width ;
	updateRect.height = event->xexpose.height ;
	XUnionRectWithRegion(&updateRect,mUpdateRegion,mUpdateRegion) ;
	if (event->xexpose.count == 0) {
		SetPort(this) ;
		if (mOffscreen == nil) 
			XSetRegion(gDisplay,mGC,mUpdateRegion) ;
		Draw() ;

		clipRect.x = mPortRect.left ;
		clipRect.y = mPortRect.top ;
		clipRect.width = mPortRect.right ;
		clipRect.height = mPortRect.bottom ;
		XSetClipRectangles(gDisplay,mGC,0,0,
				&clipRect,1,Unsorted) ;
		SetPort(tPort) ;
		XDestroyRegion(mUpdateRegion) ;
		mUpdateRegion = XCreateRegion() ;
	}

}

void		
TResidualImageDisplay::DoMouseDown(XEvent *event) {

	TWindow::DoMouseDown(event) ;

	// We will look for a double click
	// if it is, the open a new window with the residual image
	/* -----------------------------
	GrafPtr tPort ;
	Point 	Loc ;
	long	whichImage,imageNumberClicked ;
	long	TickTime ;
	long i,j,x,y ;
	char text[64] ;
	Rect	posRect ;
	
	GetPort(&tPort) ;
	SetPort((GrafPtr)mWindow) ;
	Loc = event->where ;
	TickTime = event->when ;
	GlobalToLocal(&Loc) ;
	imageNumberClicked = -1 ;
	if ((TickTime-mLastClickTime) <  GetDblTime()) {
		// Find out which image the user double clicked in
		for (whichImage=0;whichImage<mNumHorizontalImages*mNumVerticalImages;whichImage++) {
			i = whichImage%mNumHorizontalImages ;
			j = whichImage/mNumHorizontalImages ;			
			x = i*(mImageWidth + 1) ;
			y = j*(mImageHeight + 1) ;
			posRect = mImageRect ;
			OffsetRect(&posRect,x,y) ;
			if (PtInRect(Loc,&posRect)) {
				imageNumberClicked = whichImage ;
				break ;
			}
		}
	}
	
	mLastClickTime = TickTime ;
	SetPort(tPort) ;
	// If we double clicked into an image then imageNumberClicked != -1
	if (imageNumberClicked != -1 && HasExperimentalImage()) {
		SetWatchCursor() ;

		TImage *experimentalImage = CopyImage(gExperimentalImage) ;
		TImage *sigmaImage = nil ;
		TImage *residualImage ;

		// Let us see if the image dimensions are different from the simulation sampling.
		// If one of the dimensions is different, adjust to fit the calculated image
		if (experimentalImage->mWidth != gPhsgrtRec->nbeam || 
				experimentalImage->mHeight != gPhsgrtRec->mbeam) {
			if (AdjustImageSize(&experimentalImage, gPhsgrtRec->nbeam,gPhsgrtRec->mbeam))
				experimentalImage->mSimulationSampling = true ;
		}
		
		
		if (HasStandardDeviationImage()) {
			sigmaImage = CopyImage(gStandardDeviationImage) ;
			if (sigmaImage->mWidth != gPhsgrtRec->nbeam || 
					sigmaImage->mHeight != gPhsgrtRec->mbeam) {
				if (AdjustImageSize(&sigmaImage, gPhsgrtRec->nbeam,gPhsgrtRec->mbeam))
					sigmaImage->mSimulationSampling = true ;
			}
		}
		if(!open_input_file(flname->iout)) {
			DeleteImage(experimentalImage) ;
			if (HasStandardDeviationImage()) {
				DeleteImage(sigmaImage) ;
			}
			InitCursor() ;	
			return  ;
		}
		TImage calculatedImage ;
		if (!ReadImageFromCalculation(&calculatedImage,imageNumberClicked,
				experimentalImage->mSimulationSampling)) {
			DeleteImage(experimentalImage) ;
			if (HasStandardDeviationImage()) {
				DeleteImage(sigmaImage) ;
			}
			close_input_file() ;
			InitCursor() ;	
			return ;
		}
		residualImage = CopyImage(experimentalImage) ;
		if (residualImage != nil) {
			FRect actionRect = {0.,0.,1.,1.} ;
			float defocus,thickness ;
			
			ComputeResidualImage(experimentalImage,&calculatedImage,sigmaImage,&actionRect,
				 	residualImage,0) ;
			// Find the correct defocus  
			if (gDisplayRec->tfs)   
				defocus = gDisplayRec->f1 + gDisplayRec->df*i ;
			else
				defocus = gDisplayRec->f1 ;
			
			// Find the correct thickness
			if (gMsliceRec->tts)   
				thickness = gMsliceRec->t1 + gMsliceRec->dt*j ;
			else
				thickness = gMsliceRec->t1 ;
			sprintf(text,"T=%.0f,F=%.0f",thickness,defocus) ;
			SetImageName(residualImage,text) ;
			DisplayImage(residualImage) ;
			SetImageModifiedFlag(residualImage,true) ;
		}
		DeleteImage(experimentalImage) ;
		if (HasStandardDeviationImage()) {
			DeleteImage(sigmaImage) ;
		}
		close_input_file() ;
		InitCursor() ;	
	}
	----------------- */
}


Boolean
TResidualImageDisplay::AddImageToDisplay(TImage *image) {
	float	minVal,maxVal,*p1,value,range,difference,meanVal,var,red,green,blue ; 
	long i,j,x,y ;
	unsigned char *ucp ;
	TImage	*tempImage ;	
	
	i = mCurrentImageNumber%mNumHorizontalImages ;
	j = mCurrentImageNumber/mNumHorizontalImages ;
	
	x = i*(mImageWidth + 1) ;
	y = j*(mImageHeight + 1) ;

	tempImage = CopyImage(image) ;
	// Adjust the size so that it fits on the display
	AdjustImageSize(&tempImage,mImageWidth,mImageHeight) ;
	
	minVal = 0. ;
	maxVal = -100000. ;
	p1 = tempImage->mData ;
	for(i=0;i<tempImage->mWidth*tempImage->mHeight;i++) {
		value = fabs(*p1++) ;
		if (value > maxVal)  maxVal = value ;
	}
	
	range = (maxVal-minVal) ;
	
	Rect	posRect = mImageRect ;
	OffsetRect(&posRect,x,y) ;
	
	float	maxDisplayValue = gNumGrays-1 ;
	
	// Create the image values into our workspace
	unsigned char *baseAddress = (unsigned char *)gWorkSpace ;
	unsigned long *ulp = (unsigned long *)baseAddress ;
	long index ;
    	if (gMonitorDisplayType == PSEUDOCOLOR) {
		if (!mPseudoColor) {		// No color-coding
			p1 = tempImage->mData ;
			ucp = baseAddress ;
			for(i=0;i<tempImage->mHeight;i++) {
				for(j=0;j<tempImage->mWidth;j++) {
					difference = *p1++ ;
					var = fabs(difference) ;
					value = maxDisplayValue*(var - minVal)/range ;
					index = (long)value ;
					*ucp++ = (unsigned char)gLut[index] ;
				}
			}
		}
		else {	// We do a pseudocoloring of the image
			minVal = gNumGrays+1 ;
			maxVal = gNumGrays+36 ;
			meanVal = (maxVal+minVal)/2 ;
			p1 = tempImage->mData ;
			ucp = baseAddress ;
			for(i=0;i<tempImage->mHeight;i++) {
				for(j=0;j<tempImage->mWidth;j++) {
					difference = *p1++ ;
					var = fabs(difference) ;
					if (var < 1.0) {
						value = 0. ; // Everything is okay - Black
					}
					else if (var < 1.3) {	// Let's make it green
						value = gGreenPixelIndex ; // Everything is okay
					}
					else {
						if (difference > 0.) {
							value = gRedPixelIndex ; 	// make it red 
						}
						else {
							value = gBluePixelIndex ;	// make it blue 
						}
					}
					index = (long)value ;
					*ucp++ = (unsigned char)gLut[index] ;
				}
			}
		}
	}
    	if (gMonitorDisplayType == TRUECOLOR) {
		if (!mPseudoColor) {		// No color-coding
			p1 = tempImage->mData ;
			ucp = baseAddress ;
			for(i=0;i<tempImage->mHeight;i++) {
				for(j=0;j<tempImage->mWidth;j++) {
					difference = *p1++ ;
					var = fabs(difference) ;
					value = maxDisplayValue*(var - minVal)/range ;
					index = (long)value ;
					*ulp++ = gLut[index] ;
				}
			}
		}
		else {	// We do a pseudocoloring of the image
			p1 = tempImage->mData ;
			ucp = baseAddress ;
			float	tol = 2.5  ;
			for(i=0;i<tempImage->mHeight;i++) {
				for(j=0;j<tempImage->mWidth;j++) {
					difference = *p1++ ;
					var = fabs(difference) ;
					red = 0. ; 
					green = 0. ;
					blue = 0. ; 
					if (var < 1.0) {
						// Everything is okay - Black
					}
					else if (var < 1.2) {	// Let's make it green
						green = 255. ; // Everything is okay
					}
					else {
						value = var ;
						green = 255.*(1-(value-1.2)/(tol-1.2)) ; 
						if (green < 0.) green = 0. ;
						if (difference > 0.) {
							red = (value-1.2)*255/tol ;	// redish 
							if (red > 255.) red = 255 ;
						}
						else {
							blue = (value-1.2)*255/tol ;// blueish 
							if (blue > 255.) blue = 255 ;
						}
					}
					index = (long)value ;
					*ucp++ = 0 ;
					*ucp++ = red ;
					*ucp++ = green ;
					*ucp++ = blue ;
				}
			}
		}

	}
	SetOffscreenPort(mOffscreen) ;
	DrawRGBImage(baseAddress,posRect.left,posRect.top,
		tempImage->mWidth,tempImage->mHeight) ;
	UnSetOffscreenPort(mOffscreen)	;

	DeleteImage(tempImage) ;
	mCurrentImageNumber++ ;
	return true ;
}
