// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------
#include "TShiftResultTable.h"
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#include "XDrivers.h"
#include "TFile.h"

void	PutMessage(char *) ;

TShiftResultTable::TShiftResultTable()
{
	mExactMatch = 1.0 ;
	mGoodMatch = 0.9 ;
	mImage = NULL ;
}
TShiftResultTable::TShiftResultTable(TImage *result) 
	: TResultTable(result->mHeight,result->mWidth/3,sizeof(TShiftResultTableData),60,34) 
{
	long row,col ;
	long width,height ;
	char	text[5] ;
	TableCellT theCell ;
	float *valPtr ;
	TShiftResultTableData data ;
	
	width = result->mWidth/3 ;
	height = result->mHeight ;
	
	for(row=1;row <= height;row++) {
		sprintf(text,"T%d",row) ;
		SetRowLabel(row,text) ;
	}
	for(col=1;col <= width;col++) {
		sprintf(text,"F%d",col) ;
		SetColLabel(col,text) ;
	}
	
	valPtr = result->mData ;
	for(theCell.row=1;theCell.row <= height;theCell.row++) {
		for (theCell.col = 1;theCell.col <= width;theCell.col++) {
			data.xcf = *valPtr++ ;
			data.shiftn = *valPtr++ ;
			data.shiftm = *valPtr++ ;			
			SetCellData(theCell,(void *)&data) ;
		}
	}
	mExactMatch = 1.0 ;
	mGoodMatch = 0.9 ;
	mImage = result ;
}

void	
TShiftResultTable::SetSuccessValues(float exactMatch,float goodMatch) {
	float *valPtr,bestValue ;
	TableCellT theCell ;
	long width,height ;
	
	mExactMatch = exactMatch ;
	mGoodMatch = goodMatch ; 
	
	if (mImage == NULL)
		return ;
		
	width = mImage->mWidth/3 ;
	height = mImage->mHeight ;
		
	valPtr = mImage->mData ;
	bestValue  = mExactMatch > mGoodMatch ? -10000000. : 10000000. ;
	
	for(theCell.row=1;theCell.row <= height;theCell.row++) {
		for (theCell.col = 1;theCell.col <= width;theCell.col++) {
			if (mExactMatch > mGoodMatch) {
				if (*valPtr > bestValue) {
					bestValue = *valPtr ;
					mBestCell = theCell ;
				}
			}
			else {
				if (*valPtr < bestValue) {
					bestValue = *valPtr ;
					mBestCell = theCell ;
				}
			}
			valPtr += 3 ;
		}
	}
}
void TShiftResultTable::DrawCellValue(
	TableCellT &inCell,
	Rect &cellFrame)
{
	char 	str[64] ;
	float	fvalue,value ;
	long 	shiftx,shifty ;
	RGBColor theFgColor ;
	TShiftResultTableData data ;
	
	//TextFont(applFont);
	TextSize(10);
	MoveTo(cellFrame.left + 4, cellFrame.top + 16);
	
		
	FetchCellData(inCell,(void *)&data) ;
	fvalue = data.xcf ;
	shiftx = data.shiftn ;
	shifty = data.shiftm ;
	
	theFgColor.red = theFgColor.green = theFgColor.blue = 0 ;
	if (mExactMatch > mGoodMatch) {	// As in cross-correlation
		if (fvalue > mGoodMatch) {
			theFgColor.red = 30000 + (fvalue-mGoodMatch)/(mExactMatch-mGoodMatch)*25535 ;
			FormRGBForeColor(&theFgColor) ;
		}
	}
	else {	// Lower is better
		if (fvalue < mGoodMatch) {
			theFgColor.red = 30000 + (fvalue-mGoodMatch)/(mExactMatch-mGoodMatch)*25535 ;
			FormRGBForeColor(&theFgColor) ;
		}
	}
	if (inCell.col == mBestCell.col && inCell.row == mBestCell.row) {
			//TextFace(bold) ;
	}
	value = fabs(fvalue) ;
	if (value <= 1.0) {
		sprintf(str,"%.4f",fvalue) ;
	}
	else if (value < 10.0) {
		sprintf(str,"%.3f",fvalue) ;
	}
	else if (value < 100.0) {
		sprintf(str,"%.2f",fvalue) ;
	}
	else if (value < 1000.0) {
		sprintf(str,"%.1f",fvalue) ;
	}
	else {
		sprintf(str,"%6.2E",fvalue) ;
	}
	DrawString(str) ;
	
	MoveTo(cellFrame.left + 4, cellFrame.top + 30);
	sprintf(str,"%d,%d",shiftx,shifty) ;
	DrawString(str) ;
	
	
	theFgColor.red = 0 ;
	theFgColor.green = 0 ;
	theFgColor.blue = 0 ;
	TextFace(0) ;
	FormRGBForeColor(&theFgColor) ;
}

void TShiftResultTable::Save() 
{
	// ------------------------------------------------------------------ 
	// Put up the standard file dialog Box to determine the file name 
	// ------------------------------------------------------------------ 
	FSInfo 	theFInfo ;
	
	strcpy(theFInfo.mFileName,"") ;
	ChooseFile("Save Table data as:","*",theFInfo) ;
	
	if (!theFInfo.mReply) 
		return ;
	
	char buffer[256] ;
	long i ;
	long inByteCount ;
	char str[64] ;
	TableCellT theCell ;
	TShiftResultTableData data ;
	FILE	*fp ;
	
	if ((fp = fopen(theFInfo.mFileName,"w")) == nil) {
		sprintf(buffer,"Unable to open the file %s",theFInfo.mFileName) ;
		PutMessage(buffer) ;
		return ;
	}
	
	for(i=1;i<mCols;i++) {
		mColLabels.FetchItemAt(i,(void *)str) ;
		if (i == 1) {
			sprintf(buffer,"T/df\t%s\t",str) ;
		}
		else {
			sprintf(buffer,"%s\t",str) ;
		}
		inByteCount = strlen(buffer) ;
		fwrite(buffer,inByteCount,1,fp) ;
	}
	mColLabels.FetchItemAt(mCols,(void *)str) ;
	sprintf(buffer,"%s\r",str) ;
	inByteCount = strlen(buffer) ;
	fwrite(buffer,inByteCount,1,fp) ;
	
	for(i=1;i<=mRows;i++) {
		mRowLabels.FetchItemAt(i,(void *)str) ;
		sprintf(buffer,"%s\t",str) ;
		inByteCount = strlen(buffer) ;
		fwrite(buffer,inByteCount,1,fp) ;
		theCell.row = i ;
		for(theCell.col=1;theCell.col<mCols;theCell.col++) {
			FetchCellData(theCell,(void *)&data) ;
			sprintf(buffer,"%.4f(%d,%d)\t",data.xcf,data.shiftn,data.shiftm) ;
			inByteCount = strlen(buffer) ;
			fwrite(buffer,inByteCount,1,fp) ;
		}
		theCell.col = mCols ;
		FetchCellData(theCell,(void *)&data) ;
		sprintf(buffer,"%.4f(%d,%d)\r",data.xcf,data.shiftn,data.shiftm) ;
		inByteCount = strlen(buffer) ;
		fwrite(buffer,inByteCount,1,fp) ;
	}
	fclose(fp) ;
}
