// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------
#include "header.h"
#include "menudef.h"
#include "funcdefs.h"
#include "menu_index.h"
#include "storage.h"
#include "MslicRecord.h"
#include "displ.h"
#include "errors.h"
#include "devices.h"
#include "dfpdspl.h"
#include "globals.h"
extern long IMAGE,DIFFRPT,IMAGE_MODE,HISTOGRAM,MONTAGE,REQ_POS,
           DSPLY_X_LOC,DSPLY_Y_LOC,SCALED,EXIT_WAVEF,NUM_ACELL,
           NUM_BCELL,ERASE,NUM_DEFOC,NUM_THICK,no_cmnd ;
extern float RANGE,ZOOMFACT ;
extern MENUS cmndptrs[],ptr_zoom,ptr_add_cells ;
extern FPV funcpntrs[] ;
typedef void (*FPVIpIp)(long *,long *) ;
extern struct {
                long  ipot ;
                long  icmplx ;
              } flag ;
extern struct {
                long SPACE ;
                long X_SIZE ;
                long Y_SIZE ;
              } MONT ;
extern struct {
                long num_pot ;
                long num_wf ;
                long num_image ;
              } whichOne ;
char	gFileName[132] ;
long tab[100];
extern  short gPrintValues ;

typedef enum OutputSource {kPotential,kExitWave,kDiffractionPattern,kImage} OutputSource ;

/*************************************************************************/
/* Function invoked by the "DISPLAY" command in the control menu. Clears */
/*  the screen and draws a new menu                                      */
/*************************************************************************/
void r_display() { 

    eraser(0,120,1020,160) ;
    gCurrentCommand=gCurrentButton ;
    on_highlight() ; 
    off_highlight() ;
    set_menu(DSPL_SOURCE) ;
    CLEAR_DISPLAY ;
    displ_menu() ;
    set_menu(DSPL_OPER) ;
    displ_menu() ;
    set_menu(DSPL_SOURCE) ;
}

/*******************************************************************/
/* return from menu 8. Invoked by the "RETURN" command             */
/*******************************************************************/
void r_ret8() {

    long i ;
    CLEAR_DISPLAY ;
    set_menu(DSPL_SOURCE) ;
    displ_menu() ;
    set_menu(DSPL_OPER) ;
    displ_menu() ;
}  

void r_dispfft(long *np,long *mp) {
    char c ;
    GrafPtr tPort ;
    short sX,sY ;
    long i ;
 
    switch(OutDevice->type) {
		case WS:
			GetPort(&tPort) ;
			SetOffscreenPort(gOffscreenPtr) ;
			if (ERASE) ClearDisplay() ;
			dispfft(np,mp) ;
			UnSetOffscreenPort(gOffscreenPtr) ;
			SetPort(tPort) ;
			UpdateImageScreen() ;
			break ;
		case FS:
			if (ERASE) ClearDisplay() ;
			dispfft(np,mp) ;
			break ;
		case LP:
			sX = dfpdspl.xcenter ;
			sY = dfpdspl.ycenter ;
			dfpdspl.xcenter = OutDevice->xres/2 ;
			dfpdspl.ycenter = OutDevice->yres/2 ;
			dispfft(np,mp) ;
			dfpdspl.xcenter = sX ;
			dfpdspl.ycenter = sY ;
			break ;
	}
}

/*******************************************************************/
/* Set-up routine invoked by the "DISPLAY" command in the DISPLAY  */
/*  menu. Selects the type of image output routine to call         */
/*   depending upon the nature of the "image".                     */
/*******************************************************************/
void r_dsply() {
	char sreply[2] ;

	eraser(0,0,1020,32) ;
	on_highlight() ;
	if (DIFFRPT) {
		if (gPrintValues) {
			eraser(0,0,1000,32) ;
			temdtx(0,10," Name of File for Intensities/Phases -> ") ;
			ReadString(gFileName,132) ;
			eraser(0,0,1000,32) ;
		}
		cmndptrs[no_cmnd] = gCurrentButton ;
		funcpntrs[no_cmnd++] = (FPV)r_dispfft ;
	}
	else if (HISTOGRAM) {
		cmndptrs[no_cmnd] = gCurrentButton ;
		funcpntrs[no_cmnd++] = (FPV)displ_hist ;
	}
	else if (IMAGE_MODE) {
		if (gPrintValues) {
			eraser(0,0,1000,32) ;
			temdtx(0,10," Name of File for Image Data -> ") ;
			ReadString(gFileName,132) ;
			eraser(0,0,1000,32) ;
		}
		if (!requested((FPV)zoom_c))
			if (ZOOMFACT != 1.) {
				cmndptrs[no_cmnd] = ptr_zoom ;
				funcpntrs[no_cmnd++] = (FPV)zoom_c ;
			}
		if (!requested((FPV)add_unit_cells))
			if (NUM_ACELL != 1 || NUM_BCELL != 1) {
				cmndptrs[no_cmnd] = ptr_add_cells ;
				funcpntrs[no_cmnd++] = (FPV)add_unit_cells ;
			}
		cmndptrs[no_cmnd] = gCurrentButton ;
		funcpntrs[no_cmnd++] = (FPV)dsply ;
		if (REQ_POS && !MONTAGE && !gPrintValues) {
			REVERSE_VIDEO ;
			temdtx(0,10,
			" Enter the x y coordinates for upper left corner of image -> ") ;
			DSPLY_X_LOC = get_int(0) ;
			DSPLY_Y_LOC = get_int(0) ;
			NORMAL_VIDEO ;
			eraser(0,0,1020,32) ;
		}
		else {
			DSPLY_X_LOC = -1 ;
			DSPLY_Y_LOC = -1 ;
		}
	}
}

/**********************************************************************/
/* Calling routine. Calls HAFTON for output of halftone images        */
/**********************************************************************/
void dsply(long *np,long *mp) {

	long ierr=0,i,j ;
	GrafPtr	tPort ;
	float *rl_part,*cmplx_part ;

	if (OutDevice->type == WS && !gPrintValues) {
		GetPort(&tPort) ;
		SetOffscreenPort(gOffscreenPtr) ;
		if (ERASE && !MONTAGE) ClearDisplay() ;
	}
	
	if (!SCALED) {
		/* If we are looking at the exit wave function, we need to 
		take the modulus squared of the function to get the image.
		*/
		if (EXIT_WAVEF && flag.icmplx) {
			j = *np**mp ;   /* number of image points */
			rl_part = gWorkSpace ;
			cmplx_part = rl_part+1 ;
			for (i=0;i<j;i++) {
				gWorkSpace[i] = *rl_part**rl_part+*cmplx_part**cmplx_part ;
				rl_part += 2 ;
				cmplx_part += 2 ;
			}
			flag.icmplx = FALSE ;
		}
		
		if (!gPrintValues) {
			RANGE = OutDevice->range ;
			ScaleImageIntensities(np,mp,&RANGE,&ierr) ;
			SCALED = TRUE ;
		}
	}
	hafton(np,mp) ;
	if (OutDevice->type == WS && !gPrintValues) {
		UnSetOffscreenPort(gOffscreenPtr) ;
		UpdateImageScreen() ;
		SetPort(tPort) ;
	}
}

/*************************************************************************/
/* Turns off the highlights and resets parameters. Cancels all requests  */
/*  while in the DISPLAY menu                                            */
/*************************************************************************/
void r_cancel() {

    long i ;
    eraser(0,0,1020,32) ;
    on_highlight() ;
    gCurrentCommand = gCurrentButton ;
    off_highlight() ;
    for (i=0;i<no_cmnd;i++)
    {
      gCurrentCommand = cmndptrs[i] ;
      off_highlight() ;
    }
    no_cmnd = 0 ;
    IMAGE_MODE = FALSE ; DIFFRPT = FALSE ; HISTOGRAM = FALSE ; 
    SCALED = FALSE ; EXIT_WAVEF = FALSE ;
    dfpdspl.wavef  = FALSE  ;
    set_menu(DSPL_SOURCE) ;
}

/**************************************************************************/
/* Runs in sequence the string of commands that were chosen in the        */
/*  DISPLAY menu                                                          */
/**************************************************************************/
void r_exe() {

	long np,mp,imagewidth,imageheight,x_image_spacing,y_image_spacing ;
	long i,j,k,ierr,WRAP_AROUND=FALSE,i0=0 ;
	long total_number_images,LpOffset=0 ;

	eraser(0,0,1020,32) ;
	if (no_cmnd==1) {
		err_disp(1) ;
		return ;
	}
	gCurrentCommand = gCurrentButton ;      /* Turn on and off highlighting of EXECUTE */
	on_highlight() ;

	x_image_spacing = MONT.X_SIZE ;
	y_image_spacing = MONT.Y_SIZE ;

	SetWatchCursor() ;
	switch(OutDevice->type) {
		case WS:
			break ;
		case LP:
			LpOffset = 20 ;
			open_ps_file("laserprinter.ps") ;
			break ;
		case FS:
			open_semper_file("ncemss.smp") ;
			break ;
	}
	OutputSource theOutputSource ;
	if (funcpntrs[0] == (FPV)GetPotential)
		theOutputSource = kPotential ;
	else if (funcpntrs[0] == (FPV)GetExitWave)
		theOutputSource = kExitWave ;
	else if (funcpntrs[0] == (FPV)diffr)
		theOutputSource = kDiffractionPattern ;
	else if (funcpntrs[0] == (FPV)GetImage)
		theOutputSource = kImage ;

	if (IMAGE_MODE && MONTAGE && 
			(theOutputSource == kExitWave || theOutputSource == kImage) && 
			(gMsliceRec.tts || gDisplayRec.tfs)) {
		/* We will montage a set of images */
		switch(OutDevice->type) {
			case WS:
				SetOffscreenPort(gOffscreenPtr) ;
				if (ERASE) ClearDisplay() ;
				UnSetOffscreenPort(gOffscreenPtr) ;
				SetPort(gControlWindow) ;
				break ;
			case FS:
			case LP:
				break ;
		}

		DSPLY_Y_LOC = LpOffset+x_image_spacing ;
		DSPLY_X_LOC = LpOffset+y_image_spacing ;
		
		switch(theOutputSource) {
			case kExitWave:
				total_number_images = NUM_THICK ;
				break ;
			case kImage:
				total_number_images = NUM_DEFOC*NUM_THICK ;
				break ;
		}
		
		whichOne.num_image = 1 ;
		whichOne.num_wf = 1 ;
		
		if (!gCustomMontage) {
			if (theOutputSource == kImage) {
				for (j=0;j < NUM_THICK;j++) {
					if (!WRAP_AROUND) {
						DSPLY_X_LOC = LpOffset+x_image_spacing ;
					}
					for (k=0;k< NUM_DEFOC ;k++) {
						for (i=0;i<no_cmnd;i++)   {
							gCurrentCommand = cmndptrs[i] ;
							if (j==0 && k== 0) {
								off_highlight() ;
							}
							(*(FPVIpIp)(funcpntrs[i]))(&np,&mp) ;
							if (gIOError) {
								gCurrentCommand = gCurrentButton ;  
								off_highlight() ;
								err_io_ds(gIOError) ;
								InitCursor() ;
								return ;
							}
							if (gSizeError) {
								gCurrentCommand = gCurrentButton ;  
								off_highlight() ;
								err_disp(5) ;
								gSizeError = FALSE ;
								InitCursor() ;
								return ;
							}
						}    
						if (whichOne.num_image == total_number_images) 
							goto finished ;
						whichOne.num_image++ ;
						SCALED = FALSE ;
						imagewidth = np*OutDevice->pixpoint ;
						imageheight = mp*OutDevice->pixpoint ;
						DSPLY_X_LOC += (imagewidth+x_image_spacing) ;
						if ((imagewidth+DSPLY_X_LOC) > OutDevice->xres) {
							/* Wrap around */
							WRAP_AROUND = TRUE ;
							DSPLY_X_LOC = LpOffset+x_image_spacing ;
							DSPLY_Y_LOC += (imageheight+y_image_spacing) ;
							if ((DSPLY_Y_LOC+imageheight) > OutDevice->yres) {
								/* All the images will not fit. Break and issue */
								/* an informational message.....                */
								temdtx(0,10,
								" Images won't fit. Montage manually !") ;
								BELL ;
								goto finished ;
							}                         
						}
					}
					if (!WRAP_AROUND) {
						DSPLY_Y_LOC += (imageheight+y_image_spacing) ;
						if ((DSPLY_Y_LOC+imageheight) > OutDevice->yres) {
							/* All the images will not fit. Break and issue */
							/* an informational message.....                */
							/* REVERSE_VIDEO ; */
							temdtx(0,10," Images won't fit. Montage manually !") ;
							BELL ;
							/* NORMAL_VIDEO ; */
							goto finished ;
						}                         
					}
				}
			}
			else {
				for (k=0;k< NUM_THICK ;k++) {
					for (i=0;i<no_cmnd;i++)   {
						gCurrentCommand = cmndptrs[i] ;
						if (k== 0) {
							off_highlight() ;
						}
						(*(FPVIpIp)(funcpntrs[i]))(&np,&mp) ;
						if (gIOError) {
							gCurrentCommand = gCurrentButton ;  
							off_highlight() ;
							err_io_ds(gIOError) ;
							InitCursor() ;
							return ;
						}
						if (gSizeError) {
							gCurrentCommand = gCurrentButton ;  
							off_highlight() ;
							err_disp(5) ;
							gSizeError = FALSE ;
							InitCursor() ;
							return ;
						}
					}    
					if (whichOne.num_wf == total_number_images) 
						goto finished ;
					whichOne.num_wf++ ;
					SCALED = FALSE ;
					imagewidth = np*OutDevice->pixpoint ;
					imageheight = mp*OutDevice->pixpoint ;
					DSPLY_X_LOC += (imagewidth+x_image_spacing) ;
					if ((imagewidth+DSPLY_X_LOC) > OutDevice->xres) {
						/* Wrap around */
						WRAP_AROUND = TRUE ;
						DSPLY_X_LOC = LpOffset+x_image_spacing ;
						DSPLY_Y_LOC += (imageheight+y_image_spacing) ;
						if ((DSPLY_Y_LOC+imageheight) > OutDevice->yres) {
							/* All the images will not fit. Break and issue */
							/* an informational message.....                */
							temdtx(0,10,
							" Images won't fit. Montage manually !") ;
							BELL ;
							goto finished ;
						}                         
					}
				}
			}
			finished: ;   /* breaks to here if we can't fit all the images */
		} /* End Normal Montage */
		else {  /* This is the custom montage sequence part */
			for(k=1; k <= tab[0] /* Number of Images in Sequence */; k++) {
				whichOne.num_image = tab[k]; /* sequence[k] */
				for (i=0;i<no_cmnd;i++)   {
					gCurrentCommand = cmndptrs[i] ;
					if (k == 1) off_highlight() ;
					(*(FPVIpIp)(funcpntrs[i]))(&np,&mp) ;
					if (gIOError) {
						gCurrentCommand = gCurrentButton ;  
						off_highlight() ;
						err_io_ds(gIOError) ;
						InitCursor() ;
						return ;
					}
					if (gSizeError) {
						gCurrentCommand = gCurrentButton ;  
						off_highlight() ;
						err_disp(5) ;
						gSizeError = FALSE ;
						InitCursor() ;
						return ;
					}
				}    
				if (k == tab[0])  goto fin ;
				SCALED = FALSE  ;
				imagewidth = np*OutDevice->pixpoint ;
				imageheight = mp*OutDevice->pixpoint ;
				DSPLY_X_LOC += (imagewidth+x_image_spacing) ;
				if ((imagewidth+DSPLY_X_LOC) > OutDevice->xres) {
					/* Wrap around */
					WRAP_AROUND = TRUE ;
					DSPLY_X_LOC = LpOffset+x_image_spacing ;
					DSPLY_Y_LOC += (imageheight+y_image_spacing) ;
					if ((DSPLY_Y_LOC+imageheight) > OutDevice->yres) {
						/* All the images will not fit. Break and issue */
						/* an informational message.....                */
						temdtx(0,10,
						" Images won't fit. Montage manually !") ;
						BELL ;
						goto fin ;
					}                         
				}
			}
			fin:   ;   /* breaks to here if we can't fit all the images */
		} /* Custom Montage Sequence */
	} /* Montage */
	else {  /* No montage sequence */
		for (i=0;i<no_cmnd;i++) {  /* Just run through the sequence */
			gCurrentCommand = cmndptrs[i] ;
			off_highlight() ;
			(*(FPVIpIp)(funcpntrs[i]))(&np,&mp) ;
			if (gIOError) {
				err_io_ds(gIOError) ;
				gCurrentCommand = gCurrentButton ;  
				off_highlight() ;
				InitCursor() ;
				return ;
			}
			if (gSizeError) {
				gCurrentCommand = gCurrentButton ;  
				off_highlight() ;
				err_disp(5) ;
				gSizeError = FALSE ;
				InitCursor() ;
				return ;
			}
		}    
	}

	switch(OutDevice->type) {
		case WS:
			break ;
		case LP:
			close_ps_file() ;
			break ;
		case FS:
			close_semper_file() ;
			break ;
	}
	InitCursor() ;
	no_cmnd = 0 ;
	flag.ipot = FALSE ;
	DIFFRPT = FALSE ; IMAGE_MODE = FALSE ; 
	SCALED = FALSE ; EXIT_WAVEF = FALSE ;
	HISTOGRAM = FALSE ;
	dfpdspl.wavef  = FALSE  ;
	set_menu(DSPL_SOURCE) ;
	gCurrentCommand = gCurrentButton ;      /* Turn on and off highlighting of EXECUTE */
	off_highlight() ;
}

void SetCustomMontageSequence() {
/*
   This routine must fill in some table or array containing	
   the sequence of images that will be displayed when the 	
   program tries to montage a sequence of images.  This table 
   has not been declared yet and could be declared in this 	
   file or in the header file globals.h  This routine is called from the
   routine SetCustomMontage whose only function is to toggle the option
   normal/custom sequence, setting the global variable "gCustomMontage"
   to TRUE (1) (defined in globals.h)  and to call this routine.  
   In order to change the montage sequence, you must define some array
   to hold the image numbers, provide a way for the user to enter the 
   numbers and fill in the last part of the code in the routine r_exe
   above.  Just check for the occurence of PSEUDO CODE to find the code
   that must be written.  The program actually functions at the moment.
   When the flag gCustomMontage is set, image number 1 in a thru defocus
   thickness series is displayed in the upper left corner.  Thats all.
*/
	char *get_snumber(),*buffer,stext[132];	
	long lenbuf, i, nn,val;
	long i1,j, total_number_images ;
	long liste,  num, term;

	if (funcpntrs[0] == (FPV)GetPotential)
            total_number_images = 1 ;
	else
            total_number_images = NUM_DEFOC*(funcpntrs[0]==(FPV)GetImage ? 
                                             NUM_THICK : 1) ;
	term = 0;
	sprintf(stext,
 		" Select which images (1-%d) -> ",total_number_images) ;

	while (term == 0)  {
   		term = 1;
   		eraser(0,0,1000,32) ;
   		temdtx(0,10,stext) ;
   		buffer = get_snumber();
   		lenbuf = strlen(buffer);
   		eraser(0,0,1000,32) ;

   		tab[0] = 0;
   		liste = 0;
   		num = 0;
   		nn = 0;
   		for  (i=0;i<lenbuf;i++) {
			if (buffer[i]<='9' && buffer[i] >='0')  {
	     		if (num)  {
	        		val = val*10+buffer[i]-48;
	     		}
	    		else  {
	        		num = 1;
	        		val = buffer[i]-48;
	     		}  
	 		}
			else  {
	        	if (num )  {
					if (liste) {
		  				i1 = tab[nn];
	  	  				for (j=i1;j<val;j++) 
		    				tab[++nn] = j+1;
		  				tab[0] = nn;
		  				liste = 0;	    
	       	 		}
	      	 		else   {
		   				tab[++nn] = val;
		   				tab[0] = nn;
		   				if (buffer[i] == '-')  liste = 1;
		  			}
	        	}
	       		else {
		 			if ((buffer[i] == '-')&&(nn!=0)) liste = 1;
					printf("\n liste = %d",liste);
	       		}        
	      		num = 0;
	   		}
   		}
   		if (num )  {
			if (liste) {
	  			i1 = tab[nn];
  	  			for (j=i1;j<val;j++) 
	    			tab[++nn] = j+1;
	  			tab[0] = nn;
       	 	}
      	 	else   {
	   			tab[++nn] = val;
	   			tab[0] = nn;
	  		}
   		}
/* verification des numeros d'images */
   		for (i=1;i<=tab[0];i++)
		if (tab[i] >total_number_images|| tab[i]<1) {
	   		term = 0;
            BELL ;
            sprintf(stext,
                 " Only %d images available ! Re-enter -> ",total_number_images) ;
		}
	}
}

