// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------
#include "header.h"
#include "funcdefs.h"
#include "menudef.h"
#include "menu_index.h"
#include "atoms.h"
#include "PhasegratingRecord.h"
#include "devices.h"
#include "globals.h"
extern long gNumUnitCellAtoms,DIS_ATOM_INITIAL  ;

long PERSP ;

//--------------------------------------------------------------------
// Menu switching routine. Invoked by "DISPLAY" in the ATOM menu.  
// Sets up the menu for plotting of either new or old unit cell  
//--------------------------------------------------------------------
void dis_at() {

    set_menu(VIEW_UNIT_CEL) ;
    eraser(0,0,1020,100) ;
    eraser(830,300,1020,720) ;
    displ_menu() ;
}

//--------------------------------------------------------------------
// Return from the UNIT-CELL DISPLAY menu  
//--------------------------------------------------------------------
void r_ret5() {

    on_highlight() ;
    gCurrentCommand = gCurrentButton ;
    set_menu(VIEW_ATOMS) ;
    off_highlight() ;
    eraser(0,0,1020,100) ;
    DIS_ATOM_INITIAL = TRUE ;
    displ_atoms(atom) ;
    displ_menu() ;
} 
 
//--------------------------------------------------------------------
// Sets up whether the unit cells should be plotted with or   
// without perspective. Invoked by the menu-commands "PERSPECT" 
// and "NO PERSP"                                              
//--------------------------------------------------------------------
void cel_persp() {
	long ierr,i ;
	float uvw[3] ;
	char c ;

	PERSP = TRUE ;
	on_highlight() ;
	gCurrentCommand = gCurrentButton ;
	do {
		eraser(0,60,960,100) ;
		REVERSE_VIDEO ;
		temdtx(0,70," INPUT THE VIEWING INDICES U,V,W !! ") ;
		uvw[0] = get_float(0.) ;
		uvw[1] = get_float(0.) ;
		uvw[2] = get_float(1.) ;
	} while ( uvw[0]==0. && uvw[1]==0. && uvw[2]==0.) ;
	NORMAL_VIDEO ;
	setup_trnsf() ;
	SetWatchCursor() ;
	switch(OutDevice->type) {
		case WS:
			SetOffscreenPort(gOffscreenPtr) ;
			ClearDisplay() ;
			break ;
		case FS:
			return ;
			break ;
		case LP:
			open_ps_file("laserprinter.ps") ;
			break ;
	}
	ViewOriginalUnitCell(uvw,&ierr) ;
	switch(OutDevice->type) {
		case WS:
			UnSetOffscreenPort(gOffscreenPtr) ;
			UpdateImageScreen() ;
			SetPort(gControlWindow) ;
			break ;
		case LP:
			close_ps_file() ;
			break ;
	}
	off_highlight() ;
	ClearDisplay() ;
	set_menu(VIEW_UNIT_CEL) ;
	displ_menu() ;
	if (ierr == -1) err(4) ;
	if (ierr == -2) err(5) ;
	InitCursor() ;
}

void ViewNewUnitCell() {
	float	view[3]= {0.,0.,1.} ;
	long 	ierr ;
	GrafPtr	savedPort ;
	GetPort(&savedPort) ;
	PERSP = FALSE ;
	setup_trnsf() ;
	SetWatchCursor() ;
	SetOffscreenPort(gOffscreenPtr) ;
	ClearDisplay() ;
	DrawNewUnitCell(view,&ierr) ;
	UnSetOffscreenPort(gOffscreenPtr) ;
	UpdateImageScreen() ;
	SetPort(savedPort) ;
	InitCursor() ;
}

//--------------------------------------------------------------------
// Sets up whether the unit cells should be plotted with or   
// without perspective. Invoked by the menu-commands "PERSPECT"
// and "NO PERSP"                                            
//--------------------------------------------------------------------
void cel_nopersp() {
	long ierr,i ;
	float uvw[3] ;
	char c ;

	PERSP = FALSE ;
	on_highlight() ;
	gCurrentCommand = gCurrentButton ;
	do {
		eraser(0,60,960,100) ;
		REVERSE_VIDEO ;
		temdtx(0,70," INPUT THE VIEWING INDICES U,V,W !! ") ;
		uvw[0] = get_float(0.) ;
		uvw[1] = get_float(0.) ;
		uvw[2] = get_float(1.) ;
	} while ( uvw[0]==0. && uvw[1]==0. && uvw[2]==0.) ;
	
	NORMAL_VIDEO ;
	setup_trnsf() ;
	SetWatchCursor() ;
	switch(OutDevice->type) {
		case WS:
			SetOffscreenPort(gOffscreenPtr) ;
			ClearDisplay() ;
			break ;
		case FS:
			return ;
			break ;
		case LP:
			open_ps_file("laserprinter.ps") ;
			break ;
	}
	ViewOriginalUnitCell(uvw,&ierr) ;
	switch(OutDevice->type) {
		case WS:
			UnSetOffscreenPort(gOffscreenPtr) ;
			UpdateImageScreen() ;
			SetPort(gControlWindow) ;
			break ;
		case LP:
			close_ps_file() ;
			break ;
	}
	off_highlight() ;
	ClearDisplay() ;
	set_menu(VIEW_UNIT_CEL) ;
	displ_menu() ;
	if (ierr == -1) err(4) ;
	if (ierr == -2) err(5) ;
	InitCursor() ;
}

void vu_std_cell(){

    long ierr,i ;
    float uvw[3] ;
  
    eraser(0,0,240,50) ;
    set_menu(UC_PSP_NOPSP) ;
    displ_menu() ;
}

void vu_new_cell() {
    eraser(285,0,525,50) ;
    set_menu(NUC_PSP_NOPSP) ;
    displ_menu() ;
}

//--------------------------------------------------------------------
// Plot the transformed cell with perspective           
//--------------------------------------------------------------------
void nucel_persp() {
    long ierr,i ;
    float uvw[3] ;
    char c ;

    PERSP = TRUE ;
    on_highlight() ;
    gCurrentCommand = gCurrentButton ;
    uvw[0] = 0. ;
    uvw[1] = 0. ;
    uvw[2] = 1. ;
    setup_trnsf() ;
    SetWatchCursor() ;
    switch(OutDevice->type)
	{
	case WS:
    		SetOffscreenPort(gOffscreenPtr) ;
    		ClearDisplay() ;
		break ;
	case FS:
		return ;
		break ;
	case LP:
		open_ps_file("laserprinter.ps") ;
		break ;
	}
    DrawNewUnitCell(uvw,&ierr) ;
    switch(OutDevice->type)
	{
	case WS:
    		UnSetOffscreenPort(gOffscreenPtr) ;
    		UpdateImageScreen() ;
    		SetPort(gControlWindow) ;
		break ;
	case LP:
		close_ps_file() ;
		break ;
	}
    off_highlight() ;
    ClearDisplay() ;
    set_menu(VIEW_UNIT_CEL) ;
    displ_menu() ;
    if (ierr == -1) err(4) ;
    if (ierr == -2) err(5) ;
    InitCursor() ;
}

//--------------------------------------------------------------------
// Plot the transformed cell with perspective     
//--------------------------------------------------------------------
void nucel_nopersp() {
    long ierr,i ;
    float uvw[3] ;
    char c ;

    PERSP = FALSE ;
    on_highlight() ;
    gCurrentCommand = gCurrentButton ;
    uvw[0] = 0. ;
    uvw[1] = 0. ;
    uvw[2] = 1. ;
    setup_trnsf() ;
    SetWatchCursor() ;
    switch(OutDevice->type)
	{
	case WS:
    		SetOffscreenPort(gOffscreenPtr) ;
    		ClearDisplay() ;
		break ;
	case FS:
		return ;
		break ;
	case LP:
		open_ps_file("laserprinter.ps") ;
		break ;
	}
    DrawNewUnitCell(uvw,&ierr) ;
    switch(OutDevice->type)
	{
	case WS:
    		UnSetOffscreenPort(gOffscreenPtr) ;
    		UpdateImageScreen() ;
    		SetPort(gControlWindow) ;
		break ;
	case LP:
		close_ps_file() ;
		break ;
	}
    off_highlight() ;
    ClearDisplay() ;
    set_menu(VIEW_UNIT_CEL) ;
    displ_menu() ;
    if (ierr == -1) err(4) ;
    if (ierr == -2) err(5) ;
    InitCursor() ;
}

// -------------------------------------------------------
// Set up routine to call the  routine ATOM3D to draw 
// shaded atoms on the image display device           
// -------------------------------------------------------
void r_atom3d() {
    float *a,*b,*c,*al,*be,*ga,*xp,*yp,*zp,*dw,*occf ;
    long   *num_atom,*num_z,*type ;

    on_highlight() ;
    gCurrentCommand = gCurrentButton ;
    setup_trnsf()  ;      /* Get the proper assignments. Neccesary
                           * ..because we need a contiguous array to
                           * .. transfer the fractional coordinates
                           * .. in.
                           */
    a    = &gPhsgrtRec->a       ;
    b    = &gPhsgrtRec->b       ;
    c    = &gPhsgrtRec->c       ;
    al   = &gPhsgrtRec->alpha   ;
    be   = &gPhsgrtRec->beta    ;
    ga   = &gPhsgrtRec->gamma   ;
    num_atom = &gNumUnitCellAtoms   ;
    num_z    = (long *)gPhsgrtRec->atno ;
    xp       = gPhsgrtRec->x ;
    yp       = gPhsgrtRec->y ;
    zp       = gPhsgrtRec->z ;
    dw       = gPhsgrtRec->dw ;
    occf     = gPhsgrtRec->occf ;
    type     = (long *)gPhsgrtRec->ntype ;
    SetOffscreenPort(gOffscreenPtr) ;
    //atom3d(a,b,c,al,be,ga,num_atom,xp,yp,zp,num_z,type,dw,occf) ;
    UnSetOffscreenPort(gOffscreenPtr) ;
    SetPort(gControlWindow) ;
    UpdateImageScreen() ;
    off_highlight() ;
}
