// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------
// --------------------------------------------------------------------------
// Perform a complex 2-Dimensional Fourier Transform 
// DATA -- Complex input/output array        
// N	-- Number of columns in the matrix	
// M 	-- Number of rows in the Matrix		
// ISIGN - +1/-1 for FORWARD/INVERSE transform	
// --------------------------------------------------------------------------
#include <math.h>
#include <stdio.h>
#include "funcdefs.h"

#define log2(x)		log((x))/log(2.)
void fft2d(float *data,long n,long m,long isign) {
    long n2,m2,inc ;
    float table[4096] ;
    float pi=3.14159265 ,rn2,rm2;
    register long i,last ;
    register float *rp,ang,dang ;

    rn2 = n ;
    rm2 = m ;
    rn2 = log2(rn2) ;
    rm2 = log2(rm2) ;
    n2 = near(rn2) ;
    m2 = near(rm2) ;

    // Set up the table 
    last = (n-1)/2 ;
    ang = 0. ;
    dang = -isign*2.*pi/n ;
    for(i=0,rp=table;i<=last;i++,ang+=dang) {
		*rp++ = cos(ang) ;
		*rp++ = sin(ang) ;
	}
		
    inc = 1 ;
    for(i=0,rp=data;i<m;i++,rp+=2*n) {
    	binrev(rp,n2,inc) ;
    	fft1d(rp,table,n2,inc) ;
	}
	
    if (m == 1) return ; // 1-Dimensional FFT 
    if (m != n) {		// Set up a new table =
		last = (m-1)/2 ;
		ang = 0. ;
		dang = -isign*2.*pi/m ;
		for(i=0,rp=table;i<=last;i++,ang+=dang) {
			*rp++ = cos(ang) ;
			*rp++ = sin(ang) ;
		}
	}
	
    inc = n ;
    for(i=0,rp=data;i<n;i++,rp+=2) {
		binrev(rp,m2,inc) ;
		fft1d(rp,table,m2,inc) ;
	}
}

void fft1d(float *data,float *table,long m,long inc) {
    float pi=3.14159265,ct[2] ;
    long n,le,nu,ib,newNum ;
    register float *tp,*rp1,*rp2,*w,*rp,ang,dang ;
    register long i,j,l,ia,le1,inc2 ;
	
    n = pow(2.,(float)m) ;
    tp = ct ;
	
    le=2 ;
    inc2 = 2*inc ;
    for(l=0;l<m;l++) {
		le1 = le/2 ;
		nu = 0 ;
		newNum = n/le ;
		for(j=0;j<le1;j++) {
			w = table + nu + nu ;
			for(ia=j;ia<n;ia += le) {
				if(inc2==2) {
					rp1 = data+ia+ia ;
					rp2 = rp1 + le1 + le1 ;
				}
				else {
					rp1 = data + ia*inc2 ;
					rp2 = rp1 + le1*inc2 ;
				}
				*tp      = *w*(*rp2) - *(w+1)*(*(rp2+1)) ;	// T = W*Y 
				*(tp+1)  = *w*(*(rp2+1)) + *(w+1)*(*rp2) ;	//  "   "  
				*rp2     = *rp1 - *tp ;			  			// Y = X-T  
				*(rp2+1) = *(rp1+1) - *(tp+1) ;		  		//  "   "   
				*rp1     += *tp ;			  				// X = X+T  
				*(rp1+1) += *(tp+1) ;			  			//  "   "   
			}
			nu += newNum ;
		}
		le += le ;
	}
}

void binrev(float *c,long m,long inc) {
    float cmplx[2] ;
    long last,nh,n ;
    register long i,j,k,inc2 ;
    register float *rp1,*rp2,*ct ;
	
    n = pow(2.,(float)m) ;
    k = (m+1)/2 ;
    nh = n/2 ;
    last = n - 1 - pow(2.,(float)k) ;
    j = nh ;
    ct = cmplx ;
    inc2 = 2*inc ;
    for(i=1;i<=last;i++) {
		if (i < j) {
			if (inc==2) {
				rp1 = c + i + i ;
				rp2 = c + j + j ;
			}
			else {
				rp1 = c + i*inc2 ;
				rp2 = c + j*inc2 ;
			}
			*ct        = *rp1 ;
			*(ct+1)    = *(rp1+1) ;
			*rp1       = *rp2 ;
			*(rp1+1)   = *(rp2+1) ;
			*rp2       = *ct ;
			*(rp2+1)   = *(ct+1) ;
		}
		k = nh ;
		while (j>=k) {
			j -= k ;
			k /= 2 ;
		}
		j += k ;
	}
}	
