// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------
#include <sys/types.h>
#ifdef SUN
#include <sys/timeb.h>
#else
#include <time.h>
#endif
#include "funcdefs.h"
/* ---------------------------------------------------	*/
/* These are just set-up routines for calling the FFT	*/
/* on the data. Code could easily be installed to take	*/
/* advantage of Array-Processors or other fast hardware	*/
/* for doing Fourier Transforms.			*/
/* ----------------------------------------------------	*/
void mmfft(float *data,long n,long m,float *rtime) {
  /* -------------------------	*/
  /* Forward Fourier Transform	*/
  /* --------------------------	*/
    long it0,it1,itime ;
#ifdef SUN
    struct timeb tp ; 
#else
    time_t tloc ;
#endif
	
#ifdef SUN
    ftime(&tp) ;
    it0 = 1000*tp.time+tp.millitm ;
#else
    it0 = (long)time(&tloc) ;
#endif
    fft2d(data,n,m,1) ;
#ifdef SUN
    ftime(&tp) ;
    it1 = 1000*tp.time+tp.millitm ;
#else
    it1 = (long)time(&tloc) ;
#endif
    itime = it1-it0 ;
#ifdef SUN
    *rtime = itime/1000. ;
#else
    *rtime = itime ;
#endif
}

void mmffti(float *data,long n,long m,float *rtime) {
  /* --------------------------	*/
  /* Inverse Fourier Transform	*/
  /* --------------------------	*/
    long it0,it1,itime ;
#ifdef SUN
    struct timeb tp ; 
#else
    time_t tloc ;
#endif
	
#ifdef SUN
    ftime(&tp) ;
    it0 = 1000*tp.time+tp.millitm ;
#else
    it0 = (long)time(&tloc) ;
#endif
    fft2d(data,n,m,-1) ;
#ifdef SUN
    ftime(&tp) ;
    it1 = 1000*tp.time+tp.millitm ;
#else
    it1 = (long)time(&tloc) ;
#endif
    itime = it1-it0 ;
#ifdef SUN
    *rtime = itime/1000. ;
#else
    *rtime = itime ;
#endif
}	
