//-----------------------------------------------------------------------------
//	Copyright (c) 2001 Roar	Kilaas
//
//	This program is	free software; you can redistribute	it and/or modify
//	it under the terms of the GNU General Public License as	published by
//	the	Free Software Foundation; either version 2 of the License, or
//	(at	your option) any later version.
//
//	This program is	distributed	in the hope	that it	will be	useful,
//	but	WITHOUT	ANY	WARRANTY; without even the implied warranty	of
//	MERCHANTABILITY	or FITNESS FOR A PARTICULAR	PURPOSE.  See the
//	GNU	General	Public License for more	details.
//
//	You	should have	received a copy	of the GNU General Public License
//	along with this	program; if	not, write to the Free Software
//	Foundation,	Inc., 59 Temple	Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------

#include <stdio.h>

#include "TLattice.h"
#include "TImage.h"
#include "TPeakList.h"
#include "Plateau.h"
#include "XDrivers.h"
#include "TMatrix.h"

		 // This should probably be somewhere else, set to 254.99 or some other
		 // value for not byte images. Caveat: it might be that a byte image has
		 // been saved as float......
		 // Probably better, it should be a property of an image...
		 float PeakIsSaturated = 254.1 ;
		 // Peak off tolerance squared -- settable ?
		 float PeakOffTol = 16.1 ;
		 // Peak too low for XCF motif, settable ?
		 float PeakTooLow = 50.0 ;
		 // Peak too low to bother with
		 float PeakMuchTooLow = 0.05 ;
		 // Global fit parameters
		 float Bfita, Bfitb, Bfitc ;
		 // Global intensity sum of motif, to scale integration methods correctly
		 float MotifSum = 1 ;
		 // Fraction of the distance between peaks for motif
		 float Psizer = 0.75 ;
		 // Maximum number of Biweight iterations
		 long NumbBiwt = 20 ;
		 // To include a peak if it is too far off, but peak is not that different
		 float HowDifferent = 1.5 ;
		 // contraction for inner radius used in masks etc
		 float RadSize = 0.75 ;
		 // Biweight parameters
		float Btune=3.5 ; // Censor value 
		float Btune2=1./(Btune*Btune) ;
		float BiweightTol=2e-4 ;
        // ZeroBaseLevel for UseZero mode
        float ZeroBase=1.0 ;
		bool  useZero ;
		bool  QuadraticFit = false ;
		bool  CubicFit = false ;
		bool  DiffDebug = false ;
		bool  UseBiweight = true ;
        // Maximum Aperture Size for peaks ;
        float RadTot = 10000 ;
		// Maximum allowable fraction of peaks purged by biweights
		float PixPurge = 0.8 ;

extern	short	gDebugMode ;
Boolean	PtInRect(FPoint	*fPoint,Rect *rect)	;
void	PutMessage(char	*) ;
void	SetQuadMode() ;
float	ReturnOffTol() ;
void	SetOffTol(float *OffTol) ;
bool	ReturnQuadFit() ;
bool	ReturnCubicFit() ;

TPeakArray *FindPeaks(TImage *image,PeakCriteraRecord *peakCriteria,Rect *selection	) ;
float SumNeighbours(Pixel *pos,int rad,float *data,long	width,long height) ;
void SetExtremePeak(Pixel *pos,int rad,float *peak,float *image,long width,long	height,
	Boolean extremeIsMax, Pixel *newpos) ;
void SetCMPixel(Pixel *pos,int rad,float *data,long	width,long height, Pixel *newpos) ;
void SetToMaxofNeighbourhood(long width, float *datain,	float *dataout)	;
void SetToMaxofLowerNeighbourhood(long width, float	*datain, float *dataout) ;
void SetToMaxofUpperNeighbourhood(long width, float	*datain, float *dataout);
void SetToMaxofRightNeighbourhood(long width, float	*datain, float *dataout) ;
void SetToMaxofLeftNeighbourhood(long width, float *datain,	float *dataout)	;
void SetToMaxofUpperLeftNeighbourhood(long width, float	*datain, float *dataout) ;
void SetToMaxofUpperRightNeighbourhood(long	width, float *datain, float	*dataout) ;
void SetToMaxofLowerRightNeighbourhood(long	width, float *datain, float	*dataout) ;
void SetToMaxofLowerLeftNeighbourhood(long width, float	*datain, float *dataout) ;
void SetToMinofNeighbourhood(long width, float *datain,	float *dataout)	;
void SetToMinofLowerNeighbourhood(long width, float	*datain, float *dataout) ;
void SetToMinofUpperNeighbourhood(long width, float	*datain, float *dataout);
void SetToMinofRightNeighbourhood(long width, float	*datain, float *dataout) ;
void SetToMinofLeftNeighbourhood(long width, float *datain,	float *dataout)	;
void SetToMinofUpperLeftNeighbourhood(long width, float	*datain, float *dataout) ;
void SetToMinofUpperRightNeighbourhood(long	width, float *datain, float	*dataout) ;
void SetToMinofLowerRightNeighbourhood(long	width, float *datain, float	*dataout) ;
void SetToMinofLowerLeftNeighbourhood(long width, float	*datain, float *dataout) ;
void CreateMaxOfImage(float	*WorkImageData,float *PeakImageData,long Width,long	Height)	;
void CreateMinOfImage(float	*WorkImageData,float *PeakImageData,long Width,long	Height)	;
long CreateUnrefinedBinaryImage(float *WorkImageData,float *PeakImageData,
								long Width,long Height,float lowThr,float hiThr) ;
float MinOfImage(float *Data,long Width,long Height) ;
float MaxOfImage(float *Data,long Width,long Height) ;
Boolean	IsPlateau(Point	*pos,float *p2,long	Width,long Height) ;
void SetCMPlateau(Point	*pos,float *p2,long	Width,long Height) ;
void CreateMinOf1DImage(float *WorkImageData,float *PeakImageData,long Width) ;
void CreateMaxOf1DImage(float *WorkImageData,float *PeakImageData,long Width) ;
void FindCenterOfMass(TImage *image,float *xcm,float *ycm,float	cmRad) ;

float	FindFTLatticePeak(TImage *image,FPoint *peakPos) ;
float	FindTEMPeak(TImage *image,Rect *areaToSearch,FPoint	*peakPos) ;
TPeakList *FindTEMPeaksAlongLattice(TImage *image,TLattice *lattice) ;
TPeakList *FindDiffractionSpotsAtLatticePoints(TImage *image,TLattice *lattice)	;
//TPeakList *FindDiffractionSpotsAtLatticePoints(TImage *image,TLattice *lattice,TImage *peakMotif,
//							float ccThreshold,float Rfactor,float Mean)	;
TPeakList *FindDiffractionSpotsAtLatticePoints(TImage *image,TLattice *lattice,TImage *peakMotif,
							float ccThreshold,float *BX, float *BY)	;
TPeakList *FindDiffractionSpotsAtLatticePointsC(TImage *image,TLattice *lattice,TImage *peakMotif,
							float ccThreshold,float *BX, float *BY)	;

TPeak *FindDiffractionSpotAtLatticePoint(Point &mousePt,TImage *image,TLattice *lattice,TImage *peakMotif) ;
TPeak *TestDiffractionSpotAtLatticePoint(Point &mousePt,TImage *image,TLattice *lattice,TImage *peakMotif) ;

TImage *FindAveragePeakMotif(TImage	*image,TLattice	*lattice,TPeakArray	*PeakArray,short peakRadius	= 0) ;
TImage *FindAveragePeakMotifNoIter(TImage	*image,TLattice	*lattice,TPeakArray	*PeakArray,short peakRadius	= 0) ;

void FitGaussianPeak(TArray	*array,int ixc,int iyc,int irad,
		float &xc,float &yc,float &peakValue) ;
void FitSincPeak(TArray	*array,int ixc,int iyc,int irad,
		float &xc,float &yc,float &peakValue) ;
void FindCMPeak(TArray *array,int ixc,int iyc,int irad,
		float &xc,float &yc,float &peakValue) ;
void FindCMPeakWrapped(TArray *array,int ixc,int iyc,int irad,
		float &xc,float &yc,float &peakValue) ;		
bool	FitAndSubtractPeakBackground(TImage	*image,TImage *weightImage)	;
//LDM
void	FitAndSubtractPeakBackZero(TImage *image,TImage	*weightImage) ;
void	FitAndSubtractPeakBackZeroMask(TImage *image,TImage	*weightImage) ;
void	FitAndSubtractPeakBackMask(TImage *image,TImage	*weightImage) ;
void    FitBiweightBackground(TImage *image,TImage	*weightImage) ;
void	SetMotifPars (TObject *) ;
extern	void	WriteObjectInfo(char *text) ;
extern float	RealXYValueQ( int h, int k, float BX[13] ) ;
extern float	RealXYValueC(int h, int k, float BX[13] ) ;
extern float	gDiffractionAperture ;

extern	long near(float) ;
TPeak	*FindLatticePeak(TImage	*image,Point &pos,TImage *peakMotif,TImage *weightImage,TLattice *lattice) ;
TPeak	*TestLatticePeak(TImage	*image,Point &pos,TImage *peakMotif,TImage *weightImage,TLattice *lattice) ;
float	FindPeak(TImage	*image,Rect	*areaToSearch,FPoint *peakPos,
		Boolean	threshold=true) ;
TPeakList *FindPeaksAlongLattice(TImage	*image,TLattice	*lattice,
		Boolean	threshold=true) ;
TPeakList *FindPeaksAtLatticePoints(TImage *image,TLattice *lattice,
		Boolean	threshold=true) ;
void	ExtractAmplitudePhase(TArray *imageArray,long &n,long &m,float &x,float	&y)	;
float	FindFTLatticePeak(TImage *image,FPoint *peakPos) ;
float	FindLatticePeak(TImage *image,FPoint *peakPos,float	halfPeakSpacing,PeakCriteria criteria =	kThresholdOnMeanAroundSpot)	;
float	FindLatticePeak(TImage *image,FPoint *peakPos,TImage *peakMotif,TImage *weightImage,float ccThreshold,float Rfactor, float Mean) ;

bool Solve3LinearEquation(TMatrix *A,TVector *B,TVector	*C)	;

#define	MIN(x,y) ((x) <	(y)	? (x) :	(y))
#define	MAX(x,y) ((x) >	(y)	? (x) :	(y))
#define MAXZ(x)  ((x) > (0) ? (x) : (0) )

PeakForm gPeakForm = kGaussianPeak ;
#include "TDialog.h"
#include "PeakMotifDialog.h"

void	SetMotifPars(TObject *obj) {
		//fprintf(stderr,"Went to 1\n");
		PeakMotifDialog *theMotifDialog = new PeakMotifDialog() ;
		theMotifDialog->SetDialogData() ;
		if (theMotifDialog->DoModalDialog()) 
			theMotifDialog->GetDialogData() ;

		delete theMotifDialog ; 
	}

Boolean	PtInRect(FPoint	*fPoint,Rect *rect)	{
	Point 	iPoint ;
	iPoint.x = near(fPoint->x) ;
	iPoint.y = near(fPoint->y) ;
	return PtInRect(&iPoint,rect) ;
} 

TPeakList *FindPeaksAlongLattice(TImage	*image,TLattice	*lattice,
		Boolean treshold) {
	Rect 		searchRect,imageRect ;
	FPoint		peakPos,peakJ0Pos,U,V,O ;
	Point		latticePoint ;
	register	int i,j ;
	long		w2,h2 ;
	float		peakValue,radsq,aprtsq,aperture ;
	TPeakList	*peakList = new TPeakList ;
	TPeak		thePeak ;
	short		sW = 2 ; 
	
	SetRect(&imageRect,0,0,image->Width(),image->Height()) ;

	// Start at the origin of the lattice and find lattice points 
	// in a recursive fashion
	O = lattice->GetOrigin() ;
	U = lattice->GetU() ;
	V = lattice->GetV() ;
	latticePoint.x = near(O.x) ;
	latticePoint.y = near(O.y) ;

	// Ensure that we do a symmetric region about the origin
	w2 = image->Width()/2 ;
	h2 = image->Height()/2 ;
	aperture = w2 < h2 ? w2 : h2 ;
	// Smallest distance along x
	float HowFarX = 2*w2-O.x > O.x ? O.x : 2*w2-O.x ;
	// Smallest distance along y
	float HowFarY = 2*h2-O.y > O.y ? O.y : 2*h2-O.y ;
	// Smallest of these and an aperture given elsewhere
	if (aperture > HowFarX) aperture = HowFarX ;
	if (aperture > HowFarY) aperture = HowFarY ;
	if (aperture > RadTot)  aperture = RadTot  ;
	if (aperture > gDiffractionAperture) aperture = gDiffractionAperture ;
	aprtsq = aperture*aperture ;
	//fprintf(stderr,"Aperture stuff %f %f \n",aperture,gDiffractionAperture);

	// Check to make sure local peak search distance is smaller than 1/2 lattice spacing
	float	du = 0.5*sqrt(U.x*U.x+U.y*U.y) ;
	float	dv = 0.5*sqrt(V.x*V.x+V.y*V.y) ;
	float	halfSpacing = du < dv ? du : dv ;
	if (sW > near(halfSpacing)) sW = near(halfSpacing) ;

	// Positive i
	
	i = 0 ;
	SetRect(&searchRect,latticePoint.x-sW,latticePoint.y-sW,latticePoint.x+sW,latticePoint.y+sW);
	while(PtInRect(&latticePoint,&imageRect)) {
		j = 0 ;
		peakPos.x = latticePoint.x ; peakPos.y = latticePoint.y ;
		radsq = (latticePoint.x-O.x)*(latticePoint.x-O.x)+
			(latticePoint.y-O.y)*(latticePoint.y-O.y) ;
		if (radsq < aprtsq) {
			peakValue = FindPeak(image,&searchRect,&peakPos,treshold) ;
		}
		else {
			peakValue = 0. ;
		}
		if (peakValue > 0.) {
			// Add this peak
			thePeak.value = peakValue ;
			thePeak.pos = peakPos ;
			thePeak.n = i ;
			thePeak.m = j ;
			peakList->AddPeak(thePeak) ;
		}
		peakJ0Pos = peakPos ;
		// Loop over positive j
		// Start with j = 1
		j = 1 ;
		latticePoint.x = near(peakPos.x + V.x) ;
		latticePoint.y = near(peakPos.y + V.y) ;
		SetRect(&searchRect,latticePoint.x-sW,latticePoint.y-sW,
				latticePoint.x+sW,latticePoint.y+sW);
		while(PtInRect(&latticePoint,&imageRect)) {
			radsq = (latticePoint.x-O.x)*(latticePoint.x-O.x)+
				(latticePoint.y-O.y)*(latticePoint.y-O.y) ;
			if (radsq < aprtsq) {
				peakValue = FindPeak(image,&searchRect,&peakPos,treshold) ;
			}
			else {
				peakValue = 0. ;
			}
			if (peakValue > 0.) {
				// Add this peak
				thePeak.value = peakValue ;
				thePeak.pos = peakPos ;
				thePeak.n = i ;
				thePeak.m = j ;
				peakList->AddPeak(thePeak) ;
				latticePoint.x = near(peakPos.x + V.x) ;
				latticePoint.y = near(peakPos.y + V.y) ;
			}
			else {
				// Did not find a peak, but advance anyway
				latticePoint.x = near(latticePoint.x + V.x) ;
				latticePoint.y = near(latticePoint.y + V.y) ;
			}
			SetRect(&searchRect,latticePoint.x-sW,latticePoint.y-sW,
				latticePoint.x+sW,latticePoint.y+sW);
			j++ ;
		}
		// Reset to j = 0 
		peakPos = peakJ0Pos ;
		// Loop over negative j's
		// Start with j = -1
		j = -1 ;
		latticePoint.x = near(peakPos.x - V.x) ;
		latticePoint.y = near(peakPos.y - V.y) ;
		SetRect(&searchRect,latticePoint.x-sW,latticePoint.y-sW,
				latticePoint.x+sW,latticePoint.y+sW);
		while(PtInRect(&latticePoint,&imageRect)) {
			radsq = (latticePoint.x-O.x)*(latticePoint.x-O.x)+
				(latticePoint.y-O.y)*(latticePoint.y-O.y) ;
			if (radsq < aprtsq) {
				peakValue = FindPeak(image,&searchRect,&peakPos,treshold) ;
			}
			else {
				peakValue = 0. ;
			}
			if (peakValue > 0.) {
				// Add this peak
				thePeak.value = peakValue ;
				thePeak.n = i ;
				thePeak.m = j ;
				thePeak.pos = peakPos ;
				peakList->AddPeak(thePeak) ;
				latticePoint.x = near(peakPos.x - V.x) ;
				latticePoint.y = near(peakPos.y - V.y) ;
			}
			else {
				// Did not find a peak, but advance anyway
				latticePoint.x = near(latticePoint.x - V.x) ;
				latticePoint.y = near(latticePoint.y - V.y) ;
			}
			SetRect(&searchRect,latticePoint.x-sW,latticePoint.y-sW,
				latticePoint.x+sW,latticePoint.y+sW);
			j-- ;
		}
		peakPos = peakJ0Pos ;
		// go to next i
		latticePoint.x = near(peakPos.x + U.x) ;
		latticePoint.y = near(peakPos.y + U.y) ;
		SetRect(&searchRect,latticePoint.x-sW,latticePoint.y-sW,
				latticePoint.x+sW,latticePoint.y+sW);
		i++ ;
	}

	// negative i, starting with j = 0
	i = -1 ;
	j = 0 ;
	latticePoint.x = near(O.x-U.x) ;
	latticePoint.y = near(O.y-U.y) ;
	SetRect(&searchRect,latticePoint.x-sW,latticePoint.y-sW,latticePoint.x+sW,latticePoint.y+sW);
	while (PtInRect(&latticePoint,&imageRect)) {
		j = 0 ;
		peakPos.x = latticePoint.x ; peakPos.y = latticePoint.y ;
		radsq = (latticePoint.x-O.x)*(latticePoint.x-O.x)+
				(latticePoint.y-O.y)*(latticePoint.y-O.y) ;
		if (radsq < aprtsq) {
				peakValue = FindPeak(image,&searchRect,&peakPos,treshold) ;
		}
		else {
				peakValue = 0. ;
		}
		if (peakValue > 0.) {
			// Add this peak
			thePeak.value = peakValue ;
			thePeak.pos = peakPos ;
			thePeak.n = i ;
			thePeak.m = j ;
			peakList->AddPeak(thePeak) ;
		}

		peakJ0Pos = peakPos ;
		// Loop over positive j
		// Start with j = 1
		j = 1 ;
		latticePoint.x = near(peakPos.x + V.x) ;
		latticePoint.y = near(peakPos.y + V.y) ;
		SetRect(&searchRect,latticePoint.x-sW,latticePoint.y-sW,
				latticePoint.x+sW,latticePoint.y+sW);
		while(PtInRect(&latticePoint,&imageRect)) {
			radsq = (latticePoint.x-O.x)*(latticePoint.x-O.x)+
				(latticePoint.y-O.y)*(latticePoint.y-O.y) ;
			if (radsq < aprtsq) {
				peakValue = FindPeak(image,&searchRect,&peakPos,treshold) ;
			}
			else {
				peakValue = 0. ;
			}
			if (peakValue > 0.) {
				// Add this peak
				thePeak.value = peakValue ;
				thePeak.pos = peakPos ;
				thePeak.n = i ;
				thePeak.m = j ;
				peakList->AddPeak(thePeak) ;
				latticePoint.x = near(peakPos.x + V.x) ;
				latticePoint.y = near(peakPos.y + V.y) ;
			}
			else {
				// Did not find a peak, but advance anyway
				latticePoint.x = near(latticePoint.x + V.x) ;
				latticePoint.y = near(latticePoint.y + V.y) ;
			}
			SetRect(&searchRect,latticePoint.x-sW,latticePoint.y-sW,
				latticePoint.x+sW,latticePoint.y+sW);
			j++ ;
		}
		// Reset to j = 0 
		peakPos = peakJ0Pos ;
		// Loop over negative j's
		// Start with j = -1
		j = -1 ;
		latticePoint.x = near(peakPos.x - V.x) ;
		latticePoint.y = near(peakPos.y - V.y) ;
		SetRect(&searchRect,latticePoint.x-sW,latticePoint.y-sW,
				latticePoint.x+sW,latticePoint.y+sW);
		while(PtInRect(&latticePoint,&imageRect)) {
			radsq = (latticePoint.x-O.x)*(latticePoint.x-O.x)+
				(latticePoint.y-O.y)*(latticePoint.y-O.y) ;
			if (radsq < aprtsq) {
				peakValue = FindPeak(image,&searchRect,&peakPos,treshold) ;
			}
			else {
				peakValue = 0. ;
			}
			if (peakValue > 0.) {
				// Add this peak
				thePeak.value = peakValue ;
				thePeak.pos = peakPos ;
				thePeak.n = i ;
				thePeak.m = j ;
				peakList->AddPeak(thePeak) ;
				latticePoint.x = near(peakPos.x - V.x) ;
				latticePoint.y = near(peakPos.y - V.y) ;
			}
			else {
				// Did not find a peak, but advance anyway
				latticePoint.x = near(latticePoint.x - V.x) ;
				latticePoint.y = near(latticePoint.y - V.y) ;
			}
			SetRect(&searchRect,latticePoint.x-sW,latticePoint.y-sW,
				latticePoint.x+sW,latticePoint.y+sW);
			j-- ;
		}
		peakPos = peakJ0Pos ;
		// go to next i
		latticePoint.x = near(peakPos.x - U.x) ;
		latticePoint.y = near(peakPos.y - U.y) ;
		SetRect(&searchRect,latticePoint.x-sW,latticePoint.y-sW,
			latticePoint.x+sW,latticePoint.y+sW);
		i-- ;
	}
		
	if (peakList->GetNumberOfPeaks() == 0) {
		delete peakList ;
		return nil ;
	}
	return peakList ;
}

TPeakList *FindTEMPeaksAlongLattice(TImage *image,TLattice *lattice) {
	Rect 		searchRect,imageRect ;
	FPoint		peakPos,peakJ0Pos,U,V,O ;
	Point		latticePoint ;
	register	int i,j ;
	long		w2,h2 ;
	float		peakValue,radsq,aprtsq ;
	TPeakList	*peakList = new TPeakList ;
	TPeak		thePeak ;
	short		sW = 4 ; 	// This is the pixel distance to be used for background

	w2 = image->Width()/2 ;
	h2 = image->Height()/2 ;

	aprtsq = gDiffractionAperture*gDiffractionAperture ;

	SetRect(&imageRect,0,0,image->Width(),image->Height()) ;

	// Start at the origin of the lattice and find lattice points 
	// in a recursive fashion
	O = lattice->GetOrigin() ;
	U = lattice->GetU() ;
	V = lattice->GetV() ;
	latticePoint.x = near(O.x) ;
	latticePoint.y = near(O.y) ;


	// Check to make sure local peak search distance is smaller than 1/2 lattice spacing
	float	du = 0.5*sqrt(U.x*U.x+U.y*U.y) ;
	float	dv = 0.5*sqrt(V.x*V.x+V.y*V.y) ;
	float	halfSpacing = du < dv ? du : dv ;
	if (sW > near(halfSpacing)) sW = near(halfSpacing) ;

	
	// Positive i	
	i = 0 ;
	j = 0 ;
	SetRect(&searchRect,latticePoint.x-sW,latticePoint.y-sW,latticePoint.x+sW,latticePoint.y+sW);
	while(PtInRect(&latticePoint,&imageRect)) {
		peakPos.x = latticePoint.x ; peakPos.y = latticePoint.y ;
		radsq = (latticePoint.x-O.x)*(latticePoint.x-O.x)+
			(latticePoint.y-O.y)*(latticePoint.y-O.y) ;
		if (radsq < aprtsq) {
			peakValue = FindTEMPeak(image,&searchRect,&peakPos) ;
		}
		else {
			peakValue = 0. ;
		}
		if (peakValue > 0.) {
			// Add this peak
			thePeak.value = peakValue ;
			thePeak.pos = peakPos ;
			thePeak.n = i ;
			thePeak.m = j ;
			peakList->AddPeak(thePeak) ;
		}
		peakJ0Pos = peakPos ;
		// Loop over positive j
		// Start with j = 1
		j = 1 ;
		latticePoint.x = near(peakPos.x + V.x) ;
		latticePoint.y = near(peakPos.y + V.y) ;
		SetRect(&searchRect,latticePoint.x-sW,latticePoint.y-sW,
				latticePoint.x+sW,latticePoint.y+sW);
		while(PtInRect(&latticePoint,&imageRect)) {
			radsq = (latticePoint.x-O.x)*(latticePoint.x-O.x)+
				(latticePoint.y-O.y)*(latticePoint.y-O.y) ;
			if (radsq < aprtsq) {
				peakValue = FindTEMPeak(image,&searchRect,&peakPos) ;
			}
			else {
				peakValue = 0. ;
			}
			if (peakValue > 0.) {
				// Add this peak
				thePeak.value = peakValue ;
				thePeak.pos = peakPos ;
				thePeak.n = i ;
				thePeak.m = j ;
				peakList->AddPeak(thePeak) ;
				latticePoint.x = near(peakPos.x + V.x) ;
				latticePoint.y = near(peakPos.y + V.y) ;
			}
			else {
				// Did not find a peak, but advance anyway
				latticePoint.x = near(latticePoint.x + V.x) ;
				latticePoint.y = near(latticePoint.y + V.y) ;
			}
			SetRect(&searchRect,latticePoint.x-sW,latticePoint.y-sW,
				latticePoint.x+sW,latticePoint.y+sW);
			j++ ;
		}
		// Reset to j = 0 
		peakPos = peakJ0Pos ;
		// Loop over negative j's
		// Start with j = -1
		j = -1 ;
		latticePoint.x = near(peakPos.x - V.x) ;
		latticePoint.y = near(peakPos.y - V.y) ;
		SetRect(&searchRect,latticePoint.x-sW,latticePoint.y-sW,
				latticePoint.x+sW,latticePoint.y+sW);
		while(PtInRect(&latticePoint,&imageRect)) {
			radsq = (latticePoint.x-O.x)*(latticePoint.x-O.x)+
				(latticePoint.y-O.y)*(latticePoint.y-O.y) ;
			if (radsq < aprtsq) {
				peakValue = FindTEMPeak(image,&searchRect,&peakPos) ;
			}
			else {
				peakValue = 0. ;
			}
			if (peakValue > 0.) {
				// Add this peak
				thePeak.value = peakValue ;
				thePeak.n = i ;
				thePeak.m = j ;
				thePeak.pos = peakPos ;
				peakList->AddPeak(thePeak) ;
				latticePoint.x = near(peakPos.x - V.x) ;
				latticePoint.y = near(peakPos.y - V.y) ;
			}
			else {
				// Did not find a peak, but advance anyway
				latticePoint.x = near(latticePoint.x - V.x) ;
				latticePoint.y = near(latticePoint.y - V.y) ;
			}
			SetRect(&searchRect,latticePoint.x-sW,latticePoint.y-sW,
				latticePoint.x+sW,latticePoint.y+sW);
			j-- ;
		}
		peakPos = peakJ0Pos ;
		// go to next i
		latticePoint.x = near(peakPos.x + U.x) ;
		latticePoint.y = near(peakPos.y + U.y) ;
		SetRect(&searchRect,latticePoint.x-sW,latticePoint.y-sW,
				latticePoint.x+sW,latticePoint.y+sW);
		i++ ;
		j = 0 ;
	}

	// negative i, starting with j = 0
	i = -1 ;
	j = 0 ;
	latticePoint.x = near(O.x-U.x) ;
	latticePoint.y = near(O.y-U.y) ;
	SetRect(&searchRect,latticePoint.x-sW,latticePoint.y-sW,latticePoint.x+sW,latticePoint.y+sW);
	while (PtInRect(&latticePoint,&imageRect)) {
		peakPos.x = latticePoint.x ; peakPos.y = latticePoint.y ;
		radsq = (latticePoint.x-O.x)*(latticePoint.x-O.x)+
				(latticePoint.y-O.y)*(latticePoint.y-O.y) ;
		if (radsq < aprtsq) {
				peakValue = FindTEMPeak(image,&searchRect,&peakPos) ;
		}
		else {
				peakValue = 0. ;
		}
		if (peakValue > 0.) {
			// Add this peak
			thePeak.value = peakValue ;
			thePeak.pos = peakPos ;
			thePeak.n = i ;
			thePeak.m = j ;
			peakList->AddPeak(thePeak) ;
		}

		peakJ0Pos = peakPos ;
		// Loop over positive j
		// Start with j = 1
		j = 1 ;
		latticePoint.x = near(peakPos.x + V.x) ;
		latticePoint.y = near(peakPos.y + V.y) ;
		SetRect(&searchRect,latticePoint.x-sW,latticePoint.y-sW,
				latticePoint.x+sW,latticePoint.y+sW);
		while(PtInRect(&latticePoint,&imageRect)) {
			radsq = (latticePoint.x-O.x)*(latticePoint.x-O.x)+
				(latticePoint.y-O.y)*(latticePoint.y-O.y) ;
			if (radsq < aprtsq) {
				peakValue = FindTEMPeak(image,&searchRect,&peakPos) ;
			}
			else {
				peakValue = 0. ;
			}
			if (peakValue > 0.) {
				// Add this peak
				thePeak.value = peakValue ;
				thePeak.pos = peakPos ;
				thePeak.n = i ;
				thePeak.m = j ;
				peakList->AddPeak(thePeak) ;
				latticePoint.x = near(peakPos.x + V.x) ;
				latticePoint.y = near(peakPos.y + V.y) ;
			}
			else {
				// Did not find a peak, but advance anyway
				latticePoint.x = near(latticePoint.x + V.x) ;
				latticePoint.y = near(latticePoint.y + V.y) ;
			}
			SetRect(&searchRect,latticePoint.x-sW,latticePoint.y-sW,
				latticePoint.x+sW,latticePoint.y+sW);
			j++ ;
		}
		// Reset to j = 0 
		peakPos = peakJ0Pos ;
		// Loop over negative j's
		// Start with j = -1
		j = -1 ;
		latticePoint.x = near(peakPos.x - V.x) ;
		latticePoint.y = near(peakPos.y - V.y) ;
		SetRect(&searchRect,latticePoint.x-sW,latticePoint.y-sW,
				latticePoint.x+sW,latticePoint.y+sW);
		while(PtInRect(&latticePoint,&imageRect)) {
			radsq = (latticePoint.x-O.x)*(latticePoint.x-O.x)+
				(latticePoint.y-O.y)*(latticePoint.y-O.y) ;
			if (radsq < aprtsq) {
				peakValue = FindTEMPeak(image,&searchRect,&peakPos) ;
			}
			else {
				peakValue = 0. ;
			}
			if (peakValue > 0.) {
				// Add this peak
				thePeak.value = peakValue ;
				thePeak.pos = peakPos ;
				thePeak.n = i ;
				thePeak.m = j ;
				peakList->AddPeak(thePeak) ;
				latticePoint.x = near(peakPos.x - V.x) ;
				latticePoint.y = near(peakPos.y - V.y) ;
			}
			else {
				// Did not find a peak, but advance anyway
				latticePoint.x = near(latticePoint.x - V.x) ;
				latticePoint.y = near(latticePoint.y - V.y) ;
			}
			SetRect(&searchRect,latticePoint.x-sW,latticePoint.y-sW,
				latticePoint.x+sW,latticePoint.y+sW);
			j-- ;
		}
		peakPos = peakJ0Pos ;
		// go to next i
		latticePoint.x = near(peakPos.x - U.x) ;
		latticePoint.y = near(peakPos.y - U.y) ;
		SetRect(&searchRect,latticePoint.x-sW,latticePoint.y-sW,
			latticePoint.x+sW,latticePoint.y+sW);
		i-- ;
		j = 0 ;
	}

		
	if (peakList->GetNumberOfPeaks() == 0) {
		delete peakList ;
		return nil ;
	}
	return peakList ;
}

TPeakList *FindTEMPeaksAtLattice(TImage	*image,TLattice	*lattice) {
	Rect 		searchRect,imageRect ;
	FPoint		peakPos,peakJ0Pos,U,V,O ;
	FPoint		latticePoint ;
	register	int i,j ;
	long		w2,h2 ;
	float		peakValue,radsq,aprtsq ;
	TPeakList	*peakList = new TPeakList ;
	TPeak		thePeak ;
	short		sW = 1 ; 	

	w2 = image->Width()/2 ;
	h2 = image->Height()/2 ;

	aprtsq = gDiffractionAperture*gDiffractionAperture ;

	SetRect(&imageRect,0,0,image->Width(),image->Height()) ;

	// Start at the origin of the lattice and find lattice points 
	// in a recursive fashion
	O = lattice->GetOrigin() ;
	U = lattice->GetU() ;
	V = lattice->GetV() ;
	latticePoint.x = O.x ;
	latticePoint.y = O.y ;
	
	// Positive i	
	i = 0 ;
	j = 0 ;
	SetRect(&searchRect,near(latticePoint.x-sW),near(latticePoint.y-sW),near(latticePoint.x+sW),near(latticePoint.y+sW));
	while(PtInRect(&latticePoint,&imageRect)) {
		peakPos.x = latticePoint.x ; peakPos.y = latticePoint.y ;
		radsq = (latticePoint.x-O.x)*(latticePoint.x-O.x)+
			(latticePoint.y-O.y)*(latticePoint.y-O.y) ;
		if (radsq < aprtsq) {
			peakValue = FindTEMPeak(image,&searchRect,&peakPos) ;
		}
		else {
			peakValue = 0. ;
		}
		if (peakValue > 0.) {
			// Add this peak
			thePeak.value = peakValue ;
			thePeak.pos = peakPos ;
			thePeak.n = i ;
			thePeak.m = j ;
			peakList->AddPeak(thePeak) ;
		}
		peakJ0Pos = peakPos ;
		// Loop over positive j
		// Start with j = 1
		j = 1 ;
		latticePoint.x +=  V.x ;
		latticePoint.y +=  V.y ;
		SetRect(&searchRect,near(latticePoint.x-sW),near(latticePoint.y-sW),
				near(latticePoint.x+sW),near(latticePoint.y+sW));
		while(PtInRect(&latticePoint,&imageRect)) {
			radsq = (latticePoint.x-O.x)*(latticePoint.x-O.x)+
				(latticePoint.y-O.y)*(latticePoint.y-O.y) ;
			if (radsq < aprtsq) {
				peakValue = FindTEMPeak(image,&searchRect,&peakPos) ;
			}
			else {
				peakValue = 0. ;
			}
			if (peakValue > 0.) {
				// Add this peak
				thePeak.value = peakValue ;
				thePeak.pos = peakPos ;
				thePeak.n = i ;
				thePeak.m = j ;
				peakList->AddPeak(thePeak) ;
				latticePoint.x +=  V.x ;
				latticePoint.y +=  V.y ;
			}
			else {
				// Did not find a peak, but advance anyway
				latticePoint.x +=  V.x ;
				latticePoint.y +=  V.y ;
			}
			SetRect(&searchRect,near(latticePoint.x-sW),near(latticePoint.y-sW),
					near(latticePoint.x+sW),near(latticePoint.y+sW));
			j++ ;
		}
		// Reset to j = 0 
		peakPos = peakJ0Pos ;
		// Loop over negative j's
		// Start with j = -1
		j = -1 ;
		latticePoint.x = peakPos.x - V.x ;
		latticePoint.y = peakPos.y - V.y ;
		SetRect(&searchRect,near(latticePoint.x-sW),near(latticePoint.y-sW),
				near(latticePoint.x+sW),near(latticePoint.y+sW));
		while(PtInRect(&latticePoint,&imageRect)) {
			radsq = (latticePoint.x-O.x)*(latticePoint.x-O.x)+
				(latticePoint.y-O.y)*(latticePoint.y-O.y) ;
			if (radsq < aprtsq) {
				peakValue = FindTEMPeak(image,&searchRect,&peakPos) ;
			}
			else {
				peakValue = 0. ;
			}
			if (peakValue > 0.) {
				// Add this peak
				thePeak.value = peakValue ;
				thePeak.n = i ;
				thePeak.m = j ;
				thePeak.pos = peakPos ;
				peakList->AddPeak(thePeak) ;
				latticePoint.x -=  V.x ;
				latticePoint.y -=  V.y ;
			}
			else {
				// Did not find a peak, but advance anyway
				latticePoint.x -=  V.x ;
				latticePoint.y -=  V.y ;
			}
			SetRect(&searchRect,near(latticePoint.x-sW),near(latticePoint.y-sW),
				near(latticePoint.x+sW),near(latticePoint.y+sW));
			j-- ;
		}
		peakPos = peakJ0Pos ;
		// go to next i
		latticePoint.x = peakJ0Pos.x + U.x ;
		latticePoint.y = peakJ0Pos.y + U.y ;
		SetRect(&searchRect,latticePoint.x-sW,latticePoint.y-sW,
				latticePoint.x+sW,latticePoint.y+sW);
		i++ ;
		j = 0 ;
	}

	// negative i, starting with j = 0
	i = -1 ;
	j = 0 ;
	latticePoint.x = O.x-U.x ;
	latticePoint.y = O.y-U.y ;
	SetRect(&searchRect,near(latticePoint.x-sW),near(latticePoint.y-sW),
			near(latticePoint.x+sW),near(latticePoint.y+sW));
	while (PtInRect(&latticePoint,&imageRect)) {
		peakPos.x = latticePoint.x ; peakPos.y = latticePoint.y ;
		radsq = (latticePoint.x-O.x)*(latticePoint.x-O.x)+
				(latticePoint.y-O.y)*(latticePoint.y-O.y) ;
		if (radsq < aprtsq) {
				peakValue = FindTEMPeak(image,&searchRect,&peakPos) ;
		}
		else {
				peakValue = 0. ;
		}
		if (peakValue > 0.) {
			// Add this peak
			thePeak.value = peakValue ;
			thePeak.pos = peakPos ;
			thePeak.n = i ;
			thePeak.m = j ;
			peakList->AddPeak(thePeak) ;
		}

		peakJ0Pos = peakPos ;
		// Loop over positive j
		// Start with j = 1
		j = 1 ;
		latticePoint.x +=  V.x ;
		latticePoint.y +=  V.y ;
		SetRect(&searchRect,near(latticePoint.x-sW),near(latticePoint.y-sW),
				near(latticePoint.x+sW),near(latticePoint.y+sW));
		while(PtInRect(&latticePoint,&imageRect)) {
			radsq = (latticePoint.x-O.x)*(latticePoint.x-O.x)+
				(latticePoint.y-O.y)*(latticePoint.y-O.y) ;
			if (radsq < aprtsq) {
				peakValue = FindTEMPeak(image,&searchRect,&peakPos) ;
			}
			else {
				peakValue = 0. ;
			}
			if (peakValue > 0.) {
				// Add this peak
				thePeak.value = peakValue ;
				thePeak.pos = peakPos ;
				thePeak.n = i ;
				thePeak.m = j ;
				peakList->AddPeak(thePeak) ;
				latticePoint.x +=  V.x ;
				latticePoint.y +=  V.y ;
			}
			else {
				// Did not find a peak, but advance anyway
				latticePoint.x +=  V.x ;
				latticePoint.y +=  V.y ;
			}
			SetRect(&searchRect,near(latticePoint.x-sW),near(latticePoint.y-sW),
				near(latticePoint.x+sW),near(latticePoint.y+sW));
			j++ ;
		}
		// Reset to j = 0 
		peakPos = peakJ0Pos ;
		// Loop over negative j's
		// Start with j = -1
		j = -1 ;
		latticePoint.x = peakPos.x - V.x ;
		latticePoint.y = peakPos.y - V.y ;
		SetRect(&searchRect,near(latticePoint.x-sW),near(latticePoint.y-sW),
				near(latticePoint.x+sW),near(latticePoint.y+sW));
		while(PtInRect(&latticePoint,&imageRect)) {
			radsq = (latticePoint.x-O.x)*(latticePoint.x-O.x)+
				(latticePoint.y-O.y)*(latticePoint.y-O.y) ;
			if (radsq < aprtsq) {
				peakValue = FindTEMPeak(image,&searchRect,&peakPos) ;
			}
			else {
				peakValue = 0. ;
			}
			if (peakValue > 0.) {
				// Add this peak
				thePeak.value = peakValue ;
				thePeak.pos = peakPos ;
				thePeak.n = i ;
				thePeak.m = j ;
				peakList->AddPeak(thePeak) ;
				latticePoint.x -=  V.x ;
				latticePoint.y -=  V.y ;
			}
			else {
				// Did not find a peak, but advance anyway
				latticePoint.x -=  V.x ;
				latticePoint.y -=  V.y ;
			}
			SetRect(&searchRect,near(latticePoint.x-sW),near(latticePoint.y-sW),
				near(latticePoint.x+sW),near(latticePoint.y+sW));
			j-- ;
		}
		peakPos = peakJ0Pos ;
		// go to next i
		latticePoint.x = peakJ0Pos.x - U.x ;
		latticePoint.y = peakJ0Pos.y - U.y ;
		SetRect(&searchRect,near(latticePoint.x-sW),near(latticePoint.y-sW),
				near(latticePoint.x+sW),near(latticePoint.y+sW));
		i-- ;
		j = 0 ;
	}

		
	if (peakList->GetNumberOfPeaks() == 0) {
		delete peakList ;
		return nil ;
	}
	return peakList ;
}

TPeakList *FindDiffractionSpotsAtLatticePoints(TImage *image,TLattice *lattice)	{
	Rect 		imageRect ;
	FPoint		U,V,O,latticePoint ;
	register	int i,j ;
	long		h2,w2 ;
	float		peakValue,radsq,aprtsq,aprtsq1,aprtsq2,aperture ;
	TPeakList	*peakList = new TPeakList ;
	TPeak		thePeak ;
	short		sW = 3 ; 

	w2 = image->Width()/2 ;
	h2 = image->Height()/2 ;
	useZero = UseZero() ;


	// Start at the origin of the lattice and find lattice points 
	// in a recursive fashion
	O = lattice->GetOrigin() ;
	U = lattice->GetU() ;
	V = lattice->GetV() ;

	// Check to make sure local peak search distance is smaller than 1/2 lattice spacing
	float	du = 0.5*sqrt(U.x*U.x+U.y*U.y) ;
	float	dv = 0.5*sqrt(V.x*V.x+V.y*V.y) ;
	float	halfSpacing = du < dv ? du : dv ;
	if (sW > near(halfSpacing)) sW = near(halfSpacing) ;

	//float	aperture = w2 < h2 ? w2 : h2 ;
	//aperture -= 2*halfSpacing ;
	//aprtsq1 = aperture*aperture ;
	//aprtsq2 = gDiffractionAperture*gDiffractionAperture ;
	//aprtsq = aprtsq1 < aprtsq2 ? aprtsq1 : aprtsq2 ;
	SetRect(&imageRect,0,0,image->Width(),image->Height()) ;
	
	PeakCriteria criteria = GetPeakCriteria() ;
	// Ensure that we do a symmetric region about the origin
	w2 = image->Width()/2 ;
	h2 = image->Height()/2 ;
	aperture = w2 < h2 ? w2 : h2 ;
	// Smallest distance along x
	float HowFarX = 2*w2-O.x > O.x ? O.x : 2*w2-O.x ;
	// Smallest distance along y
	float HowFarY = 2*h2-O.y > O.y ? O.y : 2*h2-O.y ;
	// Smallest of these and an aperture given elsewhere
	if (aperture > HowFarX) aperture = HowFarX ;
	if (aperture > HowFarY) aperture = HowFarY ;
	if (aperture > RadTot)  aperture = RadTot  ;
	if (aperture > gDiffractionAperture) aperture = gDiffractionAperture ;
	aprtsq = aperture*aperture ;
	//fprintf(stderr,"Aperture stuff 2 %f %f \n",aperture,gDiffractionAperture);

	i = 0 ;
	j = 0 ;
	latticePoint.x = O.x ;
	latticePoint.y = O.y ;
	// Positive i
	while(PtInRect(&latticePoint,&imageRect)) {
		radsq = (latticePoint.x-O.x)*(latticePoint.x-O.x)+
			(latticePoint.y-O.y)*(latticePoint.y-O.y) ;
		if (radsq < aprtsq)
			peakValue = FindLatticePeak(image,&latticePoint,halfSpacing,criteria) ;
		else 
			peakValue = 0. ;
		if (peakValue > 0.) {
			// Add this peak
			thePeak.value = peakValue ;
			thePeak.pos = latticePoint ;
			thePeak.n = i ;
			thePeak.m = j ;
			peakList->AddPeak(thePeak) ;
		}
		// Loop over positive j
		j = 1 ;
		latticePoint.x = O.x + i*U.x + j*V.x ;
		latticePoint.y = O.y + i*U.y + j*V.y ;
		while(PtInRect(&latticePoint,&imageRect)) {
			radsq = (latticePoint.x-O.x)*(latticePoint.x-O.x)+
				(latticePoint.y-O.y)*(latticePoint.y-O.y) ;
			if (radsq < aprtsq)
				peakValue = FindLatticePeak(image,&latticePoint,halfSpacing,criteria) ;
			else 
				peakValue = 0. ;
			if (peakValue > 0.) {
				// Add this peak
				thePeak.value = peakValue ;
				thePeak.pos = latticePoint ;
				thePeak.n = i ;
				thePeak.m = j ;
				peakList->AddPeak(thePeak) ;
			}
			j++ ;
			latticePoint.x = O.x + i*U.x + j*V.x ;
			latticePoint.y = O.y + i*U.y + j*V.y ;
		}
		// Loop over negative j's
		// Start with j = -1
		j = -1 ;
		latticePoint.x = O.x + i*U.x + j*V.x ;
		latticePoint.y = O.y + i*U.y + j*V.y ;
		while(PtInRect(&latticePoint,&imageRect)) {
			radsq = (latticePoint.x-O.x)*(latticePoint.x-O.x)+
				(latticePoint.y-O.y)*(latticePoint.y-O.y) ;
			if (radsq < aprtsq)
				peakValue = FindLatticePeak(image,&latticePoint,halfSpacing,criteria) ;
			else 
				peakValue = 0. ;
			if (peakValue > 0.) {
				// Add this peak
				thePeak.value = peakValue ;
				thePeak.n = i ;
				thePeak.m = j ;
				thePeak.pos = latticePoint ;
				peakList->AddPeak(thePeak) ;
			}
			j-- ;
			latticePoint.x = O.x + i*U.x + j*V.x ;
			latticePoint.y = O.y + i*U.y + j*V.y ;
		}
		// go to next i and set j = 0 
		i++ ;
		j = 0 ;
		latticePoint.x = O.x + i*U.x + j*V.x ;
		latticePoint.y = O.y + i*U.y + j*V.y ;
	}

	// negative i, starting with j = 0
	i = -1 ;
	j = 0 ;
	latticePoint.x = O.x-U.x ;
	latticePoint.y = O.y-U.y ;
	while (PtInRect(&latticePoint,&imageRect)) {
		radsq = (latticePoint.x-O.x)*(latticePoint.x-O.x)+
			(latticePoint.y-O.y)*(latticePoint.y-O.y) ;
		if (radsq < aprtsq)
			peakValue = FindLatticePeak(image,&latticePoint,halfSpacing,criteria) ;
		else 
			peakValue = 0. ;
		if (peakValue > 0.) {
			// Add this peak
			thePeak.value = peakValue ;
			thePeak.pos = latticePoint ;
			thePeak.n = i ;
			thePeak.m = j ;
			peakList->AddPeak(thePeak) ;
		}
		// Loop over positive j
		// Start with j = 1
		j = 1 ;
		latticePoint.x = O.x + i*U.x + j*V.x ;
		latticePoint.y = O.y + i*U.y + j*V.y ;
		while(PtInRect(&latticePoint,&imageRect)) {
			radsq = (latticePoint.x-O.x)*(latticePoint.x-O.x)+
				(latticePoint.y-O.y)*(latticePoint.y-O.y) ;
			if (radsq < aprtsq)
				peakValue = FindLatticePeak(image,&latticePoint,halfSpacing,criteria) ;
			else 
				peakValue = 0. ;
			if (peakValue > 0.) {
				// Add this peak
				thePeak.value = peakValue ;
				thePeak.pos = latticePoint ;
				thePeak.n = i ;
				thePeak.m = j ;
				peakList->AddPeak(thePeak) ;
			}
			j++ ;
			latticePoint.x = O.x + i*U.x + j*V.x ;
			latticePoint.y = O.y + i*U.y + j*V.y ;
		}
		// Loop over negative j's
		// Start with j = -1
		j = -1 ;
		latticePoint.x = O.x + i*U.x + j*V.x ;
		latticePoint.y = O.y + i*U.y + j*V.y ;
		while(PtInRect(&latticePoint,&imageRect)) {
			radsq = (latticePoint.x-O.x)*(latticePoint.x-O.x)+
				(latticePoint.y-O.y)*(latticePoint.y-O.y) ;
			if (radsq < aprtsq)
				peakValue = FindLatticePeak(image,&latticePoint,halfSpacing) ;
			else 
				peakValue = 0. ;
			if (peakValue > 0.) {
				// Add this peak
				thePeak.value = peakValue ;
				thePeak.pos = latticePoint ;
				thePeak.n = i ;
				thePeak.m = j ;
				peakList->AddPeak(thePeak) ;
			}
			j-- ;
			latticePoint.x = O.x + i*U.x + j*V.x ;
			latticePoint.y = O.y + i*U.y + j*V.y ;
		}
		// go to next i and j = 0 
		i-- ;
		j = 0 ;
		latticePoint.x = O.x + i*U.x + j*V.x ;
		latticePoint.y = O.y + i*U.y + j*V.y ;
	}
		
	if (peakList->GetNumberOfPeaks() == 0) {
		delete peakList ;
		return nil ;
	}
	return peakList ;
}

TPeak *FindDiffractionSpotAtLatticePoint(Point &mousePt,TImage *image,
						TLattice *lattice,TImage *peakMotif) {
	register	int i,j ;
	float		x,y,radsq ;

	short	motifWidth = peakMotif->Width() ;
	short	motifHeight = peakMotif->Height() ;

	float	outerRadius = motifWidth/2 ;
	float	innerRadius = outerRadius*RadSize ;
	TImage	*weightImage = new TImage(motifWidth,motifHeight) ;	
	float	motifRadiusSq = innerRadius*innerRadius ;
	// Set the weights. For now just a circular mask
	register float *p1 = weightImage->GetData() ;
	for(j = 0; j < motifHeight; j++) {
		y = j - outerRadius ;
		for(i = 0 ; i < motifWidth; i++) {
			x = i - outerRadius ;
			radsq = x*x+y*y ;
			if (radsq >= motifRadiusSq) {
				*p1++ = 1. ;
			}
			else {
				*p1++ = 0. ;
			}
		}
	}
	
	TPeak	*peak =  FindLatticePeak(image,mousePt,peakMotif,weightImage,lattice) ;
	
	delete	weightImage ;
	return peak ;
}
TPeak *TestDiffractionSpotAtLatticePoint(Point &mousePt,TImage *image,TLattice *lattice,TImage *peakMotif) {
	register	int i,j ;
	float		x,y,radsq ;

	short	motifWidth = peakMotif->Width() ;
	short	motifHeight = peakMotif->Height() ;
		

	float	outerRadius = motifWidth/2 ;
	float	innerRadius = outerRadius*RadSize ;
	TImage	*weightImage = new TImage(motifWidth,motifHeight) ;	
	float	motifRadiusSq = innerRadius*innerRadius ;
	// Set the weights. For now just a circular mask
	register float *p1 = weightImage->GetData() ;
	for(j = 0; j < motifHeight; j++) {
		y = j - outerRadius ;
		for(i = 0 ; i < motifWidth; i++) {
			x = i - outerRadius ;
			radsq = x*x+y*y ;
			if (radsq >= motifRadiusSq) {
				*p1++ = 1. ;
			}
			else {
				*p1++ = 0. ;
			}
		}
	}
	
	TPeak	*peak =  TestLatticePeak(image,mousePt,peakMotif,weightImage,lattice) ;
	
	delete	weightImage ;
	return peak ;
}

TPeakList *FindDiffractionSpotsAtLatticePoints(TImage *image,
							TLattice *lattice,TImage *peakMotif,float ccThreshold)	{
	Rect 		imageRect ;
	FPoint		U,V,O,latticePoint ;
	register	int i,j ;
	long		h2,w2 ;
	float		x,y,peakValue,radsq,aprtsq,aperture,Rfactor,Mean ;
	TPeakList	*peakList = new TPeakList ;
	TPeak		thePeak ;
	short		sW = 3 ; 

	w2 = image->Width()/2 ;
	h2 = image->Height()/2 ;
	useZero = UseZero() ;


	// Start at the origin of the lattice and find lattice points 
	// in a recursive fashion
	O = lattice->GetOrigin() ;
	U = lattice->GetU() ;
	V = lattice->GetV() ;

	// Check to make sure we do not go close to any edges of the picture
	float	du = 0.5*sqrt(U.x*U.x+U.y*U.y) ;
	float	dv = 0.5*sqrt(V.x*V.x+V.y*V.y) ;
	float	halfSpacing = du < dv ? du : dv ;
	if (sW > near(halfSpacing)) sW = near(halfSpacing) ;

	short	motifWidth = peakMotif->Width() ;
	short	motifHeight = peakMotif->Height() ;
	short	edgeDistance = motifWidth > motifHeight ? motifWidth : motifHeight ;
	
	//float	aperture = w2 < h2 ? w2 : h2 ;
	//aperture -= edgeDistance ;
	
	//if (gDiffractionAperture < aperture) aperture = gDiffractionAperture ;
	// Ensure that we do a symmetric region about the origin
	w2 = image->Width()/2 ;
	h2 = image->Height()/2 ;
	aperture = w2 < h2 ? w2 : h2 ;
	// Smallest distance along x
	float HowFarX = 2*w2-O.x > O.x ? O.x : 2*w2-O.x ;
	// Smallest distance along y
	float HowFarY = 2*h2-O.y > O.y ? O.y : 2*h2-O.y ;
	// Smallest of these and an aperture given elsewhere
	if (aperture > HowFarX) aperture = HowFarX ;
	if (aperture > HowFarY) aperture = HowFarY ;
	if (aperture > RadTot)  aperture = RadTot  ;
	if (aperture > gDiffractionAperture) aperture = gDiffractionAperture ;
	aprtsq = aperture*aperture ;
	//fprintf(stderr,"Aperture stuff 3 %f %f \n",aperture,gDiffractionAperture);

	//aprtsq = aperture*aperture ;
	
	SetRect(&imageRect,0,0,image->Width(),image->Height()) ;	
	//InsetRect(&imageRect,near(w2-aperture),near(h2-aperture)) ;
	

	float	outerRadius = motifWidth/2 ;
	float	innerRadius = outerRadius*RadSize ;
	TImage	*weightImage = new TImage(motifWidth,motifHeight) ;	
	float	motifRadiusSq = innerRadius*innerRadius ;
	// Set the weights. For now just a circular mask
	register float *p1 = weightImage->GetData() ;
	for(j = 0; j < motifHeight; j++) {
		y = j - outerRadius ;
		for(i = 0 ; i < motifWidth; i++) {
			x = i - outerRadius ;
			radsq = x*x+y*y ;
			if (radsq >= motifRadiusSq) {
				*p1++ = 1. ;
			}
			else {
				*p1++ = 0. ;
			}
		}
	}
	
	i = 0 ;
	j = 0 ;
	latticePoint.x = O.x ;
	latticePoint.y = O.y ;
	// Positive i
	while(PtInRect(&latticePoint,&imageRect)) {
		radsq = (latticePoint.x-O.x)*(latticePoint.x-O.x)+
			(latticePoint.y-O.y)*(latticePoint.y-O.y) ;
		if (radsq < aprtsq)
			peakValue = FindLatticePeak(image,&latticePoint,peakMotif,weightImage,ccThreshold,Rfactor,Mean) ;
		else 
			peakValue = 0. ;
		if (peakValue > 0.) {
			// Add this peak
			thePeak.value = peakValue ;
			thePeak.pos = latticePoint ;
			thePeak.n = i ;
			thePeak.m = j ;
			thePeak.Rfactor = Rfactor ;
			peakList->AddPeak(thePeak) ;
		}
		// Loop over positive j
		j = 1 ;
		latticePoint.x = O.x + i*U.x + j*V.x ;
		latticePoint.y = O.y + i*U.y + j*V.y ;
		while(PtInRect(&latticePoint,&imageRect)) {
			radsq = (latticePoint.x-O.x)*(latticePoint.x-O.x)+
				(latticePoint.y-O.y)*(latticePoint.y-O.y) ;
			if (radsq < aprtsq)
				peakValue = FindLatticePeak(image,&latticePoint,peakMotif,weightImage,ccThreshold,Rfactor,Mean) ;
			else 
				peakValue = 0. ;
			if (peakValue > 0.) {
				// Add this peak
				thePeak.value = peakValue ;
				thePeak.pos = latticePoint ;
				thePeak.n = i ;
				thePeak.m = j ;
				thePeak.Rfactor = Rfactor ;
				thePeak.Mean = Mean ;
				peakList->AddPeak(thePeak) ;
			}
			j++ ;
			latticePoint.x = O.x + i*U.x + j*V.x ;
			latticePoint.y = O.y + i*U.y + j*V.y ;
		}
		// Loop over negative j's
		// Start with j = -1
		j = -1 ;
		latticePoint.x = O.x + i*U.x + j*V.x ;
		latticePoint.y = O.y + i*U.y + j*V.y ;
		while(PtInRect(&latticePoint,&imageRect)) {
			radsq = (latticePoint.x-O.x)*(latticePoint.x-O.x)+
				(latticePoint.y-O.y)*(latticePoint.y-O.y) ;
			if (radsq < aprtsq)
				peakValue = FindLatticePeak(image,&latticePoint,peakMotif,weightImage,ccThreshold,Rfactor,Mean) ;
			else 
				peakValue = 0. ;
			if (peakValue > 0.) {
				// Add this peak
				thePeak.value = peakValue ;
				thePeak.n = i ;
				thePeak.m = j ;
				thePeak.pos = latticePoint ;
				thePeak.Rfactor = Rfactor ;
				thePeak.Mean = Mean ;
				peakList->AddPeak(thePeak) ;
			}
			j-- ;
			latticePoint.x = O.x + i*U.x + j*V.x ;
			latticePoint.y = O.y + i*U.y + j*V.y ;
		}
		// go to next i and set j = 0 
		i++ ;
		j = 0 ;
		latticePoint.x = O.x + i*U.x + j*V.x ;
		latticePoint.y = O.y + i*U.y + j*V.y ;
	}

	// negative i, starting with j = 0
	i = -1 ;
	j = 0 ;
	latticePoint.x = O.x-U.x ;
	latticePoint.y = O.y-U.y ;
	while(PtInRect(&latticePoint,&imageRect)) {
		radsq = (latticePoint.x-O.x)*(latticePoint.x-O.x)+
			(latticePoint.y-O.y)*(latticePoint.y-O.y) ;
		if (radsq < aprtsq)
			peakValue = FindLatticePeak(image,&latticePoint,peakMotif,weightImage,ccThreshold,Rfactor,Mean) ;
		else 
			peakValue = 0. ;
		if (peakValue > 0.) {
			// Add this peak
			thePeak.value = peakValue ;
			thePeak.pos = latticePoint ;
			thePeak.n = i ;
			thePeak.m = j ;
			thePeak.Rfactor = Rfactor ;
			thePeak.Mean = Mean ;
			peakList->AddPeak(thePeak) ;
		}
		// Loop over positive j
		// Start with j = 1
		j = 1 ;
				thePeak.Rfactor = Rfactor ;
				thePeak.Mean = Mean ;
		latticePoint.x = O.x + i*U.x + j*V.x ;
		latticePoint.y = O.y + i*U.y + j*V.y ;
		while(PtInRect(&latticePoint,&imageRect)) {
			radsq = (latticePoint.x-O.x)*(latticePoint.x-O.x)+
				(latticePoint.y-O.y)*(latticePoint.y-O.y) ;
			if (radsq < aprtsq)
				peakValue = FindLatticePeak(image,&latticePoint,peakMotif,weightImage,ccThreshold,Rfactor,Mean) ;
			else 
				peakValue = 0. ;
			if (peakValue > 0.) {
				// Add this peak
				thePeak.value = peakValue ;
				thePeak.pos = latticePoint ;
				thePeak.n = i ;
				thePeak.m = j ;
				thePeak.Rfactor = Rfactor ;
				thePeak.Mean = Mean ;
				peakList->AddPeak(thePeak) ;
			}
			j++ ;
			latticePoint.x = O.x + i*U.x + j*V.x ;
			latticePoint.y = O.y + i*U.y + j*V.y ;
		}
		// Loop over negative j's
		// Start with j = -1
		j = -1 ;
		latticePoint.x = O.x + i*U.x + j*V.x ;
		latticePoint.y = O.y + i*U.y + j*V.y ;
		while(PtInRect(&latticePoint,&imageRect)) {
			radsq = (latticePoint.x-O.x)*(latticePoint.x-O.x)+
				(latticePoint.y-O.y)*(latticePoint.y-O.y) ;
			if (radsq < aprtsq)
				peakValue = FindLatticePeak(image,&latticePoint,peakMotif,weightImage,ccThreshold,Rfactor,Mean) ;
			else 
				peakValue = 0. ;
			if (peakValue > 0.) {
				// Add this peak
				thePeak.value = peakValue ;
				thePeak.pos = latticePoint ;
				thePeak.n = i ;
				thePeak.m = j ;
				thePeak.Rfactor = Rfactor ;
				thePeak.Mean = Mean ;
				peakList->AddPeak(thePeak) ;
			}
			j-- ;
			latticePoint.x = O.x + i*U.x + j*V.x ;
			latticePoint.y = O.y + i*U.y + j*V.y ;
		}
		// go to next i and j = 0 
		i-- ;
		j = 0 ;
		latticePoint.x = O.x + i*U.x + j*V.x ;
		latticePoint.y = O.y + i*U.y + j*V.y ;
	}
		
	delete weightImage ;
	if (peakList->GetNumberOfPeaks() == 0) {
		delete peakList ;
		return nil ;
	}
	return peakList ;
}
TPeakList *FindDiffractionSpotsAtLatticePointsC(TImage *image,
							TLattice *lattice,TImage *peakMotif,float ccThreshold,
							float *BX, float *BY)	{
	Rect 		imageRect ;
	FPoint		U,V,O,latticePoint ;
	register	int i,j ;
	long		h2,w2 ;
	float		x,y,peakValue,radsq,aprtsq,aperture,Rfactor,Mean ;
	TPeakList	*peakList = new TPeakList ;
	TPeak		thePeak ;
	short		sW = 3 ; 

	w2 = image->Width()/2 ;
	h2 = image->Height()/2 ;
	useZero = UseZero() ;


	// Start at the origin of the lattice and find lattice points 
	// in a recursive fashion
	O = lattice->GetOrigin() ;
	U = lattice->GetU() ;
	V = lattice->GetV() ;

	// Check to make sure we do not go close to any edges of the picture
	float	du = 0.5*sqrt(U.x*U.x+U.y*U.y) ;
	float	dv = 0.5*sqrt(V.x*V.x+V.y*V.y) ;
	float	halfSpacing = du < dv ? du : dv ;
	if (sW > near(halfSpacing)) sW = near(halfSpacing) ;

	short	motifWidth = peakMotif->Width() ;
	short	motifHeight = peakMotif->Height() ;
	short	edgeDistance = motifWidth > motifHeight ? motifWidth : motifHeight ;
	
	//float	aperture = w2 < h2 ? w2 : h2 ;
	//aperture -= edgeDistance ;
	
	//if (gDiffractionAperture < aperture) aperture = gDiffractionAperture ;
	// Ensure that we do a symmetric region about the origin
	w2 = image->Width()/2 ;
	h2 = image->Height()/2 ;
	aperture = w2 < h2 ? w2 : h2 ;
	// Smallest distance along x
	float HowFarX = 2*w2-O.x > O.x ? O.x : 2*w2-O.x ;
	// Smallest distance along y
	float HowFarY = 2*h2-O.y > O.y ? O.y : 2*h2-O.y ;
	// Smallest of these and an aperture given elsewhere
	if (aperture > HowFarX) aperture = HowFarX ;
	if (aperture > HowFarY) aperture = HowFarY ;
	if (aperture > RadTot)  aperture = RadTot  ;
	if (aperture > gDiffractionAperture) aperture = gDiffractionAperture ;
	aprtsq = aperture*aperture ;
	//fprintf(stderr,"Aperture stuff 3 %f %f \n",aperture,gDiffractionAperture);

	//aprtsq = aperture*aperture ;
	
	SetRect(&imageRect,0,0,image->Width(),image->Height()) ;	
	//InsetRect(&imageRect,near(w2-aperture),near(h2-aperture)) ;
	

	float	outerRadius = motifWidth/2 ;
	float	innerRadius = outerRadius*RadSize ;
	TImage	*weightImage = new TImage(motifWidth,motifHeight) ;	
	float	motifRadiusSq = innerRadius*innerRadius ;
	// Set the weights. For now just a circular mask
	register float *p1 = weightImage->GetData() ;
	for(j = 0; j < motifHeight; j++) {
		y = j - outerRadius ;
		for(i = 0 ; i < motifWidth; i++) {
			x = i - outerRadius ;
			radsq = x*x+y*y ;
			if (radsq >= motifRadiusSq) {
				*p1++ = 1. ;
			}
			else {
				*p1++ = 0. ;
			}
		}
	}
	
	i = 0 ;
	j = 0 ;
		latticePoint.x = RealXYValueC( i, j, BX) ; 
		latticePoint.y = RealXYValueC( i, j, BY) ;  
	// Positive i
	while(PtInRect(&latticePoint,&imageRect)) {
		radsq = (latticePoint.x-O.x)*(latticePoint.x-O.x)+
			(latticePoint.y-O.y)*(latticePoint.y-O.y) ;
		if (radsq < aprtsq)
			peakValue = FindLatticePeak(image,&latticePoint,peakMotif,weightImage,ccThreshold,Rfactor,Mean) ;
		else 
			peakValue = 0. ;
		if (peakValue > 0.) {
			// Add this peak
			thePeak.value = peakValue ;
			thePeak.pos = latticePoint ;
			thePeak.n = i ;
			thePeak.m = j ;
			thePeak.Rfactor = Rfactor ;
			peakList->AddPeak(thePeak) ;
		}
		// Loop over positive j
		j = 1 ;
		latticePoint.x = RealXYValueC( i, j, BX) ; 
		latticePoint.y = RealXYValueC( i, j, BY) ;  
			while(PtInRect(&latticePoint,&imageRect)) {
			radsq = (latticePoint.x-O.x)*(latticePoint.x-O.x)+
				(latticePoint.y-O.y)*(latticePoint.y-O.y) ;
			if (radsq < aprtsq)
				peakValue = FindLatticePeak(image,&latticePoint,peakMotif,weightImage,ccThreshold,Rfactor,Mean) ;
			else 
				peakValue = 0. ;
			if (peakValue > 0.) {
				// Add this peak
				thePeak.value = peakValue ;
				thePeak.pos = latticePoint ;
				thePeak.n = i ;
				thePeak.m = j ;
				thePeak.Rfactor = Rfactor ;
				thePeak.Mean = Mean ;
				peakList->AddPeak(thePeak) ;
			}
			j++ ;
		latticePoint.x = RealXYValueC( i, j, BX) ; 
		latticePoint.y = RealXYValueC( i, j, BY) ;  
		}
		// Loop over negative j's
		// Start with j = -1
		j = -1 ;
		latticePoint.x = RealXYValueC( i, j, BX) ; 
		latticePoint.y = RealXYValueC( i, j, BY) ;  
		while(PtInRect(&latticePoint,&imageRect)) {
			radsq = (latticePoint.x-O.x)*(latticePoint.x-O.x)+
				(latticePoint.y-O.y)*(latticePoint.y-O.y) ;
			if (radsq < aprtsq)
				peakValue = FindLatticePeak(image,&latticePoint,peakMotif,weightImage,ccThreshold,Rfactor,Mean) ;
			else 
				peakValue = 0. ;
			if (peakValue > 0.) {
				// Add this peak
				thePeak.value = peakValue ;
				thePeak.n = i ;
				thePeak.m = j ;
				thePeak.pos = latticePoint ;
				thePeak.Rfactor = Rfactor ;
				thePeak.Mean = Mean ;
				peakList->AddPeak(thePeak) ;
			}
			j-- ;
		latticePoint.x = RealXYValueC( i, j, BX) ; 
		latticePoint.y = RealXYValueC( i, j, BY) ;  
		}
		// go to next i and set j = 0 
		i++ ;
		j = 0 ;
		latticePoint.x = RealXYValueC( i, j, BX) ; 
		latticePoint.y = RealXYValueC( i, j, BY) ;  
	}

	// negative i, starting with j = 0
	i = -1 ;
	j = 0 ;
		latticePoint.x = RealXYValueC( i, j, BX) ; 
		latticePoint.y = RealXYValueC( i, j, BY) ;  
	while(PtInRect(&latticePoint,&imageRect)) {
		radsq = (latticePoint.x-O.x)*(latticePoint.x-O.x)+
			(latticePoint.y-O.y)*(latticePoint.y-O.y) ;
		if (radsq < aprtsq)
			peakValue = FindLatticePeak(image,&latticePoint,peakMotif,weightImage,ccThreshold,Rfactor,Mean) ;
		else 
			peakValue = 0. ;
		if (peakValue > 0.) {
			// Add this peak
			thePeak.value = peakValue ;
			thePeak.pos = latticePoint ;
			thePeak.n = i ;
			thePeak.m = j ;
			thePeak.Rfactor = Rfactor ;
			thePeak.Mean = Mean ;
			peakList->AddPeak(thePeak) ;
		}
		// Loop over positive j
		// Start with j = 1
		j = 1 ;
				thePeak.Rfactor = Rfactor ;
				thePeak.Mean = Mean ;
		latticePoint.x = RealXYValueC( i, j, BX) ; 
		latticePoint.y = RealXYValueC( i, j, BY) ;  
		while(PtInRect(&latticePoint,&imageRect)) {
			radsq = (latticePoint.x-O.x)*(latticePoint.x-O.x)+
				(latticePoint.y-O.y)*(latticePoint.y-O.y) ;
			if (radsq < aprtsq)
				peakValue = FindLatticePeak(image,&latticePoint,peakMotif,weightImage,ccThreshold,Rfactor,Mean) ;
			else 
				peakValue = 0. ;
			if (peakValue > 0.) {
				// Add this peak
				thePeak.value = peakValue ;
				thePeak.pos = latticePoint ;
				thePeak.n = i ;
				thePeak.m = j ;
				thePeak.Rfactor = Rfactor ;
				thePeak.Mean = Mean ;
				peakList->AddPeak(thePeak) ;
			}
			j++ ;
		latticePoint.x = RealXYValueC( i, j, BX) ; 
		latticePoint.y = RealXYValueC( i, j, BY) ;  
		}
		// Loop over negative j's
		// Start with j = -1
		j = -1 ;
		latticePoint.x = RealXYValueC( i, j, BX) ; 
		latticePoint.y = RealXYValueC( i, j, BY) ;  
		while(PtInRect(&latticePoint,&imageRect)) {
			radsq = (latticePoint.x-O.x)*(latticePoint.x-O.x)+
				(latticePoint.y-O.y)*(latticePoint.y-O.y) ;
			if (radsq < aprtsq)
				peakValue = FindLatticePeak(image,&latticePoint,peakMotif,weightImage,ccThreshold,Rfactor,Mean) ;
			else 
				peakValue = 0. ;
			if (peakValue > 0.) {
				// Add this peak
				thePeak.value = peakValue ;
				thePeak.pos = latticePoint ;
				thePeak.n = i ;
				thePeak.m = j ;
				thePeak.Rfactor = Rfactor ;
				thePeak.Mean = Mean ;
				peakList->AddPeak(thePeak) ;
			}
			j-- ;
		latticePoint.x = RealXYValueC( i, j, BX) ; 
		latticePoint.y = RealXYValueC( i, j, BY) ;  
		}
		// go to next i and j = 0 
		i-- ;
		j = 0 ;
		latticePoint.x = RealXYValueC( i, j, BX) ; 
		latticePoint.y = RealXYValueC( i, j, BY) ;  
	}
		
	delete weightImage ;
	if (peakList->GetNumberOfPeaks() == 0) {
		delete peakList ;
		return nil ;
	}
	return peakList ;
}

TPeakList *FindDiffractionSpotsAtLatticePoints(TImage *image,
							TLattice *lattice,TImage *peakMotif,float ccThreshold,
							float *BX, float *BY)	{
	Rect 		imageRect ;
	FPoint		U,V,O,latticePoint ;
	register	int i,j ;
	long		h2,w2 ;
	float		x,y,peakValue,radsq,aprtsq,aperture,Rfactor,Mean ;
	TPeakList	*peakList = new TPeakList ;
	TPeak		thePeak ;
	short		sW = 3 ; 

	w2 = image->Width()/2 ;
	h2 = image->Height()/2 ;
	useZero = UseZero() ;


	// Start at the origin of the lattice and find lattice points 
	// in a recursive fashion
	O = lattice->GetOrigin() ;
	U = lattice->GetU() ;
	V = lattice->GetV() ;

	// Check to make sure we do not go close to any edges of the picture
	float	du = 0.5*sqrt(U.x*U.x+U.y*U.y) ;
	float	dv = 0.5*sqrt(V.x*V.x+V.y*V.y) ;
	float	halfSpacing = du < dv ? du : dv ;
	if (sW > near(halfSpacing)) sW = near(halfSpacing) ;

	short	motifWidth = peakMotif->Width() ;
	short	motifHeight = peakMotif->Height() ;
	short	edgeDistance = motifWidth > motifHeight ? motifWidth : motifHeight ;
	
	//float	aperture = w2 < h2 ? w2 : h2 ;
	//aperture -= edgeDistance ;
	
	//if (gDiffractionAperture < aperture) aperture = gDiffractionAperture ;
	// Ensure that we do a symmetric region about the origin
	w2 = image->Width()/2 ;
	h2 = image->Height()/2 ;
	aperture = w2 < h2 ? w2 : h2 ;
	// Smallest distance along x
	float HowFarX = 2*w2-O.x > O.x ? O.x : 2*w2-O.x ;
	// Smallest distance along y
	float HowFarY = 2*h2-O.y > O.y ? O.y : 2*h2-O.y ;
	// Smallest of these and an aperture given elsewhere
	if (aperture > HowFarX) aperture = HowFarX ;
	if (aperture > HowFarY) aperture = HowFarY ;
	if (aperture > RadTot)  aperture = RadTot  ;
	if (aperture > gDiffractionAperture) aperture = gDiffractionAperture ;
	aprtsq = aperture*aperture ;
	//fprintf(stderr,"Aperture stuff 3 %f %f \n",aperture,gDiffractionAperture);

	//aprtsq = aperture*aperture ;
	
	SetRect(&imageRect,0,0,image->Width(),image->Height()) ;	
	//InsetRect(&imageRect,near(w2-aperture),near(h2-aperture)) ;
	

	float	outerRadius = motifWidth/2 ;
	float	innerRadius = outerRadius*RadSize ;
	TImage	*weightImage = new TImage(motifWidth,motifHeight) ;	
	float	motifRadiusSq = innerRadius*innerRadius ;
	// Set the weights. For now just a circular mask
	register float *p1 = weightImage->GetData() ;
	for(j = 0; j < motifHeight; j++) {
		y = j - outerRadius ;
		for(i = 0 ; i < motifWidth; i++) {
			x = i - outerRadius ;
			radsq = x*x+y*y ;
			if (radsq >= motifRadiusSq) {
				*p1++ = 1. ;
			}
			else {
				*p1++ = 0. ;
			}
		}
	}
	
	i = 0 ;
	j = 0 ;
		latticePoint.x = RealXYValueQ( i, j, BX ) ; //BX[0] + i*BX[1] + j*BX[2] +i*i*BX[3] + j*j*BX[4] + i*j*BX[5];
		latticePoint.y = RealXYValueQ( i, j, BY ) ; //BY[0] + i*BY[1] + j*BY[2] +i*i*BY[3] + j*j*BY[4] + i*j*BY[5] ;
	// Positive i
	while(PtInRect(&latticePoint,&imageRect)) {
		radsq = (latticePoint.x-O.x)*(latticePoint.x-O.x)+
			(latticePoint.y-O.y)*(latticePoint.y-O.y) ;
		if (radsq < aprtsq)
			peakValue = FindLatticePeak(image,&latticePoint,peakMotif,weightImage,ccThreshold,Rfactor,Mean) ;
		else 
			peakValue = 0. ;
		if (peakValue > 0.) {
			// Add this peak
			thePeak.value = peakValue ;
			thePeak.pos = latticePoint ;
			thePeak.n = i ;
			thePeak.m = j ;
			thePeak.Rfactor = Rfactor ;
			peakList->AddPeak(thePeak) ;
		}
		// Loop over positive j
		j = 1 ;
		latticePoint.x = RealXYValueQ( i, j, BX) ;
		latticePoint.y = RealXYValueQ( i, j, BY) ; 
		while(PtInRect(&latticePoint,&imageRect)) {
			radsq = (latticePoint.x-O.x)*(latticePoint.x-O.x)+
				(latticePoint.y-O.y)*(latticePoint.y-O.y) ;
			if (radsq < aprtsq)
				peakValue = FindLatticePeak(image,&latticePoint,peakMotif,weightImage,ccThreshold,Rfactor,Mean) ;
			else 
				peakValue = 0. ;
			if (peakValue > 0.) {
				// Add this peak
				thePeak.value = peakValue ;
				thePeak.pos = latticePoint ;
				thePeak.n = i ;
				thePeak.m = j ;
				thePeak.Rfactor = Rfactor ;
				thePeak.Mean = Mean ;
				peakList->AddPeak(thePeak) ;
			}
			j++ ;
		latticePoint.x = RealXYValueQ( i, j, BX) ;
		latticePoint.y = RealXYValueQ( i, j, BY) ; 
		}
		// Loop over negative j's
		// Start with j = -1
		j = -1 ;
		latticePoint.x = RealXYValueQ( i, j, BX) ;
		latticePoint.y = RealXYValueQ( i, j, BY) ; 
		while(PtInRect(&latticePoint,&imageRect)) {
			radsq = (latticePoint.x-O.x)*(latticePoint.x-O.x)+
				(latticePoint.y-O.y)*(latticePoint.y-O.y) ;
			if (radsq < aprtsq)
				peakValue = FindLatticePeak(image,&latticePoint,peakMotif,weightImage,ccThreshold,Rfactor,Mean) ;
			else 
				peakValue = 0. ;
			if (peakValue > 0.) {
				// Add this peak
				thePeak.value = peakValue ;
				thePeak.n = i ;
				thePeak.m = j ;
				thePeak.pos = latticePoint ;
				thePeak.Rfactor = Rfactor ;
				thePeak.Mean = Mean ;
				peakList->AddPeak(thePeak) ;
			}
			j-- ;
		latticePoint.x = RealXYValueQ( i, j, BX) ; 
		latticePoint.y = RealXYValueQ( i, j, BY) ; 
		}
		// go to next i and set j = 0 
		i++ ;
		j = 0 ;
		latticePoint.x = RealXYValueQ( i, j, BX) ; 
		latticePoint.y = RealXYValueQ( i, j, BY) ; 
	}

	// negative i, starting with j = 0
	i = -1 ;
	j = 0 ;
		latticePoint.x = RealXYValueQ( i, j, BX) ; 
		latticePoint.y = RealXYValueQ( i, j, BY) ; 
	while(PtInRect(&latticePoint,&imageRect)) {
		radsq = (latticePoint.x-O.x)*(latticePoint.x-O.x)+
			(latticePoint.y-O.y)*(latticePoint.y-O.y) ;
		if (radsq < aprtsq)
			peakValue = FindLatticePeak(image,&latticePoint,peakMotif,weightImage,ccThreshold,Rfactor,Mean) ;
		else 
			peakValue = 0. ;
		if (peakValue > 0.) {
			// Add this peak
			thePeak.value = peakValue ;
			thePeak.pos = latticePoint ;
			thePeak.n = i ;
			thePeak.m = j ;
			thePeak.Rfactor = Rfactor ;
			thePeak.Mean = Mean ;
			peakList->AddPeak(thePeak) ;
		}
		// Loop over positive j
		// Start with j = 1
		j = 1 ;
				thePeak.Rfactor = Rfactor ;
				thePeak.Mean = Mean ;
		latticePoint.x = RealXYValueQ( i, j, BX) ; 
		latticePoint.y = RealXYValueQ( i, j, BY) ; 
		while(PtInRect(&latticePoint,&imageRect)) {
			radsq = (latticePoint.x-O.x)*(latticePoint.x-O.x)+
				(latticePoint.y-O.y)*(latticePoint.y-O.y) ;
			if (radsq < aprtsq)
				peakValue = FindLatticePeak(image,&latticePoint,peakMotif,weightImage,ccThreshold,Rfactor,Mean) ;
			else 
				peakValue = 0. ;
			if (peakValue > 0.) {
				// Add this peak
				thePeak.value = peakValue ;
				thePeak.pos = latticePoint ;
				thePeak.n = i ;
				thePeak.m = j ;
				thePeak.Rfactor = Rfactor ;
				thePeak.Mean = Mean ;
				peakList->AddPeak(thePeak) ;
			}
			j++ ;
		latticePoint.x = RealXYValueQ( i, j, BX) ; 
		latticePoint.y = RealXYValueQ( i, j, BY) ; 
		}
		// Loop over negative j's
		// Start with j = -1
		j = -1 ;
		latticePoint.x = RealXYValueQ( i, j, BX) ; 
		latticePoint.y = RealXYValueQ( i, j, BY) ; 
		while(PtInRect(&latticePoint,&imageRect)) {
			radsq = (latticePoint.x-O.x)*(latticePoint.x-O.x)+
				(latticePoint.y-O.y)*(latticePoint.y-O.y) ;
			if (radsq < aprtsq)
				peakValue = FindLatticePeak(image,&latticePoint,peakMotif,weightImage,ccThreshold,Rfactor,Mean) ;
			else 
				peakValue = 0. ;
			if (peakValue > 0.) {
				// Add this peak
				thePeak.value = peakValue ;
				thePeak.pos = latticePoint ;
				thePeak.n = i ;
				thePeak.m = j ;
				thePeak.Rfactor = Rfactor ;
				thePeak.Mean = Mean ;
				peakList->AddPeak(thePeak) ;
			}
			j-- ;
		latticePoint.x = RealXYValueQ( i, j, BX) ; 
		latticePoint.y = RealXYValueQ( i, j, BY) ; 
		}
		// go to next i and j = 0 
		i-- ;
		j = 0 ;
		latticePoint.x = RealXYValueQ( i, j, BX) ; 
		latticePoint.y = RealXYValueQ( i, j, BY) ; 
	}
		
	delete weightImage ;
	if (peakList->GetNumberOfPeaks() == 0) {
		delete peakList ;
		return nil ;
	}
	return peakList ;
}
TPeakList *FindPeaksAtLatticePoints(TImage *image,TLattice *lattice,
		Boolean threshold) {
	Rect 		imageRect ;
	FPoint		U,V,O,latticePoint ;
	register	int i,j ;
	long		h2,w2 ;
	float		peakValue,radsq,aprtsq,aperture ;
	TPeakList	*peakList = new TPeakList ;
	TPeak		thePeak ;
	short		sW = 2 ;
	//Set these to (maybe) avoid rubbish
	thePeak.Rfactor = 1 ;
	thePeak.Mean = 0 ;
	w2 = image->Width()/2 ;
	h2 = image->Height()/2 ;

	//aprtsq = gDiffractionAperture*gDiffractionAperture ;
	SetRect(&imageRect,0,0,image->Width(),image->Height()) ;

	// Start at the origin of the lattice and find lattice points 
	// in a recursive fashion
	O = lattice->GetOrigin() ;
	U = lattice->GetU() ;
	V = lattice->GetV() ;

	// Ensure that we do a symmetric region about the origin
	w2 = image->Width()/2 ;
	h2 = image->Height()/2 ;
	aperture = w2 < h2 ? w2 : h2 ;
	// Smallest distance along x
	float HowFarX = 2*w2-O.x > O.x ? O.x : 2*w2-O.x ;
	// Smallest distance along y
	float HowFarY = 2*h2-O.y > O.y ? O.y : 2*h2-O.y ;
	// Smallest of these and an aperture given elsewhere
	if (aperture > HowFarX) aperture = HowFarX ;
	if (aperture > HowFarY) aperture = HowFarY ;
	if (aperture > RadTot)  aperture = RadTot  ;
	if (aperture > gDiffractionAperture) aperture = gDiffractionAperture ;
	aprtsq = aperture*aperture ;
	//fprintf(stderr,"Aperture stuff 5 %f %f \n",aperture,gDiffractionAperture);

	// Check to make sure local peak search distance is smaller than 1/2 lattice spacing
	float	du = 0.5*sqrt(U.x*U.x+U.y*U.y) ;
	float	dv = 0.5*sqrt(V.x*V.x+V.y*V.y) ;
	float	halfSpacing = du < dv ? du : dv ;
	if (sW > near(halfSpacing)) sW = near(halfSpacing) ;

	PeakCriteria criteria = GetPeakCriteria() ;

	i = 0 ;
	j = 0 ;
	latticePoint.x = O.x ;
	latticePoint.y = O.y ;
	// Positive i
	while(PtInRect(&latticePoint,&imageRect)) {
		radsq = (latticePoint.x-O.x)*(latticePoint.x-O.x)+
			(latticePoint.y-O.y)*(latticePoint.y-O.y) ;
		if (radsq < aprtsq)
			peakValue = FindLatticePeak(image,&latticePoint,halfSpacing,criteria) ;
		else 
			peakValue = 0. ;
		if (peakValue > 0.) {
			// Add this peak
			thePeak.value = peakValue ;
			thePeak.pos = latticePoint ;
			thePeak.n = i ;
			thePeak.m = j ;
			peakList->AddPeak(thePeak) ;
		}
		// Loop over positive j
		j = 1 ;
		latticePoint.x = O.x + i*U.x + j*V.x ;
		latticePoint.y = O.y + i*U.y + j*V.y ;
		while(PtInRect(&latticePoint,&imageRect)) {
			radsq = (latticePoint.x-O.x)*(latticePoint.x-O.x)+
				(latticePoint.y-O.y)*(latticePoint.y-O.y) ;
			if (radsq < aprtsq)
				peakValue = FindLatticePeak(image,&latticePoint,halfSpacing,criteria) ;
			else 
				peakValue = 0. ;
			if (peakValue > 0.) {
				// Add this peak
				thePeak.value = peakValue ;
				thePeak.pos = latticePoint ;
				thePeak.n = i ;
				thePeak.m = j ;
				peakList->AddPeak(thePeak) ;
			}
			j++ ;
			latticePoint.x = O.x + i*U.x + j*V.x ;
			latticePoint.y = O.y + i*U.y + j*V.y ;
		}
		// Loop over negative j's
		// Start with j = -1
		j = -1 ;
		latticePoint.x = O.x + i*U.x + j*V.x ;
		latticePoint.y = O.y + i*U.y + j*V.y ;
		while(PtInRect(&latticePoint,&imageRect)) {
			radsq = (latticePoint.x-O.x)*(latticePoint.x-O.x)+
				(latticePoint.y-O.y)*(latticePoint.y-O.y) ;
			if (radsq < aprtsq)
				peakValue = FindLatticePeak(image,&latticePoint,halfSpacing,criteria) ;
			else 
				peakValue = 0. ;
			if (peakValue > 0.) {
				// Add this peak
				thePeak.value = peakValue ;
				thePeak.n = i ;
				thePeak.m = j ;
				thePeak.pos = latticePoint ;
				peakList->AddPeak(thePeak) ;
			}
			j-- ;
			latticePoint.x = O.x + i*U.x + j*V.x ;
			latticePoint.y = O.y + i*U.y + j*V.y ;
		}
		// go to next i and set j = 0 
		i++ ;
		j = 0 ;
		latticePoint.x = O.x + i*U.x + j*V.x ;
		latticePoint.y = O.y + i*U.y + j*V.y ;
	}

	// negative i, starting with j = 0
	i = -1 ;
	j = 0 ;
	latticePoint.x = O.x-U.x ;
	latticePoint.y = O.y-U.y ;
	while (PtInRect(&latticePoint,&imageRect)) {
		radsq = (latticePoint.x-O.x)*(latticePoint.x-O.x)+
			(latticePoint.y-O.y)*(latticePoint.y-O.y) ;
		if (radsq < aprtsq)
			peakValue = FindLatticePeak(image,&latticePoint,halfSpacing,criteria) ;
		else 
			peakValue = 0. ;
		if (peakValue > 0.) {
			// Add this peak
			thePeak.value = peakValue ;
			thePeak.pos = latticePoint ;
			thePeak.n = i ;
			thePeak.m = j ;
			peakList->AddPeak(thePeak) ;
		}
		// Loop over positive j
		// Start with j = 1
		j = 1 ;
		latticePoint.x = O.x + i*U.x + j*V.x ;
		latticePoint.y = O.y + i*U.y + j*V.y ;
		while(PtInRect(&latticePoint,&imageRect)) {
			radsq = (latticePoint.x-O.x)*(latticePoint.x-O.x)+
				(latticePoint.y-O.y)*(latticePoint.y-O.y) ;
			if (radsq < aprtsq)
				peakValue = FindLatticePeak(image,&latticePoint,halfSpacing) ;
			else 
				peakValue = 0. ;
			if (peakValue > 0.) {
				// Add this peak
				thePeak.value = peakValue ;
				thePeak.pos = latticePoint ;
				thePeak.n = i ;
				thePeak.m = j ;
				peakList->AddPeak(thePeak) ;
			}
			j++ ;
			latticePoint.x = O.x + i*U.x + j*V.x ;
			latticePoint.y = O.y + i*U.y + j*V.y ;
		}
		// Loop over negative j's
		// Start with j = -1
		j = -1 ;
		latticePoint.x = O.x + i*U.x + j*V.x ;
		latticePoint.y = O.y + i*U.y + j*V.y ;
		while(PtInRect(&latticePoint,&imageRect)) {
			radsq = (latticePoint.x-O.x)*(latticePoint.x-O.x)+
				(latticePoint.y-O.y)*(latticePoint.y-O.y) ;
			if (radsq < aprtsq)
				peakValue = FindLatticePeak(image,&latticePoint,halfSpacing) ;
			else 
				peakValue = 0. ;
			if (peakValue > 0.) {
				// Add this peak
				thePeak.value = peakValue ;
				thePeak.pos = latticePoint ;
				thePeak.n = i ;
				thePeak.m = j ;
				peakList->AddPeak(thePeak) ;
			}
			j-- ;
			latticePoint.x = O.x + i*U.x + j*V.x ;
			latticePoint.y = O.y + i*U.y + j*V.y ;
		}
		// go to next i and j = 0 
		i-- ;
		j = 0 ;
		latticePoint.x = O.x + i*U.x + j*V.x ;
		latticePoint.y = O.y + i*U.y + j*V.y ;
	}
		
	if (peakList->GetNumberOfPeaks() == 0) {
		delete peakList ;
		return nil ;
	}
	return peakList ;

}

float	FindFTLatticePeak(TImage *image,FPoint *peakPos) {
	long	ix0,iy0,ixc,iyc,ix,iy ;
	long	irad=1 ;
	TArray	*array = image->GetDataArray() ;
	float	peakValue,val,maxVal ;

	
	ix0 = ixc = near(peakPos->x) ;
	iy0 = iyc = near(peakPos->y) ;

	// If we are too close to the edge of the image
	// reurn the midpoint of the area as the peak
	// and the value of zero to indicate no peak
	// This was ix, iy -- must be ix0, iy0 surely ?
	if ( ix0 <  irad+1 || ix0 > image->Width() - irad - 1)
		return 0. ;
	if ( iy0 <  irad+1 || iy0 > image->Height() - irad - 1)
		return 0. ;

	maxVal = -1.0e+10 ;

	// Find the highest pixel in the neighborhood
	for(iy = iy0-irad ; iy <= iy0+irad ; iy++) {
		for(ix = ix0-irad ; ix <= ix0+irad ; ix++) {
			val = array->GetPixelAmplitude(ix,iy) ;
			if (val > maxVal) {
				maxVal = val ;
				ixc = ix ;
				iyc = iy ;
			}
		}
	}

	peakPos->x = ixc ;
	peakPos->y = iyc ;

	peakValue = array->GetPixelAmplitude(ixc,iyc) ;
	return	peakValue ;
}	

// ------------------------------------------------------------------------------------------------------------
// Return the peak at the position clicked by the mouse
// Since we	really want	to find	a peak,	we set the ccThreshold to 0.1
// ------------------------------------------------------------------------------------------------------------
TPeak	*FindLatticePeak(TImage	*image,Point &pos,TImage *peakMotif,TImage *weightImage,TLattice *lattice) {
	long	ixc,iyc ;
	long	ix,iy ;
	//TArray	*array = image->GetDataArray() ;
	TPeak	*peak = nil ;
	char text[256] ;
	
	float ccThreshold = 0.01 ;	

	//Point	highestPixel ;
	//float val,maxVal
	float	xcf,peakValue ;
	long	sw = 1 ;
	long	irad = peakMotif->Width()/2 ;
	useZero = UseZero() ;

	// If we are too close to the edge of the image
	// reurn the midpoint of the area as the peak
	// and the value of zero to indicate no peak
	ix = pos.x  ;
	iy = pos.y ;
	if ( ix <  irad+1 || ix > image->Width() - irad -1 )
		return peak ;
	if ( iy <  irad+1 || iy > image->Height() - irad - 1)
		return peak ;
	
	if (irad/2 < sw) sw = irad/2 ;
	
	ixc =	ix ;
	iyc = 	iy ;
	
	short	motifWidth = peakMotif->Width() ;
	short	motifHeight = peakMotif->Height() ;
	short	halfWidth = motifWidth/2 ;
	short	halfHeight = motifHeight/2 ;

	short 	ix0,iy0,iyt,ixt;
	float	 roff, Rfactor, Wsum, xTol=1e-6; ;
	
	peakValue = 0. ;

		//LDM: Retain the local XCF search -- prevent dump errors	
	TImage *workImage ;
	workImage = image->Extract(ixc,iyc,motifWidth,motifHeight) ;
	//LDM Note: should we have no zero's here ?
		//Probably not - noise can be +ve & negative!
	if (workImage == nil ) {
		xcf = 0 ;
		Rfactor = 0 ;
		delete peak ;
		peak = nil ;
		return peak ;
		}
	if (!FitAndSubtractPeakBackground(workImage,weightImage)) {
		xcf = 0 ;
		Rfactor = 0 ;
		delete peak ;
		peak = nil ;
		return peak ;
	}
		switch (GetIntensityMeasurementMethod()) {
			case kUseXCFMask:
			case kUseXCFAll:
			FitAndSubtractPeakBackMask(workImage,weightImage) ;
			break ;
		}
	TImage *crossCorrelationImage ;
	crossCorrelationImage = CrossCorrelate(peakMotif,workImage) ;
	float xcf0 = crossCorrelationImage->GetPixelAmplitude(0,0) ;
	float xcf0In = xcf0 ;
	float xzero = 0 ; float yzero = 0;
	xcf = crossCorrelationImage->Max(ix0,iy0) ;
	if (ix0 > halfWidth) ix0 -= motifWidth ;
	if (iy0 > halfHeight) iy0 -= motifHeight ;

	roff = ix0*ix0 + iy0*iy0 ;
	if ( roff > 1 && roff < PeakOffTol ) {
		// Try with a shift
		delete workImage ;
		workImage = image->Extract(ixc+ix0,iyc+iy0,motifWidth,motifHeight) ;
		if (workImage == nil ) {
			xcf = 0 ;
			Rfactor = 0 ;
			delete peak ;
			peak = nil ;
			return peak ;
			}
		if (!FitAndSubtractPeakBackground(workImage,weightImage)) {
				xcf = 0 ;
				Rfactor = 0 ;
				delete peak ;
				peak = nil ;
				return peak ;
		}
		switch (GetIntensityMeasurementMethod()) {
			case kUseXCFMask:
			case kUseXCFAll:
			FitAndSubtractPeakBackMask(workImage,weightImage) ;
			break ;
		}
		delete crossCorrelationImage ;
		ixt = ix0 ;
		iyt = iy0 ;
		crossCorrelationImage = CrossCorrelate(peakMotif,workImage) ;
		//LDM Note: this is the properly normalized XCF
		float xcf0 = crossCorrelationImage->GetPixelAmplitude(0,0) ;
		float xzero = ix0 ; float yzero = iy0;
		xcf = crossCorrelationImage->Max(ix0,iy0) ;
		if (ix0 > halfWidth) ix0 -= motifWidth ;
		if (iy0 > halfHeight) iy0 -= motifHeight ;
		ix0 = ix0 + ixt ;
		iy0 = iy0 + iyt ;
		roff = ix0*ix0 + iy0*iy0 ;
	}

	delete crossCorrelationImage ;
	if( (roff >= PeakOffTol)  ) {
	// Set to a small number unless kUseXCFAll or useZero are set
		if( useZero ) {
			//Use origin value
				xcf = xcf0In; 
				ix0 = xzero ;
				iy0 = yzero ;
				roff = 0 ;
				xcf0 = xcf ; 
		}
		switch (GetIntensityMeasurementMethod()) {
			case kUseXCFAll:
				xcf = xcf0In ;
				ix0 = xzero ;
				iy0 = yzero ;
				roff = 0 ;
				xcf0 = xcf ;
				break ;
			}
		}
		Wsum = motifWidth * workImage->Stdv() ;
		Rfactor = xcf*Wsum/(Wsum*Wsum+xTol) ;
		float txcf0 = sqrt(fabsf(xcf0));
		float txcf  = sqrt(fabsf(xcf));
		if(xcf0 < 0) txcf0 = -txcf0 ;
		if (xcf < 0) txcf  = -txcf  ;
		sprintf(text,"Origin %.2f Max %.2f Rfactor %.2f\nX shift %d Y shift %d",
			txcf0,txcf,Rfactor,ix0,iy0) ;
		::WriteObjectInfo(text) ;
	if( roff >= PeakOffTol ) {
	// Set to the origin is useZero is set
		if( useZero ) {
				xcf = xcf0In; 
				ix0 = xzero ;
				iy0 = yzero ;
				roff = 0 ;
				Wsum = motifWidth * workImage->Stdv() ;
				Rfactor = xcf*Wsum/(Wsum*Wsum+xTol) ;
		}
		else {
			//Skip pixels which are too far away
			xcf = 0 ;
			Rfactor = 0 ;
		}
	}
	else {
		Wsum = motifWidth * workImage->Stdv() ;
		Rfactor = xcf*Wsum/(Wsum*Wsum+xTol) ;
		//fprintf(stderr,"Found %f %f \n",xcf,Rfactor);
		switch (GetIntensityMeasurementMethod()) {
			case kUseXCF:
			case kUseXCFMask:
			case kUseXCFAll:
				//Let's find the Rfactor (1 for a perfect match)
				if (Rfactor < ccThreshold && !useZero ) xcf = 0. ;
				//	if (Rfactor > ccThreshold) {
				//	fprintf(stderr," kUseXCF: ") ;
				//	}
				break ;
			case kUseIntegrationBZM:
				if (Rfactor > ccThreshold || useZero ) {
					FitAndSubtractPeakBackZeroMask(workImage,weightImage) ;
					Wsum = workImage->Sum() ;
					//fprintf(stderr," kUseIntegrationBZM: ") ;
					xcf=Wsum/MotifSum ;
				}
				break ;
			case kUseIntegrationBZ:
				if (Rfactor > ccThreshold || useZero ) {
					FitAndSubtractPeakBackZero(workImage,weightImage) ;
					Wsum = workImage->Sum() ;
					//fprintf(stderr," kUseIntegrationBZ: ") ;
					xcf=Wsum/MotifSum ;
				}
				break ;
			case kUseIntegration:
				if (Rfactor > ccThreshold || useZero ) {
					Wsum = workImage->Sum() ;
					//fprintf(stderr," kUseIntegration: ") ;
					xcf=Wsum/MotifSum ;
				}
				break ;
		}
	}
	
	if (xcf != 0.) {
		peak = new TPeak ;
		peak->pos.x = ixc+ix0 ;
		peak->pos.y = iyc+iy0 ;
		peak->value = xcf ;
		peak->Rfactor = Rfactor ;
		peak->Mean = workImage->Mean() ;
		// Assign the correct lattice index for this peak
		FPoint		U,V,O ;
		O = lattice->GetOrigin() ;
		U = lattice->GetU() ;
		V = lattice->GetV() ;
		float det = V.y*U.x-V.x*U.y ;
		float dx = peak->pos.x-O.x ;
		float dy = peak->pos.y-O.y ;
		if (det != 0.) {
			peak->n = near((V.y*dx-V.x*dy)/det) ;
			peak->m = near((U.x*dy-U.y*dx)/det) ;	
			if (gDebugMode) fprintf(stderr,"\nPeak (n,m) = %d,%d",peak->n,peak->m) ; fflush(stderr) ;
		}
		else {
			delete peak ;
			peak = nil ;
		}
	}
		
	delete workImage ;
	return peak ;
}
// ------------------------------------------------------------------------------------------------------------
// Return the peak at the position clicked by the mouse, unless	a real peak	is found
// ------------------------------------------------------------------------------------------------------------
TPeak	*TestLatticePeak(TImage	*image,Point &pos,TImage *peakMotif,TImage *weightImage,TLattice *lattice) {
	long	ixc,iyc ;
	long	ix,iy ;
	//TArray	*array = image->GetDataArray() ;
	TPeak	*peak = nil ;
	char text[256] ;
	
	float ccThreshold = 0.01 ;	
	//float val,maxVal
	//Point	highestPixel ;
	float	xcf,peakValue ;
	long	sw = 1 ;
	long	irad = peakMotif->Width()/2 ;

	// If we are too close to the edge of the image
	// reurn the midpoint of the area as the peak
	// and the value of zero to indicate no peak
	ix = pos.x  ;
	iy = pos.y ;
	useZero = UseZero() ;
	if ( ix <  irad+1 || ix > image->Width() - irad -1 )
		return peak ;
	if ( iy <  irad+1 || iy > image->Height() - irad - 1)
		return peak ;
	
	if (irad/2 < sw) sw = irad/2 ;
	
	ixc =	ix ;
	iyc = 	iy ;
	
	short	motifWidth = peakMotif->Width() ;
	short	motifHeight = peakMotif->Height() ;
	short	halfWidth = motifWidth/2 ;
	short	halfHeight = motifHeight/2 ;

	short 	ix0,iy0,iyt,ixt;
	float	 roff, Rfactor, Wsum, xTol=1e-6; ;
	
	peakValue = 0. ;


		//LDM: Retain the local XCF search -- prevent dump errors	
	TImage *workImage ;
	workImage = image->Extract(ixc,iyc,motifWidth,motifHeight) ;
	//LDM Note: should we have no zero's here ?
		//Probably not - noise can be +ve & negative!
	if (workImage == nil ) {
		xcf = 0 ;
		Rfactor = 0 ;
		delete peak ;
		peak = nil ;
		return peak ;
		}
	float MyMax = workImage->Max() ;
	if(!FitAndSubtractPeakBackground(workImage,weightImage)) {
			xcf = 0 ;
			Rfactor = 0 ;
			delete peak ;
			peak = nil ;
			sprintf(text,"Peak border is saturated\nNot measurable") ;
			::WriteObjectInfo(text) ;
			return peak ;
		}
	//fprintf(stderr,"Background 1 %f %f %f \n",Bfita,Bfitb,Bfitc);
	FitAndSubtractPeakBackMask(workImage,weightImage);
	//fprintf(stderr,"Background 2 %f %f %f \n",Bfita,Bfitb,Bfitc);
	TImage *crossCorrelationImage ;
	crossCorrelationImage = CrossCorrelate(peakMotif,workImage) ;
		float xcf0 = crossCorrelationImage->GetPixelAmplitude(0,0) ;
		float xcf0In = xcf0 ;
		float xzero = 0 ; float yzero = 0;
		xcf = crossCorrelationImage->Max(ix0,iy0) ;
		if (ix0 > halfWidth) ix0 -= motifWidth ;
		if (iy0 > halfHeight) iy0 -= motifHeight ;
		//fprintf(stderr,"Shift 1 %d %d \n",ix0,iy0);
		//LDM: Use 3 pixels here if (ix0 != 0 || iy0 != 0 && (abs(ix0) < 2 && abs(iy0) < 2) ) {
		//Also, update the peak position
		roff = ix0*ix0 + iy0*iy0 ;
		if ( roff > 1 && roff < PeakOffTol ) {
	// if ( ( ix0 != 0 || iy0 != 0 ) && (abs(ix0) < 4 && abs(iy0) < 4) ) {
				// Try with a shift
		delete workImage ;
		workImage = image->Extract(ixc+ix0,iyc+iy0,motifWidth,motifHeight) ;
				//fprintf(stderr,"\n Shifting %i %i ",ix0,iy0);
		if (workImage == nil ) {
			xcf = 0 ;
			Rfactor = 0 ;
			delete peak ;
			peak = nil ;
			return peak ;
			}
		MyMax = workImage->Max() ;
		if (!FitAndSubtractPeakBackground(workImage,weightImage)){ 
				xcf = 0 ;
				Rfactor = 0 ;
				delete peak ;
				peak = nil ;
				sprintf(text,"Peak border is saturated\nNot measurable") ;
				::WriteObjectInfo(text) ;
			return peak ;
			}
		//fprintf(stderr,"Background 3 %f %f %f \n",Bfita,Bfitb,Bfitc);
		FitAndSubtractPeakBackMask(workImage,weightImage);
		//fprintf(stderr,"Background 4 %f %f %f \n",Bfita,Bfitb,Bfitc);
		delete crossCorrelationImage ;
		ixt = ix0 ;
		iyt = iy0 ;
		crossCorrelationImage = CrossCorrelate(peakMotif,workImage) ;
		//LDM Note: this is the properly normalized XCF
		xcf = crossCorrelationImage->Max(ix0,iy0) ;
		if (ix0 > halfWidth) ix0 -= motifWidth ;
		if (iy0 > halfHeight) iy0 -= motifHeight ;
	    ix0 = ix0 + ixt ;
		iy0 = iy0 + iyt ;
		roff = ix0*ix0 + iy0*iy0 ;
	}

	delete crossCorrelationImage ;
	//fprintf(stderr,"\n Start xcf = %f roff= %f ix0 iy0 %i %i Zero %f ",xcf,roff,ix0,iy0,xcf0) ;
		//Wsum = workImage->VectorLength();
		Wsum = motifWidth * workImage->Stdv() ;
		Rfactor = xcf*Wsum/(Wsum*Wsum+xTol) ;
		//Output Peak test information
		float txcf0 = sqrt(fabsf(xcf0));
		float txcf  = sqrt(fabsf(xcf));
		if(xcf0 < 0) txcf0 = -txcf0 ;
		if (xcf < 0) txcf  = -txcf  ;
		sprintf(text,"Max %.2f Origin %.2f XCF %.2f Rfactor %.2f\nX shift %d Y shift %d",
			MyMax,txcf0,txcf,Rfactor,ix0,iy0) ;
		::WriteObjectInfo(text) ;
		//fprintf(stderr,"Fits %f %f %f Orig %f",Bfita,Bfitb,Bfitc,Bfita*halfWidth+Bfitb*halfHeight+Bfitc);
		if( roff >= PeakOffTol ) {
		//Force this case to origin, period
		xcf = xcf0 ;
		ix0 = 0 ;
		iy0 = 0 ;
	}
	//Analyse the peak irregardless of consequences
	//else {
		switch (GetIntensityMeasurementMethod()) {
			case kUseXCF:
			case kUseXCFMask:
			case kUseXCFAll:
				//Let's find the Rfactor (1 for a perfect match)
				if (Rfactor < ccThreshold && !useZero ) xcf = 0. ;
				//  if (Rfactor > ccThreshold) {
				//  fprintf(stderr," kUseXCF: ") ;
				//  }
				break ;
			case kUseIntegrationBZM:
				if (Rfactor > ccThreshold || useZero ) {
					FitAndSubtractPeakBackZeroMask(workImage,weightImage) ;
					Wsum = workImage->Sum() ;
					//fprintf(stderr," kUseIntegrationBZM: ") ;
					xcf=Wsum/MotifSum ;
				}
				break ;
			case kUseIntegrationBZ:
				if (Rfactor > ccThreshold || useZero ) {
					FitAndSubtractPeakBackZero(workImage,weightImage) ;
					Wsum = workImage->Sum() ;
					//fprintf(stderr," kUseIntegrationBZ: ") ;
					xcf=Wsum/MotifSum ;
				}
				break ;
			case kUseIntegration:
				if (Rfactor > ccThreshold || useZero ) {
					Wsum = workImage->Sum() ;
					//fprintf(stderr," kUseIntegration: ") ;
					xcf=Wsum/MotifSum ;
				}
				break ;
		}
	//}

	if (xcf != 0.) {
		peak = new TPeak ;
		peak->pos.x = ixc+ix0 ;
		peak->pos.y = iyc+iy0 ;
		peak->value = xcf ;
		peak->Rfactor = Rfactor ;
		peak->Mean = workImage->Mean() ;
		// Assign the correct lattice index for this peak
		FPoint		U,V,O ;
		O = lattice->GetOrigin() ;
		U = lattice->GetU() ;
		V = lattice->GetV() ;
		float det = V.y*U.x-V.x*U.y ;
		float dx = peak->pos.x-O.x ;
		float dy = peak->pos.y-O.y ;
		if (det != 0.) {
			peak->n = near((V.y*dx-V.x*dy)/det) ;
			peak->m = near((U.x*dy-U.y*dx)/det) ;	
			if (gDebugMode) fprintf(stderr,"\nPeak (n,m) = %d,%d",peak->n,peak->m) ; fflush(stderr) ;
		}
		else {
			delete peak ;
			peak = nil ;
		}
	}
		
	delete workImage ;
	return peak ;
}

float	FindLatticePeak(TImage *image,FPoint *peakPos,TImage *peakMotif,TImage *weightImage,float ccThreshold,float Rfactor,float Mean) {
	long	ixc,iyc ;
	long	ix,iy ;
	TArray	*array = image->GetDataArray() ;

	Point	highestPixel ;
	float	xcf,val,maxVal,peakValue ;
	long	sw = 3 ;
	long	irad = peakMotif->Width()/2 ;
	Mean = 0 ;

	// If we are too close to the edge of the image
	// reurn the midpoint of the area as the peak
	// and the value of zero to indicate no peak
	ix = near(peakPos->x) ;
	iy = near(peakPos->y) ;
	if ( ix <  irad+1 || ix > image->Width() - irad -1 )
		return 0. ;
	if ( iy <  irad+1 || iy > image->Height() - irad - 1)
		return 0. ;
	
	if (irad/2 < sw) sw = irad/2 ;
	
	ixc =	near(peakPos->x) ;
	iyc = 	near(peakPos->y) ;
	maxVal = -1.0e+23 ;
	for(iy = iyc-sw ; iy <= iyc+sw ; iy++) {
		for(ix = ixc-sw ; ix <= ixc+sw ; ix++) {
			val = array->GetPixelAmplitude(ix,iy) ;
			if (val > maxVal) {
				maxVal = val ;
				highestPixel.x = ix ;
				highestPixel.y = iy ;
			}
		}
	}
	ixc = highestPixel.x ;
	iyc = highestPixel.y ;
	short 	ix0,iy0,iyt,ixt;
	float   roff, Wsum, xTol=1e-6;
	if(maxVal < PeakMuchTooLow && !useZero ) {
		xcf = 0 ;
		Rfactor = 0 ;
		return xcf ;
	}
	if( maxVal > PeakIsSaturated ) {
		//Skip pixels which are saturated 
		xcf = 0 ;
		Rfactor = 0 ;
		return xcf ;
	}	
	short	motifWidth = peakMotif->Width() ;
	short	motifHeight = peakMotif->Height() ;
	short	halfWidth = motifWidth/2 ;
	short	halfHeight = motifHeight/2 ;


	
	peakValue = 0. ;
	
	TImage *workImage ;
	workImage = image->Extract(ixc,iyc,motifWidth,motifHeight) ;
	//Trap junk
	if(workImage == nil) {
		xcf = 0 ;
		return xcf ;
		}
	//LDM Note: should we have no zero's here ?
		//Probably not - noise can be +ve & negative!
	if(!FitAndSubtractPeakBackground(workImage,weightImage)) {
		// This fails if the peak is saturated -- set to zero
			xcf = 0 ;
			return xcf ;
		}
		switch (GetIntensityMeasurementMethod()) {
			case kUseXCFMask:
			case kUseXCFAll:
			FitAndSubtractPeakBackMask(workImage,weightImage) ;
			break ;
		}
	TImage *crossCorrelationImage ;
	crossCorrelationImage = CrossCorrelate(peakMotif,workImage) ;
	float xcf0 = *crossCorrelationImage->GetData(0,0) ;
	float xcf0In = xcf0 ;
	float xzero = 0 ; float yzero = 0;
	xcf = crossCorrelationImage->Max(ix0,iy0) ;
	if (ix0 > halfWidth) ix0 -= motifWidth ;
	if (iy0 > halfHeight) iy0 -= motifHeight ;
	//LDM: Use 3 pixels here if (ix0 != 0 || iy0 != 0 && ( ix0 < 2 && iy0 < 2) ) {
	//Also, update the peak position
	roff = ix0*ix0 + iy0*iy0 ;
	if ( roff > 1 && roff < PeakOffTol ) {
	// if ( ( ix0 != 0 || iy0 != 0 ) && ( abs(ix0) < 3 && abs(iy0) < 3) ) {
				// Try with a shift
		delete workImage ;
		workImage = image->Extract(ixc+ix0,iyc+iy0,motifWidth,motifHeight) ;
		//We need to trap bad extracts here
		if(workImage == nil) {
			xcf = 0 ;
			return xcf ;
			}
		if(!FitAndSubtractPeakBackground(workImage,weightImage)){
			//This fails if peak is saturated
				xcf = 0 ;
				return xcf ;
			}
		switch (GetIntensityMeasurementMethod()) {
			case kUseXCFMask:
			case kUseXCFAll:
			FitAndSubtractPeakBackMask(workImage,weightImage) ;
			break ;
		}
		delete crossCorrelationImage ;
		ixt = ix0 ;
		iyt = iy0 ;
		crossCorrelationImage = CrossCorrelate(peakMotif,workImage) ;
		//LDM Note: this is the properly normalized XCF
		float xcf0 = *crossCorrelationImage->GetData(0,0) ;
		float xzero = ix0 ; float yzero = ix0;
		xcf = crossCorrelationImage->Max(ix0,iy0) ;
		if (ix0 > halfWidth) ix0 -= motifWidth ;
		if (iy0 > halfHeight) iy0 -= motifHeight ;
		ix0 = ix0 + ixt ;
		iy0 = iy0 + iyt ;
		roff = ix0*ix0 + iy0*iy0 ;
	}

	delete crossCorrelationImage ;
	//Trap the case where we find a displaced peak 
	if( (roff >= PeakOffTol)  ) {

	// Set to the origin if useZero is set
		if( useZero ) {
				xcf = xcf0In ;
				ix0 = xzero ;
				iy0 = yzero ;
				Wsum = motifWidth * workImage->Stdv() ;
				Rfactor = xcf*Wsum/(Wsum*Wsum+xTol) ;
				roff = 0 ;
				return xcf ;
		}
		switch (GetIntensityMeasurementMethod()) {
			case kUseXCFAll:
				xcf = xcf0In ;
				ix0 = xzero ;
				iy0 = yzero ;
				roff = 0 ;
				Wsum = motifWidth * workImage->Stdv() ;
				Rfactor = xcf*Wsum/(Wsum*Wsum+xTol) ;
				return xcf ;
				break ;
			}
		}
	//Saturated ? ;
	if (maxVal > PeakIsSaturated ) {
		xcf = 0 ;
		Rfactor = 0 ;
		ix0 = xzero ;
		iy0 = yzero ;
		roff = 0 ;
		return xcf ;
	}
		if( (roff >= PeakOffTol) ) {
		xcf = 0 ;
		Rfactor = 0 ;
		if (useZero ) {
				xcf = xcf0In; 
				ix0 = xzero ;
				iy0 = yzero ;
				roff = 0 ;
				Wsum = motifWidth * workImage->Stdv() ;
				Rfactor = xcf*Wsum/(Wsum*Wsum+xTol) ;
		}
		return xcf ;
	}
	else {
		//Wsum = workImage->VectorLength();
		Wsum = motifWidth * workImage->Stdv() ;
		Rfactor = xcf*Wsum/(Wsum*Wsum+xTol) ;
		switch (GetIntensityMeasurementMethod()) {
			case kUseXCF:
			case kUseXCFMask:
			case kUseXCFAll:
				//Let's find the Rfactor (1 for a perfect match)
				if (Rfactor < ccThreshold && !useZero ) xcf = 0. ;
				//if (Rfactor > ccThreshold) {
					//fprintf(stderr," kUseXCF: ") ;
				//}
				break ;
			case kUseIntegrationBZM:
				if (Rfactor > ccThreshold || useZero ) {
					FitAndSubtractPeakBackZeroMask(workImage,weightImage) ;
					Wsum = workImage->Sum() ;
					//fprintf(stderr," kUseIntegrationBZM: ") ;
					xcf=Wsum/MotifSum ;
				}
				break ;
			case kUseIntegrationBZ:
				if (Rfactor > ccThreshold || useZero ) {
					FitAndSubtractPeakBackZero(workImage,weightImage) ;
					Wsum = workImage->Sum() ;
					//fprintf(stderr," kUseIntegrationBZ: ") ;
					xcf=Wsum/MotifSum ;
				}
				break ;
			case kUseIntegration:
				if (Rfactor > ccThreshold || useZero ) {
					Wsum = workImage->Sum() ;
					//fprintf(stderr," kUseIntegration: ") ;
					xcf=Wsum/MotifSum ;
				}
				break ;
		}
	}

	peakPos->x = ixc+ix0 ;
	peakPos->y = iyc+iy0 ;
	peakValue = xcf ;
	Mean = workImage->Mean() ;
	//fprintf(stderr," Returned xcf = %f",xcf) ;
	delete workImage ;
	return peakValue ;
}

float	FindLatticePeak(TImage *image,FPoint *peakPos,float	halfPeakSpacing,PeakCriteria peakCriteria) {
	long	ix,iy,ixc,iyc ;
	TArray	*array = image->GetDataArray() ;
	Point	highestPixel ;

	float	peakValue,meanOfImage, maxVal,val, dx,dy,roff ;
	short	sw = 3 ;	// Search only a radius of sw pixels around nominal position


	// If we are too close to the edge of the image
	// reurn the midpoint of the area as the peak
	// and the value of zero to indicate no peak
	ix = near(peakPos->x) ;
	iy = near(peakPos->y) ;
	if ( ix <  halfPeakSpacing+1 || ix > image->Width() - halfPeakSpacing - 1)
		return 0. ;
	if ( iy <  halfPeakSpacing+1 || iy > image->Height() - halfPeakSpacing - 1)
		return 0. ;

	
	if (peakCriteria == kThresholdOnMeanImageValue) {
		// ix = near(peakPos->x) ;
		// iy = near(peakPos->y) ;
		// peakValue = array->GetPixelAmplitude(ix,iy) ;
		ixc = near(peakPos->x) ;
		iyc = near(peakPos->y) ;
		maxVal = -1.0e+23 ;
		for(iy = iyc-sw ; iy <= iyc+sw ; iy++) {
			for(ix = ixc-sw ; ix <= ixc+sw ; ix++) {
				val = array->GetPixelAmplitude(ix,iy) ;
				if (val > maxVal) {
					maxVal = val ;
					highestPixel.x = ix ;
					highestPixel.y = iy ;
				}
			}
		}

				dx = ixc - highestPixel.x ;
				dy = iyc - highestPixel.y ;
				roff = dx*dx + dy*dy ;
		ixc = highestPixel.x ;
		iyc = highestPixel.y ;

		meanOfImage = image->Mean() ;
			if ( maxVal < meanOfImage || roff > PeakOffTol || maxVal > PeakIsSaturated )
			peakValue = 0. ;
		else {			
			peakPos->x =  ixc ;
			peakPos->y =  iyc ;
			peakValue = maxVal ;
		}
			return	peakValue ;
	}

	if (peakCriteria == kThresholdOnMeanAroundSpot) {

		float	val,maxVal,meanOfCenter,meanOfPerimeter ;
		short	irad = (int)halfPeakSpacing ;
		if (irad*0.5 < sw) sw = irad*0.5 ;

		
		ixc = near(peakPos->x) ;
		iyc = near(peakPos->y) ;
		maxVal = -1.0e+23 ;
		for(iy = iyc-sw ; iy <= iyc+sw ; iy++) {
			for(ix = ixc-sw ; ix <= ixc+sw ; ix++) {
				val = array->GetPixelAmplitude(ix,iy) ;
				if (val > maxVal) {
					maxVal = val ;
					highestPixel.x = ix ;
					highestPixel.y = iy ;
				}
			}
		}
				dx = ixc - highestPixel.x ;
				dy = iyc - highestPixel.y ;
				roff = dx*dx + dy*dy ;
		ixc = highestPixel.x ;
		iyc = highestPixel.y ;
		
		// Find the mean for an area of 5 pixels centered on the highest pixel
		// This should be at least twice that of the mean of the pixel values at a
		// distance of half a lattice spacing away from the spot
		meanOfCenter = (array->GetPixelAmplitude(ixc,iyc) + array->GetPixelAmplitude(ixc-1,iyc) +
			array->GetPixelAmplitude(ixc+1,iyc) + array->GetPixelAmplitude(ixc,iyc-1) + array->GetPixelAmplitude(ixc,iyc+1))/5. ;
		
		short idiag = near(0.7071136*halfPeakSpacing) ;
		meanOfPerimeter = (array->GetPixelAmplitude(ixc-irad,iyc) + array->GetPixelAmplitude(ixc+irad,iyc) +
			array->GetPixelAmplitude(ixc,iyc-irad) + array->GetPixelAmplitude(ixc,iyc+irad) +
			array->GetPixelAmplitude(ixc-idiag,iyc-idiag) + array->GetPixelAmplitude(ixc-idiag,iyc+idiag)
			+ array->GetPixelAmplitude(ixc+idiag,iyc-idiag) + array->GetPixelAmplitude(ixc+idiag,iyc+idiag))/8. ;
		
				// Trap > 8 pixels off
		if (meanOfCenter < 1.4*meanOfPerimeter || roff > PeakOffTol || meanOfCenter > PeakIsSaturated 
			|| meanOfPerimeter > PeakIsSaturated ) 
			peakValue = 0. ;
		else {			
			peakPos->x =  ixc ;
			peakPos->y =  iyc ;
			peakValue = maxVal ;
		}
		return	peakValue ;
	}

	//if (peakCriteria == kThresholdOnStdv) {
		// Think about this one
	//}

	// No criteria. Just return the value at the peak position
	ix = near(peakPos->x) ;
	iy = near(peakPos->y) ;
	peakValue = array->GetPixelAmplitude(ix,iy) ;

	return	peakValue ;

}	
	
void FindCenterOfMass(TImage *image,float *xcm,float *ycm,float	cmRad) {
	float x,y,x0,y0 ;
	
	x0 = *xcm ;
	y0 = *ycm ;
	TArray	*dataArray = image->GetDataArray() ;
	float	sumX = 0. ;
	float	sumY = 0. ;
	float	sumInt = 0. ;
	float	intensity ;
	float	minOfImage = image->Min() ;
	for(y = y0-cmRad; y <= y0+cmRad; y += 1.) {
		for(x = x0-cmRad; x <= x0+cmRad; x += 1.) {
			intensity = dataArray->GetPixelAmplitude(x,y) - minOfImage ;	// so we do not have negative values
			sumInt += intensity ;
			sumX += x*intensity ;
			sumY += y*intensity ;
		}
	}
	if (sumInt > 0.) {
		*xcm = sumX/sumInt ;
		*ycm = sumY/sumInt ;
	}
}

float	FindPeak(TImage	*image,Rect	*areaToSearch,FPoint *peakPos,
		Boolean	threshold) {
	// First find the maximum pixel in the area
	Point	highestPixel ;
	register	long ix,iy ;
	long	Top,Left,Right,Bottom,Width,Height,irad,ixc,iyc,numPixels ;
	float	tolerance,val,sumInt,xCm,yCm,maxVal = -1.0e27 ;
	TArray	*array = image->GetDataArray() ;
	float	dx,dy,x,y,rad,rCm,radsq,xc,yc,mean,stdv,
		valToBePeak,stdvOfImage,meanOfImage,minOfImage ;
	static	int numTries = 0 ;
	
	long	imageWidth = image->Width() ;
	long	imageHeight = image->Height() ;

	Top = areaToSearch->top ;
	Left = areaToSearch->left ;
	Right = areaToSearch->right ;
	Bottom = areaToSearch->bottom ;
 
	xc = peakPos->x = highestPixel.x = ixc = (Left+Right)/2 ;
	yc = peakPos->y = highestPixel.y = iyc = (Top + Bottom)/2 ;

	// If we are too close to the edge of the image
	// reurn the midpoint of the area as the peak
	// and the value of zero to indicate no peak

	if (Top < 0 ) return 0.  ;
	if (Left < 0) return 0.  ;
	if (Right >= imageWidth) return 0.  ;
	if (Bottom >= imageHeight ) return 0.  ;

	meanOfImage = image->Mean() ;

	Width = Right - Left ;
	Height = Bottom - Top ;
	if (Width == 0 || Height == 0) return 0. ;
	numPixels = 0 ;
	mean = 0 ;
	for(iy = Top ; iy <= Bottom ; iy++) {
		for(ix = Left ; ix <= Right ; ix++) {
			val = array->GetPixelAmplitude(ix,iy) ;
			numPixels++ ;
			mean += val ;
			if (val > maxVal) {
				maxVal = val ;
				highestPixel.x = ix ;
				highestPixel.y = iy ;
			}
		}
	}
	if (numPixels == 0) return 0. ;
	mean /= numPixels ;

	if (threshold) {
		if (mean < meanOfImage) return 0. ;
	}
	else {
		/*
		minOfImage = image->Min() ;
		stdvOfImage = image->Stdv() ;
		if (maxVal < minOfImage + stdvOfImage)
			return 0. ;

		// Don't know how to do a peak determination 
		// based on standard deviation yet
		stdv = 0. ;
		for(iy = Top ; iy <= Bottom ; iy++) 
			for(ix = Left ; ix <= Right ; ix++) {
				val = array->GetPixelAmplitude(ix,iy) ;
				stdv += (val-mean)*(val-mean) ;
			}
		stdv /= numPixels ;
		stdv = sqrt(stdv) ;

		valToBePeak = mean + stdv ;
		if (maxVal < valToBePeak) return 0. ;
		*/
	}

	ixc = highestPixel.x ;
	iyc = highestPixel.y ;
	irad = Width ;

	// The peak can be fitted to either a sinc function if this is a 
	// Fourier transform of a real function
	// or a Gaussian if this is an experimental diffraction pattern or
	// an "atomic" image

	switch(gPeakForm) {
		case kGaussianPeak:
			// Gaussian peak
			FitGaussianPeak(array,ixc,iyc,irad,xCm,yCm,val) ;
			break ;
		case kSincPeak:
			// Sinc Peak
			FitSincPeak(array,ixc,iyc,irad,xCm,yCm,val) ;
			break ;
		case kCMPeak:
			FindCMPeak(array,ixc,iyc,irad,xCm,yCm,val) ;
			break ;
		case kRawPeak:
			xCm = ixc ;
			yCm = iyc ;
			val = maxVal ;
			break ;
		case kHanningPeak:		
			float	f1 = 1.,f2= 1.,dk,dl,ampX,ampXLeft,ampY,ampYAbove;
			int		alpha = 1, beta = 1 ;
			float	pi = 2.0*::asin(1.0) ;
			ampXLeft = array->GetPixelAmplitude(ixc-1,iyc) ;
			ampX = array->GetPixelAmplitude(ixc+1,iyc) ;
			ampYAbove = array->GetPixelAmplitude(ixc,iyc-1) ;
			ampY = array->GetPixelAmplitude(ixc,iyc+1) ;
			if (ampXLeft > ampX) {
				ampX = ampXLeft ;
				alpha = -1 ;
			}
			if (ampYAbove > ampY) {
				ampY = ampYAbove ;
				beta = -1 ;
			}
			dk = alpha*(2*ampX-maxVal)/(ampX+maxVal) ;
			dl = beta*(2*ampY-maxVal)/(ampY+maxVal) ;
			xCm = ixc + dk ;
			yCm = iyc + dl ;
			if (fabs(dk) > 0.001)
				f1 = (1-dk*dk)*pi*dk/::sin(pi*dk) ;
			if (fabs(dl) > 0.001)
				f2 = (1-dl*dl)*pi*dl/::sin(pi*dl) ;
			val = maxVal*f1*f2 ;
			break ;
	}


	// See how far from the initial search position we were
	// We don't allow more than tolerance pixels from the starting points
	tolerance = MIN(Width/2,Height/2) ;
	dx = sqrt((xCm-xc)*(xCm-xc)) ;
	dy = sqrt((yCm-yc)*(yCm-yc)) ;
	if (dx > tolerance || dy > tolerance) {
		peakPos->x = xc ;
		peakPos->y = yc ;
		return 0. ;
	}
	peakPos->x = xCm ;
	peakPos->y = yCm ;
	return val ;
}	

float	FindTEMPeak(TImage *image,Rect *areaToSearch,FPoint	*peakPos) {
	// First find the maximum pixel in the area
	Point	highestPixel ;
	register	long ix,iy ;
	long	Top,Left,Right,Bottom,Width,Height,irad,ixc,iyc,numPixels ;
	float	tolerance,val,xCm,yCm,maxVal = -1.0e27 ;
	TArray	*array = image->GetDataArray() ;
	float	dx,dy,xc,yc,mean,meanOfImage ;
	
	long	imageWidth = image->Width() ;
	long	imageHeight = image->Height() ;

	Top = areaToSearch->top ;
	Left = areaToSearch->left ;
	Right = areaToSearch->right ;
	Bottom = areaToSearch->bottom ;
 
	xc = peakPos->x = highestPixel.x = ixc = (Left+Right)/2 ;
	yc = peakPos->y = highestPixel.y = iyc = (Top + Bottom)/2 ;

	// If we are too close to the edge of the image
	// reurn the midpoint of the area as the peak
	// and the value of zero to indicate no peak

	if (Top < 0 ) return 0.  ;
	if (Left < 0) return 0.  ;
	if (Right >= imageWidth) return 0.  ;
	if (Bottom >= imageHeight ) return 0.  ;

	meanOfImage = image->Mean() ;

	Width = Right - Left ;
	Height = Bottom - Top ;
	if (Width == 0 || Height == 0) return 0. ;
	numPixels = 0 ;
	mean = 0 ;
	for(iy = Top ; iy <= Bottom ; iy++) {
		for(ix = Left ; ix <= Right ; ix++) {
			val = array->GetPixelAmplitude(ix,iy) ;
			numPixels++ ;
			mean += val ;
			if (val > maxVal) {
				maxVal = val ;
				highestPixel.x = ix ;
				highestPixel.y = iy ;
			}
		}
	}
	if (numPixels == 0) return 0. ;
	mean /= numPixels ;

	if (mean < meanOfImage) return 0. ;
	ixc = highestPixel.x ;
	iyc = highestPixel.y ;
	irad = Width ;


	float	f1 = 1.,f2= 1.,dk,dl,ampX,ampXLeft,ampY,ampYAbove;
	int		alpha = 1, beta = 1 ;
	float	pi = 2.0*::asin(1.0) ;
	ampXLeft = array->GetPixelAmplitude(ixc-1,iyc) ;
	ampX = array->GetPixelAmplitude(ixc+1,iyc) ;
	ampYAbove = array->GetPixelAmplitude(ixc,iyc-1) ;
	ampY = array->GetPixelAmplitude(ixc,iyc+1) ;
	if (ampXLeft > ampX) {
		ampX = ampXLeft ;
		alpha = -1 ;
	}
	if (ampYAbove > ampY) {
		ampY = ampYAbove ;
		beta = -1 ;
	}
	dk = alpha*(2*ampX-maxVal)/(ampX+maxVal) ;
	dl = beta*(2*ampY-maxVal)/(ampY+maxVal) ;
	xCm = ixc + dk ;
	yCm = iyc + dl ;
	if (::fabs(dk) > 0.001)
		f1 = (1-dk*dk)*pi*dk/::sin(pi*dk) ;
	if (::fabs(dl) > 0.001)
		f2 = (1-dl*dl)*pi*dl/::sin(pi*dl) ;
	val = maxVal*f1*f2 ;

	
	// See how far from the initial search position we were
	// We don't allow more than tolerance pixels from the starting points
	tolerance = MIN(Width/2,Height/2) ;
	dx = sqrt((xCm-xc)*(xCm-xc)) ;
	dy = sqrt((yCm-yc)*(yCm-yc)) ;
	if (dx > tolerance || dy > tolerance) {
		peakPos->x = xc ;
		peakPos->y = yc ;
		return 0. ;
	}
	
	peakPos->x = xCm ;
	peakPos->y = yCm ;
	return val ;
}
	
void FindCMPeak(TArray *array,int ixc,int iyc,int irad,
		float &xc,float &yc,float &peakValue) {

	float 	xCm,yCm,rCm,sumInt,val,dx,dy ;
	long	Top,Left,Right,Bottom,ix,iy,imageWidth,imageHeight ;
	// Do a center of mass refinement on the peak using irad as the
	// radius of the refinement
	xCm = 0. ;
	yCm = 0. ;
	rCm = 0. ;
	sumInt = 0. ;

	Top = iyc - irad ;
	Left = ixc - irad ;
	Right = ixc + irad ;
	Bottom = iyc + irad ;

	imageWidth = array->GetWidth() ;
	imageHeight = array->GetHeight() ;

	if (Top < 0 ) Top = 0 ;
	if (Left < 0) Left = 0 ;
	if (Right >= imageWidth) Right = imageWidth - 1 ;
	if (Bottom >= imageHeight) Bottom = imageHeight - 1  ;

	for(iy = Top; iy <= Bottom; iy++) {
		for(ix = Left; ix <= Right; ix++) {
			val = array->GetPixelAmplitude(ix,iy) ;	
			xCm += val*ix ;
			yCm += val*iy ;
			sumInt += val ;
		}
	}
	if (sumInt > 1e-5 ) {
		xCm /= sumInt ;
		yCm /= sumInt ;
		ix = (int)xCm ;
		dx = xCm - ix ;
		iy = (int)yCm ;
		dy = yCm - iy ;
		if (ix > 0 && ix < imageWidth -2 && 
					iy > 0 && iy < imageHeight -2) {
			val  = (1.-dy)*((1-dx)*array->GetPixelAmplitude(ix,iy) 
					+ dx*array->GetPixelAmplitude(ix+1,iy)) +
				dy*((1-dx)*array->GetPixelAmplitude(ix,iy+1)+
					dx*array->GetPixelAmplitude(ix+1,iy+1)) ;
		}
		else {
			val = array->GetPixelAmplitude(ix,iy) ;
		}
	}
	else {
		xCm = (Left + Right)/2 ;
		yCm = (Top + Bottom)/2 ;
		val = 0. ;
	}
	xc = xCm ;
	yc = yCm ;
	peakValue = val ;
}
void FindCMPeakWrapped(TArray *array,int ixc,int iyc,int irad,
		float &xc,float &yc,float &peakValue) {

	float 	xCm,yCm,rCm,sumInt,val,dx,dy ;
	long	ix,iy,imageWidth,imageHeight ;
	long	ixp, iyp ;
	// Do a center of mass refinement on the peak using the irad as the
	// radius of the refinement
	xCm = 0. ;
	yCm = 0. ;
	rCm = 0. ;
	sumInt = 0. ;

	imageWidth = array->GetWidth() ;
	imageHeight = array->GetHeight() ;

	for(iy = -irad ; iy <= irad; iy++) {
			iyp = iyc +iy ;
			if ( iyp >= imageWidth )
				iyp = iyp - imageWidth ;
			if ( iyp < 0 )
				iyp = iyp + imageWidth ;
			for(ix = -irad; ix <= irad; ix++) {
			ixp = ixc +ix ;
			if ( ixp >= imageWidth )
				ixp = ixp - imageWidth ;
			if ( ixp < 0 )
				ixp = ixp + imageWidth ;
			val = array->GetPixelAmplitude(ixp,iyp) ;	
			xCm += val*ix ;
			yCm += val*iy ;
			sumInt += val ;
		}
	}
	if (sumInt >1e-3 ) {
		xCm /= sumInt ;
		yCm /= sumInt ;
		ix = (int)xCm ;
		dx = xCm - ix ;
		iy = (int)yCm ;
		dy = yCm - iy ;
		iyp = iyc +iy ;
		if ( iyp >= imageWidth )
			iyp = iyp - imageWidth ;
		if ( iyp < 0 )
			iyp = iyp + imageWidth ;
		ixp = ixc +ix ;
		if ( ixp >= imageWidth )
			ixp = ixp - imageWidth ;
		if ( ixp < 0 )
			ixp = ixp + imageWidth ;
		long ixp1 = ixp + 1 ;
		long iyp1 = iyp + 1 ;
		if ( iyp1 >= imageWidth )
			iyp1 = iyp1 - imageWidth ;
		if ( iyp1 < 0 )
			iyp1 = iyp1 + imageWidth ;
		if ( ixp1 >= imageWidth )
			ixp1 = ixp1 - imageWidth ;
		if ( ixp1 < 0 )
			ixp1 = ixp1 + imageWidth ;
			val  = (1.-dy)*((1-dx)*array->GetPixelAmplitude(ixp,iyp) 
					+ dx*array->GetPixelAmplitude(ixp1,iyp)) +
				dy*((1-dx)*array->GetPixelAmplitude(ixp,iyp1)+
					dx*array->GetPixelAmplitude(ixp1,iyp1)) ;
		}
	else {
		xc = 0 ;
		yc = 0 ;
		val = 0 ;
	}
	xc = xCm +ixc;
	yc = yCm +iyc;
	peakValue = val ;
}
// ---------------------------------------------------------------------	
// This	routine	is actually	not	implemented	yet	since it only retrieves
// the peak	at the incoming	pixel position
// ---------------------------------------------------------------------	
void FitGaussianPeak(TArray	*array,int ixc,int iyc,int irad,
		float &xc,float &yc,float &peakValue) {

	float 	xCm,yCm,rCm,sumInt,val ;
	long	Top,Left,Right,Bottom,imageWidth,imageHeight ;
	// Do a center of mass refinement on the peak using the width/2 as the
	// radius of the refinement
	xCm = 0. ;
	yCm = 0. ;
	rCm = 0. ;
	sumInt = 0. ;

	Top = iyc - irad ;
	Left = ixc - irad ;
	Right = ixc + irad ;
	Bottom = iyc + irad ;

	imageWidth = array->GetWidth() ;
	imageHeight = array->GetHeight() ;

	if (Top < 0 ) Top = 0 ;
	if (Left < 0) Left = 0 ;
	if (Right >= imageWidth) Right = imageWidth - 1 ;
	if (Bottom >= imageHeight) Bottom = imageHeight - 1  ;

	val = array->GetPixelAmplitude((long)ixc,(long)iyc) ;	

	xc = ixc ;
	yc = iyc ;
	peakValue = val ;
}

void	ExtractAmplitudePhase(TArray *array,long &n,long &m,float &amp,float &phase) {

	long	ix0,iy0,ixc,iyc,ix,iy,in,im ;
	long	irad=1 ;
	float	val,maxVal ;
	float	amplitude,phs ;

	
	ix0 = ixc = in = n ;
	iy0 = iyc = im = m ;

	maxVal = -1.0e+10 ;

	// Find the highest pixel in the neighborhood
	for(iy = iy0-irad ; iy <= iy0+irad ; iy++) {
		for(ix = ix0-irad ; ix <= ix0+irad ; ix++) {
			val = array->GetPixelAmplitude(ix,iy) ;
			if (val > maxVal) {
				maxVal = val ;
				ixc = ix ;
				iyc = iy ;
			}
		}
	}

	switch(gPeakForm) {
		case kGaussianPeak:
			// Gaussian peak
			amplitude = array->GetPixelAmplitude(ix,iy) ;
			phs = array->GetPixelPhase(ixc,iyc) ;
			break ;
		case kSincPeak:
			// Sinc Peak
			amplitude = array->GetPixelAmplitude(ix,iy) ;
			phs = array->GetPixelPhase(ixc,iyc) ;
			break ;
		case kCMPeak:
			amplitude = array->GetPixelAmplitude(ix,iy) ;
			phs = array->GetPixelPhase(ixc,iyc) ;
			break ;
		case kRawPeak:
			// Integrate the amplitude in a 3 by 3 neighborhood and subtract background
			// which lies at at the edge of a 6 by boundary
			amplitude = 0. ;
			float background ;

			for(ix = ixc-1;ix < ixc+2;ix++)
				for(iy = iyc-1; iy < iyc+2; iy++) {
					amplitude += array->GetPixelAmplitude(ix,iy) ;
			}
			// Find background
			background = (array->GetPixelAmplitude(in-6,im) + array->GetPixelAmplitude(in+6,im)
					+ array->GetPixelAmplitude(in,im-6) + array->GetPixelAmplitude(in,im+6))/4 ;

			amplitude -= background ;
			amp = amplitude ;
			phs = array->GetPixelPhase(ixc,iyc) ;
			break ;
		case kHanningPeak:		
			float	f1 = 1.,f2= 1.,dk,dl,ampX,ampXLeft,ampY,ampYAbove;
			int		alpha = 1, beta = 1 ;
			float	pi = 2.0*::asin(1.0) ;
			ampXLeft = array->GetPixelAmplitude(ixc-1,iyc) ;
			ampX = array->GetPixelAmplitude(ixc+1,iyc) ;
			ampYAbove = array->GetPixelAmplitude(ixc,iyc-1) ;
			ampY = array->GetPixelAmplitude(ixc,iyc+1) ;
			if (ampXLeft > ampX) { 
				ampX = ampXLeft ;
				alpha = -1 ;
			}
			if (ampYAbove > ampY) {
				ampY = ampYAbove ;
				beta = -1 ;
			}
			dk = alpha*(2*ampX-maxVal)/(ampX+maxVal) ;
			dl = beta*(2*ampY-maxVal)/(ampY+maxVal) ;
			if (fabs(dk) > 0.001)
				f1 = (1-dk*dk)*pi*dk/::sin(pi*dk) ;
			if (fabs(dl) > 0.001)
				f2 = (1-dl*dl)*pi*dl/::sin(pi*dl) ;
			amplitude = maxVal*f1*f2 ;
			phs = array->GetPixelPhase(ixc,iyc) + pi*dk + pi*dl ; 
			break ;
	}

	amp = amplitude ;	
	phase = phs ;	
	n = ixc ;
	m = iyc ;
}


void FitSincPeak(TArray	*array,int ixc,int iyc,int irad,
		float &xc,float &yc,float &peakValue) {
}

TImage *FindAveragePeakMotif(TImage	*image,TLattice	*lattice,TPeakArray	*peakArray,short peakRadius) {
	TImage	*averagePeakMotif = nil ;
	short	i,j,x,y,radsq,outerRadius ;
	short	ix0,iy0 ;
	float	roff, x0, y0, WMax,innerRadius,motifRadiusSq;
	float xCm, yCm,dx,dy ;
	int irad=4 ;
	
	FPoint	O = lattice->GetOrigin() ;
	FPoint	U = lattice->GetU() ;
	FPoint	V = lattice->GetV() ;

	float	du = 0.5*sqrt(U.x*U.x+U.y*U.y) ;
	float	dv = 0.5*sqrt(V.x*V.x+V.y*V.y) ;
	float	halfSpacing = du < dv ? du : dv ;
	//LDM: changed this to 2/3 of width
	if (peakRadius == 0) {
		//LDM: make inner 2/3 of outer, picture size 3/4 of distance
		outerRadius = near(halfSpacing*Psizer) ; // was 1
		//innerRadius = near(outerRadius/1.5) ; // was 1.5
	}
	else {
		outerRadius = 2*peakRadius ;
		//innerRadius = near( ((float) peakRadius)*(4./3.)) ; // was peakRadius
	}
	innerRadius = outerRadius*RadSize ;
	
	short	imageWidth = 2*outerRadius,imageHeight = 2*outerRadius ;
	short	numPeaksToAverage = 0 ;
	TImage	*workImage ;
	
	averagePeakMotif = new TImage(imageWidth,imageHeight) ;
	
	TImage	*weightImage = new TImage(imageWidth,imageHeight) ;
	TImage  *TightWeight = new TImage(imageWidth,imageHeight) ;
	motifRadiusSq = innerRadius*innerRadius ;
	float motifRadius2 = (innerRadius + 2.5)*(innerRadius + 2.5); ;
	//short motifRadius2 = near(outerRadius*outerRadius) ;
	// Set the weights. For now just a circular mask
	register float *p1 = weightImage->GetData() ;
	register float *p2 = TightWeight->GetData() ;
	for(j = 0; j < imageHeight; j++) {
		y = j - outerRadius ;
		for(i = 0 ; i < imageWidth; i++) {
			x = i - outerRadius ;
			radsq = x*x+y*y ;
			if (radsq > motifRadiusSq) {
				*p1++ = 1. ;
                                //*p2++ = 1. ;
			}
			else {
				*p1++ = 0. ;
                                //*p2++ = 0. ;
			}
		if (radsq > motifRadiusSq && radsq < motifRadius2 ) 
			*p2++ = 1. ;
		else {
			if (radsq > motifRadiusSq ) 
				*p2++ = 0.5;
			else
				*p2++ = 0. ;
			}
		}
	}
	
	//We are only getting an approximate motif here
	//Therefore, let's not be too anal about the background
	long HoldNumbBiwt = NumbBiwt ;
	NumbBiwt = 5 ;
	*averagePeakMotif = 0. ;
	//short	startPeakIndex = near(0.1*peakArray->mNumPeaks) ;
	short   startPeakIndex = 0 ;
	short	endPeakIndex = near(0.8*peakArray->mNumPeaks) ;
	for(i = startPeakIndex; i <=  endPeakIndex; i++) {
		workImage = image->Extract(peakArray->mX[i],peakArray->mY[i],imageWidth,imageHeight) ;
		if (workImage != nil) {
			//WMin = workImage->Min(ix0,iy0) ;
			WMax = workImage->Max(ix0,iy0); // -WMin ;
			TArray	*array = workImage->GetDataArray() ;
			FindCMPeak(array,ix0,iy0,irad,xCm,yCm,WMax);
			dx = xCm - imageWidth/2 ;
			dy = yCm - imageHeight/2 ;
			//dx = ix0 - imageWidth/2 ;
			//dy = iy0 - imageHeight/2 ;
			roff = dx*dx + dy*dy ;
			if ( (roff < PeakOffTol) && (WMax < PeakIsSaturated) && (WMax > PeakTooLow) ) {
				if ( (dx > 0.1) ||  (dx < -0.1) || (dy > 0.1) || (dy < -0.1) ) {
					x0 = near(peakArray->mX[i]+dx) ;
					y0 = near(peakArray->mY[i]+dy) ;
					workImage = image->Extract(x0,y0,imageWidth,imageHeight) ;
					if (workImage == nil ) {
						WMax = -99 ;
						}
					else
						WMax = workImage->Max() ;
					}
			//fprintf(stderr,"Raw extract %d %d %d %f %f\n",imageWidth,ix0,iy0,roff,WMax);
					// Trap saturated images
					if ( (WMax < PeakIsSaturated) && (WMax > PeakTooLow) ) {
							// Let there be +ve and negative values - average them
							if(FitAndSubtractPeakBackground(workImage,weightImage)) {
							numPeaksToAverage++ ;
							//Probably need two passes
							float WMax = workImage->Max() ;
							// Trap negative images
							if ( WMax > PeakTooLow ) {
									//How to add ?
									//Try sqrt weight, makes some sense... ?
									*averagePeakMotif += *workImage ; // /sqrt(WMax) ;
							}
						}
				   }
			}
		delete workImage ;
		}
	}
	// Be a bit anal about the background
	NumbBiwt = HoldNumbBiwt*2 ;
	// LDM: remove background, and mask as well
	// Apply inversion symmetry to peaks ?
    if(FitAndSubtractPeakBackground(averagePeakMotif,weightImage)) {
		}
    if(FitAndSubtractPeakBackground(averagePeakMotif,weightImage)) {
		}
    if(FitAndSubtractPeakBackground(averagePeakMotif,TightWeight)) {
		}	
	FitAndSubtractPeakBackZeroMask(averagePeakMotif,TightWeight) ;
	FitAndSubtractPeakBackZeroMask(averagePeakMotif,weightImage) ;
	//Restore NumbBiwt
	NumbBiwt = HoldNumbBiwt ;
	// LDM: normalize the motif
	//*averagePeakMotif /= averagePeakMotif->Sum() ;
	*averagePeakMotif /= averagePeakMotif->VectorLength() ;
	//Sum of intensities for normalization, so integration & xcf modes give similar numbers
	MotifSum = averagePeakMotif->Sum() ;
	//WMax = averagePeakMotif->Max(ix0,iy0) ;
	//fprintf(stderr,"Peak max %f %d %d Size %d\n",WMax,ix0,iy0,imageWidth);

	//I think I was wrong, and we really should subtract the mean!
	*averagePeakMotif -= averagePeakMotif->Mean() ;
	*averagePeakMotif /= averagePeakMotif->VectorLength() ;
	delete weightImage ;	
	delete TightWeight ;
	return averagePeakMotif ;
}

TImage *FindAveragePeakMotifNoIter(TImage	*image,TLattice	*lattice,TPeakArray	*peakArray,short peakRadius) {
	TImage	*averagePeakMotif = nil ;
	short	i,j,x,y,radsq,outerRadius ;
	long	ix0,iy0 ;
	float	WMax,innerRadius,motifRadiusSq ;
	
	FPoint	O = lattice->GetOrigin() ;
	FPoint	U = lattice->GetU() ;
	FPoint	V = lattice->GetV() ;

	float	du = 0.5*sqrt(U.x*U.x+U.y*U.y) ;
	float	dv = 0.5*sqrt(V.x*V.x+V.y*V.y) ;
	float	halfSpacing = du < dv ? du : dv ;
	//LDM: changed this to 2/3 of width
	if (peakRadius == 0) {
		//LDM: make inner 2/3 of outer, picture size 3/4 of distance
		//innerRadius = near(halfSpacing*0.5*Psizer) ; // was 1.5
		outerRadius = near(halfSpacing*Psizer) ; // was 1
	}
	else {
		//innerRadius = near( ((float) peakRadius)*(4./3.)) ; // was peakRadius
		outerRadius = 2*peakRadius ;
	}
	innerRadius = outerRadius*RadSize ;

	short	imageWidth = 2*outerRadius,imageHeight = 2*outerRadius ;
	short	numPeaksToAverage = 0 ;
	TImage	*workImage ;
	
	averagePeakMotif = new TImage(imageWidth,imageHeight) ;
	
	TImage	*weightImage = new TImage(imageWidth,imageHeight) ;
	TImage  *TightWeight = new TImage(imageWidth,imageHeight) ;
	motifRadiusSq = innerRadius*innerRadius ;
	float motifRadius2 = (innerRadius + 2)*(innerRadius + 2); ;
	//short motifRadius2 = near(outerRadius*outerRadius) ;
	// Set the weights. For now just a circular mask
	register float *p1 = weightImage->GetData() ;
	register float *p2 = TightWeight->GetData() ;
	for(j = 0; j < imageHeight; j++) {
		y = j - outerRadius ;
		for(i = 0 ; i < imageWidth; i++) {
			x = i - outerRadius ;
			radsq = x*x+y*y ;
			if (radsq > motifRadiusSq) {
				*p1++ = 1. ;
                                //*p2++ = 1. ;
			}
			else {
				*p1++ = 0. ;
                                //*p2++ = 0. ;
			}
		if (radsq > motifRadiusSq && radsq < motifRadius2 ) 
			*p2++ = 1. ;
		else {
			if (radsq > motifRadiusSq ) 
				*p2++ = 0.5;
			else
				*p2++ = 0. ;
			}
		}
	}	
	long HoldNumbBiwt = NumbBiwt ;
	*averagePeakMotif = 0. ;
	short	startPeakIndex = 0 ; //near(0.1*peakArray->mNumPeaks) ;
	short	endPeakIndex =  near(0.9*peakArray->mNumPeaks) ;
	ix0 = imageWidth/2 ; 
	iy0 = imageHeight/2 ;
	for(i = startPeakIndex; i <=  endPeakIndex; i++) {
		workImage = image->Extract(peakArray->mX[i],peakArray->mY[i],imageWidth,imageHeight) ;
		//No iterations for position here
		if (workImage == nil ) {
					WMax = -99 ;
					}
		else
			WMax = *workImage->GetData(ix0,iy0) ;
		if (  (WMax < PeakIsSaturated) && (WMax > PeakTooLow) ) {
					// Let there be +ve and negative values - average them
					if(FitAndSubtractPeakBackground(workImage,weightImage)) {
						numPeaksToAverage++ ;
						//Probably need two passes
						WMax = workImage->Max() ;
						// Trap negative images
						if ( WMax > PeakTooLow && WMax < PeakIsSaturated ) {
								//How to add ?
								//Try sqrt weight, makes some sense... ?
								*averagePeakMotif += *workImage; // /sqrt(WMax) ;
						}
					}
		}			
	}
	delete workImage ;

	// LDM: remove background, and mask as well
	// Apply inversion symmetry to peaks ?
	// Be anal about the background here
	if(NumbBiwt < 20) NumbBiwt = 20 ;
    if(FitAndSubtractPeakBackground(averagePeakMotif,weightImage)) {
		}
    if(FitAndSubtractPeakBackground(averagePeakMotif,TightWeight)) {
		}	
	FitAndSubtractPeakBackZeroMask(averagePeakMotif,TightWeight) ;
	FitAndSubtractPeakBackZeroMask(averagePeakMotif,weightImage) ;
	NumbBiwt = HoldNumbBiwt ;
	// LDM: normalize the motif
	//*averagePeakMotif /= averagePeakMotif->Sum() ;
	*averagePeakMotif /= averagePeakMotif->VectorLength() ;
	//WMax = averagePeakMotif->Max(ix0,iy0) ;
	//fprintf(stderr,"Peak max %f %d %d Size %d\n",WMax,ix0,iy0,imageWidth);
	//Sum of intensities for normalization so integration methods give similar numbers
	MotifSum = averagePeakMotif->Sum() ;

	//I think I was wrong, and we really should subtract the mean!
	*averagePeakMotif -= averagePeakMotif->Mean() ;
	*averagePeakMotif /= averagePeakMotif->VectorLength() ;

	delete weightImage ;	
	delete TightWeight ;
	return averagePeakMotif ;
}

bool	FitAndSubtractPeakBackground(TImage	*image,TImage *weightImage)	{
	// We will fit a plane of the form ax+by+c to the pixels at the outer region of the image
	// and subtract the plane from the entire image
	
	float	a,b,c,sumXI,sumYI,sumXX,sumYY,sumX,sumY,sumI,sumXY,sumW ;
	register 	short	x,y ;
	register	float 	intensity,weight ;
	short	width = image->Width() ;
	short	height = image->Height() ;
	TMatrix A ;
	TVector B,C	;
		
	register	float *p1 = image->GetData() ;
	register	float *p2 = weightImage->GetData() ;
	sumXI = sumYI = sumXX = sumYY = sumX = sumY = sumXY = sumI = sumW = 0. ;
	float pmax = -1e5 ;
	for(y = 0; y <height; y++) {
		for(x = 0 ;x < width; x++) {
			weight = *p2++ ;
			intensity = *p1++ ;
			if (intensity > pmax) pmax = intensity ;
			if(weight > 1e-3){
				sumI += intensity  ;
				sumX += x ;
				sumY += y ;
				sumXX += x*x ;
				sumYY += y*y ;
				sumXY += x*y ;
				sumXI += x*intensity ;
				sumYI += y*intensity ;
				sumW += 1 ;
				}
		}
	}
	if (pmax > PeakIsSaturated) {
		return false ;
	}
	A.m[0][0] = sumXX ;
	A.m[1][0] = sumXY ;
	A.m[2][0] = sumX ;
	A.m[0][1] = sumXY ;
	A.m[1][1] = sumYY;
	A.m[2][1] = sumY ;
	A.m[0][2] = sumX ;
	A.m[1][2] = sumY ;
	A.m[2][2] = sumW ;
	B.a[0] = sumXI ;
	B.a[1] = sumYI ;
	B.a[2] = sumI ;
	if (Solve3LinearEquation(&A,&B,&C)) {
		//Use the robust form here for better accuracy ?
		//Trap case with nothing relevant
		a = C.a[0] ;
		b = C.a[1] ;
		c = C.a[2] ;
		if ( fabsf(a) > 1e-6|| fabsf(b) > 1e-6 || fabsf(c) > 1e-6) {
		FitBiweightBackground(image,weightImage);
		a = Bfita ;		
		b = Bfitb ;
		c = Bfitc ;
			}	
		// subtract the ramp
		p1 = image->GetData() ;
		for(y = 0; y <height; y++) {
			float by = b*y + c ;
			for(x = 0 ;x < width; x++) {
				*p1++ -= (a*x+by) ;
			}
		}
	}
	return true ;
}

void	FitAndSubtractPeakBackZero(TImage *image,TImage	*weightImage) {
	// We will fit a plane of the form ax+by+c to the pixels at the outer region of the image
	// and subtract the plane from the entire image
	// Revised by LDM to enforce positivity in result
	
	float	a,b,c,sumXI,sumYI,sumXX,sumYY,sumX,sumY,sumI,sumXY,sumW ;
	register 	short	x,y ;
	register	float 	intensity,weight ;
	short	width = image->Width() ;
	short	height = image->Height() ;
	TMatrix A ;
	TVector B,C	;
		
	register	float *p1 = image->GetData() ;
	register	float *p2 = weightImage->GetData() ;
	sumXI = sumYI = sumXX = sumYY = sumX = sumY = sumXY = sumI = sumW = 0. ;
	for(y = 0; y <height; y++) {
		for(x = 0 ;x < width; x++) {
			weight = *p2++ ;
			intensity = *p1++ ;
			if(weight > 1e-3){
				sumI += intensity  ;
				sumX += x ;
				sumY += y ;
				sumXX += x*x ;
				sumYY += y*y ;
				sumXY += x*y ;
				sumXI += x*intensity ;
				sumYI += y*intensity ;
				sumW += 1 ;
				}
		}
	}
	A.m[0][0] = sumXX ;
	A.m[1][0] = sumXY ;
	A.m[2][0] = sumX ;
	A.m[0][1] = sumXY ;
	A.m[1][1] = sumYY;
	A.m[2][1] = sumY ;
	A.m[0][2] = sumX ;
	A.m[1][2] = sumY ;
	A.m[2][2] = sumW ;
	B.a[0] = sumXI ;
	B.a[1] = sumYI ;
	B.a[2] = sumI ;
	if (Solve3LinearEquation(&A,&B,&C)) {
		//Use the robust form here for better accuracy ?
		//Trap case with nothing relevant
		a = C.a[0] ;
		b = C.a[1] ;
		c = C.a[2] ;
		if ( fabsf(a) > 1e-6 || fabsf(b) > 1e-6 || fabsf(c) > 1e-6) {
		FitBiweightBackground(image,weightImage);
		a = Bfita ;		
		b = Bfitb ;
		c = Bfitc ;
			}	
		// subtract the ramp
		p1 = image->GetData() ;
		for(y = 0; y <height; y++) {
			float by = b*y+c;
			for(x = 0 ;x < width; x++) {
				//LDM      Reuse sumI here, and set with max of zero (illogical otherwise)!)
							   sumI = *p1 - (a*x+by);
							   if (sumI > 0.) {
									*p1++ = sumI ;
							   }
							   else {
									*p1++ = 0 ;
							   }
				}
			}
		}
	else
		//Trap case when the fit fails
		{
		p1 = image->GetData() ;
		for(y = 0; y <height; y++) {
			for(x = 0 ;x < width; x++) {
				//LDM      Reuse sumI here, and set with max of zero (illogical otherwise)!)
							   sumI = *p1;
							   if (sumI > 0.) {
									*p1++ = sumI ;
							   }
							   else {
									*p1++ = 0 ;
							   }
				}
			}
		}	
}
void	FitAndSubtractPeakBackZeroMask(TImage *image,TImage	*weightImage) {
	// We will fit a plane of the form ax+by+c to the pixels at the outer region of the image
	// and subtract the plane from the entire image
	// Revised by LDM to enforce positivity in result
	
	float	a,b,c,sumXI,sumYI,sumXX,sumYY,sumX,sumY,sumI,sumXY,sumW ;
	register 	short	x,y ;
	register	float 	intensity,weight ;
	short	width = image->Width() ;
	short	height = image->Height() ;
	TMatrix A ;
	TVector B,C	;
		
	register	float *p1 = image->GetData() ;
	register	float *p2 = weightImage->GetData() ;
	sumXI = sumYI = sumXX = sumYY = sumX = sumY = sumXY = sumI = sumW = 0. ;
	for(y = 0; y <height; y++) {
		for(x = 0 ;x < width; x++) {
			weight = *p2++ ;
			intensity = *p1++ ;
			if(weight > 1e-3){
				sumI += intensity  ;
				sumX += x ;
				sumY += y ;
				sumXX += x*x ;
				sumYY += y*y ;
				sumXY += x*y ;
				sumXI += x*intensity ;
				sumYI += y*intensity ;
				sumW += 1 ;
				}
		}
	}
	A.m[0][0] = sumXX ;
	A.m[1][0] = sumXY ;
	A.m[2][0] = sumX ;
	A.m[0][1] = sumXY ;
	A.m[1][1] = sumYY;
	A.m[2][1] = sumY ;
	A.m[0][2] = sumX ;
	A.m[1][2] = sumY ;
	A.m[2][2] = sumW ;
	B.a[0] = sumXI ;
	B.a[1] = sumYI ;
	B.a[2] = sumI ;
	if (Solve3LinearEquation(&A,&B,&C)) {
		//Use the robust form here for better accuracy ?
		//Trap case with nothing relevant
		a = C.a[0] ;
		b = C.a[1] ;
		c = C.a[2] ;
		if ( fabsf(a) > 1e-6 || fabsf(b) > 1e-6 || fabsf(c) > 1e-6) {
		FitBiweightBackground(image,weightImage);
		a = Bfita ;		
		b = Bfitb ;
		c = Bfitc ;
			}	
		// subtract the ramp
		p1 = image->GetData() ;
		p2 = weightImage->GetData() ;
		for(y = 0; y <height; y++) {
			float by = b*y + c ;
			for(x = 0 ;x < width; x++) {
				//LDM      Set with max of zero and mask
				sumI = *p1 - (a*x+by);
				weight = *p2++ ;
				if (sumI > 0. && weight < 0.9 ) {
						*p1++ = sumI ;
				}
				else {
						*p1++ = 0 ;
				}
			}
		}
	}
	else {
		//Trap case when no fit
		p1 = image->GetData() ;
		p2 = weightImage->GetData() ;
		for(y = 0; y <height; y++) {
			for(x = 0 ;x < width; x++) {
				//LDM      Set with max of zero and mask
				sumI = *p1;
				weight = *p2++ ;
				if (sumI > 0. && weight < 0.9 ) {
						*p1++ = sumI ;
					}
				else {
						*p1++ = 0 ;
					}
				}
			}
		}
}

void	FitAndSubtractPeakBackMask(TImage *image,TImage	*weightImage) {
	// We will fit a plane of the form ax+by+c to the pixels at the outer region of the image
	// and subtract the plane from the entire image
	// Revised by LDM to enforce positivity in result
	
	float	a,b,c,sumXI,sumYI,sumXX,sumYY,sumX,sumY,sumI,sumXY,sumW ;
	register 	short	x,y ;
	register	float 	intensity,weight ;
	short	width = image->Width() ;
	short	height = image->Height() ;
	TMatrix A ;
	TVector B,C	;
		
	register	float *p1 = image->GetData() ;
	register	float *p2 = weightImage->GetData() ;
	sumXI = sumYI = sumXX = sumYY = sumX = sumY = sumXY = sumI = sumW = 0. ;
	for(y = 0; y <height; y++) {
		for(x = 0 ;x < width; x++) {
			weight = *p2++ ;
			intensity = *p1++ ;
			if(weight > 1e-3){
				sumI += intensity  ;
				sumX += x ;
				sumY += y ;
				sumXX += x*x ;
				sumYY += y*y ;
				sumXY += x*y ;
				sumXI += x*intensity ;
				sumYI += y*intensity ;
				sumW += 1 ;
				}
		}
	}
	A.m[0][0] = sumXX ;
	A.m[1][0] = sumXY ;
	A.m[2][0] = sumX ;
	A.m[0][1] = sumXY ;
	A.m[1][1] = sumYY;
	A.m[2][1] = sumY ;
	A.m[0][2] = sumX ;
	A.m[1][2] = sumY ;
	A.m[2][2] = sumW ;
	B.a[0] = sumXI ;
	B.a[1] = sumYI ;
	B.a[2] = sumI ;
	if (Solve3LinearEquation(&A,&B,&C)) {
		//Use the robust form here for better accuracy ?
		//Trap case with nothing relevant
		a = C.a[0] ;
		b = C.a[1] ;
		c = C.a[2] ;
		if ( fabsf(a) > 1e-6 || fabsf(b) > 1e-6 || fabsf(c) > 1e-6) {
		FitBiweightBackground(image,weightImage);
		a = Bfita ;		
		b = Bfitb ;
		c = Bfitc ;
			}	
		// subtract the ramp
		p1 = image->GetData() ;
		p2 = weightImage->GetData() ;
		for(y = 0; y <height; y++) {
			float by = b*y + c ;
			for(x = 0 ;x < width; x++) {
				//LDM      Set with max of zero and mask
				sumI = *p1 - (a*x+by);
				weight = *p2++ ;
				if (weight < 0.9 ) {
						*p1++ = sumI ;
				}
				else {
						*p1++ = 0 ;
				}
			}
		}
	}
	else {
		//Trap case when no fit
		p1 = image->GetData() ;
		p2 = weightImage->GetData() ;
		for(y = 0; y <height; y++) {
			for(x = 0 ;x < width; x++) {
				//LDM      Set with max of zero and mask
				sumI = *p1;
				weight = *p2++ ;
				if (sumI > 0. && weight < 0.9 ) {
						*p1++ = sumI ;
					}
				else {
						*p1++ = 0 ;
					}
				}
			}
		}
}

void	FitBiweightBackground(TImage *image,TImage *weightImage)
	{
	// We will fit a plane of the form ax+by+c to the pixels at the outer region of the image
	// use an iterative Biweight form

	float	a,b,c,sumXI,sumYI,sumXX,sumYY,sumX,sumY,sumI,sumXY,sumW,dev,s1,wt;
	short	x, y ;
	float 	intensity, weight, stdval, s1old=1e8, totwt=0, totpt=0 ;
	short	width = image->Width() ;
	short	height = image->Height() ;
	float   xcenter, ycenter, Ramp0, BtuneU, BtuneU2 ;
	TMatrix A ;
	TVector B,C	;
	TImage	*Biweight = new TImage(width,height) ;
	TImage	*Biwold   = new TImage(width,height) ;
	// Initialize Bi-weight array to unity
	float *pbiweight = Biweight->GetData() ;
	float *pbiwold   = Biwold->GetData() ;
	for(y = 0; y <height*width; y++) {
			*pbiweight++ = 1. ;
		//	}
		}
	a = b = c = 0 ;
	BtuneU = Btune ; BtuneU2 = Btune2 ;
	xcenter = width/2 ; ycenter = height/2 ;
	// Biweight	iteration
	long Niter ;
	for	( Niter	= 0	; Niter	<= NumbBiwt ; Niter++ ) {
		sumXI = sumYI = sumXX = sumYY = sumX = sumY = sumXY = sumI = sumW = 0. ;
		pbiweight = Biweight->GetData() ;
		float *pimage = image->GetData() ;
		float *pweight = weightImage->GetData();
		for(y = 0; y <height; y++) {
			for(x = 0 ;x < width; x++) {
				weight = *pweight++  ;
				if (weight > 1e-3) {
					wt = *pbiweight++ ;
					wt = wt*wt ; 
					intensity = *pimage++ ;
					sumI += intensity * wt ;
					sumX += x*wt ;
					sumY += y*wt ;
					sumXX += x*x*wt ;
					sumYY += y*y*wt ;
					sumXY += x*y*wt ;
					sumXI += x*intensity*wt ;
					sumYI += y*intensity*wt ;
					sumW += wt ;
					}
				else {
					*pbiweight++ = 0;
					*pimage++ ;
					}
				}
			}
		A.m[0][0] =	sumXX ;
		A.m[1][0] =	sumXY ;
		A.m[2][0] =	sumX ;
		A.m[0][1] =	sumXY ;
		A.m[1][1] =	sumYY;
		A.m[2][1] =	sumY ;
		A.m[0][2] =	sumX ;
		A.m[1][2] =	sumY ;
		A.m[2][2] =	sumW ;
		B.a[0] = sumXI ;
		B.a[1] = sumYI ;
		B.a[2] = sumI ;
		sumW = 0 ;
		//fprintf(stderr,"B vector %f %f %f\n",B.a[0],B.a[1],B.a[2]);
		if (Solve3LinearEquation(&A,&B,&C))	{
			a = C.a[0] ;
			b = C.a[1] ;
			c = C.a[2] ;
			
			//Calculate the sum of absolute errors
			s1 = 0. ;
			sumW = 0 ;
			//s2 = 0 ;
			pbiweight = Biweight->GetData() ;
			pweight = weightImage->GetData() ;
			pbiwold = Biwold->GetData() ;
			for(y = 0; y <height; y++) {
				float by = b*y+c ;
				for(x = 0 ;x < width; x++) {
					weight = *pweight++ ;
					if (weight > 1e-3) {
						wt = *pbiweight ;
						dev = fabsf( *image->GetData(x,y) - (a*x+by) ) ;
						s1	  += dev*wt ;
						sumW  += wt ;
						*pbiwold++ = *pbiweight ; //Added, save old value
						*pbiweight++ = dev ;
						}
					else {
						*pbiweight++ = 0 ;
						*pbiwold++ = 0 ;
					}
				}
			}
		//Absolute deviation
		if(sumW > 1e-6) {
			s1 = s1/sumW ;
			}
		else {
			s1 = 0.5 ;
			}
			stdval =	MAX( (float) 0.01 , s1)	;
		if(DiffDebug)fprintf(stderr," %f %f %f %f %f %f \n",stdval,sumW,totpt,a,b,c);
		//Termination criteria
			if( (fabsf(s1old-s1) < BiweightTol) || !UseBiweight ) {
			//if ( (s1old-s1) < 1e-3) {
			delete Biweight ; delete Biwold ;
			Bfita = a;
			Bfitb = b;
			Bfitc = c;	
			//if ( totwt < totpt*0.5 )fprintf(stderr," Oops, too small %f %f \n",totwt,totpt) ;
			return ;
			}
		else {
			Bfita = a;
			Bfitb = b;
			Bfitc = c;
			s1old = s1 ;
			}

		//Update the Bi-weights
		pbiweight = Biweight->GetData() ;
		pbiwold   = Biwold->GetData() ;
		pweight = weightImage->GetData() ;
		totwt = 0 ;
		totpt = 0 ;
		for(y =	0; y <height*width; y++) {
					weight = *pweight++ ;
					float dev2 = 0 ;
					if ( weight > 1e-3 ) {
						dev = (*pbiweight)/stdval ;
						dev2 = *pbiwold++ ;
						if ( dev < BtuneU ) {
							dev= 1.-dev*dev*BtuneU2 ;
							//Note, dev*dev is the weight (not squared)
							dev2 = (dev2 + 2.*dev*dev)/3. ;
							}
						else {
							dev2 = dev2/3. ;
							}
						totwt += dev2 ;
						totpt += 1 ;
					}
					*pbiweight++ = dev2 ;
				}
		}
		//Trap losing too may points
		if ( totwt < totpt*PixPurge ) { // Maybe 0.75 or variable ?
			BtuneU = BtuneU*1.025 ;
			BtuneU2= 1./(BtuneU*BtuneU);
		}
		//End of Bi-weight loop
			}
		//if ( totwt < totpt*0.5 )fprintf(stderr," Oops, too small %f %f \n",totwt,totpt) ;
		delete Biweight ; delete Biwold ;
	//fprintf(stderr," %f %f %f \n",a,b,c);
}

TPeakArray *FindPeaks(TImage *image,PeakCriteraRecord *peakCriteria,Rect *selection	) {
	long 	row,col,krow,kcol, ShiftLines, ShiftPixels, 
			xpos,ypos,rightEdge,bottomEdge ;
	long 	Width, Height, Top, Bottom, Left, Right,HHeight,HWidth;
	float 	*WorkImageData,*PeakImageData,*p1,*p2 ;
	TArray	*peakArray,*workArray ;
	float 	CMRadiusSquared,R2, CMX, CMY,sum ;
	float 	SumI,SumXI,SumYI,*cmX,*cmY,*cmMass,*pkInt,*pcmX,*pcmY,*pcmMass,*ppkInt ;
	short 	numPeaks ;
	long  	i,numFloatsToAllocate ;
	float 	floatisMax, min, max ;
	Boolean isSelection;
	Rect	theSelectionRect ;
	Pixel 	pos,newpos ;
	PlateauList *plateauList  ;
	TPeakArray	*thePeakArrayList = nil ;

	float	lowThr,hiThr ;
	Boolean	isMax ;
	int	EdgeDistance,iEXRadius,iCMRadius ;

	isSelection = selection != nil ; 
	if (isSelection) {
		theSelectionRect = *selection ;
		workArray = new TArray(image->GetDataArray(),kInheritType,&theSelectionRect) ;
	}
	else {
		workArray = new TArray(image->GetDataArray()) ;
		SetRect(&theSelectionRect,0,0,image->Width(),image->Height()) ;
	}

	Top = theSelectionRect.top ;
	Left = theSelectionRect.left ;
	Bottom = theSelectionRect.bottom ;
	Right = theSelectionRect.right ;
	
	
	Height = Bottom - Top ;
	Width = Right - Left ;	
	
	min = workArray->Min() ;
	max = workArray->Max() ;

	lowThr	= peakCriteria->minValue ;
	hiThr	= peakCriteria->minValue ;
	isMax	= peakCriteria->findMax ;
	EdgeDistance	= peakCriteria->distanceFromEdgde ;
	iEXRadius		= peakCriteria->distanceFromEdgde ;
	iCMRadius		= peakCriteria->cmRadius ;

	floatisMax = isMax ;

	peakArray = new TArray(Width,Height) ;

	WorkImageData = workArray->GetData() ;
	PeakImageData = peakArray->GetData() ;

	if (isMax) {
		if (Height==1) {
			CreateMaxOf1DImage(WorkImageData,PeakImageData,Width);
		}
		else {
			CreateMaxOfImage(WorkImageData,PeakImageData,Width,Height) ;
		}
	}
	else {
		if (Height==1) {
			CreateMinOf1DImage(WorkImageData,PeakImageData,Width);
		}
		else {
			CreateMinOfImage(WorkImageData,PeakImageData,Width,Height) ;
		}
	}
		
	// -----------------------------------------------------
	// Generating the binary image of the unrefined peaks
	// -----------------------------------------------------	
	numPeaks = CreateUnrefinedBinaryImage(WorkImageData,PeakImageData,
		Width,Height,lowThr,hiThr) ;
	
	
	// --------------------------------------------------------------------------------------
	// The center of mass and exclusion radius are scaled to pixels and converted to integer
	// --------------------------------------------------------------------------------------
	CMRadiusSquared = iCMRadius * iCMRadius ;

	// ---------------------------------------------
	// The memory for the results is now allocated	
	// ---------------------------------------------
	
	numFloatsToAllocate = numPeaks ;
	numFloatsToAllocate = (numFloatsToAllocate < 8 ? 8 : numFloatsToAllocate) ;
	if ((pcmX = cmX = new float[numFloatsToAllocate]) == NULL) {
		delete workArray ;
		delete peakArray ;
		return thePeakArrayList ;
	}

	if ((pcmY = cmY = new float[numFloatsToAllocate]) == NULL) {
		delete workArray ;
		delete peakArray ;
		delete[] cmX ;
		return thePeakArrayList ;
	}

	if ((pcmMass = cmMass = new float[numFloatsToAllocate]) == NULL) {
		delete workArray ;
		delete peakArray ;
		delete[] cmX ;
		delete[] cmY ;
		return thePeakArrayList ;
	}
	
	if ((ppkInt = pkInt = new float[numFloatsToAllocate]) == NULL) {
		delete workArray ;
		delete peakArray ;
		delete[] cmX ;
		delete[] cmY ;
		delete[] cmMass ;
		return thePeakArrayList ;
	}

	// ------------------------------------------------------------
	// The main loop for the center of mass calculation starts now
	// ------------------------------------------------------------	
	p1 = WorkImageData ;
	p2 = PeakImageData ;
	numPeaks = 0 ;
	
	// Find all the plateaus
	plateauList = NewPlateauList() ;
	for (row=0; row<Height; row++) {
		for (col=0; col<Width; col++) {
			if (*p2 != 0.)	{
				pos.h = col ;
				pos.v = row ;
				if (IsPlateau(&pos,p2,Width,Height)) {
					if (!AssignPixelToPlateau(plateauList,&pos,p2,Width)) {
						delete workArray ;
						delete peakArray ;
						delete[] cmX ;
						delete[] cmY ;
						delete[] cmMass ;
						delete[] pkInt ;
						return thePeakArrayList ;
					}
				}
			}
			p2++ ;
		}
	}
	// Reduce the plateaus to single Peaks
	if (plateauList->numPlateaus != 0) {
		ReducePlateaus(plateauList,PeakImageData,Width) ;
	}

	rightEdge = Width-EdgeDistance-1 ;
	bottomEdge = Height-EdgeDistance-1 ;
	HHeight = Height/2 ;
	HWidth = Width/2 ;	
	
	for (row=0; row<Height; row++) {
		p1 = WorkImageData + row*Width ;
		p2 = PeakImageData + row*Width ;
		if (EdgeDistance > 0) {
			if (EdgeDistance < HHeight && (row < EdgeDistance || 
					row > bottomEdge)) {
				// Zero the row 
				for (col=0; col<Width; col++) {
					*p2++ = 0. ;
				}
				goto nextrow ;
			}
		}
		for (col=0; col<Width; col++) {
			if (*p2 != 0.)	{
				if (EdgeDistance > 0) {
					if (EdgeDistance < HWidth && (col < EdgeDistance || 
							col > rightEdge)) {
						*p2 = 0. ;
						goto nextcol ;
					}
				}
				pos.h = col ;
				pos.v = row ;
				newpos = pos ;
				
				// We will use this to see if the peaks should be 
				// considered part of a single peak
				// Thus the C.M. radius will serve as a radius for 
				// finding the C.M. peak for several peaks
				// constituting a single peak
				if (iCMRadius != 0) {	
					if ((sum=SumNeighbours(&pos,iCMRadius,p2,Width,Height)) > 1.) {	
						// Are the Peaks too close ?
						// Find CM of Neighboring Peaks
						SetCMPixel(&pos,iCMRadius,p2,Width,Height,&newpos) ;	
					}
				}
				
				if (newpos.h == pos.h && newpos.v == pos.v) {	
					// The peak survived the first stage
					if (iEXRadius != 0) {
						if ((sum=SumNeighbours(&pos,iEXRadius,p2,Width,Height)) > 1.) {	
							// Are the Peaks too close ?
							SetExtremePeak(&pos,iEXRadius,p2,p1,
								Width,Height,isMax,&newpos) ;	
						}
					}
					if (newpos.h == pos.h && newpos.v == pos.v) {						
						// Did the peak survive 
						numPeaks++ ; 	
						if (iCMRadius == 0 || col-iCMRadius<0 || col+iCMRadius>Width || 
								(row-iCMRadius <0 && Height != 1) || 
								(row+iCMRadius > Height && Height != 1))
						{
							// -----------------------------------
							// Finding Peaks without CM refinement
							// -----------------------------------
							*pcmX++ = col ;	// If CM area runs over the borders the 
							*pcmY++ = row ;	// peaks are left unrefined 
							*ppkInt++ = *p1 ;
							*pcmMass++ = *p1 ;
						} // End of near border pixels loop 
						
						else
						{
							// -----------------------------------
							// Finding Peaks using CM refinement
							// -----------------------------------
							SumI  = 0. ;
							SumXI = 0. ;
							SumYI = 0. ;
							for (krow=-1*iCMRadius; krow<iCMRadius+1; krow++)
							{
								ShiftLines = krow*Width ;
								ypos = row+krow ;
								if (ypos >= 0 && ypos < Height)
								{
									for(kcol=-1*iCMRadius; kcol<iCMRadius+1; kcol++)
									{
										ShiftPixels = ShiftLines + kcol ;
										xpos = col+kcol ;
										if (xpos >= 0 && xpos < Width)
										{
											if (( R2=krow*krow + kcol*kcol) <= CMRadiusSquared)
											{
												SumI += *(p1 + ShiftPixels) ;
												SumXI += xpos * (*(p1 + ShiftPixels)) ;
												SumYI += ypos * (*(p1 + ShiftPixels)) ;
											}
										}
									}
								}
							}
							CMX = SumXI/SumI ;
							CMY = SumYI/SumI ;
							*pcmX++ = CMX ;
							*pcmY++ = CMY ;
							*ppkInt++  = *p1 ;
							*pcmMass++ = SumI ;
						} // End of non near border pixels loop 
					}
				}
			} // End of peak finding loop
			
			nextcol: ;
			p2++ ;
			p1++ ;
		} // End of col loop
		nextrow: ;

	} // End of row loop

	delete	peakArray ;
	delete	workArray ;

	thePeakArrayList = new TPeakArray(numPeaks) ;
	for(i=0,pcmX=cmX,pcmY=cmY,ppkInt = pkInt;i < numPeaks;i++,pcmX++,pcmY++,ppkInt++) {
		thePeakArrayList->mX[i] = *pcmX + Left ;
		thePeakArrayList->mY[i] = *pcmY + Top ;
		thePeakArrayList->mValue[i] = *ppkInt ;
	}
	delete[] cmX ;
	delete[] cmY ;
	delete[] cmMass ;
	delete[] pkInt ;

	return thePeakArrayList ;
}


long CreateUnrefinedBinaryImage(float *WorkImageData,float *PeakImageData,
	long Width,long Height,float lowThr,float hiThr)
{
	float *p1,*p2,tolerance=0.0001,minThreshold,maxThreshold ;
	long	row,col,numPeaks ;
	
	numPeaks = 0 ;
	p1 = WorkImageData ;
	p2 = PeakImageData ;
	minThreshold = lowThr - tolerance ;
	maxThreshold = hiThr + tolerance ;
	for (row=0; row<Height; row++)
	{
		p1 = WorkImageData + row*Width ;
		for (col=0; col<Width; col++)
		{
			if (*p2 == *p1 && *p2 >= minThreshold && *p2 <= maxThreshold)
			{
				numPeaks++ ;
				*p2 = 1. ;
			}
			else 
			{
				*p2 = 0. ;
			}
			p1++ ;
			p2++ ;
		}
		
	}
	return(numPeaks) ;
}


// ---------------------------------------------------------
// Finds the Center	of Mass	for	Peaks that are too close and
// aproximates it to the nearest pixel.						
// ---------------------------------------------------------
void SetCMPixel(Pixel *pos,int rad,float *data,long	width,long height, Pixel *newpos)
{
	int ShiftLines ;
	float rsum, ri, SumXI, SumYI ;
	int i,j,x,y,ii,jj,usq,dysq,radsq ;
	
	rsum  = 0. ;
	SumXI = 0. ;
	SumYI = 0. ;
	x = pos->h ;
	y = pos->v ;
	radsq = rad*rad ;
	for (j=0; j<=rad; j++) {
		if(y+j < height) {
			ShiftLines = j*width ;
			dysq = j*j ;
			for(i= -rad; i<= rad; i++) {
				if (x+i < width && x+i >= 0) {
					usq = i*i+dysq ;
					if (usq <= radsq) {
						rsum += *(data + ShiftLines + i) ;
						SumXI += i * (*(data + ShiftLines + i));
						SumYI += j * (*(data + ShiftLines + i)) ;
					}
				}
			}
		}
	}
	ri = SumXI/rsum ;
	// Shift from current pixel to CM pixel in x-direction
	ii = (ri > 0. ? (int)(ri+0.5) : (int)(ri-0.5));	

	// Shift from current pixel to CM pixel in y-direction 
	jj = (int)(SumYI/rsum+0.5) ;					

	newpos->h += ii ;
	newpos->v += jj ;
	
	for (j=0; j<=rad; j++) {
		if(y+j < height) {
			ShiftLines = j*width ;
			dysq = j*j ;
			for(i= -rad; i<=rad; i++) {
				if (x+i < width && x+i >= 0) {
					usq = i*i+dysq ;
					if (usq <= radsq) {
						if ( i!= ii || j != jj) {
							*(data + ShiftLines + i) = 0. ;
						}
						else {
							*(data + ShiftLines + i) = 1. ;
						}
					}
				}
			}
		}
	}
}

void SetExtremePeak(Pixel *pos,int rad,float *peak,float *image,long width,long	height,
	Boolean extremeIsMax, Pixel *newpos)
{
	long ShiftLines ;
	float maxPeakValue,minPeakValue,peakIntensity,PositionValue ;
	int i,j,x,y,iExtreme,jExtreme,usq,radsq,dysq ;
	
	x = pos->h ;
	y = pos->v ;
	maxPeakValue = -1.e-27 ;
	minPeakValue =  1.e+27 ;
	radsq = rad*rad ;
	for (j= 0; j<=rad; j++) {
		if(y+j < height) {
			ShiftLines = j*width ;
			dysq = j*j ;
			for(i= -rad; i<= rad; i++) {
				if (x+i < width && x+i >= 0) {
					usq = i*i+dysq ;
					if (usq <= radsq) {
						PositionValue = *(peak + ShiftLines + i) ;
						if (PositionValue == 1.) {
							peakIntensity = *(image + ShiftLines + i) ;
							if (extremeIsMax) {
								if (peakIntensity > maxPeakValue) {
									iExtreme = i ;
									jExtreme = j ;
									maxPeakValue = peakIntensity ;
								}
							}
							else {
								if (peakIntensity < minPeakValue) {
									iExtreme = i ;
									jExtreme = j ;
									minPeakValue = peakIntensity ;
								}
							}
						}
					}
				}
			}
		}
	}
	
	newpos->h = pos->h+iExtreme ;
	newpos->v = pos->v+jExtreme ;
	
	//  Set current peak to zero if it is not the extreme peak and let others 
	//  take care of themselves later
	if (jExtreme != 0 || iExtreme != 0) {  
		*peak = 0. ;
	}
	else {  // We are the extreme peak, so set others to zero
		for (j=0; j<=rad; j++) {
			if(y+j < height) {
				ShiftLines = j*width ;
				dysq = j*j ;
				for(i= -rad; i<= rad; i++) {
					if (x+i < width && x+i >= 0) {
						usq = i*i+dysq ;
						if (usq <= radsq) {
							PositionValue = *(peak + ShiftLines + i) ;
							if (PositionValue == 1.) {
								if (i != iExtreme || j != jExtreme) {
									*(peak + ShiftLines + i) = 0. ;
								}
							}
						}
					}
				}
			}
		}
	}
}


// ---------------------------------------------------------------------
// Gets	the	sum	of the intensities inside a	neighbourhood of side 'rad'
// ---------------------------------------------------------------------

float SumNeighbours(Pixel *pos,int rad,float *data,long	width,long height)
{
	float sum ;
	int ShiftLines ;
	int i,j,x,y,dysq,usq,radsq ;
	
	if (rad==0)
	{
		return(1) ;
	}
	sum  = 0. ;
	x = pos->h ;
	y = pos->v ;
	radsq = rad*rad ;
	for (j = 0; j <= rad; j++) {
		if(y+j < height) {
			ShiftLines = j*width ;
			dysq = j*j ;
			for(i= -rad; i<=rad; i++) {
				if (x+i < width && x+i >= 0) {
					usq = i*i + dysq ;
					if (usq <= radsq) {
						sum += *(data + ShiftLines + i) ;
					}
				}
			}
		}
	}
	return(sum) ;
}


void SetToMaxofNeighbourhood(long width, float *datain,	float *dataout)
{
	float Max ;
	float *tp ;
	
	tp = datain - width - 1 ;
	Max = *datain ;
	Max= MAX(*tp, Max ) ;
	tp++ ;
	Max= MAX(*tp, Max ) ;
	tp++ ;
	Max= MAX(*tp, Max ) ;
	Max= MAX(*(datain-1), Max ) ;
	Max= MAX(*(datain+1), Max ) ;
	tp = datain + width - 1 ;
	Max= MAX(*tp, Max ) ;
	tp++ ;
	Max= MAX(*tp, Max ) ;
	tp++ ;
	Max= MAX(*tp, Max ) ;
	*dataout = Max ;
}

void SetToMinofNeighbourhood(long width, float *datain,	float *dataout)
{
	float Min ;
	float *tp ;
	
	tp = datain - width - 1 ;
	Min = *datain ;
	Min= MIN(*tp, Min ) ;
	tp++ ;
	Min= MIN(*tp, Min ) ;
	tp++ ;
	Min= MIN(*tp, Min ) ;
	Min= MIN(*(datain-1), Min ) ;
	Min= MIN(*(datain+1), Min ) ;
	tp = datain + width - 1 ;
	Min= MIN(*tp, Min ) ;
	tp++ ;
	Min= MIN(*tp, Min ) ;
	tp++ ;
	Min= MIN(*tp, Min ) ;
	*dataout = Min ;
}


void SetToMaxofLowerNeighbourhood(long width, float	*datain, float *dataout)
{
	float Max ;
	float *tp ;
	
	Max = *datain ;
	Max= MAX(*(datain-1), Max ) ;
	Max= MAX(*(datain+1), Max ) ;
	tp = datain + width - 1 ;
	Max= MAX(*tp, Max ) ;
	tp++ ;
	Max= MAX(*tp, Max ) ;
	tp++ ;
	Max= MAX(*tp, Max ) ;
	*dataout = Max ;
}

void SetToMinofLowerNeighbourhood(long width, float	*datain, float *dataout)
{
	float Min ;
	float *tp ;
	
	Min = *datain ;
	Min= MIN(*(datain-1), Min ) ;
	Min= MIN(*(datain+1), Min ) ;
	tp = datain + width - 1 ;
	Min= MIN(*tp, Min ) ;
	tp++ ;
	Min= MIN(*tp, Min ) ;
	tp++ ;
	Min= MIN(*tp, Min ) ;
	*dataout = Min ;
}

void SetToMaxofUpperNeighbourhood(long width, float	*datain, float *dataout)
{
	float Max ;
	float *tp ;
	
	tp = datain - width - 1 ;
	Max = *datain ;
	Max= MAX(*tp, Max ) ;
	tp++ ;
	Max= MAX(*tp, Max ) ;
	tp++ ;
	Max= MAX(*tp, Max ) ;
	Max= MAX(*(datain-1), Max ) ;
	Max= MAX(*(datain+1), Max ) ;
	*dataout = Max ;
}

void SetToMinofUpperNeighbourhood(long width, float	*datain, float *dataout)
{
	float Min ;
	float *tp ;
	
	tp = datain - width - 1 ;
	Min = *datain ;
	Min= MIN(*tp, Min ) ;
	tp++ ;
	Min= MIN(*tp, Min ) ;
	tp++ ;
	Min= MIN(*tp, Min ) ;
	Min= MIN(*(datain-1), Min ) ;
	Min= MIN(*(datain+1), Min ) ;
	*dataout = Min ;
}

void SetToMaxofRightNeighbourhood(long width, float	*datain, float *dataout)
{
	float Max ;
	float *tp ;
	
	tp = datain - width ;
	Max = *datain ;
	Max= MAX(*tp, Max ) ;
	tp++ ;
	Max= MAX(*tp, Max ) ;
	Max= MAX(*(datain+1), Max ) ;
	tp = datain + width  ;
	Max= MAX(*tp, Max ) ;
	tp++ ;
	Max= MAX(*tp, Max ) ;
	*dataout = Max ;
}

void SetToMinofRightNeighbourhood(long width, float	*datain, float *dataout)
{
	float Min ;
	float *tp ;
	
	tp = datain - width ;
	Min = *datain ;
	Min= MIN(*tp, Min ) ;
	tp++ ;
	Min= MIN(*tp, Min) ;
	Min= MIN(*(datain+1), Min ) ;
	tp = datain + width  ;
	Min= MIN(*tp, Min ) ;
	tp++ ;
	Min= MIN(*tp, Min ) ;
	*dataout = Min ;
}

void SetToMaxofLeftNeighbourhood(long width, float *datain,	float *dataout)
{
	float Max ;
	float *tp ;
	
	tp = datain - width -1;
	Max = *datain ;
	Max= MAX(*tp, Max ) ;
	tp++ ;
	Max= MAX(*tp, Max ) ;
	Max= MAX(*(datain-1), Max ) ;
	tp = datain + width - 1 ;
	Max= MAX(*tp, Max ) ;
	tp++ ;
	Max= MAX(*tp, Max ) ;
	*dataout = Max ;
}

void SetToMinofLeftNeighbourhood(long width, float *datain,	float *dataout)
{
	float Min ;
	float *tp ;
	
	tp = datain - width -1;
	Min = *datain ;
	Min= MIN(*tp, Min ) ;
	tp++ ;
	Min= MIN(*tp, Min ) ;
	Min= MIN(*(datain-1), Min ) ;
	tp = datain + width - 1 ;
	Min= MIN(*tp, Min ) ;
	tp++ ;
	Min= MIN(*tp, Min ) ;
	*dataout = Min ;
}
	
void SetToMaxofLowerRightNeighbourhood(long	width, float *datain, float	*dataout)
{
	float Max ;
	
	
	
	Max = *datain ;
	Max= MAX(*(datain+ 1), Max ) ;
	Max= MAX(*(datain+ width), Max ) ;
	Max= MAX(*(datain+ width + 1), Max ) ;
	*dataout = Max ;
}
	
void SetToMinofLowerRightNeighbourhood(long	width, float *datain, float	*dataout)
{
	float Min ;
	
	
	
	Min = *datain ;
	Min= MIN(*(datain+ 1), Min ) ;
	Min= MIN(*(datain+ width), Min ) ;
	Min= MIN(*(datain+ width + 1), Min ) ;
	*dataout = Min ;
}

void SetToMaxofLowerLeftNeighbourhood(long width, float	*datain, float *dataout)
{
	float Max ;
	
	
	
	Max = *datain ;
	Max= MAX(*(datain- 1), Max ) ;
	Max= MAX(*(datain+ width), Max ) ;
	Max= MAX(*(datain+ width - 1), Max ) ;
	*dataout = Max ;
}

void SetToMinofLowerLeftNeighbourhood(long width, float	*datain, float *dataout)
{
	float Min ;
	
	
	
	Min = *datain ;
	Min= MIN(*(datain- 1), Min ) ;
	Min= MIN(*(datain+ width), Min ) ;
	Min= MIN(*(datain+ width - 1), Min ) ;
	*dataout = Min ;
}

void SetToMaxofUpperRightNeighbourhood(long	width, float *datain, float	*dataout)
{
	float Max ;
	
	
	
	Max = *datain ;
	Max= MAX(*(datain+ 1), Max ) ;
	Max= MAX(*(datain- width), Max ) ;
	Max= MAX(*(datain- width + 1), Max ) ;
	*dataout = Max ;
}

void SetToMinofUpperRightNeighbourhood(long	width, float *datain, float	*dataout)
{
	float Min ;
	
	
	
	Min = *datain ;
	Min= MIN(*(datain+ 1), Min ) ;
	Min= MIN(*(datain- width), Min ) ;
	Min= MIN(*(datain- width + 1), Min ) ;
	*dataout = Min ;
}

void SetToMaxofUpperLeftNeighbourhood(long width, float	*datain, float *dataout)
{
	float Max ;
	
	
	
	Max = *datain ;
	Max= MAX(*(datain- 1), Max ) ;
	Max= MAX(*(datain- width), Max ) ;
	Max= MAX(*(datain- width - 1), Max ) ;
	*dataout = Max ;
}

void SetToMinofUpperLeftNeighbourhood(long width, float	*datain, float *dataout)
{
	float Min ;
	
	
	
	Min = *datain ;
	Min= MIN(*(datain- 1), Min ) ;
	Min= MIN(*(datain- width), Min ) ;
	Min= MIN(*(datain- width - 1), Min ) ;
	*dataout = Min ;
}


void CreateMaxOf1DImage(float *WorkImageData,float *PeakImageData,long Width)
{
	float 	*p1,*p2 ;
	long	col ;
	float 	Max ;
	
			
	// --------------------------
	// Treating the first pixel	
	// --------------------------

	p1 = WorkImageData ;
	p2 = PeakImageData ;
	*p2 = MAX(*p1,*(p1+ 1)) ;
	
	
			
	// ------------------------------------------------
	// Treating from the second to the (WIdth-1) pixel
	// ------------------------------------------------

	p1 = WorkImageData + 1 ;
	p2 = PeakImageData + 1 ;
	for (col=1;col<Width-1;col++)
	{
		Max = *p1;
		Max = MAX(Max,*(p1-1));
		Max = MAX(Max,*(p1+1));
		*p2 = Max;
		p1++;
		p2++;
		
	}

	// ------------------------
	// Treating the last pixel
	// ------------------------

	p1 = WorkImageData + Width-1;
	p2 = PeakImageData + Width-1;
	*p2 = MAX(*p1,*(p1-1)) ;
	

}


void CreateMinOf1DImage(float *WorkImageData,float *PeakImageData,long Width)
{
	float 	*p1,*p2 ;
	long	col ;
	float 	Min ;
	
			
	// --------------------------
	// Treating the first pixel	
	// --------------------------

	p1 = WorkImageData ;
	p2 = PeakImageData ;
	*p2 = MIN(*p1,*(p1+ 1)) ;
	
	
			
	// ------------------------------------------------
	// Treating from the second to the (Width-1) pixel
	// ------------------------------------------------

	p1 = WorkImageData + 1 ;
	p2 = PeakImageData + 1 ;
	for (col=1;col<Width-1;col++)
	{
		Min = *p1;
		Min = MIN(Min,*(p1-1));
		Min = MIN(Min,*(p1+1));
		*p2 = Min;
		p1++;
		p2++;
		
	}

	// ------------------------
	// Treating the last pixel
	// ------------------------

	p1 = WorkImageData + Width-1;
	p2 = PeakImageData + Width-1;
	*p2 = MIN(*p1,*(p1-1)) ;
	

}


void CreateMaxOfImage(float	*WorkImageData,float *PeakImageData,long Width,long	Height)	
{
	float *p1,*p2 ;
	long	row,col ;

	
	// ------------------------------
	// Treating the Top Left Corner	
	// ------------------------------
	p1 = WorkImageData ;
	p2 = PeakImageData ;
	SetToMaxofLowerRightNeighbourhood(Width, p1, p2) ;
	
	// -----------------------
	// Treating the first Row
	// -----------------------
	p1 = WorkImageData + 1;
	p2 = PeakImageData + 1 ;
	for (row=0,col=1; col<Width-1; col++)
	{
		SetToMaxofLowerNeighbourhood(Width, p1,p2) ;
		p1++ ;
		p2++ ;
	}

	// ------------------------------
	// Treating the Top Right Corner
	// ------------------------------
	p1 = WorkImageData + Width -1 ;
	p2 = PeakImageData + Width -1 ;
	SetToMaxofLowerLeftNeighbourhood(Width, p1, p2) ;
		
	// -----------------------
	// Treating the first Col
	// -----------------------
	p1 = WorkImageData + Width ;
	p2 = PeakImageData + Width ;
	for (row=1,col=0; row<Height-1; row++)
	{
		SetToMaxofRightNeighbourhood(Width,p1,p2) ;
		p1 += Width ;
		p2 += Width ;
	}

	// -----------------------------
	// Treating the internal pixels	
	// -----------------------------
	p2 = PeakImageData + Width -1;
	for (row=1; row<Height-1; row++)
	{
		p1 = WorkImageData + row*Width + 1 ;
		p2 += 2 ;
		for (col=1; col<Width-1; col++)
		{
			SetToMaxofNeighbourhood(Width,p1,p2) ;
			p1++ ;
			p2++ ;
		}
	}
	
	// ----------------------
	// Treating the last Col
	// ----------------------
	p1 = WorkImageData + 2*Width - 1 ;
	p2 = PeakImageData + 2*Width - 1 ;
	for (row=1,col=Width; row<Height-1; row++)
	{
		SetToMaxofLeftNeighbourhood(Width,p1,p2) ;
		p1 += Width ;
		p2 += Width ;
	}
	
	// --------------------------------
	// Treating the Bottom Left Corner
	// --------------------------------
	p1 = WorkImageData + (Height-1)*Width ;
	p2 = PeakImageData + (Height-1)*Width ;
	SetToMaxofUpperRightNeighbourhood(Width, p1, p2) ;
			
	// ----------------------
	// Treating the last Row
	// ----------------------
	p1 = WorkImageData + (Height-1)*Width + 1 ;
	p2 = PeakImageData + (Height-1)*Width + 1 ;
	for (row=Height,col=1; col<Width-1; col++)
	{
		SetToMaxofUpperNeighbourhood(Width,p1,p2) ;
		p1++ ;
		p2++ ;
	}

	// ---------------------------------
	// Treating the Bottom Right Corner
	// ---------------------------------
	p1 = WorkImageData + Height*Width - 1 ;
	p2 = PeakImageData + Height*Width - 1 ;
	SetToMaxofUpperLeftNeighbourhood(Width, p1, p2) ;
}

void CreateMinOfImage(float	*WorkImageData,float *PeakImageData,long Width,long	Height)	
{
	float *p1,*p2 ;
	long	row,col ;

	
	// -----------------------------
	// Treating the Top Left Corner
	// -----------------------------
	p1 = WorkImageData ;
	p2 = PeakImageData ;
	SetToMinofLowerRightNeighbourhood(Width, p1, p2) ;
	
	// -----------------------
	// Treating the first Row
	// -----------------------
	p1 = WorkImageData + 1;
	p2 = PeakImageData + 1 ;
	for (row=0,col=1; col<Width-1; col++)
	{
		SetToMinofLowerNeighbourhood(Width, p1,p2) ;
		p1++ ;
		p2++ ;
	}

	// ------------------------------
	// Treating the Top Right Corner
	// ------------------------------
	p1 = WorkImageData + Width -1 ;
	p2 = PeakImageData + Width -1 ;
	SetToMinofLowerLeftNeighbourhood(Width, p1, p2) ;
		
	// -----------------------
	// Treating the first Col
	// -----------------------
	p1 = WorkImageData + Width ;
	p2 = PeakImageData + Width ;
	for (row=1,col=0; row<Height-1; row++)
	{
		SetToMinofRightNeighbourhood(Width,p1,p2) ;
		p1 += Width ;
		p2 += Width ;
	}

	// -----------------------------
	// Treating the internal pixels	
	// -----------------------------
	p2 = PeakImageData + Width -1;
	for (row=1; row<Height-1; row++)
	{
		p1 = WorkImageData + row*Width + 1 ;
		p2 += 2 ;
		for (col=1; col<Width-1; col++)
		{
			SetToMinofNeighbourhood(Width,p1,p2) ;
			p1++ ;
			p2++ ;
		}
	}
	
	// ----------------------
	// Treating the last Col
	// ----------------------
	p1 = WorkImageData + 2*Width - 1 ;
	p2 = PeakImageData + 2*Width - 1 ;
	for (row=1,col=Width; row<Height-1; row++)
	{
		SetToMinofLeftNeighbourhood(Width,p1,p2) ;
		p1 += Width ;
		p2 += Width ;
	}
	
	// --------------------------------
	// Treating the Bottom Left Corner
	// --------------------------------
	p1 = WorkImageData + (Height-1)*Width ;
	p2 = PeakImageData + (Height-1)*Width ;
	SetToMinofUpperRightNeighbourhood(Width, p1, p2) ;
			
	// -----------------------
	// Treating the last Row
	// -----------------------
	p1 = WorkImageData + (Height-1)*Width + 1 ;
	p2 = PeakImageData + (Height-1)*Width + 1 ;
	for (row=Height,col=1; col<Width-1; col++)
	{
		SetToMinofUpperNeighbourhood(Width,p1,p2) ;
		p1++ ;
		p2++ ;
	}

	// ---------------------------------
	// Treating the Bottom Right Corner
	// ---------------------------------
	p1 = WorkImageData + Height*Width - 1 ;
	p2 = PeakImageData + Height*Width - 1 ;
	SetToMinofUpperLeftNeighbourhood(Width, p1, p2) ;
}

