#include "TDiffractionPeaks.h"
#include <math.h>

#define min(x,y) ((x) < (y) ? (x) : (y))
#define max(x,y) ((x) > (y) ? (x) : (y))
#define	log2(x)	log(x)/log(2.0)

extern	void PutMessage(char *) ;
extern	float	FindPeak(TImage *image,Rect *areaToSearch,FPoint *peakPos,
		Boolean threshold=true) ;
extern	TPeakArray *FindPeaksAlongLattice(TImage *image,TLattice *lattice,
		Boolean	threshold=true) ;
extern	TPeakArray *FindPeaksAtLatticePoints(TImage *image,TLattice *lattice,
		Boolean	threshold=true) ;
extern	long near(float) ;
extern	float	GetPeakThreshold() ;

void	ExtractAmplitudePhase(TArray *imageArray,long &n,long &m,float &x,float &y) ;
TPeakList *FindTEMPeaksAlongLattice(TImage *image,TLattice *lattice) ;
TPeakList *FindDiffractionSpotsAtLatticePoints(TImage *image,TLattice *lattice) ;
TPeakList *FindDiffractionSpotsAtLatticePoints(TImage *image,TLattice *lattice,TImage *peakMotif,float ccThreshold) ;
void 	FitAndSubtractPeakBackground(TImage *image,TImage *weightImage) ;


void  pvgs (long *he, long *ke, long *lse, long h, long k, long &ne, char *symm ) ;
void pgs (long *he, long *ke, long *lse, long h, long k, long &ne, char *symm ) ;
void p6symm (long *he, long *ke, long *lse, long h, long k, long &ne, char *symm ) ;
void p4symm (long *he, long *ke, long *lse, long h, long k, long &ne, char *symm ) ;
void p4mg (long *he, long *ke, long *lse, long h, long k, long &ne) ;
void p4 (long *he, long *ke, long *lse, long h, long k, long &ne, char *symm ) ;
void p3symm (long *he, long *ke, long *lse, long h, long k, long &ne, char *symm ) ;
void p2symm (long *he, long *ke, long *lse, long h, long k, long &ne, char *symm ) ; 
void p2mgs (long *he, long *ke, long *lse, long h, long k, long &ne, char *symm ) ; 
void p2ggs (long *he, long *ke, long *lse, long h, long k, long &ne, char *symm ) ;
void p1symm (long *he, long *ke, long *lse, long h, long k, long &ne, char *symm ) ; 
void ms (long *he, long *ke, long *lse, long h, long k, long &ne, char *symm ) ;
void f2symm (long *he, long *ke, long *lse, long h, long k, long &ne, char *symm ) ;
void equiv2 (long *he, long *ke, long *lse, long h, long k, long &ne, char *symm ) ; 
void averagehkl() ;

extern	double	PiHalf ;
extern	double	Pi ;
extern	double	TwoPi ;

inline	double Mod2Pi(double x) { double xx = x - TwoPi*((int)(x/TwoPi)) ; return xx ; } ;

extern	float	gDiffractionAperture ;
extern	PeakForm gPeakForm ;

TFourierPeaks::TFourierPeaks(TFourierPeaks *s) 
		: mNumPeaks(s->mNumPeaks) {
	mImageH = new long[s->mNumPeaks] ;
	mImageK = new long[s->mNumPeaks] ;
	mLatticeH = new long[s->mNumPeaks] ;
	mLatticeK = new long[s->mNumPeaks] ;
	mAmplitude = new float[s->mNumPeaks] ;
	mPhase = new float[s->mNumPeaks] ;
	memcpy(mImageH,s->mImageH,mNumPeaks*sizeof(long)) ;
	memcpy(mImageK,s->mImageK,mNumPeaks*sizeof(long)) ;
	memcpy(mLatticeH,s->mLatticeH,mNumPeaks*sizeof(long)) ;
	memcpy(mLatticeK,s->mLatticeK,mNumPeaks*sizeof(long)) ;
	memcpy(mAmplitude,s->mAmplitude,mNumPeaks*sizeof(float)) ;
	memcpy(mPhase,s->mPhase,mNumPeaks*sizeof(float)) ;
}

// -------------------------------------------------
// This routine is for HRTEM Images. The
// image was obtained by Fourier transforming a
// real image
// -------------------------------------------------

TDiffractionPeaks::TDiffractionPeaks(TImage *image,TLattice *lattice) {
	// This is used for reciprocal data which is the FT of a real object.
	// Thus we have F(h,k) = F*(-h,-k)
	// Find peaks along refined lattice
	TPeakArray	*peakArray ;
	TArray	*imageArray ;
	short	imageCenterX = image->Width()/2 ;
	short	imageCenterY = image->Height()/2 ;
	float	apertSq = gDiffractionAperture*gDiffractionAperture ;
	long	h,k,n,m,i,width,height,minimumSize ;
	float	amp,phase,radsq,radtodegree = 90./::asin(1.0) ;

	mFourierData = nil ;
	mFourierPeaks = nil ;
	
	mLattice = new TLattice(lattice) ;
	
	mImageWidth = image->Width() ;
	mImageHeight = image->Height() ;

	float	calibrationConstant = image->GetCalibration() ;
	mCalibrationUnit = image->GetCalibrationUnit() ;
	
	if (mCalibrationUnit == kInverseNanometer) {
		calibrationConstant *= 0.1 ;
		mCalibrationUnit = kInverseAngstrom ;
	}

	// Now we need to put the data in proper sampling and angle stuff
	float	ux,uy,vx,vy,u,v,gamma,siga,csga,gammastar ;
	FPoint	U,V ;
	U = lattice->GetU() ;
	V = lattice->GetV() ;

	short	iBG = min(4,min(near(sqrt(U.x*U.x+U.y*U.y)),near(sqrt(V.x*V.x+V.y*V.y)))) ;

	ux = calibrationConstant*U.x ;
	uy =  calibrationConstant*U.y ;
	vx =  calibrationConstant*V.x ;
	vy =  calibrationConstant*V.y ;
	mAStar = u  = sqrt(ux*ux+uy*uy) ;
	mBStar = v  = sqrt(vx*vx+vy*vy) ;


	gammastar = ::acos((ux*vx+uy*vy)/(u*v)) ;
	mGammaStar = radtodegree*gammastar ;

	gamma = (180-mGammaStar)*::asin(1.0)/90. ;
	siga = ::sin(gamma) ;
	csga = ::cos(gamma) ;
	gamma = 180. - mGammaStar ;

	mA = 1./(u*siga) ;
	mB = 1./(v*siga) ;
		
	minimumSize = 128 ; //Was 32
	
	imageArray = image->GetDataArray() ; 
	
	TPeakList	*peakList =  FindTEMPeaksAlongLattice(image,lattice) ;
	peakArray = new TPeakArray(peakList) ; 
	delete peakList ;
	
	// Create a new array to hold the Fourier transform coefficients
	mMaxH = 0 ;
	mMaxK = 0 ;
	for(i=0; i < peakArray->mNumPeaks; i++) {
		h = peakArray->mN[i] ; if (h < 0) h = -h ;
		k = peakArray->mM[i] ; if (k < 0) k = -k ;
		if (h > mMaxH) mMaxH = h ;
		if (k > mMaxK) mMaxK = k ;
	}
	float	gMax = sqrt((mMaxH*U.x+mMaxK*V.x)*(mMaxH*U.x+mMaxK*V.x) + (mMaxH*U.y+mMaxK*V.y)*(mMaxH*U.y+mMaxK*V.y))/
			sqrt((U.x+V.x)*(U.x+V.x) + (U.y+V.y)*(U.y+V.y)) ;
	short	minHKDim = near(gMax/siga) + 1 ;
	// Find the closest dimensions that are power of 2
	width = 2*minHKDim ;
	height = 2*minHKDim ;
	float 	sizeNP2 = log2((float)width) ; //LDM: added float
	short	iSizeNP2 = (int)sizeNP2 ;
	if ((sizeNP2-iSizeNP2) > 0.0001) iSizeNP2++ ;
	float 	sizeMP2 = log2((float)height) ; //LDM: added float
	short	iSizeMP2 = (int)sizeMP2 ;
	if ((sizeMP2-iSizeMP2) > 0.0001) iSizeMP2++ ;
	width = pow(2.0,iSizeNP2) ;
	height = pow(2.0,iSizeMP2) ;
	
	if (width < minimumSize) width = minimumSize ;
	if (height < minimumSize) height = minimumSize ;
			
	mFourierData = new TArray(width,height,kComplexArray) ;
	*mFourierData = 0. ;
	imageCenterX = image->Width()/2 ;
	imageCenterY = image->Height()/2 ;
	
	float	backGroundAmplitude ;
	mMaxH = 0 ;
	mMaxK = 0 ;
	short	hp,kp ;
	mFourierPeaks = new TFourierPeaks(peakArray->mNumPeaks) ;
	mFourierPeaks->mNumPeaks = 0 ;
	for(i=0; i < peakArray->mNumPeaks; i++) { 
		h = peakArray->mN[i] ;
		k = peakArray->mM[i] ;
		
		n = near(peakArray->mX[i]) ;
		m = near(peakArray->mY[i]) ;
		radsq = (n-imageCenterX)*(n-imageCenterX) + 
			(m-imageCenterY)*(m-imageCenterY) ;
		// Set only upper halfplane
		if ( radsq < apertSq && k >= 0) {
			ExtractAmplitudePhase(imageArray,n,m,amp,phase) ;
			if (k == 0 && h == 0) {
					phase = 0. ;
			}
			// What is the background amplitude at this area
			/* */
			backGroundAmplitude = (imageArray->GetPixelAmplitude(n-iBG,m) + imageArray->GetPixelAmplitude(n+iBG,m)
					+ imageArray->GetPixelAmplitude(n,m-iBG) + imageArray->GetPixelAmplitude(n,m+iBG))/4 ;
			amp -= backGroundAmplitude ;
			/* */
			if (amp < 0.) amp = 0. ;
			if (amp > 0.) {
				*mFourierData->GetRealHK(h,k) = amp ; // amp*scaleFactor ;
				*mFourierData->GetImagHK(h,k) = phase ;
				*mFourierData->GetRealHK(-h,-k) = amp ; // amp*scaleFactor ;
				*mFourierData->GetImagHK(-h,-k) = -phase ;
								
				hp = h >= 0 ? h : -h ;
				kp = k >= 0 ? k : -k ;
				if (hp > mMaxH) mMaxH = hp ;
				if (kp > mMaxK) mMaxK = kp ;
				mFourierPeaks->mImageH[mFourierPeaks->mNumPeaks] = n - imageCenterX ;
				mFourierPeaks->mImageK[mFourierPeaks->mNumPeaks] = m - imageCenterY ;
				mFourierPeaks->mLatticeH[mFourierPeaks->mNumPeaks] = h ;
				mFourierPeaks->mLatticeK[mFourierPeaks->mNumPeaks] = k ;
				mFourierPeaks->mAmplitude[mFourierPeaks->mNumPeaks] = amp ;
				mFourierPeaks->mPhase[mFourierPeaks->mNumPeaks] = phase ;
				mFourierPeaks->mNumPeaks++ ;
			}
			else {
				*mFourierData->GetRealHK(h,k) = 0. ;
				*mFourierData->GetImagHK(h,k) = 0. ;
				*mFourierData->GetRealHK(-h,-k) = 0. ;
				*mFourierData->GetImagHK(-h,-k) = 0. ;
			}
			
		}
	}
	delete peakArray ;

}

TDiffractionPeaks::TDiffractionPeaks(TPeakArray *peakArray,TLattice *lattice,float calibration,CalibrationUnit calibrationUnit) {
	// This routine uses experimental diffraction data
	// Thus we should not set the -h,-k based on anything
	long	h,k,i,width,height,minimumSize ;
	float	radtodegree = 90./::asin(1.0) ;
	float 	calibrationConstant ;
	
	mFourierData = nil ;
	mFourierPeaks = nil ;
	
	calibrationConstant = calibration ;
	
	mLattice = new TLattice(lattice) ;
	mCalibrationUnit = calibrationUnit ;
	if (mCalibrationUnit == kInverseNanometer) {
		calibrationConstant *= 0.1 ;
		mCalibrationUnit = kInverseAngstrom ;
	}

	// Now we need to put the data in proper sampling and angle stuff
	float	ux,uy,vx,vy,u,v,gamma,siga,csga,gammastar ;
	FPoint	U,V ;
	U = lattice->GetU() ;
	V = lattice->GetV() ;
	ux = calibrationConstant*U.x ;
	uy = calibrationConstant*U.y ;
	vx = calibrationConstant*V.x ;
	vy =  calibrationConstant*V.y ;
	mAStar = u  = sqrt(ux*ux+uy*uy) ;
	mBStar = v  = sqrt(vx*vx+vy*vy) ;

	gammastar = ::acos((ux*vx+uy*vy)/(u*v)) ;
	mGammaStar = radtodegree*gammastar ;

	gamma = (180-mGammaStar)*::asin(1.0)/90. ;
	siga = ::sin(gamma) ;
	csga = ::cos(gamma) ;
	gamma = 180. - mGammaStar ;

	mA = 1./(u*siga) ;
	mB = 1./(v*siga) ;
	
	minimumSize = 128 ; // Was 32
	
	// Create a new array to hold the Fourier amplitudes
	mMaxH = 0 ;
	mMaxK = 0 ;
	for(i=0; i < peakArray->mNumPeaks; i++) {
		h = peakArray->mN[i] ; if (h < 0) h = -h ;
		k = peakArray->mM[i] ; if (k < 0) k = -k ;
		if (h > mMaxH) mMaxH = h ;
		if (k > mMaxK) mMaxK = k ;
	}
	float	gMax = sqrt((mMaxH*U.x+mMaxK*V.x)*(mMaxH*U.x+mMaxK*V.x) + (mMaxH*U.y+mMaxK*V.y)*(mMaxH*U.y+mMaxK*V.y))/
				sqrt((U.x+V.x)*(U.x+V.x) + (U.y+V.y)*(U.y+V.y)) ;
	short	minHKDim = near(gMax/siga) + 1 ;
	
	// Find the closest dimensions that are power of 2
	width = 2*minHKDim ;
	height = 2*minHKDim ;
		
	float 	sizeNP2 = log2((float)width) ; //LDM: added float
	short	iSizeNP2 = (int)sizeNP2 ;
	if ((sizeNP2-iSizeNP2) > 0.0001) iSizeNP2++ ;
	float 	sizeMP2 = log2((float)height) ; //LDM: added float
	short	iSizeMP2 = (int)sizeMP2 ;
	if ((sizeMP2-iSizeMP2) > 0.0001) iSizeMP2++ ;
	width = pow(2.0,iSizeNP2) ;
	height = pow(2.0,iSizeMP2) ;
	
	if (width < minimumSize) width = minimumSize ;
	if (height < minimumSize) height = minimumSize ;
	
	mFourierData = new TArray(width,height,kComplexArray) ;
	*mFourierData = 0. ;
	for(i=0; i < peakArray->mNumPeaks; i++) {
		h = peakArray->mN[i] ;
		k = peakArray->mM[i] ;		
		*mFourierData->GetRealHK(h,k) = peakArray->mValue[i] ;
		*mFourierData->GetImagHK(h,k) = 0. ;
	}
	
	if (UseFriedelSymmetry()) {
		// Fill in missing reflections so that they are 'complex conjugate' of each other
		// Of course we do not have the phase since it is zero
		for(k=-mMaxK; k <= mMaxK; k++) 
			for(h=-mMaxH; h <= mMaxH; h++) {
				if (*mFourierData->GetRealHK(h,k) == 0. && *mFourierData->GetRealHK(-h,-k) != 0.)
					*mFourierData->GetRealHK(h,k) = *mFourierData->GetRealHK(-h,-k) ;
			}
	}
}

TDiffractionPeaks::TDiffractionPeaks(TDiffractionPeaks *peaks) {
	this->mA = peaks->mA ;
	this->mB = peaks->mB ;
	this->mAStar = peaks->mAStar ;
	this->mBStar = peaks->mBStar ;
	this->mGammaStar = peaks->mGammaStar ;
	this->mMaxH = peaks->mMaxH ;
	this->mMaxK = peaks->mMaxK ;
	this->mCalibrationUnit = peaks->mCalibrationUnit ;
	this->mImageWidth = peaks->mImageWidth ;
	this->mImageHeight = peaks->mImageHeight ;
	
	this->mFourierData = nil ;
	this->mFourierPeaks = nil ;
	this->mLattice = nil ;
	if (peaks->mFourierData != nil)
		this->mFourierData = new TArray(peaks->mFourierData) ;
	if (peaks->mFourierPeaks != nil) 
		this->mFourierPeaks = new TFourierPeaks(peaks->mFourierPeaks) ;
	if (peaks->mLattice != nil) 
		this->mLattice = new TLattice(peaks->mLattice) ;

}

TDiffractionPeaks::~TDiffractionPeaks() {
	if (mFourierData != nil)
		delete mFourierData ;
	if (mFourierPeaks != nil) 
		delete mFourierPeaks ;
	delete mLattice ;
		
}

TArray	*TDiffractionPeaks::CalculatePhaseResidual(short numPlaneGroup,float &phaseResidual,short &ixc,short &iyc) {
	// Find the origin that minimizes the phase residual
	// Symmetrize the phases and the amplitudes and report the
	// phase residual
	
	// ---------------------------------------------------------
	// For TEM images plane groups run from 1-17 no variants
	// ---------------------------------------------------------

	double	sumWeight,radToDegree,minPhaseResidual,phaseDifference ;
	float	argX,argY,s,w1,w2,w3,w4,p1,p2,p3,p4,dph,dpk,dn0,dm0,dn,dm,alpha,alphaSym ;
	short	n0,m0,ix0,iy0,Width,Height ;
	short	h,k,hMax,kMax ;
	TArray	*residuals = nil ;
	bool	hasWeight ;

	ix0 = n0 = 0 ;
	iy0 = m0 = 0 ;
	ixc = ix0 ;
	iyc = iy0 ;
	bool	aEqualB = false ;
	bool	gammaIs90 = false ;
	bool	gammaIs120 = false ;
	Width = mFourierData->GetWidth() ;
	Height = mFourierData->GetHeight() ;

	hMax = Width/2 ;
	kMax = Height/2 ;

	// Check to see if this planegroup is consistent with the data
	if (fabs((mAStar-mBStar)/((mAStar+mBStar)/2)) < 0.03) aEqualB = true ;	// 3 percent tolerance
	if (fabs(mGammaStar-90.0) < 4.0) gammaIs90 = true ;
	if (fabs(mGammaStar-60.0) < 4.0) gammaIs120 = true ;

	minPhaseResidual = 1.0e+7 ;
	phaseResidual = -1. ;
	switch(numPlaneGroup) {
		case 1:	// p1
			phaseResidual = 0. ;
			return residuals ;
			break ; 
		case 2: // p2
			break ;
		case 3: // pm - 
		case 4: // pg 
		case 5: // cm 
		case 6: // pmm  -
		case 7: // pmg  
		case 8: // pgg  
		case 9: // cmm  
			// Gamma = 90
			if (!gammaIs90) {
				phaseResidual = -1. ;
				return residuals ;
			}
			break ;
		case 10: // p4
		case 11: // p4m
		case 12: // p4g
			// Gamma = 90
			if (!gammaIs90 || !aEqualB) {
				phaseResidual = -1. ;
				return residuals ;
			}
			break ;
		case 13: // p3
		case 14: // p3m1
		case 15: // p31m
		case 16: // p6
		case 17: // p6m
			// Gamma = 120
			if (!gammaIs120 || !aEqualB) {
				phaseResidual = -1. ;
				return residuals ;
			}
			break ;
	}

	hasWeight = false ;
	sumWeight = 0. ;

	radToDegree = 180.0/Pi ;
	dn0 = TwoPi/Width ;
	dm0 = TwoPi/Height ;

	TArray	*IsSet = new TArray(Width,Height) ;
	residuals = new TArray(Width,Height) ;

	// Phase relation P(h,k) = -P(-h,-k) + 
	// Amplitude Relation  |F(h,k)| = |F(-h,-k)| + 
	switch(numPlaneGroup) {
		case 1:	// p1	We don't get here
			break ;
		case 2: // p2
			// Phase 0 or 180  : (h,k)
			// Check every position of the cell to find the origin
			minPhaseResidual = 1.0e+7 ;
			for(n0=0;n0<Width;n0++) {
				dn = n0*dn0 ;
				for(m0=0;m0<Height;m0++) {
					dm = m0*dm0 ;
					phaseResidual = 0. ;
					for(h=-hMax+1;h<hMax;h++) {
						dph = h*dn ;
						for(k=0;k<kMax;k++) {
							if (k == 0 && h <= 0) continue ;
							w1 = *mFourierData->GetRealHK(h,k) ;
							if (w1 == 0.) continue ;
							dpk = k*dm ;
							p1 = *mFourierData->GetImagHK(h,k) + dph + dpk ;
							argY = sin(p1) ;
							argX = cos(p1) ;
							alphaSym = atan2(argY,argX) ;
							alpha = alphaSym ;
							alphaSym = fabs(alphaSym) < PiHalf ? 0. : Pi ;
							phaseDifference = fabs(alpha-alphaSym) ;
							phaseDifference = Mod2Pi(phaseDifference) ;
							if (phaseDifference > Pi) 
								phaseDifference = TwoPi-phaseDifference ;
							phaseResidual += w1*phaseDifference ;
							if (!hasWeight) sumWeight += w1 ;
							
						}
					}
					hasWeight = true ;
					phaseResidual /= sumWeight ;
					*residuals->GetData(n0,m0) = phaseResidual ;
					if (phaseResidual < minPhaseResidual) {
						minPhaseResidual = phaseResidual ;
						ix0 = n0 ;
						iy0 = m0 ;
					}
				}
			}
			break ;
		case 3: // pm - 
			// |F(h,k)| = |F(-h,k)|
			// P(h,k) = P(-h,k) = -P(h,-k)
			// Phase 0 or 180  : (h,0)
			minPhaseResidual = 1.0e+7 ;
			for(n0=0;n0<Width;n0++) {
				dn = n0*dn0 ;
				for(m0=0;m0<Height;m0++) {
					dm = m0*dm0 ;
					phaseResidual = 0. ;
					*IsSet = 0. ;
					for(h=-hMax+1;h<hMax;h++) {
						dph = h*dn ;
						for(k=0;k<kMax;k++) {
							if (k == 0 && h <= 0) continue ;
							if (*IsSet->GetDataHK(h,k) == 1.) continue ;
							w1 = *mFourierData->GetRealHK(h,k) ;
							if (w1 == 0.) continue ;
							dpk = k*dm ;
							p1 = *mFourierData->GetImagHK(h,k) + dph + dpk ;
							p2 = *mFourierData->GetImagHK(-h,k) - dph + dpk ;
							w2 = *mFourierData->GetRealHK(-h,k) ;
							// Calculate symmetrized phase from P(h,k) and P(-h,k)
							argY = w1*sin(p1)+w2*sin(p2) ;
							argX = w1*cos(p1)+w2*cos(p2) ;
							alphaSym = atan2(argY,argX) ;
							if ( k == 0)	
								alphaSym = fabs(alpha) < PiHalf ? 0. : Pi ;
							*IsSet->GetDataHK(h,k) = 1. ;
							*IsSet->GetDataHK(-h,k) = 1. ;
							alpha = Mod2Pi(p1) ;
							phaseDifference = fabs(alpha-alphaSym) ;
							phaseDifference = Mod2Pi(phaseDifference) ;
							if (phaseDifference > Pi) 
								phaseDifference = TwoPi-phaseDifference ;
							phaseResidual += w1*phaseDifference ;
							alpha = Mod2Pi(p2) ;
							phaseDifference = fabs(alpha-alphaSym) ;
							phaseDifference = Mod2Pi(phaseDifference) ;
							if (phaseDifference > Pi) 
								phaseDifference = TwoPi-phaseDifference ;
							phaseResidual += w2*phaseDifference ;
							if (!hasWeight) sumWeight += w1 + w2 ;
						}
					}
					hasWeight = true ;
					phaseResidual /= sumWeight ;
					*residuals->GetData(n0,m0) = phaseResidual ;
					if (phaseResidual < minPhaseResidual) {
						minPhaseResidual = phaseResidual ;
						ix0 = n0 ;
						iy0 = m0 ;
					}
				}
			}
			break ;
		case 4: // pg  Systematic absence (0k): k = 2n+1
			// |F(h,k)| = |F(-h,k)|
			// P(h,k) = P(-h,k)+k*180
			// Phase 0 or 180  : (h,0)
			minPhaseResidual = 1.0e+7 ;
			for(n0=0;n0<Width;n0++) {
				dn = n0*dn0 ;
				for(m0=0;m0<Height;m0++) {
					dm = m0*dm0 ;
					phaseResidual = 0. ;
					*IsSet = 0. ;
					for(h=-hMax+1;h<hMax;h++) {
						dph = h*dn ;
						for(k=0;k<kMax;k++) {
							if (k == 0 && h <= 0) continue ;
							if (h == 0 && k%2 == 1) continue ;
							if (*IsSet->GetDataHK(h,k) == 1.) continue ;
							w1 = *mFourierData->GetRealHK(h,k) ;
							if (w1 == 0.) continue ;
							dpk = k*dm ;
							p1 = *mFourierData->GetImagHK(h,k) + dph + dpk ;
							p2 = *mFourierData->GetImagHK(-h,k) - dph + dpk ;
							w2 = *mFourierData->GetRealHK(-h,k) ;
							// Calculate symmetrized phase from P(h,k) and P(-h,k)
							s = k%2 == 0 ? 1. : -1. ;
							argY = w1*sin(p1)+s*w2*sin(p2) ;
							argX = w1*cos(p1)+s*w2*cos(p2) ;
							alphaSym = atan2(argY,argX) ;
							if ( k == 0)	
								alphaSym = fabs(alphaSym) < PiHalf ? 0. : Pi ;
							*IsSet->GetDataHK(h,k) = 1. ;
							*IsSet->GetDataHK(-h,k) = 1. ;
							*IsSet->GetDataHK(-h,-k) = 1. ;
							*IsSet->GetDataHK(h,-k) = 1. ;
							alpha = Mod2Pi(p1) ;
							phaseDifference = fabs(alpha-alphaSym) ;
							phaseDifference = Mod2Pi(phaseDifference) ;
							if (phaseDifference > Pi) 
								phaseDifference = TwoPi-phaseDifference ;
							phaseResidual += w1*phaseDifference ;
							if (s == -1) p2 += Pi ;
							alpha = Mod2Pi(p2) ;
							phaseDifference = fabs(alpha-alphaSym) ;
							phaseDifference = Mod2Pi(phaseDifference) ;
							if (phaseDifference > Pi) 
								phaseDifference = TwoPi-phaseDifference ;
							phaseResidual += w2*phaseDifference ;
							if (!hasWeight) sumWeight += w1 + w2 ;
						}
					}
					hasWeight = true ;
					phaseResidual /= sumWeight ;
					*residuals->GetData(n0,m0) = phaseResidual ;
					if (phaseResidual < minPhaseResidual) {
						minPhaseResidual = phaseResidual ;
						ix0 = n0 ;
						iy0 = m0 ;
					}
				}
			}
			break ;
		case 5: // cm  Systematic absence (hk): h+k = 2n+1
			// |F(h,k)| = |F(-h,k)|
			// P(h,k) = P(-h,k) = -P(h,-k)
			// Phase 0 or 180  : (h,0)
			minPhaseResidual = 1.0e+7 ;
			for(n0=0;n0<Width;n0++) {
				dn = n0*dn0 ;
				for(m0=0;m0<Height;m0++) {
					dm = m0*dm0 ;
					phaseResidual = 0. ;
					*IsSet = 0. ;
					for(h=-hMax+1;h<hMax;h++) {
						dph = h*dn ;
						for(k=0;k<kMax;k++) {
							if (k == 0 && h <= 0) continue ;
							if ((h+k)%2 == 1) continue ;
							if (*IsSet->GetDataHK(h,k) == 1.) continue ;
							w1 = *mFourierData->GetRealHK(h,k) ;
							if (w1 == 0.) continue ;
							dpk = k*dm ;
							p1 = *mFourierData->GetImagHK(h,k) + dph + dpk ;
							p2 = *mFourierData->GetImagHK(-h,k) - dph + dpk ;
							w2 = *mFourierData->GetRealHK(-h,k) ;
							// Calculate symmetrized phase from P(h,k) and P(-h,k)
							argY = w1*sin(p1)+w2*sin(p2) ;
							argX = w1*cos(p1)+w2*cos(p2) ;
							alphaSym = atan2(argY,argX) ;
							if ( k == 0)	
								alphaSym = fabs(alpha) < PiHalf ? 0. : Pi ;
							*IsSet->GetDataHK(h,k) = 1. ;
							*IsSet->GetDataHK(-h,k) = 1. ;
							alpha = Mod2Pi(p1) ;
							phaseDifference = fabs(alpha-alphaSym) ;
							phaseDifference = Mod2Pi(phaseDifference) ;
							if (phaseDifference > Pi) 
								phaseDifference = TwoPi-phaseDifference ;
							phaseResidual += 2*w1*phaseDifference ;
							alpha = Mod2Pi(p2) ;
							phaseDifference = fabs(alpha-alphaSym) ;
							phaseDifference = Mod2Pi(phaseDifference) ;
							if (phaseDifference > Pi) 
								phaseDifference = TwoPi-phaseDifference ;
							phaseResidual += 2*w2*phaseDifference ;
							if (!hasWeight) sumWeight += w1 + w2 ;
						}
					}
					hasWeight = true ;
					phaseResidual /= sumWeight ;
					*residuals->GetData(n0,m0) = phaseResidual ;
					if (phaseResidual < minPhaseResidual) {
						minPhaseResidual = phaseResidual ;
						ix0 = n0 ;
						iy0 = m0 ;
					}
				}
			}
			break ;
		case 6: // pmm  -
			// |F(h,k)| = |F(-h,k)|
			// P(h,k) = P(-h,k) = -P(h,-k)
			// Phase 0 or 180  : (h,k)
			minPhaseResidual = 1.0e+7 ;
			for(n0=0;n0<Width;n0++) {
				dn = n0*dn0 ;
				for(m0=0;m0<Height;m0++) {
					dm = m0*dm0 ;
					phaseResidual = 0. ;
					*IsSet = 0. ;
					for(h=-hMax+1;h<hMax;h++) {
						dph = h*dn ;
						for(k=0;k<kMax;k++) {
							if (k == 0 && h <= 0) continue ;
							if (*IsSet->GetDataHK(h,k) == 1.) continue ;
							w1 = *mFourierData->GetRealHK(h,k) ;
							if (w1 == 0.) continue ;
							dpk = k*dm ;
							p1 = *mFourierData->GetImagHK(h,k) + dph + dpk ;
							p2 = *mFourierData->GetImagHK(-h,k) - dph + dpk ;
							w2 = *mFourierData->GetRealHK(-h,k) ;
							// Calculate symmetrized phase from P(h,k) and P(-h,k)
							argY = w1*sin(p1)+w2*sin(p2) ;
							argX = w1*cos(p1)+w2*cos(p2) ;
							alphaSym = atan2(argY,argX) ;
							alphaSym = fabs(alphaSym) < PiHalf ? 0. : Pi ;
							*IsSet->GetDataHK(h,k) = 1. ;
							*IsSet->GetDataHK(-h,k) = 1. ;
							alpha = Mod2Pi(p1) ;
							phaseDifference = fabs(alpha-alphaSym) ;
							phaseDifference = Mod2Pi(phaseDifference) ;
							if (phaseDifference > Pi) 
								phaseDifference = TwoPi-phaseDifference ;
							phaseResidual += w1*phaseDifference ;
							alpha = Mod2Pi(p2) ;
							phaseDifference = fabs(alpha-alphaSym) ;
							phaseDifference = Mod2Pi(phaseDifference) ;
							if (phaseDifference > Pi) 
								phaseDifference = TwoPi-phaseDifference ;
							phaseResidual += w2*phaseDifference ;
							if (!hasWeight) sumWeight += w1 + w2 ;
						}
					}
					hasWeight = true ;
					phaseResidual /= sumWeight ;
					*residuals->GetData(n0,m0) = phaseResidual ;
					if (phaseResidual < minPhaseResidual) {
						minPhaseResidual = phaseResidual ;
						ix0 = n0 ;
						iy0 = m0 ;
					}
				}
			}
			break ;
		case 7: // pmg  Systematic absence (0k): k = 2n+1
			// |F(h,k)| = |F(-h,k)|
			// P(h,k) = P(-h,k)+k*180
			// Phase 0 or 180  : (h,k)
			minPhaseResidual = 1.0e+7 ;
			for(n0=0;n0<Width;n0++) {
				dn = n0*dn0 ;
				for(m0=0;m0<Height;m0++) {
					dm = m0*dm0 ;
					phaseResidual = 0. ;
					*IsSet = 0. ;
					for(h=-hMax+1;h<hMax;h++) {
						dph = h*dn ;
						for(k=0;k<kMax;k++) {
							if (k == 0 && h <= 0) continue ;
							if (h == 0 && k%2 == 1) continue ;
							if (*IsSet->GetDataHK(h,k) == 1.) continue ;
							w1 = *mFourierData->GetRealHK(h,k) ;
							if (w1 == 0.) continue ;
							dpk = k*dm ;
							p1 = *mFourierData->GetImagHK(h,k) + dph + dpk ;
							p2 = *mFourierData->GetImagHK(-h,k) - dph + dpk ;
							w2 = *mFourierData->GetRealHK(-h,k) ;
							// Calculate symmetrized phase from P(h,k) and P(-h,k)
							s = k%2 == 0 ? 1. : -1. ;
							argY = w1*sin(p1)+s*w2*sin(p2) ;
							argX = w1*cos(p1)+s*w2*cos(p2) ;
							alphaSym = atan2(argY,argX) ;
							alphaSym = fabs(alphaSym) < PiHalf ? 0. : Pi ;
							*IsSet->GetDataHK(h,k) = 1. ;
							*IsSet->GetDataHK(-h,k) = 1. ;
							alpha = Mod2Pi(p1) ;
							phaseDifference = fabs(alpha-alphaSym) ;
							phaseDifference = Mod2Pi(phaseDifference) ;
							if (phaseDifference > Pi) 
								phaseDifference = TwoPi-phaseDifference ;
							phaseResidual += w1*phaseDifference ;
							if (s == -1) p2 += Pi ;
							alpha = Mod2Pi(p2) ;
							phaseDifference = fabs(alpha-alphaSym) ;
							phaseDifference = Mod2Pi(phaseDifference) ;
							if (phaseDifference > Pi) 
								phaseDifference = TwoPi-phaseDifference ;
							phaseResidual += w2*phaseDifference ;
							if (!hasWeight) sumWeight += w1 + w2 ;
						}
					}
					hasWeight = true ;
					phaseResidual /= sumWeight ;
					*residuals->GetData(n0,m0) = phaseResidual ;
					if (phaseResidual < minPhaseResidual) {
						minPhaseResidual = phaseResidual ;
						ix0 = n0 ;
						iy0 = m0 ;
					}
				}
			}
			break ;
		case 8: // pgg  Systematic absence (h0): h = 2n+1 ; (0k): k = 2n+1
			// |F(h,k)| = |F(-h,k)|
			// P(h,k) = P(-h,k)+(h+k)*180
			// Phase 0 or 180  : (h,k)
			minPhaseResidual = 1.0e+7 ;
			for(n0=0;n0<Width;n0++) {
				dn = n0*dn0 ;
				for(m0=0;m0<Height;m0++) {
					dm = m0*dm0 ;
					phaseResidual = 0. ;
					*IsSet = 0. ;
					for(h=-hMax+1;h<hMax;h++) {
						dph = h*dn ;
						for(k=0;k<kMax;k++) {
							if (k == 0 && h <= 0) continue ;
							if (h == 0 && k%2 == 1) continue ;
							if (k == 0 && h%2 == 1) continue ;
							if (*IsSet->GetDataHK(h,k) == 1.) continue ;
							w1 = *mFourierData->GetRealHK(h,k) ;
							if (w1 == 0.) continue ;
							dpk = k*dm ;
							p1 = *mFourierData->GetImagHK(h,k) + dph + dpk ;
							p2 = *mFourierData->GetImagHK(-h,k) - dph + dpk ;
							w2 = *mFourierData->GetRealHK(-h,k) ;
							// Calculate symmetrized phase from P(h,k) and P(-h,k)
							s = (h+k)%2 == 0 ? 1. : -1. ;
							argY = w1*sin(p1)+s*w2*sin(p2) ;
							argX = w1*cos(p1)+s*w2*cos(p2) ;
							alphaSym = atan2(argY,argX) ;
							alphaSym = fabs(alphaSym) < PiHalf ? 0. : Pi ;
							*IsSet->GetDataHK(h,k) = 1. ;
							*IsSet->GetDataHK(-h,k) = 1. ;
							alpha = Mod2Pi(p1) ;
							phaseDifference = fabs(alpha-alphaSym) ;
							phaseDifference = Mod2Pi(phaseDifference) ;
							if (phaseDifference > Pi) 
								phaseDifference = TwoPi-phaseDifference ;
							phaseResidual += w1*phaseDifference ;
							if (s == -1) p2 += Pi ;
							alpha = Mod2Pi(p2) ;
							phaseDifference = fabs(alpha-alphaSym) ;
							phaseDifference = Mod2Pi(phaseDifference) ;
							if (phaseDifference > Pi) 
								phaseDifference = TwoPi-phaseDifference ;
							phaseResidual += w2*phaseDifference ;
							if (!hasWeight) sumWeight += w1 + w2 ;
						}
					}
					hasWeight = true ;
					phaseResidual /= sumWeight ;
					*residuals->GetData(n0,m0) = phaseResidual ;
					if (phaseResidual < minPhaseResidual) {
						minPhaseResidual = phaseResidual ;
						ix0 = n0 ;
						iy0 = m0 ;
					}
				}
			}
			break ;
		case 9: // cmm  Systematic absence (hk): h+k = 2n+1
			// |F(h,k)| = |F(-h,k)|
			// P(h,k) = P(-h,k) = -P(h,-k)
			// Phase 0 or 180  : (h,k)
			minPhaseResidual = 1.0e+7 ;
			for(n0=0;n0<Width;n0++) {
				dn = n0*dn0 ;
				for(m0=0;m0<Height;m0++) {
					dm = m0*dm0 ;
					phaseResidual = 0. ;
					*IsSet = 0. ;
					for(h=-hMax+1;h<hMax;h++) {
						dph = h*dn ;
						for(k=0;k<kMax;k++) {
							if (k == 0 && h <= 0) continue ;
							if ((h+k)%2 == 1) continue ;
							if (*IsSet->GetDataHK(h,k) == 1.) continue ;
							w1 = *mFourierData->GetRealHK(h,k) ;
							if (w1 == 0.) continue ;
							dpk = k*dm ;
							p1 = *mFourierData->GetImagHK(h,k) + dph + dpk ;
							p2 = *mFourierData->GetImagHK(-h,k) - dph + dpk ;
							w2 = *mFourierData->GetRealHK(-h,k) ;
							// Calculate symmetrized phase from P(h,k) and P(-h,k)
							argY = w1*sin(p1)+w2*sin(p2) ;
							argX = w1*cos(p1)+w2*cos(p2) ;
							alphaSym = atan2(argY,argX) ;
							alphaSym = fabs(alphaSym) < PiHalf ? 0. : Pi ;
							*IsSet->GetDataHK(h,k) = 1. ;
							*IsSet->GetDataHK(-h,k) = 1. ;
							alpha = Mod2Pi(p1) ;
							phaseDifference = fabs(alpha-alphaSym) ;
							phaseDifference = Mod2Pi(phaseDifference) ;
							if (phaseDifference > Pi) 
								phaseDifference = TwoPi-phaseDifference ;
							phaseResidual += w1*phaseDifference ;
							alpha = Mod2Pi(p2) ;
							phaseDifference = fabs(alpha-alphaSym) ;
							phaseDifference = Mod2Pi(phaseDifference) ;
							if (phaseDifference > Pi) 
								phaseDifference = TwoPi-phaseDifference ;
							phaseResidual += w2*phaseDifference ;
							if (!hasWeight) sumWeight += w1 + w2 ;
						}
					}
					hasWeight = true ;
					phaseResidual /= sumWeight ;
					*residuals->GetData(n0,m0) = phaseResidual ;
					if (phaseResidual < minPhaseResidual) {
						minPhaseResidual = phaseResidual ;
						ix0 = n0 ;
						iy0 = m0 ;
					}
				}
			}
			break ;
		case 10: // p4
			// |F(h,k)| = |F(-k,h)|
			// P(h,k) = P(-k,h) = -P(k,-h)
			// Phase 0 or 180  : (h,k)
			minPhaseResidual = 1.0e+7 ;
			for(n0=0;n0<Width;n0++) {
				dn = n0*dn0 ;
				for(m0=0;m0<Height;m0++) {
					dm = m0*dm0 ;
					phaseResidual = 0. ;
					*IsSet = 0. ;
					for(h=-hMax+1;h<hMax;h++) {
						dph = h*dn ; 
						for(k=0;k<kMax;k++) {
							if (k == 0 && h <= 0) continue ;
							if (*IsSet->GetDataHK(h,k) == 1.) continue ;
							w1 = *mFourierData->GetRealHK(h,k) ;
							if (w1 == 0.) continue ;
							dpk = k*dm ; 
							p1 = *mFourierData->GetImagHK(h,k) + dph + dpk ;
							p2 = *mFourierData->GetImagHK(-k,h) - k*dn + h*dm ;
							w2 = *mFourierData->GetRealHK(-k,h) ;
							// Calculate symmetrized phase from P(h,k) and P(-h,k)
							argY = w1*sin(p1)+w2*sin(p2) ;
							argX = w1*cos(p1)+w2*cos(p2) ;
							alphaSym = atan2(argY,argX) ;
							alphaSym = fabs(alpha) < PiHalf ? 0. : Pi ;
							*IsSet->GetDataHK(h,k) = 1. ;
							*IsSet->GetDataHK(-k,h) = 1. ;
							*IsSet->GetDataHK(k,-h) = 1. ;
							alpha = Mod2Pi(p1) ;
							phaseDifference = fabs(alpha-alphaSym) ;
							phaseDifference = Mod2Pi(phaseDifference) ;
							if (phaseDifference > Pi) 
								phaseDifference = TwoPi-phaseDifference ;
							phaseResidual += w1*phaseDifference ;
							alpha = Mod2Pi(p2) ;
							phaseDifference = fabs(alpha-alphaSym) ;
							phaseDifference = Mod2Pi(phaseDifference) ;
							if (phaseDifference > Pi) 
								phaseDifference = TwoPi-phaseDifference ;
							phaseResidual += w2*phaseDifference ;
							if (!hasWeight) sumWeight += w1 + w2 ;
						}
					}
					hasWeight = true ;
					phaseResidual /= sumWeight ;
					*residuals->GetData(n0,m0) = phaseResidual ;
					if (phaseResidual < minPhaseResidual) {
						minPhaseResidual = phaseResidual ;
						ix0 = n0 ;
						iy0 = m0 ;
					}
				}
			}
			break ;
		case 11: // p4m
			// |F(h,k)| = |F(-k,h)| = |F(-h,k)|
			// P(h,k) = P(-k,h) = P(-h,k) = -P(k,-h) = -P(h,-k)
			// Phase 0 or 180  : (h,k)
			minPhaseResidual = 1.0e+7 ;
			for(n0=0;n0<Width;n0++) {
				dn = n0*dn0 ;
				for(m0=0;m0<Height;m0++) {
					dm = m0*dm0 ;
					phaseResidual = 0. ;
					*IsSet = 0. ;
					for(h=-hMax+1;h<hMax;h++) {
						dph = h*dn ; 
						for(k=0;k<kMax;k++) {
							if (k == 0 && h <= 0) continue ;
							if (*IsSet->GetDataHK(h,k) == 1.) continue ;
							w1 = *mFourierData->GetRealHK(h,k) ;
							if (w1 == 0.) continue ;
							dpk = k*dm ; 
							p1 = *mFourierData->GetImagHK(h,k) + dph + dpk ;
							p2 = *mFourierData->GetImagHK(-h,k) - dph + dpk ;
							p3 = *mFourierData->GetImagHK(-k,h) - k*dn + h*dm ;
							w2 = *mFourierData->GetRealHK(-h,k) ;
							w3 = *mFourierData->GetRealHK(-k,h) ;
							// Calculate symmetrized phase from P(h,k) and P(-h,k)
							argY = w1*sin(p1)+w2*sin(p2)+w3*sin(p3) ;
							argX = w1*cos(p1)+w2*cos(p2)+w3*cos(p3) ;
							alphaSym = atan2(argY,argX) ;
							alphaSym = fabs(alphaSym) < PiHalf ? 0. : Pi ;
							*IsSet->GetDataHK(h,k) = 1. ;
							*IsSet->GetDataHK(-h,k) = 1. ;
							*IsSet->GetDataHK(-k,h) = 1. ;
							*IsSet->GetDataHK(-h,-k) = 1. ;
							*IsSet->GetDataHK(h,-k) = 1. ;
							*IsSet->GetDataHK(k,-h) = 1. ;
							alpha = Mod2Pi(p1) ;
							phaseDifference = fabs(alpha-alphaSym) ;
							phaseDifference = Mod2Pi(phaseDifference) ;
							if (phaseDifference > Pi) 
								phaseDifference = TwoPi-phaseDifference ;
							phaseResidual += w1*phaseDifference ;
							alpha = Mod2Pi(p2) ;
							phaseDifference = fabs(alpha-alphaSym) ;
							phaseDifference = Mod2Pi(phaseDifference) ;
							if (phaseDifference > Pi) 
								phaseDifference = TwoPi-phaseDifference ;
							phaseResidual += w2*phaseDifference ;
							alpha = Mod2Pi(p3) ;
							phaseDifference = fabs(alpha-alphaSym) ;
							phaseDifference = Mod2Pi(phaseDifference) ;
							if (phaseDifference > Pi) 
								phaseDifference = TwoPi-phaseDifference ;
							phaseResidual += w3*phaseDifference ;
							if (!hasWeight) sumWeight += w1 + w2 + w3 ;
						}
					}
					hasWeight = true ;
					phaseResidual /= sumWeight ;
					*residuals->GetData(n0,m0) = phaseResidual ;
					if (phaseResidual < minPhaseResidual) {
						minPhaseResidual = phaseResidual ;
						ix0 = n0 ;
						iy0 = m0 ;
					}
				}
			}
			break ;
		case 12: // p4g
			// |F(h,k)| = |F(-k,h)| = |F(-h,k)|
			// P(h,k) = P(-k,h) = P(-h,k)+k*180
			// Phase 0 or 180  : (h,k)
			minPhaseResidual = 1.0e+7 ;
			for(n0=0;n0<Width;n0++) {
				dn = n0*dn0 ;
				for(m0=0;m0<Height;m0++) {
					dm = m0*dm0 ;
					phaseResidual = 0. ;
					*IsSet = 0. ;
					for(h=-hMax+1;h<hMax;h++) {
						dph = h*dn ; 
						for(k=0;k<kMax;k++) {
							if (k == 0 && h <= 0) continue ;
							if (*IsSet->GetDataHK(h,k) == 1.) continue ;
							w1 = *mFourierData->GetRealHK(h,k) ;
							if (w1 == 0.) continue ;
							dpk = k*dm ; 
							p1 = *mFourierData->GetImagHK(h,k) + dph + dpk ;
							p2 = *mFourierData->GetImagHK(-h,k) - dph + dpk ;
							p3 = *mFourierData->GetImagHK(-k,h) - k*dn + h*dm ;
							w2 = *mFourierData->GetRealHK(-h,k) ;
							w3 = *mFourierData->GetRealHK(-k,h) ;
							// Calculate symmetrized phase from P(h,k) and P(-h,k)
							s = k%2 == 0 ? 1. : -1. ;
							argY = w1*sin(p1)+s*w2*sin(p2)+w3*sin(p3) ;
							argX = w1*cos(p1)+s*w2*cos(p2)+w3*cos(p3) ;
							alphaSym = atan2(argY,argX) ;
							alphaSym = fabs(alphaSym) < PiHalf ? 0. : Pi ;
							*IsSet->GetDataHK(h,k) = 1. ;
							*IsSet->GetDataHK(-h,k) = 1. ;
							*IsSet->GetDataHK(-k,h) = 1. ;
							*IsSet->GetDataHK(-h,-k) = 1. ;
							*IsSet->GetDataHK(h,-k) = 1. ;
							*IsSet->GetDataHK(k,-h) = 1. ;
							alpha = Mod2Pi(p1) ;
							phaseDifference = fabs(alpha-alphaSym) ;
							phaseDifference = Mod2Pi(phaseDifference) ;
							if (phaseDifference > Pi) 
								phaseDifference = TwoPi-phaseDifference ;
							phaseResidual += w1*phaseDifference ;
							if (k%2 != 0) p2 += Pi ;
							alpha = Mod2Pi(p2) ;
							phaseDifference = fabs(alpha-alphaSym) ;
							phaseDifference = Mod2Pi(phaseDifference) ;
							if (phaseDifference > Pi) 
								phaseDifference = TwoPi-phaseDifference ;
							phaseResidual += w2*phaseDifference ;
							alpha = Mod2Pi(p3) ;
							phaseDifference = fabs(alpha-alphaSym) ;
							phaseDifference = Mod2Pi(phaseDifference) ;
							if (phaseDifference > Pi) 
								phaseDifference = TwoPi-phaseDifference ;
							phaseResidual += w3*phaseDifference ;
							if (!hasWeight) sumWeight += w1 + w2 + w3 ;
						}
					}
					hasWeight = true ;
					phaseResidual /= sumWeight ;
					*residuals->GetData(n0,m0) = phaseResidual ;
					if (phaseResidual < minPhaseResidual) {
						minPhaseResidual = phaseResidual ;
						ix0 = n0 ;
						iy0 = m0 ;
					}
				}
			}
			break ;
		case 13: // p3
			// |F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)|
			// P(h,k) = P(k,-h-k) = P(-h-k,h)
			// Phase 0 or 180  : -
			minPhaseResidual = 1.0e+7 ;
			for(n0=0;n0<Width;n0++) {
				dn = n0*dn0 ;
				for(m0=0;m0<Height;m0++) {
					dm = m0*dm0 ;
					phaseResidual = 0. ;
					*IsSet = 0. ;
					for(h=-hMax+1;h<hMax;h++) {
						dph = h*dn ; 
						for(k=0;k<kMax;k++) {
							if (k == 0 && h <= 0) continue ;
							if (*IsSet->GetDataHK(h,k) == 1.) continue ;
							w1 = *mFourierData->GetRealHK(h,k) ;
							if (w1 == 0.) continue ;
							dpk = k*dm ; 
							p1 = *mFourierData->GetImagHK(h,k) + dph + dpk ;
							p2 = *mFourierData->GetImagHK(k,-h-k) +k*dn - (h+k)*dm ;
							p3 = *mFourierData->GetImagHK(-h-k,h) -(h+k)*dn + h*dm ;
							w2 = *mFourierData->GetRealHK(k,-h-k) ;
							w3 = *mFourierData->GetRealHK(-h-k,h) ;
							argY = w1*sin(p1)+w2*sin(p2)+w3*sin(p3) ;
							argX = w1*cos(p1)+w2*cos(p2)+w3*cos(p3) ;
							alphaSym = atan2(argY,argX) ;
							*IsSet->GetDataHK(h,k) = 1. ;
							*IsSet->GetDataHK(k,-h-k) = 1. ;
							*IsSet->GetDataHK(-h-k,h) = 1. ;
							*IsSet->GetDataHK(-h,-k) = 1. ;
							*IsSet->GetDataHK(-k,h+k) = 1. ;
							*IsSet->GetDataHK(h+k,-h) = 1. ;
							alpha = atan2(sin(p1),cos(p1)) ;
							//alpha = Mod2Pi(p1) ;
							phaseDifference = fabs(alpha-alphaSym) ;
							phaseDifference = Mod2Pi(phaseDifference) ;
							if (phaseDifference > Pi) 
								phaseDifference = TwoPi-phaseDifference ;
							phaseResidual += w1*phaseDifference ;
							alpha = atan2(sin(p2),cos(p2)) ;
							//alpha = Mod2Pi(p2) ;
							phaseDifference = fabs(alpha-alphaSym) ;
							phaseDifference = Mod2Pi(phaseDifference) ;
							if (phaseDifference > Pi) 
								phaseDifference = TwoPi-phaseDifference ;
							phaseResidual += w2*phaseDifference ;
							alpha = atan2(sin(p3),cos(p3)) ;
							//alpha = Mod2Pi(p3) ;
							phaseDifference = fabs(alpha-alphaSym) ;
							phaseDifference = Mod2Pi(phaseDifference) ;
							if (phaseDifference > Pi) 
								phaseDifference = TwoPi-phaseDifference ;
							phaseResidual += w3*phaseDifference ;
							if (!hasWeight) sumWeight += w1 + w2 + w3 ;
						}
					}
					hasWeight = true ;
					phaseResidual /= sumWeight ;
					*residuals->GetData(n0,m0) = phaseResidual ;
					if (phaseResidual < minPhaseResidual) {
						minPhaseResidual = phaseResidual ;
						ix0 = n0 ;
						iy0 = m0 ;
					}
				}
			}
			break ;
		case 14: // p3m1
			// |F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)| = |F(k,h)|
			// P(h,k) = P(k,-h-k) = P(-h-k,h) = P(-k,-h)
			// Phase 0 or 180  : (h,h) (h,-2h) (-2k,k)
			minPhaseResidual = 1.0e+7 ;
			for(n0=0;n0<Width;n0++) {
				dn = n0*dn0 ;
				for(m0=0;m0<Height;m0++) {
					dm = m0*dm0 ;
					phaseResidual = 0. ;
					*IsSet = 0. ;
					for(h=-hMax+1;h<hMax;h++) {
						dph = h*dn ; 
						for(k=0;k<kMax;k++) {
							if (k == 0 && h <= 0) continue ;
							if (*IsSet->GetDataHK(h,k) == 1.) continue ;
							w1 = *mFourierData->GetRealHK(h,k) ;
							if (w1 == 0.) continue ;
							dpk = k*dm ; 
							p1 = *mFourierData->GetImagHK(h,k) + dph + dpk ;
							p2 = *mFourierData->GetImagHK(k,-h-k) +k*dn - (h+k)*dm ;
							p3 = *mFourierData->GetImagHK(-h-k,h) -(h+k)*dn + h*dm ;
							p4 = *mFourierData->GetImagHK(-k,-h) - k*dn - h*dm ;
							w2 = *mFourierData->GetRealHK(k,-h-k) ;
							w3 = *mFourierData->GetRealHK(-h-k,h) ;
							w4 = *mFourierData->GetRealHK(-k,-h) ;
							argY = w1*sin(p1)+w2*sin(p2)+w3*sin(p3)+w4*sin(p4) ;
							argX = w1*cos(p1)+w2*cos(p2)+w3*cos(p3)+w4*cos(p4) ;
							alphaSym = atan2(argY,argX) ;
							if (h == k || k == -2*h || h == -2*k)
								alphaSym = fabs(alphaSym) < PiHalf ? 0. : Pi ;
							*IsSet->GetDataHK(h,k) = 1. ;
							*IsSet->GetDataHK(k,-h-k) = 1. ;
							*IsSet->GetDataHK(-h-k,h) = 1. ;
							*IsSet->GetDataHK(-k,-h) = 1. ;
							*IsSet->GetDataHK(-h,-k) = 1. ;
							*IsSet->GetDataHK(-k,h+k) = 1. ;
							*IsSet->GetDataHK(h+k,-h) = 1. ;
							*IsSet->GetDataHK(k,h) = 1. ;
							alpha = Mod2Pi(p1) ;
							phaseDifference = fabs(alpha-alphaSym) ;
							phaseDifference = Mod2Pi(phaseDifference) ;
							if (phaseDifference > Pi) 
								phaseDifference = TwoPi-phaseDifference ;
							phaseResidual += w1*phaseDifference ;
							alpha = Mod2Pi(p2) ;
							phaseDifference = fabs(alpha-alphaSym) ;
							phaseDifference = Mod2Pi(phaseDifference) ;
							if (phaseDifference > Pi) 
								phaseDifference = TwoPi-phaseDifference ;
							phaseResidual += w2*phaseDifference ;
							alpha = Mod2Pi(p3) ;
							phaseDifference = fabs(alpha-alphaSym) ;
							phaseDifference = Mod2Pi(phaseDifference) ;
							if (phaseDifference > Pi) 
								phaseDifference = TwoPi-phaseDifference ;
							phaseResidual += w3*phaseDifference ;
							alpha = Mod2Pi(p4) ;
							phaseDifference = fabs(alpha-alphaSym) ;
							phaseDifference = Mod2Pi(phaseDifference) ;
							if (phaseDifference > Pi) 
								phaseDifference = TwoPi-phaseDifference ;
							phaseResidual += w4*phaseDifference ;
							if (!hasWeight) sumWeight += w1 + w2 + w3 + w4 ;
						}
					}
					hasWeight = true ;
					phaseResidual /= sumWeight ;
					*residuals->GetData(n0,m0) = phaseResidual ;
					if (phaseResidual < minPhaseResidual) {
						minPhaseResidual = phaseResidual ;
						ix0 = n0 ;
						iy0 = m0 ;
					}
				}
			}
			break ;
		case 15: // p31m
			// |F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)| = |F(k,h)|
			// P(h,k) = P(k,-h-k) = P(-h-k,h) = P(k,h)
			// Phase 0 or 180  : (h,-h) (h,2h) (2k,k)
			minPhaseResidual = 1.0e+7 ;
			for(n0=0;n0<Width;n0++) {
				dn = n0*dn0 ;
				for(m0=0;m0<Height;m0++) {
					dm = m0*dm0 ;
					phaseResidual = 0. ;
					*IsSet = 0. ;
					for(h=-hMax+1;h<hMax;h++) {
						dph = h*dn ; 
						for(k=0;k<kMax;k++) {
							if (k == 0 && h <= 0) continue ;
							if (*IsSet->GetDataHK(h,k) == 1.) continue ;
							w1 = *mFourierData->GetRealHK(h,k) ;
							if (w1 == 0.) continue ;
							dpk = k*dm ; 
							p1 = *mFourierData->GetImagHK(h,k) + dph + dpk ;
							p2 = *mFourierData->GetImagHK(k,-h-k) +k*dn - (h+k)*dm ;
							p3 = *mFourierData->GetImagHK(-h-k,h) -(h+k)*dn + h*dm ;
							p4 = *mFourierData->GetImagHK(k,h) + k*dn + h*dm ;
							w2 = *mFourierData->GetRealHK(k,-h-k) ;
							w3 = *mFourierData->GetRealHK(-h-k,h) ;
       							w4 = h == k ? 0. : *mFourierData->GetRealHK(k,h) ;
							argY = w1*sin(p1)+w2*sin(p2)+w3*sin(p3)+w4*sin(p4) ;
							argX = w1*cos(p1)+w2*cos(p2)+w3*cos(p3)+w4*cos(p4) ;
							alphaSym = atan2(argY,argX) ;
							if (h == -k || h == 2*k || k == 2*h)
								alphaSym = fabs(alphaSym) < PiHalf ? 0. : Pi ;
							*IsSet->GetDataHK(h,k) = 1. ;
							*IsSet->GetDataHK(k,-h-k) = 1. ;
							*IsSet->GetDataHK(-h-k,h) = 1. ;
							*IsSet->GetDataHK(k,h) = 1. ;
							*IsSet->GetDataHK(-h,-k) = 1. ;
							*IsSet->GetDataHK(-k,h+k) = 1. ;
							*IsSet->GetDataHK(h+k,-h) = 1. ;
							*IsSet->GetDataHK(-k,-h) = 1. ;
							alpha = Mod2Pi(p1) ;
							phaseDifference = fabs(alpha-alphaSym) ;
							phaseDifference = Mod2Pi(phaseDifference) ;
							if (phaseDifference > Pi) 
								phaseDifference = TwoPi-phaseDifference ;
							phaseResidual += w1*phaseDifference ;
							alpha = Mod2Pi(p2) ;
							phaseDifference = fabs(alpha-alphaSym) ;
							phaseDifference = Mod2Pi(phaseDifference) ;
							if (phaseDifference > Pi) 
								phaseDifference = TwoPi-phaseDifference ;
							phaseResidual += w2*phaseDifference ;
							alpha = Mod2Pi(p3) ;
							phaseDifference = fabs(alpha-alphaSym) ;
							phaseDifference = Mod2Pi(phaseDifference) ;
							if (phaseDifference > Pi) 
								phaseDifference = TwoPi-phaseDifference ;
							phaseResidual += w3*phaseDifference ;
							alpha = Mod2Pi(p4) ;
							phaseDifference = fabs(alpha-alphaSym) ;
							phaseDifference = Mod2Pi(phaseDifference) ;
							if (phaseDifference > Pi) 
								phaseDifference = TwoPi-phaseDifference ;
							phaseResidual += w4*phaseDifference ;
							if (!hasWeight) sumWeight += w1 + w2 + w3 + w4 ;
						}
					}
					hasWeight = true ;
					phaseResidual /= sumWeight ;
					*residuals->GetData(n0,m0) = phaseResidual ;
					if (phaseResidual < minPhaseResidual) {
						minPhaseResidual = phaseResidual ;
						ix0 = n0 ;
						iy0 = m0 ;
					}
				}
			}
			break ;
		case 16: // p6
			// |F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)|
			// P(h,k) = P(k,-h-k) = P(-h-k,h)
			// Phase 0 or 180  : (h,k)
			minPhaseResidual = 1.0e+7 ;
			for(n0=0;n0<Width;n0++) {
				dn = n0*dn0 ;
				for(m0=0;m0<Height;m0++) {
					dm = m0*dm0 ;
					phaseResidual = 0. ;
					*IsSet = 0. ;
					for(h=-hMax+1;h<hMax;h++) {
						dph = h*dn ; 
						for(k=0;k<kMax;k++) {
							if (k == 0 && h <= 0) continue ;
							if (*IsSet->GetDataHK(h,k) == 1.) continue ;
							w1 = *mFourierData->GetRealHK(h,k) ;
							if (w1 == 0.) continue ;
							dpk = k*dm ; 
							p1 = *mFourierData->GetImagHK(h,k) + dph + dpk ;
							p1 = *mFourierData->GetImagHK(h,k) + dph + dpk ;
							p2 = *mFourierData->GetImagHK(k,-h-k) +k*dn - (h+k)*dm ;
							p3 = *mFourierData->GetImagHK(-h-k,h) -(h+k)*dn + h*dm ;
							w2 = *mFourierData->GetRealHK(k,-h-k) ;
							w3 = *mFourierData->GetRealHK(-h-k,h) ;
							argY = w1*sin(p1)+w2*sin(p2)+w3*sin(p3) ;
							argX = w1*cos(p1)+w2*cos(p2)+w3*cos(p3) ;
							alphaSym = atan2(argY,argX) ;
							alphaSym = fabs(alphaSym) < PiHalf ? 0. : Pi ;
							*IsSet->GetDataHK(h,k) = 1. ;
							*IsSet->GetDataHK(k,-h-k) = 1. ;
							*IsSet->GetDataHK(-h-k,h) = 1. ;
							*IsSet->GetDataHK(-h,-k) = 1. ;
							*IsSet->GetDataHK(-k,h+k) = 1. ;
							*IsSet->GetDataHK(h+k,-h) = 1. ;
							alpha = atan2(sin(p1),cos(p1)) ;
							//alpha = Mod2Pi(p1) ;
							phaseDifference = fabs(alpha-alphaSym) ;
							phaseDifference = Mod2Pi(phaseDifference) ;
							if (phaseDifference > Pi) 
								phaseDifference = TwoPi-phaseDifference ;
							phaseResidual += w1*phaseDifference ;
							alpha = atan2(sin(p2),cos(p2)) ;
							//alpha = Mod2Pi(p2) ;
							phaseDifference = fabs(alpha-alphaSym) ;
							phaseDifference = Mod2Pi(phaseDifference) ;
							if (phaseDifference > Pi) 
								phaseDifference = TwoPi-phaseDifference ;
							phaseResidual += w2*phaseDifference ;
							alpha = atan2(sin(p3),cos(p3)) ;
							//alpha = Mod2Pi(p3) ;
							phaseDifference = fabs(alpha-alphaSym) ;
							phaseDifference = Mod2Pi(phaseDifference) ;
							if (phaseDifference > Pi) 
								phaseDifference = TwoPi-phaseDifference ;
							phaseResidual += w3*phaseDifference ;
							if (!hasWeight) sumWeight += w1 + w2 + w3 ;
						}
					}
					hasWeight = true ;
					phaseResidual /= sumWeight ;
					*residuals->GetData(n0,m0) = phaseResidual ;
					if (phaseResidual < minPhaseResidual) {
						minPhaseResidual = phaseResidual ;
						ix0 = n0 ;
						iy0 = m0 ;
					}
				}
			}
			break ;
		case 17: // p6m
			// |F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)| = |F(k,h)|
			// P(h,k) = P(k,-h-k) = P(-h-k,h) = P(k,h)
			// Phase 0 or 180  : (h,k)
			minPhaseResidual = 1.0e+7 ;
			for(n0=0;n0<Width;n0++) {
				dn = n0*dn0 ;
				for(m0=0;m0<Height;m0++) {
					dm = m0*dm0 ;
					phaseResidual = 0. ;
					*IsSet = 0. ;
					for(h=-hMax+1;h<hMax;h++) {
						dph = h*dn ; 
						for(k=0;k<kMax;k++) {
							if (k == 0 && h <= 0) continue ;
							if (*IsSet->GetDataHK(h,k) == 1.) continue ;
							w1 = *mFourierData->GetRealHK(h,k) ;
							if (w1 == 0.) continue ;
							dpk = k*dm ; 
							p1 = *mFourierData->GetImagHK(h,k) + dph + dpk ;
							p2 = *mFourierData->GetImagHK(k,-h-k) +k*dn - (h+k)*dm ;
							p3 = *mFourierData->GetImagHK(-h-k,h) -(h+k)*dn + h*dm ;
							p4 = *mFourierData->GetImagHK(k,h) + k*dn + h*dm ;
							w2 = *mFourierData->GetRealHK(k,-h-k) ;
							w3 = *mFourierData->GetRealHK(-h-k,h) ;
							w4 = h == k ? 0. : *mFourierData->GetRealHK(k,h) ;
							argY = w1*sin(p1)+w2*sin(p2)+w3*sin(p3)+w4*sin(p4) ;
							argX = w1*cos(p1)+w2*cos(p2)+w3*cos(p3)+w4*cos(p4) ;
							alphaSym = atan2(argY,argX) ;
							alphaSym = fabs(alphaSym) < PiHalf ? 0. : Pi ;
							*IsSet->GetDataHK(h,k) = 1. ;
							*IsSet->GetDataHK(k,-h-k) = 1. ;
							*IsSet->GetDataHK(-h-k,h) = 1. ;
							*IsSet->GetDataHK(k,h) = 1. ;
							*IsSet->GetDataHK(-h,-k) = 1. ;
							*IsSet->GetDataHK(-k,h+k) = 1. ;
							*IsSet->GetDataHK(h+k,-h) = 1. ;
							*IsSet->GetDataHK(-k,-h) = 1. ;
							alpha = Mod2Pi(p1) ;
							phaseDifference = fabs(alpha-alphaSym) ;
							phaseDifference = Mod2Pi(phaseDifference) ;
							if (phaseDifference > Pi) 
								phaseDifference = TwoPi-phaseDifference ;
							phaseResidual += 2*w1*phaseDifference ;
							alpha = Mod2Pi(p2) ;
							phaseDifference = fabs(alpha-alphaSym) ;
							phaseDifference = Mod2Pi(phaseDifference) ;
							if (phaseDifference > Pi) 
								phaseDifference = TwoPi-phaseDifference ;
							phaseResidual += 2*w2*phaseDifference ;
							alpha = Mod2Pi(p3) ;
							phaseDifference = fabs(alpha-alphaSym) ;
							phaseDifference = Mod2Pi(phaseDifference) ;
							if (phaseDifference > Pi) 
								phaseDifference = TwoPi-phaseDifference ;
							phaseResidual += w3*phaseDifference ;
							alpha = Mod2Pi(p4) ;
							phaseDifference = fabs(alpha-alphaSym) ;
							phaseDifference = Mod2Pi(phaseDifference) ;
							if (phaseDifference > Pi) 
								phaseDifference = TwoPi-phaseDifference ;
							phaseResidual += w4*phaseDifference ;
							if (!hasWeight) sumWeight += w1 + w2 + w3 + w4 ;
						}
					}
					hasWeight = true ;
					phaseResidual /= sumWeight ;
					*residuals->GetData(n0,m0) = phaseResidual ;
					if (phaseResidual < minPhaseResidual) {
						minPhaseResidual = phaseResidual ;
						ix0 = n0 ;
						iy0 = m0 ;
					}
				}
			}
			break ;
	}
	
	ixc = ix0 ;
	iyc = iy0 ;
	phaseResidual = minPhaseResidual*radToDegree ;

	delete	IsSet ;
	return residuals ;
}

float	TDiffractionPeaks::CalculateAmplitudeResidual(short numPlaneGroup) {
	long	h,k,numTerms ;
	float	w1,amplitude,latticeParameter ;
	bool	aEqualB = false ;
	bool	gammaIs90 = false ;
	bool	gammaIs120 = false ;
	long	Width = mFourierData->GetWidth() ;
	long	Height = mFourierData->GetHeight() ;

	long	hMax = Width/2 ;
	long	kMax = Height/2 ;

	// Check to see if this planegroup is consistent with the
	// data

	// ---------------------------------------------------------
	// For Diffraction patterns,  plane groups run from 1-22
	// ---------------------------------------------------------

	if (fabs(mAStar-mBStar) < 0.03*mAStar) aEqualB = true ;	// 3 percent tolerance
	if (fabs(mGammaStar-90.0) < 4.0) gammaIs90 = true ;
	if (fabs(mGammaStar-60.0) < 4.0) gammaIs120 = true ;

	float 	amplitudeResidual = -1. ;
	float	rmsAmplitude = 0. ;
	float	amplitudeSum = 0. ;

	bool	useFriedelSymmetry = UseFriedelSymmetry() ;
	
	switch(numPlaneGroup) {
		case 1:	// p1
			return 0. ;
			break ; 
		case 2: // p2
			break ;
		case 3: // pm 
		case 4: // p11m 
		case 5: // pg 
		case 6: // p11g
		case 7: // cm  
		case 8: // c11m 
		case 9: // p2mm  
		case 10: // p2mg  
		case 11: // p2gm
		case 12: // pgg
		case 13: // cmm
			// Gamma = 90
			if (!gammaIs90) {
				return amplitudeResidual ;
			}
			break ;
		case 14: // p4
		case 15: // p4mm
		case 16: // p4g
		case 17: // p4mg
			// Gamma = 90
			if (!gammaIs90 || !aEqualB) {
				return amplitudeResidual ;
			}
			break ;
		case 18: // p3
		case 19: // p3m1
		case 20: // p31m
		case 21: // p6
		case 22: // p6mm
			// Gamma = 120
			if (!gammaIs120 || !aEqualB) {
				return amplitudeResidual ;
			}
			break ;
	}

	TArray	*IsSet = new TArray(Width,Height) ;
	*IsSet = 0. ;
	
	// Amplitude Relation  |F(h,k)| = |F(-h,-k)| + 
	switch(numPlaneGroup) {
		case 2: // p2
			for(h=-hMax;h<=hMax;h++) {
				for(k=0;k<=kMax;k++) {
					if (k == 0 && h <= 0) continue ;
					if (*IsSet->GetDataHK(h,k) == 1.) continue ;
					*IsSet->GetDataHK(h,k) = 1. ;
					numTerms = 0 ;
					amplitude = w1 = *mFourierData->GetRealHK(h,k) ;
					if (w1 > 0.) numTerms++ ;
					w1 = *IsSet->GetDataHK(-h,-k) == 1. ? 0. : *mFourierData->GetRealHK(-h,-k) ;
					amplitude += w1 = *IsSet->GetDataHK(-h,-k) == 1. ? 0. : *mFourierData->GetRealHK(-h,-k) ;
					*IsSet->GetDataHK(-h,-k) = 1. ;
					if (w1 > 0.) numTerms++ ;
					amplitudeSum += amplitude*amplitude ;
					if (numTerms > 0) amplitude /= numTerms ;
					rmsAmplitude += (*mFourierData->GetRealHK(h,k) - amplitude)*(*mFourierData->GetRealHK(h,k) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(-h,-k) - amplitude)*(*mFourierData->GetRealHK(-h,-k) - amplitude) ;
				}
			}
			break ;
		case 3: // pm - 
		case 4: // p11m 
			// |F(h,k)| = |F(-h,k)|
			for(h=-hMax;h<=hMax;h++) {
				for(k=0;k<=kMax;k++) {
					if (k == 0 && h <= 0) continue ;
					if (*IsSet->GetDataHK(h,k) == 1.) continue ;
					*IsSet->GetDataHK(h,k) = 1. ;
					numTerms = 0 ;
					amplitude = w1 = *mFourierData->GetRealHK(h,k) ;
					if (w1 > 0.) numTerms++ ;
					
					amplitude += w1 = *IsSet->GetDataHK(-h,-k) == 1. ? 0. : *mFourierData->GetRealHK(-h,-k) ;
					*IsSet->GetDataHK(-h,-k) = 1. ;
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(-h,k) == 1. ? 0. : *mFourierData->GetRealHK(-h,k) ;
					*IsSet->GetDataHK(-h,k) = 1. ;
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(h,-k) == 1. ? 0. : *mFourierData->GetRealHK(h,-k) ;
					*IsSet->GetDataHK(h,-k) = 1. ;
					if (w1 > 0.) numTerms++ ;
					amplitudeSum += amplitude*amplitude ;
					if (numTerms > 0) amplitude /= numTerms ;
					rmsAmplitude += (*mFourierData->GetRealHK(h,k) - amplitude)*(*mFourierData->GetRealHK(h,k) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(-h,k) - amplitude)*(*mFourierData->GetRealHK(-h,k) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(-h,-k) - amplitude)*(*mFourierData->GetRealHK(-h,-k) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(h,-k) - amplitude)*(*mFourierData->GetRealHK(h,-k) - amplitude) ;
				}
			}
			break ;
		case 5: // pg  Systematic absence (0k): k = 2n+1
			// |F(h,k)| = |F(-h,k)|
			for(h=-hMax;h<=hMax;h++) {
				for(k=0;k<=kMax;k++) {
					if (k == 0 && h <= 0) continue ;
					if (*IsSet->GetDataHK(h,k) == 1.) continue ;
					*IsSet->GetDataHK(h,k) = 1. ;
					if (h == 0 && k%2 == 1) {
						// Systematic absence
						amplitude = 0. ;
						*IsSet->GetDataHK(h,k) = 1. ;
						*IsSet->GetDataHK(-h,k) = 1. ;
						*IsSet->GetDataHK(-h,-k) = 1. ;
						*IsSet->GetDataHK(h,-k) = 1. ;
					}
					else {
						numTerms = 0 ;
						amplitude = w1 = *mFourierData->GetRealHK(h,k) ;
						if (w1 > 0.) numTerms++ ;
						amplitude += w1 = *IsSet->GetDataHK(-h,-k) == 1. ? 0. : *mFourierData->GetRealHK(-h,-k) ;
						*IsSet->GetDataHK(-h,-k) = 1. ;
						if (w1 > 0.) numTerms++ ;
						amplitude += w1 = *IsSet->GetDataHK(-h,k) == 1. ? 0. : *mFourierData->GetRealHK(-h,k) ;
						*IsSet->GetDataHK(-h,k) = 1. ;
						if (w1 > 0.) numTerms++ ;
						amplitude += w1 = *IsSet->GetDataHK(h,-k) == 1. ? 0. : *mFourierData->GetRealHK(h,-k) ;
						*IsSet->GetDataHK(h,-k) = 1. ;
						if (w1 > 0.) numTerms++ ;
						amplitudeSum += amplitude*amplitude ;
						if (numTerms > 0) amplitude /= numTerms ;
					}
					rmsAmplitude += (*mFourierData->GetRealHK(h,k) - amplitude)*(*mFourierData->GetRealHK(h,k) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(-h,k) - amplitude)*(*mFourierData->GetRealHK(-h,k) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(-h,-k) - amplitude)*(*mFourierData->GetRealHK(-h,-k) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(h,-k) - amplitude)*(*mFourierData->GetRealHK(h,-k) - amplitude) ;
				}
			}
			break ;
		case 6: // p11g
			// |F(h,k)| = |F(-h,k)|
			// Exchange h,k
			for(h=-hMax;h<=hMax;h++) {
				for(k=0;k<=kMax;k++) {
					if (k == 0 && h <= 0) continue ;
					if (*IsSet->GetDataHK(k,h) == 1.) continue ;
					*IsSet->GetDataHK(k,h) = 1. ;
					if (h == 0 && k%2 == 1) {
						// Systematic absence
						amplitude = 0. ;
						*IsSet->GetDataHK(k,h) = 1. ;
						*IsSet->GetDataHK(-k,-h) = 1. ;
						*IsSet->GetDataHK(-k,h) = 1. ;
						*IsSet->GetDataHK(k,-h) = 1. ;
					}
					else {
						numTerms = 0 ;
						amplitude = w1 = *mFourierData->GetRealHK(k,h) ;
						if (w1 > 0.) numTerms++ ;
						amplitude += w1 = *IsSet->GetDataHK(-k,-h) == 1. ? 0. : *mFourierData->GetRealHK(-k,-h) ;
						*IsSet->GetDataHK(-k,-h) = 1. ;
						if (w1 > 0.) numTerms++ ;
						amplitude += w1 = *IsSet->GetDataHK(-k,h) == 1. ? 0. : *mFourierData->GetRealHK(-k,h) ;
						*IsSet->GetDataHK(-k,h) = 1. ;
						if (w1 > 0.) numTerms++ ;
						amplitude += w1 = *IsSet->GetDataHK(k,-h) == 1. ? 0. : *mFourierData->GetRealHK(k,-h) ;
						*IsSet->GetDataHK(k,-h) = 1. ;
						if (w1 > 0.) numTerms++ ;
						amplitudeSum += amplitude*amplitude ;
						if (numTerms > 0) amplitude /= numTerms ;
					}
					rmsAmplitude += (*mFourierData->GetRealHK(k,h) - amplitude)*(*mFourierData->GetRealHK(k,h) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(-k,-h) - amplitude)*(*mFourierData->GetRealHK(-k,-h) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(-k,h) - amplitude)*(*mFourierData->GetRealHK(-k,h) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(k,-h) - amplitude)*(*mFourierData->GetRealHK(k,-h) - amplitude) ;
				}
			}
			break ;
		case 7: // cm  Systematic absence (hk): h+k = 2n+1
			// |F(h,k)| = |F(-h,k)|
			for(h=-hMax;h<=hMax;h++) {
				for(k=0;k<=kMax;k++) {
					if (k == 0 && h <= 0) continue ;
					if (*IsSet->GetDataHK(h,k) == 1.) continue ;
					*IsSet->GetDataHK(h,k) = 1. ;
					if ((h+k)%2 == 1) {
						// Systematic absence
						amplitude = 0. ;
						*IsSet->GetDataHK(h,k) = 1. ;
						*IsSet->GetDataHK(-h,k) = 1. ;
						*IsSet->GetDataHK(-h,-k) = 1. ;
						*IsSet->GetDataHK(h,-k) = 1. ;
					}
					else {
						numTerms = 0 ;
						amplitude = w1 = *mFourierData->GetRealHK(h,k) ;
						if (w1 > 0.) numTerms++ ;
						amplitude += w1 = *IsSet->GetDataHK(-h,-k) == 1. ? 0. : *mFourierData->GetRealHK(-h,-k) ;
						*IsSet->GetDataHK(-h,-k) = 1. ;
						if (w1 > 0.) numTerms++ ;
						amplitude += w1 = *IsSet->GetDataHK(-h,k) == 1. ? 0. : *mFourierData->GetRealHK(-h,k) ;
						*IsSet->GetDataHK(-h,k) = 1. ;
						if (w1 > 0.) numTerms++ ;
						amplitude += w1 = *IsSet->GetDataHK(h,-k) == 1. ? 0. : *mFourierData->GetRealHK(h,-k) ;
						*IsSet->GetDataHK(h,-k) = 1. ;
						if (w1 > 0.) numTerms++ ;
						amplitudeSum += amplitude*amplitude ;
						if (numTerms > 0) amplitude /= numTerms ;
					}
					rmsAmplitude += (*mFourierData->GetRealHK(h,k) - amplitude)*(*mFourierData->GetRealHK(h,k) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(-h,k) - amplitude)*(*mFourierData->GetRealHK(-h,k) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(-h,-k) - amplitude)*(*mFourierData->GetRealHK(-h,-k) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(h,-k) - amplitude)*(*mFourierData->GetRealHK(h,-k) - amplitude) ;
				}
			}
			break ;
		case 8: // c11m 
			// |F(h,k)| = |F(-h,k)|
			// Exchange h,k
			for(h=-hMax;h<=hMax;h++) {
				for(k=0;k<=kMax;k++) {
					if (k == 0 && h <= 0) continue ;
					if (*IsSet->GetDataHK(k,h) == 1.) continue ;
					*IsSet->GetDataHK(k,h) = 1. ;
					if ((h+k)%2 == 1) {
						// Systematic absence
						amplitude = 0. ;
						*IsSet->GetDataHK(k,h) = 1. ;
						*IsSet->GetDataHK(-k,-h) = 1. ;
						*IsSet->GetDataHK(-k,h) = 1. ;
						*IsSet->GetDataHK(k,-h) = 1. ;
					}
					else {
						numTerms = 0 ;
						amplitude = w1 = *mFourierData->GetRealHK(k,h) ;
						if (w1 > 0.) numTerms++ ;
						amplitude += w1 = *IsSet->GetDataHK(-k,-h) == 1. ? 0. : *mFourierData->GetRealHK(-k,-h) ;
						*IsSet->GetDataHK(-k,-h) = 1. ;
						if (w1 > 0.) numTerms++ ;
						amplitude += w1 = *IsSet->GetDataHK(-k,h) == 1. ? 0. : *mFourierData->GetRealHK(-k,h) ;
						*IsSet->GetDataHK(-k,h) = 1. ;
						if (w1 > 0.) numTerms++ ;
						amplitude += w1 = *IsSet->GetDataHK(k,-h) == 1. ? 0. : *mFourierData->GetRealHK(k,-h) ;
						*IsSet->GetDataHK(k,-h) = 1. ;
						if (w1 > 0.) numTerms++ ;
						amplitudeSum += amplitude*amplitude ;
						if (numTerms > 0) amplitude /= numTerms ;
					}
					rmsAmplitude += (*mFourierData->GetRealHK(k,h) - amplitude)*(*mFourierData->GetRealHK(k,h) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(-k,-h) - amplitude)*(*mFourierData->GetRealHK(-k,-h) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(-k,h) - amplitude)*(*mFourierData->GetRealHK(-k,h) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(k,-h) - amplitude)*(*mFourierData->GetRealHK(k,-h) - amplitude) ;
				}
			}
			break ;
		case 9: // pmm  -
			// |F(h,k)| = |F(-h,k)|
			for(h=-hMax;h<=hMax;h++) {
				for(k=0;k<=kMax;k++) {
					if (k == 0 && h <= 0) continue ;
					if (*IsSet->GetDataHK(h,k) == 1.) continue ;
					*IsSet->GetDataHK(h,k) = 1. ;
					numTerms = 0 ;
					amplitude = w1 = *mFourierData->GetRealHK(h,k) ;
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(-h,-k) == 1. ? 0. : *mFourierData->GetRealHK(-h,-k) ;
					*IsSet->GetDataHK(-h,-k) = 1. ;
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(-h,k) == 1. ? 0. : *mFourierData->GetRealHK(-h,k) ;
					*IsSet->GetDataHK(-h,k) = 1. ;
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(h,-k) == 1. ? 0. : *mFourierData->GetRealHK(h,-k) ;
					*IsSet->GetDataHK(h,-k) = 1. ;
					if (w1 > 0.) numTerms++ ;
					amplitudeSum += amplitude*amplitude ;
					if (numTerms > 0) amplitude /= numTerms ;
					rmsAmplitude += (*mFourierData->GetRealHK(h,k) - amplitude)*(*mFourierData->GetRealHK(h,k) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(-h,k) - amplitude)*(*mFourierData->GetRealHK(-h,k) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(-h,-k) - amplitude)*(*mFourierData->GetRealHK(-h,-k) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(h,-k) - amplitude)*(*mFourierData->GetRealHK(h,-k) - amplitude) ;
				}
			}
			break ;
		case 10: // pmg  Systematic absence (0k): k = 2n+1
			// |F(h,k)| = |F(-h,k)|
			for(h=-hMax;h<=hMax;h++) {
				for(k=0;k<=kMax;k++) {
					if (k == 0 && h <= 0) continue ;
					if (*IsSet->GetDataHK(h,k) == 1.) continue ;
					*IsSet->GetDataHK(h,k) = 1. ;
					if (h == 0 && k%2 == 1) {
						// Systematic absence
						amplitude = 0. ;
						*IsSet->GetDataHK(h,k) = 1. ;
						*IsSet->GetDataHK(-h,k) = 1. ;
						*IsSet->GetDataHK(-h,-k) = 1. ;
						*IsSet->GetDataHK(h,-k) = 1. ;
					}
					else {
						numTerms = 0 ;
						amplitude = w1 = *mFourierData->GetRealHK(h,k) ;
						if (w1 > 0.) numTerms++ ;
						amplitude += w1 = *IsSet->GetDataHK(-h,-k) == 1. ? 0. : *mFourierData->GetRealHK(-h,-k) ;
						*IsSet->GetDataHK(-h,-k) = 1. ;
						if (w1 > 0.) numTerms++ ;
						amplitude += w1 = *IsSet->GetDataHK(-h,k) == 1. ? 0. : *mFourierData->GetRealHK(-h,k) ;
						*IsSet->GetDataHK(-h,k) = 1. ;
						if (w1 > 0.) numTerms++ ;
						amplitude += w1 = *IsSet->GetDataHK(h,-k) == 1. ? 0. : *mFourierData->GetRealHK(h,-k) ;
						*IsSet->GetDataHK(h,-k) = 1. ;
						if (w1 > 0.) numTerms++ ;
						amplitudeSum += amplitude*amplitude ;
						if (numTerms > 0) amplitude /= numTerms ;
					}
					rmsAmplitude += (*mFourierData->GetRealHK(h,k) - amplitude)*(*mFourierData->GetRealHK(h,k) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(-h,k) - amplitude)*(*mFourierData->GetRealHK(-h,k) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(-h,-k) - amplitude)*(*mFourierData->GetRealHK(-h,-k) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(h,-k) - amplitude)*(*mFourierData->GetRealHK(h,-k) - amplitude) ;
				}
			}
			break ;
		case 11: // p2gm
			// |F(h,k)| = |F(-h,k)|
			// Exchange h,k
			for(h=-hMax;h<=hMax;h++) {
				for(k=0;k<=kMax;k++) {
					if (k == 0 && h <= 0) continue ;
					if (*IsSet->GetDataHK(k,h) == 1.) continue ;
					*IsSet->GetDataHK(k,h) = 1. ;
					if (h == 0 && k%2 == 1) {
						// Systematic absence
						amplitude = 0. ;
						*IsSet->GetDataHK(k,h) = 1. ;
						*IsSet->GetDataHK(-k,-h) = 1. ;
						*IsSet->GetDataHK(-k,h) = 1. ;
						*IsSet->GetDataHK(k,-h) = 1. ;
					}
					else {
						numTerms = 0 ;
						amplitude = w1 = *mFourierData->GetRealHK(k,h) ;
						if (w1 > 0.) numTerms++ ;
						amplitude += w1 = *IsSet->GetDataHK(-k,-h) == 1. ? 0. : *mFourierData->GetRealHK(-k,-h) ;
						*IsSet->GetDataHK(-k,-h) = 1. ;
						if (w1 > 0.) numTerms++ ;
						amplitude += w1 = *IsSet->GetDataHK(-k,h) == 1. ? 0. : *mFourierData->GetRealHK(-k,h) ;
						*IsSet->GetDataHK(-k,h) = 1. ;
						if (w1 > 0.) numTerms++ ;
						amplitude += w1 = *IsSet->GetDataHK(k,-h) == 1. ? 0. : *mFourierData->GetRealHK(k,-h) ;
						*IsSet->GetDataHK(k,-h) = 1. ;
						if (w1 > 0.) numTerms++ ;
						amplitudeSum += amplitude*amplitude ;
						if (numTerms > 0) amplitude /= numTerms ;
					}
					rmsAmplitude += (*mFourierData->GetRealHK(k,h) - amplitude)*(*mFourierData->GetRealHK(k,h) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(-k,-h) - amplitude)*(*mFourierData->GetRealHK(-k,-h) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(-k,h) - amplitude)*(*mFourierData->GetRealHK(-k,h) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(k,-h) - amplitude)*(*mFourierData->GetRealHK(k,-h) - amplitude) ;
				}
			}
			break ;
		case 12: // pgg  Systematic absence (h0): h = 2n+1 ; (0k): k = 2n+1
			// |F(h,k)| = |F(-h,k)|
			for(h=-hMax;h<=hMax;h++) {
				for(k=0;k<=kMax;k++) {
					if (k == 0 && h <= 0) continue ;
					if (*IsSet->GetDataHK(h,k) == 1.) continue ;
					*IsSet->GetDataHK(h,k) = 1. ;
					if ((h == 0 && k%2 == 1) || (k == 0 && h%2 == 1)) {
						// Systematic absence
						amplitude = 0. ;
						*IsSet->GetDataHK(h,k) = 1. ;
						*IsSet->GetDataHK(-h,k) = 1. ;
						*IsSet->GetDataHK(-h,-k) = 1. ;
						*IsSet->GetDataHK(h,-k) = 1. ;
					}
					else {
						numTerms = 0 ;
						amplitude = w1 = *mFourierData->GetRealHK(h,k) ;
						if (w1 > 0.) numTerms++ ;
						amplitude += w1 = *IsSet->GetDataHK(-h,-k) == 1. ? 0. : *mFourierData->GetRealHK(-h,-k) ;
						*IsSet->GetDataHK(-h,-k) = 1. ;
						if (w1 > 0.) numTerms++ ;
						amplitude += w1 = *IsSet->GetDataHK(-h,k) == 1. ? 0. : *mFourierData->GetRealHK(-h,k) ;
						*IsSet->GetDataHK(-h,k) = 1. ;
						if (w1 > 0.) numTerms++ ;
						amplitude += w1 = *IsSet->GetDataHK(h,-k) == 1. ? 0. : *mFourierData->GetRealHK(h,-k) ;
						*IsSet->GetDataHK(h,-k) = 1. ;
						if (w1 > 0.) numTerms++ ;
						amplitudeSum += amplitude*amplitude ;
						if (numTerms > 0) amplitude /= numTerms ;
					}
					rmsAmplitude += (*mFourierData->GetRealHK(h,k) - amplitude)*(*mFourierData->GetRealHK(h,k) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(-h,k) - amplitude)*(*mFourierData->GetRealHK(-h,k) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(-h,-k) - amplitude)*(*mFourierData->GetRealHK(-h,-k) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(h,-k) - amplitude)*(*mFourierData->GetRealHK(h,-k) - amplitude) ;
				}
			}
			break ;
		case 13: // cmm  Systematic absence (hk): h+k = 2n+1
			// |F(h,k)| = |F(-h,k)|
			for(h=-hMax;h<=hMax;h++) {
				for(k=0;k<=kMax;k++) {
					if (k == 0 && h <= 0) continue ;
					if (*IsSet->GetDataHK(h,k) == 1.) continue ;
					*IsSet->GetDataHK(h,k) = 1. ;
					if ((h+k)%2 == 1) {
						amplitude = 0. ;
						*IsSet->GetDataHK(h,k) = 1. ;
						*IsSet->GetDataHK(-h,k) = 1. ;
						*IsSet->GetDataHK(-h,-k) = 1. ;
						*IsSet->GetDataHK(h,-k) = 1. ;
					}
					else {
						numTerms = 0 ;
						amplitude = w1 = *mFourierData->GetRealHK(h,k) ;
						if (w1 > 0.) numTerms++ ;
						amplitude += w1 = *IsSet->GetDataHK(-h,-k) == 1. ? 0. : *mFourierData->GetRealHK(-h,-k) ;
						*IsSet->GetDataHK(-h,-k) = 1. ;
						if (w1 > 0.) numTerms++ ;
						amplitude += w1 = *IsSet->GetDataHK(-h,k) == 1. ? 0. : *mFourierData->GetRealHK(-h,k) ;
						*IsSet->GetDataHK(-h,k) = 1. ;
						if (w1 > 0.) numTerms++ ;
						amplitude += w1 = *IsSet->GetDataHK(h,-k) == 1. ? 0. : *mFourierData->GetRealHK(h,-k) ;
						*IsSet->GetDataHK(h,-k) = 1. ;
						if (w1 > 0.) numTerms++ ;
						amplitudeSum += amplitude*amplitude ;
						if (numTerms > 0) amplitude /= numTerms ;
					}
					rmsAmplitude += (*mFourierData->GetRealHK(h,k) - amplitude)*(*mFourierData->GetRealHK(h,k) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(-h,k) - amplitude)*(*mFourierData->GetRealHK(-h,k) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(-h,-k) - amplitude)*(*mFourierData->GetRealHK(-h,-k) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(h,-k) - amplitude)*(*mFourierData->GetRealHK(h,-k) - amplitude) ;
				}
			}
			break ;
		case 14: // p4
			// |F(h,k)| = |F(-k,h)|
			latticeParameter = (mA+mB)/2 ;
			mA = mB = latticeParameter ;
			latticeParameter = (mAStar+mBStar)/2 ;
			mAStar = mBStar = latticeParameter ;
			for(h=-hMax;h<=hMax;h++) {
				for(k=0;k<=kMax;k++) {
					if (k == 0 && h <= 0) continue ;
					if (*IsSet->GetDataHK(h,k) == 1.) continue ;
					*IsSet->GetDataHK(h,k) = 1. ;
					numTerms = 0 ;
					amplitude = w1 = *mFourierData->GetRealHK(h,k) ;
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(-h,-k) == 1. ? 0. : *mFourierData->GetRealHK(-h,-k) ;
					*IsSet->GetDataHK(-h,-k) = 1. ;
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(-k,h) == 1. ? 0. : *mFourierData->GetRealHK(-k,h) ;
					*IsSet->GetDataHK(-k,h) = 1. ;
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(k,-h) == 1. ? 0. : *mFourierData->GetRealHK(k,-h) ;
					*IsSet->GetDataHK(k,-h) = 1. ;
					if (w1 > 0.) numTerms++ ;
					amplitudeSum += amplitude*amplitude ;
					if (numTerms > 0) amplitude /= numTerms ;
					rmsAmplitude += (*mFourierData->GetRealHK(h,k) - amplitude)*(*mFourierData->GetRealHK(h,k) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(-k,h) - amplitude)*(*mFourierData->GetRealHK(-k,h) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(-h,-k) - amplitude)*(*mFourierData->GetRealHK(-h,-k) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(k,-h) - amplitude)*(*mFourierData->GetRealHK(k,-h) - amplitude) ;
				}
			}
			break ;
		case 15: // p4mm
		case 16: // p4g
			// |F(h,k)| = |F(-k,h)| = |F(-h,k)|
			latticeParameter = (mA+mB)/2 ;
			mA = mB = latticeParameter ;
			latticeParameter = (mAStar+mBStar)/2 ;
			mAStar = mBStar = latticeParameter ;
			for(h=-hMax;h<=hMax;h++) {
				for(k=0;k<=kMax;k++) {
					if (k == 0 && h <= 0) continue ;
					if (*IsSet->GetDataHK(h,k) == 1.) continue ;
					*IsSet->GetDataHK(h,k) = 1. ;
					numTerms = 0 ;
					amplitude = w1 = *mFourierData->GetRealHK(h,k) ;
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(-h,-k) == 1. ? 0. : *mFourierData->GetRealHK(-h,-k) ;
					*IsSet->GetDataHK(-h,-k) = 1. ;
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(-k,h) == 1. ? 0. : *mFourierData->GetRealHK(-k,h) ;
					*IsSet->GetDataHK(-k,h) = 1. ;
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(k,-h) == 1. ? 0. : *mFourierData->GetRealHK(k,-h) ;
					*IsSet->GetDataHK(k,-h) = 1. ;
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(-h,k) == 1. ? 0. : *mFourierData->GetRealHK(-h,k) ;
					*IsSet->GetDataHK(-h,k) = 1. ;
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(h,-k) == 1. ? 0. : *mFourierData->GetRealHK(h,-k) ;
					*IsSet->GetDataHK(h,-k) = 1. ;
					if (w1 > 0.) numTerms++ ;
					amplitudeSum += amplitude*amplitude ;
					if (numTerms > 0) amplitude /= numTerms ;
					rmsAmplitude += (*mFourierData->GetRealHK(h,k) - amplitude)*(*mFourierData->GetRealHK(h,k) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(-k,h) - amplitude)*(*mFourierData->GetRealHK(-k,h) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(-h,k) - amplitude)*(*mFourierData->GetRealHK(-h,k) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(-h,-k) - amplitude)*(*mFourierData->GetRealHK(-h,-k) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(k,-h) - amplitude)*(*mFourierData->GetRealHK(k,-h) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(h,-k) - amplitude)*(*mFourierData->GetRealHK(h,-k) - amplitude) ;
				}
			}
			break ;
		case 17: // p4mg
			// |F(h,k)| = |F(-k,h)| = |F(-h,k)|
			// Exchange h,k
			latticeParameter = (mA+mB)/2 ;
			mA = mB = latticeParameter ;
			latticeParameter = (mAStar+mBStar)/2 ;
			mAStar = mBStar = latticeParameter ;
			for(h=-hMax;h<=hMax;h++) {
				for(k=0;k<=kMax;k++) {
					if (k == 0 && h <= 0) continue ;
					if (*IsSet->GetDataHK(k,h) == 1.) continue ;
					*IsSet->GetDataHK(k,h) = 1. ;
					numTerms = 0 ;
					amplitude = w1 = *mFourierData->GetRealHK(k,h) ;
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(-k,-h) == 1. ? 0. : *mFourierData->GetRealHK(-k,-h) ;
					*IsSet->GetDataHK(-k,-h) = 1. ;
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(-h,k) == 1. ? 0. : *mFourierData->GetRealHK(-h,k) ;
					*IsSet->GetDataHK(-h,k) = 1. ;
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(h,-k) == 1. ? 0. : *mFourierData->GetRealHK(h,-k) ;
					*IsSet->GetDataHK(h,-k) = 1. ;
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(-k,h) == 1. ? 0. : *mFourierData->GetRealHK(-k,h) ;
					*IsSet->GetDataHK(-k,h) = 1. ;
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(k,-h) == 1. ? 0. : *mFourierData->GetRealHK(k,-h) ;
					*IsSet->GetDataHK(k,-h) = 1. ;
					if (w1 > 0.) numTerms++ ;
					amplitudeSum += amplitude*amplitude ;
					if (numTerms > 0) amplitude /= numTerms ;
					rmsAmplitude += (*mFourierData->GetRealHK(k,h) - amplitude)*(*mFourierData->GetRealHK(k,h) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(-k,-h) - amplitude)*(*mFourierData->GetRealHK(-k,-h) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(-h,k) - amplitude)*(*mFourierData->GetRealHK(-h,k) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(h,-k) - amplitude)*(*mFourierData->GetRealHK(h,-k) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(-k,h) - amplitude)*(*mFourierData->GetRealHK(-k,h) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(k,-h) - amplitude)*(*mFourierData->GetRealHK(k,-h) - amplitude) ;
				}
			}
			break ;
		case 18: // p3
			// |F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)|
			latticeParameter = (mA+mB)/2 ;
			mA = mB = latticeParameter ;
			latticeParameter = (mAStar+mBStar)/2 ;
			mAStar = mBStar = latticeParameter ;
			for(h=-hMax;h<=hMax;h++) {
				for(k=0;k<=kMax;k++) {
					if (k == 0 && h <= 0) continue ;
					if (*IsSet->GetDataHK(h,k) == 1.) continue ;
					*IsSet->GetDataHK(h,k) = 1. ;
					numTerms = 0 ;

					amplitude = w1 = *mFourierData->GetRealHK(h,k) ;
					if (w1 > 0.) numTerms++ ;
					if (useFriedelSymmetry) {
						amplitude += w1 = *IsSet->GetDataHK(-h,-k) == 1. ? 0. : *mFourierData->GetRealHK(-h,-k) ;
						*IsSet->GetDataHK(-h,-k) = 1. ;
						if (w1 > 0.) numTerms++ ;
					}
					amplitude += w1 = *IsSet->GetDataHK(k,-h-k) == 1. ? 0. : *mFourierData->GetRealHK(k,-h-k) ;
					*IsSet->GetDataHK(k,-h-k) = 1. ;
					if (w1 > 0.) numTerms++ ;
					if (useFriedelSymmetry) {
						amplitude += w1 = *IsSet->GetDataHK(-k,h+k) == 1. ? 0. : *mFourierData->GetRealHK(-k,h+k) ;
						*IsSet->GetDataHK(-k,h+k) = 1. ;
						if (w1 > 0.) numTerms++ ;
					}
					amplitude += w1 = *IsSet->GetDataHK(-h-k,h) == 1. ? 0. : *mFourierData->GetRealHK(-h-k,h) ;
					*IsSet->GetDataHK(-h-k,h) = 1. ;
					if (w1 > 0.) numTerms++ ;
					if (useFriedelSymmetry) {
						amplitude += w1 = *IsSet->GetDataHK(h+k,-h) == 1. ? 0. : *mFourierData->GetRealHK(h+k,-h) ;
						*IsSet->GetDataHK(h+k,-h) = 1. ;
						if (w1 > 0.) numTerms++ ;
					}
					amplitudeSum += amplitude*amplitude ;
					if (numTerms > 0) amplitude /= numTerms ;
					rmsAmplitude += (*mFourierData->GetRealHK(h,k) - amplitude)*(*mFourierData->GetRealHK(h,k) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(k,-h-k) - amplitude)*(*mFourierData->GetRealHK(k,-h-k) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(-h-k,h) - amplitude)*(*mFourierData->GetRealHK(-h-k,h) - amplitude) ;
					if (useFriedelSymmetry) rmsAmplitude += (*mFourierData->GetRealHK(-h,-k) - amplitude)*(*mFourierData->GetRealHK(-h,-k) - amplitude) ;
					if (useFriedelSymmetry) rmsAmplitude += (*mFourierData->GetRealHK(-k,h+k) - amplitude)*(*mFourierData->GetRealHK(-k,h+k) - amplitude) ;
					if (useFriedelSymmetry) rmsAmplitude += (*mFourierData->GetRealHK(h+k,-h) - amplitude)*(*mFourierData->GetRealHK(h+k,-h) - amplitude) ;
				}
			}
			break ;
		case 19: // p3m1
		case 20: // p31m
			// |F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)| = |F(k,h)|
			latticeParameter = (mA+mB)/2 ;
			mA = mB = latticeParameter ;
			latticeParameter = (mAStar+mBStar)/2 ;
			mAStar = mBStar = latticeParameter ;
			for(h=-hMax;h<=hMax;h++) {
				for(k=0;k<=kMax;k++) {

					if (k == 0 && h <= 0) continue ;
					if (*IsSet->GetDataHK(h,k) == 1.) continue ;
					*IsSet->GetDataHK(h,k) = 1. ;
					numTerms = 0 ;

					amplitude = w1 = *mFourierData->GetRealHK(h,k) ;
					if (w1 > 0.) numTerms++ ;
					if (useFriedelSymmetry) {
						amplitude += w1 = *IsSet->GetDataHK(-h,-k) == 1. ? 0. : *mFourierData->GetRealHK(-h,-k) ;
						*IsSet->GetDataHK(-h,-k) = 1. ;
						if (w1 > 0.) numTerms++ ;
					}
					amplitude += w1 = *IsSet->GetDataHK(k,-h-k) == 1. ? 0. : *mFourierData->GetRealHK(k,-h-k) ;
					*IsSet->GetDataHK(k,-h-k) = 1. ;
					if (w1 > 0.) numTerms++ ;
					if (useFriedelSymmetry) {
						amplitude += w1 = *IsSet->GetDataHK(-k,h+k) == 1. ? 0. : *mFourierData->GetRealHK(-k,h+k) ;
						*IsSet->GetDataHK(-k,h+k) = 1. ;
						if (w1 > 0.) numTerms++ ;
					}
					amplitude += w1 = *IsSet->GetDataHK(-h-k,h) == 1. ? 0. : *mFourierData->GetRealHK(-h-k,h) ;
					*IsSet->GetDataHK(-h-k,h) = 1. ;
					if (w1 > 0.) numTerms++ ;
					if (useFriedelSymmetry) {
						amplitude += w1 = *IsSet->GetDataHK(h+k,-h) == 1. ? 0. : *mFourierData->GetRealHK(h+k,-h) ;
						*IsSet->GetDataHK(h+k,-h) = 1. ;
						if (w1 > 0.) numTerms++ ;
					}
					amplitude += w1 = *IsSet->GetDataHK(k,h) == 1. ? 0. : *mFourierData->GetRealHK(k,h) ;
					*IsSet->GetDataHK(k,h) = 1. ;
					if (w1 > 0.) numTerms++ ;
					if (useFriedelSymmetry) {
						amplitude += w1 = *IsSet->GetDataHK(-k,-h) == 1. ? 0. : *mFourierData->GetRealHK(-k,-h) ;
						*IsSet->GetDataHK(h+k,-h) = 1. ;
						if (w1 > 0.) numTerms++ ;
					}

					amplitudeSum += amplitude*amplitude ;
					if (numTerms > 0) amplitude /= numTerms ;

					rmsAmplitude += (*mFourierData->GetRealHK(h,k) - amplitude)*(*mFourierData->GetRealHK(h,k) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(k,-h-k) - amplitude)*(*mFourierData->GetRealHK(k,-h-k) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(-h-k,h) - amplitude)*(*mFourierData->GetRealHK(-h-k,h) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(k,h) - amplitude)*(*mFourierData->GetRealHK(k,h) - amplitude) ;
					if (useFriedelSymmetry) rmsAmplitude += (*mFourierData->GetRealHK(-h,-k) - amplitude)*(*mFourierData->GetRealHK(-h,-k) - amplitude) ;
					if (useFriedelSymmetry) rmsAmplitude += (*mFourierData->GetRealHK(-k,h+k) - amplitude)*(*mFourierData->GetRealHK(-k,h+k) - amplitude) ;
					if (useFriedelSymmetry) rmsAmplitude += (*mFourierData->GetRealHK(h+k,-h) - amplitude)*(*mFourierData->GetRealHK(h+k,-h) - amplitude) ;
					if (useFriedelSymmetry) rmsAmplitude += (*mFourierData->GetRealHK(-k,-h) - amplitude)*(*mFourierData->GetRealHK(-k,-h) - amplitude) ;
				}
			}
			break ;
		case 21: // p6
			// |F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)|
			latticeParameter = (mA+mB)/2 ;
			mA = mB = latticeParameter ;
			latticeParameter = (mAStar+mBStar)/2 ;
			mAStar = mBStar = latticeParameter ;
			for(h=-hMax;h<=hMax;h++) {
				for(k=0;k<=kMax;k++) {
					if (k == 0 && h <= 0) continue ;
					if (*IsSet->GetDataHK(h,k) == 1.) continue ;
					*IsSet->GetDataHK(h,k) = 1. ;
					numTerms = 0 ;

					amplitude = w1 = *mFourierData->GetRealHK(h,k) ;
					if (w1 > 0.) numTerms++ ;
					
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(-h,-k) == 1. ? 0. : *mFourierData->GetRealHK(-h,-k) ;
					*IsSet->GetDataHK(-h,-k) = 1. ;
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(k,-h-k) == 1. ? 0. : *mFourierData->GetRealHK(k,-h-k) ;
					*IsSet->GetDataHK(k,-h-k) = 1. ;
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(-k,h+k) == 1. ? 0. : *mFourierData->GetRealHK(-k,h+k) ;
					*IsSet->GetDataHK(-k,h+k) = 1. ;
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(-h-k,h) == 1. ? 0. : *mFourierData->GetRealHK(-h-k,h) ;
					*IsSet->GetDataHK(-h-k,h) = 1. ;
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(h+k,-h) == 1. ? 0. : *mFourierData->GetRealHK(h+k,-h) ;
					*IsSet->GetDataHK(h+k,-h) = 1. ;
					if (w1 > 0.) numTerms++ ;
					amplitudeSum += amplitude*amplitude ;
					if (numTerms > 0) amplitude /= numTerms ;
					rmsAmplitude += (*mFourierData->GetRealHK(h,k) - amplitude)*(*mFourierData->GetRealHK(h,k) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(k,-h-k) - amplitude)*(*mFourierData->GetRealHK(k,-h-k) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(-h-k,h) - amplitude)*(*mFourierData->GetRealHK(-h-k,h) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(-h,-k) - amplitude)*(*mFourierData->GetRealHK(-h,-k) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(-k,h+k) - amplitude)*(*mFourierData->GetRealHK(-k,h+k) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(h+k,-h) - amplitude)*(*mFourierData->GetRealHK(h+k,-h) - amplitude) ;
				}
			}
			break ;
		case 22: // p6mm
			// |F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)| = |F(k,h)|
			latticeParameter = (mA+mB)/2 ;
			mA = mB = latticeParameter ;
			latticeParameter = (mAStar+mBStar)/2 ;
			mAStar = mBStar = latticeParameter ;
			for(h=-hMax;h<=hMax;h++) {
				for(k=0;k<=kMax;k++) {

					if (k == 0 && h <= 0) continue ;
					if (*IsSet->GetDataHK(h,k) == 1.) continue ;
					*IsSet->GetDataHK(h,k) = 1. ;
					numTerms = 0 ;

					amplitude = w1 = *mFourierData->GetRealHK(h,k) ;
					if (w1 > 0.) numTerms++ ;
					
					amplitude += w1 = *IsSet->GetDataHK(-h,-k) == 1. ? 0. : *mFourierData->GetRealHK(-h,-k) ;
					*IsSet->GetDataHK(-h,-k) = 1. ;
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(k,-h-k) == 1. ? 0. : *mFourierData->GetRealHK(k,-h-k) ;
					*IsSet->GetDataHK(k,-h-k) = 1. ;
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(-k,h+k) == 1. ? 0. : *mFourierData->GetRealHK(-k,h+k) ;
					*IsSet->GetDataHK(-k,h+k) = 1. ;
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(-h-k,h) == 1. ? 0. : *mFourierData->GetRealHK(-h-k,h) ;
					*IsSet->GetDataHK(-h-k,h) = 1. ;
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(h+k,-h) == 1. ? 0. : *mFourierData->GetRealHK(h+k,-h) ;
					*IsSet->GetDataHK(h+k,-h) = 1. ;
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(k,h) == 1. ? 0. : *mFourierData->GetRealHK(k,h) ;
					*IsSet->GetDataHK(k,h) = 1. ;
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(-k,-h) == 1. ? 0. : *mFourierData->GetRealHK(-k,-h) ;
					*IsSet->GetDataHK(-k,-h) = 1. ;
					if (w1 > 0.) numTerms++ ;
					amplitudeSum += amplitude*amplitude ;
					if (numTerms > 0) amplitude /= numTerms ;
					rmsAmplitude += (*mFourierData->GetRealHK(h,k) - amplitude)*(*mFourierData->GetRealHK(h,k) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(k,-h-k) - amplitude)*(*mFourierData->GetRealHK(k,-h-k) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(-h-k,h) - amplitude)*(*mFourierData->GetRealHK(-h-k,h) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(-h,-k) - amplitude)*(*mFourierData->GetRealHK(-h,-k) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(-k,h+k) - amplitude)*(*mFourierData->GetRealHK(-k,h+k) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(h+k,-h) - amplitude)*(*mFourierData->GetRealHK(h+k,-h) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(k,h) - amplitude)*(*mFourierData->GetRealHK(k,h) - amplitude) ;
					rmsAmplitude += (*mFourierData->GetRealHK(-k,-h) - amplitude)*(*mFourierData->GetRealHK(-k,-h) - amplitude) ;
					
				}
			}
			break ;
	}
	amplitudeResidual = sqrt(rmsAmplitude/amplitudeSum) ;
	delete IsSet ;
	return amplitudeResidual ;
}

void	TDiffractionPeaks::Symmetrize(short numPlaneGroup) {
	// Symmetrize the amplitudes. We have no phases, so we don't need to worry about these
	// The amplitudes when symmetrized is set to the mean of the measured values
	float	w1,amplitude,latticeParameter ;
	short	h,k,hMax,kMax,Width,Height,numTerms ;

	Width = mFourierData->GetWidth() ;
	Height = mFourierData->GetHeight() ;
	
	//

	TArray	*IsSet = new TArray(Width,Height) ;

	hMax = mMaxH ;
	kMax = mMaxK ;

	*IsSet = 0. ;
	bool	useFriedelSymmetry = UseFriedelSymmetry() ;
	bool	useSystematicAbsence = true ;
	

	switch(numPlaneGroup) {
		case 1:	// p1
		case 2: // p2
			break ;
		case 3: // pm 
		case 4: // p11m 
		case 5: // pg 
		case 6: // p11g
		case 7: // cm  
		case 8: // c11m 
		case 9: // p2mm  
		case 10: // p2mg  
		case 11: // p2gm
		case 12: // pgg
		case 13: // cmm
			mGammaStar = 90.0 ;
			if (!useSystematicAbsence) numPlaneGroup = 3 ;
			break ;
		case 14: // p4
		case 15: // p4mm
		case 16: // p4g
		case 17: // p4mg
			mGammaStar = 90.0 ;
			break ;
		case 18: // p3
		case 19: // p3m1
		case 20: // p31m
		case 21: // p6
		case 22: // p6mm
			mGammaStar = 60.0 ;
			break ;
	}

	
	// Amplitude Relation  |F(h,k)| = |F(-h,-k)| + 
	switch(numPlaneGroup) {
		case 1:	// p1
			// If there is no symmetry do nothing to the data 
			if (useFriedelSymmetry) {
				for(h=-hMax;h<=hMax;h++) {
					for(k=0;k<=kMax;k++) {
						if (k == 0 && h <= 0) continue ;
						if (*IsSet->GetDataHK(h,k) == 1.) continue ;
						*IsSet->GetDataHK(h,k) = 1. ;
						numTerms = 0 ;
						amplitude = w1 = *mFourierData->GetRealHK(h,k) ;
						if (w1 > 0.) numTerms++ ;
						amplitude += w1 = *IsSet->GetDataHK(-h,-k) == 1. ? 0. : *mFourierData->GetRealHK(-h,-k) ;
						*IsSet->GetDataHK(-h,-k) = 1. ;
						if (w1 > 0.) numTerms++ ;
						if (numTerms > 0) amplitude /= numTerms ;
						*mFourierData->GetRealHK(h,k) = amplitude ;
						*mFourierData->GetRealHK(-h,-k) = amplitude ;
					}
				}
			}
			break ;
		case 2: // p2
			for(h=-hMax;h<=hMax;h++) {
				for(k=0;k<=kMax;k++) {
					if (k == 0 && h <= 0) continue ;
					if (*IsSet->GetDataHK(h,k) == 1.) continue ;
					*IsSet->GetDataHK(h,k) = 1. ;
					numTerms = 0 ;
					amplitude = w1 = *mFourierData->GetRealHK(h,k) ;
					if (w1 > 0.) numTerms++ ;
					w1 = *IsSet->GetDataHK(-h,-k) == 1. ? 0. : *mFourierData->GetRealHK(-h,-k) ;
					amplitude += w1 = *IsSet->GetDataHK(-h,-k) == 1. ? 0. : *mFourierData->GetRealHK(-h,-k) ;
					*IsSet->GetDataHK(-h,-k) = 1. ;
					if (w1 > 0.) numTerms++ ;
					if (numTerms > 0) amplitude /= numTerms ;
					*mFourierData->GetRealHK(h,k) = amplitude ;
					*mFourierData->GetRealHK(-h,-k) = amplitude ;
				}
			}
			break ;
		case 3: // pm - 
		case 4: // p11m 
			// |F(h,k)| = |F(-h,k)|
			for(h=-hMax;h<=hMax;h++) {
				for(k=0;k<=kMax;k++) {
					if (k == 0 && h <= 0) continue ;
					if (*IsSet->GetDataHK(h,k) == 1.) continue ;
					*IsSet->GetDataHK(h,k) = 1. ;
					numTerms = 0 ;
					amplitude = w1 = *mFourierData->GetRealHK(h,k) ;
					if (w1 > 0.) numTerms++ ;
					
					amplitude += w1 = *IsSet->GetDataHK(-h,-k) == 1. ? 0. : *mFourierData->GetRealHK(-h,-k) ;
					*IsSet->GetDataHK(-h,-k) = 1. ;
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(-h,k) == 1. ? 0. : *mFourierData->GetRealHK(-h,k) ;
					*IsSet->GetDataHK(-h,k) = 1. ;
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(h,-k) == 1. ? 0. : *mFourierData->GetRealHK(h,-k) ;
					*IsSet->GetDataHK(h,-k) = 1. ;
					if (w1 > 0.) numTerms++ ;
					if (numTerms > 0) amplitude /= numTerms ;
					*mFourierData->GetRealHK(h,k) = amplitude ;
					*mFourierData->GetRealHK(-h,k) = amplitude ;
					*mFourierData->GetRealHK(-h,-k) = amplitude ;
					*mFourierData->GetRealHK(h,-k) = amplitude ;
				}
			}
			break ;
		case 5: // pg  Systematic absence (0k): k = 2n+1
			// |F(h,k)| = |F(-h,k)|
			for(h=-hMax;h<=hMax;h++) {
				for(k=0;k<=kMax;k++) {
					if (k == 0 && h <= 0) continue ;
					if (*IsSet->GetDataHK(h,k) == 1.) continue ;
					*IsSet->GetDataHK(h,k) = 1. ;
					if (h == 0 && k%2 == 1) {
						// Systematic absence
						amplitude = 0. ;
						*IsSet->GetDataHK(h,k) = 1. ;
						*IsSet->GetDataHK(-h,k) = 1. ;
						*IsSet->GetDataHK(-h,-k) = 1. ;
						*IsSet->GetDataHK(h,-k) = 1. ;
					}
					else {
						numTerms = 0 ;
						amplitude = w1 = *mFourierData->GetRealHK(h,k) ;
						if (w1 > 0.) numTerms++ ;
						amplitude += w1 = *IsSet->GetDataHK(-h,-k) == 1. ? 0. : *mFourierData->GetRealHK(-h,-k) ;
						*IsSet->GetDataHK(-h,-k) = 1. ;
						if (w1 > 0.) numTerms++ ;
						amplitude += w1 = *IsSet->GetDataHK(-h,k) == 1. ? 0. : *mFourierData->GetRealHK(-h,k) ;
						*IsSet->GetDataHK(-h,k) = 1. ;
						if (w1 > 0.) numTerms++ ;
						amplitude += w1 = *IsSet->GetDataHK(h,-k) == 1. ? 0. : *mFourierData->GetRealHK(h,-k) ;
						*IsSet->GetDataHK(h,-k) = 1. ;
						if (w1 > 0.) numTerms++ ;
						if (numTerms > 0) amplitude /= numTerms ;
					}
					*mFourierData->GetRealHK(h,k) = amplitude ;
					*mFourierData->GetRealHK(-h,k) = amplitude ;
					*mFourierData->GetRealHK(-h,-k) = amplitude ;
					*mFourierData->GetRealHK(h,-k) = amplitude ;
				}
			}
			break ;
		case 6: // p11g
			// Exchange h,k
			// |F(h,k)| = |F(-h,k)|
			for(h=-hMax;h<=hMax;h++) {
				for(k=0;k<=kMax;k++) {
					if (k == 0 && h <= 0) continue ;
					if (*IsSet->GetDataHK(k,h) == 1.) continue ;
					*IsSet->GetDataHK(k,h) = 1. ;
					if (h == 0 && k%2 == 1) {
						// Systematic absence
						amplitude = 0. ;
						*IsSet->GetDataHK(k,h) = 1. ;
						*IsSet->GetDataHK(-k,h) = 1. ;
						*IsSet->GetDataHK(-k,-h) = 1. ;
						*IsSet->GetDataHK(k,-h) = 1. ;
					}
					else {
						numTerms = 0 ;
						amplitude = w1 = *mFourierData->GetRealHK(k,h) ;
						if (w1 > 0.) numTerms++ ;
						amplitude += w1 = *IsSet->GetDataHK(-k,-h) == 1. ? 0. : *mFourierData->GetRealHK(-k,-h) ;
						*IsSet->GetDataHK(-k,-h) = 1. ;
						if (w1 > 0.) numTerms++ ;
						amplitude += w1 = *IsSet->GetDataHK(-k,h) == 1. ? 0. : *mFourierData->GetRealHK(-k,h) ;
						*IsSet->GetDataHK(-k,h) = 1. ;
						if (w1 > 0.) numTerms++ ;
						amplitude += w1 = *IsSet->GetDataHK(k,-h) == 1. ? 0. : *mFourierData->GetRealHK(k,-h) ;
						*IsSet->GetDataHK(k,-h) = 1. ;
						if (w1 > 0.) numTerms++ ;
						if (numTerms > 0) amplitude /= numTerms ;
					}
					*mFourierData->GetRealHK(k,h) = amplitude ;
					*mFourierData->GetRealHK(-k,h) = amplitude ;
					*mFourierData->GetRealHK(-k,-h) = amplitude ;
					*mFourierData->GetRealHK(k,-h) = amplitude ;
				}
			}
			break ;
		case 7: // cm  Systematic absence (hk): h+k = 2n+1
			// |F(h,k)| = |F(-h,k)|
			for(h=-hMax;h<=hMax;h++) {
				for(k=0;k<=kMax;k++) {
					if (k == 0 && h <= 0) continue ;
					if (*IsSet->GetDataHK(h,k) == 1.) continue ;
					*IsSet->GetDataHK(h,k) = 1. ;
					if ((h+k)%2 == 1) {
						// Systematic absence
						amplitude = 0. ;
						*IsSet->GetDataHK(h,k) = 1. ;
						*IsSet->GetDataHK(-h,k) = 1. ;
						*IsSet->GetDataHK(-h,-k) = 1. ;
						*IsSet->GetDataHK(h,-k) = 1. ;
					}
					else {
						numTerms = 0 ;
						amplitude = w1 = *mFourierData->GetRealHK(h,k) ;
						if (w1 > 0.) numTerms++ ;
						amplitude += w1 = *IsSet->GetDataHK(-h,-k) == 1. ? 0. : *mFourierData->GetRealHK(-h,-k) ;
						*IsSet->GetDataHK(-h,-k) = 1. ;
						if (w1 > 0.) numTerms++ ;
						amplitude += w1 = *IsSet->GetDataHK(-h,k) == 1. ? 0. : *mFourierData->GetRealHK(-h,k) ;
						*IsSet->GetDataHK(-h,k) = 1. ;
						if (w1 > 0.) numTerms++ ;
						amplitude += w1 = *IsSet->GetDataHK(h,-k) == 1. ? 0. : *mFourierData->GetRealHK(h,-k) ;
						*IsSet->GetDataHK(h,-k) = 1. ;
						if (w1 > 0.) numTerms++ ;
						if (numTerms > 0) amplitude /= numTerms ;
					}
					*mFourierData->GetRealHK(h,k) = amplitude ;
					*mFourierData->GetRealHK(-h,k) = amplitude ;
					*mFourierData->GetRealHK(-h,-k) = amplitude ;
					*mFourierData->GetRealHK(h,-k) = amplitude ;
				}
			}
			break ;
		case 8: // c11m 
			// Exchange h,k
			// |F(h,k)| = |F(-h,k)|
			for(h=-hMax;h<=hMax;h++) {
				for(k=0;k<=kMax;k++) {
					if (k == 0 && h <= 0) continue ;
					if (*IsSet->GetDataHK(k,h) == 1.) continue ;
					*IsSet->GetDataHK(k,h) = 1. ;
					if ((h+k)%2 == 1) {
						// Systematic absence
						amplitude = 0. ;
						*IsSet->GetDataHK(k,h) = 1. ;
						*IsSet->GetDataHK(-k,h) = 1. ;
						*IsSet->GetDataHK(-k,-h) = 1. ;
						*IsSet->GetDataHK(k,-h) = 1. ;
					}
					else {
						numTerms = 0 ;
						amplitude = w1 = *mFourierData->GetRealHK(k,h) ;
						if (w1 > 0.) numTerms++ ;
						amplitude += w1 = *IsSet->GetDataHK(-k,-h) == 1. ? 0. : *mFourierData->GetRealHK(-k,-h) ;
						*IsSet->GetDataHK(-k,-h) = 1. ;
						if (w1 > 0.) numTerms++ ;
						amplitude += w1 = *IsSet->GetDataHK(-k,h) == 1. ? 0. : *mFourierData->GetRealHK(-k,h) ;
						*IsSet->GetDataHK(-k,h) = 1. ;
						if (w1 > 0.) numTerms++ ;
						amplitude += w1 = *IsSet->GetDataHK(k,-h) == 1. ? 0. : *mFourierData->GetRealHK(k,-h) ;
						*IsSet->GetDataHK(k,-h) = 1. ;
						if (w1 > 0.) numTerms++ ;
						if (numTerms > 0) amplitude /= numTerms ;
					}
					*mFourierData->GetRealHK(k,h) = amplitude ;
					*mFourierData->GetRealHK(-k,h) = amplitude ;
					*mFourierData->GetRealHK(-k,-h) = amplitude ;
					*mFourierData->GetRealHK(k,-h) = amplitude ;
				}
			}
			break ;
		case 9: // pmm  -
			// |F(h,k)| = |F(-h,k)|
			for(h=-hMax;h<=hMax;h++) {
				for(k=0;k<=kMax;k++) {
					if (k == 0 && h <= 0) continue ;
					if (*IsSet->GetDataHK(h,k) == 1.) continue ;
					*IsSet->GetDataHK(h,k) = 1. ;
					numTerms = 0 ;
					amplitude = w1 = *mFourierData->GetRealHK(h,k) ;
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(-h,-k) == 1. ? 0. : *mFourierData->GetRealHK(-h,-k) ;
					*IsSet->GetDataHK(-h,-k) = 1. ;
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(-h,k) == 1. ? 0. : *mFourierData->GetRealHK(-h,k) ;
					*IsSet->GetDataHK(-h,k) = 1. ;
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(h,-k) == 1. ? 0. : *mFourierData->GetRealHK(h,-k) ;
					*IsSet->GetDataHK(h,-k) = 1. ;
					if (w1 > 0.) numTerms++ ;
					if (numTerms > 0) amplitude /= numTerms ;
					*mFourierData->GetRealHK(h,k) = amplitude ;
					*mFourierData->GetRealHK(-h,k) = amplitude ;
					*mFourierData->GetRealHK(-h,-k) = amplitude ;
					*mFourierData->GetRealHK(h,-k) = amplitude ;
				}
			}
			break ;
		case 10: // pmg  Systematic absence (0k): k = 2n+1
			// |F(h,k)| = |F(-h,k)|
			for(h=-hMax;h<=hMax;h++) {
				for(k=0;k<=kMax;k++) {
					if (k == 0 && h <= 0) continue ;
					if (*IsSet->GetDataHK(h,k) == 1.) continue ;
					*IsSet->GetDataHK(h,k) = 1. ;
					if (h == 0 && k%2 == 1) {
						// Systematic absence
						amplitude = 0. ;
						*IsSet->GetDataHK(h,k) = 1. ;
						*IsSet->GetDataHK(-h,k) = 1. ;
						*IsSet->GetDataHK(-h,-k) = 1. ;
						*IsSet->GetDataHK(h,-k) = 1. ;
					}
					else {
						numTerms = 0 ;
						amplitude = w1 = *mFourierData->GetRealHK(h,k) ;
						if (w1 > 0.) numTerms++ ;
						amplitude += w1 = *IsSet->GetDataHK(-h,-k) == 1. ? 0. : *mFourierData->GetRealHK(-h,-k) ;
						*IsSet->GetDataHK(-h,-k) = 1. ;
						if (w1 > 0.) numTerms++ ;
						amplitude += w1 = *IsSet->GetDataHK(-h,k) == 1. ? 0. : *mFourierData->GetRealHK(-h,k) ;
						*IsSet->GetDataHK(-h,k) = 1. ;
						if (w1 > 0.) numTerms++ ;
						amplitude += w1 = *IsSet->GetDataHK(h,-k) == 1. ? 0. : *mFourierData->GetRealHK(h,-k) ;
						*IsSet->GetDataHK(h,-k) = 1. ;
						if (w1 > 0.) numTerms++ ;
						if (numTerms > 0) amplitude /= numTerms ;
					}
					*mFourierData->GetRealHK(h,k) = amplitude ;
					*mFourierData->GetRealHK(-h,k) = amplitude ;
					*mFourierData->GetRealHK(-h,-k) = amplitude ;
					*mFourierData->GetRealHK(h,-k) = amplitude ;
				}
			}
			break ;
		case 11: // p2gm
			// Exchange h,k
			// |F(h,k)| = |F(-h,k)|
			for(h=-hMax;h<=hMax;h++) {
				for(k=0;k<=kMax;k++) {
					if (k == 0 && h <= 0) continue ;
					if (*IsSet->GetDataHK(k,h) == 1.) continue ;
					*IsSet->GetDataHK(k,h) = 1. ;
					if (h == 0 && k%2 == 1) {
						// Systematic absence
						amplitude = 0. ;
						*IsSet->GetDataHK(k,h) = 1. ;
						*IsSet->GetDataHK(-k,h) = 1. ;
						*IsSet->GetDataHK(-k,-h) = 1. ;
						*IsSet->GetDataHK(k,-h) = 1. ;
					}
					else {
						numTerms = 0 ;
						amplitude = w1 = *mFourierData->GetRealHK(k,h) ;
						if (w1 > 0.) numTerms++ ;
						amplitude += w1 = *IsSet->GetDataHK(-k,-h) == 1. ? 0. : *mFourierData->GetRealHK(-k,-h) ;
						*IsSet->GetDataHK(-k,-h) = 1. ;
						if (w1 > 0.) numTerms++ ;
						amplitude += w1 = *IsSet->GetDataHK(-k,h) == 1. ? 0. : *mFourierData->GetRealHK(-k,h) ;
						*IsSet->GetDataHK(-k,h) = 1. ;
						if (w1 > 0.) numTerms++ ;
						amplitude += w1 = *IsSet->GetDataHK(k,-h) == 1. ? 0. : *mFourierData->GetRealHK(k,-h) ;
						*IsSet->GetDataHK(k,-h) = 1. ;
						if (w1 > 0.) numTerms++ ;
						if (numTerms > 0) amplitude /= numTerms ;
					}
					*mFourierData->GetRealHK(k,h) = amplitude ;
					*mFourierData->GetRealHK(-k,h) = amplitude ;
					*mFourierData->GetRealHK(-k,-h) = amplitude ;
					*mFourierData->GetRealHK(k,-h) = amplitude ;
				}
			}
			break ;
		case 12: // pgg  Systematic absence (h0): h = 2n+1 ; (0k): k = 2n+1
			// |F(h,k)| = |F(-h,k)|
			for(h=-hMax;h<=hMax;h++) {
				for(k=0;k<=kMax;k++) {
					if (k == 0 && h <= 0) continue ;
					if (*IsSet->GetDataHK(h,k) == 1.) continue ;
					*IsSet->GetDataHK(h,k) = 1. ;
					if ((h == 0 && k%2 == 1) || (k == 0 && h%2 == 1)) {
						// Systematic absence
						amplitude = 0. ;
						*IsSet->GetDataHK(h,k) = 1. ;
						*IsSet->GetDataHK(-h,k) = 1. ;
						*IsSet->GetDataHK(-h,-k) = 1. ;
						*IsSet->GetDataHK(h,-k) = 1. ;
					}
					else {
						numTerms = 0 ;
						amplitude = w1 = *mFourierData->GetRealHK(h,k) ;
						if (w1 > 0.) numTerms++ ;
						amplitude += w1 = *IsSet->GetDataHK(-h,-k) == 1. ? 0. : *mFourierData->GetRealHK(-h,-k) ;
						*IsSet->GetDataHK(-h,-k) = 1. ;
						if (w1 > 0.) numTerms++ ;
						amplitude += w1 = *IsSet->GetDataHK(-h,k) == 1. ? 0. : *mFourierData->GetRealHK(-h,k) ;
						*IsSet->GetDataHK(-h,k) = 1. ;
						if (w1 > 0.) numTerms++ ;
						amplitude += w1 = *IsSet->GetDataHK(h,-k) == 1. ? 0. : *mFourierData->GetRealHK(h,-k) ;
						*IsSet->GetDataHK(h,-k) = 1. ;
						if (w1 > 0.) numTerms++ ;
						if (numTerms > 0) amplitude /= numTerms ;
					}
					*mFourierData->GetRealHK(h,k) = amplitude ;
					*mFourierData->GetRealHK(-h,k) = amplitude ;
					*mFourierData->GetRealHK(-h,-k) = amplitude ;
					*mFourierData->GetRealHK(h,-k) = amplitude ;
				}
			}
			break ;
		case 13: // cmm  Systematic absence (hk): h+k = 2n+1
			// |F(h,k)| = |F(-h,k)|
			for(h=-hMax;h<=hMax;h++) {
				for(k=0;k<=kMax;k++) {
					if (k == 0 && h <= 0) continue ;
					if (*IsSet->GetDataHK(h,k) == 1.) continue ;
					*IsSet->GetDataHK(h,k) = 1. ;
					if ((h+k)%2 == 1) {
						amplitude = 0. ;
						*IsSet->GetDataHK(h,k) = 1. ;
						*IsSet->GetDataHK(-h,k) = 1. ;
						*IsSet->GetDataHK(-h,-k) = 1. ;
						*IsSet->GetDataHK(h,-k) = 1. ;
					}
					else {
						numTerms = 0 ;
						amplitude = w1 = *mFourierData->GetRealHK(h,k) ;
						if (w1 > 0.) numTerms++ ;
						amplitude += w1 = *IsSet->GetDataHK(-h,-k) == 1. ? 0. : *mFourierData->GetRealHK(-h,-k) ;
						*IsSet->GetDataHK(-h,-k) = 1. ;
						if (w1 > 0.) numTerms++ ;
						amplitude += w1 = *IsSet->GetDataHK(-h,k) == 1. ? 0. : *mFourierData->GetRealHK(-h,k) ;
						*IsSet->GetDataHK(-h,k) = 1. ;
						if (w1 > 0.) numTerms++ ;
						amplitude += w1 = *IsSet->GetDataHK(h,-k) == 1. ? 0. : *mFourierData->GetRealHK(h,-k) ;
						*IsSet->GetDataHK(h,-k) = 1. ;
						if (w1 > 0.) numTerms++ ;
						if (numTerms > 0) amplitude /= numTerms ;
					}
					*mFourierData->GetRealHK(h,k) = amplitude ;
					*mFourierData->GetRealHK(-h,k) = amplitude ;
					*mFourierData->GetRealHK(-h,-k) = amplitude ;
					*mFourierData->GetRealHK(h,-k) = amplitude ;
				}
			}
			break ;
		case 14: // p4
			// |F(h,k)| = |F(-k,h)|
			latticeParameter = (mAStar+mBStar)/2 ;
			mAStar = mBStar = latticeParameter ;
			mA = mB = 1./mAStar ;
			for(h=-hMax;h<=hMax;h++) {
				for(k=0;k<=kMax;k++) {
					if (k == 0 && h <= 0) continue ;
					if (*IsSet->GetDataHK(h,k) == 1.) continue ;
					*IsSet->GetDataHK(h,k) = 1. ;
					numTerms = 0 ;
					amplitude = w1 = *mFourierData->GetRealHK(h,k) ;
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(-h,-k) == 1. ? 0. : *mFourierData->GetRealHK(-h,-k) ;
					*IsSet->GetDataHK(-h,-k) = 1. ;
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(-k,h) == 1. ? 0. : *mFourierData->GetRealHK(-k,h) ;
					*IsSet->GetDataHK(-k,h) = 1. ;
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(k,-h) == 1. ? 0. : *mFourierData->GetRealHK(k,-h) ;
					*IsSet->GetDataHK(k,-h) = 1. ;
					if (w1 > 0.) numTerms++ ;
					if (numTerms > 0) amplitude /= numTerms ;
					*mFourierData->GetRealHK(h,k) = amplitude ;
					*mFourierData->GetRealHK(-k,h) = amplitude ;
					*mFourierData->GetRealHK(-h,-k) = amplitude ;
					*mFourierData->GetRealHK(k,-h) = amplitude ;
				}
			}
			break ;
		case 15: // p4m
		case 16: // p4g
			// |F(h,k)| = |F(-k,h)| = |F(-h,k)|
			latticeParameter = (mAStar+mBStar)/2 ;
			mAStar = mBStar = latticeParameter ;
			mA = mB = 1./mAStar ;
			for(h=-hMax;h<=hMax;h++) {
				for(k=0;k<=kMax;k++) {
					if (k == 0 && h <= 0) continue ;
					if (*IsSet->GetDataHK(h,k) == 1.) continue ;
					*IsSet->GetDataHK(h,k) = 1. ;
					numTerms = 0 ;
					amplitude = w1 = *mFourierData->GetRealHK(h,k) ;
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(-h,-k) == 1. ? 0. : *mFourierData->GetRealHK(-h,-k) ;
					*IsSet->GetDataHK(-h,-k) = 1. ;
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(-k,h) == 1. ? 0. : *mFourierData->GetRealHK(-k,h) ;
					*IsSet->GetDataHK(-k,h) = 1. ;
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(k,-h) == 1. ? 0. : *mFourierData->GetRealHK(k,-h) ;
					*IsSet->GetDataHK(k,-h) = 1. ;
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(-h,k) == 1. ? 0. : *mFourierData->GetRealHK(-h,k) ;
					*IsSet->GetDataHK(-h,k) = 1. ;
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(h,-k) == 1. ? 0. : *mFourierData->GetRealHK(h,-k) ;
					*IsSet->GetDataHK(h,-k) = 1. ;
					if (w1 > 0.) numTerms++ ;
					if (numTerms > 0) amplitude /= numTerms ;
					*mFourierData->GetRealHK(h,k) = amplitude ;
					*mFourierData->GetRealHK(-k,h) = amplitude ;
					*mFourierData->GetRealHK(-h,k) = amplitude ;
					*mFourierData->GetRealHK(-h,-k) = amplitude ;
					*mFourierData->GetRealHK(k,-h) = amplitude ;
					*mFourierData->GetRealHK(h,-k) = amplitude ;
				}
			}
			break ;
		case 17: // p4mg
			// Exchange h,k
			// |F(h,k)| = |F(-k,h)| = |F(-h,k)|
			latticeParameter = (mAStar+mBStar)/2 ;
			mAStar = mBStar = latticeParameter ;
			mA = mB = 1./mAStar ;
			for(h=-hMax;h<=hMax;h++) {
				for(k=0;k<=kMax;k++) {
					if (k == 0 && h <= 0) continue ;
					if (*IsSet->GetDataHK(k,h) == 1.) continue ;
					*IsSet->GetDataHK(k,h) = 1. ;
					numTerms = 0 ;
					amplitude = w1 = *mFourierData->GetRealHK(k,h) ;
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(-k,-h) == 1. ? 0. : *mFourierData->GetRealHK(-k,-h) ;
					*IsSet->GetDataHK(-k,-h) = 1. ;
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(-h,k) == 1. ? 0. : *mFourierData->GetRealHK(-h,k) ;
					*IsSet->GetDataHK(-h,k) = 1. ;
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(h,-k) == 1. ? 0. : *mFourierData->GetRealHK(h,-k) ;
					*IsSet->GetDataHK(h,-k) = 1. ;
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(-k,h) == 1. ? 0. : *mFourierData->GetRealHK(-k,h) ;
					*IsSet->GetDataHK(-k,h) = 1. ;
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(k,-h) == 1. ? 0. : *mFourierData->GetRealHK(k,-h) ;
					*IsSet->GetDataHK(k,-h) = 1. ;
					if (w1 > 0.) numTerms++ ;
					if (numTerms > 0) amplitude /= numTerms ;
					*mFourierData->GetRealHK(k,h) = amplitude ;
					*mFourierData->GetRealHK(-h,k) = amplitude ;
					*mFourierData->GetRealHK(-k,h) = amplitude ;
					*mFourierData->GetRealHK(-k,-h) = amplitude ;
					*mFourierData->GetRealHK(h,-k) = amplitude ;
					*mFourierData->GetRealHK(k,-h) = amplitude ;
				}
			}
			break ;
		case 18: // p3
			// |F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)|
			latticeParameter = (mAStar+mBStar)/2 ;
			mAStar = mBStar = latticeParameter ;
			mA = mB = 1./(mAStar*sin(60.*asin(1.0)/90.)) ;	//  1/Asin(gamma)
			for(h=-hMax;h<=hMax;h++) {
				for(k=0;k<=kMax;k++) {
					if (k == 0 && h <= 0) continue ;
					if (*IsSet->GetDataHK(h,k) == 1.) continue ;
					*IsSet->GetDataHK(h,k) = 1. ;
					numTerms = 0 ;

					amplitude = w1 = *mFourierData->GetRealHK(h,k) ;
					if (w1 > 0.) numTerms++ ;
					if (useFriedelSymmetry) {
						amplitude += w1 = *IsSet->GetDataHK(-h,-k) == 1. ? 0. : *mFourierData->GetRealHK(-h,-k) ;
						*IsSet->GetDataHK(-h,-k) = 1. ;
						if (w1 > 0.) numTerms++ ;
					}
					amplitude += w1 = *IsSet->GetDataHK(k,-h-k) == 1. ? 0. : *mFourierData->GetRealHK(k,-h-k) ;
					*IsSet->GetDataHK(k,-h-k) = 1. ;
					if (w1 > 0.) numTerms++ ;
					if (useFriedelSymmetry) {
						amplitude += w1 = *IsSet->GetDataHK(-k,h+k) == 1. ? 0. : *mFourierData->GetRealHK(-k,h+k) ;
						*IsSet->GetDataHK(-k,h+k) = 1. ;
						if (w1 > 0.) numTerms++ ;
					}
					amplitude += w1 = *IsSet->GetDataHK(-h-k,h) == 1. ? 0. : *mFourierData->GetRealHK(-h-k,h) ;
					*IsSet->GetDataHK(-h-k,h) = 1. ;
					if (w1 > 0.) numTerms++ ;
					if (useFriedelSymmetry) {
						amplitude += w1 = *IsSet->GetDataHK(h+k,-h) == 1. ? 0. : *mFourierData->GetRealHK(h+k,-h) ;
						*IsSet->GetDataHK(h+k,-h) = 1. ;
						if (w1 > 0.) numTerms++ ;
					}
					if (numTerms > 0) amplitude /= numTerms ;
					*mFourierData->GetRealHK(h,k) = amplitude ;
					*mFourierData->GetRealHK(k,-h-k) = amplitude ;
					*mFourierData->GetRealHK(-h-k,h) = amplitude ;
					if (useFriedelSymmetry) *mFourierData->GetRealHK(-h,-k) = amplitude ;
					if (useFriedelSymmetry) *mFourierData->GetRealHK(-k,h+k) = amplitude ;
					if (useFriedelSymmetry) *mFourierData->GetRealHK(h+k,-h) = amplitude ;
				}
			}
			break ;
		case 19: // p3m1
		case 20: // p31m
			// |F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)| = |F(k,h)|
			latticeParameter = (mAStar+mBStar)/2 ;
			mAStar = mBStar = latticeParameter ;
			mA = mB = 1./(mAStar*sin(60.*asin(1.0)/90.)) ;	//  1/Asin(gamma)
			for(h=-hMax;h<=hMax;h++) {
				for(k=0;k<=kMax;k++) {

					if (k == 0 && h <= 0) continue ;
					if (*IsSet->GetDataHK(h,k) == 1.) continue ;
					*IsSet->GetDataHK(h,k) = 1. ;
					numTerms = 0 ;

					amplitude = w1 = *mFourierData->GetRealHK(h,k) ;
					if (w1 > 0.) numTerms++ ;
					if (useFriedelSymmetry) {
						amplitude += w1 = *IsSet->GetDataHK(-h,-k) == 1. ? 0. : *mFourierData->GetRealHK(-h,-k) ;
						*IsSet->GetDataHK(-h,-k) = 1. ;
						if (w1 > 0.) numTerms++ ;
					}
					amplitude += w1 = *IsSet->GetDataHK(k,-h-k) == 1. ? 0. : *mFourierData->GetRealHK(k,-h-k) ;
					*IsSet->GetDataHK(k,-h-k) = 1. ;
					if (w1 > 0.) numTerms++ ;
					if (useFriedelSymmetry) {
						amplitude += w1 = *IsSet->GetDataHK(-k,h+k) == 1. ? 0. : *mFourierData->GetRealHK(-k,h+k) ;
						*IsSet->GetDataHK(-k,h+k) = 1. ;
						if (w1 > 0.) numTerms++ ;
					}
					amplitude += w1 = *IsSet->GetDataHK(-h-k,h) == 1. ? 0. : *mFourierData->GetRealHK(-h-k,h) ;
					*IsSet->GetDataHK(-h-k,h) = 1. ;
					if (w1 > 0.) numTerms++ ;
					if (useFriedelSymmetry) {
						amplitude += w1 = *IsSet->GetDataHK(h+k,-h) == 1. ? 0. : *mFourierData->GetRealHK(h+k,-h) ;
						*IsSet->GetDataHK(h+k,-h) = 1. ;
						if (w1 > 0.) numTerms++ ;
					}
					amplitude /= numTerms ;
					amplitude += w1 = *IsSet->GetDataHK(k,h) == 1. ? 0. : *mFourierData->GetRealHK(k,h) ;
					*IsSet->GetDataHK(k,h) = 1. ;
					if (w1 > 0.) numTerms++ ;
					if (useFriedelSymmetry) {
						amplitude += w1 = *IsSet->GetDataHK(-k,-h) == 1. ? 0. : *mFourierData->GetRealHK(-k,-h) ;
						*IsSet->GetDataHK(h+k,-h) = 1. ;
						if (w1 > 0.) numTerms++ ;
					}
					if (numTerms > 0) amplitude /= numTerms ;

					*mFourierData->GetRealHK(h,k) = amplitude ;
					*mFourierData->GetRealHK(k,-h-k) = amplitude ;
					*mFourierData->GetRealHK(-h-k,h) = amplitude ;
					*mFourierData->GetRealHK(k,h) = amplitude ;
					if (useFriedelSymmetry) *mFourierData->GetRealHK(-h,-k) = amplitude ;
					if (useFriedelSymmetry) *mFourierData->GetRealHK(-k,h+k) = amplitude ;
					if (useFriedelSymmetry) *mFourierData->GetRealHK(h+k,-h) = amplitude ;
					if (useFriedelSymmetry) *mFourierData->GetRealHK(-k,-h) = amplitude ;
				}
			}
			break ;
		case 21: // p6
			// |F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)|
			latticeParameter = (mAStar+mBStar)/2 ;
			mAStar = mBStar = latticeParameter ;
			mA = mB = 1./(mAStar*sin(60.*asin(1.0)/90.)) ;	//  1/Asin(gamma)
			for(h=-hMax;h<=hMax;h++) {
				for(k=0;k<=kMax;k++) {
					if (k == 0 && h <= 0) continue ;
					if (*IsSet->GetDataHK(h,k) == 1.) continue ;
					*IsSet->GetDataHK(h,k) = 1. ;
					numTerms = 0 ;

					amplitude = w1 = *mFourierData->GetRealHK(h,k) ;
					if (w1 > 0.) numTerms++ ;
					
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(-h,-k) == 1. ? 0. : *mFourierData->GetRealHK(-h,-k) ;
					*IsSet->GetDataHK(-h,-k) = 1. ;
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(k,-h-k) == 1. ? 0. : *mFourierData->GetRealHK(k,-h-k) ;
					*IsSet->GetDataHK(k,-h-k) = 1. ;
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(-k,h+k) == 1. ? 0. : *mFourierData->GetRealHK(-k,h+k) ;
					*IsSet->GetDataHK(-k,h+k) = 1. ;
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(-h-k,h) == 1. ? 0. : *mFourierData->GetRealHK(-h-k,h) ;
					*IsSet->GetDataHK(-h-k,h) = 1. ;
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(h+k,-h) == 1. ? 0. : *mFourierData->GetRealHK(h+k,-h) ;
					*IsSet->GetDataHK(h+k,-h) = 1. ;
					if (w1 > 0.) numTerms++ ;
					if (numTerms > 0) amplitude /= numTerms ;
					*mFourierData->GetRealHK(h,k) 		= amplitude ;
					*mFourierData->GetRealHK(k,-h-k) 	= amplitude ;
					*mFourierData->GetRealHK(-h-k,h) 	= amplitude ;
					*mFourierData->GetRealHK(-h,-k) 	= amplitude ;
					*mFourierData->GetRealHK(-k,h+k) 	= amplitude ;
					*mFourierData->GetRealHK(h+k,-h) 	= amplitude ;
				}
			}
			break ;
		case 22: // p6m
			// |F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)| = |F(k,h)|
			latticeParameter = (mAStar+mBStar)/2 ;
			mAStar = mBStar = latticeParameter ;
			mA = mB = 1./(mAStar*sin(60.*asin(1.0)/90.)) ;	//  1/Asin(gamma)
			for(h=-hMax;h<=hMax;h++) {
				for(k=0;k<=kMax;k++) {

					if (k == 0 && h <= 0) continue ;
					if (*IsSet->GetDataHK(h,k) == 1.) continue ;
					*IsSet->GetDataHK(h,k) = 1. ;
					numTerms = 0 ;

					amplitude = w1 = *mFourierData->GetRealHK(h,k) ;
					if (w1 > 0.) numTerms++ ;
					
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(-h,-k) == 1. ? 0. : *mFourierData->GetRealHK(-h,-k) ;
					*IsSet->GetDataHK(-h,-k) = 1. ;
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(k,-h-k) == 1. ? 0. : *mFourierData->GetRealHK(k,-h-k) ;
					*IsSet->GetDataHK(k,-h-k) = 1. ;
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(-k,h+k) == 1. ? 0. : *mFourierData->GetRealHK(-k,h+k) ;
					*IsSet->GetDataHK(-k,h+k) = 1. ;
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(-h-k,h) == 1. ? 0. : *mFourierData->GetRealHK(-h-k,h) ;
					*IsSet->GetDataHK(-h-k,h) = 1. ;
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(h+k,-h) == 1. ? 0. : *mFourierData->GetRealHK(h+k,-h) ;
					*IsSet->GetDataHK(h+k,-h) = 1. ;
					if (w1 > 0.) numTerms++ ;
					amplitude /= numTerms ;
					amplitude += w1 = *IsSet->GetDataHK(k,h) == 1. ? 0. : *mFourierData->GetRealHK(k,h) ;
					*IsSet->GetDataHK(k,h) = 1. ;
					if (w1 > 0.) numTerms++ ;
					amplitude += w1 = *IsSet->GetDataHK(-k,-h) == 1. ? 0. : *mFourierData->GetRealHK(-k,-h) ;
					*IsSet->GetDataHK(-k,-h) = 1. ;
					if (w1 > 0.) numTerms++ ;
					if (numTerms > 0) amplitude /= numTerms ;
					*mFourierData->GetRealHK(h,k) 		= amplitude ;
					*mFourierData->GetRealHK(k,-h-k) 	= amplitude ;
					*mFourierData->GetRealHK(-h-k,h) 	= amplitude ;
					*mFourierData->GetRealHK(-h,-k) 	= amplitude ;
					*mFourierData->GetRealHK(-k,h+k) 	= amplitude ;
					*mFourierData->GetRealHK(h+k,-h) 	= amplitude ;
					*mFourierData->GetRealHK(k,h) 		= amplitude ;
					*mFourierData->GetRealHK(-k,-h) 	= amplitude ;
					
				}
			}
			break ;
	}

	delete IsSet ;

}


TArray*	TDiffractionPeaks::Symmetrize(short numPlaneGroup,short n0,short m0) {
	// Find the origin that minimizes the phase residual
	// Symmetrize the phases and the amplitudes and report the
	// phase residual

	// ---------------------------------------------------------
	// For TEM images plane groups run from 1-17 no variants
	// ---------------------------------------------------------

	float	s,argX,argY,w1,w2,w3,w4,p1,p2,p3,p4,wMax,dph,dpk,dn0,dm0,dn,dm,alphaSym ;
	short	h,k,hMax,kMax,Width,Height ;
	float	latticeParameter ;
	Cmplx	cc(0.,0.) ;

	Width = mFourierData->GetWidth() ;
	Height = mFourierData->GetHeight() ;

	TArray	*IsSet = new TArray(Width,Height) ;
	
	short	minimumDimension = 64 ;

	// We will use a minimum of minimumDimension (64) points per unit cell
	// This pretty much corresponds to minimum of a 
	short ucWidth = Width > minimumDimension ? Width : minimumDimension ;		
	short ucHeight = Height > minimumDimension ? Height : minimumDimension ;
	
	int	axisRatio = mA > mB ? near(mA/mB)/2 : near(mB/mA)/2 ;
	
	if (axisRatio > 1) {
		int scale = near(pow(2.0,axisRatio)) ;
		if (mA > mB) ucWidth *= scale ;
		if (mB > mA) ucHeight *= scale ;
	}
	
	TArray	*result = new TArray(ucWidth,ucHeight,kComplexArray) ;

	*result = cc ;
	
	dn0 = TwoPi/Width ;
	dm0 = TwoPi/Height ;

	hMax = Width/2 ;
	kMax = Height/2 ;
	dn = n0*dn0 ;
	dm = m0*dm0 ;
	
	*IsSet = 0. ;

	// Phase relation P(h,k) = -P(-h,-k) + 
	// Amplitude Relation  |F(h,k)| = |F(-h,-k)| + 
	switch(numPlaneGroup) {
		case 1:	// p1
			for(h=-hMax+1;h<hMax;h++) {
				dph = h*dn ;
				for(k=0;k<kMax;k++) {
					if (k == 0 && h <= 0) continue ;
					w1 = *mFourierData->GetRealHK(h,k) ;
					if (w1 == 0.) continue ;
					dpk = k*dm ;
					p1 = *mFourierData->GetImagHK(h,k) + dph + dpk ;
					argY = sin(p1) ;
					argX = cos(p1) ;
					alphaSym = atan2(argY,argX) ;
					*mFourierData->GetImagHK(h,k) = alphaSym ;
					*mFourierData->GetImagHK(-h,-k) = -alphaSym ;
				}
			}
			break ;
		case 2: // p2
			// Phase 0 or 180  : (h,k)
			for(h=-hMax+1;h<hMax;h++) {
				dph = h*dn ;
				for(k=0;k<kMax;k++) {
					if (k == 0 && h <= 0) continue ;
					w1 = *mFourierData->GetRealHK(h,k) ;
					if (w1 == 0.) continue ;
					dpk = k*dm ;
					p1 = *mFourierData->GetImagHK(h,k) + dph + dpk ;
					argY = sin(p1) ;
					argX = cos(p1) ;
					alphaSym = atan2(argY,argX) ;
					alphaSym = fabs(alphaSym) < PiHalf ? 0. : Pi ;
					*mFourierData->GetImagHK(h,k) = alphaSym ;
					*mFourierData->GetImagHK(-h,-k) = -alphaSym ;
				}
			}
			break ;
		case 3: // pm - 
			// |F(h,k)| = |F(-h,k)|
			// P(h,k) = P(-h,k) = -P(h,-k)
			// Phase 0 or 180  : (h,0)
			for(h=-hMax+1;h<hMax;h++) {
				dph = h*dn ;
				for(k=0;k<kMax;k++) {
					if (k == 0 && h <= 0) continue ;
					w1 = *mFourierData->GetRealHK(h,k) ;
					if (w1 == 0.) continue ;
					if (*IsSet->GetDataHK(h,k) == 1.) {
						wMax = *mFourierData->GetRealHK(h,k) ;
						alphaSym = *mFourierData->GetImagHK(h,k) ;
					}
					else {
						dpk = k*dm ;
						p1 = *mFourierData->GetImagHK(h,k) + dph + dpk ;
						p2 = *mFourierData->GetImagHK(-h,k) - dph + dpk ;
						w2 = *mFourierData->GetRealHK(-h,k) ;
						wMax = max(w1,w2) ;
						// Calculate symmetrized phase from P(h,k) and P(-h,k)
						argY = w1*sin(p1)+w2*sin(p2) ;
						argX = w1*cos(p1)+w2*cos(p2) ;
						alphaSym = atan2(argY,argX) ;
						if ( k == 0)	
							alphaSym = fabs(alphaSym) < PiHalf ? 0. : Pi ;
					}
					*mFourierData->GetImagHK(h,k) = alphaSym ;
					*mFourierData->GetImagHK(-h,k) = alphaSym ;
					*mFourierData->GetImagHK(-h,-k) = -alphaSym ;
					*mFourierData->GetImagHK(h,-k) = -alphaSym ;
					*mFourierData->GetRealHK(h,k) = wMax ;
					*mFourierData->GetRealHK(-h,k) = wMax ;
					*mFourierData->GetRealHK(-h,-k) = wMax ;
					*mFourierData->GetRealHK(h,-k) = wMax ;
					*IsSet->GetDataHK(h,k) = 1. ;
					*IsSet->GetDataHK(-h,k) = 1. ;
					*IsSet->GetDataHK(-h,-k) = 1. ;
					*IsSet->GetDataHK(h,-k) = 1. ;
				}
			}
			break ;
		case 4: // pg  Systematic absence (0k): k = 2n+1
			// |F(h,k)| = |F(-h,k)|
			// P(h,k) = P(-h,k)+k*180
			// Phase 0 or 180  : (h,0)
			for(h=-hMax+1;h<hMax;h++) {
				dph = h*dn ;
				for(k=0;k<kMax;k++) {
					w1 = *mFourierData->GetRealHK(h,k) ;
					if (!(h == 0 && k == 0) && w1 > 0.) {
						if (*IsSet->GetDataHK(h,k) == 1.) {
							wMax = *mFourierData->GetRealHK(h,k) ;
							alphaSym = *mFourierData->GetImagHK(h,k) ;
						}
						else {
							dpk = k*dm ;
							p1 = *mFourierData->GetImagHK(h,k) + dph + dpk ;
							p2 = *mFourierData->GetImagHK(-h,k) - dph + dpk ;
							w2 = *mFourierData->GetRealHK(-h,k) ;
							if (h == 0 && k%2 == 1)
								wMax = 0. ;
							else
								wMax = max(w1,w2) ;
							// Calculate symmetrized phase from P(h,k) and P(-h,k)
							s = k%2 == 0 ? 1. : -1. ;
							argY = w1*sin(p1)+s*w2*sin(p2) ;
							argX = w1*cos(p1)+s*w2*cos(p2) ;
							alphaSym = atan2(argY,argX) ;
							if ( k == 0)	
								alphaSym = fabs(alphaSym) < PiHalf ? 0. : Pi ;
						}
						*mFourierData->GetImagHK(h,k) = alphaSym ;
						*mFourierData->GetImagHK(-h,k) = Mod2Pi(alphaSym-k*Pi) ;
						*mFourierData->GetImagHK(-h,-k) = -alphaSym ;
						*mFourierData->GetImagHK(h,-k) = -Mod2Pi(alphaSym-k*Pi) ;
						*mFourierData->GetRealHK(h,k) = wMax ;
						*mFourierData->GetRealHK(-h,k) = wMax ;
						*mFourierData->GetRealHK(-h,-k) = wMax ;
						*mFourierData->GetRealHK(h,-k) = wMax ;
						*IsSet->GetDataHK(h,k) = 1. ;
						*IsSet->GetDataHK(-h,k) = 1. ;
						*IsSet->GetDataHK(-h,-k) = 1. ;
						*IsSet->GetDataHK(h,-k) = 1. ;
					}
				}
			}
			break ;
		case 5: // cm  Systematic absence (hk): h+k = 2n+1
			// |F(h,k)| = |F(-h,k)|
			// P(h,k) = P(-h,k) = -P(h,-k)
			// Phase 0 or 180  : (h,0)
			for(h=-hMax+1;h<hMax;h++) {
				dph = h*dn ;
				for(k=0;k<kMax;k++) {
					w1 = *mFourierData->GetRealHK(h,k) ;
					if (!(h == 0 && k == 0) && w1 > 0.) {
						if (*IsSet->GetDataHK(h,k) == 1.) {
							wMax = *mFourierData->GetRealHK(h,k) ;
							alphaSym = *mFourierData->GetImagHK(h,k) ;
						}
						else {
							dpk = k*dm ;
							p1 = *mFourierData->GetImagHK(h,k) + dph + dpk ;
							p2 = *mFourierData->GetImagHK(-h,k) - dph + dpk ;
							w2 = *mFourierData->GetRealHK(-h,k) ;
							if ((h+k)%2 == 1)
								wMax = 0. ;
							else
								wMax = max(w1,w2) ;
							// Calculate symmetrized phase from P(h,k) and P(-h,k)
							argY = w1*sin(p1)+w2*sin(p2) ;
							argX = w1*cos(p1)+w2*cos(p2) ;
							alphaSym = atan2(argY,argX) ;
							if ( k == 0)	
								alphaSym = fabs(alphaSym) < PiHalf ? 0. : Pi ;
						}
						*mFourierData->GetImagHK(h,k) = alphaSym ;
						*mFourierData->GetImagHK(-h,k) = Mod2Pi(alphaSym-k*Pi) ;
						*mFourierData->GetImagHK(-h,-k) = -alphaSym ;
						*mFourierData->GetImagHK(h,-k) = -Mod2Pi(alphaSym-k*Pi) ;
						*mFourierData->GetRealHK(h,k) = wMax ;
						*mFourierData->GetRealHK(-h,k) = wMax ;
						*mFourierData->GetRealHK(-h,-k) = wMax ;
						*mFourierData->GetRealHK(h,-k) = wMax ;
						*IsSet->GetDataHK(h,k) = 1. ;
						*IsSet->GetDataHK(-h,k) = 1. ;
						*IsSet->GetDataHK(-h,-k) = 1. ;
						*IsSet->GetDataHK(h,-k) = 1. ;
					}
				}
			}
			break ;
		case 6: // pmm  -
			// |F(h,k)| = |F(-h,k)|
			// P(h,k) = P(-h,k) = -P(h,-k)
			// Phase 0 or 180  : (h,k)
			for(h=-hMax+1;h<hMax;h++) {
				dph = h*dn ;
				for(k=0;k<kMax;k++) {
					w1 = *mFourierData->GetRealHK(h,k) ;
					if (!(h == 0 && k == 0) && w1 > 0.) {
						if (*IsSet->GetDataHK(h,k) == 1.) {
							wMax = *mFourierData->GetRealHK(h,k) ;
							alphaSym = *mFourierData->GetImagHK(h,k) ;
						}
						else {
							dpk = k*dm ;
							p1 = *mFourierData->GetImagHK(h,k) + dph + dpk ;
							p2 = *mFourierData->GetImagHK(-h,k) - dph + dpk ;
							w2 = *mFourierData->GetRealHK(-h,k) ;
							wMax = max(w1,w2) ;
							// Calculate symmetrized phase from P(h,k) and P(-h,k)
							argY = w1*sin(p1)+w2*sin(p2) ;
							argX = w1*cos(p1)+w2*cos(p2) ;
							alphaSym = atan2(argY,argX) ;
							alphaSym = fabs(alphaSym) < PiHalf ? 0. : Pi ;
						}
						*mFourierData->GetImagHK(h,k) = alphaSym ;
						*mFourierData->GetImagHK(-h,k) = alphaSym ;
						*mFourierData->GetImagHK(-h,-k) = -alphaSym ;
						*mFourierData->GetImagHK(h,-k) = -alphaSym ;
						*mFourierData->GetRealHK(h,k) = wMax ;
						*mFourierData->GetRealHK(-h,k) = wMax ;
						*mFourierData->GetRealHK(-h,-k) = wMax ;
						*mFourierData->GetRealHK(h,-k) = wMax ;
						*IsSet->GetDataHK(h,k) = 1. ;
						*IsSet->GetDataHK(-h,k) = 1. ;
						*IsSet->GetDataHK(-h,-k) = 1. ;
						*IsSet->GetDataHK(h,-k) = 1. ;
					}
				}
			}
			break ;
		case 7: // pmg  Systematic absence (0k): k = 2n+1
			// |F(h,k)| = |F(-h,k)|
			// P(h,k) = P(-h,k)+k*180
			// Phase 0 or 180  : (h,k)
			for(h=-hMax+1;h<hMax;h++) {
				dph = h*dn ;
				for(k=0;k<kMax;k++) {
					w1 = *mFourierData->GetRealHK(h,k) ;
					if (!(h == 0 && k == 0) && w1 > 0.) {
						if (*IsSet->GetDataHK(h,k) == 1.) {
							wMax = *mFourierData->GetRealHK(h,k) ;
							alphaSym = *mFourierData->GetImagHK(h,k) ;
						}
						else {
							dpk = k*dm ;
							p1 = *mFourierData->GetImagHK(h,k) + dph + dpk ;
							p2 = *mFourierData->GetImagHK(-h,k) - dph + dpk ;
							w2 = *mFourierData->GetRealHK(-h,k) ;
							if (h == 0 && k%2 == 1)
								wMax = 0. ;
							else
								wMax = max(w1,w2) ;
							// Calculate symmetrized phase from P(h,k) and P(-h,k)
							s = k%2 == 0 ? 1. : -1. ;
							argY = w1*sin(p1)+s*w2*sin(p2) ;
							argX = w1*cos(p1)+s*w2*cos(p2) ;
							alphaSym = atan2(argY,argX) ;
							alphaSym = fabs(alphaSym) < PiHalf ? 0. : Pi ;
						}
						*mFourierData->GetImagHK(h,k) = alphaSym ;
						*mFourierData->GetImagHK(-h,k) = Mod2Pi(alphaSym-k*Pi) ;
						*mFourierData->GetImagHK(-h,-k) = -alphaSym ;
						*mFourierData->GetImagHK(h,-k) = -Mod2Pi(alphaSym-k*Pi) ;
						*mFourierData->GetRealHK(h,k) = wMax ;
						*mFourierData->GetRealHK(-h,k) = wMax ;
						*mFourierData->GetRealHK(-h,-k) = wMax ;
						*mFourierData->GetRealHK(h,-k) = wMax ;
						*IsSet->GetDataHK(h,k) = 1. ;
						*IsSet->GetDataHK(-h,k) = 1. ;
						*IsSet->GetDataHK(-h,-k) = 1. ;
						*IsSet->GetDataHK(h,-k) = 1. ;
					}
				}
			}
			break ;
		case 8: // pgg  Systematic absence (h0): h = 2n+1 ; (0k): k = 2n+1
			// |F(h,k)| = |F(-h,k)|
			// P(h,k) = P(-h,k)+(h+k)*180
			// Phase 0 or 180  : (h,k)
			for(h=-hMax+1;h<hMax;h++) {
				dph = h*dn ;
				for(k=0;k<kMax;k++) {
					w1 = *mFourierData->GetRealHK(h,k) ;
					if (!(h == 0 && k == 0) && w1 > 0.) {
						if (*IsSet->GetDataHK(h,k) == 1.) {
							wMax = *mFourierData->GetRealHK(h,k) ;
							alphaSym = *mFourierData->GetImagHK(h,k) ;
						}
						else {
							dpk = k*dm ;
							p1 = *mFourierData->GetImagHK(h,k) + dph + dpk ;
							p2 = *mFourierData->GetImagHK(-h,k) - dph + dpk ;
							w1 = *mFourierData->GetRealHK(h,k) ;
							w2 = *mFourierData->GetRealHK(-h,k) ;
							if ((h == 0 && k%2 == 1) || (k == 0 && h%2 == 1))
								wMax = 0. ;
							else
								wMax = max(w1,w2) ;
							// Calculate symmetrized phase from P(h,k) and P(-h,k)
							s = (h+k)%2 == 0 ? 1. : -1. ;
							argY = w1*sin(p1)+s*w2*sin(p2) ;
							argX = w1*cos(p1)+s*w2*cos(p2) ;
							alphaSym = atan2(argY,argX) ;
							alphaSym = fabs(alphaSym) <PiHalf ? 0. : Pi ;
						}
						*mFourierData->GetImagHK(h,k) = alphaSym ;
						*mFourierData->GetImagHK(-h,k) = Mod2Pi(alphaSym-(h+k)*Pi) ;
						*mFourierData->GetImagHK(-h,-k) = -alphaSym ;
						*mFourierData->GetImagHK(h,-k) = -Mod2Pi(alphaSym-(h+k)*Pi) ;
						*mFourierData->GetRealHK(h,k) = wMax ;
						*mFourierData->GetRealHK(-h,k) = wMax ;
						*mFourierData->GetRealHK(-h,-k) = wMax ;
						*mFourierData->GetRealHK(h,-k) = wMax ;
						*IsSet->GetDataHK(h,k) = 1. ;
						*IsSet->GetDataHK(-h,k) = 1. ;
						*IsSet->GetDataHK(-h,-k) = 1. ;
						*IsSet->GetDataHK(h,-k) = 1. ;
					}
				}
			}
			break ;
		case 9: // cmm  Systematic absence (hk): h+k = 2n+1
			// |F(h,k)| = |F(-h,k)|
			// P(h,k) = P(-h,k) = -P(h,-k)
			// Phase 0 or 180  : (h,k)
			for(h=-hMax+1;h<hMax;h++) {
				dph = h*dn ;
				for(k=0;k<kMax;k++) {
					w1 = *mFourierData->GetRealHK(h,k) ;
					if (!(h == 0 && k == 0) && w1 > 0.) {
						if (*IsSet->GetDataHK(h,k) == 1.) {
							wMax = *mFourierData->GetRealHK(h,k) ;
							alphaSym = *mFourierData->GetImagHK(h,k) ;
						}
						else {
							dpk = k*dm ;
							p1 = *mFourierData->GetImagHK(h,k) + dph + dpk ;
							p2 = *mFourierData->GetImagHK(-h,k) - dph + dpk ;
							w2 = *mFourierData->GetRealHK(-h,k) ;
							if ((h+k)%2 == 1)
								wMax = 0. ;
							else
								wMax = max(w1,w2) ;
							// Calculate symmetrized phase from P(h,k) and P(-h,k)
							argY = w1*sin(p1)+w2*sin(p2) ;
							argX = w1*cos(p1)+w2*cos(p2) ;
							alphaSym = atan2(argY,argX) ;
							alphaSym = fabs(alphaSym) < PiHalf ? 0. : Pi ;
						}
						*mFourierData->GetImagHK(h,k) = alphaSym ;
						*mFourierData->GetImagHK(-h,k) = alphaSym ;
						*mFourierData->GetImagHK(-h,-k) = -alphaSym ;
						*mFourierData->GetImagHK(h,-k) = -alphaSym ;
						*mFourierData->GetRealHK(h,k) = wMax ;
						*mFourierData->GetRealHK(-h,k) = wMax ;
						*mFourierData->GetRealHK(-h,-k) = wMax ;
						*mFourierData->GetRealHK(h,-k) = wMax ;
						*IsSet->GetDataHK(h,k) = 1. ;
						*IsSet->GetDataHK(-h,k) = 1. ;
						*IsSet->GetDataHK(-h,-k) = 1. ;
						*IsSet->GetDataHK(h,-k) = 1. ;
					}
				}
			}
			break ;
		case 10: // p4
			// |F(h,k)| = |F(-k,h)|
			// P(h,k) = P(-k,h) = -P(k,-h)
			// Phase 0 or 180  : (h,k)
			latticeParameter = (mAStar+mBStar)/2 ;
			mAStar = mBStar = latticeParameter ;
			mA = mB = 1./mAStar ;
			for(h=-hMax+1;h<hMax;h++) {
				dph = h*dn ; 
				for(k=0;k<kMax;k++) {
					w1 = *mFourierData->GetRealHK(h,k) ;
					if (!(h == 0 && k == 0) && w1 > 0.) {
						if (*IsSet->GetDataHK(h,k) == 1.) {
							wMax = *mFourierData->GetRealHK(h,k) ;
							alphaSym = *mFourierData->GetImagHK(h,k) ;
						}
						else {
							dpk = k*dm ; 
							p1 = *mFourierData->GetImagHK(h,k) + dph + dpk ;
							p2 = *mFourierData->GetImagHK(-k,h) - k*dn + h*dm ;
							w2 = *mFourierData->GetRealHK(-k,h) ;
							wMax = max(w1,w2) ;
							// Calculate symmetrized phase from P(h,k) and P(-h,k)
							argY = w1*sin(p1)+w2*sin(p2) ;
							argX = w1*cos(p1)+w2*cos(p2) ;
							alphaSym = atan2(argY,argX) ;
							alphaSym = fabs(alphaSym) < PiHalf ? 0. : Pi ;
						}
						*mFourierData->GetImagHK(h,k) = alphaSym ;
						*mFourierData->GetImagHK(-k,h) = alphaSym ;
						*mFourierData->GetImagHK(-h,-k) = -alphaSym ;
						*mFourierData->GetImagHK(k,-h) = -alphaSym ;
						*mFourierData->GetRealHK(h,k) = wMax ;
						*mFourierData->GetRealHK(-k,h) = wMax ;
						*mFourierData->GetRealHK(-h,-k) = wMax ;
						*mFourierData->GetRealHK(k,-h) = wMax ;
						*IsSet->GetDataHK(h,k) = 1. ;
						*IsSet->GetDataHK(-k,h) = 1. ;
						*IsSet->GetDataHK(-h,-k) = 1. ;
						*IsSet->GetDataHK(k,-h) = 1. ;
					}
				}
			}
			break ;
		case 11: // p4m
			// |F(h,k)| = |F(-k,h)| = |F(-h,k)|
			// P(h,k) = P(-k,h) = P(-h,k) = -P(k,-h) = -P(h,-k)
			// Phase 0 or 180  : (h,k)
			latticeParameter = (mAStar+mBStar)/2 ;
			mAStar = mBStar = latticeParameter ;
			mA = mB = 1./mAStar ;
			for(h=-hMax+1;h<hMax;h++) {
				dph = h*dn ; 
				for(k=0;k<kMax;k++) {
					w1 = *mFourierData->GetRealHK(h,k) ;
					if (!(h == 0 && k == 0) && w1 > 0.) {
						if (*IsSet->GetDataHK(h,k) == 1.) {
							wMax = *mFourierData->GetRealHK(h,k) ;
							alphaSym = *mFourierData->GetImagHK(h,k) ;
						}
						else {
							dpk = k*dm ; 
							p1 = *mFourierData->GetImagHK(h,k) + dph + dpk ;
							p2 = *mFourierData->GetImagHK(-h,k) - dph + dpk ;
							p3 = *mFourierData->GetImagHK(-k,h) - k*dn + h*dm ;
							w2 = *mFourierData->GetRealHK(-h,k) ;
							w3 = *mFourierData->GetRealHK(-k,h) ;
							wMax = max(max(w1,w2),w3) ;
							// Calculate symmetrized phase from P(h,k) and P(-h,k)
							argY = w1*sin(p1)+w2*sin(p2)+w3*sin(p3) ;
							argX = w1*cos(p1)+w2*cos(p2)+w3*cos(p3) ;
							alphaSym = atan2(argY,argX) ;
							alphaSym = fabs(alphaSym) < PiHalf ? 0. : Pi ;
						}
						*mFourierData->GetImagHK(h,k) = alphaSym ;
						*mFourierData->GetImagHK(-k,h) = alphaSym ;
						*mFourierData->GetImagHK(-h,k) = alphaSym ;
						*mFourierData->GetImagHK(-h,-k) = -alphaSym ;
						*mFourierData->GetImagHK(k,-h) = -alphaSym ;
						*mFourierData->GetImagHK(h,-k) = -alphaSym ;
						*mFourierData->GetRealHK(h,k) = wMax ;
						*mFourierData->GetRealHK(-k,h) = wMax ;
						*mFourierData->GetRealHK(-h,k) = wMax ;
						*mFourierData->GetRealHK(-h,-k) = wMax ;
						*mFourierData->GetRealHK(k,-h) = wMax ;
						*mFourierData->GetRealHK(h,-k) = wMax ;
						*IsSet->GetDataHK(h,k) = 1. ;
						*IsSet->GetDataHK(-k,h) = 1. ;
						*IsSet->GetDataHK(-h,k) = 1. ;
						*IsSet->GetDataHK(-h,-k) = 1. ;
						*IsSet->GetDataHK(k,-h) = 1. ;
						*IsSet->GetDataHK(h,-k) = 1. ;
					}
				}
			}
			break ;
		case 12: // p4g
			// |F(h,k)| = |F(-k,h)| = |F(-h,k)|
			// P(h,k) = P(-k,h) = P(-h,k)+k*180
			// Phase 0 or 180  : (h,k)
			latticeParameter = (mAStar+mBStar)/2 ;
			mAStar = mBStar = latticeParameter ;
			mA = mB = 1./mAStar ;
			for(h=-hMax+1;h<hMax;h++) {
				dph = h*dn ; 
				for(k=0;k<kMax;k++) {
					w1 = *mFourierData->GetRealHK(h,k) ;
					if (!(h == 0 && k == 0) && w1 > 0.) {
						if (*IsSet->GetDataHK(h,k) == 1.) {
							wMax = *mFourierData->GetRealHK(h,k) ;
							alphaSym = *mFourierData->GetImagHK(h,k) ;
						}
						else {
							dpk = k*dm ; 
							p1 = *mFourierData->GetImagHK(h,k) + dph + dpk ;
							p2 = *mFourierData->GetImagHK(-h,k) - dph + dpk ;
							p3 = *mFourierData->GetImagHK(-k,h) - k*dn + h*dm ;
							w2 = *mFourierData->GetRealHK(-h,k) ;
							w3 = *mFourierData->GetRealHK(-k,h) ;
							wMax = max(max(w1,w2),w3) ;
							// Calculate symmetrized phase from P(h,k) and P(-h,k)
							s = k%2 == 0 ? 1. : -1. ;
							argY = w1*sin(p1)+s*w2*sin(p2)+w3*sin(p3) ;
							argX = w1*cos(p1)+s*w2*cos(p2)+w3*cos(p3) ;
							alphaSym = atan2(argY,argX) ;
							alphaSym = fabs(alphaSym) < PiHalf ? 0. : Pi ;
						}
						*mFourierData->GetImagHK(h,k) = alphaSym ;
						*mFourierData->GetImagHK(-k,h) = alphaSym ;
						*mFourierData->GetImagHK(-h,k) = Mod2Pi(alphaSym-k*Pi) ;
						*mFourierData->GetImagHK(-h,-k) = -alphaSym ;
						*mFourierData->GetImagHK(k,-h) = -alphaSym ;
						*mFourierData->GetImagHK(h,-k) = -Mod2Pi(alphaSym-k*Pi) ;
						*mFourierData->GetRealHK(h,k) = wMax ;
						*mFourierData->GetRealHK(-k,h) = wMax ;
						*mFourierData->GetRealHK(-h,k) = wMax ;
						*mFourierData->GetRealHK(-h,-k) = wMax ;
						*mFourierData->GetRealHK(k,-h) = wMax ;
						*mFourierData->GetRealHK(h,-k) = wMax ;
						*IsSet->GetDataHK(h,k) = 1. ;
						*IsSet->GetDataHK(-k,h) = 1. ;
						*IsSet->GetDataHK(-h,k) = 1. ;
						*IsSet->GetDataHK(-h,-k) = 1. ;
						*IsSet->GetDataHK(k,-h) = 1. ;
						*IsSet->GetDataHK(h,-k) = 1. ;
					}
				}
			}
			break ;
		case 13: // p3
			// |F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)|
			// P(h,k) = P(k,-h-k) = P(-h-k,h)
			// Phase 0 or 180  : -
			latticeParameter = (mAStar+mBStar)/2 ;
			mAStar = mBStar = latticeParameter ;
			mA = mB = 1./(mAStar*sin(60.*asin(1.0)/90.)) ;	//  1/Asin(gamma)
			for(h=-hMax+1;h<hMax;h++) {
				dph = h*dn ; 
				for(k=0;k<kMax;k++) {
					w1 = *mFourierData->GetRealHK(h,k) ;
					if (!(h == 0 && k == 0) && w1 > 0.) {
						if (*IsSet->GetDataHK(h,k) == 1.) {
							wMax = *mFourierData->GetRealHK(h,k) ;
							alphaSym = *mFourierData->GetImagHK(h,k) ;
						}
						else {
							dpk = k*dm ; 
							p1 = *mFourierData->GetImagHK(h,k) + dph + dpk ;
							p2 = *mFourierData->GetImagHK(k,-h-k) +k*dn - (h+k)*dm ;
							p3 = *mFourierData->GetImagHK(-h-k,h) -(h+k)*dn + h*dm ;
							w2 = *mFourierData->GetRealHK(k,-h-k) ;
							w3 = *mFourierData->GetRealHK(-h-k,h) ;
							wMax = max(max(w1,w2),w3) ;
							// Calculate symmetrized phase
							argY = w1*sin(p1)+w2*sin(p2)+w3*sin(p3) ;
							argX = w1*cos(p1)+w2*cos(p2)+w3*cos(p3) ;
							alphaSym = atan2(argY,argX) ;
						}
						*mFourierData->GetImagHK(h,k) = alphaSym ;
						*mFourierData->GetImagHK(k,-h-k) = alphaSym ;
						*mFourierData->GetImagHK(-h-k,h) = alphaSym ;
						*mFourierData->GetImagHK(-h,-k) = -alphaSym ;
						*mFourierData->GetImagHK(-k,h+k) = -alphaSym ;
						*mFourierData->GetImagHK(h+k,-h) = -alphaSym ;
						*mFourierData->GetRealHK(h,k) = wMax ;
						*mFourierData->GetRealHK(k,-h-k) = wMax ;
						*mFourierData->GetRealHK(-h-k,h) = wMax ;
						*mFourierData->GetRealHK(-h,-k) = wMax ;
						*mFourierData->GetRealHK(-k,h+k) = wMax ;
						*mFourierData->GetRealHK(h+k,-h) = wMax ;
						*IsSet->GetDataHK(h,k) = 1. ;
						*IsSet->GetDataHK(k,-h-k) = 1. ;
						*IsSet->GetDataHK(-h-k,h) = 1. ;
						*IsSet->GetDataHK(-h,-k) = 1. ;
						*IsSet->GetDataHK(-k,h+k) = 1. ;
						*IsSet->GetDataHK(h+k,-h) = 1. ;
					}
				}
			}
			break ;
		case 14: // p3m1
			// |F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)| = |F(k,h)|
			// P(h,k) = P(k,-h-k) = P(-h-k,h) = P(-k,-h)
			// Phase 0 or 180  : (h,h) (h,-2h) (-2k,k)
			latticeParameter = (mAStar+mBStar)/2 ;
			mAStar = mBStar = latticeParameter ;
			mA = mB = 1./(mAStar*sin(60.*asin(1.0)/90.)) ;	//  1/Asin(gamma)
			for(h=-hMax+1;h<hMax;h++) {
				dph = h*dn ; 
				for(k=0;k<kMax;k++) {
					w1 = *mFourierData->GetRealHK(h,k) ;
					if (!(h == 0 && k == 0) && w1 > 0.) {
						if (*IsSet->GetDataHK(h,k) == 1.) {
							wMax = *mFourierData->GetRealHK(h,k) ;
							alphaSym = *mFourierData->GetImagHK(h,k) ;
						}
						else {
							dpk = k*dm ; 
							p1 = *mFourierData->GetImagHK(h,k) + dph + dpk ;
							p2 = *mFourierData->GetImagHK(k,-h-k) +k*dn - (h+k)*dm ;
							p3 = *mFourierData->GetImagHK(-h-k,h) -(h+k)*dn + h*dm ;
							p4 = *mFourierData->GetImagHK(-k,-h) - k*dn - h*dm ;
							w2 = *mFourierData->GetRealHK(k,-h-k) ;
							w3 = *mFourierData->GetRealHK(-h-k,h) ;
							//w4 = *mFourierData->GetRealHK(-k,-h) ;
							w4 = h == k ? 0. : *mFourierData->GetRealHK(-k,-h) ;
							wMax = max(max(max(w1,w2),w3),w4) ;
							// Calculate symmetrized phase
							argY = w1*sin(p1)+w2*sin(p2)+w3*sin(p3)+w4*sin(p4) ;
							argX = w1*cos(p1)+w2*cos(p2)+w3*cos(p3)+w4*cos(p4) ;
							alphaSym = atan2(argY,argX) ;
							if (h == k || k == -2*h || h == -2*k)
								alphaSym = fabs(alphaSym) < PiHalf ? 0. : Pi ;
						}
						*mFourierData->GetImagHK(h,k) = alphaSym ;
						*mFourierData->GetImagHK(k,-h-k) = alphaSym ;
						*mFourierData->GetImagHK(-h-k,h) = alphaSym ;
						*mFourierData->GetImagHK(-k,-h) = alphaSym ;
						*mFourierData->GetImagHK(-h,-k) = -alphaSym ;
						*mFourierData->GetImagHK(-k,h+k) = -alphaSym ;
						*mFourierData->GetImagHK(h+k,-h) = -alphaSym ;
						*mFourierData->GetImagHK(k,h) = -alphaSym ;
						*mFourierData->GetRealHK(h,k) = wMax ;
						*mFourierData->GetRealHK(k,-h-k) = wMax ;
						*mFourierData->GetRealHK(-h-k,h) = wMax ;
						*mFourierData->GetRealHK(-k,-h) = wMax ;
						*mFourierData->GetRealHK(-h,-k) = wMax ;
						*mFourierData->GetRealHK(-k,h+k) = wMax ;
						*mFourierData->GetRealHK(h+k,-h) = wMax ;
						*mFourierData->GetRealHK(k,h) = wMax ;
						*IsSet->GetDataHK(h,k) = 1. ;
						*IsSet->GetDataHK(k,-h-k) = 1. ;
						*IsSet->GetDataHK(-h-k,h) = 1. ;
						*IsSet->GetDataHK(k,h) = 1. ;
						*IsSet->GetDataHK(-h,-k) = 1. ;
						*IsSet->GetDataHK(-k,h+k) = 1. ;
						*IsSet->GetDataHK(h+k,-h) = 1. ;
						*IsSet->GetDataHK(-k,-h) = 1. ;
					}
				}
			}
			break ;
		case 15: // p31m
			// |F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)| = |F(k,h)|
			// P(h,k) = P(k,-h-k) = P(-h-k,h) = P(k,h)
			// Phase 0 or 180  : (h,-h) (h,2h) (2k,k)
			latticeParameter = (mAStar+mBStar)/2 ;
			mAStar = mBStar = latticeParameter ;
			mA = mB = 1./(mAStar*sin(60.*asin(1.0)/90.)) ;	//  1/Asin(gamma)
			for(h=-hMax+1;h<hMax;h++) {
				dph = h*dn ; 
				for(k=0;k<kMax;k++) {
					w1 = *mFourierData->GetRealHK(h,k) ;
					if (!(h == 0 && k == 0) && w1 > 0.) {
						if (*IsSet->GetDataHK(h,k) == 1.) {
							wMax = *mFourierData->GetRealHK(h,k) ;
							alphaSym = *mFourierData->GetImagHK(h,k) ;
						}
						else {
							dpk = k*dm ; 
							p1 = *mFourierData->GetImagHK(h,k) + dph + dpk ;
							p2 = *mFourierData->GetImagHK(k,-h-k) +k*dn - (h+k)*dm ;
							p3 = *mFourierData->GetImagHK(-h-k,h) -(h+k)*dn + h*dm ;
							p4 = *mFourierData->GetImagHK(k,h) + k*dn + h*dm ;
							w2 = *mFourierData->GetRealHK(k,-h-k) ;
							w3 = *mFourierData->GetRealHK(-h-k,h) ;
							//w4 = *mFourierData->GetRealHK(k,h) ;
							w4 = h == k ? 0. : *mFourierData->GetRealHK(k,h) ;
							wMax = max(max(max(w1,w2),w3),w4) ;
							// Calculate symmetrized phase
							argY = w1*sin(p1)+w2*sin(p2)+w3*sin(p3)+w4*sin(p4) ;
							argX = w1*cos(p1)+w2*cos(p2)+w3*cos(p3)+w4*cos(p4) ;
							alphaSym = atan2(argY,argX) ;
							if (h == -k || h == 2*k || k == 2*h)
								alphaSym = fabs(alphaSym) < PiHalf ? 0. : Pi ;
						}
						*mFourierData->GetImagHK(h,k) = alphaSym ;
						*mFourierData->GetImagHK(k,-h-k) = alphaSym ;
						*mFourierData->GetImagHK(-h-k,h) = alphaSym ;
						*mFourierData->GetImagHK(k,h) = alphaSym ;
						*mFourierData->GetImagHK(-h,-k) = -alphaSym ;
						*mFourierData->GetImagHK(-k,h+k) = -alphaSym ;
						*mFourierData->GetImagHK(h+k,-h) = -alphaSym ;
						*mFourierData->GetImagHK(-k,-h) = -alphaSym ;
						*mFourierData->GetRealHK(h,k) = wMax ;
						*mFourierData->GetRealHK(k,-h-k) = wMax ;
						*mFourierData->GetRealHK(-h-k,h) = wMax ;
						*mFourierData->GetRealHK(k,h) = wMax ;
						*mFourierData->GetRealHK(-h,-k) = wMax ;
						*mFourierData->GetRealHK(-k,h+k) = wMax ;
						*mFourierData->GetRealHK(h+k,-h) = wMax ;
						*mFourierData->GetRealHK(-k,-h) = wMax ;
						*IsSet->GetDataHK(h,k) = 1. ;
						*IsSet->GetDataHK(k,-h-k) = 1. ;
						*IsSet->GetDataHK(-h-k,h) = 1. ;
						*IsSet->GetDataHK(k,h) = 1. ;
						*IsSet->GetDataHK(-h,-k) = 1. ;
						*IsSet->GetDataHK(-k,h+k) = 1. ;
						*IsSet->GetDataHK(h+k,-h) = 1. ;
						*IsSet->GetDataHK(-k,-h) = 1. ;
					}
				}
			}
			break ;
		case 16: // p6
			// |F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)|
			// P(h,k) = P(k,-h-k) = P(-h-k,h)
			// Phase 0 or 180  : (h,k)
			latticeParameter = (mAStar+mBStar)/2 ;
			mAStar = mBStar = latticeParameter ;
			mA = mB = 1./(mAStar*sin(60.*asin(1.0)/90.)) ;	//  1/Asin(gamma)
			for(h=-hMax+1;h<hMax;h++) {
				dph = h*dn ; 
				for(k=0;k<kMax;k++) {
					w1 = *mFourierData->GetRealHK(h,k) ;
					if (!(h == 0 && k == 0) && w1 > 0.) {
						if (*IsSet->GetDataHK(h,k) == 1.) {
							wMax = *mFourierData->GetRealHK(h,k) ;
							alphaSym = *mFourierData->GetImagHK(h,k) ;
						}
						else {
							dpk = k*dm ; 
							p1 = *mFourierData->GetImagHK(h,k) + dph + dpk ;
							p2 = *mFourierData->GetImagHK(k,-h-k) +k*dn - (h+k)*dm ;
							p3 = *mFourierData->GetImagHK(-h-k,h) -(h+k)*dn + h*dm ;
							w2 = *mFourierData->GetRealHK(k,-h-k) ;
							w3 = *mFourierData->GetRealHK(-h-k,h) ;
							wMax = max(max(w1,w2),w3) ;
							// Calculate symmetrized phase
							argY = w1*sin(p1)+w2*sin(p2)+w3*sin(p3) ;
							argX = w1*cos(p1)+w2*cos(p2)+w3*cos(p3) ;
							alphaSym = atan2(argY,argX) ;
							alphaSym = fabs(alphaSym) < PiHalf ? 0. : Pi ;
						}
						*mFourierData->GetImagHK(h,k) = alphaSym ;
						*mFourierData->GetImagHK(k,-h-k) = alphaSym ;
						*mFourierData->GetImagHK(-h-k,h) = alphaSym ;
						*mFourierData->GetImagHK(-h,-k) = -alphaSym ;
						*mFourierData->GetImagHK(-k,h+k) = -alphaSym ;
						*mFourierData->GetImagHK(h+k,-h) = -alphaSym ;
						*mFourierData->GetRealHK(h,k) = wMax ;
						*mFourierData->GetRealHK(k,-h-k) = wMax ;
						*mFourierData->GetRealHK(-h-k,h) = wMax ;
						*mFourierData->GetRealHK(-h,-k) = wMax ;
						*mFourierData->GetRealHK(-k,h+k) = wMax ;
						*mFourierData->GetRealHK(h+k,-h) = wMax ;
						*IsSet->GetDataHK(h,k) = 1. ;
						*IsSet->GetDataHK(k,-h-k) = 1. ;
						*IsSet->GetDataHK(-h-k,h) = 1. ;
						*IsSet->GetDataHK(-h,-k) = 1. ;
						*IsSet->GetDataHK(-k,h+k) = 1. ;
						*IsSet->GetDataHK(h+k,-h) = 1. ;
					}
				}
			}
			break ;
		case 17: // p6m
			// |F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)| = |F(k,h)|
			// P(h,k) = P(k,-h-k) = P(-h-k,h) = P(k,h)
			// Phase 0 or 180  : (h,k)
			latticeParameter = (mAStar+mBStar)/2 ;
			mAStar = mBStar = latticeParameter ;
			mA = mB = 1./(mAStar*sin(60.*asin(1.0)/90.)) ;	//  1/Asin(gamma)
			for(h=-hMax+1;h<hMax;h++) {
				dph = h*dn ; 
				for(k=0;k<kMax;k++) {
					w1 = *mFourierData->GetRealHK(h,k) ;
					if (!(h == 0 && k == 0) && w1 > 0.) {
						if (*IsSet->GetDataHK(h,k) == 1.) {
							wMax = *mFourierData->GetRealHK(h,k) ;
							alphaSym = *mFourierData->GetImagHK(h,k) ;
						}
						else {
							dpk = k*dm ; 
							p1 = *mFourierData->GetImagHK(h,k) + dph + dpk ;
							p2 = *mFourierData->GetImagHK(k,-h-k) +k*dn - (h+k)*dm ;
							p3 = *mFourierData->GetImagHK(-h-k,h) -(h+k)*dn + h*dm ;
							p4 = *mFourierData->GetImagHK(k,h) + k*dn + h*dm ;
							w2 = *mFourierData->GetRealHK(k,-h-k) ;
							w3 = *mFourierData->GetRealHK(-h-k,h) ;
							//w4 = *mFourierData->GetRealHK(k,h) ;
							w4 = h == k ? 0. : *mFourierData->GetRealHK(k,h) ;
							wMax = max(max(max(w1,w2),w3),w4) ;
							// Calculate symmetrized phase
							argY = w1*sin(p1)+w2*sin(p2)+w3*sin(p3)+w4*sin(p4) ;
							argX = w1*cos(p1)+w2*cos(p2)+w3*cos(p3)+w4*cos(p4) ;
							alphaSym = atan2(argY,argX) ;
							alphaSym = fabs(alphaSym) < PiHalf ? 0. : Pi ;
						}
						*mFourierData->GetImagHK(h,k) = alphaSym ;
						*mFourierData->GetImagHK(k,-h-k) = alphaSym ;
						*mFourierData->GetImagHK(-h-k,h) = alphaSym ;
						*mFourierData->GetImagHK(k,h) = alphaSym ;
						*mFourierData->GetImagHK(-h,-k) = -alphaSym ;
						*mFourierData->GetImagHK(-k,h+k) = -alphaSym ;
						*mFourierData->GetImagHK(h+k,-h) = -alphaSym ;
						*mFourierData->GetImagHK(-k,-h) = -alphaSym ;
						*mFourierData->GetRealHK(h,k) = wMax ;
						*mFourierData->GetRealHK(k,-h-k) = wMax ;
						*mFourierData->GetRealHK(-h-k,h) = wMax ;
						*mFourierData->GetRealHK(k,h) = wMax ;
						*mFourierData->GetRealHK(-h,-k) = wMax ;
						*mFourierData->GetRealHK(-k,h+k) = wMax ;
						*mFourierData->GetRealHK(h+k,-h) = wMax ;
						*mFourierData->GetRealHK(-k,-h) = wMax ;
						*IsSet->GetDataHK(h,k) = 1. ;
						*IsSet->GetDataHK(k,-h-k) = 1. ;
						*IsSet->GetDataHK(-h-k,h) = 1. ;
						*IsSet->GetDataHK(k,h) = 1. ;
						*IsSet->GetDataHK(-h,-k) = 1. ;
						*IsSet->GetDataHK(-k,h+k) = 1. ;
						*IsSet->GetDataHK(h+k,-h) = 1. ;
						*IsSet->GetDataHK(-k,-h) = 1. ;
					}
				}
			}
			break ;
	}

	// Transfer the data into the result array
	float	amp,phase ;
	for(h=-hMax+1;h<hMax;h++) {
		for(k=-kMax+1;k<kMax;k++) {
			amp = *mFourierData->GetRealHK(h,k) ;
			phase = *mFourierData->GetImagHK(h,k) ;
			*result->GetRealHK(h,k) = amp*::cos(phase) ;
			*result->GetImagHK(h,k) = amp*::sin(phase) ;
		}
	}

	delete IsSet ;
	return result ;			
}



// ------------------------------------------------------------------------
//  Copyright (c) 1998 L. D. Marks ; 
// ------------------------------------------------------------------------
//  This program is free software; you can redistribute it and/or modify ; 
//  it under the terms of the GNU General Public License as published by ; 
//  the Free Software Foundation; either version 2 of the License, or ; 
//  (at your option) any later version. ; 
// 
//  This program is distributed in the hope that it will be useful, ; 
//  but WITHOUT ANY WARRANTY; without even the implied warranty of ; 
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the ; 
//  GNU General Public License for more details. ; 
//  
//  You should have received a copy of the GNU General Public License ; 
//  along with this program; if not, write to the Free Software ; 
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA ; 
// 
// ------------------------------------------------------------------------

void equiv2 (long *he, long *ke, long *lse, long h, long k, long &ne, char *symm ) { 
//	generate equivalent positions for 2-fold symmetry 
//	h,k input 
//	he, ke output; ise is -ve for a conjg symm spot
//	ise > 2 for +ve or -ve values only allowed 
//	simple trap for 0,0 

	// Make sure ne gets set
	
	if (h == 0 && k == 0) {
		lse[0] = 1 ; 
		he[0] = 0 ; 
		ke[0] = 0 ; 
		ne = 1 ; 
		return ; 
	}
	
	//	Standard x-glide ; 
	if (strcmp(symm,"p11g") == 0)  // 
		pgs (he,ke,lse,h,k,ne,symm) ; 
	//	Standard y-glide ; 
	else if (strcmp(symm,"p1g1") == 0 || strcmp(symm,"pg  ") == 0) 
		pgs (ke,he,lse,k,h,ne,symm) ; 
	else if (strcmp(symm,"c11m") == 0 || strcmp(symm,"p11m") == 0) 
		ms (ke,he,lse,k,h,ne,symm) ; 
	else if (strncmp(symm,"p1",2) == 0)  
		p1symm (he,ke,lse,h,k,ne,symm) ; 
	//	Non-standard y-glide, switch h & k ; 
	//	else if(symm(1:3).eq.'pyg')then ; 
	//	pygs(he,ke,ise,h,k,ne,symm) ; 
	//	p2mg ; 
	else if (strcmp(symm,"p2mg") == 0)  
		p2mgs (he,ke,lse,h,k,ne,symm) ; 
	//     p2gm, switch h & k ; 
	else if (strcmp(symm,"p2gm") == 0)  
		p2mgs (ke,he,lse,k,h,ne,symm) ; 
	else if (strncmp(symm,"p3",2) == 0 || strcmp(symm,"p6mm") == 0)  
		p3symm (he,ke,lse,h,k,ne,symm) ; 
	else if (strncmp(symm,"p6",2) == 0)  
		p6symm (he,ke,lse,h,k,ne,symm) ; 
	else if (symm[1] == 'm' || strcmp(symm,"p1m1") == 0 || strcmp(symm,"c1m1") == 0) 
		ms (he,ke,lse,h,k,ne,symm) ; 
	else if (strcmp(symm,"p2gg") == 0 )
		p2ggs (he,ke,lse,h,k,ne,symm) ; 
	else if (strncmp(symm,"p2",2) == 0 || strncmp(symm,"c2",2) == 0) 
		p2symm (he,ke,lse,h,k,ne,symm) ; 
	else if (strncmp(symm,"f2",2) == 0) 
		f2symm (he,ke,lse,h,k,ne,symm) ; 
	else if (strcmp(symm,"p4mg") == 0 || strcmp(symm,"p4gm") == 0)  
		p4mg (he,ke,lse,h,k,ne) ; 
	else if (strncmp(symm,"p4",2) == 0 || strncmp(symm,"c4",2) == 0) 
		p4symm (he,ke,lse,h,k,ne,symm) ; 
}

void f2symm (long *he, long *ke, long *lse, long h, long k, long &ne, char *symm ) { 
//	generate equivalent positions for 2-fold symmetry 
//	h,k input 
//	he, ke output; ise is -ve for a conjg symm spot
//	ise > 2 for +ve or -ve values only allowed 
//	simple trap for 0,0 
	if (h == 0 && k == 0) {
		lse[0] = 1 ; 
		he[0] = 0 ; 
		ke[0] = 0 ; 
		ne = 1 ; 
		return ; 
	} 
	he[0] = h ; 
	ke[0] = k ; 
	lse[0] = 1 ; 
	if (strcmp(symm,"f2  ") == 0) {
		//	2-fold rotational symmetry only ;
		ne = 2 ; 
		he[1] = -h ; 
		ke[1] = -k ; 
		lse[1] = 1 ; 
		lse[0] = 1 ; 
	}
	else if (strcmp(symm,"f2mm") == 0) { 
		lse[0] = 1 ; 
		he[1] = -h ; 
		ke[1] = -k ; 
		lse[1] = 1 ; 
		ne = 2 ; 
		if (h != 0 && k != 0) { 
			 he[2] = -h ; 
			 ke[2] = k ; 
			 lse[2] = 1 ; 
			 he[3] = h ; 
			 ke[3] = -k ; 
			 lse[3] = 1 ; 
			 ne = 4 ; 
		} 
	}
}
 
void ms (long *he, long *ke, long *lse, long h, long k, long &ne, char *symm ) {
//`generate equivalent positions for symmetry pm and cm
//	h,k input 
//	he, ke output; ise is -ve for a conjg symm spot
//	ise > 2 for +ve or -ve values only allowed 
//	simple trap for 0,0 
	if (h == 0 && k == 0) {
		lse[0] = 1 ; 
		he[0] = 0 ; 
		ke[0] = 0 ; 
		ne = 1 ; 
		return ; 
	} 
	he[0] = h ; 
	ke[0] = k ; 
	lse[0] = 1 ; 
	//     mirror is along y axis 
	he[1] = -h ; 
	ke[1] = -k ; 
	lse[1] = (-1) ; 
	if (k == 0) {
		lse[0] = 3 ; 
		lse[1] = 3 ; 
		ne = 2 ; 
	}
	else if (h == 0) {
		ne = 2 ; 
	}
	else { 
		he[2] = -h ; 
		ke[2] = k ; 
		lse[2] = 1 ; 
		he[3] = h ; 
		ke[3] = -k ; 
		lse[3] = (-1) ; 
		ne = 4 ; 
	} 
}
 
void p1symm (long *he, long *ke, long *lse, long h, long k, long &ne, char *symm ) { 
//	generate equivalent positions for p1 symmetry
//	h,k input 
//	he, ke output; ise is -ve for a conjg symm spot
//	ise > 2 for +ve or -ve values only allowed 
//	simple trap for 0,0 
	if (h == 0 && k == 0) { 
		lse[0] = 1 ; 
		he[0] = 0 ; 
		ke[0] = 0 ; 
		ne = 1 ; 
		return ; 
	} 
	he[0] = h ; 
	ke[0] = k ; 
	lse[0] = 1 ; 
	he[1] = -h ; 
	ke[1] = -k ; 
	lse[1] = (-1) ; 
	ne = 2 ; 
}

void p2ggs (long *he, long *ke, long *lse, long h, long k, long &ne, char *symm ) { 
//	h,k input 
//	he, ke output; ise is -ve for a conjg symm spot
//	ise > 2 for +ve or -ve values only allowed 
//	simple trap for 0,0 
	if (h == 0 && k == 0) { 
		lse[0] = 1 ; 
		he[0] = 0 ; 
		ke[0] = 0 ; 
		ne = 1 ; 
		return ; 
	} 
	he[0] = h ; 
	ke[0] = k ; 
	lse[0] = 3 ; 
	he[1] = -h ; 
	ke[1] = -k ; 
	lse[1] = 3 ; 
	if (k == 0 || h == 0) { 
		ne = 2 ; 
	}
	else if ((h / 2) * 2 == h) {
		if ((k / 2) * 2 == k) { 
			he[2] = h ; 
			ke[2] = -k ; 
			lse[2] = 3 ; 
			he[3] = -h ; 
			ke[3] = k ; 
			lse[3] = 3 ; 
			ne = 4 ;
		} 
		else {
			he[2] = h ; 
			ke[2] = -k ; 
			lse[2] = 1000 ; 
			he[3] = -h ; 
			ke[3] = k ; 
			lse[3] = 1000 ; 
			ne = 4 ; 
		} 
	}
	else {
		if ((k / 2) * 2 == k) {
			he[2] = h ; 
			ke[2] = -k ; 
			lse[2] = 1000 ; 
			he[3] = -h ; 
			ke[3] = k ; 
			lse[3] = 1000 ; 
			ne = 4 ; 
		}
		else { 
			he[2] = h ; 
			ke[2] = -k ; 
			lse[2] = 3 ; 
			he[3] = -h ; 
			ke[3] = k ; 
			lse[3] = 3 ; 
			ne = 4 ; 
		} 
	}
}

void p2mgs (long *he, long *ke, long *lse, long h, long k, long &ne, char *symm ) { 
//	generate equivalent positions for glide symmetry along y-axis
//	h,k input 
//	he, ke output; ise is -ve for a conjg symm spot
//	ise > 2 for +ve or -ve values only allowed 
//	simple trap for 0,0 
	if (h == 0 && k == 0) {
		lse[0] = 1 ; 
		he[0] = 0 ; 
		ke[0] = 0 ; 
		ne = 1 ; 
		return ; 
	} 
	he[0] = h ; 
	ke[0] = k ; 
	lse[0] = 3 ; 
	he[1] = -h ; 
	ke[1] = -k ; 
	lse[1] = 3 ; 
	if (k == 0 || h == 0) { 
		ne = 2 ; 
	}
	else if ((h / 2) * 2 == h) {
		he[2] = h ; 
		ke[2] = -k ; 
		lse[2] = 3 ; 
		he[3] = -h ; 
		ke[3] = k ; 
		lse[3] = 3 ; 
		ne = 4 ; 
	}
	else {
		he[2] = h ; 
		ke[2] = -k ; 
		lse[2] = 1000 ; 
		he[3] = -h ; 
		ke[3] = k ; 
		lse[3] = 1000 ; 
		ne = 4 ; 
	} 
}
 
void p2symm (long *he, long *ke, long *lse, long h, long k, long &ne, char *symm ) { 
//	generate equivalent positions for 2-fold symmetry 
//	h,k input 
//	he, ke output; ise is -ve for a conjg symm spot
//	ise > 2 for +ve or -ve values only allowed 
//	simple trap for 0,0 
	if (h == 0 && k == 0) { 
		lse[0] = 1 ; 
		he[0] = 0 ; 
		ke[0] = 0 ; 
		ne = 1 ; 
		return ; 
	}
	he[0] = h ; 
	ke[0] = k ; 
	lse[0] = 1 ; 
	if (strcmp(symm,"p2  ") == 0 || strcmp(symm,"c2  ") == 0) { 
		//	2-fold rotational symmetry only ; 
		ne = 2 ; 
		he[1] = -h ; 
		ke[1] = -k ; 
		lse[1] = 3 ; 
		lse[0] = 3 ; 
	}
	else if (strcmp(symm,"p2mm") == 0 || strcmp(symm,"c2mm") == 0) {
		lse[0] = 3 ; 
		he[1] = -h ; 
		ke[1] = -k ; 
		lse[1] = 3 ; 
		ne = 2 ; 
		if (h != 0 && k != 0) { 
			 he[2] = -h ; 
			 ke[2] = k ; 
			 lse[2] = 3 ; 
			 he[3] = h ; 
			 ke[3] = -k ; 
			 lse[3] = 3 ; 
			 ne = 4 ; 
		} 
	} 
}
 
void p3symm (long *he, long *ke, long *lse, long h, long k, long &ne, char *symm ) { 
//	generate equivalent positions for 3-fold symmetry 
//	h,k input 
//	he, ke output; ise is -ve for a conjg symm spot
//	ise > 2 for +ve or -ve values only allowed 
//	simple trap for 0,0 
	Boolean check ; 
	long II1, II2, II3, II4, II5, II6, II7, II8, II9, II10 ; 
	long i,j ;

	if (h == 0 && k == 0) { 
		lse[0] = 1 ; 
		he[0] = 0 ; 
		ke[0] = 0 ; 
		ne = 1 ; 
		return ; 
	} 
	 he[0] = h ; 
	 ke[0] = k ; 
	 lse[0] = 1 ; 
	if (strcmp(symm,"p3  ") == 0) { 
		// 3-fold rotational symmetry only ; 
		ne = 6 ; 
		he[1] = k ; 
		ke[1] = -(h + k) ; 
		lse[1] = 1 ; 
		he[2] = -(h + k) ; 
		ke[2] = h ; 
		lse[2] = 1 ; 
		he[3] = -h ; 
		ke[3] = -k ; 
		lse[3] = (-1) ; 
		he[4] = -k ; 
		ke[4] = h + k ; 
		lse[4] = (-1) ; 
		he[5] = h + k ; 
		ke[5] = -h ; 
		lse[5] = (-1) ; 
	}
	else if (strcmp(symm,"p3m1") == 0) { 
		//     3-fold rotational symmetry and mirror through 220 (for rt3) ; 
		ne = 6 ; 
		he[1] = k ; 
		ke[1] = -(h + k) ; 
		lse[1] = 1 ; 
		he[2] = -(h + k) ; 
		ke[2] = h ; 
		lse[2] = 1 ; 
		he[3] = -h ; 
		ke[3] = -k ; 
		lse[3] = (-1) ; 
		he[4] = -k ; 
		ke[4] = h + k ; 
		lse[4] = (-1) ; 
		he[5] = h + k ; 
		ke[5] = -h ; 
		lse[5] = (-1) ; 
		check = false ; 
		for(j=0; j < 6; j++) {
			if (he[j] - ke[j] == 0) check = true ; 
		}
		if (check) { 
			for(j=0; j < 6; j++) lse[j] = 3 ;
		} 
		else{
			check = true ; 
			for(j=0; j < 6; j++) {
				if (he[j] == 0 || ke[j] == 0) check = false ; 
			}
			if (check) { 
				ne = 12 ; 
				for(j=6; j < 12; j++) {
					II2 = he[j-6] ; 
					II1 = ke[j-6] ; 
					II3 = lse[j-6] ; 
					he[j] = II2 + II1 ; 
					ke[j] = -II1 ; 
					lse[j] = II3 ; 
				}
			}
		} 
	}
	else if (strcmp(symm,"p31m") == 0) {
		//     3-fold rotational symmetry and mirror along 1x1 (for rt3) ; 
		ne = 6 ; 
		he[1] = k ; 
		ke[1] = -(h + k) ; 
		lse[1] = 1 ; 
		he[2] = -(h + k) ; 
		ke[2] = h ; 
		lse[2] = 1 ; 
		he[3] = -h ; 
		ke[3] = -k ; 
		lse[3] = (-1) ; 
		he[4] = -k ; 
		ke[4] = h + k ; 
		lse[4] = (-1) ; 
		he[5] = h + k ; 
		ke[5] = -h ; 
		lse[5] = (-1) ; 
		check = false ; 
		for(j=0; j < 6; j++) {
			if (he[j] == 0 || ke[j] == 0) check = true ; 
		}
		if (check) { 
			for(j=0; j < 6; j++) lse[j] = 3 ;
		} 
		else { 
			check = true ; 
			for(j=0; j < 6; j++) {
				if (he[j] - ke[j] == 0) check = false ; 
			}
			if (check) {
				ne = 12 ; 
				for(j=6; j < 12; j++) {
					II4 = ke[j-6] ; 
					II6 = he[j-6] ; 
					II7 = lse[j-6] ; 
					II5 = II4 ; 
					ke[j] = II6 ; 
					lse[j] = II7 ; 
					he[j] = II5 ; 
				} 
			} 
		} 
	}
	else if (strcmp(symm,"p3mt") == 0 || strcmp(symm,"p6mm") == 0) { 
		//     Inversion + 3-fold rotational symmetry and mirror through 220 (for rt3) ; 
		lse[0] = 3 ; 
		ne = 6 ; 
		he[1] = k ; 
		ke[1] = -(h + k) ; 
		lse[1] = 3 ; 
		he[2] = -(h + k) ; 
		ke[2] = h ; 
		lse[2] = 3 ; 
		he[3] = -h ; 
		ke[3] = -k ; 
		lse[3] = 3 ; 
		he[4] = -k ; 
		ke[4] = h + k ; 
		lse[4] = 3 ; 
		he[5] = h + k ; 
		ke[5] = -h ; 
		lse[5] = 3 ; 
		check = false ; 
		for(j=0; j < 6; j++) {
			if (he[j] - ke[j] == 0) check = true ; 
		}
		if (check) { 
			for(j=0; j < 6; j++) lse[j] = 3 ;
		} 
		else { 
			check = true ; 
			for(j=0; j < 6; j++) {
				if (he[j] == 0 || ke[j] == 0) check = false ; 
			}
			if (check) {
				ne = 12 ; 
				for(j=6; j < 12; j++) {
					II9 = he[j-6] ; 
					II8 = ke[j-6] ; 
					II10 = lse[j-6] ; 
					he[j] = II9 + II8 ; 
					ke[j] = -II8 ; 
					lse[j] = II10 ; 
				}
			}
		}
	} 
}
 
void p4 (long *he, long *ke, long *lse, long h, long k, long &ne, char *symm ) {
//	generate equivalent positions for 4-fold symmetry ; 
//	h,k input 
//	he, ke output; ise is -ve for a conjg symm spot
//	ise > 2 for +ve or -ve values only allowed 
//	simple trap for 0,0 
	if (h == 0 && k == 0) {
		he[0] = 0 ; 
		ke[0] = 0 ; 
		ne = 1 ; 
		lse[0] = 1 ; 
		return ; 
	}
	he[0] = h ; 
	ke[0] = k ; 
	lse[0] = 3 ; 
	he[1] = -k ; 
	ke[1] = h ; 
	lse[1] = 3 ; 
	he[2] = -h ; 
	ke[2] = -k ; 
	lse[2] = 3 ; 
	he[3] = k ; 
	ke[3] = -h ; 
	lse[3] = 3 ; 
	ne = 4 ; 
}

void p4mg (long *he, long *ke, long *lse, long h, long k, long &ne) { 
//	to give (h,k) -> (k,h) for symm p4gm ; 
	 long II1, II2, i, j ; 
	 if (h == 0 && k == 0) {
	   he[0] = 0 ; 
	   ke[0] = 0 ; 
	   ne = 1 ; 
	   lse[0] = 1 ; 
	   return ; 
	 } 
	 he[0] = h ; 
	 ke[0] = k ; 
	 lse[0] = 3 ; 
	 he[1] = k ; 
	 ke[1] = h ; 
	 lse[1] = 1000 ; 
	 he[2] = -k ; 
	 ke[2] = h ; 
	 lse[2] = 3 ; 
	 he[3] = -h ; 
	 ke[3] = k ; 
	 lse[3] = 1000 ; 
	 he[4] = -h ; 
	 ke[4] = -k ; 
	 lse[4] = 3 ; 
	 he[5] = -k ; 
	 ke[5] = -h ; 
	 lse[5] = 1000 ; 
	 he[6] = k ; 
	 ke[6] = -h ; 
	 lse[6] = 3 ; 
	 he[7] = h ; 
	 ke[7] = -k ; 
	 lse[7] = 1000 ; 
	 ne = 8 ; 
	 if (((h + k) / 2) * 2 == h + k) { 
	 	for(j=0; j < 8; j++) lse[j] = 3 ;
	} 
	 ne = 0 ; 

	for(i=0; i < 8; i++) {
		II1 = ke[i] ; 
		II2 = he[i] ; 
		for(j=i+1; j < 8; j++) {
			if (he[j] == II2 && ke[j] == II1) goto nexti ; 
		}
		ne = ne + 1 ; 
		he[ne-1] = he[i] ; 
		ke[ne-1] = ke[i] ; 
		nexti: ;
	}
}
void p4symm (long *he, long *ke, long *lse, long h, long k, long &ne, char *symm ) {
//	generate equivalent positions for 4-fold symmetry 
//	h,k input 
//	he, ke output; ise is -ve for a conjg symm spot
//	ise > 2 for +ve or -ve values only allowed 
//	simple trap for 0,0 
	long II1, II2,i,j ; 
	if (strcmp(symm,"p4mm") == 0 || strcmp(symm,"c4mm") == 0) { 
		if (h == 0 && k == 0) { 
			 he[0] = 0 ; 
			 ke[0] = 0 ; 
			 ne = 1 ; 
			 lse[0] = 1 ; 
			 return ; 
		} 
		he[0] = h ; 
		ke[0] = k ; 
		he[1] = -h ; 
		ke[1] = -k ; 
		he[2] = -h ; 
		ke[2] = k ; 
		he[3] = h ; 
		ke[3] = -k ; 
		he[4] = k ; 
		ke[4] = h ; 
		he[5] = -k ; 
		ke[5] = -h ; 
		he[6] = -k ; 
		ke[6] = h ; 
		he[7] = k ; 
		ke[7] = -h ; 
		ne = 8 ; 
		for(i=0; i < 8; i++) {
			 lse[i] = 3 ; 
		} 
		ne = 0 ; 
		for(i=0; i < 8; i++) {
			II1 = ke[i] ; 
			II2 = he[i] ; 
			for(j=i+1; j < 8; j++) {
				if (he[j] == II2 && ke[j] == II1) goto nexti ; 
			}
			ne = ne + 1 ; 
			he[ne-1] = he[i] ; 
			ke[ne-1] = ke[i] ; 
			nexti: ;
		}
	}
	else if (strcmp(symm,"p4  ") == 0) { 
		he[0] = h ; 
		ke[0] = k ; 
		lse[0] = 3 ; 
		he[1] = -k ; 
		ke[1] = h ; 
		lse[1] = 3 ; 
		he[2] = -h ; 
		ke[2] = -k ; 
		lse[2] = 3 ; 
		he[3] = k ; 
		ke[3] = -h ; 
		lse[3] = 3 ; 
		ne = 4 ; 
	}
}

void p6symm (long *he, long *ke, long *lse, long h, long k, long &ne, char *symm ) { 
//	generate equivalent positions for 6-fold symmetry 
//	h,k input 
//	he, ke output; ise is -ve for a conjg symm spot
//	ise > 2 for +ve or -ve values only allowed 
//	simple trap for 0,0 
	if (h == 0 && k == 0) {
		lse[0] = 1 ; 
		he[0] = 0 ; 
		ke[0] = 0 ; 
		ne = 1 ; 
		return ; 
	} 
	he[0] = h ; 
	ke[0] = k ; 
	lse[0] = 3 ; 
	he[1] = k ; 
	ke[1] = -(h + k) ; 
	lse[1] = 3 ; 
	he[2] = -(h + k) ; 
	ke[2] = h ; 
	lse[2] = 3 ; 
	he[3] = -h ; 
	ke[3] = -k ; 
	lse[3] = 3 ; 
	he[4] = -k ; 
	ke[4] = h + k ; 
	lse[4] = 3 ; 
	he[5] = h + k ; 
	ke[5] = -h ; 
	lse[5] = 3 ; 
	ne = 6 ; 
}
 
void pgs (long *he, long *ke, long *lse, long h, long k, long &ne, char *symm ) {
//	generate equivalent positions for glide symmetry along x-axis
//	(Note, not conventional ??)
//	h,k input 
//	he, ke output; ise is -ve for a conjg symm spot
//	ise > 2 for +ve or -ve values only allowed 
//	simple trap for 0,0 
	if (h == 0 && k == 0) { 
		lse[0] = 1 ; 
		he[0] = 0 ; 
		ke[0] = 0 ; 
		ne = 1 ; 
		return ; 
	} 
	he[0] = h ; 
	ke[0] = k ; 
	lse[0] = 1 ; 
	he[1] = -h ; 
	ke[1] = -k ; 
	lse[1] = (-1) ; 
	if (k == 0) { 
		ne = 2 ; 
	}
	else if (h == 0) { 
		ne = 2 ; 
		lse[0] = 3 ; 
		lse[1] = 3 ; 
	}
	else if ((h / 2) * 2 == h) { 
		he[2] = h ; 
		ke[2] = -k ; 
		lse[2] = 1 ; 
		he[3] = -h ; 
		ke[3] = k ; 
		lse[3] = (-1) ; 
		ne = 4 ; 
	}
	else { 
		he[2] = h ; 
		ke[2] = -k ; 
		lse[2] = 200 ; 
		he[3] = -h ; 
		ke[3] = k ; 
		lse[3] = 300 ; 
		ne = 4 ; 
	} 
}
 
void  pvgs (long *he, long *ke, long *lse, long h, long k, long &ne, char *symm ) {
//`generate equivalent positions for glide symmetry along x-axis
//	h,k input 
//	he, ke output; ise is -ve for a conjg symm spot
//	ise > 2 for +ve or -ve values only allowed 
//	simple trap for 0,0 
	if (h == 0 && k == 0) { 
		lse[0] = 1 ; 
		he[0] = 0 ; 
		ke[0] = 0 ; 
		ne = 1 ; 
		return ; 
	} 
	he[0] = h ; 
	ke[0] = k ; 
	lse[0] = 1 ; 
	he[1] = -h ; 
	ke[1] = -k ; 
	lse[1] = (-1) ; 
	if (h == 0) { 
		ne = 2 ; 
		if ((k / 2) * 2 == k) { 
			lse[0] = 3 ; 
			lse[1] = 3 ; 
		} 
	}
	else if (k == 0) { 
		ne = 2 ; 
	}
	else {
		he[2] = -h ; 
		ke[2] = k ; 
		lse[2] = 100 ; 
		he[3] = h ; 
		ke[3] = -k ; 
		lse[3] = 200 ; 
		ne = 4 ; 
	} 
}
